	#version 130
#extension GL_EXT_gpu_shader3 : enable

uniform sampler2D textureHeight;
uniform int channel;
uniform int textureMode;

out vec3 position;
out vec3 normal;

float GetValue(vec4 texCoordOrg){
    //vec4 texCoord = gl_TextureMatrix[0] * texCoordOrg;
    //vec4 texColor = texture2D(textureHeight, texCoord.st);
    vec4 texColor = texture2D(textureHeight, texCoordOrg.st);
    
    if (channel < 3)
	return texColor[channel];
    else if (channel == 3)
	return sqrt((0.299 * texColor[0]*texColor[0] + 0.587*texColor[1]*texColor[1] + 0.114*texColor[2]*texColor[2])/3); // luminocity
    else
	return sqrt((texColor[0]*texColor[0] + texColor[1]*texColor[1] + texColor[2]*texColor[2])/3); // norm
}

void main(void){
    //gl_TexCoord[0] = gl_TextureMatrix[0] * gl_MultiTexCoord0;
    
    gl_TexCoord[0] = gl_MultiTexCoord0; // p
    vec4 texCoord = gl_MultiTexCoord0;

    float heightZ = GetValue(texCoord);

    int width = textureSize(textureHeight, 0).x;
    int height = textureSize(textureHeight, 0).y;

    vec4 dvx = vec4(1.0/width, 0, 0, 0);
    vec4 dvy = vec4(0, 1.0/height, 0, 0);

    float dx = GetValue(texCoord+dvx) - GetValue(texCoord-dvx);
    float dy = GetValue(texCoord+dvy) - GetValue(texCoord-dvy);

    dx = dx/2 * width;
    dy = dy/2 * height;
    
    vec4 newPos = vec4(gl_Vertex[0], gl_Vertex[1], gl_Vertex[2] * heightZ*2, gl_Vertex[3]);
    
    position = vec3(gl_ModelViewMatrix * newPos);
    
    float norm = sqrt(dx*dx + dy*dy + 1);
    normal = gl_NormalMatrix * vec3(-dx/norm, -dy/norm, 1.0/norm);
    
    gl_Position = gl_ModelViewProjectionMatrix * newPos;//ftransform();

    gl_TexCoord[1] = gl_TextureMatrix[0] * gl_MultiTexCoord0; // Texturep
    if (textureMode == 1){
	gl_TexCoord[2].s = heightZ * gl_Vertex[2];
    }
}
