//#extension GL_EXT_gpu_shader4 : enable

uniform sampler3D textureVolume;
uniform sampler1D textureCC;
uniform int lightNumber;
uniform float shadowAlpha;
uniform int shadowQuality;
uniform vec3 clipBoxMin;
uniform vec3 clipBoxMax;
uniform vec3 ambientClr;
uniform vec3 diffuseClr;
uniform vec3 specularClr;
uniform int shininess;
uniform int forceDiscard0;
uniform int trim;
uniform float intervalGlobal;
uniform int subQuality;
uniform float delta;
uniform int blend;
//uniform mat4 modelMatrix;

varying vec3 position;

//##FUNC_VOLUME_MASK_BEGIN##
bool GetMask(vec3 texCoord){
    return true;
}
//##FUNC_VOLUME_MASK_END##

//##FUNC_VOLUME_COORD_BEGIN##
vec3 GetDeformCoord(vec3 vecOrg){
    return vecOrg;
}

#define GUSEMATRIX 0
mat3 GetDeformNormalMatrix(vec3 texCoord, vec3 deformedCoord){
    return mat3(1, 0, 0, 0, 1, 0, 0, 0, 1);
}
//##FUNC_VOLUME_COORD_END##

//##FUNC_SPOT_EFFECT##

void main (void)
{
    gl_FragColor = vec4(0, 0, 0 ,0);

    float factorIntegrate = 1.0/subQuality;

    vec3 viewVec = -normalize(position);
    float angleMod = sqrt (1 + (viewVec[0]*viewVec[0]+viewVec[1]*viewVec[1])/viewVec[2]*viewVec[2]);
    vec3 viewVecGlobal = normalize(mat3(gl_ModelViewMatrixInverse) * viewVec) * intervalGlobal/subQuality * 3 * angleMod * angleMod;
    vec3 viewVecTexSpace = mat3(gl_TextureMatrix[0]) * viewVecGlobal;
    vec3 srcPos = vec3(gl_TextureMatrixInverse[0] * gl_TexCoord[0]);
    
    vec3 lightVecCache[8];
    float disCache[8];
    for (int i = 0; i < lightNumber; i++){
	lightVecCache[i] = gl_LightSource[i].position.xyz - position * gl_LightSource[i].position[3];
	disCache[i] = length(lightVecCache[i]);
	lightVecCache[i] = normalize(lightVecCache[i]);
    }

    //float delta = 0.01;
    int flatData = 0;
    vec3 gradientL;

    int subCount = 0;
    for (int i = 0; i < subQuality; i++){

	vec3 tgtSubPos = srcPos + viewVecGlobal*i;

	if (any(lessThan(tgtSubPos, clipBoxMin))) continue;
	if (any(greaterThan(tgtSubPos, clipBoxMax))) continue;

	vec3 texCoord = vec3(gl_TexCoord[0] + viewVecTexSpace * i);
	vec3 deformedCoord = GetDeformCoord(texCoord);
	if (GetMask(deformedCoord) == false) continue;
	vec3 srcPos = mat3(gl_TextureMatrixInverse[0]) * texCoord;

	if (trim == 1){
	    if (any(lessThan   (deformedCoord, vec3(0, 0, 0)))) continue;
	    if (any(greaterThan(deformedCoord, vec3(1, 1, 1)))) continue;
	}

	vec4 texColor = texture3D(textureVolume, deformedCoord);
	float value = texColor[0];
	if (forceDiscard0 == 1 && value == 0) continue;
    
	//vec4 texColor2 = texture1D(textureCC, value);
	vec4 color = texture1D(textureCC, value) * gl_Color;
	if (color.a == 0) continue;
    
	//float materialAlpha = gl_Color.a;
    
#if GUSEMATRIX
	vec3 gradient = vec3(textureOffset(textureVolume, deformedCoord, ivec3(-1, 0, 0)).r - textureOffset(textureVolume, deformedCoord, ivec3(1, 0, 0)).r,
			     textureOffset(textureVolume, deformedCoord, ivec3(0, -1, 0)).r - textureOffset(textureVolume, deformedCoord, ivec3(0, 1, 0)).r,
			     textureOffset(textureVolume, deformedCoord, ivec3(0, 0, -1)).r - textureOffset(textureVolume, deformedCoord, ivec3(0, 0, 1)).r);
    	
	mat3 deformedMatrix = GetDeformNormalMatrix(texCoord, deformedCoord);
	vec3 deformedGradient = deformedMatrix * gradient;

	gradientL = vec3(gl_NormalMatrix * mat3(gl_TextureMatrixInverse[0]) * deformedGradient);
#else
        //ivec3 texSize = textureSize3D(textureVolume, 0);
    
    
	vec3 deformedCoord_x1 = GetDeformCoord(texCoord + vec3(+delta, 0, 0));
	vec3 deformedCoord_x0 = GetDeformCoord(texCoord + vec3(-delta, 0, 0));
	vec3 deformedCoord_y1 = GetDeformCoord(texCoord + vec3(0, +delta, 0));
	vec3 deformedCoord_y0 = GetDeformCoord(texCoord + vec3(0, -delta, 0));
	vec3 deformedCoord_z1 = GetDeformCoord(texCoord + vec3(0, 0, +delta));
	vec3 deformedCoord_z0 = GetDeformCoord(texCoord + vec3(0, 0, -delta));
    
	vec3 deformedGradient = -vec3
	    (texture3D(textureVolume, deformedCoord_x1).r - 
	     texture3D(textureVolume, deformedCoord_x0).r,
	     texture3D(textureVolume, deformedCoord_y1).r - 
	     texture3D(textureVolume, deformedCoord_y0).r,
	     texture3D(textureVolume, deformedCoord_z1).r - 
	     texture3D(textureVolume, deformedCoord_z0).r); 

	gradientL = vec3(gl_NormalMatrix * mat3(gl_TextureMatrixInverse[0]) * deformedGradient);
#endif
    
	if (gradientL[0] == 0 && gradientL[1] == 0 && gradientL[2] == 0)
	    flatData = 1;
    
	vec3 fnormal = normalize(gradientL);
	
	for (int i = 0; i < lightNumber; i++){
	    float diffuse = 0;
	    float specular = 0;
	    if (flatData == 0){
		
		vec3 lightVec = lightVecCache[i];
		float dis = disCache[i];
	    
		float spotAngleCos = dot(normalize(gl_LightSource[i].spotDirection),
					 -lightVec);
		float lightFactor = GetSpotEffect(i, spotAngleCos);
		// W
		lightFactor /= (gl_LightSource[i].constantAttenuation +
				gl_LightSource[i].linearAttenuation * dis +
				gl_LightSource[i].quadraticAttenuation * dis * dis);
	    	    
		vec3 lightVecGlobal = mat3(gl_ModelViewMatrixInverse) * lightVec / shadowQuality;
		vec3 lightVecTexSpace = mat3(gl_TextureMatrix[0]) * lightVecGlobal;
	    
		//vec3 srcPos = vec3(gl_TextureMatrixInverse[0] * gl_TexCoord[0]);
		//vec3 lightVecLocal = mat3(modelMatrix) * lightVecGlobal;
	    
		float maxDistance = 2;
		if (gl_LightSource[i].position[3] == 1){
		    float lightDistance = length(position - gl_LightSource[i].position.xyz);
		    maxDistance = min(maxDistance, lightDistance);
		}
		
		int sampleNum = int(maxDistance * shadowQuality);
		for (int i = 1; i <= sampleNum; i++){
		    vec3 tgtPos = tgtSubPos + lightVecGlobal * i;
		    if (any(lessThan(tgtPos , clipBoxMin))) break;
		    if (any(greaterThan(tgtPos, clipBoxMax))) break;
		    vec3 tgtPosTexSpace = texCoord + lightVecTexSpace * i;
		    vec3 tgtDeformCoord = GetDeformCoord(tgtPosTexSpace);

		    float tgtValue = texture3D(textureVolume, tgtDeformCoord).r;
		    float densityValue = texture1D(textureCC, texture3D(textureVolume, tgtDeformCoord).r).a;
		    
		    if (tgtDeformCoord.z < 0 || tgtDeformCoord.z >= 1) densityValue = 0;
		    if (GetMask(tgtDeformCoord) == false) densityValue = 0;
		    if (forceDiscard0 == 1 && tgtValue == 0) densityValue = 0;

		    lightFactor -= densityValue * shadowAlpha;
		    if (lightFactor < 0) break;
		}
		if (lightFactor < 0) lightFactor = 0;
		
		diffuse = dot(lightVec, fnormal) * 4 * lightFactor;
	    
		if (diffuse < 0) diffuse = 0;

		//˃xNg 
		vec3 reflectVec = reflect(-lightVec, fnormal);
		specular = pow(max(dot(viewVec, reflectVec), 0.0), float(shininess)) * lightFactor;
	    }
	    
	    //vec4 color = texColor2 * gl_Color;
	
	    gl_FragColor.r +=
		color.r * ambientClr.r * gl_LightSource[i].ambient.r +
		color.r * diffuse * diffuseClr.r * gl_LightSource[i].diffuse.r +
		specular * specularClr.r * gl_LightSource[i].specular.r;

	    gl_FragColor.g +=
		color.g * ambientClr.g * gl_LightSource[i].ambient.g +
		color.g * diffuse * diffuseClr.g * gl_LightSource[i].diffuse.g +
		specular * specularClr.g * gl_LightSource[i].specular.g;

	    gl_FragColor.b +=
		color.b * ambientClr.b * gl_LightSource[i].ambient.b +
		color.b * diffuse * diffuseClr.b * gl_LightSource[i].diffuse.b +
		specular * specularClr.b * gl_LightSource[i].specular.b;
	}

	gl_FragColor.a += color.a;
	subCount++;
	//float alpha = color.a;//texColor2.a * gl_Color.a;

    }

    if (gl_FragColor.a == 0) discard;

    

    //if (alpha > 1) alpha = 1;
    //gl_FragColor[3] = alpha;

    float alpha0 = gl_FragColor.a;

    float factorCount = 1.0f/subCount;

    gl_FragColor.r *= factorCount;
    gl_FragColor.g *= factorCount;
    gl_FragColor.b *= factorCount;
    gl_FragColor.a *= factorIntegrate;

    switch(blend){
    case 0:
	gl_FragColor.rgb *= alpha0;
	//gl_FragColor.a = alpha;
	break;
    case 1:
	gl_FragColor.rgb *= alpha0;
	gl_FragColor.a = 0;
	break;
    case 2:
	gl_FragColor.rgb *= alpha0;
	gl_FragColor.a *= (1 - alpha0);
	break;
    }
    
    return;
}
