//#extension GL_EXT_gpu_shader4 : enable

uniform sampler2D textureImage;
uniform sampler2D textureImage1;

uniform int quality;
uniform vec4 gamma;
uniform float projMatrix10;
uniform float projMatrix14;
uniform float focalDistance;
uniform float widthBase;

void main (void)
{
    vec2 texCoord = gl_TexCoord[0].st;

    vec4 texColorOrg = texture2D(textureImage, texCoord);
    
    texColorOrg[0] += 0.0001;
    texColorOrg[1] += 0.0001;
    texColorOrg[2] += 0.0001;
    float alpha = 0.5;

    vec4 texColor1Base = texture2D(textureImage1, texCoord);
    float z_eyeBase = - projMatrix14 / (texColor1Base[0] * -2.0 + 1.0 - projMatrix10);
    float widthAtTarget = abs((1.0 - focalDistance/z_eyeBase) / 2);
    
    gl_FragColor = vec4(0, 0, 0 ,0);

    float sum = 0;
    float fqual = float(quality);
    float widthBaseFactor = widthBase / fqual;
    float area = fqual / 1.5;
    float area2 = area * area;
    float gamma0 = gamma[0];
    for (int j = 0; j < quality*2+1; j++){
	float dy = float(j-quality) * widthBaseFactor;
	float y = float(j-quality) / fqual * 1.5;
	
	for (int i = 0; i < quality*2+1; i++){
	    //if (i == quality && j == quality) continue;

	    float dx = float(i-quality) * widthBaseFactor;
	    float x = float(i-quality) / fqual * 1.5;

	    vec2 texCoordMod = texCoord + vec2(dx, dy);
	    if (texCoordMod[0] <= 0.001 || texCoordMod[0] >= 0.999 ||
		texCoordMod[1] <= 0.001 || texCoordMod[1] >= 0.999){
		continue;//fx = 0; fy = 0;
	    }
	    
	    vec4 texColor1 = texture2D(textureImage1, texCoordMod);
	    float z_eye = - projMatrix14 / (texColor1[0] * -2.0 + 1.0 - projMatrix10);

	    float factor = 1;
	    
	    if (z_eye > z_eyeBase * (1 + widthAtTarget)){
		factor = 1;
		//continue;
	    }

	    vec4 texColor = texture2D(textureImage, texCoordMod);

	    if (texColor[3] > 1){
		texColor[0] *= texColor[3];
		texColor[1] *= texColor[3];
		texColor[2] *= texColor[3];
		texColor[3] = 1;
	    }
	    if (gamma0 != 0.0) texColor = pow(texColor, gamma);
	    
	    float width = (1.0 - focalDistance/z_eye) / 3;
	    
	    if (z_eye > z_eyeBase * (1 + widthAtTarget)){
		width = min(abs(width), abs(widthAtTarget));
	    }
	    float width2x2 = 2.0 * width * width;

	    float fy = 1.0 / sqrt(2*3.14159) * exp(-y*y / width2x2);
	    float fx = 1.0 / sqrt(2*3.14159) * exp(-x*x / width2x2);

	    sum = sum + factor * fx * fy / area2;
		
	    gl_FragColor += factor * texColor * fx * fy / area2;
	}
    }

    gl_FragColor /= sum;
    
    if (gamma0 != 0.0){
	vec4 gammaInv = vec4(1.0/gamma[0], 1.0/gamma[1], 1.0/gamma[2], 1.0/gamma[3]);
	gl_FragColor = pow(gl_FragColor, gammaInv);
    }
	
    return;
}
