//#extension GL_EXT_gpu_shader4 : enable

uniform sampler3D textureVolume;
uniform sampler1D textureCC;
uniform vec3 clipBoxMin;
uniform vec3 clipBoxMax;
uniform int forceDiscard0;
uniform int trim;
uniform float intervalGlobal;
uniform int subQuality;
uniform int blend;

varying vec3 position;

//##FUNC_VOLUME_MASK_BEGIN##
bool GetMask(vec3 texCoord){
    return true;
}
//##FUNC_VOLUME_MASK_END##

//##FUNC_VOLUME_COORD_BEGIN##
//#define GTRIM 1
vec3 GetDeformCoord(vec3 vecOrg){
    return vecOrg;
}

#define GUSEMATRIX 0
mat3 GetDeformNormalMatrix(vec3 texCoord, vec3 deformedCoord){
    return mat3(1, 0, 0, 0, 1, 0, 0, 0, 1);
}
//##FUNC_VOLUME_COORD_END##

void main (void)
{
   
    gl_FragColor = vec4(0, 0, 0 ,0);

    float factorIntegrate = 1.0/subQuality;

    vec3 viewVec = -normalize(position);
    float angleMod = sqrt (1 + (viewVec[0]*viewVec[0]+viewVec[1]*viewVec[1])/viewVec[2]*viewVec[2]);
    vec3 viewVecGlobal = normalize(mat3(gl_ModelViewMatrixInverse) * viewVec) * intervalGlobal/subQuality * 3 * angleMod * angleMod;
    vec3 viewVecTexSpace = mat3(gl_TextureMatrix[0]) * viewVecGlobal;
    vec3 srcPos = vec3(gl_TextureMatrixInverse[0] * gl_TexCoord[0]);

    

    int subCount = 0;
    for (int i = 0; i < subQuality; i++){

	vec3 tgtSubPos = srcPos + viewVecGlobal*i;

	if (any(lessThan(tgtSubPos, clipBoxMin))) continue;
	if (any(greaterThan(tgtSubPos, clipBoxMax))) continue;
			
	vec3 texCoord = vec3(gl_TexCoord[0] + viewVecTexSpace * i);
	vec3 deformedCoord = GetDeformCoord(texCoord);
	if (GetMask(deformedCoord) == false) continue;
	
	vec4 texColor = texture3D(textureVolume, deformedCoord);

	if (trim == 1){
	    if (any(lessThan   (deformedCoord, vec3(0, 0, 0)))) continue;
	    if (any(greaterThan(deformedCoord, vec3(1, 1, 1)))) continue;
	}
	
	float value = texColor[0];
	
	if (forceDiscard0 == 1 && value == 0) continue;
	
	//vec4 texColor2 = texture1D(textureCC, value);
	vec4 color = texture1D(textureCC, value) * gl_Color;
	if (color.a == 0) continue;

	gl_FragColor += color;
	subCount++;
    }

    if (gl_FragColor[3] == 0) discard;

    float factorCount = 1.0f/subCount;
    
    float alpha0 = gl_FragColor.a;
    
    gl_FragColor.r *= factorCount;
    gl_FragColor.g *= factorCount;
    gl_FragColor.b *= factorCount;
    gl_FragColor.a *= factorIntegrate;

    switch(blend){
    case 0:
	gl_FragColor.rgb *= alpha0;
	//gl_FragColor.a = alpha;
	break;
    case 1:
	gl_FragColor.rgb *= alpha0;
	gl_FragColor.a = 0;
	break;
    case 2:
	gl_FragColor.rgb *= alpha0;
	gl_FragColor.a *= (1 - alpha0);
	break;
    }
    
    return;
}
