// -*- c -*-
// "clsTensor.h"
// (setq c-basic-offset 4)

#include "clsTensorMath.h"

#include "clsVectorMath.h"

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <float.h>

#define CLS TensorMath_cls

void CLS::
LoadIdentity(double matrix[16]){
    matrix[0] = 1; matrix[4] = 0; matrix[ 8] = 0; matrix[12] = 0;
    matrix[1] = 0; matrix[5] = 1; matrix[ 9] = 0; matrix[13] = 0;
    matrix[2] = 0; matrix[6] = 0; matrix[10] = 1; matrix[14] = 0;
    matrix[3] = 0; matrix[7] = 0; matrix[11] = 0; matrix[15] = 1;
}

void CLS::
Translate(double matrix[16], double pos[3]){
    matrix[12] += pos[0];
    matrix[13] += pos[1];
    matrix[14] += pos[2];
}

void CLS::
Scale(double matrix[16], double scale[3]){
    matrix[0] *= scale[0];
    matrix[1] *= scale[1];
    matrix[2] *= scale[2];
    
    matrix[4] *= scale[0];
    matrix[5] *= scale[1];
    matrix[6] *= scale[2];

    matrix[ 8] *= scale[0];
    matrix[ 9] *= scale[1];
    matrix[10] *= scale[2];

    matrix[12] *= scale[0];
    matrix[13] *= scale[1];
    matrix[14] *= scale[2];
}

void CLS::
RotateX(double matrix[16], double rotX){
    double cs = cos(rotX);
    double sn = sin(rotX);
    double org[16];
    for (int i = 0; i < 16; i++) org[i] = matrix[i];
    matrix[0] = org[0];
    matrix[1] = cs*org[1] - sn*org[2];
    matrix[2] = sn*org[1] + cs*org[2];
    matrix[3] = org[3];

    matrix[4] = org[4];
    matrix[5] = cs*org[5] - sn*org[6];
    matrix[6] = sn*org[5] + cs*org[6];
    matrix[7] = org[7];

    matrix[8] = org[8];
    matrix[9] = cs*org[9]- sn*org[10];
    matrix[10] = sn*org[9] + cs*org[10];
    matrix[11] = org[11];

    matrix[12] = org[12];
    matrix[13] = cs*org[13]- sn*org[14];
    matrix[14] = sn*org[13] + cs*org[14];
    matrix[15] = org[15];
}

void CLS::
RotateY(double matrix[16], double rotY){
    double cs = cos(rotY);
    double sn = sin(rotY);
    double org[16];
    for (int i = 0; i < 16; i++) org[i] = matrix[i];
    matrix[0] = cs*org[0] + sn*org[2];
    matrix[1] = org[2];
    matrix[2] = -sn*org[0] + cs*org[2];
    matrix[3] = org[3];

    matrix[4] = cs*org[4] + sn*org[6];
    matrix[5] = org[5];
    matrix[6] = -sn*org[4] + cs*org[6];
    matrix[7] = org[7];

    matrix[8] = cs*org[8] + sn*org[10];
    matrix[9] = org[9];
    matrix[10] = -sn*org[8] + cs*org[10];
    matrix[11] = org[11];

    matrix[12] = cs*org[12] + sn*org[14];
    matrix[13] = org[13];
    matrix[14] = -sn*org[12] + cs*org[14];
    matrix[15] = org[15];
}

void CLS::
RotateZ(double matrix[16], double rotZ){
    double cs = cos(rotZ);
    double sn = sin(rotZ);
    double org[16];
    for (int i = 0; i < 16; i++) org[i] = matrix[i];
    matrix[0] = cs*org[0] - sn*org[1];
    matrix[1] = sn*org[0] + cs*org[1];
    matrix[2] = org[2];
    matrix[3] = org[3];

    matrix[4] = cs*org[4] - sn*org[5];
    matrix[5] = sn*org[4] + cs*org[5];
    matrix[6] = org[6];
    matrix[7] = org[7];

    matrix[8] = cs*org[8] - sn*org[9];
    matrix[9] = sn*org[8] + cs*org[9];
    matrix[10] = org[10];
    matrix[11] = org[11];

    matrix[12] = cs*org[12] - sn*org[13];
    matrix[13] = sn*org[12] + cs*org[13];
    matrix[14] = org[14];
    matrix[15] = org[15];
}

// m[3]=m[7]=m[11]=0m vec[3]=1 m̏ꍇ
void CLS::
MultiplyVFast(double m[16], double vec[4]){
    double org[3];
    //for (int i = 0; i < 4; i++) org[i] = vec[i];
    memcpy(org, vec, sizeof(double)*3);
    vec[0] = m[0]*org[0] + m[4]*org[1] + m[ 8]*org[2] + m[12];
    vec[1] = m[1]*org[0] + m[5]*org[1] + m[ 9]*org[2] + m[13];
    vec[2] = m[2]*org[0] + m[6]*org[1] + m[10]*org[2] + m[14];
    //vec[3] = m[3]*org[0] + m[7]*org[1] + m[11]*org[2] + m[15]*org[3];
}

void CLS::
MultiplyV(double m[16], double vec[4]){
    double org[4];
    //for (int i = 0; i < 4; i++) org[i] = vec[i];
    memcpy(org, vec, sizeof(double)*4);

    vec[0] = m[0]*org[0] + m[4]*org[1] + m[ 8]*org[2] + m[12]*org[3];
    vec[1] = m[1]*org[0] + m[5]*org[1] + m[ 9]*org[2] + m[13]*org[3];
    vec[2] = m[2]*org[0] + m[6]*org[1] + m[10]*org[2] + m[14]*org[3];
    vec[3] = m[3]*org[0] + m[7]*org[1] + m[11]*org[2] + m[15]*org[3];
}
    
void CLS::
Multiply(double m[16], double matrix1[16]){
    double org[16];
    for (int i = 0; i < 16; i++) org[i] = matrix1[i];

    matrix1[0] = m[0]*org[0] + m[4]*org[1] + m[ 8]*org[2] + m[12]*org[3];
    matrix1[1] = m[1]*org[0] + m[5]*org[1] + m[ 9]*org[2] + m[13]*org[3];
    matrix1[2] = m[2]*org[0] + m[6]*org[1] + m[10]*org[2] + m[14]*org[3];
    matrix1[3] = m[3]*org[0] + m[7]*org[1] + m[11]*org[2] + m[15]*org[3];

    matrix1[4] = m[0]*org[4] + m[4]*org[5] + m[ 8]*org[6] + m[12]*org[7];
    matrix1[5] = m[1]*org[4] + m[5]*org[5] + m[ 9]*org[6] + m[13]*org[7];
    matrix1[6] = m[2]*org[4] + m[6]*org[5] + m[10]*org[6] + m[14]*org[7];
    matrix1[7] = m[3]*org[4] + m[7]*org[5] + m[11]*org[6] + m[15]*org[7];

    matrix1[ 8] = m[0]*org[8] + m[4]*org[9] + m[ 8]*org[10] + m[12]*org[11];
    matrix1[ 9] = m[1]*org[8] + m[5]*org[9] + m[ 9]*org[10] + m[13]*org[11];
    matrix1[10] = m[2]*org[8] + m[6]*org[9] + m[10]*org[10] + m[14]*org[11];
    matrix1[11] = m[3]*org[8] + m[7]*org[9] + m[11]*org[10] + m[15]*org[11];

    matrix1[12] = m[0]*org[12] + m[4]*org[13] + m[ 8]*org[14] + m[12]*org[15];
    matrix1[13] = m[1]*org[12] + m[5]*org[13] + m[ 9]*org[14] + m[13]*org[15];
    matrix1[14] = m[2]*org[12] + m[6]*org[13] + m[10]*org[14] + m[14]*org[15];
    matrix1[15] = m[3]*org[12] + m[7]*org[13] + m[11]*org[14] + m[15]*org[15];
}

void CLS::
Multiply(float m[16], float matrix1[16]){
    float org[16];
    for (int i = 0; i < 16; i++) org[i] = matrix1[i];

    matrix1[0] = m[0]*org[0] + m[4]*org[1] + m[ 8]*org[2] + m[12]*org[3];
    matrix1[1] = m[1]*org[0] + m[5]*org[1] + m[ 9]*org[2] + m[13]*org[3];
    matrix1[2] = m[2]*org[0] + m[6]*org[1] + m[10]*org[2] + m[14]*org[3];
    matrix1[3] = m[3]*org[0] + m[7]*org[1] + m[11]*org[2] + m[15]*org[3];

    matrix1[4] = m[0]*org[4] + m[4]*org[5] + m[ 8]*org[6] + m[12]*org[7];
    matrix1[5] = m[1]*org[4] + m[5]*org[5] + m[ 9]*org[6] + m[13]*org[7];
    matrix1[6] = m[2]*org[4] + m[6]*org[5] + m[10]*org[6] + m[14]*org[7];
    matrix1[7] = m[3]*org[4] + m[7]*org[5] + m[11]*org[6] + m[15]*org[7];

    matrix1[ 8] = m[0]*org[8] + m[4]*org[9] + m[ 8]*org[10] + m[12]*org[11];
    matrix1[ 9] = m[1]*org[8] + m[5]*org[9] + m[ 9]*org[10] + m[13]*org[11];
    matrix1[10] = m[2]*org[8] + m[6]*org[9] + m[10]*org[10] + m[14]*org[11];
    matrix1[11] = m[3]*org[8] + m[7]*org[9] + m[11]*org[10] + m[15]*org[11];

    matrix1[12] = m[0]*org[12] + m[4]*org[13] + m[ 8]*org[14] + m[12]*org[15];
    matrix1[13] = m[1]*org[12] + m[5]*org[13] + m[ 9]*org[14] + m[13]*org[15];
    matrix1[14] = m[2]*org[12] + m[6]*org[13] + m[10]*org[14] + m[14]*org[15];
    matrix1[15] = m[3]*org[12] + m[7]*org[13] + m[11]*org[14] + m[15]*org[15];
}

/*
void CLS::
MultiplyInv(double matrix1[16], double m[16]){
    double org[16];
    for (int i = 0; i < 16; i++) org[i] = matrix1[i];


    matrix1[0] = org[0]*m[0] + org[4]*m[1] + org[ 8]*m[2] + org[12]*m[3];
    matrix1[1] = org[1]*m[0] + org[5]*m[1] + org[ 9]*m[2] + org[13]*m[3];
    matrix1[2] = org[2]*m[0] + org[6]*m[1] + org[10]*m[2] + org[14]*m[3];
    matrix1[3] = org[3]*m[0] + org[7]*m[1] + org[11]*m[2] + org[15]*m[3];

    matrix1[4] = org[0]*m[4] + org[4]*m[5] + org[ 8]*m[6] + org[12]*m[7];
    matrix1[5] = org[1]*m[4] + org[5]*m[5] + org[ 9]*m[6] + org[13]*m[7];
    matrix1[6] = org[2]*m[4] + org[6]*m[5] + org[10]*m[6] + org[14]*m[7];
    matrix1[7] = org[3]*m[4] + org[7]*m[5] + org[11]*m[6] + org[15]*m[7];

    matrix1[ 8] = org[0]*m[8] + org[4]*m[9] + org[ 8]*m[10] + org[12]*m[11];
    matrix1[ 9] = org[1]*m[8] + org[5]*m[9] + org[ 9]*m[10] + org[13]*m[11];
    matrix1[10] = org[2]*m[8] + org[6]*m[9] + org[10]*m[10] + org[14]*m[11];
    matrix1[11] = org[3]*m[8] + org[7]*m[9] + org[11]*m[10] + org[15]*m[11];

    matrix1[12] = org[0]*m[12] + org[4]*m[13] + org[ 8]*m[14] + org[12]*m[15];
    matrix1[13] = org[1]*m[12] + org[5]*m[13] + org[ 9]*m[14] + org[13]*m[15];
    matrix1[14] = org[2]*m[12] + org[6]*m[13] + org[10]*m[14] + org[14]*m[15];
    matrix1[15] = org[3]*m[12] + org[7]*m[13] + org[11]*m[14] + org[15]*m[15];
}
*/

// 3x3̍s񎮂̌vZ
double CLS::
GetDeterminant3x3(double m[9]){
    double det = m[0]*(m[4]*m[8]-m[5]*m[7])+m[1]*(m[5]*m[6]-m[3]*m[8])+m[2]*(m[3]*m[7]-m[4]*m[6]);
    return det;
}

void CLS::
GetSmallMatrix4to3(double src[16], double dst[9], int col, int row){
    int size = 4;
    int cnt = 0;
    for (int i=0; i < size; i++){
        if (i == col) continue;
        for (int j=0; j<size; j++) {
            if (j == row) continue;
            dst[cnt++] = src[i*size+j];
        }
    }
}

// ]ułɒ萔Ŋ
void CLS::
Transpose4x4(double src[16], double dst[16], double div)
{
    float mul = 1.0/div;
    dst[0] = src[0]*mul;
    dst[1] = src[4]*mul;
    dst[2] = src[8]*mul;
    dst[3] = src[12]*mul;
    dst[4] = src[1]*mul;
    dst[5] = src[5]*mul;
    dst[6] = src[9]*mul;
    dst[7] = src[13]*mul;
    dst[8] = src[2]*mul;
    dst[9] = src[6]*mul;
    dst[10] = src[10]*mul;
    dst[11] = src[14]*mul;
    dst[12] = src[3]*mul;
    dst[13] = src[7]*mul;
    dst[14] = src[11]*mul;
    dst[15] = src[15]*mul;
}

void CLS::
Transpose4x4(float src[16], float dst[16], float div)
{
    float mul = 1.0/div;
    dst[0] = src[0]*mul;
    dst[1] = src[4]*mul;
    dst[2] = src[8]*mul;
    dst[3] = src[12]*mul;
    dst[4] = src[1]*mul;
    dst[5] = src[5]*mul;
    dst[6] = src[9]*mul;
    dst[7] = src[13]*mul;
    dst[8] = src[2]*mul;
    dst[9] = src[6]*mul;
    dst[10] = src[10]*mul;
    dst[11] = src[14]*mul;
    dst[12] = src[3]*mul;
    dst[13] = src[7]*mul;
    dst[14] = src[11]*mul;
    dst[15] = src[15]*mul;
}

// 4x4̋tšvZ
double CLS::
GetInvMatrix4x4(double src[16], double dst[16]){

    float det; // s
    double c_mat[9];     // s
    double c[16];     // ]qs

    int cnt = 0;
    for (int i=0; i<4; i++) {
        for (int j=0; j<4; j++) {
            // s̎擾
            GetSmallMatrix4to3(src, c_mat, i, j);
            // ]q̌vZisign(-1)^(i+j)j
	    c[cnt++] = sign[i][j] * GetDeterminant3x3(c_mat);
        }
    }

    // s񎮂vZ
    det = src[0]*c[0]+src[1]*c[1]+src[2]*c[2]+src[3]*c[3];

    if (fabs(det) < DBL_EPSILON){
        printf("Can't calculate inverse matrix\n");
        return det;
    }
    
    // s͗]qs̓]uAts͐ss񎮂Ŋ
    Transpose4x4(c, dst, det);
    
    return det;	
}

float CLS::
GetInvMatrix4x4(float src[16], float dst[16]){
    double src_d[16], dst_d[16];
    for (int i = 0; i < 16; i++) src_d[i] = src[i];
    float det = (float)GetInvMatrix4x4(src_d, dst_d);
    ConvertToFloat(dst, dst_d, 16);
    return det;
}

void CLS::
GetRotationMatrix4x4(float dst[16], float axis_[3], float angle){
    double axis[3];
    for (int i = 0; i < 3; i++) axis[i] = axis_[i];
    VectorMath_cls::Normalize(axis, 1.0f);

    float s = sin(angle);
    float c = cos(angle);
    float oc = 1.0-c;

    dst[ 0] = oc * axis[0] * axis[0] + c;
    dst[ 1] = oc * axis[0] * axis[1] - axis[2] * s;
    dst[ 2] = oc * axis[2] * axis[0] + axis[1] * s;
    dst[ 3] = 0;

    dst[ 4] = oc * axis[0] * axis[1] + axis[2] * s;
    dst[ 5] = oc * axis[1] * axis[1] + c;
    dst[ 6] = oc * axis[1] * axis[2] - axis[0] * s;
    dst[ 7] = 0;

    dst[ 8] = oc * axis[2] * axis[0] - axis[1] * s;
    dst[ 9] = oc * axis[1] * axis[2] + axis[0] * s;
    dst[10] = oc * axis[2] * axis[2] + c;
    dst[11] = 0;

    dst[12] = 0;
    dst[13] = 0;
    dst[14] = 0;
    dst[15] = 1;
}

void CLS::
ConvertToFloat(float *pf, double *pd, int dim){
    for (int i = 0; i < dim; i++) pf[i] = (float)pd[i];
}

int CLS::sign[4][4] = {
    {1, -1, 1, -1}, 
    {-1, 1, -1, 1},
    {1, -1, 1, -1},
    {-1, 1, -1, 1}};
