// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsUseGZ.h"

#include "clsUseGZ.h"

#include <wx/filename.h>
#include <wx/strconv.h>

#include <vector>

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../PREFERENCE/clsPreference.h"

#ifdef ZLIB
#include <zlib.h>
#endif

void UseGZ_cls::
VersionCheck(){
    // version 1.2 nł邱Ƃ͉
#ifdef ZLIB
    wxString strZLib = wxString::FromAscii(zlibVersion());
    strZLib.Mid(0, 1).ToLong(&msVersionMajor);
    strZLib.Mid(2, 1).ToLong(&msVersionSub);
    strZLib.Mid(4, 1).ToLong(&msVersionSubSub);
#else
    msVersionMajor = 0; msVersionSub = 0; msVersionSubSub = 0;
#endif
}

bool UseGZ_cls::
CopyToString(std::string *pstrBuf, wxString strFile){
    wxFileName CName(strFile);
    if (CName.GetExt().IsSameAs(wxT("gz")))
	return CopyToStringGZ(pstrBuf, strFile);

#ifdef WIN32
    FILE *fp = fopen(wxFNCONV(strFile).char_str(), "rb");
#else
    FILE *fp = fopen(wxFNCONV(strFile), "rb");
#endif
    if (fp == NULL) return FALSE;

    fpos_t fsize;//
    fseek(fp,0,SEEK_END); 
    fgetpos(fp,&fsize); 
    fseek(fp,0,SEEK_SET); 

#ifdef WIN32
    long long size = fsize;
#else
    long long size = fsize.__pos;
#endif

    pstrBuf->reserve(size+1); pstrBuf->resize(size+1);
    (*pstrBuf)[size] = '\0'; // t@C̓kƌȂB

    fread(&((*pstrBuf)[0]), sizeof(char), size, fp);
    fclose(fp);
    return TRUE;
}

#ifdef ZLIB
bool UseGZ_cls::
CopyToStringGZ(std::string *pstrBuf, wxString strFile){
#ifdef WIN32
    gzFile fp = gzopen(wxFNCONV(strFile).char_str(), "rb");
#else
    gzFile fp = gzopen(wxFNCONV(strFile), "rb");
#endif
    
    if (fp == NULL) return FALSE;

    int gzBufferSize = 1024*512;

    gzbuffer(fp, gzBufferSize);

    int sum = 0;
    std::vector < char* > VpBuf;

    int index = 0;
    while (TRUE){
	VpBuf.push_back(new char [gzBufferSize]);
	int ret = gzread(fp, VpBuf[index], gzBufferSize);
	if (ret == 0 || ret == -1) break;
	sum += ret;
	index ++;
    }

    pstrBuf->reserve(sum+1); pstrBuf->resize(sum+1);
    (*pstrBuf)[sum] = '\0'; // t@C̓kƌȂB

    int current = 0;
    for (int i = 0; i < VpBuf.size(); i++){
	int size = sum - current;
	if (size > gzBufferSize) size = gzBufferSize;
	memcpy(&((*pstrBuf)[0])+current, VpBuf[i], size);
	current += size;
	delete [] VpBuf[i];
    }
    
    gzclose(fp);

    return TRUE;
}
#else
bool UseGZ_cls::
CopyToStringGZ(std::string *pstrBuf, wxString strFile){
    return FALSE;
}
#endif 

bool UseGZ_cls::
CopyToMemory(unsigned char **ppBuf, wxString strFile, long long &fileSize){
    
    wxFileName CName(strFile);
    if (CName.GetExt().IsSameAs(wxT("gz")))
	return CopyToMemoryGZ(ppBuf, strFile, fileSize);

#ifdef WIN32
    FILE *fp = fopen(wxFNCONV(strFile).char_str(), "rb");
#else
    FILE *fp = fopen(wxFNCONV(strFile), "rb");
#endif
    if (fp == NULL) return FALSE;
    
    fpos_t fsize;
    fseek(fp,0,SEEK_END); 
    fgetpos(fp,&fsize); 
    fseek(fp,0,SEEK_SET); 

#ifdef WIN32
    long long size = fsize;
#else
    long long size = fsize.__pos;
#endif
    fileSize = size + 1;
    
    *ppBuf = new unsigned char [size+1];
    (*ppBuf)[size] = '\0'; // t@C̓kƌȂB
    
    fread(*ppBuf, sizeof(char), size, fp);
    fclose(fp);
    return TRUE;
}

#ifdef ZLIB
#include <zlib.h>
bool UseGZ_cls::
CopyToMemoryGZ(unsigned char **ppBuf, wxString strFile, long long &fileSize){
#ifdef WIN32
    gzFile fp = gzopen(wxFNCONV(strFile).char_str(), "rb");
#else
    gzFile fp = gzopen(wxFNCONV(strFile), "rb");
#endif
    
    if (fp == NULL) return FALSE;
    //gzbuffer(fp, 1024*128);

    int sum = 0;
    std::vector < char* > VpBuf;

    int index = 0;
    while (TRUE){
	VpBuf.push_back(new char [8192]);
	int ret = gzread(fp, VpBuf[index], 8192);
	if (ret == 0 || ret == -1) break;
	sum += ret;
	index ++;
    }

    *ppBuf = new unsigned char [sum+1];
    (*ppBuf)[sum] = '\0'; // t@C̓kƌȂB

    fileSize = sum+1;

    int current = 0;
    for (int i = 0; i < VpBuf.size(); i++){
	int size = sum - current;
	if (size > 8192) size = 8192;
	memcpy((*ppBuf)+current, VpBuf[i], size);
	current += size;
	delete [] VpBuf[i];
    }
    
    gzclose(fp);

    return TRUE;
}
#else
bool UseGZ_cls::
CopyToMemoryGZ(unsigned char **ppBuf, wxString strFile, long long &fileSize){
    return FALSE;
}
#endif 

int UseGZ_cls::
GetLineNumber(wxString strFile){
    wxFileName CName(strFile);
    if (CName.GetExt().IsSameAs(wxT("gz")))
	return GetLineNumberGZ(strFile);

#ifdef WIN32
    FILE *fp = fopen(wxFNCONV(strFile).char_str(), "rb");
#else
    FILE *fp = fopen(wxFNCONV(strFile), "rb");
#endif
    if (fp == NULL) return 0;
	
    fseek(fp, 0, SEEK_SET );
    long ini = ftell(fp);
    fseek(fp, 0, SEEK_END );
    long fin = ftell(fp);
    long length = fin - ini;

    fseek(fp, 0, SEEK_SET );
    int count = 0;
    char buf [1024*8];
    for (long i = 0; i < length; i += 1024*8){
	int size = 1024*8;
	if (length - i < size) size = length - i;
	fread(buf, sizeof(unsigned char), size, fp);
	for (int j = 0; j < size; j++)
	    if (buf[j] == '\n') count++;
	if (size + i == length) // Ō̍s̍ŌオsȂꍇ
	    if (buf[size-1] != '\n') count++;
    }
    fclose(fp);

    return count;
}

#ifdef ZLIB
int UseGZ_cls::
GetLineNumberGZ(wxString strFile){

#ifdef WIN32
    gzFile fp = gzopen(wxFNCONV(strFile).char_str(), "rb");
#else
    gzFile fp = gzopen(wxFNCONV(strFile), "rb");
#endif
    if (fp == NULL) return 0;
	
    int sum = 0;
    std::vector < char* > VpBuf;
    
    //int index = 0;
    while (TRUE){
	char *buf = new char [8192];
	int ret = gzread(fp, buf, 8192);
	if (ret == -1 || ret == 0){ // Ō܂ōŝŋ
	    delete [] buf;
	    break;
	}
	VpBuf.push_back(buf);
	sum += ret;
	//index ++;
    }

    int count = 0;
    int current = 0;
    for (int i = 0; i < VpBuf.size(); i++){
	int size = sum - current;
	if (size > 8192) size = 8192;

	for (int j = 0; j < size; j++)
	    if (VpBuf[i][j] == '\n') count++;

	if (i == VpBuf.size() - 1) // Ō̍s̍ŌオsȂꍇ
	    if (VpBuf[i][size-1] != '\n') count++;
	
	current += size;
	delete [] VpBuf[i];
    }

    gzclose(fp);

    return count;
}
#else
int UseGZ_cls::
GetLineNumberGZ(wxString strFile){
    return 0;
}
#endif

#ifdef ZLIB
bool UseGZ_cls::
SaveMemoryToGZ(wxString strFile, char *pBuf, long long size){

    int gzQuality = (GCClassProvider.GetPCPreference()->mCGZQuality.GetBodyINT() == 1) ? 1 : 9;
    wxString strTag = wxT("wb");
    strTag << gzQuality;

#ifdef WIN32
    gzFile fp = gzopen(wxFNCONV(strFile).char_str(), strTag.char_str());
#else
    gzFile fp = gzopen(wxFNCONV(strFile), strTag.char_str());
#endif
    if (fp == NULL) return FALSE;

    gzwrite(fp, (void*)pBuf, size);

    gzclose(fp);
    
    return TRUE;
}
#else
bool UseGZ_cls::
SaveMemoryToGZ(wxString strFile, char *pBuf, long long size){
    return FALSE;
}
#endif

long UseGZ_cls::msVersionMajor;
long UseGZ_cls::msVersionSub;
long UseGZ_cls::msVersionSubSub;
