// -*- c -*-
// "clsVectorMath.cpp"
// (setq c-basic-offset 4)

#include "clsVectorMath.h"

#include "CommonDefinitions.h"

#include <math.h>

#define CLS VectorMath_cls

void CLS::
Normalize(double pos[3], double length){
    double norm = sqrt(pos[0]*pos[0] + pos[1]*pos[1] + pos[2]*pos[2]);
    LOOP_XYZ pos[xyz] = pos[xyz]/norm;
}

void CLS::
Normalize(float pos[3], float length){
    float norm = sqrt(pos[0]*pos[0] + pos[1]*pos[1] + pos[2]*pos[2]);
    LOOP_XYZ pos[xyz] = pos[xyz]/norm;
}

double CLS::
GetScalarProduct(double vec0[3], double vec1[3]){
    return vec0[X]*vec1[X] + vec0[Y]*vec1[Y] + vec0[Z]*vec1[Z];
}

float CLS::
GetScalarProduct(float vec0[3], float vec1[3]){
    return vec0[X]*vec1[X] + vec0[Y]*vec1[Y] + vec0[Z]*vec1[Z];
}

double CLS::
GetNorm(double vec[3]){
    return sqrt(GetScalarProduct(vec, vec)); }

float CLS::
GetNorm(float vec[3]){
    return sqrt(GetScalarProduct(vec, vec)); }

double CLS::
GetDistance(double vec0[3], double vec1[3]){
    double rel[3];
    LOOP_XYZ rel[xyz] = vec1[xyz] - vec0[xyz];
    return sqrt(GetScalarProduct(rel, rel));
}

float CLS::
GetDistance(float vec0[3], float vec1[3]){
    float rel[3];
    LOOP_XYZ rel[xyz] = vec1[xyz] - vec0[xyz];
    return sqrt(GetScalarProduct(rel, rel));
}

void CLS::
GetCrossedVector(double vec0[3], double vec1[3], double vecOut[3]){
    vecOut[X] = vec0[Y]*vec1[Z] - vec0[Z]*vec1[Y];
    vecOut[Y] = vec0[Z]*vec1[X] - vec0[X]*vec1[Z];
    vecOut[Z] = vec0[X]*vec1[Y] - vec0[Y]*vec1[X];
}

void CLS::
GetCrossedVector(const float vec0[3], const float vec1[3], float vecOut[3]){
    vecOut[X] = vec0[Y]*vec1[Z] - vec0[Z]*vec1[Y];
    vecOut[Y] = vec0[Z]*vec1[X] - vec0[X]*vec1[Z];
    vecOut[Z] = vec0[X]*vec1[Y] - vec0[Y]*vec1[X];
}

float CLS::
GetCrossedVectorZ(const float vec0[3], const float vec1[3]){
    return vec0[X]*vec1[Y] - vec0[Y]*vec1[X];
}


void CLS::
Rotate(double vec[3], double angle, int axis){
    double org[3];
    LOOP_XYZ org[xyz] = vec[xyz];
    double cosA = cos(angle * 3.14159 / 180);
    double sinA = sin(angle * 3.14159 / 180);
    switch (axis){
    case X:
	vec[Y] = cosA * org[Y] - sinA * org[Z];
	vec[Z] = sinA * org[Y] + cosA * org[Z];
	break;
    case Y:
	vec[Z] = cosA * org[Z] - sinA * org[X];
	vec[X] = sinA * org[Z] + cosA * org[X];
	break;
    case Z:
	vec[X] = cosA * org[X] - sinA * org[Y];
	vec[Y] = sinA * org[X] + cosA * org[Y];
	break;
    }
} 

void CLS::
RotateAround(double vec[3], double axis[3], double angle){
    double mat[3][3];
    double cosA = cos(angle * 3.14159 /180);
    double sinA = sin(angle * 3.14159 /180);
    double org[3];
    LOOP_XYZ org[xyz] = vec[xyz];
    double orgaxis[3];
    LOOP_XYZ orgaxis[xyz] = axis[xyz];
    
    mat[X][X] = (1.0-cosA) * axis[X]*axis[X] + cosA;
    mat[Y][X] = axis[X]*axis[Y] * (1.0-cosA) + axis[Z]*sinA;
    mat[Z][X] = axis[X]*axis[Z] * (1.0-cosA) - axis[Y]*sinA;
    mat[X][Y] = axis[X]*axis[Y] * (1.0-cosA) - axis[Z]*sinA;
    mat[Y][Y] = (1.0-cosA) * axis[Y]*axis[Y] + cosA;
    mat[Z][Y] = axis[Y]*axis[Z] * (1.0-cosA) + axis[X]*sinA;
    mat[X][Z] = axis[X]*axis[Z] * (1.0-cosA) + axis[Y]*sinA;
    mat[Y][Z] = axis[Y]*axis[Z] * (1.0-cosA) - axis[X]*sinA;
    mat[Z][Z] = (1.0-cosA) * axis[Z]*axis[Z] + cosA;
    
    vec[X] = mat[X][X]*org[X] + mat[Y][X]*org[Y] + mat[Z][X]*org[Z];
    vec[Y] = mat[X][Y]*org[X] + mat[Y][Y]*org[Y] + mat[Z][Y]*org[Z];
    vec[Z] = mat[X][Z]*org[X] + mat[Y][Z]*org[Y] + mat[Z][Z]*org[Z];
}
