// (setq c-basic-offset 4)
// "clsParamDialog.h"

#include "clsParamDialog.h"
#include "clsDialogBase.h"

#include "../CONFITEM/clsConfItem.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../PARAMETER/clsParameter.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../PARAMETER/clsKeyFrame.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"
#include "../COMMON/CommonDefinitions.h"

#define CLS ParamDialog_cls

CLS::
ParamDialog_cls(){
    mCPos = wxDefaultPosition;
    mpCDLG = NULL;
    mpCMNum = NULL;
}

void CLS::
OpenWindow(wxWindow *pCParent, Parameter_cls *pCParam, int keyIndex,
	   wxString strItem){

    wxString strName = pCParam->mpCParent->GetStrName() + ("::") + pCParam->GetStrName();

    mpCDLG = new DialogBase_cls(pCParent, wxID_ANY, strName, mCPos);

    int frameMax = GetPCTimeCtrl()->mCFrameMax.GetBodyINT();
    if (pCParam->mpCParent->mbForColorCurve == TRUE)
	frameMax = _COLOR_CURVE_RESOLUTION;

    double oldXNorm = pCParam->mVCKeys[keyIndex].mXNormalized;
    double oldValue = pCParam->mVCKeys[keyIndex].mValue;
    int frame = oldXNorm * frameMax;
    int oldKeyType = pCParam->mVCKeys[keyIndex].mKeyType;
    double oldEaseFactor = pCParam->mVCKeys[keyIndex].mEaseFactor;
    
    wxString strFrame; strFrame << frame;
    wxString strValue; strValue << oldValue;
    wxString strKeyType; strKeyType << oldKeyType;
    wxString strEaseFactor; strEaseFactor << oldEaseFactor;
    mCFrame.SetItem("Dummy", strFrame);
    mCValue.SetItem("Dummy", strValue);
    mCKeyType.SetItem("Dummy", strKeyType);
    mCEaseFactor.SetItem("Dummy", strEaseFactor);

    wxBoxSizer *pCSizerMain = new wxBoxSizer(wxVERTICAL);
    mpCMNum = new MultiInputPanel_cls(mpCDLG, "Parameter Key", TRUE);
    mpCMNum->CombineConfItem(mCFrame, _("Sequence Index"));
    mpCMNum->CombineConfItem(mCValue, "Value");
    
    mpCMNum->SetChoiceList(3, GetPCStringProvider()->GetAstr("KeyFrameType"));
    int indexKeyType = mpCMNum->CombineConfItem(mCKeyType, _("Interpolate Type"));
    mIndexEaseFactor = mpCMNum->CombineConfItem(mCEaseFactor, _("Ease Factor"));
    
    mpCMNum->FinishSetting();

    pCSizerMain->Add(mpCMNum, 1, wxEXPAND);
    mpCDLG->AddBasicButtons(pCSizerMain);

    mpCDLG->SetSizerAndFit(pCSizerMain);
    mpCMNum->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMBOBOX, indexKeyType);
    HideAndShow();

    if (mpCDLG->ShowModal() == wxID_OK){
	mpCMNum->SyncFromGUI();

	double newXNorm = (double)mCFrame.GetBodyINT() / frameMax;
	double newValue = mCValue.GetBodyDOUBLE();
	int newKeyType = mCKeyType.GetBodyINT();
	double newEaseFactor = mCEaseFactor.GetBodyDOUBLE();
	
	wxString strTag;
	if (strItem.Find("Scene Filter") == 0){
	    PopStrItemFromStrLine(strItem, "::");
	    strTag = GetPCFilterCtrl()->GetStrSelf() << strItem;
	}
	else
	    strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strItem;
	
	double dX = newXNorm - oldXNorm;
	double dY = newValue - oldValue;

	if (dX != 0 || dY != 0 || oldKeyType != newKeyType ||
	    oldEaseFactor != newEaseFactor){
	    GetPCUndo()->OpenBlock("KeyDialog");
	    pCParam->MoveTLKeyCommand(keyIndex, dX, dY, strTag) ;
	    pCParam->SetKeyTypeCommand(keyIndex, newKeyType, strTag);
	    pCParam->SetKeyEaseFactorCommand(keyIndex, newEaseFactor, strTag);
	    GetPCUndo()->CloseBlock("KeyDialog");
	}
    };

    mCPos = mpCDLG->GetPosition();

    delete mpCDLG;
    mpCMNum = NULL;
    mpCDLG = NULL;
}

void CLS::
HideAndShow(){
    bool bHermiteAuto = mCKeyType.GetBodyINT() == _KEY_HERMITE_AUTO;
    mpCMNum->mVCItem[mIndexEaseFactor].Show(bHermiteAuto);

    mpCDLG->ResizeDialog();
    mpCDLG->AdjustWindowFitting();
}

void CLS::
OnCombo(wxCommandEvent& event){
    HideAndShow();
}

ParamDialog_cls *GpCParamDialog = NULL;

