// (setq-c-offset 4)
// "clsBlur.h"

#include "clsBlur.h"
#include "clsFilterCtrl.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MISC/clsMisc.h"
#include "../MYGL/clsGLExtention.h"

#include "../PARAMETER/clsParamGroup.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#define CLS Blur_cls

CLS::
Blur_cls(wxString strName) : FilterBase_cls("Blur_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_SCENE";

    wxString strConfListLabel = "FilterCtrl::"; strConfListLabel << strName;
    mCConfList.SetTag(strConfListLabel);
    mCConfList.Add(&mCType, "Type=", "0", strUpdate);
    mCConfList.Add(&mCKernel, "Kernel=", "1", strUpdate);
    mCConfList.Add(&mCQuality, "Quality=", "3", strUpdate);
    mCConfList.Add(&mCbUseGamma, "UseGamma=", "1", strUpdate);

    AddOneDimParam("Width", "W", 0.01);
}

void CLS::
SyncFromBuffer(BlurBuffer_cls *pCBuffer){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VWidth;
    VWidth.push_back(pCBuffer->mCWidth.GetBodyDOUBLE());
    SyncronizeParamForFilter("Width", VWidth);
}

unsigned int CLS::
GetIterNum(){
    int exposure = 1;
    if (mCbUseGamma.GetBodyBOOL() == TRUE) exposure = 2;
    if (mCType.GetBodyINT() == 2) return 2 * exposure;
    return 1 * exposure;
} 

void CLS::
BeginExposure(){
    ExposureParams_cls CParam;
    CParam.mGamma = 0.4545;
    GetPCShaderCtrl()->BeginExposure(&CParam);
}

void CLS::
Begin(int iter){
    double time = GetPCTimeCtrl()->GetCurrentNT();

    bool bUseGamma = mCbUseGamma.GetBodyBOOL();

    if (bUseGamma == TRUE){
	if (mCType.GetBodyINT() == 2 && (iter == 1 || iter == 3))
	    return BeginExposure();
	
	if (mCType.GetBodyINT() < 2 && iter == 1)
	    return BeginExposure();
    }

    BlurParams_cls CParam;
    if (mCType.GetBodyINT() == 2){
	if (bUseGamma == TRUE){
	    if (iter == 0) CParam.mType = 0; 
	    if (iter == 2) CParam.mType = 1;
	}
	else{
	    if (iter == 0) CParam.mType = 0; 
	    if (iter == 1) CParam.mType = 1;
	}
    }
    else{
	CParam.mType = mCType.GetBodyINT();
    }

    if (bUseGamma == TRUE) CParam.mGamma = 2.2;     
    else                   CParam.mGamma = 1;
    
    CParam.mKernel = mCKernel.GetBodyINT();
    CParam.mWidth = GetPCParamGroup("Width")->GetVInterpolatedValue((float)time)[0];
    CParam.mQuality = mCQuality.GetBodyINT();

    GetPCShaderCtrl()->BeginShader("Blur", &CParam);
}

void CLS::
End(){
    GetPCShaderCtrl()->EndShader();    
}

#define CLS BlurBuffer_cls

CLS::
BlurBuffer_cls(Blur_cls *pCSrc) : FilterBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCType, "Type=", "0");
    mCConfList.Add(&mCKernel, "Kernel=", "1");
    mCConfList.Add(&mCQuality, "Quality=", "3");
    mCConfList.Add(&mCbUseGamma, "UseGamma=", "1");
    //
    SyncFromSrc();
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
		new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);
    
    pC->CombineSizer(2, _("Type"));
    wxString AstrType[3] = {_("Horizontal"), _("Vertical"), _("Both")};
    pC->SetChoiceList(3, AstrType);
    pC->CombineConfItem(mCType, _("Type"));
    wxString AstrKernel[2] = {_("Box"), _("Gaussian")};
    pC->SetChoiceList(2, AstrKernel);
    pC->CombineConfItem(mCKernel, _("Type"));
    
    pC->CombineSizer(2, _("Width/Quality"));
    pC->SetLimit(0, 0.2);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCWidth);
    pC->SetLimit(0, 32);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCQuality);

    pC->CombineSizer(1, _("Gamma Correction"));
    pC->CombineConfItem(mCbUseGamma, _("Use"));
    
    pC->FinishSetting();
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    mCWidth.SetBodyFromData(mpCSrc->GetPCParamGroup("Width")->GetVInterpolatedValue((float)time, FALSE));
}

void CLS::
ForceSrcToSyncBuffer(){ mpCSrc->SyncFromBuffer(this); }

