// (setq-c-offset 4)
// "clsExposure.h"

#include "clsExposure.h"
#include "clsFilterCtrl.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsGLExtention.h"

#include "../PARAMETER/clsParamGroup.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#define CLS Exposure_cls

CLS::
Exposure_cls(wxString strName) : FilterBase_cls("Exposure_cls", strName){

    wxString strConfListLabel = "FilterCtrl::"; strConfListLabel << strName;
    mCConfList.SetTag(strConfListLabel);
    
    ParamGroup_cls *pCLevel = GetPCNewParamGroup("Level", "Min", "Max");
    InitializeAndPushParam(pCLevel, 0, 1);

    AddOneDimParam("Gamma", "G", 1);
    AddOneDimParam("Hue", "H", 0);
    AddOneDimParam("Saturation", "S", 0);
}

void CLS::
SyncFromBuffer(ExposureBuffer_cls *pCBuffer){

    std::vector < float > VLevel = pCBuffer->mCLevel.GetBodyFLOATARRAY();
    SyncronizeParamForFilter("Level", VLevel);
    
    std::vector < float > VGamma;
    VGamma.push_back(pCBuffer->mCGamma.GetBodyDOUBLE());
    SyncronizeParamForFilter("Gamma", VGamma);

    std::vector < float > VHue;
    VHue.push_back(pCBuffer->mCHue.GetBodyDOUBLE());
    SyncronizeParamForFilter("Hue", VHue);

    std::vector < float > VSaturation;
    VSaturation.push_back(pCBuffer->mCSaturation.GetBodyDOUBLE());
    SyncronizeParamForFilter("Saturation", VSaturation);
}

void CLS::
Begin(){
    double time = GetPCTimeCtrl()->GetCurrentNT();

    ExposureParams_cls CParam;
    CParam.mMin = GetPCParamGroup("Level")->GetVInterpolatedValue((float)time)[0];
    CParam.mMax = GetPCParamGroup("Level")->GetVInterpolatedValue((float)time)[1];
    CParam.mGamma = GetPCParamGroup("Gamma")->GetVInterpolatedValue((float)time)[0];
    CParam.mHue = GetPCParamGroup("Hue")->GetVInterpolatedValue((float)time)[0];
    CParam.mSaturation = GetPCParamGroup("Saturation")->GetVInterpolatedValue((float)time)[0];

    GetPCShaderCtrl()->BeginShader("Exposure", &CParam);
}

void CLS::
End(){
    GetPCShaderCtrl()->EndShader();    
}

#define CLS ExposureBuffer_cls

CLS::
ExposureBuffer_cls(Exposure_cls *pCSrc) : FilterBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    //mCConfList.Add(&mCLevel, "Level=", "0 1");
    //
    SyncFromSrc();
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->SetLimit(0, 16);
    pC->mDragScale = 0.3;
    pC->CombineConfItem(mCLevel, _("Level"));
        
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCGamma, _("Gamma"));

    pC->CombineSizer(2, _("Hue/Saturation"));
    pC->mDragScale = 10;
    pC->CombineConfItem(mCHue);
    pC->mDragScale = 10;
    pC->CombineConfItem(mCSaturation);
        
    pC->FinishSetting();
    return pC;
}


void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    mCLevel.SetBodyFromData(mpCSrc->GetPCParamGroup("Level")->GetVInterpolatedValue((float)time, FALSE));
    mCGamma.SetBodyFromData(mpCSrc->GetPCParamGroup("Gamma")->GetVInterpolatedValue((float)time, FALSE));
    mCHue.SetBodyFromData(mpCSrc->GetPCParamGroup("Hue")->GetVInterpolatedValue((float)time, FALSE));
    mCSaturation.SetBodyFromData(mpCSrc->GetPCParamGroup("Saturation")->GetVInterpolatedValue((float)time, FALSE));
}

void CLS::
ForceSrcToSyncBuffer(){ mpCSrc->SyncFromBuffer(this); }

