// (setq c-basic-offset 4)
// "clsFilterCtrl.h"

#include "clsFilterCtrl.h"

#include "clsFilterBase.h"
#include "clsExposure.h"
#include "clsBlur.h"
#include "clsFocalBlur.h"
#include "clsGlare.h"
#include "clsRemoveAlpha.h"
#include "clsFilterOverlay.h"
#include "clsFilterLegend.h"
#include "clsSSAO.h"
#include "clsFilterDepth.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsGLExtention.h"
#include "../MYGL/clsFBO.h"

#include "../PARAMETER/clsParamGroup.h"
#include "../SCENEOBJ/clsAttribute.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../UNDO/clsUndoManager.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"

#ifdef WIN32
extern PFNGLCOPYIMAGESUBDATANVPROC glCopyImageSubData;
#endif

#define CLS FilterCtrl_cls

CLS::
FilterCtrl_cls(){
    SetToDefault();
}

CLS::
~FilterCtrl_cls(){ Clear(); }

void CLS::
SetToDefault(){
    mpCActiveFilter = NULL;
    mCConfList.Clear();

    mCConfList.SetTag("FilterCtrl");
    mCConfList.Add(&mCstrActiveItemName, "DUMMY");
}

void CLS::
Clear(){
    ClearSelection();
    Select(wxES, FALSE, FALSE);

    while( mVpCFilter.empty() == FALSE){
	Attribute_cls *pCLastOne = mVpCFilter.back();
	DeleteFilter(pCLastOne);
	//mVpCFilter.pop_back();
    }
}

void CLS::
ClearSelection(){
    /*
    for (uint i = 0; i < mVpCSelectedObject.size(); i++)
	mVpCSelectedObject[i]->ReferRelease();
    */
    mVpCSelectedFilter.clear();
}

wxString CLS::
GetStrSelf(){ return "FilterCtrl::"; }

void CLS::
GetVstrParamTag(std::vector < wxString > &VstrTag){
    uint numFilter = mVpCFilter.size();
    for (uint i = 0; i < numFilter; i++)
	GetPCFromIndex(i)->GetVstrParamTag("Scene Filter", VstrTag);
}

FilterBase_cls* CLS::
GetPCFromAddress(wxString strAddress){
    wxString strRoot = PopStrItemFromStrLine(strAddress, "::");
    if (strRoot.IsSameAs("Scene Filter"))
	return GetPCFromName(strAddress);
    return NULL;
}

FilterBase_cls* CLS::
GetPCFromName(wxString& strName){
    uint numFilter = mVpCFilter.size();
    for (uint i = 0; i < numFilter; i++){
	FilterBase_cls *pCFilter = GetPCFromIndex(i);
	if (pCFilter->mCstrName.mstrBody.IsSameAs(strName) == TRUE) return pCFilter;
    }
    return NULL;
}

FilterBase_cls* CLS::
GetPCFromIndex(int index){
    int numFilter = mVpCFilter.size();
    if (index >= numFilter || index < 0) return NULL;
    return (FilterBase_cls*)*GetIterFromIndex(index);
}

int CLS::
GetIndexFromPC(Attribute_cls *pC){
    int index = 0;
    for (std::list< Attribute_cls* >::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if (*itr == pC) return index;
	index++;
    }
    return -1;
}


std::list< Attribute_cls* >::iterator CLS::
GetIterFromIndex(int index){
    int numFilter = mVpCFilter.size();
    if (index >= numFilter || index < 0) return mVpCFilter.end();
    std::list< Attribute_cls* >::iterator itr = mVpCFilter.begin();
    for (int i = 0; i < index; i++){
	if (itr == mVpCFilter.end()) return itr;
	itr++;
    }
    return itr;
}

std::list< Attribute_cls* >::iterator CLS::
GetIterFromPC(Attribute_cls *pCAttrib){
    return GetIterFromIndex(GetIndexFromPC(pCAttrib));
}

ParamGroup_cls* CLS::
GetPCParamGroup(wxString strTag){

    wxString strFirstItem = PopStrItemFromStrLine(strTag, "::");

    if (strFirstItem.IsSameAs("Scene Filter"))
	strFirstItem = PopStrItemFromStrLine(strTag, "::");
    if (strFirstItem.IsSameAs("FilterCtrl"))
	strFirstItem = PopStrItemFromStrLine(strTag, "::");

    uint numFilter = mVpCFilter.size();
    for (uint i = 0; i < numFilter; i++){
	if (GetPCFromIndex(i)->mCstrName.GetBodySTR().IsSameAs(strFirstItem)){
	    strFirstItem = PopStrItemFromStrLine(strTag, "::");
	    return GetPCFromIndex(i)->GetPCParamGroup(strFirstItem);
	}
    }
    return NULL;
}

Parameter_cls* CLS::
GetPCParameter(wxString strTag){
    wxString strName = PopStrItemFromStrLine(strTag, wxT("::"));

    if (strName.IsSameAs("Scene Filter"))
	strName = PopStrItemFromStrLine(strTag, wxT("::"));

    Attribute_cls *pCFilter = GetPCFromName(strName);
    if (pCFilter == NULL) return NULL;
    wxString strFirstItem = PopStrItemFromStrLine(strTag, wxT("::"));
    ParamGroup_cls *pCParamGroup = pCFilter->GetPCParamGroup(strFirstItem);
    if (pCParamGroup == NULL) return NULL;

    if (strTag.Find(pCParamGroup->mCstrName.GetBodySTR()+"::") == 0)
	strTag.Replace(pCParamGroup->mCstrName.GetBodySTR()+"::", wxES);

    return pCParamGroup->GetPCParameter(strTag);
}

bool CLS::
IsSelected(Attribute_cls *pCAttrib){
    uint numFilter = mVpCSelectedFilter.size();
    for (uint i = 0; i < numFilter; i++)
	if (mVpCSelectedFilter[i] == pCAttrib) return TRUE;
    return FALSE;
}

bool CLS::
IsVaildFilterExist(wxString strAttributeName){
    uint numFilter = mVpCFilter.size();
    for (uint i = 0; i < numFilter; i++){
	FilterBase_cls *pCFilter = GetPCFromIndex(i);
	if (pCFilter->IsAttributeSameAs(strAttributeName))
	    if (pCFilter->mCbValidity.GetBodyBOOL() == TRUE) return TRUE;
    }
    return FALSE;
}

bool CLS::
IsFilterExist(wxString strName){
    uint numFilter = mVpCFilter.size();
    for (uint i = 0; i < numFilter; i++)
	if (GetPCFromIndex(i)->GetStrName().IsSameAs(strName)) return TRUE;
    return FALSE;
}

#ifdef WIN32
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
#endif

FBO_cls* CLS::
GetNewPCFilterProcessed(FBO_cls *pCFBOInSrc,
			std::vector < float > VPasspartoutInfo,
			PositionAndDirection_cls CCameraPos,
			Camera_cls *pCCamera, FBO_cls *pCFBODepth){

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int width = pCFBOInSrc->mWidthFBO;
    int height = pCFBOInSrc->mHeightFBO;

    FBO_cls *pCFBOIn = pCFBOInSrc;
    FBO_cls *pCFBOOut = NULL;
    uint numFilter = mVpCFilter.size();
    
    std::vector < FilterBase_cls* > VpCValidFilter;
    for (uint i = 0; i < numFilter; i++){
	if (GetPCFromIndex(i)->mCbValidity.GetBodyBOOL() == TRUE)
	    VpCValidFilter.push_back(GetPCFromIndex(i));
    }

    GLTexture_cls *pCDepthTexture = NULL;//new GLTexture();
    if (pCFBODepth != NULL && pCFBODepth->mpCDepthTexture != NULL){
	pCDepthTexture = new GLTexture_cls();
	if (glCopyImageSubData == NULL){
	    float *pBuffer = new float [width*height];
	    pCFBODepth->mpCDepthTexture->BindTexture();
	    glGetTexImage( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, GL_FLOAT, pBuffer );
	    pCDepthTexture->CopyFromDataFloat(pBuffer, width, height);
	    delete [] pBuffer;
	}
	else{
	    
	    glGenTextures(1, &(pCDepthTexture->mTextureID));
	    pCDepthTexture->BindTexture();
	    glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT24, width, height,
			 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	    
	    glBindTexture(GL_TEXTURE_2D, 0);

	    glCopyImageSubData(pCFBODepth->mpCDepthTexture->mTextureID, 
			       GL_TEXTURE_2D, 0, 0, 0, 0,
			       pCDepthTexture->mTextureID, 
			       GL_TEXTURE_2D, 0, 0, 0, 0,
			       width, height, 1);
	}
    }
    
    uint numValidFilter = VpCValidFilter.size();
    for (uint i = 0; i < numValidFilter; i++){

	uint numIter = 1;
	if (VpCValidFilter[i]->IsAttributeSameAs("Blur_cls"))
	    numIter = ((Blur_cls*)VpCValidFilter[i])->GetIterNum();
	/*
	if (VpCValidFilter[i]->IsAttributeSameAs("FocalBlur_cls"))
	    numIter = ((FocalBlur_cls*)VpCValidFilter[i])->GetIterNum();
	*/

	

	for (uint j = 0; j < numIter; j++){
    
	    pCFBOOut = new FBO_cls(width, height,     
				   pCFBOInSrc->mpCGLCanvasParent);

	    pCFBOOut->MakeFBO(TRUE);
	    
	    FilterBase_cls *pCFilter = VpCValidFilter[i];
	    wxString strAttributeName = pCFilter->GetStrAttributeName();

	    int depthTexture = 0;
	    //bool bUseDepth = FALSE;

	    if (strAttributeName.IsSameAs("Exposure_cls"))
		((Exposure_cls*)pCFilter)->Begin();
	    if (strAttributeName.IsSameAs("Blur_cls"))
		((Blur_cls*)pCFilter)->Begin(j);
	    if (strAttributeName.IsSameAs("Glare_cls"))
		((Glare_cls*)pCFilter)->Begin();
	    if (strAttributeName.IsSameAs("RemoveAlpha_cls"))
		((RemoveAlpha_cls*)pCFilter)->Begin();
	    if (strAttributeName.IsSameAs("Overlay_cls"))
		((Overlay_cls*)pCFilter)->Begin(VPasspartoutInfo);
	    if (strAttributeName.IsSameAs("Legend_cls")){
		((Legend_cls*)pCFilter)->MakeLegendLayer(width, height, 
							 VPasspartoutInfo);
		((Legend_cls*)pCFilter)->Begin(VPasspartoutInfo);
	    }
	    if (strAttributeName.IsSameAs("FocalBlur_cls")){
		((FocalBlur_cls*)pCFilter)->Begin(pCFBOIn, CCameraPos,
						  pCCamera, NT);
		depthTexture = GL_TEXTURE1;
	    }
	    if (strAttributeName.IsSameAs("SSAO_cls")){
		((SSAO_cls*)pCFilter)->Begin(pCFBOIn);
		depthTexture = GL_TEXTURE1;
	    }
	    if (strAttributeName.IsSameAs("FilterDepth_cls")){
		((FilterDepth_cls*)pCFilter)->Begin(pCFBOIn);
		depthTexture = GL_TEXTURE0;
	    }
	    
	    
	    pCFBOOut->Begin(TRUE);
	    GetPCGLExtentions()->ClampColor(GL_FALSE);

	    if (pCDepthTexture != NULL && depthTexture == GL_TEXTURE1){
		glActiveTexture(depthTexture);
		pCDepthTexture->BindTexture();
		glActiveTexture(GL_TEXTURE0);
	    }

	    if (pCDepthTexture != NULL && depthTexture == GL_TEXTURE0)
		pCDepthTexture->DrawOnScreen2DFit(width, height, _TEX_FIT_HORIZONTAL, FALSE, 0);
	    else
		pCFBOIn->DrawOnScreen2DFit(width, height, _TEX_FIT_HORIZONTAL, FALSE, 0);
	    
	    if (strAttributeName.IsSameAs("Exposure_cls"))
		((Exposure_cls*)pCFilter)->End();
	    if (strAttributeName.IsSameAs("Blur_cls"))
		((Blur_cls*)pCFilter)->End();
	    if (strAttributeName.IsSameAs("Glare_cls"))
		((Glare_cls*)pCFilter)->End();
	    if (strAttributeName.IsSameAs("RemoveAlpha_cls"))
		((RemoveAlpha_cls*)pCFilter)->End();
	    if (strAttributeName.IsSameAs("Overlay_cls"))
		((Overlay_cls*)pCFilter)->End();
	    if (strAttributeName.IsSameAs("Legend_cls"))
		((Legend_cls*)pCFilter)->End();
	    if (strAttributeName.IsSameAs("FocalBlur_cls"))
		((FocalBlur_cls*)pCFilter)->End();
	    if (strAttributeName.IsSameAs("SSAO_cls"))
		((SSAO_cls*)pCFilter)->End();
	    if (strAttributeName.IsSameAs("FilterDepth_cls"))
		((FilterDepth_cls*)pCFilter)->End();
	    
	    pCFBOOut->End();
	    if (pCFBOIn != pCFBOInSrc) delete pCFBOIn;
	    pCFBOIn = pCFBOOut;

	}
    }

    if (pCDepthTexture != NULL) delete pCDepthTexture;

    if (pCFBOOut == NULL){
	pCFBOOut = new FBO_cls(width, height,     
			       pCFBOInSrc->mpCGLCanvasParent);
	pCFBOOut->MakeFBO(TRUE);
	pCFBOOut->Begin(TRUE);
	GetPCShaderCtrl()->BeginShader("ClampAlpha", NULL);
	glEnable(GL_BLEND);
	pCFBOIn->DrawOnScreen2DFit(width, height, _TEX_FIT_HORIZONTAL, FALSE, 0);
	GetPCShaderCtrl()->EndShader();
	glDisable(GL_BLEND);
	pCFBOOut->End();
    }

    return pCFBOOut;
}

void CLS::
Select(wxString strName, bool bAdditional, bool bUpdateGUI){
    
    if (strName.IsEmpty() == TRUE){
	//if (mpCActiveObject != NULL) mpCActiveObject->ReferRelease();
	mpCActiveFilter = NULL;
    }
    else{
	Attribute_cls *pCPrev = mpCActiveFilter;
	
	//if (mpCActiveObject != NULL) mpCActiveObject->ReferRelease();
	mpCActiveFilter = GetPCFromName(strName);
	//if (mpCActiveObject != NULL) mpCActiveObject->Refer();
	
	if (mpCActiveFilter == pCPrev) bAdditional = TRUE;
    }

    if (bAdditional == FALSE) ClearSelection();

    if (mpCActiveFilter != NULL && IsSelected(mpCActiveFilter) == FALSE){
	mVpCSelectedFilter.push_back(mpCActiveFilter);
	//mpCActiveObject->Refer();
	GetPCMain()->ProcAfterMath("SelectFilter", strName);
    }

    if (bUpdateGUI == TRUE)
	GetPCMain()->UpdatePanels("VIEW3D");
}

wxString CLS::
GetStrNewCandidate(wxString strCandidate){
    int pos = strCandidate.Find(wxUniChar('.'), TRUE);
    if (pos == wxNOT_FOUND)
	return strCandidate+wxT(".001");
    wxString strEnum = strCandidate.Mid(pos+1);
    if (strEnum.Matches(wxT("???"))){
	long num; strEnum.ToLong(&num);
	int numInt = num;
	wxString strDigit = wxString::Format(wxT("%03d"), numInt);
	if (strEnum.IsSameAs(strDigit) == TRUE){
	    return strCandidate.Mid(0, pos+1) + wxString::Format(wxT("%03d"), numInt+1);
	}
    }
    return strCandidate+wxT(".001");
}

wxString CLS::
GetStrUniqueName(wxString strName){
    wxString strCandidate = strName;
    while(IsFilterExist(strCandidate))
	strCandidate = GetStrNewCandidate(strCandidate);
    return strCandidate;
}

void CLS::
Detatch(Attribute_cls *pCAttrib){
    // KvǂŊmF
}

void CLS::
DeleteFilter(Attribute_cls *pCAttrib){
    wxString strName = pCAttrib->GetStrName();
    
    int indexTgt = GetIndexFromPC(pCAttrib);
    if (indexTgt == -1) return;
    mVpCFilter.erase(GetIterFromIndex(indexTgt));

    Detatch(pCAttrib);
    //delete pCAttrib;
    Attribute_cls::Delete(pCAttrib);

    GetPCMain()->ProcAfterMath("DeleteFilter", strName);
}

void CLS::
DeselectKeyFrame(wxString strException){
    uint num = mVpCFilter.size();
    for (uint i = 0; i < num; i++)
	GetPCFromIndex(i)->DeselectKeyFrame(strException);
}

bool CLS::
ChangeKeyTypeByModKey(wxString strTag, int keyIndex, bool bCtrl, bool bShift, bool bAlt){
    if (bCtrl == FALSE && bShift == FALSE && bAlt == FALSE) return FALSE;

    Parameter_cls *pCParam = GetPCParameter(strTag);
    if (pCParam != NULL){
	strTag = GetStrSelf() << strTag;
	return pCParam->ChangeKeyTypeByModKey(strTag, keyIndex, bCtrl, bShift, bAlt);
    }
    return FALSE;
}

bool CLS::
SelectKeyFrame(wxString strTag, int keyIndex, bool bSelect, bool bAdditional,
	       double resolution){

    Parameter_cls *pCParam = GetPCParameter(strTag);
    if (pCParam != NULL)
	return pCParam->SelectKeyByIndex(keyIndex, bSelect, bAdditional,
					 resolution);
    return FALSE;
}

void CLS::
SetSelectedTLKeyTypeCommand(int keyIndex, double ease){
    GetPCUndo()->OpenBlock("SetSelectedTLKeyType_Filter");
    uint num = mVpCFilter.size();
    for (uint i = 0; i < num; i++)
	GetPCFromIndex(i)->SetSelectedTLKeyTypeCommand("FilterCtrl", keyIndex, ease);
    GetPCUndo()->CloseBlock("SetSelectedTLKeyType_Filter");
}

void CLS::
MoveSelectedTLKeyCommand(double dNX, double dY){
    GetPCUndo()->OpenBlock("MoveSelectedTLKey_Filter");
    uint num = mVpCFilter.size();
    for (uint i = 0; i < num; i++)
	GetPCFromIndex(i)->MoveSelectedTLKeyCommand("FilterCtrl", dNX, dY);
    GetPCUndo()->CloseBlock("MoveSelectedTLKey_Filter");
}

void CLS::
AddFilterCommand(wxString strClassName, wxString strNameCandid){
    wxString strTag = GetStrSelf();

    wxString strUndo(strTag), strRedo(strTag);
    wxString strName = GetStrUniqueName(strNameCandid);

    strUndo << "Delete::" << strName;    
    strRedo << "New::" << strClassName << "::" << strName;

    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
DeleteFilterCommand(wxString strName){
    ASSERT(IsFilterExist(strName) == TRUE);

    wxString strTag = GetStrSelf();

    wxString strUndo2(strTag), strRedo2(strTag);
    wxString strUndo3(strTag), strRedo3(strTag);
    Attribute_cls *pCFilter = GetPCFromName(strName);

    wxString strClassName = pCFilter->GetStrAttributeName();
    int index = GetIndexFromPC(pCFilter);
    ASSERT(index != -1);

    GetPCUndo()->OpenBlock("FilterDelete");
    if (index != mVpCFilter.size()-1){
	
	wxString strUndo1(strTag), strRedo1(strTag);
	strUndo1 << "Move::" << strName << "::To::" << index;
	strRedo1 << "Skip";    

	GetPCUndo()->PushCommandPair(strUndo1, strRedo1);
	GetPCUndo()->Redo();
    }

    strUndo2; pCFilter->Serialize(strUndo2); strUndo2.Trim(TRUE);
    strRedo2 << "Skip";
    GetPCUndo()->PushCommandPair(strUndo2, strRedo2);
    GetPCUndo()->Redo();

    strUndo3 << "New::" << strClassName << "::" << strName;
    strRedo3 << "Delete::" << strName;    

    GetPCUndo()->PushCommandPair(strUndo3, strRedo3);
    GetPCUndo()->Redo();
    
    GetPCUndo()->CloseBlock("FilterDelete");
}

void CLS::
MoveFilterNative(Attribute_cls *pCFrom, int indexTo){
    
    std::list < Attribute_cls* >::iterator itrTo = GetIterFromIndex(indexTo);
    std::list < Attribute_cls* >::iterator itrFrom = GetIterFromPC(pCFrom);
    
    if (itrTo == itrFrom) return;

    mVpCFilter.erase(itrFrom);
    mVpCFilter.insert(itrTo, pCFrom);
}

void CLS::
MoveFilterCommand(wxString strName, int indexTo){
    ASSERT(IsFilterExist(strName) == TRUE);

    wxString strTag = GetStrSelf();

    Attribute_cls *pCFilter = GetPCFromName(strName);
    int indexFrom = GetIndexFromPC(pCFilter);
    
    // ӖȂ̃`FbN
    if (indexTo == indexFrom+1 || indexTo == indexFrom) return; 
    if (indexTo < indexFrom) indexFrom++;

    wxString strUndo(strTag), strRedo(strTag);
    strUndo << wxT("Move::") << strName << "::To::" << indexFrom;
    strRedo << wxT("Move::") << strName << "::To::" << indexTo;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetFilterValidityCommand(wxString strName, bool bValidity){
    ASSERT(IsFilterExist(strName) == TRUE);

    FilterBase_cls *pCFilter = GetPCFromName(strName);

    if (pCFilter->mCbValidity.GetBodyBOOL() == bValidity) return;
    
    wxString strFrom, strTo;
    pCFilter->mCbValidity.SerializeSingleItem(strFrom);
    pCFilter->mCbValidity.SetBodyFromData((long long)bValidity);
    pCFilter->mCbValidity.SerializeSingleItem(strTo);

    GetPCUndo()->PushCommandPair(strFrom, strTo);
    GetPCUndo()->Redo(); 
}

bool CLS::
ProcNew(std::wistream &istream, wxString strFirstItem, wxString strLine){
    wxString strClassName = strFirstItem;
    wxString strNewName = PopStrItemFromStrLine(strLine, wxT("::"));

    Attribute_cls *pCTgt = NULL;
    if (strClassName.IsSameAs("Exposure_cls"))
	pCTgt = new Exposure_cls(strNewName);
    if (strClassName.IsSameAs("Blur_cls"))
	pCTgt = new Blur_cls(strNewName);
    if (strClassName.IsSameAs("FocalBlur_cls"))
	pCTgt = new FocalBlur_cls(strNewName);
    if (strClassName.IsSameAs("Glare_cls"))
	pCTgt = new Glare_cls(strNewName);
    if (strClassName.IsSameAs("RemoveAlpha_cls"))
	pCTgt = new RemoveAlpha_cls(strNewName);
    if (strClassName.IsSameAs("Overlay_cls"))
	pCTgt = new Overlay_cls(strNewName);
    if (strClassName.IsSameAs("Legend_cls"))
	pCTgt = new Legend_cls(strNewName);
    if (strClassName.IsSameAs("SSAO_cls"))
	pCTgt = new SSAO_cls(strNewName);
    if (strClassName.IsSameAs("FilterDepth_cls"))
	pCTgt = new FilterDepth_cls(strNewName);
    
    if (pCTgt == NULL) return FALSE;

    mVpCFilter.push_back(pCTgt);
    Select(pCTgt->GetStrName(), FALSE, FALSE);
    GetPCMain()->ProcAfterMath("FilterChange", "New");
    return TRUE;
}

bool CLS::
ProcMove(std::wistream &istream, wxString strFirstItem, wxString strLine){
    Attribute_cls *pCFrom = GetPCFromName(strFirstItem);
    if (pCFrom == NULL) return FALSE;

    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    long indexTo;
    strFirstItem.ToLong(&indexTo);
    MoveFilterNative(pCFrom, indexTo);
    return TRUE;
}

bool CLS::
ProcDelete(std::wistream &istream, wxString strFirstItem, wxString strLine){
    
    Attribute_cls *pCFilter = GetPCFromName(strFirstItem);
    if (pCFilter == NULL) return FALSE;

    DeleteFilter(pCFilter);
    return TRUE;
}

bool CLS::
ProcReadAttribute(std::wistream &istream, wxString strFirstItem, wxString strLine){
    Attribute_cls *pCFilter = GetPCFromName(strFirstItem);
    if (pCFilter == NULL) return FALSE;

    std::wstring wstrLine;
    while( std::getline(istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	
	strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));

	pCFilter->ReadStrLine(istream, strFirstItem, strLine);
	if (strFirstItem.IsSameAs("ReadAttribute")){
	    ASSERT(PopStrItemFromStrLine(strLine, wxT("::")).IsSameAs(wxT("End")));
	    break;
	}
    }
    
    return TRUE;
}

bool CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    
    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE){
	Select(mCstrActiveItemName.GetBodySTR(), FALSE);
	return TRUE;
    }

    uint numFilter = mVpCFilter.size();
    for (uint i = 0; i < numFilter; i++){
	if (GetPCFromIndex(i)->ReadStrLine(istream, strFirstItem, strLine) == TRUE){
	    GetPCMain()->UpdatePanels("VIEW3D::TIMELINE::PROPERTY_SCENE");
	    return TRUE;
	}
    }
   
    if (strFirstItem.IsSameAs("FilterCtrl") == FALSE) return FALSE;

    bool bRet = FALSE;

    // 1sR}h
    ParamGroup_cls* pCParamGroup = GetPCParamGroup(strLine);
    if (pCParamGroup != NULL){
	PopStrItemFromStrLine(strLine, "::");
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	if (strFirstItem.IsSameAs(pCParamGroup->mCstrName.GetBodySTR()))
	    strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= pCParamGroup->ReadStrOneLiner(strFirstItem, strLine);
    }

    strFirstItem = PopStrItemFromStrLine(strLine, "::");
    
    if (strFirstItem.IsSameAs("New")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcNew(istream, strFirstItem, strLine);
    }
    if (strFirstItem.IsSameAs("Delete")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcDelete(istream, strFirstItem, strLine);
    }
    if (strFirstItem.IsSameAs("Move")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcMove(istream, strFirstItem, strLine);
    }

    if (strFirstItem.IsSameAs("ReadAttribute")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet = ProcReadAttribute(istream, strFirstItem, strLine);
    }

    if (bRet == TRUE) GetPCMain()->UpdatePanels("VIEW3D::TIMELINE::PROPERTY_SCENE");
    
    return bRet;
}

void CLS::
SerializePropertiesOnly(wxString& strOut){
    
    //mCstrActiveItemName.SetBodyFromData(GetActiveFilterName());
}

void CLS::
Serialize(wxString& strOut){
    strOut << "FilterCtrl::Begin" << ENDL;

    //SerializePropertiesOnly(strOut);

    ConfListContainer_cls::Serialize(strOut);

    uint numFilter = mVpCFilter.size();
    for (uint i = 0; i < numFilter; i++){
	Attribute_cls *pCAttrib = GetPCFromIndex(i);
	strOut << "FilterCtrl::New::" << pCAttrib->GetStrAttributeName() << "::" << pCAttrib->GetStrName() << ENDL;
	pCAttrib->Serialize(strOut);
    }

    strOut << "FilterCtrl::End" << ENDL;
}
    

FilterCtrl_cls *GpCFilterCtrl = NULL;
