// (setq-c-offset 4)
// "clsFilterLegend.h"

#include "clsFilterLegend.h"
#include "clsFilterCtrl.h"

#include "../GRAPH/clsTics.h"
#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsFBO.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/DATAVIEWER/clsParticle.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../SIMDATA/clsParticleSequence.h"
#include "../SIMDATA/clsParticleSnap.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../../wxWidgetEnum.h"

#ifdef WIN32
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
#endif

#define CLS Legend_cls

CLS::
Legend_cls(wxString strName) : FilterBase_cls("Legend_cls", strName){

    wxString strConfListLabel = "FilterCtrl::"; strConfListLabel << strName;
    mCConfList.SetTag(strConfListLabel);

    wxString strUpdate = "VIEW3D::PROPERTY_SCENE";
    mCConfList.Add(&mCstrTarget, "Target=", "None", strUpdate);
#ifdef WIN32
    mCConfList.Add(&mCstrFontDesc, "TextFontDesc=", wxString::Format("%s 24 windows-932", _("Meiryo")), strUpdate);
#else
    mCConfList.Add(&mCstrFontDesc, "TextFontDesc=", "Arial 24", strUpdate);
#endif
    mCConfList.Add(&mCbTime, "UseTime=", "1", strUpdate);
    mCConfList.Add(&mCstrFormatTime, "FormatTime=", "Time=%f", strUpdate);
    mCConfList.Add(&mCTimeOffsetX, "TimeOffsetX=", "0.03", strUpdate);
    mCConfList.Add(&mCTimeOffsetY, "TimeOffsetY=", "0.03", strUpdate);
    mCConfList.Add(&mCTimeScale, "TimeScale=", "1", strUpdate);
    mCConfList.Add(&mCbClrCurve, "UseClrCurve=", "1", strUpdate);
    mCConfList.Add(&mCClrCurveOffsetX, "ClrCurveOffsetX=", "0.6", strUpdate);
    mCConfList.Add(&mCClrCurveOffsetY, "ClrCurveOffsetY=", "0.03", strUpdate);
    mCConfList.Add(&mCClrCurveScaleX, "ClrCurveScaleX=", "1", strUpdate);
    mCConfList.Add(&mCClrCurveScaleY, "ClrCurveScaleY=", "1", strUpdate);
    mCConfList.Add(&mCClrCurveTicFontSize, "ClrCurveTicFontSize=", "2", strUpdate);
    mCConfList.Add(&mCClrCurveTicScale, "ClrCurveTicScale=", "1", strUpdate);

    mpCTics = new Tics_cls();
}

CLS::
~Legend_cls(){
    delete mpCTics;
}

void CLS::
SyncFromBuffer(LegendBuffer_cls *pCBuffer){
    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
}

void CLS::
MakeTextTexture(){

    wxString strFont = mCstrFontDesc.GetBodySTR();
    wxString strFormat = mCstrFormatTime.GetBodySTR();

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    float time = (float)NT;
    SceneObj_cls *pCTgt = GetPCSceneObjCtrl()->GetPCFromName(mCstrTarget.GetBodySTR());
    if (pCTgt != NULL) {
	if (pCTgt->IsSameAs("Particle_cls")){
	    Particle_cls *pCParticle = (Particle_cls*)pCTgt;
	    int frame0, frame1, frame2, frame3, subframe;
	
	    double interpFactor = 
		pCParticle->GetFrameInfo(frame0, frame1, frame2, frame3, subframe, NT);
	    bool bTimeDepend = pCParticle->mpCParticleSequence->mbTimeDepend;
	    ParticleSnap_cls *pCSnap1 = pCParticle->mpCParticleSequence->GetPCSnap(frame1);
	    if (pCSnap1 == NULL) return;

	    ParticleSnap_cls *pCSnap2 = NULL;
	    if (bTimeDepend == TRUE && interpFactor != 0)
		pCSnap2 = pCParticle->mpCParticleSequence->GetPCSnap(frame2);
	    
	    if (pCSnap2 != NULL) 
		time = (1.0-interpFactor) * pCSnap1->mSnapTime + interpFactor * pCSnap2->mSnapTime;
	    else
		time = pCSnap1->mSnapTime;
	}
    }

    wxString strText = wxString::Format(mCstrFormatTime.GetBodySTR(), time);
    
    if (mstrFontUsed.IsSameAs(strFont) &&
	mstrUsed.IsSameAs(strText)) return;

    //long style = wxFONTWEIGHT_NORMAL;
    //loang weight = wxFONTWEIGHT_NORMAL;
    //int fontSize = 36;
    wxFont CFont;
    CFont.SetNativeFontInfoUserDesc(strFont);

    mCGLTexture.SetEmptyImage();
    mCGLTexture.ClearImage();

    bool bSucceed = TRUE; // ȂsƂ̂sc
    wxScreenDC DC;
    DC.SetFont(CFont);

    wxCoord singleW, singleH, singleD;
    wxCoord w,h;
    
    DC.GetTextExtent(strText, &singleW, &singleH, &singleD);
    DC.GetMultiLineTextExtent(strText, &w, &h, &singleH);
    
    wxRect CRectCanvas(0, 0, w, h);
    wxRect CRectBounding;
    wxMemoryDC *pMemDC = mCGLTexture.BeginDC(w, h);
    
    pMemDC->SetBackground(wxColour(0, 0, 0));
    pMemDC->Clear();
    pMemDC->SetBackground(wxColour(255, 0, 0));
    pMemDC->SetTextForeground(*wxWHITE);
    pMemDC->SetFont(CFont);
    pMemDC->DrawLabel(strText, wxNullBitmap, CRectCanvas, wxALIGN_LEFT | wxALIGN_TOP, -1, &CRectBounding);
    
    mCGLTexture.EndDC(TRUE, TRUE);
    
    mWidth = w;
    mHeight = h;
    mDescend = (float)singleD/singleH;

    mstrUsed = strText;
    mstrFontUsed = mCstrFontDesc.GetBodySTR();
}

void CLS::
DrawColorCurve(ColorCurve_cls *pCClr,
	       int offsetX, int offsetY, int width, int height){
    pCClr->PrepareTexture(FALSE, 1, FALSE);

    glDisable(GL_TEXTURE_2D);
    glEnable(GL_TEXTURE_1D);
    glBindTexture(GL_TEXTURE_1D, pCClr->mTexture1D);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    glBegin(GL_QUADS);
    glTexCoord1f(0); glVertex2d( offsetX,         offsetY);
    glTexCoord1f(1); glVertex2d( offsetX + width, offsetY);
    glTexCoord1f(1); glVertex2d( offsetX + width, offsetY + height);
    glTexCoord1f(0); glVertex2d( offsetX,         offsetY + height);
    glEnd();
}

void CLS::
DrawTics(ColorCurve_cls *pCClr,
	 int offsetX, int offsetY, int width, int height, 
	 int ticHeight, float aspect, int ticFontSize, int ticScale){
    
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < double > VRange = 
	pCClr->GetPCParamGroup("Range")->GetVInterpolatedValue(NT);
    
    mpCTics->SetParameter(wxSize(width*1.2/ticScale, 
				 ticHeight * aspect/ticScale),
			  VRange[0], VRange[1]);
    mpCTics->SetBorderX(width * 0.1, width * 0.1);

    wxMemoryDC *pMemDC = mCGLTic.BeginDC(width*1.2/ticScale,
					 ticHeight * aspect/ticScale);

    pMemDC->SetBackground(wxColour(0, 0, 0));
    pMemDC->Clear();

    mpCTics->DrawTics(*pMemDC, FALSE, TRUE, ticFontSize);

    mCGLTic.EndDC(TRUE, TRUE);
    
    int ticOffsetY = height;
    glDisable(GL_TEXTURE_1D);
    glEnable(GL_TEXTURE_2D);
    mCGLTic.BindTexture();
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0); glVertex2d( offsetX - width*0.1, offsetY + ticOffsetY);
    glTexCoord2f(1, 0); glVertex2d( offsetX + width*1.1, offsetY + ticOffsetY);
    glTexCoord2f(1, 1); glVertex2d( offsetX + width*1.1, offsetY + ticHeight + ticOffsetY);
    glTexCoord2f(0, 1); glVertex2d( offsetX - width*0.1, offsetY + ticHeight + ticOffsetY);
    glEnd();
}

void CLS::
MakeLegendLayer(int width, int height,
		std::vector < float > VPasspartoutInfo){

    float aspect = 1;
    if (VPasspartoutInfo[0] != -1){
	float xscale = VPasspartoutInfo[0]/VPasspartoutInfo[2];
	float yscale = VPasspartoutInfo[1]/VPasspartoutInfo[2];
	aspect = yscale/xscale;
    }
    
    MakeTextTexture();

    mpCFBOLegendLayer = new FBO_cls(width, height, NULL);
    mpCFBOLegendLayer->MakeFBO(TRUE);

    mpCFBOLegendLayer->Begin(TRUE);

    if (mCbTime.GetBodyBOOL() == TRUE){
	float timeScale = mCTimeScale.GetBodyDOUBLE();
	float timeOffsetX = mCTimeOffsetX.GetBodyDOUBLE();
	float timeOffsetY = mCTimeOffsetY.GetBodyDOUBLE();
	mCGLTexture.DrawOnScreen2D(width, height,
				   mWidth*timeScale, mHeight*timeScale/aspect, 
				   width*timeOffsetX, height*timeOffsetY,
				   FALSE, GL_MODULATE);
    }

    SceneObj_cls *pCTgt = GetPCSceneObjCtrl()->GetPCFromName(mCstrTarget.GetBodySTR());
    if (mCbClrCurve.GetBodyBOOL() == TRUE && pCTgt != NULL) {
	
	float clrCurveScaleX = mCClrCurveScaleX.GetBodyDOUBLE();
	float clrCurveScaleY = mCClrCurveScaleY.GetBodyDOUBLE();
	float clrCurveOffsetX = width  * mCClrCurveOffsetX.GetBodyDOUBLE();
	float clrCurveOffsetY = height * mCClrCurveOffsetY.GetBodyDOUBLE();
	int clrWidth = width * 0.3 * clrCurveScaleX;
	int clrHeight = height * 0.05 * clrCurveScaleY;
	int ticHeight = height * 0.05;
	int ticFontSize = mCClrCurveTicFontSize.GetBodyINT();
	int ticScale = mCClrCurveTicScale.GetBodyINT();

	Push2DMatrix(width, height);
	if (pCTgt->IsSameAs("Particle_cls")){
	    DataViewerBase_cls *pCDataViewer = (DataViewerBase_cls*)pCTgt;
	    
	    //Particle_cls *pCParticle = (Particle_cls*)pCTgt;

	    int clrCurveNum = pCDataViewer->GetNumAttribute("ColorCurve_cls");
	    for (int i = 0; i < clrCurveNum; i++){
		ColorCurve_cls *pCClrCureve = pCDataViewer->GetPCColorCurve(i);

		DrawColorCurve(pCClrCureve,
			       clrCurveOffsetX, clrCurveOffsetY,
			       clrWidth, clrHeight);
		DrawTics(pCClrCureve,
			 clrCurveOffsetX, clrCurveOffsetY,
			 clrWidth, clrHeight,
			 ticHeight, aspect, ticFontSize, ticScale);

		clrCurveOffsetY += clrHeight + ticHeight;
	    }
	}

	if (pCTgt->IsSameAs("VolumeBox_cls") ||
	    pCTgt->IsSameAs("HeightField_cls") || 
	    pCTgt->IsSameAs("Streamline_cls")){
	    DataViewerBase_cls *pCDataViewer = (DataViewerBase_cls*)pCTgt;
	    ColorCurve_cls *pCClrCureve = /*pCVolume->*/pCDataViewer->GetPCColorCurve(0);
	    DrawColorCurve(pCClrCureve,
			   clrCurveOffsetX, clrCurveOffsetY, 
			   clrWidth, clrHeight);
	    DrawTics(pCClrCureve,
		     clrCurveOffsetX, clrCurveOffsetY, clrWidth, clrHeight,
		     ticHeight, aspect, ticFontSize, ticScale);
	}

	Pop2DMatrix();
    }
    

    mpCFBOLegendLayer->End();
}

void CLS::
Begin(std::vector < float > VPasspartoutInfo){

    glEnable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE1);

    //mCGLTexture.BindTexture();
    mpCFBOLegendLayer->BindTexture();

    OverlayParam_cls CParam;
    CParam.mAlpha = 1;
    
    if (VPasspartoutInfo[0] != -1){
	CParam.mXScale = VPasspartoutInfo[0]/VPasspartoutInfo[2];
	CParam.mYScale = VPasspartoutInfo[1]/VPasspartoutInfo[2];
    }
    else{
	CParam.mXScale = 1;
	CParam.mYScale = 1;
    }

    GetPCShaderCtrl()->BeginShader("Overlay", &CParam);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP_TO_EDGE);
    glActiveTexture(GL_TEXTURE0);
}

void CLS::
End(){
    GetPCShaderCtrl()->EndShader();
    delete mpCFBOLegendLayer;
}

void CLS::
SelectObject(wxString strObjName){
    mCstrTarget.PushCommandPair(strObjName);
}

#define CLS LegendBuffer_cls

CLS::
LegendBuffer_cls(Legend_cls *pCSrc) : FilterBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);

    mCConfList.Add(&mCstrTarget, "Target=", "None");
#ifdef WIN32
    mCConfList.Add(&mCstrFontDesc, "TextFontDesc=", wxString::Format("%s 24 windows-932", _("Meiryo")));
#else
    mCConfList.Add(&mCstrFontDesc, "TextFontDesc=", "Arial 24");
#endif
    mCConfList.Add(&mCbTime, "UseTime=", "1");
    mCConfList.Add(&mCstrFormatTime, "FormatTime=", "Time=%f");
    mCConfList.Add(&mCTimeOffsetX, "TimeOffsetX=", "0.03");
    mCConfList.Add(&mCTimeOffsetY, "TimeOffsetY=", "0.03");
    mCConfList.Add(&mCTimeScale, "TimeScale=", "1");
    mCConfList.Add(&mCbClrCurve, "UseClrCurve=", "1");
    mCConfList.Add(&mCClrCurveOffsetX, "ClrCurveOffsetX=", "0.6");
    mCConfList.Add(&mCClrCurveOffsetY, "ClrCurveOffsetY=", "0.03");
    mCConfList.Add(&mCClrCurveScaleX, "ClrCurveScaleX=", "1");
    mCConfList.Add(&mCClrCurveScaleY, "ClrCurveScaleY=", "1");
    mCConfList.Add(&mCClrCurveTicFontSize, "ClrCurveTicFontSize=", "2");
    mCConfList.Add(&mCClrCurveTicScale, "ClrCurveTicScale=", "1");
    
    SyncFromSrc();
}

/*
void CLS::
HideAndShow(){
}
*/

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineSizer(3, _("Target"));
    pC->CombineBitmapPanel(btnDropdown, wxES, 14);
    pC->mVCItem.back().mBitmapTemplate = _BITMAP_BTN_DROPDOWN;
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = mCstrName.GetBodySTR();
    pC->CombineConfItem(mCstrTarget);
    int indexTargetCancel = pC->CombineBitmapPanel(btnLegendTargetClear, "Clear", 24);
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = mCstrName.GetBodySTR();
        
    pC->CombineSizer(3, _("Time"));
    pC->CombineConfItem(mCbTime, _("Display"));
    pC->CombineConfItem(mCstrFormatTime);
    
    int indexFont = pC->CombineBitmapPanel(btnLegendFont, "Font", 24);
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = mCstrName.GetBodySTR();
    
    pC->CombineSizer(2, _("Offset"));
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCTimeOffsetX);
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCTimeOffsetY);

    pC->SetLimit(0.2, 4);
    pC->mDragScale = 0.2;
    pC->CombineConfItem(mCTimeScale, _("Scale"));

    pC->CombineSizer(1, _("Color Curve"));
    pC->CombineConfItem(mCbClrCurve, _("Display"));

    pC->CombineSizer(2, _("Offset"));
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCClrCurveOffsetX);
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCClrCurveOffsetY);
    pC->CombineSizer(2, _("Scale"));
    pC->SetLimit(0.2, 4);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCClrCurveScaleX);
    pC->SetLimit(0.2, 4);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCClrCurveScaleY);

    pC->CombineSizer(2, _("Tics Font Size"));
    pC->SetLimit(0, 6);
    pC->mDragScale = 3;
    pC->CombineConfItem(mCClrCurveTicFontSize);
    pC->SetLimit(1, 4);
    pC->mDragScale = 3;
    pC->CombineConfItem(mCClrCurveTicScale);
    
    pC->FinishSetting();

    BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[indexFont].mpCWindowLink;
    pCBitmap->mbButton = TRUE;
    pCBitmap->AddBitmapStr("wxART_FONT", "wxART_FONT_P");
    pCBitmap->mBorderX = 4;
    pCBitmap->mBorderY = 4;

    BitmapPanel_cls *pCBitmap2 = (BitmapPanel_cls*)pC->mVCItem[indexTargetCancel].mpCWindowLink;
    pCBitmap2->mbButton = TRUE;
    pCBitmap2->AddBitmapStr("wxART_CLOSE_PANEL", "wxART_CLOSE_PANEL_P");
    pCBitmap2->SetBitmapStrDisable("wxART_CLOSE_PANEL_D");
    pCBitmap2->mBorderX = 4;
    pCBitmap2->mBorderY = 4;

    return pC;
}


void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
}

void CLS::
ForceSrcToSyncBuffer(){ mpCSrc->SyncFromBuffer(this); }

