// (setq-c-offset 4)
// "clsFilterOverlay.h"

#include "clsFilterOverlay.h"
#include "clsFilterCtrl.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MISC/clsMisc.h"
#include "../MYGL/clsTextureProvider.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../SIMDATA/clsImageSequence.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/EX/wxButtonEX.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitionsGL.h"

#include "../../wxWidgetEnum.h"

#ifdef WIN32
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
#endif

#define CLS Overlay_cls

CLS::
Overlay_cls(wxString strName) : FilterBase_cls("Overlay_cls", strName){

    wxString strConfListLabel = "FilterCtrl::"; strConfListLabel << strName;
    mCConfList.SetTag(strConfListLabel);

    wxString strUpdate = "VIEW3D::PROPERTY_SCENE";
    mCConfList.Add(&mCstrPath, "strPath=", "NONE", strUpdate);
    mCConfList.Add(&mCSequenceOffset, "sequenceOffset=", "0", strUpdate);
    mCConfList.Add(&mCbSequence, "Sequence=", "0", strUpdate);
    mCConfList.Add(&mCbFit, "Fit=", "0", strUpdate);

    AddOneDimParam("Alpha", "A", 0.5);

    mpCImageSequence = new ImageSequence_cls;
}

CLS::
~Overlay_cls(){
    delete mpCImageSequence;
}

void CLS::
SyncFromBuffer(OverlayBuffer_cls *pCBuffer){
    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VAlpha;
    VAlpha.push_back(pCBuffer->mCAlpha.GetBodyDOUBLE());
    SyncronizeParamForFilter("Alpha", VAlpha);
}

void CLS::
CheckFileSequence(){
    int offset = mCSequenceOffset.GetBodyINT();
    mpCImageSequence->Set(mCstrPath.GetBodySTR(), offset);

    mpCImageSequence->CalcMaxOffset(mCbSequence.GetBodyBOOL());
    if (mpCImageSequence->mSeqLen != 1)
	mpCImageSequence->mbTimeDepend = TRUE;

    mstrPathChecked = mCstrPath.GetBodySTR();
    GetPCMain()->mCLastOpenedFileIMG.SetBodyFromData(mstrPathChecked);
}

void CLS::
ModFrame(int &frame, int seqLen, bool bRepeat){
    if (frame >= seqLen){
	if (bRepeat) frame = Misc_cls::Surplus(frame, seqLen);
	//else         frame = seqLen - 1;
    }
}

void CLS::
Begin(std::vector < float > VPasspartoutInfo){
    if (mstrPathChecked != mCstrPath.GetBodySTR())
	CheckFileSequence();

    int seqLen = mpCImageSequence->mSeqLen;

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NT, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);

    bool bRepeat = FALSE;

    if (interpFactor < 1e-8) interpFactor = 0;
    
    ModFrame(frame1, seqLen, bRepeat);

    ImageSnap_cls *pCSnap = mpCImageSequence->GenerateIfNot(frame1);
    
    wxString strImg;
    if (mpCImageSequence->mbTimeDepend == TRUE){
	strImg = mpCImageSequence->GetModifiedStr(frame1);
	if (frame1 < 0 || frame1 >= seqLen)
	    strImg = "None";
    }
    else
	strImg = mstrPathChecked;

    glEnable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE1);

    float alpha = GetPCParamGroup("Alpha")->GetVInterpolatedValue((float)NT)[0];

    int indexImage = -1;
    if (strImg.IsSameAs("None")){
	GetPCTextureProvider()->BindBlankTexture(FALSE);
	alpha = 0;
    }
    else
	indexImage = GetPCTextureProvider()->BindTexture(strImg);

    OverlayParam_cls CParam;
    CParam.mAlpha = alpha;
    if (VPasspartoutInfo[0] != -1){
	if (mCbFit.GetBodyBOOL() == TRUE){
	    CParam.mXScale = VPasspartoutInfo[0]/VPasspartoutInfo[2];
	    CParam.mYScale = VPasspartoutInfo[1]/VPasspartoutInfo[2];
	}
	else{
	    wxSize CSize = GetPCTextureProvider()->GetTextureSize(indexImage);
	    float aspect = (float)CSize.GetWidth() / CSize.GetHeight();
	    CParam.mXScale = VPasspartoutInfo[0]/VPasspartoutInfo[2];
	    CParam.mYScale = CParam.mXScale / aspect;
	}
    }
    else{
	if (mCbFit.GetBodyBOOL() == TRUE){
	    wxSize CSize = GetPCTextureProvider()->GetTextureSize(indexImage);
	    float aspect = (float)CSize.GetWidth() / CSize.GetHeight();
	    CParam.mXScale = 1;
	    CParam.mYScale = 1.0f/aspect;
	}
	else{
	    CParam.mXScale = 1;
	    CParam.mYScale = 1;
	}
    }
    
    GetPCShaderCtrl()->BeginShader("Overlay", &CParam);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP_TO_EDGE);
    glActiveTexture(GL_TEXTURE0);
}

void CLS::
End(){
    GetPCShaderCtrl()->EndShader();
}

#define CLS OverlayBuffer_cls

CLS::
OverlayBuffer_cls(Overlay_cls *pCSrc) : FilterBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    
    mCConfList.Add(&mCstrPath, "strPath=", "NONE");
    mCConfList.Add(&mCSequenceOffset, "sequenceOffset=", "0");
    mCConfList.Add(&mCbSequence, "Sequence=", "0");
    mCConfList.Add(&mCbFit, "Fit=", "0");
    mCConfList.Add(&mCAlpha, "Alpha=", "0.5");

    SyncFromSrc();
}

void CLS::
HideAndShow(){
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    mIndexPath = pC->CombineSizer(2, _("File Path"));
    pC->CombineConfItem(mCstrPath);
    pC->CombineButton(btnOverlayPath, "..", 30);
    
    pC->CombineSizer(2, _("Offset"));
    pC->CombineConfItem(mCSequenceOffset);
    pC->CombineConfItem(mCbSequence, _("Image Sequence"));

    pC->CombineSizer(2, _("Alpha"));
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.2;
    pC->CombineConfItem(mCAlpha);
    pC->CombineConfItem(mCbFit, _("Fit"));

    pC->FinishSetting();

    ((wxButtonEX*)pC->mVCItem[mIndexPath+2].mpCWindowLink)->mstrCommandStr = 
	GetStrName();
    
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();
    
    mCAlpha.SetBodyFromData(mpCSrc->GetPCParamGroup("Alpha")->GetVInterpolatedValue((float)time, FALSE));
}

void CLS::
ForceSrcToSyncBuffer(){ mpCSrc->SyncFromBuffer(this); }
