// (setq-c-offset 4)
// "clsFocalBlur.h"

#include "clsFocalBlur.h"
#include "clsFilterCtrl.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsFBO.h"
#include "../MYGL/clsGLExtention.h"
#include "../MYGL/clsTextureProvider.h"
#include "../SCENEOBJ/clsCamera.h"
#include "../SCENEOBJ/PROP/clsFocalBlurProp.h"

#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
#endif

#define CLS FocalBlur_cls

CLS::
FocalBlur_cls(wxString strName) : FilterBase_cls("FocalBlur_cls", strName){

    wxString strUpdate = "VIEW3D::PROPERTY_SCENE";
 
    wxString strConfListLabel = "FilterCtrl::"; strConfListLabel << strName;
    mCConfList.SetTag(strConfListLabel);
    mCConfList.Add(&mCQuality, "Quality=", "3", strUpdate);
    mCConfList.Add(&mCType, "Type=", "1", strUpdate);
    mCConfList.Add(&mCbUseGamma, "UseGamma=", "1", strUpdate);
}

void CLS::
SyncFromBuffer(FocalBlurBuffer_cls *pCBuffer){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
}

/*
unsigned int CLS::
GetIterNum(){
    int exposure = 1;
    if (mCbUseGamma.GetBodyBOOL() == TRUE) exposure = 2;
    return 2 * exposure;
}

void CLS::
BeginExposure(){
    ExposureParams_cls CParam;
    CParam.mGamma = 0.4545;
    GetPCShaderCtrl()->BeginExposure(&CParam);
}
*/

void CLS::
Begin(FBO_cls *pCFBO, PositionAndDirection_cls CCameraPos, 
      Camera_cls *pCCamera, double NT){

    bool bUseGamma = mCbUseGamma.GetBodyBOOL();

    float focalDistance = 10;
    float widthBase = 0.03;
    if (pCCamera != NULL){
	focalDistance = pCCamera->mpCFocalBlurProp->GetFocalDistance(CCameraPos, NT);
	widthBase = pCCamera->mpCFocalBlurProp->GetWidthBase(NT);
    }
    /*
    if (bUseGamma == TRUE){
	if (itr == 1) return BeginExposure();
    }
    */

    FocalBlurParams_cls CParam;
    CParam.mQuality = mCQuality.GetBodyINT();
    CParam.mGamma = (bUseGamma == TRUE) ? 2.2 : 1;
    CParam.mFocalDistance = focalDistance;
    CParam.mWidthBase = widthBase / focalDistance;
    
    if (mCType.GetBodyINT() == 1)
	GetPCShaderCtrl()->BeginShader("FocalBlurB", &CParam);
    else
	GetPCShaderCtrl()->BeginShader("FocalBlur", &CParam);


    //glEnable(GL_TEXTURE_2D);
    //glActiveTexture(GL_TEXTURE1);
    //pCFBO->mpCDepthTexture->BindTexture();

    //GetPCTextureProvider()->BindBlankTexture(TRUE);
    //GetPCTextureProvider()->BindTexture("z:/Z103a.png");

    /*
    OverlayParam_cls CParam;
    CParam.mAlpha = 0.5;
    CParam.mXScale = 1;
    CParam.mYScale = 1;
    GetPCShaderCtrl()->BeginShader("Overlay", &CParam);
    */
    
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP_TO_EDGE);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP_TO_EDGE);
    //glActiveTexture(GL_TEXTURE0);
}

void CLS::
End(){
    GetPCShaderCtrl()->EndShader();
}

#define CLS FocalBlurBuffer_cls

CLS::
FocalBlurBuffer_cls(FocalBlur_cls *pCSrc) : FilterBase_cls(*pCSrc){
    mpCSrc = pCSrc;

    mCConfList.Add(&mCQuality, "Quality=", "3");
    mCConfList.Add(&mCType, "Type=", "1");
    mCConfList.Add(&mCbUseGamma, "UseGamma=", "1");
    //
    SyncFromSrc();
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
		new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineSizer(2, _("Quality"));
    pC->mDragScale = 1;
    pC->CombineConfItem(mCQuality);
    
    wxString AstrType[2] = {_("Simple"), _("High Quality")};
    pC->SetChoiceList(2, AstrType);
    pC->CombineConfItem(mCType);
    
    pC->CombineSizer(1, _("Gamma Correction"));
    pC->CombineConfItem(mCbUseGamma, _("Use"));

    pC->FinishSetting();
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);

}

void CLS::
ForceSrcToSyncBuffer(){ mpCSrc->SyncFromBuffer(this); }
