// (setq-c-offset 4)
// "clsGlare.h"

#include "clsGlare.h"
#include "clsFilterCtrl.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MISC/clsMisc.h"
#include "../MYGL/clsGLExtention.h"

#include "../PARAMETER/clsParamGroup.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#define CLS Glare_cls

CLS::
Glare_cls(wxString strName) : FilterBase_cls("Glare_cls", strName){
    
    wxString strConfListLabel = "FilterCtrl::"; strConfListLabel << strName;
    wxString strUpdate = "VIEW3D::PROPERTY_SCENE";
    mCConfList.SetTag(strConfListLabel);
    mCConfList.Add(&mCType, "Type=", "0", strUpdate);
    mCConfList.Add(&mCQuality, "Quality=", "3", strUpdate);
    mCConfList.Add(&mCRayNumber, "Number=", "3", strUpdate);
    
    AddOneDimParam("Width", "W", 0.01);
    AddOneDimParam("Offset", "O", 0);
    AddOneDimParam("Crit", "C", 0.7);
    //AddOneDimParam("Enhance", "E", 1);

    ParamGroup_cls *pCParamEnhance = GetPCNewParamGroup("Enh./Fade", "En", "Fa");
    InitializeAndPushParam(pCParamEnhance, 1, 0.9);
}

void CLS::
SyncFromBuffer(GlareBuffer_cls *pCBuffer){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VWidth;
    VWidth.push_back(pCBuffer->mCWidth.GetBodyDOUBLE());
    SyncronizeParamForFilter("Width", VWidth);

    std::vector < float > VOffset;
    VOffset.push_back(pCBuffer->mCRayOffset.GetBodyDOUBLE());
    SyncronizeParamForFilter("Offset", VOffset);

    std::vector < float > VCrit;
    VCrit.push_back(pCBuffer->mCCrit.GetBodyDOUBLE());
    SyncronizeParamForFilter("Crit", VCrit);

    std::vector < float > VEnhance;
    VEnhance.push_back(pCBuffer->mCEnhance.GetBodyDOUBLE());
    VEnhance.push_back(pCBuffer->mCFade.GetBodyDOUBLE());
    SyncronizeParamForFilter("Enh./Fade", VEnhance);
}

void CLS::
Begin(){
    double time = GetPCTimeCtrl()->GetCurrentNT();

    GlareParams_cls CParam;

    CParam.mType = mCType.GetBodyINT();
    CParam.mWidth = GetPCParamGroup("Width")->GetVInterpolatedValue((float)time)[0];
    CParam.mOffset = GetPCParamGroup("Offset")->GetVInterpolatedValue((float)time)[0];
    CParam.mCrit = GetPCParamGroup("Crit")->GetVInterpolatedValue((float)time)[0];
    CParam.mEnhance = GetPCParamGroup("Enh./Fade")->GetVInterpolatedValue((float)time)[0];
    CParam.mFade = GetPCParamGroup("Enh./Fade")->GetVInterpolatedValue((float)time)[1];

    CParam.mQuality = mCQuality.GetBodyINT();
    CParam.mRayNumber = mCRayNumber.GetBodyINT();

    GetPCShaderCtrl()->BeginShader("Glare", &CParam);
}

void CLS::
End(){
    GetPCShaderCtrl()->EndShader();    
}

#define CLS GlareBuffer_cls

CLS::
GlareBuffer_cls(Glare_cls *pCSrc) : FilterBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCType, "Type=", "0");
    mCConfList.Add(&mCQuality, "Quality=", "3");
    mCConfList.Add(&mCRayNumber, "Number=", "3");

    SyncFromSrc();
}

void CLS::
HideAndShow(){
    bool bChange = FALSE;
    switch (mCType.GetBodyINT()){
    case 0:
	bChange |= mpCMNumLink->mVCItem[mIndexNumber].Show();
	break;
    case 1:
	bChange |= mpCMNumLink->mVCItem[mIndexNumber].Hide();
	break;
    }

    if (bChange == TRUE){
	wxWindow *pCPain = mpCMNumLink->GetParent();
	pCPain->GetParent()->GetParent()->Freeze();
	wxSize CBestSize = pCPain->GetParent()->GetBestSize();
	pCPain->GetParent()->SetSize(wxSize(-1, CBestSize.GetHeight()));
	pCPain->GetParent()->GetParent()->Layout();
	pCPain->GetParent()->GetParent()->PostSizeEvent();
	pCPain->GetParent()->GetParent()->Thaw();
    }
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    wxString AstrType[2] = {_("Streaks"), _("Fog glow")};
    pC->SetChoiceList(2, AstrType);
    pC->CombineConfItem(mCType, _("Type"));

    mIndexNumber = pC->CombineSizer(2, _("Number/Offset"));
    pC->CombineConfItem(mCRayNumber);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCRayOffset);

    pC->CombineSizer(2, _("Width/Quality"));
    pC->SetLimit(0, 0.2);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCWidth);
    pC->SetLimit(1, 32);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCQuality);
    
    //pC->CombineSizer(2, _("Crit/Enhance"));
    pC->SetLimit(0, 16);
    pC->mDragScale = 0.1;

    pC->CombineConfItem(mCCrit, _("Criterion"));

    pC->CombineSizer(2, _("Enhance/Fade"));
    pC->SetLimit(0, 16);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCEnhance);
    
    pC->CombineConfItem(mCFade);

    pC->FinishSetting();
    
    pC->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_COMBOBOX_SELECTED);
    
    mpCMNumLink = pC;

    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    mCWidth.SetBodyFromData(mpCSrc->GetPCParamGroup("Width")->GetVInterpolatedValue((float)time, FALSE));
    mCRayOffset.SetBodyFromData(mpCSrc->GetPCParamGroup("Offset")->GetVInterpolatedValue((float)time, FALSE));
    mCCrit.SetBodyFromData(mpCSrc->GetPCParamGroup("Crit")->GetVInterpolatedValue((float)time, FALSE));
    mCEnhance.SetBodyFromData(mpCSrc->GetPCParamGroup("Enh./Fade")->GetVInterpolatedValue((float)time, FALSE)[0]);
    mCFade.SetBodyFromData(mpCSrc->GetPCParamGroup("Enh./Fade")->GetVInterpolatedValue((float)time, FALSE)[1]);

}

void CLS::
ForceSrcToSyncBuffer(){ mpCSrc->SyncFromBuffer(this); }

