// (setq-c-offset 4)
// "clsSSAO.h"

#include "clsSSAO.h"
#include "clsFilterCtrl.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsFBO.h"
#include "../MYGL/clsGLExtention.h"

#include "../TIMELINE/clsTimeCtrl.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/EX/wxButtonEX.h"

#include "../../wxWidgetEnum.h"

#define CLS SSAO_cls

CLS::
SSAO_cls(wxString strName) : FilterBase_cls("SSAO_cls", strName){

    wxString strUpdate = "VIEW3D::PROPERTY_SCENE";
 
    wxString strConfListLabel = "FilterCtrl::"; strConfListLabel << strName;
    mCConfList.SetTag(strConfListLabel);
    mCConfList.Add(&mCQuality, "Quality=", "3", strUpdate);
    mCConfList.Add(&mCFactor, "Factor=", "0.5", strUpdate);
    mCConfList.Add(&mCRadius, "Radius=", "1", strUpdate);
    mCConfList.Add(&mCType, "Type=", "0", strUpdate);

    ParamGroup_cls *pCBaseColor = GetPCNewParamGroup("BaseColor", "R", "G", "B");
    InitializeAndPushParam(pCBaseColor, 0, 0, 0);
}

wxColor CLS::
GetWXColor(){

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < double > VColor = GetPCParamGroup("BaseColor")->GetVInterpolatedValue(NT);
    wxColor CColor(VColor[0]*255, VColor[1]*255, VColor[2]*255);
    return CColor;
}

void CLS::
SetBaseColor(std::vector < float > VClr){
    
    ParamGroup_cls *pCParamGroup = GetPCParamGroup("BaseColor");
    double time = GetPCTimeCtrl()->GetCurrentNT();

    int keyType;
    double easeFactor;
    GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);

    GetPCUndo()->OpenBlock("ClrChange");

    for (int i = 0; i < 3; i++){
	float newValue = VClr[i];
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	float oldValue = CParam.GetValue(time);
	if (oldValue != newValue){
	    wxString strTag = GetPCFilterCtrl()->GetStrSelf() << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam.GetStrName();
	    CParam.SetValueAtCommand(time, newValue, strTag,
				     keyType, easeFactor);
	}
    }

    GetPCUndo()->CloseBlock();
    
}

void CLS::
SyncFromBuffer(SSAOBuffer_cls *pCBuffer){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VBaseColor = pCBuffer->mCBaseColor.GetBodyFLOATARRAY();
    SyncronizeParamForFilter("BaseColor", VBaseColor);
}

void CLS::
Begin(FBO_cls *pCFBO){
    double NT = GetPCTimeCtrl()->GetCurrentNT();

    SSAOParam_cls CParam;
    CParam.mQuality = mCQuality.GetBodyINT();
    CParam.mFactor = mCFactor.GetBodyDOUBLE();
    CParam.mRadius = mCRadius.GetBodyDOUBLE();
    CParam.mType = mCType.GetBodyINT();
    std::vector < double > VColor = GetPCParamGroup("BaseColor")->GetVInterpolatedValue(NT);
    CParam.mBaseColor[0] = VColor[0];
    CParam.mBaseColor[1] = VColor[1];
    CParam.mBaseColor[2] = VColor[2];

    GetPCShaderCtrl()->BeginShader("SSAO", &CParam);
}

void CLS::
End(){
    GetPCShaderCtrl()->EndShader();
}

#define CLS SSAOBuffer_cls

CLS::
SSAOBuffer_cls(SSAO_cls *pCSrc) : FilterBase_cls(*pCSrc){
    mpCSrc = pCSrc;

    mCConfList.Add(&mCQuality, "Quality=", "3");
    mCConfList.Add(&mCFactor, "Factor=", "0.5");
    mCConfList.Add(&mCRadius, "Radius=", "1");
    mCConfList.Add(&mCType, "Type=", "0");
    //
    SyncFromSrc();
}

void CLS::
UpdateButtonColor(MultiInputPanel_cls *pCMNum){
    pCMNum->mVCItem[mIndexBtnColor].mpCWindowLink->SetBackgroundColour(mCBaseColor.GetBodyWXColor());
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
		new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    wxString AstrType[3] = {_("Darken"), _("Lighten"), _("Both")};
    pC->SetChoiceList(3, AstrType);
    pC->CombineConfItem(mCType, _("Type"));
    
    pC->CombineSizer(2, _("Quality/Factor"));
    pC->mDragScale = 1;
    pC->SetLimit(1, 64);
    pC->CombineConfItem(mCQuality);
    pC->mDragScale = 1;
    pC->SetLimit(0, 4);
    pC->CombineConfItem(mCFactor);

    pC->CombineConfItem(mCRadius, _("Sampling Radius"));

    pC->CombineSizer(2, _("Color"));
    pC->mDragScale = 0.1;
    pC->SetLimit(0, 1);
    pC->CombineConfItem(mCBaseColor);
    pC->mContextMenuType = _ITEM_MENU_COLOR;
    mIndexBtnColor = pC->CombineButton(btnColorSSAO, " ", 30);
    
    pC->FinishSetting();

    pC->mVCItem[mIndexBtnColor].mpCWindowLink->SetBackgroundStyle(wxBG_STYLE_PAINT);
    ((wxButtonEX*)pC->mVCItem[mIndexBtnColor].mpCWindowLink)->mstrCommandStr = GetStrName();
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);

    double time = GetPCTimeCtrl()->GetCurrentNT();

    mCBaseColor.SetBodyFromData(mpCSrc->GetPCParamGroup("BaseColor")->GetVInterpolatedValue((float)time, FALSE));

}

void CLS::
ForceSrcToSyncBuffer(){ mpCSrc->SyncFromBuffer(this); }
