// "clsGLButton.h"
// (setq c-basic-offset 4)

#include "clsGLButton.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../WIDGETCTRL/WX/myArtProv.h"

#define CLS GLButton_cls

CLS::
GLButton_cls(int offsetType, int offsetX, int offsetY){
    mOffsetType = offsetType;
    mOffsetX = offsetX;
    mOffsetY = offsetY;
    mTexID = 0;
    mbMouseDown = FALSE;
}

void CLS::
MakeTextureFromImage(wxImage& CImage, bool bFlip){
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &mTexID);

    if (bFlip == TRUE) CImage = CImage.Mirror(FALSE);
    mCSizeImg = CImage.GetSize();
    int width = mCSizeImg.GetWidth();
    int height = mCSizeImg.GetHeight();

    bool bMask = CImage.HasMask();
    if (bMask == TRUE) MyArtProvider::ConvertMaskToAlpha(CImage);
    
    unsigned char *pRGB = CImage.GetData();
    unsigned char *pAlpha = CImage.GetAlpha();

    int size = width * height;
    
    unsigned char *pTgt = new unsigned char [size * 4];

    for (int i = 0; i < size; i++){
	pTgt[i * 4] = pRGB[i*3];
	pTgt[i * 4 + 1] = pRGB[i*3 + 1];
	pTgt[i * 4 + 2] = pRGB[i*3 + 2];
	pTgt[i * 4 + 3] = pAlpha[i];
    }

    glBindTexture(GL_TEXTURE_2D, mTexID);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, 
		 GL_RGBA, GL_UNSIGNED_BYTE, pTgt);

    delete [] pTgt;
}

int CLS::
GetCoordYFromGLtoWX(int y){
    return mCSizeCanvas.GetHeight() - y;
}

void CLS::
DrawButton(bool bState){

    int x0 = 0, y0 = 0;
    switch(mOffsetType){
    case _TOP_LEFT:
	x0 = mOffsetX;
	y0 = mCSizeCanvas.GetHeight() - mCSizeImg.GetHeight() - mOffsetY;
	break;
    case _BTM_LEFT:
	x0 = mOffsetX;
	y0 = mOffsetY;
	break;
    case _TOP_RIGHT:
	x0 = mCSizeCanvas.GetWidth() - mCSizeImg.GetWidth() - mOffsetX;
	y0 = mCSizeCanvas.GetHeight() - mCSizeImg.GetHeight() - mOffsetY;
	break;
    case _BTM_RIGHT:
	x0 = mCSizeCanvas.GetWidth() - mCSizeImg.GetWidth() - mOffsetX;
	y0 = mOffsetY;
	break;
    }
    int x1 = x0 + mCSizeImg.GetWidth();
    int y1 = y0 + mCSizeImg.GetHeight();

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    
    if (mbMouseOn == TRUE){
	glDisable(GL_TEXTURE_2D);
	glColor4f(0.2f, 0.2f, 0.3f, 1.0f);
	glBegin(GL_QUADS);
	glVertex2d(x0, y0); glVertex2d(x1, y0);
	glVertex2d(x1, y1); glVertex2d(x0, y1);
	glEnd();
	glColor4f(0.3f, 0.3f, 0.5f, 1.0f);
	glBegin(GL_LINE_LOOP);
	glVertex2d(x0, y0); glVertex2d(x1, y0);
	glVertex2d(x1, y1); glVertex2d(x0, y1);
	glEnd();
    }

    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //At@

    glBindTexture( GL_TEXTURE_2D, mTexID);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glBegin(GL_QUADS);
    /*
    if (bState == FALSE)
	glColor4f(0.3f, 0.3f, 0.28f, 1.0f);
    else
	glColor4f(0.9f, 0.9f, 1.0f, 1.0f);
    */
    if (mbMouseDown == TRUE)
	glColor4f(0.8, 0.8, 0.8, 1.0);
    else
	glColor4f(1.0, 1.0, 1.1, 1.0);
    
    glTexCoord2f(0, 0); glVertex2d(x0, y0);
    glTexCoord2f(1, 0); glVertex2d(x1, y0);
    glTexCoord2f(1, 1); glVertex2d(x1, y1);
    glTexCoord2f(0, 1); glVertex2d(x0, y1);

    glEnd();

    glPopAttrib();
}

bool CLS::
CheckButtonDown(int mouseX, int mouseY){
    mbMouseDown = IsMouseOnButton(mouseX, mouseY);
    return mbMouseDown;
}

bool CLS::
CheckButtonUp(int mouseX, int mouseY){
    if (mbMouseDown == FALSE) return FALSE;
    mbMouseDown = FALSE;
    return IsMouseOnButton(mouseX, mouseY);
}

bool CLS::
IsMouseOnButton(int mouseX, int mouseY){

    int x0 = 0, y0 = 0;
    switch(mOffsetType){
    case _TOP_LEFT:
	x0 = mOffsetX;
	y0 = mCSizeCanvas.GetHeight() - mCSizeImg.GetHeight() - mOffsetY;
	break;
    case _BTM_LEFT:
	x0 = mOffsetX;
	y0 = mOffsetY;
	break;
    case _TOP_RIGHT:
	x0 = mCSizeCanvas.GetWidth() - mCSizeImg.GetWidth() - mOffsetX;
	y0 = mCSizeCanvas.GetHeight() - mCSizeImg.GetHeight() - mOffsetY;
	break;
    case _BTM_RIGHT:
	x0 = mCSizeCanvas.GetWidth() - mCSizeImg.GetWidth() - mOffsetX;
	y0 = mOffsetY;
	break;
    }
    
    y0 = GetCoordYFromGLtoWX(y0);
    
    int x1 = x0 + mCSizeImg.GetWidth();
    int y1 = y0;
    y0 = y0 - mCSizeImg.GetHeight();

    if (mouseX > x0 && mouseX < x1 &&
	mouseY > y0 && mouseY < y1)
	return TRUE;

    return FALSE;
}

bool CLS::
UpdateStatus(wxSize CSizeCanvas, 
	     int mouseX, int mouseY){
    mCSizeCanvas = CSizeCanvas;
    
    bool bMouseOn = IsMouseOnButton(mouseX, mouseY);
    if (bMouseOn == FALSE && mbMouseDown == TRUE){
	mbMouseOn = bMouseOn;
	mbMouseDown = FALSE;
	return TRUE;
    }
	
    if (mbMouseOn != bMouseOn){
	mbMouseOn = bMouseOn;
	return TRUE;
    }
    return FALSE;
}

