// (setq c-basic-offset 4)
// "clsNavigator.h"

#include "clsNavigator.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../../GLOBJ/clsVBOCone.h"
#include "../../GLOBJ/clsVBOCube.h"
#include "../../MYGL/clsFBO.h"

#include "../../COMMON/CommonDefinitions.h"

#define CLS Navigator_cls

CLS::
Navigator_cls(){
    mpCNaviFBO = NULL;
}

CLS::
~Navigator_cls(){
    if (mpCNaviFBO != NULL) delete mpCNaviFBO;
}

void CLS::
SetColorAxis(int axis){
    switch(axis){
    case X: glColor4f(1.0, 0.1, 0.1, 1.0); break;
    case Y: glColor4f(0.1, 1.0, 0.1, 1.0); break;
    case Z: glColor4f(0.2, 0.2, 1.0, 1.0); break;
    }
}

void CLS::
ShowArrow(bool bLong){
    glBegin(GL_LINES);
    if (bLong == TRUE){
	glVertex3d(-10, 0, 0);
	glVertex3d(+10, 0, 0);
    }
    else{
	glVertex3d(0, 0, 0);
	glVertex3d(+1, 0, 0);
    }
    glEnd();
    glTranslated(0.85, 0, 0);
    glScaled(0.2, 0.12, 0.12);
    glRotatef(90, 0, 1, 0);
    GCClassProvider.GetPCVBOCone()->Draw(_VBO_OBJECT, 0);
}

void CLS::
ShowBlock(bool bLong){
    glBegin(GL_LINES);
    if (bLong == TRUE){
	glVertex3d(-10, 0, 0);
	glVertex3d(+10, 0, 0);
    }
    else{
	glVertex3d(0, 0, 0);
	glVertex3d(+1, 0, 0);
    }
    glEnd();
    glTranslated(0.85, 0, 0);
    glScaled(0.2, 0.12, 0.12);
    glRotatef(90, 0, 1, 0);
    GCClassProvider.GetPCVBOCube()->Draw(_VBO_OBJECT, 0);
}

void CLS::
ShowRing(bool bLong){
    
    if (bLong == TRUE){
	glBegin(GL_LINES);
	glVertex3d(-10, 0, 0);
	glVertex3d(10, 0, 0);
	glEnd();
    }
    
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= 36; i++){
	float theta = 3.14159 * i * 10 / 180;
	glVertex3d(0, cos(theta), sin(theta));
    }
    glEnd();
}

void CLS::
ShowNaviTranslate(bool bAxisX, bool bAxisY, bool bAxisZ){
    SetColorAxis(X);
    glPushMatrix(); // +1
    ShowArrow(bAxisX);
    glPopMatrix(); // +-0

    SetColorAxis(Y);
    glPushMatrix(); // +1
    glRotatef(90, 0, 0, 1);
    ShowArrow(bAxisY);
    glPopMatrix(); // +-0

    SetColorAxis(Z);
    glPushMatrix(); // +1
    glRotatef(-90, 0, 1, 0);
    ShowArrow(bAxisZ);
    glPopMatrix(); // +-0
}

void CLS::
ShowNaviRotate(bool bAxisX, bool bAxisY, bool bAxisZ){
    SetColorAxis(X);
    glPushMatrix();
    ShowRing(bAxisX);
    glPopMatrix();

    SetColorAxis(Y);
    glPushMatrix();
    glRotatef(90, 0, 0, 1);
    ShowRing(bAxisY);
    glPopMatrix();
    
    SetColorAxis(Z);
    glPushMatrix();
    glRotatef(-90, 0, 1, 0);
    ShowRing(bAxisZ);
    glPopMatrix();
}

void CLS::
ShowNaviScale(bool bAxisX, bool bAxisY, bool bAxisZ){
    SetColorAxis(X);
    glPushMatrix(); // +1
    ShowBlock(bAxisX);
    glPopMatrix(); // +-0

    SetColorAxis(Y);
    glPushMatrix(); // +1
    glRotatef(90, 0, 0, 1);
    ShowBlock(bAxisY);
    glPopMatrix(); // +-0

    SetColorAxis(Z);
    glPushMatrix(); // +1
    glRotatef(-90, 0, 1, 0);
    ShowBlock(bAxisZ);
    glPopMatrix(); // +-0
}

void CLS::
MakeImage(PositionAndDirection_cls CCamPos){

    int sizeFBO = 192;

    if (mpCNaviFBO == NULL){
	mpCNaviFBO = new FBO_cls (sizeFBO, sizeFBO, NULL);
	mpCNaviFBO->MakeFBO(TRUE);
    }

    mpCNaviFBO->Begin(TRUE);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_DEPTH_TEST);

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    GLPerspective_cls::SetProjection(sizeFBO, sizeFBO);
    GLPerspective_cls::SetFrustum(sizeFBO, sizeFBO, 15, 0.1, 100);
    
    PositionAndDirection_cls CNaviCameraPos;
    CNaviCameraPos = CCamPos;
    LOOP_XYZ CNaviCameraPos.mPos[xyz] = -6 * CCamPos.mTgt[xyz];

    GLPerspective_cls::SetModelView(CNaviCameraPos, CNaviCameraPos);

    
    glLineWidth(5);
    glTranslated(-0.25, -0.25, -0.25);

    ShowNaviTranslate();
    
    mpCNaviFBO->End();
}

void CLS::
Draw(wxSize CSizeCanvas, wxSize CSizeImg){

    int x0 = 16;
    int y0 = 16;

    int x1 = x0 + CSizeImg.GetWidth();
    int y1 = y0 + CSizeImg.GetHeight();

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glEnable(GL_TEXTURE_2D);
    //glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //At@

    mpCNaviFBO->BindTexture();

    glColor4f(1.0, 1.0, 1.0, 1.0);

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_REPEAT);

    glBegin(GL_QUADS);

    glTexCoord2f(0, 0); glVertex2d(x0, y0);
    glTexCoord2f(1, 0); glVertex2d(x1, y0);
    glTexCoord2f(1, 1); glVertex2d(x1, y1);
    glTexCoord2f(0, 1); glVertex2d(x0, y1);

    glEnd();

    glPopAttrib();
}

