// "clsGLCanvas.h"
// (setq c-basic-offset 4)

#include "clsGLCanvas.h"

#include "clsImageComposer.h"
#include "clsGLNumericalPanel.h"
#include "clsStereogram.h"
#include "clsView3D.h"
#include "NAVIGATORS/clsGLButton.h"
#include "NAVIGATORS/clsHitTest.h"
#include "NAVIGATORS/clsNavigator.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../FILTER/clsFilterCtrl.h"
#include "../GRAPHIC/clsImageIO.h"
#include "../GRAPHIC/clsImageOutput.h"
#include "../GLOBJ/clsVBOCameraBox.h"
#include "../GLOBJ/clsVBOCone.h"
#include "../GLOBJ/clsVBOCube.h"
#include "../GLOBJ/clsVBOCylinder.h"
#include "../GLOBJ/clsVBOEmpty.h"
#include "../GLOBJ/clsVBOPlane.h"
#include "../GLOBJ/clsVBOSphere.h"
#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsFBO.h"
#include "../MYGL/clsGLExtention.h"
#include "../MYGL/clsGLInfo.h"
#include "../MYGL/clsGLPerspective.h"
#include "../MYGL/clsGLSL.h"
#include "../MYGL/clsGLString2D.h"
#include "../MYGL/clsShadowVolumeCtrl.h"
#include "../MYGL/clsWideCamera.h"
#include "../PREFERENCE/clsPreference.h"
#include "../SCENEOBJ/clsCamera.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/clsWorld.h"
#include "../SCENEOBJ/PROP/clsCameraProp.h"
#include "../SCENEOBJ/PROP/clsFocalBlurProp.h"
#include "../SCENEOBJ/PROP/clsFogProp.h"
#include "../SCENEOBJ/PROP/clsStereoProp.h"
#include "../TIMELINE/clsEnsembleCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/WX/myArtProv.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"

#include "../COMMON/CommonDefinitionsGL.h"
#include "../COMMON/CommonDefinitions.h"

#include "../../clsFrameMain.h"

#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#ifdef WIN32
extern PFNGLCLAMPCOLORPROC glClampColor;
#endif

//static bool GbGLDrawingLock = FALSE;

#define CLS GLCanvas_cls

BEGIN_EVENT_TABLE(CLS, GLCanvasBase_cls)
EVT_LEFT_DOWN(CLS::OnMouse)
EVT_LEFT_UP(CLS::OnMouse)
EVT_MIDDLE_DOWN(CLS::OnMouse)
EVT_MIDDLE_UP(CLS::OnMouse)
EVT_RIGHT_DOWN(CLS::OnMouse)
EVT_RIGHT_UP(CLS::OnMouse)
EVT_LEFT_DCLICK(CLS::OnMouse)
EVT_MIDDLE_DCLICK(CLS::OnMouse)
EVT_RIGHT_DCLICK(CLS::OnMouse)
EVT_MOUSEWHEEL(CLS::OnMouse)
EVT_MOTION(CLS::OnMouseMotion)
EVT_CHAR_HOOK(CLS::OnKeyDown)
EVT_PAINT (CLS::OnPaint)
END_EVENT_TABLE()

CLS::
GLCanvas_cls(View3D_cls *pCView3D, int *attribList,
	     long style, wxString name/*, unsigned int selfIndex*/) :
GLCanvasBase_cls(pCView3D, attribList, style, name)
{
    mpCParent = pCView3D;
    Initialize();
}

#ifdef WIN32
CLS::
GLCanvas_cls(View3D_cls *pCView3D, wxGLAttributes& CGLAttribute,
	     long style, wxString name/*, unsigned int selfIndex = 0*/) :
GLCanvasBase_cls(pCView3D, CGLAttribute, style, name)
{
    mpCParent = pCView3D;
    Initialize();
}
#endif

CLS::
~GLCanvas_cls(){
    delete mpCBtnNumPanel;
    delete mpCGLString2D;
    delete mpCMouseCtrl;
    delete mpCMouseInfoInertia;
    delete mpCNavigator;
    delete mpCHitTest;
    if (mpCFBODepth != NULL) delete mpCFBODepth;
    if (mpCFBORenderOut != NULL) delete mpCFBORenderOut;
    if (mpCFBOStereoBuffer[0] != NULL) delete mpCFBOStereoBuffer[0];
    if (mpCFBOStereoBuffer[1] != NULL) delete mpCFBOStereoBuffer[1];
    if (mpCFBOStereoDepth[0] != NULL) delete mpCFBOStereoDepth[0];
    if (mpCFBOStereoDepth[1] != NULL) delete mpCFBOStereoDepth[1];
    if (mpCFBOImageCompositBuffer != NULL) delete mpCFBOImageCompositBuffer;
}

void CLS::
Initialize(){
    SetBackgroundColour(*wxBLACK);
    SetDoubleBuffered(FALSE);

    // LINUXłGLContext ̓EBhE\ԂoĂȂƓȂ̂
    // RXgN^ł͂Ȃō쐬
    /*
#ifdef WIN32
    SetGLContext(FALSE); // 쐬邾SetCurrentȂ
#endif
    */

    static const int attribs[] = { WX_GL_RGBA, WX_GL_DOUBLEBUFFER, 0 };
    bool bSupport = wxGLCanvas::IsDisplaySupported(attribs);
    wxString strSupport = "bSupport=";
    strSupport << bSupport << ENDL;
    wxLogMessage(strSupport);
    
    mpCBtnNumPanel = new GLButton_cls(_TOP_RIGHT, 0, 0);

    mbInitialized    = FALSE;
    mbInertia        = FALSE;
    mbViewDragged    = FALSE;
    mScalingCurrent = mScalingInit = 0;
    mInputMode       = 0;
    mbStatusDrawing  = FALSE;
    ClearQueue();
    mbRedrawQueue    = TRUE;
    mDrag[X] = mDrag[Y] = mDrag[Z] = 0;

    LoadIdentity(mMat16NavigatorModelview);
    LoadIdentity(mMat16NavigatorModelOnly);

    mpCGLString2D = new GLStrings2D_cls;
    mpCMouseCtrl  = new MouseCtrl_cls;
    mpCMouseInfoInertia = new MouseCtrl_cls;
    mpCNavigator = new Navigator_cls;
    mpCHitTest = new Hittest_cls;
    mpCFBOCurrentLink = NULL;
    mpCFBOFinalOut = NULL;
    mpCFBODepth = NULL;
    mpCFBOStereoFiltered[0] = NULL;
    mpCFBOStereoFiltered[1] = NULL;
    mpCFBOStereoDepth[0] = NULL;
    mpCFBOStereoDepth[1] = NULL;
    
    mMillisecondIndex = 0;
    for (int i = 0; i < 10; i++) mMillisecondList[i] = 0;

    printf("Initialized\n");
}

void CLS::
ClearQueue(){
    mbRedrawQueue    = FALSE;
    mbHittestQueue   = FALSE;
    mbHittestQueueAdditive = FALSE;
}

void CLS::
ClearBuffer(wxString strClrTag){
    int BGAlpha = GetPCColorProvider()->mCBGAlphaType.GetBodyINT();
    if (BGAlpha == 1)
	ColorProvider_cls::SetGLClearColor(strClrTag, FALSE, 0);
    else
	ColorProvider_cls::SetGLClearColor(strClrTag);

    if (strClrTag.IsSameAs("GLBG")){
	World_cls *pCWorld = (World_cls*)GetPCSceneObjCtrl()->GetPCWorld();
	if (pCWorld != NULL &&
	    pCWorld->mpCFogProp->mCFogType.GetBodyINT() != _FOG_NONE &&
	    pCWorld->mpCFogProp->mCbOverwriteBG.GetBodyBOOL() == TRUE){
	    float clrFog[4];
	    pCWorld->GetFogColor(clrFog);
	    if (BGAlpha == 1)
		glClearColor(clrFog[0], clrFog[1], clrFog[2], 0);
	    else
		glClearColor(clrFog[0], clrFog[1], clrFog[2], 1);
	}
    }

    glClear( GL_COLOR_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void CLS::
InitGLUT(){

    int argc = 1;
    char **argv = new char* [1];
    argv[0] = new char [6];
    sprintf(argv[0], "dummy");
    glutInit(&argc, argv);
    delete [] argv[0]; 
    delete [] argv;

}

bool CLS::
InitializeGLCommonIfNot(){
    if (msbInitializedCommon == TRUE) return FALSE;
    InitGLUT();
    msbInitializedCommon = TRUE;
    GCClassProvider.GetPCGLExtentions()->InitializeFunctions();
    GCClassProvider.GetPCGLSL()->InitializeFunctions();
    GCClassProvider.GetPCVBOCameraBox()->Initialize();
    GCClassProvider.GetPCVBOCone()->Initialize();
    GCClassProvider.GetPCVBOCube()->Initialize();
    GCClassProvider.GetPCVBOCylinder()->Initialize();
    GCClassProvider.GetPCVBOEmpty()->Initialize();
    GCClassProvider.GetPCVBOPlane()->Initialize();
    GCClassProvider.GetPCVBOSphere()->Initialize();
    GCClassProvider.GetPCGLInfo()->Initialize();
    return TRUE;
}

bool CLS::
InitializeGLIfNot(){
    if (mbInitialized == TRUE) return FALSE;

    wxImage CImage = ART("wxART_SHOW_GL").ConvertToImage();
    mpCBtnNumPanel->MakeTextureFromImage(CImage);

    mpCFBORenderOut = new FBO_cls(2048, 2048, this);
    mpCFBODepth = new FBO_cls(16, 16, this);
    mpCFBOStereoBuffer[0] = new FBO_cls(16, 16, this);
    mpCFBOStereoBuffer[1] = new FBO_cls(16, 16, this);
    mpCFBOStereoDepth[0] = new FBO_cls(16, 16, this); 
    mpCFBOStereoDepth[1] = new FBO_cls(16, 16, this);
    mpCFBOImageCompositBuffer = new FBO_cls(16, 16, this);

    glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
    
    mbInitialized = TRUE;
    return TRUE;
}

void CLS::
RefreshEX(){
    mbRedrawQueue = TRUE;

#ifdef WIN32
    //HWND hWnd = GetHandle();
    //InvalidateRect(hWnd, NULL, FALSE);
    Refresh();
#else
    Refresh();
#endif
    
}

int CLS::
GetPercent(int fbo, int fboNum, int SVIdx, int SVSampleNum,
	   int ensemble, int ensembleSampleNum){
    int total = fboNum * SVSampleNum * ensembleSampleNum;
    int curr = (fbo*SVSampleNum + SVIdx)*ensembleSampleNum + ensemble;
    return 100 * curr / total;
}

double CLS::
GetAspectRatio(){
    return (double)mCWindowSize.GetWidth() / mCWindowSize.GetHeight();
}

double CLS::
GetModByAspectX(){
    double aspect = GetAspectRatio();
    if (aspect > 1) return 1.0;
    return aspect;
}

double CLS::
GetModByAspectY(){
    double aspect = GetAspectRatio();
    if (aspect < 1) return 1.0;
    return 1.0 / aspect;
}

Camera_cls* CLS::
GetPCCamera(){
    wxString strCamera = mpCParent->mCstrCamera.GetBodySTR();
    SceneObj_cls *pCObj = GetPCSceneObjCtrl()->GetPCFromName(strCamera);
    if (pCObj == NULL) return NULL;
    if (pCObj->mstrClassName.IsSameAs("Camera_cls") == FALSE) return NULL;

    return (Camera_cls*)pCObj;
}

int CLS::
GetCameraType(){
    Camera_cls *pCCamera = GetPCCamera();
    if (mpCParent->IsUseCamera() == TRUE && pCCamera != NULL){
	return pCCamera->mCType.GetBodyINT();
    }
    return mpCParent->mCWideCameraType.GetBodyINT();
}

bool CLS::
DoesUseDepthBuffer(){
    bool bRet = FALSE;
    bRet |= GetPCFilterCtrl()->IsVaildFilterExist("FocalBlur_cls");
    bRet |= GetPCFilterCtrl()->IsVaildFilterExist("SSAO_cls");
    bRet |= GetPCFilterCtrl()->IsVaildFilterExist("FilterDepth_cls");
    return bRet;
}

int CLS::
GetStereoNum(){
    Camera_cls *pCCamera = GetPCCamera();
    if (mpCParent->IsUseCamera() == TRUE && pCCamera != NULL){
	return pCCamera->mpCStereoProp->GetStereoNum();
    }
    return 1;
}

int CLS::
GetEyeType(){
    Camera_cls *pCCamera = GetPCCamera();
    if (mpCParent->IsUseCamera() == TRUE && pCCamera != NULL){
	return pCCamera->mpCStereoProp->GetEyeType();
    }
    return _STEREO_NONE;
}

int CLS::
GetImageCompositeNum(){
    return GetPCTimeCtrl()->GetImageCompositeNum();
}

double CLS::
GetViewAngle(){
    Camera_cls *pCCamera = GetPCCamera();
    if (mpCParent->IsUseCamera() == TRUE && pCCamera != NULL){
	if (pCCamera->mCType.GetBodyINT() > 1)
	    return pCCamera->mpCCameraProp->mCAngleV.GetBodyDOUBLE();
	if (GetPCTimeCtrl()->mbRecordRender == TRUE)
	    return pCCamera->GetAngleL();
    }
    return mpCParent->mCViewAngleL.GetBodyDOUBLE();
}

double CLS::
GetSizeOrtho(){
    Camera_cls *pCCamera = GetPCCamera();
    if (mpCParent->IsUseCamera() == TRUE && pCCamera != NULL){
	if (GetPCTimeCtrl()->mbRecordRender == TRUE)
	    return pCCamera->GetSizeOrthoL();
    }
    double distance = GetDistance(mCCameraBase.mPos, mCPivot.mPos);
    return distance * GetViewAngle() * 3.14159 / 180;

    //return mpCParent->mCSizeOrthoL.GetBodyDOUBLE();
}

double CLS::
GetRelativeAngleDeg(bool bForceSnap){
    double relative = - mAngularMod * 3.14159 * 2 + mAngularCurrent - mAngularInit;
    double deg = 180 * relative / 3.14159;

    if (mpCMouseCtrl->IsMoveSnap() == TRUE || bForceSnap){
	deg += 7.5;
	deg -= fmod(deg, 15);
	mbDragSnapping = TRUE;
    }
    else
	mbDragSnapping = FALSE; 
    
    return deg;
}

std::vector < double > CLS::
GetVMinMax(){
    Camera_cls *pCCamera = GetPCCamera();
    if (mpCParent->IsUseCamera() == TRUE && pCCamera != NULL){
	double NT = GetPCTimeCtrl()->GetCurrentNT();
	return pCCamera->GetVMinMax(NT);
    }
    std::vector < double > VMinMax;
    VMinMax.push_back(GetPCPreference()->mCGLDepthMin.GetBodyDOUBLE());
    VMinMax.push_back(GetPCPreference()->mCGLDepthMax.GetBodyDOUBLE());
    return VMinMax;
}

double CLS::
GetDomeTilt(){
    Camera_cls *pCCamera = GetPCCamera();
    if (mpCParent->IsUseCamera() == TRUE && pCCamera != NULL){
	return pCCamera->mpCCameraProp->mCTiltAngle.GetBodyDOUBLE();
    }
    return 0;
}

void CLS::
GetSampling(bool *pbUseEnsemble, bool *pbUseSV, bool *pbUseBlur,
	    int *pEnsembleSampleNum, int *pSVSampleIni, int *pSVSampleNum,
	    int *pBlurSampleNum){
    *pbUseEnsemble = GetPCEnsembleCtrl()->mCbUseEnsemble.GetBodyBOOL();
    //*pbUseSV = GetPCShadowVolumeCtrl()->mCbUseShadowVolume.GetBodyBOOL();
    *pbUseBlur = GetPCBlurCtrl()->mCbUseMotionBlur.GetBodyBOOL();
    *pEnsembleSampleNum = GetPCEnsembleCtrl()->GetEnsembleSampleNum();
    *pSVSampleIni = GetPCShadowVolumeCtrl()->GetShadowVolumeSampleIni();
    *pSVSampleNum = GetPCShadowVolumeCtrl()->GetShadowVolumeSampleNum();
    *pBlurSampleNum = GetPCBlurCtrl()->GetBlurSampleNum();
    if (*pSVSampleNum > 1) *pbUseSV = TRUE;
    else                   *pbUseSV = FALSE;
}

bool CLS::
IsMouseOnClient(){
    wxRect  CScreenRect = GetScreenRect();
    wxPoint CMousePos   = wxGetMousePosition();
    return CScreenRect.Contains(CMousePos);
}

bool CLS::
ProcInertia(){
    if (mbInertia == FALSE) return FALSE;

    MouseNormalized_cls CMP(mpCMouseInfoInertia, mCWindowSize);
    mCCameraBase.Gyration(CMP, GetModByAspectX(), GetModByAspectY());

    mpCParent->SetPositionView(mCCameraBase);
    mpCParent->SetDirectionView(mCCameraBase);

    return TRUE;
}

bool CLS::
ProcViewOrItemDragging(){
    bool trigRefresh = FALSE;
    MouseNormalized_cls CMP(mpCMouseCtrl, mCWindowSize);

    if (mpCMouseCtrl->IsInertia()){
	mbInertia = TRUE;
	*mpCMouseInfoInertia = *mpCMouseCtrl;
    }
    if (mpCMouseCtrl->IsGyration()) 
	mbInertia = FALSE;

    bool Trigrefresh = FALSE;

    if (mpCMouseCtrl->IsGyration() || mpCMouseCtrl->IsInertia()){
	mCCameraBase.mSensitivity = 1.3;
	trigRefresh |= mCCameraBase.Gyration
	    (CMP, GetModByAspectX(), GetModByAspectY());
	/*
	if (IsAlongAxis() == TRUE)
	    trigRefresh |= mpCDolly->MoveCenter(GetPCCameraPosition(),
						GetPCAnchorPosition());
	else
	    trigRefresh |= mpCDolly->Gyration  (GetPCCameraPosition());
	*/
	mbViewDragged = TRUE;
    }
    
    if (mpCMouseCtrl->IsZoom()){
	mCCameraBase.mSensitivity = 1;
	trigRefresh |= mCCameraBase.Zoom(CMP);
	mbViewDragged = TRUE;
    }

    if (mpCMouseCtrl->IsMoveCenter()){
	double distance = GetDistance(mCCameraBase.mPos, mCPivot.mPos);

	mCCameraBase.mSensitivity = distance * 0.5;
	mCPivot.mSensitivity = distance * 0.5;

	mCCameraBase.Translate(CMP.GetDX()*GetAspectRatio(), CMP.GetDY());
	mCPivot.Translate(CMP.GetDX()*GetAspectRatio(), CMP.GetDY());
	trigRefresh = TRUE;
	mbViewDragged = TRUE;
    }

    SceneObj_cls *pCActiveObj = GetPCSceneObjCtrl()->mpCActiveObject;
    
    if (mpCMouseCtrl->IsMoveItem() && pCActiveObj != NULL){
	double dragX = CMP.GetDX()*GetAspectRatio();
	double dragY = CMP.GetDY();

	double degreeL = GetViewAngle();
	double degreeV = degreeL;

	double mat16[16], mat16inv[16], mat16Model[16], mat16ModelInv[16];
	memcpy(mat16, mMat16NavigatorModelview, sizeof(double)*16);
	memcpy(mat16Model, mMat16NavigatorModelOnly, sizeof(double)*16);
	GetInvMatrix4x4(mat16, mat16inv);
	GetInvMatrix4x4(mat16Model, mat16ModelInv);
	
	double scale = mSizeOrtho;
	if (mbOrtho == FALSE){
	    double depth = -1.0 * mat16[14] / GetAspectRatio();
	    scale = depth * sin(degreeL/2 * 3.14159 / 180);
	}

	double dVec[3];
	dVec[X] = (mat16inv[0] * dragX + mat16inv[4] * dragY) * scale;
	dVec[Y] = (mat16inv[1] * dragX + mat16inv[5] * dragY) * scale;
	dVec[Z] = (mat16inv[2] * dragX + mat16inv[6] * dragY) * scale;

	if (mpCParent->IsSnapNavi()){
	    double dVecOrg[3];
	    LOOP_XYZ { dVecOrg[xyz] = dVec[xyz]; dVec[xyz] = 0; }
	    
	    double lockVec_[3][3], lockVec[3];
	    mpCParent->GetSnapAxis(lockVec_[X], lockVec_[Y], lockVec_[Z]);

	    LOOP_XYZ {
		lockVec[0] = mat16ModelInv[0]*lockVec_[xyz][0] + mat16ModelInv[4]*lockVec_[xyz][1] + mat16ModelInv[8]*lockVec_[xyz][2];
		lockVec[1] = mat16ModelInv[1]*lockVec_[xyz][0] + mat16ModelInv[5]*lockVec_[xyz][1] + mat16ModelInv[9]*lockVec_[xyz][2];
		lockVec[2] = mat16ModelInv[2]*lockVec_[xyz][0] + mat16ModelInv[6]*lockVec_[xyz][1] + mat16ModelInv[10]*lockVec_[xyz][2];
		
		double norm = dVecOrg[X]*lockVec[X] + dVecOrg[Y]*lockVec[Y] + dVecOrg[Z]*lockVec[Z];
		LOOP_XYZ dVec[xyz] += norm * lockVec[xyz];
	    }
	}


	LOOP_XYZ mDrag[xyz] += dVec[xyz];

	Refresh(); // redraw 𔺂Ȃ
    }

    return trigRefresh;
}

bool CLS::
ProcItemDragEnd(){
    SceneObj_cls *pCActiveObj = GetPCSceneObjCtrl()->mpCActiveObject;
    if (pCActiveObj == NULL) return FALSE;

    if (mpCParent->mNavigatorMode == _NAVI_TRANSLATE){
	GetPCUndo()->OpenBlock("ItemDrag");
	pCActiveObj->SetParamD3Command("Position", mDrag);
	GetPCUndo()->CloseBlock("ItemDrag");
    }

    if (mpCParent->mNavigatorMode == _NAVI_ROTATION){
	if (mScalingInit == 0) mScalingInit = mScalingCurrent;
	double relative = GetRelativeAngleDeg(mbDragSnapping);
	double mDRot[3] = {relative, relative, relative};
	
	if (mpCParent->IsSnapNavi()){
	    mDRot[X] *= mpCParent->mbSnapX;
	    mDRot[Y] *= mpCParent->mbSnapY;
	    mDRot[Z] *= mpCParent->mbSnapZ;
	}
	else{
	    mDRot[X] = 0; mDRot[Y] = 0;
	}
	GetPCUndo()->OpenBlock("ItemRot");
	pCActiveObj->SetParamD3Command("RotationEuler", mDRot);
	GetPCUndo()->CloseBlock("ItemRot");

	mScalingInit = 0; 
    }

    if (mpCParent->mNavigatorMode == _NAVI_SCALE){
	if (mScalingInit == 0) mScalingInit = mScalingCurrent;
	double scaleFactor = mScalingCurrent / mScalingInit - 1;
	double mDScale[3];
	double NT = GetPCTimeCtrl()->GetCurrentNT();
	std::vector < double > VScale = pCActiveObj->GetPCParamGroup("Scale")->GetVInterpolatedValue(NT);
	LOOP_XYZ mDScale[xyz] = VScale[xyz] * scaleFactor;
	if (mpCParent->IsSnapNavi()){
	    mDScale[X] *= mpCParent->mbSnapX;
	    mDScale[Y] *= mpCParent->mbSnapY;
	    mDScale[Z] *= mpCParent->mbSnapZ;
	}
	GetPCUndo()->OpenBlock("ItemScale");
	pCActiveObj->SetParamD3Command("Scale", mDScale);
	GetPCUndo()->CloseBlock("ItemScale");

	mScalingInit = 0;
    }

    LOOP_XYZ mDrag[xyz] = 0;

    return TRUE;
}

void CLS::
ProcViewDragEnd(){
    if (mpCParent->mCbLockToCamera.GetBodyBOOL() == TRUE){
	
	//wxCommandEvent dummy;
	//mpCParent->OnSetPosRot(dummy);
	mbViewDragged = FALSE;
	SendMenuEvent(this, menuView3DSetPosRot);
	mpCParent->mCbFollow.SetBodyFromData((long long)1);
	return;
	
    }
    mbViewDragged = FALSE;
    mpCParent->UpdatePanels("ALL");
}

bool CLS::
ProcInputCommand(){
    SceneObj_cls *pCActiveObj = GetPCSceneObjCtrl()->mpCActiveObject;
    if (pCActiveObj == NULL) return FALSE;

    wxString strCommand = mstrInputCommand.Left(1);
    mstrInputCommand = mstrInputCommand.Mid(1);
    int axis = -1;
    if (strCommand.IsSameAs("G") || strCommand.IsSameAs("S")){
	wxString strAxis = mstrInputCommand.Left(1);
	if (strAxis.IsSameAs("X") || strAxis.IsSameAs("Y") || strAxis.IsSameAs("Z")){
	    mstrInputCommand = mstrInputCommand.Mid(1);
	    if (strAxis.IsSameAs("X")) axis = 0;
	    if (strAxis.IsSameAs("Y")) axis = 1;
	    if (strAxis.IsSameAs("Z")) axis = 2;
	}

	if (mstrInputCommand.IsEmpty() == TRUE) return FALSE;
	
	if (strCommand.IsSameAs("G")){
	    if (axis == -1) axis = 0;
	    float value = atof(mstrInputCommand.ToAscii());
	    double drag[3] = {0, 0, 0};
	    drag[axis] = value;
	    GetPCUndo()->OpenBlock("ItemDrag");
	    pCActiveObj->SetParamD3Command("Position", drag);
	    GetPCUndo()->CloseBlock("ItemDrag");
	}
	if (strCommand.IsSameAs("S")){
	    double NT = GetPCTimeCtrl()->GetCurrentNT();
	    std::vector < double > VScale = pCActiveObj->GetPCParamGroup("Scale")->GetVInterpolatedValue(NT);
	    
	    float value = atof(mstrInputCommand.ToAscii());
	    double scale[3] = {0, 0, 0};
	    if (axis == -1){
		scale[0] = value * VScale[0] - VScale[0];
		scale[1] = value * VScale[1] - VScale[1];
		scale[2] = value * VScale[2] - VScale[2];
	    }
	    else scale[axis] = value * VScale[axis] - VScale[axis];
	    GetPCUndo()->OpenBlock("ItemScale");
	    pCActiveObj->SetParamD3Command("Scale", scale);
	    GetPCUndo()->CloseBlock("ItemScale");
	}
	return TRUE;
    }

    return FALSE;
}

void CLS::
ProcessMouseEvent(wxMouseEvent& event){
    bool trigRefresh = FALSE;
    bool refreshForce = FALSE;
    //static bool sbLockDragging = FALSE; // GL{^DragȂ

    //Navigators_cls::msbQuickInput = FALSE;

    mpCMouseCtrl->SetMouseInformation(event);

    int mouseX = event.GetX();
    int mouseY = event.GetY();

    // j[
    if (mpCMouseCtrl->IsShowMenu() == TRUE){
	//ShowMenu(mpCMouseCtrl->mMouseX, mpCMouseCtrl->mMouseY);
	event.Skip();
	return; }

    bool bHitTest = mpCMouseCtrl->IsHitTest();
    bool bHitTestAdditive = mpCMouseCtrl->IsHitTestAdditive();
    if (bHitTest == TRUE || bHitTestAdditive == TRUE){
	mbHittestQueue = TRUE;
	if (bHitTestAdditive == TRUE) mbHittestQueueAdditive = TRUE;
	event.Skip();
	Refresh();
	return;
    }

    if (event.ButtonDown() == TRUE){
	if (mpCParent->IsActive() == FALSE) mpCParent->SetToActiveGLView();

	

	// L[̓[h̏ꍇ
	if (mstrInputCommand.IsEmpty() == FALSE){
	    if (ProcInputCommand() == TRUE) Refresh();
	    mInputMode = 0;
	    mstrInputCommand.Clear();
	    return;
	}

	if (mbInertia == TRUE && mpCParent->mCbLockToCamera.GetBodyBOOL() == TRUE)
	    SendMenuEvent(this, menuView3DSetPosRot);
	mbInertia = FALSE;
    }

    if (event.ButtonUp() == TRUE){
    }
    
    // ʏ̃{^
    if (mpCMouseCtrl->IsButton(_MOUSE_BTN_LEFT) == TRUE &&
	mpCMouseCtrl->IsButtonDown() == TRUE){
	if (mpCBtnNumPanel->CheckButtonDown(mouseX, mouseY) == TRUE){
	    RefreshEX();
	    return; }
    }
    
    if (mpCMouseCtrl->IsButtonUp() == TRUE){
	if (mpCBtnNumPanel->CheckButtonUp(mouseX, mouseY) == TRUE){
	    SendBtnEvent(mpCParent, btnNumPanelHide);
	    return; }
    }

    if (mpCMouseCtrl->IsButtonUp() == TRUE){
	if (HasCapture() == TRUE) ReleaseMouse();

	if (mDrag[X] != 0 || mDrag[Y] != 0 || mDrag[Z] != 0){
	    trigRefresh |= ProcItemDragEnd();
	}

	if (mbViewDragged == TRUE) ProcViewDragEnd();
    }

    if (mpCMouseCtrl->IsButtonDown() == TRUE){
	if (HasCapture() == FALSE) CaptureMouse();
	mCAutoPivot = mpCParent->GetAutoPivotPos(mCCameraBase);
    }

    // r[܂̓ACẽhbO
    if (mpCMouseCtrl->IsMouseDragging() == TRUE && msbLockDragging == FALSE){
	
	if (mpCParent->mCbFollow.GetBodyBOOL() == TRUE &&
	    mpCParent->mCbLockToCamera.GetBodyBOOL() == FALSE)
	    mpCParent->mCbFollow.SetBodyFromData((long long)0);
		
	wxMouseState mouseState = ::wxGetMouseState();
	bool isLeftDown = mouseState.LeftIsDown();
	bool isMiddleDown = mouseState.MiddleIsDown();
	bool isRightDown = mouseState.RightIsDown();
	
	
	// L[܂Ă܂oỎ
	// 3.0 ŕKvǂ͂܂mF
	if ((isLeftDown | isMiddleDown | isRightDown) == TRUE )
	    trigRefresh |= ProcViewOrItemDragging();
	
    }

    // }EXzC[
    if (mpCMouseCtrl->IsWheel() == TRUE){
	if (IsMouseOnClient() == TRUE){
	    if (mpCParent->IsUseCamera() == FALSE)
		trigRefresh |= mCCameraBase.ZoomWheel(mpCMouseCtrl->mWheelRotationNormalized);
	    
	    /*
	    UpdateAnc();
	    mpCDolly->mDisplayMode = mCDisplayMode.mBody;
	    trigRefresh |= mpCDolly->ZoomWheel
		(GetPCCameraPosition(), mpCMouseCtrl->mWheelRotationNormalized,
		 &mCGLCameraProperty.mCGLOrthoScale.mBody);
	    mpCMouseCtrl->SetPreviousPosition();
	    SetThisToBeActiveGL();
	    */
	}
	else
	    /*
	    trigRefresh |= SetOtherCanvasToBeActiveGL();
	    */
	    ;
    }

    if (trigRefresh == TRUE){
	mpCParent->SetPositionView(mCCameraBase);
	mpCParent->SetDirectionView(mCCameraBase);
	mpCParent->SetPositionPivot(mCPivot);
    }

    if (trigRefresh == TRUE){
	RefreshEX();
	mpCParent->mpCNumPanel->UpdateValues();
    }

    msbLockDragging = FALSE;

    mpCMouseCtrl->SetPreviousPosition();
}

void CLS::
SetCameraInfo(){
    mCCameraBase.SetPos(mpCParent->GetPositionView());
    mCCameraBase.SetDir(mpCParent->GetDirectionView());
    mCPivot.SetPos(mpCParent->GetPositionPivot());
    mCPivot.SetDir(mpCParent->GetDirectionView());

    if (mbViewDragged == TRUE && mpCParent->mCbLockToCamera.GetBodyBOOL() == TRUE){
	mCCameraBase.SetPivotPos(mCAutoPivot);//mpCParent->GetAutoPivotPos(mCCameraBase));
    }
    else
	mCCameraBase.SetPivotPos(mCPivot);

    if (mbViewDragged == FALSE || mpCParent->mCbLockToCamera.GetBodyBOOL() == FALSE)
	mCCameraBase.LookAtPivot();

    mCCameraBaseNext = mCCameraBase;
    if (mpCParent->mCbFollow.GetBodyBOOL() == TRUE && mbInertia == FALSE){

	Camera_cls *pCCam = GetPCCamera();
	if (pCCam != NULL) pCCam->mbSkipDraw = TRUE;

	// Jɏ]킸ɏo͂ꍇ̂
	// o͒̓Jɏ]悤
	if (GetPCTimeCtrl()->mbRecordRender == TRUE ||
	    mpCMouseCtrl->IsMouseDragging() == FALSE)
	    FollowCamera(pCCam);
    }
}

void CLS::
AdjustAngleByScreenAspect(double& degreeV){
    int width = mCWindowSize.GetWidth();
    int height = mCWindowSize.GetHeight();

    if (height > width){
	degreeV = atan((double)height/width * tan(degreeV/2)) * 2;
    }
}

void CLS::
FollowCamera(Camera_cls *pCCam){
    if (pCCam == NULL) return;
    
    //pCCam->mbSkipDraw = TRUE;
    
    double NT = GetPCTimeCtrl()->GetCurrentNT();

    double matrix[16];
    pCCam->LoadIdentity(matrix);
    pCCam->GetMatrix16(matrix, NT);
    
    PositionAndDirection_cls CCam(matrix);
    mCCameraBase.CopyPosition(CCam);

    double NTNext = GetPCTimeCtrl()->GetNextNT();
    
    pCCam->LoadIdentity(matrix);
    pCCam->GetMatrix16(matrix, NTNext);
    
    PositionAndDirection_cls CCamNext(matrix);
    mCCameraBaseNext.CopyPosition(CCamNext);
}

void CLS::
RepairCamera(){
    Camera_cls *pCCam = GetPCCamera();
    if (pCCam == NULL) return;
    
    pCCam->mbSkipDraw = FALSE;
}

void CLS::
DefaultMatrix(){
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void CLS::
SetProjection(){
    GLPerspective_cls::SetProjection(GetSize().GetWidth(), GetSize().GetHeight());
}

/*
void CLS::
SetFrustum(double degreeV){

    int width = GetSize().GetWidth();
    int height = GetSize().GetHeight();

    std::vector < double > VMinMax = GetVMinMax();
    
    int tileMaxX = 1, tileMaxY = 1;
    int tileIndexX = 0, tileIndexY = 0;
    
    GLPerspective_cls::SetFrustumTiled(width, height, degreeV,
				       VMinMax[0], VMinMax[1],
				       tileMaxX, tileMaxY,
				       tileIndexX, tileIndexY);
}
*/

void CLS::
DrawPivot(){
    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
    if (pCSV->IsFirstStep() == FALSE) return;
    if (pCSV->mSVState != 0 && pCSV->mSVState != 3) return;

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    ColorProvider_cls::SetGLColor("GLGrid", TRUE);
    
    glPushMatrix();
    glTranslated(mCPivot.mPos[0], mCPivot.mPos[1], mCPivot.mPos[2]);
    GCClassProvider.GetPCVBOEmpty()->Draw();
    glPopMatrix();
}

void CLS::
DrawGrid(){
  
    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
    if (pCSV->IsFirstStep() == FALSE) return;
    if (pCSV->mSVState != 0 && pCSV->mSVState != 3) return;

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    ColorProvider_cls::SetGLColor("GLGrid", TRUE);
    glBegin(GL_LINES);
    for (int x = -10; x <= 10; x++){
	glVertex3d(x, -10, 0);
	glVertex3d(x, 10, 0);
    }
    
    for (int y = -10; y <= 10; y++){
	glVertex3d(-10, y, 0);
	glVertex3d(10, y, 0);
    }
    glEnd();
}

extern wxString GstrVolumeSnap;

void CLS::
DrawMessage(){

    void* font = GLUT_BITMAP_HELVETICA_12;

    wxString strTime;
    
    int millisec = mCStopWatchSnap.Time();
    mCStopWatchSnap.Start();

    strTime << wxT(":") << millisec << wxT("ms");

    mMillisecondList[mMillisecondIndex] = millisec;
    
    mMillisecondIndex++;
    if (mMillisecondIndex >= 10) mMillisecondIndex = 0;
    int millisec10 = 0;
    for (int i = 0; i < 10; i++) millisec10 += mMillisecondList[i];
    
    float fps = 10.0/millisec10 * 1000;
    strTime << wxT(":") << wxString::Format(wxT("%.2f"), fps) << wxT("fps");

    strTime << wxT("::DebugInfo::") << GstrVolumeSnap;

    MouseNormalized_cls CMP(mpCMouseCtrl, mCWindowSize);
    
    strTime << wxT("::") << CMP.mCurrX << wxT("::") << CMP.mCurrY;

    mpCGLString2D->Draw(font, strTime.ToAscii(), 5, 10, _STRINGS_TOP_LEFT);

    if (mstrInputCommand.IsEmpty() == FALSE)
	mpCGLString2D->Draw(font, mstrInputCommand.ToAscii(), 5, 25, _STRINGS_TOP_LEFT);
    
    mMillisecondPrev = mCStopWatchSnap.Time();
}

void CLS::
DrawDragginMessage(){

    void* font = GLUT_BITMAP_HELVETICA_12;

    if (mpCMouseCtrl->IsMouseDragging() == FALSE) return;
    if (mDrag[X] == 0 && mDrag[Y] == 0 && mDrag[Z] == 0) return;
    
    wxString strMessage;
    if (mpCParent->mNavigatorMode == _NAVI_ROTATION){
	//strMessage << mAngularMod << "+" << mAngularCurrent;
	strMessage << GetRelativeAngleDeg() <<  "deg";
    }
    
    if (mpCParent->mNavigatorMode == _NAVI_SCALE){
	strMessage << mScalingCurrent << ":" << mScalingInit;
    }

    mpCGLString2D->Draw(font, strMessage.ToAscii(), 5, 10, _STRINGS_BOTTOM_LEFT);
}

void CLS::
DrawProtoProcess(FBO_cls *pCFBORenderTgt, int typeWideCamera){

    pCFBORenderTgt->Begin(FALSE);

    double degreeL = GetViewAngle();
    double degreeV = degreeL;
    if (typeWideCamera <= _WC_ORTHOGONAL && 
	(mpCParent->mCbFollow.GetBodyBOOL() == TRUE ||
	 mpCParent->mCbLockToCamera.GetBodyBOOL() == TRUE)){

	if (GetPCTimeCtrl()->mbRecordRender == FALSE){
	    Camera_cls *pCCam = GetPCCamera();
	    if (pCCam != NULL){
		if (typeWideCamera == _WC_PERSPECTIVE)
		    pCCam->GLDrawPassepartout(degreeV);
		if (typeWideCamera == _WC_ORTHOGONAL)
		    pCCam->GLDrawPassepartoutOrtho(mSizeOrtho);
	    }
	}
    }

    pCFBORenderTgt->End();
}

std::vector < float > CLS::
GetVPasspartoutInfo(){
    std::vector < float > VPasspartoutInfo = {-1, -1};

    double degreeL = GetViewAngle();
    double degreeV = degreeL;

    Camera_cls *pCCamera = GetPCCamera();
    if (/*mpCParent->mCbFollow.GetBodyBOOL() == TRUE &&*/
	pCCamera != NULL)
	VPasspartoutInfo = pCCamera->GetVPasspartoutInfo();

    float angleCanvas = tan(degreeV*3.14159/180 / 2);
    VPasspartoutInfo.push_back(angleCanvas);
    
    return VPasspartoutInfo;
}

void CLS::
SetModelViewForEachFBO(int eyeType, int stereo, int fbo, int ensemble,
		       int typeWideCamera){
    mCCameraInUse = mCCameraBase;
    mCCameraInUseNext = mCCameraBaseNext;

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    double NTNext = GetPCTimeCtrl()->GetNextNT();

    Camera_cls *pCCamera = GetPCCamera();
    if (pCCamera != NULL && eyeType != _STEREO_NONE){
	pCCamera->mpCStereoProp->StereoRotate(mCCameraInUse, stereo, NT);
	pCCamera->mpCStereoProp->StereoRotate(mCCameraInUseNext, stereo, NTNext);
    }

    if (pCCamera != NULL && mpCParent->IsUseCamera() == TRUE){
	pCCamera->mpCFocalBlurProp->BlurRotate(mCCameraInUse, ensemble, NT);
	pCCamera->mpCFocalBlurProp->BlurRotate(mCCameraInUseNext, ensemble, NTNext);
    }
    
    double tilt = GetDomeTilt();
    GetPCWideCamera()->ModifyCamera(&mCCameraInUse, fbo, typeWideCamera, tilt);
    GetPCWideCamera()->ModifyCamera(&mCCameraInUseNext, fbo, typeWideCamera, tilt);

    GLPerspective_cls::SetModelView(mCCameraInUse, mCCameraInUse);
}

void CLS::
DrawEachFBO(int ensemble, int composit, 
	    float baseMat16[16], float baseMat16Inv[16],
	    float viewMat16[16], float viewMat16Inv[16],
	    bool bHitTest){

    glEnable(GL_DEPTH_TEST);
    
    glLineWidth(mpCParent->mCLineWidth.GetBodyINT());

    if (bHitTest == FALSE &&
	mpCParent->mCbDisplayPivot.GetBodyBOOL() == TRUE) DrawPivot();

    if (bHitTest == FALSE &&
	mpCParent->mCbDisplayGrid.GetBodyBOOL() == TRUE) DrawGrid();
    
    bool bDrawGuideObject = mpCParent->mCbDisplayGuideObjects.GetBodyBOOL();
    bool bDrawLightCam = mpCParent->mCbDisplayLightCam.GetBodyBOOL();
    if (GetPCShadowVolumeCtrl()->IsFirstStep() == FALSE){
	bDrawGuideObject = FALSE;
	bDrawLightCam = FALSE;
    }
    bool bDrawData = mpCParent->mCbDisplayData.GetBodyBOOL();

    glDisable(GL_FOG);
    GetPCShaderCtrl()->mFogMode = 0;
    
    glDisable(GL_TEXTURE_2D);
    
    int lightMax = GetPCGLInfo()->mMaxLightNum;
    glEnable(GL_LIGHTING);
    for (int light = GL_LIGHT0; light < GL_LIGHT0+lightMax; light++)
	glDisable(light);
    int lightCurrent = GL_LIGHT0;
    glEnable(GL_LIGHT0); // ftHg
    
    glLineWidth(mpCParent->mCLineWidth.GetBodyINT() * 2);
    GetPCSceneObjCtrl()->GLDrawObjects(bDrawGuideObject, bDrawLightCam,
				       bDrawData,
				       lightCurrent, ensemble, composit,
				       mCCameraInUse, mCCameraInUseNext,
				       baseMat16, baseMat16Inv,
				       viewMat16, viewMat16Inv,
				       mpCHitTest, bHitTest);
    
    CheckGLError("END OBJECT DRAW");
}

FBO_cls* CLS::
GetPCNewFBOForRecord(int bufferSize){
    FBO_cls *pCFBOForRecord = NULL;
    bool bRecord = GetPCTimeCtrl()->mbRecordRender & mpCParent->IsActive();
    
    if (bRecord == TRUE){
	pCFBOForRecord = new FBO_cls(bufferSize, bufferSize, this);
	pCFBOForRecord->MakeFBO(TRUE);
	GetPCTimeCtrl()->mbRecordConfirmed = TRUE;
    }
    return pCFBOForRecord;
}

void CLS::
BeginFBOAndSetStatus(FBO_cls *pCFBO, wxString strClearColor, double degreeV,
		     double min, double max, bool bUseDepthBuffer){
    pCFBO->MakeFBO(_FBO_USE_RGBA, bUseDepthBuffer);
    if (strClearColor.IsEmpty() == FALSE){
	pCFBO->Begin(TRUE);
	ClearBuffer(strClearColor);
    }
    else
	pCFBO->Begin(FALSE);
    /*
      pCFBO->Begin(bClear);
      if (bClear == TRUE) ClearBuffer("GLBG");
    */
    DefaultMatrix();
    glEnable(GL_NORMALIZE);
    GetPCGLExtentions()->ClampColor(GL_FALSE);

    if (mpCFBOCurrentLink != NULL){
	mpCFBOCurrentLink->SetProjection();
	if (mbOrtho == TRUE)
	    mpCFBOCurrentLink->SetOrtho(mSizeOrtho, min, max);
	else
	    mpCFBOCurrentLink->SetFrustum(degreeV, min, max);
    }
    else{
	pCFBO->SetProjection();
	if (mbOrtho == TRUE)
	    pCFBO->SetOrtho(mSizeOrtho, min, max);
	else
	    pCFBO->SetFrustum(degreeV, min, max);
    }
}

void CLS::
DummyDraw(){
    static float a = 0;
    glClearColor(a, 0.5, 0.2, 0);
    glClear(GL_COLOR_BUFFER_BIT);
    glFlush();
    SwapBuffers();

    a += 0.1;
    if (a > 1) a = 0;
}

void CLS::
Paint(){
    SetGLContext();
    wxPaintDC(this);
    InitializeGLCommonIfNot();
    InitializeGLIfNot();
    ClearBuffer("MaskBG");

    PaintMain();

    GetPCFrameMain()->SetGauge(100, TRUE);
    if (GetPCTimeCtrl()->msbForceAbort == TRUE) 
	GetPCFrameMain()->SetStatusAborted();
}

void CLS::
PaintMain(){

    bool bUseDepthBuffer = DoesUseDepthBuffer();

    int bufferSize = mpCParent->GetBufferSize();
    mpCFBORenderOut->SetGeometryFBO(bufferSize, bufferSize);
    mpCFBORenderOut->MakeFBO(TRUE);
    if (bUseDepthBuffer == TRUE){
	mpCFBODepth->SetGeometryFBO(bufferSize, bufferSize);
	mpCFBODepth->SetColorDepth(8); // ߖ
	mpCFBODepth->MakeFBO(TRUE, TRUE);
    }

    double degreeL = GetViewAngle();
    std::vector < double > VMinMax = GetVMinMax();
    
    int width, height, clientWidth, clientHeight;
    GetWindowSize(&width, &height, &clientWidth, &clientHeight);
    mpCGLString2D->SetWindowGeometry(clientWidth, clientHeight);

    bool bUseEnsemble, bUseSV, bUseBlur;
    int ensembleSampleNum, SVIni, SVSampleNum, blurSampleNum;
    GetSampling(&bUseEnsemble, &bUseSV, &bUseBlur,
		&ensembleSampleNum, &SVIni, &SVSampleNum, &blurSampleNum);
    World_cls::msSVSampleNum = SVSampleNum;

    if (bUseEnsemble == FALSE) GetPCEnsembleCtrl()->ClearFBOBuffer();
    if (bUseBlur == FALSE) GetPCBlurCtrl()->ClearFBOBuffer();

    
    int numStereo = GetStereoNum();
    int eyeType = GetEyeType();
    int stereoOutType = GetPCImageOutput()->mCStereoOutput.GetBodyINT();
    int numOutput = 1;
    if (numStereo == 2 && stereoOutType == 1)
	numOutput = 2;
    if (numStereo == 2 && eyeType == _STEREO_SCANLINE)
	numOutput = 2;

    int numImageComposit = GetImageCompositeNum();
    if (numImageComposit == 2){
	mpCFBOImageCompositBuffer->SetGeometryFBO(bufferSize, bufferSize);
	mpCFBOImageCompositBuffer->MakeFBO(TRUE);
    }
    
    std::vector < float > VPasspartoutInfo = GetVPasspartoutInfo();
    int typeWideCamera = GetCameraType();
    mbOrtho = (typeWideCamera == _WC_ORTHOGONAL) ? TRUE : FALSE;
    mSizeOrtho = GetSizeOrtho();
    if (mbOrtho == TRUE)
	VMinMax[0] = -VMinMax[1];
    
    FBO_cls *pCFBOForRecord[2] = {NULL, NULL};
    for (int output = 0; output < numOutput; output++)
	pCFBOForRecord[output] = GetPCNewFBOForRecord(bufferSize);
        
    WideCamera_cls *pCWC = GetPCWideCamera();
    double degreeV = pCWC->GetDegreeV(typeWideCamera, degreeL);
    int fboNum = pCWC->GetFBONum(typeWideCamera);
    int bufferSize1st = mpCParent->GetBufferSize1st();
    pCWC->SetRenderPartParam(bufferSize1st, 16, fboNum, bUseEnsemble);
    if (numStereo == 2){
	mpCFBOStereoBuffer[0]->SetGeometryFBO(bufferSize, bufferSize);
	mpCFBOStereoBuffer[1]->SetGeometryFBO(bufferSize, bufferSize);
	mpCFBOStereoBuffer[0]->MakeFBO(TRUE);
	mpCFBOStereoBuffer[1]->MakeFBO(TRUE);
	if (bUseDepthBuffer ==TRUE){
	    mpCFBOStereoDepth[0]->SetGeometryFBO(bufferSize, bufferSize);
	    mpCFBOStereoDepth[1]->SetGeometryFBO(bufferSize, bufferSize);
	    mpCFBOStereoDepth[0]->SetColorDepth(8); // ߖ
	    mpCFBOStereoDepth[1]->SetColorDepth(8); // ߖ
	    mpCFBOStereoDepth[0]->MakeFBO(TRUE, TRUE);
	    mpCFBOStereoDepth[1]->MakeFBO(TRUE, TRUE);
	}
    }
    
    int ZdepNum = 1;
    if (bUseDepthBuffer == TRUE)
	ZdepNum = 2;

    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();

    for (int stereo = 0; stereo < numStereo; stereo++){

	FBO_cls *pCFBORenderTgt = mpCFBORenderOut;
	if (numStereo == 2)
	    pCFBORenderTgt = mpCFBOStereoBuffer[stereo];

	if (mbRedrawQueue == FALSE)
	    goto SKIP_REDRAW_TARGET;

	

	for (int composit = 0; composit < numImageComposit; composit++){
	    for (int blurStep = 0; blurStep < blurSampleNum; blurStep++){
		
		if (bUseBlur == TRUE && blurStep == 0)
		    GetPCBlurCtrl()->PrepareBuffer(pCFBORenderTgt);

		GetPCBlurCtrl()->SetBlurStep(blurStep);
		
		float baseMat16[16], baseMat16Inv[16];
		SetCameraInfo();
		GLCamera_cls CCamTmp = mCCameraBase;
		CCamTmp.NormalizeAndOrthogonalize();
		//CCamTmp.RotatePitch(-60);

		glPushMatrix();
		glLoadIdentity();
		GLPerspective_cls::SetModelView(CCamTmp, CCamTmp);

		glGetFloatv(GL_MODELVIEW_MATRIX, baseMat16);
		TensorMath_cls::GetInvMatrix4x4(baseMat16, baseMat16Inv);
		glPopMatrix();

		wxStopWatch CSW;
		CSW.Start();
		wxString strLog;

		for (int fbo = 0; fbo < fboNum; fbo++){

		    for (int ZdepIdex = 0; ZdepIdex < ZdepNum; ZdepIdex++){

			FBO_cls *pCFBOToDraw = 
			    pCWC->GetPCFBO(typeWideCamera, fbo, pCFBORenderTgt, bUseEnsemble);

			bool _bUseEnsemble = bUseEnsemble;
			int _ensembleSampleNum = ensembleSampleNum;
			bool _bUseSV = bUseSV;
			int _SVNum = SVSampleNum;
			int _SVIni = SVIni;
			if (ZdepIdex == 1){
			    pCFBOToDraw = mpCFBODepth;
			    _bUseSV = FALSE;
			    _SVNum = 1;
			    _SVIni = 0;
			    _bUseEnsemble = FALSE;
			    _ensembleSampleNum = 1;
			}
					    
			for (int SVIdx = SVIni; SVIdx < SVSampleNum; SVIdx++){
			    
			    if (_bUseSV == TRUE && SVIdx == _SVIni)
				pCSV->PrepareBuffer(pCFBOToDraw);
			    pCSV->SetShadowVolumeStep(SVIdx, _SVIni);
			    //pCSV->SetEnsembleSampleNum(ensembleSampleNum);
			    
			    for (int ensemble = 0; ensemble < _ensembleSampleNum; ensemble++){
				if (_bUseEnsemble == TRUE && ensemble == 0)
				    GetPCEnsembleCtrl()->PrepareBuffer(pCFBOToDraw);
				GetPCEnsembleCtrl()->SetEnsembleStep(ensemble);
				
				wxString strBGClr = "Black";
				if (SVIdx == SVIni) strBGClr = "GLBG";
				
				BeginFBOAndSetStatus(pCFBOToDraw, strBGClr, degreeV,
						     VMinMax[0], VMinMax[1], ZdepIdex == 1);
				pCWC->ModifyViewportIfNeeded(typeWideCamera, fbo);
				
				SetModelViewForEachFBO(eyeType, stereo, fbo, ensemble, typeWideCamera);
				float viewMat16[16];
				glGetFloatv(GL_MODELVIEW_MATRIX, viewMat16);
				float viewMat16Inv[16];
				TensorMath_cls::GetInvMatrix4x4(viewMat16, viewMat16Inv);
				
				GetPCShaderCtrl()->PrepareViewportInfo();
				GetPCImageIO()->PrepareViewportInfo();
				
				if (GetPCTimeCtrl()->msbForceAbort == FALSE){
				    
				    strLog << "fbo" << fbo << " Start " << CSW.Time() << " ms" << ENDL;
				    DrawEachFBO(ensemble, composit,
						baseMat16, baseMat16Inv,
						viewMat16, viewMat16Inv);
				    strLog << "fbo" << fbo << " END " << CSW.Time() << " ms" << ENDL;
				    
				    if (SVIdx != 0){
					pCSV->PushState();
					
					// SetStencil
					pCSV->SetStencilState();
					//pCSV->SettingFront();
					DrawEachFBO(ensemble, composit, 
						    baseMat16, baseMat16Inv, 
						    viewMat16, viewMat16Inv,
						    FALSE);
					
					//pCSV->SettingBack();
					// Shadowed Area
					pCSV->SetShadowedArea();
					DrawEachFBO(ensemble, composit,
						    baseMat16, baseMat16Inv,
						    viewMat16, viewMat16Inv,
						    FALSE);
					pCSV->PopState();
				    }
				}
				
				pCFBOToDraw->End();
				
				if (_bUseEnsemble == TRUE)
				    GetPCEnsembleCtrl()->DrawOnBuffer(pCFBOToDraw, ensembleSampleNum);
				
				if (SVSampleNum != 1 || fboNum != 1 ||
				    ensembleSampleNum != 1){
				    int percent = GetPercent(fbo, fboNum, SVIdx, SVSampleNum, ensemble, ensembleSampleNum);
				    GetPCFrameMain(0)->SetGauge(percent);
				}
			    }
			
			
			    if (_bUseEnsemble == TRUE){
				if (GetPCTimeCtrl()->msbForceAbort == FALSE)
				    GetPCEnsembleCtrl()->DrawFromBuffer(pCFBOToDraw);
				GetPCEnsembleCtrl()->SetEnsembleStep(0);
			    }
			    
			    if (_bUseSV == TRUE){
				if (GetPCTimeCtrl()->msbForceAbort == FALSE)
				    pCSV->DrawOnBuffer(pCFBOToDraw);
			    }
			}
			

			if (_bUseSV == TRUE){
			    if (GetPCTimeCtrl()->msbForceAbort == FALSE)
				pCSV->DrawFromBuffer(pCFBOToDraw);
			    pCSV->SetShadowVolumeStep(0, 0);
			}
		    }
		}

		//wxLogMessage(strLog);
		
		if (fboNum != 1)
		    pCWC->GatherPart(pCFBORenderTgt, typeWideCamera, degreeL);
		
		DrawProtoProcess(pCFBORenderTgt, typeWideCamera);
		
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_BLEND);
		glDisable(GL_LIGHTING);
		glColor4f(1, 1, 1, 1);
		
		if (bUseBlur == TRUE)
		    GetPCBlurCtrl()->DrawOnBuffer(pCFBORenderTgt, blurSampleNum);
		
	    }
	    
	    if (bUseBlur == TRUE){
		GetPCBlurCtrl()->DrawFromBuffer(pCFBORenderTgt);
		GetPCBlurCtrl()->SetBlurStep(0);
	    }
	    
	    if (numImageComposit == 2){
		if (composit == 0){
		    GetPCImageComposer()->StoreInBuffer
			(mpCFBOImageCompositBuffer, pCFBORenderTgt);
		}
		if (composit == 1){
		    GetPCImageComposer()->OverlapBuffer
			(pCFBORenderTgt, mpCFBOImageCompositBuffer);
		}
	    }
	}
	
    SKIP_REDRAW_TARGET:
	if (numStereo == 2){
	    mpCFBOStereoFiltered[stereo] = 
		GetPCFilterCtrl()->GetNewPCFilterProcessed(mpCFBOStereoBuffer[stereo], VPasspartoutInfo, mCCameraBase, GetPCCamera(), mpCFBODepth);
	}
	
    }

    if (numStereo == 2){
	mpCFBOFinalOut = new FBO_cls(bufferSize, bufferSize, this);
	mpCFBOFinalOut->MakeFBO(TRUE);
	GetPCStereogram()->MergeStereogram(mpCFBOStereoFiltered, mpCFBOFinalOut, eyeType);
    }
    else{
	mpCFBOFinalOut = 
	    GetPCFilterCtrl()->GetNewPCFilterProcessed(mpCFBORenderOut, VPasspartoutInfo, mCCameraBase, GetPCCamera(), mpCFBODepth);
    }
    
    int fit = pCWC->GetFitIndex(typeWideCamera, GetAspectRatio());
    
    // ۑpFBO 쐬
    if (pCFBOForRecord[0] != NULL){
	PrepareFBOForRecord(pCFBOForRecord, fit, numOutput);
    }
    
    SceneObj_cls *pCActiveObj = GetPCSceneObjCtrl()->mpCActiveObject;
    if (numStereo == 2){
	for (int stereo = 0; stereo < numStereo; stereo++)
	    DrawNavigator(mpCFBOStereoFiltered[stereo], pCActiveObj, degreeV, VMinMax[0], VMinMax[1]);
	GetPCStereogram()->MergeStereogram(mpCFBOStereoFiltered, mpCFBOFinalOut, eyeType);
    }
    else
	DrawNavigator(mpCFBOFinalOut, pCActiveObj, degreeV, VMinMax[0], VMinMax[1]);
        
    // \pFBOʂɕ\
    if (numStereo == 2 && eyeType == _STEREO_SCANLINE){
	GetPCStereogram()->DrawScanLineOnScreen(mpCFBOStereoFiltered, clientWidth, clientHeight, fit, (clientHeight - height) * 0.5);
    }
    else
	mpCFBOFinalOut->DrawOnScreen2DFit(clientWidth, clientHeight, fit, FALSE, (clientHeight - height) * 0.5);
    
    if (numStereo == 2){
	delete mpCFBOStereoFiltered[0]; mpCFBOStereoFiltered[0] = NULL;
	delete mpCFBOStereoFiltered[1]; mpCFBOStereoFiltered[1] = NULL;
    }
    
    delete mpCFBOFinalOut;
    mpCFBOFinalOut = NULL;

    CheckGLError("PAINT END");

    bool bLoupe = mpCParent->mCbDisplayLoupe.GetBodyBOOL();
    bool bHitTest = mbHittestQueue;
    
    GLCamera_cls CCamTmp = mCCameraBase;
    CCamTmp.NormalizeAndOrthogonalize();

    mpCNavigator->MakeImage(CCamTmp);
    if (bLoupe == TRUE || bHitTest == TRUE){
	MouseNormalized_cls CMP(mpCMouseCtrl, mCWindowSize);
	mpCHitTest->HitTestBegin(mCCameraBase, degreeV, 
				 0.5 + CMP.mCurrX * 0.5 * GetModByAspectX(),
				 0.5 + CMP.mCurrY * 0.5 * GetModByAspectY(),
				 bHitTest);
	// HitTest  baseMat16͕KvȂ͂
	DrawEachFBO(0, 0, NULL, NULL, NULL, NULL, bHitTest); 
	SceneObj_cls *pCHit = mpCHitTest->HitTestEnd(bHitTest);
	if (pCHit != NULL){
	    GetPCSceneObjCtrl()->Select
		(pCHit->GetStrName(), mbHittestQueueAdditive, TRUE, TRUE);
	}
    }

    GLPerspective_cls::Push2DMatrix(width, height);
    if (mpCParent->mCbNumPanelShown.GetBodyBOOL() == FALSE){
	mpCBtnNumPanel->mCSizeCanvas = GetClientRect().GetSize();
	mpCBtnNumPanel->DrawButton(FALSE);
    }
    mpCNavigator->Draw(GetClientRect().GetSize(), wxSize(48, 48));
    if (bLoupe == TRUE)
	mpCHitTest->Draw(GetClientRect().GetSize(), wxSize(128, 128));
    GLPerspective_cls::Pop2DMatrix();
    
    CheckGLError("DRAW ON SCREEN");

    DrawMessage();
    DrawDragginMessage();

    glFlush();
    SwapBuffers();
    ClearQueue();

    ProcSaveImage(numOutput, pCFBOForRecord, eyeType, stereoOutType);
}

void CLS::
ProcSaveImage(int numOutput, FBO_cls *pCFBOForRecord[2],
	      int eyeType, int stereoOutType){
    if (numOutput == 1){
	if (pCFBOForRecord[0] != NULL){
	    SaveImageFromFBO(pCFBOForRecord[0], 0);
	    delete pCFBOForRecord[0];
	}
    }
    if (numOutput == 2){
	if (eyeType == _STEREO_SCANLINE && stereoOutType == 0){
	    if (pCFBOForRecord[0] != NULL){
		SaveImageFromFBOScanLine(pCFBOForRecord);
		delete pCFBOForRecord[0];
		delete pCFBOForRecord[1];
	    }
		
	}
	else{
	    if (pCFBOForRecord[0] != NULL){
		SaveImageFromFBO(pCFBOForRecord[0], 1);
		delete pCFBOForRecord[0];
	    }
	    if (pCFBOForRecord[1] != NULL){
		SaveImageFromFBO(pCFBOForRecord[1], 2);
		delete pCFBOForRecord[1];
	    }
	}
    }

    if (pCFBOForRecord[0] != NULL && GetPCTimeCtrl()->mbRecordSnap == TRUE)
	GetPCTimeCtrl()->StopPlayCommand();
}

void CLS::
DrawGuideLineForRotScale(FBO_cls *pCFBORenderTgt, SceneObj_cls *pCActiveObj){
    
    if (mpCMouseCtrl->IsMouseDragging() == FALSE) return;
    if (mDrag[X] == 0 && mDrag[Y] == 0 && mDrag[Z] == 0) return;
    
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    
    std::vector < double > VPosObjGlobal = pCActiveObj->GetPositionGlobal(NT);
    double winPos[3];
    GLPerspective_cls::GetWindowCoordv(VPosObjGlobal.data(), winPos);
    
    int width, height, clientWidth, clientHeight;
    GetWindowSize(&width, &height, &clientWidth, &clientHeight);
    int fit = GetPCWideCamera()->GetFitIndex(0, GetAspectRatio());
    
    MouseNormalized_cls CMP(mpCMouseCtrl, mCWindowSize);
    double texX, texY;
    pCFBORenderTgt->ConvertScrCoordToTexCoord
	(clientWidth, clientHeight, fit, (clientHeight - height) * 0.5,
	 (CMP.mCurrX + 1)/2, (CMP.mCurrY + 1)/2, texX, texY);
    
    pCFBORenderTgt->Push2DMatrix();
    glEnable(GL_LINE_STIPPLE);
    glColor4f(1, 1, 1, 1);
    glLineStipple(2, 0x9999);
    glBegin(GL_LINES);
    glVertex2d(texX, texY);
    glVertex2d(winPos[X], winPos[Y]);
    glEnd();
    glColor4f(0, 0, 0, 1);
    glLineStipple(2, 0x3333);
    glBegin(GL_LINES);
    glVertex2d(texX, texY);
    glVertex2d(winPos[X], winPos[Y]);
    glEnd();
    glDisable(GL_LINE_STIPPLE);
    pCFBORenderTgt->Pop2DMatrix();
    
    mScalingCurrent = sqrt((winPos[X]-texX)*(winPos[X]-texX)+
			   (winPos[Y]-texY)*(winPos[Y]-texY));
    mAngularCurrent = atan2(-(winPos[Y]-texY), -(winPos[X]-texX));
    
    if (mScalingInit == 0){
	mScalingInit = mScalingCurrent;
	mAngularInit = mAngularCurrent;
	mAngularMod = 0;
	mAngularPre = mAngularCurrent;
    }

    if (mAngularPre < 0 && mAngularCurrent > 0){
	if (mAngularCurrent > 3.14159 / 2 || mAngularCurrent < -3.14159 / 2)
	    mAngularMod ++;
    }
    if (mAngularPre > 0 && mAngularCurrent < 0){
	if (mAngularCurrent > 3.14159 / 2 || mAngularCurrent < -3.14159 / 2)
	    mAngularMod --;
    }
    mAngularPre = mAngularCurrent;
}

void CLS::
DrawNavigator(FBO_cls *pCFBORenderTgt,
	      SceneObj_cls *pCActiveObj, double degreeV, double min, double max){
    if (pCActiveObj == NULL) return;
    
    if (mpCParent->mCbFollow.GetBodyBOOL() == TRUE &&
	pCActiveObj == GetPCCamera()) return; 
    
    BeginFBOAndSetStatus(pCFBORenderTgt, wxES, degreeV, min, max, FALSE);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    SetModelViewForEachFBO(0, 0, 0, 0, 0);

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    if (pCActiveObj != NULL &&
	mpCParent->mCbDisplayNavigator.GetBodyBOOL() == TRUE){

	glLineWidth(mpCParent->mCLineWidth.GetBodyINT());
	
	bool bGlobal = mpCParent->mCNavigatorCoord.GetBodyINT() == 1;
	bool bParent = mpCParent->mCNavigatorCoord.GetBodyINT() == 2;

	if (mpCParent->mNavigatorMode == _NAVI_SCALE ||
	    mpCParent->mNavigatorMode == _NAVI_ROTATION){ 
	    bGlobal = FALSE;
	    bParent = FALSE;

	    DrawGuideLineForRotScale(pCFBORenderTgt, pCActiveObj);
	}
	
	double matrix[16], matrixParent[16];
	
	// e̍WňvZĕۑ
	LoadIdentity(matrix);
	LoadIdentity(matrixParent);
	if (pCActiveObj->mpCParentObj->mParentObjType != _PARENTOBJ_ROOT){
	    SceneObj_cls *pCParent = (SceneObj_cls*)pCActiveObj->mpCParentObj;
	    pCParent->GetMatrix16(matrixParent, NT);
	}
	
	pCActiveObj->ApplyTranslate(matrix, NT);
	Multiply(matrixParent, matrix);
	
	if (bGlobal == TRUE) glPushMatrix();
	glMultMatrixd(matrix);
	glGetDoublev(GL_MODELVIEW_MATRIX, mMat16NavigatorModelview);
	if (bGlobal == TRUE) glPopMatrix();

	LoadIdentity(mMat16NavigatorModelOnly);
	if (bGlobal == TRUE)
	    for (int i = 0; i < 16; i++) mMat16NavigatorModelOnly[i] = matrix[i];
	
	if (mpCParent->mNavigatorMode == _NAVI_TRANSLATE){ 
	    // \́AGlobal̏ꍇʒu񂾂ł
	    if (bGlobal == TRUE){
		LoadIdentity(matrix);
		std::vector < double > VPosGlobal = pCActiveObj->GetPositionGlobal(NT, mDrag);
		glTranslated(VPosGlobal[0], VPosGlobal[1], VPosGlobal[2]);
	    }
	    else{
		glTranslated(mDrag[X], mDrag[Y], mDrag[Z]);
	    }
	}

	if (mpCParent->mNavigatorMode != _NAVI_ROTATION){
	    if (bGlobal == FALSE && bParent == FALSE){
		double matrix[16], matrixinv[16];
		LoadIdentity(matrix);
		pCActiveObj->ApplyRotate(matrix, NT);
		glMultMatrixd(matrix);
		GetInvMatrix4x4(matrix, matrixinv);
		Multiply(matrixinv, mMat16NavigatorModelOnly);
	    }
	}

	GLfloat matrixA[16], matrixB[16];
	double scale = GLPerspective_cls::GetScaleForFixSizedObject
	    (matrixA, matrixB);

	glScaled(scale*0.05, scale*0.05, scale*0.05);
	
	bool bSnapX = mpCParent->mbSnapX, bSnapY = mpCParent->mbSnapY, bSnapZ = mpCParent->mbSnapZ;
	glLineWidth(mpCParent->mCLineWidth.GetBodyINT()*2);
	switch(mpCParent->mNavigatorMode){
	case 0:
	    mpCNavigator->ShowNaviTranslate(bSnapX, bSnapY, bSnapZ); break;
	case 1:
	    mpCNavigator->ShowNaviRotate(bSnapX, bSnapY, bSnapZ); break;
	case 2:
	    mpCNavigator->ShowNaviScale(bSnapX, bSnapY, bSnapZ); break;
	}
    }
    pCFBORenderTgt->End();
}

void CLS::
PrepareFBOForRecord(FBO_cls *pCFBOForRecord[2], int fit, int numOutput){

    int recordBufferWidth = pCFBOForRecord[0]->mWidthFBO;
    int recordBufferHeight = pCFBOForRecord[0]->mHeightFBO;

    if (numOutput == 2){
	for (int stereo = 0; stereo < numOutput; stereo++){
	    pCFBOForRecord[stereo]->Begin();
	    ClearBuffer("GLBG");
	    glEnable(GL_BLEND);
	    GetPCShaderCtrl()->BeginShader("ClampAlpha", NULL);
	    mpCFBOStereoFiltered[stereo]->DrawOnScreen2DFit
		(recordBufferWidth, recordBufferHeight, fit, FALSE, 0);
	    GetPCShaderCtrl()->EndShader();
	    glDisable(GL_BLEND);
	    pCFBOForRecord[stereo]->End();
	}
    }
    else{
	pCFBOForRecord[0]->Begin();
	ClearBuffer("GLBG");
	glEnable(GL_BLEND);
	GetPCShaderCtrl()->BeginShader("ClampAlpha", NULL);
	mpCFBOFinalOut->DrawOnScreen2DFit
	    (recordBufferWidth, recordBufferHeight, fit, FALSE, 0);
	GetPCShaderCtrl()->EndShader();
	glDisable(GL_BLEND);
	pCFBOForRecord[0]->End();
    }
}

wxSize CLS::
GetOutputImageSize(){
    Camera_cls *pCCam = GetPCCamera();
    if (pCCam != NULL){
	if (pCCam->mpCCameraProp->mCbOverwriteGeometry.GetBodyBOOL() == TRUE)
	    return pCCam->GetSize();
    }
    return GetPCImageOutput()->GetSize();
}

void CLS::
SaveImageFromFBO(FBO_cls *pCFBO, int stereo){
    int step = GetPCTimeCtrl()->GetCurrentStep();
    wxString strFilename = GetPCImageOutput()->GetModifiedStr(step, stereo);
    wxSize CSizeImage = GetOutputImageSize();
    bool bAlpha = GetPCImageOutput()->mCImageAlpha.GetBodyBOOL();
    GetPCImageIO()->SaveImageFromFBO(pCFBO, CSizeImage, strFilename, bAlpha);
}

void CLS::
SaveImageFromFBOScanLine(FBO_cls *pCFBO[2]){
    int step = GetPCTimeCtrl()->GetCurrentStep();
    wxString strFilename = GetPCImageOutput()->GetModifiedStr(step, 0);
    wxSize CSizeImage = GetOutputImageSize();
    GetPCImageIO()->SaveImageFromFBOScanLine(pCFBO, CSizeImage, strFilename);
}

void CLS::
OnPaint(wxPaintEvent &event){

    if (IsShown() == FALSE) return;
    
    if (mCWindowSize.GetWidth() == 0) UpdateSize(); 

    if (mbStatusDrawing == TRUE){
	wxLogMessage("Drawing"); wxLogMessage(ENDL);
	return;
    }

    mbStatusDrawing = TRUE;
    wxRect CRect = GetClientRect();

    Paint();

    if (mpCParent->mCbFollow.GetBodyBOOL() == TRUE && mbInertia == FALSE)
	RepairCamera();

    wxString strPosPrev = mpCParent->mCViewPos.GetBodySTR();
    wxString strDirPrevTgt = mpCParent->mCViewDirTgt.GetBodySTR();
    wxString strDirPrevTop = mpCParent->mCViewDirTop.GetBodySTR();
    wxString strDirPrevRgt = mpCParent->mCViewDirRgt.GetBodySTR();
    wxString strDirPrevEuler = mpCParent->mCViewEuler.GetBodySTR();
    
    //mpCParent->SetPositionView(mCCameraBase);
    //mpCParent->SetDirectionView(mCCameraBase);
    //mpCParent->SetPositionPivot(mCPivot);

    wxString strPosNew = mpCParent->mCViewPos.GetBodySTR();
    wxString strDirNewTgt = mpCParent->mCViewDirTgt.GetBodySTR();
    wxString strDirNewTop = mpCParent->mCViewDirTop.GetBodySTR();
    wxString strDirNewRgt = mpCParent->mCViewDirRgt.GetBodySTR();
    wxString strDirNewEuler = mpCParent->mCViewEuler.GetBodySTR();

    bool bSame = TRUE;
    bSame &= strPosPrev.IsSameAs(strPosNew);
    bSame &= strDirPrevTgt.IsSameAs(strDirNewTgt);
    bSame &= strDirPrevTop.IsSameAs(strDirNewTop);
    bSame &= strDirPrevRgt.IsSameAs(strDirNewRgt);
    bSame &= strDirPrevEuler.IsSameAs(strDirNewEuler);

    if (bSame == FALSE)
	mpCParent->mpCNumPanel->UpdateValues();

    FinishPaint(event);
}

void CLS::
FinishPaint(wxPaintEvent& event){
    //event.Skip();
    mbStatusDrawing = FALSE;

    //Refresh();
}

void CLS::
OnMouse(wxMouseEvent& event){
    if (HasFocus() == FALSE) SetFocus();
    //Refresh();
    ProcessMouseEvent(event);
    //event.Skip();
}

void CLS::
OnMouseMotion(wxMouseEvent& event){
    int mouseX = event.GetX();
    int mouseY = event.GetY();
    if (mpCBtnNumPanel->UpdateStatus(GetClientRect().GetSize(), mouseX, mouseY) == TRUE)
	mbRedrawQueue = TRUE;

    if (mpCParent->mCbDisplayLoupe.GetBodyBOOL() == TRUE)
	Refresh();
    
    ProcessMouseEvent(event);
}

void CLS::
OnKeyDown(wxKeyEvent& event){
    if (mInputMode == 0){
	event.Skip();
	return;
    }

    int keyCode = event.GetKeyCode();
    if (keyCode >= '0' && keyCode <= '9' || keyCode == '-' || keyCode == '.'){
	mInputMode = 2;
	mstrInputCommand << (char)keyCode;
	Refresh();
	return;
    }
    if (keyCode == WXK_BACK){
	mstrInputCommand.RemoveLast(1);
	if (mstrInputCommand.IsEmpty() == TRUE) mInputMode = 0;
	Refresh();
	return;
    }
    
    if (mInputMode == 1 && keyCode >= 'X' && keyCode <= 'Z'){
	if (mstrInputCommand.Length() > 1)
	    mstrInputCommand = mstrInputCommand.Left(1);
	mstrInputCommand << (char)keyCode;
	Refresh();//event.Skip();
	return;
    }
    if (keyCode == WXK_RETURN){
	if (ProcInputCommand() == TRUE) Refresh();
	mInputMode = 0;
	mstrInputCommand.Clear();
	return;
    }
    
    mstrInputCommand.Clear();
    mInputMode = 0;
    Refresh();
    event.Skip();
}

bool GLCanvas_cls::msbInitializedCommon = FALSE;
bool GLCanvas_cls::msbLockDragging = FALSE;
