// (setq c-basic-offset 4)
// "clsGLNumericalPanel.h"

#include "clsGLNumericalPanel.h"

#include "clsView3D.h"

#include "../MYGL/clsWideCamera.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/MULTIINPUT/clsHideAndShowButton.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsBuf.h"

#include "../../wxWidgetEnum.h"

#define CLS GLNumericalPanel_cls

CLS::
GLNumericalPanel_cls(View3D_cls *pCParent, wxSize& size) : 
    MNumPainBase_cls(pCParent, size)
{
    mpCParent = pCParent;

    wxBoxSizer *pCBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    
    SetMNumViewPos();
    SetMNumActiveObj();
    SetMNumFollow();
    SetMNumDisplay();

    mAstrClosedItem.push_back("follow camera");
    mAstrClosedItem.push_back("display");

    SetMNum(mpCMNumViewPos, mAstrClosedItem, _("View"));
    SetMNum(mpCMNumActiveObj, mAstrClosedItem, _("Active Object"));
    SetMNum(mpCMNumFollow, mAstrClosedItem, _("Camera"));
    SetMNum(mpCMNumDisplay, mAstrClosedItem, _("Display"));

    InitLayoutNumPanels(pCBoxSizerSrc);
    SetSizerAndFit(pCBoxSizerSrc);
    SetMinClientSize(size);
    SetMaxClientSize(size);

    SyncStateToData();
    LayoutNumPanels();

    HideAndShow();
}

void CLS::
SetMNumViewPos(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNumViewPos, this,
					"View3DNumericalPanel", TRUE);

    pC->mMinWidth = 180;
    pC->mTextWidth = 20;
    
    pC->mDragScale = 2;
    int indexP = pC->CombineConfItem(mpCParent->mCViewPos, "P");
    int indexTg = pC->CombineConfItem(mpCParent->mCViewDirTgt, "Tg");
    int indexTp = pC->CombineConfItem(mpCParent->mCViewDirTop, "Tp");
    int indexRt = pC->CombineConfItem(mpCParent->mCViewDirRgt, "Rt");
    int indexEu = pC->CombineConfItem(mpCParent->mCViewEuler, "Eu");
    
    pC->mTextWidth = 110;
    mIndexAngleL = pC->CombineSizer(2, _("AngleL/Pivot"));
    pC->SetLimit(1, 180);
    pC->mDragScale = 10;
    pC->CombineConfItem(mpCParent->mCViewAngleL, "AngleL");
    pC->CombineConfItem(mpCParent->mCstrPivotObject, "Pivot");

    mIndexSizeOrthoL = pC->CombineSizer(2, "SizeL/Pivot");
    pC->mDragScale = 1;
    pC->CombineConfItem(mpCParent->mCSizeOrthoL, "Size");
    pC->CombineConfItem(mpCParent->mCstrPivotObject, "Pivot");
    
    pC->mTextWidth = 20;
    pC->mDragScale = 2;
    mIndexPivotPos = pC->CombineConfItem(mpCParent->mCPivotPos, "P");
    
    pC->FinishSetting();

    pC->SetColorForXYZ(indexP);
    pC->SetColorForXYZ(indexTg);
    pC->SetColorForXYZ(indexTp);
    pC->SetColorForXYZ(indexRt);
    pC->SetColorForXYZ(indexEu);
    pC->SetColorForXYZ(mIndexPivotPos);

    pC->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_TEXT_ENTER);
}


void CLS::
SetMNumActiveObj(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNumActiveObj, this,
					"View3DNumericalPanel", TRUE);
    
    pC->mMinWidth = 180;
    pC->mTextWidth = 60;
    pC->CombineSizer(2, "Object");
    pC->CombineBitmapPanel(btnDropdown, wxES, 14);
    pC->mVCItem.back().mBitmapTemplate = _BITMAP_BTN_DROPDOWN;
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = "ActiveObject";
    pC->CombineConfItem(mpCParent->mCstrActiveObject);
    
    pC->mTextWidth = 20;
    pC->mDragScale = 2;
    mIndexAOPos = pC->CombineConfItem(mpCParent->mCObjectPos, "P");
    pC->mDragScale = 10;
    mIndexAORot = pC->CombineConfItem(mpCParent->mCObjectRot, "R");
    pC->mDragScale = 1;
    mIndexAOScale = pC->CombineConfItem(mpCParent->mCObjectScale, "S");
    
    pC->FinishSetting();

    pC->SetColorForXYZ(mIndexAOPos);
    pC->SetColorForXYZ(mIndexAORot);
    pC->SetColorForXYZ(mIndexAOScale);

    pC->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER);
}

void CLS::
SetMNumFollow(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNumFollow, this,
					"View3DNumericalPanel", TRUE);

    pC->mTextWidth = 60;
    pC->CombineSizer(3, "Camera");
    pC->CombineBitmapPanel(btnDropdown, wxES, 14);
    pC->mVCItem.back().mBitmapTemplate = _BITMAP_BTN_DROPDOWN;
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = "Camera";
    mIndexCameraName = pC->CombineConfItem(mpCParent->mCstrCamera);
    pC->CombineConfItem(mpCParent->mCbFollow, _("Follow"));
    pC->mTextWidth = 120;
    mIndexLock = 
	pC->CombineConfItem(mpCParent->mCbLockToCamera, _("Lock to Camera"));

    pC->FinishSetting();

    pC->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_TEXT_ENTER);
}

void CLS::
SetMNumDisplay(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNumDisplay, this,
					"View3DNumericalPanel", TRUE);
    
    pC->mTextWidth = 100;

    pC->SetChoiceList(_CAMERA_MAX, GetPCStringProvider()->GetAstr("CameraType"));
    mIndexCameraType = 
	pC->CombineConfItem(mpCParent->mCWideCameraType, _("Camera Type"));

    pC->SetChoiceList(_BUFFER_INDEX_MAX-2, GetPCStringProvider()->GetAstr("BufferSize"));
    mIndex1stBuffer = 
	pC->CombineConfItem(mpCParent->mC1stBufferSizeIndex, _("1st Buffer"));

    pC->SetChoiceList(_BUFFER_INDEX_MAX, GetPCStringProvider()->GetAstr("BufferSize"));
    pC->CombineConfItem(mpCParent->mCBufferSizeIndex, _("Buffer Size"));
    
    pC->mDragScale = 4;
    pC->SetLimit(1, 12);
    pC->CombineConfItem(mpCParent->mCLineWidth, _("Line Width"));
 
    pC->FinishSetting();
}

void CLS::
HideAndShow(){

    bool bChange = FALSE;
    if (mpCParent->mCWideCameraType.GetBodyINT() == _WC_ORTHOGONAL){
	bChange |= mpCMNumViewPos->mVCItem[mIndexAngleL].Hide();
	bChange |= mpCMNumViewPos->mVCItem[mIndexSizeOrthoL].Show();
    }
    else{
	bChange |= mpCMNumViewPos->mVCItem[mIndexAngleL].Show();
	bChange |= mpCMNumViewPos->mVCItem[mIndexSizeOrthoL].Hide();
    }


    if (mpCParent->mCstrPivotObject.GetBodySTR().IsSameAs("None")){
	bChange |= mpCMNumViewPos->mVCItem[mIndexPivotPos].Show();
    }
    else{
	bChange |= mpCMNumViewPos->mVCItem[mIndexPivotPos].Hide();
    }

    if (mpCParent->mCstrActiveObject.GetBodySTR().IsSameAs("None")){
	bChange |= mpCMNumActiveObj->mVCItem[mIndexAOPos].Hide();
	bChange |= mpCMNumActiveObj->mVCItem[mIndexAORot].Hide();
	bChange |= mpCMNumActiveObj->mVCItem[mIndexAOScale].Hide();
    }
    else{
	bChange |= mpCMNumActiveObj->mVCItem[mIndexAOPos].Show();
	bChange |= mpCMNumActiveObj->mVCItem[mIndexAORot].Show();
	bChange |= mpCMNumActiveObj->mVCItem[mIndexAOScale].Show();
    }

    bool bFollow = mpCParent->mCbFollow.GetBodyBOOL();
    if (mpCParent->GetPCCamera() == NULL){
	bFollow = FALSE;
	wxColor clr(GetPCColorProvider()->GetColor("Error"));
	mpCMNumFollow->mVCItem[mIndexCameraName].SetBackgroundColour(clr);
    }
    else{
	mpCMNumFollow->mVCItem[mIndexCameraName].SetBackgroundColour(*wxWHITE);
    }
    
    if (bFollow == TRUE){
	bChange |= mpCMNumFollow->mVCItem[mIndexLock].Show();
	bChange |= mpCMNumDisplay->mVCItem[mIndexCameraType].Hide();
	bChange |= mpCMNumDisplay->mVCItem[mIndex1stBuffer].Hide();
    }
    else{
	bool bLock = mpCParent->mCbLockToCamera.GetBodyBOOL();
	if (bLock == TRUE)
	    bChange |= mpCMNumFollow->mVCItem[mIndexLock].Show();
	else
	    bChange |= mpCMNumFollow->mVCItem[mIndexLock].Hide();
	bChange |= mpCMNumDisplay->mVCItem[mIndexCameraType].Show();
	if (mpCParent->mCWideCameraType.GetBodyINT() > 1){
	    bChange |= mpCMNumDisplay->mVCItem[mIndex1stBuffer].Show();
	}
	else{
	    bChange |= mpCMNumDisplay->mVCItem[mIndex1stBuffer].Hide();
	}
    }
        
    if (bChange == TRUE){
	Layout();
	GetParent()->Layout();
    }
}

void CLS::
OnViewChange(wxCommandEvent& event){
    mpCParent->UpdatePanels("ALL");
}

void CLS::
OnActiveObjChange(wxCommandEvent& event){
    mpCParent->SyncronizeObjToPanel();
}
