// -*- c -*-
// "clsVBOBaseObject.h"
// (setq c-basic-offset 4)

#include "clsVBOBaseObject.h"
//#include "../FILTER/clsFilterObject.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLDELETEBUFFERSPROC glDeleteBuffers;
extern PFNGLGENBUFFERSPROC glGenBuffers;
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
#endif

#define CLS VBOBaseObject_cls

CLS::
VBOBaseObject_cls(){
    mVBOVertex = 0;
    mVBONormal = 0;
    for (int i = 0; i < _MAX_MAPPING_TYPE; i++) mVBOUV[i] = 0;
    mVVBOIndiceTriangle.clear();
    mVVBOIndiceAdjacent.clear();
    mVIndiceTriangle.clear();
    mVIndiceTriangle.clear();
    //mVBOIndicesTriangle = 0;
    //mVBOIndicesAdjacent = 0;
    mbUseNormal = FALSE;
}

CLS::
~VBOBaseObject_cls(){
    Clear();
}

void CLS::
Clear(){
    DeleteVBOIfGenerated(mVBOVertex);
    DeleteVBOIfGenerated(mVBONormal);
    for (int i = 0; i < _MAX_MAPPING_TYPE; i++) 
	DeleteVBOIfGenerated(mVBOUV[i]);

    DeleteVVBOIfGenerated(mVVBOIndiceTriangle);
    DeleteVVBOIfGenerated(mVVBOIndiceAdjacent);
    mbUseNormal = FALSE;
}

void CLS::
DeleteVBOIfGenerated(unsigned int &VBO){
    if (VBO != 0){
	glDeleteBuffers(1, &VBO);
	VBO = 0;
    }
}

void CLS::
DeleteVVBOIfGenerated(std::vector < unsigned int > &VVBO){
    unsigned int numVBO = VVBO.size();
    if (numVBO == 0) return;
    glDeleteBuffers(numVBO, VVBO.data());
    VVBO.clear();
}

void CLS::
GenBuffersVertex(int UVNumber){
    glGenBuffers(1, &mVBOVertex);
    glGenBuffers(1, &mVBONormal);
    glGenBuffers(UVNumber, mVBOUV);
    //glGenBuffers(1, &(mVVBOIndiceTriangle[0]));
    //glGenBuffers(1, &(mVVBOIndiceAdjacent[0]));
}

int CLS::
GetNumVBOTriangle(){
    int numVBOTriangle = 1 + mTotalIndiceTriangle / _VBO_INDICE_UNIT;
    if (mTotalIndiceTriangle == 0) numVBOTriangle = 0;
    return numVBOTriangle;
}

int CLS::
GetNumVBOAdjacent(){
    int numVBOAdjacent = 1 + mTotalIndiceAdjacent / _VBO_INDICE_UNIT;
    if (mTotalIndiceAdjacent == 0) numVBOAdjacent = 0;
    return numVBOAdjacent;
}

void CLS::
GenBuffersIndice(){
   
    int numVBOTriangle = GetNumVBOTriangle();
    int numVBOAdjacent = GetNumVBOAdjacent();
    
    mVVBOIndiceTriangle.resize(numVBOTriangle);
    mVVBOIndiceAdjacent.resize(numVBOAdjacent);
    mVIndiceTriangle.resize(numVBOTriangle);
    mVIndiceAdjacent.resize(numVBOAdjacent);
    if (numVBOTriangle != 0)
	glGenBuffers(numVBOTriangle, mVVBOIndiceTriangle.data());
    if (numVBOAdjacent != 0)
	glGenBuffers(numVBOAdjacent, mVVBOIndiceAdjacent.data());
    for (int i = 0; i < numVBOTriangle; i++){
	int num = mTotalIndiceTriangle - i * _VBO_INDICE_UNIT;
	if (num > _VBO_INDICE_UNIT) num = _VBO_INDICE_UNIT;
	mVIndiceTriangle[i] = num;
    }
    for (int i = 0; i < numVBOAdjacent; i++){
	int num = mTotalIndiceAdjacent - i * _VBO_INDICE_UNIT;
	if (num > _VBO_INDICE_UNIT) num = _VBO_INDICE_UNIT;
	mVIndiceAdjacent[i] = num;
    }
}

void CLS::
MakeVBO(float (*pVertex)[3], float (*pNormal)[3], float (*pUValongX)[2], float (*pUValongY)[2], float (*pUValongZ)[2], float (*pUVUV)[2], int totalVertex){
    
    glBindBuffer(GL_ARRAY_BUFFER, mVBOVertex);
    glBufferData(GL_ARRAY_BUFFER, 3*totalVertex * sizeof(float), pVertex, GL_STATIC_DRAW);

    if (pNormal != NULL){
	mbUseNormal = TRUE;
	glBindBuffer(GL_ARRAY_BUFFER, mVBONormal);
	glBufferData(GL_ARRAY_BUFFER, 3*totalVertex * sizeof(float), pNormal, GL_STATIC_DRAW);
    }

    if (pUValongX != NULL){
	glBindBuffer(GL_ARRAY_BUFFER, mVBOUV[0]);
	glBufferData(GL_ARRAY_BUFFER, 2*totalVertex * sizeof(float), pUValongX, GL_STATIC_DRAW);
    }

    if (pUValongY != NULL){
	glBindBuffer(GL_ARRAY_BUFFER, mVBOUV[1]);
	glBufferData(GL_ARRAY_BUFFER, 2*totalVertex * sizeof(float), pUValongY, GL_STATIC_DRAW);
    }

    if (pUValongZ != NULL){
	glBindBuffer(GL_ARRAY_BUFFER, mVBOUV[2]);
	glBufferData(GL_ARRAY_BUFFER, 2*totalVertex * sizeof(float), pUValongZ, GL_STATIC_DRAW);
    }

    if (pUVUV != NULL){
	glBindBuffer(GL_ARRAY_BUFFER, mVBOUV[3]);
	glBufferData(GL_ARRAY_BUFFER, 2*totalVertex * sizeof(float), pUVUV, GL_STATIC_DRAW);
    }
}

unsigned int* CLS::
GetPNewTriangle(){
    unsigned int *pIndiceTriangle = new unsigned int [mTotalIndiceTriangle];
    return pIndiceTriangle;
}

unsigned int* CLS::
GetPNewAdjucent(){
    unsigned int *pIndiceAdjucent = new unsigned int [mTotalIndiceAdjacent];
    for (int i = 0; i < mTotalIndiceAdjacent; i++)
	pIndiceAdjucent[i] = UINT_MAX;
    return pIndiceAdjucent;
}

void CLS::
MakeVBOIndicesTriangle(unsigned int *pIndicesTriangle){
    int numVBOTriangle = GetNumVBOTriangle();
    for (int i = 0; i < numVBOTriangle; i++){
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVVBOIndiceTriangle[i]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, mVIndiceTriangle[i] * sizeof(unsigned int), pIndicesTriangle + i * _VBO_INDICE_UNIT, GL_STATIC_DRAW);
    }
}

void CLS::
MakeVBOIndicesAdjacent(unsigned int *pIndicesAdjucent){
    int numVBOAdjacent = GetNumVBOAdjacent();
    for (int i = 0; i < numVBOAdjacent; i++){
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVVBOIndiceAdjacent[i]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, mVIndiceAdjacent[i] * sizeof(unsigned int), pIndicesAdjucent + i * _VBO_INDICE_UNIT, GL_STATIC_DRAW);
    }
}

void CLS::
PrepareClientState(bool bTexture, int mapType, bool bShadow, int VBOindex){
    bool bColor = FALSE;

    if (bTexture == TRUE && mVBOUV[mapType] != 0){
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, mVBOUV[mapType]);
	glTexCoordPointer(2, GL_FLOAT, 0, 0);
    }
    else{
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
    
    glEnableClientState(GL_VERTEX_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, mVBOVertex);
    glVertexPointer(3, GL_FLOAT, 0, 0);
    
    if (bShadow == TRUE){
	glDisableClientState(GL_NORMAL_ARRAY);
	
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVVBOIndiceAdjacent[VBOindex]);
    }
    else{
	if (mbUseNormal == TRUE){
	    glEnableClientState(GL_NORMAL_ARRAY);
	    glBindBuffer(GL_ARRAY_BUFFER, mVBONormal);
	    glNormalPointer(GL_FLOAT, 0, 0);
	}
	else
	    glDisableClientState(GL_NORMAL_ARRAY);
	
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVVBOIndiceTriangle[VBOindex]);
    }
}

void CLS::
Draw(int objectState, int mapType, bool bPrepareClientState){
    if (objectState == _VBO_SHADOW_VOLUME_HF){
	uint numVBO = mVVBOIndiceAdjacent.size();
	//if (numVBO > 1) numVBO = 1;
	for (uint i = 0; i < numVBO; i++){
	    if (bPrepareClientState == TRUE) 
		PrepareClientState(TRUE, 0, TRUE, i);
	    glDrawElements(GL_TRIANGLES_ADJACENCY_EXT, mVIndiceAdjacent[i],
			   GL_UNSIGNED_INT, NULL);
	}
    }
    else if (objectState == _VBO_SHADOW_VOLUME){
	uint numVBO = mVVBOIndiceAdjacent.size();
	for (uint i = 0; i < numVBO; i++){
	    if (bPrepareClientState == TRUE) 
		PrepareClientState(FALSE, 0, TRUE, i);
	    glDrawElements(GL_TRIANGLES_ADJACENCY_EXT, mVIndiceAdjacent[i],
			   GL_UNSIGNED_INT, NULL);
	}
    }
    else{
	uint numVBO = mVVBOIndiceTriangle.size();
	for (uint i = 0; i < numVBO; i++){
	    if (bPrepareClientState == TRUE) 
		PrepareClientState(TRUE, mapType, FALSE, i);
	    glDrawElements(GL_TRIANGLES, mVIndiceTriangle[i], 
			   GL_UNSIGNED_INT, NULL);
	}
    }
}

void CLS::
DrawFast(){
    glDrawElements(GL_TRIANGLES, mVIndiceTriangle[0], 
		   GL_UNSIGNED_INT, NULL);
}

void CLS::
FinishClientState(){
    glDisableClientState(GL_COLOR_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
