// -*- c -*-
// "clsVBOCommon.h"
// (setq c-basic-offset 4)

#include "clsVBOCommon.h"

#include "../COMMON/clsArrayAndIndex.h"
#include "../COMMON/clsSortParallel.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include <algorithm>

#ifdef WIN32
//extern PFNGLBINDBUFFERARBPROC glBindBufferARB;
//extern PFNGLBUFFERDATAARBPROC glBufferDataARB;
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
#endif

#include <limits.h>

int VBOCommon_cls::
SearchAdjTriangleB(int ibase_, unsigned int * const pIndiceTriangle, 
		   ArrayAndIndex_cls * const pCIdxAndZ,  
		   const int vertex0, const int vertex1, 
		   const int vertex2,
		   const int triangleNum, const int searchWidth){
    int ini = ibase_ - searchWidth;
    int fin = ibase_ + searchWidth;
    if (ini < 0) ini = 0;
    if (fin > triangleNum) fin = triangleNum;
    
    for (int i_ = ini; i_ < fin; i_++){
	int i = pCIdxAndZ[i_].index;
	int vertexA = pIndiceTriangle[i*3];
	int vertexB = pIndiceTriangle[i*3+1];
	int vertexC = pIndiceTriangle[i*3+2];
	if (vertexB == vertex0 && vertexA == vertex1 &&
	    vertexC != vertex2) return vertexC;
	if (vertexC == vertex0 && vertexB == vertex1 &&
	    vertexA != vertex2) return vertexA;
	if (vertexA == vertex0 && vertexC == vertex1 &&
	    vertexB != vertex2) return vertexB;
    }
    return -1;
}

void VBOCommon_cls::
ConstructAdjStructure(float (* const pVertex)[3], 
		      unsigned int * const pIndiceTriangle,
		      unsigned int * const pIndiceAdj, 
		      const int totalTriangle,
		      const int searchAxis, const int searchWidth){
    
    ArrayAndIndex_cls *pCIdxAndZ = new ArrayAndIndex_cls [totalTriangle];

    for (int i = 0; i < totalTriangle; i++){
	pCIdxAndZ[i].index = i;
	int dummy0 = pIndiceTriangle[i*3];
	float z0 = pVertex[pIndiceTriangle[i*3]][searchAxis];
	float z1 = pVertex[pIndiceTriangle[i*3+1]][searchAxis];
	float z2 = pVertex[pIndiceTriangle[i*3+2]][searchAxis];
	pCIdxAndZ[i].z = (z0 + z1 + z2) / 3;
    }
    
    if (totalTriangle > 640000){
	SortParallel_cls CSortParallel;
	CSortParallel.SortParallel(pCIdxAndZ, totalTriangle);
    }
    else
	std::sort(pCIdxAndZ, pCIdxAndZ+totalTriangle);

    // pIndiceAdj != UINT_MAX ̂̂͌vZς݂Ƃ
    for (int i_ = 0; i_ < totalTriangle; i_++){
	int i = pCIdxAndZ[i_].index;
	if (pIndiceAdj[i*6] != UINT_MAX) continue;
	int vertexA = pIndiceTriangle[i*3];
	int vertexB = pIndiceTriangle[i*3+1];
	int vertexC = pIndiceTriangle[i*3+2];
	int vertexABdash = SearchAdjTriangleB
	    (i_, pIndiceTriangle, pCIdxAndZ, vertexA, vertexB, vertexC,
	     totalTriangle, searchWidth);
	int vertexBCdash = SearchAdjTriangleB
	    (i_, pIndiceTriangle, pCIdxAndZ, vertexB, vertexC, vertexA,
	     totalTriangle, searchWidth);
	int vertexCAdash = SearchAdjTriangleB
	    (i_, pIndiceTriangle, pCIdxAndZ, vertexC, vertexA, vertexB,
	     totalTriangle, searchWidth);
	if (vertexABdash == -1) vertexABdash = vertexC;
	if (vertexBCdash == -1) vertexBCdash = vertexA;
	if (vertexCAdash == -1) vertexCAdash = vertexB;

	pIndiceAdj[i*6+0] = vertexA;
	pIndiceAdj[i*6+1] = vertexABdash;
	pIndiceAdj[i*6+2] = vertexB;
	pIndiceAdj[i*6+3] = vertexBCdash;
	pIndiceAdj[i*6+4] = vertexC;
	pIndiceAdj[i*6+5] = vertexCAdash;
	
	/*
	mVFaceIndexVertexAdj.push_back(vertexA);
	mVFaceIndexVertexAdj.push_back(vertexABdash);
	mVFaceIndexVertexAdj.push_back(vertexB);
	mVFaceIndexVertexAdj.push_back(vertexBCdash);
	mVFaceIndexVertexAdj.push_back(vertexC);
	mVFaceIndexVertexAdj.push_back(vertexCAdash);
	*/
    }

    delete [] pCIdxAndZ;
}
