// -*- c -*-
// "clsVBOCone.h"
// (setq c-basic-offset 4)

#include "clsVBOCone.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
#endif

#define CLS VBOCone_cls

CLS::
VBOCone_cls(){
}

void CLS::
Initialize(){
    SetGeometry(&mCObject);
}

void CLS::
Draw(int objectState, int mapType){
    //mapType = 0;
    mCObject.Draw(objectState, mapType);
}

int CLS::
GetTotalVertex(int face){
    return (face+1)*2 + 1;
}

int CLS::
GetTotalTriangle(int face){
    return face*2 - 2;    
}

void CLS::
ConstructTriangles(float (*pVertex)[3], float (*pNormal)[3], 
		   float (*pUValongX)[2], float (*pUValongY)[2], 
		   float (*pUValongZ)[2], float (*pUVUV)[2],
		   unsigned int *pIndiceTriangle, int face){

    float factor = M_PI * 2 / face;

    int totalVertex = GetTotalVertex(face);
    int totalTriangle = GetTotalTriangle(face);

    int vertexOffsetBTM = (face+1)+1;
    int indiceOffsetBTM = (face)*3;

    // a1ō2
    float cosAngle = 2.0f/sqrt(5.0);
    float sinAngle = 1.0f/sqrt(5.0);

    // _
    float vert[3] = {0, 0, 1}, normal[3] = {0, 0, 1};
    LOOP_XYZ pVertex[0][xyz] = vert[xyz];
    LOOP_XYZ pNormal[0][xyz] = normal[xyz];

    if (pUValongX != NULL){
	pUValongX[0][0] = 0.5*vert[Y]+0.5;
	pUValongX[0][1] = 0.5*vert[Z]+0.5; 
	pUValongY[0][0] = 0.5*vert[X]+0.5;
	pUValongY[0][1] = 0.5*vert[Z]+0.5;
	pUValongZ[0][0] = 0.5*vert[X]+0.5; 
	pUValongZ[0][1] = 0.5*vert[Y]+0.5; 
	pUVUV[0][0] = 0;
	pUVUV[0][1] = 1;
    }
    
    // ǖ
    for (int i = 0; i <= face; i++){
	float vert0[3] = {cos(factor*i), sin(factor*i), -1};
	float normal0[3] = {cos(factor*i)*cosAngle, sin(factor*i)*cosAngle, -sinAngle};
	LOOP_XYZ pVertex[i+1][xyz] = vert0[xyz];
	LOOP_XYZ pNormal[i+1][xyz] = normal0[xyz];

	if (pUValongX != NULL){
	    pUValongX[i+1][0] = 0.5*vert0[Y]+0.5;
	    pUValongX[i+1][1] = 0.5*vert0[Z]+0.5; 
	    pUValongY[i+1][0] = 0.5*vert0[X]+0.5;
	    pUValongY[i+1][1] = 0.5*vert0[Z]+0.5;
	    pUValongZ[i+1][0] = 0.5*vert0[X]+0.5; 
	    pUValongZ[i+1][1] = 0.5*vert0[Y]+0.5; 
	    pUVUV[i+1][0] = factor*i/M_PI/2;
	    pUVUV[i+1][1] = 0;
	}

    }

    // 
    for (int i = 0; i <= face; i++){
	float vert0[3] = {cos(factor*i), -sin(factor*i), -1};
	float normal0[3] = {0, 0, -1};

	int index = vertexOffsetBTM + i;

	LOOP_XYZ pVertex[index][xyz] = vert0[xyz];
	LOOP_XYZ pNormal[index][xyz] = normal0[xyz];

	if (pUValongX != NULL){
	    pUValongX[index][0] = 0.5*vert0[Y]+0.5;
	    pUValongX[index][1] = 0.5*vert0[Z]+0.5; 
	    pUValongY[index][0] = 0.5*vert0[X]+0.5;
	    pUValongY[index][1] = 0.5*vert0[Z]+0.5;
	    pUValongZ[index][0] = 0.5*vert0[X]+0.5; 
	    pUValongZ[index][1] = 0.5*vert0[Y]+0.5; 
	    pUVUV[index][0] = 0;
	    pUVUV[index][1] = 0;
	}
	
    }

    // ǖ
    for (int i = 0; i < face; i++){
	pIndiceTriangle[i*3+0] = (i+1+1);
	pIndiceTriangle[i*3+1] = 0;
	pIndiceTriangle[i*3+2] = (i+1);
    }
    // 
    for (int i = 0; i < face-2; i++){
	pIndiceTriangle[indiceOffsetBTM+i*3+0] = vertexOffsetBTM + 0;
	pIndiceTriangle[indiceOffsetBTM+i*3+1] = vertexOffsetBTM + i+1;
	pIndiceTriangle[indiceOffsetBTM+i*3+2] = vertexOffsetBTM + i+2;
    }
}

#define LOOP_UV for (int uv = 0; uv < 2; uv++)

void CLS::
SetGeometry(VBOBaseObject_cls *pCTgt){

    pCTgt->Clear();
    pCTgt->GenBuffersVertex(4);

    int face = 48;
    float factor = M_PI * 2 / face;

    int totalVertex = GetTotalVertex(face);
    int totalTriangle = GetTotalTriangle(face);
    pCTgt->mTotalIndiceTriangle = totalTriangle*3;

    float (*pVertex)[3] = new float [totalVertex][3];
    float (*pNormal)[3] = new float [totalVertex][3];
    float (*pUValongX)[2] = new float [totalVertex][2];
    float (*pUValongY)[2] = new float [totalVertex][2];
    float (*pUValongZ)[2] = new float [totalVertex][2];
    float (*pUVUV)[2] = new float [totalVertex][2];
    
    unsigned int *pIndiceTriangle = new unsigned int [pCTgt->mTotalIndiceTriangle];
    
    ConstructTriangles(pVertex, pNormal, pUValongX, pUValongY, pUValongZ, pUVUV,
		       pIndiceTriangle, face);

    pCTgt->MakeVBO(pVertex, pNormal, pUValongX, pUValongY, pUValongZ, pUVUV, totalVertex);
    
    pCTgt->mTotalIndiceAdjacent = totalTriangle*3*2;
    unsigned int *pIndiceAdjacent = pCTgt->GetPNewAdjucent();
    
    ConstructAdjStructure(pVertex, pIndiceTriangle, pIndiceAdjacent, totalTriangle);

    pCTgt->GenBuffersIndice();
    pCTgt->MakeVBOIndicesTriangle(pIndiceTriangle);
    pCTgt->MakeVBOIndicesAdjacent(pIndiceAdjacent);
    
    delete [] pVertex;
    delete [] pNormal;
    delete [] pUValongX;
    delete [] pUValongY;
    delete [] pUValongZ;
    delete [] pUVUV;
    
    delete [] pIndiceTriangle;
    delete [] pIndiceAdjacent;
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

CLS *GpCVBOCone = NULL;


