// -*- c -*-
// "clsVBOCube.h"
// (setq c-basic-offset 4)

#include "clsVBOCube.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
#endif

#define CLS VBOCube_cls

void CLS::
Initialize(){
    SetGeometry(&mCObject);
}

void CLS::
Draw(int objectState, int mapType){
    if (mapType > _UV_MAPPING_UV) mapType = _UV_MAPPING_UV;
    mCObject.Draw(objectState, mapType);
}

int CLS::
GetTotalVertex(int face){
    return face*4;
}

int CLS::
GetTotalTriangle(int face){
    return face*2;
}

void CLS::
ConstructTriangles(float (*pVertex)[3], float (*pNormal)[3], 
		   unsigned int *pIndiceTriangle){

    float vert000[3] = {-1, -1, -1};
    float vert001[3] = {+1, -1, -1};
    float vert010[3] = {-1, +1, -1};
    float vert011[3] = {+1, +1, -1};
    float vert100[3] = {-1, -1, +1};
    float vert101[3] = {+1, -1, +1};
    float vert110[3] = {-1, +1, +1};
    float vert111[3] = {+1, +1, +1};

    float normalzp[3] = {0, 0, +1};
    float normalzm[3] = {0, 0, -1};
    float normalxp[3] = {+1, 0, 0};
    float normalxm[3] = {-1, 0, 0};
    float normalyp[3] = {0, +1, 0};
    float normalym[3] = {0, -1, 0};

    LOOP_XYZ pVertex[0*4+0][xyz] = vert100[xyz]; //
    LOOP_XYZ pVertex[0*4+1][xyz] = vert101[xyz]; 
    LOOP_XYZ pVertex[0*4+2][xyz] = vert111[xyz]; 
    LOOP_XYZ pVertex[0*4+3][xyz] = vert110[xyz]; 
    for (int i = 0; i < 4; i++)
	LOOP_XYZ pNormal[0*4+i][xyz] = normalzp[xyz];

    LOOP_XYZ pVertex[1*4+0][xyz] = vert000[xyz]; //
    LOOP_XYZ pVertex[1*4+1][xyz] = vert010[xyz]; 
    LOOP_XYZ pVertex[1*4+2][xyz] = vert011[xyz]; 
    LOOP_XYZ pVertex[1*4+3][xyz] = vert001[xyz]; 
    for (int i = 0; i < 4; i++)
	LOOP_XYZ pNormal[1*4+i][xyz] = normalzm[xyz];

    LOOP_XYZ pVertex[2*4+0][xyz] = vert001[xyz]; //E
    LOOP_XYZ pVertex[2*4+1][xyz] = vert011[xyz]; 
    LOOP_XYZ pVertex[2*4+2][xyz] = vert111[xyz]; 
    LOOP_XYZ pVertex[2*4+3][xyz] = vert101[xyz]; 
    for (int i = 0; i < 4; i++)
	LOOP_XYZ pNormal[2*4+i][xyz] = normalxp[xyz];
    
    LOOP_XYZ pVertex[3*4+0][xyz] = vert000[xyz]; //
    LOOP_XYZ pVertex[3*4+1][xyz] = vert100[xyz]; 
    LOOP_XYZ pVertex[3*4+2][xyz] = vert110[xyz]; 
    LOOP_XYZ pVertex[3*4+3][xyz] = vert010[xyz]; 
    for (int i = 0; i < 4; i++)
	LOOP_XYZ pNormal[3*4+i][xyz] = normalxm[xyz];

    LOOP_XYZ pVertex[4*4+0][xyz] = vert000[xyz]; //
    LOOP_XYZ pVertex[4*4+1][xyz] = vert001[xyz]; 
    LOOP_XYZ pVertex[4*4+2][xyz] = vert101[xyz]; 
    LOOP_XYZ pVertex[4*4+3][xyz] = vert100[xyz]; 
    for (int i = 0; i < 4; i++)
	LOOP_XYZ pNormal[4*4+i][xyz] = normalym[xyz];

    LOOP_XYZ pVertex[5*4+0][xyz] = vert010[xyz]; //O
    LOOP_XYZ pVertex[5*4+1][xyz] = vert110[xyz]; 
    LOOP_XYZ pVertex[5*4+2][xyz] = vert111[xyz]; 
    LOOP_XYZ pVertex[5*4+3][xyz] = vert011[xyz]; 
    for (int i = 0; i < 4; i++)
	LOOP_XYZ pNormal[5*4+i][xyz] = normalyp[xyz];

    int triangles[6] = {0, 1, 2, 2, 3, 0};
    for (int j = 0; j < 6; j++)
	for (int i = 0; i < 6; i++) 
	    pIndiceTriangle[j*6+i] = 4*j + triangles[i]; // 
}

#define LOOP_UV for (int uv = 0; uv < 2; uv++)

void CLS::
ConstructUV(float (*pUV)[2], int mapType){

    float UV00[2] = {0, 0};
    float UV01[2] = {1, 0};
    float UV10[2] = {0, 1};
    float UV11[2] = {1, 1};
    
    switch(mapType){
    case _UV_MAPPING_X:
	LOOP_UV pUV[0*4+0][uv] = UV10[uv]; // 
	LOOP_UV pUV[0*4+1][uv] = UV10[uv];
	LOOP_UV pUV[0*4+2][uv] = UV11[uv];
	LOOP_UV pUV[0*4+3][uv] = UV11[uv];
	LOOP_UV pUV[1*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[1*4+1][uv] = UV01[uv];
	LOOP_UV pUV[1*4+2][uv] = UV01[uv];
	LOOP_UV pUV[1*4+3][uv] = UV00[uv];
	LOOP_UV pUV[2*4+0][uv] = UV00[uv]; // E
	LOOP_UV pUV[2*4+1][uv] = UV01[uv];
	LOOP_UV pUV[2*4+2][uv] = UV11[uv];
	LOOP_UV pUV[2*4+3][uv] = UV10[uv];
	LOOP_UV pUV[3*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[3*4+1][uv] = UV10[uv];
	LOOP_UV pUV[3*4+2][uv] = UV11[uv];
	LOOP_UV pUV[3*4+3][uv] = UV01[uv];
	LOOP_UV pUV[4*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[4*4+1][uv] = UV00[uv];
	LOOP_UV pUV[4*4+2][uv] = UV10[uv];
	LOOP_UV pUV[4*4+3][uv] = UV10[uv];
	LOOP_UV pUV[5*4+0][uv] = UV01[uv]; // O
	LOOP_UV pUV[5*4+1][uv] = UV11[uv];
	LOOP_UV pUV[5*4+2][uv] = UV11[uv];
	LOOP_UV pUV[5*4+3][uv] = UV01[uv];
	break;
    case _UV_MAPPING_Y:
	LOOP_UV pUV[0*4+0][uv] = UV10[uv]; // 
	LOOP_UV pUV[0*4+1][uv] = UV11[uv];
	LOOP_UV pUV[0*4+2][uv] = UV11[uv];
	LOOP_UV pUV[0*4+3][uv] = UV10[uv];
	LOOP_UV pUV[1*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[1*4+1][uv] = UV00[uv];
	LOOP_UV pUV[1*4+2][uv] = UV01[uv];
	LOOP_UV pUV[1*4+3][uv] = UV01[uv];
	LOOP_UV pUV[2*4+0][uv] = UV01[uv]; // E
	LOOP_UV pUV[2*4+1][uv] = UV01[uv];
	LOOP_UV pUV[2*4+2][uv] = UV11[uv];
	LOOP_UV pUV[2*4+3][uv] = UV11[uv];
	LOOP_UV pUV[3*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[3*4+1][uv] = UV10[uv];
	LOOP_UV pUV[3*4+2][uv] = UV10[uv];
	LOOP_UV pUV[3*4+3][uv] = UV00[uv];
	LOOP_UV pUV[4*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[4*4+1][uv] = UV01[uv];
	LOOP_UV pUV[4*4+2][uv] = UV11[uv];
	LOOP_UV pUV[4*4+3][uv] = UV10[uv];
	LOOP_UV pUV[5*4+0][uv] = UV00[uv]; // O
	LOOP_UV pUV[5*4+1][uv] = UV10[uv];
	LOOP_UV pUV[5*4+2][uv] = UV11[uv];
	LOOP_UV pUV[5*4+3][uv] = UV01[uv];
	break;
    case _UV_MAPPING_Z:
	LOOP_UV pUV[0*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[0*4+1][uv] = UV01[uv];
	LOOP_UV pUV[0*4+2][uv] = UV11[uv];
	LOOP_UV pUV[0*4+3][uv] = UV10[uv];
	LOOP_UV pUV[1*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[1*4+1][uv] = UV10[uv];
	LOOP_UV pUV[1*4+2][uv] = UV11[uv];
	LOOP_UV pUV[1*4+3][uv] = UV01[uv];
	LOOP_UV pUV[2*4+0][uv] = UV01[uv]; // E
	LOOP_UV pUV[2*4+1][uv] = UV11[uv];
	LOOP_UV pUV[2*4+2][uv] = UV11[uv];
	LOOP_UV pUV[2*4+3][uv] = UV01[uv];
	LOOP_UV pUV[3*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[3*4+1][uv] = UV00[uv];
	LOOP_UV pUV[3*4+2][uv] = UV10[uv];
	LOOP_UV pUV[3*4+3][uv] = UV10[uv];
	LOOP_UV pUV[4*4+0][uv] = UV00[uv]; // 
	LOOP_UV pUV[4*4+1][uv] = UV01[uv];
	LOOP_UV pUV[4*4+2][uv] = UV01[uv];
	LOOP_UV pUV[4*4+3][uv] = UV00[uv];
	LOOP_UV pUV[5*4+0][uv] = UV10[uv]; // O
	LOOP_UV pUV[5*4+1][uv] = UV10[uv];
	LOOP_UV pUV[5*4+2][uv] = UV11[uv];
	LOOP_UV pUV[5*4+3][uv] = UV11[uv];
	break;
    case 3:{
	float UVOffset00[2] = {-1.0/8, -1.0/6};
	float UVOffset01[2] = {+1.0/8, -1.0/6};
	float UVOffset10[2] = {-1.0/8, +1.0/6};
	float UVOffset11[2] = {+1.0/8, +1.0/6};
	float UVCenterTop[2] = {3.0/8, +5.0/6};
	float UVCenterBtm[2] = {3.0/8, +1.0/6};
	float UVCenterRgt[2] = {5.0/8, +3.0/6};
	float UVCenterLft[2] = {1.0/8, +3.0/6};
	float UVCenterBck[2] = {3.0/8, +3.0/6};
	float UVCenterFrt[2] = {7.0/8, +3.0/6};
	LOOP_UV pUV[0*4+0][uv] = UVOffset00[uv] +  UVCenterTop[uv];
	LOOP_UV pUV[0*4+1][uv] = UVOffset01[uv] +  UVCenterTop[uv];
	LOOP_UV pUV[0*4+2][uv] = UVOffset11[uv] +  UVCenterTop[uv];
	LOOP_UV pUV[0*4+3][uv] = UVOffset10[uv] +  UVCenterTop[uv];
	LOOP_UV pUV[1*4+0][uv] = UVOffset10[uv] +  UVCenterBtm[uv];
	LOOP_UV pUV[1*4+1][uv] = UVOffset00[uv] +  UVCenterBtm[uv];
	LOOP_UV pUV[1*4+2][uv] = UVOffset01[uv] +  UVCenterBtm[uv];
	LOOP_UV pUV[1*4+3][uv] = UVOffset11[uv] +  UVCenterBtm[uv];
	LOOP_UV pUV[2*4+0][uv] = UVOffset00[uv] +  UVCenterRgt[uv];
	LOOP_UV pUV[2*4+1][uv] = UVOffset01[uv] +  UVCenterRgt[uv];
	LOOP_UV pUV[2*4+2][uv] = UVOffset11[uv] +  UVCenterRgt[uv];
	LOOP_UV pUV[2*4+3][uv] = UVOffset10[uv] +  UVCenterRgt[uv];
	LOOP_UV pUV[3*4+0][uv] = UVOffset01[uv] +  UVCenterLft[uv];
	LOOP_UV pUV[3*4+1][uv] = UVOffset11[uv] +  UVCenterLft[uv];
	LOOP_UV pUV[3*4+2][uv] = UVOffset10[uv] +  UVCenterLft[uv];
	LOOP_UV pUV[3*4+3][uv] = UVOffset00[uv] +  UVCenterLft[uv];
	LOOP_UV pUV[4*4+0][uv] = UVOffset00[uv] +  UVCenterBck[uv];
	LOOP_UV pUV[4*4+1][uv] = UVOffset01[uv] +  UVCenterBck[uv];
	LOOP_UV pUV[4*4+2][uv] = UVOffset11[uv] +  UVCenterBck[uv];
	LOOP_UV pUV[4*4+3][uv] = UVOffset10[uv] +  UVCenterBck[uv];
	LOOP_UV pUV[5*4+0][uv] = UVOffset01[uv] +  UVCenterFrt[uv];
	LOOP_UV pUV[5*4+1][uv] = UVOffset11[uv] +  UVCenterFrt[uv];
	LOOP_UV pUV[5*4+2][uv] = UVOffset10[uv] +  UVCenterFrt[uv];
	LOOP_UV pUV[5*4+3][uv] = UVOffset00[uv] +  UVCenterFrt[uv];
	break; }
    }
}

void CLS::
SetGeometry(VBOBaseObject_cls *pCTgt){
    
    pCTgt->Clear();
    pCTgt->GenBuffersVertex(4);

    int face = 6;

    int totalVertex = GetTotalVertex(face);
    int totalTriangle = GetTotalTriangle(face);
    pCTgt->mTotalIndiceTriangle = totalTriangle*3;

    float (*pVertex)[3] = new float [totalVertex][3];
    float (*pNormal)[3] = new float [totalVertex][3];
    float (*pUValongX)[2] = new float [totalVertex][2];
    float (*pUValongY)[2] = new float [totalVertex][2];
    float (*pUValongZ)[2] = new float [totalVertex][2];
    float (*pUVUV)[2] = new float [totalVertex][2];
    
    unsigned int *pIndiceTriangle = pCTgt->GetPNewTriangle();
    
    ConstructTriangles(pVertex, pNormal, pIndiceTriangle);
    ConstructUV(pUValongX, 0);
    ConstructUV(pUValongY, 1);
    ConstructUV(pUValongZ, 2);
    ConstructUV(pUVUV, 3);

    pCTgt->MakeVBO(pVertex, pNormal, pUValongX, pUValongY, pUValongZ, pUVUV, totalVertex);
    
    pCTgt->mTotalIndiceAdjacent = totalTriangle*3*2;
    unsigned int *pIndiceAdjacent = pCTgt->GetPNewAdjucent();
    
    ConstructAdjStructure(pVertex, pIndiceTriangle, pIndiceAdjacent, totalTriangle);

    pCTgt->GenBuffersIndice();
    pCTgt->MakeVBOIndicesTriangle(pIndiceTriangle);
    pCTgt->MakeVBOIndicesAdjacent(pIndiceAdjacent);
    

    delete [] pVertex;
    delete [] pNormal;
    delete [] pUValongX;
    delete [] pUValongY;
    delete [] pUValongZ;
    delete [] pUVUV;

    delete [] pIndiceTriangle;
    delete [] pIndiceAdjacent;

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

}

VBOCube_cls *GpCVBOCube = NULL;
