// -*- c -*-
// "clsVBOCylinder.h"
// (setq c-basic-offset 4)

#include "clsVBOCylinder.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
#endif

#define CLS VBOCylinder_cls

CLS::
VBOCylinder_cls(){
}

void CLS::
Initialize(){
    SetGeometry(&mCObject);
}

void CLS::
Draw(int objectState, int mapType){
    //mapType = 0;
    mCObject.Draw(objectState, mapType);
}

int CLS::
GetTotalVertex(int face){
    return (face+1)*2 + (face+1)*2;
}

int CLS::
GetTotalTriangle(int face){
    return face*2 + (face-2)*2;
}

void CLS::
ConstructTriangles(float (*pVertex)[3], float (*pNormal)[3], 
		   float (*pUValongX)[2], float (*pUValongY)[2], 
		   float (*pUValongZ)[2], float (*pUVUV)[2],
		   unsigned int *pIndiceTriangle, int face){

    float factor = M_PI * 2 / face;

    int totalVertex = GetTotalVertex(face);
    int totalTriangle = GetTotalTriangle(face);

    int vertexOffsetBTM = face*2 + 2;
    int vertexOffsetTOP = face*3 + 3;
    int indiceOffsetBTM = (face*2)*3;
    int indiceOffsetTOP = (face*2+(face-2))*3;

    // ǖ
    for (int i = 0; i <= face; i++){
	float vert0[3] = {cos(factor*i), sin(factor*i), -1};
	float normal0[3] = {cos(factor*i), sin(factor*i), 0};
	LOOP_XYZ pVertex[i*2][xyz] = vert0[xyz];
	LOOP_XYZ pNormal[i*2][xyz] = normal0[xyz];
	
	
	float vert1[3] = {cos(factor*i), sin(factor*i), +1};
	float normal1[3] = {cos(factor*i), sin(factor*i), 0};
	LOOP_XYZ pVertex[i*2+1][xyz] = vert1[xyz];
	LOOP_XYZ pNormal[i*2+1][xyz] = normal1[xyz];
	
	if (pUValongX != NULL){
	    pUValongX[i*2][0] = 0.5*vert0[Y]+0.5;
	    pUValongX[i*2][1] = 0.5*vert0[Z]+0.5; 
	    pUValongY[i*2][0] = 0.5*vert0[X]+0.5;
	    pUValongY[i*2][1] = 0.5*vert0[Z]+0.5;
	    pUValongZ[i*2][0] = 0.5*vert0[X]+0.5; 
	    pUValongZ[i*2][1] = 0.5*vert0[Y]+0.5; 
	    pUVUV[i*2][0] = factor*i/M_PI/2;
	    pUVUV[i*2][1] = 0.5*vert0[Z] + 0.5;
	    
	    pUValongX[i*2+1][0] = 0.5*vert1[Y]+0.5; 
	    pUValongX[i*2+1][1] = 0.5*vert1[Z]+0.5; 
	    pUValongY[i*2+1][0] = 0.5*vert1[X]+0.5;
	    pUValongY[i*2+1][1] = 0.5*vert1[Z]+0.5; 
	    pUValongZ[i*2+1][0] = 0.5*vert1[X]+0.5; 
	    pUValongZ[i*2+1][1] = 0.5*vert1[Y]+0.5; 
	    pUVUV[i*2+1][0] = factor*i/M_PI/2;
	    pUVUV[i*2+1][1] = 0.5*vert1[Z] + 0.5;
	}
    }

    // 
    for (int i = 0; i <= face; i++){
	float vert0[3] = {cos(factor*i), -sin(factor*i), -1};
	float normal0[3] = {0, 0, -1};
	int index = vertexOffsetBTM + i;
	LOOP_XYZ pVertex[index][xyz] = vert0[xyz];
	LOOP_XYZ pNormal[index][xyz] = normal0[xyz];
	if (pUValongX != NULL){
	    pUValongX[index][0] = 0.5*vert0[Y]+0.5; 
	    pUValongX[index][1] = 0.5*vert0[Z]+0.5; 
	    pUValongY[index][0] = 0.5*vert0[X]+0.5; 
	    pUValongY[index][1] = 0.5*vert0[Z]+0.5; 
	    pUValongZ[index][0] = 0.5*vert0[X]+0.5; 
	    pUValongZ[index][1] = 0.5*vert0[Y]+0.5; 
	    pUVUV[index][0] = 0;//factor*i/M_PI/2;
	    pUVUV[index][1] = 0.5*vert0[Z] + 0.5;
	}
    }

    // 
    for (int i = 0; i <= face; i++){
	float vert0[3] = {cos(factor*i), sin(factor*i), +1};
	float normal0[3] = {0, 0, +1};
	int index = vertexOffsetTOP + i;
	LOOP_XYZ pVertex[index][xyz] = vert0[xyz];
	LOOP_XYZ pNormal[index][xyz] = normal0[xyz];
	if (pUValongX != NULL){
	    pUValongX[index][0] = 0.5*vert0[Y]+0.5;
	    pUValongX[index][1] = 0.5*vert0[Z]+0.5; 
	    pUValongY[index][0] = 0.5*vert0[X]+0.5;
	    pUValongY[index][1] = 0.5*vert0[Z]+0.5; 
	    pUValongZ[index][0] = 0.5*vert0[X]+0.5;
	    pUValongZ[index][1] = 0.5*vert0[Y]+0.5; 
	    pUVUV[index][0] = 1;//factor*i/M_PI/2;
	    pUVUV[index][1] = 0.5*vert0[Z] + 0.5;
	}
    }

    // ǖ
    for (int i = 0; i < face; i++){
	pIndiceTriangle[i*6+0] = (i*2+1)     ;//% (face * 2);
	pIndiceTriangle[i*6+1] = (i*2)       ;//% (face * 2);
	pIndiceTriangle[i*6+2] = ((i+1)*2)   ;//% (face * 2);
	pIndiceTriangle[i*6+3] = (i*2+1)     ;//% (face * 2);
	pIndiceTriangle[i*6+4] = ((i+1)*2)   ;//% (face * 2);
	pIndiceTriangle[i*6+5] = ((i+1)*2+1) ;//% (face * 2);
    }

    // 
    for (int i = 0; i < face-2; i++){
	pIndiceTriangle[indiceOffsetBTM+i*3+0] = vertexOffsetBTM + 0;
	pIndiceTriangle[indiceOffsetBTM+i*3+1] = vertexOffsetBTM + i+1;
	pIndiceTriangle[indiceOffsetBTM+i*3+2] = vertexOffsetBTM + i+2;
    }

    // 
    for (int i = 0; i < face-2; i++){
	pIndiceTriangle[indiceOffsetTOP+i*3+0] = vertexOffsetTOP + 0;
	pIndiceTriangle[indiceOffsetTOP+i*3+1] = vertexOffsetTOP + i+1;
	pIndiceTriangle[indiceOffsetTOP+i*3+2] = vertexOffsetTOP + i+2;
    }
}

#define LOOP_UV for (int uv = 0; uv < 2; uv++)

void CLS::
SetGeometry(VBOBaseObject_cls *pCTgt){

    pCTgt->Clear();
    pCTgt->GenBuffersVertex(4);

    int face = 48;
    float factor = M_PI * 2 / face;

    int totalVertex = GetTotalVertex(face);
    int totalTriangle = GetTotalTriangle(face);
    pCTgt->mTotalIndiceTriangle = totalTriangle*3;

    float (*pVertex)[3] = new float [totalVertex][3];
    float (*pNormal)[3] = new float [totalVertex][3];
    float (*pUValongX)[2] = new float [totalVertex][2];
    float (*pUValongY)[2] = new float [totalVertex][2];
    float (*pUValongZ)[2] = new float [totalVertex][2];
    float (*pUVUV)[2] = new float [totalVertex][2];
    unsigned int *pIndiceTriangle = new unsigned int [pCTgt->mTotalIndiceTriangle];
    
    ConstructTriangles(pVertex, pNormal, pUValongX, pUValongY, pUValongZ, pUVUV,
		       pIndiceTriangle, face);

    pCTgt->MakeVBO(pVertex, pNormal, pUValongX, pUValongY, pUValongZ, pUVUV, totalVertex);
    
    pCTgt->mTotalIndiceAdjacent = totalTriangle*3*2;
    unsigned int *pIndiceAdjacent = pCTgt->GetPNewAdjucent();

    ConstructAdjStructure(pVertex, pIndiceTriangle, pIndiceAdjacent, totalTriangle);
    
    pCTgt->GenBuffersIndice();
    pCTgt->MakeVBOIndicesTriangle(pIndiceTriangle);
    pCTgt->MakeVBOIndicesAdjacent(pIndiceAdjacent);
    
    delete [] pVertex;
    delete [] pNormal;
    delete [] pUValongX;
    delete [] pUValongY;
    delete [] pUValongZ;
    delete [] pUVUV;
    delete [] pIndiceTriangle;
    delete [] pIndiceAdjacent;
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

CLS *GpCVBOCylinder = NULL;
