// -*- c -*-
// "clsVBOPlane.h"
// (setq c-basic-offset 4)

#include "clsVBOPlane.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
#endif

#define CLS VBOPlane_cls

CLS::
VBOPlane_cls(){
}

void CLS::
Initialize(){
    SetGeometry(&mCObject);
}

void CLS::
Draw(int objectState, int mapType){
    mapType = 0;
    mCObject.Draw(objectState, mapType);
}

int CLS::
GetTotalVertex(int face){
    return face*4;
}

int CLS::
GetTotalTriangle(int face){
    return face*2;
}

void CLS::
ConstructTriangles(float (*pVertex)[3], float (*pNormal)[3], 
		   unsigned int *pIndiceTriangle){

    float vert00[3] = {-1, -1, 0};
    float vert01[3] = {+1, -1, 0};
    float vert10[3] = {-1, +1, 0};
    float vert11[3] = {+1, +1, 0};
    
    float normalzp[3] = {0, 0, +1};

    LOOP_XYZ pVertex[0*4+0][xyz] = vert00[xyz]; 
    LOOP_XYZ pVertex[0*4+1][xyz] = vert01[xyz]; 
    LOOP_XYZ pVertex[0*4+2][xyz] = vert11[xyz]; 
    LOOP_XYZ pVertex[0*4+3][xyz] = vert10[xyz]; 
    for (int i = 0; i < 4; i++)
	LOOP_XYZ pNormal[0*4+i][xyz] = normalzp[xyz];

    int triangles[6] = {0, 1, 2, 2, 3, 0};
    for (int j = 0; j < 1; j++)
	for (int i = 0; i < 6; i++) 
	    pIndiceTriangle[j*6+i] = 4*j + triangles[i]; 
}

#define LOOP_UV for (int uv = 0; uv < 2; uv++)

void CLS::
ConstructUV(float (*pUV)[2]){

    float UV00[2] = {0, 0};
    float UV01[2] = {1, 0};
    float UV10[2] = {0, 1};
    float UV11[2] = {1, 1};
    
    LOOP_UV pUV[0*4+0][uv] = UV00[uv]; 
    LOOP_UV pUV[0*4+1][uv] = UV01[uv];
    LOOP_UV pUV[0*4+2][uv] = UV11[uv];
    LOOP_UV pUV[0*4+3][uv] = UV10[uv];
}

void CLS::
SetGeometry(VBOBaseObject_cls *pCTgt){
    
    pCTgt->Clear();
    pCTgt->GenBuffersVertex(1);

    int face = 1;

    int totalVertex = GetTotalVertex(face);
    int totalTriangle = GetTotalTriangle(face);
    pCTgt->mTotalIndiceTriangle = totalTriangle*3;

    float (*pVertex)[3] = new float [totalVertex][3];
    float (*pNormal)[3] = new float [totalVertex][3];
    float (*pUV)[2] = new float [totalVertex][2];
        
    unsigned int *pIndiceTriangle = new unsigned int [pCTgt->mTotalIndiceTriangle];
    
    ConstructTriangles(pVertex, pNormal, pIndiceTriangle);
    ConstructUV(pUV);

    pCTgt->MakeVBO(pVertex, pNormal, pUV, NULL, NULL, NULL, totalVertex);
        
    pCTgt->mTotalIndiceAdjacent = totalTriangle*3*2;
    unsigned int *pIndiceAdjacent = pCTgt->GetPNewAdjucent();
    
    ConstructAdjStructure(pVertex, pIndiceTriangle, pIndiceAdjacent, totalTriangle);
    
    pCTgt->GenBuffersIndice();
    pCTgt->MakeVBOIndicesTriangle(pIndiceTriangle);
    pCTgt->MakeVBOIndicesAdjacent(pIndiceAdjacent);

    delete [] pVertex;
    delete [] pNormal;
    delete [] pUV;

    delete [] pIndiceTriangle;
    delete [] pIndiceAdjacent;
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

CLS *GpCVBOPlane = NULL;


