// -*- c -*-
// "clsVBOSphere.h"
// (setq c-basic-offset 4)

#include "clsVBOSphere.h"
#include "clsVBOBaseObject.h"

//#include "../COMMON/MyVector.h"
#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
#endif

#include <math.h>
#include <vector>
#include <algorithm>

#define GLOBJ_SPHERE1 0.447213595f
#define GLOBJ_SPHERE2 0.894427190f
#define GLOBJ_SPHERE3 0.276393202f
#define GLOBJ_SPHERE4 0.723606797f
#define GLOBJ_SPHERE5 0.525731112f
#define GLOBJ_SPHERE6 0.850650808f

#ifndef M_PI
#define M_PI 3.14159
#endif 

#define LOOP_UV for (int uv = 0; uv < 2; uv++)

#define CLS VBOSphere_cls

CLS::
VBOSphere_cls(){
    GLfloat l_icosahedronv[12][3] = {
	{             0,             0,             1},
	{             0, GLOBJ_SPHERE2, GLOBJ_SPHERE1},
	{-GLOBJ_SPHERE6, GLOBJ_SPHERE3, GLOBJ_SPHERE1},
	{-GLOBJ_SPHERE5,-GLOBJ_SPHERE4, GLOBJ_SPHERE1},
	{ GLOBJ_SPHERE5,-GLOBJ_SPHERE4, GLOBJ_SPHERE1},
	{ GLOBJ_SPHERE6, GLOBJ_SPHERE3, GLOBJ_SPHERE1}, 
	{-GLOBJ_SPHERE5, GLOBJ_SPHERE4,-GLOBJ_SPHERE1},
	{-GLOBJ_SPHERE6, GLOBJ_SPHERE3,-GLOBJ_SPHERE1},
	{             0,-GLOBJ_SPHERE2,-GLOBJ_SPHERE1},
	{ GLOBJ_SPHERE6,-GLOBJ_SPHERE3,-GLOBJ_SPHERE1},
	{ GLOBJ_SPHERE5, GLOBJ_SPHERE4,-GLOBJ_SPHERE1},
	{             0,             0,            -1} };

    GLint l_icosahedronface[20][3] = {
	{1,2,0},{2,3,0},{3,4,0},{4,5,0},{5,1,0},
	{1,6,2},{2,6,7},{2,7,3},{3,7,8},{4,3,8},
	{4,8,9},{4,9,5},{5,9,10},{5,10,1},{1,10,6},
	{6,10,11},{7,6,11},{8,7,11},{9,8,11},{10,9,11} };

    for (int i = 0; i < 12; i++)
	LOOP_XYZ
	    mVerexIcosahedron[i][xyz] = l_icosahedronv[i][xyz];

    for (int i = 0; i < 20; i++)
	for (int j = 0; j < 3; j++)
	    mFaceIcosahedron[i][j] = l_icosahedronface[i][j];

    mDummytex2D[0] = 0;
    mDummytex2D[1] = 0;

}

void CLS::
Initialize(){
    for (int depth = 0; depth < 5; depth++)
	SetGeometry(&mCObject[depth], depth);
}

void CLS::
Draw(int objectState, int mapType, int depth){
    if (mapType > _UV_MAPPING_UV) mapType = _UV_MAPPING_UV;
    mCObject[depth].Draw(objectState, mapType);
}

void CLS::
DrawFast(int depth){ // 炦(PrepareClientStateς)
    mCObject[depth].DrawFast();
}

// ----

#include "../COMMON/clsArrayAndIndex.h"

int CLS::
SearchSameVertex(int ibase_, int iComp, ArrayAndIndex_cls* pCIdxAndZ,
		 float (*pVertex)[3]){
    int ini = ibase_ - 999;
    int fin = ibase_;
    if (ini < 0) ini = 0;
    
    for (int i_ = ini; i_ < fin; i_++){
	int i = pCIdxAndZ[i_].index;
	if (pVertex[i][0] == pVertex[iComp][0] &&
	    pVertex[i][1] == pVertex[iComp][1] &&
	    pVertex[i][2] == pVertex[iComp][2] 
	    ){
	    if (i < iComp) return i;
	    return iComp;
	}
    }
    return -1;
}

int CLS::
SearchSameVertex(int ibase_, int iComp, ArrayAndIndex_cls* pCIdxAndZ,
		 float (*pVertex)[3], 
		 float (*pUValongX)[2], float (*pUValongY)[2], 
		 float (*pUValongZ)[2], float (*pUVUV)[2]){
    int ini = ibase_ - 999;
    int fin = ibase_;
    if (ini < 0) ini = 0;
    
    for (int i_ = ini; i_ < fin; i_++){
	int i = pCIdxAndZ[i_].index;
	if (pVertex[i][0] == pVertex[iComp][0] &&
	    pVertex[i][1] == pVertex[iComp][1] &&
	    pVertex[i][2] == pVertex[iComp][2] &&
	    pUValongX[i][0] == pUValongX[iComp][0] &&
	    pUValongX[i][1] == pUValongX[iComp][1] &&
	    pUValongY[i][0] == pUValongY[iComp][0] &&
	    pUValongY[i][1] == pUValongY[iComp][1] &&
	    pUValongZ[i][0] == pUValongZ[iComp][0] &&
	    pUValongZ[i][1] == pUValongZ[iComp][1] &&
	    pUVUV[i][0] == pUVUV[iComp][0] &&
	    pUVUV[i][1] == pUVUV[iComp][1] ){
	    if (i < iComp) return i;
	    return iComp;
	}
    }
    return -1;
}

int CLS::
RemoveDouble(unsigned int *pIndiceTriangle, 
	     int &numIndicesTriangle, int &numVertex, 
	     float (*pVertex)[3], float (*pNormal)[3], 
	     float (*pUValongX)[2], float (*pUValongY)[2], 
	     float (*pUValongZ)[2], float (*pUVUV)[2]){

    ArrayAndIndex_cls *pCIdxAndZ = new ArrayAndIndex_cls [numVertex];
    for (int i = 0; i < numVertex; i++){
	pCIdxAndZ[i].index = i;
	pCIdxAndZ[i].z = pVertex[i][2];
    }

    float (*pVertexCache)[3] = new float [numVertex][3];
    float (*pNormalCache)[3] = new float [numVertex][3];
    for (int i = 0; i < numVertex; i++){
	LOOP_XYZ pVertexCache[i][xyz] = pVertex[i][xyz];
	LOOP_XYZ pNormalCache[i][xyz] = pNormal[i][xyz];
    }

    float (*pUValongXCache)[2] = NULL;
    float (*pUValongYCache)[2] = NULL;
    float (*pUValongZCache)[2] = NULL;
    float (*pUVUVCache)[2] = NULL;
    if (pUValongX != NULL){
	pUValongXCache = new float [numVertex][2];
	pUValongYCache = new float [numVertex][2];
	pUValongZCache = new float [numVertex][2];
	pUVUVCache = new float [numVertex][2];
	for (int i = 0; i < numVertex; i++){
	    LOOP_UV pUValongXCache[i][uv] = pUValongX[i][uv];
	    LOOP_UV pUValongYCache[i][uv] = pUValongY[i][uv];
	    LOOP_UV pUValongZCache[i][uv] = pUValongZ[i][uv];
	    LOOP_UV pUVUVCache[i][uv] = pUVUVCache[i][uv];
	}
    }
    
    std::sort(pCIdxAndZ, pCIdxAndZ+numVertex);

    std::vector < int > VvertexTable;
    VvertexTable.resize(numVertex);
    
    // ̂̃Xg
    if (pUValongX != NULL)
	for (int i_ = 0; i_ < numVertex; i_++){
	    int index = pCIdxAndZ[i_].index;
	    int same = SearchSameVertex(i_, index, pCIdxAndZ, pVertex,
					pUValongX, pUValongY, pUValongZ,
					pUVUV);
	    if (same == -1) same = index;
	    VvertexTable[index] = same;
	}
    else
	for (int i_ = 0; i_ < numVertex; i_++){
	    int index = pCIdxAndZ[i_].index;
	    int same = SearchSameVertex(i_, index, pCIdxAndZ, pVertex);
	    if (same == -1) same = index;
	    VvertexTable[index] = same;
	}

    std::vector < int > VvertexTableOrg = VvertexTable;

    // XgAgĂȂԍ͂Ȃ̂ŁA
    // ߂ȂƂȂ
    std::vector < int > VvertexTableSorted = VvertexTable;
    std::sort(VvertexTableSorted.begin(), VvertexTableSorted.end());
    std::vector < int > VvertexTableSortedOrg = VvertexTableSorted;

    // ԍl߂
    int indexFilled = 0;
    for (int i = 0; i < numVertex; i++){
	int indexOrg = VvertexTableSortedOrg[i];
	if (VvertexTableSorted[i] > indexFilled){
	    int offset = VvertexTableSorted[i] - indexFilled;
	    for (int j = i; j < numVertex; j++)
		VvertexTableSorted[j] -= offset;
	    for (int j = 0; j < numVertex; j++)
		if (VvertexTable[j] > indexFilled)
		    VvertexTable[j] -= offset;
	}

	if (VvertexTableSorted[i] < indexFilled)
	    indexFilled = VvertexTableSorted[i];

	indexFilled ++;
    }

    for (int i = 0; i < numVertex; i++){
	int newIndex = VvertexTable[i];
	for (int j = 0; j < numIndicesTriangle; j++){
	    if (pIndiceTriangle[j] == i) pIndiceTriangle[j] = newIndex;
	}
	LOOP_XYZ pVertex[newIndex][xyz] = pVertexCache[i][xyz];
	LOOP_XYZ pNormal[newIndex][xyz] = pNormalCache[i][xyz];
    }
    if (pUValongX != NULL)
	for (int i = 0; i < numVertex; i++){
	    int newIndex = VvertexTable[i];
	    LOOP_UV pUValongX[newIndex][uv] = pUValongX[i][uv];
	    LOOP_UV pUValongY[newIndex][uv] = pUValongY[i][uv];
	    LOOP_UV pUValongZ[newIndex][uv] = pUValongZ[i][uv];
	    LOOP_UV pUVUV[newIndex][uv] = pUVUV[i][uv];
	}
    
    delete [] pCIdxAndZ;
    delete [] pVertexCache;
    delete [] pNormalCache;
    if (pUValongXCache != NULL){
	delete [] pUValongXCache;
	delete [] pUValongYCache;
	delete [] pUValongZCache;
	delete [] pUVUVCache;
    }

    return indexFilled;
}

/*!
  @brief XYZW theta-phi Wɕϊ
*/
void CLS::
ChangeCoordinate3Dto2D(float vec3D[3], float vec2D[2]){
    int phi = 0, theta = 1;
    vec2D[phi] = atan2(vec3D[Y], +vec3D[X]);
    vec2D[phi] = (vec2D[phi] + M_PI)/(2.0 * M_PI);
    vec2D[theta] = atan2(vec3D[Z], 
			 sqrt(vec3D[X]*vec3D[X]+vec3D[Y]*vec3D[Y]));
    vec2D[theta] = (vec2D[theta]+M_PI/2.0)/M_PI;
}

/*!
  @brief x[0,1] y[0,1] ͈̔͂͂ݏoW̏C
*/
void CLS::
AdjustVec2DOverBoundary(float vector[2], float vectormem[2]){
    LOOP_UV{
	while (vector[uv]-vectormem[uv] > 0.5)
	    vector[uv] = vector[uv] - 1.0f;
	while (vector[uv]-vectormem[uv] < -0.5)
	    vector[uv] = vector[uv] + 1.0f; 
    }
}

int CLS::
GetTotalVertexCandid(int depth){
    int vertex = 12;
    int currentDepth = 0;
    while (currentDepth < depth){
	int face = GetTotalTriangle(currentDepth);
	// 1ɂ_3
	// 肪̂ŔɂȂ͂
	vertex += face * 1.5;
	currentDepth++;
    }
    return vertex * 2;
}

int CLS::
GetTotalTriangle(int depth){
    int face = 20;
    int currentDepth = 0;
    while (currentDepth < depth){
	// ʂ4{ɑ
	face *= 4;
	currentDepth++;
    }
    return face;
}

void CLS::
SetTriangle(unsigned int *pIndiceTriangle, int &currentFace,
	    int vertex1, int vertex2, int vertex3){
    pIndiceTriangle[currentFace*3+0] = vertex1;
    pIndiceTriangle[currentFace*3+1] = vertex2;
    pIndiceTriangle[currentFace*3+2] = vertex3;
    currentFace ++;
    return;
}

int CLS::
GetVertex(int currentVertex, float (*pVertex)[3], float vertex[3],
	  float (*pUVUV)[2], float texCoord[2]){
    float d = 1e-5;
    if (pUVUV == NULL)
	for (int i = 0; i < currentVertex; i++){
	    if (pVertex[i][X] < vertex[X]+d && pVertex[i][X] > vertex[X]-d &&
		pVertex[i][Y] < vertex[Y]+d && pVertex[i][Y] > vertex[Y]-d &&
		pVertex[i][Z] < vertex[Z]+d && pVertex[i][Z] > vertex[Z]-d)
		return i;
	}
    else
	for (int i = 0; i < currentVertex; i++){
	    if (pVertex[i][X] < vertex[X]+d && pVertex[i][X] > vertex[X]-d &&
		pVertex[i][Y] < vertex[Y]+d && pVertex[i][Y] > vertex[Y]-d &&
		pVertex[i][Z] < vertex[Z]+d && pVertex[i][Z] > vertex[Z]-d &&
		pUVUV[i][0] < texCoord[0]+d && pUVUV[i][0] > texCoord[0]-d &&
		pUVUV[i][1] < texCoord[1]+d && pUVUV[i][1] > texCoord[1]-d)
		return i;
	}

    return -1;
}

void CLS::
DivideTriangle(float (*pVertex)[3], float (*pNormal)[3], 
	       float (*pUValongX)[2], float (*pUValongY)[2], 
	       float (*pUValongZ)[2], float (*pUVUV)[2],
	       unsigned int *pIndiceTriangle, 
	       int depth, int &currentVertex, int &currentFace,
	       int vertex1, int vertex2, int vertex3){
	       
    if (depth <= 0){
	SetTriangle(pIndiceTriangle, currentFace, 
		    vertex1, vertex2, vertex3);
	return;
    }

    float v1[3], v2[3], v3[3];
    LOOP_XYZ v1[xyz] = pVertex[vertex1][xyz];
    LOOP_XYZ v2[xyz] = pVertex[vertex2][xyz];
    LOOP_XYZ v3[xyz] = pVertex[vertex3][xyz];
    
    float texCoord1[2], texCoord2[2], texCoord3[2];
    ChangeCoordinate3Dto2D(v1, texCoord1);
    ChangeCoordinate3Dto2D(v2, texCoord2);
    ChangeCoordinate3Dto2D(v3, texCoord3);

    float texCoordMod1[2], texCoordMod2[2], texCoordMod3[2];
    LOOP_UV texCoordMod1[uv] = texCoord1[uv];
    LOOP_UV texCoordMod2[uv] = texCoord2[uv];
    LOOP_UV texCoordMod3[uv] = texCoord3[uv];

    AdjustVec2DOverBoundary(texCoordMod2, texCoordMod1);
    if (texCoordMod3[Y] > 0.99 || texCoordMod3[Y] < 0.01) 
	texCoordMod3[X] = (texCoordMod1[X] + texCoordMod2[X])/2.0f; 
    else 
	AdjustVec2DOverBoundary(texCoordMod3, texCoordMod1);
    
    vertex1 = GetVertex(currentVertex, pVertex, v1, pUVUV, texCoordMod1);
    if (vertex1 == -1){
	LOOP_XYZ pVertex[currentVertex][xyz] = v1[xyz];
	LOOP_XYZ pNormal[currentVertex][xyz] = v1[xyz];
	if (pUValongX != NULL){
	    pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
	    pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
	    pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
	    pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
	    pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
	    pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
	    //ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
	    pUVUV[currentVertex][0] = texCoordMod1[0];
	    pUVUV[currentVertex][1] = texCoordMod1[1];
	}
	vertex1 = currentVertex;
	currentVertex++;
    }
    
    vertex2 = GetVertex(currentVertex, pVertex, v2, pUVUV, texCoordMod2);
    if (vertex2 == -1){
	LOOP_XYZ pVertex[currentVertex][xyz] = v2[xyz];
	LOOP_XYZ pNormal[currentVertex][xyz] = v2[xyz];
	if (pUValongX != NULL){
	    pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
	    pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
	    pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
	    pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
	    pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
	    pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
	    //ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
	    pUVUV[currentVertex][0] = texCoordMod2[0];
	    pUVUV[currentVertex][1] = texCoordMod2[1];
	}
	vertex2 = currentVertex;
	currentVertex++;
    }
    
    vertex3 = GetVertex(currentVertex, pVertex, v3, pUVUV, texCoordMod3);
    if (vertex3 == -1){
	LOOP_XYZ pVertex[currentVertex][xyz] = v3[xyz];
	LOOP_XYZ pNormal[currentVertex][xyz] = v3[xyz];
	if (pUValongX != NULL){
	    pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
	    pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
	    pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
	    pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
	    pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
	    pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
	    //ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
	    pUVUV[currentVertex][0] = texCoordMod3[0];
	    pUVUV[currentVertex][1] = texCoordMod3[1];
	}
	vertex3 = currentVertex;
	currentVertex++;
    }
    
    float v12[3], v23[3], v31[3];
    LOOP_XYZ v12[xyz] = (v1[xyz] + v2[xyz]);
    LOOP_XYZ v23[xyz] = (v2[xyz] + v3[xyz]);
    LOOP_XYZ v31[xyz] = (v3[xyz] + v1[xyz]); 
    
    Normalize(v12, 1.0f);
    Normalize(v23, 1.0f);
    Normalize(v31, 1.0f);

    float texCoord12[2], texCoord23[2], texCoord31[2];
    ChangeCoordinate3Dto2D(v12, texCoord12);
    ChangeCoordinate3Dto2D(v23, texCoord23);
    ChangeCoordinate3Dto2D(v31, texCoord31);
    float texCoordMod12[2], texCoordMod23[2], texCoordMod31[2];
    LOOP_UV texCoordMod12[uv] = texCoord12[uv];
    LOOP_UV texCoordMod23[uv] = texCoord23[uv];
    LOOP_UV texCoordMod31[uv] = texCoord31[uv];

    AdjustVec2DOverBoundary(texCoordMod12, texCoordMod1);
    AdjustVec2DOverBoundary(texCoordMod23, texCoordMod1);
    if (texCoordMod31[Y] > 0.99 || texCoordMod31[Y] < 0.01) 
	texCoordMod31[X] = (texCoordMod12[X] + texCoordMod23[X])/2.0f; 
    else 
	AdjustVec2DOverBoundary(texCoordMod31, texCoordMod1);
    
    int vertex12 = GetVertex(currentVertex, pVertex, v12, pUVUV, texCoordMod12);
    if (vertex12 == -1){
	LOOP_XYZ pVertex[currentVertex][xyz] = v12[xyz];
	LOOP_XYZ pNormal[currentVertex][xyz] = v12[xyz];
	if (pUValongX != NULL){
	    pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
	    pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
	    pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
	    pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
	    pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
	    pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
	    //ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
	    pUVUV[currentVertex][0] = texCoordMod12[0];
	    pUVUV[currentVertex][1] = texCoordMod12[1];
	}
	vertex12 = currentVertex;
	currentVertex++;
    }

    int vertex23 = GetVertex(currentVertex, pVertex, v23, pUVUV, texCoordMod23);
    if (vertex23 == -1){
	LOOP_XYZ pVertex[currentVertex][xyz] = v23[xyz];
	LOOP_XYZ pNormal[currentVertex][xyz] = v23[xyz];
	if (pUValongX != NULL){
	    pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
	    pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
	    pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
	    pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
	    pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
	    pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
	    //ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
	    pUVUV[currentVertex][0] = texCoordMod23[0];
	    pUVUV[currentVertex][1] = texCoordMod23[1];
	}
	vertex23 = currentVertex;
	currentVertex++;
    }

    int vertex31 = GetVertex(currentVertex, pVertex, v31, pUVUV, texCoordMod31);
    if (vertex31 == -1){
	LOOP_XYZ pVertex[currentVertex][xyz] = v31[xyz];
	LOOP_XYZ pNormal[currentVertex][xyz] = v31[xyz];
	if (pUValongX != NULL){
	    pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
	    pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
	    pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
	    pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
	    pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
	    pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
	    //ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
	    pUVUV[currentVertex][0] = texCoordMod31[0];
	    pUVUV[currentVertex][1] = texCoordMod31[1];
	}
	vertex31 = currentVertex;
	currentVertex++;
    }

    DivideTriangle(pVertex, pNormal, 
		   pUValongX, pUValongY, pUValongZ, pUVUV, 
		   pIndiceTriangle, depth-1, currentVertex, currentFace, 
		   vertex1, vertex12, vertex31);
    DivideTriangle(pVertex, pNormal, 
		   pUValongX, pUValongY, pUValongZ, pUVUV, 
		   pIndiceTriangle, depth-1, currentVertex, currentFace, 
		   vertex2, vertex23, vertex12);
    DivideTriangle(pVertex, pNormal, 
		   pUValongX, pUValongY, pUValongZ, pUVUV, 
		   pIndiceTriangle, depth-1, currentVertex, currentFace, 
		   vertex12, vertex23, vertex31);
    DivideTriangle(pVertex, pNormal, 
		   pUValongX, pUValongY, pUValongZ, pUVUV, 
		   pIndiceTriangle, depth-1, currentVertex, currentFace, 
		   vertex31, vertex23, vertex3);
}

int CLS::
ConstructTriangles(float (*pVertex)[3], float (*pNormal)[3], 
		   float (*pUValongX)[2], float (*pUValongY)[2], 
		   float (*pUValongZ)[2], float (*pUVUV)[2],
		   unsigned int *pIndiceTriangle, int depth){

    int currentVertex = 12;
    int currentFace = 0;
    
    // ŏ12_
    for (int i = 0; i <12; i++){
	LOOP_XYZ pVertex[i][xyz] = mVerexIcosahedron[i][xyz];
	LOOP_XYZ pNormal[i][xyz] = pVertex[i][xyz];
	if (pUValongX != NULL){
	    pUValongX[i][0] = 0.5*pVertex[i][Y]+0.5;
	    pUValongX[i][1] = 0.5*pVertex[i][Z]+0.5; 
	    pUValongY[i][0] = 0.5*pVertex[i][X]+0.5;
	    pUValongY[i][1] = 0.5*pVertex[i][Z]+0.5;
	    pUValongZ[i][0] = 0.5*pVertex[i][X]+0.5; 
	    pUValongZ[i][1] = 0.5*pVertex[i][Y]+0.5; 
	    ChangeCoordinate3Dto2D(pVertex[i], pUVUV[i]);
	}
    }

    // ŏ20|SX^[g
    for (int i = 0; i < 20; i++){
	int vertex1 = mFaceIcosahedron[i][0];
	int vertex2 = mFaceIcosahedron[i][1];
	int vertex3 = mFaceIcosahedron[i][2];

	float v1[3], v2[3], v3[3];
	LOOP_XYZ v1[xyz] = pVertex[vertex1][xyz];
	LOOP_XYZ v2[xyz] = pVertex[vertex2][xyz];
	LOOP_XYZ v3[xyz] = pVertex[vertex3][xyz];

	float texCoord1[2], texCoord2[2], texCoord3[2];
	ChangeCoordinate3Dto2D(v1, texCoord1);
	ChangeCoordinate3Dto2D(v2, texCoord2);
	ChangeCoordinate3Dto2D(v3, texCoord3);

	AdjustVec2DOverBoundary(texCoord2, texCoord1);
	if (texCoord3[Y] > 0.99 || texCoord3[Y] < 0.01) 
	    texCoord3[X] = (texCoord1[X] + texCoord2[X])/2.0f; 
	else 
	    AdjustVec2DOverBoundary(texCoord3, texCoord2);
	
	vertex1 = GetVertex(currentVertex, pVertex, v1, pUVUV, texCoord1);
	if (vertex1 == -1){
	    LOOP_XYZ pVertex[currentVertex][xyz] = v1[xyz];
	    LOOP_XYZ pNormal[currentVertex][xyz] = v1[xyz];
	    if (pUValongX != NULL){
		pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
		pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
		pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
		pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
		pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
		pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
		//ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
		pUVUV[currentVertex][0] = texCoord1[0];
		pUVUV[currentVertex][1] = texCoord1[1];
	    }
	    vertex1 = currentVertex;
	    currentVertex++;
	}

	vertex2 = GetVertex(currentVertex, pVertex, v2, pUVUV, texCoord2);
	if (vertex2 == -1){
	    LOOP_XYZ pVertex[currentVertex][xyz] = v2[xyz];
	    LOOP_XYZ pNormal[currentVertex][xyz] = v2[xyz];
	    if (pUValongX != NULL){
		pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
		pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
		pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
		pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
		pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
		pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
		//ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
		pUVUV[currentVertex][0] = texCoord2[0];
		pUVUV[currentVertex][1] = texCoord2[1];
	    }
	    vertex2 = currentVertex;
	    currentVertex++;
	}

	vertex3 = GetVertex(currentVertex, pVertex, v3, pUVUV, texCoord3);
	if (vertex3 == -1){
	    LOOP_XYZ pVertex[currentVertex][xyz] = v3[xyz];
	    LOOP_XYZ pNormal[currentVertex][xyz] = v3[xyz];
	    if (pUValongX != NULL){
		pUValongX[currentVertex][0] = 0.5*pVertex[currentVertex][Y]+0.5;
		pUValongX[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5; 
		pUValongY[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5;
		pUValongY[currentVertex][1] = 0.5*pVertex[currentVertex][Z]+0.5;
		pUValongZ[currentVertex][0] = 0.5*pVertex[currentVertex][X]+0.5; 
		pUValongZ[currentVertex][1] = 0.5*pVertex[currentVertex][Y]+0.5; 
		//ChangeCoordinate3Dto2D(pVertex[currentVertex], pUVUV[currentVertex]);
		pUVUV[currentVertex][0] = texCoord3[0];
		pUVUV[currentVertex][1] = texCoord3[1];
	    }
	    vertex3 = currentVertex;
	    currentVertex++;
	}
	
	DivideTriangle(pVertex, pNormal,
		       pUValongX, pUValongY, pUValongZ, pUVUV, 
		       pIndiceTriangle,
		       depth, currentVertex, currentFace, 
		       vertex1, vertex2, vertex3);
    }

    return currentVertex;
}

void CLS::
SetGeometry(VBOBaseObject_cls *pCTgt, int depth){
    
    pCTgt->Clear();
    pCTgt->GenBuffersVertex(4);

    //int depth = 4;

    int totalVertexCandid = GetTotalVertexCandid(depth);
    int totalTriangle = GetTotalTriangle(depth);
    pCTgt->mTotalIndiceTriangle = totalTriangle*3;

    float (*pVertex)[3] = new float [totalVertexCandid][3];
    float (*pNormal)[3] = new float [totalVertexCandid][3];
    float (*pUValongX)[2] = new float [totalVertexCandid][2];
    float (*pUValongY)[2] = new float [totalVertexCandid][2];
    float (*pUValongZ)[2] = new float [totalVertexCandid][2];
    float (*pUVUV)[2] = new float [totalVertexCandid][2];
    unsigned int *pIndiceTriangle = new unsigned int [pCTgt->mTotalIndiceTriangle];

    int vertexNum = ConstructTriangles(pVertex, pNormal, pUValongX, pUValongY, pUValongZ, pUVUV,
				       pIndiceTriangle, depth);

    int totalIndiceTriangle = totalTriangle*3;
    int newVertexNum = 
	RemoveDouble(pIndiceTriangle, totalIndiceTriangle, vertexNum,
		     pVertex, pNormal, 
		     pUValongX, pUValongY, pUValongZ, pUVUV);
    
    pCTgt->MakeVBO(pVertex, pNormal, pUValongX, pUValongY, pUValongZ, pUVUV, newVertexNum);

    pCTgt->mTotalIndiceAdjacent = totalTriangle*3*2;
    unsigned int *pIndiceAdjacent = pCTgt->GetPNewAdjucent();

    ConstructAdjStructure(pVertex, pIndiceTriangle, pIndiceAdjacent, totalTriangle);

    pCTgt->GenBuffersIndice();
    pCTgt->MakeVBOIndicesTriangle(pIndiceTriangle);
    pCTgt->MakeVBOIndicesAdjacent(pIndiceAdjacent);
    
    delete [] pVertex;
    delete [] pNormal;
    delete [] pUValongX;
    delete [] pUValongY;
    delete [] pUValongZ;
    delete [] pUVUV;
    delete [] pIndiceTriangle;
    delete [] pIndiceAdjacent;
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

VBOSphere_cls *GpCVBOSphere = NULL;
