// -*- c -*- 
// "clsShaderCtrl.h"

#include "clsShaderCtrl.h"

#include "../../GL/glsl.h"

#include "../COMMON/clsTensorMath.h"

#include "../COMMON/CommonDefinitions.h"

ExposureParams_cls::
ExposureParams_cls(){
    mMin = 0; mMax = 1;
    mGamma = 1;
    mHue = 0;
    mSaturation = 0;
}

ShaderParamEnsemble_cls::
ShaderParamEnsemble_cls(int ensemble, int dissolveMode,
			int lightNumber, int textureMode){
    mEnsemble = ensemble;
    mDissolveMode = dissolveMode;
    mLightNumber = lightNumber;
    mTextureMode = textureMode;
}

void ShaderParamEnsemble_cls::
CopyFrom(ShaderParamEnsemble_cls &CFrom){
    mEnsemble = CFrom.mEnsemble;
    mDissolveMode = CFrom.mDissolveMode;
    mLightNumber = CFrom.mLightNumber;
    mTextureMode = CFrom.mTextureMode;
}

StreamlineBase_cls::
StreamlineBase_cls(ShaderParamEnsemble_cls &CParamEnsemble){
    ShaderParamEnsemble_cls::CopyFrom(CParamEnsemble);
}

void StreamlineBase_cls::
CopyFrom(StreamlineBase_cls &CFrom){
    ShaderParamEnsemble_cls::CopyFrom(CFrom);
    mpClipBoxMin = CFrom.mpClipBoxMin;
    mpClipBoxMax = CFrom.mpClipBoxMax;
    mpMatrix16 = CFrom.mpMatrix16;
    mSmoothWidth = CFrom.mSmoothWidth;
    mThinningFactor = CFrom.mThinningFactor;
}

#define CLS ShaderCtrl_cls

CLS::
ShaderCtrl_cls(){
    mCurrentShader = 0;
    mFogMode = 0;

    ShaderItem_cls CClampAlpha("ClampAlpha");
    CClampAlpha.SetShaderPath("GL/ClampAlpha");
    CClampAlpha.mVstrUniform.push_back("textureImage");
    mVCShader.push_back(CClampAlpha);
    
    ShaderItem_cls CExposure("Exposure");
    CExposure.mVstrUniform.push_back("textureImage");
    CExposure.mVstrUniform.push_back("levelMin");
    CExposure.mVstrUniform.push_back("levelMax");
    CExposure.mVstrUniform.push_back("gamma");
    CExposure.mVstrUniform.push_back("hueOffset");
    CExposure.mVstrUniform.push_back("saturation");
    CExposure.mVstrUniform.push_back("blendMode");
    CExposure.mVstrUniform.push_back("alpha");
    CExposure.SetShaderPath("GL/Exposure");
    mVCShader.push_back(CExposure);

    ShaderItem_cls CBlur("Blur");
    CBlur.mVstrUniform.push_back("textureImage");
    CBlur.mVstrUniform.push_back("type");
    CBlur.mVstrUniform.push_back("kernel");
    CBlur.mVstrUniform.push_back("width");
    CBlur.mVstrUniform.push_back("quality");
    CBlur.mVstrUniform.push_back("gamma");
    CBlur.SetShaderPath("GL/Blur");
    mVCShader.push_back(CBlur);

    ShaderItem_cls CFocalBlur("FocalBlur");
    CFocalBlur.mVstrUniform.push_back("textureImage");
    CFocalBlur.mVstrUniform.push_back("textureImage1");
    CFocalBlur.mVstrUniform.push_back("quality");
    CFocalBlur.mVstrUniform.push_back("gamma");
    CFocalBlur.mVstrUniform.push_back("projMatrix10");
    CFocalBlur.mVstrUniform.push_back("projMatrix14");
    CFocalBlur.mVstrUniform.push_back("focalDistance");
    CFocalBlur.mVstrUniform.push_back("widthBase");
    CFocalBlur.SetShaderPath("GL/FocalBlur");
    mVCShader.push_back(CFocalBlur);

    ShaderItem_cls CFocalBlurB("FocalBlurB");
    CFocalBlurB.mVstrUniform.push_back("textureImage");
    CFocalBlurB.mVstrUniform.push_back("textureImage1");
    CFocalBlurB.mVstrUniform.push_back("quality");
    CFocalBlurB.mVstrUniform.push_back("gamma");
    CFocalBlurB.mVstrUniform.push_back("projMatrix10");
    CFocalBlurB.mVstrUniform.push_back("projMatrix14");
    CFocalBlurB.mVstrUniform.push_back("focalDistance");
    CFocalBlurB.mVstrUniform.push_back("widthBase");
    CFocalBlurB.SetShaderPath("GL/FocalBlurB");
    mVCShader.push_back(CFocalBlurB);
    

    ShaderItem_cls CGlare("Glare");
    CGlare.mVstrUniform.push_back("textureImage");
    CGlare.mVstrUniform.push_back("type");
    CGlare.mVstrUniform.push_back("width");
    CGlare.mVstrUniform.push_back("quality");
    CGlare.mVstrUniform.push_back("rayNum");
    CGlare.mVstrUniform.push_back("offset");
    CGlare.mVstrUniform.push_back("crit");
    CGlare.mVstrUniform.push_back("enhance");
    CGlare.mVstrUniform.push_back("fade");
    CGlare.SetShaderPath("GL/Glare");
    mVCShader.push_back(CGlare);

    ShaderItem_cls CRemoveAlpha("RemoveAlpha");
    CRemoveAlpha.mVstrUniform.push_back("textureImage");
    CRemoveAlpha.mVstrUniform.push_back("extremeAlpha");
    CRemoveAlpha.mVstrUniform.push_back("convert");
    CRemoveAlpha.mVstrUniform.push_back("backgroundTile");
    CRemoveAlpha.SetShaderPath("GL/RemoveAlpha");
    mVCShader.push_back(CRemoveAlpha);

    ShaderItem_cls COverlay("Overlay");
    COverlay.mVstrUniform.push_back("textureImage");
    COverlay.mVstrUniform.push_back("textureImage1");
    COverlay.mVstrUniform.push_back("texAlpha");
    COverlay.mVstrUniform.push_back("texXScale");
    COverlay.mVstrUniform.push_back("texYScale");
    COverlay.SetShaderPath("GL/Overlay");
    mVCShader.push_back(COverlay);

    ShaderItem_cls CSSAO("SSAO");
    CSSAO.mVstrUniform.push_back("textureImage");
    CSSAO.mVstrUniform.push_back("textureImage1");
    CSSAO.mVstrUniform.push_back("quality");
    CSSAO.mVstrUniform.push_back("projMatrix10");
    CSSAO.mVstrUniform.push_back("projMatrix14");
    CSSAO.mVstrUniform.push_back("factor");
    CSSAO.mVstrUniform.push_back("radius");
    CSSAO.mVstrUniform.push_back("baseColor");
    CSSAO.mVstrUniform.push_back("type");
    CSSAO.SetShaderPath("GL/SSAO");
    mVCShader.push_back(CSSAO);

    ShaderItem_cls CDepth("Depth");
    CDepth.mVstrUniform.push_back("textureImage");
    CDepth.mVstrUniform.push_back("projMatrix10");
    CDepth.mVstrUniform.push_back("projMatrix14");
    CDepth.SetShaderPath("GL/ConvertDepth");
    mVCShader.push_back(CDepth);

    ShaderItem_cls CVolumeBasic("VolumeBasic");
    CVolumeBasic.mVstrUniform.push_back("texturVolume");
    CVolumeBasic.mVstrUniform.push_back("textureCC");
    CVolumeBasic.mVstrUniform.push_back("clipBoxMin");
    CVolumeBasic.mVstrUniform.push_back("clipBoxMax");
    CVolumeBasic.mVstrUniform.push_back("forceDiscard0");
    CVolumeBasic.mVstrUniform.push_back("trim");
    CVolumeBasic.mVstrUniform.push_back("intervalGlobal");
    CVolumeBasic.mVstrUniform.push_back("subQuality");
    CVolumeBasic.mVstrUniform.push_back("blend");
    CVolumeBasic.SetShaderPath("GL/VolumeBasic");
    mVCShader.push_back(CVolumeBasic);

    ShaderItem_cls CVolumePhong("VolumePhong");
    CVolumePhong.mVstrUniform.push_back("texturVolume");
    CVolumePhong.mVstrUniform.push_back("textureCC");
    CVolumePhong.mVstrUniform.push_back("lightNumber");
    CVolumePhong.mVstrUniform.push_back("clipBoxMin");
    CVolumePhong.mVstrUniform.push_back("clipBoxMax");
    CVolumePhong.mVstrUniform.push_back("ambientClr");
    CVolumePhong.mVstrUniform.push_back("diffuseClr");
    CVolumePhong.mVstrUniform.push_back("specularClr");
    CVolumePhong.mVstrUniform.push_back("shininess");
    CVolumePhong.mVstrUniform.push_back("forceDiscard0");
    CVolumePhong.mVstrUniform.push_back("trim");
    CVolumePhong.mVstrUniform.push_back("intervalGlobal");
    CVolumePhong.mVstrUniform.push_back("subQuality");
    CVolumePhong.mVstrUniform.push_back("delta");
    CVolumePhong.mVstrUniform.push_back("blend");
    CVolumePhong.SetShaderPath("GL/VolumePhong");
    mVCShader.push_back(CVolumePhong);
    
    ShaderItem_cls CVolumePhongShadow("VolumePhongShadow");
    CVolumePhongShadow.mVstrUniform.push_back("texturVolume");
    CVolumePhongShadow.mVstrUniform.push_back("textureCC");
    CVolumePhongShadow.mVstrUniform.push_back("lightNumber");
    CVolumePhongShadow.mVstrUniform.push_back("shadowAlpha");
    CVolumePhongShadow.mVstrUniform.push_back("shadowQuality");
    CVolumePhongShadow.mVstrUniform.push_back("clipBoxMin");
    CVolumePhongShadow.mVstrUniform.push_back("clipBoxMax");
    CVolumePhongShadow.mVstrUniform.push_back("ambientClr");
    CVolumePhongShadow.mVstrUniform.push_back("diffuseClr");
    CVolumePhongShadow.mVstrUniform.push_back("specularClr");
    CVolumePhongShadow.mVstrUniform.push_back("shininess");
    CVolumePhongShadow.mVstrUniform.push_back("forceDiscard0");
    CVolumePhongShadow.mVstrUniform.push_back("trim");
    CVolumePhongShadow.mVstrUniform.push_back("intervalGlobal");
    CVolumePhongShadow.mVstrUniform.push_back("subQuality");
    CVolumePhongShadow.mVstrUniform.push_back("delta");
    CVolumePhongShadow.mVstrUniform.push_back("blend");
    CVolumePhongShadow.SetShaderPath("GL/VolumePhongShadow");
    mVCShader.push_back(CVolumePhongShadow);

    ShaderItem_cls CVolumeScatter("VolumeScatter");
    CVolumeScatter.mVstrUniform.push_back("texturVolume");
    CVolumeScatter.mVstrUniform.push_back("textureCC");
    CVolumeScatter.mVstrUniform.push_back("lightNumber");
    CVolumeScatter.mVstrUniform.push_back("clipBoxMin");
    CVolumeScatter.mVstrUniform.push_back("clipBoxMax");
    CVolumeScatter.mVstrUniform.push_back("ambientClr");
    CVolumeScatter.mVstrUniform.push_back("diffuseClr");
    CVolumeScatter.mVstrUniform.push_back("specularClr");
    CVolumeScatter.mVstrUniform.push_back("shininess");
    CVolumeScatter.mVstrUniform.push_back("scatterFactor");
    CVolumeScatter.mVstrUniform.push_back("forceDiscard0");
    CVolumeScatter.mVstrUniform.push_back("trim");
    CVolumeScatter.mVstrUniform.push_back("intervalGlobal");
    CVolumeScatter.mVstrUniform.push_back("subQuality");
    CVolumeScatter.mVstrUniform.push_back("blend");
    CVolumeScatter.SetShaderPath("GL/VolumeScatter");
    mVCShader.push_back(CVolumeScatter);

    ShaderItem_cls CVolumeScatterShadow("VolumeScatterShadow");
    CVolumeScatterShadow.mVstrUniform.push_back("texturVolume");
    CVolumeScatterShadow.mVstrUniform.push_back("textureCC");
    CVolumeScatterShadow.mVstrUniform.push_back("lightNumber");
    CVolumeScatterShadow.mVstrUniform.push_back("shadowAlpha");
    CVolumeScatterShadow.mVstrUniform.push_back("shadowQuality");
    CVolumeScatterShadow.mVstrUniform.push_back("clipBoxMin");
    CVolumeScatterShadow.mVstrUniform.push_back("clipBoxMax");
    CVolumeScatterShadow.mVstrUniform.push_back("ambientClr");
    CVolumeScatterShadow.mVstrUniform.push_back("diffuseClr");
    CVolumeScatterShadow.mVstrUniform.push_back("specularClr");
    CVolumeScatterShadow.mVstrUniform.push_back("shininess");
    CVolumeScatterShadow.mVstrUniform.push_back("scatterFactor");
    CVolumeScatterShadow.mVstrUniform.push_back("forceDiscard0");
    CVolumeScatterShadow.mVstrUniform.push_back("trim");
    CVolumeScatterShadow.mVstrUniform.push_back("intervalGlobal");
    CVolumeScatterShadow.mVstrUniform.push_back("subQuality");
    CVolumeScatterShadow.mVstrUniform.push_back("blend");
    CVolumeScatterShadow.SetShaderPath("GL/VolumeScatterShadow");
    mVCShader.push_back(CVolumeScatterShadow);

    ShaderItem_cls CObjectGouraud("ObjectGouraud");
    CObjectGouraud.mbDissolve = TRUE;
    CObjectGouraud.mbFog = TRUE;
    CObjectGouraud.mVstrUniform.push_back("lightNumber");
    CObjectGouraud.mVstrUniform.push_back("lightRecieve");
    CObjectGouraud.mVstrUniform.push_back("textureImage");
    CObjectGouraud.mVstrUniform.push_back("textureMode");
    CObjectGouraud.mVstrUniform.push_back("useTexAlpha");
    CObjectGouraud.mVstrUniform.push_back("dissolveMode");
    CObjectGouraud.SetShaderPath("GL/Gouraud");
    mVCShader.push_back(CObjectGouraud);

    ShaderItem_cls CObjectPhong("ObjectPhong");
    CObjectPhong.mbDissolve = TRUE;
    CObjectPhong.mbFog = TRUE;
    CObjectPhong.mVstrUniform.push_back("lightNumber");
    CObjectPhong.mVstrUniform.push_back("lightRecieve");
    CObjectPhong.mVstrUniform.push_back("textureImage");
    CObjectPhong.mVstrUniform.push_back("textureMode");
    CObjectPhong.mVstrUniform.push_back("useTexAlpha");
    CObjectPhong.mVstrUniform.push_back("dissolveMode");
    CObjectPhong.mVstrUniform.push_back("bFlat");
    CObjectPhong.SetShaderPath("GL/Phong");
    mVCShader.push_back(CObjectPhong);

    ShaderItem_cls CObjectToon("ObjectToon");
    CObjectToon.mbDissolve = TRUE;
    CObjectToon.mbFog = TRUE;
    CObjectToon.mVstrUniform.push_back("boldness");
    CObjectToon.mVstrUniform.push_back("lightNumber");
    CObjectToon.mVstrUniform.push_back("lightRecieve");
    CObjectToon.mVstrUniform.push_back("textureImage");
    CObjectToon.mVstrUniform.push_back("textureMode");
    CObjectToon.mVstrUniform.push_back("useTexAlpha");
    CObjectToon.mVstrUniform.push_back("dissolveMode");
    CObjectToon.mVstrUniform.push_back("shadowVolumeMode");
    CObjectToon.mVstrUniform.push_back("edgeThreshold");
    CObjectToon.mVstrUniform.push_back("bFlat");
    CObjectToon.SetShaderPath("GL/Outline");
    mVCShader.push_back(CObjectToon);

    ShaderItem_cls CObjectProjectNormal("ObjectProjectNormal");
    CObjectProjectNormal.mVstrUniform.push_back("bFlat");
    CObjectProjectNormal.SetShaderPath("GL/ProjectedNormal");
    mVCShader.push_back(CObjectProjectNormal);

    ShaderItem_cls CLineBasic("LineBasic");
    CLineBasic.mbDissolve = TRUE;
    CLineBasic.mVstrUniform.push_back("fogMode");
    CLineBasic.mVstrUniform.push_back("dissolveMode");
    CLineBasic.SetShaderPath("GL/LineBasic");
    mVCShader.push_back(CLineBasic);

    ShaderItem_cls CPointBasic("PointBasic");
    CPointBasic.mbDissolve = TRUE;
    CPointBasic.mbFog = TRUE;
    CPointBasic.mVstrUniform.push_back("lightNumber");
    CPointBasic.mVstrUniform.push_back("lightRecieve");
    CPointBasic.mVstrUniform.push_back("textureColorCurve");
    CPointBasic.mVstrUniform.push_back("textureImage");
    CPointBasic.mVstrUniform.push_back("viewport");
    CPointBasic.mVstrUniform.push_back("baseDistance");
    CPointBasic.mVstrUniform.push_back("dissolveMode");
    CPointBasic.mVstrAttrib.push_back("pointSize");
    CPointBasic.mVstrAttrib.push_back("colorCurveS");
    CPointBasic.mVstrAttrib.push_back("colorCurveT");
    CPointBasic.mVstrAttrib.push_back("filterAlpha");
    CPointBasic.mVstrAttrib.push_back("paramBit");
    CPointBasic.SetShaderPath("GL/PointBasic");
    mVCShader.push_back(CPointBasic);

    ShaderItem_cls CPointBillboard("PointBillboard");
    CPointBillboard.mbDissolve = TRUE;
    CPointBillboard.mbFog = TRUE;
    CPointBillboard.mVstrUniform.push_back("lightNumber");
    CPointBillboard.mVstrUniform.push_back("lightRecieve");
    CPointBillboard.mVstrUniform.push_back("textureColorCurve");
    CPointBillboard.mVstrUniform.push_back("textureImage");
    CPointBillboard.mVstrUniform.push_back("dissolveMode");
    CPointBillboard.mVstrUniform.push_back("baseInvModelMatrix");
    CPointBillboard.mVstrUniform.push_back("modelInvBaseMatrix");
    CPointBillboard.mVstrUniform.push_back("baseInvMatrix");
    CPointBillboard.mVstrUniform.push_back("rotationMatrix");
    CPointBillboard.mVstrUniform.push_back("twistMode");
    CPointBillboard.mVstrAttrib.push_back("pointSize");
    CPointBillboard.mVstrAttrib.push_back("colorCurveS");
    CPointBillboard.mVstrAttrib.push_back("colorCurveT");
    CPointBillboard.mVstrAttrib.push_back("filterAlpha");
    CPointBillboard.mVstrAttrib.push_back("paramBit");
    CPointBillboard.mVstrAttrib.push_back("twistID");
    CPointBillboard.SetShaderPath("GL/PointBillboard");
    mVCShader.push_back(CPointBillboard);
    
    ShaderItem_cls CSphereGouraud("SphereGouraud");
    CSphereGouraud.mbDissolve = TRUE;
    CSphereGouraud.mbFog = TRUE;
    CSphereGouraud.mVstrUniform.push_back("lightNumber");
    CSphereGouraud.mVstrUniform.push_back("lightRecieve");
    CSphereGouraud.mVstrUniform.push_back("textureColorCurve");
    CSphereGouraud.mVstrUniform.push_back("textureImage");
    CSphereGouraud.mVstrUniform.push_back("useTexAlpha");
    CSphereGouraud.mVstrUniform.push_back("dissolveMode");
    CSphereGouraud.mVstrAttrib.push_back("colorCurveS");
    CSphereGouraud.mVstrAttrib.push_back("colorCurveT");
    CSphereGouraud.SetShaderPath("GL/SphereGouraud");
    mVCShader.push_back(CSphereGouraud);

    ShaderItem_cls CSpherePhong("SpherePhong");
    CSpherePhong.mbDissolve = TRUE;
    CSpherePhong.mbFog = TRUE;
    CSpherePhong.mVstrUniform.push_back("lightNumber");
    CSpherePhong.mVstrUniform.push_back("lightRecieve");
    CSpherePhong.mVstrUniform.push_back("textureColorCurve");
    CSpherePhong.mVstrUniform.push_back("textureImage");
    CSpherePhong.mVstrUniform.push_back("useTexAlpha");
    CSpherePhong.mVstrUniform.push_back("dissolveMode");
    CSpherePhong.mVstrUniform.push_back("bFlat");
    CSpherePhong.mVstrAttrib.push_back("colorCurveS");
    CSpherePhong.mVstrAttrib.push_back("colorCurveT");
    CSpherePhong.SetShaderPath("GL/SpherePhong");
    mVCShader.push_back(CSpherePhong);

    ShaderItem_cls CSphereToon("SphereToon");
    CSphereToon.mbDissolve = TRUE;
    CSphereToon.mbFog = TRUE;
    CSphereToon.mVstrUniform.push_back("boldness");
    CSphereToon.mVstrUniform.push_back("lightNumber");
    CSphereToon.mVstrUniform.push_back("lightRecieve");
    CSphereToon.mVstrUniform.push_back("textureColorCurve");
    CSphereToon.mVstrUniform.push_back("textureImage");
    CSphereToon.mVstrUniform.push_back("useTexAlpha");
    CSphereToon.mVstrUniform.push_back("dissolveMode");
    CSphereToon.mVstrUniform.push_back("edgeThreshold");
    CSphereToon.mVstrUniform.push_back("bFlat");
    CSphereToon.mVstrAttrib.push_back("colorCurveS");
    CSphereToon.mVstrAttrib.push_back("colorCurveT");
    CSphereToon.SetShaderPath("GL/SphereToon");
    mVCShader.push_back(CSphereToon);

    ShaderItem_cls CSVTriangleMesh("SVTriangleMesh");
    CSVTriangleMesh.mbDissolve = FALSE;
    CSVTriangleMesh.mVstrUniform.push_back("shadowLength");
    CSVTriangleMesh.mVstrUniform.push_back("offsetFactor");
    CSVTriangleMesh.mVstrUniform.push_back("cullingSide");
    CSVTriangleMesh.mVstrUniform.push_back("forClosed");
    CSVTriangleMesh.SetShaderPath("GL/SVTriangleMesh");
    mVCShader.push_back(CSVTriangleMesh);

    ShaderItem_cls CHF("HF");
    CHF.mbDissolve = TRUE;
    CHF.mbFog = TRUE;
    CHF.mVstrUniform.push_back("lightNumber");
    CHF.mVstrUniform.push_back("lightRecieve");
    CHF.mVstrUniform.push_back("textureHeight");
    CHF.mVstrUniform.push_back("channel");
    CHF.mVstrUniform.push_back("textureImage");
    CHF.mVstrUniform.push_back("textureCC");
    CHF.mVstrUniform.push_back("textureMode");
    CHF.mVstrUniform.push_back("useTexAlpha");
    CHF.mVstrUniform.push_back("dissolveMode");
    CHF.mVstrUniform.push_back("bFlat");
    CHF.SetShaderPath("GL/HF");
    mVCShader.push_back(CHF);

    ShaderItem_cls CSVTriangleMeshHF("SVTriangleMeshHF");
    CSVTriangleMeshHF.mbDissolve = FALSE;
    CSVTriangleMeshHF.mVstrUniform.push_back("shadowLength");
    CSVTriangleMeshHF.mVstrUniform.push_back("offsetFactor");
    CSVTriangleMeshHF.mVstrUniform.push_back("cullingSide");
    CSVTriangleMeshHF.mVstrUniform.push_back("forClosed");
    CSVTriangleMeshHF.mVstrUniform.push_back("textureHeight");
    CSVTriangleMeshHF.mVstrUniform.push_back("channel");
    CSVTriangleMeshHF.SetShaderPath("GL/SVTriangleMeshHF");
    mVCShader.push_back(CSVTriangleMeshHF);

    ShaderItem_cls CStreamlineBasic("StreamlineBasic");
    CStreamlineBasic.mbDissolve = TRUE;
    CStreamlineBasic.mbFog = TRUE;
    CStreamlineBasic.mVstrUniform.push_back("lightNumber");
    CStreamlineBasic.mVstrUniform.push_back("textureMode");
    CStreamlineBasic.mVstrUniform.push_back("textureColorCurve");
    CStreamlineBasic.mVstrUniform.push_back("dissolveMode");
    CStreamlineBasic.mVstrUniform.push_back("useClip");
    CStreamlineBasic.mVstrUniform.push_back("clipBoxMin");
    CStreamlineBasic.mVstrUniform.push_back("clipBoxMax");
    CStreamlineBasic.mVstrUniform.push_back("matrix16");
    CStreamlineBasic.mVstrUniform.push_back("smoothWidth");
    CStreamlineBasic.mVstrUniform.push_back("thinningFactor");
    CStreamlineBasic.SetShaderPath("GL/StreamlineBasic");
    mVCShader.push_back(CStreamlineBasic);
    
    ShaderItem_cls CStreamlineTriangles("StreamlineTriangles");
    CStreamlineTriangles.mbDissolve = TRUE;
    CStreamlineTriangles.mbFog = TRUE;
    CStreamlineTriangles.mVstrUniform.push_back("lightNumber");
    CStreamlineTriangles.mVstrUniform.push_back("textureMode");
    CStreamlineTriangles.mVstrUniform.push_back("textureColorCurve");
    CStreamlineTriangles.mVstrUniform.push_back("textureSample");
    CStreamlineTriangles.mVstrUniform.push_back("dissolveMode");
    CStreamlineTriangles.mVstrUniform.push_back("lineWidth");
    CStreamlineTriangles.mVstrUniform.push_back("viewport");
    CStreamlineTriangles.mVstrUniform.push_back("mappingType");
    CStreamlineTriangles.mVstrUniform.push_back("useClip");
    CStreamlineTriangles.mVstrUniform.push_back("clipBoxMin");
    CStreamlineTriangles.mVstrUniform.push_back("clipBoxMax");
    CStreamlineTriangles.mVstrUniform.push_back("matrix16");
    CStreamlineTriangles.mVstrUniform.push_back("smoothWidth");
    CStreamlineTriangles.mVstrUniform.push_back("thinningFactor");
    CStreamlineTriangles.SetShaderPath("GL/StreamlineTriangles");
    mVCShader.push_back(CStreamlineTriangles);

    wxString strPathCommonDefinitions("GL/CommonDefinitionsShader.h");
    if (wxFileExists(strPathCommonDefinitions) == TRUE)
	GetStrFullFromFile(strPathCommonDefinitions, mstrCommonDefinitions, TRUE);
        
    wxString strPathSpotEffect("GL/FUNCTION/SpotEffect.txt");
    if (wxFileExists(strPathSpotEffect) == TRUE)
	GetStrFullFromFile(strPathSpotEffect, mstrSpotEffectFunc, TRUE);

    wxString strPathFogEffect("GL/FUNCTION/FogEffect.txt");
    if (wxFileExists(strPathFogEffect) == TRUE)
	GetStrFullFromFile(strPathFogEffect, mstrFogEffectFunc, TRUE);

    wxString strPathNormal("GL/FUNCTION/Normal.txt");
    if (wxFileExists(strPathNormal) == TRUE)
	GetStrFullFromFile(strPathNormal, mstrNormalFunc, TRUE);

    wxString strPathEnsembleDithering("GL/FUNCTION/EnsembleDithering.txt");
    if (wxFileExists(strPathEnsembleDithering) == TRUE)
	GetStrFullFromFile(strPathEnsembleDithering, mstrEnsembleDitheringFunc, TRUE);
    
    uint shaderNum = mVCShader.size();
    for (uint i = 0; i < shaderNum; i++){
	mVCShader[i].mstrVertSrc.Replace("//##COMMON_DEFINITIONS_SHADER##", mstrCommonDefinitions);
	mVCShader[i].mstrVertSrc.Replace("//##FUNC_SPOT_EFFECT##", mstrSpotEffectFunc);
	mVCShader[i].mstrVertSrc.Replace("//##FUNC_FOG_EFFECT##", mstrFogEffectFunc);

	mVCShader[i].mstrFragSrc.Replace("//##FUNC_SPOT_EFFECT##", mstrSpotEffectFunc);
	mVCShader[i].mstrFragSrc.Replace("//##FUNC_FOG_EFFECT##", mstrFogEffectFunc);
	mVCShader[i].mstrFragSrc.Replace("//##FUNC_NORMAL##", mstrNormalFunc);
	mVCShader[i].mstrFragSrc.Replace("//##FUNC_ENSEMBLE_DITHERING", mstrEnsembleDitheringFunc);

	if (mstrEnsembleDitheringFunc.IsEmpty() == FALSE)
	    mVCShader[i].mstrFragSrc.Replace("//##APPLY_ENSEMBLE_DITHERING##", ("EnsembleDithering(alpha);"));
    }
    
}

void CLS::
PrepareViewportInfo(){
    int viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
	
    double op[16]; // original projection matrix
    glGetDoublev(GL_PROJECTION_MATRIX, op);

    mProjMatrix10 = op[10];
    mProjMatrix14 = op[14];

    double inverseProject[16] = { 0.0 };

    mProjInv11 = 1.0/op[14];
    mProjInv15 = -op[10] / (op[14] * op[11]);
}

void CLS::
UseProgram(ShaderItem_cls &CItem){
    GLuint program = CItem.mProgram;
    if (mCurrentShader == program) return;

    glUseProgram(program);
    mCurrentShader = program;
}

void CLS::
SetEnsembleShaderSetting(ShaderItem_cls &CItem, int ensemble){

    float ensembleParamA[2] = {ensemble, ensemble*0.5};
    float ensembleParamB = 43758.5453+ensemble;
    float ensembleParamC[2] = {sin(0.1 * (ensemble + 1)) * 100, cos(0.1 * (ensemble + 1)) * 100};
    glUniform2fv(CItem.mIdEnsembleParamA, 1, ensembleParamA);
    glUniform1f(CItem.mIdEnsembleParamB, ensembleParamB);
    glUniform2fv(CItem.mIdEnsembleParamC, 1, ensembleParamC);
}

void CLS::
SetFogSetting(ShaderItem_cls &CItem){
    glUniform1i(CItem.mIdFogMode, mFogMode);
    glUniform1f(CItem.mIdProjInv11, mProjInv11);
    glUniform1f(CItem.mIdProjInv15, mProjInv15);
}

/*
void CLS::
BeginShader(){
    ShaderItem_cls& CItem = mVCShader[0];

    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();

    UseProgram(CItem);

    if (CItem.mstrShaderName.IsSameAs("Exposure")){
	glUniform1i(CItem.mVIdUniform[0], 0);
	float min[4] = {0, 0, 0, 0};
	float max[4] = {1, 1, 1 ,1};
	float gamma[4] = {2, 2, 2, 2};
	glUniform4fv(CItem.mVIdUniform[1], 1, min);
	glUniform4fv(CItem.mVIdUniform[2], 1, max);
	glUniform4fv(CItem.mVIdUniform[3], 1, gamma);
	glUniform1i(CItem.mVIdUniform[4], 120);
	glUniform1i(CItem.mVIdUniform[5], 0);
	glUniform1i(CItem.mVIdUniform[6], 0);
	glUniform1f(CItem.mVIdUniform[7], 1);
    }
}
*/

void CLS::
BeginShader(wxString strShader, ShaderParams_cls *pCParam){
    if (strShader.IsSameAs("ClampAlpha"))
	BeginClampAlpha(pCParam);
    if (strShader.IsSameAs("Exposure"))
	BeginExposure((ExposureParams_cls*)pCParam);
    if (strShader.IsSameAs("Blur"))
	BeginBlur((BlurParams_cls*)pCParam);
    if (strShader.IsSameAs("FocalBlur"))
	BeginFocalBlur((FocalBlurParams_cls*)pCParam);
    if (strShader.IsSameAs("FocalBlurB"))
	BeginFocalBlurB((FocalBlurParamsB_cls*)pCParam);
    if (strShader.IsSameAs("Glare"))
	BeginGlare((GlareParams_cls*)pCParam);
    if (strShader.IsSameAs("RemoveAlpha"))
	BeginRemoveAlpha((RemoveAlphaParam_cls*)pCParam);
    if (strShader.IsSameAs("Overlay"))
	BeginOverlay((OverlayParam_cls*)pCParam);
    if (strShader.IsSameAs("SSAO"))
	BeginSSAO((SSAOParam_cls*)pCParam);
    if (strShader.IsSameAs("Depth"))
	BeginDepth(pCParam);
    if (strShader.IsSameAs("VolumeBasic"))
	BeginVolumeBasic((VolumeBasicParams_cls*)pCParam);
    if (strShader.IsSameAs("VolumePhong"))
	BeginVolumePhong((VolumePhongParams_cls*)pCParam);
    if (strShader.IsSameAs("VolumePhongShadow"))
	BeginVolumePhongShadow((VolumePhongShadowParams_cls*)pCParam);
}

void CLS::
BeginClampAlpha(ShaderParams_cls *pCDummy){
    ShaderItem_cls& CItem = mVCShader[_SHADER_CLAMP_ALPHA];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;
    UseProgram(CItem);

    glUniform1i(CItem.mVIdUniform[0], 0);
}

void CLS::
BeginExposure(ExposureParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_EXPOSURE];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;
    UseProgram(CItem);
    
    glUniform1i(CItem.mVIdUniform[0], 0);
    float min = pCParam->mMin;
    float max = pCParam->mMax;
    float gamma = pCParam->mGamma;
    float Amin[4] = {min, min, min, 0/*min*/};
    float Amax[4] = {max, max, max, 1/*max*/};
    float Agamma[4] = {gamma, gamma, gamma, gamma};
    int hue = pCParam->mHue;
    int saturation = pCParam->mSaturation;
    
    glUniform4fv(CItem.mVIdUniform[1], 1, Amin);
    glUniform4fv(CItem.mVIdUniform[2], 1, Amax);
    glUniform4fv(CItem.mVIdUniform[3], 1, Agamma);
    glUniform1i (CItem.mVIdUniform[4], hue);
    glUniform1i (CItem.mVIdUniform[5], saturation);
    glUniform1i (CItem.mVIdUniform[6], 1);
    glUniform1f (CItem.mVIdUniform[7], 1);
}

void CLS::
BeginBlur(BlurParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_BLUR];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    
    glUniform1i(CItem.mVIdUniform[0], 0);
    int type = pCParam->mType;
    int kernel = pCParam->mKernel;
    float width = pCParam->mWidth;
    int quality = pCParam->mQuality;
    float gamma = pCParam->mGamma;
    float Agamma[4] = {gamma, gamma, gamma, gamma};
    glUniform1i (CItem.mVIdUniform[1], type);
    glUniform1i (CItem.mVIdUniform[2], kernel);
    glUniform1f (CItem.mVIdUniform[3], width);
    glUniform1i (CItem.mVIdUniform[4], quality);
    glUniform4fv(CItem.mVIdUniform[5], 1, Agamma);
}

void CLS::
BeginFocalBlur(FocalBlurParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_FOCAL_BLUR];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    int textureImage = 0, textureImage1 = 1;
    int quality = pCParam->mQuality;
    float gamma = pCParam->mGamma;
    float Agamma[4] = {gamma, gamma, gamma, gamma};
    float focalDistance = pCParam->mFocalDistance;
    float widthBase = pCParam->mWidthBase;
    UseProgram(CItem);

    glUniform1i(CItem.mVIdUniform[0], textureImage);
    glUniform1i(CItem.mVIdUniform[1], textureImage1);
    glUniform1i(CItem.mVIdUniform[2], quality);
    glUniform4fv(CItem.mVIdUniform[3], 1, Agamma);
    glUniform1f(CItem.mVIdUniform[4], mProjMatrix10);
    glUniform1f(CItem.mVIdUniform[5], mProjMatrix14);
    glUniform1f(CItem.mVIdUniform[6], focalDistance);
    glUniform1f(CItem.mVIdUniform[7], widthBase);
}

void CLS::
BeginFocalBlurB(FocalBlurParamsB_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_FOCAL_BLUR_B];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    int textureImage = 0, textureImage1 = 1;
    int quality = pCParam->mQuality;
    float gamma = pCParam->mGamma;
    float Agamma[4] = {gamma, gamma, gamma, gamma};
    float focalDistance = pCParam->mFocalDistance;
    float widthBase = pCParam->mWidthBase;
    UseProgram(CItem);

    glUniform1i(CItem.mVIdUniform[0], textureImage);
    glUniform1i(CItem.mVIdUniform[1], textureImage1);
    glUniform1i(CItem.mVIdUniform[2], quality);
    glUniform4fv(CItem.mVIdUniform[3], 1, Agamma);
    glUniform1f(CItem.mVIdUniform[4], mProjMatrix10);
    glUniform1f(CItem.mVIdUniform[5], mProjMatrix14);
    glUniform1f(CItem.mVIdUniform[6], focalDistance);
    glUniform1f(CItem.mVIdUniform[7], widthBase);
}

void CLS::
BeginGlare(GlareParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_GLARE];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    
    glUniform1i(CItem.mVIdUniform[0], 0);
    int type = pCParam->mType;
    float width = pCParam->mWidth;
    int quality = pCParam->mQuality;
    int rayNum = pCParam->mRayNumber;
    float offset = pCParam->mOffset;
    float crit = pCParam->mCrit;
    float enhance = pCParam->mEnhance;
    float fade = pCParam->mFade;
    glUniform1i(CItem.mVIdUniform[1], type);
    glUniform1f(CItem.mVIdUniform[2], width);
    glUniform1i(CItem.mVIdUniform[3], quality);
    glUniform1i(CItem.mVIdUniform[4], rayNum);
    glUniform1f(CItem.mVIdUniform[5], offset);
    glUniform1f(CItem.mVIdUniform[6], crit);
    glUniform1f(CItem.mVIdUniform[7], enhance);
    glUniform1f(CItem.mVIdUniform[8], fade);

}

void CLS::
BeginRemoveAlpha(RemoveAlphaParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_REMOVE_ALPHA];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int extremeAlpha = pCParam->mExtremeAlpha;
    int convert = pCParam->mConvert;
    int backgroundTile = pCParam->mBackgroundTile;
    glUniform1i(CItem.mVIdUniform[0], 0);
    glUniform1i(CItem.mVIdUniform[1], extremeAlpha);
    glUniform1i(CItem.mVIdUniform[2], convert);
    glUniform1i(CItem.mVIdUniform[3], backgroundTile);
}

void CLS::
BeginOverlay(OverlayParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_OVERLAY];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    int textureImage = 0, textureImage1 = 1;
    float alpha = pCParam->mAlpha;
    float XScale = pCParam->mXScale, YScale = pCParam->mYScale;
    UseProgram(CItem);
    glUniform1i(CItem.mVIdUniform[0], textureImage);
    glUniform1i(CItem.mVIdUniform[1], textureImage1);
    glUniform1f(CItem.mVIdUniform[2], alpha);
    glUniform1f(CItem.mVIdUniform[3], XScale);
    glUniform1f(CItem.mVIdUniform[4], YScale);
}

void CLS::
BeginSSAO(SSAOParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_SSAO];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    int textureImage = 0, textureImage1 = 1;
    int quality = pCParam->mQuality;
    float factor = pCParam->mFactor;
    float radius = pCParam->mRadius;
    int type = pCParam->mType;
    UseProgram(CItem);

    glUniform1i (CItem.mVIdUniform[0], textureImage);
    glUniform1i (CItem.mVIdUniform[1], textureImage1);
    glUniform1i (CItem.mVIdUniform[2], quality);
    glUniform1f (CItem.mVIdUniform[3], mProjMatrix10);
    glUniform1f (CItem.mVIdUniform[4], mProjMatrix14);
    glUniform1f (CItem.mVIdUniform[5], factor);
    glUniform1f (CItem.mVIdUniform[6], radius);
    glUniform3fv(CItem.mVIdUniform[7], 1, pCParam->mBaseColor);
    glUniform1i (CItem.mVIdUniform[8], type);
}

void CLS::
BeginDepth(ShaderParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_DEPTH];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    int textureImage = 0;
    UseProgram(CItem);

    glUniform1i (CItem.mVIdUniform[0], textureImage);
    glUniform1f (CItem.mVIdUniform[1], mProjMatrix10);
    glUniform1f (CItem.mVIdUniform[2], mProjMatrix14);
}

void CLS::
BeginVolumeBasic(VolumeBasicParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_VOLUME_BASIC];

    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    int discard0 = pCParam->mDiscard0;
    int trim = pCParam->mTrim;
    float intervalGlobal = pCParam->mIntervalGlobal;
    int subQuality = pCParam->mSubQuality;
    int blend = pCParam->mBlendType;

    UseProgram(CItem);
    glUniform1i (CItem.mVIdUniform[0], 0);
    glUniform1i (CItem.mVIdUniform[1], 1);
    glUniform3fv(CItem.mVIdUniform[2], 1, pCParam->mClipMin);
    glUniform3fv(CItem.mVIdUniform[3], 1, pCParam->mClipMax);
    glUniform1i (CItem.mVIdUniform[4], discard0);
    glUniform1i (CItem.mVIdUniform[5], trim);
    glUniform1f (CItem.mVIdUniform[6], intervalGlobal);
    glUniform1i (CItem.mVIdUniform[7], subQuality);
    glUniform1i (CItem.mVIdUniform[8], blend);
}

void CLS::
BeginVolumePhong(VolumePhongParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_VOLUME_PHONG];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int lightNumber = pCParam->mLightNumber;
    int shininess = pCParam->mShininess;
    int discard0 = pCParam->mDiscard0;
    int trim = pCParam->mTrim;
    float intervalGlobal = pCParam->mIntervalGlobal;
    int subQuality = pCParam->mSubQuality;
    float delta = pCParam->mDelta;
    int blend = pCParam->mBlendType;

    glUniform1i (CItem.mVIdUniform[0], 0);
    glUniform1i (CItem.mVIdUniform[1], 1);
    glUniform1i (CItem.mVIdUniform[2], lightNumber);
    glUniform3fv(CItem.mVIdUniform[3], 1, pCParam->mClipMin);
    glUniform3fv(CItem.mVIdUniform[4], 1, pCParam->mClipMax);
    glUniform3fv(CItem.mVIdUniform[5], 1, pCParam->mAmbient);
    glUniform3fv(CItem.mVIdUniform[6], 1, pCParam->mDiffise);
    glUniform3fv(CItem.mVIdUniform[7], 1, pCParam->mSpecular);
    glUniform1i (CItem.mVIdUniform[8], shininess);
    glUniform1i (CItem.mVIdUniform[9], discard0);
    glUniform1i (CItem.mVIdUniform[10], trim);
    glUniform1f (CItem.mVIdUniform[11], intervalGlobal);
    glUniform1i (CItem.mVIdUniform[12], subQuality);
    glUniform1f (CItem.mVIdUniform[13], delta);
    glUniform1i (CItem.mVIdUniform[14], blend);
}

void CLS::
BeginVolumePhongShadow(VolumePhongShadowParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_VOLUME_SHADOW];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;
    
    UseProgram(CItem);
    int lightNumber = pCParam->mLightNumber;
    int shadowQuality = pCParam->mShadowQuality;
    int shininess = pCParam->mShininess;
    int discard0 = pCParam->mDiscard0;
    int trim = pCParam->mTrim;
    float intervalGlobal = pCParam->mIntervalGlobal;
    int subQuality = pCParam->mSubQuality;
    float shadowAlpha = pCParam->mBaseAlpha / shadowQuality * pCParam->mShadowFactor;
    float delta = pCParam->mDelta;
    int blend = pCParam->mBlendType;

    glUniform1i (CItem.mVIdUniform[0], 0);
    glUniform1i (CItem.mVIdUniform[1], 1);
    glUniform1i (CItem.mVIdUniform[2], lightNumber);
    glUniform1f (CItem.mVIdUniform[3], shadowAlpha);
    glUniform1i (CItem.mVIdUniform[4], shadowQuality);
    glUniform3fv(CItem.mVIdUniform[5], 1, pCParam->mClipMin);
    glUniform3fv(CItem.mVIdUniform[6], 1, pCParam->mClipMax);
    glUniform3fv(CItem.mVIdUniform[7], 1, pCParam->mAmbient);
    glUniform3fv(CItem.mVIdUniform[8], 1, pCParam->mDiffise);
    glUniform3fv(CItem.mVIdUniform[9], 1, pCParam->mSpecular);
    glUniform1i (CItem.mVIdUniform[10], shininess);
    glUniform1i (CItem.mVIdUniform[11], discard0);
    glUniform1i (CItem.mVIdUniform[12], trim);
    glUniform1f (CItem.mVIdUniform[13], intervalGlobal);
    glUniform1i (CItem.mVIdUniform[14], subQuality);
    glUniform1f (CItem.mVIdUniform[15], delta);
    glUniform1i (CItem.mVIdUniform[16], blend);
    //glUniform4fv(CItem.mVIdUniform[11], 1, pCParam->mModelMatrix);
}

void CLS::
BeginVolumeScatter(VolumeScatterParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_VOLUME_SCATTER];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int lightNumber = pCParam->mLightNumber;
    int shininess = pCParam->mShininess;
    float scatterFactor = pCParam->mScatterFatcor;
    int discard0 = pCParam->mDiscard0;
    int trim = pCParam->mTrim;
    float intervalGlobal = pCParam->mIntervalGlobal;
    int subQuality = pCParam->mSubQuality;
    int blend = pCParam->mBlendType;
    glUniform1i (CItem.mVIdUniform[0], 0);
    glUniform1i (CItem.mVIdUniform[1], 1);
    glUniform1i (CItem.mVIdUniform[2], lightNumber);
    glUniform3fv(CItem.mVIdUniform[3], 1, pCParam->mClipMin);
    glUniform3fv(CItem.mVIdUniform[4], 1, pCParam->mClipMax);
    glUniform3fv(CItem.mVIdUniform[5], 1, pCParam->mAmbient);
    glUniform3fv(CItem.mVIdUniform[6], 1, pCParam->mDiffise);
    glUniform3fv(CItem.mVIdUniform[7], 1, pCParam->mSpecular);
    glUniform1i (CItem.mVIdUniform[8], shininess);
    glUniform1f (CItem.mVIdUniform[9], scatterFactor);
    glUniform1i (CItem.mVIdUniform[10], discard0);
    glUniform1i (CItem.mVIdUniform[11], trim);
    glUniform1f (CItem.mVIdUniform[12], intervalGlobal);
    glUniform1i (CItem.mVIdUniform[13], subQuality);
    glUniform1i (CItem.mVIdUniform[14], blend);
}

void CLS::
BeginVolumeScatterShadow(VolumeScatterShadowParams_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_VOLUME_SCATTER_SHADOW];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int lightNumber = pCParam->mLightNumber;
    int shadowQuality = pCParam->mShadowQuality;
    float shadowAlpha = pCParam->mBaseAlpha / pCParam->mShadowQuality * pCParam->mShadowFactor;
    int shininess = pCParam->mShininess;
    float scatterFactor = pCParam->mScatterFatcor;
    int discard0 = pCParam->mDiscard0;
    int trim = pCParam->mTrim;
    float intervalGlobal = pCParam->mIntervalGlobal;
    int subQuality = pCParam->mSubQuality;
    int blend = pCParam->mBlendType;

    glUniform1i (CItem.mVIdUniform[0], 0);
    glUniform1i (CItem.mVIdUniform[1], 1);
    glUniform1i (CItem.mVIdUniform[2], lightNumber);
    glUniform1f (CItem.mVIdUniform[3], shadowAlpha);
    glUniform1i (CItem.mVIdUniform[4], shadowQuality);
    glUniform3fv(CItem.mVIdUniform[5], 1, pCParam->mClipMin);
    glUniform3fv(CItem.mVIdUniform[6], 1, pCParam->mClipMax);
    glUniform3fv(CItem.mVIdUniform[7], 1, pCParam->mAmbient);
    glUniform3fv(CItem.mVIdUniform[8], 1, pCParam->mDiffise);
    glUniform3fv(CItem.mVIdUniform[9], 1, pCParam->mSpecular);
    glUniform1i (CItem.mVIdUniform[10], shininess);
    glUniform1f (CItem.mVIdUniform[11], scatterFactor);
    glUniform1i (CItem.mVIdUniform[12], discard0);
    glUniform1i (CItem.mVIdUniform[13], trim);
    glUniform1f (CItem.mVIdUniform[14], intervalGlobal);
    glUniform1i (CItem.mVIdUniform[15], subQuality);
    glUniform1i (CItem.mVIdUniform[16], blend);
}

void CLS::
BeginObjectGouraud(ObjectGouraudParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_OBJECT_GOURAUD];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureImage = 0;
    int textureMode = pCParam->mTextureMode;
    //int fogMode = 0;
    int useTexAlpha = 0;//
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);

    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1iv(CItem.mVIdUniform[1], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[2], textureImage);
    glUniform1i (CItem.mVIdUniform[3], textureMode);
    //glUniform1i (CItem.mVIdUniform[4], fogMode);
    glUniform1i (CItem.mVIdUniform[4], useTexAlpha);
    glUniform1i (CItem.mVIdUniform[5], dissolveMode);
}

void CLS::
BeginObjectPhong(ObjectPhongParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_OBJECT_PHONG];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureImage = 0;
    int textureMode = pCParam->mTextureMode;
    //int fogMode = mFogMode;
    int useTexAlpha = pCParam->mUseAlpha;
    int dissolveMode = pCParam->mDissolveMode;
    int bFlat = pCParam->mbFlat;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);
    
    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1iv(CItem.mVIdUniform[1], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[2], textureImage);
    glUniform1i (CItem.mVIdUniform[3], textureMode);
    //glUniform1i (CItem.mVIdUniform[4], fogMode);
    glUniform1i (CItem.mVIdUniform[4], useTexAlpha);
    glUniform1i (CItem.mVIdUniform[5], dissolveMode);
    glUniform1i (CItem.mVIdUniform[6], bFlat);
}

void CLS::
BeginObjectToon(ObjectToonParam_cls* pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_OBJECT_TOON];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    float boldness = pCParam->mBoldness;
    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureImage = 0;
    int textureMode = pCParam->mTextureMode;
    //int fogMode = 0;
    int useTexAlpha = 0;
    int dissolveMode = pCParam->mDissolveMode;
    int shadowVolumeMode = 1;
    float edgeThreshold = pCParam->mThreshold;
    int bFlat = pCParam->mbFlat;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);
    
    glUniform1f (CItem.mVIdUniform[0], boldness);
    glUniform1i (CItem.mVIdUniform[1], lightNumber);
    glUniform1iv(CItem.mVIdUniform[2], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[3], textureImage);
    glUniform1i (CItem.mVIdUniform[4], textureMode);
    //glUniform1i (CItem.mVIdUniform[5], fogMode);
    glUniform1i (CItem.mVIdUniform[5], useTexAlpha);
    glUniform1i (CItem.mVIdUniform[6], dissolveMode);
    glUniform1i (CItem.mVIdUniform[7], shadowVolumeMode);
    glUniform1f (CItem.mVIdUniform[8], edgeThreshold);
    glUniform1i (CItem.mVIdUniform[9], bFlat);
}

void CLS::
BeginObjectProjectNormal(ObjectNormalParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_OBJECT_NORMAL];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int fogMode = 0;
    int bFlat = pCParam->mbFlat;
    glUniform1i(CItem.mVIdUniform[0], bFlat);
}

void CLS::
BeginLineBasic(LineBasicParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_LINE_BASIC];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int fogMode = 0;
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    glUniform1i(CItem.mVIdUniform[0], fogMode);
    glUniform1i(CItem.mVIdUniform[1], dissolveMode);
}

void CLS::
BeginPointBasic(PointBasicParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_POINT_BASIC];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;
    
    UseProgram(CItem);
    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureImage[3] = {1, 2, 3};
    float viewportf[4] = {pCParam->mViewport[0], pCParam->mViewport[1],
			  pCParam->mViewport[2], pCParam->mViewport[3]};
    float distanceBase = pCParam->mDistanceBase;
    int fogMode = 0;
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);

    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1iv(CItem.mVIdUniform[1], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[2], 0);
    glUniform1iv(CItem.mVIdUniform[3], 3, textureImage);
    glUniform4fv(CItem.mVIdUniform[4], 1, viewportf);
    glUniform1f (CItem.mVIdUniform[5], distanceBase);
    glUniform1i (CItem.mVIdUniform[6], dissolveMode);
}

void CLS::
BeginPointBillboard(PointBillboardParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_POINT_BILLBOARD];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureImage[3] = {1, 2, 3};
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    int twistMode = pCParam->mTwistMode;

    //float *pBaseMatrix    = pCParam->mpBaseMatrix16;
    //float *pBaseMatrixInv = pCParam->mpBaseMatrix16Inv;
    
    //float *pViewMatrixInv = pCParam->mpViewMatrix16Inv;
    float *pBaseInvModelMatrix = pCParam->mpBaseInvModel;
    float *pModelInvBaseMatrix = pCParam->mpModelInvBase;
    float *pBaseInvMatrix      = pCParam->mpBaseInv;
    float pRotateMatrix[16];
    float axis[3] = {1, 0, 0};
    TensorMath_cls::GetRotationMatrix4x4(pRotateMatrix, axis, -3.14159/180*30);

    SetFogSetting(CItem);

    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1iv(CItem.mVIdUniform[1], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[2], 0);
    glUniform1iv(CItem.mVIdUniform[3], 3, textureImage);
    glUniform1i (CItem.mVIdUniform[4], dissolveMode);
    glUniformMatrix4fv(CItem.mVIdUniform[5],  1, FALSE, pBaseInvModelMatrix);
    glUniformMatrix4fv(CItem.mVIdUniform[6], 1, FALSE, pModelInvBaseMatrix);
    glUniformMatrix4fv(CItem.mVIdUniform[7], 1, FALSE, pBaseInvMatrix);
    glUniformMatrix4fv(CItem.mVIdUniform[8], 1, FALSE, pRotateMatrix);
    glUniform1i (CItem.mVIdUniform[9], twistMode);
}

void CLS::
BeginSphereGouraud(SphereGouraudParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_SPHERE_GOURAUD];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    if (pCParam == NULL) return;
    
    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureImage[3] = {1, 2, 3};
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);

    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1iv(CItem.mVIdUniform[1], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[2], 0);
    glUniform1iv(CItem.mVIdUniform[3], 3, textureImage);
    glUniform1i (CItem.mVIdUniform[4], 1);
    glUniform1i (CItem.mVIdUniform[5], dissolveMode);
}

void CLS::
BeginSpherePhong(SpherePhongParam_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_SPHERE_PHONG];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    if (pCParam == NULL) return;

    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureImage[3] = {1, 2, 3};
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);
    int bFlat = (int)pCParam->mbFlat;

    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1iv(CItem.mVIdUniform[1], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[2], 0);
    glUniform1iv(CItem.mVIdUniform[3], 3, textureImage);
    glUniform1i (CItem.mVIdUniform[4], 1);
    glUniform1i (CItem.mVIdUniform[5], dissolveMode);
    glUniform1i (CItem.mVIdUniform[6], bFlat);
}

void CLS::
BeginSphereToon(SphereToonParam_cls* pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_SPHERE_TOON];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    if (pCParam == NULL) return;
				    
    float boldness = pCParam->mBoldness;
    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureImage[3] = {1, 2, 3};
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);
    float edgeThreshold = pCParam->mThreshold;
    int bFlat = (int)pCParam->mbFlat;

    glUniform1f (CItem.mVIdUniform[0], boldness);
    glUniform1i (CItem.mVIdUniform[1], lightNumber);
    glUniform1iv(CItem.mVIdUniform[2], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[3], 0);
    glUniform1iv(CItem.mVIdUniform[4], 3, textureImage);
    glUniform1i (CItem.mVIdUniform[5], 1);
    glUniform1i (CItem.mVIdUniform[6], dissolveMode);
    glUniform1f (CItem.mVIdUniform[7], edgeThreshold);
    glUniform1i (CItem.mVIdUniform[8], bFlat);
}

void CLS::
BeginSVTrinagleMesh(ShaderSVTriangleMesh_cls* pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_SV_TRIANGLE_MESH];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);

    float shadowLength = pCParam->mShadowLength;
    float offsetFactor = pCParam->mOffsetFactor;
    //int cullingSide = pCParam->mCullingSide;
    int forClosed = pCParam->mForClosed; //1 for Clesed, -1 for Opened

    glUniform1f (CItem.mVIdUniform[0], shadowLength);
    glUniform1f (CItem.mVIdUniform[1], offsetFactor);
    //glUniform1i (CItem.mVIdUniform[2], cullingSide);
    glUniform1i (CItem.mVIdUniform[3], forClosed);
}

void CLS::
BeginHF(ShaderHF_cls *pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_HF];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);

    int lightNumber = pCParam->mLightNumber;
    int lightRecieve[7] = {1, 1, 1, 1, 1, 1, 1};
    int textureHeight = 1;
    int channel = pCParam->mChannel;
    int textureImage = 0, textureCC = 2;
    int textureMode = pCParam->mTextureMode;
    int useTexAlpha = 1;
    int dissolveMode = pCParam->mDissolveMode;
    int bFlat = pCParam->mbFlat;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);

    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1iv(CItem.mVIdUniform[1], 7, lightRecieve);
    glUniform1i (CItem.mVIdUniform[2], textureHeight);
    glUniform1i (CItem.mVIdUniform[3], channel);
    glUniform1i (CItem.mVIdUniform[4], textureImage);
    glUniform1i (CItem.mVIdUniform[5], textureCC);
    glUniform1i (CItem.mVIdUniform[6], textureMode);
    glUniform1i (CItem.mVIdUniform[7], useTexAlpha);
    glUniform1i (CItem.mVIdUniform[8], dissolveMode);
    glUniform1i (CItem.mVIdUniform[9], bFlat);
}

void CLS::
BeginSVTrinagleMeshHF(ShaderSVTriangleMesh_cls* pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_SV_TRIANGLE_MESH_HF];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);

    float shadowLength = pCParam->mShadowLength;
    float offsetFactor = pCParam->mOffsetFactor;
    //int cullingSide = pCParam->mCullingSide;
    int forClosed = pCParam->mForClosed; //1 for Clesed, -1 for Opened
    int textureHeight = 1;
    int channel = 0;

    glUniform1f (CItem.mVIdUniform[0], shadowLength);
    glUniform1f (CItem.mVIdUniform[1], offsetFactor);
    //glUniform1i (CItem.mVIdUniform[2], cullingSide);
    glUniform1i (CItem.mVIdUniform[3], forClosed);
    glUniform1i (CItem.mVIdUniform[4], textureHeight);
    glUniform1i (CItem.mVIdUniform[5], channel);
}

void CLS::
BeginStreamlineBasic(StreamlineBasicParam_cls* pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_STREAMLINE_BASIC];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;

    UseProgram(CItem);
    int lightNumber = 1;
    int textureMode = pCParam->mTextureMode;
    int textureColorCurve = 1;
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    SetFogSetting(CItem);

     int useClip = 1;
    float clipBoxMin[3] = {pCParam->mpClipBoxMin[0],
			   pCParam->mpClipBoxMin[1],
			   pCParam->mpClipBoxMin[2]};
    float clipBoxMax[3] = {pCParam->mpClipBoxMax[0],
			   pCParam->mpClipBoxMax[1],
			   pCParam->mpClipBoxMax[2]};
    if (clipBoxMin[0] == -9999 && clipBoxMin[1] == -9999 && clipBoxMin[2] == -9999 &&
	clipBoxMax[0] == 9999 && clipBoxMax[1] == 9999 && clipBoxMax[2] == 9999)
	useClip = 0;
    float matrix16[16];
    for (int i = 0; i < 16; i++) matrix16[i] = pCParam->mpMatrix16[i];
    float smoothWidth = pCParam->mSmoothWidth;
    float thinningFactor = pCParam->mThinningFactor;

    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1i (CItem.mVIdUniform[1], textureMode);
    glUniform1i (CItem.mVIdUniform[2], textureColorCurve);
    glUniform1i (CItem.mVIdUniform[3], dissolveMode);
    glUniform1i (CItem.mVIdUniform[4], useClip);
    glUniform3fv(CItem.mVIdUniform[5], 1, clipBoxMin);
    glUniform3fv(CItem.mVIdUniform[6], 1, clipBoxMax);
    glUniformMatrix4fv(CItem.mVIdUniform[7], 1, GL_FALSE, matrix16);
    glUniform1f (CItem.mVIdUniform[8], smoothWidth);
    glUniform1f (CItem.mVIdUniform[9], thinningFactor);
}

void CLS::
BeginStreamlineTriangle(StreamlineTriangleParam_cls* pCParam){
    ShaderItem_cls& CItem = mVCShader[_SHADER_STREAMLINE_TRIANGLE];
    if (CItem.mbCompileTried == FALSE) CItem.MakeShader();
    if (CItem.mbShaderAvailable == FALSE) return;
    
    UseProgram(CItem);
    int lightNumber = 1;
    int textureMode = pCParam->mTextureMode;
    int textureColorCurve = 1;
    int textureSample = 0;
    int dissolveMode = pCParam->mDissolveMode;
    if (dissolveMode >= 1)
	SetEnsembleShaderSetting(CItem, pCParam->mEnsemble);
    float lineWidth = pCParam->mLineWidth;
    float viewportf[4] = {pCParam->mViewport[0], pCParam->mViewport[1],
			  pCParam->mViewport[2], pCParam->mViewport[3]};
    int mappingType = pCParam->mMappingType;
    int useClip = 1;
    float clipBoxMin[3] = {pCParam->mpClipBoxMin[0],
			   pCParam->mpClipBoxMin[1],
			   pCParam->mpClipBoxMin[2]};
    float clipBoxMax[3] = {pCParam->mpClipBoxMax[0],
			   pCParam->mpClipBoxMax[1],
			   pCParam->mpClipBoxMax[2]};
    if (clipBoxMin[0] == -9999 && clipBoxMin[1] == -9999 && clipBoxMin[2] == -9999 &&
	clipBoxMax[0] == 9999 && clipBoxMax[1] == 9999 && clipBoxMax[2] == 9999)
	useClip = 0;
    float matrix16[16];
    for (int i = 0; i < 16; i++) matrix16[i] = pCParam->mpMatrix16[i];
    float smoothWidth = pCParam->mSmoothWidth;
    float thinningFactor = pCParam->mThinningFactor;
	
    SetFogSetting(CItem);

    glUniform1i (CItem.mVIdUniform[0], lightNumber);
    glUniform1i (CItem.mVIdUniform[1], textureMode);
    glUniform1i (CItem.mVIdUniform[2], textureColorCurve);
    glUniform1i (CItem.mVIdUniform[3], textureSample);
    glUniform1i (CItem.mVIdUniform[4], dissolveMode);
    glUniform1f (CItem.mVIdUniform[5], lineWidth);
    glUniform4fv(CItem.mVIdUniform[6], 1, viewportf);
    glUniform1i (CItem.mVIdUniform[7], mappingType);
    glUniform1i (CItem.mVIdUniform[8], useClip);
    glUniform3fv(CItem.mVIdUniform[9], 1, clipBoxMin);
    glUniform3fv(CItem.mVIdUniform[10], 1, clipBoxMax);
    glUniformMatrix4fv(CItem.mVIdUniform[11], 1, GL_FALSE, matrix16);
    glUniform1f (CItem.mVIdUniform[12], smoothWidth);
    glUniform1f (CItem.mVIdUniform[13], thinningFactor);
}

void CLS::
SetCullingSide(int cullingSide){
    ShaderItem_cls& CItem = mVCShader[_SHADER_SV_TRIANGLE_MESH];
    if (CItem.mbShaderAvailable == FALSE) return;
    glUniform1i (CItem.mVIdUniform[2], cullingSide);
}

void CLS::
SetCullingSideHF(int cullingSide){
    ShaderItem_cls& CItem = mVCShader[_SHADER_SV_TRIANGLE_MESH_HF];
    if (CItem.mbShaderAvailable == FALSE) return;
    glUniform1i (CItem.mVIdUniform[2], cullingSide);
}

void CLS::
SetHFFlat(bool bFlat){
    ShaderItem_cls& CItem = mVCShader[_SHADER_HF];
    if (CItem.mbShaderAvailable == FALSE) return;
    glUniform1i (CItem.mVIdUniform[9], bFlat);
}

void CLS::
EndShader(){
    if (mCurrentShader != 0){
	glUseProgram(0);
	mCurrentShader = 0;
    }
}

ShaderCtrl_cls *GpCShaderCtrl;
