// -*- c -*- 
// "clsShaderCtrl.cpp"

#ifndef _SHADER_CTRL_H_
#define _SHADER_CTRL_H_

#include "clsShaderItem.h"
#include "../CONFITEM/clsConfItem.h"
#include "../FILEIO/clsAsciiReader.h"

#include "../COMMON/CommonDefinitionsShader.h"

#define _TEXTURE_DISABLED 0
#define _TEXTURE_ENABLED  1

#include <vector>

class ShaderParams_cls{};

class ShaderParamEnsemble_cls : ShaderParams_cls {
public:
    ShaderParamEnsemble_cls(){};
    ShaderParamEnsemble_cls(int ensemble, int dissolveMode, 
			    int lightNumber, int textureMode);
    
    int mEnsemble, mDissolveMode;
    int mLightNumber, mTextureMode;

    void CopyFrom(ShaderParamEnsemble_cls &CFrom);
};

class ExposureParams_cls : public ShaderParams_cls {
public:
    ExposureParams_cls();
    float mMin, mMax;
    float mGamma;
    int mHue, mSaturation;
};

class BlurParams_cls : public ShaderParams_cls {
public:
    int mType, mKernel, mQuality;
    float mWidth, mGamma;
};

class FocalBlurParams_cls : public ShaderParams_cls {
public:
    int mQuality;
    float mFocalDistance, mWidthBase;
    float mGamma;
};

class FocalBlurParamsB_cls : public ShaderParams_cls {
public:
    int mQuality;
    float mFocalDistance, mWidthBase;
    float mGamma;
};

class GlareParams_cls : public ShaderParams_cls {
public:
    int mType, mQuality, mRayNumber;
    float mWidth, mOffset, mCrit, mEnhance;
    float mFade;
};

class RemoveAlphaParam_cls : public ShaderParams_cls {
public:
    int mExtremeAlpha, mConvert, mBackgroundTile;
};

class OverlayParam_cls : public ShaderParams_cls { 
public:
    float mAlpha;
    float mXScale, mYScale;
};

class SSAOParam_cls : public ShaderParams_cls{
public:
    int mQuality;
    int mType;
    float mFactor, mRadius;
    float mBaseColor[3];
};

class VolumeBasicParams_cls : public ShaderParams_cls {
public:
    int mDiscard0, mTrim, mSubQuality;
    float mClipMin[3], mClipMax[3];
    float mIntervalGlobal;
    int mBlendType;
};

class VolumePhongParams_cls : public ShaderParams_cls {
public:
    int mLightNumber;
    int mDiscard0, mTrim, mSubQuality;
    float mAmbient[3], mDiffise[3], mSpecular[3], mShininess;
    float mClipMin[3], mClipMax[3];
    float mIntervalGlobal;
    float mDelta;
    int mBlendType;
};

class VolumePhongShadowParams_cls : public ShaderParams_cls {
public:
    int mLightNumber;
    int mDiscard0, mTrim, mSubQuality;
    int mShadowQuality;
    float mShadowFactor;
    float mBaseAlpha;
    float mAmbient[3], mDiffise[3], mSpecular[3], mShininess;
    float mClipMin[3], mClipMax[3];
    float mIntervalGlobal;
    float mDelta;
    int mBlendType;
    //float mModelMatrix[16];
};

class VolumeScatterParams_cls : public ShaderParams_cls {
 public:
    int mLightNumber;
    float mScatterFatcor;
    int mDiscard0, mTrim, mSubQuality;
    float mAmbient[3], mDiffise[3], mSpecular[3], mShininess;
    float mClipMin[3], mClipMax[3];
    float mIntervalGlobal;
    int mBlendType;
};

class VolumeScatterShadowParams_cls : public ShaderParams_cls {
 public:
    int mLightNumber;
    int mShadowQuality;
    float mShadowFactor;
    float mBaseAlpha;
    float mScatterFatcor;
    int mDiscard0, mTrim, mSubQuality;
    float mAmbient[3], mDiffise[3], mSpecular[3], mShininess;
    float mClipMin[3], mClipMax[3];
    float mIntervalGlobal;
    int mBlendType;
};

class ObjectGouraudParam_cls : public ShaderParamEnsemble_cls {
 public:
};

class ObjectPhongParam_cls : public ShaderParamEnsemble_cls {
 public:
    bool mbFlat;
    int mUseAlpha;
};

class ObjectToonParam_cls : public ShaderParamEnsemble_cls {
 public:
    float mThreshold, mBoldness;
    bool mbFlat;
};

class ObjectNormalParam_cls : public ShaderParams_cls {
 public:
    bool mbFlat;
};

class LineBasicParam_cls : public ShaderParams_cls {
public:
    int mEnsemble, mDissolveMode;
};

class PointBasicParam_cls : public ShaderParams_cls {
public:
    int mLightNumber, mEnsemble, mDissolveMode, mSizeMode;
    float mDistanceBase;
    int mViewport[4];
};

class PointBillboardParam_cls : public ShaderParams_cls{
public:
    int mLightNumber, mEnsemble, mDissolveMode, mTwistMode;
    //float *mpBaseMatrix16, *mpBaseMatrix16Inv;
    //float *mpViewMatrix16, *mpViewMatrix16Inv;
    float *mpBaseInvModel, *mpModelInvBase;
    float *mpBaseInv;
};

class SphereGouraudParam_cls : public ShaderParams_cls {
    public:
    int mLightNumber, mEnsemble, mDissolveMode;
};

class SpherePhongParam_cls : public ShaderParams_cls {
public:
    int mLightNumber, mEnsemble, mDissolveMode;
    bool mbFlat;
};

class SphereToonParam_cls : public ShaderParams_cls {
public:
    int mLightNumber, mEnsemble, mDissolveMode;//, mTextureMode;
    float mThreshold, mBoldness;
    bool mbFlat;
};

class ShaderSVTriangleMesh_cls : public ShaderParams_cls {
public:
    float mShadowLength;
    float mOffsetFactor;
    int mForClosed;
};

class ShaderHF_cls : public ShaderParamEnsemble_cls {
public:
    int mChannel;
    bool mbFlat;
};

class StreamlineBase_cls : public ShaderParamEnsemble_cls {
public:
    StreamlineBase_cls(){};
    StreamlineBase_cls(ShaderParamEnsemble_cls &CPrarmEnsemble);
    double *mpClipBoxMin, *mpClipBoxMax;
    double *mpMatrix16;
    float mSmoothWidth, mThinningFactor;
    void CopyFrom(StreamlineBase_cls &StreamlineBase_cls);
};

class StreamlineBasicParam_cls : public StreamlineBase_cls {
 public:
};

class StreamlineTriangleParam_cls : public StreamlineBase_cls {
public:
    float mLineWidth;
    int mMappingType;
    int mViewport[4];
};

class ShaderCtrl_cls : public AsciiReader_cls {
 public:
    ShaderCtrl_cls();
    int mCurrentShader;
    wxString mstrCommonDefinitions;
    wxString mstrSpotEffectFunc, mstrFogEffectFunc, mstrNormalFunc;
    wxString mstrEnsembleDitheringFunc;

    int mFogMode;
    float mProjInv11, mProjInv15;
    float mProjMatrix10, mProjMatrix14;
    void PrepareViewportInfo();

    std::vector < ShaderItem_cls > mVCShader;
    
    void UseProgram(ShaderItem_cls &CItem);

    void SetEnsembleShaderSetting(ShaderItem_cls &CItem, int ensemble);
    void SetFogSetting(ShaderItem_cls &CItem);

    //void BeginShader();
    void BeginShader(wxString strShader, ShaderParams_cls *pCParam);

    void BeginClampAlpha(ShaderParams_cls* pCDummy);
    void BeginExposure(ExposureParams_cls* pCParam);
    void BeginBlur(BlurParams_cls* pCParam);
    void BeginFocalBlur(FocalBlurParams_cls* pCParam);
    void BeginFocalBlurB(FocalBlurParamsB_cls* pCParam);
    void BeginGlare(GlareParams_cls* pCParam);
    void BeginRemoveAlpha(RemoveAlphaParam_cls* pCParam);
    void BeginOverlay(OverlayParam_cls* pCParam);
    void BeginSSAO(SSAOParam_cls* pCParam);
    void BeginDepth(ShaderParams_cls* pCParam);
    void BeginVolumeBasic(VolumeBasicParams_cls* pCParam);
    void BeginVolumePhong(VolumePhongParams_cls* pCParam);
    void BeginVolumePhongShadow(VolumePhongShadowParams_cls* pCParam);
    void BeginVolumeScatter(VolumeScatterParams_cls* pCParam);
    void BeginVolumeScatterShadow(VolumeScatterShadowParams_cls* pCParam);
    void BeginObjectGouraud(ObjectGouraudParam_cls* pCParam);
    void BeginObjectPhong(ObjectPhongParam_cls* pCParam);
    void BeginObjectToon(ObjectToonParam_cls* pCParam);
    void BeginObjectProjectNormal(ObjectNormalParam_cls* pCParam);
    void BeginLineBasic(LineBasicParam_cls* pCParam);
    void BeginPointBasic(PointBasicParam_cls* pCParam);
    void BeginPointBillboard(PointBillboardParam_cls* pCParam);
    void BeginSphereGouraud(SphereGouraudParam_cls* pCParam);
    void BeginSpherePhong(SpherePhongParam_cls* pCParam);
    void BeginSphereToon(SphereToonParam_cls* pCParam);
    void BeginSVTrinagleMesh(ShaderSVTriangleMesh_cls* pCParam);
    void BeginHF(ShaderHF_cls* pCParam);
    void BeginSVTrinagleMeshHF(ShaderSVTriangleMesh_cls* pCParam);
    void BeginStreamlineBasic(StreamlineBasicParam_cls* pCParam);
    void BeginStreamlineTriangle(StreamlineTriangleParam_cls* pCParam);

    void SetCullingSide(int cullingSide);
    void SetCullingSideHF(int cullingSide);
    void SetHFFlat(bool bFlat);
    
    void EndShader();
};

#endif
