// -*- c -*-
// "clsShaderItem.h"

#include "clsShaderItem.h"

#include "../../GL/glsl.h"

#include "../COMMON/CommonDefinitions.h"

#include <wx/filefn.h> 

#ifdef WIN32
extern PFNGLPROGRAMPARAMETERIEXTPROC glProgramParameteri;
extern PFNGLBINDATTRIBLOCATIONPROC glBindAttribLocation;
#endif

ShaderStatusVolumeCoord_cls::
ShaderStatusVolumeCoord_cls(){ Clear(); }

void ShaderStatusVolumeCoord_cls::
Clear(){
    mbUse = FALSE; mbUseCompiled = FALSE;
    mType = mTypeCompiled = 0;
    mRMin = mRMinCompiled = 0; mRMax = mRMaxCompiled = 1;
    mCustom0 = mCustom0Compiled = 0;
    mCustom1 = mCustom1Compiled = 1;
    mCustom2 = mCustom2Compiled = 2;
    mstrCustom = mstrCustomCompiled = wxES;
}

void ShaderStatusVolumeCoord_cls::
Update(){
    mbUseCompiled = mbUse;
    mTypeCompiled = mType;
    mRMinCompiled = mRMin; mRMaxCompiled = mRMax;
    mCustom0Compiled = mCustom0;
    mCustom1Compiled = mCustom1;
    mCustom2Compiled = mCustom2;
    mstrCustomCompiled = mstrCustom;
}

// -- 

ShaderStatusVolumeMask_cls::
ShaderStatusVolumeMask_cls(){ Clear(); }

void ShaderStatusVolumeMask_cls::
Clear(){
    mbUse = FALSE; mbUseCompiled = FALSE;
    mMaskType = mMaskTypeCompiled = 0;
    mXMin = mXMinCompiled = 0; mXMax = mXMaxCompiled = 1;
    mYMin = mYMinCompiled = 0; mYMax = mYMaxCompiled = 1;
    mZMin = mZMinCompiled = 0; mZMax = mZMaxCompiled = 1;
}

void ShaderStatusVolumeMask_cls::
Update(){
    mbUseCompiled = mbUse;
    mMaskTypeCompiled = mMaskType;
    mXMinCompiled = mXMin; mXMaxCompiled = mXMax;
    mYMinCompiled = mYMin; mYMaxCompiled = mYMax;
    mZMinCompiled = mZMin; mZMaxCompiled = mZMax;
}

#define CLS ShaderItem_cls

CLS::
ShaderItem_cls(wxString strName){
    mProgram = 0;
    mbDissolve = FALSE;
    mbFog = FALSE;
    mstrShaderName = strName;
    Clear();
}

void CLS::
Clear(){
    if (mProgram != 0) glDeleteProgram(mProgram);
    mProgram = 0;
    mbShaderAvailable = FALSE;
    mbCompileTried = FALSE;
    mCStatusVolumeCoord.Clear();
    mCStatusVolumeMask.Clear();
    mstrVert.clear();
    mstrGeom.clear();
    mstrFrag.clear();
}

void CLS::
SetShaderPath(wxString strPathBase){
    mstrShaderPathBase = strPathBase;

    wxString strPathVert = strPathBase; strPathVert << ".vert";
    wxString strPathGeom = strPathBase; strPathGeom << ".geom";
    wxString strPathFrag = strPathBase; strPathFrag << ".frag";

    if (wxFileExists(strPathVert) == TRUE)
	GetStrFullFromFile(strPathVert, mstrVertSrc);

    if (wxFileExists(strPathGeom) == TRUE)
	GetStrFullFromFile(strPathGeom, mstrGeomSrc);

    if (wxFileExists(strPathFrag) == TRUE)
	GetStrFullFromFile(strPathFrag, mstrFragSrc);
}

int ShaderItem_cls::
GetGLGeometryConstant(wxString strWord){
    if (strWord.IsSameAs(wxT("GL_LINES_ADJACENCY_EXT")))
	return GL_LINES_ADJACENCY_EXT;
    if (strWord.IsSameAs(wxT("GL_LINE_STRIP")))
	return GL_LINE_STRIP;

    if (strWord.IsSameAs(wxT("GL_TRIANGLES_ADJACENCY_EXT"))) 
	return GL_TRIANGLES_ADJACENCY_EXT;
    if (strWord.IsSameAs(wxT("GL_TRIANGLES"))) return GL_TRIANGLES;
    if (strWord.IsSameAs(wxT("GL_TRIANGLE_STRIP"))) return GL_TRIANGLE_STRIP;
    if (strWord.IsSameAs(wxT("GL_POINTS"))) return GL_POINTS;
    if (strWord.IsSameAs(wxT("GL_LINES"))) return GL_LINES;
    return GL_TRIANGLES;
}

void ShaderItem_cls::
ReadGeometryShaderStrForParameter(wxString strLine,
				  int &geometryInput,
				  int &geometryOutput,
				  long &geometryVertices){

    int posSetParam = strLine.Find(wxT("SetParameter"));
    int posEqual = strLine.Find(wxT("="));
    if (posSetParam == wxNOT_FOUND) return;
    if (posEqual == wxNOT_FOUND) return;
    int posKey1 = posSetParam + 13;
    wxString strKey = strLine.Mid(posKey1, posEqual-posKey1);
    wxString strValue = strLine.Mid(posEqual+1);

    if (strKey.IsSameAs(wxT("GL_GEOMETRY_INPUT_TYPE_EXT")))
	geometryInput = GetGLGeometryConstant(strValue);
    if (strKey.IsSameAs(wxT("GL_GEOMETRY_OUTPUT_TYPE_EXT")))
	geometryOutput = GetGLGeometryConstant(strValue);
    if (strKey.IsSameAs(wxT("GL_GEOMETRY_VERTICES_OUT_EXT")))
	strValue.ToLong(&geometryVertices);
}



bool CLS::
ReadShaderString(GLuint& shader, wxString& strShader){
    wxCharBuffer buffer = strShader.ToUTF8();
    int length = strlen(buffer.data());
    const GLchar *pBuffer = buffer.data();
    glShaderSource(shader, 1, &pBuffer, &length);
    return TRUE;
}

void CLS::
MakeShader(){
    static bool bLog = TRUE;
    if (bLog == TRUE)
	wxLogMessage(mstrShaderName + "::Try\n");

    if (mstrVert.IsEmpty() == TRUE) mstrVert = mstrVertSrc;
    if (mstrGeom.IsEmpty() == TRUE) mstrGeom = mstrGeomSrc;
    if (mstrFrag.IsEmpty() == TRUE) mstrFrag = mstrFragSrc;
       
    if (mstrVert.IsEmpty() == TRUE || mstrFrag.IsEmpty() == TRUE) return;
    
    mCStatusVolumeCoord.Update();
    mCStatusVolumeMask.Update();
    mbCompileTried = TRUE;
    //mbCompiledStatusForCoord = mbStatusForCoord;
    
    GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
    GLuint geomShader = 0;
    if (mstrGeom.IsEmpty() == FALSE) 
	geomShader = glCreateShader(GL_GEOMETRY_SHADER);
    GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);

    if (vertShader != 0) ReadShaderString(vertShader, mstrVert);
    if (geomShader != 0) ReadShaderString(geomShader, mstrGeom);
    if (fragShader != 0) ReadShaderString(fragShader, mstrFrag);
    
    // o[ebNXVF[_̃\[XvÕRpC 
    GLint compiled;
    glCompileShader(vertShader);
    glGetShaderiv(vertShader, GL_COMPILE_STATUS, &compiled);
    if (compiled == GL_FALSE) {
	AppendLog(vertShader, mstrShaderName + "::vertex\n", _GLSL_SHADER);
	mbShaderAvailable = FALSE; 
	return;
    }
    
    if (bLog == TRUE)
	wxLogMessage(mstrShaderName + "::vertex::Success\n");

    // WIgVF[_[̃\[XvÕRpC 
    if (geomShader != 0){
	glCompileShader(geomShader);
	glGetShaderiv(geomShader, GL_COMPILE_STATUS, &compiled);
	if (compiled == GL_FALSE) {
	    AppendLog(geomShader, mstrShaderName + "::geom\n", _GLSL_SHADER);
	    mbShaderAvailable = FALSE; 
	    return;
	}
	if (bLog == TRUE)
	    wxLogMessage(mstrShaderName + "::geom::Success\n");
    }
    
    // tOgVF[_̃\[XvÕRpC 
    glCompileShader(fragShader);
    glGetShaderiv(fragShader, GL_COMPILE_STATUS, &compiled);
    if (compiled == GL_FALSE) {
	AppendLog(fragShader, mstrShaderName + "::fragment\n", _GLSL_SHADER);
	//wxLogMessage(mstrFrag);
	mbShaderAvailable = FALSE; 
	return;
    }
    if (bLog == TRUE)
	wxLogMessage(mstrShaderName + "::fragment::Success\n");
    
    // vOIuWFNg̍쐬 
    mProgram = glCreateProgram();

    // VF[_IuWFNg̃VF[_vOւ̓o^ 
    glAttachShader(mProgram, vertShader);

    // WIgVF[_[̃VF[_vOւ̓o^ 
    if (geomShader != 0){
	
	int geometryInput = GL_TRIANGLES;
	int geometryOutput = GL_TRIANGLES;
	long geometryVertices = 6;

	wxString strBuffer = mstrGeom;
	int pos = 0;
	while (TRUE){
	    int length = strBuffer.Mid(pos).Find("\n");
	    if (length == wxNOT_FOUND) break;
	    wxString strLine = strBuffer.Mid(pos, length);
	    ReadGeometryShaderStrForParameter
		(strLine, geometryInput, geometryOutput, geometryVertices);
	    pos += length + 1;
	}

	/*
	for (wxString strLine = CFile.GetFirstLine(); !CFile.Eof();
	     strLine = CFile.GetNextLine() ){
	    ReadGeometryShaderStrForParameter
		(strLine, geometryInput, geometryOutput, geometryVertices);
	}
	*/

	glAttachShader(mProgram, geomShader);

	int max;
	glGetIntegerv(GL_MAX_GEOMETRY_OUTPUT_VERTICES_EXT, &max);

	if (glProgramParameteri != 0){
	    //glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	    glProgramParameteri(mProgram, GL_GEOMETRY_VERTICES_OUT_EXT, geometryVertices);
	    glProgramParameteri(mProgram, GL_GEOMETRY_INPUT_TYPE_EXT, geometryInput);
	    glProgramParameteri(mProgram, GL_GEOMETRY_OUTPUT_TYPE_EXT, geometryOutput );
	    //CheckGLError("GL_GEOMETRY_OUTPUT_TYPE_EXT", TRUE); // somehow returns error other than GL_POINTS.
	}
    }
	
    
    glAttachShader(mProgram, fragShader);

    // VF[_vÕN 
    GLint linked;
    glLinkProgram(mProgram);
    glGetProgramiv(mProgram, GL_LINK_STATUS, &linked);

    if (linked == GL_FALSE) {
	AppendLog(mProgram, mstrShaderName + "::Link\n", _GLSL_PROGRAM);
	mbShaderAvailable = FALSE;
	return;
    }

    if (bLog == TRUE)
	wxLogMessage(mstrShaderName + "::Link::Success\n");

    // VF[_IuWFNg̍폜 
    glDeleteShader(vertShader);
    if (geomShader != 0) glDeleteShader(geomShader);
    glDeleteShader(fragShader);

    glUseProgram(mProgram);
    uint numUniform = mVstrUniform.size();
    for (uint i = 0; i < numUniform; i++){
	wxCharBuffer buffer = mVstrUniform[i].ToUTF8();
	mVIdUniform.push_back(glGetUniformLocation(mProgram, buffer.data()));
    }
    if (mbDissolve == TRUE){
	mIdEnsembleParamA = glGetUniformLocation(mProgram, "ensembleParamA");
	mIdEnsembleParamB = glGetUniformLocation(mProgram, "ensembleParamB");
	mIdEnsembleParamC = glGetUniformLocation(mProgram, "ensembleParamC");
    }
    if (mbFog == TRUE){
	mIdFogMode = glGetUniformLocation(mProgram, "fogMode");
	mIdProjInv11 = glGetUniformLocation(mProgram, "projInv11");
	mIdProjInv15 = glGetUniformLocation(mProgram, "projInv15");
    }

    uint numAttrib = mVstrAttrib.size();
    for (uint i = 0; i < numAttrib; i++){
	wxCharBuffer buffer = mVstrAttrib[i].ToUTF8();
	glBindAttribLocation(mProgram, i, buffer.data());
    }

    glUseProgram(0);

    if (bLog == TRUE)
	wxLogMessage(mstrShaderName + "::FullSuccess\n");
    

    mbShaderAvailable = TRUE;
}

void CLS::
AppendLog(int shader_or_program, wxString strTag, int type){
    int logSize;
    int length;
    wxString strMes;

    if (shader_or_program == -1){
	wxLogMessage(strTag + "::Unrecognized Error\n");
	return;
    }
    
    if (type == _GLSL_PROGRAM)
	glGetProgramiv(shader_or_program, GL_INFO_LOG_LENGTH, &logSize);
    else
	glGetShaderiv(shader_or_program, GL_INFO_LOG_LENGTH , &logSize);

    char *pLog = new char [logSize];

    if (type == _GLSL_PROGRAM)
	glGetProgramInfoLog(shader_or_program, logSize, &length, pLog);
    else
	glGetShaderInfoLog(shader_or_program, logSize, &length, pLog);

    wxString strLog(pLog, wxConvFile);

    wxLogMessage(strTag + strLog);
    
    delete [] pLog;
    
}
