// (setq c-basic-offset 4)
// "clsColorCurveEditor.h"

#include "clsColorCurveEditor.h"

#include "clsGraphColorCurve.h"
#include "clsGraphEditorTics.h"
#include "clsGraphNaviCurrent.h"
#include "clsGraphNumericalPanel.h"
#include "clsGraphPanel.h"

#include "../COMMON/clsQueue.h"
#include "../MISC/clsMisc.h"
#include "../TIMELINE/clsTimelineAA.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../PREFERENCE/clsShortcut.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/EX/clsPanelBar.h"
#include "../WIDGETCTRL/EX/clsColourDialogEX.h"
#include "../WIDGETCTRL/INPUTITEM/clsTextButton.h"
#include "../WIDGETCTRL/INPUTITEM/clsVCenter.h"
#include "../WIDGETCTRL/INPUTITEM/clsIconSelector.h"
#include "../WIDGETCTRL/WX/clsColorCurveProvider.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsSizeAdjuster.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include <wx/clipbrd.h>
#include <wx/bmpcbox.h>

#define CLS ColorCurveEditor_cls

#define EVT_BUTTON_RANGE(id1, id2, func) wx__DECLARE_EVT2(wxEVT_COMMAND_BUTTON_CLICKED, id1, id2, wxCommandEventHandler(func))

BEGIN_EVENT_TABLE(CLS, EditorBase_cls)
EVT_BUTTON(btnNumPanelHide, CLS::OnHideAndShowNumPanel)
EVT_BUTTON(menuPanelView, CLS::OnMenuView)
EVT_BUTTON(menuPanelEdit, CLS::OnMenuEdit)
EVT_BUTTON(menuPanelKey, CLS::OnMenuKey)
EVT_MENU(menuShowNumPanel, CLS::OnHideAndShowNumPanel)
EVT_MENU(menuCopyToClipboard, CLS::OnCopyToClipboard)
EVT_MENU(menuPasteFromClipboard, CLS::OnPasteFromClipboard)
EVT_MENU(menuCopyColor, CLS::OnCopyColor)
EVT_MENU(menuPasteColor, CLS::OnPasteColor)
EVT_BUTTON(btnDropdown, CLS::OnDropDown)
EVT_BUTTON(btnTicsMode, CLS::OnTicsMode)
EVT_BUTTON(btnGraphRescale, CLS::OnRescale)
EVT_BUTTON(btnCtrlDefaultKey, CLS::OnButtonDefaultKey)
EVT_MENU_RANGE(menuZoomPlus, menuZoomMinus, CLS::OnMenuZoom)
EVT_MENU_RANGE(menuPop0, menuPop9, CLS::OnPopMenu)
EVT_MENU(menuDelete, CLS::OnDeleteSelectedKey)
EVT_BUTTON(btnGraphColor, CLS::OnKeyColor)
EVT_BUTTON(btnGraphFlipColor, CLS::OnFlipColor)
EVT_BUTTON(btnGraphSwapRGB, CLS::OnSwapRGB)
EVT_COMBOBOX(cmbDefaultKey, CLS::OnCmbDefaultKey)
END_EVENT_TABLE()

CLS::
ColorCurveEditor_cls(wxWindow *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
EditorBase_cls("ColorCurveEditor_cls", pCParent) {

#ifdef WIN32
    Freeze();
#endif

    SetToDefaultParam();
    SetPanelType(_PANEL_COLOR_CURVE);
    
    mpCPanelBar = new PanelBar_cls(this);
    mpCPanelBar->SetPanelType(_PANEL_CHILD, _PANEL_COLOR_CURVE);

    wxBoxSizer *pBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerBar = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerInfo = new wxBoxSizer(wxHORIZONTAL);

    wxBoxSizer *pBoxSizerSplit = new wxBoxSizer(wxHORIZONTAL);
    
    wxBoxSizer *pBoxSizerNumPanel = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerMainGraph = new wxBoxSizer(wxVERTICAL);

    wxBoxSizer *pBoxSizerAA = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerTicX = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerColorCurve = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerNavi = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerLock = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerH = new wxBoxSizer(wxHORIZONTAL);
    
    wxBoxSizer *pBoxSizerKeyCtrl = new wxBoxSizer(wxHORIZONTAL);

    ConstructTypeSelector(pBoxSizerBar);
    
    mpCTextMenuView = new TextButton_cls
	(mpCPanelBar, menuPanelView, GetPCStringProvider()->GetStrMenu("View"));
    mpCTextMenuView->SetEventTarget(this);
    VCenter_cls *pCMenuView = new VCenter_cls(mpCTextMenuView);

    mpCTextMenuEdit = new TextButton_cls
	(mpCPanelBar, menuPanelEdit, GetPCStringProvider()->GetStrMenu("Edit"));
    mpCTextMenuEdit->SetEventTarget(this);
    VCenter_cls *pCMenuEdit = new VCenter_cls(mpCTextMenuEdit);
	 
    mpCTextMenuKey = new TextButton_cls
	 (mpCPanelBar, menuPanelKey, GetPCStringProvider()->GetStrMenu("Key"));
    mpCTextMenuKey->SetEventTarget(this);
    VCenter_cls *pCMenuKey = new VCenter_cls(mpCTextMenuKey);

    pBoxSizerBar->Add(pCMenuView, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuEdit, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuKey, 0, wxEXPAND);

    pBoxSizerBar->AddStretchSpacer(1);
    mpCPanelBar->ConstructBarItemDefault(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemHideShow(pBoxSizerBar);

    mpCPanelBar->SetSizer(pBoxSizerBar);

    wxSize CSizeNumPanel = wxSize(265, -1);
    mpCNumPanel = new GraphNumericalPanel_cls(this, CSizeNumPanel);
    mpCNumPanel->Hide();

    mCClosedItem.SetLink(&mpCNumPanel->mAstrClosedItem);
    
    BitmapPanel_cls *pCBtnPinMode = BitmapPanel_cls::GetNewPCButtonPin(this);
    BitmapPanel_cls *pCBtnDropDown = BitmapPanel_cls::GetNewPCButtonDropDown(this);
    VCenter_cls *pCVCenterDropDown = new VCenter_cls(pCBtnDropDown);

    mpCLinkedParamGroup = new wxStaticText(this, wxID_ANY, "dummy");
    mpCBtnNumPanelHide = BitmapPanel_cls::GetNewPCButtonPanelHide(this);
	
    pBoxSizerInfo->Add(pCBtnPinMode, 0, wxEXPAND);
    pBoxSizerInfo->Add(mpCLinkedParamGroup, 0, wxEXPAND);
    pBoxSizerInfo->AddSpacer(4);
    pBoxSizerInfo->Add(pCVCenterDropDown, 0, wxEXPAND);
    pBoxSizerInfo->AddStretchSpacer(1);
    pBoxSizerInfo->Add(mpCBtnNumPanelHide, 0, wxEXPAND);
        
    mpCBtnLinked = BitmapPanel_cls::GetNewPCButtonLink(this); 
    mpCTimelineAA = new TimelineAA_cls((Timeline_cls*)this);
    mpCTimelineAA->mbForColorCurve = TRUE;
    pBoxSizerAA->Add(mpCBtnLinked, 0, 0);
    pBoxSizerAA->AddSpacer(120 - 18);
    pBoxSizerAA->Add(mpCTimelineAA, 1, wxEXPAND);

    BitmapPanel_cls *pCBtnTicsMode = new BitmapPanel_cls(this, btnTicsMode, wxDP, wxSize(16, 16));
    pCBtnTicsMode->AddBitmapStr("wxART_SELECTOR_N", "wxART_SELECTOR_P");
    pCBtnTicsMode->mbButton = TRUE;
    mpCTicsMode = new wxStaticText(this, wxID_ANY, Tics_cls::msAstrTicsModeCC[0], wxDP, wxSize(102, 16));
    /*
    wxStaticText *pCTicsMode = new wxStaticText(this, wxID_ANY, "X-axis value", wxDP, wxSize(16+102, 16));
    */

    mpCTicsX     =   new GraphEditorTics_cls(this); 
    mpCTicsY     =   new GraphEditorTics_cls(this, TRUE);
    mpCColorCurve = new GraphColorCurve_cls(this);
    mpCNaviCurrent = new GraphNaviCurrent_cls(this);
    mpCGraphPanel = new GraphPanel_cls(this, TRUE, FALSE);
    mpCTicsX->mbForColorCurve = TRUE;
    mpCColorCurve->mbForColorCurve = TRUE;
    mpCNaviCurrent->mbForColorCurve = TRUE;

    mpCTicsX->mbLocked = TRUE;
    //mpCTicsY->mbLocked = TRUE;

    mpCTicsX->SetXViewRange(0, 1);
    mpCTicsY->SetYViewRange(0, 1);
    mpCTicsY->SetRegion(-0.02, 1.05);
    mpCTicsY->SetBorderY(0, 2);

    //pBoxSizerTicX->Add(pCTicsMode, 0, wxEXPAND);
    pBoxSizerTicX->Add(pCBtnTicsMode, 0, wxEXPAND);
    pBoxSizerTicX->Add(mpCTicsMode, 0, wxEXPAND);
    pBoxSizerTicX->Add(mpCTicsX, 1, wxEXPAND);

    std::vector < wxString > VstrART;
    VstrART.push_back("wxART_ALPHA_OFF");
    VstrART.push_back("wxART_ALPHA_ON");

    mpCIconSelectorAlpha = new IconSelector_cls
	(this, wxID_ANY, 1, VstrART, wxDP, wxSize(118, 24));
    mpCIconSelectorAlpha->SetItemSize(wxSize(22, 22));
    mpCIconSelectorAlpha->SetBackgroundColour(wxColour(188, 218, 254));
    mpCIconSelectorAlpha->mVstrTag.push_back("RGB");
    mpCIconSelectorAlpha->mVstrTag.push_back("RGBA");

    pBoxSizerColorCurve->Add(mpCIconSelectorAlpha);
    //pBoxSizerColorCurve->AddSpacer(120-2);
    pBoxSizerColorCurve->Add(mpCColorCurve, 1, wxEXPAND);
    
    pBoxSizerNavi->AddSpacer(120-2);
    pBoxSizerNavi->Add(mpCNaviCurrent, 1, wxEXPAND);
    
    wxBoxSizer *pBoxSizerLockItem[4];
    for (int i = 0; i < 4; i++){
	pBoxSizerLockItem[i] = new wxBoxSizer(wxHORIZONTAL);
	mpCLockGroup[i] = new wxStaticText(this, wxID_ANY, "--", wxDP, wxSize(24, -1));
	mpCBtnLock[i] = BitmapPanel_cls::GetNewPCButtonLock(this, btnGraphLock0+i);
	VCenter_cls *pCVCenter = new VCenter_cls(mpCBtnLock[i]);
	
	mpCLockGroup[i]->Disable();
	pBoxSizerLockItem[i]->Add(mpCLockGroup[i], 0, wxEXPAND);
	pBoxSizerLockItem[i]->Add(pCVCenter, 0, wxEXPAND);
    }

    pBoxSizerLock->Add(pBoxSizerLockItem[0], 0, wxEXPAND);
    pBoxSizerLock->Add(pBoxSizerLockItem[1], 0, wxEXPAND);
    pBoxSizerLock->Add(pBoxSizerLockItem[2], 0, wxEXPAND);
    pBoxSizerLock->Add(pBoxSizerLockItem[3], 0, wxEXPAND);
    pBoxSizerLock->AddStretchSpacer(1);

    pBoxSizerH->AddSpacer(1);
    pBoxSizerH->Add(pBoxSizerLock, 0, wxEXPAND);
    pBoxSizerH->Add(mpCTicsY, 0, wxEXPAND);
    pBoxSizerH->Add(mpCGraphPanel, 1, wxEXPAND);
    pBoxSizerH->AddSpacer(2);

    wxSize CSize1(40, 28);
    wxSize CSize2(-1, 28);
    mpCBtnRescale = new wxButton (this, btnGraphRescale, _("Rescale"), wxDP, CSize2);
    mpCBtnAddKey = new wxButton (this, btnGraphAddKey, _("Insert"), wxDP, CSize2);
    mpCBtnRemKey = new wxButton (this, btnGraphRemKey, _("Delete"), wxDP, CSize2);
    mpCBtnLft = new wxButton(this, btnGraphLft, wxT("<"), wxDP, CSize1);
    mpCBtnRgt = new wxButton(this, btnGraphRgt, wxT(">"), wxDP, CSize1);

    VCenter_cls *pCDefaultKeyIcon = SetDefaultKey(this);

    //mpCBtnColor = new wxButton(this, btnGraphColor, " ", wxDP, CSize1);
    mpCBtnColor = new BitmapPanel_cls(this, btnGraphColor, wxDP, CSize1);
    mpCBtnColor->Connect(wxEVT_RIGHT_DOWN, wxMouseEventHandler(ColorCurveEditor_cls::OnBtnColorRight), NULL, this);

    mpCBtnFlipColor = new wxButton(this, btnGraphFlipColor, _("Flip Color"), wxDP, CSize2);
    mpCBtnSwapRGB = new wxButton(this, btnGraphSwapRGB, _("Swap RGB"), wxDP, CSize2);

        //pBoxSizerKeyCtrl->AddSpacer(160);
    pBoxSizerKeyCtrl->Add(mpCBtnRescale, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddStretchSpacer(1);
    pBoxSizerKeyCtrl->Add(mpCBtnLft, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddSpacer(6);
    pBoxSizerKeyCtrl->Add(mpCBtnAddKey, 0, wxEXPAND);
    pBoxSizerKeyCtrl->Add(mpCBtnRemKey, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddSpacer(6);
    pBoxSizerKeyCtrl->Add(mpCBtnRgt, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddSpacer(12);
    pBoxSizerKeyCtrl->Add(pCDefaultKeyIcon, 0, wxEXPAND);
    pBoxSizerKeyCtrl->Add(mpCDefaultKey, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddSpacer(12);
    pBoxSizerKeyCtrl->Add(mpCBtnColor, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddStretchSpacer(1);
    pBoxSizerKeyCtrl->Add(mpCBtnFlipColor, 0, wxEXPAND);
    pBoxSizerKeyCtrl->Add(mpCBtnSwapRGB, 0, wxEXPAND);

    pBoxSizerSrc->Add(mpCPanelBar, 0, wxEXPAND);

    pBoxSizerMainGraph->Add(pBoxSizerInfo, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerAA, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerTicX, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerColorCurve, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerNavi, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerH, 1, wxEXPAND);
    pBoxSizerNumPanel->Add(mpCNumPanel, 1, wxEXPAND);

    pBoxSizerSplit->Add(pBoxSizerMainGraph, 1, wxEXPAND);
    pBoxSizerSplit->Add(pBoxSizerNumPanel, 0, wxEXPAND);

    pBoxSizerSrc->Add(pBoxSizerSplit, 1, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerKeyCtrl, 0, wxEXPAND);

    UpdateDefaultKeyIcon();

#ifndef WIN32
    Freeze();
#endif

    SetSizerAndFit(pBoxSizerSrc);   

    Connect(mpCIconSelectorAlpha->GetId(), wxEVT_COMMAND_COMBOBOX_SELECTED, 
	    wxCommandEventHandler(CLS::OnSelector));

    Thaw();
}

void CLS::
SetToDefaultParam(){
    mCConfList.SetTag("ColorCuerveEditor");
   
    wxString strAAMax; strAAMax << _COLOR_CURVE_RESOLUTION;
    mCConfList.Add(&mCColorCurveMode, "ColorCurveMode=", "1");
    mCConfList.Add(&mCbTemplateAlpha, "TemplateAlpha=", "0");
    mCConfList.Add(&mCAAMin, "AAMin=", "0");
    mCConfList.Add(&mCAAMax, "AAMax=", strAAMax);

    mCConfList.Add(&mCXRange[0], "XRangeMin=", "0");
    mCConfList.Add(&mCXRange[1], "XRangeMax=", "1");

    // XRangeMin, XRangeMax́AvWFNgt@CɏoKv
    // ͂Ȃ̂ŁAo^ȂĂǂB

    mCTicsMode.SetBodyFromData((long long)2);
}

void CLS::
SyncronizeParamToPanelXRange(){
    ColorCurve_cls *pCColorCurve = (ColorCurve_cls*)GetPCParamGroup();
    if (pCColorCurve == NULL) return;

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    ParamGroup_cls *pCRange = pCColorCurve->GetPCParamGroup("Range");
    std::vector < double > VRange = pCRange->GetVInterpolatedValue(NT);

    bool bUndoGroup = pCRange->mCbLinked.GetBodyBOOL();
    bool bGroupOpened = FALSE;
    for (int i = 0; i < 2; i++){
	double newValue = mCXRange[i].GetBodyDOUBLE();
	Parameter_cls &CParam = pCRange->mVCParam[i];
	if (Misc_cls::IsNearlySame(VRange[i], newValue, 1e-6) == FALSE){
	    if (bUndoGroup == TRUE && bGroupOpened == FALSE){
		GetPCUndo()->OpenBlock("SyncronizeRange");
		bGroupOpened = TRUE;
	    }
	    wxString strTag = mCstrLinkedParamGroup.GetBodySTR();
	    //PopLastItemFromStrLine(strTag, "::");
	    strTag << "::Range::" << CParam.GetStrName();
	    CParam.SetValueAtCommand(NT, newValue, strTag, 1, 1);
	}
    }
	
    if (bGroupOpened == TRUE)
	GetPCUndo()->CloseBlock("SyncronizeRange");

    /*
    float min = VRange[0];//pCColorCurve->mCXRangeMin.GetBodyDOUBLE();
    float minPanel = mCXRangeMin.GetBodyDOUBLE();
    
    if (Misc_cls::IsNearlySame(min, minPanel, 1e-6) == FALSE){
	//wxString strNewMin; strNewMin << minPanel;
	//pCColorCurve->mCXRangeMin.PushCommandPair(strNewMin);
    }

    float max = VRange[1];//pCColorCurve->mCXRangeMax.GetBodyDOUBLE();
    float maxPanel = mCXRangeMax.GetBodyDOUBLE();
    
    if (Misc_cls::IsNearlySame(max, maxPanel, 1e-6) == FALSE){
	//wxString strNewMax; strNewMax << maxPanel;
	//pCColorCurve->mCXRangeMax.PushCommandPair(strNewMax);
    }
    */
}

void CLS::
SetColorCurveTemplate(int index){
    ColorCurve_cls *pCColorCurve = (ColorCurve_cls*)GetPCParamGroup();
    if (pCColorCurve == NULL) return;
    wxString strTag = mCstrLinkedParamGroup.GetBodySTR();
    PopLastItemFromStrLine(strTag, "::");
    bool bAlpha = mCbTemplateAlpha.GetBodyBOOL();
    pCColorCurve->CopyColorCurveCommand(GetPCColorCurveProvider()->GetPCColorCurve(index), strTag, bAlpha);
}

void CLS::
ProcAfterMath(const wxString& strCommand, wxString strItem, bool bForce){
    if (strCommand.IsSameAs("DeleteObject")){
	wxString strTgt = mCstrLinkedParamGroup.GetBodySTR();
	wxString strController = PopStrItemFromStrLine(strTgt, "::");
	if (strController.IsSameAs("SceneObjCtrl")){
	    PopStrItemFromStrLine(strTgt, "::");
	    wxString strObj = PopStrItemFromStrLine(strTgt, "::");
	    if (strObj.IsSameAs(strItem)){
		mCstrLinkedParamGroup.SetBodyFromData("None");
		mpCLinkedParamGroup->SetLabel(mCstrLinkedParamGroup.GetBodySTR());
		Layout();
	    }
	}
    }

    if (strCommand.IsSameAs("RenameObject")){
	wxString strNameOld = PopStrItemFromStrLine(strItem, "::");
	wxString strNameNew = strItem;

	wxString strTgt = mCstrLinkedParamGroup.GetBodySTR();
	wxString strController = PopStrItemFromStrLine(strTgt, "::");
	if (strController.IsSameAs("SceneObjCtrl")){
	    PopStrItemFromStrLine(strTgt, "::");
	    wxString strObj = PopStrItemFromStrLine(strTgt, "::");
	    if (strObj.IsSameAs(strNameOld)){
		mCstrLinkedParamGroup.mstrBody.Replace(strNameOld, strNameNew);
		mpCLinkedParamGroup->SetLabel(mCstrLinkedParamGroup.GetBodySTR());
		UpdatePanels("ALL");
	    }
	}
    }

    if (strCommand.IsSameAs("SelectColorCurve") ||
	strCommand.IsSameAs("ColorCurveChange")){
	if (bForce == FALSE && mCbPinned.GetBodyBOOL() == TRUE) return;
	mCstrLinkedParamGroup.SetBodyFromData(strItem);
	mpCLinkedParamGroup->SetLabel(strItem);
	UpdatePanels("ALL");
    }
}

void CLS::
HideAndShowNumPanel(){
    if (mCbNumPanelShown.GetBodyBOOL() == TRUE &&
	mpCNumPanel->IsShown() == FALSE){
	mpCPanelBar->Freeze();
	mpCBtnNumPanelHide->Hide();
	mpCNumPanel->Show();
	Layout();
	UpdateGraphs();
	mpCPanelBar->Thaw();
    }
    if (mCbNumPanelShown.GetBodyBOOL() == FALSE &&
	mpCNumPanel->IsShown() == TRUE){
	mpCPanelBar->Freeze();
	mpCBtnNumPanelHide->Show();
	mpCNumPanel->Hide();
	Layout();
	UpdateGraphs();
	mpCPanelBar->Thaw();
    }
}

void CLS::
OnHideAndShowNumPanel(wxComEvt& event){
    if (mpCNumPanel->IsShown() == TRUE)
	mCbNumPanelShown.SetBodyFromData((long long)0);
    else
	mCbNumPanelShown.SetBodyFromData((long long)1);
    HideAndShowNumPanel();
}

void CLS::
OnMenuView(wxComEvt& event){
    Shortcut_cls *pCSC = GetPCShortcut();
    wxMenu CMenu;
    pCSC->SetMenuItem(&CMenu, menuZoomPlus);
    pCSC->SetMenuItem(&CMenu, menuZoomMinus);
    //CMenu.Append(menuZoomPlus, GetPCShortcut()->GetMenuString(menuZoomPlus));
    //CMenu.Append(menuZoomMinus, GetPCShortcut()->GetMenuString(menuZoomMinus));
    CMenu.AppendSeparator();
    pCSC->SetMenuCheckItem(&CMenu, menuShowNumPanel, mCbNumPanelShown.GetBodyBOOL());
    //bool bPanel = mCbNumPanelShown.GetBodyBOOL();
    //CMenu.AppendCheckItem(menuShowNumPanel, GetPCShortcut()->GetMenuString(menuShowNumPanel))->Check(bPanel);
    CMenu.AppendSeparator();
    pCSC->SetMenuCheckItem(&CMenu, menuShowPanelManage, mpCPanelBar->mbShowItem);
    //bool bBarItem = mpCPanelBar->mbShowItem;
    //CMenu.AppendCheckItem(menuShowPanelManage, GetPCShortcut()->GetMenuString(menuShowPanelManage))->Check(bBarItem);

    mpCTextMenuView->PopMenu(this, &CMenu);
}

void CLS::
OnMenuEdit(wxComEvt& event){
    Shortcut_cls *pCSC = GetPCShortcut();
    wxMenu CMenu;
    pCSC->SetMenuItem(&CMenu, menuCopyToClipboard);
    pCSC->SetMenuItem(&CMenu, menuPasteFromClipboard);
    //CMenu.Append(menuCopyToClipboard, GetPCShortcut()->GetMenuString(menuCopyToClipboard));
    //CMenu.Append(menuPasteFromClipboard, GetPCShortcut()->GetMenuString(menuPasteFromClipboard));
    mpCTextMenuEdit->PopMenu(this, &CMenu);
}

void CLS::
OnMenuKey(wxComEvt& event){
    wxMenu CMenu;
    CMenu.Append(menuDelete, GetPCShortcut()->GetMenuString(menuDelete));
    
    mpCTextMenuKey->PopMenu(this, &CMenu);
}

void CLS::
OnMenuZoom(wxComEvt& event){
    bool bExpand = (event.GetId() == menuZoomPlus);
    ExpandAAForColor(bExpand, GetNaviCurrent() * _COLOR_CURVE_RESOLUTION);
    UpdatePanels("ALL");
}

void CLS::
OnDropDown(wxComEvt& event){
    mCSelector.PopupColorCurveSelector(this); }

void CLS::
OnTicsMode(wxComEvt& event){
    long long mode = mCTicsMode.GetBodyINT() - 2;
    mode++;
    if (mode > 1) mode = 0;
    mpCTicsMode->SetLabel(Tics_cls::msAstrTicsModeCC[mode]);
    mCTicsMode.SetBodyFromData(mode + 2);
    mpCTicsX->Refresh();
    mpCTicsY->Refresh();
}

void CLS::
OnRescale(wxComEvt& event){
    mpCTicsY->SetRegion(-0.05, 1.05);
    UpdateGraphs();
}

void CLS::
OnPopMenu(wxComEvt& event){
    wxString strPopMenu = mCSelector.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == TRUE) return;
    
    wxString strCommand = "SelectColorCurve";
    wxString strItem;
    strItem << GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strPopMenu;
    ProcAfterMath(strCommand, strItem, TRUE);
}

/*
void CLS::
OnDeleteSelectedKey(wxComEvt& event){
}
*/

void CLS::
OnKeyColor(wxComEvt& event){
    
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;

    float Xnavi = GetNaviCurrent();
    std::vector < float > Vclr = pCParamGroup->GetVInterpolatedValue(Xnavi);
    wxColourDialogEX *pCDLG = GetPCColourDialog();

    pCDLG->SetColourFloat(Vclr.data());

#ifdef WIN32
    wxPoint CPoint = wxGetMousePosition();
    wxRect CRect(CPoint-wxPoint(250, 250), wxSize(545, 385));
    GetPCSizeAdjuster()->AdjustWindowFitting(pCDLG, CRect);
#endif

    if (pCDLG->ShowModal() == wxID_CANCEL) return;

    mCCurrent[0].SetBodyFromData((double)(pCDLG->GetR()) / 255);
    mCCurrent[1].SetBodyFromData((double)(pCDLG->GetG()) / 255);
    mCCurrent[2].SetBodyFromData((double)(pCDLG->GetB()) / 255);
    SyncronizeParamToPanel(TRUE);
}

void CLS::
OnFlipColor(wxComEvt& event){

    ColorCurve_cls *pCColorCurve = (ColorCurve_cls*)GetPCParamGroup();
    if (pCColorCurve == NULL) return;
    wxString strTag = mCstrLinkedParamGroup.GetBodySTR();
    //PopLastItemFromStrLine(strTag, "::");
    PopLastItemFromStrLine(strTag, "::");

    ColorCurve_cls *pCColorCurveNew = new ColorCurve_cls(pCColorCurve->GetStrName(), GetPCStringProvider()->GetArrayString("RGBA"));
    pCColorCurveNew->CopyColorCurveFromP(pCColorCurve);
    pCColorCurveNew->Flip();

    pCColorCurve->CopyColorCurveCommand(pCColorCurveNew, strTag, TRUE);

    delete pCColorCurveNew;
}

void CLS::
OnSwapRGB(wxComEvt& event){

    ColorCurve_cls *pCColorCurve = (ColorCurve_cls*)GetPCParamGroup();
    if (pCColorCurve == NULL) return;
    wxString strTag = mCstrLinkedParamGroup.GetBodySTR();
    PopLastItemFromStrLine(strTag, "::");

    ColorCurve_cls *pCColorCurveNew = new ColorCurve_cls(pCColorCurve->GetStrName(), GetPCStringProvider()->GetArrayString("RGBA"));
    pCColorCurveNew->CopyColorCurveFromP(pCColorCurve);
    pCColorCurveNew->SwapRGB();

    pCColorCurve->CopyColorCurveCommand(pCColorCurveNew, strTag, TRUE);

    delete pCColorCurveNew;
}

void CLS::
OnCopyToClipboard(wxComEvt& event){
    ColorCurve_cls *pCColorCurve = (ColorCurve_cls*)GetPCParamGroup();
    if (pCColorCurve == NULL) return;

    wxString strSerialize;
    pCColorCurve->Serialize(strSerialize);
    
    if (wxTheClipboard->Open()){    
	wxTheClipboard->SetData( new wxTextDataObject(strSerialize) );
	wxTheClipboard->Close();
    }
}

void CLS::
OnPasteFromClipboard(wxComEvt& event){

    ColorCurve_cls *pCColorCurve = (ColorCurve_cls*)GetPCParamGroup();
    if (pCColorCurve == NULL) return;

    wxString strSerialize;
    if (wxTheClipboard->Open()){
	if (wxTheClipboard->IsSupported( wxDF_TEXT )){
	    wxTextDataObject data;
	    wxTheClipboard->GetData( data );
	    strSerialize = data.GetText();
	}
	wxTheClipboard->Close();
    }

    if (strSerialize.IsEmpty() == TRUE) return;
    if (strSerialize.Mid(0, 14).IsSameAs("ReadColorCurve") == FALSE) return;

    wxString strTag = mCstrLinkedParamGroup.GetBodySTR();
    PopLastItemFromStrLine(strTag, "::");

    pCColorCurve->Paste(strSerialize, strTag);
}

void CLS::
OnCopyColor(wxComEvt& event){
    wxString strSerialize = "COLOR=";
    strSerialize << mCCurrent[0].GetBodyDOUBLE() << " " << mCCurrent[1].GetBodyDOUBLE() << " " << mCCurrent[2].GetBodyDOUBLE() << ENDL;
    
    wxTextDataObject *pCData = new wxTextDataObject(strSerialize);
    wxTheClipboard->Open();
    wxTheClipboard->SetData(pCData);
    wxTheClipboard->Close();
}

void CLS::
OnPasteColor(wxComEvt& event){
    if (wxTheClipboard->IsSupported(wxDataFormat(wxDF_TEXT)) == FALSE) return;
    wxTextDataObject CData;
    wxTheClipboard->Open();
    wxTheClipboard->GetData(CData);
    wxTheClipboard->Close();
    wxString strData = CData.GetText();

    if (strData.Matches(wxT("COLOR=*")) == TRUE){
	ConfItemFLOATARRAY_cls CArray(wxT("COLOR="), wxT("0.5 0.5 0.5 1"));
	CArray.ReadStrLine(strData);
	
	std::vector < float > VColor = CArray.GetBodyFLOATARRAY();
	
	mCCurrent[0].SetBodyFromData(VColor[0]);
	mCCurrent[1].SetBodyFromData(VColor[1]);
	mCCurrent[2].SetBodyFromData(VColor[2]);
	SyncronizeParamToPanel(TRUE);
    }
}

void CLS::
OnBtnColorRight(wxMouseEvent& event){
    wxMenu CMenu;
    CMenu.Append(menuCopyColor, _("Copy Color"));
    CMenu.Append(menuPasteColor, _("Paste Color"));
    PopupMenu(&CMenu);
}

void CLS::
OnSelector(wxComEvt& event){
    mCColorCurveMode.SetBodyFromData((long long)mpCIconSelectorAlpha->GetValueInt());
    mpCColorCurve->Refresh(FALSE);
}

void CLS::
OnCmbDefaultKey(wxComEvt& event){
    ProcDefaultKey(event);
}

void CLS::
OnButtonDefaultKey(wxComEvt& event){
    ProcButtonDefaultKey();
}

void CLS::
UpdateColorBar(){
    //mpCColorCurve->Refresh(FALSE)
    mpCColorCurve->CheckRefresh();
}

void CLS::
UpdateColorButton(){
    wxColour CClr(mCCurrent[0].GetBodyDOUBLE() * 255,
		  mCCurrent[1].GetBodyDOUBLE() * 255,
		  mCCurrent[2].GetBodyDOUBLE() * 255);
    mpCBtnColor->SetBackgroundColour(CClr);
    mpCBtnColor->Refresh();
}

void CLS::
UpdatePanels(const wxString& strPanelTag, bool bUpdateBar){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs(wxT("ALL"))){

	/*
	if (Queue_cls::msbQueueSyncronizeGraphNavi == TRUE){
	    mCNaviCurrent.SetBodyFromData(GetPCTimeCtrl()->GetCurrentNT());
	}
	*/

	//if (mCbZoomAuto.GetBodyBOOL() == TRUE) Rescale(FALSE);
	if (IsUpdateTooFast() == FALSE){
	    UpdateGraphs();
	    UpdateLock();
	    UpdateNumPanels();
	    UpdateColorBar();
	    UpdateColorButton();
	    GetPCMain()->UpdatePanels((std::string)GetName());
	    mPastUpdateTimeStamp = mCStopWatch.Time();
	}
	
	if (bUpdateBar == TRUE) mpCPanelBar->Refresh();

	ParamGroup_cls *pCParamGroup = GetPCParamGroup();
	if (pCParamGroup == NULL)
	    mpCBtnLinked->Disable();
	else{
	    mpCBtnLinked->Enable();
	    mpCBtnLinked->SetStatus(pCParamGroup->mCbLinked.GetBodyINT());
	}

	UpdateDefaultKeyIcon();

	mpCBtnLinked->Refresh();
    }
}

void CLS::
Serialize(wxString& strOut){
    wxString strViewY; 
    strViewY << mpCGraphPanel->mViewMinY << " " << mpCGraphPanel->mViewMaxY; 
    mCViewY.SetBodyFromData(strViewY);
    ConfListContainer_cls::Serialize(strOut);
}

void CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == FALSE)
	return;
    
    mpCLinkedParamGroup->SetLabel(mCstrLinkedParamGroup.GetBodySTR());
    
    mpCNumPanel->SyncStateToData();
    mpCTicsY->mRegionMin = mCViewY.GetBodyFLOATARRAY()[0];
    mpCTicsY->mRegionMax = mCViewY.GetBodyFLOATARRAY()[1];
    mpCNumPanel->UpdateTemplateButtons();
    HideAndShowNumPanel();
}

