// (setq c-basic-offset 4)
// "clsEditorBase.h"

#include "clsEditorBase.h"

#include "clsGraphColorCurve.h"
#include "clsGraphEditorTics.h"
#include "clsGraphEditor.h"
#include "clsGraphNaviCurrent.h"
#include "clsGraphNumericalPanel.h"
#include "clsGraphPanel.h"
#include "clsColorCurveEditor.h"

#include "../MISC/clsMisc.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../PREFERENCE/clsShortcut.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../TIMELINE/clsTimelineAA.h"
#include "../TIMELINE/clsTimeline.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"

#include "../../clsMain.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS EditorBase_cls

#define EVT_BUTTON_RANGE(id1, id2, func) wx__DECLARE_EVT2(wxEVT_COMMAND_BUTTON_CLICKED, id1, id2, wxCommandEventHandler(func))

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
EVT_BUTTON(btnGraphPin, CLS::OnPinMode)
EVT_BUTTON(btnGraphLinked, CLS::OnLinked)
EVT_BUTTON_RANGE(btnGraphLock0, btnGraphLock3, CLS::OnLock)
EVT_BUTTON_RANGE(btnGraphLft, btnGraphRgt, CLS::OnLR)
EVT_BUTTON(btnGraphAddKey, CLS::OnAddKey)
EVT_BUTTON(btnGraphRemKey, CLS::OnRemKey)
EVT_SIZE(CLS::OnSize)
EVT_CHAR_HOOK(CLS::OnKeyDown)
END_EVENT_TABLE()

CLS::
EditorBase_cls(wxString strClassName, wxWindow *pCParent) : Timeline_cls(pCParent) {
    mstrClassName = strClassName;
    SetToDefaultParamBase();
    mCStopWatch.Start();
    mPastUpdateTimeStamp = 0;
}

void CLS::
SetToDefaultParamBase(){
    mCConfList.Clear();
    mCConfList.Add(&mCstrLinkedParamGroup, "LinkedParamGroup=", "None");
    mCConfList.Add(&mCbPinned, "Pinned=", "0");
    mCConfList.Add(&mCNaviCurrent, "NaviCurrent=", "0");
    mCConfList.Add(&mCbNumPanelShown, "NumPanelShown=", "0");
    mCConfList.Add(&mCClosedItem, "ClosedItem=", wxES);

    mCConfList.Add(&mCViewY, "ViewY=", "0 1");

    mCConfList.Add(&mCCurrent[0], "Current0=", "0");
    mCConfList.Add(&mCCurrent[1], "Current1=", "0");
    mCConfList.Add(&mCCurrent[2], "Current2=", "0");
    mCConfList.Add(&mCCurrent[3], "Current3=", "0");

    mCConfList.Add(&mCActiveParamIndex, "ActiveParamIndex=", "-1");
    mCConfList.Add(&mCActiveKeyXnorm, "ActiveKeyXnorm=", "0");
    mCConfList.Add(&mCActiveKeyValue, "ActiveKeyValue=", "0");
    mCConfList.Add(&mCActiveKeyType, "ActiveKeyType=", "0");
    mCConfList.Add(&mCActiveKeyEaseFactor, "ActiveKeyEaseFactor=", "1");
    // ActiveKeyCurrent́AvWFNgt@CɏoKv
    // ͂Ȃ̂ŁAo^ȂĂǂB
}

ParamGroup_cls* CLS::
GetPCParamGroup(){
    wxString strTgt = mCstrLinkedParamGroup.GetBodySTR();
    return GetPCMain()->GetPCParamGroup(strTgt);
}

Parameter_cls* CLS::
GetPCParameterActive(){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return NULL;
    int paramIndex = mCActiveParamIndex.GetBodyINT();
    if (paramIndex == -1) return NULL;
    return &pCParamGroup->mVCParam[paramIndex];
}

double CLS::
GetNaviCurrent(){
    if (mDraggingItem == _TL_SEL_GRAPH_NAVI) return mDraggingNormalTime;
    return mCNaviCurrent.GetBodyDOUBLE();
}

void CLS::
MoveNextKey(int direction){
    double XNormalized = mCNaviCurrent.GetBodyDOUBLE();
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL){ 
	if (direction == -1) mCNaviCurrent.SetBodyFromData((double)0);
	if (direction == +1) mCNaviCurrent.SetBodyFromData((double)1);
	UpdateGraphs();
	return;
    }
    uint num = pCParamGroup->mVCParam.size();
    double nextNX = 1;
    if (direction == -1) nextNX = 0;
    for (uint i = 0; i < num; i++){
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	int adjuscent = CParam.GetAdjuscentKeyIndex(direction, XNormalized);
	double candidNX = CParam.GetKeyXNormalizeEX(adjuscent);
	if (direction == -1)
	    nextNX = Misc_cls::Max(nextNX, candidNX);
	if (direction == +1)
	    nextNX = Misc_cls::Min(nextNX, candidNX);
    }
    mCNaviCurrent.SetBodyFromData(nextNX);
    UpdateGraphs();
    UpdateNumPanels();
    if (mstrClassName.IsSameAs("ColorCurveEditor_cls")){
	((ColorCurveEditor_cls*)this)->UpdateColorButton();
    }
}

bool CLS::
IsUpdateTooFast(){
    if (mCStopWatch.Time() < mPastUpdateTimeStamp + 30) return TRUE;
    return FALSE;
}

void CLS::
UpdateNumPanels(){
    
    ParamGroup_cls* pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL){
	mCCurrent[0].SetBodyFromData(0.0);
	mCCurrent[1].SetBodyFromData(0.0);
	mCCurrent[2].SetBodyFromData(0.0);
	mCCurrent[3].SetBodyFromData(0.0);

	mCActiveKeyXnorm.SetBodyFromData(0.0);
	mCActiveKeyValue.SetBodyFromData(0.0);
	mCActiveKeyEaseFactor.SetBodyFromData(0.0);
    }
    else{

	if (mstrClassName.IsSameAs("ColorCurveEditor_cls")){
	    ColorCurveEditor_cls *pCCCE = (ColorCurveEditor_cls*)this;
	    ColorCurve_cls *pCCC = (ColorCurve_cls*)pCParamGroup;
	    ParamGroup_cls *pCRange = pCCC->GetPCParamGroup("Range");
	    double NT = GetPCTimeCtrl()->GetCurrentNT();
	    std::vector < double > VRange = pCRange->GetVInterpolatedValue(NT);
	    pCCCE->mCXRange[0].SetBodyFromData(VRange[0]);
	    pCCCE->mCXRange[1].SetBodyFromData(VRange[1]);
	}

	double Xnavi = GetNaviCurrent();
	uint paramNum = pCParamGroup->mVCParam.size();
	for (int i = 0; i < paramNum; i++)
	    mCCurrent[i].SetBodyFromData(pCParamGroup->mVCParam[i].GetValue(Xnavi));
	if (pCParamGroup != mpCParamGroupUsed){
	    mpCNumPanel->UpdateLabels(pCParamGroup);
	    mCActiveParamIndex.SetBodyFromData((long long)-1); 
	}

	int paramIndex = mCActiveParamIndex.GetBodyINT();
	if (paramIndex != -1){
	    Parameter_cls &CParam = pCParamGroup->mVCParam[paramIndex];
	    int activeKeyIndex = CParam.GetActiveKeyIndex();
	    if (activeKeyIndex != -1){
		mCActiveKeyXnorm.SetBodyFromData(CParam.mVCKeys[activeKeyIndex].mXNormalized);
		mCActiveKeyValue.SetBodyFromData(CParam.mVCKeys[activeKeyIndex].mValue);
		mCActiveKeyType.SetBodyFromData((long long)CParam.mVCKeys[activeKeyIndex].mKeyType);
		mCActiveKeyEaseFactor.SetBodyFromData(CParam.mVCKeys[activeKeyIndex].mEaseFactor);
	    }
	}
    }

    mpCParamGroupUsed = pCParamGroup;
    
    mpCNumPanel->UpdateValues();
}

void CLS::
UpdateGraphs(){
    mpCTimelineAA->Refresh();
    mpCTicsX->Refresh();
    mpCTicsY->Refresh();
    mpCNaviCurrent->Refresh();

    double viewXMin = (double)mCAAMin.GetBodyINT() / _COLOR_CURVE_RESOLUTION;
    double viewXMax = (double)mCAAMax.GetBodyINT() / _COLOR_CURVE_RESOLUTION;
    
    if (mstrClassName.IsSameAs("GraphEditor_cls")){
	viewXMin = GetPCTimeCtrl()->GetNormalizedValue
	    (mCAAMin.GetBodyINT(), 0);
	viewXMax = GetPCTimeCtrl()->GetNormalizedValue
	    (mCAAMax.GetBodyINT(), 0);
    }
    
    mpCGraphPanel->SetXViewRange(viewXMin, viewXMax);
    mpCGraphPanel->SetYViewRange(mpCTicsY->mRegionMin,mpCTicsY->mRegionMax);
    mpCGraphPanel->SetYValueRange(mpCTicsY->mRegionMin,mpCTicsY->mRegionMax);
    mpCGraphPanel->Refresh();
    
    if (mstrClassName.IsSameAs("ColorCurveEditor_cls"))
	return ((ColorCurveEditor_cls*)this)->mpCColorCurve->CheckRefresh();
}

void CLS::
UpdateLock(){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return DisableLock();

    uint size = pCParamGroup->mVCParam.size();
    for (uint i = 0; i < size; i++){
	wxString strLabel = pCParamGroup->mVCParam[i].GetStrName();
	if (mpCLockGroup[i]->GetLabel().IsSameAs(strLabel) == FALSE)
	    mpCLockGroup[i]->SetLabel(strLabel);
	mpCLockGroup[i]->Enable();
	mpCBtnLock[i]->Enable();
	mpCBtnLock[i]->SetStatus(pCParamGroup->mVCParam[i].mCbLocked.GetBodyINT());
    }
    for (uint i = size; i < 4; i++){
	wxString strLabel = "--";
	mpCLockGroup[i]->SetLabel("--");
	if (mpCLockGroup[i]->GetLabel().IsSameAs(strLabel) == FALSE)
	    mpCLockGroup[i]->SetLabel(strLabel);
	mpCLockGroup[i]->Disable();
	mpCBtnLock[i]->Disable();
	mpCBtnLock[i]->Refresh(TRUE);
    }
}

void CLS::
DisableLock(){
    for (int i = 0; i < 4; i++){
	mpCLockGroup[i]->SetLabel("--");
	mpCLockGroup[i]->Disable();
	mpCBtnLock[i]->Disable();
	mpCBtnLock[i]->Refresh(TRUE);
    }
}

void CLS::
OnPinMode(wxCommandEvent& event){
    long long pinned = event.GetInt();
    mCbPinned.SetBodyFromData(pinned);
}

void CLS::
OnLinked(wxCommandEvent& event){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;
    GetPCMain()->SetParamGroupConfCommand(mCstrLinkedParamGroup.GetBodySTR(), "Linked=", event.GetInt());
}

void CLS::
OnLock(wxCommandEvent& event){
    int i = event.GetId() - btnGraphLock0;
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;
    Parameter_cls &pCParamer = pCParamGroup->mVCParam[i];
    
    wxString strItem = mCstrLinkedParamGroup.GetBodySTR() + "::" + pCParamer.mCstrName.GetBodySTR();

    GetPCMain()->SetParameterConfCommand(strItem, "Locked=", mpCBtnLock[i]->mState);
    UpdateLock();
}

void CLS::
OnLR(wxCommandEvent& event){
    if (event.GetId() == btnGraphLft) MoveNextKey(-1);
    if (event.GetId() == btnGraphRgt) MoveNextKey(+1);
}

void CLS::
OnAddKey(wxCommandEvent& event){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;

    int keyType;
    double easeFactor;
    GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);

    pCParamGroup->InsertKeyCommand(mCNaviCurrent.GetBodyDOUBLE(), mCstrLinkedParamGroup.GetBodySTR(),
				   keyType, easeFactor);
}

void CLS::
OnRemKey(wxCommandEvent& event){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;
    pCParamGroup->DeleteKeyCommand(mCNaviCurrent.GetBodyDOUBLE(), mCstrLinkedParamGroup.GetBodySTR());
}

void CLS::
OnDeleteSelectedKey(wxCommandEvent& event){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;
    pCParamGroup->DeleteSelectedTLKeyCommand(mCstrLinkedParamGroup.GetBodySTR());
}

void CLS::
OnSize(wxSizeEvent& event){
    Layout();
    UpdateGraphs(); 

    event.Skip();
}

void CLS::
OnKeyDown(wxKeyEvent& event){
    if (GetPCShortcut()->ProcShortcutKey(event, this) == FALSE)
	event.Skip();
}

void CLS::
SyncronizeParamToPanel(bool bUndoGroup){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;

    if (mstrClassName.IsSameAs("ColorCurveEditor_cls"))
	((ColorCurveEditor_cls*)this)->SyncronizeParamToPanelXRange();

    bool bGroupOpened = FALSE;
    double Xnavi = GetNaviCurrent();
    int paramNum = pCParamGroup->mVCParam.size();
    for (int i = 0; i < paramNum; i++){
	double newValue = mCCurrent[i].GetBodyDOUBLE();
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	if (Misc_cls::IsNearlySame(CParam.GetValue(Xnavi), newValue, 1e-6) == FALSE){
	    if (bUndoGroup == TRUE && bGroupOpened == FALSE){
		GetPCUndo()->OpenBlock("SyncronizeParam");
		bGroupOpened = TRUE;
	    }
	    wxString strTag = mCstrLinkedParamGroup.GetBodySTR() << "::" << CParam.GetStrName();
	    
	    CParam.SetValueAtCommand(Xnavi, newValue, strTag);
	}
    }
    
    if (bGroupOpened == TRUE)
	GetPCUndo()->CloseBlock("SyncronizeParam");
}

void CLS::
SyncronizeKeyToPanel(){
    Parameter_cls *pCParam = GetPCParameterActive();
    if (pCParam != NULL){
	int activeKeyIndex = pCParam->GetActiveKeyIndex();
	if (activeKeyIndex != -1){
	    int newType = mCActiveKeyType.GetBodyINT();
	    wxString strTag = mCstrLinkedParamGroup.GetBodySTR() << "::" << pCParam->GetStrName();
	    if (pCParam->mVCKeys[activeKeyIndex].mKeyType != newType)
		pCParam->SetKeyTypeCommand(activeKeyIndex, newType, strTag);

	    double newValue = mCActiveKeyValue.GetBodyDOUBLE();
	    double oldValue = pCParam->mVCKeys[activeKeyIndex].mValue;

	    double newX = mCActiveKeyXnorm.GetBodyDOUBLE();
	    double oldX = pCParam->mVCKeys[activeKeyIndex].mXNormalized;
	    if (Misc_cls::IsNearlySame(newX, oldX, 1e-6) == FALSE ||
		Misc_cls::IsNearlySame(newValue, oldValue, 1e-6) == FALSE)
		pCParam->MoveTLKeyCommand(activeKeyIndex, newX - oldX, newValue - oldValue, strTag);

	    double newEF = mCActiveKeyEaseFactor.GetBodyDOUBLE();
	    double oldEF = pCParam->mVCKeys[activeKeyIndex].mEaseFactor;
	    if (Misc_cls::IsNearlySame(newEF, oldEF, 1e-6) == FALSE)
		pCParam->SetKeyEaseFactorCommand(activeKeyIndex, newEF, strTag);
	}
    }
}

void CLS::
SyncronizeTicsToGraph(){
    mpCTicsY->mRegionMin = mpCGraphPanel->mViewMinY;
    mpCTicsY->mRegionMax = mpCGraphPanel->mViewMaxY;
}


