// -*- c++ -*-
// "clsGraphColorCurve.h"

#include "clsGraphColorCurve.h"
#include "clsColorCurveEditor.h"
#include "clsEditorBase.h"

#include "../PARAMETER/clsParamGroup.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsFontProvider.h"

#include "../COMMON/CommonDefinitions.h"

#include "../../wxWidgetEnumMenu.h"

#define CLS GraphColorCurve_cls

BEGIN_EVENT_TABLE(CLS, TimelineBase_cls)
EVT_PAINT(CLS::OnPaint)
EVT_SIZE(CLS::OnSize)
EVT_RIGHT_DOWN(CLS::OnBtnRight)
END_EVENT_TABLE()

CLS::
GraphColorCurve_cls(EditorBase_cls *pCParent) :
TimelineBase_cls(pCParent, _TL_GRAPH, wxSize(-1, 25)){
    mpCParent = pCParent;
    //mCIcon.CopyFromBitmap(ART("wxART_NAVISLIDER"));
}

void CLS::
CheckRefresh(){
    ColorCurve_cls *pCColorCurve = (ColorCurve_cls*)mpCParent->GetPCParamGroup();
    if (pCColorCurve == NULL){
	if (mpCColorCurveDrawed == NULL) return;
	return Refresh(FALSE);
    }
    
    UpdateState();

    int mode = ((ColorCurveEditor_cls*)mpCParent)->mCColorCurveMode.GetBodyINT();
    wxString strCounter = pCColorCurve->GetStrProcCounter();
    strCounter << mode;
    strCounter << mAAMinNormalized << mAAMaxNormalized;
    if (strCounter.IsSameAs(mstrCounterDrawed)) return;

    mstrCounterDrawed = strCounter;
    Refresh(FALSE);
}

void CLS::
UpdateDisplay(){

    //dc.SetBackground(wxBrush(wxColour(216,216,216), wxSOLID));
    
    UpdateState();
    SetTLRegion(mAAMinNormalized, mAAMaxNormalized);
    
    //DrawCurrentTime(dc, 0);

    int draggingX = 0;
    int sliderX = GetDevCoordFromN(mpCParent->mCNaviCurrent.GetBodyDOUBLE(), 0)-7, sliderY = 4;

    ColorCurve_cls *pCColorCurve = (ColorCurve_cls*)mpCParent->GetPCParamGroup();
    mpCColorCurveDrawed = pCColorCurve;

    if (pCColorCurve == NULL){
	wxAutoBufferedPaintDC dc(this);
	ColorProvider_cls::SetBGColorDC(dc, "BasePanel");
	dc.Clear();
	dc.SetFont(FontProvider_cls::GetFont(0));
	ColorProvider_cls::SetTextColorWeakDC(dc);
	wxString strMes = "Not Defined";
	int width = FontProvider_cls::GetTextWidth(0, strMes);
	dc.DrawText(strMes, (mCSize.GetWidth() - width) / 2, 4);
	int posX0 = GetDevCoordFromN(0, 0);
	int posX1 = GetDevCoordFromN(1, 0);
	dc.SetPen(ColorProvider_cls::GetColor("TextWeak"));
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	dc.DrawRectangle(posX0, 0, posX1-posX0, 24);
    }
    else{
	int mode = ((ColorCurveEditor_cls*)mpCParent)->mCColorCurveMode.GetBodyINT();

	/*
#ifdef WIN32
	wxString strCounter = pCColorCurve->GetStrProcCounter();
	strCounter << mode;
	if (strCounter.IsSameAs(mstrCounterDrawed)) return;
	mstrCounterDrawed = strCounter;
#endif
	*/

	wxAutoBufferedPaintDC dc(this);
	ColorProvider_cls::SetBGColorDC(dc, "BasePanel");
	dc.Clear();
	
	pCColorCurve->PrecalcBegin();
	int width = mCSize.GetWidth() - mBorderXL - mBorderXR;
	if (mode == _COLOR_CURVE_RGB)
	    for (int i = 0; i < width; i++){
		double x0_ = (double)i / width;
		double x1_ = (double)(i+1) / width;
		double x0 = x0_ * (mTLRegionMax-mTLRegionMin) + mTLRegionMin;
		double x1 = x1_ * (mTLRegionMax-mTLRegionMin) + mTLRegionMin;
		int posX0 = GetDevCoordFromN(x0, 0);
		int posX1 = GetDevCoordFromN(x1, 0);
		
		wxColour CClr = pCColorCurve->GetColor(x0);
		dc.SetPen(CClr);
		dc.SetBrush(CClr);
		dc.DrawRectangle(posX0, 0, posX1-posX0, 24);
	    }
	if (mode == _COLOR_CURVE_RGBA)
	    for (int i = 0; i < width; i++){
		double x0_ = (double)i / width;
		double x1_ = (double)(i+1) / width;
		double x0 = x0_ * (mTLRegionMax-mTLRegionMin) + mTLRegionMin;
		double x1 = x1_ * (mTLRegionMax-mTLRegionMin) + mTLRegionMin;
		int posX0 = GetDevCoordFromN(x0, 0);
		int posX1 = GetDevCoordFromN(x1, 0);
		
		double alpha;
		wxColour CClr = pCColorCurve->GetColor(x0, &alpha);

		int bg0 = (posX0 / 14) % 2 * 64 + 98;
		int bg1 = ((posX0 / 14) + 1) % 2 * 64 + 98;

		wxColour CClrBG0(bg0, bg0, bg0);
		wxColour CClrBG1(bg1, bg1, bg1);
		
		wxColour CClr0 =
		    ColorCurve_cls::AlphaBlend(CClrBG0, CClr, alpha);
		
		dc.SetPen(CClr0);
		dc.SetBrush(CClr0);
		dc.DrawRectangle(posX0, 0, posX1-posX0, 12);

		wxColour CClr1 =
		    ColorCurve_cls::AlphaBlend(CClrBG1, CClr, alpha);
		/*
		wxColour CClr2(bg1 * (1.0-alpha) + alpha * VClr[0] * 255,
			       bg1 * (1.0-alpha) + alpha * VClr[1] * 255,
			       bg1 * (1.0-alpha) + alpha * VClr[2] * 255);
		*/
		dc.SetPen(CClr1);
		dc.SetBrush(CClr1);
		dc.DrawRectangle(posX0, 12, posX1-posX0, 12);
	    }
	pCColorCurve->PrecalcEnd();

	//mTLRegionMinDrown = mTLRegionMin;
	//mTLRegionMaxDrown = mTLRegionMax;
    }

    /*
    if (mSelected == _TL_SEL_SLIDER){
	sliderX = GetDevCoordFromN(mQuantizedTimeDragging, 0)-7;
	draggingX = sliderX; }
    */

    /*
    //dc.DrawIcon(mCIcon, sliderX, sliderY);
    //mRectSlider = wxRect(sliderX, sliderY, 15, 18);
    
    // bZ[W`
    int offset = 0;
    if (draggingX > mPanelWidth * 0.7) offset = -40;

    if (HasCapture() == TRUE){
	wxString Mes;
	Mes << mFrameDragging;
	dc.SetFont(*wxNORMAL_FONT);
	dc.SetTextForeground(*wxWHITE);
	dc.DrawText(Mes, draggingX + 15 + offset, 0);
    }
    */
}

void CLS::
OnPaint(wxPaintEvent& event){
    //wxAutoBufferedPaintDC dc(this);
    UpdateDisplay();
}

void CLS::
OnSize(wxSizeEvent& event){
    mCSize = event.GetSize();
    Refresh();
    event.Skip();
}

void CLS::
OnBtnRight(wxMouseEvent& event){
    wxMenu CMenu;
    CMenu.Append(menuCopyToClipboard, _("Copy Color Curve"));
    CMenu.Append(menuPasteFromClipboard, _("Paste Color Curve"));
    PopupMenu(&CMenu);
}
