// (setq c-basic-offset 4)
// "clsGraphEditor.h"

#include "clsGraphEditor.h"

#include "clsGraphEditorTics.h"
#include "clsGraphNaviCurrent.h"
#include "clsGraphNumericalPanel.h"
#include "clsGraphPanel.h"

#include "../COMMON/clsQueue.h"
#include "../TIMELINE/clsTimelineAA.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../PREFERENCE/clsShortcut.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/EX/clsPanelBar.h"
#include "../WIDGETCTRL/INPUTITEM/clsTextButton.h"
#include "../WIDGETCTRL/INPUTITEM/clsVCenter.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"
#include "../WIDGETCTRL/WX/myArtProv.h"
#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include <wx/bmpcbox.h>

#define CLS GraphEditor_cls

#define EVT_BUTTON_RANGE(id1, id2, func) wx__DECLARE_EVT2(wxEVT_COMMAND_BUTTON_CLICKED, id1, id2, wxCommandEventHandler(func))

BEGIN_EVENT_TABLE(CLS, EditorBase_cls)
EVT_BUTTON(btnNumPanelHide, CLS::OnHideAndShowNumPanel)
EVT_BUTTON(menuPanelView, CLS::OnMenuView)
EVT_BUTTON(menuPanelChannel, CLS::OnMenuChannel)
EVT_BUTTON(menuPanelKey, CLS::OnMenuKey)
EVT_MENU(menuShowNumPanel, CLS::OnHideAndShowNumPanel)
EVT_BUTTON(btnDropdown, CLS::OnDropDown)
EVT_BUTTON(btnTicsMode, CLS::OnTicsMode)
EVT_BUTTON(btnGraphRescale, CLS::OnRescale)
EVT_BUTTON(btnCtrlDefaultKey, CLS::OnButtonDefaultKey)
EVT_MENU_RANGE(menuExtrapolateOff, menuExtrapolateAllOn, CLS::OnMenuExtrapolate)
EVT_MENU_RANGE(menuZoomPlus, menuZoomMinus, CLS::OnMenuZoom)
EVT_MENU_RANGE(menuPop0, menuPop9, CLS::OnPopMenu)
EVT_MENU(menuDelete, CLS::OnDeleteSelectedKey)
EVT_BUTTON_RANGE(btnGraphZoomPlus, btnGraphZoomMinus, CLS::OnZoom)
EVT_BUTTON(btnGraphZoomAuto, CLS::OnZoomAuto)
EVT_COMBOBOX(cmbDefaultKey, CLS::OnCmbDefaultKey)

END_EVENT_TABLE()

CLS::
GraphEditor_cls(wxWindow *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
EditorBase_cls("GraphEditor_cls", pCParent) {
    //Freeze();

    SetToDefaultParam();
    SetPanelType(_PANEL_GRAPH);

    mpCPanelBar = new PanelBar_cls(this);
    mpCPanelBar->SetPanelType(_PANEL_CHILD, _PANEL_GRAPH);

    wxBoxSizer *pBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerBar = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerInfo = new wxBoxSizer(wxHORIZONTAL);

    wxBoxSizer *pBoxSizerSplit = new wxBoxSizer(wxHORIZONTAL);
    
    wxBoxSizer *pBoxSizerNumPanel = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerMainGraph = new wxBoxSizer(wxVERTICAL);

    wxBoxSizer *pBoxSizerAA = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerTicX = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerNavi = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerLock = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerH = new wxBoxSizer(wxHORIZONTAL);

    wxBoxSizer *pBoxSizerKeyCtrl = new wxBoxSizer(wxHORIZONTAL);

    ConstructTypeSelector(pBoxSizerBar);
    
    mpCTextMenuView = new TextButton_cls
	(mpCPanelBar, menuPanelView, GetPCStringProvider()->GetStrMenu("View"));
    mpCTextMenuView->SetEventTarget(this);
    VCenter_cls *pCMenuView = new VCenter_cls(mpCTextMenuView);

    mpCTextMenuChannel = new TextButton_cls
	(mpCPanelBar, menuPanelChannel, GetPCStringProvider()->GetStrMenu("Channel"));
    mpCTextMenuChannel->SetEventTarget(this);
    VCenter_cls *pCMenuChannel = new VCenter_cls(mpCTextMenuChannel);
    
    mpCTextMenuKey = new TextButton_cls
	(mpCPanelBar, menuPanelKey, GetPCStringProvider()->GetStrMenu("Key"));
    mpCTextMenuKey->SetEventTarget(this);
    VCenter_cls *pCMenuKey = new VCenter_cls(mpCTextMenuKey);

    pBoxSizerBar->Add(pCMenuView, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuChannel, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuKey, 0, wxEXPAND);
    pBoxSizerBar->AddStretchSpacer(1);
    mpCPanelBar->ConstructBarItemDefault(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemHideShow(pBoxSizerBar);

    mpCPanelBar->SetSizer(pBoxSizerBar);

    wxSize CSizeNumPanel = wxSize(265, -1);
    mpCNumPanel = new GraphNumericalPanel_cls(this, CSizeNumPanel);
    mpCNumPanel->Hide();

    mCClosedItem.SetLink(&mpCNumPanel->mAstrClosedItem);

    BitmapPanel_cls *pCBtnPinMode = BitmapPanel_cls::GetNewPCButtonPin(this);
    BitmapPanel_cls *pCBtnDropDown = BitmapPanel_cls::GetNewPCButtonDropDown(this);
    VCenter_cls *pCVCenterDropDown = new VCenter_cls(pCBtnDropDown);

    mpCLinkedParamGroup = new wxStaticText(this, wxID_ANY, "dummy");
    mpCBtnNumPanelHide = BitmapPanel_cls::GetNewPCButtonPanelHide(this);
   
    pBoxSizerInfo->Add(pCBtnPinMode, 0, wxEXPAND);
    pBoxSizerInfo->Add(mpCLinkedParamGroup, 0, wxEXPAND);
    pBoxSizerInfo->AddSpacer(4);
    pBoxSizerInfo->Add(pCVCenterDropDown, 0, wxEXPAND);
    pBoxSizerInfo->AddStretchSpacer(1);
    pBoxSizerInfo->Add(mpCBtnNumPanelHide, 0, wxEXPAND);

    mpCBtnLinked = BitmapPanel_cls::GetNewPCButtonLink(this); 
    mpCTimelineAA = new TimelineAA_cls((Timeline_cls*)this);
        
    pBoxSizerAA->Add(mpCBtnLinked, 0, 0);
    pBoxSizerAA->AddSpacer(120 - 18);
    pBoxSizerAA->Add(mpCTimelineAA, 1, wxEXPAND);

    BitmapPanel_cls *pCBtnTicsMode = new BitmapPanel_cls(this, btnTicsMode, wxDP, wxSize(16, 16));
    pCBtnTicsMode->AddBitmapStr("wxART_SELECTOR_N", "wxART_SELECTOR_P");
    pCBtnTicsMode->mbButton = TRUE;
    mpCTicsMode = new wxStaticText(this, wxID_ANY, Tics_cls::msAstrTicsMode[0], wxDP, wxSize(102, 16));
    mpCTicsX     =   new GraphEditorTics_cls(this); 
    mpCTicsY     =   new GraphEditorTics_cls(this, TRUE);
    mpCNaviCurrent = new GraphNaviCurrent_cls(this);
    mpCGraphPanel = new GraphPanel_cls(this, FALSE, FALSE);
    
    mpCTicsX->SetXViewRange(0, 1);
    mpCTicsY->SetYViewRange(0, 1);
    mpCTicsY->SetRegion(-0.5, 0.5);
    mpCTicsY->SetBorderY(0, 2);

    pBoxSizerTicX->Add(pCBtnTicsMode, 0, wxEXPAND);
    pBoxSizerTicX->Add(mpCTicsMode, 0, wxEXPAND);
    pBoxSizerTicX->Add(mpCTicsX, 1, wxEXPAND);

    mpCCheckSnap = new wxCheckBox(this, cbGraphSnap, _("Snapping"), wxDP, wxSize(118, -1));

    //pBoxSizerNavi->AddSpacer(120 - 2);
    pBoxSizerNavi->Add(mpCCheckSnap, 0, wxEXPAND);
    pBoxSizerNavi->Add(mpCNaviCurrent, 1, wxEXPAND);
    
    wxBoxSizer *pBoxSizerLockItem[4];
    for (int i = 0; i < 4; i++){
	pBoxSizerLockItem[i] = new wxBoxSizer(wxHORIZONTAL);
	mpCLockGroup[i] = new wxStaticText(this, wxID_ANY, "--", wxDP, wxSize(24, -1));
	mpCBtnLock[i] = BitmapPanel_cls::GetNewPCButtonLock(this, btnGraphLock0+i);
	VCenter_cls *pCVCenter = new VCenter_cls(mpCBtnLock[i]);

	mpCLockGroup[i]->Disable();
	pBoxSizerLockItem[i]->Add(mpCLockGroup[i], 0, wxEXPAND);
	pBoxSizerLockItem[i]->Add(pCVCenter, 0, wxEXPAND);
    }
       
    pBoxSizerLock->Add(pBoxSizerLockItem[0], 0, wxEXPAND);
    pBoxSizerLock->Add(pBoxSizerLockItem[1], 0, wxEXPAND);
    pBoxSizerLock->Add(pBoxSizerLockItem[2], 0, wxEXPAND);
    pBoxSizerLock->Add(pBoxSizerLockItem[3], 0, wxEXPAND);
    pBoxSizerLock->AddStretchSpacer(1);

    pBoxSizerH->AddSpacer(1);
    pBoxSizerH->Add(pBoxSizerLock, 0, wxEXPAND);
    pBoxSizerH->Add(mpCTicsY, 0, wxEXPAND);
    pBoxSizerH->Add(mpCGraphPanel, 1, wxEXPAND);
    pBoxSizerH->AddSpacer(2);
    

    wxSize sizeBtn2(40, 28);
    mpCBtnZoomPlus = new wxBitmapButton
	(this, btnGraphZoomPlus, ARTS("wxART_ZOOM_PLUS", wxSize(24, 24)),
	 wxDP, sizeBtn2);
    mpCBtnZoomMinus = new wxBitmapButton
	(this, btnGraphZoomMinus, ARTS("wxART_ZOOM_MINUS", wxSize(24, 24)),
	 wxDP, sizeBtn2);
    mpCBtnZoomAuto = new wxBitmapButton
	(this, btnGraphZoomAuto, ARTS("wxART_ZOOM_AUTO", wxSize(24, 24)),
	 wxDP, sizeBtn2);

    wxSize CSize1(40, 28);
    wxSize CSize2(-1, 28);    
    mpCBtnRescale = new wxButton (this, btnGraphRescale, _("Rescale"), wxDP, CSize2);
    mpCBtnAddKey = new wxButton (this, btnGraphAddKey, _("Insert"), wxDP, CSize2);
    mpCBtnRemKey = new wxButton (this, btnGraphRemKey, _("Delete"), wxDP, CSize2);
    mpCBtnLft = new wxButton(this, btnGraphLft, wxT("<"), wxDP, CSize1);
    mpCBtnRgt = new wxButton(this, btnGraphRgt, wxT(">"), wxDP, CSize1);

    VCenter_cls *pCDefaultKeyIcon = SetDefaultKey(this);

    pBoxSizerKeyCtrl->Add(mpCBtnZoomPlus, 0, wxEXPAND);
    pBoxSizerKeyCtrl->Add(mpCBtnZoomMinus, 0, wxEXPAND);
    pBoxSizerKeyCtrl->Add(mpCBtnZoomAuto, 0, wxEXPAND);    
    pBoxSizerKeyCtrl->Add(mpCBtnRescale, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddStretchSpacer(1);
    pBoxSizerKeyCtrl->Add(mpCBtnLft, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddSpacer(6);
    pBoxSizerKeyCtrl->Add(mpCBtnAddKey, 0, wxEXPAND);
    pBoxSizerKeyCtrl->Add(mpCBtnRemKey, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddSpacer(6);
    pBoxSizerKeyCtrl->Add(mpCBtnRgt, 0, wxEXPAND);
    pBoxSizerKeyCtrl->AddSpacer(12);
    pBoxSizerKeyCtrl->Add(pCDefaultKeyIcon, 0, wxEXPAND);
    pBoxSizerKeyCtrl->Add(mpCDefaultKey, 0, wxEXPAND);
    pBoxSizerKeyCtrl->Add(10, -1);
    pBoxSizerKeyCtrl->AddStretchSpacer(1);
    
    pBoxSizerSrc->Add(mpCPanelBar, 0, wxEXPAND);

    pBoxSizerMainGraph->Add(pBoxSizerInfo, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerAA, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerTicX, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerNavi, 0, wxEXPAND);
    pBoxSizerMainGraph->Add(pBoxSizerH, 1, wxEXPAND);
    pBoxSizerNumPanel->Add(mpCNumPanel, 1, wxEXPAND);

    pBoxSizerSplit->Add(pBoxSizerMainGraph, 1, wxEXPAND);
    pBoxSizerSplit->Add(pBoxSizerNumPanel, 0, wxEXPAND);
    
    pBoxSizerSrc->Add(pBoxSizerSplit, 1, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerKeyCtrl, 0, wxEXPAND);

    UpdateDefaultKeyIcon();

    Freeze();
        
    SetSizerAndFit(pBoxSizerSrc);

    Thaw();
}

void CLS::
SetToDefaultParam(){
    mCConfList.SetTag("GraphEditor");
    mCConfList.Add(&mCbZoomAuto, "ZoomAuto=", "0");

    int AAMax = GetPCTimeCtrl()->mCFrameMax.GetBodyINT();
    wxString strAAMax; strAAMax << AAMax;
    mCConfList.Add(&mCAAMin, "AAMin=", "0");
    mCConfList.Add(&mCAAMax, "AAMax=", strAAMax);
}

void CLS::
Rescale(bool bCallUpdate){
    ParamGroup_cls* pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;
    mpCGraphPanel->UpdateYValueRange(GetPCParamGroup());
    double minY = mpCGraphPanel->mValueMinY;
    double maxY = mpCGraphPanel->mValueMaxY;
    float center = (minY + maxY) / 2;
    if (minY == maxY){
	minY = center - 0.5;
	maxY = center + 0.5;
    }
    else {
	minY = center + (minY - center) * 1.05;
	maxY = center + (maxY - center) * 1.05;
    }
    mpCTicsY->SetRegion(minY, maxY);
    if (bCallUpdate == TRUE) UpdateGraphs();
}

void CLS::
ProcAfterMath(const wxString& strCommand, wxString strItem, bool bForce){
    if (strCommand.IsSameAs("DeleteObject")){
	wxString strTgt = mCstrLinkedParamGroup.GetBodySTR();
	wxString strController = PopStrItemFromStrLine(strTgt, "::");
	if (strController.IsSameAs("SceneObjCtrl")){
	    PopStrItemFromStrLine(strTgt, "::");
	    wxString strObj = PopStrItemFromStrLine(strTgt, "::");
	    if (strObj.IsSameAs(strItem)){
		mCstrLinkedParamGroup.SetBodyFromData("None");
		mpCLinkedParamGroup->SetLabel(mCstrLinkedParamGroup.GetBodySTR());
		
	    }
	}
    }
    if (strCommand.IsSameAs("RenameObject")){
	wxString strNameOld = PopStrItemFromStrLine(strItem, "::");
	wxString strNameNew = strItem;

	wxString strTgt = mCstrLinkedParamGroup.GetBodySTR();
	wxString strController = PopStrItemFromStrLine(strTgt, "::");
	if (strController.IsSameAs("SceneObjCtrl")){
	    PopStrItemFromStrLine(strTgt, "::");
	    wxString strObj = PopStrItemFromStrLine(strTgt, "::");
	    if (strObj.IsSameAs(strNameOld)){
		mCstrLinkedParamGroup.mstrBody.Replace(strNameOld, strNameNew);
		mpCLinkedParamGroup->SetLabel(mCstrLinkedParamGroup.GetBodySTR());
	    }
	}
    }
    if (strCommand.IsSameAs("SelectParamGroup")){
	if (bForce == FALSE && mCbPinned.GetBodyBOOL() == TRUE) return;
	mCstrLinkedParamGroup.SetBodyFromData(strItem);
	mpCLinkedParamGroup->SetLabel(strItem);
	UpdatePanels("ALL");
    }
}

void CLS::
HideAndShowNumPanel(){
    if (mCbNumPanelShown.GetBodyBOOL() == TRUE &&
	mpCNumPanel->IsShown() == FALSE){
	//mpCPanelBar->Freeze();
	mpCBtnNumPanelHide->Hide();
	mpCNumPanel->Show();
	Layout();
	UpdateGraphs();
	//mpCPanelBar->Thaw();
    }
    if (mCbNumPanelShown.GetBodyBOOL() == FALSE &&
	mpCNumPanel->IsShown() == TRUE){
	//mpCPanelBar->Freeze();
	mpCBtnNumPanelHide->Show();
	mpCNumPanel->Hide();
	Layout();
	UpdateGraphs();
	//mpCPanelBar->Thaw();
    }
}

void CLS::
OnHideAndShowNumPanel(wxCommandEvent& event){
    if (mpCNumPanel->IsShown() == TRUE)
	mCbNumPanelShown.SetBodyFromData((long long)0);
    else
	mCbNumPanelShown.SetBodyFromData((long long)1);
    HideAndShowNumPanel();
}

void CLS::
OnMenuView(wxCommandEvent& event){
    Shortcut_cls *pCSC = GetPCShortcut();
    wxMenu CMenu;
    pCSC->SetMenuItem(&CMenu, menuZoomPlus);
    pCSC->SetMenuItem(&CMenu, menuZoomMinus);
    //CMenu.Append(menuZoomPlus, GetPCShortcut()->GetMenuString(menuZoomPlus));
    //CMenu.Append(menuZoomMinus, GetPCShortcut()->GetMenuString(menuZoomMinus));

    CMenu.AppendSeparator();
    pCSC->SetMenuCheckItem(&CMenu, menuShowNumPanel, mCbNumPanelShown.GetBodyBOOL());
    CMenu.AppendSeparator();
    pCSC->SetMenuCheckItem(&CMenu, menuShowPanelManage, mpCPanelBar->mbShowItem);

    //bool bPanel = mCbNumPanelShown.GetBodyBOOL();
    //CMenu.AppendCheckItem(menuShowNumPanel, GetPCShortcut()->GetMenuString(menuShowNumPanel))->Check(bPanel);
    //CMenu.AppendSeparator();
    //bool bBarItem = mpCPanelBar->mbShowItem;
    //CMenu.AppendCheckItem(menuShowPanelManage, GetPCShortcut()->GetMenuString(menuShowPanelManage))->Check(bBarItem);

    mpCTextMenuView->PopMenu(this, &CMenu);
}

void CLS::
OnMenuChannel(wxCommandEvent& event){
    wxMenu CMenu;

    wxMenu *pCMenuExtrapolateAll = new wxMenu;
    pCMenuExtrapolateAll->Append(menuExtrapolateAllOff, "Off");
    pCMenuExtrapolateAll->Append(menuExtrapolateAllOn, "On");
    CMenu.Append(wxID_ANY, _("Extrapolate All"), pCMenuExtrapolateAll);

    wxMenu *pCMenuExtrapolateSelected = new wxMenu;
    pCMenuExtrapolateSelected->Append(menuExtrapolateOff, "Off");
    pCMenuExtrapolateSelected->Append(menuExtrapolateOn, "On");
    CMenu.Append(wxID_ANY, _("Extrapolate Selected"), pCMenuExtrapolateSelected);

    mpCTextMenuChannel->PopMenu(this, &CMenu);
}

void CLS::
OnMenuKey(wxCommandEvent& event){
    wxMenu CMenu;
    CMenu.Append(menuDelete, GetPCShortcut()->GetMenuString(menuDelete));
    
    mpCTextMenuKey->PopMenu(this, &CMenu);
}

void CLS::
OnTicsMode(wxCommandEvent& event){
    long long mode = mCTicsMode.GetBodyINT();
    mode++;
    if (mode > 2) mode = 0;
    mpCTicsMode->SetLabel(Tics_cls::msAstrTicsMode[mode]);
    mCTicsMode.SetBodyFromData(mode);
    mpCTicsX->Refresh();
    mpCTicsY->Refresh();
}

void CLS::
OnRescale(wxCommandEvent& event){ Rescale(); }

void CLS::
OnMenuExtrapolate(wxCommandEvent& event){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;

    bool bOn = (event.GetId() - menuExtrapolateOff) % 2;
    bool bSelectedOnly = ! ((event.GetId() - menuExtrapolateOff) / 2);
    uint numParam = pCParamGroup->mVCParam.size();
    GetPCUndo()->OpenBlock("SetExtrapolate");
    for (uint i = 0; i < numParam; i++){
	GetPCMain()->SetParameterConfCommand(mCstrLinkedParamGroup.GetBodySTR() + "::" + pCParamGroup->mVCParam[i].GetStrName(), "Extrapolate=", bOn, bSelectedOnly);
    }
    GetPCUndo()->CloseBlock("SetExtrapolate");
}

void CLS::
OnMenuZoom(wxCommandEvent& event){
    bool bExpand = (event.GetId() == menuZoomPlus);
    ExpandAA(bExpand);
    UpdatePanels("ALL");
}

void CLS::
OnDropDown(wxCommandEvent& event){
    mCSelector.PopupParamGroupSelector(this); }

void CLS::
OnPopMenu(wxCommandEvent& event){
    wxString strPopMenu = mCSelector.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == TRUE) return;
    
    wxString strCommand = "SelectParamGroup";
    wxString strItem;
    strItem << GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strPopMenu;
    ProcAfterMath(strCommand, strItem, TRUE);
}

void CLS::
OnZoom(wxCommandEvent& event){
    float rate = -2;
    int mode = event.GetId() - btnGraphZoomPlus;
    if (mode == 1) rate = 2;
    mpCTicsY->ProcessZoom(rate);
    mpCTicsY->Refresh(FALSE);
    UpdateGraphs();
}

void GraphEditor_cls::
OnZoomAuto(wxCommandEvent& event){
    mCbZoomAuto.SetBodyFromData((long long)!mCbZoomAuto.GetBodyBOOL());
    if (mCbZoomAuto.GetBodyBOOL() == TRUE) Rescale();
    UpdateButtonsXYZ();
}

void CLS::
OnCmbDefaultKey(wxComEvt& event){
    ProcDefaultKey(event);
}

void CLS::
OnButtonDefaultKey(wxComEvt& event){
    ProcButtonDefaultKey();
}

/*
void CLS::
SyncronizeParamToPanel(){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup();
    if (pCParamGroup == NULL) return;
    double Xnavi = GetNaviCurrent();
    int paramNum = pCParamGroup->mVCParam.size();
    for (int i = 0; i < paramNum; i++){
	double newValue = mCCurrent[i].GetBodyDOUBLE();
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	if (CParam.GetValue(Xnavi) != newValue){
	    wxString strTag = mCstrLinkedParamGroup.GetBodySTR() << "::" << CParam.GetStrName();
	    CParam.SetValueAtCommand(Xnavi, newValue, strTag);
	}
    }
}
*/

void CLS::
SyncronizeKeyToPanel(){
    Parameter_cls *pCParam = GetPCParameterActive();
    if (pCParam != NULL){
	int activeKeyIndex = pCParam->GetActiveKeyIndex();
	if (activeKeyIndex != -1){
	    int newType = mCActiveKeyType.GetBodyINT();
	    wxString strTag = mCstrLinkedParamGroup.GetBodySTR() << "::" << pCParam->GetStrName();
	    if (pCParam->mVCKeys[activeKeyIndex].mKeyType != newType)
		pCParam->SetKeyTypeCommand(activeKeyIndex, newType, strTag);
	}
    }
}

void CLS::
SyncronizeTicsToGraph(){
    mpCTicsY->mRegionMin = mpCGraphPanel->mViewMinY;
    mpCTicsY->mRegionMax = mpCGraphPanel->mViewMaxY;
}

void CLS::
UpdateButtonsXYZ(){
    if (mCbZoomAuto.GetBodyBOOL() == TRUE)
	mpCBtnZoomAuto->SetBackgroundColour(wxColour(128, 128, 255));
    else
	mpCBtnZoomAuto->SetBackgroundColour(wxNullColour);
}

void CLS::
UpdatePanels(const wxString& strPanelTag, bool bUpdateBar){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs(wxT("ALL"))){

	if (GetPCQueue()->mbQueueAdjustAADuetoFrameMax == TRUE){
	    long long frameMax = GetPCTimeCtrl()->mCFrameMax.GetBodyINT();
	    if (mCAAMax.GetBodyINT() == GetPCQueue()->mQueueFrameMaxPrev)
		mCAAMax.SetBodyFromData(frameMax);
	    else if (mCAAMax.GetBodyINT() > frameMax)
		mCAAMax.SetBodyFromData(frameMax);
	}
	
	if (GetPCQueue()->mbQueueSyncronizeGraphNavi == TRUE){
	    mCNaviCurrent.SetBodyFromData(GetPCTimeCtrl()->GetCurrentNT());
	}

	if (mCbZoomAuto.GetBodyBOOL() == TRUE) Rescale(FALSE);
	UpdateNumPanels();
	UpdateGraphs();
	UpdateLock();
	UpdateDefaultKeyIcon();
	
	if (bUpdateBar == TRUE) mpCPanelBar->Refresh();
	ParamGroup_cls *pCParamGroup = GetPCParamGroup();
	if (pCParamGroup == NULL)
	    mpCBtnLinked->Disable();
	else{
	    mpCBtnLinked->Enable();
	    mpCBtnLinked->SetStatus(pCParamGroup->mCbLinked.GetBodyINT());
	}
	mpCBtnLinked->Refresh();
    }
}

void CLS::
Serialize(wxString& strOut){
    wxString strViewY; 
    strViewY << mpCGraphPanel->mViewMinY << " " << mpCGraphPanel->mViewMaxY; 
    mCViewY.SetBodyFromData(strViewY);
    ConfListContainer_cls::Serialize(strOut);
}

void CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == FALSE)
	return;
    mpCLinkedParamGroup->SetLabel(mCstrLinkedParamGroup.GetBodySTR());
    mpCNumPanel->SyncStateToData();
    mpCTicsY->mRegionMin = mCViewY.GetBodyFLOATARRAY()[0];
    mpCTicsY->mRegionMax = mCViewY.GetBodyFLOATARRAY()[1];
    HideAndShowNumPanel();
}

