// -*- c++ -*-
// "clsGraphEditoreTics.h"

#include "clsGraphEditorTics.h"
#include "clsGraphEditor.h"
#include "clsColorCurveEditor.h"

#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../PREFERENCE/clsMousePreference.h"
#include "../INPUTMETHOD/clsMouseCtrl.h"

#include "../TIMELINE/clsTimeCtrl.h"

#define _VERTICAL_TIC_WIDTH 80

#define CLS GraphEditorTics_cls

BEGIN_EVENT_TABLE(CLS, TimelineBase_cls)
EVT_PAINT(CLS::OnPaint)
EVT_MOUSE_EVENTS(CLS::OnMouse)
EVT_LEFT_DOWN(CLS::OnMouseDown)
EVT_LEFT_UP(CLS::OnMouseUp)
EVT_RIGHT_DOWN(CLS::OnMouseDown)
EVT_RIGHT_UP(CLS::OnMouseUp)
EVT_MIDDLE_DOWN(CLS::OnMouseDown)
EVT_MIDDLE_UP(CLS::OnMouseUp)
END_EVENT_TABLE()

CLS::
GraphEditorTics_cls(EditorBase_cls *pCParent, bool bVertical) :
TimelineBase_cls (pCParent, _TL_TICS, (bVertical == TRUE) ? wxSize(_VERTICAL_TIC_WIDTH, -1) : wxSize(-1, 16)), Tics_cls(bVertical)
{
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    mpCParent = pCParent;
    SetBorderX(6, 6);
    mpCMouseCtrl  = new MouseCtrl_cls();
    mRegionMin = 0;
    mRegionMax = 1;
    mbLocked = FALSE;
}

CLS::
~CLS(){
    delete mpCMouseCtrl;
}

void CLS::
ProcessZoom(float rate){
    float regionCenter = (mRegionMax + mRegionMin) / 2;
    float regionHeight = mRegionMax - mRegionMin;
    float regionMax = regionCenter + (regionHeight/2) * (1.0f + rate*0.2);
    float regionMin = regionCenter - (regionHeight/2) * (1.0f + rate*0.2);
    mRegionMin = regionMin;
    mRegionMax = regionMax;
}

void CLS::
UpdateDisplaySetting(){

    UpdateState();

    SetTLRegion(mAAMinNormalized, mAAMaxNormalized);
    
    switch(mpCParent->mCTicsMode.GetBodyINT()){
    case 0:
	SetParameter(GetClientSize(), 
		     mpCAALink->mCAAMin.GetBodyINT(),
		     mpCAALink->mCAAMax.GetBodyINT());
	break;
    case 1: {
	int factor = GetPCTimeCtrl()->mCSubFrameMax.GetBodyINT();
	SetParameter(GetClientSize(), 
		     mpCAALink->mCAAMin.GetBodyINT() * factor,
		     mpCAALink->mCAAMax.GetBodyINT() * factor);
	break; }
    case 2: 
	SetParameter(GetClientSize(), mAAMinNormalized, mAAMaxNormalized);
	break; 
    case 3:{
	ColorCurveEditor_cls* pCCCE = (ColorCurveEditor_cls*)mpCParent;
	double min = pCCCE->mCXRange[0].GetBodyDOUBLE();
	double max = pCCCE->mCXRange[1].GetBodyDOUBLE();
	double width = max - min;
	
	SetParameter(GetClientSize(), 
		     min + width * mpCAALink->mCAAMin.GetBodyINT() / _COLOR_CURVE_RESOLUTION,
		     min + width * mpCAALink->mCAAMax.GetBodyINT() / _COLOR_CURVE_RESOLUTION);
	break; }
    }
}

void CLS::
OnPaint(wxPaintEvent& event){
    wxAutoBufferedPaintDC dc(this);

    float min = 0, max = 1;
    if (mbVertical == TRUE){
	/*
	min = mpCParentPanelLink->mpCMainGraph->mViewMinY;
	max = mpCParentPanelLink->mpCMainGraph->mViewMaxY;
	*/
	SetParameter(GetClientSize(), mRegionMin, mRegionMax);
    }
    else
	UpdateDisplaySetting();
	//SetParameter(GetClientSize(), min, max/*, 1*/);

    DrawTics(dc);

    if (mbVertical == FALSE && mbForColorCurve == FALSE) DrawCurrentTime(dc, 0);
}

void CLS::
OnMouse(wxMouseEvent& event){
    if (mbLocked == TRUE) return event.Skip();

    mpCMouseCtrl->SetMouseInformation(event);
    MouseNormalized_cls CMP(mpCMouseCtrl, GetSize());

    bool bRedraw = FALSE;
    bool bRedrawButton = FALSE;

    if (mbVertical == TRUE && mpCMouseCtrl->IsGyration()){
	float dY = CMP.GetDY();
	float regionCenter = (mRegionMax + mRegionMin) / 2;
	float regionHeight = mRegionMax - mRegionMin;
	float regionMax = mRegionMax - regionHeight * dY * 0.5;
	float regionMin = mRegionMin - regionHeight * dY * 0.5;
	mRegionMin = regionMin;
	mRegionMax = regionMax;
	bRedraw = TRUE;
	bRedrawButton = TRUE;
    }

    if (mbVertical == TRUE && mpCMouseCtrl->IsZoom()){
	float rate = CMP.GetDY() * 3;
	
	if (GetPCMousePreference()->mCbZoomInvert.GetBodyBOOL() == TRUE) 
	    rate *= -1;
	
	ProcessZoom(rate);
	bRedraw = TRUE;
	bRedrawButton = TRUE;
    }

    mpCMouseCtrl->SetPreviousPosition();

    if (bRedraw == TRUE){
	Refresh(FALSE);
	mpCParent->UpdateGraphs();
    }
    /*
    if (bRedrawButton == TRUE) 
	SendMenuEvent(GetParent(), menuRefreshFromTics);
    */
    event.Skip();
}

void CLS::
OnMouseDown(wxMouseEvent& event){
    if (HasCapture() == FALSE) CaptureMouse();
}

void CLS::
OnMouseUp(wxMouseEvent& event){
    if (HasCapture() == TRUE) ReleaseMouse();
}
