// (setq c-basic-offset 4)
// "clsGraphNumericalPanel.h"

#include "clsGraphNumericalPanel.h"

#include "clsEditorBase.h"
#include "clsColorCurveEditor.h"

#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/MULTIINPUT/clsHideAndShowButton.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"

#include "../COMMON/CommonDefinitions.h"

#include "../../wxWidgetEnum.h"

#define CLS GraphNumericalPanel_cls

#define EVT_BUTTON_RANGE(id1, id2, func) wx__DECLARE_EVT2(wxEVT_COMMAND_BUTTON_CLICKED, id1, id2, wxCommandEventHandler(func))

BEGIN_EVENT_TABLE(CLS, wxPanel)
EVT_BUTTON_RANGE(btnColorTemplate0, btnColorTemplate11, CLS::OnTemplate)
END_EVENT_TABLE()

CLS::
GraphNumericalPanel_cls(EditorBase_cls *pCParent, wxSize &size) : 
    MNumPainBase_cls(pCParent, size)
{
    SetBackgroundColour(GetPCColorProvider()->GetColor("EditPanel"));
    mpCParent = pCParent;

    wxBoxSizer *pCBoxSizerSrc = new wxBoxSizer(wxVERTICAL);

    mIndexRange = -1;
    mIndexRGB = -1;

    SetMNumValue();
    SetMNumKey();
    
    SetMNum(mpCMNumValue, mAstrClosedItem, "parameter");
    SetMNum(mpCMNumKey, mAstrClosedItem, "active key");

    if (pCParent->mstrClassName.IsSameAs("ColorCurveEditor_cls")){
	SetMNumTemplate();
	SetMNum(mpCMNumTemplate, mAstrClosedItem, "template");
    }

    InitLayoutNumPanels(pCBoxSizerSrc);
    
    UpdateLabels(NULL);
    SetSizerAndFit(pCBoxSizerSrc);
    SetMinSize(size);
}

void CLS::
SetMNumValue(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNumValue, this,
					"GraphNumericalPanel", TRUE);

    pC->mMinWidth = 180;
    pC->mTextWidth = 80;

    if (mpCParent->mstrClassName.IsSameAs("ColorCurveEditor_cls")){
	ColorCurveEditor_cls *pCParentCCE = (ColorCurveEditor_cls*)mpCParent;
	mIndexRange = pC->CombineSizer(2, "Mix/Max");
	pC->mDragScale = 0.1;
	pC->CombineConfItem(pCParentCCE->mCXRange[0]);
	pC->mDragScale = 0.1;
	pC->CombineConfItem(pCParentCCE->mCXRange[1]);

	pC->mTextWidth = 30;
	pC->mDragScale = 0.1;
	mIndexRGB = pC->CombineSizer(3, "RGB");
	mIndexCurrent = pC->CombineConfItem(mpCParent->mCCurrent[0]);
	pC->CombineConfItem(mpCParent->mCCurrent[1]);
	pC->CombineConfItem(mpCParent->mCCurrent[2]);
	pC->CombineConfItem(mpCParent->mCCurrent[3], "alpha");
	    
    }
    else{
	pC->mDragScale = 0.1;
	mIndexCurrent = 
	    pC->CombineConfItem(mpCParent->mCCurrent[0], "0");
	pC->mDragScale = 0.1;
	pC->CombineConfItem(mpCParent->mCCurrent[1], "1");
	pC->mDragScale = 0.1;
	pC->CombineConfItem(mpCParent->mCCurrent[2], "2");
	pC->mDragScale = 0.1;
	pC->CombineConfItem(mpCParent->mCCurrent[3], "3");
    }

    pC->FinishSetting();

    pC->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_TEXT_ENTER);
}

void CLS::
SetMNumKey(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNumKey, this,
					"GraphNumericalPanel", TRUE);
    pC->mMinWidth = 180;
    pC->mTextWidth = 80;
 
    mIndexKey = pC->CombineSizer(2, _("NT/Value"));
    pC->CombineConfItem(mpCParent->mCActiveKeyXnorm);
    pC->CombineConfItem(mpCParent->mCActiveKeyValue);

    pC->SetChoiceList(3, GetPCStringProvider()->GetAstr("KeyFrameType"));
    mIndexKeyFrameType = pC->CombineConfItem(mpCParent->mCActiveKeyType, _("Type"));
    mIndexEaseFactor = pC->CombineConfItem(mpCParent->mCActiveKeyEaseFactor, _("Ease Factor"));
    
    pC->FinishSetting();

    pC->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER);
    pC->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_COMBOBOX_SELECTED);
}

void CLS::
SetMNumTemplate(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNumTemplate, this,
					"GraphNumericalPanel", TRUE);

    ColorCurveEditor_cls *pCParentCCE = (ColorCurveEditor_cls*)mpCParent;

    pC->mMinWidth = 180;

    pC->CombineConfItem(pCParentCCE->mCbTemplateAlpha, _("Use Alpha"));
    
    pC->CombineSizer(3);
    mIndexTemplate[0] = pC->CombineBitmapPanel(btnColorTemplate0, "template0");
    mIndexTemplate[1] = pC->CombineBitmapPanel(btnColorTemplate1, "template1");
    mIndexTemplate[2] = pC->CombineBitmapPanel(btnColorTemplate2, "template2");

    pC->CombineSizer(3);
    mIndexTemplate[3] = pC->CombineBitmapPanel(btnColorTemplate3, "template3");
    mIndexTemplate[4] = pC->CombineBitmapPanel(btnColorTemplate4, "template4");
    mIndexTemplate[5] = pC->CombineBitmapPanel(btnColorTemplate5, "template5");

    pC->CombineSizer(3);
    mIndexTemplate[6] = pC->CombineBitmapPanel(btnColorTemplate6, "template6");
    mIndexTemplate[7] = pC->CombineBitmapPanel(btnColorTemplate7, "template7");
    mIndexTemplate[8] = pC->CombineBitmapPanel(btnColorTemplate8, "template8");

    pC->CombineSizer(3);
    mIndexTemplate[9] = pC->CombineBitmapPanel(btnColorTemplate9, "template9");
    mIndexTemplate[10] = pC->CombineBitmapPanel(btnColorTemplate10, "template10");
    mIndexTemplate[11] = pC->CombineBitmapPanel(btnColorTemplate11, "template11");
    
    pC->FinishSetting();
    pC->SetHook(this, &HookItem_cls::HookFunc03, wxEVT_CHECKBOX);

    for (int i = 0;i < _COLORCURVE_TEMPLATE_MAX; i++){
	BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[mIndexTemplate[i]].mpCWindowLink;
	pCBitmap->mbButton = TRUE;
	pCBitmap->SetColorCurveLink(GetPCColorCurveProvider()->GetPCColorCurve(i));
	pCBitmap->mBorderX = 1;
	pCBitmap->mBorderY = 1;
    }
}

void CLS::
UpdateLabels(ParamGroup_cls *pCParamGroup){
    if (pCParamGroup == NULL){
	if (mIndexRange != -1) mpCMNumValue->mVCItem[mIndexRange].Hide();
	if (mIndexRGB != -1) mpCMNumValue->mVCItem[mIndexRGB].Hide();
	for (int i = 0; i < 4; i++)
	    mpCMNumValue->mVCItem[mIndexCurrent+i].Hide();

	mpCMNumKey->mVCItem[mIndexKey].Hide();
	mpCMNumKey->mVCItem[mIndexKeyFrameType].Hide();
	mpCMNumKey->mVCItem[mIndexEaseFactor].Hide();
    }
    else{
	if (mIndexRange != -1) mpCMNumValue->mVCItem[mIndexRange].Show();
	if (mIndexRGB != -1) mpCMNumValue->mVCItem[mIndexRGB].Show();

	uint paramNum = pCParamGroup->mVCParam.size();
	for (int i = 0; i < paramNum; i++){
	    mpCMNumValue->mVCItem[mIndexCurrent+i].mstrLabel = pCParamGroup->mVCParam[i].GetStrName();
	    mpCMNumValue->mVCItem[mIndexCurrent+i].UpdateLabel();
	    mpCMNumValue->mVCItem[mIndexCurrent+i].Show();
	}
	for (int i = paramNum; i < 4; i++)
	    mpCMNumValue->mVCItem[mIndexCurrent+i].Hide();

	mpCMNumKey->mVCItem[mIndexKey].Show();
	mpCMNumKey->mVCItem[mIndexKeyFrameType].Show();
	mpCMNumKey->mVCItem[mIndexEaseFactor].Show();
    }
    LayoutNumPanels();
}

void CLS::
UpdateTemplateButtons(){
    bool bAlpha = ((ColorCurveEditor_cls*)mpCParent)->mCbTemplateAlpha.GetBodyBOOL();
    for (int i = 0;i < _COLORCURVE_TEMPLATE_MAX; i++){
	BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)mpCMNumTemplate->mVCItem[mIndexTemplate[i]].mpCWindowLink;
	pCBitmap->mColorCurveAlpha = (bAlpha == TRUE) ? _COLOR_CURVE_RGBA : _COLOR_CURVE_RGB;
	pCBitmap->Refresh();
    }
}

void CLS::
OnNumChange(wxCommandEvent& event){
    mpCParent->SyncronizeParamToPanel();
}

void CLS::
OnKeyChange(wxCommandEvent& event){
    mpCParent->SyncronizeKeyToPanel();
}

void CLS::
OnTemplateChange(wxCommandEvent& event){
    bool bAlpha = event.IsChecked();
    ((ColorCurveEditor_cls*)mpCParent)->mCbTemplateAlpha.SetBodyFromDataBOOL(bAlpha);
    
    UpdateTemplateButtons();
    
    //mpCParent->SyncronizeKeyToPanel
}

void CLS::
OnTemplate(wxCommandEvent& event){
    int index = event.GetId() - btnColorTemplate0;
    ((ColorCurveEditor_cls*)mpCParent)->SetColorCurveTemplate(index);
}
