// (setq c-basic-offset 4)
// "clsGraphPanel.h"


#include "clsGraphPanel.h"
#include "clsGraphEditor.h"
#include "clsEditorBase.h"

#include "../DIALOG/clsParamDialog.h"
#include "../DIALOG/clsParamGroupDialog.h"
#include "../INPUTMETHOD/clsMouseCtrl.h"
#include "../PREFERENCE/clsMousePreference.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsPenBrushProvider.h"
#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS GraphPanel_cls

BEGIN_EVENT_TABLE(CLS, wxPanel)
EVT_PAINT(CLS::OnPaint)
EVT_LEFT_DOWN(CLS::OnLeftDown)
EVT_RIGHT_DOWN(CLS::OnRightDown)
EVT_LEFT_UP(CLS::OnLeftUp)
EVT_MOTION(CLS::OnMotion)
EVT_LEFT_DCLICK(CLS::OnDoubleClick)
EVT_MOUSE_EVENTS(CLS::OnMotion)
END_EVENT_TABLE()

CLS::
GraphPanel_cls(EditorBase_cls *pCParent,
	       bool bColor, bool bLimitYValue) :
wxPanel(pCParent, wxID_ANY, wxDPS/*, wxFULL_REPAINT_ON_RESIZE*/)
{
    mpCParent = pCParent;
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    ColorProvider_cls::SetBGColor(this, "EditPanel");
    
    mbForColor = bColor;
    mColorType = _GRAPH_RGBA;
    mbLimitYValue = bLimitYValue;
    mbKeySelecting = FALSE;

    mpCMouseCtrl  = new MouseCtrl_cls();

    mXMarginMin = 2; mXMarginMax = 2;
    mValueLimitMinX = 0;
    mValueLimitMaxX = 1; 
    mViewMinX = 0; mViewMaxX = 1;
    mViewMinY = -0.5; mViewMaxY = +0.5;
}

CLS::
~GraphPanel_cls(){
    delete mpCMouseCtrl;
}

void GraphPanel_cls::
SetXViewRange(double minX, double maxX){
    mViewMinX = minX; mViewMaxX = maxX;
}

void GraphPanel_cls::
SetYViewRange(double minY, double maxY){
    mViewMinY = minY; mViewMaxY = maxY;
}

void GraphPanel_cls::
SetYValueRange(double minY, double maxY){
    mValueLimitMinY = minY;
    mValueLimitMaxY = maxY;
}

float CLS::
GetViewWidth(){
    return mViewMaxX - mViewMinX;
}

float CLS::
GetViewHeight(){
    return mViewMaxY - mViewMinY;
}

int CLS::
GetCoordX(double valueX){
    double doubleX = (valueX - mViewMinX) / GetViewWidth();
    return doubleX * (mPanelWidth - 3 - mXMarginMin - mXMarginMax) + 1 + mXMarginMin;
}

int CLS::
GetCoordY(double valueY){
    double doubleY = (valueY - mViewMinY) / GetViewHeight();
    return (mPanelHeight - 2) - doubleY * (mPanelHeight - 2) + 1;
}

void CLS::
GetMouseDPos(int &dx, int &dy){
    dx = mMouseXCurrent - mMouseXAtDown;
    dy = mMouseYCurrent - mMouseYAtDown;

    bool bSnapX = TRUE, bSnapY = TRUE;
    if (bSnapY == TRUE){
	if (dx < -2 || dx > 2) 
	    if (dy > -5 && dy < 5) dy = 0;
    }
    if (bSnapX == TRUE)
	if (dy < -2 || dy > 2)
	    if (dx > -5 && dx < 5) dx = 0;
}

void CLS::
GetMouseDNPos(double &dNX, double &dNY){
    int dX, dY;
    GetMouseDPos(dX, dY);

    double NXMouseCurrent = GetXNormalized(mMouseXAtDown + dX);
    double NXMouseAtDown = GetXNormalized(mMouseXAtDown);
    double NYMouseCurrent = GetYNormalized(mMouseYAtDown + dY);
    double NYMouseAtDown = GetYNormalized(mMouseYAtDown);
    dNX = NXMouseCurrent - NXMouseAtDown;
    dNY = NYMouseCurrent - NYMouseAtDown;
}

double CLS::
GetXNormalized(int coordX){
    double XinPanel = ((double)coordX - 0.5 - mXMarginMin) / (mPanelWidth - 2 - mXMarginMin - mXMarginMax);
    double XValue = XinPanel * GetViewWidth() + mViewMinX;
    if (XValue < mValueLimitMinX) XValue = mValueLimitMinX;
    if (XValue > mValueLimitMaxX) XValue = mValueLimitMaxX;
    return XValue;
}

double CLS::
GetYNormalized(int coordY, bool bLimit){
    double YinPanel =1.0 - ((double)coordY - 1) / (mPanelHeight - 2);
    double YValue = YinPanel * GetViewHeight() + mViewMinY;
    if (bLimit == TRUE){
	if (YValue < mValueLimitMinY) YValue = mValueLimitMinY;
	if (YValue > mValueLimitMaxY) YValue = mValueLimitMaxY;
    }
    return YValue;
}

void GraphPanel_cls::
SetPenColor(wxAutoBufferedPaintDC &dc, ParamGroup_cls *pCParamGroup, int i){
    wxPen pen;
    if (pCParamGroup->mVCParam[i].mCbLocked.GetBodyBOOL() == FALSE)
	pen = wxPen(msColorRGBA[i]);
    else
	pen = wxPen(msColorLockedRGBA[i]);
    //if (i == mpCParamGroupLinked->mControlMode) pen.SetWidth(2);
    if (mColorType == _GRAPH_RGBA) dc.SetPen(pen);
}

void GraphPanel_cls::
SetBrushColor(wxAutoBufferedPaintDC &dc, int i){
    if (mColorType == _GRAPH_RGBA) dc.SetBrush(wxBrush(msColorRGBA[i]));
}

void GraphPanel_cls::
UpdateYValueRange(ParamGroup_cls *pCParamGroup){
    // `OYl̃WKvɂȂꍇ̂߃Abvf[gs
    ASSERT(pCParamGroup != NULL);
    
    int dim = pCParamGroup->mVCParam.size();
    int xstep = 1;
    if (mPanelWidth > 320) xstep = 2;
    if (mPanelWidth > 640) xstep = 4;

    bool bInit = TRUE;
    for (int x = 0; x < mPanelWidth; x += xstep){
	double XNormalized = GetXNormalized(x);

	std::vector < double > VValue = 
	    pCParamGroup->GetVInterpolatedValue(XNormalized);
	/*
	std::vector < double > VValue;
	if (mbShowVelocity == TRUE)
	    VValue = 
		(mpCParamGroupLinked->GetVelocityValue(valueX));
	else
	    VValue = 
		(mpCParamGroupLinked->CalcInterpolatedValue(valueX, 0));
	*/

	for (int i = 0; i < dim; i++){
	    if (bInit == TRUE){
		mValueMinY = VValue[i];
		mValueMaxY = VValue[i];
		bInit = FALSE;
	    }
	    else{
		if (VValue[i] > mValueMaxY)	mValueMaxY = VValue[i];
		if (VValue[i] < mValueMinY)	mValueMinY = VValue[i];
	    }
	}
    }
}

void CLS::
DrawCurrentTime(wxDC& dc){
    int x0 = GetCoordX(mpCParent->GetTimeNormalized()); 
    dc.SetPen(wxPen(wxColour(255, 0, 0)));
    dc.DrawRectangle(x0, 0, 1, mPanelHeight);
}

void CLS::
DrawNaviCurrent(wxDC& dc){
    int x0 = GetCoordX(mpCParent->GetNaviCurrent());
    dc.SetPen(wxPen(wxColour(255, 255, 255)));
    dc.DrawRectangle(x0, 0, 1, mPanelHeight);
}

void CLS::
DrawGraphOfParamGroup(wxAutoBufferedPaintDC &dc,
		      ParamGroup_cls *pCParamGroup){

    mAAKeyFrameRect.clear();
    mAAKeyFrameIndex.clear();

    int dim = pCParamGroup->mVCParam.size();
    std::vector < std::vector < wxPoint > > VVPoint(dim);

    int xstep = 1;
    if (mPanelWidth > 320) xstep = 1;
    if (mPanelWidth > 640) xstep = 2;

    pCParamGroup->PrecalcBegin();
    for (int x = 0; x < mPanelWidth; x += xstep){
	double XNormalized = GetXNormalized(x);
	std::vector < double > VValue = 
	    pCParamGroup->GetVInterpolatedValue(XNormalized);

	for (int i = 0; i < dim; i++){
	    wxPoint CP = wxPoint(GetCoordX(XNormalized), GetCoordY(VValue[i]));
	    VVPoint[i].push_back(CP);
	    
	}
    }
    pCParamGroup->PrecalcEnd();

    for (int i = 0; i < dim; i++){
	SetPenColor(dc, pCParamGroup, i);
	//for (int x = 0; x < mPanelWidth; x+=xstep)
	dc.DrawLines(mPanelWidth / xstep, VVPoint[i].data());
    }

    if (dim == 3 || dim == 4){
	for (int i = 0; i < 2; i++){
	    SetPenColor(dc, pCParamGroup, i);
	    for (int x = i*xstep; x < mPanelWidth - 6; x+=xstep*3)
		dc.DrawLines(2, VVPoint[i].data()+x/xstep);
	}
    }

    int dX, dY;
    GetMouseDPos(dX, dY);

    for (int i = 0; i < dim; i++){
	
	SetPenColor(dc, pCParamGroup, i);
	Parameter_cls* pCParam = &(pCParamGroup->mVCParam[i]);
	
	uint keyNum= pCParam->mVCKeys.size();
	if (pCParam->mCbTimeDepend.GetBodyBOOL() == FALSE) continue;

	bool bLocked = pCParam->mCbLocked.GetBodyBOOL();
	std::vector < wxRect > AKeyFrameRect;
	std::vector < int > AKeyFrameIndex;
	for (uint key = 0; key < keyNum; key++){
	    KeyFrame_cls &CKey = pCParam->mVCKeys[key];
	    int coordX = GetCoordX(CKey.mXNormalized);
	    int coordY;
	    coordY = GetCoordY(CKey.mValue);
	    
	    wxRect CRect(wxPoint(coordX-3, coordY-3), wxSize(7, 7));

	    CKey.mbLocked = bLocked;
	    CKey.DrawKey(dc, coordX, coordY);
	    /*
	    if (mbForColor == FALSE)
		CKey.DrawKey(dc, coordX, coordY);
	    else
		dc.DrawRectangle(CRect);
	    */

	    if (mbKeySelecting == TRUE && CKey.mbSelected == TRUE &&
		(dX != 0 || dY != 0)){
		CKey.DrawKeyGhost(dc, coordX + dX, coordY + dY);
		if (dY == 0){
		    dc.SetPen(GetPCPenBrushProvider()->GetPen(_PEN_DASH_SHORT));
		    std::vector < wxPoint > VPoint;
		    VPoint.push_back(wxPoint(coordX, coordY));
		    VPoint.push_back(wxPoint(coordX+dX, coordY));
		    dc.DrawLines(2, VPoint.data());
		}
		if (dX == 0){
		    dc.SetPen(GetPCPenBrushProvider()->GetPen(_PEN_DASH_SHORT));
		    std::vector < wxPoint > VPoint;
		    VPoint.push_back(wxPoint(coordX, coordY));
		    VPoint.push_back(wxPoint(coordX, coordY+dY));
		    dc.DrawLines(2, VPoint.data());
		}
		    
	    }

	    if (pCParam->mCbLocked.GetBodyBOOL() == FALSE){
		AKeyFrameRect.push_back(CRect);
		AKeyFrameIndex.push_back(key);
	    }
	}

	mAAKeyFrameRect.push_back(AKeyFrameRect);
	mAAKeyFrameIndex.push_back(AKeyFrameIndex);
    }

    
}

void CLS::
SendUpdate(){
    GetPCMain()->UpdatePanels("TIMELINE::GRAPH::COLOR_CURVE::VIEW3D::PROPERTY_MATERIAL");
}

bool CLS::
SelectItemByMouse(bool bAdditive){
    
    mbKeySelecting = FALSE;

    ParamGroup_cls *pCParamGroup = mpCParent->GetPCParamGroup();
    if (pCParamGroup == NULL) return FALSE;
	
    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    
    bool bSelectedKeySelected = FALSE;
    bool bKeySelected = FALSE;

    double resolution = _RES_DEFAULT;
    if (mbForColor == TRUE)
	resolution = 1.0 / _COLOR_CURVE_RESOLUTION;
    
    uint num = pCParamGroup->mVCParam.size();
    if (mAAKeyFrameRect.size() == 0) num = 0; // timedepend 0 
    for (uint param = 0; param < num; param++){
	Parameter_cls &CParam = pCParamGroup->mVCParam[param];
	wxString strTag = mpCParent->mCstrLinkedParamGroup.GetBodySTR();
	strTag << "::" << CParam.GetStrName();
	uint keyNum = mAAKeyFrameRect[param].size();
	for (uint keyIndex = 0; keyIndex < keyNum; keyIndex++){
	    if (mAAKeyFrameRect[param][keyIndex].Contains(CAtDown)){
		bool bBreak = FALSE;
		bool bSelectedAlready = CParam.mVCKeys[keyIndex].mbSelected;
		if (bSelectedAlready == TRUE && bAdditive == FALSE){
		    bKeySelected = TRUE;
		    bSelectedKeySelected = TRUE;
		}
		else{
		    bKeySelected |= GetPCSceneObjCtrl()->
			SelectKeyFrame(strTag, keyIndex, ! bSelectedAlready, bAdditive, resolution);
		    mpCParent->mCActiveParamIndex.SetBodyFromData((long long)param);
		    bBreak = TRUE;
		}

		GetPCSceneObjCtrl()->
		    ChangeKeyTypeByModKey(strTag, keyIndex, mbControlDown, mbShiftDown, mbAltDown);
		
		if (bBreak == TRUE)
		    goto BREAK_DUAL_LOOP;

		if (CParam.mVCKeys[keyIndex].mbSelected == FALSE){
		    Refresh();
		    return TRUE;
		}
	    }
	}
    }
BREAK_DUAL_LOOP:
    
    if (bKeySelected == FALSE){
	if (bAdditive == FALSE){
	    GetPCSceneObjCtrl()->DeselectKeyFrame();
	    SendUpdate();
	}
	return FALSE;
    }
    else{
	if (bAdditive == FALSE && bSelectedKeySelected == FALSE){
	    wxString strTag = mpCParent->mCstrLinkedParamGroup.GetBodySTR();
	    GetPCSceneObjCtrl()->DeselectKeyFrame(strTag);
	}
	SendUpdate();
    }

    mbKeySelecting = bKeySelected;

    return TRUE;
}

void CLS::
SelectItemByRegion(){
    ParamGroup_cls *pCParamGroup = mpCParent->GetPCParamGroup();
    if (pCParamGroup == NULL) return;

    double resolution = _RES_DEFAULT;
    if (mbForColor == TRUE)
	resolution = 1.0 / _COLOR_CURVE_RESOLUTION;

    uint num = pCParamGroup->mVCParam.size();
    if (mAAKeyFrameRect.size() == 0) num = 0; // timedepend 0 
    for (uint param = 0; param < num; param++){
	Parameter_cls &CParam = pCParamGroup->mVCParam[param];
	wxString strTag = mpCParent->mCstrLinkedParamGroup.GetBodySTR();
	strTag << "::" << CParam.GetStrName();
	uint keyNum = mAAKeyFrameRect[param].size();
	for (uint keyIndex = 0; keyIndex < keyNum; keyIndex++){
	    wxRect CRectKey = mAAKeyFrameRect[param][keyIndex];
	    CRectKey.Deflate(1, 1);
	    if (mCRectSelectRegion.Contains(CRectKey)){
		GetPCSceneObjCtrl()->
		    SelectKeyFrame(strTag, keyIndex, TRUE, TRUE, resolution);
	    }
	}
    }
    SendUpdate();
}

void CLS::
OnPaint(wxPaintEvent &event){
    wxAutoBufferedPaintDC dc(this);
    mPanelWidth = GetClientSize().GetWidth();
    mPanelHeight = GetClientSize().GetHeight();
    wxRect rectToDraw (0, 0, mPanelWidth, mPanelHeight);

    dc.SetPen(*wxBLACK_PEN);
    dc.SetBrush(wxBrush(*wxLIGHT_GREY));
    dc.DrawRectangle(rectToDraw);

    ParamGroup_cls *pCParamGroup = mpCParent->GetPCParamGroup();
    if (pCParamGroup != NULL)
	DrawGraphOfParamGroup(dc, pCParamGroup);

    DrawDraggingRegion(dc);
    if (mbForColor == FALSE) DrawCurrentTime(dc);
    DrawNaviCurrent(dc);
}

void CLS::
OnLeftDown(wxMouseEvent &event){
    LeftDownBase(event);

    bool bAdditional = event.m_shiftDown;
    mbDraggingRegion = FALSE;
    if (SelectItemByMouse(bAdditional) == FALSE)
	mbDraggingRegion = TRUE;

    if (HasCapture() == FALSE) CaptureMouse();

    SetFocus();
}

bool CLS::
ProcKeyDialogCheck(wxPoint CAtDown){

    ParamGroup_cls *pCParamGroup = mpCParent->GetPCParamGroup();
    if (pCParamGroup == NULL) return FALSE;
    
    uint dim = pCParamGroup->mVCParam.size();
    
    for (uint index = 0; index < dim; index++){
	uint keyNum = mAAKeyFrameRect[index].size();
	Parameter_cls *pCParam = &pCParamGroup->mVCParam[index];
	wxString strItem = mpCParent->mCstrLinkedParamGroup.GetBodySTR();

	//wb_͗vȂ
	if (PopStrItemFromStrLine(strItem, "::").IsSameAs("SceneObjCtrl"))
	    PopStrItemFromStrLine(strItem, "::"); 
	
	strItem << "::" << pCParam->GetStrName();
	
	for (int i = 0; i < keyNum; i++){
	    if (mAAKeyFrameRect[index][i].Contains(CAtDown)){
		int keyIndex = mAAKeyFrameIndex[index][i]; 
		
		GetPCParamDialog()->mCPos = 
		    GetScreenPosition() + GetPointAtDown() - wxPoint(20, 30);
		GetPCParamGroupDialog()->mCPos = 
		    GetScreenPosition() + GetPointAtDown() - wxPoint(20, 30);
		
		if (pCParam->IsKeyLinkedAll(keyIndex) == TRUE)
		    GetPCParamGroupDialog()->OpenWindow
			(this, pCParam, keyIndex, strItem);
		else
		    GetPCParamDialog()->OpenWindow
			(this, pCParam, keyIndex, strItem);
		return TRUE;
	    }
	}
    }
    return FALSE;
}

void CLS::
OnRightDown(wxMouseEvent &event){
    LeftDownBase(event);

    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    ProcKeyDialogCheck(CAtDown);
}

void CLS::
OnLeftUp(wxMouseEvent &event){
    if (HasCapture() == TRUE) ReleaseMouse();

    if (mbDraggingRegion == TRUE){
	SelectItemByRegion();
	mbDraggingRegion = FALSE;
	mpCParent->UpdateGraphs();
    }
    if (mbKeySelecting == TRUE){
	double dNX, dNY;
	GetMouseDNPos(dNX, dNY);
	GetPCSceneObjCtrl()->MoveSelectedTLKeyCommand(dNX, dNY);
	mbKeySelecting = FALSE;
    }
}

void CLS::
OnDoubleClick(wxMouseEvent &event){
    LeftDownBase(event);

    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    if (mbShiftDown == FALSE && mbControlDown == FALSE && mbAltDown == FALSE)
	ProcKeyDialogCheck(CAtDown);
}

void CLS::
MoveCenter(MouseNormalized_cls &CMP){
    float dY = CMP.GetDY();
    float viewCenter = (mViewMaxY + mViewMinY) / 2;
    float viewHeight = mViewMaxY - mViewMinY;
    float viewYmax = mViewMaxY - viewHeight * dY * 0.5;
    float viewYmin = mViewMinY - viewHeight * dY * 0.5;
    SetYViewRange(viewYmin, viewYmax);
    SetYValueRange(viewYmin, viewYmax);
    mpCParent->SyncronizeTicsToGraph();
}

void CLS::
Zoom(MouseNormalized_cls &CMP){
    float rate = 0;
    if (mpCMouseCtrl->IsWheel() == TRUE)
	rate = mpCMouseCtrl->mWheelRotationNormalized;
    if (mpCMouseCtrl->IsZoom() == TRUE)
	rate = CMP.GetDY() * 3;
    
    if (GetPCMousePreference()->mCbZoomInvert.GetBodyBOOL() == TRUE) rate *= -1;
    
    float viewCenter = (mViewMaxY + mViewMinY) / 2;
    float viewHeight = mViewMaxY - mViewMinY;
    float viewYmax = viewCenter + (viewHeight / 2) * (1.0f + rate*0.2);
    float viewYmin = viewCenter - (viewHeight / 2) * (1.0f + rate*0.2);
    SetYViewRange(viewYmin, viewYmax);
    SetYValueRange(viewYmin, viewYmax);
    mpCParent->SyncronizeTicsToGraph();
}

void CLS::
OnMotion(wxMouseEvent &event){
    
    mMouseXCurrent = event.GetX();
    mMouseYCurrent = event.GetY();
    
    mpCMouseCtrl->SetMouseInformation(event);
    MouseNormalized_cls CMP(mpCMouseCtrl, GetSize());

    int x = event.GetX();
    int y = event.GetY();

    bool bRedraw = FALSE;
    //bool bRedrawButton = mbSelectGuide;

    if (event.Dragging() == TRUE){
	bRedraw = TRUE;
    }

    /*
    if (event.ButtonDown(wxMOUSE_BTN_LEFT) && 
	mpCMouseCtrl->IsMoveCenter() == FALSE){
	
	SelectionCheck(event.m_shiftDown);
	
	CaptureMouse();
	SetFocus();
    }
    */

    if (/*mbLockYRange == FALSE &&*/ mpCMouseCtrl->IsMoveCenter()){
	mbDraggingRegion = FALSE;
	MoveCenter(CMP);
	bRedraw = TRUE;
	//bRedrawButton = TRUE;
    }

    if (/*mbLockYRange == FALSE && */
	(mpCMouseCtrl->IsWheel() == TRUE || mpCMouseCtrl->IsZoom())){
	mbDraggingRegion = FALSE;
	Zoom(CMP);
	bRedraw = TRUE;
	//bRedrawButton = TRUE;
    }

    mpCMouseCtrl->SetPreviousPosition();

    if (bRedraw == TRUE) mpCParent->UpdateGraphs();

    event.Skip();
}

wxColour GraphPanel_cls::msColorRGBA[4] = {
    wxColour(255, 0, 0), wxColour(0, 255, 0), wxColour(0, 0, 255), 
    wxColour(255,255,255) };

wxColour GraphPanel_cls::msColorLockedRGBA[4] = {
    wxColour(208, 164, 164), wxColour(164, 208, 164), wxColour(164, 164, 208),
    wxColour(208, 208, 208) };

