// -*- c -*- 
// "clsGraphPanel.cpp"

#ifndef _GRAPH_PANEL_H_
#define _GRAPH_PANEL_H_

//#include "clsHookCtrl.h"
#include "../INPUTMETHOD/clsClickInfoCtrl.h"
#include "../WIDGETCTRL/clsWidgetCommon.h"

#include <wx/dcbuffer.h>
#include <vector>

#define _GRAPH_MONOCHROME 0
#define _GRAPH_RGBA       1

#include "../COMMON/clsGetPC.h"
#include "../FILEIO/clsParser.h"

class EditorBase_cls;
class Parameters_cls;
class ParamGroup_cls;
class MouseCtrl_cls;
class MouseNormalized_cls;
class GraphEditor_cls;

class GraphPanel_cls : public wxPanel, public ClickInfoCtrl_cls, public WidgetCommon_cls, public GetPC_cls, public Parser_cls {
 public:
    GraphPanel_cls(EditorBase_cls *pCParent,
		   bool bColor = TRUE, bool bLimitYValue = FALSE);
    ~GraphPanel_cls();

    EditorBase_cls* mpCParent;
    MouseCtrl_cls*  mpCMouseCtrl;
    int mPanelWidth, mPanelHeight;
    int mXMarginMin, mXMarginMax;
    bool mbLimitYValue;
    bool mbForColor, mbLockYRange, mbShowVelocity, mbUseSnap;
    bool mbKeySelecting;
    int  mColorType;

    double mViewMinX, mViewMaxX;
    double mValueLimitMinX, mValueLimitMaxX;
    double mViewMinY, mViewMaxY;
    double mValueLimitMinY, mValueLimitMaxY;
    double mValueMinY, mValueMaxY; // l

    std::vector < std::vector < wxRect > > mAAKeyFrameRect;
    std::vector < std::vector < int > > mAAKeyFrameIndex;
    
    static wxColour msColorRGBA[4];
    static wxColour msColorLockedRGBA[4];

    void SetXViewRange(double minX, double maxX);
    void SetYViewRange(double minY, double maxY);
    void SetYValueRange(double minY, double maxY);

    float GetViewWidth();
    float GetViewHeight();
    int  GetCoordX(double valueX);
    int  GetCoordY(double valueY);
    void GetMouseDPos(int &dx, int &dy);
    void GetMouseDNPos(double &dNX, double &dNY);
    double GetXNormalized(int coordx);
    double GetYNormalized(int coordy, bool bLimit = TRUE);
    
    void SetPenColor(wxAutoBufferedPaintDC &dc, ParamGroup_cls *pCParamGroup, int i);
    void SetBrushColor(wxAutoBufferedPaintDC &dc, int i);
    
    void UpdateYValueRange(ParamGroup_cls *pCParamGroup);
    void DrawCurrentTime(wxDC &dc);
    void DrawNaviCurrent(wxDC &dc);
    void DrawGraphOfParamGroup(wxAutoBufferedPaintDC &dc, ParamGroup_cls *pCParamGroup);
    bool SelectItemByMouse(bool bAdditive/*, bool bRightBtn = false*/);
    void SelectItemByRegion();
    void OnPaint(wxPaintEvent &event);
    void OnLeftDown(wxMouseEvent &event);
    bool ProcKeyDialogCheck(wxPoint CAtDown);
    void OnRightDown(wxMouseEvent &event);
    void OnLeftUp(wxMouseEvent &event);
    void OnDoubleClick(wxMouseEvent &event);
    void MoveCenter(MouseNormalized_cls& CMP);
    void Zoom(MouseNormalized_cls& CMP);
    void OnMotion(wxMouseEvent &event);


    void SendUpdate();
    /*
    ParamGroup_cls* mpCParamGroupLinked;
    bool mbInitialized, mbGuideSnip;
    bool mbSelectGuide;//

    wxMenu *mpCMenu;
    wxIcon mCIcon;

    std::vector < double > mVTicsX;
    std::vector < double > mVTicsY;

    void SetParamGroup(ParamGroup_cls *pCParamGroup);
    void GrayOutUnusedMenu();
    void ShowMenu();

    
    void GetSnippedDelta(int &dx, int &dy);
    void GetCurrentColor(float clr[3]);
    void UpdateCurrentNT(double doubleIndex);
    void SetGuide(double doubleIndex);
    
    

    void DrawTicsBox(wxAutoBufferedPaintDC& dc);
    void DrawSelectionBox(wxAutoBufferedPaintDC& dc);
        
    bool   mbShowCurrentNT;
    
       
    

    
    
    double mCurrentNT;

    int    mGuideX;
    int    mCoordX, mCoordY;
    int    mCoordXAtClick, mCoordYAtClick;
    double mValueX, mValueY;
    double mValueXAtClick, mValueYAtClick;

    double mCurrentDoubleIndex;
    Parameters_cls *mpCParam;
    int mTypeCache;
    double mDValue1Cache, mDValue2Cache;
    int mIndexHermiteParam, mIndexVelocity;

    

    void InsertKey();
    void DeleteKey();
    void EditKey();
    void HideAndShowParameters(void *pArg);
    void HookFuncHS(int id, void *pArg);
    void SetKeyColor();
    void OnInsertKey(wxCommandEvent& event);
    void OnDeleteKey(wxCommandEvent& event);
    void OnEditKey(wxCommandEvent& event);
    void OnColor(wxCommandEvent& event);
    void SetMode(int mode);
    void OnKeyDown(wxKeyEvent& event);
    */
        
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

#endif
