// (setq c-basic-offset 4)
// "clsTics.h"

#include "clsTics.h"
#include "../MISC/clsMisc.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsFontProvider.h"

Tics_cls::
Tics_cls(bool bVertical){
    //mBGColor  = wxColour(216, 216, 216);
    mFGColor1 = wxColour(32, 32, 32);
    mFGColor2 = wxColour(128, 128, 128);
    mbVertical = bVertical;
    mTicBorderXmin = 0;
    mTicBorderXmax = 0;
    mTicBorderYmin = 0;
    mTicBorderYmax = 0;
}

void Tics_cls::
SetParameter(wxSize size, float min, float max){
    
    mTicWidth  = size.GetWidth();
    mTicHeight = size.GetHeight();
    
    mRegionMin = min;
    mRegionMax = max;
    float rangeWidth = max - min;

    rangeWidth *= (mViewMaxX - mViewMinX);

    if (fabs(rangeWidth) < 2) mbUseFloat = TRUE;
    else                      mbUseFloat = FALSE;
    
    float ratio = fabs(rangeWidth);
    if (mbVertical == TRUE) ratio /= mTicHeight;
    else                    ratio /= mTicWidth;

    if (mRegionMax > 1000) ratio *= 2;
		       
    mStepLarge = 1;

    mStepFloatOrder = 1;
    float stepOrder = 1; 
    while (fabs(rangeWidth) < stepOrder * 1.01){
	mStepFloatOrder ++;
	stepOrder /= 10;
    }

    mStepFloatScaling = 0.1f;
    float floatScaling = 0.1f;
    while (fabs(rangeWidth) < mStepFloatScaling * 5){
	mStepFloatScaling /= 10;
    }
    if (rangeWidth / mStepFloatScaling > 10) mStepFloatScaling *= 2;
    if (rangeWidth / mStepFloatScaling > 10) mStepFloatScaling *= 5;
    
    if (ratio > 0.021)  mStepLarge = 2;
    if (ratio > 0.051)  mStepLarge = 5;
    if (ratio > 0.101)  mStepLarge = 10;
    if (ratio > 0.201)  mStepLarge = 20;
    if (ratio > 0.501)  mStepLarge = 50;
    if (ratio > 1.001)  mStepLarge = 100;
    if (ratio > 2.001)  mStepLarge = 200;
    if (ratio > 5.001)  mStepLarge = 500;
    if (ratio > 10.001)  mStepLarge = 1000;
    if (ratio > 20.001)  mStepLarge = 2000;
    if (ratio > 100.001)  mStepLarge = 10000;
    if (ratio > 200.001)  mStepLarge = 20000;
    if (ratio > 500.001)  mStepLarge = 50000;
    if (ratio > 1000.001)  mStepLarge = 100000;
    if (ratio > 2000.001)  mStepLarge = 200000;
    if (ratio > 5000.001)  mStepLarge = 500000;
    if (ratio > 10000.001)  mStepLarge = 1000000;
    if (ratio > 20000.001)  mStepLarge = 2000000;
    if (ratio > 50000.001)  mStepLarge = 5000000;
    if (ratio > 100000.01)  mStepLarge = 10000000;
    if (ratio > 200000.01)  mStepLarge = 20000000;
    if (ratio > 500000.01)  mStepLarge = 50000000;
    if (ratio > 1000000.01)  mStepLarge = 100000000;
    if (ratio > 2000000.01)  mStepLarge = 200000000;
    if (ratio > 5000000.01)  mStepLarge = 500000000;
    if (ratio > 10000000.1) mStepLarge = 1000000000;
    if (ratio > 20000000.1) mStepLarge = 2000000000;
    if (ratio > 50000000.1) mStepLarge = 5000000000;
    if (ratio > 100000000.1) mStepLarge = 10000000000;
    if (ratio > 200000000.1) mStepLarge = 20000000000;
    if (ratio > 500000000.1) mStepLarge = 50000000000;
    
    mStepMiddle = mStepLarge / 2;
    mStepSmall = mStepLarge / 10; 
    if (mStepLarge == 5) mStepMiddle = 5;
    if (mStepSmall == 0) mStepSmall = 1;
    if (mStepMiddle == 0) mStepMiddle = 1;
}

void Tics_cls::
SetBorderX(int borderXmin, int borderXmax){
    mTicBorderXmin = borderXmin;
    mTicBorderXmax = borderXmax;
}

void Tics_cls::
SetBorderY(int borderYmin, int borderYmax){
    mTicBorderYmin = borderYmin;
    mTicBorderYmax = borderYmax;
}

void Tics_cls::
SetRegion(double regionMin, double regionMax){
    mRegionMin = regionMin;
    mRegionMax = regionMax;
}

void Tics_cls::
SetColorParameter(/*wxColour BGColor,*/wxColour FGColor1, wxColour FGColor2){
    //mBGColor = BGColor;
    mFGColor1 = FGColor1;
    mFGColor2 = FGColor2;
}

int Tics_cls::
GetTicDeviceX(float x){
    return (int)(GetMappedX(x) * (mTicWidth - mTicBorderXmin - mTicBorderXmax) + mTicBorderXmin);
}

int Tics_cls::
GetTicDeviceY(float y){
    return (mTicHeight - mTicBorderYmax) - (int)(GetMappedY(y) * (mTicHeight - mTicBorderYmin - mTicBorderYmax) + mTicBorderYmin);
}

void Tics_cls::
CalcTicDeviceCoordinateX(int *pX0, int *pX1,
			 float frame0, float frame1){
    if (mRegionMax == mRegionMin){
	*pX0 = 0; *pX1 = 1;
	return;
    }

    float diff0 = frame0 - mRegionMin;
    float diff1 = frame1 - mRegionMin;
    float width = mRegionMax - mRegionMin;
    float x0 = diff0/width;
    float x1 = diff1/width;
    *pX0 = GetTicDeviceX(x0); *pX1 = GetTicDeviceX(x1);
}

void Tics_cls::
CalcTicDeviceCoordinateY(int *pY0, int *pY1,
			 float frame0, float frame1){
    if (mRegionMax == mRegionMin){
	*pY0 = 0; *pY1 = 1;
	return;
    }

    float diff0 = frame0 - mRegionMin;
    float diff1 = frame1 - mRegionMin;
    float width = mRegionMax - mRegionMin;
    float y0 = diff0/width;
    float y1 = diff1/width;
    *pY0 = GetTicDeviceY(y0); *pY1 = GetTicDeviceY(y1);
}

void Tics_cls::
DrawTextVerticalTic(wxDC &dc, wxString strMes, int ticsY, int mesOffset){
    wxCoord w, h, singleH;
    dc.GetMultiLineTextExtent(strMes, &w, &h, &singleH);
    
    dc.DrawText(strMes, mTicWidth*7/8 - w, ticsY - h + mesOffset);
    dc.DrawRectangle(mTicWidth*7/8, ticsY, mTicWidth/8, 1);
    
    dc.GetMultiLineTextExtent(strMes, &w, &h, &singleH);
}

void Tics_cls::
DrawTicsVertical(wxDC &dc){
    
    dc.SetFont(FontProvider_cls::GetFont(0));

    dc.SetPen(mFGColor2);
    dc.SetTextForeground(mFGColor2);
    //int prevTexEnd = -1;
    if (mbUseFloat == TRUE){
	float scale = mStepFloatScaling;
	long long int   tics0 = (long long int)(mRegionMin / scale);
	long long int   tics1 = (long long int)(mRegionMax / scale);
	if (tics1 < tics0) Misc_cls::Swap (tics0, tics1);
	tics1++;
	for (long long int i = tics0; i <= tics1; i++){
	    float value = i * scale;
	    
	    int ticsY0, ticsY1;

	    CalcTicDeviceCoordinateY(&ticsY0, &ticsY1, value, value);
	    dc.SetPen(mFGColor1);
		    
	    wxString strMes, strFormat;
	    if (mStepFloatOrder < 5)
		strFormat << wxT("%0.") << mStepFloatOrder << wxT("f");
	    else{
		strFormat << wxT("%.1e");
	    }
	    strMes << wxString::Format(strFormat, value);
	    if (strMes.Last() == '0') strMes.RemoveLast();
	    strMes.Replace(wxT("e+0"), wxT("e+"));
	    strMes.Replace(wxT("e-0"), wxT("e-"));

	    int mesOffset = 2;
	    DrawTextVerticalTic(dc, strMes, ticsY0, mesOffset);
	}
    }
    else{
	long long int   increment = 1;
	long long int   tics0 = (int)mRegionMin;
	long long int   tics1 = (int)mRegionMax;
	if (tics1 < tics0) Misc_cls::Swap (tics0, tics1);
	tics1++;
	for (long long int i = tics0; i <= tics1; i+= increment){

	    long long modIndex = i % mStepSmall;
	    if (modIndex != 0){ 
		if (modIndex > 0){
		    if (mStepSmall - modIndex > 10)
			i += (mStepSmall - modIndex) / 2;
		}
		if (modIndex < 0){
		    if (modIndex < -10)
			i += - modIndex / 2;
		}
		continue;
	    }
		
	    int ticsY0, ticsY1;
	    
	    CalcTicDeviceCoordinateY(&ticsY0, &ticsY1, i, i);
	    if (ticsY0 < 0 || ticsY0 > mTicHeight) continue;
	    if (i % mStepMiddle == 0) 
		dc.SetPen(mFGColor1);
	    else
		dc.SetPen(mFGColor2);
	    if (i % mStepLarge == 0){
		wxString strMes;
		if (mStepLarge <= 1000)
		    strMes << i;
		else
		    strMes << wxString::Format(wxT("%.1e"), (float)i);
		strMes.Replace(wxT("e+0"), wxT("e+"));
		strMes.Replace(wxT("e-0"), wxT("e-"));

		int mesOffset = 2;
		DrawTextVerticalTic(dc, strMes, ticsY0, mesOffset);
		continue;
	    }
	    dc.DrawRectangle(mTicWidth-3, ticsY0, 3, 1);

	    if (mStepLarge > 100) increment = mStepLarge / 10;
	}
    }
    
}

void Tics_cls::
DrawTics(wxDC &dc, bool bClear, bool bUseWhite, int fontSize){
    ColorProvider_cls::SetBGColorDC(dc, "BasePanel");
    //dc.SetBackground(wxBrush(mBGColor, wxSOLID));
    //dc.SetBrush(mBGColor);
    //dc.SetPen(*wxTRANSPARENT_PEN);
    if (bClear == TRUE) dc.Clear();
    
    if (mbVertical == TRUE) return DrawTicsVertical(dc);

    /*
    if (mTicHeight > 100)
	dc.SetFont(FontProvider_cls::GetFont(7));
    if (mTicHeight > 50)
	dc.SetFont(FontProvider_cls::GetFont(6));
    if (mTicHeight > 30)
	dc.SetFont(FontProvider_cls::GetFont(5));
    else if (mTicHeight > 25)
	dc.SetFont(FontProvider_cls::GetFont(4));
    else if (mTicHeight > 20)
	dc.SetFont(FontProvider_cls::GetFont(1));
    else
	dc.SetFont(FontProvider_cls::GetFont(0));
    */
    dc.SetFont(FontProvider_cls::GetFont(fontSize));

    wxCoord w, h, singleH;
    dc.GetMultiLineTextExtent(wxT("0.0"), &w, &h, &singleH);

#ifdef WIN32
    int texOffsetY = singleH * 7 / 8;
#else
    int texOffsetY = singleH;
#endif

    if (bUseWhite == TRUE){
	dc.SetPen(*wxWHITE);
	dc.SetTextForeground(*wxWHITE);
    }
    else{
	dc.SetPen(mFGColor2);
	dc.SetTextForeground(mFGColor2);
    }
    int prevTexEnd = -1;
    if (mbUseFloat == TRUE){
	float scale = mStepFloatScaling;
	long long int   tics0 = (long long int)(mRegionMin / scale);
	long long int   tics1 = (long long int)(mRegionMax / scale);
	if (tics1 < tics0) Misc_cls::Swap (tics0, tics1);
	tics1++;
	for (long long int i = tics0; i <= tics1; i++){
	    float value = i * scale;
	    
	    int ticsX0, ticsX1;
	    CalcTicDeviceCoordinateX(&ticsX0, &ticsX1, value, value);
	    if (bUseWhite == TRUE)
		dc.SetPen(*wxWHITE);
	    else
		dc.SetPen(mFGColor1);
	    
	    wxString strMes, strFormat;
	    if (mStepFloatOrder < 5)
		strFormat << wxT("%0.") << mStepFloatOrder << wxT("f");
	    else{
		strFormat << wxT("%.1e");
	    }
	    strMes << wxString::Format(strFormat, value);
	    if (strMes.Last() == '0') strMes.RemoveLast();
	    strMes.Replace(wxT("e+0"), wxT("e+"));
	    strMes.Replace(wxT("e-0"), wxT("e-"));
	    int mesOffset = 2;
	    if (prevTexEnd < ticsX0 + mesOffset){
		dc.DrawText(strMes, ticsX0 + mesOffset, mTicHeight-texOffsetY);
		dc.DrawRectangle(ticsX0, mTicHeight-singleH/4, 1, singleH/4);
		dc.GetMultiLineTextExtent(strMes, &w, &h, &singleH);
		prevTexEnd = ticsX0+mesOffset+w;
	    }
	}
    }
    else{
	long long int   increment = 1;
	long long int   tics0 = (long long int)mRegionMin;
	long long int   tics1 = (long long int)mRegionMax;
	if (tics1 < tics0) Misc_cls::Swap (tics0, tics1);
	tics1++;
	for (long long int i = tics0; i <= tics1; i+= increment){
	    long long modIndex = i % mStepSmall;
	    if (modIndex != 0){ 
		if (modIndex > 0){
		    if (mStepSmall - modIndex > 10)
			i += (mStepSmall - modIndex) / 2;
		}
		if (modIndex < 0){
		    if (modIndex < -10)
			i += - modIndex / 2;
		}
		continue;
	    }
		

	    int ticsX0, ticsX1;
	    CalcTicDeviceCoordinateX(&ticsX0, &ticsX1, i, i);
	    if (ticsX1 < 0 || ticsX0 > mTicWidth) continue;
	    if (bUseWhite == TRUE)
		dc.SetPen(*wxWHITE);
	    else{
		if (i % mStepMiddle == 0) 
		    dc.SetPen(mFGColor1);
		else
		    dc.SetPen(mFGColor2);
	    }
	    if (i % mStepLarge == 0){
		wxString strMes;
		if (mStepLarge <= 1000)
		    strMes << i;
		else
		    strMes << wxString::Format(wxT("%.1e"), (float)i);
		strMes.Replace(wxT("e+0"), wxT("e+"));
		strMes.Replace(wxT("e-0"), wxT("e-"));

		int mesOffset = 2;
		if (prevTexEnd < ticsX0 + mesOffset){
		    dc.DrawText(strMes, ticsX0+mesOffset,mTicHeight-texOffsetY);
		    dc.DrawRectangle(ticsX0, mTicHeight-singleH/4, 1,singleH/4);
		    dc.GetMultiLineTextExtent(strMes, &w, &h, &singleH);
		    prevTexEnd = ticsX0+mesOffset+w;
		}
		continue;
	    }
	    dc.DrawRectangle(ticsX0, mTicHeight-3, 1, 3);
	    
	    if (mStepLarge > 100) increment = mStepLarge / 10;
	}
    }
}

wxString Tics_cls::msAstrTicsMode[3] = 
{wxT("Sequence Index"), wxT("Scene Frame"), wxT("Normalized Time")};

wxString Tics_cls::msAstrTicsModeCC[2] = 
{wxT("Clamped [0-1]"), wxT("[Min, Max]")};
