// "clsHDRI.h"
// (setq c-basic-offset 4)

#ifdef _USE_EXR

#include "clsHDRI.h"

//#include "../MYAPI/MsgBox.h"

#include <ImfRgbaFile.h>
#include <ImfCRgbaFile.h>
#include <ImfChannelList.h>

#include "../COMMON/CommonDefinitions.h"

ThreadHDRISave_cls::
ThreadHDRISave_cls(bool *pbFinishedFlag) : wxThread(wxTHREAD_DETACHED)
{
    mpbFinishedFlagLink = pbFinishedFlag;
}

void ThreadHDRISave_cls::
SetParameters(float *pBufferSrc, wxSize CSizeFBO, wxSize CSizeImage,
	      wxString strFileOut, int channelSRC,
	      int depth, int compress,
	      float projMatrix14, float projMatrix10){
    
    mpBufferSRC = pBufferSrc;
    mBufferWidth  = CSizeFBO.GetWidth();
    mBufferHeight = CSizeFBO.GetHeight();
    mWidthOut  = CSizeImage.GetWidth();
    mHeightOut = CSizeImage.GetHeight();
    mstrFileOut = strFileOut;
    
    mChannelSRC = channelSRC;
    mDepth = depth; mCompress = compress;
    mProjMatrix14 = projMatrix14;
    mProjMatrix10 = projMatrix10;

    ASSERT(Create() == wxTHREAD_NO_ERROR);
}

OPENEXR_IMF_NAMESPACE::Compression ThreadHDRISave_cls::
GetCompress(){
    OPENEXR_IMF_NAMESPACE::Compression compress;
    switch(mCompress){
    case 0: compress = OPENEXR_IMF_NAMESPACE::ZIP_COMPRESSION; break;
    case 1: compress = OPENEXR_IMF_NAMESPACE::PIZ_COMPRESSION; break;
    case 2: compress = OPENEXR_IMF_NAMESPACE::RLE_COMPRESSION; break;
    case 3: compress = OPENEXR_IMF_NAMESPACE::PXR24_COMPRESSION; break;
    }
    return compress;
}

float ThreadHDRISave_cls::
GetAverage(double bufferX, double bufferY,
	   double bufferX2, double bufferY2, int rgb, int bufferChannel){
    double sum = 0, areaSum = 0;
    if (bufferY2 >= mBufferHeight) bufferY2 = mBufferHeight - 0.001;
    if (bufferX2 >= mBufferWidth) bufferX2 = mBufferWidth - 0.001;
    
    for (int bufferiY = (int)bufferY; bufferiY <= bufferY2; bufferiY++){
	double factorY = 1;
	if (bufferY - bufferiY > 0)
	    factorY -= bufferY - bufferiY;
	if (bufferY2 - bufferiY < 1)
	    factorY -= (bufferiY + 1 - bufferY2);
	int indexY = bufferiY * mBufferWidth;
	for (int bufferiX = (int)bufferX; bufferiX <= bufferX2; bufferiX++){
	    double factorX = 1;
	    if (bufferX - bufferiX > 0)
		factorX -= bufferX - bufferiX;
	    if (bufferX2 - bufferiX < 1)
		factorX -= (bufferiX + 1 - bufferX2);
	    int index = (indexY + bufferiX) * bufferChannel;
	    double area = factorX * factorY;
	    sum += mpBufferSRC[index + rgb] * area;
	    areaSum += area;
	}
    }

    return sum / areaSum;
}

void ThreadHDRISave_cls::
Save1chHalf(){
    
    half *luminance = new half [mHeightOut * mWidthOut];
    if (mWidthOut < mHeightOut){
	double dWidth = 1.0 / mHeightOut * mBufferWidth;
	double dHeight = 1.0 / mHeightOut * mBufferHeight;
	double dOffsetX = (1.0 - (double)mWidthOut/mHeightOut) / 2;
	for (int y = 0; y < mHeightOut; y++){
	    double bufferY = (double)y / mHeightOut * mBufferHeight;
	    int indexyInv = (mHeightOut - y - 1)* mWidthOut;
	    for (int x = 0; x < mWidthOut; x++){
		double bufferX = ((double)x/mHeightOut + dOffsetX) * mBufferWidth;
		float average = 
		    GetAverage(bufferX, bufferY, 
			       bufferX + dWidth, bufferY + dHeight, 0, 1);
		float z_eye = mProjMatrix14/(average * -2.0 + 1.0 - mProjMatrix10);
		luminance[indexyInv+x] = -z_eye;
	    }
	}
    }
    else{
	double dWidth = 1.0 / mWidthOut * mBufferWidth;
	double dHeight = 1.0 / mWidthOut * mBufferHeight;
	double dOffsetY = (1.0 - (double)mHeightOut/mWidthOut) / 2;

	for (int y = 0; y < mHeightOut; y++){
	    double bufferY = ((double)y / mWidthOut + dOffsetY) * mBufferHeight;
	    int indexyInv = (mHeightOut - y - 1) * mWidthOut;

	    for (int x = 0; x < mWidthOut; x++){
		double bufferX = (double)x /mWidthOut * mBufferWidth;

		float average = 
		    GetAverage(bufferX, bufferY, 
			       bufferX + dWidth, bufferY + dHeight, 0, 1);
		
		float z_eye = mProjMatrix14/(average * -2.0 + 1.0 - mProjMatrix10);
		luminance[indexyInv+x] = -z_eye;
	    }
	}
    }

    OPENEXR_IMF_NAMESPACE::Header header(mWidthOut, mHeightOut);
    header.compression() = GetCompress();

    header.channels().insert("Y", OPENEXR_IMF_NAMESPACE::Channel(OPENEXR_IMF_NAMESPACE::HALF));
    OPENEXR_IMF_NAMESPACE::OutputFile File2(mstrFileOut.char_str(), header);
    
    OPENEXR_IMF_NAMESPACE::FrameBuffer frameBuffer;
    
    frameBuffer.insert ("Y", // name 
			OPENEXR_IMF_NAMESPACE::Slice 
			(OPENEXR_IMF_NAMESPACE::HALF, // type 
			 (char *) luminance,      // base 
			 sizeof (*luminance) * 1, // xStride
			 sizeof (*luminance) * mWidthOut)); // yStride
    
    File2.setFrameBuffer(frameBuffer);
    File2.writePixels(mHeightOut);
    
    delete [] luminance;
}

void ThreadHDRISave_cls::
Save1chFloat(){

    float *luminance = new float [mHeightOut * mWidthOut];    
    if (mWidthOut < mHeightOut){
	double dWidth = 1.0 / mHeightOut * mBufferWidth;
	double dHeight = 1.0 / mHeightOut * mBufferHeight;
	double dOffsetX = (1.0 - (double)mWidthOut/mHeightOut) / 2;
	for (int y = 0; y < mHeightOut; y++){
	    double bufferY = (double)y / mHeightOut * mBufferHeight;
	    int indexyInv = (mHeightOut - y - 1) * mWidthOut;
	    for (int x = 0; x < mWidthOut; x++){
		double bufferX = ((double)x/mHeightOut + dOffsetX) * mBufferWidth;
		float average = 
		    GetAverage(bufferX, bufferY, 
			       bufferX + dWidth, bufferY + dHeight,  0, 1);
		float z_eye = mProjMatrix14/(average * -2.0 + 1.0 - mProjMatrix10);
		luminance[indexyInv+x] = -z_eye;
	    }
	}
    }
    else{
	double dWidth = 1.0 / mWidthOut * mBufferWidth;
	double dHeight = 1.0 / mWidthOut * mBufferHeight;
	double dOffsetY = (1.0 - (double)mHeightOut/mWidthOut) / 2;

	for (int y = 0; y < mHeightOut; y++){
	    double bufferY = ((double)y / mWidthOut + dOffsetY) * mBufferHeight;
	    int indexyInv = (mHeightOut - y - 1) * mWidthOut;

	    for (int x = 0; x < mWidthOut; x++){
		double bufferX = (double)x /mWidthOut * mBufferWidth;

		float average = 
		    GetAverage(bufferX, bufferY, 
			       bufferX + dWidth, bufferY + dHeight, 0, 1);
		
		float z_eye = mProjMatrix14/(average * -2.0 + 1.0 - mProjMatrix10);
		luminance[indexyInv+x] = -z_eye;
	    }
	}
    }

    OPENEXR_IMF_NAMESPACE::Header header(mWidthOut, mHeightOut);
    header.compression() = GetCompress();
    
    header.channels().insert("Y", OPENEXR_IMF_NAMESPACE::Channel(OPENEXR_IMF_NAMESPACE::FLOAT));
    OPENEXR_IMF_NAMESPACE::OutputFile File2(mstrFileOut.ToAscii(), header);
    
    OPENEXR_IMF_NAMESPACE::FrameBuffer frameBuffer;
    
    frameBuffer.insert ("Y", // name 
			OPENEXR_IMF_NAMESPACE::Slice 
			(OPENEXR_IMF_NAMESPACE::FLOAT, // type 
			 (char *) luminance, // base 
			 sizeof (*luminance) * 1, // xStride
			 sizeof (*luminance) * mWidthOut)); // yStride
    
    File2.setFrameBuffer(frameBuffer);
    File2.writePixels(mHeightOut);
    
    delete [] luminance;
}

void ThreadHDRISave_cls::
Save3or4ch(){
    float *pBufferSrc = mpBufferSRC;
    
    OPENEXR_IMF_NAMESPACE::Rgba *pixels = new OPENEXR_IMF_NAMESPACE::Rgba [mHeightOut * mWidthOut];
    
    if (mWidthOut < mHeightOut){
	double dWidth = 1.0 / mHeightOut * mBufferWidth;
	double dHeight = 1.0 / mHeightOut * mBufferHeight;
	double dOffsetX = (1.0 - (double)mWidthOut/mHeightOut) / 2;
	for (int y = 0; y < mHeightOut; y++){
	    double bufferY = (double)y / mHeightOut * mBufferHeight;
	    int indexyInv = (mHeightOut - y - 1) * mWidthOut;
	    if (mChannelSRC == 3)
		for (int x = 0; x < mWidthOut; x++){
		    double bufferX = ((double)x/mHeightOut + dOffsetX) * mBufferWidth;
		    pixels[indexyInv+x].r = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 0, 3);
		    pixels[indexyInv+x].g = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 1, 3);
		    pixels[indexyInv+x].b = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 2, 3);
		    pixels[indexyInv+x].a = 1.0f;
		}
	    if (mChannelSRC == 4)
		for (int x = 0; x < mWidthOut; x++){
		    double bufferX = ((double)x/mHeightOut + dOffsetX) * mBufferWidth;
		    pixels[indexyInv+x].r = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 0, 4);
		    pixels[indexyInv+x].g = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 1, 4);
		    pixels[indexyInv+x].b = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 2, 4);
		    pixels[indexyInv+x].a = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 3, 4);
		}
	    
	}
    }
    else{
	double dWidth = 1.0 / mWidthOut * mBufferWidth;
	double dHeight = 1.0 / mWidthOut * mBufferHeight;
	double dOffsetY = (1.0 - (double)mHeightOut/mWidthOut) / 2;

	for (int y = 0; y < mHeightOut; y++){
	    double bufferY = ((double)y / mWidthOut + dOffsetY) * mBufferHeight;
	    int indexyInv = (mHeightOut - y - 1) * mWidthOut;
	    
	    if (mChannelSRC == 3)
		for (int x = 0; x < mWidthOut; x++){
		    double bufferX = (double)x /mWidthOut * mBufferWidth;
		    pixels[indexyInv+x].r = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 0, 3);
		    pixels[indexyInv+x].g = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 1, 3);
		    pixels[indexyInv+x].b = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 2, 3);
		    pixels[indexyInv+x].a = 1.0f;
		}
	    if (mChannelSRC == 4)
		for (int x = 0; x < mWidthOut; x++){
		    double bufferX = (double)x/mWidthOut * mBufferWidth;
		    pixels[indexyInv+x].r = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 0, 4);
		    pixels[indexyInv+x].g = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 1, 4);
		    pixels[indexyInv+x].b = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 2, 4);
		    pixels[indexyInv+x].a = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight, 3, 4);
		}
	}
    }
    
    OPENEXR_IMF_NAMESPACE::RgbaChannels type = OPENEXR_IMF_NAMESPACE::RgbaChannels::WRITE_RGB;
    if (mChannelSRC == 4)	
	type = OPENEXR_IMF_NAMESPACE::RgbaChannels::WRITE_RGBA;

    OPENEXR_IMF_NAMESPACE::RgbaOutputFile File(mstrFileOut.ToAscii(), mWidthOut, mHeightOut, type, 1, IMATH_NAMESPACE::V2f (0, 0), 1, OPENEXR_IMF_NAMESPACE::INCREASING_Y, GetCompress());
    File.setFrameBuffer(pixels, 1, mWidthOut);
    File.writePixels(mHeightOut);
    
    delete [] pixels;
}

void ThreadHDRISave_cls::
Save3or4chFloat(){
    
    float *pixR = new float [mHeightOut * mWidthOut];
    float *pixG = new float [mHeightOut * mWidthOut];
    float *pixB = new float [mHeightOut * mWidthOut];
    float *pixA = NULL;
    if (mChannelSRC == 4)
	pixA = new float [mHeightOut * mWidthOut];

    if (mWidthOut < mHeightOut){
	double dWidth = 1.0 / mHeightOut * mBufferWidth;
	double dHeight = 1.0 / mHeightOut * mBufferHeight;
	double dOffsetX = (1.0 - (double)mWidthOut/mHeightOut) / 2;
	for (int y = 0; y < mHeightOut; y++){
	    double bufferY = (double)y / mHeightOut * mBufferHeight;
	    int indexyInv = (mHeightOut - y - 1) * mWidthOut;
	    if (mChannelSRC == 3)
		for (int x = 0; x < mWidthOut; x++){
		    double bufferX = ((double)x/mHeightOut + dOffsetX) * mBufferWidth;
		    pixR[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  0, 3);
		    pixG[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  1, 3);
		    pixB[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  2, 3);
		}
	    if (mChannelSRC == 4)
		for (int x = 0; x < mWidthOut; x++){
		    double bufferX = ((double)x/mHeightOut + dOffsetX) * mBufferWidth;
		    pixR[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  0, 4);
		    pixG[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  1, 4);
		    pixB[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  2, 4);
		    pixA[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  3, 4);
		}
	}
    }
    else{
	double dWidth = 1.0 / mWidthOut * mBufferWidth;
	double dHeight = 1.0 / mWidthOut * mBufferHeight;
	double dOffsetY = (1.0 - (double)mHeightOut/mWidthOut) / 2;

	for (int y = 0; y < mHeightOut; y++){
	    double bufferY = ((double)y / mWidthOut + dOffsetY) * mBufferHeight;
	    int indexyInv = (mHeightOut - y - 1) * mWidthOut;
	    if (mChannelSRC == 3)
		for (int x = 0; x < mWidthOut; x++){
		    double bufferX = (double)x/mWidthOut * mBufferWidth;
		    pixR[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  0, 3);
		    pixG[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  1, 3);
		    pixB[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  2, 3);
		}
	    if (mChannelSRC == 4)
		for (int x = 0; x < mWidthOut; x++){
		    double bufferX = (double)x/mWidthOut * mBufferWidth;
		    pixR[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  0, 4);
		    pixG[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  1, 4);
		    pixB[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  2, 4);
		    pixA[indexyInv+x] = 
			GetAverage(bufferX, bufferY, 
				   bufferX + dWidth, bufferY + dHeight,  3, 4);
		}
	}
    }

    OPENEXR_IMF_NAMESPACE::Header header(mWidthOut, mHeightOut);
    header.compression() = GetCompress();
    
    header.channels().insert("R", OPENEXR_IMF_NAMESPACE::Channel(OPENEXR_IMF_NAMESPACE::FLOAT));
    header.channels().insert("G", OPENEXR_IMF_NAMESPACE::Channel(OPENEXR_IMF_NAMESPACE::FLOAT));
    header.channels().insert("B", OPENEXR_IMF_NAMESPACE::Channel(OPENEXR_IMF_NAMESPACE::FLOAT));
    if (mChannelSRC == 4)
	header.channels().insert("A", OPENEXR_IMF_NAMESPACE::Channel(OPENEXR_IMF_NAMESPACE::FLOAT));

    OPENEXR_IMF_NAMESPACE::OutputFile File2(mstrFileOut.ToAscii(), header);
	
    OPENEXR_IMF_NAMESPACE::FrameBuffer frameBuffer;
    
    frameBuffer.insert ("R", // name 
			OPENEXR_IMF_NAMESPACE::Slice 
			(OPENEXR_IMF_NAMESPACE::FLOAT, // type 
			 (char *) pixR,      // base 
			 sizeof (*pixR) * 1, // xStride
			 sizeof (*pixR) * mWidthOut)); // yStride
    frameBuffer.insert ("G", // name 
			OPENEXR_IMF_NAMESPACE::Slice 
			(OPENEXR_IMF_NAMESPACE::FLOAT, // type 
			 (char *) pixG,      // base 
			 sizeof (*pixG) * 1, // xStride
			 sizeof (*pixG) * mWidthOut)); // yStride
    frameBuffer.insert ("B", // name 
			OPENEXR_IMF_NAMESPACE::Slice 
			(OPENEXR_IMF_NAMESPACE::FLOAT, // type 
			 (char *) pixB,      // base 
			 sizeof (*pixB) * 1, // xStride
			 sizeof (*pixB) * mWidthOut)); // yStride
    if (mChannelSRC == 4)
	    frameBuffer.insert ("A", // name 
			OPENEXR_IMF_NAMESPACE::Slice 
			(OPENEXR_IMF_NAMESPACE::FLOAT, // type 
			 (char *) pixA, // base 
			 sizeof (*pixA) * 1, // xStride
			 sizeof (*pixA) * mWidthOut)); // yStride
	
    
    File2.setFrameBuffer(frameBuffer);
    File2.writePixels(mHeightOut);
    
    delete [] pixR;
    delete [] pixG;
    delete [] pixB;
    if (mChannelSRC == 4)
	delete [] pixA;
}

wxThread::ExitCode ThreadHDRISave_cls::
Entry(){
    //unsigned char *pBufferSrc = mpBufferSRC;
    
    switch(mChannelSRC){
    case 1: 
	if (mDepth == _HDRI_HALF_FLOAT) Save1chHalf();
	else                            Save1chFloat();
	break;
    case 3: case 4:
	if (mDepth == _HDRI_HALF_FLOAT) Save3or4ch();
	else                            Save3or4chFloat();
	break;
    }
    delete [] mpBufferSRC;
    *mpbFinishedFlagLink = TRUE;
    return 0;
}
    
#endif
