// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsImageIO.h"

#include "clsImageIO.h"
#include "clsHDRI.h"

#include "../MYGL/clsFBO.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include <wx/filename.h>

#define CLS ThreadImageSave_cls

CLS::
ThreadImageSave_cls(bool *pbFinishedFlag) : wxThread(wxTHREAD_DETACHED){
    mpbFinishedFlagLink = pbFinishedFlag;
    mpBufferSRC = NULL;
    mpBufferOUT = NULL;
    mpBufferAlphaSRC = NULL;
    mpBufferAlphaOUT = NULL;
}

void CLS::
SetParameters(wxSize CSizeSrc, wxSize CSizeTgt, wxString strFileName,
	      bool bAlpha){
    mCSizeSrc = CSizeSrc;
    mCSizeTgt = CSizeTgt;
    mstrFileName = strFileName;

    int size1ch = mCSizeSrc.GetWidth() * mCSizeSrc.GetHeight();
    mpBufferSRC = (unsigned char*)malloc(size1ch * 3);
    if (bAlpha == TRUE)
	mpBufferAlphaSRC = (unsigned char*)malloc(size1ch);
    
    ASSERT(Create() == wxTHREAD_NO_ERROR);
}

void CLS::
GetAverageRGB(int widthSrc, int heightSrc,
	      double srcX1, double srcY1, double srcX2, double srcY2,
	      unsigned char* pBufferSRC, unsigned char* pBufferTGT){
    double RSum = 0, GSum = 0, BSum = 0;
    double areaSum = 0;
    if (srcY2 >= heightSrc) srcY2 = heightSrc - 0.001;
    if (srcX2 >= widthSrc) srcX2 = widthSrc - 0.001;
    for (int bufferiY = (int)srcY1; bufferiY <= srcY2; bufferiY++){
	double factorY = 1;
	if (srcY1 - bufferiY > 0)
	    factorY -= srcY1 - bufferiY;
	if (srcY2 - bufferiY < 1)
	    factorY -= (bufferiY + 1 - srcY2);
	int indexY = bufferiY * widthSrc;
	for (int bufferiX = (int)srcX1; bufferiX <= srcX2; bufferiX++){
	    double factorX = 1;
	    if (srcX1 - bufferiX > 0) factorX -= srcX1 - bufferiX;
	    if (srcX2 - bufferiX < 1) factorX -= (bufferiX + 1 - srcX2);
	    int index = (indexY + bufferiX) * 3;
	    double area = factorX * factorY;
	    RSum += pBufferSRC[index + 0] * area;
	    GSum += pBufferSRC[index + 1] * area;
	    BSum += pBufferSRC[index + 2] * area;
	    areaSum += area;
	}
    }
    
    pBufferTGT[0] = (unsigned char)(RSum / areaSum);
    pBufferTGT[1] = (unsigned char)(GSum / areaSum);
    pBufferTGT[2] = (unsigned char)(BSum / areaSum);
}

void CLS::
ResizeImageRGB(){
    
    int widthSrc = mCSizeSrc.GetWidth(), heightSrc = mCSizeSrc.GetHeight();
    int widthTgt = mCSizeTgt.GetWidth(), heightTgt = mCSizeTgt.GetHeight();

    if (widthTgt < heightTgt){
	double dWidth = 1.0 / heightTgt * widthSrc;
	double dHeight = 1.0 / heightTgt * heightSrc;
	double dOffsetX = (1.0 - (double)widthTgt/heightTgt) / 2;
	for (int y = 0; y < heightTgt; y++){
	    double srcY = (double)y / heightTgt * heightSrc;
	    int indexy = y * widthTgt;
	    for (int x = 0; x < widthTgt; x++){
		double srcX = ((double)x/heightTgt + dOffsetX) * widthSrc;
		GetAverageRGB(widthSrc, heightSrc,
			      srcX, srcY, srcX+dWidth, srcY+dHeight, 
			      mpBufferSRC, mpBufferOUT + (indexy + x) * 3); }
	}
    }
    else{
	double dWidth = 1.0 / widthTgt * widthSrc;
	double dHeight = 1.0 / widthTgt * heightSrc;
	double dOffsetY = (1.0 - (double)heightTgt/widthTgt) / 2;
	for (int y = 0; y < heightTgt; y++){
	    double srcY = ((double)y/widthTgt + dOffsetY) * heightSrc;
	    int indexy = y * widthTgt;
	    for (int x = 0; x < widthTgt; x++){
		double srcX = (double)x/widthTgt * heightSrc;
		GetAverageRGB(widthSrc, heightSrc,
			      srcX, srcY, srcX+dWidth, srcY+dHeight, 
			      mpBufferSRC, mpBufferOUT + (indexy + x) * 3); }
	}
    }
}

void CLS::
GetAverageAlpha(int widthSrc, int heightSrc,
		double srcX1, double srcY1, double srcX2, double srcY2,
		unsigned char* pBufferSRC, unsigned char* pBufferTGT){
    double AlphaSum = 0;
    double areaSum = 0;
    if (srcY2 >= heightSrc) srcY2 = heightSrc - 0.001;
    if (srcX2 >= widthSrc) srcX2 = widthSrc - 0.001;
    for (int bufferiY = (int)srcY1; bufferiY <= srcY2; bufferiY++){
	double factorY = 1;
	if (srcY1 - bufferiY > 0)
	    factorY -= srcY1 - bufferiY;
	if (srcY2 - bufferiY < 1)
	    factorY -= (bufferiY + 1 - srcY2);
	int indexY = bufferiY * widthSrc;
	for (int bufferiX = (int)srcX1; bufferiX <= srcX2; bufferiX++){
	    double factorX = 1;
	    if (srcX1 - bufferiX > 0) factorX -= srcX1 - bufferiX;
	    if (srcX2 - bufferiX < 1) factorX -= (bufferiX + 1 - srcX2);
	    int index = indexY + bufferiX;
	    double area = factorX * factorY;
	    AlphaSum += pBufferSRC[index] * area;
	    areaSum += area;
	}
    }
    
    *pBufferTGT = (unsigned char)(AlphaSum / areaSum);
}

void CLS::
ResizeImageAlpha(){
    
    int widthSrc = mCSizeSrc.GetWidth(), heightSrc = mCSizeSrc.GetHeight();
    int widthTgt = mCSizeTgt.GetWidth(), heightTgt = mCSizeTgt.GetHeight();

    if (widthTgt < heightTgt){
	double dWidth = 1.0 / heightTgt * widthSrc;
	double dHeight = 1.0 / heightTgt * heightSrc;
	double dOffsetX = (1.0 - (double)widthTgt/heightTgt) / 2;
	for (int y = 0; y < heightTgt; y++){
	    double srcY = (double)y / heightTgt * heightSrc;
	    int indexy = y * widthTgt;
	    for (int x = 0; x < widthTgt; x++){
		double srcX = ((double)x/heightTgt + dOffsetX) * widthSrc;
		GetAverageAlpha(widthSrc, heightSrc,
				srcX, srcY, srcX+dWidth, srcY+dHeight, 
				mpBufferAlphaSRC, 
				mpBufferAlphaOUT + (indexy + x)); }
	}
    }
    else{
	double dWidth = 1.0 / widthTgt * widthSrc;
	double dHeight = 1.0 / widthTgt * heightSrc;
	double dOffsetY = (1.0 - (double)heightTgt/widthTgt) / 2;
	for (int y = 0; y < heightTgt; y++){
	    double srcY = ((double)y/widthTgt + dOffsetY) * heightSrc;
	    int indexy = y * widthTgt;
	    for (int x = 0; x < widthTgt; x++){
		double srcX = (double)x/widthTgt * heightSrc;
		GetAverageAlpha(widthSrc, heightSrc,
				srcX, srcY, srcX+dWidth, srcY+dHeight, 
				mpBufferAlphaSRC,
				mpBufferAlphaOUT + (indexy + x)); }
	}
    }
}

wxThread::ExitCode CLS::
Entry(){

    int imageAreaSize = mCSizeTgt.GetWidth() * mCSizeTgt.GetHeight();
    
    if (mCSizeTgt == mCSizeSrc){
	mpBufferOUT = mpBufferSRC;
	mpBufferAlphaOUT = mpBufferAlphaSRC;
    }
    else{
	mpBufferOUT = (unsigned char*)malloc(imageAreaSize * 3);
	ResizeImageRGB();
	free(mpBufferSRC);

	if (mpBufferAlphaSRC != NULL){
	    mpBufferAlphaOUT = (unsigned char*)malloc(imageAreaSize);
	    ResizeImageAlpha();
	    free(mpBufferAlphaSRC);
	}
    }

    if (mpBufferAlphaOUT != NULL){
	for (int idx = 0; idx < imageAreaSize; idx++)
	    if (mpBufferAlphaOUT[idx] == 0){
		mpBufferOUT[idx*3+0] = 0;
		mpBufferOUT[idx*3+1] = 0;
		mpBufferOUT[idx*3+2] = 255;
	    }
    }
    
    wxImage CImage(mCSizeTgt, mpBufferOUT); // mpBufferOUT̊Ǘ͈p
    if (mpBufferAlphaOUT != NULL)
	CImage.SetAlpha(mpBufferAlphaOUT);
    
    wxImage CMirror(CImage.Mirror(FALSE)); // ]ftH

    CMirror.SaveFile(mstrFileName);
    
    *mpbFinishedFlagLink = TRUE;
    return 0;
}

// ------

#define CLS ThreadImageSaveScanline_cls

CLS::
ThreadImageSaveScanline_cls(bool *pbFinishedFlag) :
    ThreadImageSave_cls(pbFinishedFlag){
}

void CLS::
AllocateBufferSRC1(){
    mpBufferSRC_1 = (unsigned char*)malloc(mCSizeSrc.GetWidth() * mCSizeSrc.GetHeight() * 3);
}

void CLS::
ResizeImageRGBScanline(){
    
    int widthSrc = mCSizeSrc.GetWidth(), heightSrc = mCSizeSrc.GetHeight();
    int widthTgt = mCSizeTgt.GetWidth(), heightTgt = mCSizeTgt.GetHeight();

    if (widthTgt < heightTgt){
	double dWidth = 1.0 / heightTgt * widthSrc;
	double dHeight = 1.0 / heightTgt * heightSrc;
	double dOffsetX = (1.0 - (double)widthTgt/heightTgt) / 2;
	for (int y = 0; y < heightTgt; y += 2){
	    double srcY = (double)y / heightTgt * heightSrc;
	    int indexy = y * widthTgt;
	    for (int x = 0; x < widthTgt; x++){
		double srcX = ((double)x/heightTgt + dOffsetX) * widthSrc;
		GetAverageRGB(widthSrc, heightSrc,
			      srcX, srcY, srcX+dWidth, srcY+dHeight, 
			      mpBufferSRC, mpBufferOUT + (indexy + x) * 3); }
	}
	for (int y = 1; y < heightTgt; y += 2){
	    double srcY = (double)y / heightTgt * heightSrc;
	    int indexy = y * widthTgt;
	    for (int x = 0; x < widthTgt; x++){
		double srcX = ((double)x/heightTgt + dOffsetX) * widthSrc;
		GetAverageRGB(widthSrc, heightSrc,
			      srcX, srcY, srcX+dWidth, srcY+dHeight, 
			      mpBufferSRC_1, mpBufferOUT + (indexy + x) * 3); }
	}
    }
    else{
	double dWidth = 1.0 / widthTgt * widthSrc;
	double dHeight = 1.0 / widthTgt * heightSrc;
	double dOffsetY = (1.0 - (double)heightTgt/widthTgt) / 2;
	for (int y = 0; y < heightTgt; y+= 2){
	    double srcY = ((double)y/widthTgt + dOffsetY) * heightSrc;
	    int indexy = y * widthTgt;
	    for (int x = 0; x < widthTgt; x++){
		double srcX = (double)x/widthTgt * heightSrc;
		GetAverageRGB(widthSrc, heightSrc,
			      srcX, srcY, srcX+dWidth, srcY+dHeight, 
			      mpBufferSRC, mpBufferOUT + (indexy + x) * 3); }
	}
	for (int y = 1; y < heightTgt; y+= 2){
	    double srcY = ((double)y/widthTgt + dOffsetY) * heightSrc;
	    int indexy = y * widthTgt;
	    for (int x = 0; x < widthTgt; x++){
		double srcX = (double)x/widthTgt * heightSrc;
		GetAverageRGB(widthSrc, heightSrc,
			      srcX, srcY, srcX+dWidth, srcY+dHeight, 
			      mpBufferSRC_1, mpBufferOUT + (indexy + x) * 3); }
	}
    }
}

wxThread::ExitCode CLS::
Entry(){
    mpBufferOUT = (unsigned char*)malloc(mCSizeTgt.GetWidth() * mCSizeTgt.GetHeight() * 3);
    ResizeImageRGBScanline();
    free(mpBufferSRC);
    free(mpBufferSRC_1);

    wxImage CImage(mCSizeTgt, mpBufferOUT); // mpBufferOUT̊Ǘ͈p
    wxImage CMirror(CImage.Mirror(FALSE)); // ]ftH

    CMirror.SaveFile(mstrFileName);

    *mpbFinishedFlagLink = TRUE;
    return 0;
}

// ------

#define CLS ImageIO_cls

CLS::
ImageIO_cls(){
    for (int i = 0; i < _IMAGE_OUT_MAX_THREAD; i++){
	mbThreadActive[i] = FALSE;
	mbThreadFinished[i] = FALSE;
	mpCThread[i] = NULL;
	mpCThreadHDRI[i] = NULL;
    }
    mActiveThread = 0;
}

#ifdef _USE_EXR
bool ImageIO_cls::
SaveEXRFromFBO(FBO_cls *pCFBOMain, wxSize CSizeImage,
	       wxString strFileName, bool bAlpha){

    int widthFBO =  pCFBOMain->mWidthFBO;
    int heightFBO = pCFBOMain->mHeightFBO;
    wxSize CSizeFBO(widthFBO, heightFBO);

    float *pFloatImage = NULL;

    bool bDepthBuffer = FALSE;
    int channelSRC = 3;
    if (bAlpha == TRUE) channelSRC = 4;

    if (bDepthBuffer == TRUE){
	pCFBOMain->mpCDepthTexture->BindTexture();
	pFloatImage = new float [heightFBO * widthFBO];
	glGetTexImage( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, GL_FLOAT, pFloatImage );
    }
    else{
	pCFBOMain->BindTexture();
	pFloatImage = new float [heightFBO * widthFBO * channelSRC];
	if (channelSRC == 3)
	    glGetTexImage( GL_TEXTURE_2D, 0, GL_RGB, GL_FLOAT, pFloatImage );
	if (channelSRC == 4)
	    glGetTexImage( GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, pFloatImage );
    }

    // gXbht
    if (mbThreadActive[mActiveThread] == TRUE){
	while (mbThreadFinished[mActiveThread] == FALSE)
	    wxMilliSleep(3);
	mbThreadActive[mActiveThread] = FALSE;
	mbThreadFinished[mActiveThread] = FALSE;
    }
    
    if (bDepthBuffer == TRUE){
	mpCThreadHDRI[mActiveThread] = 
	    new ThreadHDRISave_cls(&mbThreadFinished[mActiveThread]);
	mpCThreadHDRI[mActiveThread]->
	    SetParameters(pFloatImage, CSizeFBO, CSizeImage,
			  strFileName, 1, 
			  _HDRI_HALF_FLOAT, _HDRI_ZIP,  
			  mProjMatrix14, mProjMatrix10);
    }
    else{
	mpCThreadHDRI[mActiveThread] =
	    new ThreadHDRISave_cls(&mbThreadFinished[mActiveThread]);
	
	mpCThreadHDRI[mActiveThread]->
	    SetParameters(pFloatImage, CSizeFBO, CSizeImage,
			  strFileName, channelSRC, 
			  _HDRI_HALF_FLOAT, _HDRI_ZIP, 
			  mProjMatrix14, mProjMatrix10);
    }

    mbThreadActive[mActiveThread] = TRUE; 
    mpCThreadHDRI[mActiveThread]->Run();

    mActiveThread++;
    if (mActiveThread >= _IMAGE_OUT_MAX_THREAD) mActiveThread = 0;
    
    return TRUE;
}
#endif

void CLS::
PrepareViewportInfo(){
    double op[16]; // original projection matrix
    glGetDoublev(GL_PROJECTION_MATRIX, op);

    mProjMatrix10 = op[10];
    mProjMatrix14 = op[14];
}

bool CLS::
SaveImageFromFBO(FBO_cls *pCFBOMain, wxSize CSizeImage,
		 wxString strFileName, bool bAlpha){

    wxFileName CFile(strFileName);
    if (wxDirExists(CFile.GetPath()) == FALSE) return FALSE;

#ifdef _USE_EXR
    if (CFile.GetExt().IsSameAs(wxT("exr")) ||
	CFile.GetExt().IsSameAs(wxT("EXR"))){
	return SaveEXRFromFBO(pCFBOMain, CSizeImage,
			      strFileName, bAlpha);
    }
#endif

    if (mbThreadActive[mActiveThread] == TRUE){
	while (mbThreadFinished[mActiveThread] == FALSE)
	    wxMilliSleep(3);
	mbThreadActive[mActiveThread] = FALSE;
	mbThreadFinished[mActiveThread] = FALSE;
    }

    int widthFBO = pCFBOMain->mWidthFBO;
    int heightFBO = pCFBOMain->mHeightFBO;
    int texID = pCFBOMain->mTextureID;

    mpCThread[mActiveThread] = 
	new ThreadImageSave_cls(&mbThreadFinished[mActiveThread]);
    mpCThread[mActiveThread]->
	SetParameters(wxSize(widthFBO, heightFBO), CSizeImage, 
		      strFileName, bAlpha);
    
    unsigned char *pBuffer      = mpCThread[mActiveThread]->mpBufferSRC;
    unsigned char *pBufferAlpha = mpCThread[mActiveThread]->mpBufferAlphaSRC;
	
    glBindTexture(GL_TEXTURE_2D, texID);

    if (bAlpha == TRUE){
	if (pBuffer != NULL && pBufferAlpha != NULL){
	    int size1ch = widthFBO * heightFBO;
	    unsigned char *pBuffer4ch = new unsigned char [size1ch * 4];
	    glGetTexImage( GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, pBuffer4ch);
	    for (int i = 0; i < size1ch; i++){
		pBuffer[i*3 + 0] = pBuffer4ch[i*4 + 0];
		pBuffer[i*3 + 1] = pBuffer4ch[i*4 + 1];
		pBuffer[i*3 + 2] = pBuffer4ch[i*4 + 2];
		pBufferAlpha[i]  = pBuffer4ch[i*4 + 3];
	    }
	    delete pBuffer4ch;
	}
	else
	    printf("*pBuffer == NULL\n");
    }
    else{
	if (pBuffer != NULL){
	    glGetTexImage( GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, pBuffer);
	}
	else
	    printf("*pBuffer == NULL\n");
    }
    
    mbThreadActive[mActiveThread]   = TRUE;
    mbThreadFinished[mActiveThread] = FALSE;
    mpCThread[mActiveThread]->Run();

    mActiveThread++;
    if (mActiveThread >= _IMAGE_OUT_MAX_THREAD) mActiveThread = 0;

    return TRUE;
}

bool CLS::
SaveImageFromFBOScanLine(FBO_cls *pCFBOMain[2], wxSize CSizeImage,
			 wxString strFileName){
    
    if (mbThreadActive[mActiveThread] == TRUE){
	while (mbThreadFinished[mActiveThread] == FALSE)
	    wxMilliSleep(3);
	mbThreadActive[mActiveThread] = FALSE;
	mbThreadFinished[mActiveThread] = FALSE;
    }

    wxFileName CFile(strFileName);
    if (wxDirExists(CFile.GetPath()) == FALSE) return FALSE;

    int widthFBO = pCFBOMain[0]->mWidthFBO;
    int heightFBO = pCFBOMain[0]->mHeightFBO;
    int texID_0 = pCFBOMain[0]->mTextureID;
    int texID_1 = pCFBOMain[1]->mTextureID;

    mpCThread[mActiveThread] = 
	new ThreadImageSaveScanline_cls(&mbThreadFinished[mActiveThread]);
    mpCThread[mActiveThread]->
	SetParameters(wxSize(widthFBO, heightFBO), CSizeImage, strFileName, FALSE);
    ((ThreadImageSaveScanline_cls*)mpCThread[mActiveThread])->AllocateBufferSRC1();
    
    unsigned char *pBuffer = mpCThread[mActiveThread]->mpBufferSRC;

    bool bDirectRGB = TRUE;
    
    glBindTexture(GL_TEXTURE_2D, texID_0);
    if (bDirectRGB == TRUE){
	if (pBuffer != NULL){
	    glGetTexImage( GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, pBuffer);
	}
	if (pBuffer == NULL)
	    printf("*pBuffer == NULL\n");
    }

    unsigned char *pBuffer_1 = ((ThreadImageSaveScanline_cls*)mpCThread[mActiveThread])->mpBufferSRC_1;
    glBindTexture(GL_TEXTURE_2D, texID_1);
    if (bDirectRGB == TRUE){
	if (pBuffer_1 != NULL){
	    glGetTexImage( GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, pBuffer_1);
	}
	if (pBuffer_1 == NULL)
	    printf("*pBuffer == NULL\n");
    }
    
    mbThreadActive[mActiveThread]   = TRUE;
    mbThreadFinished[mActiveThread] = FALSE;
    ((ThreadImageSaveScanline_cls*)mpCThread[mActiveThread])->Run();

    mActiveThread++;
    if (mActiveThread >= _IMAGE_OUT_MAX_THREAD) mActiveThread = 0;

    return TRUE;
}


ImageIO_cls *GpCImageIO = NULL;
