// -*- c -*-
// "clsMouseCtrl.h"

#include "clsMouseCtrl.h"
#include "../PREFERENCE/clsMousePreference.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#define CLS MouseCtrl_cls

void CLS::
SetMouseInformation(wxMouseEvent &event, bool bModulate){
    
    mbAModifier[0] = event.m_shiftDown;
    mbAModifier[1] = event.m_controlDown;
    mbAModifier[2] = event.m_altDown;
    mbAModifier[3] = FALSE;
    mbAMouseBTN[0] = event.LeftIsDown() || event.LeftDClick();
    mbAMouseBTN[1] = event.MiddleIsDown() || event.MiddleDClick();
    mbAMouseBTN[2] = event.RightIsDown() || event.RightDClick();
    mbAMouseBTN[3] = FALSE;

    mMouseX = event.GetX();
    mMouseY = event.GetY();

    mbButtonDown = event.ButtonDown();
    mbButtonDown |= event.LeftDClick();
    mbButtonDown |= event.MiddleDClick();
    mbButtonDown |= event.RightDClick();

    mbButtonUp   = event.ButtonUp();
    

    mbMouseWheel = (event.GetEventType() == wxEVT_MOUSEWHEEL);
    mbMouseWheel &= (event.GetWheelAxis() == wxMOUSE_WHEEL_VERTICAL);
    if (mbMouseWheel){
	int wheelRotaion = event.GetWheelRotation();
	int wheelDelta = event.GetWheelDelta();
	if (wheelRotaion == wheelDelta) mWheelRotationNormalized = 1;
	if (wheelRotaion == -wheelDelta) mWheelRotationNormalized = -1;
	/*
	if (wheelRotaion > 1)
	    mWheelRotationNormalized =
		(float)(wheelRotaion)/ event.GetWheelDelta();
	if (wheelRotaion < -1)
	    mWheelRotationNormalized =
		- ((float)(-wheelRotaion)/ event.GetWheelDelta());
	*/
    }
    	    
    mbMouseDragging = event.Dragging() & 
	((mPrevX != mMouseX) | (mPrevY != mMouseY));

    if (mbButtonDown == TRUE){
	mMouseAtDownX = mMouseX;
	mMouseAtDownY = mMouseY;
    }
    
    if (bModulate == TRUE)
	GetPCMousePreference()->ModulateBTN(mbAModifier, mbAMouseBTN);
}

void CLS::
SetPreviousPosition(){
    mPrevX = mMouseX;
    mPrevY = mMouseY;
}

void MouseCtrl_cls::
SetGyrationMode(wxSize sz){
    float lX = (float)(2.0 * mMouseX - sz.x)     / sz.x;
    float lY = (float)(sz.y - 2.0 * mMouseY)     / sz.y;
    if (lX > 0.85 || lX < -0.85 || lY > 0.85 || lY < -0.85)
	mGyrationMode = _GYRATION_FIXPOS;
    else
	mGyrationMode = _GYRATION_FREE;
}

bool CLS::
IsGyration(){
    return GetPCMousePreference()->
	IsMouse4Gyration(mbAModifier, mbAMouseBTN); }

bool CLS::
IsZoom(){
    return GetPCMousePreference()->
	IsMouse4Zoom(mbAModifier, mbAMouseBTN); }

bool CLS::
IsWheel(){
    return mbMouseWheel;
}

bool CLS::
IsMoveCenter(){
    return GetPCMousePreference()->
	IsMouse4MoveCenter(mbAModifier, mbAMouseBTN); }

bool CLS::
IsMoveItem(){
    return GetPCMousePreference()->
	IsMouse4MoveItem(mbAModifier, mbAMouseBTN); }

bool CLS::
IsMoveSnap(){
    return GetPCMousePreference()->
	IsMouse4MoveItemSnap(mbAModifier, mbAMouseBTN); }

bool CLS::
IsInertia(){
    return GetPCMousePreference()->
	IsMouse4Inertia(mbAModifier, mbAMouseBTN); }

bool CLS::
IsShowMenu(){
    if (mbButtonDown == FALSE) 
	return FALSE;
    
    if (GCClassProvider.GetPCMousePreference()->IsMouse4Menu(mbAModifier, mbAMouseBTN) == FALSE)
	return FALSE;

    
    /*
    if (mbAMouseBTN[2] == FALSE) return FALSE;
    */
    
    return TRUE;
}

bool MouseCtrl_cls::
IsHitTest(){
    return GCClassProvider.GetPCMousePreference()->
	IsMouse4HitTest(mbAModifier, mbAMouseBTN) && IsButtonDown();
}

bool MouseCtrl_cls::
IsHitTestAdditive(){
    return GCClassProvider.GetPCMousePreference()->
	IsMouse4HitTestAdditive(mbAModifier, mbAMouseBTN) && IsButtonDown();
}

bool MouseCtrl_cls::
IsButton(int btnType){
    return mbAMouseBTN[btnType];
}

bool MouseCtrl_cls::
IsButtonDown(){
    return mbButtonDown;
}

bool MouseCtrl_cls::
IsButtonUp(){
    return mbButtonUp;
}

bool MouseCtrl_cls::
IsMouseDragging(){
    return mbMouseDragging;
}

// -----------------------------

MouseNormalized_cls::
MouseNormalized_cls(MouseCtrl_cls *pCMouseCtrl, wxSize sz){
    mPrevX = (float)(2.0 * pCMouseCtrl->mPrevX - sz.x)  / sz.x;
    mPrevY = (float)(sz.y - 2.0 * pCMouseCtrl->mPrevY)  / sz.y;
    mCurrX = (float)(2.0 * pCMouseCtrl->mMouseX - sz.x) / sz.x;
    mCurrY = (float)(sz.y - 2.0 * pCMouseCtrl->mMouseY) / sz.y;
    mDownX = (float)(2.0 * pCMouseCtrl->mMouseAtDownX - sz.x) / sz.x;
    mDownY = (float)(sz.y - 2.0 * pCMouseCtrl->mMouseAtDownY) / sz.y;
    mGyrationMode = pCMouseCtrl->mGyrationMode;
}

void MouseNormalized_cls::
LimitMove(){
    if (mPrevX < mCurrX - 0.2) mPrevX = mCurrX - 0.2;
    if (mPrevX > mCurrX + 0.2) mPrevX = mCurrX + 0.2;
    if (mPrevY < mCurrY - 0.2) mPrevY = mCurrY - 0.2;
    if (mPrevY > mCurrY + 0.2) mPrevY = mCurrY + 0.2;
}

void MouseNormalized_cls::
Invert(){
    mPrevX *= -1; mCurrX *= -1;
    mPrevY *= -1; mCurrY *= -1;
}

double MouseNormalized_cls::
GetDX(){ return mCurrX - mPrevX; }

double MouseNormalized_cls::
GetDY(){ return mCurrY - mPrevY; }
