// -*- c -*-
// "clsMeshBase.h"

#include "clsMeshBase.h"

#include "../GLOBJ/clsVBOBaseObject.h"
#include "../COMMON/clsArrayAndIndex.h"

#include <algorithm>

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
#endif

MeshBase_cls::
MeshBase_cls(CallBackMaterial_cls *pCCallBackParent){
    mpCCallBackParent = pCCallBackParent;
    mbFullyTriangular = FALSE;
}

void MeshBase_cls::
ClearAll(){
    std::vector < float >().swap(mVVertex);
    std::vector < float >().swap(mVNormal);
    std::vector < float >().swap(mVUV2);
    std::vector < unsigned int >().swap(mVFaceIndexVertex);
    std::vector < unsigned int >().swap(mVFaceIndexNormal);
    std::vector < unsigned int >().swap(mVFaceIndexUV2);
    std::vector < int >().swap(mVMaterialAndPolygon);

    std::vector < unsigned int >().swap(mVFaceIndexVertexAdj);
    mbFullyTriangular = FALSE;
}

void MeshBase_cls::
MinimizeMemoryUse(){
    std::vector < float >(mVVertex).swap(mVVertex);
    std::vector < float >(mVNormal).swap(mVNormal);
    std::vector < float >(mVUV2).swap(mVUV2);
    std::vector < unsigned int >(mVFaceIndexVertex).swap(mVFaceIndexVertex);
    std::vector < unsigned int >(mVFaceIndexNormal).swap(mVFaceIndexNormal);
    std::vector < unsigned int >(mVFaceIndexUV2).swap(mVFaceIndexUV2);
    std::vector < int >(mVMaterialAndPolygon).swap(mVMaterialAndPolygon);

    std::vector < unsigned int >(mVFaceIndexVertexAdj).swap(mVFaceIndexVertexAdj);
}

bool MeshBase_cls::
IsFullyTriangular(){
    int faceIndexOffset = 0;
    int size = (int)mVMaterialAndPolygon.size();
    for (int i = 0; i < size; i++){
	int ngon = mVMaterialAndPolygon[i];
	if (ngon == _MESH_MATERIAL) {
	    i++; 
	    continue;
	}
	if (ngon == _MESH_POLYGONS_ARRAY) {
	    i++; int polyType = mVMaterialAndPolygon[i];
	    if (polyType == GL_QUADS) return FALSE;
	    i++; 
	    continue;
	}

	if (ngon != 3) return FALSE;
    }
    return TRUE;
}

void MeshBase_cls::
ReserveVertexNumber(int num){
    mVVertex.reserve(num*3);
    mVNormal.reserve(num*3);
    mVUV2.reserve(num*2);
}

void MeshBase_cls::
ReserveFaceNumber(int num){
    mVFaceIndexVertex.reserve(num*4);
    mVFaceIndexNormal.reserve(num*4);
    mVFaceIndexUV2.reserve(num*4);
}

void MeshBase_cls::
SetVertexNumber(int num){
    mVVertex.resize(num*3);
    mVNormal.resize(num*3);
    mVUV2.resize(num*2);
}

void MeshBase_cls::
SetFaceNumber(int num){
    mVFaceIndexVertex.resize(num*4);
    mVFaceIndexNormal.resize(num*4);
    mVFaceIndexUV2.resize(num*4);
}

int MeshBase_cls::
SearchAdjTriangle(int ibase, int vertex0, int vertex1){
    int size = mVFaceIndexVertex.size();
    int ini = ibase - 9996;
    int fin = ibase + 9999;
    if (ini < 0) ini = 0;
    if (fin > size) fin = size;

    for (int i = ini; i < fin; i+=3){
	int vertexA = mVFaceIndexVertex[i];
	int vertexB = mVFaceIndexVertex[i+1];
	int vertexC = mVFaceIndexVertex[i+2];
	if (vertexB == vertex0 && vertexA == vertex1) return vertexC;
	if (vertexC == vertex0 && vertexB == vertex1) return vertexA;
	if (vertexA == vertex0 && vertexC == vertex1) return vertexB;
    }
    return -1;
}

int MeshBase_cls::
SearchAdjTriangleB(int ibase_, ArrayAndIndex_cls *pCIdxAndZ, 
		   int vertex0, int vertex1){
    int triangleNum = mVFaceIndexVertex.size() / 3;
    int ini = ibase_ - 999;
    int fin = ibase_ + 999;
    if (ini < 0) ini = 0;
    if (fin > triangleNum) fin = triangleNum;
    
    for (int i_ = ini; i_ < fin; i_++){
	int i = pCIdxAndZ[i_].index;
	int vertexA = mVFaceIndexVertex[i*3];
	int vertexB = mVFaceIndexVertex[i*3+1];
	int vertexC = mVFaceIndexVertex[i*3+2];
	if (vertexB == vertex0 && vertexA == vertex1) return vertexC;
	if (vertexC == vertex0 && vertexB == vertex1) return vertexA;
	if (vertexA == vertex0 && vertexC == vertex1) return vertexB;
    }
    return -1;
}

void MeshBase_cls::
ConstructAdjStructure(){
    
    mVFaceIndexVertexAdj.clear();

    // traiangle only ł邱Ƃ肷
    mbFullyTriangular = IsFullyTriangular();
    if (mbFullyTriangular == FALSE) return;

    int size = mVFaceIndexVertex.size();
    int triangleNum = size / 3;

    ArrayAndIndex_cls *pCIdxAndZ = new ArrayAndIndex_cls [triangleNum];
    
    for (int i = 0; i < triangleNum; i++){
	pCIdxAndZ[i].index = i;
	// ŏ̒_zWg
	float z0 = mVVertex[mVFaceIndexVertex[i*3]*3+2];
	float z1 = mVVertex[mVFaceIndexVertex[i*3+1]*3+2];
	float z2 = mVVertex[mVFaceIndexVertex[i*3+2]*3+2];
	pCIdxAndZ[i].z = (z0 + z1 + z2) / 3;
    }
    
    std::sort(pCIdxAndZ, pCIdxAndZ+triangleNum);

    for (int i_ = 0; i_ < triangleNum; i_++){
	int i = pCIdxAndZ[i_].index;
	int vertexA = mVFaceIndexVertex[i*3];
	int vertexB = mVFaceIndexVertex[i*3+1];
	int vertexC = mVFaceIndexVertex[i*3+2];
	int vertexABdash = SearchAdjTriangleB(i_, pCIdxAndZ, vertexA, vertexB);
	int vertexBCdash = SearchAdjTriangleB(i_, pCIdxAndZ, vertexB, vertexC);
	int vertexCAdash = SearchAdjTriangleB(i_, pCIdxAndZ, vertexC, vertexA);
	if (vertexABdash == -1) vertexABdash = vertexC;
	if (vertexBCdash == -1) vertexBCdash = vertexA;
	if (vertexCAdash == -1) vertexCAdash = vertexB;
	mVFaceIndexVertexAdj.push_back(vertexA);
	mVFaceIndexVertexAdj.push_back(vertexABdash);
	mVFaceIndexVertexAdj.push_back(vertexB);
	mVFaceIndexVertexAdj.push_back(vertexBCdash);
	mVFaceIndexVertexAdj.push_back(vertexC);
	mVFaceIndexVertexAdj.push_back(vertexCAdash);
    }

    /*
    for (int i = 0; i < size; i+=3){
	int vertexA = mVFaceIndexVertex[i];
	int vertexB = mVFaceIndexVertex[i+1];
	int vertexC = mVFaceIndexVertex[i+2];
	int vertexABdash = SearchAdjTriangle(i, vertexA, vertexB);
	int vertexBCdash = SearchAdjTriangle(i, vertexB, vertexC);
	int vertexCAdash = SearchAdjTriangle(i, vertexC, vertexA);
	if (vertexABdash == -1) vertexABdash = vertexC;
	if (vertexBCdash == -1) vertexBCdash = vertexA;
	if (vertexCAdash == -1) vertexCAdash = vertexB;
	mVFaceIndexVertexAdj.push_back(vertexA);
	mVFaceIndexVertexAdj.push_back(vertexABdash);
	mVFaceIndexVertexAdj.push_back(vertexB);
	mVFaceIndexVertexAdj.push_back(vertexBCdash);
	mVFaceIndexVertexAdj.push_back(vertexC);
	mVFaceIndexVertexAdj.push_back(vertexCAdash);
    }
    */

    delete [] pCIdxAndZ;
}

void MeshBase_cls::
GLDrawUseNormalUseTexture(){
    std::vector < unsigned int > &VFaceIndexNormal = mVFaceIndexNormal;
    std::vector < unsigned int > &VFaceIndexUV2 = mVFaceIndexUV2;

    if (VFaceIndexNormal.size() == 0) VFaceIndexNormal = mVFaceIndexVertex;
    if (VFaceIndexUV2.size() == 0) VFaceIndexUV2 = mVFaceIndexVertex;

    int faceIndexOffset = 0;
    int size = (int)mVMaterialAndPolygon.size();

    for (int i = 0; i < size; i++){
	int ngon = mVMaterialAndPolygon[i];
	if (ngon == _MESH_MATERIAL) {
	    i++; int materialType = mVMaterialAndPolygon[i];
	    mpCCallBackParent->CallMaterial(materialType);
	    continue;
	}
	if (ngon == _MESH_POLYGONS_ARRAY) {
	    i++; int polyType = mVMaterialAndPolygon[i];
	    i++; int faceNum = mVMaterialAndPolygon[i];
	    GLDrawArrayMesh(polyType, faceNum, faceIndexOffset, TRUE, TRUE);
	    continue;
	}
	
	switch(ngon){
	case 3: glBegin(GL_TRIANGLES); break;
	case 4: glBegin(GL_QUADS); break;
	default: glBegin(GL_POLYGON); break;
	}
	for (int i = 0; i < ngon; i++){
	    int indexVertex = mVFaceIndexVertex[faceIndexOffset + i];
	    int indexNormal = VFaceIndexNormal[faceIndexOffset + i];
	    int indexUV = VFaceIndexUV2[faceIndexOffset + i];
	    glTexCoord2fv(&mVUV2[indexUV*2]);
	    glNormal3fv(&mVNormal[indexNormal*3]);
	    glVertex3fv(&mVVertex[indexVertex*3]);
	}
	glEnd();

	faceIndexOffset += ngon;
    }
}

void MeshBase_cls::
GLDrawUseNormal(){
    std::vector < unsigned int > &VFaceIndexNormal = mVFaceIndexNormal;

    if (VFaceIndexNormal.size() == 0) VFaceIndexNormal = mVFaceIndexVertex;

    int faceIndexOffset = 0;
    int size = (int)mVMaterialAndPolygon.size();
    for (int i = 0; i < size; i++){
	int ngon = mVMaterialAndPolygon[i];
	if (ngon == _MESH_MATERIAL) {
	    i++; int materialType = mVMaterialAndPolygon[i];
	    mpCCallBackParent->CallMaterial(materialType);
	    continue;
	}
	if (ngon == _MESH_POLYGONS_ARRAY) {
	    i++; int polyType = mVMaterialAndPolygon[i];
	    i++; int faceNum = mVMaterialAndPolygon[i];
	    GLDrawArrayMesh(polyType, faceNum, faceIndexOffset, TRUE, FALSE);
	    continue;
	}

	switch(ngon){
	case 3: glBegin(GL_TRIANGLES); break;
	case 4: glBegin(GL_QUADS); break;
	default: glBegin(GL_POLYGON); break;
	}
	for (int i = 0; i < ngon; i++){
	    int indexVertex = mVFaceIndexVertex[faceIndexOffset + i];
	    int indexNormal = VFaceIndexNormal[faceIndexOffset + i];
	    glNormal3fv(&mVNormal[indexNormal*3]);
	    glVertex3fv(&mVVertex[indexVertex*3]);
	}
	glEnd();

	faceIndexOffset += ngon;
    }
}

void MeshBase_cls::
GLDraw(bool bUseNornal, bool bUseUV){

    if (bUseNornal == TRUE && bUseUV == TRUE){
	GLDrawUseNormalUseTexture();
	return;
    }

    if (bUseNornal == TRUE && bUseUV == FALSE){
	GLDrawUseNormal();
	return;
    }

    int faceIndexOffset = 0;
    int size = (int)mVMaterialAndPolygon.size();
    for (int i = 0; i < size; i++){
	int ngon = mVMaterialAndPolygon[i];
	if (ngon == _MESH_MATERIAL) {
	    i++; int materialType = mVMaterialAndPolygon[i];
	    mpCCallBackParent->CallMaterial(materialType);
	    continue;
	}
	if (ngon == _MESH_POLYGONS_ARRAY) {
	    i++; int polyType = mVMaterialAndPolygon[i];
	    i++; int faceNum = mVMaterialAndPolygon[i];
	    GLDrawArrayMesh(polyType, faceNum, faceIndexOffset,
			    bUseNornal, bUseUV);
	    continue;
	}

	switch(ngon){
	case 3: glBegin(GL_TRIANGLES); break;
	case 4: glBegin(GL_QUADS); break;
	default: glBegin(GL_POLYGON); break;
	}
	for (int i = 0; i < ngon; i++){
	    int index = mVFaceIndexVertex[faceIndexOffset + i];
	    glNormal3fv(&mVNormal[index*3]);
	}
	glEnd();

	faceIndexOffset += ngon;
    }
}

void MeshBase_cls::
GLDrawArrayMesh(int polyType, int faceNum, int &faceIndexOffset,
		bool bUseNornal, bool bUseUV){
    int poly = 3;
    if (polyType == GL_QUADS) poly = 4;

    glEnableClientState(GL_VERTEX_ARRAY);
    if (bUseNornal == TRUE) glEnableClientState(GL_NORMAL_ARRAY);
    else                    glDisableClientState(GL_NORMAL_ARRAY);
    if (bUseUV == TRUE) glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    else                glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    
    glVertexPointer(3, GL_FLOAT, 0, mVVertex.data());
    if (bUseNornal == TRUE) glNormalPointer(GL_FLOAT, 0, mVNormal.data());
    if (bUseUV == TRUE)     glTexCoordPointer(2, GL_FLOAT, 0, mVUV2.data());

    glDisableClientState(GL_COLOR_ARRAY);
    
    
    glDrawElements(polyType, faceNum * poly, GL_UNSIGNED_INT,
		   mVFaceIndexVertex.data() + faceIndexOffset);
    
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);

    faceIndexOffset += faceNum * poly;
}

void MeshBase_cls::
SetVBOShadowGeometry(VBOBaseObject_cls *pCTgt){

    pCTgt->Clear();
    pCTgt->GenBuffersVertex(1);

    int totalVertex = mVVertex.size();
    glBindBuffer(GL_ARRAY_BUFFER, pCTgt->mVBOVertex);
    glBufferData(GL_ARRAY_BUFFER, totalVertex * sizeof(float),
		    mVVertex.data(), GL_STATIC_DRAW);

    //int totalIndices = mVFaceIndexVertexAdj.size() / 6;
    //pCTgt->mTotalIndiceTriangle = 0; // ʏ핔͎͕`Fêݎg
    pCTgt->mTotalIndiceTriangle = mVFaceIndexVertex.size();
    pCTgt->mTotalIndiceAdjacent = mVFaceIndexVertexAdj.size();

    /*
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, pCTgt->mVBOIndicesAdjacent);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, pCTgt->mTotalIndiceAdjacent * sizeof(unsigned int), mVFaceIndexVertexAdj.data(), GL_STATIC_DRAW);
    */

    pCTgt->GenBuffersIndice();
    pCTgt->MakeVBOIndicesTriangle(mVFaceIndexVertex.data());
    pCTgt->MakeVBOIndicesAdjacent(mVFaceIndexVertexAdj.data());
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
