// (setq c-basic-offset 4)
// "clsLuaInterpreter.cpp"

#ifndef _LUA_INTERPRETER_H_
#define _LUA_INTERPRETER_H_

#include <wx/wx.h>
#include <vector>

extern "C" {
#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"
}

class Main_cls;
class SceneObj_cls;
class SceneObjCtrl_cls;
class UndoManager_cls;

class LuaInterpreter_cls {
 public:
    LuaInterpreter_cls();
    
    std::vector < luaL_Reg > mVLibTableOps;
    std::vector < luaL_Reg > mVLibTableOpsActiveObj;
    std::vector < luaL_Reg > mVLibTableOpsSelectedObj;
    std::vector < luaL_Reg > mVLibTableData;
    std::vector < luaL_Reg > mVLibTableParamGroup;
    std::vector < luaL_Reg > mVLibTableObject;
    std::vector < luaL_Reg > mVLibTableThis;
    
    static SceneObjCtrl_cls* GetPCSceneObjCtrl();
    static SceneObj_cls* GetPCSceneObjFromFirstArg(lua_State *L);
    static UndoManager_cls* GetPCUndo();
    static Main_cls* GetPCMain();
    
    static int ReturnInteger(lua_State *L, int num);
    static int ReturnFalse(lua_State *L);
    static int ReturnTrue(lua_State *L);
    
    static bool GetBool(wxString strArg);

    static void AddSceneObjBase(lua_State *L, wxString strBaseName);
    static int AddPrimitive(lua_State *L);
    static int AddCamera(lua_State *L);
    static int AddLight(lua_State *L);
    static int AddMesh(lua_State *L);
    static int AddEmpty(lua_State *L);
    static int AddWorld(lua_State *L);
    static int AddVolumeBox(lua_State *L);
    static int AddParticle(lua_State *L);
    static int AddStreamline(lua_State *L);
    static int AddHeightField(lua_State *L);
    static int DeleteObject(lua_State *L);
    static int SelectObject(lua_State *L);
    //static int SetVisible(lua_State *L);
    //static int SetValid(lua_State *L);
    static int DeselectObject(lua_State *L);

    static int DeleteActive(lua_State *L);
    static int RenameActive(lua_State *L);
    static int GetVisibilityActive(lua_State *L);
    static int SetVisibilityActive(lua_State *L);
    static int GetValidityActive(lua_State *L);
    static int SetValidityActive(lua_State *L);
    static int GetShyActive(lua_State *L);
    static int SetShyActive(lua_State *L);
    static int GetParameterLinkActive(lua_State *L);
    static int SetParameterLinkActive(lua_State *L);
    
    static int ObjectsMoveTo(lua_State *L);
    static int ObjectsSetChildTo(lua_State *L);
    static int DeleteSelected(lua_State *L);
    static int SetVisibilitySelected(lua_State *L);
    static int SetValiditySelected(lua_State *L);

    static int GetName(lua_State *L);
    static int GetActiveObject(lua_State *L);
    static int GetParent(lua_State *L);
    static int GetChild(lua_State *L);
    static int GetChildList(lua_State *L);
    static int GetVisibility(lua_State *L);
    static int SetVisibility(lua_State *L);
    static int GetValidity(lua_State *L);
    static int SetValidity(lua_State *L);
    static int GetPosition(lua_State *L);
    static int GetRotationEuler(lua_State *L);
    static int GetScale(lua_State *L);
    static int GetParameterValue(lua_State *L, 
				 wxString strParamGroup = wxEmptyString);
    
    static int GetDimension(lua_State *L);
    static int GetValueAt(lua_State *L);
    static int SetValueAt(lua_State *L);
    static int GetLink(lua_State *L);
    static int SetLink(lua_State *L);
    static int GetParamGroupShy(lua_State *L);
    static int SetParamGroupShy(lua_State *L);
    static int GetUseExpression(lua_State *L);
    static int SetUseExpression(lua_State *L);
    static int GetExpression(lua_State *L);
    static int SetExpression(lua_State *L);
    static int GetTimeDependency(lua_State *L);
    static int SetTimeDependency(lua_State *L);

    static int GetObjectShy(lua_State *L);
    static int SetObjectShy(lua_State *L);
    static int GetObjectConfList(lua_State *L);
    static int GetObjectConfItem(lua_State *L);
    static int GetObjectAttribList(lua_State *L);
    static int GetObjectAttribConfList(lua_State *L);
    static int GetObjectAttribConfItem(lua_State *L);

    static int ThisGetValueAt(lua_State *L);
    static int ThisGetKeyNum(lua_State *L);
};

#endif
