//(setq c-basic-offset 4)
// "clsPropertyFormat.h"

#include "clsPropertyFormat.h"
#include "../clsProperty.h"

#include "../../SCENEOBJ/clsMeshObj.h"
#include "../../SCENEOBJ/clsSceneObj.h"
#include "../../SCENEOBJ/clsSceneObjCtrl.h"
#include "../../SCENEOBJ/DATAVIEWER/clsParticle.h"
#include "../../SCENEOBJ/DATAVIEWER/clsStreamline.h"

#include "../../SIMDATA/clsParticleFormat.h"
#include "../../SIMDATA/clsStreamlineFormat.h"

#include "../../WIDGETCTRL/clsBitmapPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"

#include "../../../clsMain.h"

#include "../../COMMON/CommonDefinitions.h"

#include "../../../wxWidgetEnumMenu.h"

#define CLS PropertyFormat_cls

#define EVT_BUTTON_RANGE(id1, id2, func) wx__DECLARE_EVT2(wxEVT_COMMAND_BUTTON_CLICKED, id1, id2, wxCommandEventHandler(func))

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
EVT_BUTTON(btnGraphPin, CLS::OnPinMode)
EVT_BUTTON(btnDropdown, CLS::OnDropDown)
EVT_MENU_RANGE(menuPop0, menuPop9, CLS::OnPopMenu)
EVT_BUTTON_RANGE(btnFormatHeaderDelete0, btnFormatHeaderDelete9, CLS::OnDeleteHeader)
EVT_BUTTON_RANGE(btnFormatHeaderAdd0, btnFormatHeaderAdd9, CLS::OnAddHeader)
EVT_BUTTON_RANGE(btnFormatHeaderUp0, btnFormatHeaderUp9, CLS::OnUpHeader)
EVT_BUTTON_RANGE(btnFormatHeaderDown0, btnFormatHeaderDown9, CLS::OnDownHeader)
EVT_BUTTON_RANGE(btnFormatBodyDelete0, btnFormatBodyDelete9, CLS::OnDeleteBody)
EVT_BUTTON_RANGE(btnFormatBodyAdd0, btnFormatBodyAdd9, CLS::OnAddBody)
EVT_BUTTON_RANGE(btnFormatBodyUp0, btnFormatBodyUp9, CLS::OnUpBody)
EVT_BUTTON_RANGE(btnFormatBodyDown0, btnFormatBodyDown9, CLS::OnDownBody)
END_EVENT_TABLE()

PropertyFormat_cls::
PropertyFormat_cls(Property_cls *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxPanelEX(pCParent, id, pos, size, style), PropertyBase_cls(pCParent)
{
    SetPanelType(_PANEL_PROPERTY_FORMAT);

    SetToDefaultParam();

    mpCBoxSizerSrc  = new wxBoxSizer(wxVERTICAL);

    mpCPain = GetPCNewPain(this);

    //mpCTextFormat = new wxStaticText(this, wxID_ANY, wxT("Format"));

    wxBoxSizer *pBoxSizerInfo = GetNewPCBoxSizerInfoWithItems(this);

    mpCBoxSizerSrc->Add(pBoxSizerInfo, 0, wxEXPAND);
    mpCBoxSizerSrc->Add(mpCPain);
    //mpCBoxSizerSrc->Add(mpCTextFormat, 1, wxEXPAND);

    SetSizerAndFit(mpCBoxSizerSrc);

    wxString strActiveObj = GetPCSceneObjCtrl()->GetActiveObjectName();
    if (strActiveObj.IsEmpty())
	Disconnect();
    else{
	wxString strTag;
	strTag << GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strActiveObj;
	mCstrLinkedObject.SetBodyFromData(strTag);
	mpCLinkedObject->SetLabel(strTag);
	Reconnect();
    }
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();
    mCConfList.SetTag("PropertyFormat");
    mCConfList.Add(&mCstrLinkedObject, "LinkedObject=", "None");
    mCConfList.Add(&mCbPinned, "Pinned=", "0");
}

void CLS::
ClearPanels(){
    if (mpCPain != NULL){
	mpCBoxSizerSrc->Detach(mpCPain);
	uint num = mpCPain->mVpCMNumPanelLink.size();
	for (uint i = 0; i < num; i++)
	    mpCPain->mVpCMNumPanelLink[i]->mbVolatile = TRUE;
	delete mpCPain;
	mpCPain = NULL;
    }
}

void CLS::
Disconnect(){

    SetMinSize(wxSize(-1, -1));
    
    ClearPanels();
    Layout();
}

void CLS::
Reconnect(){
    Disconnect();

    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return ;

    SetMinSize(wxSize(-1, -1));

    wxSize CSize = wxDS;
    if (mpCPain == NULL)
	mpCPain = GetPCNewPain(this);

    wxString strClassName = pCObj->mstrClassName;
    if (strClassName.IsSameAs("Particle_cls")) 
	ConnectParticle((Particle_cls*)pCObj);
    if (strClassName.IsSameAs("Streamline_cls"))
	ConnectStreamline((Streamline_cls*)pCObj);
    if (strClassName.IsSameAs("MeshObj_cls"))
	ConnectMesh((MeshObj_cls*)pCObj);

    wxBoxSizer *pCSizerForPain = new wxBoxSizer(wxVERTICAL);
    mpCPain->InitLayoutNumPanels(pCSizerForPain);
    mpCPain->SetSizer(pCSizerForPain);

    mpCBoxSizerSrc->Add(mpCPain, 0, wxEXPAND);
}

void CLS::
ConnectParticle(Particle_cls* pCObj){
    MultiInputPanel_cls *pCMNumFormatMain = pCObj->GetPCNewPanelFormatMain(mpCPain);
    mpCPain->SetMNum(pCMNumFormatMain, mAstrClosedItem, "Format.Main", TRUE, TRUE);

    MultiInputPanel_cls *pCMNumFormatHeader = pCObj->GetPCNewPanelFormatHeader(mpCPain);
    pCMNumFormatHeader->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_COMBOBOX_SELECTED);
    mpCPain->SetMNum(pCMNumFormatHeader, mAstrClosedItem, "Header", TRUE, TRUE);

    MultiInputPanel_cls *pCMNumFormatBody = pCObj->GetPCNewPanelFormatBody(mpCPain);
    pCMNumFormatBody->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_COMBOBOX_SELECTED);
    mpCPain->SetMNum(pCMNumFormatBody, mAstrClosedItem, "Body", TRUE, TRUE);
    mCurrentDataType = pCObj->mpCParticleFormat->mCDataType.GetBodyINT();

    mpCPain->LayoutNumPanels(FALSE);
    //HideAndShow();
    Layout();
}

void CLS::
ConnectStreamline(Streamline_cls* pCObj){
    MultiInputPanel_cls *pCMNumFormatMain = pCObj->GetPCNewPanelFormatMain(mpCPain);
    mpCPain->SetMNum(pCMNumFormatMain, mAstrClosedItem, "Format.Main", TRUE, TRUE);

    mpCPain->LayoutNumPanels(FALSE);
    Layout();
}

void CLS::
ConnectMesh(MeshObj_cls *pCObj){
    MultiInputPanel_cls *pCMNumFormatMain = pCObj->GetPCNewPanelFormatWavefrontObj(mpCPain);
    mpCPain->SetMNum(pCMNumFormatMain, mAstrClosedItem, "Format.Wavefront", TRUE, TRUE);

    mpCPain->LayoutNumPanels(FALSE);
    Layout();
}

void CLS::
CheckFormat(){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;
    
    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls *pCParticle = (Particle_cls*)pCObj;
	
	if (pCParticle->CheckFormat(mCurrentDataType) == FALSE)
	    GetPCMain()->ProcAfterMath("SelectObject", mCstrLinkedObject.GetBodySTR());
    }
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs("ALL")){
	
	SceneObj_cls *pCObj = GetPCLinkedObj();
	if (pCObj == NULL) return Disconnect();

	CheckFormat();
	
	if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	    ((Particle_cls*)pCObj)->UpdateFormatPanel(mpCPain);
	    ((Particle_cls*)pCObj)->UpdateHeaderPanel(mpCPain);
	    ((Particle_cls*)pCObj)->UpdateBodyPanel(mpCPain);
	}

	/*
	wxString strInfo = "Empty";
	if (pCObj->mstrClassName.IsSameAs("VolumeBox_cls"))
	    strInfo = ((VolumeBox_cls*)pCObj)->GetStrInfo();

	if (mpCTextInfo->GetLabel().IsSameAs(strInfo) == FALSE){
	    Freeze();
	    mpCTextInfo->SetLabel(strInfo);
	    Thaw();
	    CheckLineNumber(strInfo);
	}
	*/
	
	//Disconnect();
    }
}

void CLS::
OnHeaderChange(wxCommandEvent& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;
    
    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls *pCParticle = (Particle_cls*)pCObj;
	pCParticle->UpdateHeaderCommand();
    }
}

void CLS::
OnBodyChange(wxCommandEvent& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;
    
    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls *pCParticle = (Particle_cls*)pCObj;
	pCParticle->UpdateBodyCommand();
    }
}

void CLS::
OnDropDown(wxCommandEvent& event){
    mCSelector.PopupObjectSelector(this, wxES, "ALL");
}


void CLS::
OnPinMode(wxCommandEvent& event){
    long long pinned = event.GetInt();
    mCbPinned.SetBodyFromData(pinned);
}

void CLS::
OnPopMenu(wxCommandEvent& event){
    wxString strPopMenu = mCSelector.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == TRUE) return;
    mpCParent->SelectObject(strPopMenu);
}

void CLS::
OnDeleteHeader(wxCommandEvent& event){
    int index = event.GetId() - btnFormatHeaderDelete0;
    SceneObj_cls* pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls* pCParticle = (Particle_cls*)pCObj;
	pCParticle->DeleteHeaderCommand(index);
    }
    CheckFormat();
}

void CLS::
OnAddHeader(wxCommandEvent& event){
    int index = event.GetId() - btnFormatHeaderAdd0;
    SceneObj_cls* pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls* pCParticle = (Particle_cls*)pCObj;
	pCParticle->AddHeaderCommand(index);
    }
    CheckFormat();
}

void CLS::
OnUpHeader(wxCommandEvent& event){
    int index = event.GetId() - btnFormatHeaderUp0;
    SceneObj_cls* pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls* pCParticle = (Particle_cls*)pCObj;
	pCParticle->UpDownHeaderCommand(index, -1);

	// ACe̕ς炸A\XVmȂ̂Œ
	UpdatePanels("ALL");
    }
    //CheckFormat(); 
}

void CLS::
OnDownHeader(wxCommandEvent& event){
    int index = event.GetId() - btnFormatHeaderDown0;
    SceneObj_cls* pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls* pCParticle = (Particle_cls*)pCObj;
	pCParticle->UpDownHeaderCommand(index, +1);

	UpdatePanels("ALL");
    }
    //CheckFormat(); 
}


void CLS::
OnDeleteBody(wxCommandEvent& event){
    int index = event.GetId() - btnFormatBodyDelete0;
    SceneObj_cls* pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls* pCParticle = (Particle_cls*)pCObj;
	pCParticle->DeleteBodyCommand(index);
    }
    CheckFormat();
}

void CLS::
OnAddBody(wxCommandEvent& event){
    int index = event.GetId() - btnFormatBodyAdd0;
    SceneObj_cls* pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls* pCParticle = (Particle_cls*)pCObj;
	pCParticle->AddBodyCommand(index);
    }
    CheckFormat();
}

void CLS::
OnUpBody(wxCommandEvent& event){
    int index = event.GetId() - btnFormatBodyUp0;
    SceneObj_cls* pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls* pCParticle = (Particle_cls*)pCObj;
	pCParticle->UpDownBodyCommand(index, -1);

	UpdatePanels("ALL");
    }
}

void CLS::
OnDownBody(wxCommandEvent& event){
    int index = event.GetId() - btnFormatBodyDown0;
    SceneObj_cls* pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls* pCParticle = (Particle_cls*)pCObj;
	pCParticle->UpDownBodyCommand(index, +1);

	UpdatePanels("ALL");
    }
}
