//(setq c-basic-offset 4)
// "clsPropertyInfo.h"

#include "clsPropertyInfo.h"
#include "../clsProperty.h"

#include "../../SCENEOBJ/clsSceneObj.h"
#include "../../SCENEOBJ/clsSceneObjCtrl.h"
#include "../../SCENEOBJ/clsMeshObj.h"
#include "../../SCENEOBJ/DATAVIEWER/clsHeightField.h"
#include "../../SCENEOBJ/DATAVIEWER/clsParticle.h"
#include "../../SCENEOBJ/DATAVIEWER/clsStreamline.h"
#include "../../SCENEOBJ/DATAVIEWER/clsVolumeBox.h"

#include "../../WIDGETCTRL/clsBitmapPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"

#include "../../COMMON/CommonDefinitions.h"

#include "../../../wxWidgetEnumMenu.h"

#define CLS PropertyInfo_cls

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
EVT_BUTTON(btnGraphPin, CLS::OnPinMode)
EVT_BUTTON(btnDropdown, CLS::OnDropDown)
EVT_BUTTON(btnInfoIDSearch, CLS::OnIDSearch)
EVT_MENU_RANGE(menuPop0, menuPop9, CLS::OnPopMenu)
END_EVENT_TABLE()

PropertyInfo_cls::
PropertyInfo_cls(Property_cls *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxPanelEX(pCParent, id, pos, size, style), PropertyBase_cls(pCParent)
{
    SetPanelType(_PANEL_PROPERTY_INFO);

    SetToDefaultParam();

    mpCBoxSizerSrc  = new wxBoxSizer(wxVERTICAL);

    mpCPain = GetPCNewPain(this);

    mpCTextInfo = new wxStaticText(this, wxID_ANY, wxT("Info"));

    wxBoxSizer *pBoxSizerInfo = GetNewPCBoxSizerInfoWithItems(this);

    mpCBoxSizerSrc->Add(pBoxSizerInfo, 0, wxEXPAND);
    mpCBoxSizerSrc->Add(mpCPain);
    mpCBoxSizerSrc->Add(mpCTextInfo, 1, wxEXPAND);

    SetSizerAndFit(mpCBoxSizerSrc);

    wxString strActiveObj = GetPCSceneObjCtrl()->GetActiveObjectName();
    if (strActiveObj.IsEmpty())
	Disconnect();
    else{
	wxString strTag;
	strTag << GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strActiveObj;
	mCstrLinkedObject.SetBodyFromData(strTag);
	mpCLinkedObject->SetLabel(strTag);
	Reconnect();
    }
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();
    mCConfList.SetTag("PropertyInfo");
    mCConfList.Add(&mCstrLinkedObject, "LinkedObject=", "None");
    mCConfList.Add(&mCbPinned, "Pinned=", "0");
}

void CLS::
ClearPanels(){
    if (mpCPain != NULL){
	mpCBoxSizerSrc->Detach(mpCPain);
	uint num = mpCPain->mVpCMNumPanelLink.size();
	for (uint i = 0; i < num; i++)
	    mpCPain->mVpCMNumPanelLink[i]->mbVolatile = TRUE;
	delete mpCPain;
	mpCPain = NULL;
    }
}

void CLS::
Disconnect(){
    if (mpCTextInfo->GetLabel().IsEmpty() == FALSE)
	mpCTextInfo->SetLabel(wxES);

    SetMinSize(wxSize(-1, -1));
    
    ClearPanels();
    Layout();
}

void CLS::
Reconnect(){
    Disconnect();

    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return ;

    SetMinSize(wxSize(-1, -1));

    wxSize CSize = wxDS;
    if (mpCPain == NULL)
	mpCPain = GetPCNewPain(this);

    wxString strClassName = pCObj->mstrClassName;
    if (strClassName.IsSameAs("Particle_cls"))
	ConnectParticle((Particle_cls*)pCObj);
    if (strClassName.IsSameAs("VolumeBox_cls"))
	ConnectVolume((VolumeBox_cls*)pCObj);

    wxBoxSizer *pCSizerForPain = new wxBoxSizer(wxVERTICAL);
    mpCPain->InitLayoutNumPanels(pCSizerForPain);
    mpCPain->SetSizer(pCSizerForPain);

    mpCBoxSizerSrc->Insert(0, mpCPain, 0, wxEXPAND);
}

void CLS::
ConnectParticle(Particle_cls* pCObj){
    MultiInputPanel_cls *pCMNumInfo = pCObj->GetPCNewPanelInfo(mpCPain);
    mpCPain->SetMNum(pCMNumInfo, mAstrClosedItem, "Particle.InfoIndex", TRUE, TRUE);

    mpCPain->LayoutNumPanels(FALSE);
    //HideAndShow();
    Layout();
}

void CLS::
ConnectVolume(VolumeBox_cls* pCObj){
    MultiInputPanel_cls *pCMNumInfo = pCObj->GetPCNewPanelInfo(mpCPain);
    mpCPain->SetMNum(pCMNumInfo, mAstrClosedItem, "Volume.InfoIndex", TRUE, TRUE);
    mpCPain->LayoutNumPanels(FALSE);
    Layout();
}

void CLS::
CheckLineNumber(wxString strInfo){
    int number = strInfo.Replace("\n", "\n");
    if (mLineNumber != number) {
	wxSize CCurrentSize = GetSize();
	wxSize CBestSize = GetBestSize();
	SetSize(wxSize(-1, CBestSize.GetHeight()));
	GetParent()->Layout();
	GetParent()->PostSizeEvent();
    }
    mLineNumber = number;
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs("ALL")){
	
	SceneObj_cls *pCObj = GetPCLinkedObj();
	if (pCObj == NULL) return Disconnect();

	wxString strClassName = pCObj->mstrClassName;
	wxString strInfo = "Empty";
	if (strClassName.IsSameAs("VolumeBox_cls"))
	    strInfo = ((VolumeBox_cls*)pCObj)->GetStrInfo();
	if (strClassName.IsSameAs("Particle_cls"))
	    strInfo = ((Particle_cls*)pCObj)->GetStrInfo();
	if (strClassName.IsSameAs("Streamline_cls"))
	    strInfo = ((Streamline_cls*)pCObj)->GetStrInfo();
	if (strClassName.IsSameAs("HeightField_cls"))
	    strInfo = ((HeightField_cls*)pCObj)->GetStrInfo();
	if (strClassName.IsSameAs("MeshObj_cls"))
	    strInfo = ((MeshObj_cls*)pCObj)->GetStrInfo();

	if (mpCTextInfo->GetLabel().IsSameAs(strInfo) == FALSE){
	    Freeze();
	    mpCTextInfo->SetLabel(strInfo);
	    Thaw();
	    CheckLineNumber(strInfo);
	}
	
	
	//Disconnect();
    }
}

void CLS::
OnPinMode(wxCommandEvent& event){
    long long pinned = event.GetInt();
    mCbPinned.SetBodyFromData(pinned);
}

void CLS::
OnDropDown(wxCommandEvent& event){
    mCSelector.PopupObjectSelector(this, wxES, "ALL");
}

void CLS::
OnPopMenu(wxCommandEvent& event){
    wxString strPopMenu = mCSelector.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == TRUE) return;
    mpCParent->SelectObject(strPopMenu);
}

void CLS::
OnIDSearch(wxCommandEvent& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return ;

    if (pCObj->IsSameAs("Particle_cls")){
	Particle_cls *pCParticle = (Particle_cls*)pCObj;
	long long index = pCParticle->SearchID();
	wxString strInfo;
	if (index < 0)
	    strInfo << _("Not Found");
	else
	    strInfo << "index:" << index << " ID:" << pCParticle->mCInfoID.GetBodyINT();
	wxMessageBox(strInfo, _("ID Search"));
    }
}
