// (setq c-basic-offset 4)
// "clsPropertyMaterial.cpp"

#ifndef _PROPERTY_MATERIAL_H
#define _PROPERTY_MATERIAL_H

#include "clsPropertyBase.h"

#include "../../CONFITEM/clsConfListContainer.h"
#include "../../INPUTMETHOD/clsSelectorPopup.h"
#include "../../WIDGETCTRL/EX/wxPanelEX.h"
#include "../../WIDGETCTRL/INPUTITEM/clsVCenter.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsHookCtrl.h"

class ColorCurve_cls;
class HeightField_cls;
class Isosurface_cls;
class LightSource_cls;
class Material_cls;
class MeshObj_cls;
class MultiInputPanel_cls;
class MNumPainBase_cls;
class Particle_cls;
class Primitive_cls;
class Property_cls;
class SceneObj_cls;
class Streamline_cls;
class TextObj_cls;
class VolumeBox_cls;

#include "../../COMMON/CommonDefinitionsWX.h"

class PropertyMaterial_cls : public wxPanelEX, public PropertyBase_cls, public ConfListContainer_cls, public HookItem_cls {
public:
    PropertyMaterial_cls(Property_cls *pCParent, wxWindowID id=wxID_ANY, const wxPoint &pos=wxDP, const wxSize &size=wxDS, long style=wxTAB_TRAVERSAL);

    ConfItemSTR_cls mCDummySTR;

    ConfItemFLOATARRAY_cls mCAmbient, mCDiffuse, mCSpecular;
    ConfItemFLOAT_cls mCShininess, mCAlpha, mCStrength;
    
    SelectorPopup_cls mCSelector;

    wxArrayString mAstrClosedItem;

    void SetToDefaultParam();

    int mIndexBtnAmbient, mIndexBtnDiffuse, mIndexBtnSpecular;
    int mIndexAmbient, mIndexDiffuse, mIndexSpecular;
    int mIndexShineAlpha, mIndexStrength;
    int mIndexLink, mIndexLinkCancel, mIndexTexturePath;
    int mIndexColorCurve0;
    int mIndexFlat, mIndexThreshold;

    int mCountMaterial, mCountColorCurve, mCountTexture;
    
    //wxButton *mpCBtnAddItem;
    MultiInputPanel_cls* mpCMNumLink;
    
    void SetMNumLink();
    MultiInputPanel_cls* GetPCNewPanelMaterial(wxWindow *pCWindow);
    MultiInputPanel_cls* GetPCNewPanelShader(wxWindow *pCWindow, Material_cls *pCMaterial);
    	
    void HideAndShow();
    bool HideAndShowShader(MultiInputPanel_cls *pCMNum, Material_cls *pCMaterial);

    void ClearPanels();
    void Disconnect();
    void UseLink(bool bUse);
    void ConnectPrimitive(Primitive_cls* pCObj);
    void ConnectMeshObj(MeshObj_cls* pCObj);
    void ConnectTextObj(TextObj_cls* pCObj);
    void ConnectLightSource(LightSource_cls* pCObj);
    void ConnectVolumeBox(VolumeBox_cls* pCObj);
    void ConnectIsosurface(Isosurface_cls* pCObj);
    void ConnectParticle(Particle_cls* pCObj);
    void ConnectStreamline(Streamline_cls* pCObj);
    void ConnectHF(HeightField_cls* pCObj);
    void Reconnect();
    void UpdateButtonColor();
    void UpdateButtonColorParticle();
    void UpdateButtonColorStreamline();
    void SetMaterialInfo();
    void SetMaterialInfoLightSource();
    void SetMaterialInfoParticle();
    void SetMaterialInfoStreamline();
    void CheckColorCurveAndTexture();
    
    void UpdatePanels(const wxString& strPanelTag);

    void HookFunc01(wxComEvt& event){OnLinkChange(event);}
    void OnLinkChange(wxComEvt& event);

    void HookFunc02(wxComEvt& event){OnMaterialChange(event);}
    void OnMaterialChange(wxComEvt& event);
    void SyncMaterial(bool bGroup = FALSE);
    void SyncMaterialParticle(bool bGroup = FALSE);
    void SyncMaterialStreamline(bool bGroup = FALSE);

    //std::vector < float > GetColorFromDialog(wxColour CInitColor);

    void OnDropDown    (wxComEvt& event);
    void OnClearLink   (wxComEvt& event);
    void OnBtnAmbient  (wxComEvt& event);
    void OnBtnDiffuse  (wxComEvt& event);
    void OnBtnSpecular (wxComEvt& event);
    void OnTexture     (wxComEvt& event);
    void OnClearTexture(wxComEvt& event);
    void OnParticleTexture     (wxComEvt& event);
    void OnClearParticleTexture(wxComEvt& event);
    void OnBtnMaterialEmit     (wxComEvt& event);
    void OnBtnMaterialDiffuse  (wxComEvt& event);
    void OnBtnMaterialSpecular (wxComEvt& event);
    void OnColorCurve          (wxComEvt& event);
    void OnPinMode(wxComEvt& event);
    void OnAddItem(wxComEvt& event);
    void OnPopMenu(wxComEvt& event);
    void OnKeyDown(wxKeyEvent& event);
    ConfItemFLOATARRAY_cls* GetPCItemColor(int buttonId);
    void OnCopyColor       (wxComEvt& event);
    void OnPasteColor      (wxComEvt& event);
    ColorCurve_cls* GetPCColorCurve(int curveIdx);
    void OnCopyColorCurve  (wxComEvt& event);
    void OnPasteColorCurve (wxComEvt& event);
    void OnAddColorCurve   (wxComEvt& event);
    void OnDeleteColorCurve(wxComEvt& event);
    void OnAddTexture      (wxComEvt& event);
    void OnDeleteTexture   (wxComEvt& event);
    void OnDeleteMaterial  (wxComEvt& event);

    DECLARE_EVENT_TABLE()
};

#endif
