//(setq c-basic-offset 4)
// "clsPropertyObject.h"

#include "clsPropertyObject.h"

#include "../clsProperty.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../../SCENEOBJ/clsCamera.h"
#include "../../SCENEOBJ/clsLightSource.h"
#include "../../SCENEOBJ/clsMaterial.h"
#include "../../SCENEOBJ/clsMaterialLight.h"
#include "../../SCENEOBJ/clsMeshObj.h"
#include "../../SCENEOBJ/clsPrimitive.h"
#include "../../SCENEOBJ/clsSceneObj.h"
#include "../../SCENEOBJ/clsSceneObjCtrl.h"
#include "../../SCENEOBJ/clsTextObj.h"
#include "../../SCENEOBJ/clsWorld.h"
#include "../../SCENEOBJ/DATAVIEWER/clsHeightField.h"
#include "../../SCENEOBJ/DATAVIEWER/clsVolumeBox.h"
#include "../../SCENEOBJ/DATAVIEWER/clsIsosurface.h"
#include "../../SCENEOBJ/DATAVIEWER/clsParticle.h"
#include "../../SCENEOBJ/DATAVIEWER/clsStreamline.h"
#include "../../SCENEOBJ/PROP/clsCameraProp.h"
#include "../../SCENEOBJ/PROP/clsFocalBlurProp.h"
#include "../../SCENEOBJ/PROP/clsFogProp.h"
#include "../../SCENEOBJ/PROP/clsIsosurfProp.h"
#include "../../SCENEOBJ/PROP/clsStreamlineProp.h"
#include "../../SCENEOBJ/PROP/clsStereoProp.h"
#include "../../SCENEOBJ/PROP/clsTextProp.h"
#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../UNDO/clsUndoManager.h"
#include "../../WIDGETCTRL/clsBitmapPanel.h"
#include "../../WIDGETCTRL/INPUTITEM/clsVCenter.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"
#include "../../WIDGETCTRL/EX/clsFileDialogEX.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"
#include "../../../clsMain.h"

#include <wx/filename.h>
#include <wx/fontdlg.h>

#include "../../COMMON/CommonDefinitions.h"
#include "../../../wxWidgetEnum.h"
#include "../../../wxWidgetEnumMenu.h"

#define CLS PropertyObject_cls

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
EVT_BUTTON(btnGraphPin, CLS::OnPinMode)
EVT_BUTTON(btnDropdown, CLS::OnDropDown)
EVT_BUTTON(btnAutoscale, CLS::OnAutoscale)
EVT_BUTTON(btnReload, CLS::OnReload)
EVT_BUTTON(btnExporter, CLS::OnExport)
EVT_BUTTON(btnExportFormat, CLS::OnExportFormat)
EVT_BUTTON(btnClassify, CLS::OnClassify)
EVT_BUTTON(btnVolumePath, CLS::OnVolumePath)
EVT_BUTTON(btnParticlePath, CLS::OnParticlePath)
EVT_BUTTON(btnParticlePathExporter, CLS::OnParticlePathExporter)
EVT_BUTTON(btnStreamlinePath, CLS::OnStreamlinePath)
EVT_BUTTON(btnHFPath, CLS::OnHFPath)
EVT_BUTTON(btnMeshPath, CLS::OnMeshPath)
EVT_BUTTON(btnTextFont, CLS::OnTextFont)
EVT_BUTTON(btnFogColor, CLS::OnBtnFogColor)
EVT_MENU_RANGE(menuPop0, menuPop9, CLS::OnPopMenu)
//EVT_CHAR (PropertyRender_cls::OnKeyDown)
END_EVENT_TABLE()

CLS::
PropertyObject_cls(Property_cls *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) : 
wxPanelEX(pCParent, id, pos, size, style), PropertyBase_cls(pCParent)
{
    
    SetPanelType(_PANEL_PROPERTY_OBJECT);

    SetToDefaultParam();

    mpCBoxSizerSrc  = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerInfo = GetNewPCBoxSizerInfoWithItems(this);
    
    SetMNum();
    
    mpCPain = GetPCNewPain(this);
    
    mpCBoxSizerSrc->Add(pBoxSizerInfo, 0, wxEXPAND);
    mpCBoxSizerSrc->Add(mpCMNum, 0, wxEXPAND);
    mpCBoxSizerSrc->Add(mpCPain);

    SetSizerAndFit(mpCBoxSizerSrc);

    wxString strActiveObj = GetPCSceneObjCtrl()->GetActiveObjectName();
    if (strActiveObj.IsEmpty())
	Disconnect();
    else{
	wxString strTag;
	strTag << GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strActiveObj;
	mCstrLinkedObject.SetBodyFromData(strTag);
	mpCLinkedObject->SetLabel(strTag);
	Reconnect();
    }
}

CLS::
~PropertyObject_cls(){
    Disconnect();
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();
    mCConfList.SetTag("PropertyObject");
    mCConfList.Add(&mCstrLinkedObject, "LinkedObject=", "None");
    mCConfList.Add(&mCbPinned, "Pinned=", "0");
}

void CLS::
SetMNum(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNum, this,
					wxT("PropertyObject"));

    pC->mbItemVolatile = TRUE;
    mIndexName = pC->CombineConfItem(mCDummyName, _("Name"));
    
    pC->FinishSetting();
    pC->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_TEXT_ENTER, mIndexName);
}

void CLS::
HideAndShow(){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    bool bChange = FALSE;
    if (pCObj->IsSameAs("LightSource_cls"))
	bChange |= ((LightSource_cls*)pCObj)->HideAndShow(mpCPain->GetPCMNum("LightSource"));

    if (pCObj->IsSameAs("Primitive_cls"))
	bChange |= ((Primitive_cls*)pCObj)->HideAndShow(mpCPain->GetPCMNum("Primitive"));
    if (pCObj->IsSameAs("Camera_cls")){
	Camera_cls *pCCam = (Camera_cls*)pCObj;
	bChange |= pCCam->HideAndShowCamera(mpCPain->GetPCMNum(_("Camera")));
	bChange |= pCCam->HideAndShowStereo(mpCPain->GetPCMNum(_("Stereogram")));
	bChange |= pCCam->HideAndShowFocalBlur(mpCPain->GetPCMNum(_("Focal Blur")));
    }
    if (pCObj->IsSameAs("Particle_cls"))
	bChange |= ((Particle_cls*)pCObj)->HideAndShow(mpCPain);
    if (pCObj->IsSameAs("World_cls"))
	bChange |= ((World_cls*)pCObj)->HideAndShow(mpCPain->GetPCMNum(_("Fog")));
    
    if (bChange == TRUE){
	Layout();
	GetParent()->Layout();
    }
}

void CLS::
ClearPanels(){
    if (mpCPain != NULL){
	mpCBoxSizerSrc->Detach(mpCPain);
	uint num = mpCPain->mVpCMNumPanelLink.size();
	for (uint i = 0; i < num; i++)
	    mpCPain->mVpCMNumPanelLink[i]->mbVolatile = TRUE;
	delete mpCPain;
	mpCPain = NULL;
    }
}

void CLS::
Disconnect(){
    SetMinSize(wxSize(-1, -1));

    mpCMNum->mVCItem[mIndexName].Hide();
    ClearPanels();
    
    Layout();
}

void CLS::
LayoutForConnect(){
    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectPrimitive(Primitive_cls* pCObj){

    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "Primitive", TRUE, TRUE);

    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectMesh(MeshObj_cls* pCObj){
    
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "Mesh", TRUE, TRUE);

    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectTextObj(TextObj_cls* pCObj){
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "Text", TRUE, TRUE);

    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexScale);

    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectLightSource(LightSource_cls* pCObj){
    
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "LightSource", TRUE, TRUE);

    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexSpot+1);
    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexSpot+2);
    
    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectCamera(Camera_cls* pCObj){
   
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanelCamera(mpCPain);
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, _("Camera"), TRUE, TRUE);

    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexAngleV + 2);
    
    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexClipMinMax+1);
    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexClipMinMax+2);
    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexTilt);

    MultiInputPanel_cls *pCMNumFocalBlur = pCObj->GetPCNewPanelFocalBlur(mpCPain);
    mpCPain->SetMNum(pCMNumFocalBlur, mAstrClosedItem, _("Focal Blur"), TRUE, TRUE);

    pCMNumFocalBlur->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexBlurSize+1);
    pCMNumFocalBlur->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexBlurSize+2);

    MultiInputPanel_cls *pCMNumStereo = pCObj->GetPCNewPanelStereo(mpCPain);
    mpCPain->SetMNum(pCMNumStereo, mAstrClosedItem, _("Stereogram"), TRUE, TRUE);
    
    pCMNumStereo->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexParallax+2);
    
    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectWorld(World_cls* pCObj){

    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanelFog(mpCPain);
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, _("Fog"), TRUE, TRUE);
    
    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER);

    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectVolumeBox(VolumeBox_cls* pCObj){
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "VolumeBox", TRUE, TRUE);

    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectIsosurface(Isosurface_cls* pCObj){
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "Isosurface", TRUE, TRUE);

    pCMNum->SetHook(this, &HookItem_cls::HookFunc02, wxEVT_COMMAND_TEXT_ENTER, pCObj->mIndexThreshold);

    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectParticle(Particle_cls* pCObj){
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);
    
    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "Particle", TRUE, TRUE);
    
    MultiInputPanel_cls *pCMNumExporter = pCObj->GetPCNewPanelExporter(mpCPain);
    mpCPain->SetMNum(pCMNumExporter, mAstrClosedItem, "Exporter", TRUE, TRUE);
    
    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectStreamline(Streamline_cls* pCObj){
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);

    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "Streamline", TRUE, TRUE);

    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
ConnectHeightField(HeightField_cls* pCObj){
    MultiInputPanel_cls *pCMNum = pCObj->GetPCNewPanel(mpCPain);

    mpCPain->SetMNum(pCMNum, mAstrClosedItem, "HeightField", TRUE, TRUE);
    
    mpCPain->LayoutNumPanels();
    HideAndShow();
    Layout();
}

void CLS::
Reconnect(){
    Disconnect();
  
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    mpCMNum->mVCItem[mIndexName].Show();
    
    SetMinSize(wxSize(-1, -1));

    wxSize CSize = wxDS;
    if (mpCPain == NULL){
	mpCPain = new MNumPainBase_cls(this, CSize, FALSE);
	mpCPain->mbHideButton = FALSE;
	
	wxString strActiveObj = GetPCSceneObjCtrl()->GetActiveObjectName();
	mpCPain->mstrTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strActiveObj << "::";
    }

    mpCPain->Freeze();

    if (pCObj->IsSameAs("Primitive_cls"))
	ConnectPrimitive((Primitive_cls*)pCObj);
    if (pCObj->IsSameAs("MeshObj_cls"))
	ConnectMesh((MeshObj_cls*)pCObj);
    if (pCObj->IsSameAs("TextObj_cls"))
	ConnectTextObj((TextObj_cls*)pCObj);
    if (pCObj->IsSameAs("LightSource_cls"))
	ConnectLightSource((LightSource_cls*)pCObj);
    if (pCObj->IsSameAs("Camera_cls"))
	ConnectCamera((Camera_cls*)pCObj);
    if (pCObj->IsSameAs("World_cls"))
	ConnectWorld((World_cls*)pCObj);
    if (pCObj->IsSameAs("VolumeBox_cls"))
	ConnectVolumeBox((VolumeBox_cls*)pCObj);
    if (pCObj->IsSameAs("Isosurface_cls"))
	ConnectIsosurface((Isosurface_cls*)pCObj);
    if (pCObj->IsSameAs("Particle_cls"))
	ConnectParticle((Particle_cls*)pCObj);
    if (pCObj->IsSameAs("Streamline_cls"))
	ConnectStreamline((Streamline_cls*)pCObj);
    if (pCObj->IsSameAs("HeightField_cls"))
	ConnectHeightField((HeightField_cls*)pCObj);
        
    wxBoxSizer *pCSizerForPain = new wxBoxSizer(wxVERTICAL);

    mpCPain->InitLayoutNumPanels(pCSizerForPain);
    mpCPain->SetSizer(pCSizerForPain);

    mpCBoxSizerSrc->Add(mpCPain, 0, wxEXPAND);
    mpCPain->Thaw();

    mpCMNum->UpdateValues(_MN_ALLITEM);
}

void CLS::
UpdateButtonColor(){
    if (mpCPain == NULL) return;

    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;
    if (pCObj->IsSameAs("World_cls"))
	return UpdateButtonColorWorld();
    if (pCObj->IsSameAs("Particle_cls"))
	return UpdateButtonColorParticle();
}

void CLS::
UpdateButtonColorWorld(){
    World_cls *pCObj = (World_cls*)GetPCLinkedObj();
    pCObj->UpdateButtonColor(mpCPain);
}

void CLS::
UpdateButtonColorParticle(){
    Particle_cls *pCObj = (Particle_cls*)GetPCLinkedObj();
    pCObj->UpdateButtonColorExporter(mpCPain);
}

void CLS::
SetObjectInfo(){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    mCDummyName.SetBodyFromData(pCObj->GetStrName());

    double time = GetPCTimeCtrl()->GetCurrentNT();
    if (pCObj->IsSameAs("Camera_cls")){
	Camera_cls *pCCamera = (Camera_cls*)pCObj;
	pCCamera->mpCCameraProp    ->SyncronizeConf(time);
	pCCamera->mpCFocalBlurProp ->SyncronizeConf(time);
	pCCamera->mpCStereoProp    ->SyncronizeConf(time);
    }

    if (pCObj->IsSameAs("LightSource_cls")){
	LightSource_cls *pCLight = (LightSource_cls*)pCObj;
	MaterialLight_cls *pCLightProp = pCLight->mpCMaterial;
	
	std::vector < double > VSpot = pCLightProp->GetPCParamGroup("SpotParam")->GetVInterpolatedValue(time, FALSE);
	pCLight->mCSpotAngle.SetBodyFromData(VSpot[0]);
	pCLight->mCSpotExponent.SetBodyFromData(VSpot[1]);
    }

    if (pCObj->IsSameAs("VolumeBox_cls")){
	VolumeBox_cls *pCVolumeBox = (VolumeBox_cls*)pCObj;
	pCVolumeBox->UpdatePathColor(mpCPain, pCVolumeBox->IsDataValid());
    }

    if (pCObj->IsSameAs("Isosurface_cls")){
	Isosurface_cls *pCIsosurface = (Isosurface_cls*)pCObj;
	pCIsosurface->mpCIsosurfProp->SyncronizeConf(time);
    }    

    if (pCObj->IsSameAs("HeightField_cls")){
	HeightField_cls *pCHeightField = (HeightField_cls*)pCObj;
	pCHeightField->UpdatePathColor(mpCPain, pCHeightField->IsDataValid());
    }

    if (pCObj->IsSameAs("Particle_cls")){
	Particle_cls *pCParticle = (Particle_cls*)pCObj;
	pCParticle->UpdatePathColor(mpCPain, pCParticle->IsDataValid());
	pCParticle->UpdatePathColorExporter(mpCPain);
    }

    if (pCObj->IsSameAs("Streamline_cls")){
	Streamline_cls *pCStreamline = (Streamline_cls*)pCObj;
	pCStreamline->UpdatePathColor(mpCPain, pCStreamline->IsDataValid());
	pCStreamline->mpCStreamlineProp->SyncronizeConf(time);
    }

    if (pCObj->IsSameAs("TextObj_cls")){
	TextObj_cls *pCTextObj = (TextObj_cls*)pCObj;
	pCTextObj->mpCTextProp->SyncronizeConf(time);
    }

    if (pCObj->IsSameAs("World_cls")){
	World_cls *pCWorld = (World_cls*)pCObj;
	pCWorld->SetPropInfo();
    }
}

void CLS::
OnObjectChange(wxComEvt& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    wxString strNewName = mCDummyName.GetBodySTR();
    wxString strOldName = pCObj->GetStrName();
    GetPCSceneObjCtrl()->RenameCommand(strOldName, strNewName);

    UpdatePanels("ALL");
}

void CLS::
CheckParamChange(SceneObj_cls *pCObj, Attribute_cls *pCProp,
		 ParamGroup_cls *pCPG, int paramIndex, 
		 float time, float newValue){
    
    Parameter_cls &CParam = pCPG->mVCParam[paramIndex];
    float oldValue = CParam.GetValue(time);
    
    int keyType;
    double easeFactor;
    GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);

    if (oldValue != newValue){
	wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << pCObj->GetStrName() << "::" << pCProp->GetStrName() << "::" << pCPG->GetStrName() << "::" << CParam.GetStrName();
	CParam.SetValueAtCommand(time, newValue, strTag, 
				 keyType, easeFactor);
    }

}

void CLS::
OnParamChange(wxComEvt& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    wxString strObj = pCObj->GetStrName();
    float time = GetPCTimeCtrl()->GetCurrentNT();

    if (pCObj->IsSameAs("Camera_cls")){
	Camera_cls *pCCamera = (Camera_cls*)pCObj;
	pCCamera->mpCFocalBlurProp->SyncronizeParam(strObj);
	pCCamera->mpCStereoProp->SyncronizeParam(strObj);
	pCCamera->mpCCameraProp->SyncronizeParam(strObj);
    }

    // mpCMaterial に ParamGroup があって、LightSource_cls にmCConfがある変則
    if (pCObj->IsSameAs("LightSource_cls")){
	LightSource_cls *pCLight = (LightSource_cls*)pCObj;
	MaterialLight_cls *pCLightProp = pCLight->mpCMaterial;

	ParamGroup_cls *pCPGSpotParam = pCLightProp->GetPCParamGroup("SpotParam");
	CheckParamChange(pCObj, pCLightProp, pCPGSpotParam, 0, time, pCLight->mCSpotAngle.GetBodyDOUBLE());
	CheckParamChange(pCObj, pCLightProp, pCPGSpotParam, 1, time, pCLight->mCSpotExponent.GetBodyDOUBLE());
    }

    if (pCObj->IsSameAs("Isosurface_cls")){
	Isosurface_cls *pCIsosurface = (Isosurface_cls*)pCObj;
	pCIsosurface->mpCIsosurfProp->SyncronizeParam(strObj);
    }

    if (pCObj->IsSameAs("TextObj_cls")){
	TextObj_cls *pCTextObj = (TextObj_cls*)pCObj;
	pCTextObj->mpCTextProp->SyncronizeParam(strObj);
    }
    
    SyncProp(FALSE);
}

void CLS::
SyncProp(bool bGroup){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->IsSameAs("World_cls"))
	return SyncPropWorld(bGroup);
}

void CLS::
SyncPropWorld(bool bGroup){
    World_cls* pCObj = (World_cls*)GetPCLinkedObj();
    
    if (bGroup == TRUE) GetPCUndo()->OpenBlock("ClrChange");
    pCObj->SyncPropWorld();
    if (bGroup == TRUE) GetPCUndo()->CloseBlock("ClrChange");
}

void CLS::
OnDropDown(wxComEvt& event){
    wxString strCommand = event.GetString();
    if (strCommand.IsSameAs("Isosurface_cls"))
	return mCSelector.PopupObjectSelector(this, strCommand, "VolumeBox_cls");
    mCSelector.PopupObjectSelector(this, strCommand, "ALL"); }

void CLS::
OnAutoscale(wxComEvt& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->IsSameAs("VolumeBox_cls"))
	((VolumeBox_cls*)pCObj)->Rescale();
}

void CLS::
OnReload(wxComEvt& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->IsSameAs("VolumeBox_cls"))   ((VolumeBox_cls*)pCObj)->Reload();
    if (pCObj->IsSameAs("Particle_cls"))    ((Particle_cls*)pCObj)->Reload();
    if (pCObj->IsSameAs("HeightField_cls")) ((HeightField_cls*)pCObj)->Reload();
    if (pCObj->IsSameAs("MeshObj_cls"))     ((MeshObj_cls*)pCObj)->Reload();
    if (pCObj->IsSameAs("Streamline_cls"))  ((Streamline_cls*)pCObj)->Reload();

    GetPCMain()->UpdatePanels("VIEW3D");
}

void CLS::
OnExport(wxComEvt& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->IsSameAs("Particle_cls")) ((Particle_cls*)pCObj)->Export();

    GetPCTimeCtrl()->ExportCommand();
}

void CLS::
OnExportFormat(wxComEvt& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;

    if (pCObj->IsSameAs("Particle_cls")) ((Particle_cls*)pCObj)->ExportFormat();
}

void CLS::
OnClassify(wxComEvt& event){
    SceneObj_cls *pCObj = GetPCLinkedObj();
    if (pCObj == NULL) return;
    
    if (pCObj->IsSameAs("Particle_cls")) ((Particle_cls*)pCObj)->Expression();
}

void CLS::
OnVolumePath(wxComEvt& event){
    wxFileDialogEX *pCDLG = GCClassProvider.GetPCFileDialog();
    VolumeBox_cls *pCObj = (VolumeBox_cls*)GetPCLinkedObj();

    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select Texture");
    if (pCObj->mCbZStack.GetBodyBOOL() == TRUE)
	pCDLG->SetWildcard("IMAGE");
    else
	pCDLG->SetWildcard("df3|*.df3");

    if (!(pCObj->mstrPathChecked.IsEmpty() == TRUE ||
	  pCObj->mstrPathChecked.IsSameAs("None"))){
	/*
	wxFileName CName(pCObj->mstrPathChecked);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
	*/
	wxString strCurrentFile = pCObj->GetStrPathCandid();
	wxFileName CName(strCurrentFile);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    else if (GetPCMain()->mCLastOpenedFileVOLUME.IsEmpty() == FALSE){
	wxFileName CName(GetPCMain()->mCLastOpenedFileVOLUME.mstrBody);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }

    

    pCDLG->mpCDialog->SetPosition(wxDP);
    pCDLG->mpCDialog->SetSize(wxDS);
    
    if (pCDLG->ShowModal() == wxID_OK)
	pCObj->SetPathGUI(mpCPain->mVpCMNumPanelLink[0], pCDLG->GetPath());

    pCDLG->DestroyDialog();
}

void CLS::
OnParticlePath(wxComEvt& event){
    wxFileDialogEX *pCDLG = GCClassProvider.GetPCFileDialog();
    Particle_cls *pCObj = (Particle_cls*)GetPCLinkedObj();

    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select File Sequence");
    pCDLG->SetWildcard("*|*.*");

    if (!(pCObj->mstrPathChecked.IsEmpty() == TRUE ||
	  pCObj->mstrPathChecked.IsSameAs("None"))){
	wxString strCurrentFile = pCObj->GetStrPathCandid();
	wxFileName CName(strCurrentFile);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    else if (GetPCMain()->mCLastOpenedFilePARTICLE.IsEmpty() == FALSE){
	wxFileName CName(GetPCMain()->mCLastOpenedFilePARTICLE.mstrBody);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }

    pCDLG->mpCDialog->SetPosition(wxDP);
    pCDLG->mpCDialog->SetSize(wxDS);

    if (pCDLG->ShowModal() == wxID_OK)
	pCObj->SetPathGUI(mpCPain->mVpCMNumPanelLink[0], pCDLG->GetPath());
    
    pCDLG->DestroyDialog();
}

void CLS::
OnParticlePathExporter(wxComEvt& event){
    wxFileDialogEX *pCDLG = GCClassProvider.GetPCFileDialog();
    Particle_cls *pCObj = (Particle_cls*)GetPCLinkedObj();

    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select File Sequence");
    pCDLG->SetWildcard("*|*.*");

    pCDLG->mpCDialog->SetPosition(wxDP);
    pCDLG->mpCDialog->SetSize(wxDS);
    
    if (pCDLG->ShowModal() == wxID_OK)
	pCObj->SetPathGUIExporter(mpCPain->mVpCMNumPanelLink[1], pCDLG->GetPath());
    
    pCDLG->DestroyDialog();
}

void CLS::
OnStreamlinePath(wxComEvt& event){
    wxFileDialogEX *pCDLG = GCClassProvider.GetPCFileDialog();
    Streamline_cls *pCObj = (Streamline_cls*)GetPCLinkedObj();

    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select File Sequence");
    pCDLG->SetWildcard("*|*.*");

    if (!(pCObj->mstrPathChecked.IsEmpty() == TRUE ||
	  pCObj->mstrPathChecked.IsSameAs("None"))){
	/*
	wxFileName CName(pCObj->mstrPathChecked);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
	*/
	wxString strCurrentFile = pCObj->GetStrPathCandid();
	wxFileName CName(strCurrentFile);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    else if (GetPCMain()->mCLastOpenedFileSTREAM.IsEmpty() == FALSE){
	wxFileName CName(GetPCMain()->mCLastOpenedFileSTREAM.mstrBody);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }

    pCDLG->mpCDialog->SetPosition(wxDP);
    pCDLG->mpCDialog->SetSize(wxDS);

    if (pCDLG->ShowModal() == wxID_OK)
	pCObj->SetPathGUI(mpCPain->mVpCMNumPanelLink[0], pCDLG->GetPath());
    
    pCDLG->DestroyDialog();
}

void CLS::
OnHFPath(wxComEvt& event){
    wxFileDialogEX *pCDLG = GCClassProvider.GetPCFileDialog();
    HeightField_cls *pCObj = (HeightField_cls*)GetPCLinkedObj();

    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select File Sequence");
    pCDLG->SetWildcard("IMAGE");

    if (!(pCObj->mstrPathChecked.IsEmpty() == TRUE ||
	  pCObj->mstrPathChecked.IsSameAs("None"))){
	/*
	wxFileName CName(pCObj->mstrPathChecked);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
	*/
	wxString strCurrentFile = pCObj->GetStrPathCandid();
	wxFileName CName(strCurrentFile);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    else if (GetPCMain()->mCLastOpenedFileHF.IsEmpty() == FALSE){
	wxFileName CName(GetPCMain()->mCLastOpenedFileHF.mstrBody);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    
    pCDLG->mpCDialog->SetPosition(wxDP);
    pCDLG->mpCDialog->SetSize(wxDS);

    if (pCDLG->ShowModal() == wxID_OK)
	pCObj->SetPathGUI(mpCPain->mVpCMNumPanelLink[0], pCDLG->GetPath());
    
    pCDLG->DestroyDialog();
}

void CLS::
OnMeshPath(wxComEvt& event){
    wxFileDialogEX *pCDLG = GCClassProvider.GetPCFileDialog();
    MeshObj_cls *pCObj = (MeshObj_cls*)GetPCLinkedObj();

    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select Wavefront .obj file");
    pCDLG->SetWildcard("obj|*.obj");
    
    if (!(pCObj->mstrLoaded.IsEmpty() == TRUE ||
	  pCObj->mstrLoaded.IsSameAs("None"))){
	wxFileName CName(pCObj->mstrLoaded);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    else if (GetPCMain()->mCLastOpenedFileMESH.IsEmpty() == FALSE){
	wxFileName CName(GetPCMain()->mCLastOpenedFileMESH.mstrBody);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    
    pCDLG->mpCDialog->SetPosition(wxDP);
    pCDLG->mpCDialog->SetSize(wxDS);
    
    if (pCDLG->ShowModal() == wxID_OK) {
	MeshObj_cls *pCObj = (MeshObj_cls*)GetPCLinkedObj();
	pCObj->SetPathGUI(mpCPain->mVpCMNumPanelLink[0], pCDLG->GetPath());
	//mpCMNum->SetItemSTR(mIndexMeshPath+1, pCDLG->GetPath());
    }

    pCDLG->DestroyDialog();
}

void CLS::
OnTextFont(wxComEvt& event){
    TextObj_cls *pCObj = (TextObj_cls*)GetPCLinkedObj();

    wxFont CFont;
    CFont.SetNativeFontInfoUserDesc(pCObj->mpCTextProp->mCstrFontDesc.GetBodySTR());

    CFont = wxGetFontFromUser(this, CFont, wxES);

    if (CFont.IsOk() == TRUE)
	pCObj->mpCTextProp->mCstrFontDesc.PushCommandPair(CFont.GetNativeFontInfoUserDesc());
}

void CLS::
OnBtnFogColor(wxComEvt& event){
    World_cls* pCWorld = (World_cls*)GetPCLinkedObj();
    if (pCWorld == NULL) return;
    
    wxColour strClr = pCWorld->mpCFogProp->mCFogColor.GetBodyWXColor();
    std::vector < float > VClr = GetColorFromDialog(strClr);
    if (VClr.size() == 0) return;

    pCWorld->mpCFogProp->mCFogColor.SetBodyFromData(VClr);
    SyncPropWorld(TRUE);
}

void CLS::
OnPinMode(wxComEvt& event){
    long long pinned = event.GetInt();
    mCbPinned.SetBodyFromData(pinned);
}

void CLS::
OnPopMenu(wxComEvt& event){
    wxString strPopMenu = mCSelector.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == TRUE) return;

    wxString strTag = mCSelector.mstrTag;
    if (strTag.IsSameAs("FocalBlurObj")){
	((Camera_cls*)GetPCLinkedObj())->mpCFocalBlurProp->mCstrFocusObj.PushCommandPair(strPopMenu);
	UpdatePanels("ALL");
    }
    else if (strTag.IsSameAs("Isosurface_cls")){
	Isosurface_cls* pCIsosurface = (Isosurface_cls*)GetPCLinkedObj();
	if (pCIsosurface == NULL) return;
	
	pCIsosurface->mCstrVolumeLink.PushCommandPair(strPopMenu);
	UpdatePanels("ALL");
    }
    else
	mpCParent->SelectObject(strPopMenu);
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    if (mpCMNum->GetName().IsSameAs(strPanelTag) || 
	GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs("ALL")){
	SetObjectInfo();
	HideAndShow();
	mpCMNum->UpdateValues(_MN_ALLITEM);
	if (mpCPain != NULL)
	    mpCPain->UpdateValues();
	UpdateButtonColor();
    }
}
