//(setq c-basic-offset 4)
// "clsPropertyRender.h"

#include "clsPropertyRender.h"
#include "../clsProperty.h"

#include "../../GLCANVAS/clsView3D.h"
#include "../../GRAPHIC/clsImageOutput.h"
#include "../../MYGL/clsShadowVolumeCtrl.h"
#include "../../SCENEOBJ/clsSceneObjCtrl.h"
#include "../../SCENEOBJ/clsWorld.h"
#include "../../SCENEOBJ/PROP/clsFogProp.h"
#include "../../TIMELINE/clsEnsembleCtrl.h"
#include "../../TIMELINE/clsMotionBlurCtrl.h"
#include "../../TIMELINE/clsTimeCtrl.h"

#include "../../WIDGETCTRL/EX/clsColourDialogEX.h"
#include "../../WIDGETCTRL/EX/clsFileDialogEX.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"
#include "../../WIDGETCTRL/WX/clsSizeAdjuster.h"

#include "../../../clsMain.h"
#include "../../../clsFrameMain.h"

#include <wx/filename.h>
#include <wx/clipbrd.h>

#include "../../COMMON/CommonDefinitions.h"

#include "../../../wxWidgetEnum.h"
#include "../../../wxWidgetEnumMenu.h"

#define CLS PropertyRender_cls

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
EVT_BUTTON(btnPropRenderSnap, CLS::OnRenderSnap)
EVT_BUTTON(btnPropRenderAnim, CLS::OnRenderAnim)
EVT_BUTTON(btnPropRenderPath, CLS::OnRenderPath)
EVT_BUTTON(btnPropGLBG, CLS::OnGLBG)
EVT_BUTTON(btnPropMaskBG, CLS::OnMaskBG)
EVT_MENU(menuCopyColor, CLS::OnCopyColor)
EVT_MENU(menuPasteColor, CLS::OnPasteColor)
END_EVENT_TABLE()

CLS::
PropertyRender_cls(Property_cls *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxPanelEX(pCParent, id, pos, size, style), PropertyBase_cls(pCParent)
{
    SetPanelType(_PANEL_PROPERTY_RENDER);

    SetToDefaultParam();

    mpCBoxSizerSrc  = new wxBoxSizer(wxVERTICAL);
    
    wxSize CSize = wxDS;
    mpCPain = new MNumPainBase_cls(this, CSize, FALSE);
    mpCPain->mbHideButton = FALSE;
    mpCPain->mstrTag = "Property_Render";
    
    MultiInputPanel_cls *pCMNumBG = GetPCNewPanelBG(mpCPain);
    mpCPain->SetMNum(pCMNumBG, mAstrClosedItem, _("BG"), TRUE, TRUE);

    MultiInputPanel_cls *pCMNumOutput = GetPCNewPanelOutput(mpCPain);
    mpCPain->SetMNum(pCMNumOutput, mAstrClosedItem, _("Output"), TRUE, TRUE);

    MultiInputPanel_cls *pCMNumSV = GetPCNewPanelShadowVolume(mpCPain);
    mpCPain->SetMNum(pCMNumSV, mAstrClosedItem, _("Shadow Volume"), TRUE, TRUE);

    MultiInputPanel_cls *pCMNumBlur = GetPCNewPanelMotionBlur(mpCPain);
    mpCPain->SetMNum(pCMNumBlur, mAstrClosedItem, _("Motion Blur"), TRUE, TRUE);

    MultiInputPanel_cls *pCMNumEnsemble = GetPCNewPanelEnsemble(mpCPain);
    mpCPain->SetMNum(pCMNumEnsemble, mAstrClosedItem, _("Ensemble"), TRUE, TRUE);
    
    mpCPain->LayoutNumPanels();
    Layout();

    wxBoxSizer *pCSizerForPain = new wxBoxSizer(wxVERTICAL);
    mpCPain->InitLayoutNumPanels(pCSizerForPain);
    mpCPain->SetSizer(pCSizerForPain);
    
    mpCBoxSizerSrc->Add(mpCPain, 0, wxEXPAND);

    SetMNum();
    mpCBoxSizerSrc->Add(mpCMNum, 0, wxEXPAND);

    mpCPain->LayoutNumPanels();

    HideAndShow();
    SetSizerAndFit(mpCBoxSizerSrc);

    SetMinSize(wxSize(-1, -1));
    Layout();
    GetParent()->Layout();
}

void CLS::
SetToDefaultParam(){
    mCDummyWidth.SetItem("Width=", "640");
    mCDummyHeight.SetItem("Height=", "360");
}

void CLS::
SetMNum(){
    MultiInputPanel_cls *pC = 
	MultiInputPanel_cls::Initialize(&mpCMNum, this,
					wxT("PropertyRender"));
    
    pC->CombineSizer(2);
    mIndexRenderSnap = pC->CombineButton(btnPropRenderSnap, _("Render Snap"));
    mIndexRenderAnim = pC->CombineButton(btnPropRenderAnim, _("Render Anim"));
    
    pC->FinishSetting();
    UpdateButtonColor();
}

MultiInputPanel_cls* CLS::
GetPCNewPanelBG(wxWindow* pCWindow){
    
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, _("BG")/*, TRUE*/);

    pC->mTextWidth = 120;

    pC->CombineSizer(2, _("BG Color"));
    pC->mContextMenuType = _ITEM_MENU_COLOR;
    mIndexGLBG = pC->CombineBitmapPanel(btnPropGLBG, " ");
    pC->mContextMenuType = _ITEM_MENU_COLOR;
    mIndexMaskBG = pC->CombineBitmapPanel(btnPropMaskBG, " ");

    wxString AstrBGAlpha[2] = {_("Not Transparent"), _("Transparent")};
    pC->SetChoiceList(2, AstrBGAlpha);
    pC->CombineConfItem(GetPCColorProvider()->mCBGAlphaType, _("BG Alpha"));
    
    pC->FinishSetting();

    //pC->mVCItem[mIndexGLBG].mpCWindowLink->SetBackgroundStyle(wxBG_STYLE_PAINT);
    //pC->mVCItem[mIndexMaskBG].mpCWindowLink->SetBackgroundStyle(wxBG_STYLE_PAINT);
    BitmapPanel_cls *pCBitmapGLBG = (BitmapPanel_cls*)pC->mVCItem[mIndexGLBG].mpCWindowLink;
    pCBitmapGLBG->mbButton = TRUE;
    
    BitmapPanel_cls *pCBitmapMaskBG = (BitmapPanel_cls*)pC->mVCItem[mIndexMaskBG].mpCWindowLink;
    pCBitmapMaskBG->mbButton = TRUE;

    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelOutput(wxWindow* pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, _("Output")/*, TRUE*/);

    ImageOutput_cls *pCImgOut = GetPCImageOutput();
    pC->mTextWidth = 120;

    pC->CombineSizer(5, _("Template"));
    wxString AstrTemplateType[4] = {"4:3", "16:9", "1:1", "2:1"};
    pC->SetChoiceList(4, AstrTemplateType);
    pC->CombineConfItem(pCImgOut->mCSizeType);

    for (int j = 0; j < 4; j++){
	//int sizeType = pCImgOut->mCSizeType.GetBodyINT();
	wxString AstrTemplate[_IMG_SIZE_MAX];
	for (int i = 0; i < _IMG_SIZE_MAX; i++) AstrTemplate[i] = GetStrGeom(j, i);
	pC->SetChoiceList(_IMG_SIZE_MAX, AstrTemplate);
	mIndexTemplate[j] = pC->CombineConfItem(pCImgOut->mCSizeIndex);
    }
    
    pC->CombineSizer(4, _("size"));
    mIndexDummyWidth = pC->CombineConfItem(mCDummyWidth, wxES, FALSE);
    mIndexDummyHeight = pC->CombineConfItem(mCDummyHeight, wxES, FALSE);
    mIndexWidth = pC->CombineConfItem(pCImgOut->mCWidth);
    mIndexHeight = pC->CombineConfItem(pCImgOut->mCHeight);

    
    pC->mTextWidth = 120;
    
    mIndexRenderPath = pC->CombineSizer(2);
    pC->CombineConfItem(pCImgOut->mCstrFileFormat);
    pC->CombineButton(btnPropRenderPath, "..", 30);

    pC->CombineSizer(2, _("ImageType"));
    wxString AstrChoiceFormat[3] = {"png", "jpg", "exr"};
    pC->SetChoiceList(3, AstrChoiceFormat);
    pC->CombineConfItem(pCImgOut->mCImageType);
    wxString AstrChoiceAlpha[2] = {"RGB", "RGBA"};
    pC->SetChoiceList(2, AstrChoiceAlpha);
    pC->CombineConfItem(pCImgOut->mCImageAlpha);
    
    wxString AstrChoiceStereo[2] = {_("Combined"), _("Separately")};
    pC->SetChoiceList(2, AstrChoiceStereo);
    mIndexStereoOutput = pC->CombineConfItem(pCImgOut->mCStereoOutput, _("Stereo Output"));

    pC->CombineSizer(1, _("Option"));
    pC->CombineConfItem(pCImgOut->mCbRenderFromWorkAreaIn, _("Render From Start Frame"));

    pC->FinishSetting();

    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelShadowVolume(wxWindow* pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, _("Shadow Volume")/*, TRUE*/);

    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
    pC->CombineSizer(1, _("Shadow Volume"));
    pC->CombineConfItem(pCSV->mCbUseShadowVolume, _("Use"));

    mIndexLengthSV = pC->CombineSizer(2, _("Length/Offset"));
    pC->CombineConfItem(pCSV->mCShadowLength);
    pC->SetLimit(0.00001, 0.01);
    pC->mDragScale = 0.001;
    pC->CombineConfItem(pCSV->mCOffsetFactor);

    mIndexOptionSV = pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(pCSV->mCbClipCheck, _("Check Clipping"));
    pC->CombineConfItem(pCSV->mCbForClosed, _("Closed Object Only"));
    
    //mIndexGammaSV = pC->CombineConfItem(pCSV->mCbUseGammaForSV, _("Gamma Correction"));

    mIndexTransparentShadow = pC->CombineSizer(1, " ");
    pC->CombineConfItem(pCSV->mCbTransparentShadowForEnsemble, _("Transp. Shadow"));
    
    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelMotionBlur(wxWindow* pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, _("Motion Blur")/*, TRUE*/);

    MotionBlurCtrl_cls *pCBlurCtrl = GetPCBlurCtrl();
    pC->CombineSizer(1, _("Motion Blur"));
    pC->CombineConfItem(pCBlurCtrl->mCbUseMotionBlur, _("Use"));

    mIndexBlur = pC->CombineSizer(2, _("Lenghth/Quality"));
    pC->SetLimit(0, 4);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(pCBlurCtrl->mCBlurLength);
    pC->SetLimit(1, 32);
    pC->mDragScale = 1;
    pC->CombineConfItem(pCBlurCtrl->mCBlurQuality);

    mIndexGammaBlur = pC->CombineConfItem(pCBlurCtrl->mCbUseGammaForBlur, _("Gamma Correction"));

    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelEnsemble(wxWindow* pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, _("Ensemble")/*, TRUE*/);

    EnsembleCtrl_cls *pCEnsembleCtrl = GetPCEnsembleCtrl();
    pC->CombineSizer(1, _("Ensemble"));
    pC->CombineConfItem(pCEnsembleCtrl->mCbUseEnsemble, _("Use"));

    mIndexSample = pC->CombineSizer(2, _("Sample Number"));
    pC->SetLimit(1, 511);
    pC->mDragScale = 4;
    pC->CombineConfItem(pCEnsembleCtrl->mCEnsembleSambleNum);//, _("Sample Number"));
    mIndexGammaEnsemble = pC->CombineConfItem(pCEnsembleCtrl->mCbUseGammaForEnsemble, _("Gamma Correction"));

    pC->FinishSetting();
    return pC;
}

void CLS::
Reconnect(){
}

void CLS::
HideAndShow(){
    bool bChange = FALSE;

    /*
    wxString AstrTemplate[_IMG_SIZE_MAX];
    for (int i = 0; i < _IMG_SIZE_MAX; i++) AstrTemplate[i] = GetStrGeom(sizeType, i);
    pC->SetChoiceList(_IMG_SIZE_MAX, AstrTemplate);
    pC->CombineConfItem(pCImgOut->mCSizeIndex);
    */

    ImageOutput_cls *pCImgOut = GetPCImageOutput();
    MultiInputPanel_cls *pCMNum = mpCPain->GetPCMNum(_("Output"));
    if (pCImgOut->mCSizeIndex.GetBodyINT() == 0){
	bChange |= pCMNum->mVCItem[mIndexDummyWidth].Hide();
	bChange |= pCMNum->mVCItem[mIndexDummyHeight].Hide();
	bChange |= pCMNum->mVCItem[mIndexWidth].Show();
	bChange |= pCMNum->mVCItem[mIndexHeight].Show();
	
    }
    else {
	bChange |= pCMNum->mVCItem[mIndexDummyWidth].Show();
	bChange |= pCMNum->mVCItem[mIndexDummyHeight].Show();
	bChange |= pCMNum->mVCItem[mIndexWidth].Hide();
	bChange |= pCMNum->mVCItem[mIndexHeight].Hide();
    }
    for (int i = 0; i < 4; i++){
	if (pCImgOut->mCSizeType.GetBodyINT() == i)
	    bChange |= pCMNum->mVCItem[mIndexTemplate[i]].Show();
	else
	    bChange |= pCMNum->mVCItem[mIndexTemplate[i]].Hide();
    }
    
    
    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
    MotionBlurCtrl_cls *pCBlurCtrl = GetPCBlurCtrl();
    EnsembleCtrl_cls *pCEnsembleCtrl = GetPCEnsembleCtrl();

    MultiInputPanel_cls *pCMNumSV = mpCPain->GetPCMNum(_("Shadow Volume"));
    bool bSV = pCSV->mCbUseShadowVolume.GetBodyBOOL();
    bChange |= pCMNumSV->mVCItem[mIndexLengthSV].Show(bSV);
    bChange |= pCMNumSV->mVCItem[mIndexOptionSV].Show(bSV);
    bChange |= pCMNumSV->mVCItem[mIndexTransparentShadow].Show(bSV);
    //bChange |= pCMNumSV->mVCItem[mIndexGammaSV].Show(bSV);

    MultiInputPanel_cls *pCMNumBlur = mpCPain->GetPCMNum(_("Motion Blur"));
    bool bBlur = pCBlurCtrl->mCbUseMotionBlur.GetBodyBOOL();
    bChange |= pCMNumBlur->mVCItem[mIndexBlur].Show(bBlur);
    bChange |= pCMNumBlur->mVCItem[mIndexGammaBlur].Show(bBlur);

    bool bEnsemble = pCEnsembleCtrl->mCbUseEnsemble.GetBodyBOOL();
    MultiInputPanel_cls *pCMNumEnsemble = mpCPain->GetPCMNum(_("Ensemble"));
    bChange |= pCMNumEnsemble->mVCItem[mIndexSample].Show(bEnsemble);
    bChange |= pCMNumEnsemble->mVCItem[mIndexGammaEnsemble].Show(bEnsemble);
    //bChange |= pCMNumEnsemble->mVCItem[mIndexTransparentShadow].Show(bEnsemble);

    if (bChange == TRUE){
	Layout();
	GetParent()->Layout();
	Refresh();
    }
}

bool CLS::
IsBGTransparent(){
    bool bBGTransparent = TRUE;
    if (GetPCColorProvider()->mCBGAlphaType.GetBodyINT() == 0) 
	bBGTransparent = FALSE;
    World_cls *pCWorld = (World_cls*)GetPCSceneObjCtrl()->GetPCWorld();
    if (pCWorld != NULL &&
	pCWorld->mpCFogProp->mCFogType.GetBodyINT() != _FOG_NONE &&
	pCWorld->mpCFogProp->mCbOverwriteBG.GetBodyBOOL() == TRUE)
	bBGTransparent = FALSE;

    return bBGTransparent;
}

void CLS::
SyncDummySize(){
    ImageOutput_cls *pCImgOut = GetPCImageOutput();
    
    long long width = -1, height = -1;
    GetGeometryFromIndex(pCImgOut->mCSizeType.GetBodyINT(), pCImgOut->mCSizeIndex.GetBodyINT(), width, height);
    if (width != -1){
	mCDummyWidth.SetBodyFromData(width);
	mCDummyHeight.SetBodyFromData(height);
	pCImgOut->mCWidth.SetBodyFromData(width);
	pCImgOut->mCHeight.SetBodyFromData(height);
    }
}

void CLS::
UpdateButtonColor(){

    MultiInputPanel_cls *pCMNumBG = mpCPain->GetPCMNum(_("BG"));
    pCMNumBG->mVCItem[mIndexGLBG].mpCWindowLink->SetBackgroundColour(GetPCColorProvider()->GetColor("GLBG"));
    pCMNumBG->mVCItem[mIndexMaskBG].mpCWindowLink->SetBackgroundColour(GetPCColorProvider()->GetColor("MaskBG"));
    
    if (GetPCTimeCtrl()->mbRecordRender == TRUE)
	mpCMNum->mVCItem[mIndexRenderAnim].mpCWindowLink->SetBackgroundColour(GetPCColorProvider()->GetColor("Active"));
    else
	mpCMNum->mVCItem[mIndexRenderAnim].mpCWindowLink->SetBackgroundColour(GetPCColorProvider()->GetColor("Null"));

    if (GetPCTimeCtrl()->mbRecordSnap == TRUE)
	mpCMNum->mVCItem[mIndexRenderSnap].mpCWindowLink->SetBackgroundColour(GetPCColorProvider()->GetColor("Active"));
    else
	mpCMNum->mVCItem[mIndexRenderSnap].mpCWindowLink->SetBackgroundColour(GetPCColorProvider()->GetColor("Null"));

    MultiInputPanel_cls *pCMNumOutput = mpCPain->GetPCMNum(_("Output"));
    ImageOutput_cls *pCImgOut = GetPCImageOutput();

    if (pCImgOut->mCstrFileFormat.DoesDirExist() == FALSE)
	pCMNumOutput->mVCItem[mIndexRenderPath+1].SetBackgroundColour(GetPCColorProvider()->GetColor("Error"));
    else
	pCMNumOutput->mVCItem[mIndexRenderPath+1].SetBackgroundColour(wxNullColour);
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    if (mpCMNum->GetName().IsSameAs(strPanelTag) ||
	GetName().IsSameAs(strPanelTag) || 
	strPanelTag.IsSameAs(wxT("ALL"))){
	HideAndShow();
	SyncDummySize();
	mpCMNum->UpdateValues(_MN_ALLITEM);
	UpdateButtonColor();
	if (mpCPain != NULL)
	    mpCPain->UpdateValues();
    }
}

void CLS::
OpenColorDialog(wxString strClrTag){
    wxColourDialogEX *pCDLG = GetPCColourDialog();

    pCDLG->SetColour(GetPCColorProvider()->GetColor(strClrTag));
    
    wxRect CRect(wxGetMousePosition()-wxPoint(250, 250), wxSize(545, 385));
    GetPCSizeAdjuster()->AdjustWindowFitting(pCDLG, CRect);

    if (pCDLG->ShowModal() == wxID_CANCEL) return;

    wxString strColor;

    strColor << (int)pCDLG->GetR(); strColor << " ";
    strColor << (int)pCDLG->GetG(); strColor << " ";
    strColor << (int)pCDLG->GetB();

    GetPCColorProvider()->SetColourCommand(strClrTag, strColor);

    UpdateButtonColor();
    GetPCMain()->UpdatePanels("VIEW3D");
}


void CLS::
OnGLBG(wxCommandEvent& event){
    OpenColorDialog("GLBG");
}

void CLS::
OnMaskBG(wxCommandEvent& event){
    OpenColorDialog("MaskBG");
}

void CLS::
OnRenderSnap(wxCommandEvent& event){
    ImageOutput_cls* pCImgOut = GetPCImageOutput();
    
    if (pCImgOut->CheckFileSequence() == FALSE){
	wxMessageBox(_("Filepath is invalid"), _("Warning"), wxICON_EXCLAMATION);
	return;
    }

    View3D_cls *pCView3DActive = GetPCFrameMain(0)->GetPCActiveView3D();
    if (pCView3DActive == NULL){
	wxMessageBox(_("No Active 3DView Panel"), _("Warning"), wxICON_ERROR);
	return;
    }

    if (IsBGTransparent() == TRUE &&
	pCImgOut->mCImageAlpha.GetBodyINT() == 0){
	if (wxMessageBox(_("BG is transparent, but File format is not RGBA. Continue?"), _("Warning"), wxYES_NO) == wxNO)
	    return;
    }

    if (pCView3DActive->IsStereogram() == TRUE &&
	pCImgOut->mCStereoOutput.GetBodyINT() == 0){
	if (wxMessageBox(_("Output combined stereogram ?"),
			 _("Warning"), wxYES_NO) == wxNO){
	    pCImgOut->mCStereoOutput.SetBodyFromData("1");
	    mpCPain->GetPCMNum(_("Output"))->UpdateValues(_MN_ALLITEM);
	}
    }

    GetPCTimeCtrl()->AnimCommand(TRUE);
    UpdateButtonColor();
}

void CLS::
OnRenderAnim(wxCommandEvent& event){
    ImageOutput_cls* pCImgOut = GetPCImageOutput();

    if (pCImgOut->CheckFileSequence() == FALSE){
	wxMessageBox(_("Filepath is invalid"), _("Warning"), wxOK_DEFAULT);
	return;
    }

    if (pCImgOut->mCFileSequence.mDigitLen == -1){
	if (wxMessageBox(_("Not Sequence. Append 4-digit tag ?"),
			 _("Warning"), wxYES_NO) == wxNO) return;
	pCImgOut->mCFileSequence.AppendDefaultSequenceTag();
	pCImgOut->mCFileSequence.InitFormat(0);
	mpCPain->GetPCMNum(_("Output"))->SetItemSTR(mIndexRenderPath+1, pCImgOut->mCFileSequence.GetFormat());
    }

    if (IsBGTransparent() == TRUE &&
	pCImgOut->mCImageAlpha.GetBodyINT() == 0){
	if (wxMessageBox(_("BG is transparent, but File format is not RGBA. Continue?"), _("Warning"), wxYES_NO) == wxNO)
	    return;
    }
    
    View3D_cls *pCView3DActive = GetPCFrameMain(0)->GetPCActiveView3D();
    if (pCView3DActive == NULL){
	wxMessageBox(_("No Active 3DView Panel"), _("Warning"), wxICON_ERROR);
	return;
    }

    if (pCView3DActive->IsStereogram() == TRUE &&
	pCImgOut->mCStereoOutput.GetBodyINT() == 0){
	if (wxMessageBox(_("Output combined stereogram ?"),
			 _("Warning"), wxYES_NO) == wxNO){
	    pCImgOut->mCStereoOutput.SetBodyFromData("1");
	    mpCPain->GetPCMNum(_("Output"))->UpdateValues(_MN_ALLITEM);
	}
    }
    
    GetPCTimeCtrl()->AnimCommand();
}

void CLS::
OnRenderPath(wxCommandEvent& event){
    wxFileDialogEX *pCDLG = GetPCFileDialog();

    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select output file path");
    pCDLG->SetWildcard("IMAGE");

    if (GetPCMain()->mCLastOpenedFileIMG.mstrBody.IsEmpty() == FALSE){
	wxFileName CName(GetPCMain()->mCLastOpenedFileIMG.mstrBody);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }

    pCDLG->mpCDialog->SetPosition(wxDP);
    pCDLG->mpCDialog->SetSize(wxDS);

    if (pCDLG->ShowModal() == wxID_OK){
	mpCPain->GetPCMNum(_("Output"))->SetItemSTR(mIndexRenderPath+1, pCDLG->GetPath());
    }

    pCDLG->DestroyDialog();
}

/*
void PropertyRender_cls::
OnKeyDown(wxKeyEvent& event){
    event.Skip();
}
*/

void CLS::
OnCopyColor(wxCommandEvent& event){
    int id = event.GetId();
    int buttonId = event.GetInt();

    wxString strColor, strSerialize;
    std::vector < float > VClr;
    if (buttonId == btnPropGLBG) VClr = GetPCColorProvider()->GetVGLColor("GLBG");
    if (buttonId == btnPropMaskBG) VClr = GetPCColorProvider()->GetVGLColor("MaskBG");
    
    wxString strTag = wxT("COLOR=");
    strSerialize << strTag << VClr[0] << " " << VClr[1] << " " << VClr[2];
    
    wxTextDataObject *pCData = new wxTextDataObject(strSerialize);
    wxTheClipboard->Open();
    wxTheClipboard->SetData(pCData);
    wxTheClipboard->Close();
}

void CLS::
OnPasteColor(wxCommandEvent& event){
    int id = event.GetId();
    int buttonId = event.GetInt();

    wxString strClrTag;
    if (buttonId == btnPropGLBG) strClrTag = "GLBG";
    if (buttonId == btnPropMaskBG) strClrTag = "MaskBG";

    if (wxTheClipboard->IsSupported(wxDataFormat(wxDF_TEXT)) == FALSE) return;
    wxTextDataObject CData;
    wxTheClipboard->Open();
    wxTheClipboard->GetData(CData);
    wxTheClipboard->Close();
    wxString strData = CData.GetText();

    if (strData.Matches(wxT("COLOR=*")) == TRUE){
	ConfItemFLOATARRAY_cls CArray(wxT("COLOR="), wxT("0.5 0.5 0.5 1"));
	CArray.ReadStrLine(strData);
	
	std::vector < float > VClr = CArray.GetBodyFLOATARRAY();
	if (VClr.size() != 3) return;
	
	wxString strColor;
	strColor << (int)(VClr[0] * 255); strColor << " ";
	strColor << (int)(VClr[1] * 255); strColor << " ";
	strColor << (int)(VClr[2] * 255);
	
	GetPCColorProvider()->SetColourCommand(strClrTag, strColor);

	UpdateButtonColor();
	GetPCMain()->UpdatePanels("VIEW3D");
    }
}
