//(setq c-basic-offset 4)
// "clsPropertyScene.h"

#include "clsPropertyScene.h"
#include "../clsProperty.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../../FILTER/clsBlur.h"
#include "../../FILTER/clsFocalBlur.h"
#include "../../FILTER/clsExposure.h"
#include "../../FILTER/clsFilterOverlay.h"
#include "../../FILTER/clsFilterLegend.h"
#include "../../FILTER/clsFilterCtrl.h"
#include "../../FILTER/clsGlare.h"
#include "../../FILTER/clsRemoveAlpha.h"
#include "../../FILTER/clsSSAO.h"
#include "../../FILTER/clsFilterDepth.h"

#include "../../WIDGETCTRL/EX/clsFileDialogEX.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"

#include "../../../clsMain.h"

#include <wx/filename.h>
#include <wx/fontdlg.h>

#include "../../COMMON/CommonDefinitions.h"
#include "../../../wxWidgetEnum.h"
#include "../../../wxWidgetEnumMenu.h"

#define CLS PropertyScene_cls

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
EVT_BUTTON(btnDropdown, CLS::OnDropDown)
EVT_BUTTON(btnColorSSAO, CLS::OnBtnColorSSAO)
EVT_BUTTON(btnAddFilter, CLS::OnAddFilter)
EVT_BUTTON(btnNumPanelDelete, CLS::OnDeleteFilter)
EVT_BUTTON(btnNumPanelValidity, CLS::OnToggleValidity)
EVT_BUTTON(btnOverlayPath, CLS::OnOverlayPath)
EVT_BUTTON(btnLegendTargetClear, CLS::OnLegendTargetClear)
EVT_BUTTON(btnLegendFont, CLS::OnLegendFont)
EVT_MENU(menuFilterExposure, CLS::OnAddExposure)
EVT_MENU(menuFilterBlur, CLS::OnAddBlur)
EVT_MENU(menuFilterFocalBlur, CLS::OnAddFocalBlur)
EVT_MENU(menuFilterGlare, CLS::OnAddGlare)
EVT_MENU(menuFilterRemoveAlpha, CLS::OnAddRemoveAlpha)
EVT_MENU(menuFilterOverlay, CLS::OnAddOverlay)
EVT_MENU(menuFilterLegend, CLS::OnAddLegend)
EVT_MENU(menuFilterSSAO, CLS::OnAddSSAO)
EVT_MENU(menuFilterDepth, CLS::OnAddFilterDepth)
EVT_MENU_RANGE(menuPop0, menuPop9, CLS::OnPopMenu)
END_EVENT_TABLE()

CLS::
PropertyScene_cls(Property_cls *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxPanelEX(pCParent, id, pos, size, style), PropertyBase_cls(pCParent)
{
    SetPanelType(_PANEL_PROPERTY_SCENE);

    SetToDefaultParam();

    mpCBoxSizerSrc  = new wxBoxSizer(wxVERTICAL);

    mpCBtnAddFilter = new wxButton(this, btnAddFilter, _("Add Filter"));
    
    mpCPain = GetPCNewPain(this, FALSE);
    mpCPain->mbDeleteButton = TRUE;
    mpCPain->mbInsertTarget = TRUE;
    
    mpCPainInRefuge = NULL;

    mpCBoxSizerSrc->Add(mpCBtnAddFilter, 0, wxEXPAND);
    mpCBoxSizerSrc->Add(mpCPain, 0, wxEXPAND);

    SetSizer(mpCBoxSizerSrc);
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();
    mCConfList.SetTag("PropertyScene");
}

void CLS::
ClearPanels(){
    if (mpCPain != NULL){
	mpCBoxSizerSrc->Detach(mpCPain);
	delete mpCPain;
	mpCPain = NULL;
    }

    mVpCMNum.clear();
    int numBuffer = mVpCAttributeBuffer.size(); 
    for (int i = 0; i < numBuffer; i++){
	wxString strAttrName = mVpCAttributeBuffer[i]->GetStrAttributeName();
	if (strAttrName.IsSameAs("Exposure_cls"))
	    delete (ExposureBuffer_cls*)mVpCAttributeBuffer[i];
	if (strAttrName.IsSameAs("Blur_cls"))
	    delete (BlurBuffer_cls*)mVpCAttributeBuffer[i];
	if (strAttrName.IsSameAs("FocalBlur_cls"))
	    delete (FocalBlurBuffer_cls*)mVpCAttributeBuffer[i];
	if (strAttrName.IsSameAs("Glare_cls"))
	    delete (GlareBuffer_cls*)mVpCAttributeBuffer[i];
	if (strAttrName.IsSameAs("RemoveAlpha_cls"))
	    delete (RemoveAlphaBuffer_cls*)mVpCAttributeBuffer[i];
	if (strAttrName.IsSameAs("Overlay_cls"))
	    delete (OverlayBuffer_cls*)mVpCAttributeBuffer[i];
	if (strAttrName.IsSameAs("Legend_cls"))
	    delete (LegendBuffer_cls*)mVpCAttributeBuffer[i];
	if (strAttrName.IsSameAs("SSAO_cls"))
	    delete (SSAOBuffer_cls*)mVpCAttributeBuffer[i];
	if (strAttrName.IsSameAs("FilterDepth_cls"))
	    delete (FilterDepthBuffer_cls*)mVpCAttributeBuffer[i];
    }
    mVpCAttributeBuffer.clear();
}

void CLS::
SavePanelInfoToRefuge(){
    if (mpCPain != NULL){
	mpCBoxSizerSrc->Detach(mpCPain);
	mpCPainInRefuge = mpCPain;
	mpCPain = NULL;
    }

    mVpCMNumInRefuge = mVpCMNum;
    mVpCMNum.clear();

    mVpCAttributeBufferInRefuge = mVpCAttributeBuffer;
    mVpCAttributeBuffer.clear();
}

void CLS::
ClearRefuge(){
    if (mpCPainInRefuge != NULL){
	delete mpCPainInRefuge;
	mpCPainInRefuge = NULL;
    }

    mVpCMNumInRefuge.clear();

    int numBuffer = mVpCAttributeBufferInRefuge.size(); 
    for (int i = 0; i < numBuffer; i++){
	wxString strAttrName = mVpCAttributeBufferInRefuge[i]->GetStrAttributeName();
	if (strAttrName.IsSameAs("Exposure_cls"))
	    delete (ExposureBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	if (strAttrName.IsSameAs("Blur_cls"))
	    delete (BlurBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	if (strAttrName.IsSameAs("FocalBlur_cls"))
	    delete (FocalBlurBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	if (strAttrName.IsSameAs("Glare_cls"))
	    delete (GlareBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	if (strAttrName.IsSameAs("RemoveAlpha_cls"))
	    delete (RemoveAlphaBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	if (strAttrName.IsSameAs("Overlay_cls"))
	    delete (OverlayBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	if (strAttrName.IsSameAs("Legend_cls"))
	    delete (LegendBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	if (strAttrName.IsSameAs("SSAO_cls"))
	    delete (SSAOBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	if (strAttrName.IsSameAs("FilterDepth_cls"))
	    delete (FilterDepthBuffer_cls*)mVpCAttributeBufferInRefuge[i];
	
    }
    mVpCAttributeBufferInRefuge.clear();
}

MultiInputPanel_cls* CLS::
GetPCMNumForReuse(Attribute_cls *pCAttrib){
    wxString strAttribName = pCAttrib->GetStrName();
    uint numAttribInRefuge = mVpCMNumInRefuge.size();
    for (uint i = 0; i < numAttribInRefuge; i++){
	if (mVpCMNumInRefuge[i]->GetName().IsSameAs(strAttribName)){
	    Attribute_cls *pCBuffer = mVpCAttributeBufferInRefuge[i];
	    MultiInputPanel_cls *pCMNum = mVpCMNumInRefuge[i];
	    
	    mVpCMNum.push_back(pCMNum);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	    
	    mVpCAttributeBufferInRefuge.erase(mVpCAttributeBufferInRefuge.begin() + i);
	    mVpCMNumInRefuge.erase(mVpCMNumInRefuge.begin() + i);
	    return pCMNum;
	}
    }
    
    return NULL;
}

std::vector < bool > CLS::
GetVbValidity(){
    std::vector < bool > VbValidity;

    int numFilter = GetPCFilterCtrl()->mVpCFilter.size();
    
    for (int i = 0; i < numFilter; i++)
	VbValidity.push_back(GetPCFilterCtrl()->GetPCFromIndex(i)->mCbValidity.GetBodyBOOL());
    
    return VbValidity;
}


void CLS::
Disconnect(){
    ClearPanels();
    mpCBtnAddFilter->Disable();

    Layout();
}

void CLS::
Reconnect(){
    GetParent()->Freeze();

    //ClearPanels();
    SavePanelInfoToRefuge();

    wxSize CSize = wxDS;
    if (mpCPain == NULL){
	mpCPain = new MNumPainBase_cls(this, CSize, FALSE);
	mpCPain->mbHideButton = FALSE;
	mpCPain->mbDeleteButton = TRUE;
	mpCPain->mbInsertTarget = TRUE;
	mpCPain->mstrTag = GetPCFilterCtrl()->GetStrSelf();
	mAstrClosedItem.clear();
    }
    
    int numFilter = GetPCFilterCtrl()->mVpCFilter.size();
    for (int i = 0; i < numFilter; i++){
	FilterBase_cls *pCFilter = (FilterBase_cls*)GetPCFilterCtrl()->GetPCFromIndex(i);
	MultiInputPanel_cls *pCReuse = GetPCMNumForReuse(pCFilter);
	if (pCReuse != NULL){
	    ASSERT(mpCPainInRefuge != NULL);
	    mpCPainInRefuge->GetSizer()->Detach(pCReuse);
	    pCReuse->Reparent(mpCPain);
	    //!! etւ
	    mpCPain->SetMNum(pCReuse, mAstrClosedItem, pCReuse->GetName(), pCFilter->mCbValidity.GetBodyBOOL(), TRUE);
	    continue;
	}

	MultiInputPanel_cls *pCMNum = NULL;
	wxString strFilterName;
	if (pCFilter->IsAttributeSameAs("Exposure_cls")){
	    Exposure_cls *pCExposure = (Exposure_cls*)(pCFilter);
	    ExposureBuffer_cls *pCBuffer = new ExposureBuffer_cls(pCExposure);
	    
	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);
	    
	    pCMNum->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_TEXT_ENTER, wxID_ANY, pCExposure->GetStrName());
	    //mVpCMNum.push_back(pCMNum);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	    
	    //mpCPain->SetMNum(pCMNum, mAstrClosedItem, pCExposure->GetStrName(), pCFilter->mCbValidity.GetBodyBOOL(), TRUE);
	}

	if (pCFilter->IsAttributeSameAs("Blur_cls")){
	    Blur_cls *pCBlur = (Blur_cls*)(pCFilter);
	    BlurBuffer_cls *pCBuffer = new BlurBuffer_cls(pCBlur);

	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	}

	if (pCFilter->IsAttributeSameAs("FocalBlur_cls")){
	    FocalBlur_cls *pCFocalBlur = (FocalBlur_cls*)(pCFilter);
	    FocalBlurBuffer_cls *pCBuffer = new FocalBlurBuffer_cls(pCFocalBlur);

	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	}

	if (pCFilter->IsAttributeSameAs("Glare_cls")){
	    Glare_cls *pCGlare = (Glare_cls*)(pCFilter);
	    
	    GlareBuffer_cls *pCBuffer = new GlareBuffer_cls(pCGlare);
	    
	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);

	    //pCMNum->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_TEXT_ENTER, wxID_ANY, pCGlare->GetStrName());
	    //pCMNum->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_COMBOBOX_SELECTED, wxID_ANY, pCGlare->GetStrName());
	    //mVpCMNum.push_back(pCMNum);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	    
	    //mpCPain->SetMNum(pCMNum, mAstrClosedItem, pCGlare->GetStrName(), pCFilter->mCbValidity.GetBodyBOOL(), TRUE);
	}

	if (pCFilter->IsAttributeSameAs("RemoveAlpha_cls")){
	    RemoveAlpha_cls *pCRemoveAlpha = (RemoveAlpha_cls*)(pCFilter);
	    
	    RemoveAlphaBuffer_cls *pCBuffer = new RemoveAlphaBuffer_cls(pCRemoveAlpha);
	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	}

	if (pCFilter->IsAttributeSameAs("Overlay_cls")){
	    Overlay_cls *pCOverlay = (Overlay_cls*)(pCFilter);
	    OverlayBuffer_cls *pCBuffer = new OverlayBuffer_cls(pCOverlay);
	    
	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	}

	if (pCFilter->IsAttributeSameAs("Legend_cls")){
	    Legend_cls *pCLegend = (Legend_cls*)(pCFilter);
	    LegendBuffer_cls *pCBuffer = new LegendBuffer_cls(pCLegend);
	    
	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	}

	if (pCFilter->IsAttributeSameAs("SSAO_cls")){
	    SSAO_cls *pCSSAO = (SSAO_cls*)(pCFilter);
	    SSAOBuffer_cls *pCBuffer = new SSAOBuffer_cls(pCSSAO);
	    
	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	}

	if (pCFilter->IsAttributeSameAs("FilterDepth_cls")){
	    FilterDepth_cls *pCFilterDepth = (FilterDepth_cls*)(pCFilter);
	    FilterDepthBuffer_cls *pCBuffer = new FilterDepthBuffer_cls(pCFilterDepth);
	    
	    pCMNum = pCBuffer->GetPCNewPanel(mpCPain);
	    mVpCAttributeBuffer.push_back(pCBuffer);
	}

	if (pCMNum != NULL){ // ʏ
	    strFilterName = pCFilter->GetStrName();
	    pCMNum->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_TEXT_ENTER, wxID_ANY, strFilterName);
	    pCMNum->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_COMBOBOX_SELECTED, wxID_ANY, strFilterName);
	    pCMNum->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_CHECKBOX_CLICKED, wxID_ANY, strFilterName);
	    mVpCMNum.push_back(pCMNum);
	    mpCPain->SetMNum(pCMNum, mAstrClosedItem, strFilterName, pCFilter->mCbValidity.GetBodyBOOL(), TRUE);
	}
	
    }
    
    wxBoxSizer *pCSizerForPain = new wxBoxSizer(wxVERTICAL);
    mpCPain->InitLayoutNumPanels(pCSizerForPain);
    mpCPain->SetSizer(pCSizerForPain);

    ClearRefuge();

    uint numPanel = mVpCMNum.size();
    for (uint i = 0; i < numPanel; i++)
	mVpCMNum[i]->UpdateValues(_MN_ALLITEM);
    
    mpCBoxSizerSrc->Add(mpCPain, 0, wxEXPAND);
    Layout();

    GetParent()->Thaw();

    wxSize CCurrentSize = GetSize();
    wxSize CBestSize = GetBestSize();
    SetSize(wxSize(-1, CBestSize.GetHeight()));
    //GetParent()->Layout();
    //GetParent()->PostSizeEvent();
}


void CLS::
SetAttributeInfo(){
    uint numBuffer = mVpCAttributeBuffer.size();
    for (uint i = 0; i < numBuffer; i++){
	wxString strAttribName = mVpCAttributeBuffer[i]->GetStrAttributeName();
	if (strAttribName.IsSameAs("Exposure_cls"))
	    ((ExposureBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
	if (strAttribName.IsSameAs("Blur_cls"))
	    ((BlurBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
	if (strAttribName.IsSameAs("FocalBlur_cls"))
	    ((FocalBlurBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
	if (strAttribName.IsSameAs("Glare_cls"))
	    ((GlareBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
	if (strAttribName.IsSameAs("RemoveAlpha_cls"))
	    ((RemoveAlphaBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
	if (strAttribName.IsSameAs("Overlay_cls"))
	    ((OverlayBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
	if (strAttribName.IsSameAs("Legend_cls"))
	    ((LegendBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
	if (strAttribName.IsSameAs("SSAO_cls"))
	    ((SSAOBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
	if (strAttribName.IsSameAs("FilterDepth_cls"))
	    ((FilterDepthBuffer_cls*)mVpCAttributeBuffer[i])->SyncFromSrc();
    }
}

void CLS::
ProcAfterMathScene(wxString strCommand, wxString strTag){
    if (strCommand.IsSameAs("FilterChange") ||
	strCommand.IsSameAs("DeleteFilter")){
	Reconnect();
    }
}

void CLS::
CheckPanelOrder(){
    uint numPanel = mVpCMNum.size();
    uint numFilter = GetPCFilterCtrl()->mVpCFilter.size();
    ASSERT(numPanel == numFilter);
    bool bDiff = FALSE;
    for (uint i = 0; i < numPanel; i++){
	if (mVpCMNum[i]->GetName().IsSameAs(GetPCFilterCtrl()->GetPCFromIndex(i)->GetStrName()) == FALSE)
	    bDiff = TRUE;
    }
    if (bDiff == TRUE) Reconnect();
}

void CLS::
UpdateButtonColor(){
    uint numBuffer = mVpCAttributeBuffer.size();
    for (uint i = 0; i < numBuffer; i++){
	wxString strAttribName = mVpCAttributeBuffer[i]->GetStrAttributeName();
	if (strAttribName.IsSameAs("SSAO_cls"))
	    ((SSAOBuffer_cls*)mVpCAttributeBuffer[i])->UpdateButtonColor(mVpCMNum[i]);
    }
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs("ALL")){
	CheckPanelOrder();
	SetAttributeInfo();
	/*
	uint numPanel = mVpCMNum.size();
	for (uint i = 0; i < numPanel; i++)
	    mVpCMNum[i]->UpdateValues(_MN_ALLITEM);
	*/
	if (mpCPain != NULL){
	    mpCPain->SyncValidityToData(GetVbValidity());
	    mpCPain->UpdateValues();
	}
	UpdateButtonColor();
    }
    
}

void CLS::
OnAttributeChange(wxComEvt& event){

    wxString strName = event.GetString();

    uint numAttribBuffer = mVpCAttributeBuffer.size();
    for (uint i = 0; i < numAttribBuffer; i++){
	Attribute_cls *pCAttrib = mVpCAttributeBuffer[i];
	if (pCAttrib->GetStrName().IsSameAs(strName)){
	    wxString strAttribName = pCAttrib->GetStrAttributeName();
	    if (strAttribName.IsSameAs("Exposure_cls"))
		((ExposureBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	    if (strAttribName.IsSameAs("Blur_cls"))
		((BlurBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	    if (strAttribName.IsSameAs("FocalBlur_cls"))
		((FocalBlurBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	    if (strAttribName.IsSameAs("Glare_cls"))
		((GlareBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	    if (strAttribName.IsSameAs("RemoveAlpha_cls"))
		((RemoveAlphaBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	    if (strAttribName.IsSameAs("Overlay_cls"))
		((OverlayBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	    if (strAttribName.IsSameAs("Legend_cls"))
		((LegendBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	    if (strAttribName.IsSameAs("SSAO_cls"))
		((SSAOBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	    if (strAttribName.IsSameAs("FilterDepth_cls"))
		((FilterDepthBuffer_cls*)pCAttrib)->ForceSrcToSyncBuffer();
	}
    }
}

void CLS::
OnDropDown(wxComEvt& event){
    wxString strCommand = event.GetString();
    wxString strType = "ALL";
    //bool bSolidOnly = FALSE;
    strType = "Particle_cls||VolumeBox_cls||HeightField_cls||Streamline_cls";
    mCSelector.PopupObjectSelector(this, strCommand, strType);
}

void CLS::
OnBtnColorSSAO(wxComEvt& event){
    wxString strName = event.GetString();
    SSAO_cls *pCSSAO = (SSAO_cls*)(GetPCFilterCtrl()->GetPCFromName(strName));
    if (pCSSAO != NULL){
	std::vector < float > VClr = GetColorFromDialog(pCSSAO->GetWXColor());
	if (VClr.size() == 0) return;
	
	pCSSAO->SetBaseColor(VClr);
    }
}

void CLS::
OnAddFilter(wxComEvt& event){
    wxMenu *pCPopMenu = new wxMenu;

    pCPopMenu->Append(menuFilterExposure, _("Exposure"));
    pCPopMenu->Append(menuFilterBlur, _("Blur"));
    pCPopMenu->Append(menuFilterFocalBlur, _("Focal Blur"));
    pCPopMenu->Append(menuFilterGlare, _("Glare"));
    pCPopMenu->Append(menuFilterRemoveAlpha, _("Remove Alpha"));
    pCPopMenu->Append(menuFilterOverlay, _("Overlay"));
    pCPopMenu->Append(menuFilterLegend, _("Legend"));
    pCPopMenu->Append(menuFilterSSAO, _("SSAO"));
    pCPopMenu->Append(menuFilterDepth, _("Depth Buffer"));
    PopupMenu(pCPopMenu);

    delete pCPopMenu;
}

void CLS::
OnAddExposure(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("Exposure_cls", "Exposure"); }

void CLS::
OnAddBlur(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("Blur_cls", "Blur"); }

void CLS::
OnAddFocalBlur(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("FocalBlur_cls", "FocalBlur"); }

void CLS::
OnAddGlare(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("Glare_cls", "Glare"); }

void CLS::
OnAddRemoveAlpha(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("RemoveAlpha_cls", "Remove Alpha"); }

void CLS::
OnAddOverlay(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("Overlay_cls", "Overlay"); }

void CLS::
OnAddLegend(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("Legend_cls", "Legend"); }

void CLS::
OnAddSSAO(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("SSAO_cls", "SSAO"); }

void CLS::
OnAddFilterDepth(wxComEvt& event)
{ GetPCFilterCtrl()->AddFilterCommand("FilterDepth_cls", "Depth Buffer"); }

void CLS::
OnDeleteFilter(wxComEvt& event){
    wxString strTgt = event.GetString();
    GetPCFilterCtrl()->DeleteFilterCommand(strTgt);
}

void CLS::
OnToggleValidity(wxComEvt& event){
    wxString strTgt = event.GetString();
    int state = event.GetInt();

    GetPCFilterCtrl()->SetFilterValidityCommand(strTgt, state == 1);
}

void CLS::
OnOverlayPath(wxComEvt& event){
    wxString strName = event.GetString();
    wxFileDialogEX *pCDLG = GCClassProvider.GetPCFileDialog();

    Overlay_cls* pCOverlay = (Overlay_cls*)(GetPCFilterCtrl()->GetPCFromName(strName));
    if (pCOverlay == NULL) return;

    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select Image File (Sequence)");
    pCDLG->SetWildcard("IMAGE");

    if (pCOverlay->mstrPathChecked.IsEmpty() == TRUE ||
	pCOverlay->mstrPathChecked.IsSameAs("None")){
	wxFileName CName(pCOverlay->mstrPathChecked);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    else if (GetPCMain()->mCLastOpenedFileIMG.mstrBody.IsEmpty() == FALSE){
	wxFileName CName(GetPCMain()->mCLastOpenedFileIMG.mstrBody);
	pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
    }
    
    pCDLG->mpCDialog->SetPosition(wxDP);
    pCDLG->mpCDialog->SetSize(wxDS);

    if (pCDLG->ShowModal() == wxID_OK)
	pCOverlay->mCstrPath.PushCommandPair(pCDLG->GetPath());

    pCDLG->DestroyDialog();
}

void CLS::
OnLegendTargetClear(wxComEvt& event){
    wxString strEventFrom = event.GetString();
    Legend_cls* pCLegend = (Legend_cls*)(GetPCFilterCtrl()->GetPCFromName(strEventFrom));
    if (pCLegend == NULL) return;

    pCLegend->mCstrTarget.PushCommandPair("None");
}


void CLS::
OnLegendFont(wxComEvt& event){
    wxString strEventFrom = event.GetString();
    Legend_cls* pCLegend = (Legend_cls*)(GetPCFilterCtrl()->GetPCFromName(strEventFrom));
    if (pCLegend == NULL) return;

    wxFont CFont;
    CFont.SetNativeFontInfoUserDesc(pCLegend->mCstrFontDesc.GetBodySTR());

    CFont = wxGetFontFromUser(this, CFont, wxES);

    if (CFont.IsOk() == TRUE)
	pCLegend->mCstrFontDesc.PushCommandPair(CFont.GetNativeFontInfoUserDesc());
}

void CLS::
OnPopMenu(wxComEvt& event){
    wxString strPopMenu = mCSelector.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == TRUE) return;
    
    wxString strTag = mCSelector.mstrTag;
    if (strTag.IsEmpty() == FALSE){
	
	Attribute_cls* pCAttrib = GetPCFilterCtrl()->GetPCFromName(strTag);
	
	if (pCAttrib != NULL){
	    if (pCAttrib->IsAttributeSameAs("Legend_cls"))
		((Legend_cls*)pCAttrib)->SelectObject(strPopMenu);
	    return;
	}
    }
    
    //mpCParent->SelectObject(strPopMenu);
}
