//(setq c-basic-offset 4)
// "clsProperty.h"

#include "clsProperty.h"

#include "PROPERTY/clsPropertyRender.h"
#include "PROPERTY/clsPropertyScene.h"
#include "PROPERTY/clsPropertyObject.h"
#include "PROPERTY/clsPropertyMaterial.h"
#include "PROPERTY/clsPropertyModifier.h"
#include "PROPERTY/clsPropertyFormat.h"
#include "PROPERTY/clsPropertyInfo.h"
#include "../PREFERENCE/clsShortcut.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../WIDGETCTRL/EX/clsPanelBar.h"
#include "../WIDGETCTRL/INPUTITEM/clsIconSelector.h"
#include "../WIDGETCTRL/INPUTITEM/clsTextButton.h"
#include "../WIDGETCTRL/INPUTITEM/clsVCenter.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"

#include "../COMMON/CommonDefinitions.h"

#include "../../wxWidgetEnumMenu.h"

#define CLS Property_cls

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
//EVT_CHAR (CLS::OnKeyDown)
EVT_BUTTON(menuPanelView, CLS::OnMenuView)
EVT_LEFT_DOWN(CLS::OnLeftDown)
END_EVENT_TABLE()

CLS::
Property_cls(wxWindow *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxPanelEX(pCParent, id, pos, size, style) {

    SetPanelType(_PANEL_PROPERTY);

    Freeze();

    SetBackgroundColour(wxColour(128, 96, 96));

    SetDefault();

    mpCPanelBar = new PanelBar_cls(this);
    mpCPanelBar->Freeze();
    mpCPanelBar->SetPanelType(_PANEL_CHILD, _PANEL_PROPERTY);

    wxBoxSizer *pBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerBar = new wxBoxSizer(wxHORIZONTAL);

    ConstructTypeSelector(pBoxSizerBar);
    
    mpCTextMenuView = new TextButton_cls
	(mpCPanelBar, menuPanelView, GetPCStringProvider()->GetStrMenu("View"));
    mpCTextMenuView->SetEventTarget(this);
    VCenter_cls *pCMenuView = new VCenter_cls(mpCTextMenuView);

    pBoxSizerBar->Add(pCMenuView, 0, wxEXPAND);
    pBoxSizerBar->AddStretchSpacer(1);
    mpCPanelBar->ConstructBarItemDefault(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemHideShow(pBoxSizerBar);

    mpCPanelBar->SetSizer(pBoxSizerBar);

    std::vector < wxString > VstrArt;
    VstrArt.push_back("wxART_RENDER");
    VstrArt.push_back("wxART_FILTER");
    VstrArt.push_back("wxART_OBJECT_SQUARE");
    VstrArt.push_back("wxART_TEXTURE");
    VstrArt.push_back("wxART_MODIFIER");
    VstrArt.push_back("wxART_EXECUTE");
    VstrArt.push_back("wxART_TEXT_ICON");
    VstrArt.push_back("wxART_INFO");
  
    mpCIconSelector = new IconSelector_cls
	(this, wxID_ANY, 0, /*3,*/VstrArt, wxDP, wxSize(-1, 40));
    mpCIconSelector->SetItemSize(wxSize(36, 36));
    mpCIconSelector->SetBackgroundColour(wxColour(188, 218, 254));

    mpCIconSelector->mVstrTag.push_back("render");
    mpCIconSelector->mVstrTag.push_back("scene filter");
    mpCIconSelector->mVstrTag.push_back("object");
    mpCIconSelector->mVstrTag.push_back("material");
    mpCIconSelector->mVstrTag.push_back("modifier");
    mpCIconSelector->mVstrTag.push_back("data process");
    mpCIconSelector->mVstrTag.push_back("format");
    mpCIconSelector->mVstrTag.push_back("information");
    
    mpCRender = new PropertyRender_cls(this);
    mpCScene = new PropertyScene_cls(this);
    mpCObject = new PropertyObject_cls(this);
    mpCMaterial = new PropertyMaterial_cls(this);
    mpCModifier = new PropertyModifier_cls(this, _MOD_MODIFIER);
    mpCDataProc = new PropertyModifier_cls(this, _MOD_DATAPROC);
    mpCFormat = new PropertyFormat_cls(this);
    mpCInfo = new PropertyInfo_cls(this);

    //pBoxSizerBar->Add(pCDummy, 0, wxEXPAND);
    //mpCPanelBar->SetSizer(pBoxSizerBar);

    pBoxSizerSrc->Add(mpCPanelBar, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCIconSelector, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCRender, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCScene, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCObject, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCMaterial, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCModifier, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCDataProc, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCFormat, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCInfo, 0, wxEXPAND);
    
    pBoxSizerSrc->AddSpacer(1);
    
    SetSizer(pBoxSizerSrc);

    Connect(mpCIconSelector->GetId(), wxEVT_COMMAND_COMBOBOX_SELECTED, 
	    wxCommandEventHandler(CLS::OnSelector));

    HideAndShow();

    mpCPanelBar->Thaw();
    Thaw();
}

void CLS::
SetDefault(){
    mCConfList.Clear();
    mCConfList.SetTag(wxT("PropertyPanel"));
    mCConfList.Add(&mCSelectorMode, wxT("Selector="), wxT("0"));
}

void CLS::
HideAndShow(){
    int mode = mpCIconSelector->mValue;

    //Freeze();
    if (mpCRender->IsShown()) mpCRender->Hide();
    if (mpCScene->IsShown()) mpCScene->Hide();
    if (mpCObject->IsShown()) mpCObject->Hide();
    if (mpCMaterial->IsShown()) mpCMaterial->Hide();
    if (mpCModifier->IsShown()) mpCModifier->Hide();
    if (mpCDataProc->IsShown()) mpCDataProc->Hide();
    if (mpCFormat->IsShown()) mpCFormat->Hide();
    if (mpCInfo->IsShown()) mpCInfo->Hide();

    switch(mode){
    case _PROPERTY_RENDER:
	mpCRender->Show(); break;
    case _PROPERTY_SCENE:
	mpCScene->Show(); break;
    case _PROPERTY_OBEJCT:
	mpCObject->Show(); break;
    case _PROPERTY_MATERIAL:
	mpCMaterial->Show(); break;
    case _PROPERTY_MODIFIER:
	mpCModifier->Show(); break;
    case _PROPERTY_DATAPROC:
	mpCDataProc->Show(); break;
    case _PROPERTY_FORMAT:
	mpCFormat->Show(); break;
    case _PROPERTY_INFORMATION:
	mpCInfo->Show(); break;
    }
    
    Layout();
    //mpCModifier->Layout();//NumPanels();
    //mpCModifier->FitInside();
    //Thaw();
    PostSizeEvent();
}

void CLS::
SelectObject(wxString strObj){
    wxString strCommand = "SelectObject";
    wxString strTag;
    strTag << GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj;
    ProcAfterMath(strCommand, strTag, TRUE);
}

void CLS::
ProcAfterMath(const wxString& strCommand, wxString strTag, bool bForce){
    	
    if (strCommand.IsSameAs("DeleteObject") ||
	strCommand.IsSameAs("RenameObject") ||
	strCommand.IsSameAs("SelectObject")){
	
	mpCObject->ProcAfterMathProperty(strCommand, strTag, bForce);
	mpCMaterial->ProcAfterMathProperty(strCommand, strTag, bForce);
	mpCModifier->ProcAfterMathProperty(strCommand, strTag, bForce);
	mpCDataProc->ProcAfterMathProperty(strCommand, strTag, bForce);
	mpCFormat->ProcAfterMathProperty(strCommand, strTag, bForce);
	mpCInfo->ProcAfterMathProperty(strCommand, strTag, bForce);
	// 上の一連の処理で新しいパネルに付け替えるので、
	// Freeze-Thaw の間に挟めない(LINUX)
	Freeze(); 
	Layout();
	PostSizeEvent();
	Thaw();
    }
    if (strCommand.IsSameAs("ModifierChange"))
	mpCModifier->ProcAfterMathModifier(strCommand, strTag);
    if (strCommand.IsSameAs("DataProcChange"))
	mpCDataProc->ProcAfterMathModifier(strCommand, strTag);
    
    if (strCommand.IsSameAs("FilterChange") ||
	strCommand.IsSameAs("DeleteFilter"))
	mpCScene->ProcAfterMathScene(strCommand, strTag);

    if (strCommand.IsSameAs("ModifierChange") ||
	strCommand.IsSameAs("FilterChange") ||
	strCommand.IsSameAs("DeleteFilter")){
	Layout();
	PostSizeEvent();
    }
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    mpCRender->UpdatePanels(strPanelTag);
    mpCScene->UpdatePanels(strPanelTag);
    mpCObject->UpdatePanels(strPanelTag);
    mpCMaterial->UpdatePanels(strPanelTag);
    mpCModifier->UpdatePanels(strPanelTag);
    mpCDataProc->UpdatePanels(strPanelTag);
    mpCFormat->UpdatePanels(strPanelTag);
    mpCInfo->UpdatePanels(strPanelTag);
}

void CLS::
OnMenuView(wxCommandEvent& event){
    wxMenu CMenu;
    bool bBarItem = mpCPanelBar->mbShowItem;
    CMenu.AppendCheckItem(menuShowPanelManage, GetPCShortcut()->GetMenuString(menuShowPanelManage))->Check(bBarItem);
    
    mpCTextMenuView->PopMenu(this, &CMenu);
}

/*
void CLS::
OnKeyDown(wxKeyEvent& event){
    event.Skip();
}
*/

void CLS::
OnSelector(wxCommandEvent& event){
    HideAndShow();
}

void CLS::
Serialize(wxString& strOut){
    mCSelectorMode.SetBodyFromData((long long)mpCIconSelector->mValue);
    ConfListContainer_cls::Serialize(strOut);
}

void CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == FALSE)
	return;
    mpCIconSelector->mValue = mCSelectorMode.GetBodyINT();
    HideAndShow();
}
