// -*- c++ -*- 
// "clsRibbonUI.h"
// (setq c-basic-offset 4)

#include "clsRibbonUI.h"

#include <wx/wx.h>
#include <wx/tooltip.h>
#include <wx/ribbon/bar.h>


#include "../PREFERENCE/clsShortcut.h"
#include "../PREFERENCE/clsPreference.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#include "../COMMON/CommonDefinitions.h"
#include "../../wxWidgetEnum.h"

RibbonPage_cls::
RibbonPage_cls(wxRibbonBar* pCParentBar, 
	       wxString strTab, std::vector < wxString > VstrPanels) :
    wxRibbonPage(pCParentBar, wxID_ANY, strTab)
{
    
    for (uint i = 0; i < VstrPanels.size(); i++){
	mVpCRbnPanel.push_back
	    (new wxRibbonPanel(this, wxID_ANY, VstrPanels[i],
			       wxNullBitmap, wxDPS, wxRIBBON_PANEL_NO_AUTO_MINIMISE));
	mVpCRbnToolBar.push_back
	    (new wxRibbonToolBar(mVpCRbnPanel[i], wxID_ANY));
    }
}

#define CLS RibbonUI_cls

CLS::
RibbonUI_cls(wxWindow *pCParentLink){
    mpCParentLink = pCParentLink;

    mpCRbnBar = new wxRibbonBar(mpCParentLink, wxID_ANY,  wxDPS,
				wxRIBBON_BAR_DEFAULT_STYLE |
				wxRIBBON_BAR_SHOW_PANEL_EXT_BUTTONS);
    std::vector < wxString > VstrHome;
    std::vector < wxString > VstrHelp;

    VstrHome.push_back(_("Home"));
    VstrHome.push_back(_("Data"));
    VstrHome.push_back(_("Object"));
    VstrHome.push_back(_("Preference"));

    VstrHelp.push_back(_("Help"));
    

    mVpCRibbonPage.push_back
	(new RibbonPage_cls(mpCRbnBar, _("Home"), VstrHome));
    mVpCRibbonPage.push_back
	(new RibbonPage_cls(mpCRbnBar, _("Help"), VstrHelp));

    
    
    wxSize sz = wxSize(36, 36);
    
    mCurrentTabIndex   = _RBN_HOME;
    mCurrentPanelIndex = _RBN_PANEL_HOME;
    GetPToolBar()->AddTool(rBtnOpenProject, ARTS("wxART_OPEN_FILE", sz));
    GetPToolBar()->AddHybridTool(rBtnOpenRecent, ARTS("wxART_OPEN_HISTORY", sz));
    GetPToolBar()->AddHybridTool(rBtnSaveProject, ARTS("wxART_SAVE_FILE", sz));
    GetPToolBar()->AddSeparator();
    GetPToolBar()->AddTool(rBtnScene, ARTS("wxART_SCENE", sz));
    GetPToolBar()->AddSeparator();
    GetPToolBar()->AddTool(rBtnUndo, ARTS("wxART_UNDO", sz));
    GetPToolBar()->AddTool(rBtnRedo, ARTS("wxART_REDO", sz));

    // V[gJbgL[ݒ肳ĂȂ{^HelpStringݒ
    // L[ݒ肪ꍇ̓V[gJbgŃwv񂪏Ă̂
    // Őݒ肵ȂĂ悢B

    mCurrentPanelIndex = _RBN_PANEL_DATA;
    wxString strAdd = _("Add") + ":";
    AddTool(rBtnVolumeBox,   "wxART_VOLUMEBOX",   sz, strAdd+_("Volume Box"));
    AddTool(rBtnIsosurface,  "wxART_ISOSURFACE",  sz, strAdd+_("Isosurface"));
    AddTool(rBtnParticle,    "wxART_PARTICLE",    sz, strAdd+_("Particle"));
    AddTool(rBtnStreamline,  "wxART_STREAMLINE",  sz, strAdd+_("Streamline"));
    AddTool(rBtnHeightField, "wxART_HEIGHTFIELD", sz, strAdd+_("Height Field"));
    
    mCurrentPanelIndex = _RBN_PANEL_OBJECT;
    GetPToolBar()->AddHybridTool(rBtnObjects, ARTS("wxART_NEW_OBJ", sz));
    GetPToolBar()->AddTool(rBtnDuplicate, ARTS("wxART_DUPLICATE", sz));
    GetPToolBar()->AddTool(rBtnDelete, ARTS("wxART_CLEAR", sz));

    mCurrentPanelIndex = _RBN_PANEL_PREFERENCE;
    GetPToolBar()->AddTool(rBtnPreference, ARTS("wxART_PREFERENCE", sz));
    GetPToolBar()->AddHybridTool(rBtnCache, ARTS("wxART_MEMORY", sz));
    
    mCurrentTabIndex   = _RBN_HELP;
    mCurrentPanelIndex = _RBN_PANEL_HELP;
    GetPToolBar()->AddToggleTool(rBtnToolTip, ARTS("wxART_TIPS", sz));    
    GetPToolBar()->AddSeparator();
    GetPToolBar()->AddHybridTool(rBtnInfo, ARTS("wxART_INFO", sz));
    GetPToolBar()->AddTool(rBtnHelp, ARTS("wxART_HELP", sz));


    mpCRbnBar->Realize();
}

void CLS::
UpdateRibbonHelp(bool bUse){
    GetPCShortcut()->SetRbnHelpString
	(mVpCRibbonPage[_RBN_HOME]->mVpCRbnToolBar[_RBN_PANEL_HOME], bUse);
    GetPCShortcut()->SetRbnHelpString
	(mVpCRibbonPage[_RBN_HOME]->mVpCRbnToolBar[_RBN_PANEL_DATA], bUse);
    GetPCShortcut()->SetRbnHelpString
	(mVpCRibbonPage[_RBN_HOME]->mVpCRbnToolBar[_RBN_PANEL_OBJECT], bUse);
    GetPCShortcut()->SetRbnHelpString
	(mVpCRibbonPage[_RBN_HOME]->mVpCRbnToolBar[_RBN_PANEL_PREFERENCE], bUse);
    GetPCShortcut()->SetRbnHelpString
	(mVpCRibbonPage[_RBN_HELP]->mVpCRbnToolBar[_RBN_PANEL_HELP], bUse);
}

void CLS::
UpdateRibbonState(){
    bool bTip = mVpCRibbonPage[_RBN_HELP]->mVpCRbnToolBar[_RBN_PANEL_HELP]->GetToolState(rBtnToolTip);
    if (bTip != GetPCPreference()->mCbToolTip.GetBodyBOOL()){
	mVpCRibbonPage[_RBN_HELP]->mVpCRbnToolBar[_RBN_PANEL_HELP]->ToggleTool(rBtnToolTip, !bTip);
	if (bTip == TRUE)
	    UpdateRibbonHelp(FALSE);
	else
	    UpdateRibbonHelp(TRUE);
    }
    wxToolTip::Enable(bTip);
}

wxRibbonToolBar* CLS::
GetPToolBar(){
    return mVpCRibbonPage[mCurrentTabIndex]->mVpCRbnToolBar[mCurrentPanelIndex];
}

void CLS::
AddTool(int id, wxString strIcon, wxSize sz, wxString strHelpString){
    GetPToolBar()->AddTool(id, ARTS(strIcon, sz));
    GetPToolBar()->SetToolHelpString(id, strHelpString);
}

