// -*- c -*-
// "clsMeshBase2.cpp"

#ifndef _MESH_BASE2_H
#define _MESH_BASE2_H

#include "clsWavefrontMaterial.h"
#include "clsTree2D.h"

#include "../COMMON/clsVectorMath.h"
#include "../FILEIO/clsAsciiReader.h"
#include "../GLOBJ/clsVBOBaseObject.h"

#include <vector>
#include <wx/wx.h>

class ArrayAndIndex_cls;

class MaterialStock_cls {
 public:
    MaterialStock_cls(int matIndex, int vertIndexNum);
    int mMatIndex;
    int mVertIndexNum;
};

class MeshBase2_cls : public VectorMath_cls, public AsciiReader_cls {
 public:
    MeshBase2_cls();

    VBOBaseObject_cls mCVBO;
    WavefrontMaterialCtrl_cls mCWFMaterial;

    Tree2D_cls mCTree;
    wxString mstrTreeChecked;
    
    std::vector < float > mVVertex;
    std::vector < float > mVNormal;
    std::vector < float > mVUV2;
    
    std::vector < unsigned int > mVFaceIndexVertex;
    std::vector < unsigned int > mVFaceIndexNormal;
    std::vector < unsigned int > mVFaceIndexUV2;

    std::vector < MaterialStock_cls > mVMaterialStock;
    std::vector < unsigned int > mVFaceIndexVertexAdj;

    float mRangeMin[3], mRangeMax[3];
    wxString mstrRangeChecked;
    
    wxString mstrLoaded;
    wxString mstrLoadedObjectName; // .objł̃bV

    bool IsInsideRegionBox(float fX, float fY, float fZ);
    int GetTagBufferSize();
    bool IsInside(float fX, float fY, float fZ, 
		  int tagBufferSize, int *pVertexTag);
    
    void ClearAll();
    void MinimizeMemoryUse();
    void CheckRange();
    void CheckTree();

    void ConvertCoord(int order[3]);
    void ConstructNormalVectors();
    int TriangulateNgon(int ngon);
    int TriangulateLastItem(int ngon);
    void DuplicateDouble();
    void ReconstructStructure();

    bool ReadObjFile(wxString strFile, int coord, bool bUseEmit);
    
    int mVertexOffset;
    void ReadBeginFromData();
    void ReadAppendFromData(int vertexSize, int faceSize, float *pVertices, float *pNormals, int *pFaceIndex);
    bool ReadFinishFromData(bool bUseSV);
   
    void ReserveVertexNumber(int num);
    void ReserveFaceNumber(int num);
    void SetVertexNumber(int num);
    void SetFaceNumber(int num);
    
    int  SearchAdjTriangle(int ibase, int vertex0, int vertex1);
    int  SearchAdjTriangleB(int ibase_, ArrayAndIndex_cls *pCIndexAndZ, 
			    int vertex0, int vertex1);
    void ConstructAdjStructure();

    void GLDraw(bool bUseNormal = true, bool bUseUV = true, bool bUseMatrial = true,
		float alphaFactor = 1, float shininess = 20);

    void SetVBO(bool bUseSV = TRUE);
};

#endif
