// -*- c -*-
// "clsTree2D.h"
// (setq c-basic-offset 4)

#include "clsTree2D.h"

#include <cstddef>

#include "../COMMON/CommonDefinitions.h"

#define CLS Tree2D_cls

CLS::
Tree2D_cls(int depth, int depthMax){

    mDepth = depth;
    
    for (int i = 0; i < 4; i++) mpCChild[i] = NULL;
    if (depth >= depthMax) return;

    mpCChild[0] = new Tree2D_cls(depth+1, depthMax);
    mpCChild[1] = new Tree2D_cls(depth+1, depthMax);
    mpCChild[2] = new Tree2D_cls(depth+1, depthMax);
    mpCChild[3] = new Tree2D_cls(depth+1, depthMax);
}

CLS::
~Tree2D_cls(){
    for (int i = 0; i < 4; i++)
	if (mpCChild[i] != NULL) delete mpCChild[i];
}

void CLS::
Clear(){
    std::vector < unsigned >().swap(mVItemIndex); // 
    mItemNum = 0;
    
    for (int i = 0; i < 4; i++)
	if (mpCChild[i] != NULL) mpCChild[i]->Clear();
}

void CLS::
SetRange(float minX, float maxX, float minY, float maxY){
    mMinX = minX; mMaxX = maxX;
    mMinY = minY; mMaxY = maxY;

    if (mpCChild[0] == NULL) return;
    mpCChild[0]->SetRange(minX, (maxX-minX)*0.5+minX,
			  minY, (maxY-minY)*0.5+minY);
    mpCChild[1]->SetRange((maxX-minX)*0.5+minX, maxX,
			  minY, (maxY-minY)*0.5+minY);

    mpCChild[2]->SetRange(minX, (maxX-minX)*0.5+minX,
			  (maxY-minY)*0.5+minY, maxY);
    mpCChild[3]->SetRange((maxX-minX)*0.5+minX, maxX,
			  (maxY-minY)*0.5+minY, maxY);
}

Tree2D_cls* CLS::
GetPCTree(float fX, float fY){
    if (mpCChild[0] == NULL) return this;
    if (mItemNum == 0) return this;
    
    // ѕ̏ƂāAmMax ̃`FbN
    // vȂ͂
    for (int i = 0; i < 3; i++){
	Tree2D_cls *pC = mpCChild[i];
	if (fX <= pC->mMaxX && fY <= pC->mMaxY)
	    return pC->GetPCTree(fX, fY);
    }
    return mpCChild[3]->GetPCTree(fX, fY);

    /*
    for (int i = 0; i < 4; i++){
	Tree2D_cls *pC = mpCChild[i];
	    if ((fX >= pC->mMinX && fX <= pC->mMaxX) &&
		(fY >= pC->mMinY && fY <= pC->mMaxY))
	return pC->GetPCTree(fX, fY);
    }

    ASSERT(FALSE); //ɂ͗Ȃ͂
    return this;
    */
}

bool CLS::
IsVerticesFullyOutsideBox(float *posA, float *posB, float *posC){
    if (posA[0] < mMinX && posB[0] < mMinX && posC[0] < mMinX) return TRUE;
    if (posA[0] > mMaxX && posB[0] > mMaxX && posC[0] > mMaxX) return TRUE;
    if (posA[1] < mMinY && posB[1] < mMinY && posC[1] < mMinY) return TRUE;
    if (posA[1] > mMaxY && posB[1] > mMaxY && posC[1] > mMaxY) return TRUE;
    return FALSE;
}

bool CLS::
IsVertexInsideBox(float *posA){
    bool bInside = TRUE;
    if      (posA[0] < mMinX) bInside = FALSE;
    else if (posA[0] > mMaxX) bInside = FALSE;
    else if (posA[1] < mMinY) bInside = FALSE;
    else if (posA[1] > mMaxY) bInside = FALSE;
    return bInside;
}

bool CLS::
IsVerticesInsideBox(float *posA, float *posB, float *posC){
    if (IsVertexInsideBox(posA) == TRUE) return TRUE;
    if (IsVertexInsideBox(posB) == TRUE) return TRUE;
    if (IsVertexInsideBox(posC) == TRUE) return TRUE;
    return FALSE;
}

bool CLS::
IsVertexInsidePolygon(float *posTgt, float *posA, float *posB, float *posC,
		      float *vecBA, float *vecCB, float *vecAC){
    float vecAP[3] = {posTgt[0]-posA[0], posTgt[1]-posA[1], 0};
    float vecBP[3] = {posTgt[0]-posB[0], posTgt[1]-posB[1], 0};
    float vecCP[3] = {posTgt[0]-posC[0], posTgt[1]-posC[1], 0};
    float vecCrossAZ = GetCrossedVectorZ(vecBA, vecAP);
    float vecCrossBZ = GetCrossedVectorZ(vecCB, vecBP);
    float vecCrossCZ = GetCrossedVectorZ(vecAC, vecCP);
    if ((vecCrossAZ >= 0 && vecCrossBZ >= 0 && vecCrossCZ >= 0) ||
	(vecCrossAZ <= 0 && vecCrossBZ <= 0 && vecCrossCZ <= 0)) return TRUE;
    return FALSE;
}

bool CLS::
IsBoxInsidePolygon(float *posA, float *posB, float *posC){
    float posTgt[4][2] = 
	{{mMinX, mMinY}, {mMaxX, mMinY}, {mMinX, mMaxY}, {mMaxX, mMaxY}};
    float vecBA[3] = {posA[0]-posB[0],posA[1]-posB[1], 0};
    float vecCB[3] = {posB[0]-posC[0],posB[1]-posC[1], 0};
    float vecAC[3] = {posC[0]-posA[0],posC[1]-posA[1], 0};
    if (IsVertexInsidePolygon
	(posTgt[0], posA, posB, posC, vecBA, vecCB, vecAC)) return TRUE;
    if (IsVertexInsidePolygon
	(posTgt[1], posA, posB, posC, vecBA, vecCB, vecAC)) return TRUE;
    if (IsVertexInsidePolygon
	(posTgt[2], posA, posB, posC, vecBA, vecCB, vecAC)) return TRUE;
    if (IsVertexInsidePolygon
	(posTgt[3], posA, posB, posC, vecBA, vecCB, vecAC)) return TRUE;
    return FALSE;
}

bool CLS::
IsLineSegmentIntersect(float *posA0, float *posA1, 
		       float *posB0, float *posB1){
    float ta = (posB0[0]-posB1[0])*(posA0[1]-posB0[1]) + (posB0[1]-posB1[1])*(posB0[0]-posA0[0]);
    float tb = (posB0[0]-posB1[0])*(posA1[1]-posB0[1]) + (posB0[1]-posB1[1])*(posB0[0]-posA1[0]);
    float tc = (posA0[0]-posA1[0])*(posB0[1]-posA0[1]) + (posA0[1]-posA1[1])*(posA0[0]-posB0[0]);
    float td = (posA0[0]-posA1[0])*(posB1[1]-posA0[1]) + (posA0[1]-posA1[1])*(posA0[0]-posB1[0]);
    
    return (tc*td < 0) && (ta*tb) < 0;
}

bool CLS::
IsBoxAndVerticesIntersect(float *posA, float *posB, float *posC){
    float box[4][2] = 
	{{mMinX, mMinY}, {mMaxX, mMinY}, {mMaxX, mMaxY}, {mMinX, mMaxY}};
    if (IsLineSegmentIntersect(posA, posB, box[0], box[1])) return TRUE;
    if (IsLineSegmentIntersect(posA, posB, box[1], box[2])) return TRUE;
    if (IsLineSegmentIntersect(posA, posB, box[2], box[3])) return TRUE;
    if (IsLineSegmentIntersect(posA, posB, box[3], box[0])) return TRUE;
    if (IsLineSegmentIntersect(posB, posC, box[0], box[1])) return TRUE;
    if (IsLineSegmentIntersect(posB, posC, box[1], box[2])) return TRUE;
    if (IsLineSegmentIntersect(posB, posC, box[2], box[3])) return TRUE;
    if (IsLineSegmentIntersect(posB, posC, box[3], box[0])) return TRUE;
    if (IsLineSegmentIntersect(posC, posA, box[0], box[1])) return TRUE;
    if (IsLineSegmentIntersect(posC, posA, box[1], box[2])) return TRUE;
    if (IsLineSegmentIntersect(posC, posA, box[2], box[3])) return TRUE;
    if (IsLineSegmentIntersect(posC, posA, box[3], box[0])) return TRUE;
    return FALSE;
}

void CLS::
CheckTreeItem(std::vector < unsigned int > &mVItemParent,
	      std::vector < unsigned int > &VFaceIndexVertex,
	      std::vector < float > &VVertex, int depth){

    if (depth == 0)
	mVItemIndex.swap(mVItemParent);

    else{
	uint numParent = mVItemParent.size();
	for (uint i_ = 0; i_ < numParent; i_++){
	    int i = mVItemParent[i_];
	    int indexA = VFaceIndexVertex[i*3+0];
	    int indexB = VFaceIndexVertex[i*3+1];
	    int indexC = VFaceIndexVertex[i*3+2];
	    
	    float *posA = &VVertex[indexA*3];
	    float *posB = &VVertex[indexB*3];
	    float *posC = &VVertex[indexC*3];
	    
	    /*
	    if (IsVerticesFullyOutsideBox(posA, posB, posC) == TRUE)
		continue;
	    */
	    
	    if (IsVerticesInsideBox(posA, posB, posC) == TRUE){
		mVItemIndex.push_back(i); continue; }
	    if (IsBoxInsidePolygon(posA, posB, posC) == TRUE){
		mVItemIndex.push_back(i); continue; }
	    if (IsBoxAndVerticesIntersect(posA, posB, posC) == TRUE){
		mVItemIndex.push_back(i); continue; }
	}
    }

    mItemNum = mVItemIndex.size();
	
    if (mpCChild[0] == NULL) return;
    
    for (int i = 0; i < 4; i++)
	if (mpCChild[i] != NULL) 
	    mpCChild[i]->CheckTreeItem(mVItemIndex, VFaceIndexVertex, VVertex, depth+1);
}
