// "clsWavefrontMaterial.h"
// (c-basic-offset 4)

#include "clsWavefrontMaterial.h"

#include "../COMMON/CommonDefinitionsGL.h"
#include "../COMMON/CommonDefinitions.h"

#include "../MYGL/clsTextureProvider.h"

//#include "../GLCANVAS/clsMainGLCanvas.h"
//#include "../clsPovRay.h"
//#include "../clsPovMaterialColor.h"

//#include "../OPTIONAL/clsPovAdvanced.h"
//#include "../OPTIONAL/clsPovFormat.h"

#include <string.h>

#define CLS WavefrontMaterialItem_cls

CLS::
WavefrontMaterialItem_cls(wxString strName, wxString strDir, const char **ptr, char *bufferEnd, bool bUseEmit){
    mstrName = strName;

    LOOP_RGBA mDiffuse[rgba] = 0.5;
    LOOP_RGBA mAmbient[rgba] = 0.5;
    LOOP_RGBA mSpecular[rgba] = 0.5;

    mAlphaFactor = 1;
    mShininess = 20;

    if (strName.IsEmpty() == TRUE) return;

    char charAmbinent = 'a';
    if (bUseEmit == TRUE) charAmbinent = 'e';
    
    float alpha = 1;
    while (TRUE){
	if (*ptr >= bufferEnd) break;
	if (**ptr == 'n')
	    if (SearchTagStr(ptr, "newmtl", bufferEnd, FALSE) == TRUE){
		(*ptr)--; //
		break;
	    }
	if (**ptr == 'K'){
	    if (*((*ptr)+1) == 'd'){
		(*ptr)++;(*ptr)++; (*ptr)++; 
		mDiffuse[0] = ReadFloat(ptr, ' '); (*ptr)++;
		mDiffuse[1] = ReadFloat(ptr, ' '); (*ptr)++;
		mDiffuse[2] = ReadFloat(ptr, '\n');
	    }
	    if (*((*ptr)+1) == charAmbinent){
		(*ptr)++;(*ptr)++; (*ptr)++; 
		mAmbient[0] = ReadFloat(ptr, ' '); (*ptr)++;
		mAmbient[1] = ReadFloat(ptr, ' '); (*ptr)++;
		mAmbient[2] = ReadFloat(ptr, '\n');
	    }
	    if (*((*ptr)+1) == 's'){
		(*ptr)++;(*ptr)++; (*ptr)++; 
		mSpecular[0] = ReadFloat(ptr, ' '); (*ptr)++;
		mSpecular[1] = ReadFloat(ptr, ' '); (*ptr)++;
		mSpecular[2] = ReadFloat(ptr, '\n');
	    }
	}
	if (**ptr == 'd'){ 
	    (*ptr)++; (*ptr)++;
	    alpha = ReadFloat(ptr, '\n');
	}
	if (**ptr == 'T' && *((*ptr)+1) == 'r'){ 
	    (*ptr)++;(*ptr)++; (*ptr)++;
	    alpha = ReadFloat(ptr, '\n');
	}
	if (**ptr == 'T' && *((*ptr)+1) == 'f'){ 
	    (*ptr)++;(*ptr)++; (*ptr)++;
	    mTransmission[0] = ReadFloat(ptr, ' ');
	    mTransmission[1] = ReadFloat(ptr, ' ');
	    
	}
	if (**ptr == 'N' && *((*ptr)+1) == 'i'){ 
	    (*ptr)++;(*ptr)++;(*ptr)++;
	    mIOR = ReadFloat(ptr, '\n');
	}
	if (**ptr == 'm'){
	    if (SearchTagStr(ptr, "map_Kd", bufferEnd) ||
		SearchTagStr(ptr, "map_Ka", bufferEnd) ){
		(*ptr)++;
		char strMtlFileBuf[512];
		GetStrUptoLineEnd(ptr, strMtlFileBuf, bufferEnd);
		wxString strMtlFile(wxString::FromAscii(strMtlFileBuf));
		strMtlFile.Trim(FALSE); // [̋󔒏

		wxString strCandidA = 
		    strDir + wxT("textures") + _WXDELIM_ + strMtlFile;
		wxString strCandidB = 
		    strDir + _WXDELIM_ + strMtlFile;
		if (::wxFileExists(strCandidA) == TRUE)
		    mstrTexDiffuse = strCandidA;
		else if (::wxFileExists(strCandidB) == TRUE)
		    mstrTexDiffuse = strCandidB;
		else
		    mstrTexDiffuse = strMtlFile;
	    }
	    else{
		char strMtlFile[512];
		GetStrUptoLineEnd(ptr, strMtlFile, bufferEnd); // dummy
	    }
	}
	if (**ptr == 'i'){ // illum
	    mIllumMode = ReadInt(ptr, '\n');
	}
		
	(*ptr)++;
    }

    mAmbient[3] = alpha;
    mDiffuse[3] = alpha;
    mSpecular[3] = alpha;
}

wxString CLS::
GetStrTexDiffuse(){
    if (mstrOverwriteTexture.IsEmpty() == FALSE)
	return mstrOverwriteTexture;
    return mstrTexDiffuse;
}

void CLS::
GLSetMaterial(bool bUseMaterialColor){

    wxString strTexture = GetStrTexDiffuse();
    if (strTexture.IsEmpty() == TRUE)
	GetPCTextureProvider()->BindBlankTexture();
    else{
	GetPCTextureProvider()->BindTexture(strTexture);
	//mCTexture.LoadImage(strTexture, TRUE);
	//mCTexture.BindTexture();
    }

    if (bUseMaterialColor == FALSE) return;

    float ambient[4];
    float diffuse[4];
    float specular[4];

    LOOP_RGB ambient[rgb] = mAmbient[rgb];
    LOOP_RGB diffuse[rgb] = mDiffuse[rgb];
    LOOP_RGB specular[rgb] = mSpecular[rgb];
    ambient [3] = mAmbient[3]  * mAlphaFactor;
    diffuse [3] = mDiffuse[3]  * mAlphaFactor;
    specular[3] = mSpecular[3] * mAlphaFactor;
    
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mShininess);
}

void CLS::
GLOverwriteTexture(wxString strTexture){
    mstrOverwriteTexture = strTexture;
    mCTexture.LoadImage(strTexture, TRUE);
    mCTexture.BindTexture();
}

/*
void CLS::
GetAlphaForMultiLayer2(float& alpha){
    // eNX`2d˂̏ꍇ
    float tr = 1.0-alpha;
    float trsq = sqrt(tr);
    alpha = 1.0 - trsq;
}
*/

void CLS::
GetClrInfo(float clr0[3][4], float *clrOvwr){
    if (clrOvwr != NULL){
	LOOP_RGB clr0[0][rgb] = clrOvwr[rgb]  ; clr0[0][3] = clrOvwr[0+3];
	LOOP_RGB clr0[1][rgb] = clrOvwr[4+rgb]; clr0[1][3] = clrOvwr[4+3];
	LOOP_RGB clr0[2][rgb] = clrOvwr[8+rgb]; clr0[2][3] = clrOvwr[8+3];
    }
    else{
	LOOP_RGB clr0[0][rgb] = mAmbient[rgb] ; clr0[0][3] = mAmbient[3];
	LOOP_RGB clr0[1][rgb] = mDiffuse[rgb] ; clr0[1][3] = mDiffuse[3];
	LOOP_RGB clr0[2][rgb] = mSpecular[rgb]; clr0[2][3] = mSpecular[3];
    }
}

// ---------------------------

#define CLS WavefrontMaterialCtrl_cls

CLS::
WavefrontMaterialCtrl_cls(){
    mbUseEmitAsAmbient = FALSE;
}

void CLS::
Clear(){
    mVCItems.clear();
}

void CLS::
ReadMtllib(wxString strDir, wxString strFile){
    
    mstrDirectory = strDir + _WXDELIM_;

    long fileSize;
    char *pBuffer = GetPNewBufferFromFile(mstrDirectory + strFile, fileSize);

    if (pBuffer == NULL) return;

    const char *ptr = pBuffer;
    char *pEnd = pBuffer + fileSize;

    while (TRUE){
	if (ptr >= pEnd) break;	
	
	if (*ptr == '#'){
	    if (SearchChar(&ptr, '\n', pEnd) == FALSE) return;
	}
	if (*ptr == 'n')
	    if (SearchTagStr(&ptr, "newmtl", pEnd) == TRUE){
		ptr++;
		char strItemName[512];
		GetStrUptoLineEnd(&ptr, strItemName, pEnd);
		
		mVCItems.push_back
		    (WavefrontMaterialItem_cls
		     (wxString::FromAscii(strItemName), mstrDirectory, 
		      &ptr, pEnd, mbUseEmitAsAmbient));
	    }
	ptr++;
    }
    
    delete pBuffer;
}

void CLS::
MakeMaterialIfEmpty(){
    if (mVCItems.size() == 0)
	mVCItems.push_back(WavefrontMaterialItem_cls(wxES, wxES, NULL, NULL, FALSE));
}

int CLS::
GetIndexFromTag(wxString strTag){
    int size = (int)mVCItems.size();
    for (int i = 0; i < size; i++){
	if (mVCItems[i].mstrName.IsSameAs(strTag) == TRUE)
	    return i;
    }
    return 0;
}

void CLS::
GLSetMaterial(int index, bool bUseMaterialColor){
    if (index < -1 || index >= (int)mVCItems.size()) return;
    if (index == -1){
	for (int i = 0; i < mVCItems.size(); i++){
	    mVCItems[i].mstrOverwriteTexture = mstrOverwriteTexture;
	    mVCItems[i].mAlphaFactor = mAlphaFactor;
	    mVCItems[i].mShininess = mShininess;
	    mVCItems[i].GLSetMaterial(bUseMaterialColor);
	}
	return;
    }
    mVCItems[index].mstrOverwriteTexture = mstrOverwriteTexture;
    mVCItems[index].mAlphaFactor = mAlphaFactor;
    mVCItems[index].mShininess = mShininess;
    mVCItems[index].GLSetMaterial(bUseMaterialColor);
}

void CLS::
GLOverwriteTexture(wxString strTexture){
    for (unsigned int i = 0; i < mVCItems.size(); i++){
	mVCItems[i].GLOverwriteTexture(strTexture);
    }
}

void CLS::
SetTexWrap(int texWrap){
    for (unsigned int i = 0; i < mVCItems.size(); i++){
	mVCItems[i].mCTexture.mTexWrap = texWrap;
    }
}

