// (setq-c-offset 4)
// "clsArrayCopy.h"

#include "clsArrayCopy.h"

#include "../MYGL/clsPosition.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#define CLS ArrayCopy_cls

CLS::
ArrayCopy_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("ArrayCopy_cls", strName){

    wxString strUpdate = "VIEW3D::PROPERTY_MODIFIER";
    
    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCstrTarget, "Target=", "None", strUpdate);
    mCConfList.Add(&mCCoord, "Coord=", "0", strUpdate);
    
    ParamGroup_cls *pCX = GetPCNewParamGroup("X Min/Max", "Min", "Max");
    InitializeAndPushParam(pCX, 0, 1);
    
    ParamGroup_cls *pCY = GetPCNewParamGroup("Y Min/Max", "Min", "Max");
    InitializeAndPushParam(pCY, 0, 1);
        
    ParamGroup_cls *pCZ = GetPCNewParamGroup("Z Min/Max", "Min", "Max");
    InitializeAndPushParam(pCZ, 0, 1);

    ParamGroup_cls *pCAisxX = GetPCNewParamGroup("X' Axis", "x", "y", "z");
    InitializeAndPushParam(pCAisxX, 2, 0, 0);
    
    ParamGroup_cls *pCAisxY = GetPCNewParamGroup("Y' Axis", "x", "y", "z");
    InitializeAndPushParam(pCAisxY, 0, 2, 0);

    ParamGroup_cls *pCAisxZ = GetPCNewParamGroup("Z' Axis", "x", "y", "z");
    InitializeAndPushParam(pCAisxZ, 0, 0, 2);
};

void CLS::
Execute(SceneObj_cls *pCObj, 
	std::vector < SceneObjInstance_cls* > &VPCInstance,
	int modifierDepth){
    if (mCbValidity.GetBodyBOOL() == FALSE) return;

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int minX = floor(GetPCParamGroup("X Min/Max")->GetVInterpolatedValue(NT)[0] + 0.5);
    int maxX = floor(GetPCParamGroup("X Min/Max")->GetVInterpolatedValue(NT)[1] + 0.5);
    int minY = floor(GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue(NT)[0] + 0.5);
    int maxY = floor(GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue(NT)[1] + 0.5);
    int minZ = floor(GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue(NT)[0] + 0.5);
    int maxZ = floor(GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue(NT)[1] + 0.5);

    float Xaxis[3] = {
	GetPCParamGroup("X' Axis")->GetVInterpolatedValue(NT)[0],
	GetPCParamGroup("X' Axis")->GetVInterpolatedValue(NT)[1],
	GetPCParamGroup("X' Axis")->GetVInterpolatedValue(NT)[2],
    };
    float Yaxis[3] = {
	GetPCParamGroup("Y' Axis")->GetVInterpolatedValue(NT)[0],
	GetPCParamGroup("Y' Axis")->GetVInterpolatedValue(NT)[1],
	GetPCParamGroup("Y' Axis")->GetVInterpolatedValue(NT)[2],
    };
    float Zaxis[3] = {
	GetPCParamGroup("Z' Axis")->GetVInterpolatedValue(NT)[0],
	GetPCParamGroup("Z' Axis")->GetVInterpolatedValue(NT)[1],
	GetPCParamGroup("Z' Axis")->GetVInterpolatedValue(NT)[2],
    };
    
    if (mCCoord.GetBodyINT() == 0){
	double matrix[16], matrixA[16];
	LoadIdentity(matrix);
	pCObj->ApplyRotate(matrix, NT);
	GetInvMatrix4x4(matrix, matrixA);
	
	float XaxisDash[3] = {Xaxis[0], Xaxis[1], Xaxis[2]};
	float YaxisDash[3] = {Yaxis[0], Yaxis[1], Yaxis[2]};
	float ZaxisDash[3] = {Zaxis[0], Zaxis[1], Zaxis[2]};

	//float XaxisDash[3] = {Xaxis[0], Yaxis[0], Zaxis[0]};
	//float YaxisDash[3] = {Xaxis[1], Yaxis[1], Zaxis[1]};
	//float ZaxisDash[3] = {Xaxis[2], Yaxis[2], Zaxis[2]};

	Xaxis[0] = matrixA[ 0]*XaxisDash[0] + matrixA[ 1]*XaxisDash[1] + matrixA[ 2]*XaxisDash[2];
	Xaxis[1] = matrixA[ 4]*XaxisDash[0] + matrixA[ 5]*XaxisDash[1] + matrixA[ 6]*XaxisDash[2];
	Xaxis[2] = matrixA[ 8]*XaxisDash[0] + matrixA[ 9]*XaxisDash[1] + matrixA[10]*XaxisDash[2];

	Yaxis[0] = matrixA[ 0]*YaxisDash[0] + matrixA[ 1]*YaxisDash[1] + matrixA[ 2]*YaxisDash[2];
	Yaxis[1] = matrixA[ 4]*YaxisDash[0] + matrixA[ 5]*YaxisDash[1] + matrixA[ 6]*YaxisDash[2];
	Yaxis[2] = matrixA[ 8]*YaxisDash[0] + matrixA[ 9]*YaxisDash[1] + matrixA[10]*YaxisDash[2];

	Zaxis[0] = matrixA[ 0]*ZaxisDash[0] + matrixA[ 1]*ZaxisDash[1] + matrixA[ 2]*ZaxisDash[2];
	Zaxis[1] = matrixA[ 4]*ZaxisDash[0] + matrixA[ 5]*ZaxisDash[1] + matrixA[ 6]*ZaxisDash[2];
	Zaxis[2] = matrixA[ 8]*ZaxisDash[0] + matrixA[ 9]*ZaxisDash[1] + matrixA[10]*ZaxisDash[2];
	
    }

    for (int iZ = minZ; iZ < maxZ; iZ++)
	for (int iY = minY; iY < maxY; iY++)
	    for (int iX = minX; iX < maxX; iX++){
		if (iX == 0 && iY == 0 && iZ == 0) continue;
		SceneObjInstance_cls *pCNew = 
		    new SceneObjInstance_cls(pCObj);
		pCNew->mMatrix[12] = iX*Xaxis[0] + iY*Yaxis[0] + iZ*Zaxis[0];
		pCNew->mMatrix[13] = iX*Xaxis[1] + iY*Yaxis[1] + iZ*Zaxis[1];
		pCNew->mMatrix[14] = iX*Xaxis[2] + iY*Yaxis[2] + iZ*Zaxis[2];
		pCNew->mModifierDepth = modifierDepth;
		VPCInstance.push_back(pCNew);
	    }
}

void CLS::
SyncFromBuffer(ArrayCopyBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
    
    std::vector < float > VXMinMax;
    VXMinMax.push_back(pCBuffer->mCXMin.GetBodyDOUBLE());
    VXMinMax.push_back(pCBuffer->mCXMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("X Min/Max", VXMinMax, pCObj);

    std::vector < float > VYMinMax;
    VYMinMax.push_back(pCBuffer->mCYMin.GetBodyDOUBLE());
    VYMinMax.push_back(pCBuffer->mCYMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("Y Min/Max", VYMinMax, pCObj);

    std::vector < float > VZMinMax;
    VZMinMax.push_back(pCBuffer->mCZMin.GetBodyDOUBLE());
    VZMinMax.push_back(pCBuffer->mCZMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("Z Min/Max", VZMinMax, pCObj);

    std::vector < float > VAxisX = pCBuffer->mCAxisX.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("X' Axis", VAxisX, pCObj);

    std::vector < float > VAxisY = pCBuffer->mCAxisY.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("Y' Axis", VAxisY, pCObj);

    std::vector < float > VAxisZ = pCBuffer->mCAxisZ.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("Z' Axis", VAxisZ, pCObj);
}

#define CLS ArrayCopyBuffer_cls

CLS::
ArrayCopyBuffer_cls(ArrayCopy_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCstrTarget, "Target=", "None");
    mCConfList.Add(&mCAxisX, "AxisX=", "2 0 0");
    mCConfList.Add(&mCAxisY, "AxisY=", "0 2 0");
    mCConfList.Add(&mCAxisZ, "AxisZ=", "0 0 2");
    mCConfList.Add(&mCCoord, "Coord=", "0");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
		new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);
    pC->CombineSizer(2, _("Target/Coord"));
    pC->CombineConfItem(mCstrTarget);

    wxString AstrCoord[2] = {"local", "parent"};
    pC->SetChoiceList(2, AstrCoord);
    pC->CombineConfItem(mCCoord);
    
    int indexX = pC->CombineSizer(2, "X Min/Max");
    pC->mDragScale = 4;
    pC->CombineConfItem(mCXMin);
    pC->mDragScale = 4;
    pC->CombineConfItem(mCXMax);
    
    int indexY = pC->CombineSizer(2, "Y Min/Max");
    pC->mDragScale = 4;
    pC->CombineConfItem(mCYMin);
    pC->mDragScale = 4;
    pC->CombineConfItem(mCYMax);
    
    int indexZ = pC->CombineSizer(2, "Z Min/Max");
    pC->mDragScale = 4;
    pC->CombineConfItem(mCZMin);
    pC->mDragScale = 4;
    pC->CombineConfItem(mCZMax);

    pC->mDragScale = 0.4;
    int indexXdash = pC->CombineConfItem(mCAxisX, "X' Axis");
    pC->mDragScale = 0.4;
    int indexYdash = pC->CombineConfItem(mCAxisY, "Y' Axis");
    pC->mDragScale = 0.4;
    int indexZdash = pC->CombineConfItem(mCAxisZ, "Z' Axis");
	    
    pC->FinishSetting();

    pC->SetColorForXYZMinMax(indexX, indexY, indexZ);
    pC->SetColorForXYZ(indexXdash);
    pC->SetColorForXYZ(indexYdash);
    pC->SetColorForXYZ(indexZdash);
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VX = mpCSrc->GetPCParamGroup("X Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    std::vector < float > VY = mpCSrc->GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    std::vector < float > VZ = mpCSrc->GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    
    mCXMin.SetBodyFromData((long long)VX[0]);
    mCXMax.SetBodyFromData((long long)VX[1]);
    mCYMin.SetBodyFromData((long long)VY[0]);
    mCYMax.SetBodyFromData((long long)VY[1]);
    mCZMin.SetBodyFromData((long long)VZ[0]);
    mCZMax.SetBodyFromData((long long)VZ[1]);
    
    mCAxisX.SetBodyFromData(mpCSrc->GetPCParamGroup("X' Axis")->GetVInterpolatedValue((float)time, FALSE));
    mCAxisY.SetBodyFromData(mpCSrc->GetPCParamGroup("Y' Axis")->GetVInterpolatedValue((float)time, FALSE));
    mCAxisZ.SetBodyFromData(mpCSrc->GetPCParamGroup("Z' Axis")->GetVInterpolatedValue((float)time, FALSE));
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}
