// (setq-c-offset 4)
// "clsBooleanBase.h"

#include "clsBooleanBase.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../SCENEOBJ/clsMeshObj.h"
#include "../SCENEOBJ/clsPrimitive.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/DATAVIEWER/clsViewerClone.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../../wxWidgetEnum.h"

#define CLS BooleanBase_cls

CLS::
BooleanBase_cls(wxString strClassName, wxString strName, wxString strObjName) :
    ModifierBase_cls(strClassName, strName){
   
    wxString strUpdate = "VIEW3D::PROPERTY_DATAPROC";
    
    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCBooleanType, "BooleanType=", "0", strUpdate);
    mCConfList.Add(&mCShapeType, "BooleanShape=", "0", strUpdate);
    mCConfList.Add(&mCstrBoolObject, "BooleanObject=", "None", strUpdate);
    
    ParamGroup_cls *pCX = GetPCNewParamGroup("X Min/Max", "Min", "Max");
    InitializeAndPushParam(pCX, 0, 1);
    ParamGroup_cls *pCY = GetPCNewParamGroup("Y Min/Max", "Min", "Max");
    InitializeAndPushParam(pCY, 0, 1);
    ParamGroup_cls *pCZ = GetPCNewParamGroup("Z Min/Max", "Min", "Max");
    InitializeAndPushParam(pCZ, 0, 1);
    ParamGroup_cls *pCRadius = GetPCNewParamGroup("Radius", "R");
    InitializeAndPushParam(pCRadius, 1);
    
}

void CLS::
SyncFromBuffer(BooleanBufferBase_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VXMinMax, VYMinMax, VZMinMax;
    std::vector < float > VRadius;

    VXMinMax.push_back(pCBuffer->mCXMin.GetBodyDOUBLE());
    VXMinMax.push_back(pCBuffer->mCXMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("X Min/Max", VXMinMax, pCObj);
    VYMinMax.push_back(pCBuffer->mCYMin.GetBodyDOUBLE());
    VYMinMax.push_back(pCBuffer->mCYMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("Y Min/Max", VYMinMax, pCObj);
    VZMinMax.push_back(pCBuffer->mCZMin.GetBodyDOUBLE());
    VZMinMax.push_back(pCBuffer->mCZMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("Z Min/Max", VZMinMax, pCObj);
    VRadius.push_back(pCBuffer->mCRadius.GetBodyDOUBLE());
    SyncronizeParamForModifier("Radius", VRadius, pCObj);
}

void CLS::
SelectObject(wxString strObjName){
    mCstrBoolObject.PushCommandPair(strObjName);
}

wxString CLS::
GetStrProcCounterParam(SceneObj_cls *pCObj, double NT){
    wxString strCounter;
    int shapeType = mCShapeType.GetBodyINT();
    switch(shapeType){
    case 0:
	strCounter << GetPCParamGroup("X Min/Max")->GetStrValues(NT);
	strCounter << GetPCParamGroup("Y Min/Max")->GetStrValues(NT);
	strCounter << GetPCParamGroup("Z Min/Max")->GetStrValues(NT);
	break;
    case 1: {
	strCounter << GetPCParamGroup("Radius")->GetStrValues(NT);
	wxString strTgtObj = mCstrBoolObject.GetBodySTR();
	strCounter << GetStrProcCounterForTgtObj(strTgtObj, NT);
	
	break; }
    case 2:{
	double matrix16[16];
	pCObj->GetMatrix16(matrix16, NT);
	for (int i = 0; i < 16; i++)
	    strCounter << matrix16[i];
	
	wxString strTgtObj = mCstrBoolObject.GetBodySTR();
	strCounter << GetStrProcCounterForTgtObj(strTgtObj, NT);
	break; }
    }
    
    return strCounter;
}

void CLS::
GetMatrix16TgtInv(AttributeClone_cls* pCMod,
		  double matrix16TgtInv[16], double NT){
    
    wxString strObj = pCMod->GetPCItem("BooleanObject=")->GetBodySTR();
    SceneObj_cls* pCObj = GCClassProvider.GetPCSceneObjCtrl()->GetPCFromName(strObj);
    if (pCObj == NULL)
	return LoadIdentity(matrix16TgtInv);
    double matrix16[16];
    pCObj->GetMatrix16(matrix16, NT);
    GetInvMatrix4x4(matrix16, matrix16TgtInv);
}

wxString CLS::
GetStrProcCounterForTgtObj(wxString strObj, double NT){
    // XṼ`FbNp̕(ΏۃIuWFNgԂ)
    wxString strProcCounter;

    double matrix16TgtInv[16];
    SceneObj_cls *pCObj = GCClassProvider.GetPCSceneObjCtrl()->GetPCFromName(strObj);

    if (pCObj == NULL)
	LoadIdentity(matrix16TgtInv);
    else{

	if (pCObj->mstrClassName.IsSameAs("MeshObj_cls"))
	    strProcCounter << ((MeshObj_cls*)pCObj)->mstrLoaded;
	if (pCObj->mstrClassName.IsSameAs("Primitive_cls"))
	    strProcCounter << ((Primitive_cls*)pCObj)->mCType.GetBodyINT();

	double matrix16[16];
	pCObj->GetMatrix16(matrix16, NT);
	GetInvMatrix4x4(matrix16, matrix16TgtInv);
    }
    for (int i = 0; i < 16; i++)
	strProcCounter << matrix16TgtInv[i];

    return strProcCounter;
}

wxString CLS::
GetStrProcCounterClone(ViewerClone_cls *pCViewerClone, AttributeClone_cls *pC, double NT){
    wxString strCounter;
    if (pC->GetPCItem("Validity=")->GetBodyBOOL() == FALSE)
	return strCounter;

    strCounter << pC->mCConfList.GetStrProcCounter();
	    
    // VolumeBoolean_cls::GetStrProcCounterParam
    // Ɠ̕CloneBo֐ɂɂB
    int shapeType = pC->GetPCItem("BooleanShape=")->GetBodyINT();
    switch(shapeType){
    case 0:
	strCounter << pC->GetPCParamGroup("X Min/Max")->GetStrValues(NT);
	strCounter << pC->GetPCParamGroup("Y Min/Max")->GetStrValues(NT);
	strCounter << pC->GetPCParamGroup("Z Min/Max")->GetStrValues(NT);
	break;
    case 1: {
	strCounter << pC->GetPCParamGroup("Radius")->GetStrValues(NT);
	
	wxString strTgtObj = pC->GetPCItem("BooleanObject=")->GetBodySTR();
	strCounter << 
	    GetStrProcCounterForTgtObj(strTgtObj, NT);
	break; }
    case 2: {
	for (int i = 0; i < 16; i++)
	    strCounter << pCViewerClone->mMatrix16[i];
	
	wxString strTgtObj = pC->GetPCItem("BooleanObject=")->GetBodySTR();
	strCounter << 
	    GetStrProcCounterForTgtObj(strTgtObj, NT);
	break;  }
    }

    return strCounter;
}

#define CLS BooleanBufferBase_cls

CLS::
BooleanBufferBase_cls(BooleanBase_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCShapeType, "BooleanShape=", "0", "VIEW3D::PROPERTY_DATAPROC");
    mCConfList.Add(&mCBooleanType, "BooleanType=", "0", "VIEW3D");
    mCConfList.Add(&mCstrBoolObject, "BooleanObject=", "None", "VIEW3D");
};

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineSizer(2, _("BooleanType"));

    wxString AstrShapeType[3] = 
	{"Box", "Sphere", "Object"};
    pC->SetChoiceList(3, AstrShapeType);
    pC->CombineConfItem(mCShapeType);

    wxString AstrFilterType[3] =
	{"Intersect", "Diffierence", "Add"};
    pC->SetChoiceList(3, AstrFilterType);
    pC->CombineConfItem(mCBooleanType);

    mIndexTgtObj = pC->CombineSizer(2, _("Object"));
    pC->CombineBitmapPanel(btnDropdown, wxES, 14);
    pC->mVCItem.back().mBitmapTemplate = _BITMAP_BTN_DROPDOWN;
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = mCstrName.GetBodySTR();
    pC->CombineConfItem(mCstrBoolObject);
    
    mIndexRangeX = pC->CombineSizer(2, "X Min/Max");
    pC->mDragScale = 1;
    pC->CombineConfItem(mCXMin);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCXMax);

    mIndexRangeY = pC->CombineSizer(2, "Y Min/Max");
    pC->mDragScale = 1;
    pC->CombineConfItem(mCYMin);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCYMax);

    mIndexRangeZ = pC->CombineSizer(2, "Z Min/Max");
    pC->mDragScale = 1;
    pC->CombineConfItem(mCZMin);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCZMax);

    pC->mDragScale = 1;
    pC->SetLimit(0, 2);
    mIndexRadius = pC->CombineConfItem(mCRadius, _("Radius"));

    pC->FinishSetting();

    pC->SetColorForXYZMinMax(mIndexRangeX, mIndexRangeY, mIndexRangeZ);
    
    // BitmapPanel_cls *pCBitmapObjSelect = (BitmapPanel_cls*)pC->mVCItem[indexObjSelector].mpCWindowLink;
    // BitmapPanel_cls::SetButtonDropDown(pCBitmapObjSelect);

    mpCMNumPanelLink = pC;
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);

    double time = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VX = mpCSrc->GetPCParamGroup("X Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    std::vector < float > VY = mpCSrc->GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    std::vector < float > VZ = mpCSrc->GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    
    mCXMin.SetBodyFromData(VX[0]);
    mCXMax.SetBodyFromData(VX[1]);

    mCYMin.SetBodyFromData(VY[0]);
    mCYMax.SetBodyFromData(VY[1]);

    mCZMin.SetBodyFromData(VZ[0]);
    mCZMax.SetBodyFromData(VZ[1]);

    mCRadius.SetBodyFromData(mpCSrc->GetPCParamGroup("Radius")->GetVInterpolatedValue((float)time, FALSE));
}
    
void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}

bool CLS::
HideAndShow(){
    bool bChange = FALSE;

    if (mCShapeType.GetBodyINT() == _BOOL_BOX){
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRangeX].Show(TRUE);
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRangeY].Show(TRUE);
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRangeZ].Show(TRUE);
    }
    else{
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRangeX].Show(FALSE);
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRangeY].Show(FALSE);
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRangeZ].Show(FALSE);
    }

    if (mCShapeType.GetBodyINT() == _BOOL_SPHERE)
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRadius].Show(TRUE);
    else
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRadius].Show(FALSE);

    if (mCShapeType.GetBodyINT() == _BOOL_SPHERE ||
	mCShapeType.GetBodyINT() == _BOOL_OBJECT){
	bChange |= mpCMNumPanelLink->mVCItem[mIndexTgtObj].Show(TRUE);
    }
    else
	bChange |= mpCMNumPanelLink->mVCItem[mIndexTgtObj].Show(FALSE);

    return bChange;
}
