// (setq-c-offset 4)
// "clsModifyParticleSize.h"

#include "clsModifyParticleSize.h"

#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/DATAVIEWER/clsParticle.h"
#include "../SCENEOBJ/DATAVIEWER/clsViewerClone.h"
#include "../SIMDATA/clsArrayAndIndex.h"
#include "../SIMDATA/clsParticleSnap.h"
#include "../SIMDATA/clsParticleForRender.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#define CLS ModifyParticleSize_cls

CLS::
ModifyParticleSize_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("ModifyParticleSize_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_DATAPROC";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCTargetType, "TargetType=", "0", strUpdate);
    mCConfList.Add(&mCTargetAttrib, "TargetAttrib=", "0", strUpdate);
    mCConfList.Add(&mCDependType, "DependType=", "3", strUpdate);
    mCConfList.Add(&mCColorCurveIndex, "ColorCurveIndex=", "0", strUpdate);
    mCConfList.Add(&mCChannelType, "ChannelType=", "3", strUpdate);
}

void CLS::
SyncFromBuffer(ModifyParticleSizeBuffer_cls *pCBuffer, SceneObj_cls *pCObj){
    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
}

#define CLS ModifyParticleSizeBuffer_cls 

CLS::
ModifyParticleSizeBuffer_cls(ModifyParticleSize_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCTargetType, "TargetType=", "0");
    mCConfList.Add(&mCTargetAttrib, "TargetAttrib=", "0");
    mCConfList.Add(&mCDependType, "DependType=", "3");
    mCConfList.Add(&mCColorCurveIndex, "ColorCurveIndex=", "0");
    mCConfList.Add(&mCChannelType, "ChannelType=", "3");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineSizer(2, _("Target Type"));
    wxString strTarget[5] = {_("All"), "type 0", "type 1", "type 2","type 3" };
    pC->SetChoiceList(5, strTarget);
    pC->CombineConfItem(mCTargetType);
    wxString strAttrib[2] = {_("Size"), _("Alpha")};
    pC->SetChoiceList(2, strAttrib);
    pC->CombineConfItem(mCTargetAttrib);

    wxString strDependType[9] = { "Index depend", "Size depend", "ID depend", "Value0", "Value1", "Value2", "X", "Y", "Z" };
    pC->SetChoiceList(9, strDependType);
    pC->CombineConfItem(mCDependType, _("Dependency"));
    
    pC->CombineSizer(2, _("Color Curve"));
    pC->SetLimit(0, _MAX_COLOR_CURVE - 1);
    pC->CombineConfItem(mCColorCurveIndex);

    wxString strChannelType[6] = { "R", "G", "B", "A", _("Luminance"), _("Brightness")};
    pC->SetChoiceList(6, strChannelType);
    pC->CombineConfItem(mCChannelType);
    
    pC->FinishSetting();
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}

#define CLS ThreadModifyParticleSize_cls

void CLS::
SetParameters(int dependType, int *pID, long long *pIDLongLong,
	      float (*pX)[3], float *pValue, 
	      float *pFilterAlpha, float *pSizeMod, unsigned char *pType,
	      double factor, int type, int attrib,
	      float *pTable,
	      std::vector < double > VRange){
    mDependType = dependType;
    mpID = pID;
    mpIDLongLong = pIDLongLong;
    mpX = pX;
    mpValue = pValue;
    mpFilterAlpha = pFilterAlpha;
    mpSizeMod = pSizeMod;
    mpType = pType;
    mFactor = factor;
    mType = type;
    mAttrib = attrib;
    mpTable = pTable;
    mVRange = VRange;
}

double CLS::
GetRemapX(const double width, const double min, const double x0){
    return (x0 - min)/width;
}

wxThread::ExitCode CLS::
Entry(){
    DummyEntry();
    return 0;
}

void CLS::
DummyEntry(){
    
    switch(mDependType){
    case 0: ProcessByIndex(); break;
    case 2: ProcessByID(); break;
    case 1: case 3: case 4: case 5: ProcessByValue(); break;
    case 6: ProcessByX(0); break;
    case 7: ProcessByX(1); break;
    case 8: ProcessByX(2); break;
    }
        
    *mpbFinishedFlagLink = TRUE;
}

void CLS::
ProcessByValue(){
    const int PIni = mPIni, PFin = mPFin;
    const float * const pValue = mpValue;
    float * const pFilterAlpha = mpFilterAlpha;
    float * const pSizeMod = mpSizeMod;
    const unsigned char * const pType = mpType;
    const double factor = mFactor;
    const double factorInv = 1.0 - mFactor;
    const float * const pTable = mpTable;
    
    double width0 = mVRange[1] - mVRange[0];
    if (width0 == 0) width0 = 1;
    const double width = width0;
    const double rangeMin = mVRange[0];
    
    float * const pTgtAttrib = (mAttrib == 1) ? pFilterAlpha : pSizeMod;
    if (pValue != NULL){
	for (int i = PIni; i < PFin; i++){
	    if (mType != -1 && pType[i] != mType) continue;
	    if (pFilterAlpha[i] == 0) continue;
	    
	    float x = GetRemapX(width, rangeMin, pValue[i]);
	    if (x < 0) x = 0;
	    if (x > 1) x = 1;
	    pTgtAttrib[i] *= pTable[(int)(x*1023)]; //alpha
	}
    }
}

void CLS::
ProcessByIndex(){
    const int PIni = mPIni, PFin = mPFin;
    
    float * const pFilterAlpha = mpFilterAlpha;
    float * const pSizeMod = mpSizeMod;
    const unsigned char * const pType = mpType;

    const float * const pTable = mpTable;
    std::vector < double > VRange = mVRange;

    double width0 = mVRange[1] - mVRange[0];
    if (width0 == 0) width0 = 1;
    const double width = width0;
    const double rangeMin = mVRange[0];
    
    float * const pTgtAttrib = (mAttrib == 1) ? pFilterAlpha : pSizeMod;
    for (int i = PIni; i < PFin; i++){
	if (mType != -1 && pType[i] != mType) continue;
	if (pFilterAlpha[i] == 0) continue;
	
	float value = (float)i;
	
	float x = GetRemapX(width, rangeMin, value);
	if (x < 0) x = 0;
	if (x > 1) x = 1;
	pTgtAttrib[i] *= pTable[(int)(x*1023)]; //alpha
    }
}

void CLS::
ProcessByID(){
    const int PIni = mPIni, PFin = mPFin;
    const int * const pID = mpID;
    const long long * const pIDLongLong = mpIDLongLong;

    float * const pFilterAlpha = mpFilterAlpha;
    float * const pSizeMod = mpSizeMod;
    const unsigned char * const pType = mpType;

    const float * const pTable = mpTable;
    std::vector < double > VRange = mVRange;

    double width0 = mVRange[1] - mVRange[0];
    if (width0 == 0) width0 = 1;
    const double width = width0;
    const double rangeMin = mVRange[0];

    if (pIDLongLong == NULL && pID == NULL)
	return ProcessByIndex();

    float * const pTgtAttrib = (mAttrib == 1) ? pFilterAlpha : pSizeMod;
    if (pIDLongLong != NULL)
	for (int i = PIni; i < PFin; i++){
	    if (mType != -1 && pType[i] != mType) continue;
	    if (pFilterAlpha[i] == 0) continue;
	    
	    float value = (float)pIDLongLong[i];
	    
	    float x = GetRemapX(width, rangeMin, value);
	    if (x < 0) x = 0;
	    if (x > 1) x = 1;
	    pTgtAttrib[i] *= pTable[(int)(x*1023)]; //alpha
	}
    
    if (pID != NULL)
	for (int i = PIni; i < PFin; i++){
	    if (mType != -1 && pType[i] != mType) continue;
	    if (pFilterAlpha[i] == 0) continue;
	    
	    float value = (float)pID[i];
	    
	    float x = GetRemapX(width, rangeMin, value);
	    if (x < 0) x = 0;
	    if (x > 1) x = 1;
	    pTgtAttrib[i] *= pTable[(int)(x*1023)]; //alpha
	}
}

void CLS::
ProcessByX(const int axis){
    const int PIni = mPIni, PFin = mPFin;
    const float (* const pX)[3] = mpX;
    float * const pFilterAlpha = mpFilterAlpha;
    float * const pSizeMod = mpSizeMod;
    const unsigned char * const pType = mpType;
    const double factor = mFactor;
    const double factorInv = 1.0 - mFactor;
    const float * const pTable = mpTable;
        
    double width0 = mVRange[1] - mVRange[0];
    if (width0 == 0) width0 = 1;
    const double width = width0;
    const double rangeMin = mVRange[0];
    
    float * const pTgtAttrib = (mAttrib == 1) ? pFilterAlpha : pSizeMod;
    if (pX != NULL){
	for (int i = PIni; i < PFin; i++){
	    if (mType != -1 && pType[i] != mType) continue;
	    if (pFilterAlpha[i] == 0) continue;
	    
	    float x = GetRemapX(width, rangeMin, pX[i][axis]);
	    if (x < 0) x = 0;
	    if (x > 1) x = 1;
	    pTgtAttrib[i] *= pTable[(int)(x*1023)]; //alpha
	}
    }
}


#define CLS ModifyParticleSizeExec_cls

CLS::
ModifyParticleSizeExec_cls(AttributeClone_cls* pCSrc){
    mbValid = pCSrc->GetPCItem("Validity=")->GetBodyBOOL();
    mDependType = pCSrc->GetPCItem("DependType=")->GetBodyINT();
    mType = pCSrc->GetPCItem("TargetType=")->GetBodyINT() - 1; 
    mAttrib = pCSrc->GetPCItem("TargetAttrib=")->GetBodyINT();
    mColorIndex = pCSrc->GetPCItem("ColorCurveIndex=")->GetBodyINT();
    mChannelType = pCSrc->GetPCItem("ChannelType=")->GetBodyINT();
}

void CLS::
Execute(ViewerClone_cls* pCParticle, ParticleForRender_cls *pCCurrent,
	double factor, double NT, bool bNext){
    if (mbValid == FALSE) return;

    float *pFilterAlpha = pCCurrent->mpFilterAlpha;
    float *pSizeMod = pCCurrent->mpSizeMod;
    unsigned char *pType = pCCurrent->mpType;
    long long numParticle = pCCurrent->mPNum;
    
    bool bUseID = (pCParticle->GetPCItem("InterpolateBy=")->GetBodyINT() == 1);
    float (*pX)[3] = NULL;
    float *pValue = NULL;
    int       *pIDSnap1 = NULL;
    long long *pIDSnap1LongLong = NULL;

    switch(mDependType){
    case 1:
	pValue = pCCurrent->mpSize; break;
    case 2:
	pIDSnap1 = pCCurrent->mpID;
	pIDSnap1LongLong = pCCurrent->mpIDLongLong;
	break;
    case 3:
	pValue = pCCurrent->mpValue[0]; break;
    case 4:
	pValue = pCCurrent->mpValue[1];	break;
    case 5:
	pValue = pCCurrent->mpValue[2]; break;
    case 6: case 7: case 8:
	pX = pCCurrent->mpX; break;
    }

    if (pValue == NULL){
	if (mDependType == 1 ||
	    mDependType == 3 || mDependType == 4 || mDependType == 5)
	    return;
    }
    
    AttributeClone_cls *pCClrCurve = pCParticle->GetPCAttribCloneFromAttribName("ColorCurve_cls", mColorIndex);
    if (pCClrCurve == NULL) return;
    
    wxString strClr = pCClrCurve->mCstrName.GetBodySTR();
    ParamGroup_cls *pCClr = pCParticle->GetPCParamGroup(strClr, wxES);
    ParamGroup_cls *pCRange = pCParticle->GetPCParamGroup(strClr, "Range");

    std::vector < double > VRange = pCRange->GetVInterpolatedValue(NT);

    if (mChannelType < 4)
	for (int ix = 0; ix < 1024; ix++){
	    float x = (float)ix/1024;
	    mTable[ix] = pCClr->GetVInterpolatedValue(x)[mChannelType];
	}

    if (mChannelType == 4)
	for (int ix = 0; ix < 1024; ix++){
	    float x = (float)ix/1024;
	    std::vector < float > VClr = pCClr->GetVInterpolatedValue(x);
	    mTable[ix] = (VClr[0] * 0.298912 + VClr[1] *  0.586611 + VClr[2] * 0.114478) * VClr[3];
	}
    if (mChannelType == 5)
	for (int ix = 0; ix < 1024; ix++){
	    float x = (float)ix/1024;
	    std::vector < float > VClr = pCClr->GetVInterpolatedValue(x);
	    float bright = VClr[0];
	    if (VClr[1] > bright) bright = VClr[1];
	    if (VClr[2] > bright) bright = VClr[2];
	    mTable[ix] = bright * VClr[3];
	}

    int threadNumber;
    bool *bAFinished = GetNewPCbFinishList(threadNumber, numParticle, bNext);
    ThreadModifyParticleSize_cls **ppCThread = 
	new ThreadModifyParticleSize_cls* [threadNumber];

    for (int core = 0; core < threadNumber; core++){
	ppCThread[core] = 
	    new ThreadModifyParticleSize_cls;

	int PIni, PFin;
	GetIniAndFin(PIni, PFin, core, threadNumber, numParticle);

	ppCThread[core]->SetParameters(mDependType, pIDSnap1, pIDSnap1LongLong,
				       pX, pValue,
				       pFilterAlpha, pSizeMod, pType,
				       factor, mType, mAttrib, 
				       mTable, VRange);

	ppCThread[core]->CreateThread(&bAFinished[core], PIni, PFin);
	if (threadNumber == 1){
	    ppCThread[core]->DummyEntry();
	    delete ppCThread[core];
	}
	else
	    ppCThread[core]->Run();
    }

    WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 1);
    delete [] ppCThread;
}
