// (setq-c-offset 4)
// "clsParticleBoolean.h"

#include "clsParticleBoolean.h"

#include "../PARAMETER/clsParamGroup.h"
#include "../SCENEOBJ/clsMeshObj.h"
#include "../SCENEOBJ/clsPrimitive.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS ParticleBoolean_cls

CLS::
ParticleBoolean_cls(wxString strName, wxString strObjName) : BooleanBase_cls ("ParticleBoolean_cls", strName, strObjName){
}

#define CLS ParticleBooleanBuffer_cls

CLS::
ParticleBooleanBuffer_cls(BooleanBase_cls *pCSrc) : BooleanBufferBase_cls(pCSrc){
}

#define CLS ThreadParticleBoolean_cls

void CLS::
SetParameters(int booleanType, int booleanShape,
	      float XMin, float XMax, float YMin, float YMax, float ZMin, float ZMax,
	      float radius,
	      float *pFilterAlpha, float (*pX)[3],
	      double matrix16[16], double matrix16TgtInv[16],
	      SceneObj_cls *pCObj){
    mBooleanType = booleanType;
    mBooleanShape = booleanShape;
    mXMin = XMin; mXMax = XMax;
    mYMin = YMin; mYMax = YMax;
    mZMin = ZMin; mZMax = ZMax;
    mRadius = radius;
    mpFilterAlpha = pFilterAlpha;
    mpX = pX;
    memcpy(mMatrix16,       matrix16,       sizeof(double)*16);
    memcpy(mMatrix16TgtInv, matrix16TgtInv, sizeof(double)*16);
    mpCObj = pCObj;
}

wxThread::ExitCode CLS::
Entry(){
    int PIni = mPIni, PFin = mPFin;
    float *pFilterAlpha = mpFilterAlpha;
    float (*pX)[3] = mpX;
    float XMin = mXMin, XMax = mXMax;
    float YMin = mYMin, YMax = mYMax;
    float ZMin = mZMin, ZMax = mZMax;
    float radius2 = mRadius * mRadius;

    Multiply(mMatrix16TgtInv, mMatrix16);

    if (mBooleanType == _BOOL_INTERSECT){ 
	if (mBooleanShape == _BOOL_BOX)
	    for (int i = PIni; i < PFin; i++){
		if (pX[i][X] < XMin || pX[i][X] > XMax) pFilterAlpha[i] *= 0;
		if (pX[i][Y] < YMin || pX[i][Y] > YMax) pFilterAlpha[i] *= 0;
		if (pX[i][Z] < ZMin || pX[i][Z] > ZMax) pFilterAlpha[i] *= 0;
	    }
	if (mBooleanShape == _BOOL_SPHERE)
	    for (int i = PIni; i < PFin; i++){
		double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
		MultiplyVFast(mMatrix16, pos);
		float distance2 = (pos[X]*pos[X] + pos[Y]*pos[Y] + pos[Z]*pos[Z]);
		if (distance2 > radius2) pFilterAlpha[i] *= 0;
	    }
	if (mBooleanShape == _BOOL_OBJECT){
	    if (mpCObj != NULL){
		if (mpCObj->mstrClassName.IsSameAs("Primitive_cls")){
		    Primitive_cls *pCPrimitive = (Primitive_cls*)mpCObj;
		    for (int i = PIni; i < PFin; i++){
			double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
			MultiplyVFast(mMatrix16, pos);
			if (pCPrimitive->IsInside(pos[X], pos[Y], pos[Z]) == FALSE)
			    pFilterAlpha[i] *= 0;
		    }
		}
		if (mpCObj->mstrClassName.IsSameAs("MeshObj_cls")){
		    MeshObj_cls *pCMesh = (MeshObj_cls*)mpCObj;
		    MeshBase2_cls &CMesh2 = pCMesh->mCMesh2;

		    int tagBufferSize = CMesh2.GetTagBufferSize();
		    int *pVertexTag = new int [tagBufferSize];

		    for (int i = PIni; i < PFin; i++){
			double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
			MultiplyVFast(mMatrix16, pos);
			if (CMesh2.IsInside(pos[X], pos[Y], pos[Z],
					    tagBufferSize, pVertexTag) == FALSE)
			    pFilterAlpha[i] *= 0;
		    }

		    delete [] pVertexTag;
		}
	    }
	}
    }

    if (mBooleanType == _BOOL_DIFFERECE){ 
	if (mBooleanShape == _BOOL_BOX)
	    for (int i = PIni; i < PFin; i++){
		if (pX[i][X] > XMin && pX[i][X] < XMax &&
		    pX[i][Y] > YMin && pX[i][Y] < YMax &&
		    pX[i][Z] > XMin && pX[i][Z] < XMax) pFilterAlpha[i] *= 0;
	    }
	
	if (mBooleanShape == _BOOL_SPHERE){
	    if (mBooleanShape == _BOOL_SPHERE)
		for (int i = PIni; i < PFin; i++){
		    double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
		    MultiplyVFast(mMatrix16, pos);
		    float distance2 = (pos[X]*pos[X] + pos[Y]*pos[Y] + pos[Z]*pos[Z]);
		    if (distance2 < radius2) pFilterAlpha[i] *= 0;
		}
	}
	if (mBooleanShape == _BOOL_OBJECT){
	    if (mpCObj != NULL){
		if (mpCObj->mstrClassName.IsSameAs("Primitive_cls")){
		    Primitive_cls *pCPrimitive = (Primitive_cls*)mpCObj;
		    for (int i = PIni; i < PFin; i++){
			double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
			MultiplyVFast(mMatrix16, pos);
			if (pCPrimitive->IsInside(pos[X], pos[Y], pos[Z]) == TRUE)
			    pFilterAlpha[i] *= 0;
		    }
		}
		if (mpCObj->mstrClassName.IsSameAs("MeshObj_cls")){
		    MeshObj_cls *pCMesh = (MeshObj_cls*)mpCObj;
		    MeshBase2_cls &CMesh2 = pCMesh->mCMesh2;

		    int tagBufferSize = CMesh2.GetTagBufferSize();
		    int *pVertexTag = new int [tagBufferSize];

		    for (int i = PIni; i < PFin; i++){
			double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
			MultiplyVFast(mMatrix16, pos);
			if (CMesh2.IsInside(pos[X], pos[Y], pos[Z],
					    tagBufferSize, pVertexTag) == TRUE)
			    pFilterAlpha[i] *= 0;
		    }

		    delete [] pVertexTag;
		}
	    }
	}
    }
    if (mBooleanType == _BOOL_ADD){ 
	if (mBooleanShape == _BOOL_BOX)
	    for (int i = PIni; i < PFin; i++){
		if (pX[i][X] > XMin && pX[i][X] < XMax &&
		    pX[i][Y] > YMin && pX[i][Y] < YMax &&
		    pX[i][Z] > XMin && pX[i][Z] < XMax) pFilterAlpha[i] = 1;
	    }
	
	if (mBooleanShape == _BOOL_SPHERE){
	    if (mBooleanShape == _BOOL_SPHERE)
		for (int i = PIni; i < PFin; i++){
		    double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
		    MultiplyVFast(mMatrix16, pos);
		    float distance2 = (pos[X]*pos[X] + pos[Y]*pos[Y] + pos[Z]*pos[Z]);
		    if (distance2 < radius2) pFilterAlpha[i] = 1;
		}
	}
	if (mBooleanShape == _BOOL_OBJECT){
	    if (mpCObj != NULL){
		if (mpCObj->mstrClassName.IsSameAs("Primitive_cls")){
		    Primitive_cls *pCPrimitive = (Primitive_cls*)mpCObj;
		    for (int i = PIni; i < PFin; i++){
			double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
			MultiplyVFast(mMatrix16, pos);
			if (pCPrimitive->IsInside(pos[X], pos[Y], pos[Z]) == TRUE)
			    pFilterAlpha[i] = 1;
		    }
		}
		if (mpCObj->mstrClassName.IsSameAs("MeshObj_cls")){
		    MeshObj_cls *pCMesh = (MeshObj_cls*)mpCObj;
		    MeshBase2_cls &CMesh2 = pCMesh->mCMesh2;

		    int tagBufferSize = CMesh2.GetTagBufferSize();
		    int *pVertexTag = new int [tagBufferSize];

		    for (int i = PIni; i < PFin; i++){
			double pos[4] = {pX[i][X], pX[i][Y], pX[i][Z], 1};
			MultiplyVFast(mMatrix16, pos);
			if (CMesh2.IsInside(pos[X], pos[Y], pos[Z],
					    tagBufferSize, pVertexTag) == TRUE)
			    pFilterAlpha[i] = 1;
		    }
		    
		    delete [] pVertexTag;
		}
	    }
	}
    }
   
    *mpbFinishedFlagLink = TRUE;
    return 0;
}

#define CLS ParticleBooleanExec_cls

CLS::
ParticleBooleanExec_cls(AttributeClone_cls* pCSrc, double NT,
			double matrix16[16], double matrix16TgtInv[16], 
			SceneObj_cls *pCObj){
    mpCSrc = pCSrc;
    mNT = NT;
    memcpy(mMatrix16,       matrix16,       sizeof(double)*16);
    memcpy(mMatrix16TgtInv, matrix16TgtInv, sizeof(double)*16);
    mpCObj = pCObj;

    mbValid = pCSrc->GetPCItem("Validity=")->GetBodyBOOL();
    mBooleanType = pCSrc->GetPCItem("BooleanType=")->GetBodyINT();
    mBooleanShape = pCSrc->GetPCItem("BooleanShape=")->GetBodyINT();
    mstrBooleanObject = pCSrc->GetPCItem("BooleanObject")->GetBodySTR();
        
    mXMin = pCSrc->GetPCParamGroup("X Min/Max")->GetVInterpolatedValue(NT)[0];
    mXMax = pCSrc->GetPCParamGroup("X Min/Max")->GetVInterpolatedValue(NT)[1];
    mYMin = pCSrc->GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue(NT)[0];
    mYMax = pCSrc->GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue(NT)[1];
    mZMin = pCSrc->GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue(NT)[0];
    mZMax = pCSrc->GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue(NT)[1];
    mRadius = pCSrc->GetPCParamGroup("Radius")->GetVInterpolatedValue(NT)[0];
}

void CLS::
CheckBoolean(float *pFilterAlpha, float (*pX)[3],
	     long long numParticle, bool bNext){
    if (mbValid == FALSE) return;

    int threadNumber;
    bool *bAFinished = GetNewPCbFinishList(threadNumber, numParticle, bNext);

    ThreadParticleBoolean_cls **ppCThreadBoolean = 
	new ThreadParticleBoolean_cls* [threadNumber];

    for (int core = 0; core < threadNumber; core++){
	ppCThreadBoolean[core] = 
	    new ThreadParticleBoolean_cls;

	int PIni, PFin;
	GetIniAndFin(PIni, PFin, core, threadNumber, numParticle);

	ppCThreadBoolean[core]->SetParameters
	    (mBooleanType, mBooleanShape,
	     mXMin, mXMax, mYMin, mYMax, mZMin, mZMax, mRadius, 
	     pFilterAlpha, pX,
	     mMatrix16, mMatrix16TgtInv, mpCObj);
	
	ppCThreadBoolean[core]->CreateThread
	    (&bAFinished[core], PIni, PFin);

	ppCThreadBoolean[core]->Run();

    }

    WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 3);

    delete [] ppCThreadBoolean;
}
