// (setq-c-offset 4)
// "clsPartilceDownsize.h"

#include "clsParticleExclude.h"

#include "../SCENEOBJ/clsSceneObj.h"
#include "../SIMDATA/clsParticleForRender.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"

#include <algorithm>

#define CLS ParticleExclude_cls

CLS::
ParticleExclude_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("ParticleExclude_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_DATAPROC";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCstrTarget, "Target=", "None", strUpdate);
}

void CLS::
SyncFromBuffer(ParticleExcludeBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
}

#define CLS ParticleExcludeBuffer_cls

CLS::
ParticleExcludeBuffer_cls(ParticleExclude_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCstrTarget, "Target=", "None");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineConfItem(mCstrTarget, _("Target"));

    pC->FinishSetting();
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}

#define CLS ParticleExcludeExe_cls

CLS::
ParticleExcludeExe_cls(AttributeClone_cls* pCSrc, ParticleForRender_cls *pCTo, ParticleForRender_cls *pCFrom){
    mpCTo = pCTo;
    mpCFrom = pCFrom;
}

void CLS::
Execute(){
    ParticleForRender_cls *pCTo = mpCTo;
    ParticleForRender_cls *pCFrom = mpCFrom;

    wxString strLog;
    strLog << "Exclude" << pCTo->mPNum << ":" << pCFrom->mPNum <<ENDL;
    if (pCFrom->mpIDLongLong != NULL) 
	strLog << "pCFrom->mpIDLongLong != NULL" << ENDL;
    if (pCTo->mpIDLongLong != NULL) 
	strLog << "pCTo->mpIDLongLong != NULL" << ENDL;
    wxLogMessage(strLog);
    
    if (pCFrom->mPNum <= 0) return;
    if (pCTo->mPNum <=0) return;

    if (pCTo->mpIDLongLong != NULL) return ExecuteIDLongLong();
    
    return ExecuteID();
}

void CLS::
ExecuteID(){
    ParticleForRender_cls *pCTo = mpCTo;
    ParticleForRender_cls *pCFrom = mpCFrom;
    int numTo = mpCTo->mPNum;
    int numFrom = mpCFrom->mPNum;

    ASSERT(mpCFrom->mpID != NULL);
    if (mpCFrom->mpID == NULL) return;

    class IDArray{
    public:
	int mIndex;
	int mID;
	
	int operator < (IDArray rhs) const {
	    return mID < rhs.mID; }
	int operator <= (IDArray rhs) const {
	    return mID <= rhs.mID; }
	int operator > (IDArray rhs) const {
	    return mID > rhs.mID; }
	int operator >= (IDArray rhs) const {
	    return mID >= rhs.mID; }
    };

    IDArray* pIDArrayTo = new IDArray [numTo];
    IDArray* pIDArrayFrom = new IDArray [numFrom];

    for (int i = 0; i < numTo; i++){
	pIDArrayTo[i].mIndex = i;
	pIDArrayTo[i].mID = pCTo->mpID[i];
    }
    for (int i = 0; i < numFrom; i++){
	pIDArrayFrom[i].mIndex = i;
	pIDArrayFrom[i].mID = pCFrom->mpID[i];
    }

    /*
    // \[g̓f[^ǂݍݎɍŝŕKvȂ͂iHj
    std::sort(pIDArrayTo, pIDArrayTo+numTo);
    std::sort(pIDArrayFrom, pIDArrayFrom+numFrom);
    */

    int tgt = 0;
    for (int i = 0; i < numTo; i++){
	int idx = pIDArrayTo[i].mIndex;
	int id = pIDArrayTo[i].mID;

	while(TRUE){
	    if (tgt >= numFrom) break;

	    if (pIDArrayFrom[tgt].mID == id){
		if (pCFrom->mpSize[tgt] != 0)
		    pCTo->mpFilterAlpha[idx] = 0;
		break;
	    }
	    if (pIDArrayFrom[tgt].mID > id) break;
	    tgt++;
	}
    }
    
    delete [] pIDArrayTo;
    delete [] pIDArrayFrom;
}

void CLS::
ExecuteIDLongLong(){
    ParticleForRender_cls *pCTo = mpCTo;
    ParticleForRender_cls *pCFrom = mpCFrom;
    int numTo = mpCTo->mPNum;
    int numFrom = mpCFrom->mPNum;

    ASSERT(mpCFrom->mpIDLongLong != NULL);
    if (mpCFrom->mpIDLongLong == NULL) return;

    wxString strLog;
    strLog << "Exclude" << pCTo->mpIDLongLong[0] << ":" << pCFrom->mpIDLongLong[0] <<ENDL;
    wxLogMessage(strLog);
    
    class IDArrayLongLong{
    public:
	int mIndex;
	long long mID;

	int operator < (IDArrayLongLong rhs) const {
	    return mID < rhs.mID; }
	int operator <= (IDArrayLongLong rhs) const {
	    return mID <= rhs.mID; }
	int operator > (IDArrayLongLong rhs) const {
	    return mID > rhs.mID; }
	int operator >= (IDArrayLongLong rhs) const {
	    return mID >= rhs.mID; }
    };

    IDArrayLongLong* pIDArrayTo = new IDArrayLongLong [numTo];
    IDArrayLongLong* pIDArrayFrom = new IDArrayLongLong [numFrom];

    for (int i = 0; i < numTo; i++){
	pIDArrayTo[i].mIndex = i;
	pIDArrayTo[i].mID = pCTo->mpIDLongLong[i];
    }

    for (int i = 0; i < numFrom; i++){
	pIDArrayFrom[i].mIndex = i;
	pIDArrayFrom[i].mID = pCFrom->mpIDLongLong[i];
    }
    
    /*
    // \[g̓f[^ǂݍݎɍŝŕKvȂ͂iHj
    std::sort(pIDArrayTo, pIDArrayTo+numTo);
    std::sort(pIDArrayFrom, pIDArrayFrom+numFrom);
    */

    int tgt = 0;
    for (int i = 0; i < numTo; i++){
	int idx = pIDArrayTo[i].mIndex;
	long long id = pIDArrayTo[i].mID;

	while(TRUE){
	    if (tgt >= numFrom) break;

	    if (pIDArrayFrom[tgt].mID == id){
		if (pCFrom->mpSize[tgt] != 0)
		    pCTo->mpFilterAlpha[idx] = 0;
		break;
	    }
	    if (pIDArrayFrom[tgt].mID > id) break;
	    tgt++;
	}
    }
    
    delete [] pIDArrayTo;
    delete [] pIDArrayFrom;
}
