// (setq-c-offset 4)
// "clsPartilceFiltering.h"

#include "clsParticleFiltering.h"

#include "../SCENEOBJ/clsSceneObj.h"
#include "../SIMDATA/clsArrayAndIndex.h"
#include "../SIMDATA/clsParticleForRender.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#define CLS ParticleFiltering_cls

CLS::
ParticleFiltering_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("ParticleFiltering_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_DATAPROC";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCFilteringType, "FilteringType=", "0", strUpdate);
    mCConfList.Add(&mCIndexType, "IndexType=", "0", "VIEW3D");
    mCConfList.Add(&mCstrString, "FilteringString=", "0", "VIEW3D");
}

void CLS::
SyncFromBuffer(ParticleFilteringBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
}

#define CLS ParticleFilteringBuffer_cls

CLS::
ParticleFilteringBuffer_cls(ParticleFiltering_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCFilteringType, "FilteringType=", "0");
    mCConfList.Add(&mCIndexType, "IndexType=", "0");
    mCConfList.Add(&mCstrString, "FilteringString=", "0");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    wxString AstrFilterType[3] =
	{_("Show following particles"), _("Hide following particles"), _("Force showing")};
    pC->SetChoiceList(3, AstrFilterType);
    pC->CombineConfItem(mCFilteringType, _("FilteringType"));

    wxString AstrIndexType[2] = {_("Particle index"), _("Particle ID")};
    pC->SetChoiceList(2, AstrIndexType);
    pC->CombineConfItem(mCIndexType, _("Classify by"));

    pC->CombineConfItem(mCstrString, _("Particles"));
    
    pC->FinishSetting();
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}

#define CLS ParticleFilteringExec_cls

CLS::
ParticleFilteringExec_cls(AttributeClone_cls* pCSrc){
    mbValid = pCSrc->GetPCItem("Validity=")->GetBodyBOOL();
    mType = pCSrc->GetPCItem("FilteringType=")->GetBodyINT();
    mIndexType = pCSrc->GetPCItem("IndexType=")->GetBodyINT();
    mstrString = pCSrc->GetPCItem("FilteringString=")->GetBodySTR();
}

void CLS::
CheckFilteringIDLongLong(float *pFilterAlpha, ArrayAndIndex_cls *pIndexAndZ, 
			 bool *pbFilterList,
			 long long numParticle, long long *pIDLongLong){
    switch(mType){
    case _FILTERING_SHOW: {
	for (int i = 0; i < numParticle; i++){
	    long long index = pIndexAndZ[i].mIndex;
	    long long id = pIDLongLong[index];
	    if (id >= mFilterListMax)
		pFilterAlpha[i] *= 0;
	    else
		pFilterAlpha[i] *= pbFilterList[id];
	}
	break; }
    case _FILTERING_HIDE: {
	for (int i = 0; i < numParticle; i++){
	    long long index = pIndexAndZ[i].mIndex;
	    long long id = pIDLongLong[index];
	    if (id >= mFilterListMax)
		pFilterAlpha[i] *= 1;
	    else
		pFilterAlpha[i] *= !pbFilterList[id];
	}
	break; }
    case _FILTERING_ADD: {
	for (int i = 0; i < numParticle; i++){
	    long long index = pIndexAndZ[i].mIndex;
	    long long id = pIDLongLong[index];
	    if (id >= mFilterListMax)
		continue;
	    else
		if (pbFilterList[id] == TRUE)
		    pFilterAlpha[i] = 1;
	}
	break; }
    }
}

void CLS::
CheckFilteringID(float *pFilterAlpha, ArrayAndIndex_cls *pIndexAndZ, 
		 bool *pbFilterList, long long numParticle, int *pID){
    switch(mType){
    case _FILTERING_SHOW: {
	for (int i = 0; i < numParticle; i++){
	    long long index = pIndexAndZ[i].mIndex;
	    int id = pID[index];
	    if (id >= mFilterListMax)
		pFilterAlpha[i] *= 0;
	    else
		pFilterAlpha[i] *= pbFilterList[id];
	}
	break; }
    case _FILTERING_HIDE: {
	for (int i = 0; i < numParticle; i++){
	    long long index = pIndexAndZ[i].mIndex;
	    int id = pID[index];
	    if (id >= mFilterListMax)
		pFilterAlpha[i] *= 1;
	    else
		pFilterAlpha[i] *= !pbFilterList[id];
	}
	break; }
    case _FILTERING_ADD: {
	for (int i = 0; i < numParticle; i++){
	    long long index = pIndexAndZ[i].mIndex;
	    int id = pID[index];
	    if (id >= mFilterListMax)
		continue;
	    else
		if (pbFilterList[id] == TRUE)
		    pFilterAlpha[i] = 1;
	}
	break; }
    }
}

void CLS::
CheckFiltering(float *pFilterAlpha, ArrayAndIndex_cls *pIndexAndZ, 
	       ParticleForRender_cls *pCParticle){
    if (mbValid == FALSE) return;

    long long numParticle = pCParticle->mPNum;

    bool *pbFilterList = GetNewPbFilterList(numParticle);

    if (pbFilterList == NULL) return;
    
    if (mIndexType == 0){
	switch(mType){
	case _FILTERING_SHOW: {
	    for (int i = 0; i < numParticle; i++){
		long long index = pIndexAndZ[i].mIndex;
		if (index >= mFilterListMax)
		    pFilterAlpha[i] *= 0;
		else
		    pFilterAlpha[i] *= pbFilterList[index];
	    }
	    break; }
	case _FILTERING_HIDE: {
	    for (int i = 0; i < numParticle; i++){
		long long index = pIndexAndZ[i].mIndex;
		if (index >= mFilterListMax)
		    pFilterAlpha[i] *= 1;
		else
		    pFilterAlpha[i] *= !pbFilterList[index];
	    }
	    break; }
	case _FILTERING_ADD: {
	    for (int i = 0; i < numParticle; i++){
		long long index = pIndexAndZ[i].mIndex;
		if (index >= mFilterListMax)
		    continue;
		else
		    if (pbFilterList[index] == TRUE)
			pFilterAlpha[i] = 1;
	    }
	    break; }
	}
    }

    if (mIndexType == 1){
	if (pCParticle->mpIDLongLong != NULL)
	    CheckFilteringIDLongLong(pFilterAlpha, pIndexAndZ, pbFilterList, numParticle, pCParticle->mpIDLongLong);
	else if (pCParticle->mpID != NULL)
	    CheckFilteringID(pFilterAlpha, pIndexAndZ, pbFilterList, numParticle, pCParticle->mpID);
    }

    delete [] pbFilterList;
}

bool* CLS::
GetNewPbFilterList(int numParticle){

    wxString strAll = mstrString;
    
    long long indexMax = 0;

    while (strAll.IsEmpty() == FALSE){
	wxString strBlock = GetStringBlock(&strAll);
	long long ini;GetStrIni(strBlock).ToLongLong(&ini);

	long long fin;
	wxString strFin = GetStrFin(strBlock);
	if (strFin.IsSameAs(wxT("-")) )
	    fin = ini;
	else
	    GetStrFin(strBlock).ToLongLong(&fin);
	
	if (indexMax < fin+1) indexMax = fin+1;
    }

    if (mIndexType == 0)
	if (indexMax > numParticle) indexMax = numParticle;

    if (indexMax >= 41942784) return NULL;
    
    bool * const pbFilterList = new bool [indexMax];

    for (long long i = 0; i < indexMax; i++) pbFilterList[i] = FALSE;

    strAll = mstrString;
    while (strAll.IsEmpty() == FALSE){
	wxString strBlock = GetStringBlock(&strAll);
	long long ini;GetStrIni(strBlock).ToLongLong(&ini);
	if (ini < 0) ini = 0;

	long long fin;
	wxString strFin = GetStrFin(strBlock);
	if (strFin.IsSameAs(wxT("-")) )
	    fin = ini;
	else
	    GetStrFin(strBlock).ToLongLong(&fin);
	if (fin > indexMax-1) fin = indexMax-1;

	long long inc;GetStrInc(strBlock).ToLongLong(&inc);
	if (inc < 1) inc = 1;
	
	for (long long i = ini; i <= fin; i += inc)
	    pbFilterList[i] = TRUE;
    }

    mFilterListMax = indexMax;
    return pbFilterList;
}
