// (setq-c-offset 4)
// "clsPartilceTweak.h"

#include "clsParticleTweak.h"

#include "../SCENEOBJ/clsSceneObj.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#define CLS ParticleTweak_cls

CLS::
ParticleTweak_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("ParticleTweak_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_MODIFIER";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCBillboardFocusRotOrder, "BillboardRotOrder=", "0", strUpdate);
    mCConfList.Add(&mCbStoreValue, "StoreValue=", "0", strUpdate);
    mCConfList.Add(&mCbStoreIDLongLong, "StoreIDLongLong=", "0", strUpdate);
    
    ParamGroup_cls *pCParamGroupBillboardFocus = GetPCNewParamGroup("BillboardFocus", "Y", "P", "R");
    InitializeAndPushParam(pCParamGroupBillboardFocus, 0, -60, 0);
}

void CLS::
SyncFromBuffer(ParticleTweakBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VBillboardFocus;
    VBillboardFocus.push_back(pCBuffer->mCBillboardFocusYaw.GetBodyDOUBLE());
    VBillboardFocus.push_back(pCBuffer->mCBillboardFocusPitch.GetBodyDOUBLE());
    VBillboardFocus.push_back(pCBuffer->mCBillboardFocusRaw.GetBodyDOUBLE());
    SyncronizeParamForModifier("BillboardFocus", VBillboardFocus, pCObj);
}

void CLS::
GetBillboartFocus(float &yaw, float &pitch, float &row, int &rotOrder){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    yaw = GetPCParamGroup("BillboardFocus")->GetVInterpolatedValue(NT, TRUE)[0];
    pitch = GetPCParamGroup("BillboardFocus")->GetVInterpolatedValue(NT, TRUE)[1];
    row = GetPCParamGroup("BillboardFocus")->GetVInterpolatedValue(NT, TRUE)[2];
    rotOrder = mCBillboardFocusRotOrder.GetBodyINT();
}

#define CLS ParticleTweakBuffer_cls

CLS::
ParticleTweakBuffer_cls(ParticleTweak_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCBillboardFocusRotOrder, "BillboardRotOrder=", "0");
    mCConfList.Add(&mCbStoreValue, "StoreValue=", "0");
    mCConfList.Add(&mCbStoreIDLongLong, "StoreIDLongLong=", "0");
    
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    /*
    wxString AstrDownsizeType[3] =
	{_("by index"), _("by ID")};
    pC->SetChoiceList(2, AstrDownsizeType);
    pC->CombineConfItem(mCDownsizeType, _("Tweak Type"));
    */
    
    //pC->CombineConfItem(mCDownsize);

    pC->mTextWidth = 160;
    pC->CombineSizer(3, _("Billboard Focus (YPR)"));
    pC->mDragScale = 10;
    pC->CombineConfItem(mCBillboardFocusYaw);
    pC->mDragScale = 10;
    pC->CombineConfItem(mCBillboardFocusPitch);
    pC->mDragScale = 10;
    pC->CombineConfItem(mCBillboardFocusRaw);

    wxString AstrOrder[2] = {"YPR", "RPY"};
    pC->SetChoiceList(2, AstrOrder);
    pC->CombineConfItem(mCBillboardFocusRotOrder, _("Rotate Order"));

    pC->CombineSizer(1, _("Store Values*"));
    int indexStore = pC->CombineConfItem(mCbStoreValue);

    pC->CombineSizer(1, _("Store ID(long long)"));
    int indexIDLongLong = pC->CombineConfItem(mCbStoreIDLongLong);

    pC->FinishSetting();

    wxString strTipCommon(_("A few modifiers use them. If unchecked, you can save memory usage, but some modifiers become invalid."));
    pC->mVCItem[indexStore].mpCWindowLink->SetToolTip((wxString)_("Store values* in interpolated data structure.") + " " + strTipCommon);
    pC->mVCItem[indexIDLongLong].mpCWindowLink->SetToolTip((wxString)_("Store ID(long long) in interpolated data structure.") + " " + strTipCommon);

    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);

    double time = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VYPR = mpCSrc->GetPCParamGroup("BillboardFocus")->GetVInterpolatedValue((float)time, FALSE);
    mCBillboardFocusYaw.SetBodyFromData(VYPR[0]);
    mCBillboardFocusPitch.SetBodyFromData(VYPR[1]);
    mCBillboardFocusRaw.SetBodyFromData(VYPR[2]);
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}
