// (setq-c-offset 4)
// "clsSequenceRemap.h"

#include "clsSequenceRemap.h"

#include "../MISC/clsMisc.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS SequenceRemap_cls

CLS::
SequenceRemap_cls(wxString strName, wxString strObjName) : ModifierBase_cls("SequenceRemap_cls", strName){

    wxString strUpdate = "VIEW3D::PROPERTY_MODIFIER";
    
    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCbClamp, "Clamp=", "1", strUpdate);
    mCConfList.Add(&mCbRepeat, "Repeat=", "0", strUpdate);
    
    ParamGroup_cls *pCRemap = GetPCNewParamGroup("Remap", "idx");
    InitializeAndPushParam(pCRemap, 0);
}

int CLS::
GetRemappedFrame(double NT, int frame, int seqLen){
    if (mCbValidity.GetBodyBOOL() == FALSE) return frame;
    double dframe = GetPCParamGroup("Remap")->GetVInterpolatedValue(NT)[0];
    frame = floor(dframe + 1e-7);

    if (seqLen != -1 && frame >= seqLen){
	bool bRepeat = mCbRepeat.GetBodyBOOL(); 
	if (bRepeat) frame = Misc_cls::Surplus(frame, seqLen);
    }

    if (mCbClamp.GetBodyBOOL() == TRUE){
	if (frame >= seqLen) frame = seqLen - 1;
	if (frame < 0) frame = 0;
    }

    return frame;
}

double CLS::
GetInterpFactor(double NT, int frame, double interpFactor){
    if (mCbValidity.GetBodyBOOL() == FALSE) return interpFactor;
    double dframe = GetPCParamGroup("Remap")->GetVInterpolatedValue(NT)[0];
    double interpFrame = dframe - floor(dframe + 1e-7);
    if (interpFrame < 1e-7) interpFrame = 0;
    return interpFrame;
}
    
void CLS::
SyncFromBuffer(SequenceRemapBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
    
    std::vector < float > VRemap;
    VRemap.push_back(pCBuffer->mCRemap.GetBodyDOUBLE());
    SyncronizeParamForModifier("Remap", VRemap, pCObj);
}

#define CLS SequenceRemapBuffer_cls

CLS::
SequenceRemapBuffer_cls(SequenceRemap_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCbClamp, "Clamp=", "1");
    mCConfList.Add(&mCbRepeat, "Repeat=", "0");
    mCConfList.Add(&mCRemap, "Remap=", "0");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
		new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(mCbClamp, _("Clamp"));
    pC->CombineConfItem(mCbRepeat, _("Repeat"));
    pC->CombineConfItem(mCRemap, _("Remap"));

    pC->FinishSetting();
    
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();
    mCRemap.SetBodyFromData(mpCSrc->GetPCParamGroup("Remap")->GetVInterpolatedValue((float)time, FALSE));
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}
