// (setq-c-offset 4)
// "clsStreamFiltering.h"

#include "clsStreamFiltering.h"

#include "../SCENEOBJ/clsSceneObj.h"
#include "../SIMDATA/clsArrayAndIndex.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#define CLS StreamFiltering_cls

CLS::
StreamFiltering_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("StreamFiltering_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_DATAPROC";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCFilteringType, "FilteringType=", "0", strUpdate);
    //mCConfList.Add(&mCIndexType, "IndexType=", "0", "VIEW3D");
    mCConfList.Add(&mCstrString, "FilteringString=", "0", "VIEW3D");
}

void CLS::
SyncFromBuffer(StreamFilteringBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
}

#define CLS StreamFilteringBuffer_cls

CLS::
StreamFilteringBuffer_cls(StreamFiltering_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCFilteringType, "FilteringType=", "0");
    //mCConfList.Add(&mCIndexType, "IndexType=", "0");
    mCConfList.Add(&mCstrString, "FilteringString=", "0");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    wxString AstrFilterType[3] =
	{_("Show following Streams"), _("Hide following Streams"), _("Force showing")};
    pC->SetChoiceList(3, AstrFilterType);
    pC->CombineConfItem(mCFilteringType, _("FilteringType"));

    pC->CombineConfItem(mCstrString, _("Streams"));
    
    pC->FinishSetting();
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}

#define CLS StreamFilteringExec_cls

CLS::
StreamFilteringExec_cls(AttributeClone_cls* pCSrc){
    mbValid = pCSrc->GetPCItem("Validity=")->GetBodyBOOL();
    mType = pCSrc->GetPCItem("FilteringType=")->GetBodyINT();
    //mIndexType = pCSrc->GetPCItem("IndexType=")->GetBodyINT();
    mstrString = pCSrc->GetPCItem("FilteringString=")->GetBodySTR();
}

void CLS::
CheckFiltering(float *pFilterAlpha, int *mpLineIndex,
	       long long numStreamVertex){
    if (mbValid == FALSE) return;

    bool *pbFilterList = GetNewPbFilterList(numStreamVertex);
    
    switch(mType){
    case _FILTERING_SHOW: {
	for (int i = 0; i < numStreamVertex; i++){
	    long long index = mpLineIndex[i];
	    //long long index = i;
	    if (index >= mFilterListMax)
		pFilterAlpha[i] *= 0;
	    else
		pFilterAlpha[i] *= pbFilterList[index];
	}
	break; }
    case 1: {
	for (int i = 0; i < numStreamVertex; i++){
	    long long index = mpLineIndex[i];
	    //long long index = i;
	    if (index >= mFilterListMax)
		pFilterAlpha[i] *= 1;
	    else
		pFilterAlpha[i] *= !pbFilterList[index];
	}
	break; }
    case 2: {
	for (int i = 0; i < numStreamVertex; i++){
	    long long index = mpLineIndex[i];
	    //long long index = i;
	    if (index >= mFilterListMax)
		continue;
	    else
		if (pbFilterList[index] == TRUE)
		    pFilterAlpha[i] = 1;
	}
	break; }
    }

    delete [] pbFilterList;
}

bool* CLS::
GetNewPbFilterList(int numStreamVertex){

    wxString strAll = mstrString;
    
    long long indexMax = 0;

    while (strAll.IsEmpty() == FALSE){
	wxString strBlock = GetStringBlock(&strAll);
	long long ini;GetStrIni(strBlock).ToLongLong(&ini);

	long long fin;
	wxString strFin = GetStrFin(strBlock);
	if (strFin.IsSameAs(wxT("-")) )
	    fin = ini;
	else
	    GetStrFin(strBlock).ToLongLong(&fin);
	
	if (indexMax < fin+1) indexMax = fin+1;
    }

    /*
    if (mIndexType == 0)
	if (indexMax > numStream) indexMax = numStream;
    */

    bool * const pbFilterList = new bool [indexMax];

    for (long long i = 0; i < indexMax; i++) pbFilterList[i] = FALSE;

    strAll = mstrString;
    while (strAll.IsEmpty() == FALSE){
	wxString strBlock = GetStringBlock(&strAll);
	long long ini;GetStrIni(strBlock).ToLongLong(&ini);
	if (ini < 0) ini = 0;

	long long fin;
	wxString strFin = GetStrFin(strBlock);
	if (strFin.IsSameAs(wxT("-")) )
	    fin = ini;
	else
	    GetStrFin(strBlock).ToLongLong(&fin);
	if (fin > indexMax-1) fin = indexMax-1;

	long long inc;GetStrInc(strBlock).ToLongLong(&inc);
	if (inc < 1) inc = 1;
	
	for (long long i = ini; i <= fin; i += inc)
	    pbFilterList[i] = TRUE;
    }

    mFilterListMax = indexMax;
    return pbFilterList;
}
