// (setq-c-offset 4)
// "clsStreamTransform.h"

#include "clsStreamTransform.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SIMDATA/clsArrayAndIndex.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS StreamTransform_cls

CLS::
StreamTransform_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("StreamTransform_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_DATAPROC";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCOffset, "Offset=", "0 0 0", strUpdate);
    mCConfList.Add(&mCRotate, "Rotate=", "0 0 0", strUpdate);
    mCConfList.Add(&mCRotateOrder, "RotateOrder=", "0", strUpdate);
    mCConfList.Add(&mCScale, "Scale=", "1 1 1", strUpdate);
    
    ParamGroup_cls *pCOffset = GetPCNewParamGroup("Offset", "x", "y", "z");
    InitializeAndPushParam(pCOffset, 0, 0, 0);

    ParamGroup_cls *pCRotate = GetPCNewParamGroup("Rotate", "x", "y", "z");
    InitializeAndPushParam(pCRotate, 0, 0, 0);

    ParamGroup_cls *pCScale = GetPCNewParamGroup("Scale", "x", "y", "z");
    InitializeAndPushParam(pCScale, 1, 1, 1);
}

void CLS::
SyncFromBuffer(StreamTransformBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VOffset = pCBuffer->mCOffset.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("Offset", VOffset, pCObj);
    
    std::vector < float > VRotate = pCBuffer->mCRotate.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("Rotate", VRotate, pCObj);

    std::vector < float > VScale = pCBuffer->mCScale.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("Scale", VScale, pCObj);
}

#define CLS StreamTransformBuffer_cls

CLS::
StreamTransformBuffer_cls(StreamTransform_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.Add(&mCOffset, "Offset=", "0 0 0", "VIEW3D::PROPERTY_DATAPROC");
    mCConfList.Add(&mCRotate, "Rotate=", "0 0 0", "VIEW3D::PROPERTY_DATAPROC");
    mCConfList.Add(&mCRotateOrder, "RotateOrder=", "0", "VIEW3D::PROPERTY_DATAPROC");
    mCConfList.Add(&mCScale, "Scale=", "1 1 1", "VIEW3D::PROPERTY_DATAPROC");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->mDragScale = 1;
    int indexT = pC->CombineConfItem(mCOffset, _("Offset"));
			
    pC->mDragScale = 3;
    int indexR = pC->CombineConfItem(mCRotate, _("Rotate"));
    wxString AstrRotateOrder[6] =
	{"Z-Y-X", "Z-X-Y", "X-Y-Z", "X-Z-Y", "Y-X-Z", "Y-Z-X"};
    pC->SetChoiceList(6, AstrRotateOrder);
    pC->CombineConfItem(mCRotateOrder, _("Rotate Order"));

    pC->mDragScale = 1;
    int indexS = pC->CombineConfItem(mCScale, _("Scale"));
    
    pC->FinishSetting();

    pC->SetColorForXYZ(indexT);
    pC->SetColorForXYZ(indexR);
    pC->SetColorForXYZ(indexS);

    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    mCOffset.SetBodyFromData(mpCSrc->GetPCParamGroup("Offset")->GetVInterpolatedValue((float)time, FALSE));
    mCRotate.SetBodyFromData(mpCSrc->GetPCParamGroup("Rotate")->GetVInterpolatedValue((float)time, FALSE));
    mCScale.SetBodyFromData(mpCSrc->GetPCParamGroup("Scale")->GetVInterpolatedValue((float)time, FALSE));
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}

#define CLS StreamTransformExec_cls

CLS::
StreamTransformExec_cls(AttributeClone_cls* pCSrc){
    mbValid = pCSrc->GetPCItem("Validity=")->GetBodyBOOL();
    //mVOffset = pCSrc->GetPCItem("Offset=")->GetBodyFLOATARRAY();
    //mVRotate = pCSrc->GetPCItem("Rotate=")->GetBodyFLOATARRAY();
    //mVScale  = pCSrc->GetPCItem("Scale=")->GetBodyFLOATARRAY();

    double NT = GCClassProvider.GetPCTimeCtrl()->GetCurrentNT();

    mVOffset = pCSrc->GetPCParamGroup("Offset")->GetVInterpolatedValue(NT);
    mVRotate = pCSrc->GetPCParamGroup("Rotate")->GetVInterpolatedValue(NT);
    mVScale  = pCSrc->GetPCParamGroup("Scale")->GetVInterpolatedValue(NT);
    mRotateOrder = pCSrc->GetPCItem("RotateOrder=")->GetBodyINT();
}

void CLS::
RotateX(float (*pPos)[3], int SNum, float radianX){
    float cosX = cos(radianX);
    float sinX = sin(radianX);

    for (int i = 0; i < SNum; i++){
	//kfloat xorg = pPos[i][X];
	float yorg = pPos[i][Y];
	float zorg = pPos[i][Z];
	
	pPos[i][Y] = + cosX*yorg - sinX*zorg;
	pPos[i][Z] = + sinX*yorg + cosX*zorg;
    }
}

void CLS::
RotateY(float (*pPos)[3], int SNum, float radianY){
    float cosY = cos(radianY);
    float sinY = sin(radianY);
    
    for (int i = 0; i < SNum; i++){
	float xorg = pPos[i][X];
	//float yorg = pPos[i][Y];
	float zorg = pPos[i][Z];

	pPos[i][Z] = + cosY*zorg - sinY*xorg;
	pPos[i][X] = + sinY*zorg + cosY*xorg;
    }
}

void CLS::
RotateZ(float (*pPos)[3], int SNum, float radianZ){
    float cosZ = cos(radianZ);
    float sinZ = sin(radianZ);
    
    for (int i = 0; i < SNum; i++){
	float xorg = pPos[i][X];
	float yorg = pPos[i][Y];
	//float zorg = pPos[i][Z];

	pPos[i][X] = + cosZ*xorg - sinZ*yorg;
	pPos[i][Y] = + sinZ*xorg + cosZ*yorg;
    }
}

void CLS::
Transform(float (*pPos)[3], int SNum){
    if (mbValid == FALSE) return;

    if (mVOffset[0] != 0 || mVOffset[1] != 0 || mVOffset[2] != 0)
	for (int i = 0; i < SNum; i++){
	    pPos[i][X] += mVOffset[X];
	    pPos[i][Y] += mVOffset[Y];
	    pPos[i][Z] += mVOffset[Z];
	}

    if (mVScale[0] != 1 || mVScale[1] != 1 || mVScale[2] != 1)
	for (int i = 0; i < SNum; i++){
	    pPos[i][X] *= mVScale[X];
	    pPos[i][Y] *= mVScale[Y];
	    pPos[i][Z] *= mVScale[Z];
	}
    
    if (mVRotate[0] != 0 || mVRotate[1] != 0 || mVRotate[2] != 0){
	float radianX = mVRotate[X] * 3.14159 / 180;
	float radianY = mVRotate[Y] * 3.14159 / 180;
	float radianZ = mVRotate[Z] * 3.14159 / 180;

	switch(mRotateOrder){
	case 0:
	    RotateZ(pPos, SNum, radianZ);
	    RotateY(pPos, SNum, radianY);
	    RotateX(pPos, SNum, radianX);
	    break;
	case 1:
	    RotateZ(pPos, SNum, radianZ);
	    RotateX(pPos, SNum, radianY);
	    RotateY(pPos, SNum, radianX);
	    break;
	case 2:
	    RotateX(pPos, SNum, radianZ);
	    RotateY(pPos, SNum, radianY);
	    RotateZ(pPos, SNum, radianX);
	    break;
	case 3:
	    RotateX(pPos, SNum, radianZ);
	    RotateZ(pPos, SNum, radianY);
	    RotateY(pPos, SNum, radianX);
	    break;
	case 4:
	    RotateY(pPos, SNum, radianZ);
	    RotateX(pPos, SNum, radianY);
	    RotateZ(pPos, SNum, radianX);
	    break;
	case 5:
	    RotateY(pPos, SNum, radianZ);
	    RotateZ(pPos, SNum, radianY);
	    RotateX(pPos, SNum, radianX);
	    break;
	}
    }
}
