// (setq-c-offset 4)
// "clsTexSpaceMod.h"

#include "clsTexSpaceMod.h"

#include "../MYGL/clsPosition.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#define CLS TexSpaceMod_cls

CLS::
TexSpaceMod_cls(wxString strName, wxString strObjName) : ModifierBase_cls("TexSpaceMod_cls", strName){

    wxString strUpdate = "VIEW3D::PROPERTY_MODIFIER";
    
    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCbAdjustScale, "AdjustScale=", "1", strUpdate);

    ParamGroup_cls *pCTrans = GetPCNewParamGroup("translate", "x", "y", "z");
    InitializeAndPushParam(pCTrans, 0, 0, 0);

    ParamGroup_cls *pCRot = GetPCNewParamGroup("rotate", "x", "y", "z");
    InitializeAndPushParam(pCRot, 0, 0, 0);

    ParamGroup_cls *pCScale = GetPCNewParamGroup("scale", "x", "y", "z");
    InitializeAndPushParam(pCScale, 1, 1, 1);
};

void CLS::
Execute(std::vector < double > VScaleObj){
    if (mCbValidity.GetBodyBOOL() == FALSE) return;
    
    double NT = GetPCTimeCtrl()->GetCurrentNT();

    std::vector < double > VTrans = GetPCParamGroup("translate")->GetVInterpolatedValue(NT);
    std::vector < double > VRotate = GetPCParamGroup("rotate")->GetVInterpolatedValue(NT);
    std::vector < double > VScale = GetPCParamGroup("scale")->GetVInterpolatedValue(NT);

    if (mCbAdjustScale.GetBodyBOOL() == TRUE)
	glScaled(1.0/VScaleObj[0], 1.0/VScaleObj[1], 1.0/VScaleObj[2]);

    glScaled(VScale[0], VScale[1], VScale[2]);    
    glRotated(VRotate[Z], 0, 0, 1);
    glRotated(VRotate[Y], 0, 1, 0);
    glRotated(VRotate[X], 1, 0, 0);
    glTranslated(VTrans[0], VTrans[1], VTrans[2]);


    if (mCbAdjustScale.GetBodyBOOL() == TRUE)
	glScaled(VScaleObj[0], VScaleObj[1], VScaleObj[2]);
}

void CLS::
SyncFromBuffer(TexSpaceModBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
    
    float time = GetPCTimeCtrl()->GetCurrentNT();

    std::vector < float > VTrance = pCBuffer->mCTranslate.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("translate", VTrance, pCObj);

    std::vector < float > VRot = pCBuffer->mCRotate.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("rotate", VRot, pCObj);

    std::vector < float > VScale = pCBuffer->mCScale.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("scale", VScale, pCObj);
}

#define CLS TexSpaceModBuffer_cls

CLS::
TexSpaceModBuffer_cls(TexSpaceMod_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCbAdjustScale, "AdjustScale=", "1");

    mCConfList.Add(&mCTranslate, "Translate=", "0 0 0");
    mCConfList.Add(&mCRotate, "Rotate=", "0 0 0");
    mCConfList.Add(&mCScale, "Scale=", "1 1 1");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
		new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineConfItem(mCbAdjustScale, _("Adjust Scale"));
    pC->mDragScale = 0.05;
    int indexT = pC->CombineConfItem(mCTranslate, "Translate");
    pC->mDragScale = 10;
    int indexR = pC->CombineConfItem(mCRotate, "Rotate");
    pC->mDragScale = 0.1;
    int indexS = pC->CombineConfItem(mCScale, "Scale");

    pC->FinishSetting();

    pC->SetColorForXYZ(indexT);
    pC->SetColorForXYZ(indexR);
    pC->SetColorForXYZ(indexS);

    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    mCTranslate.SetBodyFromData(mpCSrc->GetPCParamGroup("translate")->GetVInterpolatedValue((float)time, FALSE));
    mCRotate.SetBodyFromData(mpCSrc->GetPCParamGroup("rotate")->GetVInterpolatedValue((float)time, FALSE));
    mCScale.SetBodyFromData(mpCSrc->GetPCParamGroup("scale")->GetVInterpolatedValue((float)time, FALSE));
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}
