// (setq-c-offset 4)
// "clsTracking.h"

#include "clsTracking.h"

#include "../MYGL/clsPosition.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"

#include "../../wxWidgetEnum.h"

#define CLS Tracking_cls

CLS::
Tracking_cls(wxString strName, wxString strObjName) : ModifierBase_cls("Tracking_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_MODIFIER";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCstrTarget, "Target=", "None", strUpdate);
    mCConfList.Add(&mCBaseDirection, "BaseDirection=", "4", strUpdate);
    mCConfList.Add(&mCBaseTopDirection, "BaseTopDirection=", "1", strUpdate);
    
    AddOneDimParam("Influence", "I", 1);
    AddOneDimParam("Rotate", "R", 0);
};

void CLS::
Execute(double matrix[16], double matrixLocal[16]){
    if (mCbValidity.GetBodyBOOL() == FALSE) return;

    SceneObj_cls *pCTgt = GetPCSceneObjCtrl()->GetPCFromName(mCstrTarget.GetBodySTR());
    if (pCTgt == NULL) return;

    if (mLoopDetection == 1) return;
    mLoopDetection = 1;

    PositionAndDirection_cls CPosOrg(matrix);
    PositionAndDirection_cls CPosTgt(matrix);

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    pCTgt->ClearModifierMatrix();

    //std::vector < double > VPosGlobal = pCTgt->GetPositionGlobal(NT);
    double matrixTgt[16];
    pCTgt->GetMatrix16(matrixTgt, NT);
    std::vector < double > VPosGlobal, VTopGlobal;
    VPosGlobal.push_back(matrixTgt[12]);
    VPosGlobal.push_back(matrixTgt[13]);
    VPosGlobal.push_back(matrixTgt[14]);
    VPosGlobal.push_back(1);

    VTopGlobal.push_back(matrixTgt[ 8]);
    VTopGlobal.push_back(matrixTgt[ 9]);
    VTopGlobal.push_back(matrixTgt[10]);
    VTopGlobal.push_back(1);
    
    double influence = GetPCParamGroup("Influence")->GetVInterpolatedValue(NT)[0];
    double rotate = GetPCParamGroup("Rotate")->GetVInterpolatedValue(NT)[0];
    
    double matrixLocalInv[16];
    GetInvMatrix4x4(matrixLocal, matrixLocalInv);

    // VPosGlobal  [Jł̈ʒuɕϊ
    MultiplyV(matrixLocalInv, VPosGlobal.data());
    MultiplyV(matrixLocalInv, VTopGlobal.data());
    
    // CPosTgt  VPos Global 
    Position_cls CTgt(VPosGlobal);
    
    
    CPosTgt.SetPivotPos(CTgt);
    CPosTgt.LookAtPivot();

    int topType = mCBaseTopDirection.GetBodyINT();
    if (topType == 1){
	LOOP_XYZ CPosTgt.mTop[xyz] = VTopGlobal[xyz];
	CPosTgt.NormalizeAndOrthogonalize();
    }
    if (topType == 2){
	CPosTgt.mTop[0] = 0; CPosTgt.mTop[1] = 0; CPosTgt.mTop[2] = 1;
	CPosTgt.NormalizeAndOrthogonalize();
    }
    
    // Influence ɉ mix 
    PositionAndDirection_cls CPos;
    CPos.SetMixed(CPosOrg, CPosTgt, influence);

    CPos.RotateRoll(rotate); 

    int baseDirection = mCBaseDirection.GetBodyINT();
    switch(baseDirection){
    case 0: break;
    case 1: CPos.RotateYaw(180); break;
    case 2: CPos.RotateYaw(90); break;
    case 3: CPos.RotateYaw(-90); break;
    case 4: CPos.RotatePitch(90); CPos.RotateYaw(90); break;
    case 5: CPos.RotatePitch(-90); CPos.RotateYaw(-90); break;
    }
    
    // matrix ɕԂ
    mLoopDetection = 0;
    CPos.GetMatrix16(matrix);
}

void CLS::
SyncFromBuffer(TrackingBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
    
    std::vector < float > VInfluence;
    VInfluence.push_back(pCBuffer->mCInfluence.GetBodyDOUBLE());
    SyncronizeParamForModifier("Influence", VInfluence, pCObj);

    std::vector < float > VRotate;
    VRotate.push_back(pCBuffer->mCRotate.GetBodyDOUBLE());
    SyncronizeParamForModifier("Rotate", VRotate, pCObj);
}

void CLS::
SelectObject(wxString strObjName){
    mCstrTarget.PushCommandPair(strObjName);
}

#define CLS TrackingBuffer_cls

CLS::
TrackingBuffer_cls(Tracking_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCstrTarget, "Target=", "None");
    mCConfList.Add(&mCBaseDirection, "BaseDirection=", "4");
    mCConfList.Add(&mCBaseTopDirection, "BaseTopDirection=", "1");
    mCConfList.Add(&mCInfluence, "Influence=", "1");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
		new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineSizer(2, _("Target"));
    pC->CombineBitmapPanel(btnDropdown, wxES, 14);
    pC->mVCItem.back().mBitmapTemplate = _BITMAP_BTN_DROPDOWN;
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = mCstrName.GetBodySTR();

    pC->CombineConfItem(mCstrTarget);
    
    pC->CombineSizer(2, _("Base Direction"));
    wxString AstrDir[6] = {"+X", "-X", "+Y", "-Y", "+Z", "-Z"};
    pC->SetChoiceList(6, AstrDir);
    pC->CombineConfItem(mCBaseDirection);

    wxString AstrTop[6] = {"Object Z", "Target Z", "Global Z"};
    pC->SetChoiceList(3, AstrTop);
    pC->CombineConfItem(mCBaseTopDirection);
    
    pC->CombineSizer(2, _("Influence/Rotate"));
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.3;
    pC->CombineConfItem(mCInfluence);
    pC->mDragScale = 10;
    pC->CombineConfItem(mCRotate);
    pC->FinishSetting();

    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();
    mCInfluence.SetBodyFromData(mpCSrc->GetPCParamGroup("Influence")->GetVInterpolatedValue((float)time, FALSE));
    mCRotate.SetBodyFromData(mpCSrc->GetPCParamGroup("Rotate")->GetVInterpolatedValue((float)time, FALSE));
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}
