// (setq-c-offset 4)
// "clsTrimming.h"

#include "clsTrimming.h"

#include "../MYGL/clsPosition.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
//#include "../WIDGETCTRL/WX/clsColorProvider.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS Trimming_cls

CLS::
Trimming_cls(wxString strName, wxString strObjName) : ModifierBase_cls("Trimming_cls", strName){
    
    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCbGlobal, "Global=", "0");
    mCConfList.Add(&mCbSmooth, "Smooth=", "0");
    //mCConfList.Add(&mCTrimType, "TrimType=", "0");
    mCConfList.Add(&mCSmoothWidth, "SmoothWidth=", "0.1");
    mCConfList.Add(&mCThinningFactor, "ThinningFactor=", "0");
    
    ParamGroup_cls *pCX = GetPCNewParamGroup("X Min/Max", "Min", "Max");
    InitializeAndPushParam(pCX, 0, 1);
    
    ParamGroup_cls *pCY = GetPCNewParamGroup("Y Min/Max", "Min", "Max");
    InitializeAndPushParam(pCY, 0, 1);

    ParamGroup_cls *pCZ = GetPCNewParamGroup("Z Min/Max", "Min", "Max");
    InitializeAndPushParam(pCZ, 0, 1);

    ParamGroup_cls *pCThinning = GetPCNewParamGroup("Thinning", "W", "F");
    InitializeAndPushParam(pCThinning, 0.1, 0);
};

void CLS::
Execute(double clipBoxMin[3], double clipBoxMax[3]){
    if (mCbValidity.GetBodyBOOL() == FALSE) return;

    double NT = GetPCTimeCtrl()->GetCurrentNT();

    std::vector < double > CVX = GetPCParamGroup("X Min/Max")->GetVInterpolatedValue(NT);
    std::vector < double > CVY = GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue(NT);
    std::vector < double > CVZ = GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue(NT);

    clipBoxMin[X] = CVX[0]; clipBoxMax[X] = CVX[1];
    clipBoxMin[Y] = CVY[0]; clipBoxMax[Y] = CVY[1];
    clipBoxMin[Z] = CVZ[0]; clipBoxMax[Z] = CVZ[1];

    /*
    double widthX = clipBoxMax[X] - clipBoxMin[X];
    clipBoxMin[X] += widthX * CVX[0];
    clipBoxMax[X] -= widthX * (1.0 - CVX[1]);

    double widthY = clipBoxMax[Y] - clipBoxMin[Y];
    clipBoxMin[Y] += widthY * CVY[0];
    clipBoxMax[Y] -= widthY * (1.0 - CVY[1]);

    double widthZ = clipBoxMax[Z] - clipBoxMin[Z];
    clipBoxMin[Z] += widthZ * CVZ[0];
    clipBoxMax[Z] -= widthZ * (1.0 - CVZ[1]);    
    */
}


void CLS::
SyncFromBuffer(TrimmingBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);
    
    std::vector < float > VXMinMax = pCBuffer->mCXRange.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("X Min/Max", VXMinMax, pCObj);

    std::vector < float > VYMinMax = pCBuffer->mCYRange.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("Y Min/Max", VYMinMax, pCObj);

    std::vector < float > VZMinMax = pCBuffer->mCZRange.GetBodyFLOATARRAY();
    SyncronizeParamForModifier("Z Min/Max", VZMinMax, pCObj);

    std::vector < float > VThinning(2);
    VThinning[0] = pCBuffer->mCSmoothWidth.GetBodyDOUBLE();
    VThinning[1] = pCBuffer->mCThinningFactor.GetBodyDOUBLE();
    SyncronizeParamForModifier("Thinning", VThinning, pCObj);
}

#define CLS TrimmingBuffer_cls

CLS::
TrimmingBuffer_cls(Trimming_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCXRange, "XRange=", "0 1");
    mCConfList.Add(&mCYRange, "YRange=", "0 1");
    mCConfList.Add(&mCZRange, "ZRange=", "0 1");
    mCConfList.Add(&mCbGlobal, "Global=", "0");
    mCConfList.Add(&mCbSmooth, "Smooth=", "0");
    //mCConfList.Add(&mCTrimType, "TrimType=", "0");
    mCConfList.Add(&mCSmoothWidth, "SmoothWidth=", "0.1");
    mCConfList.Add(&mCThinningFactor, "ThinningFactor=", "0.5");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow, wxString strParentClassName){
    MultiInputPanel_cls *pC =
		new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    bool bStream = FALSE;
    if (strParentClassName.IsSameAs("Streamline_cls")) bStream = TRUE;
    
    pC->mDragScale = 0.1;
    if (bStream == FALSE) pC->SetLimit(0, 1);
    int indexX = pC->CombineConfItem(mCXRange, "X Min/Max");
    
    pC->mDragScale = 0.1;
    if (bStream == FALSE) pC->SetLimit(0, 1);
    int indexY = pC->CombineConfItem(mCYRange, "Y Min/Max");
    
    pC->mDragScale = 0.1;
    if (bStream == FALSE) pC->SetLimit(0, 1);
    int indexZ = pC->CombineConfItem(mCZRange, "Z Min/Max");

    if (bStream == TRUE){
	pC->CombineSizer(2, _("Option"));
	pC->CombineConfItem(mCbGlobal, _("Global"));
	pC->CombineConfItem(mCbSmooth, _("Smooth"));
	/*
	wxString AstrList[3] = {"Trim", "Smooth", "Enhance"};
	pC->SetChoiceList(3, AstrList);
	pC->CombineConfItem(mCTrimType);
	*/
	pC->CombineSizer(2, _("Width/Factor"));
	pC->CombineConfItem(mCSmoothWidth, _("Smooth"));
	pC->CombineConfItem(mCThinningFactor, _("Smooth"));
    }
    
    pC->FinishSetting();

    pC->SetColorForXYZMinMax(indexX, indexY, indexZ);
    
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    mCXRange.SetBodyFromData(mpCSrc->GetPCParamGroup("X Min/Max")->GetVInterpolatedValue((float)time, FALSE));
    mCYRange.SetBodyFromData(mpCSrc->GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue((float)time, FALSE));
    mCZRange.SetBodyFromData(mpCSrc->GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue((float)time, FALSE));

    ParamGroup_cls *pCThinning = mpCSrc->GetPCParamGroup("Thinning");
    mCSmoothWidth.SetBodyFromData(pCThinning->GetVInterpolatedValue((float)time, FALSE)[0]);
    mCThinningFactor.SetBodyFromData(pCThinning->GetVInterpolatedValue((float)time, FALSE)[1]);
}

void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}
