// (setq-c-offset 4)
// "clsVolumeCoord.h"

#include "clsVolumeCoord.h"

#include "../GLSL/clsShaderItem.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/EX/wxButtonEX.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../../wxWidgetEnum.h"

#define CLS VolumeCoord_cls

CLS::
VolumeCoord_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("VolumeCoord_cls", strName){

    wxString strUpdate = "VIEW3D::PROPERTY_DATAPROC";
    
    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCCoordType, "CoordType=", "0", strUpdate);
    mCConfList.Add(&mCstrCustomCoord, "CustomCoord=", "vec3 GetDeformCoord(vec3 vecOrg){\n\treturn vecOrg;\n}\n\n#define GUSEMATRIX 0\nmat3 GetDeformNormalMatrix(vec3 vecOrg, vec3 deformedCoord){\n\treturn mat3(1, 0, 0, 0, 1, 0, 0, 0, 1);\n}\n", strUpdate);
    
    ParamGroup_cls *pCRadius = GetPCNewParamGroup("R Min/Max", "Min", "Max");
    InitializeAndPushParam(pCRadius, 0, 1);
    ParamGroup_cls *pCCustomParam = GetPCNewParamGroup("Custom", "p0", "p1", "p2");
    InitializeAndPushParam(pCCustomParam, 0, 0, 0);
}

void CLS::
SyncFromBuffer(VolumeCoordBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VRMinMax;
    VRMinMax.push_back(pCBuffer->mCRMin.GetBodyDOUBLE());
    VRMinMax.push_back(pCBuffer->mCRMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("R Min/Max", VRMinMax, pCObj);

    std::vector < float > VCustom = 
	pCBuffer->mCCustomParam.GetBodyFLOATARRAY();
    
    SyncronizeParamForModifier("Custom", VCustom, pCObj);
}

bool CLS::
IsStatusChanged(ShaderStatusVolumeCoord_cls &CStatus){
    
    float NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VR = GetPCParamGroup("R Min/Max")->GetVInterpolatedValue(NT);
    std::vector < float > VCustom = GetPCParamGroup("Custom")->GetVInterpolatedValue(NT);
    if (CStatus.mbUseCompiled == FALSE ||
	CStatus.mTypeCompiled != mCCoordType.GetBodyINT() ||
	CStatus.mRMinCompiled != VR[0] ||
	CStatus.mRMaxCompiled != VR[1] ||
	CStatus.mCustom0Compiled != VCustom[0] ||
	CStatus.mCustom1Compiled != VCustom[1] ||
	CStatus.mCustom2Compiled != VCustom[2] || 
	CStatus.mstrCustomCompiled.IsSameAs(mCstrCustomCoord.GetBodySTR()) == FALSE){
	return TRUE;
    }
return FALSE;
}

void CLS::
SetStatus(ShaderStatusVolumeCoord_cls &CStatus){
    float NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VR = GetPCParamGroup("R Min/Max")->GetVInterpolatedValue(NT);
    std::vector < float > VCustom = GetPCParamGroup("Custom")->GetVInterpolatedValue(NT);
    CStatus.mbUse = TRUE;
    CStatus.mType = mCCoordType.GetBodyINT();
    CStatus.mRMin = VR[0];
    CStatus.mRMax = VR[1];
    CStatus.mCustom0 = VCustom[0];
    CStatus.mCustom1 = VCustom[1];
    CStatus.mCustom2 = VCustom[2];
    CStatus.mstrCustom = mCstrCustomCoord.GetBodySTR();
}

wxString CLS::
GetStrSpherical(wxString &strFrag, wxString strTagFin, int ini, int fin){
    
    if (smstrFunc.IsEmpty() == TRUE)	
	GetStrFullFromFile("GL/FUNCTION/VolumeCoord.txt", smstrFunc, TRUE);

    wxString strFunc = smstrFunc;    
    float NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VR = GetPCParamGroup("R Min/Max")->GetVInterpolatedValue(NT);
    wxString strRMin; strRMin << VR[0];
    wxString strRMax; strRMax << VR[1];
    strFunc.Replace("#RMIN#", strRMin);
    strFunc.Replace("#RMAX#", strRMax);

    fin += strTagFin.Len();
    strFrag.Remove(ini, fin-ini);
    strFrag.insert(ini, strFunc);	
    
    return strFrag;
}

wxString CLS::
GetStrCustom(wxString &strFrag, wxString strTagFin, int ini, int fin){

    wxString strFunc = mCstrCustomCoord.GetBodySTR();

    float NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VCustom = GetPCParamGroup("Custom")->GetVInterpolatedValue(NT);

    wxString str0; str0 << VCustom[0];
    wxString str1; str1 << VCustom[1];
    wxString str2; str2 << VCustom[2];

    strFunc.Replace("#p0#", str0);
    strFunc.Replace("#p1#", str1);
    strFunc.Replace("#p2#", str2);

    fin += strTagFin.Len();
    strFrag.Remove(ini, fin-ini);
    strFrag.insert(ini, strFunc);
    
    return strFrag;
}

wxString CLS::
GetStrModifiedShader(wxString strFragSrc){
    wxString strFrag = strFragSrc;
    wxString strTagIni = "//##FUNC_VOLUME_COORD_BEGIN##";
    wxString strTagFin = "//##FUNC_VOLUME_COORD_END##";
    int ini = strFrag.Find(strTagIni);
    int fin = strFrag.Find(strTagFin);
    if (ini == wxNOT_FOUND) return strFragSrc;
    
    if (mCCoordType.GetBodyINT() == 0)
	GetStrSpherical(strFrag, strTagFin, ini, fin);
    if (mCCoordType.GetBodyINT() == 1)
	GetStrCustom(strFrag, strTagFin, ini, fin);

    return strFrag;
}

wxString VolumeCoord_cls::smstrFunc;

#define CLS VolumeCoordBuffer_cls

CLS::
VolumeCoordBuffer_cls(VolumeCoord_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCCoordType, "CoordType=", "0");
    
    mCCustomParam.SetItem("CustomParam=", "0 0 0");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    pC->CombineSizer(2, _("CoordType"));
    wxString AstrCoordType[2] =
	{_("Spherical"), _("Custom")};
    pC->SetChoiceList(2, AstrCoordType);
    pC->CombineConfItem(mCCoordType/*, _("CoordType")*/);
    mIndexEdit = pC->CombineButton(btnVolumeCoordEdit, _("Edit"));
    
    mIndexRMinMax = pC->CombineSizer(2, "R Min/Max");
    pC->mDragScale = 1;
    pC->CombineConfItem(mCRMin);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCRMax);

    mIndexCustomParam = pC->CombineConfItem(mCCustomParam, _("Parameters"));

    pC->FinishSetting();
    ((wxButtonEX*)pC->mVCItem[mIndexEdit].mpCWindowLink)->mstrCommandStr = 
	mpCSrc->mCConfList.mstrTag + "::CustomCoord";
    //pC->SetHook(this, &HookItem_cls::HookFunc01, wxEVT_COMMAND_TEXT_ENTER);

    mpCMNumPanelLink = pC;
    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);

    double time = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VR = mpCSrc->GetPCParamGroup("R Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    
    mCRMin.SetBodyFromData(VR[0]);
    mCRMax.SetBodyFromData(VR[1]);
}
    
void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}

bool CLS::
HideAndShow(){
    bool bChange = FALSE;
    if (mCCoordType.GetBodyINT() == 0){
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRMinMax].Show(TRUE);
	bChange |= mpCMNumPanelLink->mVCItem[mIndexCustomParam].Show(FALSE);
	bChange |= mpCMNumPanelLink->mVCItem[mIndexEdit].Show(FALSE);
    }
    else{
	bChange |= mpCMNumPanelLink->mVCItem[mIndexRMinMax].Show(FALSE);
	bChange |= mpCMNumPanelLink->mVCItem[mIndexCustomParam].Show(TRUE);
	bChange |= mpCMNumPanelLink->mVCItem[mIndexEdit].Show(TRUE);
    }

    if (bChange == TRUE){
	mpCMNumPanelLink->Layout();
    }
    
    return bChange;
}
