// (setq-c-offset 4)
// "clsVolumeMask.h"

#include "clsVolumeMask.h"

#include "../GLSL/clsShaderItem.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS VolumeMask_cls

CLS::
VolumeMask_cls(wxString strName, wxString strObjName) : ModifierBase_cls ("VolumeMask_cls", strName){
    
    wxString strUpdate = "VIEW3D::PROPERTY_DATAPROC";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    mCConfList.Add(&mCMaskType, "MaskType=", "0", strUpdate);
    
    ParamGroup_cls *pCX = GetPCNewParamGroup("X Min/Max", "Min", "Max");
    InitializeAndPushParam(pCX, 0, 1);
    ParamGroup_cls *pCY = GetPCNewParamGroup("Y Min/Max", "Min", "Max");
    InitializeAndPushParam(pCY, 0, 1);
    ParamGroup_cls *pCZ = GetPCNewParamGroup("Z Min/Max", "Min", "Max");
    InitializeAndPushParam(pCZ, 0, 1);
}

void CLS::
SyncFromBuffer(VolumeMaskBuffer_cls *pCBuffer, SceneObj_cls *pCObj){

    mCConfList.SyncFromCommand(pCBuffer->mCConfList);

    std::vector < float > VXMinMax, VYMinMax, VZMinMax;

    VXMinMax.push_back(pCBuffer->mCXMin.GetBodyDOUBLE());
    VXMinMax.push_back(pCBuffer->mCXMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("X Min/Max", VXMinMax, pCObj);
    VYMinMax.push_back(pCBuffer->mCYMin.GetBodyDOUBLE());
    VYMinMax.push_back(pCBuffer->mCYMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("Y Min/Max", VYMinMax, pCObj);
    VZMinMax.push_back(pCBuffer->mCZMin.GetBodyDOUBLE());
    VZMinMax.push_back(pCBuffer->mCZMax.GetBodyDOUBLE());
    SyncronizeParamForModifier("Z Min/Max", VZMinMax, pCObj);
}

bool CLS::
IsStatusChanged(ShaderStatusVolumeMask_cls &CStatus){
    
    float NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VX = GetPCParamGroup("X Min/Max")->GetVInterpolatedValue(NT);
    std::vector < float > VY = GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue(NT);
    std::vector < float > VZ = GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue(NT);
    
    if (CStatus.mbUseCompiled == FALSE ||
	CStatus.mXMinCompiled != VX[0] ||
	CStatus.mXMaxCompiled != VX[1] ||
	CStatus.mYMinCompiled != VY[0] ||
	CStatus.mYMaxCompiled != VY[1] ||
	CStatus.mZMinCompiled != VZ[0] ||
	CStatus.mZMaxCompiled != VZ[1] ||
	CStatus.mMaskTypeCompiled != mCMaskType.GetBodyINT()){
	return TRUE;
    }
    return FALSE;
}

void CLS::
SetStatus(ShaderStatusVolumeMask_cls &CStatus){
    float NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VX = GetPCParamGroup("X Min/Max")->GetVInterpolatedValue(NT);
    std::vector < float > VY = GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue(NT);
    std::vector < float > VZ = GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue(NT);
    
    CStatus.mbUse = TRUE;
    CStatus.mMaskType = mCMaskType.GetBodyINT();
    CStatus.mXMin = VX[0];
    CStatus.mXMax = VX[1];
    CStatus.mYMin = VY[0];
    CStatus.mYMax = VY[1];
    CStatus.mZMin = VZ[0];
    CStatus.mZMax = VZ[1];
}

wxString CLS::
GetStrModifiedShader(wxString strFragSrc){

    wxString strFrag = strFragSrc;
    wxString strTagIni = "//##FUNC_VOLUME_MASK_BEGIN##";
    wxString strTagFin = "//##FUNC_VOLUME_MASK_END##";
    int ini = strFrag.Find(strTagIni);
    int fin = strFrag.Find(strTagFin);
    if (ini == wxNOT_FOUND) return strFragSrc;

    if (smstrFunc.IsEmpty() == TRUE)
	GetStrFullFromFile("GL/FUNCTION/VolumeMask.txt", smstrFunc, TRUE);
    
    wxString strFunc = smstrFunc;    
    float NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VXMinMax = GetPCParamGroup("X Min/Max")->GetVInterpolatedValue(NT);
    std::vector < float > VYMinMax = GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue(NT);
    std::vector < float > VZMinMax = GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue(NT);
    wxString strXMin; strXMin << VXMinMax[0];
    wxString strXMax; strXMax << VXMinMax[1];
    wxString strYMin; strYMin << VYMinMax[0];
    wxString strYMax; strYMax << VYMinMax[1];
    wxString strZMin; strZMin << VZMinMax[0];
    wxString strZMax; strZMax << VZMinMax[1];
    strFunc.Replace("#XMIN#", strXMin);
    strFunc.Replace("#XMAX#", strXMax);
    strFunc.Replace("#YMIN#", strYMin);
    strFunc.Replace("#YMAX#", strYMax);
    strFunc.Replace("#ZMIN#", strZMin);
    strFunc.Replace("#ZMAX#", strZMax);

    wxString strCondition;
    if (mCMaskType.GetBodyINT() == 1){
	strCondition << "    if (texCoord.x > xmin && texCoord.x < xmax &&" << ENDL;
	strCondition << "        texCoord.y > ymin && texCoord.y < ymax &&" << ENDL;
	strCondition << "        texCoord.z > zmin && texCoord.z < zmax) return false;" << ENDL;
    }
    else{
	strCondition << "    if (texCoord.x < xmin || texCoord.x > xmax) return false;" << ENDL;
	strCondition << "    if (texCoord.y < ymin || texCoord.y > ymax) return false;" << ENDL;
	strCondition << "    if (texCoord.z < zmin || texCoord.z > zmax) return false;" << ENDL;
    }
	
    strFunc.Replace("#CONDITION#", strCondition);

    fin += strTagFin.Len();
    strFrag.Remove(ini, fin-ini);
    strFrag.insert(ini, strFunc);
    return strFrag;
}

wxString VolumeMask_cls::smstrFunc;

#define CLS VolumeMaskBuffer_cls

CLS::
VolumeMaskBuffer_cls(VolumeMask_cls *pCSrc) : ModifierBase_cls(*pCSrc){
    mpCSrc = pCSrc;
    mCConfList.SetTag(pCSrc->mCConfList.mstrTag);
    mCConfList.Add(&mCMaskType, "MaskType=", "0");
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName(), TRUE);

    wxString AstrFilterType[3] =
	{"Intersect", "Diffierence", "Add"};
    pC->SetChoiceList(3, AstrFilterType);
    pC->CombineConfItem(mCMaskType, _("MaskType"));
    
    int indexX = pC->CombineSizer(2, "X Min/Max");
    pC->mDragScale = 1;
    pC->CombineConfItem(mCXMin);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCXMax);

    int indexY = pC->CombineSizer(2, "Y Min/Max");
    pC->mDragScale = 1;
    pC->CombineConfItem(mCYMin);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCYMax);

    int indexZ = pC->CombineSizer(2, "Z Min/Max");
    pC->mDragScale = 1;
    pC->CombineConfItem(mCZMin);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCZMax);

    pC->FinishSetting();

    pC->SetColorForXYZMinMax(indexX, indexY, indexZ);

    return pC;
}

void CLS::
SyncFromSrc(){
    mCConfList.SyncFrom(mpCSrc->mCConfList);

    double time = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VX = mpCSrc->GetPCParamGroup("X Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    std::vector < float > VY = mpCSrc->GetPCParamGroup("Y Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    std::vector < float > VZ = mpCSrc->GetPCParamGroup("Z Min/Max")->GetVInterpolatedValue((float)time, FALSE);
    
    mCXMin.SetBodyFromData(VX[0]);
    mCXMax.SetBodyFromData(VX[1]);

    mCYMin.SetBodyFromData(VY[0]);
    mCYMax.SetBodyFromData(VY[1]);

    mCZMin.SetBodyFromData(VZ[0]);
    mCZMax.SetBodyFromData(VZ[1]);
}
    
void CLS::
ForceSrcToSyncBuffer(SceneObj_cls *pCObj){
    mpCSrc->SyncFromBuffer(this, pCObj);
}

