// "clsFBO.h"
// (setq c-basic-offset 4)

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
extern PFNGLGENFRAMEBUFFERSPROC glGenFramebuffers;
extern PFNGLDELETEFRAMEBUFFERSPROC glDeleteFramebuffers;
extern PFNGLBINDFRAMEBUFFERPROC glBindFramebuffer;
extern PFNGLFRAMEBUFFERTEXTURE2DPROC glFramebufferTexture2D;
extern PFNGLGENRENDERBUFFERSPROC glGenRenderbuffers;
extern PFNGLDELETERENDERBUFFERSPROC glDeleteRenderbuffers;
extern PFNGLBINDRENDERBUFFERPROC glBindRenderbuffer;
extern PFNGLRENDERBUFFERSTORAGEPROC glRenderbufferStorage;
extern PFNGLFRAMEBUFFERRENDERBUFFERPROC glFramebufferRenderbuffer;
extern PFNGLCHECKFRAMEBUFFERSTATUSPROC glCheckFramebufferStatus;
extern PFNGLBLITFRAMEBUFFERPROC glBlitFramebuffer;
#elif defined __WXMAC__
#define GL_GLEXT_PROTOTYPES 
#include <OpenGL/glu.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include "clsFBO.h"
//#include "clsPositionAndDirection.h"

#include "../GLCANVAS/clsGLCanvas.h"
#include "../MYGL/clsGLPerspective.h"
#include "../PREFERENCE/clsPreference.h"

#include "../COMMON/CommonDefinitions.h"

FBO_cls::
FBO_cls(int width, int height, GLCanvas_cls *pCGLCanvasParent){
    if (msbInitialized == FALSE) Initialize();
    mWidthFBO = width; mHeightFBO = height;
    mbGenerated = FALSE;
    mColorDepth = 16;
    mColorDepthGenerated = 0;
    mFBOId = 0;
    mpCDepthTexture = NULL;
    mbDepthTextureUsed = FALSE;
    mDepthIndexGenerated = -1;
    mpCGLCanvasParent = pCGLCanvasParent;
    mpCStackFBO = NULL;
}

FBO_cls::
~FBO_cls(){
    // 쐬ς FBO 폜 eNX`Baseō폜
    if (mbGenerated == TRUE){
	glDeleteRenderbuffers(1, &mRBIdDepthStencil);
	glDeleteFramebuffers(1, &mFBOId);
	if (mpCDepthTexture != NULL)
	    delete mpCDepthTexture;
    }
}

void FBO_cls::
Initialize(){

#ifdef WIN32
    if (glCheckFramebufferStatus == NULL) msbSupported = FALSE;
#endif
    
    if (msbSupported == TRUE){
	GLenum status;                                            
	status = glCheckFramebufferStatus(GL_FRAMEBUFFER); 
	if (status != GL_FRAMEBUFFER_COMPLETE)
	    msbSupported = FALSE;
    }

    msbInitialized = TRUE;
}

void FBO_cls::
SetGeometryFBO(int width, int height){
    mWidthFBO = width;
    mHeightFBO = height;
}

void FBO_cls::
SetColorDepth(int depth){
    mColorDepth = depth;
}

void FBO_cls::
MakeFBO(bool bUseRGBA, bool bUseDepthTexture){
    //if (mFBOId != 0) return;
    if (msbSupported == FALSE) return;

    if (CheckGLError("MakeFBO") == FALSE){
	printf("dummy");
    }

    int depthIndex = GetPCPreference()->mCGLDepthIndex.GetBodyINT();
    if (mbGenerated == TRUE){
	if (mWidthTex == mWidthFBO &&
	    mHeightTex == mHeightFBO &&
	    mColorDepth == mColorDepthGenerated &&
	    depthIndex == mDepthIndexGenerated &&
	    mbAlphaGenerated == bUseRGBA &&
	    mbDepthTextureUsed == bUseDepthTexture) return;
	else{
	    glDeleteTextures(1, &mTextureID);
	    glDeleteRenderbuffers(1, &mRBIdDepthStencil);
	    glDeleteFramebuffers(1, &mFBOId);
	    if (mpCDepthTexture != NULL){
		delete mpCDepthTexture;
		mpCDepthTexture = NULL;
	    }
	}
    }

    GenerateAndBindTexture();
        
    switch (mColorDepth){
    case 16:
	if (bUseRGBA == TRUE)
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA16F_ARB, 
			  mWidthFBO, mHeightFBO, 0,
			  GL_RGBA, GL_HALF_FLOAT_ARB, 0);
	else
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB16F_ARB, 
			  mWidthFBO, mHeightFBO, 0,
			  GL_RGB, GL_HALF_FLOAT_ARB, 0);
	break;
    case 8:
	if (bUseRGBA == TRUE)
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 
			  mWidthFBO, mHeightFBO, 0,
			  GL_RGBA, GL_UNSIGNED_BYTE, 0);
	else
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, 
			  mWidthFBO, mHeightFBO, 0,
			  GL_RGB, GL_UNSIGNED_BYTE, 0);
	break;
    }
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

    if (bUseDepthTexture == TRUE && mpCDepthTexture == NULL){
	mpCDepthTexture = new TextureBase_cls;
	mpCDepthTexture->mWidthTex = mWidthFBO;
	mpCDepthTexture->mHeightTex = mHeightFBO;
	mpCDepthTexture->GenerateAndBindTexture();

	
	glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT24, mWidthFBO, mHeightFBO,
		      0, GL_DEPTH_COMPONENT, GL_FLOAT , 0);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    }

    CheckGLError("INIT");

    // FBO̍쐬
    glGenFramebuffers(1, &mFBOId);

    // Depth  Stencil ێ郌_[obt@𐶐č쐬
    glGenRenderbuffers(1, &mRBIdDepthStencil); 

    glBindRenderbuffer(GL_RENDERBUFFER, mRBIdDepthStencil);

    if (bUseDepthTexture == FALSE){
	GLenum GLDepth = GL_DEPTH24_STENCIL8;
	if (depthIndex == 1) GLDepth = GL_DEPTH32F_STENCIL8;
	glRenderbufferStorage(GL_RENDERBUFFER, 
			      /*GL_DEPTH32F_STENCIL8_NV*//*GL_DEPTH_COMPONENT32F*/
			      GLDepth, mWidthFBO, mHeightFBO);
    }
    else{
	glRenderbufferStorage(GL_RENDERBUFFER,
			      GL_STENCIL_INDEX8, mWidthFBO, mHeightFBO);
    }
    CheckGLError("STORAGE DEPTH STENCIL");

    glBindRenderbuffer(GL_RENDERBUFFER, 0);

    
    // FBOɃeNX`IuWFNgƃ_[obt@ 
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, mFBOId);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
			   GL_TEXTURE_2D, mTextureID, 0);
    
    if (bUseDepthTexture == TRUE){
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
			       GL_TEXTURE_2D, mpCDepthTexture->mTextureID, 0);
	/*
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT,
				  GL_RENDERBUFFER, mRBIdDepthStencil);
	*/
	/*
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT,
			       GL_TEXTURE_2D, mpCDepthTexture->mTextureID, 0);
	*/
    }
    
    else{
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT,
				  GL_RENDERBUFFER, mRBIdDepthStencil);
	CheckGLError("ATTATCH DEPTH STENCIL");
    }

    mbGenerated = TRUE;
    mColorDepthGenerated = mColorDepth;
    mDepthIndexGenerated = depthIndex;
    mbAlphaGenerated = bUseRGBA;
    mbDepthTextureUsed = bUseDepthTexture;
    mWidthTex = mWidthFBO;
    mHeightTex = mHeightFBO;

    glClearColor(0,0,0,0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void FBO_cls::
ClearFBO(){
    CheckGLError("CLEAR BEGIN");
    
    Begin(TRUE);
    End();

    CheckGLError("CLEAR END");
}

void FBO_cls::
Begin(bool bClear, bool bCalledInFBOStack){
    mbClear = bClear;
    
    if (mpCGLCanvasParent != NULL){
	if (mpCGLCanvasParent->mpCFBOCurrentLink != NULL)
	    mpCStackFBO = mpCGLCanvasParent->mpCFBOCurrentLink;
	
	mpCGLCanvasParent->mpCFBOCurrentLink = this;
    }
    
    if (bCalledInFBOStack == FALSE) 
	glPushAttrib(GL_COLOR_BUFFER_BIT | GL_CURRENT_BIT | GL_ENABLE_BIT | GL_FOG_BIT | GL_LIGHTING_BIT | GL_LINE_BIT | GL_STENCIL_BUFFER_BIT | GL_TEXTURE_BIT | GL_TRANSFORM_BIT | GL_VIEWPORT_BIT);

    if (msbSupported == FALSE) return;

    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, mFBOId);

    CheckGLError("BIND FBO");
        
    //glBindRenderbuffer(GL_RENDERBUFFER, 0);

    // ̂܂ܕ`FBOɈڂ
    glViewport(0, 0, mWidthFBO, mHeightFBO);
    
    glClearColor(0,0,0,0);
    
    if (bClear == TRUE)
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    else
	glClear(GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    CheckGLError("BEGIN END");
}

void FBO_cls::
End(){

    if (msbSupported == TRUE)
	glBindFramebuffer( GL_FRAMEBUFFER, 0 );
    
    glPopAttrib();
    
    if (mpCStackFBO != NULL){
	mpCStackFBO->Begin(FALSE, TRUE);
	mpCStackFBO = NULL;
    }
    else{
	if (mpCGLCanvasParent != NULL)
	    mpCGLCanvasParent->mpCFBOCurrentLink = NULL;
    }
}

void FBO_cls::
SetProjection(){
    GLPerspective_cls::SetProjection(mWidthFBO, mHeightFBO);
}

void FBO_cls::
SetFrustum(double degreeV, double min, double max){
    //double min = 0.1, max = 100;
    int tileMaxX = 1, tileMaxY = 1;
    int tileIndexX = 0, tileIndexY = 0;
    GLPerspective_cls::SetFrustumTiled
	(mWidthFBO, mHeightFBO, degreeV, min, max,
	 tileMaxX, tileMaxY, tileIndexX, tileIndexY);
}

void FBO_cls::
SetOrtho(double sizeOthro, double min, double max){
    int tileMaxX = 1, tileMaxY = 1;
    int tileIndexX = 0, tileIndexY = 0;
    GLPerspective_cls::SetOrthoTiled
	(sizeOthro, min, max,
	 tileMaxX, tileMaxY, tileIndexX, tileIndexY);
}

float FBO_cls::
GetLineWidthFactor(){
    switch(mWidthFBO){
    case 256:
	return 0.125;
    case 512:
	return 0.25;
    case 1024: 
	return 0.5;
    case 2048:
	return 1.0;
    case 4096:
	return 2.0;
    }
    return 1;
}

bool FBO_cls::msbInitialized = FALSE;
bool FBO_cls::msbSupported = TRUE;

//std::vector < FBO_cls* > FBO_cls::msVpCStack;
