// -*- c -*- 
// (setq c-basic-offset 4)
// "clsGLInfo.h"

#include "clsGLInfo.h"
#include <stdlib.h>
#include <stdio.h>

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

void GLInfo_cls::
Initialize(){
    mstrGLVersion    = (char*)glGetString(GL_VERSION);
    mstrGLExtentions = (char*)glGetString(GL_EXTENSIONS);
    
    sscanf(mstrGLVersion.c_str(), "%d.%d", &mMajorVersion, &mMinorVersion);

    if (mMajorVersion >= 2)
	mstrGLSLVersion  = (char*)glGetString(GL_SHADING_LANGUAGE_VERSION);
    else
	mstrGLSLVersion = "0.0";
    
    sscanf(mstrGLSLVersion.c_str(), "%d.%d", &mGLSLMajorVersion, &mGLSLMinorVersion);

    mMaxElementsVertices = 0;
    mMaxElementsIndices = 0;
    mStencilBit = 0;

    glGetIntegerv(GL_MAX_ELEMENTS_VERTICES, &mMaxElementsVertices);
    glGetIntegerv(GL_MAX_ELEMENTS_INDICES, &mMaxElementsIndices);

    glGetIntegerv(GL_MAX_TEXTURE_UNITS, &mMaxTextureUnit);
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &mMaxTextureSize);
    glGetIntegerv(GL_MAX_LIGHTS, &mMaxLightNum); 

    glGetIntegerv(GL_STENCIL_BITS, &mStencilBit);
    glGetIntegerv(GL_DEPTH_BITS, &mDepthBit);
}

wxString GLInfo_cls::
GetGLVersionStr(){
    return wxString::Format(_T("OpenGL ver. %d.%d"), mMajorVersion, mMinorVersion);
}

wxString GLInfo_cls::
GetGLSLVersionStr(){
    if (mGLSLMajorVersion == 0)
	return wxT("GLSL Not Implemented");
    else
	return wxString::Format(_T("GLSL ver. %d.%d"), mGLSLMajorVersion, mGLSLMinorVersion);
}

wxString GLInfo_cls::
GetGLMaxElementsVertices(){
    wxString strRet = wxT("Max Elements Vertices: ");
    strRet << mMaxElementsVertices;
    return strRet;
}

wxString GLInfo_cls::
GetGLMaxElementsIndices(){
    wxString strRet = wxT("Max Elements Indices: ");
    strRet << mMaxElementsIndices;
    return strRet;
}

wxString GLInfo_cls::
GetGLMaxTextureUnit(){
    wxString strRet = wxT("Max Texture Unit: ");
    strRet << mMaxTextureUnit;
    return strRet;
}

wxString GLInfo_cls::
GetGLMaxTextureSize(){
    wxString strRet = wxT("Max Texture Size: ");
    strRet << mMaxTextureSize;
    return strRet;
}

wxString GLInfo_cls::
GetGLMaxLightNum(){
    wxString strRet = wxT("Max Light Num: ");
    strRet << mMaxLightNum;
    return strRet;
}

wxString GLInfo_cls::
GetGLStencilBit(){
    wxString strRet = wxT("Stencil Bit: ");
    strRet << mStencilBit;
    return strRet;
}

wxString GLInfo_cls::
GetGLDepthBit(){
    wxString strRet = wxT("Depth Bit: ");
    strRet << mDepthBit;
    return strRet;
}

void GLInfo_cls::
GetAllGLInfo(wxString &strInfo){
    strInfo << GetGLVersionStr() << ENDL;
    strInfo << GetGLSLVersionStr() << ENDL;
    strInfo << GetGLMaxElementsVertices() << ENDL;
    strInfo << GetGLMaxElementsIndices() << ENDL;
    strInfo << GetGLMaxTextureUnit() << ENDL;
    strInfo << GetGLMaxTextureSize() << ENDL;
    strInfo << GetGLMaxLightNum() << ENDL;
    strInfo << GetGLStencilBit() << ENDL;
    strInfo << GetGLDepthBit() << ENDL;
}

bool GLInfo_cls::
IsGeometryShaderAvailable(){
    if (mGLSLMajorVersion >= 2) return TRUE;
    if (mGLSLMinorVersion >= 5) return TRUE;
    return FALSE;
}

GLInfo_cls *GpCGLInfo = NULL;
