
#include "clsGLPerspective.h"

#define CLS GLPerspective_cls

#include "../COMMON/CommonDefinitionsGL.h"
#include "../COMMON/CommonDefinitions.h"

#include <math.h>

void CLS::
GetWindowCoordv(double pos[3], double winPos[3]){
    double modelview[16], projection[16];
    int viewport[4];
    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);
    gluProject(pos[0], pos[1], pos[2],
	       modelview, projection, viewport,
	       &winPos[0], &winPos[1], &winPos[2]);
}

void CLS::
SetFrustum(int width, int height, double degreeV,
	   double min, double max){
    
    double baseSizeV = min * tan(3.14159 * degreeV / 180 / 2);
    double baseSizeH = baseSizeV * (double) width / height;
    
    double Xmin = -baseSizeH, Xmax = baseSizeH;
    double Ymin = -baseSizeV, Ymax = baseSizeV;

    glFrustum(Xmin, Xmax, Ymin, Ymax, min, max);
}

void CLS::
SetFrustumTiled(int width, int height, double degreeV,
		double min, double max,
		int tileMaxX, int tileMaxY,
		int tileIndexX, int tileIndexY){
    
    double baseSizeV = min * tan(3.14159 * degreeV / 180 / 2);
    double baseSizeH = baseSizeV * (double) width / height;
    
    double Xmin = -baseSizeH, Xmax = baseSizeH;
    double Ymin = -baseSizeV, Ymax = baseSizeV;

    double X0 = (Xmax-Xmin) * tileIndexX/tileMaxX + Xmin;
    double X1 = (Xmax-Xmin) * (tileIndexX+1)/tileMaxX + Xmin;
    double Y0 = (Ymax-Ymin) * tileIndexY/tileMaxY + Ymin;
    double Y1 = (Ymax-Ymin) * (tileIndexY+1)/tileMaxY + Ymin;

    glFrustum(X0, X1, Y0, Y1, min, max);
}

void CLS::
SetFrustumPartial(int width, int height, double degreeV,
		  double min, double max,
		  double posX, double posY, double part){

    double baseSizeV = min * tan(3.14159 * degreeV / 180 / 2);
    double baseSizeH = baseSizeV * (double) width / height;
    
    double Xmin = -baseSizeH, Xmax = baseSizeH;
    double Ymin = -baseSizeV, Ymax = baseSizeV;
    
    double X0 = (Xmax-Xmin) * (posX - part * 0.5) + Xmin;
    double X1 = (Xmax-Xmin) * (posX + part * 0.5) + Xmin;
    double Y0 = (Ymax-Ymin) * (posY - part * 0.5) + Ymin;
    double Y1 = (Ymax-Ymin) * (posY + part * 0.5) + Ymin;

    glFrustum(X0, X1, Y0, Y1, min, max);
}

void CLS::
SetOrthoTiled(double sizeOrtho,
	      double min, double max,
	      int tileMaxX, int tileMaxY,
	      int tileIndexX, int tileIndexY){

    double Xmin = -sizeOrtho, Xmax = sizeOrtho;
    double Ymin = -sizeOrtho, Ymax = sizeOrtho;

    double X0 = (Xmax-Xmin) * tileIndexX/tileMaxX + Xmin;
    double X1 = (Xmax-Xmin) * (tileIndexX+1)/tileMaxX + Xmin;
    double Y0 = (Ymax-Ymin) * tileIndexY/tileMaxY + Ymin;
    double Y1 = (Ymax-Ymin) * (tileIndexY+1)/tileMaxY + Ymin;

    glOrtho(X0, X1, Y0, Y1, min, max);
}

double CLS::
GetScaleForFixSizedObject(float matrixA[16], float matrixB[16]){
    glGetFloatv(GL_PROJECTION_MATRIX, matrixA);
    glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, matrixB);
    double scale = 1;
    if (matrixA[15] == 1)
	    scale = 1.0/matrixA[5] * 2;
    else
	scale = -matrixB[11];
    return scale;
}


void CLS::
SetProjection(int width, int height){
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
}

void CLS::
SetModelView(Position_cls CPos, Direction_cls CDir){
    double mat[16];
    glMatrixMode(GL_MODELVIEW);
    mat[0]  =  CDir.mRgt[X];
    mat[1]  =  CDir.mTop[X];
    mat[2]  = -CDir.mTgt[X];
    mat[3]  = 0;
    mat[4]  =  CDir.mRgt[Y];
    mat[5]  =  CDir.mTop[Y];
    mat[6]  = -CDir.mTgt[Y];
    mat[7]  = 0;
    mat[8]  =  CDir.mRgt[Z];
    mat[9]  =  CDir.mTop[Z];
    mat[10] = -CDir.mTgt[Z];
    mat[11] = mat[12] = mat[13] = mat[14] = 0;
    mat[15] = 1;
    
    glMultMatrixd(mat);
    
    glTranslated(-CPos.mPos[X],
		 -CPos.mPos[Y],
		 -CPos.mPos[Z]);
}

void CLS::
Push2DMatrix(int width, int height){
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glViewport( 0, 0, width, height);
    glLoadIdentity();
    gluOrtho2D(0, width, 0, height);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
}

void CLS::
Pop2DMatrix(){
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW); 
}

