// -*- c++ -*- 
// "clsFrameMain.h"
// (setq c-basic-offset 4)

#include "clsGLTexture.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../GRAPHIC/clsReadPNG.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#define CLS GLTexture_cls

CLS::
GLTexture_cls(){
}

CLS::
~GLTexture_cls(){
    ClearImage();
}

wxMemoryDC* CLS::
BeginDC(int width, int height){
    mWidthTex  = width;
    mHeightTex = height;
    mpCBitmap = new wxBitmap(width, height, -1);
    mpCMemoryDC = new wxMemoryDC(*mpCBitmap);

    return mpCMemoryDC;
}

void CLS::
EndDC(bool bUseAlpha, bool bFlip){
    if (mTextureID == 0)
	glGenTextures(1, &mTextureID);
    
    wxImage CImg;
    if (bFlip == TRUE) 
	CImg = mpCBitmap->ConvertToImage().Mirror(FALSE);
    else
	CImg = mpCBitmap->ConvertToImage();
        
    unsigned char *pDataMain  = CImg.GetData();

    unsigned char *pBufferLA = new unsigned char [mWidthTex * mHeightTex *  2];

    if (bUseAlpha == TRUE)
	for (int y = 0; y < mHeightTex; y++){
	    int indexy = mWidthTex * y;
	    for (int x = 0; x < mWidthTex; x++){
		int index = indexy + x;
		pBufferLA[index * 2 + 1] = pDataMain[index * 3 + 0];
		pBufferLA[index * 2 + 0] = 255;
	    }
	}
    else
	for (int y = 0; y < mHeightTex; y++){
	    int indexy = mWidthTex * y;
	    for (int x = 0; x < mWidthTex; x++){
		int index = indexy + x;
		pBufferLA[index * 2 + 1] = 255;
		pBufferLA[index * 2 + 0] = pDataMain[index * 3 + 0];
	    }
	}

    glBindTexture(GL_TEXTURE_2D, mTextureID);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE_ALPHA, mWidthTex, mHeightTex,
		 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, pBufferLA);

    delete[] pBufferLA;
    delete mpCMemoryDC;
    delete mpCBitmap;
}

bool CLS::
LoadImage(wxString strFile, bool bFlip, wxImage &CImg, wxImage &CImgDW){
    
    if (mstrImageLoaded.IsSameAs(strFile) &&
	mMinFilterUsed == mMinFilter &&
	mMagFilterUsed == mMagFilter &&
	mTexWrapUsed == mTexWrap) return FALSE;
    
    if (strFile.IsSameAs(wxT("NONE"))){
	CImg = wxImage(1, 1);
	CImg.SetRGB(0, 0, 255, 255, 255);
	mstrImageLoaded = wxT("NONE");
	return FALSE;
    }
    
    mMinFilterUsed = mMinFilter;
    mMagFilterUsed = mMagFilter;
    mTexWrapUsed = mTexWrap;

    FILE *fp = fopen(/*wxFNCONV(strFile).*/strFile.char_str(), "rb");
    if (fp == NULL){
	CImg = wxImage(1, 1);
	CImg.SetRGB(0, 0, 255, 255, 255);
	mstrImageLoaded = strFile;
	wxLogMessage(strFile+"::Load Failed\n");
	return FALSE;
    }
    wxLogMessage(strFile+"::Load Success\n");

    if (GCClassProvider.GetPCReadPNG()->ReadPNG(fp, strFile, CImg, CImgDW) == FALSE)
	CImg.LoadFile(strFile);
    
    // png_init_io  delete  fp=NULL ŕԂ悤ɂ
    if (fp != NULL) fclose(fp);
    
    bool bOK = CImg.IsOk();
    if (bOK == FALSE){
	CImg = wxImage(1, 1);
	CImg.SetRGB(0, 0, 255, 255, 255);
	//return FALSE; // !!_IMPLEMENT_ME_!!
    }

    if (CImgDW.IsOk() == TRUE){
	if (CImgDW.HasMask() == TRUE && CImgDW.HasAlpha() == FALSE)
	    CImgDW.InitAlpha();

	if (bFlip == TRUE)
	    CopyFromImage16bit(CImg.Mirror(FALSE), CImgDW.Mirror(FALSE));
	else
	    CopyFromImage16bit(CImg, CImgDW);

	if (&CImg == &msCImgWorkSpace)     CImg.Destroy();
	if (&CImgDW == &msCImgWorkSpaceDW) CImgDW.Destroy();
    
	mstrImageLoaded = strFile;
	return bOK;
    }
    
    if (CImg.HasMask() == TRUE && CImg.HasAlpha() == FALSE)
	CImg.InitAlpha();

    if (bFlip == TRUE)// CImg = CImg.Mirror(FALSE);
	CopyFromImage(CImg.Mirror(FALSE));
    else
	CopyFromImage(CImg);

    if (&CImg == &msCImgWorkSpace) CImg.Destroy();
    
    mstrImageLoaded = strFile;
    return bOK;
}

void CLS::
glTexImage2D_RGBSafe(const wxImage &CImg){
    int width  = CImg.GetWidth();
    int height = CImg.GetHeight();

    unsigned char *buf = new unsigned char [width * height * 3];
    for (int y = 0; y < height; y++){
	int indexY = y * width;
	for (int x = 0; x < width; x++){
	    int index = indexY + x;
	    buf[index*3 + 0] = CImg.GetRed(x, y);
	    buf[index*3 + 1] = CImg.GetGreen(x, y);
	    buf[index*3 + 2] = CImg.GetBlue(x, y);
	}
    }
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
		 width, height, 0,
		 GL_RGB, GL_UNSIGNED_BYTE, buf);
    delete[] buf;
}

void CLS::
CopyFromImage(const wxImage &CImg){
    if (mTextureID == 0)
	glGenTextures(1, &mTextureID);
    glBindTexture(GL_TEXTURE_2D, mTextureID);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mMagFilter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mMinFilter);

    mWidthTex = CImg.GetWidth();
    mHeightTex= CImg.GetHeight();

    mDepth = 8;

    if (CImg.HasAlpha() == FALSE){
	if (mWidthTex % 4 == 0){
	    SetAlignment(mWidthTex);
	    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
			 mWidthTex, mHeightTex, 0,
			 GL_RGB, GL_UNSIGNED_BYTE, CImg.GetData());
	}
	else
	    glTexImage2D_RGBSafe(CImg);
	glBindTexture(GL_TEXTURE_2D, 0);
	return; 
    }
    
    int width = mWidthTex;
    int height = mHeightTex;
    unsigned char *buf = new unsigned char [width * height * 4];
    for (int y = 0; y < height; y++){
	int indexY = y * width;
	for (int x = 0; x < width; x++){
	    int index = indexY + x;
	    buf[index*4 + 0] = CImg.GetRed(x,y);
	    buf[index*4 + 1] = CImg.GetGreen(x,y);
	    buf[index*4 + 2] = CImg.GetBlue(x,y);
	    buf[index*4 + 3] = CImg.GetAlpha(x,y);
	}
    }
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
		 width, height, 0,
		 GL_RGBA, GL_UNSIGNED_BYTE, buf);
    delete[] buf;

    glBindTexture(GL_TEXTURE_2D, 0);
}

void CLS::
CopyFromImage16bit(const wxImage &CImg, const wxImage &CImgDW){
    if (mTextureID == 0)
	glGenTextures(1, &mTextureID);
    glBindTexture(GL_TEXTURE_2D, mTextureID);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mMagFilter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mMinFilter);

    mWidthTex = CImg.GetWidth();
    mHeightTex= CImg.GetHeight();
    mDepth = 16;

    if (CImg.HasAlpha() == FALSE){
	int width  = mWidthTex;
	int height = mHeightTex;

	unsigned short *buf = new unsigned short [width * height * 3];
	for (int y = 0; y < height; y++){
	    int indexY = y * width;
	    for (int x = 0; x < width; x++){
		int index = indexY + x;
		buf[index*3 + 0] = CImg.GetRed  (x,y)*256 + CImgDW.GetRed  (x,y);
		buf[index*3 + 1] = CImg.GetGreen(x,y)*256 + CImgDW.GetGreen(x,y);
		buf[index*3 + 2] = CImg.GetBlue (x,y)*256 + CImgDW.GetBlue (x,y);
	    }
	}
	SetAlignment(width);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F_ARB, 
		     width, height, 0,
		     GL_RGB, GL_UNSIGNED_SHORT, buf);
	delete[] buf;
	
	glBindTexture(GL_TEXTURE_2D, 0);
	return; 
    }

    // !!_FIX_ME_!!
    // At@t̏ꍇ̏܂
}

void CLS::
CopyFromDataRGBA(unsigned char *pData, int width, int height){
    if (mTextureID == 0)
	glGenTextures(1, &mTextureID);
    glBindTexture(GL_TEXTURE_2D, mTextureID);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mMagFilter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mMinFilter);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
		 width, height, 0,
		 GL_RGBA, GL_UNSIGNED_BYTE, pData);

    mWidthTex = width;
    mHeightTex = height;
    mDepth = 8;

    glBindTexture(GL_TEXTURE_2D, 0);
}

void CLS::
CopyFromDataRGBA16(unsigned short *pData, int width, int height){
    if (mTextureID == 0)
	glGenTextures(1, &mTextureID);
    glBindTexture(GL_TEXTURE_2D, mTextureID);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mMagFilter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mMinFilter);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F_ARB,
		 width, height, 0,
		 GL_RGBA, GL_UNSIGNED_BYTE, pData);

    mWidthTex = width;
    mHeightTex = height;
    mDepth = 16;
    
    glBindTexture(GL_TEXTURE_2D, 0);
}

void CLS::
CopyFromDataRGBA16HF(unsigned short *pData, int width, int height){
    if (mTextureID == 0)
	glGenTextures(1, &mTextureID);
    glBindTexture(GL_TEXTURE_2D, mTextureID);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mMagFilter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mMinFilter);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F_ARB,
		 width, height, 0,
		 GL_RGBA, GL_HALF_FLOAT_ARB, pData);

    mWidthTex = width;
    mHeightTex = height;
    mDepth = 16;

    glBindTexture(GL_TEXTURE_2D, 0);
}

void CLS::
CopyFromDataFloat(float *pData, int width, int height){
    if (mTextureID == 0)
	glGenTextures(1, &mTextureID);
    glBindTexture(GL_TEXTURE_2D, mTextureID);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, mTexWrap);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mMagFilter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mMinFilter);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT,
		 width, height, 0,
		 GL_DEPTH_COMPONENT, GL_FLOAT, pData);

    mWidthTex = width;
    mHeightTex = height;
    mDepth = 16;

    glBindTexture(GL_TEXTURE_2D, 0);
}

wxImage CLS::msCImgWorkSpace;
wxImage CLS::msCImgWorkSpaceDW;
