// -*- c -*-
// "clsPosition.h"

#include "clsPosition.h"

#include "../COMMON/CommonDefinitions.h"

#include <math.h>

#define CLS Direction_cls

CLS::
Direction_cls(double tgtX, double tgtY, double tgtZ,
	      double topX, double topY, double topZ,
	      double rgtX, double rgtY, double rgtZ){
    mTgt[X] = tgtX; mTgt[Y] = tgtY; mTgt[Z] = tgtZ;
    mTop[X] = topX; mTop[Y] = topY; mTop[Z] = topZ;
    mRgt[X] = rgtX; mRgt[Y] = rgtY; mRgt[Z] = rgtZ;
}

CLS::
Direction_cls(std::vector < float > VIniTgt,
	      std::vector < float > VIniTop,
	      std::vector < float > VIniRgt){
    LOOP_XYZ mTgt[xyz] = VIniTgt[xyz];
    LOOP_XYZ mTop[xyz] = VIniTop[xyz];
    LOOP_XYZ mRgt[xyz] = VIniRgt[xyz];
}

void CLS::
SetDir(Direction_cls CDir){
    LOOP_XYZ mTgt[xyz] = CDir.mTgt[xyz];
    LOOP_XYZ mTop[xyz] = CDir.mTop[xyz];
    LOOP_XYZ mRgt[xyz] = CDir.mRgt[xyz];
}

bool CLS::
IsSimilarTgtDirection(Direction_cls& CDir, float criterion){
    float dotProduct = mTgt[X]*CDir.mTgt[X] + mTgt[Y]*CDir.mTgt[Y] + mTgt[Z]*CDir.mTgt[Z];
    if (dotProduct >= criterion) return TRUE;
    
    return FALSE;
}

bool CLS::
Orthogonalize(double vec0[3], double vec1[3]){
    float product = GetScalarProduct(vec0, vec1);
    LOOP_XYZ
	vec1[xyz] -= vec0[xyz] * product;
    bool bValid = FALSE;
    LOOP_XYZ if (vec1[xyz] > 1e-12 || vec1[xyz] < -1e-12) bValid = TRUE;
    return bValid;
}

void CLS::
NormalizeAndOrthogonalize(){
    Normalize(mTgt, 1.0);
    if (Orthogonalize(mTgt, mTop) == FALSE){
	mTgt[X] = 0; mTgt[Y] = 0; mTgt[Z] = -1;
	mTop[X] = 0; mTop[Y] = 1; mTop[Z] = 0;
    }
    Normalize(mTop, 1.0);
    GetCrossedVector(mTgt, mTop, mRgt);
}

void CLS::
RotateYPR(double Yaw, double Pitch, double Roll){
    RotateYaw(Yaw);
    NormalizeAndOrthogonalize();
    RotatePitch(Pitch);
    NormalizeAndOrthogonalize();
    RotateRoll(Roll);
}

void CLS::
RotateYaw(double Yaw){
    double YawRad = Yaw * 3.1415926 / 180;
    double cosYaw = cos(YawRad);
    double sinYaw = sin(YawRad);

    double tgtN[3] = {
	cosYaw * mTgt[X] - sinYaw * mRgt[X],
	cosYaw * mTgt[Y] - sinYaw * mRgt[Y],
	cosYaw * mTgt[Z] - sinYaw * mRgt[Z] };
    double rgtN[3] = {
	cosYaw * mRgt[X] + sinYaw * mTgt[X],
	cosYaw * mRgt[Y] + sinYaw * mTgt[Y],
	cosYaw * mRgt[Z] + sinYaw * mTgt[Z] };
    
    LOOP_XYZ mTgt[xyz] = tgtN[xyz];
    LOOP_XYZ mRgt[xyz] = rgtN[xyz];
}

void CLS::
RotatePitch(double Pitch){
    double PitchRad = Pitch * 3.1415926 / 180;
    double cosPitch = cos(PitchRad);
    double sinPitch = sin(PitchRad);

    double tgtN[3] = {
	cosPitch * mTgt[X] - sinPitch * mTop[X],
	cosPitch * mTgt[Y] - sinPitch * mTop[Y],
	cosPitch * mTgt[Z] - sinPitch * mTop[Z] };
    double topN[3] = {
	cosPitch * mTop[X] + sinPitch * mTgt[X],
	cosPitch * mTop[Y] + sinPitch * mTgt[Y],
	cosPitch * mTop[Z] + sinPitch * mTgt[Z] };
    
    LOOP_XYZ mTgt[xyz] = tgtN[xyz];
    LOOP_XYZ mTop[xyz] = topN[xyz];
}

void CLS::
RotateRoll(double Roll){
    double RollRad = Roll * 3.1415926 / 180;
    double cosRoll = cos(RollRad);
    double sinRoll = sin(RollRad);

    double topN[3] = {
	cosRoll * mTop[X] + sinRoll * mRgt[X],
	cosRoll * mTop[Y] + sinRoll * mRgt[Y],
	cosRoll * mTop[Z] + sinRoll * mRgt[Z] };

    double rgtN[3] = {
	cosRoll * mRgt[X] - sinRoll * mTop[X],
	cosRoll * mRgt[Y] - sinRoll * mTop[Y],
	cosRoll * mRgt[Z] - sinRoll * mTop[Z] };

    LOOP_XYZ mTop[xyz] = topN[xyz];
    LOOP_XYZ mRgt[xyz] = rgtN[xyz];
}

std::vector < float > CLS::
GetYPR(){
    float yaw, pitch, roll;
    
    if (abs(mTgt[X]) < 1e-4 && abs(mTgt[Y]) < 1e-4){
	yaw = 0;
	pitch = asin(-mTgt[Z]);
	roll = atan2(mTop[Y], mTop[X]) - 3.1415926f;
    }
    else{
	yaw   = atan2(-mTgt[Y], -mTgt[X]) + 3.1415926f; 
	pitch = asin(-mTgt[Z]);
	roll  = -asin(mRgt[Z]/cos(pitch)); 
	if (mTop[Z] < 0) roll = 3.1415926f - roll;
    }
    
    if (roll > 3.1415926 ) roll -= 2*3.1415926;
    if (roll < - 3.1415926 ) roll += 2*3.1415926;
    if (yaw > 3.1415926) yaw -= 2*3.1415926;
    if (yaw < - 3.1415926) yaw += 2*3.1415926;

    yaw   *= 180 / 3.1415926f;
    pitch *= 180 / 3.1415926f;
    roll  *= 180 / 3.1415926f;



    std::vector < float > VYPR;
    VYPR.push_back(yaw); VYPR.push_back(pitch); VYPR.push_back(roll);
    return VYPR;
}

std::vector < float > CLS::
GetRPY(){
    std::vector < float > VYPR = GetYPR();
    std::vector < float > VRPY;
    VRPY.push_back(VYPR[2]);
    VRPY.push_back(VYPR[1]);
    VRPY.push_back(VYPR[0]);
    return VRPY;
}

#define CLS Position_cls

CLS::
Position_cls(double posX, double posY, double posZ){
    mPos[X] = posX; mPos[Y] = posY; mPos[Z] = posZ;
}

CLS::
Position_cls(std::vector < float > VIni){
    LOOP_XYZ mPos[xyz] = VIni[xyz]; }

CLS::
Position_cls(std::vector < double > VIni){ 
    LOOP_XYZ mPos[xyz] = VIni[xyz]; }

void CLS::
SetPos(Position_cls CPos){ 
    LOOP_XYZ mPos[xyz] = CPos.mPos[xyz]; }

bool CLS::
IsSimilarPosition(Position_cls &CPos, double criterion){
    bool bSimilar = TRUE;
    LOOP_XYZ if (mPos[xyz] < CPos.mPos[xyz] - criterion || mPos[xyz] > CPos.mPos[xyz] + criterion) bSimilar = FALSE;
    return bSimilar;
}

double CLS::
GetLength(){
    return sqrt(mPos[0]*mPos[0] + mPos[1]*mPos[1] + mPos[2]*mPos[2]);
}

void CLS::
NormalizePos(double length){
    double norm = GetLength();
    LOOP_XYZ mPos[xyz] /= norm;
    LOOP_XYZ mPos[xyz] *= length;
}

#define CLS PositionAndDirection_cls

CLS::
PositionAndDirection_cls(){
    mbLock[X] = mbLock[Y] = mbLock[Z] = FALSE;
    mPivotPos[X] = mPivotPos[Y] = mPivotPos[Z] = 0;
    mSensitivity = 1;
}

// modelview matrix  inverted 
CLS::
PositionAndDirection_cls(double matrix[16], bool bInvert){
    if (bInvert == TRUE){
	mRgt[X] = matrix[0]; mTop[X] = matrix[4]; mTgt[X] = -matrix[ 8];
	mRgt[Y] = matrix[1]; mTop[Y] = matrix[5]; mTgt[Y] = -matrix[ 9];
	mRgt[Z] = matrix[2]; mTop[Z] = matrix[6]; mTgt[Z] = -matrix[10];
    }
    else{
	mTgt[X] = matrix[0]; mRgt[X] = -matrix[4]; mTop[X] = matrix[ 8];
	mTgt[Y] = matrix[1]; mRgt[Y] = -matrix[5]; mTop[Y] = matrix[ 9];
	mTgt[Z] = matrix[2]; mRgt[Z] = -matrix[6]; mTop[Z] = matrix[10];
    }
    mPos[X] = matrix[12];
    mPos[Y] = matrix[13];
    mPos[Z] = matrix[14];
}

void CLS::
GetMatrix16(double matrix[16]){
    matrix[0] = mTgt[X]; matrix[4] = -mRgt[X]; matrix[ 8] = mTop[X];
    matrix[1] = mTgt[Y]; matrix[5] = -mRgt[Y]; matrix[ 9] = mTop[Y];
    matrix[2] = mTgt[Z]; matrix[6] = -mRgt[Z]; matrix[10] = mTop[Z];
    matrix[3] = 0;       matrix[7] = 0;       matrix[11] = 0;

    matrix[12] = mPos[X];
    matrix[13] = mPos[Y];
    matrix[14] = mPos[Z];
    matrix[15] = 1;
}

void CLS::
GetMatrix16Inv(double matrix[16]){
    double matrixSrc[16];
    GetMatrix16(matrixSrc);
    GetInvMatrix4x4(matrixSrc, matrix);
}

void CLS::
GetMatrix16f(float matrix[16]){
    double matrix_d[16];
    GetMatrix16(matrix_d);
    ConvertToFloat(matrix, matrix_d, 16);
}

void CLS::
GetMatrix16fInv(float matrix[16]){
    double matrix_d[16], matrixSrc[16];
    GetMatrix16(matrixSrc);
    GetInvMatrix4x4(matrixSrc, matrix_d);
    ConvertToFloat(matrix, matrix_d, 16);
}

void CLS::
FromModelView(float matrix[16]){
    mRgt[X] = matrix[0]; mTop[X] = matrix[1]; mTgt[X] = -matrix[2];
    mRgt[Y] = matrix[4]; mTop[Y] = matrix[5]; mTgt[Y] = -matrix[6];
    mRgt[Z] = matrix[8] ;mTop[Z] = matrix[9]; mTgt[Z] = -matrix[10];

    //mPos[X] = -matrix[12]; mPos[Y] = -matrix[13]; mPos[Z] = -matrix[14];

    float matrixInv[16];
    GetInvMatrix4x4(matrix, matrixInv);

    matrixInv[12] = 0; matrixInv[13] = 0; matrixInv[14] = 0;
    float matrixTmp[16];
    for (int i = 0; i < 16; i++) matrixTmp[i] = matrix[i];
    Multiply(matrixInv, matrixTmp);
    mPos[X] = -matrixTmp[12]; mPos[Y] = -matrixTmp[13]; mPos[Z] = -matrixTmp[14];
}

void CLS::
ToModelView(float matrix[16]){
    matrix[0]  =  mRgt[X];
    matrix[1]  =  mTop[X];
    matrix[2]  = -mTgt[X];
    matrix[3]  = 0;
    matrix[4]  =  mRgt[Y];
    matrix[5]  =  mTop[Y];
    matrix[6]  = -mTgt[Y];
    matrix[7]  = 0;
    matrix[8]  =  mRgt[Z];
    matrix[9]  =  mTop[Z];
    matrix[10] = -mTgt[Z];
    matrix[11] = 0;
    matrix[12] = -mPos[X]*matrix[0] - mPos[Y]*matrix[4] - mPos[Z]*matrix[8];
    matrix[13] = -mPos[X]*matrix[1] - mPos[Y]*matrix[5] - mPos[Z]*matrix[9];
    matrix[14] = -mPos[X]*matrix[2] - mPos[Y]*matrix[6] - mPos[Z]*matrix[10];
    matrix[15] = 1;
}

void CLS::
SetPivotPos(Position_cls CPos){
    LOOP_XYZ mPivotPos[xyz] = CPos.mPos[xyz];
}

void CLS::
TrackBall(float currX, float currY, float dx, float dy){
    
    if (dx > 0.2) dx = 0.2f;
    if (dx < -0.2) dx = -0.2f;
    if (dy > 0.2) dy = 0.2f;
    if (dy < -0.2) dy = -0.2f;

    //float mAncPos[3] = {0, 0, 0};
    
    LOOP_XYZ mPos[xyz] -= mPivotPos[xyz];

    float rgt[3], top[3];
    LOOP_XYZ rgt[xyz] = mRgt[xyz];
    LOOP_XYZ top[xyz] = mTop[xyz];
    VectorMath_cls::Normalize(rgt, 1.0f);
    VectorMath_cls::Normalize(top, 1.0f);
    
    float distance = GetLength();

    float dxMass = 1 - currY * currY; 
    float dyMass = 1 - currX * currX; 
    if (dxMass < 0) dxMass = 0;
    if (dyMass < 0) dyMass = 0;

    LOOP_XYZ{
	mPos[xyz]   -= rgt[xyz] * dx * dxMass * distance ;
	mTgt[xyz]   += rgt[xyz] * dx * dxMass; }
    
    LOOP_XYZ{
	mPos[xyz]   -= top[xyz] * dy * dyMass * distance;
	mTgt[xyz]   += top[xyz] * dy * dyMass; }

    LOOP_XYZ
	mTop[xyz]   += (currX * dy - currY * dx) * rgt[xyz] ;

    NormalizePos(distance);

    LOOP_XYZ mPos[xyz] += mPivotPos[xyz];

    NormalizeAndOrthogonalize();
}

void CLS::
TranslateInWC(float dx, float dy){
    LOOP_XYZ{
	mPos[xyz] -= mRgt[xyz] * dx;
	mPos[xyz] -= mTop[xyz] * dy;
    }
}

void CLS::
Translate(float dx, float dy, PositionAndDirection_cls *pCParentCoord){
    float mATranslateScale[3] = {1, 1, 1};

    if (pCParentCoord != NULL){
	LOOP_XYZ{
	    mPos[xyz]   -= mRgt[xyz] * dx * mSensitivity * (!mbLock[xyz]) * mATranslateScale[xyz];
	    mPos[xyz]   -= mTop[xyz] * dy * mSensitivity * (!mbLock[xyz]) * mATranslateScale[xyz];
	}
	
    }
    else
	LOOP_XYZ{
	    mPos[xyz] -= mRgt[xyz] * dx * mSensitivity * (!mbLock[xyz]) * mATranslateScale[xyz];
	    mPos[xyz] -= mTop[xyz] * dy * mSensitivity * (!mbLock[xyz]) * mATranslateScale[xyz];
	}
}

void CLS::
Zoom(float dx, float dy){

    LOOP_XYZ mPos[xyz] -= mPivotPos[xyz];

    float distance = GetLength();
    distance *= (1.0 + dy - dx * 0.2);
    
    NormalizePos(distance);

    LOOP_XYZ mPos[xyz] += mPivotPos[xyz];

    NormalizeAndOrthogonalize();
}

void CLS::
SetMixed(PositionAndDirection_cls& CPos1, PositionAndDirection_cls& CPos2, double rate){
    LOOP_XYZ mPos[xyz] = (1.0-rate) * CPos1.mPos[xyz] + rate * CPos2.mPos[xyz];
    LOOP_XYZ mTgt[xyz] = (1.0-rate) * CPos1.mTgt[xyz] + rate * CPos2.mTgt[xyz];
    LOOP_XYZ mRgt[xyz] = (1.0-rate) * CPos1.mRgt[xyz] + rate * CPos2.mRgt[xyz];
    LOOP_XYZ mTop[xyz] = (1.0-rate) * CPos1.mTop[xyz] + rate * CPos2.mTop[xyz];
    NormalizeAndOrthogonalize();
}

void CLS::
LookAtPivot(){
    if (mPivotPos[X] == mPos[X] &&
	mPivotPos[Y] == mPos[Y] &&
	mPivotPos[Z] == mPos[Z]) return;
    LOOP_XYZ mTgt[xyz] = mPivotPos[xyz] - mPos[xyz];
    NormalizeAndOrthogonalize();
}

bool operator==(const PositionAndDirection_cls& a, const PositionAndDirection_cls& b){
    LOOP_XYZ {
	if (a.mPos[xyz] != b.mPos[xyz]) return FALSE;
	if (a.mTgt[xyz] != b.mTgt[xyz]) return FALSE;
	if (a.mTop[xyz] != b.mTop[xyz]) return FALSE;
	if (a.mRgt[xyz] != b.mRgt[xyz]) return FALSE;
    }
    return TRUE;
}
