// -*- c -*-
// "clsPosition.cpp"

#ifndef _POSITION_H_
#define _POSITION_H_

#include "../COMMON/clsVectorMath.h"
#include "../COMMON/clsTensorMath.h"

#include <stdlib.h>
#include <vector>

class Direction_cls : public VectorMath_cls, public TensorMath_cls {
public:
    Direction_cls(double tgtX =  0, double tgtY =  0, double tgtZ = -1,
		  double topX =  0, double topY =  1, double topZ =  0,
		  double rgtX =  1, double rgtY =  0, double rgtZ =  0);
    Direction_cls(std::vector < float > VIniTgt,
		  std::vector < float > VIniTop,
		  std::vector < float > VIniRgt);
    
    double mTgt [3], mTop [3], mRgt [3];
    
    void SetDir(Direction_cls CDir);
    bool IsSimilarTgtDirection(Direction_cls &CDir, float criterion);
    bool Orthogonalize(double vec0[3], double vec1[3]);
    void NormalizeAndOrthogonalize();

    void   RotateYPR(double Yaw, double Pitch, double Roll);
    void   RotateYaw(double Yaw);
    void   RotatePitch(double Pitch);
    void   RotateRoll(double Roll);
    std::vector < float > GetYPR();
    std::vector < float > GetRPY();
};

class Position_cls{
public:
    Position_cls(double posX =  0, double posY =  0, double posZ = 0);
    Position_cls(std::vector < float > VIni);
    Position_cls(std::vector < double > VIni);
    double mPos[3];

    void SetPos(Position_cls CPos);
    bool IsSimilarPosition(Position_cls &CPos, double criterion);
    double GetLength();
    void NormalizePos(double length);
    
};

class PositionAndDirection_cls : public Position_cls, public Direction_cls{
 public:
    PositionAndDirection_cls();
    PositionAndDirection_cls(double matrix[16], bool bInvert = false);
    
    bool mbLock[3];
    float mPivotPos[3];
    float mScale[3];
    double mSensitivity;
    
    // ΉModelViewMatrix̂̂ł͂Ȃ̂Œ
    void GetMatrix16    (double matrix[16]);
    void GetMatrix16Inv (double matrix[16]);
    void GetMatrix16f   (float matrix[16]);
    void GetMatrix16fInv(float matrix[16]);

    // 
    void FromModelView(float matrix[16]);
    void ToModelView(float matrix[16]);

    void SetPivotPos(Position_cls CPos);    

    void TrackBall(float currX, float currY, float dx, float dy);
    void TranslateInWC(float dx, float dy);
    void Translate(float dx, float dy, 
		   PositionAndDirection_cls *pCParentCoord = NULL);
    void Zoom(float dx, float dy);

    void SetMixed(PositionAndDirection_cls &CPos1, PositionAndDirection_cls& CPos2, double rate);

    void LookAtPivot();

    friend bool operator==(const PositionAndDirection_cls& a, const PositionAndDirection_cls& b);
    
};

#endif
