// -*- c++ -*-
// (setq c-basic-offset 4)
// "clsShadowVolumeCtrl.h"

#include "clsShadowVolumeCtrl.h"

#include "../COMMON/clsHalton.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsFBO.h"

#define CLS ShadowVolumeCtrl_cls

CLS::
ShadowVolumeCtrl_cls(){
    mSVStep = 0;
    mSVIni = 0;
    mSVState = 0;
    mpCFBOBuffer = NULL;
    SetToDefaultParam();
}

CLS::
~ShadowVolumeCtrl_cls(){
    ClearFBOBuffer();
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();

    wxString strUpdate = "VIEW3D::PROPERTY_RENDER";

    mCConfList.SetTag("ShadowVolumeCtrl");
    mCConfList.Add(&mCbUseShadowVolume, "UseShadowVolume=", "0", strUpdate);
    mCConfList.Add(&mCShadowLength, "ShadowVolumeLength=", "20", strUpdate);
    mCConfList.Add(&mCOffsetFactor, "ShadowVolumeOffset=", "0.0003", strUpdate);
    mCConfList.Add(&mCbUseGammaForSV, "UseGammaForShadowVolume=", "0", strUpdate);
    mCConfList.Add(&mCbForClosed, "ForClosed=", "0", strUpdate);
    mCConfList.Add(&mCbClipCheck, "ShadowVolumeClipCheck=", "1", strUpdate);
    mCConfList.Add(&mCbTransparentShadowForEnsemble, "UseTransparentShadow=", "0", strUpdate);
}

void CLS::
ClearFBOBuffer(){
    if (mpCFBOBuffer != NULL){
	delete mpCFBOBuffer;
	mpCFBOBuffer = NULL;
    }
}

void CLS::
SetShadowVolumeStep(int shadowVolumeStep, int shadowVolumeIni){
    mSVStep = shadowVolumeStep;
    mSVIni = shadowVolumeIni;
}

/*
void CLS::
SetEnsembleSampleNum(int ensembleSampleNum){
    mEnsembleSampleNum = ensembleSampleNum;
}
*/

int CLS::
GetShadowVolumeSampleIni(){
    if (mCbUseShadowVolume.GetBodyBOOL() == 0) return 0;
    int lightNum = GetPCSceneObjCtrl()->GetActiveLightNumRoot(FALSE);
    int ShadowNum = GetPCSceneObjCtrl()->GetActiveLightNumRoot(TRUE);
    if (lightNum == ShadowNum) return 1;
    return 0;
}

int CLS::
GetShadowVolumeSampleNum(){
    if (mCbUseShadowVolume.GetBodyBOOL() == 0) return 1;
    int shadowNum = GetPCSceneObjCtrl()->GetActiveLightNumRoot(TRUE);
    return shadowNum + 1;
}

bool CLS::
IsFirstStep(){
    if (mSVStep == mSVIni) return TRUE;
    return FALSE;
}

bool CLS::
IsSkip(float alpha, int ensemble, SceneObj_cls *pCObj, int offsetBase){
    if (mCbTransparentShadowForEnsemble.GetBodyBOOL() == FALSE) return FALSE;
    int offset = (((long long)pCObj / 8) + offsetBase) % 256;
    float crit = GetPCHalton()->GetX(ensemble + offset);
    if (crit > alpha) return TRUE;
    return FALSE;
}

void CLS::
PushState(){
    //_ASSERT(mSVState == 0);
    
    glClearStencil(0);
    glClear(GL_STENCIL_BUFFER_BIT);
    
    glEnable(GL_STENCIL_TEST);
    glStencilMask(~0);
}

void CLS::
SetStencilState(){
    mSVState = 1;
    glDisable(GL_CULL_FACE);
    glDepthMask(GL_FALSE);
    glStencilFunc(GL_ALWAYS, 0, ~0);
    glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
    //glColorMask(GL_TRUE, GL_FALSE, GL_FALSE, GL_FALSE);
    glDisable(GL_BLEND);
}

void CLS::
SettingFront(){
    glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
}

void CLS::
SettingBack(){
    glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
}

void CLS::
SetShadowedArea(){
    mSVState = 2;
    glDepthMask(GL_TRUE);
    glClear(GL_DEPTH_BUFFER_BIT);
    glEnable(GL_STENCIL_TEST);
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    //glStencilFunc(GL_ALWAYS, 0, 255);
    glStencilFunc(GL_LESS, 0, ~0);
}

void CLS::
PopState(){
    mSVState = 0;
    
    glStencilFunc(GL_ALWAYS, 0, 255);
    glDisable(GL_STENCIL_TEST);

    glClearStencil(0);
    glClear(GL_STENCIL_BUFFER_BIT);
}

void CLS::
BeginShader(){
    ShaderSVTriangleMesh_cls CParam;
    CParam.mShadowLength = mCShadowLength.GetBodyDOUBLE();
    CParam.mOffsetFactor = mCOffsetFactor.GetBodyDOUBLE();
    if (mCbForClosed.GetBodyBOOL() == TRUE)
	CParam.mForClosed = 1;
    else
	CParam.mForClosed = -1;
    
    GetPCShaderCtrl()->BeginSVTrinagleMesh(&CParam);
}

void CLS::
BeginShaderHF(){
    ShaderSVTriangleMesh_cls CParam;
    CParam.mShadowLength = mCShadowLength.GetBodyDOUBLE();
    CParam.mOffsetFactor = mCOffsetFactor.GetBodyDOUBLE();

    if (mCbForClosed.GetBodyBOOL() == TRUE)
	CParam.mForClosed = 1;
    else
	CParam.mForClosed = -1;
    
    GetPCShaderCtrl()->BeginSVTrinagleMeshHF(&CParam);
}

void CLS::
SetCullingSide(int step){
    int cullingSide = 0;
    if (mCbForClosed.GetBodyBOOL() == TRUE){
	if (step == 1) cullingSide = 1;
	if (step == 2) cullingSide = 0; 
    }
    else {
	if (step == 1) cullingSide = 0;
	if (step == 2) cullingSide = 1; 
    }
    GetPCShaderCtrl()->SetCullingSide(cullingSide);
}

void CLS::
SetCullingSideHF(int step){
    int cullingSide = 0;
    if (mCbForClosed.GetBodyBOOL() == TRUE){
	if (step == 1) cullingSide = 1;
	if (step == 2) cullingSide = 0; 
    }
    else {
	if (step == 1) cullingSide = 0;
	if (step == 2) cullingSide = 1; 
    }
    GetPCShaderCtrl()->SetCullingSideHF(cullingSide);
}

void CLS::
EndShader(){
    GetPCShaderCtrl()->EndShader();
}

void CLS::
PrepareBuffer(FBO_cls *pCFBOSrc){
    int size = pCFBOSrc->mWidthFBO;
    int depth = pCFBOSrc->mColorDepth;
    
    if (mpCFBOBuffer == NULL){
	mpCFBOBuffer = new FBO_cls(size, size, NULL);
	mpCFBOBuffer->SetColorDepth(depth);
    }
    else{
	mpCFBOBuffer->SetGeometryFBO(size, size);
	mpCFBOBuffer->SetColorDepth(depth);
    }

    mpCFBOBuffer->MakeFBO(_FBO_USE_RGBA, FALSE);
}

void CLS::
DrawOnBuffer(FBO_cls *pCFBOSrc){
    
    int size = pCFBOSrc->mWidthTex;

    glClearColor(0, 0, 0, 0);
    bool bClear = IsFirstStep();
    mpCFBOBuffer->Begin(bClear);

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    GetPCGLExtentions()->ClampColor(GL_FALSE);

    glColor4f(1.0, 1.0, 1.0, 1.0f);

    glEnable(GL_TEXTURE_2D);

    ExposureParams_cls CParam;
    if (mCbUseGammaForSV.GetBodyBOOL() == TRUE) CParam.mGamma = 2.2;
    GetPCShaderCtrl()->BeginExposure(&CParam);

    glEnable(GL_BLEND);
    glDisable(GL_LIGHTING);
    GetPCGLExtentions()->
	BlendFuncSeparate(GL_SRC_ALPHA, GL_ONE,
			  GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    pCFBOSrc->DrawOnScreen2D(size, size,
			     size, size,
			     0, 0, TRUE);

    GetPCShaderCtrl()->EndShader();

    glPopAttrib();

    mpCFBOBuffer->End();
}

void CLS::
DrawFromBuffer(FBO_cls *pCFBOTgt){
    
    int size = pCFBOTgt->mWidthTex;

    glClearColor(0, 0, 0, 0);
    //bool bClear = (mSVStep == 0) ? TRUE : FALSE;
    pCFBOTgt->Begin(TRUE);
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glColor4f(1.0, 1.0, 1.0, 1.0f);

    glEnable(GL_TEXTURE_2D);

    ExposureParams_cls CParam;
    if (mCbUseGammaForSV.GetBodyBOOL() == TRUE) CParam.mGamma = 0.4545;
    GetPCShaderCtrl()->BeginExposure(&CParam);

    glEnable(GL_BLEND);
    glDisable(GL_LIGHTING);
    glBlendFunc(GL_ONE, GL_ZERO);

    mpCFBOBuffer->DrawOnScreen2D(size, size,
				 size, size,
				 0, 0, TRUE);

    GetPCShaderCtrl()->EndShader();

    glPopAttrib();

    pCFBOTgt->End();
}

ShadowVolumeCtrl_cls *GpCShadowVolumeCtrl = NULL;
