// -*- c++ -*- 
// "clsTexturePrivoder.h"
// (setq c-basic-offset 4)

#include "clsTextureProvider.h"

#include "../MYGL/clsGLTexture.h"
#include "../PREFERENCE/clsPreference.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
#endif

#define CLS TextureProvider_cls

CLS::
TextureProvider_cls(){
    mCounter = 0;
    mUsedMemoryByte = 0;
}

CLS::
~TextureProvider_cls(){
    ClearAll();
}

void CLS::
ClearAll(){
    uint num = mVpCGLTexture.size();
    for (uint i = 0; i < num; i++)
	delete mVpCGLTexture[i];
    mVpCGLTexture.clear();
    mUsedMemoryByte = 0;
}


void CLS::
ReleaseOldImages(){
    int memoryLimit = GetPCPreference()->mCMemoryLimitTexMB.GetBodyINT();
    while (TRUE){
	int usedMemoryMB = GetUsedMemoryMB();
	if (usedMemoryMB < memoryLimit) break;
	
	uint num = mVpCGLTexture.size();
	int candidate = -1;
	int oldest = mCounter;
	for (uint i = 0; i < num; i++){
	    if (mVpCGLTexture[i]->mTagCounter < oldest){
		oldest = mVpCGLTexture[i]->mTagCounter;
		candidate = i;
	    }
	}

	if (candidate != -1){
	    mUsedMemoryByte -= mVpCGLTexture[candidate]->GetByte();
	    delete mVpCGLTexture[candidate];
	    mVpCGLTexture.erase(mVpCGLTexture.begin() + candidate);
	}
    }
}

wxSize CLS::
GetTextureSize(int index){
    if (index < 0) return wxSize(1, 1);
    wxSize CSize(mVpCGLTexture[index]->mWidthTex,
		 mVpCGLTexture[index]->mHeightTex);
    return CSize;
}

int CLS::
BindTexture(wxString strImagePath){
    uint num = mVpCGLTexture.size();
    for (uint i = 0; i < num; i++){
	if (mVpCGLTexture[i]->mstrImageLoaded.IsSameAs(strImagePath)){
	    mVpCGLTexture[i]->BindTexture();
	    return i;
	}
    }

    GLTexture_cls *pCGLTexture = new GLTexture_cls;

    pCGLTexture->LoadImage(strImagePath, TRUE);
    
    mVpCGLTexture.push_back(pCGLTexture);

    mUsedMemoryByte += pCGLTexture->GetByte(); 

    pCGLTexture->BindTexture();
    pCGLTexture->mTagCounter = mCounter;
    mCounter++;

    ReleaseOldImages(); 

    return mVpCGLTexture.size()-1;
}

void CLS::
BindBlankTexture(bool bWhite, int active){

    switch(active){
    case 0:
	glActiveTexture(GL_TEXTURE0); break;
    case 1:
	glActiveTexture(GL_TEXTURE1); break;
    case 2:
	glActiveTexture(GL_TEXTURE2); break;
    case 3:
	glActiveTexture(GL_TEXTURE3); break;
    }
    
    if (bWhite == TRUE)
	BindTexture("Resource/TEXTURE/White.png");
    else
	BindTexture("Resource/TEXTURE/Black.png");
    // gƂ͑SʂłׂȂ̂

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP_TO_EDGE);
    glActiveTexture(GL_TEXTURE0);
}

TextureProvider_cls *GpCTextureProvider = NULL;
