// "clsWideCamera.h"
// (setq c-basic-offset 4)

#include "clsWideCamera.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"

#include "../COMMON/CommonDefinitions.h"
#include "clsFBO.h"

#define CLS WideCamera_cls

CLS::
WideCamera_cls(){
    for (int i = 0; i < _FBO_MAX; i++){
	mpCFBOParts[i] = NULL;
    }
    for (int i = 0; i < 4; i++) mDListFisheye4[i] = 0;
    for (int i = 0; i < 3; i++) mDListFisheye3[i] = 0;
    for (int i = 0; i < 6; i++) mDListER[i] = 0;
    for (int i = 0; i < 6; i++) mDListSpherical[i] = 0;
    for (int i = 0; i < 4; i++) mDListDualFW[i] = 0;
    for (int i = 0; i < 2; i++) mDListDualBW[i] = 0;

    mbERGenerated = FALSE;
    mbDualGenerated = FALSE;
    mbSphericalGenerated = FALSE;
}

CLS::
~WideCamera_cls(){
     for (int i = 0; i < _FBO_MAX; i++){
	 if (mpCFBOParts[i] != NULL) delete mpCFBOParts[i];
	 mpCFBOParts[i] = NULL;
     }
}

int CLS::
GetFBONum(int type){
    switch(type){
    case _WC_FISHEYE4: return 4;
    case _WC_FISHEYE3: return 3;
    case _WC_DUAL_FISHEYE: return 6;
    case _WC_SPHERICAL_FISHEYE: return 6;
    case _WC_EQUIRECTANGULAR: return 6;
    }
    return 1;
}

int CLS::
GetFitIndex(int type, double aspect){
    int fit = _TEX_FIT_HORIZONTAL;
    if (aspect < 1) fit = _TEX_FIT_VERTICAL;
    if (type == _WC_FISHEYE4 ||
	type == _WC_FISHEYE3 ||
	type == _WC_SPHERICAL_FISHEYE){
	fit = _TEX_FIT_VERTICAL;
	if (aspect < 1) fit = _TEX_FIT_HORIZONTAL;
    }
    return fit;
}

void CLS::
DrawDomeTexSegment(double coordu, double coordv0, double coordv1,
		   double x0, double x1, double y0, double y1, 
		   double z0, double z1, double deg){
    
    double r0 = sqrt(x0*x0 + y0*y0 + z0*z0), r1 = sqrt(x1*x1 + y1*y1 + z1*z1);
    double theta0 = asin(z0/r0), theta1 = asin(z1/r1);
    double phi0 = atan2(y0, x0), phi1 = atan2(y1, x1);

    glTexCoord2f(coordu, coordv1);
    glVertex3d((1.570796 - theta1) / 3.1415926 * 180 / deg * cos(phi1) * 2, 
	       (1.570796 - theta1) / 3.1415926 * 180 / deg * sin(phi1) * 2, 0);
    glTexCoord2f(coordu, coordv0);
    glVertex3d((1.570796 - theta0) / 3.1415926 * 180 / deg * cos(phi0) * 2, 
	       (1.570796 - theta0) / 3.1415926 * 180 / deg * sin(phi0) * 2, 0);
}

void CLS::
DrawDomeTexSegmentB(double coordu0, double coordu1,
		    double coordv0, double coordv1,
		   double x0, double x1, double y0, double y1, 
		   double z0, double z1, double deg){
    double r0 = sqrt(x0*x0 + y0*y0 + z0*z0);
    double r1 = sqrt(x1*x1 + y1*y1 + z1*z1);
    double theta0 = asin(z0/r0);
    double theta1 = asin(z1/r1);
    double phi0 = atan2(y0, x0);
    double phi1 = atan2(y1, x1);

    if (x0 == 0 && y0 == 0) phi0 = phi1;

    glTexCoord2f(coordu1, coordv1);
    glVertex3d((1.570796 - theta1) / 3.1415926 * 180 / deg * cos(phi1) * 2, 
	       (1.570796 - theta1) / 3.1415926 * 180 / deg * sin(phi1) * 2, 0);
    glTexCoord2f(coordu0, coordv0);
    glVertex3d((1.570796 - theta0) / 3.1415926 * 180 / deg * cos(phi0) * 2, 
	       (1.570796 - theta0) / 3.1415926 * 180 / deg * sin(phi0) * 2, 0);
}

void CLS::
AdjustUV(double &coordu, double &coordv0, double &coordv1,
	 double coordu_, double coordv0_, double &coordv1_){
    float factor = 
	tan(45*3.14159/180) / tan((45+_FISHEYE_BORDER_DEG)*3.14159/180);
    coordu = (coordu_ * 2 - 1) * factor * 0.5 + 0.5;
    coordv0 = (coordv0_ * 2 - 1) * factor * 0.5 + 0.5;
    coordv1 = (coordv1_ * 2 - 1) * factor * 0.5 + 0.5;
}

void CLS::
AdjustUVB(double &coordu0, double &coordu1, double &coordv0, double &coordv1,
	  double coordu0_, double coordu1_, double coordv0_, double &coordv1_){
    float factor = 
	tan(45*3.14159/180) / tan((45+_FISHEYE_BORDER_DEG)*3.14159/180);
    coordu0 = (coordu0_ * 2 - 1) * factor * 0.5 + 0.5;
    coordu1 = (coordu1_ * 2 - 1) * factor * 0.5 + 0.5;
    coordv0 = (coordv0_ * 2 - 1) * factor * 0.5 + 0.5;
    coordv1 = (coordv1_ * 2 - 1) * factor * 0.5 + 0.5;
}

void CLS::
MakeListFisheye4(double deg){
    
    for (int i = 0; i < 4; i++) 
	if (mDListFisheye4[i] != 0)
	    glDeleteLists(mDListFisheye4[i], 1);

    for (int i = 0; i < 4; i++)
	mDListFisheye4[i] = glGenLists(1);
    
    double sqrt2 = sqrt(2.0);

    glNewList(mDListFisheye4[0], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x  = (coordu_ - 1) * sqrt2; 
	    double y0 = coordv0_ * 2 - 1;
	    double y1 = coordv1_ * 2 - 1;
	    double z =  coordu_ * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    x += 0.001;
	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x, x, y0, y1, z, z, deg);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListFisheye4[1], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x  = coordu_ * sqrt2; 
	    double y0 = coordv0_ * 2 - 1;
	    double y1 = coordv1_ * 2 - 1;
	    double z = (1.0 - coordu_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x, x, y0, y1, z, z, deg);
	}
	glEnd();
    }
    glEndList();
    
    glNewList(mDListFisheye4[2], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x0 = (coordu_ + coordv0_ - 1) * sqrt2; 
	    double x1 = (coordu_ + coordv1_ - 1) * sqrt2; 
	    double y = -1;
	    double z0 = (-coordu_ + coordv0_) * sqrt2;
	    double z1 = (-coordu_ + coordv1_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x0, x1, y, y, z0, z1, deg);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListFisheye4[3], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x0 = (coordu_ + coordv0_ - 1) * sqrt2; 
	    double x1 = (coordu_ + coordv1_ - 1) * sqrt2; 
	    double y = 1;
	    double z0 = (coordu_ - coordv0_) * sqrt2;
	    double z1 = (coordu_ - coordv1_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x0, x1, y, y, z0, z1, deg);
	}
	glEnd();
    }
    glEndList();
}

void CLS::
MakeListFisheye3(double deg){
    
    for (int i = 0; i < 3; i++) 
	if (mDListFisheye3[i] != 0)
	    glDeleteLists(mDListFisheye3[i], 1);

    for (int i = 0; i < 3; i++)
	mDListFisheye3[i] = glGenLists(1);
    
    double sqrt2 = sqrt(2.0);

    glNewList(mDListFisheye3[0], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0 = (double)v / _FISHEYE_MAX;
	double coordv1 = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu = (double)u / _FISHEYE_MAX;
	    double x00 = -1,             x01 = -1;
	    double y00 = coordv0 * 3 -1, y01 = coordv1 * 3 - 1;
	    double z00 = coordu * 3 -2,  z01 = coordu * 3 - 2; 

	    double pos0[3] = {x00, y00, z00};
	    double pos1[3] = {x01, y01, z01};
	    
	    Rotate(pos0, +45, 1);
	    Rotate(pos1, +45, 1);
	    Rotate(pos0, -35.2643, 0);
	    Rotate(pos1, -35.2643, 0);
	    
	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       pos0[0], pos1[0],
			       pos0[1], pos1[1],
			       pos0[2], pos1[2], deg);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListFisheye3[1], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0 = (double)v / _FISHEYE_MAX;
	double coordv1 = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu = (double)u / _FISHEYE_MAX;
	    double x00 = coordu * 3 -1, x01 = coordu * 3 - 1;
	    double y00 = coordv0 * 3 -1, y01 = coordv1 * 3 - 1;
	    double z00 = 1,             z01 = 1; 

	    double pos0[3] = {x00, y00, z00};
	    double pos1[3] = {x01, y01, z01};
	    Rotate(pos0, +45, 1);
	    Rotate(pos1, +45, 1);
	    Rotate(pos0, -35.2643, 0);
	    Rotate(pos1, -35.2643, 0);
	    
	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       pos0[0], pos1[0],
			       pos0[1], pos1[1],
			       pos0[2], pos1[2], deg);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListFisheye3[2], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0 = (double)v / _FISHEYE_MAX;
	double coordv1 = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu = (double)u / _FISHEYE_MAX;
	    double x00 = coordu * 3 -1,  x01 = coordu * 3 -1;
	    double y00 = -1,             y01 = -1;
	    double z00 = coordv0 * 3 -2, z01 = coordv1 * 3 - 2;

	    double pos0[3] = {x00, y00, z00};
	    double pos1[3] = {x01, y01, z01};
	    Rotate(pos0, +45, 1);
	    Rotate(pos1, +45, 1);
	    Rotate(pos0, -35.2643, 0);
	    Rotate(pos1, -35.2643, 0);
	    
	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       pos0[0], pos1[0],
			       pos0[1], pos1[1],
			       pos0[2], pos1[2], deg);
	}
	glEnd();
    }
    glEndList();
}

void CLS::
MakeListsER(){
    for (int i = 0; i < 6; i++)
	mDListER[i] = glGenLists(1);

    double x00, x01, y00, y01, z00, z01;
    
    glNewList(mDListER[0], GL_COMPILE);
    y00 = 0.5;
    y01 = 0.5;
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0 = (double)v / _FISHEYE_MAX;
	double coordv1 = (double)(v+1) / _FISHEYE_MAX;
	double z00 = coordv0 - 0.5;
	double z01 = coordv1 - 0.5;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu = (double)u / _FISHEYE_MAX;
	    double x00 = coordu - 0.5;
	    double x01 = coordu - 0.5;
	    double phi0 = atan2(x00, y00) / 3.14159 ;
	    double phi1 = atan2(x01, y01) / 3.14159 ;
	    double theta0 = atan(z00/sqrt(x00*x00+y00*y00)) / 3.14159 ;
	    double theta1 = atan(z01/sqrt(x01*x01+y01*y01)) / 3.14159 ;
	    glTexCoord2f(coordu, coordv1);
	    glVertex3d(phi1, theta1, 0);
	    
	    glTexCoord2f(coordu, coordv0);
	    glVertex3d(phi0, theta0, 0);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListER[1], GL_COMPILE);
    x00 = -0.5;
    x01 = -0.5;
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0 = (double)v / _FISHEYE_MAX;
	double coordv1 = (double)(v+1) / _FISHEYE_MAX;
	double z00 = coordv0 - 0.5;
	double z01 = coordv1 - 0.5;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu = (double)u / _FISHEYE_MAX;
	    double y00 = coordu - 0.5;
	    double y01 = coordu - 0.5;
	    double phi0 = atan2(x00, y00) / 3.14159 ;
	    double phi1 = atan2(x01, y01) / 3.14159 ;
	    double theta0 = atan(z00/sqrt(x00*x00+y00*y00)) / 3.14159 ;
	    double theta1 = atan(z01/sqrt(x01*x01+y01*y01)) / 3.14159 ;
	    glTexCoord2f(coordu, coordv1);
	    glVertex3d(phi1, theta1, 0);

	    glTexCoord2f(coordu, coordv0);
	    glVertex3d(phi0, theta0,0);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListER[2], GL_COMPILE);
    y00 = -0.5;
    y01 = -0.5;
    for (int u = 0; u < _FISHEYE_MAX; u++){
	double coordu0 = (double)u / _FISHEYE_MAX;
	double coordu1 = (double)(u+1) / _FISHEYE_MAX;
	double x00 = 0.5 - coordu0;
	double x01 = 0.5 - coordu1;
	double phi0 = atan2(x00, y00) / 3.14159 ;
	double phi1 = atan2(x01, y01) / 3.14159 ;
	if (phi0 > phi1) phi0 -=2; 
	
	glBegin(GL_QUAD_STRIP);
	for (int v = 0; v <= _FISHEYE_MAX; v++){
	    double coordv = (double)v / _FISHEYE_MAX;
	    double z00 = coordv - 0.5;
	    double z01 = coordv - 0.5;
	    
	    double theta0 = atan(z00/sqrt(x00*x00+y00*y00)) / 3.14159 ;
	    double theta1 = atan(z01/sqrt(x01*x01+y01*y01)) / 3.14159 ;

	    glTexCoord2f(coordu0, coordv);
	    glVertex3d(phi0, theta0,0);

	    glTexCoord2f(coordu1, coordv);
	    glVertex3d(phi1, theta1, 0);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListER[3], GL_COMPILE);
    x00 = +0.5;
    x01 = +0.5;
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0 = (double)v / _FISHEYE_MAX;
	double coordv1 = (double)(v+1) / _FISHEYE_MAX;
	double z00 = coordv0 - 0.5;
	double z01 = coordv1 - 0.5;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu = (double)u / _FISHEYE_MAX;
	    double y00 = 0.5 - coordu;
	    double y01 = 0.5 - coordu;
	    double phi0 = atan2(x00, y00) / 3.14159;
	    double phi1 = atan2(x01, y01) / 3.14159;
	    double theta0 = atan(z00/sqrt(x00*x00+y00*y00)) / 3.14159 ;
	    double theta1 = atan(z01/sqrt(x01*x01+y01*y01)) / 3.14159 ;
	    glTexCoord2f(coordu, coordv1);
	    glVertex3d(phi1, theta1, 0);

	    glTexCoord2f(coordu, coordv0);
	    glVertex3d(phi0, theta0,0);
	}
	glEnd();
    }
    glEndList();

    double boarder = 0.002; // ㉺̃}[W

    int offset = -_FISHEYE_MAX/8 + _FISHEYE_MAX;
    double squareu[_FISHEYE_MAX+1], squarev[_FISHEYE_MAX+1];
    for (int i = 0; i < _FISHEYE_MAX/4; i++){
	squareu[(i+offset)%_FISHEYE_MAX] = -0.5 + (double)i/_FISHEYE_MAX*4;
	squarev[(i+offset)%_FISHEYE_MAX] = -0.5;
    }
    for (int i = _FISHEYE_MAX/4; i < 2*_FISHEYE_MAX/4; i++){
	squareu[(i+offset)%_FISHEYE_MAX] = +0.5;
	squarev[(i+offset)%_FISHEYE_MAX] = -0.5 + (double)(i-_FISHEYE_MAX/4)/_FISHEYE_MAX*4;
    }
    for (int i = 2*_FISHEYE_MAX/4; i < 3*_FISHEYE_MAX/4; i++){
	squareu[(i+offset)%_FISHEYE_MAX] = 0.5 - (double)(i-2*_FISHEYE_MAX/4)/_FISHEYE_MAX*4;
	squarev[(i+offset)%_FISHEYE_MAX] = +0.5;
    }
    for (int i = 3*_FISHEYE_MAX/4; i < 4*_FISHEYE_MAX/4; i++){
	squareu[(i+offset)%_FISHEYE_MAX] = -0.5;
	squarev[(i+offset)%_FISHEYE_MAX] = +0.5 - (double)(i-3*_FISHEYE_MAX/4)/_FISHEYE_MAX*4;
    }
    squareu[_FISHEYE_MAX] = 0;
    squarev[_FISHEYE_MAX] = -0.5;

    glNewList(mDListER[4], GL_COMPILE);
    z00 = -0.5;
    z01 = -0.5;
    for (int uvthetaIndex = 0; uvthetaIndex < _FISHEYE_MAX; uvthetaIndex++){
	double uvtheta0 = (double)uvthetaIndex / _FISHEYE_MAX;
	double uvtheta1 = (double)(uvthetaIndex+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int uvphiIndex = 0; uvphiIndex <= _FISHEYE_MAX; uvphiIndex++){
	    double coordu0 = squareu[uvphiIndex] * uvtheta0 + 0.5;
	    double coordv0 = squarev[uvphiIndex] * uvtheta0 + 0.5;
	    double coordu1 = squareu[uvphiIndex] * uvtheta1 + 0.5;
	    double coordv1 = squarev[uvphiIndex] * uvtheta1 + 0.5;
	    	    
	    double x00 = squareu[uvphiIndex] * uvtheta0;
	    double x01 = squareu[uvphiIndex] * uvtheta1;
	    double y00 = squarev[uvphiIndex] * uvtheta0;
	    double y01 = squarev[uvphiIndex] * uvtheta1;
	    
	    double phi0 = (uvphiIndex == 0) ? 1 : atan2(x00, y00) / 3.14159;
	    double phi1 = (uvphiIndex == 0) ? 1 : atan2(x01, y01) / 3.14159;

	    if (uvphiIndex == _FISHEYE_MAX){ phi0 = -1; phi1 = -1; }
	    if (uvthetaIndex == 0) phi0 = phi1;
	    
	    double r0 = sqrt(x00*x00+y00*y00);
	    double theta0 = (uvthetaIndex == 0) ? -0.25-boarder : atan(z00/r0) / 3.14159 ;
	    double theta1 = atan(z01/sqrt(x01*x01+y01*y01)) / 3.14159 ;

	    if (theta0 < -0.498) theta0 = -0.5;
	    if (theta1 > +0.498) theta1 = +0.5;

	    glTexCoord2f(coordu0, coordv0);
	    glVertex3d(phi0, theta0, 0);
	    
	    glTexCoord2f(coordu1, coordv1);
	    glVertex3d(phi1, theta1, 0);
    
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListER[5], GL_COMPILE);
    z00 = +0.5;
    z01 = +0.5;
    for (int uvthetaIndex = 0; uvthetaIndex < _FISHEYE_MAX; uvthetaIndex++){
	double uvtheta0 = (double)uvthetaIndex / _FISHEYE_MAX;
	double uvtheta1 = (double)(uvthetaIndex+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int uvphiIndex = 0; uvphiIndex <= _FISHEYE_MAX; uvphiIndex++){
	    double coordu0 = squareu[uvphiIndex] * uvtheta0 + 0.5;
	    double coordv0 = 0.5 - squarev[uvphiIndex] * uvtheta0;
	    double coordu1 = squareu[uvphiIndex] * uvtheta1 + 0.5;
	    double coordv1 = 0.5 - squarev[uvphiIndex] * uvtheta1;
	    	    
	    double x00 = squareu[uvphiIndex] * uvtheta0;
	    double x01 = squareu[uvphiIndex] * uvtheta1;
	    double y00 = squarev[uvphiIndex] * uvtheta0;
	    double y01 = squarev[uvphiIndex] * uvtheta1;

	    double phi0 = (uvphiIndex == 0) ? 1 : atan2(x00, y00) / 3.14159;
	    double phi1 = (uvphiIndex == 0) ? 1 : atan2(x01, y01) / 3.14159;

	    if (uvphiIndex == _FISHEYE_MAX){ phi0 = -1; phi1 = -1; }
	    if (uvthetaIndex == 0) phi0 = phi1;
	    
	    double r0 = sqrt(x00*x00+y00*y00);
	    double theta0 = (uvthetaIndex == 0) ? 0.25+boarder : atan(z00/r0) / 3.14159 ;
	    double theta1 = atan(z01/sqrt(x01*x01+y01*y01)) / 3.14159 ;

	    if (theta0 < -0.498) theta0 = -0.5;
	    if (theta1 > +0.498) theta1 = +0.5;

	    glTexCoord2f(coordu1, coordv1);
	    glVertex3d(phi1, theta1, 0);

	    glTexCoord2f(coordu0, coordv0);
	    glVertex3d(phi0, theta0,0);
	    
	}
	glEnd();
    }
    glEndList();

    mbERGenerated = TRUE;
}

void CLS::
MakeListsDual(){
        
    for (int i = 0; i < 4; i++) mDListDualFW[i] = glGenLists(1);
    for (int i = 0; i < 2; i++) mDListDualBW[i] = glGenLists(1);

    double sqrt2 = sqrt(2.0);
    
    glNewList(mDListDualFW[0], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = -2; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x  = (coordu_ - 1) * sqrt2; 
	    double y0 = coordv0_ * 2 - 1;
	    double y1 = coordv1_ * 2 - 1;
	    double z =  coordu_ * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    x += 0.001;
	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x, x, y0, y1, z, z, 180);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListDualFW[1], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX + 2; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x  = coordu_ * sqrt2; 
	    double y0 = coordv0_ * 2 - 1;
	    double y1 = coordv1_ * 2 - 1;
	    double z = (1.0 - coordu_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x, x, y0, y1, z, z, 180);
	}
	glEnd();
    }
    glEndList();
    
    glNewList(mDListDualFW[2], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x0 = (coordu_ + coordv0_ - 1) * sqrt2; 
	    double x1 = (coordu_ + coordv1_ - 1) * sqrt2; 
	    double y = -1;
	    double z0 = (-coordu_ + coordv0_) * sqrt2;
	    double z1 = (-coordu_ + coordv1_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x0, x1, y, y, z0, z1, 180);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListDualFW[3], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x0 = (coordu_ + coordv0_ - 1) * sqrt2; 
	    double x1 = (coordu_ + coordv1_ - 1) * sqrt2; 
	    double y = 1;
	    double z0 = (coordu_ - coordv0_) * sqrt2;
	    double z1 = (coordu_ - coordv1_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x0, x1, y, y, z0, z1, 180);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListDualBW[0], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x0 = -(coordu_ + coordv0_ - 1) * sqrt2; 
	    double x1 = -(coordu_ + coordv1_ - 1) * sqrt2; 
	    double y = -1;
	    double z0 = -(-coordu_ + coordv0_) * sqrt2;
	    double z1 = -(-coordu_ + coordv1_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x0, x1, y, y, z0, z1, 180);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListDualBW[1], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x0 = -(coordu_ + coordv0_ - 1) * sqrt2; 
	    double x1 = -(coordu_ + coordv1_ - 1) * sqrt2; 
	    double y = 1;
	    double z0 = -(coordu_ - coordv0_) * sqrt2;
	    double z1 = -(coordu_ - coordv1_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x0, x1, y, y, z0, z1, 180);
	}
	glEnd();
    }
    glEndList();

    mbDualGenerated = TRUE;
}

void CLS::
MakeListsSpherical(){
        
    for (int i = 0; i < 6; i++) mDListSpherical[i] = glGenLists(1);
    double sqrt2 = sqrt(2.0);
    
    glNewList(mDListSpherical[0], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x  = (coordu_ - 1) * sqrt2; 
	    double y0 = coordv0_ * 2 - 1;
	    double y1 = coordv1_ * 2 - 1;
	    double z =  coordu_ * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    x += 0.001;
	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x, x, y0, y1, z, z, 360);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListSpherical[1], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x  = coordu_ * sqrt2; 
	    double y0 = coordv0_ * 2 - 1;
	    double y1 = coordv1_ * 2 - 1;
	    double z = (1.0 - coordu_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x, x, y0, y1, z, z, 360);
	}
	glEnd();
    }
    glEndList();
    
    glNewList(mDListSpherical[2], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x0 = (coordu_ + coordv0_ - 1) * sqrt2; 
	    double x1 = (coordu_ + coordv1_ - 1) * sqrt2; 
	    double y = -1;
	    double z0 = (-coordu_ + coordv0_) * sqrt2;
	    double z1 = (-coordu_ + coordv1_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x0, x1, y, y, z0, z1, 360);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListSpherical[3], GL_COMPILE);
    for (int v = 0; v < _FISHEYE_MAX; v++){
	double coordv0_ = (double)v / _FISHEYE_MAX;
	double coordv1_ = (double)(v+1) / _FISHEYE_MAX;
	glBegin(GL_QUAD_STRIP);
	for (int u = 0; u <= _FISHEYE_MAX; u++){
	    double coordu_ = (double)u / _FISHEYE_MAX;
	    double coordu, coordv0, coordv1;
	    
	    double x0 = (coordu_ + coordv0_ - 1) * sqrt2; 
	    double x1 = (coordu_ + coordv1_ - 1) * sqrt2; 
	    double y = 1;
	    double z0 = (coordu_ - coordv0_) * sqrt2;
	    double z1 = (coordu_ - coordv1_) * sqrt2;

	    AdjustUV(coordu, coordv0, coordv1,
		     coordu_, coordv0_, coordv1_);

	    DrawDomeTexSegment(coordu, coordv0, coordv1,
			       x0, x1, y, y, z0, z1, 360);
	}
	glEnd();
    }
    glEndList();

    const int max = _FISHEYE_MAX/4 * 3;
    double squareu[max+1], squarev[max+1];
    for (int i = 0; i < max/3; i++){
	squareu[i] = -0.5 + (double)i/max*3;
	squarev[i] = -0.5;
    }
    for (int i = max/3; i < 2*max/3; i++){
	squareu[i] = 0.5;
	squarev[i] = -0.5 + (double)(i-max/3)/max*3;
    }
    for (int i = 2*max/3; i <= max; i++){
	squareu[i] = 0.5 - (double)(i-2*max/3)/max*3;
	squarev[i] = 0.5;
    }

    glNewList(mDListSpherical[4], GL_COMPILE);
    for (int uvthetaIndex = 0; uvthetaIndex < max; uvthetaIndex++){
	double uvtheta0 = (double)uvthetaIndex / max;
	double uvtheta1 = (double)(uvthetaIndex+1) / max;
	glBegin(GL_QUAD_STRIP);
	for (int uvphiIndex = 0; uvphiIndex <= max; uvphiIndex++){
	    double coordu0_ = -(squareu[uvphiIndex]+0.5)*uvtheta0 + 1;
	    double coordv0_ = squarev[uvphiIndex] * uvtheta0 + 0.5;
	    double coordu1_ = -(squareu[uvphiIndex]+0.5)*uvtheta1 + 1;
	    double coordv1_ = squarev[uvphiIndex] * uvtheta1 + 0.5;
	    
	    double x0 = -(coordu0_ - 1) * sqrt2; 
	    double x1 = -(coordu1_ - 1) * sqrt2;
	    double y0 = (coordv0_ * 2 - 1);
	    double y1 = (coordv1_ * 2 - 1);
	    double z0 =  -coordu0_ * sqrt2;
	    double z1 =  -coordu1_ * sqrt2;

	    double coordu0, coordu1, coordv0, coordv1;
	    AdjustUVB(coordu0, coordu1, coordv0, coordv1,
		      coordu0_, coordu1_, coordv0_, coordv1_);
	    
	    DrawDomeTexSegmentB(coordu0, coordu1, coordv0, coordv1,
				x0, x1, y0, y1, z0, z1, 360);
	}
	glEnd();
    }
    glEndList();

    glNewList(mDListSpherical[5], GL_COMPILE);
    for (int uvthetaIndex = 0; uvthetaIndex < max; uvthetaIndex++){
	double uvtheta0 = (double)uvthetaIndex / max;
	double uvtheta1 = (double)(uvthetaIndex+1) / max;
	glBegin(GL_QUAD_STRIP);
	for (int uvphiIndex = 0; uvphiIndex <= max; uvphiIndex++){
	    double coordu0_ = (squareu[uvphiIndex]+0.5)*uvtheta0;// + 0.5;
	    double coordv0_ = squarev[uvphiIndex] * uvtheta0 + 0.5;
	    double coordu1_ = (squareu[uvphiIndex]+0.5)*uvtheta1;// + 0.5;
	    double coordv1_ = squarev[uvphiIndex] * uvtheta1 + 0.5;
	    
	    double x0  = -coordu0_ * sqrt2; 
	    double x1  = -coordu1_ * sqrt2; 
	    double y0 = coordv0_ * 2 - 1;
	    double y1 = coordv1_ * 2 - 1;
	    double z0 = -(1.0 - coordu0_) * sqrt2;
	    double z1 = -(1.0 - coordu1_) * sqrt2;
	    
	    double coordu0, coordu1, coordv0, coordv1;
	    AdjustUVB(coordu0, coordu1, coordv0, coordv1,
		      coordu0_, coordu1_, coordv0_, coordv1_);

	    DrawDomeTexSegmentB(coordu1, coordu0, coordv1, coordv0,
				x1, x0, y1, y0, z1, z0, 360);
	}
	glEnd();
    }
    glEndList();

    mbSphericalGenerated = TRUE;
}

double CLS::
GetDegreeV(int typeWideCamera, double degreeL){
    switch(typeWideCamera){
    case _WC_PERSPECTIVE:
	if (degreeL > 120) return 120;
	return degreeL;
    case _WC_FISHEYE4:
    case _WC_DUAL_FISHEYE:
    case _WC_SPHERICAL_FISHEYE:
	return 90 + _FISHEYE_BORDER_DEG * 2;
    case _WC_FISHEYE3:
	return 63.4349 * 2;
    case _WC_EQUIRECTANGULAR:
	return 90;
    }
    return degreeL;
}

FBO_cls *CLS::
GetPCFBO(int type, int fboIndex, FBO_cls *pCFBORenderOut, bool bUseEnsemble){
    //if (bUseEnsemble == TRUE) return mpCFBOEnsemble;
    //return pCFBORenderOut;
    
    if (type == _WC_PERSPECTIVE ||
	type == _WC_ORTHOGONAL) return pCFBORenderOut;
    
    return mpCFBOParts[fboIndex];
}

void CLS::
ModifyViewportIfNeeded(int type, int fboIndex){
    if (type != _WC_FISHEYE3) return;
    int block = mPartSize / 3;
    switch (fboIndex){
    case 0:
	glViewport(0, -block, block*4, block*4);
	break;
    case 1:
	glViewport(-block, -block, block*4, block*4);
	break;
    case 2:
	glViewport(-block, 0, block*4, block*4);
	break;
    }
    //glMatrixMode(GL_PROJECTION);
    //glLoadIdentity();
    //glMatrixMode(GL_MODELVIEW);
}

void CLS::
MakeStencilBegin(){
    //ꂩ`悷̂̃XeVlɂׂĂP^O    
    glStencilFunc( GL_ALWAYS, 1, ~0); 
    glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
    //XeVȊOɂ͏܂Ȃ
    glColorMask(1,0,0,0);
    glDepthMask(0);
}

void CLS::
MakeStencilEnd(){
    //̉
    glColorMask(1,1,1,1);
    glDepthMask(1);
    glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
}

void CLS::
MakeCirclePatternOnStencil(){
    MakeStencilBegin();
    //glColor3f(1.0f, 1.0f, 0);
    double pixelSize = 1.0/mBufferSize4Gather;
    double radius = 1.0 + pixelSize*4;

    glDisable(GL_DEPTH_TEST);
    glBegin(GL_TRIANGLE_FAN);
    glVertex3d(0, 0, 0);
    for (int i = 0; i <= 72; i++){
	double rad = (double)i / 72 * 2 * 3.1416;
	glVertex3d(radius * cos(rad), radius * sin(rad), 0);
    }
    glEnd();
    MakeStencilEnd();
}

void CLS::
GatherPart(FBO_cls *pCFBOToBake, int type, double degree){
    
    pCFBOToBake->Begin();

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glPushMatrix();

    int width = pCFBOToBake->mWidthFBO, height = pCFBOToBake->mHeightFBO;
    mBufferSize4Gather = height;
    
    switch(type){
    case _WC_FISHEYE4:
	GatherFishEye(degree, FALSE, width, height, 0, FALSE, 4);
	break;
    case _WC_FISHEYE3:
	GatherFishEye(degree, FALSE, width, height, 0, FALSE, 3);
	break;
    case _WC_EQUIRECTANGULAR:
	GatherEquirectangular(FALSE, width, height, 0, FALSE);
	break;
    case _WC_DUAL_FISHEYE:
	GatherDualFishEye(FALSE, width, height, 0, FALSE);
	break;
    case _WC_SPHERICAL_FISHEYE:
	GatherSphericalFishEye(FALSE, width, height, 0, FALSE);
	break;
    }

    glPopMatrix();
    glPopAttrib();

    pCFBOToBake->End();
}

void CLS::
GatherGLSetting(bool bAdd, int tgtWidth, int tgtHeight){
    //glClearColor(0.5, 0.5, 0, 1);
    ColorProvider_cls::SetGLClearColor("MaskBG");
    if (bAdd == FALSE)
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|
		GL_ACCUM_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
    else
	glClear(GL_DEPTH_BUFFER_BIT|
		GL_ACCUM_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
    
    glLoadIdentity();
    GLPerspective_cls::SetProjection(tgtWidth, tgtHeight);
    GLPerspective_cls::SetFrustum(tgtWidth, tgtHeight, 90, 0.1, 10);
    //GLPerspective_cls CGLPerspective;
    //CGLPerspective.SetWindowSize(clipWidth, clipHeight);
    //CGLPerspective.SetClipSize(clipWidth, clipHeight);
    Position_cls CPos(0, 0, -1);
    Direction_cls CDir(0, 0, 1);
    GLPerspective_cls::SetModelView(CPos, CDir);

    //CGLPerspective.mbPerspective = FALSE;
    //CGLPerspective.SetViewportOnGL();
    //CGLPerspective.SetPerspectiveOnGL(&CPos);
    //CGLPerspective.SetModelOrProjectionViewOnGL(&CPos, GL_MODELVIEW);
   
    glEnable(GL_TEXTURE_2D);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
}

void CLS::
GatherFishEye(double deg, bool bAdd, int clipWidth, int clipHeight, 
	      int stereo, bool bUseDepthTexture, int fboNum){
    
    if (deg != mDeg){
	MakeListFisheye4(deg);
	MakeListFisheye3(deg);
    }
    mDeg = deg;
    
    GatherGLSetting(bAdd, clipWidth, clipHeight);

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_STENCIL_TEST);
    MakeCirclePatternOnStencil();
    glStencilFunc( GL_EQUAL, 1, ~0); //1̃sNZɕ`
    glDisable(GL_POLYGON_SMOOTH);
    
    if (bAdd == FALSE){
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	GetPCColorProvider()->SetGLColor("MaskBG", TRUE);
	//glColor4f(0, 0, 0, 1);
	glBegin(GL_QUADS);
	glVertex3d(-10, -10, 0); glVertex3d(+10, -10, 0);
	glVertex3d(+10, +10, 0); glVertex3d(-10, +10, 0);
	glEnd(); // ͈͓œhԂ
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
    }
    
    if (bAdd == TRUE)
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    else
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    //GetPCStereogram()->ColorMaskBegin(stereo);

    //GLAttribute_cls::mspCShaderCtrl->BeginBaseColorShader();
    //GLAttribute_cls::mspCShaderCtrl->SetTextureState(TRUE);
    
    float baseColor[4] = {1, 1, 1, 1};
    glMaterialfv(GL_FRONT, GL_DIFFUSE, baseColor);
    glColor4f(1, 1, 1, 1);

    for (int i = 0; i < fboNum; i++){
	
	if (bUseDepthTexture == TRUE && mpCFBOParts[i]->mpCDepthTexture != NULL)
	    mpCFBOParts[i]->mpCDepthTexture->BindTexture();
	else
	    mpCFBOParts[i]->BindTexture();
	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		
	if (fboNum == 3)
	    glCallList(mDListFisheye3[i]);
	if (fboNum == 4)
	    glCallList(mDListFisheye4[i]);
    }

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    
    glDisable(GL_STENCIL_TEST);
}

void CLS::
GatherEquirectangular(bool bAdd, int clipWidth, int clipHeight, 
		      int stereo, bool bUseDepthTexture){
    
    if (mbERGenerated == FALSE) MakeListsER();

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glPushMatrix();
    GatherGLSetting(bAdd, clipWidth, clipHeight);

    glDisable(GL_DEPTH_TEST);    
    //glEnable(GL_STENCIL_TEST);
    //MakeCirclePatternOnStencil();
    //glStencilFunc( GL_EQUAL, 1, ~0); //1̃sNZɕ`
    glDisable(GL_POLYGON_SMOOTH);

    glDisable(GL_DEPTH_TEST);
    if (bAdd == FALSE){
	glDisable(GL_BLEND);
	GetPCColorProvider()->SetGLColor("MaskBG", TRUE);
	glBegin(GL_QUADS);
	glVertex3d(-10, -10, 0); glVertex3d(+10, -10, 0);
	glVertex3d(+10, +10, 0); glVertex3d(-10, +10, 0);
	glEnd(); // ͈͓œhԂ
	glEnable(GL_BLEND);
    }
    if (bAdd == TRUE){
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    }

    //GetPCStereogram()->ColorMaskBegin(stereo);

    //GLAttribute_cls::mspCShaderCtrl->BeginBaseColorShader();
    //GLAttribute_cls::mspCShaderCtrl->SetTextureState(TRUE);

    float baseColor[4] = {1, 1, 1, 1};
    glMaterialfv(GL_FRONT, GL_DIFFUSE, baseColor);

    glColor4f(1, 1, 1, 1);
    for (int i = 0; i < 6; i++){
	if (bUseDepthTexture == TRUE && mpCFBOParts[i]->mpCDepthTexture != NULL)
	    mpCFBOParts[i]->mpCDepthTexture->BindTexture();
	else
	    mpCFBOParts[i]->BindTexture();
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glCallList(mDListER[i]);
    }

    //GLAttribute_cls::mspCShaderCtrl->EndShader(0);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    
    //GetPCStereogram()->ColorMaskEnd();
    //glDisable(GL_STENCIL_TEST);

    glPopMatrix();
    glPopAttrib();
}

void CLS::
GatherDualFishEye(bool bAdd, int clipWidth, int clipHeight, 
		  int stereo, bool bUseDepthTexture){

    if (mbDualGenerated == FALSE) MakeListsDual();

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glPushMatrix();
    
    GatherGLSetting(bAdd, clipWidth, clipHeight);

    glDisable(GL_DEPTH_TEST);    
    glEnable(GL_STENCIL_TEST);
    glPushMatrix();
    glScalef(0.5, 0.5, 1);
    glTranslatef(-1, 0, 0);
    MakeCirclePatternOnStencil();
    glTranslatef(+2.0, 0, 0);
    MakeCirclePatternOnStencil();
    glPopMatrix();

    glStencilFunc( GL_EQUAL, 1, ~0); //1̃sNZɕ`
    glDisable(GL_POLYGON_SMOOTH);
    
    if (bAdd == FALSE){
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	GetPCColorProvider()->SetGLColor("MaskBG", TRUE);
	glBegin(GL_QUADS);
	glVertex3d(-10, -10, 0); glVertex3d(+10, -10, 0);
	glVertex3d(+10, +10, 0); glVertex3d(-10, +10, 0);
	glEnd(); // ͈͓œhԂ
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
    }
    
    if (bAdd == TRUE)
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    else
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    //GetPCStereogram()->ColorMaskBegin(stereo);

    int fboMax = 6;
    
    //GLAttribute_cls::mspCShaderCtrl->BeginBaseColorShader();
    //GLAttribute_cls::mspCShaderCtrl->SetTextureState(TRUE);

    float baseColor[4] = {1, 1, 1, 1};
    glMaterialfv(GL_FRONT, GL_DIFFUSE, baseColor);

    glColor4f(1, 1, 1, 1);
    for (int i = 0; i < fboMax; i++){
	
	if (bUseDepthTexture == TRUE && mpCFBOParts[i]->mpCDepthTexture != NULL)
	    mpCFBOParts[i]->mpCDepthTexture->BindTexture();
	else
	    mpCFBOParts[i]->BindTexture();
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	glPushMatrix();
	switch(i){
	case 0: case 1:
	    glScalef(0.5, 0.5, 0.5);
	    glTranslatef(-1, 0, 0);
	    glCallList(mDListDualFW[i]);
	    break;
	case 2: case 3:
	    glScalef(0.5, 0.5, 0.5);
	    glTranslatef(-1, 0, 0);
	    glCallList(mDListDualFW[i]);
	    glTranslatef(+2, 0 , 0);
	    glCallList(mDListDualBW[i-2]);
	    break;
	case 4: case 5:
	    glScalef(0.5, 0.5, 0.5);
	    glTranslatef(+1, 0, 0);
	    glCallList(mDListDualFW[i-4]);
	    break;
	}
	glPopMatrix();
    }

    //GLAttribute_cls::mspCShaderCtrl->EndShader(0);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    
    //GetPCStereogram()->ColorMaskEnd();
    glDisable(GL_STENCIL_TEST);

    glPopMatrix();
    glPopAttrib();
}

void CLS::
GatherSphericalFishEye(bool bAdd, int clipWidth, int clipHeight, 
		       int stereo, bool bUseDepthTexture){

    if (mbSphericalGenerated == FALSE) MakeListsSpherical();

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glPushMatrix();
    
    GatherGLSetting(bAdd, clipWidth, clipHeight);

    glDisable(GL_DEPTH_TEST);    
    glEnable(GL_STENCIL_TEST);
    MakeCirclePatternOnStencil();
    glStencilFunc( GL_EQUAL, 1, ~0); //1̃sNZɕ`
    glDisable(GL_POLYGON_SMOOTH);
    
    if (bAdd == FALSE){
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	GetPCColorProvider()->SetGLColor("MaskBG", TRUE);
	glBegin(GL_QUADS);
	glVertex3d(-10, -10, 0); glVertex3d(+10, -10, 0);
	glVertex3d(+10, +10, 0); glVertex3d(-10, +10, 0);
	glEnd(); // ͈͓œhԂ
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
    }
    
    if (bAdd == TRUE)
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    else
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    //GetPCStereogram()->ColorMaskBegin(stereo);

    int fboMax = 6;
    
    //GLAttribute_cls::mspCShaderCtrl->BeginBaseColorShader();
    //GLAttribute_cls::mspCShaderCtrl->SetTextureState(TRUE);

    float baseColor[4] = {1, 1, 1, 1};
    glMaterialfv(GL_FRONT, GL_DIFFUSE, baseColor);

    glColor4f(1, 1, 1, 1);
    for (int i = 0; i < fboMax; i++){
	
	if (bUseDepthTexture == TRUE && mpCFBOParts[i]->mpCDepthTexture != NULL)
	    mpCFBOParts[i]->mpCDepthTexture->BindTexture();
	else
	    mpCFBOParts[i]->BindTexture();
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	glCallList(mDListSpherical[i]);
    }

    //GLAttribute_cls::mspCShaderCtrl->EndShader(0);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    
    //GetPCStereogram()->ColorMaskEnd();
    glDisable(GL_STENCIL_TEST);

    glPopMatrix();
    glPopAttrib();
}

void CLS::
SetRenderPartParam(int partSize, int depth, int fboMax, bool bUseEnsemble){
    mPartSize = partSize;
    
    for (int i = 0; i < fboMax; i++){ 
	if (mpCFBOParts[i] == NULL){
	    mpCFBOParts[i] = new FBO_cls(mPartSize, mPartSize, NULL);
	    mpCFBOParts[i]->SetColorDepth(depth);
	}
	else{
	    mpCFBOParts[i]->SetGeometryFBO(mPartSize, mPartSize);
	    mpCFBOParts[i]->SetColorDepth(depth);
	}
    }
}

void CLS::
ModifyCamera(PositionAndDirection_cls *pCCam, 
	     int fboIndex, int mode, double tilt){
    pCCam->RotateYPR(0, tilt, 0);

    switch(mode){
    case _WC_FISHEYE4:
    case _WC_DUAL_FISHEYE:
    case _WC_SPHERICAL_FISHEYE:
	ModifyFoFishEye(pCCam, fboIndex); break;
    case _WC_FISHEYE3:
	pCCam->RotateYPR(0, -35.2643, 0);
	ModifyFoFishEye(pCCam, fboIndex); break;
    case _WC_EQUIRECTANGULAR:
	ModifyForEquirectangular(pCCam, fboIndex); break;
    }
}

void CLS::
ModifyFoFishEye(PositionAndDirection_cls *pCCam, int fboIndex){
    switch(fboIndex){
    case 0: pCCam->RotateYPR(45, 0 , 0);  break;
    case 1: pCCam->RotateYPR(-45, 0, 0);  break;
    case 2: pCCam->RotateYPR(-45, +90, 0); break;
    case 3: pCCam->RotateYPR(+45, -90, 0); break;
    case 4: pCCam->RotateYPR(-135, 0, 0); break;
    case 5: pCCam->RotateYPR(+135, 0, 0); break;
    }
}

void CLS::
ModifyForEquirectangular(PositionAndDirection_cls *pCCam, int fisheyeIndex){
    switch(fisheyeIndex){
    case 0: pCCam->RotateYPR(0, 0, 0); break;
    case 1: pCCam->RotateYPR(90, 0, 0); break;
    case 2: pCCam->RotateYPR(180, 0, 0); break;
    case 3: pCCam->RotateYPR(270, 0, 0); break;
    case 4: pCCam->RotateYPR(0, 90, 0); break;
    case 5: pCCam->RotateYPR(0, -90, 0); break;
 
    }
}

WideCamera_cls *GpCWideCamera = NULL;
