// "clsWideCamera.cpp"
// (setq c-basic-offset 4)

#ifndef _WIDECAMERA_H_
#define _WIDECAMERA_H_

#define _WC_PERSPECTIVE 0
#define _WC_ORTHOGONAL 1
#define _WC_FISHEYE4 2
#define _WC_FISHEYE3 3
#define _WC_DUAL_FISHEYE 4
#define _WC_SPHERICAL_FISHEYE 5
#define _WC_EQUIRECTANGULAR 6

#include "clsPosition.h"
#include "clsGLPerspective.h"

#include "../COMMON/clsGetPC.h"
#include "../COMMON/clsVectorMath.h"

class FBO_cls;

#define _FISHEYE_MAX 320 // lpɂĝ 4 ̔{
#define _FISHEYE_BORDER_DEG 2

#define _FBO_MAX 6

class WideCamera_cls : public GetPC_cls, public VectorMath_cls {
public:
    WideCamera_cls();
    ~WideCamera_cls();

    int mDeg;
    int mPartSize, mBufferSize4Gather;
    FBO_cls *mpCFBOParts[_FBO_MAX];
    
    double GetDegreeV(int type, double degreeL);
    FBO_cls *GetPCFBO(int type, int fboIndex, FBO_cls *pCFBORenderOut, bool bUseEnsemble);

    void ModifyViewportIfNeeded(int type, int fboIndex);

    bool mbERGenerated, mbDualGenerated, mbSphericalGenerated;

    int  mDListFisheye4[4], mDListFisheye3[3];
    int  mDListER[6], mDListSpherical[6];
    int  mDListDualFW[4], mDListDualBW[2];

    int GetFBONum(int type);
    int GetFitIndex(int type, double aspect);

    void DrawDomeTexSegment(double coordu, double coordv0, double coordv1,
			    double x0, double x1, double y0, double y1, 
			    double z0, double z1, double deg);
    void DrawDomeTexSegmentB(double coordu0, double coordu1,
			     double coordv0, double coordv1,
			     double x0, double x1, double y0, double y1, 
			     double z0, double z1, double deg);
    void AdjustUV(double &coordu, double &coordv0, double &coordv1,
		  double coordu_, double coordv0_, double &coordv1_);
    void AdjustUVB
	(double &coordu0, double &coordu1, double &coordv0, double &coordv1,
	 double coordu0_, double coordu1_, double coordv0_, double &coordv1_);
    
    void MakeListFisheye4(double deg);
    void MakeListFisheye3(double deg);
    void MakeListsER();
    void MakeListsDual();
    void MakeListsSpherical();
    
    void MakeStencilBegin();
    void MakeStencilEnd();
    void MakeCirclePatternOnStencil();

    void GatherPart(FBO_cls *pCFBOToBake, int type, double degree);
    void GatherGLSetting(bool bAdd, int clipWidth, int clipHeight);
    void GatherFishEye
	(double deg, bool bAdd, int clipWidth, int clipHeight,
	 int stereo, bool bUseDepthTexture, int fboNum);
    void GatherEquirectangular
	(bool bAdd, int clipWidth, int clipHeight,
	 int stereo, bool bUseDepthTexture);
    void GatherDualFishEye
	(bool bAdd, int clipWidth, int clipHeight,
	 int stereo, bool bUseDepthTexture);
    void GatherSphericalFishEye
	(bool bAdd, int clipWidth, int clipHeight,
	 int stereo, bool bUseDepthTexture);
    
    void SetRenderPartParam(int partSize, int depth, int fboMax, 
			    bool bUseEnsemble);

    void ModifyCamera(PositionAndDirection_cls *pCCam, 
		      int fboIndex, int mode, double tilt);
    void ModifyFoFishEye(PositionAndDirection_cls *pCCam, int fboIndex);
    void ModifyForEquirectangular(PositionAndDirection_cls *pCCam, int fboIndex);
};

#endif
