// (setq c-basic-offset 4)
// "clsKeyFrame.h"

#include "clsKeyFrame.h"


#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../WIDGETCTRL/WX/clsIconProvider.h"
#include "../MISC/clsMisc.h"
#include "../TIMELINE/clsTimelineBase.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS KeyFrame_cls

CLS::
KeyFrame_cls(){
    mbSelected = FALSE;
    mbActive = FALSE;
    mbLocked = FALSE;
    mXNormalized = 0;
    mValue = 0;
    mEaseFactor = 1;
    mKeyType = _KEY_LINEAR;
}

CLS::
KeyFrame_cls(wxString& strInit){
    ReadStrLine(strInit);
}

wxIcon* CLS::
GetPCIcon(bool bSelected, bool bLocked){
    int subIndex = 0;
    if (bSelected == TRUE) subIndex = 1;
    if (bLocked == TRUE) subIndex = 2;

    if (mKeyType == _KEY_STEP)
	return GetPCIconProvider()->GetPCIcon("TL_STEP", subIndex);
    if (mKeyType == _KEY_HERMITE_AUTO){
	if (mEaseFactor == 0)
	    return GetPCIconProvider()->GetPCIcon("TL_HEMITE_EASE", subIndex);
	return GetPCIconProvider()->GetPCIcon("TL_HEMITE_AUTO", subIndex);
    }
    return GetPCIconProvider()->GetPCIcon("TL_LINEAR", subIndex);
    
}

bool CLS::
Select(double posX, double resolution, bool bSelect){
    if (posX-resolution < mXNormalized && mXNormalized < posX+resolution){
	mbSelected = bSelect;
	return TRUE;
    }
    return FALSE;
}

bool CLS::
DrawTL(wxDC &dcMem, TLPanelGeom_cls* pCGeom,
       int sashX, int panelWidth, int height, int& posX, int index){
    posX = pCGeom->GetDevCoordFromN(mXNormalized, sashX);
    dcMem.SetPen(wxColor(0, 0 ,0));
    if (pCGeom->IsWithinRange(posX, 4) == TRUE){
	dcMem.DrawIcon(*GetPCIcon(mbSelected, FALSE), posX-4, height/2-4);
	//wxString strIndex; strIndex << index;
	//dcMem.DrawText(strIndex, posX+4, 0);
	return TRUE;
    }
    return FALSE;
}

void CLS::
DrawTLGhost(wxDC &dcMem, TLPanelGeom_cls* pCGeom,
	  int sashX, int panelWidth, int height, double dX){
    double newX = GCClassProvider.GetPCTimeCtrl()->GetQuantized(mXNormalized+dX, _MISC_ROUND_ROUND);
    if (Misc_cls::IsNearlySame(mXNormalized, newX, 1e-6) == TRUE) return;
    int posX = pCGeom->GetDevCoordFromN(newX, sashX);
    dcMem.SetPen(wxColor(0, 0, 0));
    if (pCGeom->IsWithinRange(posX, 4) == TRUE)
	dcMem.DrawIcon(*GetPCIcon(FALSE, FALSE), posX-4, height/2-4);
}

void CLS::
DrawKey(wxDC &dc, int posX, int posY){
    dc.DrawIcon(*GetPCIcon(mbSelected, mbLocked),posX-4, posY-4);
}

void CLS::
DrawKeyGhost(wxDC &dc, int posX, int posY){
    dc.DrawIcon(*GetPCIcon(FALSE, FALSE), posX-4, posY-4);
}

bool CLS::
ReadStrOneLiner(wxString strFirstItem, wxString strLine){
    if (strFirstItem.IsSameAs("SetValue")){
	strLine.ToDouble(&mValue); return TRUE;
    }
    if (strFirstItem.IsSameAs("SetX")){
	strLine.ToDouble(&mXNormalized); return TRUE;
    }
    if (strFirstItem.IsSameAs("SetType")){
	strLine.ToLong(&mKeyType); return TRUE;
    }
    if (strFirstItem.IsSameAs("SetEaseFactor")){
	strLine.ToDouble(&mEaseFactor); return TRUE;
    }
    return FALSE;
}

void CLS::
Serialize(wxString& strOut){
    strOut << mXNormalized << "::" << mbSelected << "::" << mValue << "::" << mKeyType << "::" << mEaseFactor << ENDL;
}

void CLS::
ReadStrLine(wxString strLine){
    PopStrItemFromStrLine(strLine, "::").ToDouble(&mXNormalized);
    long selected; 
    PopStrItemFromStrLine(strLine, "::").ToLong(&selected);
    mbSelected = (selected != 0);
    PopStrItemFromStrLine(strLine, "::").ToDouble(&mValue);
    PopStrItemFromStrLine(strLine, "::").ToLong(&mKeyType);
    PopStrItemFromStrLine(strLine, "::").ToDouble(&mEaseFactor);
}
