// (setq c-basic-offset 4)
// "clsParamGroup.h"

#include "clsParamGroup.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../LUA/clsLuaInterpreter.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../TIMELINE/clsTimelineBase.h"
#include "../UNDO/clsUndoManager.h"

#include "../../clsFrameMain.h"

#include <wx/dcbuffer.h>
#include <wx/dcmemory.h>

#include <sstream>

#include "../COMMON/CommonDefinitions.h"

#include "../../wxWidgetEnumMenu.h"

#define CLS ParamGroup_cls

CLS::
ParamGroup_cls(wxString strName, wxArrayString AstrItemName){
    
    mCConfList.Add(&mCstrName, "Name=", strName);
    mCConfList.Add(&mCbLinked, "Linked=", "1");
    mCConfList.Add(&mCbShy, "Shy=", "0");
    mCConfList.Add(&mCbExpression, "UseExpr=", "0");
    mCConfList.Add(&mCstrExpression, "Expression=", "return ", "TEXT_EDITOR");     
    int num = AstrItemName.GetCount();
    for (int i = 0; i < num; i++){
	mVCParam.push_back(Parameter_cls(this, AstrItemName[i], i));
	mCstrExpression.mstrBody << AstrItemName[i];
	if (i != num - 1) mCstrExpression.mstrBody << ",";
    }

    mCColorBase = wxColor(204, 204, 196);
    mCColorWA = wxColor(248, 248, 240);

    mCConfList.SetTag(strName);
    mbForColorCurve = FALSE;
    mReferenceCounter = 0;
}

CLS::
~ParamGroup_cls(){
}

Parameter_cls* CLS::
GetPCParameter(wxString strTag){
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++){
	if (mVCParam[i].mCstrName.GetBodySTR().IsSameAs(strTag))
	    return &mVCParam[i];
    }
    return NULL;
}

wxString CLS::
GetStrProcCounter(){
    wxString str;
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++) str << mVCParam[i].mProcCounter;
    return str;
}

wxString CLS::
GetStrMemberListCSV(){
    wxString strCSV;
    for (int i = 0; i < mVCParam.size(); i++){
	if (i != 0) strCSV << wxT(",");
	strCSV << mVCParam[i].mCstrName.GetBodySTR();
    }
    return strCSV;
}

wxString CLS::
GetStrExpressionMessage(){
    wxString str;
    str << "nt, seq, sub, frame = Normalized Time, Index, Subindex, Scene Frame" << ENDL;
    str << GetStrMemberListCSV() << " = members"; 
    return str;
}

void CLS::
CheckTag(wxString &strTag){
    wxString strLastItem = GetLastItem(strTag, "::");
    if (strLastItem.IsSameAs(GetStrName()) == FALSE)
	strTag << "::" << GetStrName();
}

void CLS::
PrecalcBegin(){
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++) mVCParam[i].PrecalcBegin();
}

void CLS::
PrecalcEnd(){
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++) mVCParam[i].PrecalcEnd();
}

void CLS::
OpenLua(ParamGroup_cls *pCInit){
    if (mpLua != NULL) CloseLua();
    
    wxString strFunction;
    
    strFunction << wxT("function script(nt, seq, sub, frame, ") << GetStrMemberListCSV() << wxT(")") << ENDL;
    strFunction << mCstrExpression.mstrBody << ENDL;
    strFunction << wxT("end") << ENDL;

    mstrExpressionDefined = mCstrExpression.mstrBody;
    
    luaL_Reg *LuaTables[2] = {
	GCClassProvider.GetPCLuaInterpreter()->mVLibTableData.data(),
	GCClassProvider.GetPCLuaInterpreter()->mVLibTableThis.data(),
    };
    
    wxArrayString AstrLibraryName;
    AstrLibraryName.push_back(wxT("Data"));
    AstrLibraryName.push_back(wxT("This"));

    OpenLuaBase(strFunction, 2, AstrLibraryName, LuaTables);

    lua_pushlightuserdata(mpLua, this);
    lua_setglobal(mpLua, "ParamGroup");
}

void CLS::
CloseLua(){
    mReferenceCounter = 0;
    CloseLuaBase();
}

std::vector < double > CLS::
CallLuaExpression(std::vector < double > VOrg, double NT){
    std::vector < double > VValue;
    if (mpLua == NULL) return VOrg;

    lua_getglobal(mpLua, "ParamGroupChain");
    
    mReferenceCounter++;

    if (mReferenceCounter > 1){

	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuLog);
	wxString strMessage = "Reference Loop";
	strMessage << "::" << GetStrName();
	
	com.SetString(strMessage);
	GetPCFrameMain(0)->GetEventHandler()->AddPendingEvent(com);
    
	mReferenceCounter--;
	return VValue;
    }

    //double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frame, subframe);
    int step = GetPCTimeCtrl()->GetCurrentStep();

    lua_getglobal(mpLua, "ParamGroupInit");
    if (lua_touserdata(mpLua, -1) == NULL){
	lua_pushlightuserdata(mpLua, this);
	lua_setglobal(mpLua, "ParamGroupInit");
    }

    int num0 = lua_gettop(mpLua);
    lua_getglobal(mpLua, "script");
    lua_pushnumber(mpLua, NT);
    lua_pushnumber(mpLua, frame);
    lua_pushnumber(mpLua, subframe);
    lua_pushnumber(mpLua, step);
    int num1 = lua_gettop(mpLua);

    int dimension = (int)VOrg.size();
    for (int i = 0; i < dimension; i++) lua_pushnumber(mpLua, VOrg[i]);

    if (lua_pcall(mpLua, 4+dimension, dimension, 0) != 0){
	int num2 = lua_gettop(mpLua);
	CheckErrorMessage(mpLua, mCstrName.GetBodySTR());
	CloseLua();
	
	return VOrg;
    }

    for (int i = 0; i < dimension; i++)
	VValue.push_back(lua_tonumber(mpLua, -dimension+i));
    lua_pop(mpLua, dimension);

    mReferenceCounter--;
    return VValue;
}

std::vector < double > CLS::
GetVExpression(double Xnormalized, std::vector < double > VOrg, ParamGroup_cls *pCInit, bool bFromExpression){

    if (mstrExpressionDefined.IsSameAs(mCstrExpression.GetBodySTR()) == FALSE)
	OpenLua(pCInit);

    std::vector < double > VValue =  CallLuaExpression(VOrg, Xnormalized/*, pCInit*/);

    return VValue;
}

std::vector < double > CLS::
GetVInterpolatedValue(double Xnormalized, bool bUseExpression,
		      ParamGroup_cls *pCInit, bool bFromExpression){
    

    std::vector < double > VValue;
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++)
	VValue.push_back(mVCParam[i].GetValue(Xnormalized));

    if (bUseExpression == TRUE && mCbExpression.GetBodyBOOL() == TRUE){
	std::vector < double > VExp =  GetVExpression(Xnormalized, VValue, pCInit, bFromExpression);
	return VExp;
    }

    return VValue;
}

std::vector < float > CLS::
GetVInterpolatedValue(float Xnormalized, bool bUseExpression, 
		      ParamGroup_cls *pCInit, bool bFromExpression){
    
    std::vector < double > VValueDouble = 
	GetVInterpolatedValue((double)Xnormalized, bUseExpression, pCInit, bFromExpression);

    std::vector < float > VValue;
    for (int i = 0; i < VValueDouble.size(); i++)
	VValue.push_back((float)VValueDouble[i]);

    return VValue;
}

bool CLS::
IsTimeDepend(){
    if (mCbExpression.GetBodyBOOL() == TRUE) return TRUE;
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++){
	if (mVCParam[i].mCbTimeDepend.GetBodyBOOL() == TRUE) return TRUE;
    }
    return FALSE;
}

bool CLS::
IsSelectedKeyExist(){
    bool bExist = FALSE;
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++)
	bExist |= mVCParam[i].IsSelectedKeyExist();
    return bExist;
}

bool CLS::
IsKeyExistAll(double Xnormalized){
    bool bExist = TRUE;
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++)
	bExist &= mVCParam[i].IsKeyExist(Xnormalized);
    return bExist;
}

wxString CLS::
GetStrValues(double Xnormalized, bool bUseExpression){
    std::vector < double > VValue = 
	GetVInterpolatedValue(Xnormalized, bUseExpression);
    wxString strValue;
    uint num = VValue.size();
    for (uint i = 0; i < num; i++){
	if (i != 0) strValue << " ";
	strValue << VValue[i];
    }
    return strValue;
}

void CLS::
SetInitValue(double v0, double v1, double v2, double v3){
    uint numParam = mVCParam.size();
    if (numParam >= 1) mVCParam[0].mVCKeys[0].mValue = v0;
    if (numParam >= 2) mVCParam[1].mVCKeys[0].mValue = v1;
    if (numParam >= 3) mVCParam[2].mVCKeys[0].mValue = v2;
    if (numParam >= 4) mVCParam[3].mVCKeys[0].mValue = v3;
}

void CLS::
SetInitKeyframes(int keynum, double (*comp)[5] = NULL){
    uint num = mVCParam.size();
    ASSERT(num == 4);
    for (uint i = 0; i < num; i++){
	Parameter_cls &CParam = mVCParam[i];
	CParam.mCbTimeDepend.SetBodyFromData("1");
	CParam.mVCKeys.clear();
	for (int j = 0; j < keynum; j++){
	    wxString strLine;
	    strLine << comp[j][0] << "::0::" << comp[j][i+1] << "::1::1";
	    KeyFrame_cls CKey(strLine);
	    CParam.mVCKeys.insert(CParam.mVCKeys.begin()+j, CKey);
	}
    }
}

bool CLS::
CheckCalledAsLink(wxString& strTag){
    bool bCalledAsLinked = FALSE;
    wxString lastItem = GetLastItem(strTag, "::");
    if (lastItem.IsSameAs(GetStrName()) == FALSE){
	PopLastItemFromStrLine(strTag, "::");
	ASSERT(GetLastItem(strTag, "::").IsSameAs(GetStrName()) == TRUE);
	bCalledAsLinked = TRUE;
    }
    return bCalledAsLinked;
}

void CLS::
SetNewKeyCommand(double newValue, double Xnormalized, wxString strTag, 
		 int keyType, double easeFactor){
    
    wxString strLastItem = GetLastItem(strTag, "::");
    bool bCalledAsLinked = CheckCalledAsLink(strTag);

    GetPCUndo()->OpenBlock("SetNewKeyAsGroup");
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++){
	if (mVCParam[i].IsKeyExist(Xnormalized) == TRUE) continue;
	double newValueParam = newValue;
	if (bCalledAsLinked == TRUE &&
	    mVCParam[i].GetStrName().IsSameAs(strLastItem) == FALSE)
	    newValueParam = mVCParam[i].GetValue(Xnormalized);
	wxString strTagParam = strTag;
	strTagParam << "::" << mVCParam[i].GetStrName();
	mVCParam[i].SetNewKeyCommand(newValueParam, Xnormalized, strTagParam, 
				     keyType, easeFactor, TRUE);
    }
    GetPCUndo()->CloseBlock("SetNewKeyAsGroup");
}

void CLS::
InsertKeyCommand(double Xnormalized, wxString strTag, 
		 int keyType, double easeFactor){
    bool bCalledAsLinked = CheckCalledAsLink(strTag);
    GetPCUndo()->OpenBlock("InsertKeyAsGroup");
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++){
	if (mVCParam[i].mCbTimeDepend.GetBodyBOOL() == FALSE){
	    wxString strTagParam = strTag;
	    strTagParam << "::" << mVCParam[i].GetStrName();
	    wxString strUndo(strTagParam), strRedo(strTagParam);
	    strUndo << "::TimeDepend=0";
	    strRedo << "::TimeDepend=1";
	    GetPCUndo()->PushCommandPair(strUndo, strRedo); 
	    GetPCUndo()->Redo();
	}
    }
    for (uint i = 0; i < num; i++){
	if (mVCParam[i].IsKeyExist(Xnormalized) == TRUE) continue;
	wxString strTagParam = strTag;
	strTagParam << "::" << mVCParam[i].GetStrName();
	mVCParam[i].InsertKeyCommand(Xnormalized, strTagParam, keyType, easeFactor, TRUE);
    }
    GetPCUndo()->CloseBlock("InsertKeyAsGroup");
}

void CLS::
DeleteKeyCommand(double Xnormalized, wxString strTag){
    bool bCalledAsLinked = CheckCalledAsLink(strTag);
    GetPCUndo()->OpenBlock("DeleteKeyAsGroup");
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++){
	if (mVCParam[i].IsKeyExist(Xnormalized) == FALSE) continue;
	if (mVCParam[i].mCbTimeDepend.GetBodyBOOL() == FALSE) continue;
	wxString strTagParam = strTag;
	strTagParam << "::" << mVCParam[i].GetStrName();
	mVCParam[i].DeleteKeyCommand(Xnormalized, strTagParam, TRUE);
    }
    GetPCUndo()->CloseBlock("DeleteKeyAsGroup");
}

void CLS::
DeleteSelectedTLKeyCommand(wxString strTag){
    CheckTag(strTag);
    uint num = mVCParam.size();
    int keyNum = 0;
    for (uint i = 0; i < num; i++)
	keyNum += mVCParam[i].GetSelectedKeyNum();
    if (keyNum == 0) return;

    if (keyNum > 1) GetPCUndo()->OpenBlock("DeleteSelectedKeys");
    for (uint i = 0; i < num; i++)
	mVCParam[i].DeleteSelectedTLKeyCommand(strTag);
    if (keyNum > 1) GetPCUndo()->CloseBlock("DeleteSelectedKeys");
}

void CLS::
SetSelectedTLKeyTypeCommand(wxString strTag, int keyType, double ease){
    
    CheckTag(strTag);
    uint num = mVCParam.size();
    int keyNum = 0;
    for (uint i = 0; i < num; i++)
	keyNum += mVCParam[i].GetSelectedKeyNum();
    if (keyNum == 0) return;

    if (keyNum > 1) GetPCUndo()->OpenBlock("SetSelectedKeysType");
    for (uint i = 0; i < num; i++)
	mVCParam[i].SetSelectedTLKeyTypeCommand(strTag, keyType, ease);
    if (keyNum > 1) GetPCUndo()->CloseBlock("SetSelectedKeysType");
}

void CLS::
MoveSelectedTLKeyCommand(double dNX, double dY, wxString strTag){
    uint num = mVCParam.size();
    strTag << "::" << GetStrName();
    for (uint i = 0; i < num; i++)
	mVCParam[i].MoveSelectedTLKeyCommand(dNX, dY, strTag);
}

/*
void CLS::
SetExpression(wxString strNewExpression, wxString strTag){
    
}
*/

void CLS::
RescaleParametersCommand(double scale, wxString strTag){
    uint num = mVCParam.size();
    strTag << "::" << GetStrName();
    for (uint i = 0; i < num; i++)
	mVCParam[i].RescaleCommand(scale, strTag);
}

/*
void CLS::
CopyParamGroupCommand(ParamGroup_cls *pCNew, wxString strTag){
    strTag << "::";

    wxString strNameOrg = pCNew->mCstrName.GetBodySTR();
    pCNew->mCstrName.SetBodyFromData(mCstrName.GetBodySTR());

    wxString strUndo(strTag), strRedo(strTag);

    Serialize(strUndo);
    pCNew->Serialize(strRedo); 
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();

    pCNew->mCstrName.SetBodyFromData(strNameOrg);
}
*/

bool CLS::
ChangeKeyTypeByModKey(wxString strTag, double posX, int newKeyType,
		      double resolution){
    if (resolution == _RES_DEFAULT)
	resolution = GetPCTimeCtrl()->GetDefaultResolution();
    
    wxString strLastItem = GetLastItem(strTag, "::");
    bool bCalledAsLinked = CheckCalledAsLink(strTag);

    bool bChanged = FALSE;
    uint num = mVCParam.size();
    GetPCUndo()->OpenBlock("ChangeKeyTypeGroup");
    for (uint i = 0; i < num; i++){

	int keyIndex = mVCParam[i].GetKeyIndex(posX);
	if (keyIndex == -1) continue;

	wxString strTagParam = strTag;
	strTagParam << "::" << mVCParam[i].GetStrName();

	bChanged |= mVCParam[i].ChangeKeyType(strTagParam, keyIndex, newKeyType);
    }
    GetPCUndo()->CloseBlock("ChangeKeyTypeGroup");
    return bChanged;
}

void CLS::
DeselectKey(wxString strException){
    if (mCstrName.GetBodySTR().IsSameAs(strException) == TRUE) return;
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++)
	mVCParam[i].DeselectKey();
}

bool CLS::
SelectKey(double posX, bool bSelect, bool bAdditional, double resolution){
    if (resolution == _RES_DEFAULT)
	resolution = GetPCTimeCtrl()->GetDefaultResolution();
    uint num = mVCParam.size();
    bool bSelected = FALSE;
    for (uint i = 0; i < num; i++)
	bSelected |= mVCParam[i].SelectKey(posX, bSelect, bAdditional, resolution);
    return bSelected;
}

void CLS::
SelectKeyAll(){
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++) mVCParam[i].SelectKeyAll(TRUE);
}

void CLS::
Flip(){
    uint num = mVCParam.size();
    for (uint i = 0; i < num; i++) mVCParam[i].Flip();
}

void CLS::
SwapRGB(){
    wxString strR; mVCParam[0].Serialize(strR);
    wxString strG; mVCParam[1].Serialize(strG);
    wxString strB; mVCParam[2].Serialize(strB);

    std::wstring wstrLine;

    std::wstring strBufR(strR.c_str());
    std::wistringstream sstreamR(strBufR);
    while ( std::getline (sstreamR, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
	mVCParam[1].ReadStrLine(sstreamR, strFirstItem, strLine);
    }

    std::wstring strBufG(strG.c_str());
    std::wistringstream sstreamG(strBufG);
    while ( std::getline (sstreamG, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
	mVCParam[2].ReadStrLine(sstreamG, strFirstItem, strLine);
    }

    std::wstring strBufB(strB.c_str());
    std::wistringstream sstreamB(strBufB);
    while ( std::getline (sstreamB, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
	mVCParam[0].ReadStrLine(sstreamB, strFirstItem, strLine);
    }
}

void CLS::
DrawTL(wxDC& dc, TLPanelGeom_cls* pCGeom,
       wxBitmap &CBitmapBuffer,
       int sashX, int posY, int panelWidth, int height,
       double normalizedWAMin, double normalizedWAMax){

    if (CBitmapBuffer.GetWidth() != panelWidth ||
	CBitmapBuffer.GetHeight() != height)
	CBitmapBuffer.Create(panelWidth, height); 
    wxMemoryDC CMemDC(CBitmapBuffer);
    
    DrawTLBackground(CMemDC, pCGeom, sashX, panelWidth, height, 
		     normalizedWAMin, normalizedWAMax);

    dc.Blit(sashX, posY, panelWidth-sashX, height, &CMemDC, sashX, 0);
}

void CLS::
Serialize(wxString& strOut){
    strOut << wxT("ReadParamGroup::") << mCstrName.GetBodySTR() << ENDL;
    ConfListContainer_cls::Serialize(strOut);
    int parameterNum = mVCParam.size();
    for (int i = 0; i < parameterNum; i++)
	mVCParam[i].Serialize(strOut);
    strOut << wxT("ReadParamGroup::End") << ENDL;
}

bool CLS::
ReadStrOneLiner(wxString strFirstItem, wxString strLine){
    if (mCConfList.ReadStrLineAll(strFirstItem) == TRUE) return TRUE;

    int parameterNum = mVCParam.size();
    for (int i = 0; i < parameterNum; i++)
	if (mVCParam[i].GetStrName().IsSameAs(strFirstItem)){
	    strFirstItem = PopStrItemFromStrLine(strLine, "::");
	    return mVCParam[i].ReadStrOneLiner(strFirstItem, strLine);
	}
    return FALSE;
}

bool CLS::
ReadStrLine(std::wistream &istream,
	    wxString strFirstItem, wxString strLine){

    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE) return TRUE;
    
    ConfItemSTR_cls CstrName;
    CstrName.SetItem("Name=", "dummy");
    if (CstrName.ReadStrLine(strFirstItem) == TRUE){
	int parameterNum = mVCParam.size();
	for (int i = 0; i < parameterNum; i++){
	    if (CstrName.GetBodySTR().IsSameAs(mVCParam[i].GetStrName())){
		std::wstring wstrLine;
		std::getline(istream, wstrLine);
		wxString strLine = wstrLine.c_str();
		wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
		mVCParam[i].ReadStrLine(istream, strFirstItem, strLine);
	    }
	}
    }

    return TRUE;
}
