// (setq c-basic-offset 4)
// "clsParameter.h"

#include "clsParameter.h"
#include "clsParamGroup.h"

#include "../MISC/clsMisc.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../TIMELINE/clsTimelineBase.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"

#include "../COMMON/CommonDefinitions.h"

#include <wx/dcbuffer.h>
#include <wx/dcmemory.h>
#include <algorithm>

#define CLS Parameter_cls

CLS::
Parameter_cls(ParamGroup_cls* pCParent, wxString strName, int selfIndex){
    mSelfIndex = selfIndex;
    mProcCounter = 0;
    mbUseVelocityCache = FALSE;
    mpCParent = pCParent;
    mCstrName.     SetItem("Name=", strName);
    mCbTimeDepend. SetItem("TimeDepend=", "0");
    mCbLocked.     SetItem("Locked=", "0");
    mCbExtrapolate.SetItem("Extrapolate=", "0");
    
    mCColorBase = wxColor(204, 196, 196);
    mCColorWA = wxColor(248, 240, 240);

    mVCKeys.push_back(KeyFrame_cls());
}

wxString CLS::
GetStrName(){ return mCstrName.GetBodySTR(); }

ConfItemBOOL_cls* CLS::
GetPCItem(wxString strHeader){
    if (strHeader.IsSameAs("TimeDepend=")) return &mCbTimeDepend;
    if (strHeader.IsSameAs("Extrapolate=")) return &mCbExtrapolate;
    if (strHeader.IsSameAs("Locked=")) return &mCbLocked;
    return NULL;
}

void CLS::
CheckTag(wxString& strTag){
    wxString strLastItem = GetLastItem(strTag, "::");
    if (strLastItem.IsSameAs(GetStrName()) == FALSE)
	strTag << "::" << GetStrName();
    /*
    wxString strLastItem = PopLastItemFromStrLine(strTag, "::");
    if (strLastItem.IsSameAs(GetStrName()))
	strTag << "::" << strLastItem;
    else
	strTag << "::" << strLastItem << "::" << GetStrName();
    */
}

void CLS::
PrecalcBegin(){
    uint size = mVCKeys.size();
    for (uint i = 0; i < size; i++){
	mVCKeys[i].mVelocityCache = GetKeyVelocityAuto(i);
    }
    mbUseVelocityCache = TRUE;    
}

void CLS::
PrecalcEnd(){
    mbUseVelocityCache = FALSE;
}

int CLS::
GetActiveKeyIndex(){
    uint size = mVCKeys.size();
    for (uint i = 0; i < size; i++)
	if (mVCKeys[i].mbActive == TRUE) return (int)i;
    return -1;
}

int CLS::
GetCurrentKeyIndex(){
    double currentNT = GetPCTimeCtrl()->GetCurrentNT();

    uint size = mVCKeys.size();
    for (uint i = 0; i < size; i++)
	if (Misc_cls::IsNearlySame(currentNT, mVCKeys[i].mXNormalized, 1e-6)) return i;
    return -1;
}

int CLS::
GetKeyIndex(double Xnormalized){
    uint size = mVCKeys.size();
    for (uint i = 0; i < size; i++)
	if (Misc_cls::IsNearlySame(Xnormalized, mVCKeys[i].mXNormalized, 1e-6)) return i;
    return -1;
}

int CLS::
GetControllingKey(double Xnormalized){
    int active = -1;
    uint size = mVCKeys.size();
    for (uint i = 0; i < size; i++){
	if (Xnormalized >= mVCKeys[i].mXNormalized - 1e-6) active = i;
	else break;
    }
    return active;
}

int CLS::
GetAdjuscentKeyIndex(int direction, double XNormalized){
    if (XNormalized == -1) XNormalized = GetPCTimeCtrl()->GetCurrentNT();
    bool bOnKey = IsKeyExist(XNormalized);
    int controlling = GetControllingKey(XNormalized);
    if (bOnKey == TRUE) controlling += direction;
    else
	if (direction == +1) controlling += direction;

    if (controlling < -1) controlling = -1;
    if (controlling >= (int)mVCKeys.size()) controlling = mVCKeys.size();
    return controlling;
}

double CLS::
GetKeyXNormalizeEX(int index){
    if (index < 0) return 0;
    if (index >= mVCKeys.size()) return 1;
    return mVCKeys[index].mXNormalized;
}

bool CLS::
IsSelectedKeyExist(){
    uint size = mVCKeys.size();
    for (uint i = 0; i < size; i++)
	if (mVCKeys[i].mbSelected == TRUE) return TRUE;
    return FALSE;
}

bool CLS::
IsKeyExist(double XNormalized){
    uint size = mVCKeys.size();
    for (uint i = 0; i < size; i++){
	if (Misc_cls::IsNearlySame(mVCKeys[i].mXNormalized, XNormalized, 1e-6))
	    return TRUE;
    }
    return FALSE;
}

bool CLS::
IsKeyLinkedAll(int keyIndex){
    if (mpCParent->mCbLinked.GetBodyBOOL() == FALSE) return FALSE;
    return mpCParent->IsKeyExistAll(mVCKeys[keyIndex].mXNormalized);
}

int CLS::
GetSelectedKeyNum(){
    int num = 0;
    uint size = mVCKeys.size();
    for (uint i = 0; i < size; i++)
	if (mVCKeys[i].mbSelected == TRUE) num++;
    return num;
}

double CLS::
GetCurrentValue(){
    return GetValue(GetPCTimeCtrl()->GetCurrentNT()); }

double CLS::
GetValueExtrapolate0(double XNormalized){
    int idx0 = 0;
    int idx1 = 1;
    double v = GetKeyVelocityAuto(idx0);
    double dX1 = XNormalized - mVCKeys[idx0].mXNormalized;
    return mVCKeys[idx0].mValue + v * dX1;
}

double CLS::
GetValueExtrapolate1(double XNormalized){
    int idx0 = mVCKeys.size() - 2;
    int idx1 = idx0 + 1;
    double v = GetKeyVelocityAuto(idx1);
    
    double dX1 = XNormalized - mVCKeys[idx1].mXNormalized;
    return mVCKeys[idx1].mValue + v * dX1;
}

double CLS::
GetKeyVelocityLinear(int index0, int index1){
    double width = mVCKeys[index1].mXNormalized - mVCKeys[index0].mXNormalized;
    return (mVCKeys[index1].mValue - mVCKeys[index0].mValue)/width;
}

double CLS::
GetKeyVelocityAuto(int index){
    if (mbUseVelocityCache == TRUE) return mVCKeys[index].mVelocityCache;

    if (mCbTimeDepend.GetBodyBOOL() == FALSE)
	return 0;
    
    uint size = mVCKeys.size();
    if (size == 1) return 0;

    if (mVCKeys[index].mKeyType == _KEY_STEP) return 0;
    if (mVCKeys[index].mKeyType == _KEY_LINEAR){
	if (index == size - 1){
	    if (mCbExtrapolate.GetBodyBOOL() == FALSE)
		return 0;
	    else
		return GetKeyVelocityLinear(index-1, index);
	}
	return GetKeyVelocityLinear(index, index+1);
    }
    if (mVCKeys[index].mKeyType == _KEY_HERMITE_AUTO){
	if (index == size - 1){
	    if (mCbExtrapolate.GetBodyBOOL() == FALSE)
		return 0;
	    else
		return GetKeyVelocityLinear(index-1, index) * mVCKeys[index].mEaseFactor;
	}
	else if (index == 0){
	    if (mCbExtrapolate.GetBodyBOOL() == FALSE)
		return 0;
	    else
		return GetKeyVelocityLinear(index, index+1) * mVCKeys[index].mEaseFactor;
	}
	else
	    return GetKeyVelocityLinear(index-1, index+1) * mVCKeys[index].mEaseFactor;
    }
    
    if (index == 0) return GetKeyVelocityLinear(index, index+1);
    if (index == size - 1) return GetKeyVelocityLinear(index-1, index);
    
    return GetKeyVelocityLinear(index-1, index+1);
}

double CLS::
GetKeyEaseFactor(int index){
    if (mVCKeys[index].mKeyType == _KEY_STEP) return 0;
    if (mVCKeys[index].mKeyType == _KEY_LINEAR) return 1;
    return mVCKeys[index].mEaseFactor;
}

double CLS::
GetValue(double XNormalized){
    if (mCbTimeDepend.GetBodyBOOL() == FALSE)
	return mVCKeys[0].mValue;

    bool bOnKey = IsKeyExist(XNormalized);
    int controlling = GetControllingKey(XNormalized);
    
    if (bOnKey == TRUE) return mVCKeys[controlling].mValue;

    uint size = mVCKeys.size();
    
    if (controlling == -1){
	if (mVCKeys[0].mKeyType == _KEY_STEP) return mVCKeys[0].mValue;
	if (mCbExtrapolate.GetBodyBOOL() == FALSE) return mVCKeys[0].mValue;
	if (mVCKeys.size() == 1) return mVCKeys[0].mValue;
	return GetValueExtrapolate0(XNormalized);
    }

    if (mVCKeys[controlling].mKeyType == _KEY_STEP) return mVCKeys[controlling].mValue; 
    if (controlling == size - 1){
	if (mCbExtrapolate.GetBodyBOOL() == FALSE) return mVCKeys[controlling].mValue;
	if (mVCKeys.size() == 1) return mVCKeys[0].mValue;
	return GetValueExtrapolate1(XNormalized);
    }
    
    if (mVCKeys[controlling].mKeyType == _KEY_LINEAR)
	return InterpolateLinear(controlling, XNormalized);

    if (mVCKeys[controlling].mKeyType == _KEY_HERMITE_AUTO)
	return InterpolateHermiteAuto(controlling, XNormalized);

    return 0;
}

double CLS::
InterpolateLinear(int index, double NT){
    double width = mVCKeys[index+1].mXNormalized-mVCKeys[index].mXNormalized;
    double rate = (NT - mVCKeys[index].mXNormalized) / width;
    return mVCKeys[index].mValue*(1.0-rate) + mVCKeys[index+1].mValue*rate;
}

double CLS::
InterpolateHermiteAuto(int index, double NT){
    double width = mVCKeys[index+1].mXNormalized-mVCKeys[index].mXNormalized;
    double rate = (NT - mVCKeys[index].mXNormalized) / width;

    double vin = GetKeyVelocityAuto(index);
    double vout = GetKeyVelocityAuto(index+1);

    vin  *= GetKeyEaseFactor(index);
    vout *= GetKeyEaseFactor(index+1);
    
    double value0 = mVCKeys[index].mValue;
    double value1 = mVCKeys[index+1].mValue;
    
    double interpolatedValue = 
	value0 + 
	vin * width * rate + 
	(-3 * value0 - 2 * vin * width + 3 * value1 - vout * width) * rate * rate + 
	(2 * value0 + vin * width - 2 * value1 + vout * width) * rate * rate * rate;
    //mInterpolatedValue = interpolatedValue;
    return interpolatedValue;
    
}

void CLS::
SortKey(){
    std::sort(mVCKeys.begin(), mVCKeys.end());
}

void CLS::
SetNewKeyCommand(double newValue, double XNormalized, wxString strTag,
		 int keyType, double easeFactor, bool bCalledFromGroup){
    if (bCalledFromGroup == FALSE && mpCParent->mCbLinked.GetBodyBOOL() == TRUE){
	mpCParent->SetNewKeyCommand(newValue, XNormalized, strTag,
				    keyType, easeFactor);
	return; }

    int controlling = GetControllingKey(XNormalized);

    if (keyType == -1){
	keyType = 0;
	if (controlling == -1 && mVCKeys.size() > 0) keyType = mVCKeys[0].mKeyType; 
	if (controlling != -1){
	    keyType = mVCKeys[controlling].mKeyType;
	    if (keyType == _KEY_HERMITE_AUTO)
		if (mVCKeys[controlling].mEaseFactor == 0) easeFactor = 0;
	}
    }
    
    // controlling + 1  index Ƃđ}͂
    wxString strUndo(strTag);
    wxString strRedo(strTag);
    strUndo << "::DeleteKey::" << controlling+1;
    strRedo << "::InsertKey::" << controlling+1 << "::" << XNormalized << "::0::" << newValue << "::" << keyType << "::" << easeFactor;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
InsertKeyCommand(double XNormalized, wxString strTag, 
		 int keyType, double easeFactor,
		 bool bCalledFromGroup){
    if (bCalledFromGroup == FALSE && mpCParent->mCbLinked.GetBodyBOOL() == TRUE){
	mpCParent->InsertKeyCommand(XNormalized, strTag, keyType, easeFactor);
	return; }

    double currentValue = GetValue(XNormalized);
    SetNewKeyCommand(currentValue, XNormalized, strTag, 
		     keyType, easeFactor, FALSE);
}

void CLS::
DeleteKeyCommand(double XNormalized, wxString strTag,
		 bool bCalledFromGroup){
    if (bCalledFromGroup == FALSE && mpCParent->mCbLinked.GetBodyBOOL() == TRUE){
	mpCParent->DeleteKeyCommand(XNormalized, strTag);
	return; }

    if (mVCKeys.size() == 1) return; // Do not remeve last one.
    int index = GetControllingKey(XNormalized);
    int type = mVCKeys[index].mKeyType;
    double currentValue = mVCKeys[index].mValue;
    bool bSelected = mVCKeys[index].mbSelected;
    wxString strUndo(strTag);
    wxString strRedo(strTag);
    strUndo << "::InsertKey::" << index << "::" << XNormalized << "::" << bSelected << "::" << currentValue << "::" << type;
    strRedo << "::DeleteKey::" << index;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetKeyValueCommand(int index, double newValue, wxString strTag){
    double oldValue = mVCKeys[index].mValue;
    if (oldValue == newValue) return;
    wxString strUndo(strTag);
    wxString strRedo(strTag);
    strUndo << "::Key::" << index << "::SetValue::" << oldValue;
    strRedo << "::Key::" << index << "::SetValue::" << newValue;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetKeyTypeCommand(int index, int newType, wxString strTag){
    int oldType = mVCKeys[index].mKeyType;
    if (oldType == newType) return;

    CheckTag(strTag);
    wxString strUndo(strTag);
    wxString strRedo(strTag);
    strUndo << "::Key::" << index << "::SetType::" << oldType;
    strRedo << "::Key::" << index << "::SetType::" << newType;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetKeyEaseFactorCommand(int index, double newValue, wxString strTag){
    double oldValue = mVCKeys[index].mEaseFactor;
    if (oldValue == newValue) return;

    CheckTag(strTag);
    wxString strUndo(strTag);
    wxString strRedo(strTag);
    strUndo << "::Key::" << index << "::SetEaseFactor::" << oldValue;
    strRedo << "::Key::" << index << "::SetEaseFactor::" << newValue;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetValueCommand(double newValue, wxString strTag, 
		int keyType, double easeFactor){
    if (mCbTimeDepend.GetBodyBOOL() == FALSE){
	SetKeyValueCommand(0, newValue, strTag);
	return; }

    int index = GetCurrentKeyIndex();
    if (index != -1){
	SetKeyValueCommand(index, newValue, strTag);
	return; }

    if (keyType == -1){
	GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);
    }

    SetNewKeyCommand(newValue, GetPCTimeCtrl()->GetCurrentNT(), strTag, keyType, easeFactor, FALSE);
}

void CLS::
SetValueAtCommand(double XNormalized, double newValue, wxString strTag,
		  int keyType, double easeFactor){


    if (mCbTimeDepend.GetBodyBOOL() == FALSE){
	SetKeyValueCommand(0, newValue, strTag);
	return; }

    int index = GetKeyIndex(XNormalized);
    if (index != -1){
	SetKeyValueCommand(index, newValue, strTag);
	return; }

    if (keyType == -1){
	GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);
    }

    SetNewKeyCommand(newValue, XNormalized, strTag, keyType, easeFactor, FALSE);
}


void CLS::
DeleteSelectedTLKeyCommand(wxString strTag){
    if (IsSelectedKeyExist() == FALSE) return;

    uint keyNum = mVCKeys.size();
    CheckTag(strTag);

    for (int i = keyNum - 1; i >= 0; i--)
	if (mVCKeys[i].mbSelected == TRUE)
	    DeleteKeyCommand(mVCKeys[i].mXNormalized, strTag, TRUE);
}

void CLS::
SetSelectedTLKeyTypeCommand(wxString strTag, int keyType, double ease){
    if (IsSelectedKeyExist() == FALSE) return;

    uint keyNum = mVCKeys.size();
    CheckTag(strTag);

    for (int i = 0; i < keyNum; i++)
	if (mVCKeys[i].mbSelected == TRUE){
	    SetKeyTypeCommand(i, keyType, strTag);
	    SetKeyEaseFactorCommand(i, ease, strTag);
	}
}

void CLS::
MoveSelectedTLKeyCommand(double dNX, double dY, wxString strTag){
    if (IsSelectedKeyExist() == FALSE) return;
    
    //if (mCbLocked.GetBodyBOOL() == TRUE) return;

    CheckTag(strTag);
    uint num = mVCKeys.size();
    
    int ini = 0, inc = 1;
    if (dNX > 0){ ini = num-1; inc = -1; }

    GetPCUndo()->OpenBlock("MoveKeyAndSort");
    for (int i = ini; (0 <= i) && (i < num); i += inc){
	if (mVCKeys[i].mbSelected == FALSE) continue;
	
	double oldY = mVCKeys[i].mValue;
	double newY = oldY + dY;
	    
	double oldX = mVCKeys[i].mXNormalized;
	double newX;
	if (mpCParent->mbForColorCurve == TRUE) 
	    newX = GetPCTimeCtrl()->GetQuantizedForColor(oldX+dNX, _MISC_ROUND_ROUND);
	else 
	    newX = GetPCTimeCtrl()->GetQuantized(oldX+dNX, _MISC_ROUND_ROUND);
	    
	if (Misc_cls::IsNearlySame(oldX, newX, 1e-6)
	    && oldY == newY) continue;

	if (Misc_cls::IsNearlySame(oldX, newX, 1e-6)){
	    wxString strUndo(strTag), strRedo(strTag);
	    strUndo << "::Key::" << i << "::SetValue::" << oldY;
	    strRedo << "::Key::" << i << "::SetValue::" << newY;
	    GetPCUndo()->PushCommandPair(strUndo, strRedo);
	    GetPCUndo()->Redo();
	    continue;
	}
	
	if (IsKeyExist(newX) == TRUE){
	    DeleteKeyCommand(newX, strTag, TRUE);
	    num = mVCKeys.size();
	    if (dNX < 0) i--;
	}

	int newi = GetControllingKey(newX);
	if (dNX < 0) newi++;

	wxString strUndo1(strTag), strUndo2(strTag), strUndo3(strTag);
	wxString strRedo1(strTag), strRedo2(strTag), strRedo3(strTag);
	strUndo1 << "::SortKey";
	strUndo2 << "::Key::" << newi << "::SetValue::" << oldY;
	strUndo3 << "::Key::" << newi << "::SetX::" << oldX;
	strRedo1 << "::Key::" << i << "::SetX::" << newX;
	strRedo2 << "::Key::" << i << "::SetValue::" << newY;
	strRedo3 << "::SortKey";
	GetPCUndo()->PushCommandPair(strUndo1, strRedo1);
	GetPCUndo()->Redo();
	if (oldY != newY){
	    GetPCUndo()->PushCommandPair(strUndo2, strRedo2);
	    GetPCUndo()->Redo();
	}
	GetPCUndo()->PushCommandPair(strUndo3, strRedo3);
	GetPCUndo()->Redo();
    }
    GetPCUndo()->CloseBlock("MoveKeyAndSort");
}

void CLS::
MoveTLKeyCommand(int i, double dNX, double dY, wxString strTag){

    uint num = mVCKeys.size();
    CheckTag(strTag);

    double oldY = mVCKeys[i].mValue;
    double newY = oldY + dY;
    
    double oldX = mVCKeys[i].mXNormalized;
    double newX;
    if (mpCParent->mbForColorCurve == TRUE) 
	newX = GetPCTimeCtrl()->GetQuantizedForColor(oldX+dNX, _MISC_ROUND_ROUND);
    else 
	newX = GetPCTimeCtrl()->GetQuantized(oldX+dNX, _MISC_ROUND_ROUND);
    
    if (Misc_cls::IsNearlySame(oldX, newX, 1e-6)
	&& oldY == newY) return;
    
    if (Misc_cls::IsNearlySame(oldX, newX, 1e-6)){
	wxString strUndo(strTag), strRedo(strTag);
	strUndo << "::Key::" << i << "::SetValue::" << oldY;
	strRedo << "::Key::" << i << "::SetValue::" << newY;
	GetPCUndo()->PushCommandPair(strUndo, strRedo);
	GetPCUndo()->Redo();
	return;
    }

    GetPCUndo()->OpenBlock("MoveKeyAndSort");
    
    if (IsKeyExist(newX) == TRUE){
	DeleteKeyCommand(newX, strTag, TRUE);
	num = mVCKeys.size();
	if (dNX < 0) i--;
    }
    
    int newi = GetControllingKey(newX);
    if (dNX < 0) newi++;
    
    wxString strUndo1(strTag), strUndo2(strTag), strUndo3(strTag);
    wxString strRedo1(strTag), strRedo2(strTag), strRedo3(strTag);
    strUndo1 << "::SortKey";
    strUndo2 << "::Key::" << newi << "::SetValue::" << oldY;
    strUndo3 << "::Key::" << newi << "::SetX::" << oldX;
    strRedo1 << "::Key::" << i << "::SetX::" << newX;
    strRedo2 << "::Key::" << i << "::SetValue::" << newY;
    strRedo3 << "::SortKey";
    GetPCUndo()->PushCommandPair(strUndo1, strRedo1);
    GetPCUndo()->Redo();
    if (oldY != newY){
	GetPCUndo()->PushCommandPair(strUndo2, strRedo2);
	GetPCUndo()->Redo();
    }
    GetPCUndo()->PushCommandPair(strUndo3, strRedo3);
    GetPCUndo()->Redo();

    GetPCUndo()->CloseBlock("MoveKeyAndSort");
}

void CLS::
RescaleCommand(double scale, wxString strTag){
    if (mCbTimeDepend.GetBodyBOOL() == FALSE) return;
    int num = (int)mVCKeys.size();
    if (scale > 1){
	for (int i = num - 1; i >= 0; i--){
	    double nx = mVCKeys[i].mXNormalized;
	    MoveTLKeyCommand(i, nx*(scale-1), 0, strTag);
	}
    }
    if (scale < 1){
	for (int i = 0; i < num; i++){
	    double nx = mVCKeys[i].mXNormalized;
	    MoveTLKeyCommand(i, nx*(scale-1), 0, strTag);
	}
    }
}

bool CLS::
ChangeKeyType(wxString strTag, int keyIndex, int newKeyType){
    
    int oldKeyType = mVCKeys[keyIndex].mKeyType;
    if (oldKeyType == newKeyType) return FALSE;

    SetKeyTypeCommand(keyIndex, newKeyType, strTag);

    return TRUE;
}

bool CLS::
ChangeKeyTypeByModKey(wxString strTag, int keyIndex, bool bCtrl, bool bShift, bool bAlt){
    int oldKeyType = mVCKeys[keyIndex].mKeyType;
    int newKeyType = _KEY_LINEAR;
    if (bCtrl == TRUE && bShift == FALSE && bAlt == FALSE){
	if (oldKeyType == _KEY_HERMITE_AUTO) newKeyType = _KEY_LINEAR;
	else newKeyType = _KEY_HERMITE_AUTO;
    }
    if (bCtrl == TRUE && bShift == FALSE && bAlt == TRUE){
	newKeyType = _KEY_STEP;
    }

    if (oldKeyType == newKeyType) return FALSE;

    if (mpCParent->mCbLinked.GetBodyBOOL() == TRUE)
	return mpCParent->
	    ChangeKeyTypeByModKey(strTag, mVCKeys[keyIndex].mXNormalized, newKeyType);

    SetKeyTypeCommand(keyIndex, newKeyType, strTag);

    return TRUE;
}

void CLS::DeactivateKeys(){
    uint num = mVCKeys.size();
    for (uint i = 0; i < num; i++) mVCKeys[i].mbActive = FALSE;
}

void CLS::DeselectKey(){
    uint num = mVCKeys.size();
    for (uint i = 0; i < num; i++) mVCKeys[i].mbSelected = FALSE;
}

bool CLS::
SelectKey(double posX, bool bSelect, bool bAdditional, double resolution){
    if (resolution == _RES_DEFAULT)
	resolution = GetPCTimeCtrl()->GetDefaultResolution();

    //if (mCbLocked.GetBodyBOOL() == TRUE) return FALSE;

    if (bAdditional == FALSE) DeselectKey();
    uint num = mVCKeys.size();
    bool bSelected = FALSE;
    //int selected = -1;
    for (uint i = 0; i < num; i++){
	bSelected |= mVCKeys[i].Select(posX, resolution, bSelect);
	if (bSelected == TRUE){
	    //selected = i;
	    DeactivateKeys();
	    mVCKeys[i].mbActive = TRUE;
	    break;
	}
    }
    
    return bSelected;
}

bool CLS::
SelectKeyByIndex(int keyIndex, bool bSelect, bool bAdditional,
		 double resolution){
    if (keyIndex < 0 || keyIndex >= mVCKeys.size()) return FALSE;
    
    if (mpCParent->mCbLinked.GetBodyBOOL() == TRUE)
	return mpCParent->
	    SelectKey(mVCKeys[keyIndex].mXNormalized, bSelect, bAdditional, resolution);

    //if (mCbLocked.GetBodyBOOL() == TRUE) bSelect = FALSE;

    if (bAdditional == FALSE) mpCParent->DeselectKey(wxES);
    mVCKeys[keyIndex].mbSelected = bSelect;
    DeactivateKeys();
    mVCKeys[keyIndex].mbActive = TRUE;
    return TRUE;
}

void CLS::
SelectKeyAll(bool bCalledFromGroup){
    if (mpCParent->mCbLinked.GetBodyBOOL() == TRUE && bCalledFromGroup == FALSE)
	mpCParent->SelectKeyAll();
    
    if (mCbTimeDepend.GetBodyBOOL() == FALSE) return;
    uint num = mVCKeys.size();
    for (uint i = 0; i < num; i++) mVCKeys[i].mbSelected = TRUE;
}

void CLS::
Flip(){
    std::vector < KeyFrame_cls > mVCKeysOrg = mVCKeys;
    uint num = mVCKeys.size();
    for (uint i = 0; i < num; i++){
	mVCKeys[i] = mVCKeysOrg[num-1-i];
	mVCKeys[i].mXNormalized = 1 - mVCKeys[i].mXNormalized;
    }
}

void CLS::
DrawTL(wxDC& dc, TLPanelGeom_cls* pCGeom,
       wxBitmap &CBitmapBuffer, double dX, 
       int sashX, int posY, int panelWidth, int height,
       double normalizedWAMin, double normalizedWAMax,
       std::vector < wxRect > &AKeyFrameRect,
       std::vector < int > &AKeyFrameIndex){

    if (CBitmapBuffer.GetWidth() != panelWidth ||
	CBitmapBuffer.GetHeight() != height)
	CBitmapBuffer.Create(panelWidth, height); 
    wxMemoryDC CMemDC(CBitmapBuffer);
    CMemDC.SetFont(dc.GetFont());

    DrawTLBackground(CMemDC, pCGeom, sashX, panelWidth, height, 
		     normalizedWAMin, normalizedWAMax);

    if (mpCParent->mCbExpression.GetBodyBOOL() == TRUE){
	double NT = GetPCTimeCtrl()->GetCurrentNT();
	float value = (float)mpCParent->GetVInterpolatedValue(NT)[mSelfIndex];
	wxString strValue; strValue << value;
	//int posX = pCGeom->GetDevCoordFromN(NT, sashX);
	CMemDC.SetTextForeground(GetPCColorProvider()->GetColor("TextWeak"));
	CMemDC.DrawText(strValue, sashX + 16, 0);
    }

    uint keyNum = mVCKeys.size();
    if (mCbTimeDepend.GetBodyBOOL() == FALSE) keyNum = 0;

    if (dX != 0)
	for (int i = 0; i < keyNum; i++){
	    if (mVCKeys[i].mbSelected == TRUE)
		mVCKeys[i].DrawTLGhost(CMemDC, pCGeom, sashX, panelWidth, height, dX);
	}
    
    for (int i = 0; i < keyNum; i++){
	int posX;
	bool bDrawn = mVCKeys[i].DrawTL(CMemDC, pCGeom, sashX, panelWidth, height, posX, i);
	if (bDrawn == FALSE) continue;
	AKeyFrameRect.push_back(wxRect(wxPoint(posX-4, posY + height/2-4), wxSize(9, 9)));
	AKeyFrameIndex.push_back(i);
    }

    dc.Blit(sashX, posY, panelWidth-sashX, height, &CMemDC, sashX, 0);
}

void CLS::
Serialize(wxString &strOut){
    mCstrName.Serialize(strOut);
    mCbTimeDepend.Serialize(strOut);
    mCbLocked.Serialize(strOut);
    mCbExtrapolate.Serialize(strOut);
    int keyNum = mVCKeys.size();
    strOut << wxT("keyNum=") << keyNum << ENDL;
    for (int i = 0; i < keyNum; i++)
	mVCKeys[i].Serialize(strOut);
}

bool CLS::
ReadStrOneLiner(wxString strFirstItem, wxString strLine){
    mProcCounter++;
    if (mCbTimeDepend.ReadStrLine(strFirstItem) == TRUE) return TRUE;
    if (mCbLocked.ReadStrLine(strFirstItem) == TRUE) return TRUE;
    if (mCbExtrapolate.ReadStrLine(strFirstItem) == TRUE) return TRUE;

    // !! eKeyɊւ鑀
    if (strFirstItem.IsSameAs("Key")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	long index = 0; strFirstItem.ToLong(&index);
	ASSERT(index >= 0);
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	return mVCKeys[index].ReadStrOneLiner(strFirstItem, strLine);
    }

    if (strFirstItem.IsSameAs("InsertKey")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	long index = 0; strFirstItem.ToLong(&index);
	KeyFrame_cls CKey(strLine);
	mVCKeys.insert(mVCKeys.begin()+index, CKey);
	return TRUE;
    }

    if (strFirstItem.IsSameAs("DeleteKey")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	long index = 0; strFirstItem.ToLong(&index);
	mVCKeys.erase(mVCKeys.begin()+index);
	return TRUE;
    }

    if (strFirstItem.IsSameAs("SortKey")){
	SortKey();
	return TRUE;
    }

    return FALSE;
}

void CLS::
ReadStrLine(std::wistream& istream, wxString strFirstItem, wxString strLine){
    mProcCounter++;
    
    std::wstring wstrLine;
    if (mCbTimeDepend.ReadStrLine(strFirstItem) == TRUE){
	std::getline(istream, wstrLine);
	wxString strLine = wstrLine.c_str();
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
    }
    if (mCbLocked.ReadStrLine(strFirstItem) == TRUE){
	std::getline(istream, wstrLine);
	wxString strLine = wstrLine.c_str();
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
    }
    if (mCbExtrapolate.ReadStrLine(strFirstItem) == TRUE){
	std::getline(istream, wstrLine);
	wxString strLine = wstrLine.c_str();
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
    }
    
    wxString strLineKeyNum = strFirstItem;
    ConfItemINT_cls CKeyNum;
    CKeyNum.SetItem("keyNum=", "0");
    if (CKeyNum.ReadStrLine(strLineKeyNum) == TRUE){
	mVCKeys.clear();
	int keyNum = CKeyNum.GetBodyINT();
	for (int i = 0; i < keyNum; i++){
	    std::getline(istream, wstrLine);
	    wxString strLineKey = wstrLine.c_str();
	    mVCKeys.push_back(KeyFrame_cls(strLineKey));
	}
    }
}
