// (setq c-basic-offset 4)
// "clsParameter.cpp"

#ifndef _PARAMETER_H_
#define _PARAMETER_H_

#include "clsKeyFrame.h"
#include "clsParamDrawTLBase.h"

#include "../CONFITEM/clsConfItem.h"
#include "../COMMON/clsGetPC.h"
#include "../FILEIO/clsParser.h"

class TLPanelGeom_cls;
class ParamGroup_cls;

#include <vector>

#define _RES_DEFAULT -1

class Parameter_cls : public ParamDrawTLBase_cls, public Parser_cls, public GetPC_cls {
public:
    Parameter_cls ( ParamGroup_cls *pCParent, wxString strName, int selfIndex);
    ConfItemSTR_cls mCstrName;
    ConfItemBOOL_cls mCbTimeDepend;
    ConfItemBOOL_cls mCbLocked;
    ConfItemBOOL_cls mCbExtrapolate;
    
    int mSelfIndex;
    int mProcCounter;
    bool mbUseVelocityCache;

    ParamGroup_cls *mpCParent;
    std::vector < KeyFrame_cls > mVCKeys;

    wxString GetStrName();
    ConfItemBOOL_cls *GetPCItem(wxString strHeader);
    void CheckTag(wxString& strTag);
    void PrecalcBegin();
    void PrecalcEnd();
    bool IsSelectedKeyExist();
    bool IsKeyExist(double Xnormalized);
    bool IsKeyLinkedAll(int keyIndex);
    int GetSelectedKeyNum();
    int GetActiveKeyIndex();
    int GetCurrentKeyIndex();
    int GetKeyIndex(double Xnormalized);
    int GetControllingKey(double Xnormalized);
    int GetAdjuscentKeyIndex(int direction, double XNormalized = -1);
    double GetKeyXNormalizeEX(int keyIndex);
    double GetCurrentValue();
    double GetValueExtrapolate0(double Xnormalized);
    double GetValueExtrapolate1(double Xnormalized);
    double GetKeyVelocityLinear(int index0, int index1);
    double GetKeyVelocityAuto(int index);
    double GetKeyEaseFactor(int index);
    double GetValue(double Xnormalized);
    double InterpolateLinear(int active, double NT);
    double InterpolateHermiteAuto(int active, double NT);
    void SortKey();

    void SetNewKeyCommand(double newValue, double XNormalized, wxString strTag, 
			  int keyType, double easeFactor, 
			  bool bCalledFromGroup);
    void InsertKeyCommand(double XNormalized, wxString strTag, 
			  int keyType, double easeFactor,
			  bool bCalledFromGroup);
    void DeleteKeyCommand(double XNormalized, wxString strTag,
			  bool bCalledFromGroup);
    void SetKeyValueCommand(int index, double newValue, wxString strTag);
    void SetKeyTypeCommand(int index, int type, wxString strTag);
    void SetKeyEaseFactorCommand(int index, double newEaseFactor, wxString strTag);
    void SetValueCommand(double newValue, wxString strTag,
			 int keyType = -1, double easeFactor = 1);
    void SetValueAtCommand(double XNormalized, double newValue, wxString strTag,
			   int keyType = -1, double easeFactor = 1);
    void DeleteSelectedTLKeyCommand(wxString strTag);
    void SetSelectedTLKeyTypeCommand(wxString strTag, int keyType, double ease);
    void MoveSelectedTLKeyCommand(double dNX, double dY, wxString strTag);
    void MoveTLKeyCommand(int i, double dNX, double dY, wxString strTag);
    void RescaleCommand(double scale, wxString strTag);

    bool ChangeKeyType(wxString strTag, int keyIndex, int newKeyType);
    bool ChangeKeyTypeByModKey(wxString strTag, int keyIndex, bool bCtrl, bool bShift, bool bAlt);
    void DeactivateKeys();
    void DeselectKey();
    bool SelectKey(double posX, bool bSelect, bool bAdditional, double resolution = _RES_DEFAULT);
    bool SelectKeyByIndex(int keyIndex, bool bSelect, bool bAdditional, double resolution = _RES_DEFAULT);
    void SelectKeyAll(bool bCalledFromGroup);
    void Flip();
    
    void DrawTL(wxDC& dc, TLPanelGeom_cls* pCGeom,
		wxBitmap &CBitmapBuffer, double dX, 
		int sashX, int posY, int width, int height,
		double normalizedWAMin, double normalizedWAMax,
		std::vector < wxRect > &AKeyFrameRect,
		std::vector < int > &AKeyFrameIndex);

    void Serialize(wxString &strOut);
    bool ReadStrOneLiner(wxString strFirstItem, wxString strLine);
    void ReadStrLine(std::wistream& istream, wxString strFirstItem, wxString strLine);
};

#endif
