// (setq c-basic-offset 4)
// "clsMCThread.h"

#include "clsMCThread.h"

#include "../../../COMMON/CommonDefinitions.h"

#define CLS MCThread_cls

CLS::
MCThread_cls(unsigned char *pDataLink,
	     int *pAllocSize, int *pFaceNum,
	     float **ppVertices, float **ppNormals, int **ppFaceIndex,
	     int bufferSize) : mpData(pDataLink), mpAllocSize(pAllocSize), mpFaceNum(pFaceNum) {
    mppVertices  = ppVertices;
    mppNormals   = ppNormals;
    mppFaceIndex = ppFaceIndex;
    mBufferSize  = bufferSize;
    mArrayIndex = 0;
}

void CLS::
SetParameters(int size[3], 
	      int initZ, int finZ, int initY, int finY, int initX, int finX,
	      float targetValue){
    LOOP_XYZ mSize[xyz] = size[xyz];
    mDataBoxFullSize = size[X] * size[Y] * size[Z];
    mInitZ = initZ; mFinZ = finZ;
    mInitY = initY; mFinY = finY;
    mInitX = initX; mFinX = finX;
    mTargetValue = targetValue;
}

// finds the point of intersection between two points
GLfloat CLS::
GetOffset(GLfloat fValue1, GLfloat fValue2, GLfloat fValueDesired){
    GLdouble fDelta = fValue2 - fValue1;
    if (fDelta == 0.0) return 0.5;

    GLfloat offset = (fValueDesired - fValue1)/fDelta;

    if (offset < 0.02) offset = 0.02;
    if (offset > 0.98) offset = 0.98;

    return offset;
}

float CLS::
GetInterpolatedValue(const unsigned char *pData,
		     const int lBaseIndex, 
		     const float fx, const float fy, const float fz){

    // ŌNomalizêŁANormalizesv
    // Ǐ֐ǂނقo֐ǂ݂ɍs葬íHj
    
    int lOffsetZ(mOffsetZ);
    int lOffsetY(mOffsetY);

    if (lBaseIndex + lOffsetZ > mDataBoxFullSize) 
	lOffsetZ = -mOffsetZ * (mSize[Z] -1 );
    
    const float P000(pData[lBaseIndex]);
    const float P001(pData[lBaseIndex+1]);
    const float P010(pData[lBaseIndex+lOffsetY]);
    const float P011(pData[lBaseIndex+lOffsetY+1]);
    const float P100(pData[lBaseIndex+lOffsetZ]); 
    const float P101(pData[lBaseIndex+lOffsetZ+1]);
    const float P110(pData[lBaseIndex+lOffsetZ+lOffsetY]);
    const float P111(pData[lBaseIndex+lOffsetZ+lOffsetY+1]);
    
    return  
	(1.0f-fz) * ( (1.0f-fy) * ((1.0f-fx)*P000 + fx*P001)  +
		      fy        * ((1.0f-fx)*P010 + fx*P011)) +
	fz        * ( (1.0f-fy) * ((1.0f-fx)*P100 + fx*P101)  +
		      fy        * ((1.0f-fx)*P110 + fx*P111));
}

float CLS::
GetInterpolatedValue2(const unsigned char *pData,
		      int x, int y, int z,
		      const float fx, const float fy, const float fz){

    // ŌNomalizêŁANormalizesv
    // Ǐ֐ǂނقo֐ǂ݂ɍs葬íHj

    if (x > mSize[X] - 1)
	x -= mSize[X];
    if (y > mSize[Y] - 1)
	y -= mSize[Y];
    if (z > mSize[Z] - 1)
	z -= mSize[Z];
    
    const int lBaseIndex = mOffsetZ * z + mOffsetY * y + x;
    
    int offsetX = 1;
    int offsetY = mOffsetY;
    int offsetZ = mOffsetZ;
    if (x == mSize[X] - 1) offsetX = -x;
    if (y == mSize[Y] - 1) offsetY = -mOffsetY * y;
    if (z == mSize[Z] - 1) offsetZ = -mOffsetZ * z;

    const float P000(pData[lBaseIndex                               ]);
    const float P001(pData[lBaseIndex                      + offsetX]);
    const float P010(pData[lBaseIndex           + offsetY           ]);
    const float P011(pData[lBaseIndex           + offsetY  + offsetX]);
    const float P100(pData[lBaseIndex + offsetZ                     ]); 
    const float P101(pData[lBaseIndex + offsetZ            + offsetX]);
    const float P110(pData[lBaseIndex + offsetZ + offsetY           ]);
    const float P111(pData[lBaseIndex + offsetZ + offsetY  + offsetX]);
    
    return  
	(1.0f-fz) * ( (1.0f-fy) * ((1.0f-fx)*P000 + fx*P001)  +
		      fy        * ((1.0f-fx)*P010 + fx*P011)) +
	fz        * ( (1.0f-fy) * ((1.0f-fx)*P100 + fx*P101)  +
		      fy        * ((1.0f-fx)*P110 + fx*P111));
}

void CLS::
GetNormalPrecise2(GLfloat ANorm[4], const unsigned char *pData, 
		 const int x, const GLfloat fX, 
		 const int y, const GLfloat fY,
		 const int z, const GLfloat fZ){
    
    GLfloat fxplus = fX + 0.5, fxminus = fX - 0.5; 
    int xplus = x, xminus = x;
    if (fxplus > 1) {
	fxplus  -= 1; xplus++; }
    if (fxminus < 0) {
	fxminus += 1; xminus--; 
	if (xminus < 0) xminus += mSize[X];
    }
    ANorm[X] = 
	(GetInterpolatedValue2(pData, xminus, y, z, fxminus, fY, fZ)
	 - GetInterpolatedValue2(pData, xplus, y, z, fxplus, fY, fZ));

    GLfloat fyplus = fY + 0.5, fyminus = fY - 0.5; 
    int yplus = y, yminus = y;
    if (fyplus > 1) {
	fyplus  -= 1; yplus++; }
    if (fyminus < 0) {
	fyminus += 1; yminus--; 
	if (yminus < 0) yminus += mSize[Y];
    }
    ANorm[Y] = 
	(GetInterpolatedValue2(pData, x, yminus, z, fX, fyminus, fZ)
	 - GetInterpolatedValue2(pData, x, yplus, z, fX, fyplus, fZ));

    GLfloat fzplus = fZ + 0.5, fzminus = fZ - 0.5; 
    int zplus = z, zminus = z;
    if (fzplus > 1) {
	fzplus  -= 1; zplus++; }
    if (fzminus < 0) {
	fzminus += 1; zminus--; 
	if (zminus < 0) zminus += mSize[Z];
    }
    ANorm[Z] = 
	(GetInterpolatedValue2(pData, x, y, zminus, fX, fY, fzminus)
	 - GetInterpolatedValue2(pData, x, y, zplus, fX, fY, fzplus));
    
    LOOP_XYZ ANorm[xyz] *= mSize[xyz];

    float norm = sqrt(ANorm[X]*ANorm[X] + ANorm[Y]*ANorm[Y] + ANorm[Z]*ANorm[Z]);

    if (norm != 0){ // ̏ꂵ̂if 
	ANorm[X] /= norm;
	ANorm[Y] /= norm;
	ANorm[Z] /= norm;
    }
    else
	ANorm[Z] = 1;
}

void CLS::
GetNormalPrecise(GLfloat ANorm[4], const unsigned char *pData, 
		 const int x, const GLfloat fX, 
		 const int y, const GLfloat fY,
		 const int z, const GLfloat fZ){
    // FIX_ME!! ŜƂĂȂ肻̏ꂵ̂
    const float baseValue
	(GetInterpolatedValue(pData, mBaseIndex, fX, fY, fZ));
    const int baseIndex = mBaseIndex;

    ANorm[X] = ANorm[Y] = ANorm[Z] = baseValue;
    
    ANorm[X] -= GetInterpolatedValue2(pData, x+1, y, z, fX, fY, fZ);
    ANorm[Y] -= GetInterpolatedValue2(pData, x, y+1, z, fX, fY, fZ);
    ANorm[Z] -= GetInterpolatedValue2(pData, x, y, z+1, fX, fY, fZ);

    LOOP_XYZ ANorm[xyz] *= mSize[xyz];

    float norm = sqrt(ANorm[X]*ANorm[X] + ANorm[Y]*ANorm[Y] + ANorm[Z]*ANorm[Z]);

    if (norm != 0){ // ̏ꂵ̂if 
	ANorm[X] /= norm;
	ANorm[Y] /= norm;
	ANorm[Z] /= norm;
    }
    else
	ANorm[Z] = 1;
}

void CLS::
ProcessBorder(const int x, const int y, const int z,
	      const unsigned char *pDataIndex, GLfloat afCubeValue[8]){
    GLint *pOffsetIndex;
    GLint (*pOffsetIndexPos)[3];

    int lOffsetZ(mOffsetZ);
    int lOffsetY(mOffsetY);

    if (x < mSize[X] - 1 && y < mSize[Y] - 1 && z < mSize[Z] - 1){
	pOffsetIndexPos = VOffset;
    }
    else{
	pOffsetIndexPos = new GLint[8][3];
	memcpy(pOffsetIndexPos, &VOffset[0][0], sizeof(GLint) * 24);
	if (x == mSize[X] - 1){
	    pOffsetIndexPos[1][0] = -(mSize[X] - 1);
	    pOffsetIndexPos[2][0] = -(mSize[X] - 1);
	    pOffsetIndexPos[5][0] = -(mSize[X] - 1);
	    pOffsetIndexPos[6][0] = -(mSize[X] - 1);
	}
	if (y == mSize[Y] - 1){
	    pOffsetIndexPos[2][1] = -(mSize[Y] - 1);
	    pOffsetIndexPos[3][1] = -(mSize[Y] - 1);
	    pOffsetIndexPos[6][1] = -(mSize[Y] - 1);
	    pOffsetIndexPos[7][1] = -(mSize[Y] - 1);
	}
	if (z == mSize[Z] - 1){
	    pOffsetIndexPos[4][2] = -(mSize[Z] - 1);
	    pOffsetIndexPos[5][2] = -(mSize[Z] - 1);
	    pOffsetIndexPos[6][2] = -(mSize[Z] - 1);
	    pOffsetIndexPos[7][2] = -(mSize[Z] - 1);
	}
    }
    
    pOffsetIndex = &pOffsetIndexPos[0][0];

    const GLint *pOffsetIndexEnd = pOffsetIndex + 24;
    GLfloat *pfCubeValueIndex = afCubeValue;
    for (; pOffsetIndex < pOffsetIndexEnd; 
	 pOffsetIndex+=3, pfCubeValueIndex++){
	*pfCubeValueIndex = (GLfloat) *(pDataIndex + 
					pOffsetIndex[Z]*lOffsetZ +
					pOffsetIndex[Y]*lOffsetY +
					pOffsetIndex[X]);
    }

    if (pOffsetIndexPos != VOffset) delete [] pOffsetIndexPos;
}

int CLS::
MarchCubeMainThread(const unsigned char *pData, 
		    const int x, const int y, const int z){
    
    int *pSize = mSize;
    int arrayIndex = mArrayIndex;
    float * const pTmpVertices = mpTmpVertices;
    float * const pTmpNormal = mpTmpNormal;
    const float targetValue255 = mTargetValue255;
    
    GLint iFlagIndex(0), iEdgeFlags;
    GLfloat afCubeValue[8];

    const int lOffsetZ(mOffsetZ);
    const int lOffsetY(mOffsetY);
    mBaseIndex = z*lOffsetZ + y*lOffsetY + x;
	
    //Make a local copy of the values at the cube's corners
    const unsigned char *pDataIndex = pData + mBaseIndex;

    // afcube Ƀf[^BE͎E
    ProcessBorder(x, y, z, pDataIndex, afCubeValue);

    //Find which vertices are inside of the surface and which are outside
    if(afCubeValue[0] <= targetValue255) iFlagIndex |= 1;
    if(afCubeValue[1] <= targetValue255) iFlagIndex |= 2;
    if(afCubeValue[2] <= targetValue255) iFlagIndex |= 4;
    if(afCubeValue[3] <= targetValue255) iFlagIndex |= 8;
    if(afCubeValue[4] <= targetValue255) iFlagIndex |= 16;
    if(afCubeValue[5] <= targetValue255) iFlagIndex |= 32;
    if(afCubeValue[6] <= targetValue255) iFlagIndex |= 64;
    if(afCubeValue[7] <= targetValue255) iFlagIndex |= 128;
    
    iEdgeFlags = aiCubeEdgeFlags[iFlagIndex];
    if(iEdgeFlags == 0) {
	if (afCubeValue[0] > targetValue255)
	    return 1;
	else
	    return 2;
    }

    GLfloat fOffset;
    GLfloat AAEdgeVertex[12][3];
    GLfloat AAEdgeNorm[12][4]; // SSÊ߂4߂̓_~[

    //Find the point of intersection of the surface with each edge
    //Then find the normal to the surface at those points

    // if [v̊O
    bool mbPrecise = TRUE;
    if (mbPrecise == TRUE)
	for(int iEdge = 0; iEdge < 12; iEdge++) {
	    const GLint *pEdgeConnection = a2iEdgeConnection[iEdge];
	    if(iEdgeFlags & (1<<iEdge)) { //if intersection on this edge
		fOffset = 
		    GetOffset(afCubeValue[ pEdgeConnection[0] ], 
			      afCubeValue[ pEdgeConnection[1] ], 
			      targetValue255);
		
		// 0-1͈̔͂ł̍W
		float fX = 
		    VOffset[pEdgeConnection[0]][X] +
		    fOffset * a2fEdgeDirection[iEdge][X];
		float fY = 
		    VOffset[pEdgeConnection[0]][Y] +
		    fOffset * a2fEdgeDirection[iEdge][Y];
		float fZ = 
		    VOffset[pEdgeConnection[0]][Z] +
		    fOffset * a2fEdgeDirection[iEdge][2];
		
		AAEdgeVertex[iEdge][X] = (float)(x + fX) / pSize[X];
		AAEdgeVertex[iEdge][Y] = (float)(y + fY) / pSize[Y];
		AAEdgeVertex[iEdge][Z] = (float)(z + fZ) / pSize[Z];
		
		GetNormalPrecise2(AAEdgeNorm[iEdge], pData,
				  x, fX, y, fY, z, fZ);
	    }
	}
    else
	for(int iEdge = 0; iEdge < 12; iEdge++) {
	    const GLint *pEdgeConnection = a2iEdgeConnection[iEdge];
	    if(iEdgeFlags & (1<<iEdge)) { //if intersection on this edge
		fOffset = 
		    GetOffset(afCubeValue[ pEdgeConnection[0] ], 
			      afCubeValue[ pEdgeConnection[1] ], 
			      targetValue255);
		
		// 0-1͈̔͂ł̍W
		float fX = 
		    VOffset[pEdgeConnection[0]][X] +
		    fOffset * a2fEdgeDirection[iEdge][X];
		float fY = 
		    VOffset[pEdgeConnection[0]][Y] +
		    fOffset * a2fEdgeDirection[iEdge][Y];
		float fZ = 
		    VOffset[pEdgeConnection[0]][Z] +
		    fOffset * a2fEdgeDirection[iEdge][2];
		
		AAEdgeVertex[iEdge][X] = (float)(x + fX) / pSize[X];
		AAEdgeVertex[iEdge][Y] = (float)(y + fY) / pSize[Y];
		AAEdgeVertex[iEdge][Z] = (float)(z + fZ) / pSize[Z];
		
		GetNormalPrecise(AAEdgeNorm[iEdge], pData,
				 x, fX, y, fY, z, fZ);
	    }
	}
    
    //Store triangles that were found.  There can be up to five per cube

    const GLint *pConnectionTable = a2iTriangleConnectionTable[iFlagIndex];
    for(int iTriangleIndex = 0; iTriangleIndex < 15; iTriangleIndex+=3) {
	if(pConnectionTable[iTriangleIndex] < 0) break;

	// Corei7ł͌vZANZX肪̂,+3Ƃ+6̌vZĂ
	// ANZX܂Ƃ߂ɑB//Pen4Ɣɒx
	GLint iVertexX = pConnectionTable[iTriangleIndex];
	GLint iVertexY = pConnectionTable[iTriangleIndex+1];
	GLint iVertexZ = pConnectionTable[iTriangleIndex+2];

	memcpy(pTmpVertices + arrayIndex, AAEdgeVertex[iVertexX], 12);
	memcpy(pTmpVertices + arrayIndex+3, AAEdgeVertex[iVertexY], 12);
	memcpy(pTmpVertices + arrayIndex+6, AAEdgeVertex[iVertexZ], 12);
	
	memcpy(pTmpNormal + arrayIndex, AAEdgeNorm[iVertexX], 12);
	memcpy(pTmpNormal + arrayIndex+3, AAEdgeNorm[iVertexY], 12);
	memcpy(pTmpNormal + arrayIndex+6, AAEdgeNorm[iVertexZ], 12);

	arrayIndex+=9;
    }
    
    mArrayIndex = arrayIndex;
    
    return 0; // EʗL
}

int CLS::
MarchCubeLOOPX1(const unsigned char *pData, int y, int z, 
		int startX, int finX){
    // 0EL 1,2  E & targetValue ̏㉺ bitFiledI
    //int FinX = mSize[X] - 1;
    int prevStateX = 0, currStateX = 0;
    int boxnum = 0;
    
    for (int x = startX + 1; x < finX; x+=2){
	prevStateX = currStateX;
	currStateX = MarchCubeMainThread(mpData, x, y, z);
	boxnum += (currStateX == 0);
	if (currStateX != 0 &&
	    prevStateX == currStateX) continue;
	boxnum += (MarchCubeMainThread(mpData, x-1, y, z) == 0);
    }
    if ((finX - startX) % 2 == 1)
	boxnum += (MarchCubeMainThread(mpData, finX-1, y, z) == 0);
    
    if (boxnum == 0) return currStateX;
    else return 0;
}

int CLS::
MarchCubeLOOPY(const unsigned char *pData, int z,
	       int startY, int finY, int startX, int finX){
    // 0,1,2  & Ă̂ŁA12̒lɂȂ̂ݏȗ
    int prevStateY = 3, currStateY = 3;
    int linenum = 0;
    for (int y = startY + 1; y < finY; y+=2){
	prevStateY = currStateY;
	currStateY = MarchCubeLOOPX1(mpData, y, z, startX, finX);
	linenum += (currStateY == 0);
	if ((prevStateY == 1 && currStateY == 1) ||
	    (prevStateY == 2 && currStateY == 2)) {
	    continue; }
	
	linenum += (MarchCubeLOOPX1(mpData, y-1, z, startX, finX) == 0);
    }

    if ((finY - startY) % 2 == 1)
	MarchCubeLOOPX1(mpData, finY-1, z, startX, finX);
    
    if (linenum == 0) return currStateY;
    else return 0;
}

// ʂ̏߂XbhŊȂƑʖ
bool CLS::
CopyResults(){
    if (mArrayIndex == 0){
	*mpAllocSize = 0;
	*mpFaceNum = 0;
	return TRUE;
    }

    bool bMallocVertex = TRUE, bMallocNormal = TRUE;
    
    // ߂Ă҂̗̈m
    // xǁA̍ĔzuĂB
    *mppVertices = new float [mArrayIndex];
    *mppNormals = new float [mArrayIndex];
    *mppFaceIndex = new int [mArrayIndex];
    
    memcpy(*mppVertices, mpTmpVertices, sizeof(float) * mArrayIndex);
    memcpy(*mppNormals, mpTmpNormal, sizeof(float) * mArrayIndex);

    for (int i = 0; i < mArrayIndex; i++)
	(*mppFaceIndex)[i] = i;
    
    // arrayIndex = float̐ = triangle(Face)̐*3 * 3(xyz)
    *mpAllocSize = mArrayIndex; 
    *mpFaceNum = mArrayIndex/9;
    
    return TRUE;
}

    
wxThread::ExitCode CLS::
Entry(){
    *mpStatusFlagLink = _THREAD_RUNNING;
    
    mOffsetZ = mSize[X] * mSize[Y];
    mOffsetY = mSize[X];
    mTargetValue255 = mTargetValue * 255.0f;

    mpTmpVertices = new float [mBufferSize];
    mpTmpNormal = new float [mBufferSize];

    for (int z = mInitZ; z < mFinZ; z++)
	MarchCubeLOOPY(mpData, z, mInitY, mFinY, mInitX, mFinX);
    
    CopyResults();

    delete [] mpTmpVertices;
    delete [] mpTmpNormal;

    *mpStatusFlagLink = _THREAD_FINISHED;

    return 0;
}

//lists the positions, relative to vertex0, of each of the 8 vertices
GLint CLS::VOffset[8][3] = {
    {0, 0, 0},{1, 0, 0},{1, 1, 0},{0, 1, 0},
    {0, 0, 1},{1, 0, 1},{1, 1, 1},{0, 1, 1}
};

//lists the index of the endpoint vertices for each of the 12 edges
const GLint CLS::a2iEdgeConnection[12][2] = {
    {0,1}, {1,2}, {2,3}, {3,0},
    {4,5}, {5,6}, {6,7}, {7,4},
    {0,4}, {1,5}, {2,6}, {3,7}
};

//lists the direction vector (vertex1-vertex0) for each edge
const GLfloat CLS::a2fEdgeDirection[12][3] = {
    {1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{-1.0, 0.0, 0.0},{0.0, -1.0, 0.0},
    {1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{-1.0, 0.0, 0.0},{0.0, -1.0, 0.0},
    {0.0, 0.0, 1.0},{0.0, 0.0, 1.0},{ 0.0, 0.0, 1.0},{0.0,  0.0, 1.0}
};

// _[8]  256 ʂ̏ԂA_ׂ̂[12]bitmaskœ
const GLint CLS::aiCubeEdgeFlags[256]= {
    0x000, 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c, 0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00, 
    0x190, 0x099, 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c, 0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90, 
    0x230, 0x339, 0x033, 0x13a, 0x636, 0x73f, 0x435, 0x53c, 0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30, 
    0x3a0, 0x2a9, 0x1a3, 0x0aa, 0x7a6, 0x6af, 0x5a5, 0x4ac, 0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0, 
    0x460, 0x569, 0x663, 0x76a, 0x066, 0x16f, 0x265, 0x36c, 0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60, 
    0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0x0ff, 0x3f5, 0x2fc, 0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0, 
    0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x055, 0x15c, 0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950, 
    0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0x0cc, 0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0, 
    0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc, 0x0cc, 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0, 
    0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c, 0x15c, 0x055, 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650, 
    0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc, 0x2fc, 0x3f5, 0x0ff, 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0, 
    0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c, 0x36c, 0x265, 0x16f, 0x066, 0x76a, 0x663, 0x569, 0x460, 
    0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac, 0x4ac, 0x5a5, 0x6af, 0x7a6, 0x0aa, 0x1a3, 0x2a9, 0x3a0, 
    0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c, 0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x033, 0x339, 0x230, 
    0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c, 0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x099, 0x190, 
    0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c, 0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x000
};

//  For example: a2iTriangleConnectionTable[3] list the 2 triangles formed when corner[0] 
//  and corner[1] are inside of the surface, but the rest of the cube is not.
const GLint CLS::a2iTriangleConnectionTable[256][16] =  {
    {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
    {3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
    {3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
    {3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
    {9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
    {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
    {8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
    {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
    {3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
    {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
    {4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
    {4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
    {5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
    {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
    {9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
    {0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
    {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
    {10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
    {5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
    {5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
    {9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
    {1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
    {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
    {8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
    {2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
    {7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
    {2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
    {11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
    {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
    {11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
    {11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
    {9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
    {2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
    {6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
    {3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
    {6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
    {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
    {6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
    {8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
    {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
    {3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
    {0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
    {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
    {8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
    {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
    {0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
    {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
    {10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
    {10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
    {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
    {1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
    {0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
    {10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
    {3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
    {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
    {9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
    {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
    {3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
    {6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
    {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
    {10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
    {10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
    {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
    {7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
    {7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
    {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
    {1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
    {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
    {8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
    {0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
    {7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
    {10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
    {2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
    {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
    {7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
    {2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
    {10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
    {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
    {0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
    {7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
    {6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
    {8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
    {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
    {6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
    {4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
    {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
    {8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
    {0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
    {1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
    {8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
    {10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
    {4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
    {10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
    {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
    {9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
    {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
    {7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
    {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
    {7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
    {3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
    {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
    {9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
    {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
    {4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
    {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
    {6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
    {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
    {0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
    {6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
    {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
    {11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
    {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
    {5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
    {9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
    {1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
    {1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
    {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
    {0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
    {5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
    {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
    {11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
    {9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
    {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
    {2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
    {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
    {9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
    {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
    {1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
    {9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
    {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
    {0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
    {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
    {2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
    {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
    {0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
    {9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
    {5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
    {3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
    {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
    {8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
    {0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
    {9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
    {1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
    {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
    {4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
    {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
    {11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
    {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
    {2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
    {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
    {3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
    {1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
    {4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
    {3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
    {0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
    {1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
};
