// (setq c-basic-offset 4)
// "clsMarchingCube.h"

#include "clsMarchingCube.h"
#include "clsMCThread.h"
#include "clsOptimizeThread.h"

#include "../../../SIMDATA/clsVolumeForRender.h"

#include "../../../COMMON/CommonDefinitions.h"

#define CLS MarchingCube_cls

CLS::
MarchingCube_cls(){
}

CLS::
~MarchingCube_cls(){
    Clear();
}

void CLS::
Clear(){
    uint sectionNum = mVpVertices.size();
    for (uint section = 0; section < sectionNum; section++){
	if (mVAllocSize[section] != 0){
	    delete [] mVpVertices[section];
	    delete [] mVpNormals[section];
	    delete [] mVpFaceIndex[section];
	    mVAllocSize[section] = 0;
	    mVFaceNum[section] = 0;
	}
    }

    mVpVertices.clear();
    mVpNormals.clear();
    mVpFaceIndex.clear();
    mVAllocSize.clear();
    mVFaceNum.clear();
    mVVertexNumOrg.clear();

    int section = 8*8*8;
    mVpVertices.resize(section);
    mVpNormals.resize(section);
    mVpFaceIndex.resize(section);
    mVAllocSize.resize(section);
    mVFaceNum.resize(section);
    mVVertexNumOrg.resize(section);

    for (int i = 0; i < section; i++){
	mVpVertices[i] = NULL;
	mVpNormals[i] = NULL;
	mVpFaceIndex[i] = NULL;
	mVAllocSize[i] = 0;
	mVFaceNum[i] = 0;
	mVVertexNumOrg[i] = 0;
    }
}

std::vector < unsigned char > CLS::
GetMinMaxValue(unsigned char *pData, 
	       int initZ, int finZ, int initY, int finY, int initX, int finX){

    unsigned char max = 0;
    unsigned char min = 255;
    int sizeX = mSize[X], sizeY = mSize[Y], sizeZ = mSize[Z];

    for (int z = initZ; z < finZ; z++){
	int offsetZ = z * sizeX * sizeY;
	for (int y = initY; y < finY; y++){
	    int offsetY = offsetZ + y * sizeX;
	    for (int x = initX; x < finX; x++){
		int offset = offsetY + x;
		if (pData[offset] > max) max = pData[offset];
		if (pData[offset] < min) min = pData[offset];
	    }
	}
    }

    std::vector < unsigned char > retVec; 
    retVec.push_back(min); 
    retVec.push_back(max);
    
    return retVec;
}

void CLS::
CalcMaxAndMin(unsigned char *pDataLink){
    int verticesIndex = 0;
    
    for (int indexZ = 0; indexZ < 8; indexZ++){
	int initZ = mSize[Z] * indexZ / 8;
	int finZ  = mSize[Z] * (indexZ+1) / 8;
	if (indexZ != 7) finZ++; // ̃I[o[bv

	for (int indexY = 0; indexY < 8; indexY++){
	    int initY = mSize[Y] * indexY /8;
	    int finY  = mSize[Y] * (indexY+1) /8;
	    if (indexY != 7) finY++;
	    
	    for (int indexX = 0; indexX < 8; indexX++){
		int initX = mSize[X] * indexX / 8;
		int finX  = mSize[X] * (indexX+1) / 8;
		if (indexX != 7) finX++;
		
		std::vector < unsigned char> VMinMax = GetMinMaxValue
		    (pDataLink, initZ, finZ, initY, finY, initX, finX);
		mSectionMinValue[indexZ][indexY][indexX] = VMinMax[0];
		mSectionMaxValue[indexZ][indexY][indexX] = VMinMax[1];
	    }
	}
    }
}

bool CLS::
IsItSameVolumeData(VolumeFoRender_cls *pCVolumeForRender){
    if (mstrSnap1LoadedBody == pCVolumeForRender->mstrSnap1LoadedBody &&
	mstrSnap2LoadedBody == pCVolumeForRender->mstrSnap2LoadedBody){
	if (mstrSnap2LoadedBody.IsEmpty() == TRUE)
	    return TRUE;
	if (mFactor == pCVolumeForRender->mFactor)
	    return TRUE;
    }
    return FALSE;
}

void CLS::
ConstructVertices(VolumeFoRender_cls *pCVolumeForRender,
		  float threshold, bool bOptimize){

    mTargetValue = threshold;
    
    // f[^ςȂĂ悢
    if (IsItSameVolumeData(pCVolumeForRender) == TRUE &&
	mPrepareCounterUsed == pCVolumeForRender->mPrepareCounter &&
	mTargetValueUsed == mTargetValue &&
	mbOptimizeUsed == bOptimize)
	return;

    mpCVolumeForRenderUsed = pCVolumeForRender;
    mstrSnap1LoadedBody = pCVolumeForRender->mstrSnap1LoadedBody;
    mstrSnap2LoadedBody = pCVolumeForRender->mstrSnap2LoadedBody;
    mFactor             = pCVolumeForRender->mFactor;
    mPrepareCounterUsed = pCVolumeForRender->mPrepareCounter;
    mTargetValueUsed = mTargetValue;
    mbOptimizeUsed   = bOptimize;

    Clear();

    LOOP_XYZ mSize[xyz] = pCVolumeForRender->mSize[xyz];
    unsigned char *pDataLink = pCVolumeForRender->mpData_uchar;

    if (pDataLink == NULL) return;

    // őlƍŏl𓾂
    bool bRecalcMax = TRUE;
    if (bRecalcMax == TRUE)
	CalcMaxAndMin(pDataLink);

    ConstructThreadControl(pDataLink);

    for (int i = 0; i < 8*8*8; i++)
	mCMesh2[i].ClearAll();
    
    if (bOptimize == TRUE)
	OptimizeThreadControl();
}

void CLS::
SetVBO(bool bUseSV){

    // gf[^ςĂAAhX
    // PrepareCounterωĂ͂
    // AhX̂͊ɔjāAԒnŃ`FbNĂ̂ɒ
    if (mpCVolumeForRenderVBO == mpCVolumeForRenderUsed &&
	mPrepareCounterVBO == mPrepareCounterUsed &&
	mTargetValueVBO == mTargetValueUsed &&
	mbOptimizeVBO == mbOptimizeUsed &&
	mbUseSVVBO == bUseSV) 
	return;

    mpCVolumeForRenderVBO = mpCVolumeForRenderUsed;
    mPrepareCounterVBO = mPrepareCounterUsed;
    mTargetValueVBO = mTargetValueUsed;
    mbOptimizeVBO = mbOptimizeUsed;
    mbUseSVVBO = bUseSV;
   
    int sectionNum = mVpVertices.size();
    for (int section = 0; section < sectionNum; section++){
	if (mVAllocSize[section] != 0){
	    mCMesh2[section].ReadBeginFromData();
	    mCMesh2[section].ReadAppendFromData(mVAllocSize[section], mVFaceNum[section], mVpVertices[section], mVpNormals[section], mVpFaceIndex[section]);
	    mCMesh2[section].ReadFinishFromData(bUseSV);
	    mCMesh2[section].SetVBO(bUseSV);
	}
    }
}

void CLS::
ConstructThreadControl(unsigned char* pDataLink){

    // Kvobt@ _}bNX1box5*3(*3)iႤHjǁA
    int bufferSize = 
	sizeof(float) * (mSize[X]/8)*(mSize[Y]/8)*(mSize[Z]/8)* 12;

    int *pStatusList = GetNewPStatusList(8*8*8);
    SetThreadNumber(12);
    
    int section = 0;
    for (int indexZ = 0; indexZ < 8; indexZ++){
	int initZ = mSize[Z] * indexZ / 8;
	int finZ  = mSize[Z] * (indexZ+1) / 8;
	if (finZ > mSize[Z]) finZ = mSize[Z]; 

	for (int indexY = 0; indexY < 8; indexY++){
	    int initY = mSize[Y] * indexY /8;
	    int finY  = mSize[Y] * (indexY+1) /8;
	    if (finY > mSize[Y]) finY = mSize[Y];
	    
	    for (int indexX = 0; indexX < 8; indexX++){
		
		int min = mSectionMinValue[indexZ][indexY][indexX];
		int max = mSectionMaxValue[indexZ][indexY][indexX];
		int tgt = mTargetValue * 255;
		if (tgt < min || tgt > max){
		    pStatusList[section] = _THREAD_FINISHED;
		    section++;
		    continue;
		}

		int initX = mSize[X] * indexX / 8;
		int finX  = mSize[X] * (indexX+1) / 8;
		if (finX > mSize[X]) finX = mSize[X];
		
		MCThread_cls* pCThread = new MCThread_cls
		    (pDataLink, 
		     &mVAllocSize[section], &mVFaceNum[section],
		     &mVpVertices[section], &mVpNormals[section], 
		     &mVpFaceIndex[section], bufferSize);
		pCThread->SetParameters(mSize,
					initZ, finZ, initY, finY, initX, finX,
					mTargetValue);
		pCThread->CreateThread(&pStatusList[section]);
		
		WaitForThread(pStatusList);
		pCThread->Run();
		
		section++;
	    }

	}
    }

    WaitFinishAndDeleteFlag(pStatusList);
}

void CLS::
OptimizeThreadControl(){

    int *pStatusList = GetNewPStatusList(8*8*8);
    SetThreadNumber(12);
    
    int sectionNum = mVpVertices.size();
    for (int section = 0; section < sectionNum; section++){

	if (mVAllocSize[section] == 0){
	    pStatusList[section] = _THREAD_FINISHED;
	    continue;
	}

	OptimizeThread_cls* pCThread = new OptimizeThread_cls
	    (&mVAllocSize[section], &mVFaceNum[section],
	     &mVpVertices[section], &mVpNormals[section], 
	     &mVpFaceIndex[section]);

	//pCThread->SetParameters();
	pCThread->CreateThread(&pStatusList[section]);
	     
	WaitForThread(pStatusList);
	pCThread->Run();
    }

    WaitFinishAndDeleteFlag(pStatusList);
}

void CLS::
GLDraw(bool bUseVBO){

    if (bUseVBO == TRUE){
	for (int i = 0; i < 8*8*8; i++)
	    mCMesh2[i].GLDraw(TRUE, FALSE, FALSE);
	return;
    }

    int sectionNum = mVpVertices.size();
    for (int section = 0; section < sectionNum; section++){

	int faceNum = mVFaceNum[section];
	if (faceNum == 0) continue;
	glBegin(GL_TRIANGLES);
	for (int i = 0; i < faceNum; i++){
	    int index0 = mVpFaceIndex[section][i*3];
	    int index1 = mVpFaceIndex[section][i*3+1];
	    int index2 = mVpFaceIndex[section][i*3+2];
	    glNormal3fv(&mVpNormals[section][index0*3]);
	    glVertex3fv(&mVpVertices[section][index0*3]);
	    glNormal3fv(&mVpNormals[section][index1*3]);
	    glVertex3fv(&mVpVertices[section][index1*3]);
	    glNormal3fv(&mVpNormals[section][index2*3]);
	    glVertex3fv(&mVpVertices[section][index2*3]);
	}
	glEnd();
    }
}

void CLS::
GLDrawSV(int section){
    mCMesh2[section].mCVBO.Draw(_VBO_SHADOW_VOLUME, TRUE);
}
 
