// (setq c-basic-offset 4)
// "clsOptimizeThread.h"

#include "clsOptimizeThread.h"

#include <algorithm>

#include "../../../MISC/clsMisc.h"

#include "../../../COMMON/clsArrayAndIndex.h"
#include "../../../COMMON/CommonDefinitions.h"

#define CLS OptimizeThread_cls

CLS::
OptimizeThread_cls(int *pAllocSize, int *pFaceNum,
		   float **ppVertices, float **ppNormals, int **ppFaceIndex) :
    mpAllocSize(pAllocSize), mpFaceNum(pFaceNum) {
    mppVertices  = ppVertices;
    mppNormals   = ppNormals;
    mppFaceIndex = ppFaceIndex;
}

wxThread::ExitCode CLS::
Entry(){
    *mpStatusFlagLink = _THREAD_RUNNING;

    int vertexSize = *mpAllocSize / 3;
    float *pVertices = *mppVertices;
    float *pNormals = *mppNormals;

    ArrayAndIndex_cls *pCIdxAndZ = new ArrayAndIndex_cls[vertexSize];
    
    float *pVertexTmp = new float [vertexSize*3];
    float *pNormalTmp = new float [vertexSize*3];
    
    for (int i = 0; i < vertexSize; i++){
	pCIdxAndZ[i].index = i;
	pCIdxAndZ[i].z = pVertices[i*3+0] + pVertices[i*3+1] + pVertices[i*3+2];
    }

    std::sort(pCIdxAndZ, pCIdxAndZ+vertexSize);

    pVertexTmp[0] = -1;
    pVertexTmp[1] = -1;
    pVertexTmp[2] = -1;

    int packIndex = 0, packIndexPrev = 0;
    for (int i_ = 0; i_ < vertexSize; i_++){
	int i = pCIdxAndZ[i_].index;
	
	if ((Misc_cls::IsNearlySame
	     (pVertices[i*3+0], pVertexTmp[(packIndex-1)*3+0], (float)1e-7) == TRUE &&
	     Misc_cls::IsNearlySame
	     (pVertices[i*3+1], pVertexTmp[(packIndex-1)*3+1], (float)1e-7) == TRUE &&
	     Misc_cls::IsNearlySame
	     (pVertices[i*3+2], pVertexTmp[(packIndex-1)*3+2], (float)1e-7) == TRUE)){
	    (*mppFaceIndex)[i] = packIndexPrev;
	}
	else{
	    pVertexTmp[packIndex*3+0] = pVertices[i*3+0];
	    pVertexTmp[packIndex*3+1] = pVertices[i*3+1];
	    pVertexTmp[packIndex*3+2] = pVertices[i*3+2];
	    pNormalTmp[packIndex*3+0] = pNormals[i*3+0];
	    pNormalTmp[packIndex*3+1] = pNormals[i*3+1];
	    pNormalTmp[packIndex*3+2] = pNormals[i*3+2];
	    (*mppFaceIndex)[i] = packIndex;
	    packIndexPrev = packIndex;
	    packIndex++;
	}
    }

    float *pVerticesRep = new float [packIndex*3];
    float *pNormalsRep = new float [packIndex*3];
    
    /*
    for (int i = 0; i < packIndex; i++){
	pVerticesRep[i*3+0] = pVertexTmp[i*3+0];
	pVerticesRep[i*3+1] = pVertexTmp[i*3+1];
	pVerticesRep[i*3+2] = pVertexTmp[i*3+2];
	pNormalsRep[i*3+0] = pNormalTmp[i*3+0];
	pNormalsRep[i*3+1] = pNormalTmp[i*3+1];
	pNormalsRep[i*3+2] = pNormalTmp[i*3+2];
    }
    */
    
    memcpy(pVerticesRep, pVertexTmp, sizeof(float) * 3 * packIndex);
    memcpy(pNormalsRep, pNormalTmp, sizeof(float) * 3 * packIndex);

    delete [] *mppVertices;
    *mppVertices = pVerticesRep;

    delete [] *mppNormals;
    *mppNormals = pNormalsRep;

    delete [] pCIdxAndZ;

    *mpAllocSize = packIndex*3;

    delete [] pVertexTmp;
    delete [] pNormalTmp;
    
    *mpStatusFlagLink = _THREAD_FINISHED;
    return 0;
}
