// (setq-c-offset 4)
// "clsPrimitive.cpp"

#include "clsColorCurve.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"
#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../UNDO/clsUndoManager.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"

#ifdef WIN32
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
#endif

#include <sstream>

#define CLS ColorCurve_cls

CLS::
ColorCurve_cls(wxString strName, wxArrayString AstrItemName) : Attribute_cls("ColorCurve_cls", strName),  ParamGroup_cls(strName, AstrItemName) {
    wxString strConfListLabel = "SceneObj::"; strConfListLabel << strName << "::ColorCurve";

    mCConfList.SetTag(strConfListLabel);
    
    ParamGroup_cls *pCXRange = GetPCNewParamGroup("Range", "min", "max");
    InitializeAndPushParam(pCXRange, 0, 1);
    
    mTexture1D = 0;
    mbForColorCurve = TRUE;
}

CLS::
~ColorCurve_cls(){
    ClearTexture();
}

void CLS::
Paste(wxString& strData, wxString strTag){
    // ̖ÕJ[J[uR̃f[^ïႢĵŔώG

    ColorCurve_cls *pCNewColorCurve = new ColorCurve_cls(GetStrName(), GetPCStringProvider()->GetArrayString("RGBA"));
    
    pCNewColorCurve->CopyColorCurveFromString(strData, FALSE);

    CopyColorCurveCommand(pCNewColorCurve, strTag, TRUE);

    delete pCNewColorCurve;
}

void CLS::
CopyColorCurveCommand(ColorCurve_cls *pCNew, wxString strTag, bool bAlpha){
    strTag << "::";
    
    std::vector < double > VAlpha;
    if (bAlpha == FALSE){
	uint keyNum = pCNew->mVCParam[3].mVCKeys.size();
	for (uint i = 0; i < keyNum; i++){
	    VAlpha.push_back(pCNew->mVCParam[3].mVCKeys[i].mValue);
	    pCNew->mVCParam[3].mVCKeys[i].mValue = 1;
	}
    }

    wxString strRange;
    GetPCParamGroup("Range")->Serialize(strRange);
    
    std::wstring strBuf(strRange.c_str());
    std::wistringstream sstream(strBuf);
    std::wstring wstrLine;
    std::getline (sstream, wstrLine);
    while ( std::getline (sstream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
	pCNew->GetPCParamGroup("Range")->ReadStrLine(sstream, strFirstItem, strLine);
    }
	
    wxString strNameOrg = pCNew->mCstrName.GetBodySTR();
    pCNew->mCstrName.SetBodyFromData(mCstrName.GetBodySTR());

    wxString strUndo(strTag), strRedo(strTag);

    Serialize(strUndo);
    pCNew->Serialize(strRedo); 
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();

    if (bAlpha == FALSE){
	uint keyNum = pCNew->mVCParam[3].mVCKeys.size();
	for (uint i = 0; i < keyNum; i++){
	    pCNew->mVCParam[3].mVCKeys[i].mValue = VAlpha[i];
	}
    }
}

void CLS::
CopyColorCurveFromP(ColorCurve_cls *pCFrom){
    wxString strOut;
    pCFrom->Serialize(strOut);

    CopyColorCurveFromString(strOut);
    
    /*
    std::wstring strBuf(strOut.c_str());
    std::wistringstream sstream(strBuf);
    std::wstring wstrLine;
    while ( std::getline (sstream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
	ReadStrLine(sstream, strFirstItem, strLine);
    }
    */
}

void CLS::
CopyColorCurveFromString(wxString strSerialize, bool bSelf){
    std::wstring strBuf(strSerialize.c_str());
    std::wistringstream sstream(strBuf);
    std::wstring wstrLine;
    while ( std::getline (sstream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
	ReadStrLine(sstream, strFirstItem, strLine, bSelf);
    }
}

void CLS::
ClearTexture(){
    if (mTexture1D != 0){
	glDeleteTextures(1, &mTexture1D);
	mTexture1D = 0;
    }
}

double CLS::
GetRemapX(double x0){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < double > VRange = GetPCParamGroup("Range")->GetVInterpolatedValue(NT);
    double width = VRange[1] - VRange[0];
    if (width == 0) return 1;
    return (x0 - VRange[0])/width;
}

wxColour CLS::
GetColor(double x0, double *pAlpha){
    std::vector < double > VClr = GetVInterpolatedValue(x0);
    for (int i = 0; i < 4; i++){
	if (VClr[i] < 0) VClr[i] = 0;
	if (VClr[i] > 1) VClr[i] = 1;
    }

    if (pAlpha != NULL) *pAlpha = VClr[3];

    wxColour CClr(VClr[0] * 255, VClr[1] * 255, VClr[2] * 255);
    return CClr;
}

void CLS::
TrimColorVector(std::vector < double > &VClr){
    uint num = VClr.size();
    for (int i = 0; i < num; i++){
	if (VClr[i] < 0) VClr[i] = 0;
	if (VClr[i] > 1) VClr[i] = 1;
    }
}

void CLS::
PrepareTexture(bool bRepeat, float alphaFactor, bool bUseRemap){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < double > VRange = GetPCParamGroup("Range")->GetVInterpolatedValue(NT);
    wxString strProcCounter = GetStrProcCounter();
    strProcCounter << VRange[0] << VRange[1] << bRepeat << alphaFactor << bUseRemap;
    if (strProcCounter.IsSameAs(strCounterAtTexGen) == FALSE){
	ClearTexture();
	strCounterAtTexGen = strProcCounter;
    }

    if (mTexture1D != 0){
	return;
    }

    glGenTextures(1, &mTexture1D);

    int size = _COLOR_CURVE_RESOLUTION+1;
    int alignment = 8;

    unsigned short *pData = new unsigned short [size * 4];
    for (int x = 0; x < size; x++){
	double x0 = (double)x/size;
	double x1 = x0;
	if (bUseRemap == TRUE) x1 = GetRemapX(x0);
	if (bRepeat == TRUE)
	    x1 = x1 - floor(x1);
	std::vector < double > VClr = GetVInterpolatedValue(x1);
	TrimColorVector(VClr);
	pData[x*4 + 0] = VClr[0] * 65535;
	pData[x*4 + 1] = VClr[1] * 65535;
	pData[x*4 + 2] = VClr[2] * 65535;
	float alpha = VClr[3] * alphaFactor;
	if (alpha > 1) alpha = 1;
	   
	pData[x*4 + 3] = alpha * 65535;
    }
    
    glEnable(GL_TEXTURE_1D);
    glBindTexture(GL_TEXTURE_1D, mTexture1D);

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    
    glPixelStorei(GL_UNPACK_ALIGNMENT, alignment);

    bool bCompress = FALSE;
    //int format =  bCompress ? GL_COMPRESSED_ALPHA : GL_ALPHA;
    int format = GL_RGBA;

    glTexImage1D(GL_TEXTURE_1D, 0, format, 
		 size, 0,
		 format, GL_UNSIGNED_SHORT, pData);

    delete [] pData;
}

void CLS::
BindTexture(int active, bool bRepeat){
    glEnable(GL_TEXTURE_1D);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    switch(active){
    case 0:
	glActiveTexture(GL_TEXTURE0); break;
    case 1:
	glActiveTexture(GL_TEXTURE1); break;
    case 2:
	glActiveTexture(GL_TEXTURE2); break;
    case 3:
	glActiveTexture(GL_TEXTURE3); break;
    }
	
    glBindTexture(GL_TEXTURE_1D, mTexture1D);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    if (bRepeat == FALSE)
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    else
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glActiveTexture(GL_TEXTURE0);
}

wxColour CLS::
AlphaBlend(wxColour ClrBase, wxColour ClrBlend, double alpha){
    int r = ClrBase.Red() * (1.0 - alpha) + alpha * ClrBlend.Red();
    int g = ClrBase.Green() * (1.0 - alpha) + alpha * ClrBlend.Green();
    int b = ClrBase.Blue() * (1.0 - alpha) + alpha * ClrBlend.Blue();
    wxColour Clr(r, g, b);
    return Clr;
}

bool CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine,
	    bool bSelf){
    
    // f[^̃Rsył́AO֘AႤf[^̕񂪗̂
    // bSelf őΉ
    
    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE){
	mVCParam[0].mProcCounter++; // Rő\
	return TRUE;
    }
    
    if (strFirstItem.IsSameAs("ReadColorCurve") == FALSE) return FALSE;
    
    //wxString strAttribName = PopStrItemFromStrLine(strLine, "::");
    wxString strName = PopStrItemFromStrLine(strLine, "::");
    
    if (bSelf == TRUE){
	if (strName.IsSameAs(GetStrName()) != TRUE) return FALSE;
    }
    else 
	mCstrName.SetBodyFromData(strName);

    std::wstring wstrLine;
    while (std::getline(istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");

	if (Attribute_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE)
	    continue;

	// Sp[^ǂ
	if (strFirstItem.IsSameAs("ReadParamGroup")){
	    wxString strName = PopStrItemFromStrLine(strLine, "::");
	    if (bSelf == TRUE)
		ASSERT(strName.IsSameAs(mCstrName.GetBodySTR()));

	    std::wstring wstrLine;
	    while (std::getline(istream, wstrLine)){
		wxString strLine = wstrLine.c_str();
		wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
		ParamGroup_cls::ReadStrLine(istream, strFirstItem, strLine);
		if (strFirstItem.IsSameAs("ReadParamGroup")){
		    ASSERT(PopStrItemFromStrLine(strLine, wxT("::")).IsSameAs(wxT("End")));
		    break;
		}
	    }
	    continue;
	}

	if (strFirstItem.IsSameAs("ReadColorCurve")){
	    ASSERT(strLine.IsSameAs("End"));
	    break;
	}
    }

    return TRUE;
}

void CLS::
Serialize(wxString& strOut){
    strOut << "ReadColorCurve::" << GetStrName() << "::Begin" << ENDL;
    Attribute_cls::Serialize(strOut);
    ParamGroup_cls::Serialize(strOut);
    strOut << "ReadColorCurve::End" << ENDL;
}
