// (setq-c-offset 4)
// "clsParticle.cpp"

#include "clsDataViewerBase.h"
#include "clsColorCurve.h"
#include "clsViewerClone.h"

#include "../clsAttribute.h"

#include "../../MISC/clsMisc.h"
#include "../../WIDGETCTRL/clsBitmapPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"

#include "../../COMMON/CommonDefinitions.h"

#include "../../../wxWidgetEnum.h"

#define CLS DataViewerBase_cls

CLS::
DataViewerBase_cls(ParentObj_cls* pCParentObj) : 
    SceneObj_cls(pCParentObj){

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";

    mCConfList.Add(&mCstrPath, "Path=", "None", strUpdate);
    mCConfList.Add(&mCbSequence, "DataSequence=", "1", strUpdate);
    mCConfList.Add(&mCSequenceOffset, "SeqOffset=", "0", strUpdate);

    mCstrPath.mbFilePath = TRUE;
}

void CLS::
ClearTexture(){
    ColorCurve_cls* pCColorCurve = GetPCColorCurve(0);

    pCColorCurve->ClearTexture();
}

ColorCurve_cls* CLS::
GetPCColorCurve(int clrCurveIndex){
    int index = -1;
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->GetStrAttributeName().IsSameAs("ColorCurve_cls")){
	    index++;
	    if (index == clrCurveIndex) return (ColorCurve_cls*)pCAttrib;
	}

    }
    return NULL;
}

MultiInputPanel_cls *CLS::
GetPCNewPanelColorCurve(wxWindow *pCWindow, ConfItemBOOL_cls *pCConfUse,
			ConfItemBOOL_cls *pCConfRepeat){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "ColorCurve"/*, TRUE*/);

    pC->mTextWidth = 100;

    int indexClrCurve;
    if (pCConfUse != NULL){
	pC->CombineSizer(2, _("ColorCurve"));
	pC->mContextMenuType = _ITEM_MENU_COLOR_CURVE;
	indexClrCurve = pC->CombineBitmapPanel(btnMaterialColorCurve0, "Send to Editor");
	pC->mItemSize = 0;
	pC->CombineConfItem(*pCConfUse);
	pC->mItemSize = 1;
    }
    else{
        pC->CombineSizer(1, _("ColorCurve"));
	pC->mContextMenuType = _ITEM_MENU_COLOR_CURVE;
	indexClrCurve = pC->CombineBitmapPanel(btnMaterialColorCurve0, "Send to Editor");
    }

    if (pCConfRepeat != NULL)
	pC->CombineConfItem(*pCConfRepeat, _("Repeat"));
    
    
    
    pC->FinishSetting();
    
    BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[indexClrCurve].mpCWindowLink;
    pCBitmap->mbButton = TRUE;
    ColorCurve_cls *pCClrCurve = (ColorCurve_cls*)GetPCAttributeFromAttribName("ColorCurve_cls", 0);
    pCBitmap->SetColorCurveLink(pCClrCurve);
    pCBitmap->mColorCurveAlpha = _COLOR_CURVE_RGBA;
    pCBitmap->mBorderX = 1;
    pCBitmap->mBorderY = 1;

    return pC;
}

void CLS::
UpdatePathColor(MNumPainBase_cls* pCPain, bool bValid){
    MultiInputPanel_cls* pCMNum = pCPain->mVpCMNumPanelLink[0];
    int index = mIndexPath+1;
    if (bValid == FALSE){
	wxColor clr(GetPCColorProvider()->GetColor("Error"));
	pCMNum->mVCItem[index].SetBackgroundColour(clr);
    }
    else
	pCMNum->mVCItem[index].SetBackgroundColour(*wxWHITE);
    pCMNum->mVCItem[index].mpCWindowLink->Refresh();
}

void CLS::
ModFrame(int &frame, int seqLen, bool bRepeat){
    if (frame >= seqLen){
	if (bRepeat) frame = Misc_cls::Surplus(frame, seqLen);
	else         frame = seqLen - 1;
    }
    if (frame < 0){
	if (bRepeat) frame = Misc_cls::Surplus(frame, seqLen);
	else         frame = 0;
    }
    
}

void CLS::
Clone(ViewerClone_cls &CClone, double NT){
    CClone.CloneConfListFrom(*this);

    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	AttributeClone_cls *pCAttribClone = new AttributeClone_cls;
	Attribute_cls* pCTgt = GetPCAttributeFromIndex(i);
	if (pCTgt->GetStrAttributeName().IsSameAs("MaterialParticle_cls"))
	    ((MaterialParticle_cls*)pCTgt)->SetMaterialInfo();
	pCTgt->Clone(*pCAttribClone);
	CClone.mVpCAttribClone.push_back(pCAttribClone);
    }
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	AttributeClone_cls *pCAttribClone = new AttributeClone_cls;
	Attribute_cls* pCTgt = GetPCDataProcFromIndex(i);
	pCTgt->Clone(*pCAttribClone);
	CClone.mVpCAttribClone.push_back(pCAttribClone);
    }

    double matrix16[16];
    GetMatrix16(matrix16, NT);
    memcpy(CClone.mMatrix16, matrix16, sizeof(double)*16);
}
