// -*- c -*-
// "clsDrawParticle.h"

#include "clsParticle.h"
#include "clsDrawParticle.h"
#include "clsMaterialParticle.h"

#include "../../GLOBJ/clsVBOSphere.h"
#include "../../GLSL/clsShaderCtrl.h"
#include "../../MYGL/clsGLExtention.h"
#include "../../MYGL/clsShadowVolumeCtrl.h"
#include "../../MYGL/clsTextureProvider.h"
#include "../../SIMDATA/clsParticleForRender.h"
#include "../../TIMELINE/clsTimeCtrl.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"
#include "../../COMMON/CommonDefinitionsShader.h"
#include "../../COMMON/clsTensorMath.h"

#include "../../../GL/glsl.h"

#ifdef WIN32
extern PFNGLDELETEBUFFERSPROC glDeleteBuffers;
extern PFNGLGENBUFFERSPROC glGenBuffers;
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
extern PFNGLVERTEXATTRIBIPOINTERPROC glVertexAttribIPointer;
extern PFNGLVERTEXATTRIBI1IPROC glVertexAttribI1i;
#endif

#define CLS DrawParticle_cls

#define BUFFER_OFFSET(bytes) ((GLubyte *)NULL + (bytes))

CLS::
DrawParticle_cls(){
    mpElementArrayBuffer = NULL;
    mpParamBitArrayBuffer = NULL;
    mAllocateNumArrayBuffer = -1;
    mAllocateNumParamBitBuffer = -1;

    for (int i = 0; i < _MAX_SPHERESHADER; i++){
	mIDClrCurveST[i] = -1;
	mIDFilterAlpha[i] = -1;
	mIDParamBit[i] = -1;
    }
}

CLS::
~DrawParticle_cls(){
    ClearVBOBase();
    if (mpElementArrayBuffer != NULL)
	delete [] mpElementArrayBuffer;
    if (mpParamBitArrayBuffer != NULL)
	delete [] mpParamBitArrayBuffer;
}

void CLS::
AllocateElemenyArray(int num){
    if (num == mAllocateNumArrayBuffer) return;
    delete [] mpElementArrayBuffer;
    mpElementArrayBuffer = new unsigned int [num];
    mAllocateNumArrayBuffer = num;
}

void CLS::
AllocateParamBitArray(int num){
    if (num == mAllocateNumParamBitBuffer) return;
    delete [] mpParamBitArrayBuffer;
    mpParamBitArrayBuffer = new int [num];
    mAllocateNumParamBitBuffer = num;
}

void CLS::
SetVBOElementArrayAll(int EAindex, int PNum, ArrayAndIndex_cls *pIndexAndZ,
		      float *pFilterAlpha){
    
    int numValid = 0;
    for (int i = 0; i < PNum; i++){
	if (pFilterAlpha[i] != 0) numValid++;
    }
    
    int index = 0;
    //unsigned int *pElementArray = new unsigned int [numValid];
    AllocateElemenyArray(numValid);
    unsigned int *pElementArray = mpElementArrayBuffer;
    for (int i = 0; i < PNum; i++){
	int particleIndex = pIndexAndZ[i].mIndex;
	if (pFilterAlpha[particleIndex] != 0){ 
	    pElementArray[index] = particleIndex;
	    index++;
	}
    }
    
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[EAindex]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, index * sizeof(unsigned int), pElementArray, GL_STATIC_DRAW);
	
    //delete [] pElementArray;
    mNumValid[EAindex] = index;
}

void CLS::
SetVBOElementArraySelective(int EAindex, int PNum, ArrayAndIndex_cls *pIndexAndZ,
			    float *pFilterAlpha, 
			    unsigned char *pParamBit, unsigned char blendType){
    int numValid = 0;
    for (int i = 0; i < PNum; i++)
	if (pFilterAlpha[i] != 0) numValid++;
    
    int index = 0;
    //unsigned int *pElementArray = new unsigned int [numValid];
    AllocateElemenyArray(numValid);
    unsigned int *pElementArray = mpElementArrayBuffer;
    for (int i = 0; i < PNum; i++){
	int particleIndex = pIndexAndZ[i].mIndex;
	if (pFilterAlpha[particleIndex] != 0 &&
	    pParamBit[particleIndex] % 2 == blendType){ 
	    pElementArray[index] = particleIndex;
	    index++;
	}
    }
	
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[EAindex]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, index * sizeof(unsigned int), pElementArray, GL_STATIC_DRAW);
	
    //delete [] pElementArray;
    mNumValid[EAindex] = index;
}

void CLS::
SetVBOBuffer(ParticleForRender_cls *pCSnapForRender, bool bTwoPass){
    
    int PNum = pCSnapForRender->mPNum;
    
    const float (* const pX)[3] = pCSnapForRender->mpX                  + 0;
    const float * const pSize = pCSnapForRender->mpSize                 + 0;
    const float * const pSizeMod = pCSnapForRender->mpSizeMod           + 0;
    const float * const pClrCurveS = pCSnapForRender->mpClrCurveS       + 0;
    const float * const pClrCurveT = pCSnapForRender->mpClrCurveT       + 0;
    float * const pFilterAlpha = pCSnapForRender->mpFilterAlpha         + 0;
    const unsigned char(* const pType) = pCSnapForRender->mpType        + 0;
    unsigned char(* const pParamBit) = pCSnapForRender->mpParamBit      + 0;
    const int * const pTwistID = pCSnapForRender->mpID                  + 0;
    ArrayAndIndex_cls * const pIndexAndZ = pCSnapForRender->mpIndexAndZ + 0;
    
    CheckGLError("DrawParticle::SetVBOBuffer");
    
    bool bRemakeVertex = FALSE;
    if (mpCSnapForRenderUsed != pCSnapForRender ||
	mDrawCountUsed != pCSnapForRender->mDrawCounter) bRemakeVertex = TRUE;
    
    bool bRemakeColor = bRemakeVertex;
    if (mpCSnapForRenderUsed != pCSnapForRender ||
	mDrawCountIndexUsed != pCSnapForRender->mDrawCounterIndex) bRemakeColor = TRUE;
    
    bool bRemakeAlpha = bRemakeVertex;
    bool bRemakeSize = bRemakeVertex;
    bool bRemakeParamBit = bRemakeVertex;
    bool bRemakeTwistID = bRemakeVertex;
    bool bRemakeElementArray = bRemakeVertex;
    if (mpCSnapForRenderUsed != pCSnapForRender ||
	mDrawCountIndexUsed != pCSnapForRender->mDrawCounterIndex) bRemakeElementArray = TRUE;
    
    mpCSnapForRenderUsed = pCSnapForRender;
    mDrawCountUsed = pCSnapForRender->mDrawCounter;
    mDrawCountIndexUsed = pCSnapForRender->mDrawCounterIndex;
    
    if (bRemakeVertex == TRUE){
	if (mVBOVertexPoint != 0) glDeleteBuffers(1, &mVBOVertexPoint);
	glGenBuffers(1, &mVBOVertexPoint);

	glBindBuffer(GL_ARRAY_BUFFER, mVBOVertexPoint);
	glBufferData(GL_ARRAY_BUFFER, 3 * PNum * sizeof(float), pX, GL_STATIC_DRAW);
    }

    if (bRemakeColor == TRUE){
	if (mVBOClrCurveSPoint != 0) glDeleteBuffers(1, &mVBOClrCurveSPoint);
	if (mVBOClrCurveTPoint != 0) glDeleteBuffers(1, &mVBOClrCurveTPoint);
	glGenBuffers(1, &mVBOClrCurveSPoint);
	glGenBuffers(1, &mVBOClrCurveTPoint);

	glBindBuffer(GL_ARRAY_BUFFER, mVBOClrCurveSPoint);
	glBufferData(GL_ARRAY_BUFFER, 1 * PNum * sizeof(float), pClrCurveS, GL_STATIC_DRAW);
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOClrCurveTPoint);
	glBufferData(GL_ARRAY_BUFFER, 1 * PNum * sizeof(float), pClrCurveT, GL_STATIC_DRAW);
    }

    if (bRemakeAlpha == TRUE){
	if (mVBOFilterAlpha != 0) glDeleteBuffers(1, &mVBOFilterAlpha);
	glGenBuffers(1, &mVBOFilterAlpha);

	glBindBuffer(GL_ARRAY_BUFFER, mVBOFilterAlpha);
	glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(float), pFilterAlpha, GL_STATIC_DRAW);
    }

    if (bRemakeSize == TRUE){
	if (mVBOSizePoint != 0) glDeleteBuffers(1, &mVBOSizePoint);
	glGenBuffers(1, &mVBOSizePoint);
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOSizePoint);
	if (pSizeMod == NULL)
	    glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(float), pSize, GL_STATIC_DRAW);
	else{
	    float *pSizeModified = new float [PNum];
	    for (int i = 0; i < PNum; i++) pSizeModified[i] = pSize[i] * pSizeMod[i];
	    glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(float), pSizeModified, GL_STATIC_DRAW);
	    delete [] pSizeModified;
	}
    }

    //int *pParamBitArray = NULL;
    if (bRemakeParamBit == TRUE){
	if (mVBOParamBitPoint != 0) glDeleteBuffers(1, &mVBOParamBitPoint);
	glGenBuffers(1, &mVBOParamBitPoint);

	AllocateParamBitArray(PNum);
	int *pParamBitArray = mpParamBitArrayBuffer;//new int [PNum];
	for (int i = 0; i < PNum; i++)
	    pParamBitArray[i] = (int)pParamBit[i];
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOParamBitPoint);
	glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(int), pParamBitArray, GL_STATIC_DRAW);
    }

    if (bRemakeTwistID == TRUE){
	if (mVBOTwistID != 0) glDeleteBuffers(1, &mVBOTwistID);
	glGenBuffers(1, &mVBOTwistID);

	glBindBuffer(GL_ARRAY_BUFFER, mVBOTwistID);
	glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(int), pTwistID, GL_STATIC_DRAW);
    }
    
    if (bRemakeElementArray == TRUE){
	if (mVBOElementArrayPoint[0] != 0) glDeleteBuffers(1, &mVBOElementArrayPoint[0]);
	glGenBuffers(1, &mVBOElementArrayPoint[0]);

	if (mVBOElementArrayPoint[1] != 0) glDeleteBuffers(1, &mVBOElementArrayPoint[1]);
	mVBOElementArrayPoint[1] = 0;
	
	if (bTwoPass == TRUE){
	    glGenBuffers(1, &mVBOElementArrayPoint[1]);
	    SetVBOElementArraySelective(0, PNum, pIndexAndZ, pFilterAlpha, pParamBit, 0);
	    SetVBOElementArraySelective(1, PNum, pIndexAndZ, pFilterAlpha, pParamBit, 1);
	}
	else{
	    SetVBOElementArrayAll(0, PNum, pIndexAndZ, pFilterAlpha);
	}
    }

    /*
    if (pParamBitArray != NULL)
	delete [] pParamBitArray;
    */
}



void CLS::
DrawSphereSV(ParticleForRender_cls *pCSnapForRender, int sphereDepth, int ensemble,
	     MaterialParticle_cls *pCMatParticle[_MAX_PARTICLE_TYPE]){
    
    if (pCSnapForRender->mbCastShadow == FALSE) return;

    int PNum = pCSnapForRender->mPNum;
    const float (* const pX)[3] = pCSnapForRender->mpX;
    const float * const pSize = pCSnapForRender->mpSize;
    const float * const pSizeMod = pCSnapForRender->mpSizeMod;
    float * const pFilterAlpha = pCSnapForRender->mpFilterAlpha;
    ArrayAndIndex_cls * const pIndexAndZ = pCSnapForRender->mpIndexAndZ;
    unsigned char * const pType = pCSnapForRender->mpType;

    VBOSphere_cls *pCVBOSphere = GetPCVBOSphere();

    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
    bool bTransparentSV = pCSV->mCbTransparentShadowForEnsemble.GetBodyBOOL();
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    
    pCSV->BeginShader();

    const int packNum = 500;
    for (int i_Base = 0; i_Base < PNum; i_Base+=packNum){
	int i_Max = i_Base + packNum;
	if (i_Max > PNum) i_Max = PNum;

	//float filterAlphaPrev = -1;
	//int paramBitPrev = -1;
	float posPrev[3] = {0, 0, 0};
	float scalePrevInv = 1;

	glPushMatrix();
	for (int i_ = i_Base; i_ < i_Max; i_++){
	    int i = pIndexAndZ[i_].mIndex;

	    float filterAlpha = pFilterAlpha[i];
	    if (filterAlpha == 0) continue;

	    float scale = pSize[i];
	    if (pSizeMod != NULL) scale *= pSizeMod[i];
	    if (scale == 0) continue;

	    if (bTransparentSV == TRUE){
		int type = pType[i];
		
		// !!_FIX_ME_!! X̗q̃At@͂ǂēH
		// ͉
		float alpha = pCMatParticle[type]->GetAlpha(NT);
		bool bSkip = pCSV->IsSkip(alpha, ensemble, NULL, i);
		if (bSkip == TRUE) continue;
	    }

	    glTranslatef((pX[i][0]-posPrev[0]) * scalePrevInv,
			 (pX[i][1]-posPrev[1]) * scalePrevInv,
			 (pX[i][2]-posPrev[2]) * scalePrevInv);
	    
	    glScalef(scale*scalePrevInv, scale*scalePrevInv, scale*scalePrevInv);

	    memcpy(posPrev, pX[i], sizeof(float)*3);
	    scalePrevInv = 1.0/scale;

	    
	    if (pCSV->mCbClipCheck.GetBodyBOOL() == TRUE){
		glDisable(GL_DEPTH_TEST);    
		pCSV->SettingFront();
		pCSV->SetCullingSide(2);
		pCVBOSphere->Draw(_VBO_SHADOW_VOLUME, 0, sphereDepth);

		pCSV->SettingBack();
		pCSV->SetCullingSide(1);
		pCVBOSphere->Draw(_VBO_SHADOW_VOLUME, 0, sphereDepth);
	    }
	    
	    glEnable(GL_DEPTH_TEST);
	    pCSV->SettingFront();
	    pCSV->SetCullingSide(1);
	    pCVBOSphere->Draw(_VBO_SHADOW_VOLUME, 0, sphereDepth);
	    
	    pCSV->SettingBack();
	    pCSV->SetCullingSide(2);
	    pCVBOSphere->Draw(_VBO_SHADOW_VOLUME, 0, sphereDepth);
	}
	glPopMatrix();
    }

    pCSV->EndShader();
}

SphereGouraudParam_cls* CLS::
GetPCNewGouraudParam(int lightNum, int ensemble, int dissolve){
    SphereGouraudParam_cls *pCParam = new SphereGouraudParam_cls;
    pCParam->mLightNumber = lightNum;
    pCParam->mEnsemble = ensemble;
    pCParam->mDissolveMode = dissolve;
    
    GetPCShaderCtrl()->BeginSphereGouraud(pCParam);
    if (mIDClrCurveST[_SPHERESHADER_GOURAUD] == -1){
	ShaderItem_cls &CItemGouraud = GetPCShaderCtrl()->mVCShader[_SHADER_SPHERE_GOURAUD];
	mIDClrCurveST [_SPHERESHADER_GOURAUD] =
	    glGetAttribLocation(CItemGouraud.mProgram, "colorCurveST");
	mIDFilterAlpha[_SPHERESHADER_GOURAUD] =
	    glGetAttribLocation(CItemGouraud.mProgram, "filterAlpha");
	mIDParamBit   [_SPHERESHADER_GOURAUD] =
	    glGetAttribLocation(CItemGouraud.mProgram, "paramBit");
    }
    return pCParam;
}

SpherePhongParam_cls* CLS::
GetPCNewPhongParam(int lightNum, int ensemble, int dissolve, bool bFlat){
    SpherePhongParam_cls *pCParam = new SpherePhongParam_cls;
    pCParam->mLightNumber = lightNum;
    pCParam->mEnsemble = ensemble;
    pCParam->mDissolveMode = dissolve;
    pCParam->mbFlat = bFlat;
	    
    if (mIDClrCurveST[_SPHERESHADER_PHONG] == -1){
	GetPCShaderCtrl()->BeginSpherePhong(pCParam);
	ShaderItem_cls &CItemPhong = GetPCShaderCtrl()->mVCShader[_SHADER_SPHERE_PHONG];
	mIDClrCurveST [_SPHERESHADER_PHONG] = 
	    glGetAttribLocation(CItemPhong.mProgram, "colorCurveST");
	mIDFilterAlpha[_SPHERESHADER_PHONG] =
	    glGetAttribLocation(CItemPhong.mProgram, "filterAlpha");
	mIDParamBit   [_SPHERESHADER_PHONG] =
	    glGetAttribLocation(CItemPhong.mProgram, "paramBit");
    }
    return pCParam;
}

SphereToonParam_cls* CLS::
GetPCNewToonParam(int lightNum, int ensemble, int dissolve, float threshold, float boldness, bool bFlat){
    SphereToonParam_cls *pCParam = new SphereToonParam_cls;
    pCParam->mLightNumber = lightNum;
    pCParam->mEnsemble = ensemble;
    pCParam->mDissolveMode = dissolve;
    pCParam->mThreshold = threshold;
    pCParam->mBoldness = boldness;
    pCParam->mbFlat = bFlat;
	    
    if (mIDClrCurveST[_SPHERESHADER_TOON] == -1){
	GetPCShaderCtrl()->BeginSphereToon(pCParam);
	ShaderItem_cls &CItemToon = GetPCShaderCtrl()->mVCShader[_SHADER_SPHERE_TOON];
	mIDClrCurveST [_SPHERESHADER_TOON] = 
	    glGetAttribLocation(CItemToon.mProgram, "colorCurveST");
	mIDFilterAlpha[_SPHERESHADER_TOON] =
		    glGetAttribLocation(CItemToon.mProgram, "filterAlpha");
	mIDParamBit   [_SPHERESHADER_TOON] =
	    glGetAttribLocation(CItemToon.mProgram, "paramBit");
	
    }
    return pCParam;
}


void CLS::
DrawSphere(ParticleForRender_cls *pCSnapForRender, int lightNum, int ensemble, int dissolve, int sphereDepth, bool bSingle, 
	   MaterialParticle_cls *pCMatParticle[_MAX_PARTICLE_TYPE]){
    // bSingle = TRUE  Matrial 1̂݁B
    // ȒPȍ̕@͂邩H (܂)

    if (GetPCShadowVolumeCtrl()->mSVState == 1) {
	
	return DrawSphereSV(pCSnapForRender, sphereDepth, ensemble,
			    pCMatParticle);
    }

    BindParticleTexture(pCSnapForRender);
    glActiveTexture(GL_TEXTURE0);
    pCSnapForRender->mCColorCurveTexture.BindTexture();

    int PNum = pCSnapForRender->mPNum;
    float const (*pX)[3] = pCSnapForRender->mpX;
    float const *pSize = pCSnapForRender->mpSize;
    float const *pSizeMod = pCSnapForRender->mpSizeMod;
    float const *pClrCurveS = pCSnapForRender->mpClrCurveS;
    float const *pClrCurveT = pCSnapForRender->mpClrCurveT;
    float const *pFilterAlpha = pCSnapForRender->mpFilterAlpha;
    unsigned char(*pType) = pCSnapForRender->mpType;
    unsigned char(*pParamBit) = pCSnapForRender->mpParamBit;
    ArrayAndIndex_cls *pIndexAndZ = pCSnapForRender->mpIndexAndZ;

    VBOSphere_cls *pCVBOSphere = GetPCVBOSphere();

    const int packNum = 500;

    int sphereShader[_MAX_PARTICLE_TYPE] = {};
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	if (pCMatParticle[i] != NULL)
	    sphereShader[i] = pCMatParticle[i]->mCShaderTypeSphere.GetBodyINT();
    }
    
    bool bUseInvertedHull = FALSE;

    ShaderParams_cls *pCSphereShaderParams[_MAX_PARTICLE_TYPE] = {};
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	MaterialParticle_cls *pCMat = pCMatParticle[i];
	if (sphereShader[i] == _SPHERESHADER_GOURAUD){
	    pCSphereShaderParams[i] = 
		GetPCNewGouraudParam(lightNum, ensemble, dissolve);
	}
	if (sphereShader[i] == _SPHERESHADER_PHONG){
	    
	    pCSphereShaderParams[i] = 
		GetPCNewPhongParam(lightNum, ensemble, dissolve,
				   pCMat->mCbShaderFlat.GetBodyBOOL());
	}	    

	if (sphereShader[i] == _SPHERESHADER_TOON){
	    pCSphereShaderParams[i] = 
		GetPCNewToonParam(lightNum, ensemble, dissolve,
				  pCMat->mCToonThreshold.GetBodyDOUBLE(),
				  pCMat->mCToonBoldness.GetBodyDOUBLE(),
				  pCMat->mCbShaderFlat.GetBodyBOOL());
	    
	    if (pCMat->mCToonBoldness.GetBodyDOUBLE() != 0)
		bUseInvertedHull = TRUE;
	}
    }
    
    int currentType = -1;
    int currentShader = -1;

    if (bUseInvertedHull == FALSE){
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
    }
    else
	glDisable(GL_CULL_FACE);
    
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glColor4f(1, 1, 1, 1);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    pCVBOSphere->mCObject[sphereDepth].
	PrepareClientState(TRUE, _UV_MAPPING_UV, FALSE);

    for (int i_Base = 0; i_Base < PNum; i_Base+=packNum){
	int i_Max = i_Base + packNum;
	if (i_Max > PNum) i_Max = PNum;

	float filterAlphaPrev = -1;
	int paramBitPrev = -1;
	float posPrev[3] = {0, 0, 0};
	float scalePrevInv = 1;
	
	glPushMatrix();
	for (int i_ = i_Base; i_ < i_Max; i_++){
	    int i = pIndexAndZ[i_].mIndex;

	    float filterAlpha = pFilterAlpha[i];
	    if (filterAlpha == 0) continue;

	    float scale = pSize[i];
	    if (pSizeMod != NULL) scale *= pSizeMod[i];
	    if (scale == 0) continue;

	    int type = pType[i];
	    int shaderType = sphereShader[type];

	    if (type != currentType/*shaderType != currentShader*/){
		currentShader = shaderType;
		currentType = type;
		GetPCShaderCtrl()->EndShader();
		switch(currentShader){
		case 0: 
		    GetPCShaderCtrl()->BeginSphereGouraud((SphereGouraudParam_cls*)pCSphereShaderParams[type]); 
		    break;
		case 1: 
		    GetPCShaderCtrl()->BeginSpherePhong((SpherePhongParam_cls*)pCSphereShaderParams[type]); 
		    break;
		case 2:
		    GetPCShaderCtrl()->BeginSphereToon((SphereToonParam_cls*)pCSphereShaderParams[type]); 
		    break;
		}
	    }
	    
	    glTranslatef((pX[i][0]-posPrev[0]) * scalePrevInv,
			 (pX[i][1]-posPrev[1]) * scalePrevInv,
			 (pX[i][2]-posPrev[2]) * scalePrevInv);
	    
	    glScalef(scale*scalePrevInv, scale*scalePrevInv, scale*scalePrevInv);

	    memcpy(posPrev, pX[i], sizeof(float)*3);
	    scalePrevInv = 1.0/scale;

	    glVertexAttrib2f(mIDClrCurveST[shaderType], pClrCurveS[i], pClrCurveT[i]);
	    
	    
	    if (filterAlpha != filterAlphaPrev)
		glVertexAttrib1f(mIDFilterAlpha[shaderType], pFilterAlpha[i]);
	    filterAlphaPrev = filterAlpha;
	    
	    int paramBit = pParamBit[i];
	    if (paramBit != paramBitPrev)
		glVertexAttribI1i(mIDParamBit[shaderType], pParamBit[i]);
	    paramBitPrev = paramBit;
	    	    
	    pCVBOSphere->DrawFast(sphereDepth);
	}
	glPopMatrix();
    }

    GetPCShaderCtrl()->EndShader();
    
    CheckGLError("DrawParticle::DrawSphere");

    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	if (pCSphereShaderParams[i] != NULL)
	    delete pCSphereShaderParams[i];
    
}

void CLS::
DrawSereal(ParticleForRender_cls *pCSnapForRender, int lightNum, int ensemble, int dissolve, int sphereDepth, int twist,
	   MaterialParticle_cls *pCMatParticle[_MAX_PARTICLE_TYPE],
	   float baseMatrix16[16], float baseMatrix16Inv[16],
	   float viewMatrix16[16], float viewMatrix16Inv[16]){
    // bSingle = TRUE  Matrial 1̂݁B
    // ȒPȍ̕@͂邩H (܂)

    if (GetPCShadowVolumeCtrl()->mSVState == 1) {
	
	return; /*DrawSphereSV(pCSnapForRender, sphereDepth, ensemble,
		  pCMatParticle);*/
    }

    BindParticleTexture(pCSnapForRender);
    glActiveTexture(GL_TEXTURE0);
    pCSnapForRender->mCColorCurveTexture.BindTexture();

    glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);

    int PNum = pCSnapForRender->mPNum;
    float const (*pX)[3] = pCSnapForRender->mpX;
    float const *pSize = pCSnapForRender->mpSize;
    float const *pSizeMod = pCSnapForRender->mpSizeMod;
    float const *pClrCurveS = pCSnapForRender->mpClrCurveS;
    float const *pClrCurveT = pCSnapForRender->mpClrCurveT;
    float const *pFilterAlpha = pCSnapForRender->mpFilterAlpha;
    unsigned char(*pType) = pCSnapForRender->mpType;
    unsigned char(*pParamBit) = pCSnapForRender->mpParamBit;
    int * const pTwistID = pCSnapForRender->mpID                  + 0;
    ArrayAndIndex_cls *pIndexAndZ = pCSnapForRender->mpIndexAndZ;

    float matrixA[16], matrixB[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, matrixA);
    
    TensorMath_cls::Multiply(viewMatrix16Inv, matrixA);
    TensorMath_cls::Multiply(baseMatrix16, matrixA);
    
    TensorMath_cls::GetInvMatrix4x4(matrixA, matrixB);

    VBOSphere_cls *pCVBOSphere = GetPCVBOSphere();

    int sphereShader[_MAX_PARTICLE_TYPE] = {};
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	if (pCMatParticle[i] != NULL)
	    sphereShader[i] = pCMatParticle[i]->mCShaderTypeSphere.GetBodyINT();
    }

    bool bUseInvertedHull = FALSE;

    ShaderParams_cls *pCSphereShaderParams[_MAX_PARTICLE_TYPE] = {};
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	MaterialParticle_cls *pCMat = pCMatParticle[i];
	if (sphereShader[i] == _SPHERESHADER_GOURAUD){
	    pCSphereShaderParams[i] = 
		GetPCNewGouraudParam(lightNum, ensemble, dissolve);
	}
	if (sphereShader[i] == _SPHERESHADER_PHONG){
	    
	    pCSphereShaderParams[i] = 
		GetPCNewPhongParam(lightNum, ensemble, dissolve,
				   pCMat->mCbShaderFlat.GetBodyBOOL());
	}	    

	if (sphereShader[i] == _SPHERESHADER_TOON){
	    pCSphereShaderParams[i] = 
		GetPCNewToonParam(lightNum, ensemble, dissolve,
				  pCMat->mCToonThreshold.GetBodyDOUBLE(),
				  pCMat->mCToonBoldness.GetBodyDOUBLE(),
				  pCMat->mCbShaderFlat.GetBodyBOOL());
	    
	    if (pCMat->mCToonBoldness.GetBodyDOUBLE() != 0)
		bUseInvertedHull = TRUE;
	}
    }

    if (bUseInvertedHull == FALSE){
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
    }
    else
	glDisable(GL_CULL_FACE);

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glPointSize(12);
    glColor4f(1, 1, 1, 1);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    int currentBaseType = -1;
    int currentType = -1;
    int currentShader = -1;
    for (int i_ = 0; i_ < PNum; i_++){
	int i = pIndexAndZ[i_].mIndex;

	float filterAlpha = pFilterAlpha[i];
	if (filterAlpha == 0) continue;
	
	float scale = pSize[i];
	if (pSizeMod != NULL) scale *= pSizeMod[i];
	if (scale == 0) continue;

	int type = pType[i];
	
	MaterialParticle_cls *pCMat = pCMatParticle[type];
	int baseType = pCMat->mCBaseType.GetBodyINT();

	if (baseType == _PARTICLE_POINT){
	    if (type != currentType){
		if (currentBaseType == _PARTICLE_SPHERE)
		    pCVBOSphere->mCObject[sphereDepth].FinishClientState();
		currentType = type;

		PointBasicParam_cls CParam;
		CParam.mLightNumber = lightNum;
		CParam.mEnsemble = ensemble;
		CParam.mDissolveMode = dissolve;
		CParam.mDistanceBase = pCSnapForRender->mDistanceBase;
		glGetIntegerv(GL_VIEWPORT, CParam.mViewport);
		
		GetPCShaderCtrl()->BeginPointBasic(&CParam);

		ShaderItem_cls &CItem = GetPCShaderCtrl()->mVCShader[_SHADER_POINT_BASIC];
		if (mIDPointSize == -1) SetPointAttribLocation(CItem);

		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
		glEnableClientState(GL_VERTEX_ARRAY);
		glDisableVertexAttribArray(mIDPointSize);
		glDisableVertexAttribArray(mIDPointClrCurveS);
		glDisableVertexAttribArray(mIDPointClrCurveT);
		glDisableVertexAttribArray(mIDPointFilterAlpha);
		glDisableVertexAttribArray(mIDPointParamBit);
	    }
	    
	    glVertexAttrib1f (mIDPointSize, pSize[i]);
	    glVertexAttrib1f (mIDPointClrCurveS, pClrCurveS[i]);
	    glVertexAttrib1f (mIDPointClrCurveT, pClrCurveT[i]);
	    glVertexAttrib1f (mIDPointFilterAlpha, pFilterAlpha[i]);
	    glVertexAttribI1i(mIDPointParamBit, pParamBit[i]);
	    glVertexPointer(3, GL_FLOAT, 0, pX[i]);

	    glDrawArrays(GL_POINTS, 0, 1);
	}

	if (baseType == _PARTICLE_BILLBOARD){
	    if (type != currentType){
		if (currentBaseType == _PARTICLE_SPHERE)
		    pCVBOSphere->mCObject[sphereDepth].FinishClientState();

		currentType = type;
		
		PointBillboardParam_cls CParam;
		CParam.mLightNumber = lightNum;
		CParam.mEnsemble = ensemble;
		CParam.mDissolveMode = dissolve;
		CParam.mTwistMode = twist;
		CParam.mpBaseInvModel = matrixA;
		CParam.mpModelInvBase = matrixB;
		CParam.mpBaseInv = baseMatrix16Inv;
		
		GetPCShaderCtrl()->BeginPointBillboard(&CParam);

		ShaderItem_cls &CItem = GetPCShaderCtrl()->mVCShader[_SHADER_POINT_BILLBOARD];
		if (mIDBillboardSize == -1) SetBillboardAttribLocation(CItem);

		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
		glEnableClientState(GL_VERTEX_ARRAY);
		glDisableVertexAttribArray(mIDBillboardSize);
		glDisableVertexAttribArray(mIDBillboardClrCurveS);
		glDisableVertexAttribArray(mIDBillboardClrCurveT);
		glDisableVertexAttribArray(mIDBillboardFilterAlpha);
		glDisableVertexAttribArray(mIDBillboardParamBit);
	    }
	    
	    glVertexAttrib1f (mIDBillboardSize, pSize[i]);
	    glVertexAttrib1f (mIDBillboardClrCurveS, pClrCurveS[i]);
	    glVertexAttrib1f (mIDBillboardClrCurveT, pClrCurveT[i]);
	    glVertexAttrib1f (mIDBillboardFilterAlpha, pFilterAlpha[i]);
	    glVertexAttribI1i(mIDBillboardParamBit, pParamBit[i]);
	    if (mIDBillboardTwistID != 0)
		glVertexAttribI1i(mIDBillboardTwistID, pTwistID[i]);
	    glVertexPointer(3, GL_FLOAT, 0, pX[i]);

	    glDrawArrays(GL_POINTS, 0, 1);
	}
	    
	if (baseType == _PARTICLE_SPHERE){
	    int shaderType = sphereShader[type];
	    
	    glPushMatrix();

	    if (type != currentType){
		currentShader = shaderType;
		currentType = type;
		
		GetPCShaderCtrl()->EndShader();
		switch(currentShader){
		case 0: 
		    GetPCShaderCtrl()->BeginSphereGouraud((SphereGouraudParam_cls*)pCSphereShaderParams[type]); 
		    break;
		case 1: 
		    GetPCShaderCtrl()->BeginSpherePhong((SpherePhongParam_cls*)pCSphereShaderParams[type]); 
		    break;
		case 2:
		    GetPCShaderCtrl()->BeginSphereToon((SphereToonParam_cls*)pCSphereShaderParams[type]); 
		    break;
		}
		
		pCVBOSphere->mCObject[sphereDepth].
		    PrepareClientState(TRUE, _UV_MAPPING_UV, FALSE);
	    }

	    glTranslatef(pX[i][0], pX[i][1], pX[i][2]);
	    glScalef(scale, scale, scale);

	    glVertexAttrib2f(mIDClrCurveST[shaderType], pClrCurveS[i], pClrCurveT[i]);
	    glVertexAttrib1f(mIDFilterAlpha[shaderType], pFilterAlpha[i]);
	    
	    int paramBit = pParamBit[i];
	    glVertexAttribI1i(mIDParamBit[shaderType], pParamBit[i]);
	    
	    pCVBOSphere->DrawFast(sphereDepth);

	    glPopMatrix();
	}

	currentBaseType = baseType;
    }

    GetPCShaderCtrl()->EndShader();

    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	if (pCSphereShaderParams[i] != NULL)
	    delete pCSphereShaderParams[i];
}


