// -*- c -*-
// "clsDrawParticleBase.h"

#include "clsDrawParticleBase.h"

#include "../../GLSL/clsShaderCtrl.h"
#include "../../MYGL/clsTextureProvider.h"
#include "../../MYGL/clsGLExtention.h"
#include "../../SIMDATA/clsParticleForRender.h"

#include "../../COMMON/CommonDefinitionsGL.h"
#include "../../COMMON/CommonDefinitionsShader.h"

#include "../../../GL/glsl.h"

#ifdef WIN32
extern PFNGLDELETEBUFFERSPROC glDeleteBuffers;
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
extern PFNGLVERTEXATTRIBIPOINTERPROC glVertexAttribIPointer;
#endif

#define CLS DrawParticleBase_cls

#define BUFFER_OFFSET(bytes) ((GLubyte *)NULL + (bytes))

CLS::
DrawParticleBase_cls(){
    mpCSnapForRenderUsed = NULL;
    mDrawCountUsed = -1;
    mDrawCountIndexUsed = -1;
    mVBOVertexPoint = 0;
    mVBOClrCurveSPoint = 0; mVBOClrCurveTPoint = 0;
    mVBOFilterAlpha = 0;
    mVBOSizePoint = 0;
    mVBOParamBitPoint = 0;
    mVBOTwistID = 0;
    for (int i = 0; i < 2; i++){
	mNumValid[i] = 0;
	mVBOElementArrayPoint[i] = 0;
    }
    
    mIDPointSize        = -1;
    mIDPointClrCurveS   = -1;
    mIDPointClrCurveT   = -1;
    mIDPointFilterAlpha = -1;
    mIDPointParamBit    = -1;
    
    mIDBillboardSize        = -1;
    mIDBillboardClrCurveS   = -1;
    mIDBillboardClrCurveT   = -1;
    mIDBillboardFilterAlpha = -1;
    mIDBillboardParamBit    = -1;
    mIDBillboardTwistID     = -1;
}

void CLS::
ClearVBOBase(){
    if (mVBOVertexPoint != 0){
	glDeleteBuffers(1, &mVBOVertexPoint); mVBOVertexPoint = 0; }
    if (mVBOClrCurveSPoint != 0){
	glDeleteBuffers(1, &mVBOClrCurveSPoint); mVBOClrCurveSPoint = 0; }
    if (mVBOClrCurveTPoint != 0){
	glDeleteBuffers(1, &mVBOClrCurveTPoint); mVBOClrCurveTPoint = 0; }
    if (mVBOFilterAlpha != 0){
	glDeleteBuffers(1, &mVBOFilterAlpha); mVBOFilterAlpha = 0; }
    if (mVBOSizePoint != 0){
	glDeleteBuffers(1, &mVBOSizePoint); mVBOSizePoint = 0; }
    if (mVBOParamBitPoint != 0){
	glDeleteBuffers(1, &mVBOParamBitPoint); mVBOParamBitPoint = 0; }
    for (int i = 0; i < 2; i++){
	if (mVBOElementArrayPoint[i] != 0){
	    glDeleteBuffers(1, &mVBOElementArrayPoint[i]);
	    mVBOElementArrayPoint[i] = 0; }
    }
}

void CLS::
BindParticleTextureMain(wxString strPath, int textureUnit){
    if (textureUnit == 1) glActiveTexture(GL_TEXTURE1);
    if (textureUnit == 2) glActiveTexture(GL_TEXTURE2);
    if (textureUnit == 3) glActiveTexture(GL_TEXTURE3);
    if (strPath.IsSameAs("None")){
	//glDisable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, 0);
    }
    else{
	glEnable(GL_TEXTURE_2D);
	GetPCTextureProvider()->BindTexture(strPath);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }
}

void CLS::
BindParticleTexture(ParticleForRender_cls *pCSnapForRender){
    for (int i = 0; i < 3; i++){
	wxString strPath = pCSnapForRender->mstrParticleTexture[i];
	BindParticleTextureMain(strPath, i+1);
    }
}

void CLS::
SetPointAttribLocation(ShaderItem_cls &CItem){
    int prog = CItem.mProgram;
    mIDPointSize        = glGetAttribLocation(prog, "pointSize");
    mIDPointClrCurveS   = glGetAttribLocation(prog, "colorCurveS");
    mIDPointClrCurveT   = glGetAttribLocation(prog, "colorCurveT");
    mIDPointFilterAlpha = glGetAttribLocation(prog, "filterAlpha");
    mIDPointParamBit    = glGetAttribLocation(prog, "paramBit");
}

void CLS::
SetBillboardAttribLocation(ShaderItem_cls &CItem){
    int prog = CItem.mProgram;
    mIDBillboardSize        = glGetAttribLocation(prog, "pointSize");
    mIDBillboardClrCurveS   = glGetAttribLocation(prog, "colorCurveS");
    mIDBillboardClrCurveT   = glGetAttribLocation(prog, "colorCurveT");
    mIDBillboardFilterAlpha = glGetAttribLocation(prog, "filterAlpha");
    mIDBillboardParamBit    = glGetAttribLocation(prog, "paramBit");
    mIDBillboardTwistID     = glGetAttribLocation(prog, "twistID");
}

void CLS::
DrawPoints(ParticleForRender_cls *pCSnapForRender, int lightNum, int ensemble, int dissolve){
    
    BindParticleTexture(pCSnapForRender);
    glActiveTexture(GL_TEXTURE0);
    pCSnapForRender->mCColorCurveTexture.BindTexture();
    
    glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);

    CheckGLError("DrawParticle::DrawPoints");

    if (mNumValid[0] != 0 || mNumValid[1] != 0){

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glPointSize(12);
	glColor4f(1, 1, 1, 1);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

	PointBasicParam_cls CParam;
	CParam.mLightNumber = lightNum;
	CParam.mEnsemble = ensemble;
	CParam.mDissolveMode = dissolve;
	CParam.mDistanceBase = pCSnapForRender->mDistanceBase;
	glGetIntegerv(GL_VIEWPORT, CParam.mViewport);

	GetPCShaderCtrl()->BeginPointBasic(&CParam);
    
	glEnableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	ShaderItem_cls &CItem = GetPCShaderCtrl()->mVCShader[_SHADER_POINT_BASIC];

	if (mIDPointSize == -1) SetPointAttribLocation(CItem);

	glEnableVertexAttribArray(mIDPointSize);
	glEnableVertexAttribArray(mIDPointClrCurveS);
	glEnableVertexAttribArray(mIDPointClrCurveT);
	glEnableVertexAttribArray(mIDPointFilterAlpha);
	glEnableVertexAttribArray(mIDPointParamBit);
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOVertexPoint);
	glVertexPointer(3, GL_FLOAT, 0, BUFFER_OFFSET(0));
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOSizePoint);
	glVertexAttribPointer(mIDPointSize, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOClrCurveSPoint);
	glVertexAttribPointer(mIDPointClrCurveS, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));

	glBindBuffer(GL_ARRAY_BUFFER, mVBOClrCurveTPoint);
	glVertexAttribPointer(mIDPointClrCurveT, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));

	glBindBuffer(GL_ARRAY_BUFFER, mVBOFilterAlpha);
	glVertexAttribPointer(mIDPointFilterAlpha, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));

	glBindBuffer(GL_ARRAY_BUFFER, mVBOParamBitPoint);
	glVertexAttribIPointer(mIDPointParamBit, 1, GL_INT, 0, BUFFER_OFFSET(0));
	
	if (mNumValid[0] != 0){
	    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[0]);
	    glDrawElements(GL_POINTS, mNumValid[0], GL_UNSIGNED_INT, BUFFER_OFFSET(0));
	    CheckGLError("Tag");
	}
	
	if (mVBOElementArrayPoint[1] != 0 &&
	    mNumValid[1] != 0){
	    glDepthMask(GL_FALSE);
	    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[1]);
	    glDrawElements(GL_POINTS, mNumValid[1], GL_UNSIGNED_INT, BUFFER_OFFSET(0));
	}
	
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableVertexAttribArray(mIDPointSize);
	glDisableVertexAttribArray(mIDPointClrCurveS);
	glDisableVertexAttribArray(mIDPointClrCurveT);
	glDisableVertexAttribArray(mIDPointFilterAlpha);
	glDisableVertexAttribArray(mIDPointParamBit);
	
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	GetPCShaderCtrl()->EndShader();
    }
}

void CLS::
DrawBillboards(ParticleForRender_cls *pCSnapForRender, int lightNum, int ensemble, int dissolve, int twist,
	       float baseMatrix16[16], float baseMatrix16Inv[16],
	       float viewMatrix16[16], float viewMatrix16Inv[16]){
    
    BindParticleTexture(pCSnapForRender);
    glActiveTexture(GL_TEXTURE0);
    pCSnapForRender->mCColorCurveTexture.BindTexture();
    
    //glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
    
    CheckGLError("DrawParticle::DrawBillboards");

    if (mNumValid[0] != 0 || mNumValid[1] != 0){

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glPointSize(12);
	glColor4f(1, 1, 1, 1);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

	float matrixA[16], matrixB[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, matrixA);

	TensorMath_cls::Multiply(viewMatrix16Inv, matrixA);
	TensorMath_cls::Multiply(baseMatrix16, matrixA);

	TensorMath_cls::GetInvMatrix4x4(matrixA, matrixB);

	/*
	float matrixTranspose[16];
	glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, matrixTranspose);
		
	float view[16], viewInv[16];
	float viewT[16], viewInvT[16];
	float test0[16], test1[16];
	float test2[16], test3[16];
	for (int i = 0; i < 16; i++) test0[i] = matrix[i];
	for (int i = 0; i < 16; i++) test1[i] = matrix[i];
	for (int i = 0; i < 16; i++) test2[i] = matrixTranspose[i];
	for (int i = 0; i < 16; i++) test3[i] = matrixTranspose[i];
	for (int i = 0; i < 16; i++) view[i] = viewMatrix16[i];
	for (int i = 0; i < 16; i++) viewInv[i] = viewMatrix16Inv[i];
	TensorMath_cls::Transpose4x4(viewMatrix16, viewT, 1.0f);
	TensorMath_cls::Transpose4x4(viewMatrix16Inv, viewInvT, 1.0f);
	
	TensorMath_cls::Multiply(test0, viewT);
	TensorMath_cls::Multiply(test1, viewInvT);
	TensorMath_cls::Multiply(viewMatrix16,    test2);
	TensorMath_cls::Multiply(viewMatrix16Inv, test3);
	*/
	
	PointBillboardParam_cls CParam;
	CParam.mLightNumber = lightNum;
	CParam.mEnsemble = ensemble;
	CParam.mDissolveMode = dissolve;
	CParam.mTwistMode = twist;
	CParam.mpBaseInvModel = matrixA;
	CParam.mpModelInvBase = matrixB;
	CParam.mpBaseInv = baseMatrix16Inv;
	
	GetPCShaderCtrl()->BeginPointBillboard(&CParam);
	
	glEnableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	ShaderItem_cls &CItem = GetPCShaderCtrl()->mVCShader[_SHADER_POINT_BILLBOARD];

	if (mIDBillboardSize == -1) SetBillboardAttribLocation(CItem);
	
	glEnableVertexAttribArray(mIDBillboardSize);
	glEnableVertexAttribArray(mIDBillboardClrCurveS);
	glEnableVertexAttribArray(mIDBillboardClrCurveT);
	glEnableVertexAttribArray(mIDBillboardFilterAlpha);
	glEnableVertexAttribArray(mIDBillboardParamBit);
	if (mIDBillboardTwistID != 0)
	    glEnableVertexAttribArray(mIDBillboardTwistID);
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOVertexPoint);
	glVertexPointer(3, GL_FLOAT, 0, BUFFER_OFFSET(0));
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOSizePoint);
	glVertexAttribPointer(mIDBillboardSize, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOClrCurveSPoint);
	glVertexAttribPointer(mIDBillboardClrCurveS, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOClrCurveTPoint);
	glVertexAttribPointer(mIDBillboardClrCurveT, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOFilterAlpha);
	glVertexAttribPointer(mIDBillboardFilterAlpha, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
	
	glBindBuffer(GL_ARRAY_BUFFER, mVBOParamBitPoint);
	glVertexAttribIPointer(mIDBillboardParamBit, 1, GL_INT, 0, BUFFER_OFFSET(0));

	if (mIDBillboardTwistID != 0){
	    glBindBuffer(GL_ARRAY_BUFFER, mVBOTwistID);
	    glVertexAttribIPointer(mIDBillboardTwistID, 1, GL_INT, 0, BUFFER_OFFSET(0));
	}
	
	if (mNumValid[0] != 0){
	    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[0]);
	    glDrawElements(GL_POINTS, mNumValid[0], GL_UNSIGNED_INT, BUFFER_OFFSET(0));
	}
	
	if (mVBOElementArrayPoint[1] != 0 &&
	    mNumValid[1] != 0){
	    glDepthMask(GL_FALSE);
	    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[1]);
	    glDrawElements(GL_POINTS, mNumValid[1], GL_UNSIGNED_INT, BUFFER_OFFSET(0));
	}
	
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableVertexAttribArray(mIDBillboardSize);
	glDisableVertexAttribArray(mIDBillboardClrCurveS);
	glDisableVertexAttribArray(mIDBillboardClrCurveT);
	glDisableVertexAttribArray(mIDBillboardFilterAlpha);
	glDisableVertexAttribArray(mIDBillboardParamBit);
	if (mIDBillboardTwistID != 0)
	    glDisableVertexAttribArray(mIDBillboardTwistID);
	
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	GetPCShaderCtrl()->EndShader();
    }
}
