// -*- c -*-
// "clsDrawParticleLargeVBO.h"

#include "clsDrawParticleLargeVBO.h"
#include "clsMaterialParticle.h"

#include "../../GLOBJ/clsVBOSphere.h"
#include "../../GLSL/clsShaderCtrl.h"
#include "../../MYGL/clsGLExtention.h"
#include "../../MYGL/clsShadowVolumeCtrl.h"
#include "../../MYGL/clsTextureProvider.h"
#include "../../SIMDATA/clsParticleForRender.h"
#include "../../TIMELINE/clsTimeCtrl.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"
#include "../../COMMON/CommonDefinitionsShader.h"
#include "../../COMMON/clsTensorMath.h"

#include "../../../GL/glsl.h"

#ifdef WIN32
extern PFNGLDELETEBUFFERSPROC glDeleteBuffers;
extern PFNGLGENBUFFERSPROC glGenBuffers;
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
extern PFNGLVERTEXATTRIBIPOINTERPROC glVertexAttribIPointer;
extern PFNGLVERTEXATTRIBI1IPROC glVertexAttribI1i;
#endif

#define CLS ThreadPrepareNextVBOParticleCount_cls

void CLS::
SetParameters(ParticleForRender_cls *pCSnapForRender,
	      int ini, int fin, int offset, int *pCounter){
    mpCSnapForRender = pCSnapForRender;
    mIni = ini; mFin = fin;
    mOffset = offset;
    mpCounter = pCounter;
}

wxThread::ExitCode CLS::
Entry(){
    int ini = mIni + mOffset;
    int fin = mFin + mOffset;
    //int offset = mOffset;
    float const * const pFilterAlpha = mpCSnapForRender->mpFilterAlpha;
    const ArrayAndIndex_cls * const pIndexAndZ = mpCSnapForRender->mpIndexAndZ;

    int count = 0;
    for (int i = ini; i < fin; i++){
	if (pFilterAlpha[pIndexAndZ[i].mIndex] != 0) count++;
    }
    
    *mpCounter = count;
    
    *mpStatusFlagLink = 2;
    return 0;
}

#define CLS ThreadPrepareNextBase_cls

void CLS::
SetParametersBase(ParticleForRender_cls *pCSnapForRender, 
		  DrawParticleLargeVBO_cls *pCCtrl,
		  int PNum, int offset, int tgtOffset, bool bFullCopy){
    mpCSnapForRender = pCSnapForRender;
    mpCCtrl = pCCtrl;
    mPNum = PNum;
    mOffset = offset;
    mTgtOffset = tgtOffset;
    mbFullCopy = bFullCopy;
}

#define CLS ThreadPrepareNextVBOParticleDataPos_cls

void CLS::
SetParameters(ParticleForRender_cls *pCSnapForRender, 
	      DrawParticleLargeVBO_cls *pCCtrl,
	      int PNum, int offset, int tgtOffset, 
	      bool *pbFullCopy, bool bFinThread){
    SetParametersBase(pCSnapForRender, pCCtrl,
		      PNum, offset, tgtOffset, FALSE);
    mbFinThread = bFinThread;
    mpbFullCopy = pbFullCopy;
}

wxThread::ExitCode CLS::
Entry(){
    int PNum = mPNum;
    int offset = mOffset;
    
    float (* const pXCtrl)[3]      = mpCCtrl->mpX;
    float * const pFilterAlphaCtrl = mpCCtrl->mpFilterAlpha;
    
    float const (* const pX)[3]      = mpCSnapForRender->mpX;
    float const * const pFilterAlpha = mpCSnapForRender->mpFilterAlpha;
    ArrayAndIndex_cls const * const pIndexAndZ = mpCSnapForRender->mpIndexAndZ;
    
    int ini = offset;
    int fin = ini + PNum;
    int iTgt = mTgtOffset;
    for (int i = ini; i < fin; i++){
	int index = pIndexAndZ[i].mIndex;
	if (pFilterAlpha[index] == 0) continue;
	
	pXCtrl[iTgt][0]        = pX[index][0];
	pXCtrl[iTgt][1]        = pX[index][1];
	pXCtrl[iTgt][2]        = pX[index][2];
	iTgt++;
    }
    
    // Ō̃XbhŐm
    if (mbFinThread == TRUE)
	mpCCtrl->mNumNext = iTgt;

    if (iTgt - mTgtOffset == PNum) *mpbFullCopy = TRUE;
    else                           *mpbFullCopy = FALSE;
				       
    
    *mpStatusFlagLink = 2;
    return 0;
}

#define CLS ThreadPrepareNextVBOParticleDataST_cls

void CLS::
SetParameters(ParticleForRender_cls *pCSnapForRender, 
	      DrawParticleLargeVBO_cls *pCCtrl,
	      int PNum, int offset, int tgtOffset, bool bFullCopy) {
    SetParametersBase(pCSnapForRender, pCCtrl, PNum, 
		      offset, tgtOffset, bFullCopy);
}

wxThread::ExitCode CLS::
Entry(){
    int PNum = mPNum;
    int offset = mOffset;

    float * const pClrCurveSCtrl   = mpCCtrl->mpClrCurveS;
    float * const pClrCurveTCtrl   = mpCCtrl->mpClrCurveT;
    float * const pFilterAlphaCtrl = mpCCtrl->mpFilterAlpha;
    
    float const * const pClrCurveS   = mpCSnapForRender->mpClrCurveS;
    float const * const pClrCurveT   = mpCSnapForRender->mpClrCurveT;
    float const * const pFilterAlpha = mpCSnapForRender->mpFilterAlpha;
    ArrayAndIndex_cls const * const pIndexAndZ = mpCSnapForRender->mpIndexAndZ;

    int ini = offset;
    int fin = ini + PNum;
    int iTgt = mTgtOffset;
    if (mbFullCopy == TRUE){
	for (int i = ini; i < fin; i++){
	    int index = pIndexAndZ[i].mIndex;
	    pClrCurveSCtrl[iTgt]   = pClrCurveS[index];
	    pClrCurveTCtrl[iTgt]   = pClrCurveT[index];
	    iTgt++;
	}
    }
    else{
	for (int i = ini; i < fin; i++){
	    int index = pIndexAndZ[i].mIndex;
	    if (pFilterAlpha[index] == 0) continue;
	    
	    pClrCurveSCtrl[iTgt]   = pClrCurveS[index];
	    pClrCurveTCtrl[iTgt]   = pClrCurveT[index];
	    iTgt++;
	}
    }

    *mpStatusFlagLink = 2;
    return 0;
}

#define CLS ThreadPrepareNextVBOParticleDataAlpha_cls

void CLS::
SetParameters(ParticleForRender_cls *pCSnapForRender, 
	      DrawParticleLargeVBO_cls *pCCtrl,
	      int PNum, int offset, int tgtOffset, bool bFullCopy) {
    SetParametersBase(pCSnapForRender, pCCtrl, PNum, 
		      offset, tgtOffset, bFullCopy);
}

wxThread::ExitCode CLS::
Entry(){
    int PNum = mPNum;
    int offset = mOffset;

    float * const pSizeCtrl        = mpCCtrl->mpSize;
    float * const pFilterAlphaCtrl = mpCCtrl->mpFilterAlpha;
    
    float const * const pSize        = mpCSnapForRender->mpSize;
    float const * const pSizeMod     = mpCSnapForRender->mpSizeMod;
    float const * const pFilterAlpha = mpCSnapForRender->mpFilterAlpha;
    ArrayAndIndex_cls const * const pIndexAndZ = mpCSnapForRender->mpIndexAndZ;
    
    int ini = offset;
    int fin = ini + PNum;
    int iTgt = mTgtOffset;
    if (pSizeMod == NULL){
	if (mbFullCopy == TRUE){
	    for (int i = ini; i < fin; i++){
		int index = pIndexAndZ[i].mIndex;
		pSizeCtrl[iTgt]        = pSize[index];
		pFilterAlphaCtrl[iTgt] = pFilterAlpha[index];
		iTgt++;
	    }
	}
	else{
	    for (int i = ini; i < fin; i++){
		int index = pIndexAndZ[i].mIndex;
		float filterAlpha = pFilterAlpha[index];
		if (filterAlpha == 0) continue;
		
		pSizeCtrl[iTgt]        = pSize[index];
		pFilterAlphaCtrl[iTgt] = filterAlpha;
		iTgt++;
	    }
	}
    }
    else{
	if (mbFullCopy == TRUE){
	    for (int i = ini; i < fin; i++){
		int index = pIndexAndZ[i].mIndex;
		pSizeCtrl[iTgt]        = pSize[index] * pSizeMod[index];
		pFilterAlphaCtrl[iTgt] = pFilterAlpha[index];
		iTgt++;
	    }
	}
	else{
	    for (int i = ini; i < fin; i++){
		int index = pIndexAndZ[i].mIndex;
		float filterAlpha = pFilterAlpha[index];
		if (filterAlpha == 0) continue;
		
		pSizeCtrl[iTgt]        = pSize[index] * pSizeMod[index];
		pFilterAlphaCtrl[iTgt] = filterAlpha;
		iTgt++;
	    }
	}
    }

    *mpStatusFlagLink = 2;
    return 0;
}

#define CLS ThreadPrepareNextVBOParticleDataOther_cls

void CLS::
SetParameters(ParticleForRender_cls *pCSnapForRender, 
	      DrawParticleLargeVBO_cls *pCCtrl,
	      int PNum, int offset, int tgtOffset, bool bFullCopy) {
    SetParametersBase(pCSnapForRender, pCCtrl, PNum, 
		      offset, tgtOffset, bFullCopy);
}

wxThread::ExitCode CLS::
Entry(){
    int PNum = mPNum;
    int offset = mOffset;

    float * const pFilterAlphaCtrl = mpCCtrl->mpFilterAlpha;
    unsigned char * const pTypeCtrl     = mpCCtrl->mpType;
    int * const pParamBitCtrl = mpCCtrl->mpParamBit;
    int * const pTwistIDCtrl = mpCCtrl->mpTwistID;
    
    float const * const pFilterAlpha = mpCSnapForRender->mpFilterAlpha;
    unsigned char const * const pType     = mpCSnapForRender->mpType;
    unsigned char const * const pParamBit = mpCSnapForRender->mpParamBit;
    ArrayAndIndex_cls const * const pIndexAndZ = mpCSnapForRender->mpIndexAndZ;
    int const * const pTwistID = mpCSnapForRender->mpID; 
    
    int ini = offset;
    int fin = ini + PNum;
    int iTgt = mTgtOffset;
    if (mbFullCopy == TRUE){
	for (int i = ini; i < fin; i++){
	    int index = pIndexAndZ[i].mIndex;
	    pTypeCtrl[iTgt]        = pType[index];
	    pParamBitCtrl[iTgt]    = (int)pParamBit[index];
	    pTwistIDCtrl[iTgt]     = pTwistID[index];
	    iTgt++;
	}
    }
    else{
	for (int i = ini; i < fin; i++){
	    int index = pIndexAndZ[i].mIndex;
	    if (pFilterAlpha[index] == 0) continue;

	    pTypeCtrl[iTgt]        = pType[index];
	    pParamBitCtrl[iTgt]    = (int)pParamBit[index];
	    pTwistIDCtrl[iTgt]     = pTwistID[index];
	    iTgt++;
	}
    }

    *mpStatusFlagLink = 2;
    return 0;
}

#define CLS ThreadPrepareNextVBOParticleDataElement_cls

void CLS::
SetParameters(DrawParticleLargeVBO_cls *pCCtrl, int PNum, bool bTwoPass){
    mpCCtrl = pCCtrl;
    mPNum = PNum;
    mbTwoPass = bTwoPass;
}

wxThread::ExitCode CLS::
Entry(){
    int PNum = mPNum;

    float * const pFilterAlphaCtrl = mpCCtrl->mpFilterAlpha;
    int * const pParamBitCtrl = mpCCtrl->mpParamBit;
    unsigned int * const pElementArray0 = mpCCtrl->mpElementArray0;
    unsigned int * const pElementArray1 = mpCCtrl->mpElementArray1;

    if (mbTwoPass == TRUE){
		
	int index0 = 0, index1 = 0;
	for (int i = 0; i < PNum; i++){
	    if (pFilterAlphaCtrl[i] != 0)
		if (pParamBitCtrl[i] % 2 == 0){
		    pElementArray0[index0] = i;
		    index0++;
		}
		else{
		    pElementArray1[index1] = i;
		    index1++;
		}
	    }
	
	mpCCtrl->mNumValidNext[0] = index0;
	mpCCtrl->mNumValidNext[1] = index1;
    }
    else{
	int index = 0;
	for (int i = 0; i < PNum; i++){
	    if (pFilterAlphaCtrl[i] == 0) continue;
	    pElementArray0[index] = i;
	    index++;
	}
	
	mpCCtrl->mNumValidNext[0] = index;
    }

    *mpStatusFlagLink = 2;
    return 0;
}

#define CLS DrawParticleLargeVBO_cls

#define BUFFER_OFFSET(bytes) ((GLubyte *)NULL + (bytes))

CLS::
DrawParticleLargeVBO_cls(){
    mNumPack = 0;
        
    mpX    = NULL;
    mpSize = NULL;
    mpClrCurveS = NULL; mpClrCurveT = NULL;
    mpFilterAlpha = NULL;
    mpType     = NULL;
    mpParamBit = NULL;
    mpTwistID = NULL;
    mpElementArray0 = NULL; mpElementArray1 = NULL;
}

CLS::
~DrawParticleLargeVBO_cls(){
    ClearVBO();
    ReleaseMemoryData();
}

void CLS::
ClearVBO(){
    ClearVBOBase();
    for (int i = 0; i < 2; i++){
	mNumValid[i]     = 0;
	mNumValidNext[i] = 0;
    }
    mNumNext = 0;
}

void CLS::
Allocate(int numPack, bool bTwoPass){
    wxString strLog;
    wxStopWatch CSW;
    CSW.Start();

    mNumPack = numPack;
    mpX    = new float [numPack][3];
    mpSize = new float [numPack];
    mpClrCurveS   = new float [numPack];
    mpClrCurveT   = new float [numPack];
    mpFilterAlpha = new float [numPack];
    mpType     = new unsigned char [numPack];
    mpParamBit = new int [numPack];
    mpTwistID = new int [numPack];
    mpElementArray0 = new unsigned int [numPack];
    if (bTwoPass == TRUE)
	mpElementArray1 = new unsigned int [numPack];

    strLog << "Large VBO Allocation " << CSW.Time() << " ms" << ENDL;
    wxLogMessage(strLog);
}

void CLS::
ReleaseMemoryData(){
    if (mpX != NULL){
	delete [] mpX; mpX = NULL; }
    if (mpSize != NULL){
	delete [] mpSize; mpSize = NULL; }
    if (mpClrCurveS != NULL){
	delete [] mpClrCurveS; mpClrCurveS = NULL; }
    if (mpClrCurveT != NULL){
	delete [] mpClrCurveT; mpClrCurveT = NULL; }
    if (mpFilterAlpha != NULL){
	delete [] mpFilterAlpha; mpFilterAlpha = NULL; }
    if (mpType != NULL){
	delete [] mpType; mpType = NULL; }
    if (mpParamBit != NULL){
	delete [] mpParamBit; mpParamBit = NULL; }
    if (mpTwistID != NULL){
	delete [] mpTwistID; mpTwistID = NULL; }
    if (mpElementArray0 != NULL){
	delete [] mpElementArray0; mpElementArray0 = NULL; }
    if (mpElementArray1 != NULL){
	delete [] mpElementArray1; mpElementArray1 = NULL; }
}

#include <wx/stopwatch.h>

void CLS::
PrepareNextDataPos(ParticleForRender_cls *pCSnapForRender, int offset){
    int PNum = mNumPack;
    int PNumSRC = pCSnapForRender->mPNum;
    if (offset + PNum > PNumSRC) PNum = PNumSRC - offset;

    ArrayAndIndex_cls *pIndexAndZ = pCSnapForRender->mpIndexAndZ;
    float const *pFilterAlpha = pCSnapForRender->mpFilterAlpha;

    wxStopWatch CSW;
    CSW.Start();

    int counter[3];
    mpStatusFlagListCount = GetNewPStatusList(3);
    ThreadPrepareNextVBOParticleCount_cls *pCThreadCount[3];
    for (int i = 0; i < 3; i++)
	pCThreadCount[i] = new ThreadPrepareNextVBOParticleCount_cls;

    mpStatusFlagListDataPos = GetNewPStatusList(4);
    ThreadPrepareNextVBOParticleDataPos_cls *pCThread[4];
    for (int i = 0; i < 4; i++)
	pCThread[i] = new ThreadPrepareNextVBOParticleDataPos_cls;

    // ŏ̃XbhcounterKvȂ̂
    // tCOŃX^[gėǂ
    pCThread[0]->SetParameters(pCSnapForRender, this,
			       PNum/4       , offset           , 0, 
			       &mbFullCopy[0]);
    
    pCThread[0]->CreateThread(mpStatusFlagListDataPos+0);
    pCThread[0]->Run();
        
    
    pCThreadCount[0]->SetParameters(pCSnapForRender, 
				    0       , PNum/4  , offset, &counter[0]);
    pCThreadCount[1]->SetParameters(pCSnapForRender,
				    PNum/4  , PNum/4*2, offset, &counter[1]);
    pCThreadCount[2]->SetParameters(pCSnapForRender,
				    PNum/4*2, PNum/4*3, offset, &counter[2]);
    
    for (int i = 0; i < 3; i++){
	pCThreadCount[i]->CreateThread(mpStatusFlagListCount+i);
	pCThreadCount[i]->Run();
    }
    
    WaitFinishAndDeleteFlag(mpStatusFlagListCount, 3);

    int tgtOffset1 = counter[0];
    int tgtOffset2 = tgtOffset1 + counter[1];
    int tgtOffset3 = tgtOffset2 + counter[2];

    pCThread[1]->SetParameters(pCSnapForRender, this         , PNum/4,
			       offset+PNum/4  , tgtOffset1    , &mbFullCopy[1]);
    pCThread[2]->SetParameters(pCSnapForRender, this         , PNum/4,
			       offset+PNum/4*2, tgtOffset2   , &mbFullCopy[2]);
    pCThread[3]->SetParameters(pCSnapForRender, this         , PNum-PNum/4*3,
			       offset+PNum/4*3, tgtOffset3   , &mbFullCopy[3],
			       TRUE);
    
    for (int i = 1; i < 4; i++){
	pCThread[i]->CreateThread(mpStatusFlagListDataPos+i);
	pCThread[i]->Run();
    }

    mTgtOffset1 = tgtOffset1;
    mTgtOffset2 = tgtOffset2;
    mTgtOffset3 = tgtOffset3;

}

void CLS::
PrepareNextDataST(ParticleForRender_cls *pCSnapForRender, int offset){

    int PNum = mNumPack;
    int PNumSRC = pCSnapForRender->mPNum;
    if (offset + PNum > PNumSRC) PNum = PNumSRC - offset;

    ArrayAndIndex_cls *pIndexAndZ = pCSnapForRender->mpIndexAndZ;
    float const *pFilterAlpha = pCSnapForRender->mpFilterAlpha;

    mpStatusFlagListDataST = GetNewPStatusList(4);
    ThreadPrepareNextVBOParticleDataST_cls *pCThread[4];
    for (int i = 0; i < 4; i++)
	pCThread[i] = new ThreadPrepareNextVBOParticleDataST_cls;

    pCThread[0]->SetParameters(pCSnapForRender, this       , PNum/4,
			       offset         , 0          , mbFullCopy[0]);
    pCThread[1]->SetParameters(pCSnapForRender, this       , PNum/4,
			       offset+PNum/4  , mTgtOffset1, mbFullCopy[1]);
    pCThread[2]->SetParameters(pCSnapForRender, this       , PNum/4,
			       offset+PNum/4*2, mTgtOffset2, mbFullCopy[2]);
    pCThread[3]->SetParameters(pCSnapForRender, this       , PNum-PNum/4*3,
			       offset+PNum/4*3, mTgtOffset3, mbFullCopy[3]);

    for (int i = 0; i < 4; i++){
	pCThread[i]->CreateThread(mpStatusFlagListDataST+i);
	pCThread[i]->Run();
    }
}

void CLS::
PrepareNextDataAlpha(ParticleForRender_cls *pCSnapForRender, int offset){

    int PNum = mNumPack;
    int PNumSRC = pCSnapForRender->mPNum;
    if (offset + PNum > PNumSRC) PNum = PNumSRC - offset;

    ArrayAndIndex_cls *pIndexAndZ = pCSnapForRender->mpIndexAndZ;
    float const *pFilterAlpha = pCSnapForRender->mpFilterAlpha;

    mpStatusFlagListDataAlpha = GetNewPStatusList(4);
    ThreadPrepareNextVBOParticleDataAlpha_cls *pCThread[4];
    for (int i = 0; i < 4; i++)
	pCThread[i] = new ThreadPrepareNextVBOParticleDataAlpha_cls;

    pCThread[0]->SetParameters(pCSnapForRender, this       , PNum/4,
			       offset         , 0          , mbFullCopy[0]);
    pCThread[1]->SetParameters(pCSnapForRender, this       , PNum/4,
			       offset+PNum/4  , mTgtOffset1, mbFullCopy[1]);
    pCThread[2]->SetParameters(pCSnapForRender, this       , PNum/4,
			       offset+PNum/4*2, mTgtOffset2, mbFullCopy[2]);
    pCThread[3]->SetParameters(pCSnapForRender, this       , PNum-PNum/4*3,
			       offset+PNum/4*3, mTgtOffset3, mbFullCopy[3]);

    for (int i = 0; i < 4; i++){
	pCThread[i]->CreateThread(mpStatusFlagListDataAlpha+i);
	pCThread[i]->Run();
    }
}

void CLS::
PrepareNextDataOther(ParticleForRender_cls *pCSnapForRender, int offset){

    int PNum = mNumPack;
    int PNumSRC = pCSnapForRender->mPNum;
    if (offset + PNum > PNumSRC) PNum = PNumSRC - offset;

    ArrayAndIndex_cls *pIndexAndZ = pCSnapForRender->mpIndexAndZ;
    float const *pFilterAlpha = pCSnapForRender->mpFilterAlpha;

    mpStatusFlagListDataOther = GetNewPStatusList(4);
    ThreadPrepareNextVBOParticleDataOther_cls *pCThread[4];
    for (int i = 0; i < 4; i++)
	pCThread[i] = new ThreadPrepareNextVBOParticleDataOther_cls;

    pCThread[0]->SetParameters(pCSnapForRender, this         , PNum/4,
			       offset         , 0            , mbFullCopy[0]);
    pCThread[1]->SetParameters(pCSnapForRender, this         , PNum/4,
			       offset+PNum/4  , mTgtOffset1  , mbFullCopy[1]);
    pCThread[2]->SetParameters(pCSnapForRender, this         , PNum/4,
			       offset+PNum/4*2, mTgtOffset2  , mbFullCopy[2]);
    pCThread[3]->SetParameters(pCSnapForRender, this         , PNum-PNum/4*3,
			       offset+PNum/4*3, mTgtOffset3  , mbFullCopy[3]);

    for (int i = 0; i < 4; i++){
	pCThread[i]->CreateThread(mpStatusFlagListDataOther+i);
	pCThread[i]->Run();
    }
}

void CLS::
PrepareNextDataElement(bool bTwoPass){
    int PNum = mNumNext;

    ThreadPrepareNextVBOParticleDataElement_cls *pCThread = 
	new ThreadPrepareNextVBOParticleDataElement_cls;

    pCThread->SetParameters(this, PNum, bTwoPass);
    
    mpStatusFlagListDataElement = GetNewPStatusList(1);
    pCThread->CreateThread(mpStatusFlagListDataElement+0);
    pCThread->Run();
}

/*
void CLS::
SetVBOElementArrayAll(int EAindex, int PNum){

    float *pFilterAlpha = mpFilterAlpha;
    int numValid = 0;
    for (int i = 0; i < PNum; i++){
	if (pFilterAlpha[i] != 0) numValid++;
    }

    int index = 0;
    unsigned int *pElementArray = new unsigned int [numValid];
    for (int i = 0; i < PNum; i++){
	if (pFilterAlpha[i] == 0) continue;
	pElementArray[index] = i;
	index++;
    }
    
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[EAindex]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, index * sizeof(unsigned int), pElementArray, GL_STATIC_READ);
	
    delete [] pElementArray;
    mNumValid[EAindex] = index;
}

void CLS::
SetVBOElementArraySelective(int EAindex, int PNum, unsigned char blendType){

    float *pFilterAlpha = mpFilterAlpha;
    int *pParamBit = mpParamBit;
    int numValid = 0;
    for (int i = 0; i < PNum; i++){
	if (pFilterAlpha[i] != 0) numValid++;
    }

    int index = 0;
    unsigned int *pElementArray = new unsigned int [numValid];
    for (int i = 0; i < PNum; i++){
	if (pFilterAlpha[i] != 0 &&
	    pParamBit[i] % 2 == blendType){ 
	    pElementArray[index] = i;
	    index++;
	}
    }
    
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[EAindex]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, index * sizeof(unsigned int), pElementArray, GL_STATIC_READ);
    
    delete [] pElementArray;
    mNumValid[EAindex] = index;
}
*/

void CLS::
SetNextVBOPos(){
    int PNum = mNumNext;

    if (mVBOVertexPoint != 0) glDeleteBuffers(1, &mVBOVertexPoint);
    glGenBuffers(1, &mVBOVertexPoint);
    glBindBuffer(GL_ARRAY_BUFFER, mVBOVertexPoint);
    glBufferData(GL_ARRAY_BUFFER, 3 * PNum * sizeof(float), mpX, GL_STATIC_DRAW);
}

void CLS::
SetNextVBOST(){
    int PNum = mNumNext;

    if (mVBOClrCurveSPoint != 0) glDeleteBuffers(1, &mVBOClrCurveSPoint);
    if (mVBOClrCurveTPoint != 0) glDeleteBuffers(1, &mVBOClrCurveTPoint);
    glGenBuffers(1, &mVBOClrCurveSPoint);
    glGenBuffers(1, &mVBOClrCurveTPoint);
    glBindBuffer(GL_ARRAY_BUFFER, mVBOClrCurveSPoint);
    glBufferData(GL_ARRAY_BUFFER, 1 * PNum * sizeof(float), mpClrCurveS, GL_STATIC_READ);
    glBindBuffer(GL_ARRAY_BUFFER, mVBOClrCurveTPoint);
    glBufferData(GL_ARRAY_BUFFER, 1 * PNum * sizeof(float), mpClrCurveT, GL_STATIC_READ);
}

void CLS::
SetNextVBOAlpha(){
    int PNum = mNumNext;

    if (mVBOFilterAlpha != 0) glDeleteBuffers(1, &mVBOFilterAlpha);
    glGenBuffers(1, &mVBOFilterAlpha);
    glBindBuffer(GL_ARRAY_BUFFER, mVBOFilterAlpha);
    glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(float), mpFilterAlpha, GL_STATIC_READ);

    if (mVBOSizePoint != 0) glDeleteBuffers(1, &mVBOSizePoint);
    glGenBuffers(1, &mVBOSizePoint);
    glBindBuffer(GL_ARRAY_BUFFER, mVBOSizePoint);
    glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(float), mpSize, GL_STATIC_READ);
}

void CLS::
SetNextVBOOther(bool bTwoPass){
    int PNum = mNumNext;

    if (mVBOParamBitPoint != 0) glDeleteBuffers(1, &mVBOParamBitPoint);
    glGenBuffers(1, &mVBOParamBitPoint);
        
    glBindBuffer(GL_ARRAY_BUFFER, mVBOParamBitPoint);
    glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(int), mpParamBit, GL_STATIC_READ);

    if (mVBOTwistID != 0) glDeleteBuffers(1, &mVBOTwistID);
    glGenBuffers(1, &mVBOTwistID);

    glBindBuffer(GL_ARRAY_BUFFER, mVBOTwistID);
    glBufferData(GL_ARRAY_BUFFER, PNum * sizeof(int), mpTwistID, GL_STATIC_READ);

    /*
    if (mVBOElementArrayPoint[0] != 0) glDeleteBuffers(1, &mVBOElementArrayPoint[0]);
    glGenBuffers(1, &mVBOElementArrayPoint[0]);
    
    if (mVBOElementArrayPoint[1] != 0) glDeleteBuffers(1, &mVBOElementArrayPoint[1]);
    mVBOElementArrayPoint[1] = 0;
    */
    
    /*
    if (bTwoPass == TRUE){
	glGenBuffers(1, &mVBOElementArrayPoint[1]);
	SetVBOElementArraySelective(0, PNum, 0);
	SetVBOElementArraySelective(1, PNum, 1);
    }
    else{
	SetVBOElementArrayAll(0, PNum);
    }
    */
}

void CLS::
SetNextVBOElement(bool bTwoPass){
    if (mVBOElementArrayPoint[0] != 0) glDeleteBuffers(1, &mVBOElementArrayPoint[0]);
    glGenBuffers(1, &mVBOElementArrayPoint[0]);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[0]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, mNumValidNext[0] * sizeof(unsigned int), mpElementArray0, GL_STATIC_READ);
    mNumValid[0] = mNumValidNext[0];

    if (mVBOElementArrayPoint[1] != 0) glDeleteBuffers(1, &mVBOElementArrayPoint[1]);
    mVBOElementArrayPoint[1] = 0;
    if (bTwoPass == TRUE){
	glGenBuffers(1, &mVBOElementArrayPoint[1]);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArrayPoint[1]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, mNumValidNext[1] * sizeof(unsigned int), mpElementArray1, GL_STATIC_READ);
	mNumValid[1] = mNumValidNext[1];
    }
    
    //wxString strLog;
    //strLog << "mNumValid = " << mNumValid[0] << ENDL;
    //wxLogMessage(strLog);
}

void CLS::
FinishPrepareNextPos(){
    WaitFinishAndDeleteFlag(mpStatusFlagListDataPos, 4);
}

void CLS::
FinishPrepareNextST(){
    WaitFinishAndDeleteFlag(mpStatusFlagListDataST, 4);
}

void CLS::
FinishPrepareNextAlpha(){
    WaitFinishAndDeleteFlag(mpStatusFlagListDataAlpha, 4);
}

void CLS::
FinishPrepareNextOther(){
    WaitFinishAndDeleteFlag(mpStatusFlagListDataOther, 4);
}

void CLS::
FinishPrepareNextElement(){
    WaitFinishAndDeleteFlag(mpStatusFlagListDataElement, 1);
}
