// -*- c -*-
// "clsDrawStreamline.h"

#include "clsDrawStreamline.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../../GLSL/clsShaderCtrl.h"
#include "../../SIMDATA/clsStreamlineForRender.h"
#include "../../MYGL/clsGLExtention.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"
#include "../../COMMON/CommonDefinitionsShader.h"

#include "../../../GL/glsl.h"

#ifdef WIN32
extern PFNGLDELETEBUFFERSPROC glDeleteBuffers;
extern PFNGLGENBUFFERSPROC glGenBuffers;
extern PFNGLBINDBUFFERPROC glBindBuffer;
extern PFNGLBUFFERDATAPROC glBufferData;
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
extern PFNGLVERTEXATTRIBIPOINTERPROC glVertexAttribIPointer;
extern PFNGLVERTEXATTRIBI1IPROC glVertexAttribI1i;
#endif

#define CLS DrawStreamline_cls

#define BUFFER_OFFSET(bytes) ((GLubyte *)NULL + (bytes))


CLS::
DrawStreamline_cls(){
    mpCSnapForRenderUsed = NULL;
    mDrawCountUsed = -1;
    mVBOVertex = 0;
    mVBOU = 0;
    mVBOLineIndex = 0;
    mVBOAlpha = 0;
    mVBOElementArray = 0;
}

CLS::
~DrawStreamline_cls(){
}

void CLS::
SetVBOElementArray(int SNum){
    unsigned int *pElementArray = new unsigned int [SNum+2];

    pElementArray[0] = 0;
    for (int i = 1; i < SNum+1; i++)
	pElementArray[i] = i-1;
    pElementArray[SNum+1] = SNum - 1;

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArray);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, (SNum+2) * sizeof(unsigned int), pElementArray, GL_STATIC_DRAW);

    delete [] pElementArray;
    
    mSNum = SNum;
}

void CLS::
SetVBOBuffer(StreamlineForRender_cls *pCSnapForRender){
    int SNum = pCSnapForRender->mSNum;
    float (*pX)[3] = pCSnapForRender->mpPos;
    float *pU = pCSnapForRender->mpU;
    int   *pLineIndex = pCSnapForRender->mpLineIndex;
    float *pFilterAlpha = pCSnapForRender->mpFilterAlpha;

    bool bRemakeVertex = FALSE;
    if (mpCSnapForRenderUsed != pCSnapForRender ||
	mDrawCountUsed != pCSnapForRender->mDrawCounter) bRemakeVertex = TRUE;

    mpCSnapForRenderUsed = pCSnapForRender;
    mDrawCountUsed = pCSnapForRender->mDrawCounter;
    
    bool bRemakeU =            bRemakeVertex;
    bool bRemakeLineIndex    = bRemakeVertex;
    bool bRemakeAlpha        = bRemakeVertex;
    bool bRemakeElementArray = bRemakeVertex;

    if (bRemakeVertex == TRUE){
	if (mVBOVertex != 0) glDeleteBuffers(1, &mVBOVertex);
	glGenBuffers(1, &mVBOVertex);

	glBindBuffer(GL_ARRAY_BUFFER, mVBOVertex);
	glBufferData(GL_ARRAY_BUFFER, 3 * SNum * sizeof(float), pX, GL_STATIC_DRAW);
    }

    if (bRemakeU == TRUE){
	if (mVBOU != 0) glDeleteBuffers(1, &mVBOU);
	glGenBuffers(1, &mVBOU);

	glBindBuffer(GL_ARRAY_BUFFER, mVBOU);
	glBufferData(GL_ARRAY_BUFFER, 1 * SNum * sizeof(float), pU, GL_STATIC_DRAW);
    }

    if (bRemakeLineIndex == TRUE){
	if (mVBOLineIndex != 0) glDeleteBuffers(1, &mVBOLineIndex);
	glGenBuffers(1, &mVBOLineIndex);

	glBindBuffer(GL_ARRAY_BUFFER, mVBOLineIndex);
	glBufferData(GL_ARRAY_BUFFER, 1 * SNum * sizeof(int), pLineIndex, GL_STATIC_DRAW);
    }

    if (bRemakeVertex == TRUE){
	if (mVBOAlpha != 0) glDeleteBuffers(1, &mVBOAlpha);
	glGenBuffers(1, &mVBOAlpha);
	glBindBuffer(GL_ARRAY_BUFFER, mVBOAlpha);
	glBufferData(GL_ARRAY_BUFFER, 1 * SNum * sizeof(int), pFilterAlpha, GL_STATIC_DRAW);
    }
    
    if (bRemakeElementArray == TRUE){
	if (mVBOElementArray != 0) glDeleteBuffers(1, &mVBOElementArray);
	glGenBuffers(1, &mVBOElementArray);
	
	SetVBOElementArray(SNum);
    }
}

void CLS::
Draw(StreamlineForRender_cls *pCSnapForRender, 
     int shaderType, float lineWidth, int mappingType,
     int textureMode, int lightNum, int ensemble, int dissolve,
     double clipBoxMin[3], double clipBoxMax[3],
     double matrix16[16], float smoothWidth, float thinningFactor){
    
    glLineWidth(lineWidth);

    CheckGLError("DrawStreamline");

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    
    glColor4f(1, 1, 1, 1);
    //glEnable(GL_BLEND);
    //glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    int shaderListIndex = 0;
    ShaderParamEnsemble_cls CParamEnsemble(ensemble, dissolve,
					   lightNum, textureMode);
    StreamlineBase_cls CParamStreamline(CParamEnsemble);
    CParamStreamline.mpClipBoxMin = clipBoxMin;
    CParamStreamline.mpClipBoxMax = clipBoxMax;
    CParamStreamline.mpMatrix16 = matrix16;
    CParamStreamline.mSmoothWidth = smoothWidth;
    CParamStreamline.mThinningFactor = thinningFactor;
    switch(shaderType){
    case 0:{
	StreamlineBasicParam_cls CParam;
	CParam.CopyFrom(CParamStreamline);
	GetPCShaderCtrl()->BeginStreamlineBasic(&CParam);
	shaderListIndex = _SHADER_STREAMLINE_BASIC;
	break; }
    case 1:{
	StreamlineTriangleParam_cls CParam;
	CParam.mLineWidth = lineWidth;
	CParam.mMappingType = mappingType;
	CParam.CopyFrom(CParamStreamline);
	glGetIntegerv(GL_VIEWPORT, CParam.mViewport);
	GetPCShaderCtrl()->BeginStreamlineTriangle(&CParam);
	shaderListIndex = _SHADER_STREAMLINE_TRIANGLE;
	break; }
    }
    
    if (shaderType == 1){
	CheckGLError("!!");
	//GetPCShaderCtrl()->EndShader();
    }

    glEnableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    glBindBuffer(GL_ARRAY_BUFFER, mVBOVertex);
    glVertexPointer(3, GL_FLOAT, 0, BUFFER_OFFSET(0));
    glBindBuffer(GL_ARRAY_BUFFER, mVBOU);
    glTexCoordPointer(1, GL_FLOAT, 0, BUFFER_OFFSET(0));

    ShaderItem_cls &CItem = GetPCShaderCtrl()->mVCShader[shaderListIndex];
    int IDlineIndex = glGetAttribLocation(CItem.mProgram, "lineIndexAttrib");
    glEnableVertexAttribArray(IDlineIndex);

    int IDFilterAlpha = glGetAttribLocation(CItem.mProgram, "filterAlphaAttrib");
    glEnableVertexAttribArray(IDFilterAlpha);

    glBindBuffer(GL_ARRAY_BUFFER, mVBOLineIndex);
    glVertexAttribIPointer(IDlineIndex, 1, GL_INT, 0, BUFFER_OFFSET(0));

    glBindBuffer(GL_ARRAY_BUFFER, mVBOAlpha);
    glVertexAttribPointer(IDFilterAlpha, 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
    
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mVBOElementArray);
    switch(shaderType){
    case 0:
	glDrawElements(GL_LINE_STRIP, mSNum, GL_UNSIGNED_INT, BUFFER_OFFSET(sizeof(unsigned int)));
	break;
    case 1:
	glDrawElements(GL_LINE_STRIP_ADJACENCY_EXT, mSNum+2, GL_UNSIGNED_INT, BUFFER_OFFSET(0));
	break;
    }

    CheckGLError("DrawStreamLine!!"); 
    
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableVertexAttribArray(IDlineIndex);
    glDisableVertexAttribArray(IDFilterAlpha);
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

    GetPCShaderCtrl()->EndShader();

    CheckGLError("DrawStreamLine!!");

}


