// -*- c -*-
// "clsDrawVolumeBox.h"

#include "clsDrawVolumeBox.h"

#include "../../MYGL/clsGLExtention.h"
#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../../clsFrameMain.h"

#include <math.h>
#include <algorithm>

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#define S 0
#define T 1

PositionEX_cls::
PositionEX_cls(double posX, double posY, double posZ, unsigned int index) :
    Position_cls(posX, posY, posZ)
{
    mIndex = index;
    mAngleTan = 0;
}

PositionEX_cls::
PositionEX_cls(const PositionEX_cls &CPos){
    LOOP_XYZ mPos[xyz] = CPos.mPos[xyz];
    //LOOP_XYZ mVel[xyz] = CPos.mVel[xyz];
    mIndex = CPos.mIndex;
    mAngleTan = CPos.mAngleTan;
}

bool operator<(const PositionEX_cls& CLeft, const PositionEX_cls& CRight)
{ return CLeft.mAngleTan < CRight.mAngleTan; }

// --------------

#define CLS DrawVolumeBox_cls

float CLS::
CalcAllSlices(Direction_cls CCam, int quality,
	      double clipBoxMin[3], double clipBoxMax[3]){

    mVVCIntersect.clear();
    
    double scaleZdash = sqrt(CCam.mTgt[X]*CCam.mTgt[X]+
			     CCam.mTgt[Y]*CCam.mTgt[Y]+
			     CCam.mTgt[Z]*CCam.mTgt[Z]);
    double interval = 2.0 / quality / scaleZdash;

    //CCam.NormalizeAndOrthogonalize();
    
    for (int i = -quality * 0.87 * scaleZdash ; i <= +quality * 0.87 * scaleZdash; i++){
	double basePos[3] = 
	    {-CCam.mTgt[X] * interval * i / scaleZdash,
	     -CCam.mTgt[Y] * interval * i / scaleZdash,
	     -CCam.mTgt[Z] * interval * i / scaleZdash};
	
	CalcSingleSlice(CCam, basePos, clipBoxMin, clipBoxMax); 
    }
    
    return 2.0/quality;//interval;
}


void CLS::
CalcSingleSlice(Direction_cls CCam, double basePos[3],
		double clipBoxMin[3], double clipBoxMax[3]){

    std::vector < Position_cls > VCIntersect;

    //double clipBoxMin[3] = {0, 0, 0};
    //double clipBoxMax[3] = {1, 1, 1};

    double APos[4][2];
    
    APos[0][S] = clipBoxMin[X]-0.5;    APos[0][T] = clipBoxMin[Y]-0.5;
    APos[1][S] = clipBoxMax[X]-0.5;    APos[1][T] = clipBoxMin[Y]-0.5;
    APos[2][S] = clipBoxMax[X]-0.5;    APos[2][T] = clipBoxMax[Y]-0.5;
    APos[3][S] = clipBoxMin[X]-0.5;    APos[3][T] = clipBoxMax[Y]-0.5;

    if (CCam.mTgt[Z] != 0)
	for (int i = 0; i < 4; i++){
	    float z = 
		(CCam.mTgt[X] * (basePos[X]-APos[i][S]) +
		 CCam.mTgt[Y] * (basePos[Y]-APos[i][T])) / CCam.mTgt[Z] +
		basePos[Z];
	    if (z > clipBoxMin[Z] - 0.5 && z < clipBoxMax[Z] - 0.5)
		VCIntersect.push_back
		    (Position_cls(APos[i][S], APos[i][T], z));
	}

    APos[0][S] = clipBoxMin[Y]-0.5;    APos[0][T] = clipBoxMin[Z]-0.5;
    APos[1][S] = clipBoxMax[Y]-0.5;    APos[1][T] = clipBoxMin[Z]-0.5;
    APos[2][S] = clipBoxMax[Y]-0.5;    APos[2][T] = clipBoxMax[Z]-0.5;
    APos[3][S] = clipBoxMin[Y]-0.5;    APos[3][T] = clipBoxMax[Z]-0.5;

    if (CCam.mTgt[X] != 0)
	for (int i = 0; i < 4; i++){
	    float x = 
		(CCam.mTgt[Y] * (basePos[Y]-APos[i][S]) +
		 CCam.mTgt[Z] * (basePos[Z]-APos[i][T])) / CCam.mTgt[X] +
		basePos[X];
	    if (x > clipBoxMin[X] - 0.5 && x < clipBoxMax[X] - 0.5)
		VCIntersect.push_back
		    (Position_cls(x, APos[i][S], APos[i][T]));
	}

    APos[0][S] = clipBoxMin[Z]-0.5;    APos[0][T] = clipBoxMin[X]-0.5;
    APos[1][S] = clipBoxMax[Z]-0.5;    APos[1][T] = clipBoxMin[X]-0.5;
    APos[2][S] = clipBoxMax[Z]-0.5;    APos[2][T] = clipBoxMax[X]-0.5;
    APos[3][S] = clipBoxMin[Z]-0.5;    APos[3][T] = clipBoxMax[X]-0.5;

    if (CCam.mTgt[Y] != 0)
	for (int i = 0; i < 4; i++){
	    float y =
		(CCam.mTgt[Z] * (basePos[Z]-APos[i][S]) +
		 CCam.mTgt[X] * (basePos[X]-APos[i][T])) / CCam.mTgt[Y] +
		basePos[Y];
	    if (y > clipBoxMin[Y] - 0.5 && y < clipBoxMax[Y] - 0.5)
		VCIntersect.push_back
		    (Position_cls(APos[i][T], y, APos[i][S]));
	}
    
    if (VCIntersect.size() == 0) return;
    
    std::vector < PositionEX_cls > VCProjected;
    for (unsigned int i = 0; i < VCIntersect.size(); i++){

	float x = 
	    CCam.mRgt[X] * VCIntersect[i].mPos[X] + 
	    CCam.mRgt[Y] * VCIntersect[i].mPos[Y] + 
	    CCam.mRgt[Z] * VCIntersect[i].mPos[Z];
	float y = 
	    CCam.mTop[X] * VCIntersect[i].mPos[X] + 
	    CCam.mTop[Y] * VCIntersect[i].mPos[Y] + 
	    CCam.mTop[Z] * VCIntersect[i].mPos[Z];
	float z = 
	    CCam.mTgt[X] * VCIntersect[i].mPos[X] + 
	    CCam.mTgt[Y] * VCIntersect[i].mPos[Y] + 
	    CCam.mTgt[Z] * VCIntersect[i].mPos[Z];
	
	VCProjected.push_back
	    (PositionEX_cls(x, y, z, i));
    }

    int baseIndex = 0; //\[gp̃x[X
    double xMin = VCProjected[0].mPos[X];
    for (unsigned int i = 0; i < VCProjected.size(); i++)
	if (VCProjected[i].mPos[X] < xMin){
	    xMin = VCProjected[i].mPos[X];
	    baseIndex = i;
	}

    for (unsigned int i = 0; i < VCProjected.size(); i++){
	if (i != baseIndex){
	    VCProjected[i].mPos[X] -= VCProjected[baseIndex].mPos[X] ;
	    VCProjected[i].mPos[Y] -= VCProjected[baseIndex].mPos[Y] ;
	}
    }

    VCProjected[baseIndex].mPos[X] = 0;
    VCProjected[baseIndex].mPos[Y] = 0;
    // [baseIndex]_ɂ xyʏɓ_zuꂽ    
    
    // e_̊px
    for (unsigned int i = 0; i < VCProjected.size(); i++){
	if (i == baseIndex)
	    VCProjected[i].mAngleTan = -9.99e9;
	else {
	    if (VCProjected[i].mPos[X] == 0){
		if (VCProjected[i].mPos[Y] < 0 ){
		    VCProjected[i].mAngleTan = -9.99e9;
		    VCProjected[baseIndex].mAngleTan = +9.99e9;
		    // baseႢ_base͍Ō
		}
		else
		    VCProjected[i].mAngleTan = +9.90e9;
	    }
	    else 
		VCProjected[i].mAngleTan = VCProjected[i].mPos[Y] / VCProjected[i].mPos[X];
	}
    }
    
    std::sort(VCProjected.begin(), VCProjected.end());

    std::vector < Position_cls > VCIntrsectSorted;

    for (unsigned int i = 0; i < VCProjected.size(); i++)
	VCIntrsectSorted.push_back(VCIntersect[VCProjected[i].mIndex]);

    mVVCIntersect.push_back(VCIntrsectSorted);
}

void CLS::
DrawAllSliceLineForDebug(int quality){
    glColor4f(1, 1, 1, 1);
    
    int layerNum = (int)mVVCIntersect.size();
    for (int index = 0; index < layerNum; index++){

	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_3D);
	glColor4f(1, 1, 0, 1);
	glBegin(GL_LINE_LOOP);
	size_t iSize = mVVCIntersect[index].size();
	for (size_t i = 0; i < iSize; i++){
	    glVertex3d(mVVCIntersect[index][i].mPos[X],
		       mVVCIntersect[index][i].mPos[Y],
		       mVVCIntersect[index][i].mPos[Z]);
	}
	glEnd();
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_3D);
	//glColor4f(1, 1, 1, 1);
    }
}

void CLS::
DrawAllSlice(int quality, float alpha){

    alpha = alpha / quality * 8;

    float luminance = 1;
    if (alpha > 1){ 
	luminance = alpha;
	alpha = 1;
    }

    glColor4f(luminance, luminance, luminance, alpha);
    
    int layerNum = (int)mVVCIntersect.size();
    for (int index = 0; index < layerNum; index++){
	if (index % 20 == 0)
	    GetPCFrameMain()->CheckAbortion();
	if (GetPCTimeCtrl()->msbForceAbort == TRUE) continue;
	DrawSingleSlice(index); 
    }

    GetPCGLExtentions()->
	BlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
			  GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void CLS::
DrawSingleSlice(int index){
    int vertexNum = (int)mVVCIntersect[index].size();
    glBegin(GL_TRIANGLE_FAN);
    for (int i = 0; i < vertexNum; i++){
	glTexCoord3d(mVVCIntersect[index][i].mPos[X] + 0.5,
		     mVVCIntersect[index][i].mPos[Y] + 0.5,
		     mVVCIntersect[index][i].mPos[Z] + 0.5);
	glVertex3dv(mVVCIntersect[index][i].mPos);
    }
    glEnd();
}
