// (setq-c-offset 4)
// "clsHeightField.cpp"

#include "clsHeightField.h"
#include "clsColorCurve.h"
#include "../clsMaterial.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../../GLOBJ/clsVBOCameraBox.h"
#include "../../GLOBJ/clsVBOHFProvider.h"
#include "../../GLSL/clsShaderCtrl.h"
#include "../../MODIFIER/clsSequenceRemap.h"

#include "../../MYGL/clsShadowVolumeCtrl.h"
#include "../../MYGL/clsTextureProvider.h"

#include "../../SIMDATA/clsImageSequence.h"
#include "../../SIMDATA/clsImageSnap.h"
#include "../../TIMELINE/clsTimeCtrl.h"

#include "../../WIDGETCTRL/INPUTITEM/clsAtoSTextCtrl.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"
#include "../../../clsMain.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#include "../../../wxWidgetEnum.h"

#ifdef WIN32
extern PFNGLACTIVETEXTUREPROC glActiveTexture;
#endif

#define CLS HeightField_cls

CLS::
HeightField_cls(ParentObj_cls *pCParentObj, wxString strName) :
    DataViewerBase_cls(pCParentObj)
{
    mstrClassName = "HeightField_cls";
    
    mpCImageSequence = new ImageSequence_cls;
    
    mpCMaterial = new Material_cls(strName, mstrClassName);
    
    Initialize(strName);

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";

    mCConfList.Add(&mCChannel, "Channel=", "0", strUpdate);
    mCConfList.Add(&mCbUseClrCurve, "UseColorCurve=", "0", strUpdate);
    mCConfList.Add(&mCbRim, "UseRim=", "0", strUpdate);
    mCConfList.Add(&mCbCastShadow, "CastShadow=",  "1", strUpdate);
    mCConfList.Add(&mCbXRay, "XRay=", "0", strUpdate);
    
    ColorCurve_cls *pCColorCurve = new ColorCurve_cls(_("ColorCurve"), GetPCStringProvider()->GetArrayString("RGBA"));
    
    double clrInit[2][5] = {{0, 0, 1, 0, 1},
			    {1, 1, 0, 0, 1}};
    pCColorCurve->SetInitKeyframes(2, clrInit);

    mVpCAttribute.push_back(mpCMaterial);
    mVpCAttribute.push_back(pCColorCurve);
}

CLS::
~HeightField_cls(){
    delete mpCImageSequence;
}

void CLS::
SyncConfListTagHeightField(){
    wxString strObj = mCstrName.GetBodySTR();
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }

    mpCMaterial->SyncConfListTag(strObj);
}

wxString CLS::
GetStrInfo(){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frame, subframe);

    wxString strInfo;
    strInfo << mpCImageSequence->GetStrInfo(FALSE);

    int seqLen = mpCImageSequence->mSeqLen;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");
    if (pCSeqRemap != NULL)
	frame = (int)pCSeqRemap->GetRemappedFrame(NT, frame, seqLen);

    ImageSnap_cls *pCSnap = mpCImageSequence->GetPCSnap(frame);
    if (pCSnap == NULL) return strInfo;
    
    strInfo << pCSnap->GetStrInfo();
    return strInfo;
}

wxString CLS::
GetStrPathCandid(){
    return mpCImageSequence->GetFormat(); }

bool CLS::
IsDataValid(){
    return mpCImageSequence->IsDataValid(); }

void CLS::
CheckFileSequence(){
    int offset = mCSequenceOffset.GetBodyINT();
    mpCImageSequence->Set(mCstrPath.GetBodySTR(), offset);

    mpCImageSequence->CalcMaxOffset(mCbSequence.GetBodyBOOL());
    if (mpCImageSequence->mSeqLen != 1)
	mpCImageSequence->mbTimeDepend = TRUE;

    //mpCCurrent->Clear();
    mstrPathChecked = mCstrPath.GetBodySTR();
    if (mpCImageSequence->IsDataValid() == TRUE)
	GetPCMain()->mCLastOpenedFileHF.SetBodyFromData(mstrPathChecked);
}

void CLS::
GLDrawBox(int status){
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    glColor4f(0.2f, 0.7f, 0.2f, 1.0f);
    if (status == 1) ColorProvider_cls::SetGLColor("Selected");
    if (status == 2) ColorProvider_cls::SetGLColor("Active");

    glTranslated(0, 0, 1);
    GCClassProvider.GetPCVBOCameraBox()->Draw();
    glTranslated(0, 0, -1);
}

void CLS::
ConstructVBO(wxSize CSizeImage, VBOBaseObject_cls *pCVBOBasePlane){
    //pCVBOBasePlane->Clear();
    pCVBOBasePlane->GenBuffersVertex(1);

    int width = CSizeImage.GetWidth(), height = CSizeImage.GetHeight();

    int vertexNum = (width+1)*(height+1);

    float (*pVertex)[3] = new float [vertexNum][3];
    float (*pUV)[2] = new float[vertexNum][2];

    for (int y = 0; y < height+1; y++){
	int indexy = y * (width+1);
	for (int x = 0; x < width+1; x++){
	    int index = (indexy + x);
	    pVertex[index][0] = (float)x/width * 2 - 1;
	    pVertex[index][1] = (float)y/height * 2 - 1;
	    pVertex[index][2] = 1;

	    pUV[index][0] = (float)x/width;
	    pUV[index][1] = (float)y/height;
	}
    }
    
    /*
    for (int y = 0; y < height+1; y++){
	int indexy = y * (width+1);
	for (int x = 0; x < width+1; x++){
	    int index = (indexy + x);
	    pUV[index][0] = (float)x/width;
	    pUV[index][1] = (float)y/height;
	}
    }
    */

    pCVBOBasePlane->MakeVBO(pVertex, NULL, pUV, NULL, NULL, NULL, vertexNum);
    
    int totalIndiceTriangle = width*height * 2 * 3;
    pCVBOBasePlane->mTotalIndiceTriangle = totalIndiceTriangle;
    pCVBOBasePlane->mTotalIndiceAdjacent = totalIndiceTriangle * 2;

    unsigned int *pIndiceTriangle = pCVBOBasePlane->GetPNewTriangle();
    unsigned int *pIndiceAdjacent = pCVBOBasePlane->GetPNewAdjucent();
    
    wxStopWatch CStopWatch;
    CStopWatch.Start();
    for (int y = 0; y < height; y++){
	int indexy = y * width;
	int vindexy = y * (width+1);
	for (int x = 0; x < width; x++){
	    int index = (indexy + x) * 6;
	    int vindex = vindexy + x;
	    pIndiceTriangle[index+0] = vindex;
	    pIndiceTriangle[index+1] = vindex + 1;
	    pIndiceTriangle[index+2] = vindex + (width + 1);
	    pIndiceTriangle[index+3] = vindex + 1;
	    pIndiceTriangle[index+4] = vindex + (width + 1) + 1;
	    pIndiceTriangle[index+5] = vindex + (width + 1);

	    int indexa = (indexy + x) * 12;

	    if (y == 0){
		pIndiceAdjacent[indexa+ 0] = vindex;
		pIndiceAdjacent[indexa+ 1] = vindex + (width + 1);
		pIndiceAdjacent[indexa+ 2] = vindex + 1;
		pIndiceAdjacent[indexa+ 3] = vindex + (width + 1) + 1;
		pIndiceAdjacent[indexa+ 4] = vindex + (width + 1);
		if (x == 0)
		    pIndiceAdjacent[indexa + 5] = vindex + 1;
		else
		    pIndiceAdjacent[indexa+ 5] = vindex + (width + 1) - 1;
		
		pIndiceAdjacent[indexa+ 6] = vindex + 1;
		if (x == width - 1)
		    pIndiceAdjacent[indexa+ 7] = vindex;
		else
		    pIndiceAdjacent[indexa+ 7] = vindex + 1 + 1;
		pIndiceAdjacent[indexa+ 8] = vindex + (width + 1) + 1;
		pIndiceAdjacent[indexa+ 9] = vindex + (width + 1) + (width + 1);
		pIndiceAdjacent[indexa+10] = vindex + (width + 1);
		pIndiceAdjacent[indexa+11] = vindex;
		continue;
	    }

	    if (y == height - 1){
		pIndiceAdjacent[indexa+ 0] = vindex;
		pIndiceAdjacent[indexa+ 1] = vindex - (width + 1) + 1;
		pIndiceAdjacent[indexa+ 2] = vindex + 1;
		pIndiceAdjacent[indexa+ 3] = vindex + (width + 1) + 1;
		pIndiceAdjacent[indexa+ 4] = vindex + (width + 1);
		if (x == 0)
		    pIndiceAdjacent[indexa+ 5] = vindex + 1;
		else
		    pIndiceAdjacent[indexa+ 5] = vindex + (width + 1) - 1;
		
		pIndiceAdjacent[indexa+ 6] = vindex + 1;
		if (x == width - 1)
		    pIndiceAdjacent[indexa+ 7] = vindex + (width + 1);
		else
		    pIndiceAdjacent[indexa+ 7] = vindex + 1 + 1;
		pIndiceAdjacent[indexa+ 8] = vindex + (width + 1) + 1;
		pIndiceAdjacent[indexa+ 9] = vindex + 1;
		pIndiceAdjacent[indexa+10] = vindex + (width + 1);
		pIndiceAdjacent[indexa+11] = vindex;

		continue;
	    }

	    if (x == 0){
		pIndiceAdjacent[indexa+ 0] = vindex;
		pIndiceAdjacent[indexa+ 1] = vindex - (width + 1) + 1;
		pIndiceAdjacent[indexa+ 2] = vindex + 1;
		pIndiceAdjacent[indexa+ 3] = vindex + (width + 1) + 1;
		pIndiceAdjacent[indexa+ 4] = vindex + (width + 1);
		pIndiceAdjacent[indexa+ 5] = vindex + 1;
		
		pIndiceAdjacent[indexa+ 6] = vindex + 1;
		pIndiceAdjacent[indexa+ 7] = vindex + 1 + 1;
		pIndiceAdjacent[indexa+ 8] = vindex + (width + 1) + 1;
		pIndiceAdjacent[indexa+ 9] = vindex + (width + 1) + (width + 1);
		pIndiceAdjacent[indexa+10] = vindex + (width + 1);
		pIndiceAdjacent[indexa+11] = vindex;

		continue;
	    }

	    if (x == width-1){
		pIndiceAdjacent[indexa+ 0] = vindex;
		pIndiceAdjacent[indexa+ 1] = vindex - (width + 1) + 1;
		pIndiceAdjacent[indexa+ 2] = vindex + 1;
		pIndiceAdjacent[indexa+ 3] = vindex + (width + 1) + 1;
		pIndiceAdjacent[indexa+ 4] = vindex + (width + 1);
		pIndiceAdjacent[indexa+ 5] = vindex + (width + 1) - 1;
		
		pIndiceAdjacent[indexa+ 6] = vindex + 1;
		pIndiceAdjacent[indexa+ 7] = vindex + (width + 1);
		pIndiceAdjacent[indexa+ 8] = vindex + (width + 1) + 1;
		pIndiceAdjacent[indexa+ 9] = vindex + (width + 1) + (width + 1);
		pIndiceAdjacent[indexa+10] = vindex + (width + 1);
		pIndiceAdjacent[indexa+11] = vindex;

		continue;
	    }

	    pIndiceAdjacent[indexa+ 0] = vindex;
	    pIndiceAdjacent[indexa+ 1] = vindex - (width + 1) + 1;
	    pIndiceAdjacent[indexa+ 2] = vindex + 1;
	    pIndiceAdjacent[indexa+ 3] = vindex + (width + 1) + 1;
	    pIndiceAdjacent[indexa+ 4] = vindex + (width + 1);
	    pIndiceAdjacent[indexa+ 5] = vindex + (width + 1) - 1;
	    
	    pIndiceAdjacent[indexa+ 6] = vindex + 1;
	    pIndiceAdjacent[indexa+ 7] = vindex + 1 + 1;
	    pIndiceAdjacent[indexa+ 8] = vindex + (width + 1) + 1;
	    pIndiceAdjacent[indexa+ 9] = vindex + (width + 1) + (width + 1);
	    pIndiceAdjacent[indexa+10] = vindex + (width + 1);
	    pIndiceAdjacent[indexa+11] = vindex;
	}
    }

    int time = CStopWatch.Time();
    wxString strTime = "HFConstruct::"; strTime << time << "ms\n";
    wxLogMessage(strTime);
	
    //ConstructAdjStructure(pVertex, pIndiceTriangle, pIndiceAdjacent, totalIndiceTriangle / 3, 0, height * 3);

    int time2 = CStopWatch.Time();
    wxString strTime2 = "withAdjacentConstruct::"; strTime2 << time2 << "ms\n";
    wxLogMessage(strTime2);
    
    pCVBOBasePlane->GenBuffersIndice();
    pCVBOBasePlane->MakeVBOIndicesTriangle(pIndiceTriangle);
    pCVBOBasePlane->MakeVBOIndicesAdjacent(pIndiceAdjacent);
    	    
    delete [] pVertex;
    delete [] pUV;
    delete [] pIndiceTriangle;
    delete [] pIndiceAdjacent;
}

void CLS::
ConstructVBORim(wxSize CSizeImage){
    mCVBOBaseRim.Clear();
    mCVBOBaseRim.GenBuffersVertex(1);

    int width = CSizeImage.GetWidth(), height = CSizeImage.GetHeight();
    
    // 4[̒_͏dƂ
    int vertexNum = ((width+1)*2 + (height+1)*2) * 2;

    float (*pVertex)[3] = new float [vertexNum][3];
    float (*pUV)[2] = new float[vertexNum][2];

    // 
    for (int x = 0; x < width+1; x++){
	int index = x * 2;
	pVertex[index][0] = (float)x/width * 2 - 1;
	pVertex[index][1] = -1;
	pVertex[index][2] = 0;
	pVertex[index+1][0] = (float)x/width * 2 - 1;
	pVertex[index+1][1] = -1;
	pVertex[index+1][2] = 1;
	pUV[index][0] = (float)x/width;
	pUV[index][1] = 0;
	pUV[index+1][0] = (float)x/width;
	pUV[index+1][1] = 0;
    }

    // 
    for (int x = 0; x < width+1; x++){
	int index = x * 2 + (width+1) * 2;
	pVertex[index][0] = (float)x/width * 2 - 1;
	pVertex[index][1] = +1;
	pVertex[index][2] = 0;
	pVertex[index+1][0] = (float)x/width * 2 - 1;
	pVertex[index+1][1] = +1;
	pVertex[index+1][2] = 1;
	pUV[index][0] = (float)x/width;
	pUV[index][1] = 1;
	pUV[index+1][0] = (float)x/width;
	pUV[index+1][1] = 1;
    }
    
    // 
    for (int y = 0; y < height+1; y++){
	int index = y * 2 + (width+1) * 2 * 2;
	pVertex[index][0] = -1;
	pVertex[index][1] = (float)y/height * 2 - 1;
	pVertex[index][2] = 0;
	pVertex[index+1][0] = -1;
	pVertex[index+1][1] = (float)y/height * 2 - 1;
	pVertex[index+1][2] = 1;
	pUV[index][0] = 0;
	pUV[index][1] = (float)y/height;
	pUV[index+1][0] = 0;
	pUV[index+1][1] = (float)y/height;
    }

    // E
    for (int y = 0; y < height+1; y++){
	int index = y * 2 + (width+1) * 2 * 2 + (height+1) * 2;
	pVertex[index][0] = 1;
	pVertex[index][1] = (float)y/height * 2 - 1;
	pVertex[index][2] = 0;
	pVertex[index+1][0] = 1;
	pVertex[index+1][1] = (float)y/height * 2 - 1;
	pVertex[index+1][2] = 1;
	pUV[index][0] = 1;
	pUV[index][1] = (float)y/height;
	pUV[index+1][0] = 1;
	pUV[index+1][1] = (float)y/height;
    }
    
    mCVBOBaseRim.MakeVBO(pVertex, NULL, pUV, NULL, NULL, NULL, vertexNum);

    int totalIndiceTriangle = (width*2+height*2)*2 * 3;
    mCVBOBaseRim.mTotalIndiceTriangle = totalIndiceTriangle;
    mCVBOBaseRim.mTotalIndiceAdjacent = totalIndiceTriangle * 2;
    
    unsigned int *pIndiceTriangle = mCVBOBaseRim.GetPNewTriangle();
    unsigned int *pIndiceAdjacent = mCVBOBaseRim.GetPNewAdjucent();

    // 
    for (int x = 0; x < width; x++){
	int index = x * 6;
	int vindex = x * 2;
	pIndiceTriangle[index+0] = vindex;
	pIndiceTriangle[index+1] = vindex+2;
	pIndiceTriangle[index+2] = vindex+1;
	pIndiceTriangle[index+3] = vindex+2;
	pIndiceTriangle[index+4] = vindex+3;
	pIndiceTriangle[index+5] = vindex+1;
    }

    // 
    for (int x = 0; x < width; x++){
	int index = x * 6 + (width)*6;
	int vindex = x * 2 + (width+1)*2;
	pIndiceTriangle[index+0] = vindex;
	pIndiceTriangle[index+1] = vindex+1;
	pIndiceTriangle[index+2] = vindex+2;
	pIndiceTriangle[index+3] = vindex+2;
	pIndiceTriangle[index+4] = vindex+1;
	pIndiceTriangle[index+5] = vindex+3;
    }

    // 
    for (int y = 0; y < height; y++){
	int index = y * 6 + (width)*6*2 ;
	int vindex = y * 2 + (width+1)*2*2;
	pIndiceTriangle[index+0] = vindex;
	pIndiceTriangle[index+1] = vindex+1;
	pIndiceTriangle[index+2] = vindex+2;
	pIndiceTriangle[index+3] = vindex+2;
	pIndiceTriangle[index+4] = vindex+1;
	pIndiceTriangle[index+5] = vindex+3;
    }
    
    // E
    for (int y = 0; y < height; y++){
	int index = y * 6 + (width)*6*2 + (height)*6;
	int vindex = y * 2 + (width+1)*2*2 + (height+1)*2;
	pIndiceTriangle[index+0] = vindex;
	pIndiceTriangle[index+1] = vindex+2;
	pIndiceTriangle[index+2] = vindex+1;
	pIndiceTriangle[index+3] = vindex+2;
	pIndiceTriangle[index+4] = vindex+3;
	pIndiceTriangle[index+5] = vindex+1;
    }

    ConstructAdjStructure(pVertex, pIndiceTriangle, pIndiceAdjacent, totalIndiceTriangle / 3, 0, height * 3);

    mCVBOBaseRim.GenBuffersIndice();
    mCVBOBaseRim.MakeVBOIndicesTriangle(pIndiceTriangle);
    mCVBOBaseRim.MakeVBOIndicesAdjacent(pIndiceAdjacent);
    
    delete [] pVertex;
    delete [] pUV;
    delete [] pIndiceTriangle;
    delete [] pIndiceAdjacent;
}

void CLS::
ConstructVBOFloor(){
    mCVBOBaseFloor.Clear();
    mCVBOBaseFloor.GenBuffersVertex(1);

    int vertexNum = 4;
    float (*pVertex)[3] = new float [vertexNum][3];
    float (*pUV)[2] = new float[vertexNum][2];

    pVertex[0][0] = -1; pVertex[0][1] = -1; pVertex[0][2] = 0;
    pVertex[1][0] = +1; pVertex[1][1] = -1; pVertex[1][2] = 0;
    pVertex[2][0] = -1; pVertex[2][1] = +1; pVertex[2][2] = 0;
    pVertex[3][0] = +1; pVertex[3][1] = +1; pVertex[3][2] = 0;
    pUV[0][0] = 0; pUV[0][1] = 0;
    pUV[1][0] = 1; pUV[1][1] = 0;
    pUV[2][0] = 0; pUV[2][1] = 1;
    pUV[3][0] = 1; pUV[3][1] = 1;

    mCVBOBaseFloor.MakeVBO(pVertex, NULL, pUV, NULL, NULL, NULL, vertexNum);

    int totalIndiceTriangle = 2 * 3;
    mCVBOBaseFloor.mTotalIndiceTriangle = totalIndiceTriangle;
    mCVBOBaseFloor.mTotalIndiceAdjacent = totalIndiceTriangle * 2;

    unsigned int *pIndiceTriangle = mCVBOBaseFloor.GetPNewTriangle();
    unsigned int *pIndiceAdjacent = mCVBOBaseFloor.GetPNewAdjucent();

    pIndiceTriangle[0*3+0] = 0;
    pIndiceTriangle[0*3+1] = 2;
    pIndiceTriangle[0*3+2] = 1;
    pIndiceTriangle[1*3+0] = 1;
    pIndiceTriangle[1*3+1] = 2;
    pIndiceTriangle[1*3+2] = 3;
    
    ConstructAdjStructure(pVertex, pIndiceTriangle, pIndiceAdjacent, totalIndiceTriangle / 3);

    mCVBOBaseFloor.GenBuffersIndice();
    mCVBOBaseFloor.MakeVBOIndicesTriangle(pIndiceTriangle);
    mCVBOBaseFloor.MakeVBOIndicesAdjacent(pIndiceAdjacent);

    delete [] pVertex;
    delete [] pUV;
    delete [] pIndiceTriangle;
    delete [] pIndiceAdjacent;
}

void CLS::
GLDrawSVUnit(VBOBaseObject_cls *pCVBOBasePlane){
    pCVBOBasePlane->Draw(_VBO_SHADOW_VOLUME_HF, 0, TRUE);
    if (mCbRim.GetBodyBOOL() == TRUE){
	mCVBOBaseRim.Draw(_VBO_SHADOW_VOLUME_HF, 0, TRUE);
	mCVBOBaseFloor.Draw(_VBO_SHADOW_VOLUME_HF, 0, TRUE);
    }
}

void CLS::
GLDrawSV(wxString strImg, VBOBaseObject_cls *pCVBOBasePlane){

    if (mCbCastShadow.GetBodyBOOL() == FALSE){
	glPopAttrib();
	return;
    }

    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();

    pCSV->BeginShaderHF();

    if (pCSV->mCbClipCheck.GetBodyBOOL() == TRUE){
	glDisable(GL_DEPTH_TEST);    
	pCSV->SettingFront();
	pCSV->SetCullingSideHF(2);
	GLDrawSVUnit(pCVBOBasePlane);
	
	pCSV->SettingBack();
	pCSV->SetCullingSideHF(1);
	GLDrawSVUnit(pCVBOBasePlane);
    }

    glEnable(GL_DEPTH_TEST);
    pCSV->SettingFront();
    pCSV->SetCullingSideHF(1);
    GLDrawSVUnit(pCVBOBasePlane);

    pCSV->SettingBack();
    pCSV->SetCullingSideHF(2);
    GLDrawSVUnit(pCVBOBasePlane);

    pCSV->EndShader();

    glPopAttrib();
}

void CLS::
GLDraw(bool bDrawGuideObject, 
       int lightNum, int ensemble, int composit, int status, bool bHittest){

    glEnable(GL_DEPTH_TEST);

    Material_cls *pCMaterial = mpCMaterial->GetPCLinked("HeightField_cls::Primitive_cls::MeshObj_cls::Isosurface_cls");
    bool bWireFrame = pCMaterial->mCbWireframe.GetBodyBOOL();
    
    if (ensemble == 0 && mProcessedInInstanceLoop == 0){
	if (mstrPathChecked != mCstrPath.GetBodySTR())
	    CheckFileSequence();
	pCMaterial->CheckTextureSequence();
    }

    int seqLen = mpCImageSequence->mSeqLen;

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NT, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);

    bool bRepeat = FALSE;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    if (pCSeqRemap != NULL){
	frame1 = pCSeqRemap->GetRemappedFrame(NT, frame1, seqLen);
	interpFactor = pCSeqRemap->GetInterpFactor(NT, frame1, interpFactor);
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    if (interpFactor < 1e-8) interpFactor = 0;
    
    int frame2 = frame1+1;
    ModFrame(frame1, seqLen, bRepeat);
    ModFrame(frame2, seqLen, bRepeat);

    GetPCTimeCtrl()->StoreInterpFactorForComposit(interpFactor);
    GetPCTimeCtrl()->AdjustFrameForComposer
	(&frame1, &frame2, &interpFactor, composit);

    if (mpCImageSequence->mbTimeDepend == TRUE)
	if (frame1 < 0 || frame1 >= seqLen) return;	

    wxString strImg;
    if (mpCImageSequence->mbTimeDepend == TRUE)
	strImg = mpCImageSequence->GetModifiedStr(frame1);
    else
	strImg = mstrPathChecked;
    ImageSnap_cls *pCSnap = mpCImageSequence->GenerateIfNot(frame1);

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glEnable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE1);
    
    int indexImage = -1;
    if (strImg.IsSameAs("None"))
	GetPCTextureProvider()->BindBlankTexture(FALSE);
    else
	indexImage = GetPCTextureProvider()->BindTexture(strImg);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP_TO_EDGE);
    glActiveTexture(GL_TEXTURE0);

    wxSize CSizeImage(1,1);
    if (indexImage != -1){
	CSizeImage = GetPCTextureProvider()->GetTextureSize(indexImage);
	if (pCSnap != NULL)
	    pCSnap->mCSizeLoaded = CSizeImage;
    }

    VBOBaseObject_cls *pCVBOBasePlane = 
	GCClassProvider.GetPCVBOHFProvider()->GetPCVBO(CSizeImage);

    if (GetPCShadowVolumeCtrl()->mSVState == 1 && bWireFrame == FALSE) {
	// OglPopAttrib()͕Kv
	float alpha = mpCMaterial->GetAlpha();
	bool bSkip = GetPCShadowVolumeCtrl()->IsSkip(alpha, ensemble, this);
	if (bSkip == TRUE) return glPopAttrib();

	// GLDrawSVňxPopAttrib
	return GLDrawSV(strImg, pCVBOBasePlane);
    }

    if (bDrawGuideObject == TRUE)
	GLDrawBox(status);

    //if (mpCImageSequence->IsDataValid() == FALSE) return;

    int textureMode = 0;
    pCMaterial->BindTexture(frame1);
    
    if (mCbUseClrCurve.GetBodyBOOL() == TRUE){
	ColorCurve_cls* pCColorCurve = GetPCColorCurve(0);
	pCColorCurve->PrepareTexture();
	pCColorCurve->BindTexture(2);
	textureMode = 1;
    }
    else{
	//glActiveTexture(GL_TEXTURE2);
	GetPCTextureProvider()->BindBlankTexture(FALSE, 2);
	//glActiveTexture(GL_TEXTURE0);
    }

    if (pCVBOBasePlane->mVBOVertex == 0) 
	ConstructVBO(CSizeImage, pCVBOBasePlane);
	
    if (mCVBOBaseFloor.mVBOVertex == 0) ConstructVBOFloor();
    if (CSizeImage != mCVBOSize){
	ConstructVBORim(CSizeImage);
	mCVBOSize = CSizeImage;
    }

    if (bHittest == FALSE){ // łƏxȂ̂H

	if (bWireFrame == TRUE)
	    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	
	ProcTexSpaceModBegin();

	ShaderHF_cls *pCParam = new ShaderHF_cls;
	pCParam->mLightNumber = lightNum;
	pCParam->mEnsemble = ensemble;
	pCParam->mDissolveMode = pCMaterial->mCShaderDissolve.GetBodyINT();
	pCParam->mTextureMode = textureMode;
	pCParam->mbFlat = pCMaterial->mCbShaderFlat.GetBodyBOOL();
	pCParam->mChannel = mCChannel.GetBodyINT();
	GetPCShaderCtrl()->BeginHF(pCParam);
	delete pCParam;
	
	CheckGLError("HeightField::A");
	
	glEnable(GL_BLEND);
	pCMaterial->SetBlendFunc();
	pCMaterial->SetDepthMask();
	
	pCMaterial->SetGLMaterialSetting();
	
	pCVBOBasePlane->Draw(_VBO_OBJECT, 0, TRUE);

	if (mCbRim.GetBodyBOOL() == TRUE){
	    GetPCShaderCtrl()->SetHFFlat(TRUE);
	    mCVBOBaseRim.Draw(_VBO_OBJECT, 0, TRUE);
	    mCVBOBaseFloor.Draw(_VBO_OBJECT, 0, TRUE);
	}
	
	CheckGLError("HeightField::B");
	
	GetPCShaderCtrl()->EndShader();

	ProcTexSpaceModEnd();

	if (bWireFrame == TRUE)
	    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    glDisable(GL_TEXTURE_2D);

    glPopAttrib();
}

void CLS::
SetPathGUI(MultiInputPanel_cls *pCMNum, wxString strPath){
    InputItem_cls* pCInputItem = pCMNum->GetPCInputItem(&mCstrPath);
    if (pCInputItem == NULL) return;
    ((AtoSTextCtrl_cls*)(pCInputItem->mpCWindowLink))->SetValue(strPath);
    pCInputItem->OnChanged(FALSE);
}

void CLS::
Reload(){
    mpCImageSequence->ClearSequence();
    mstrPathChecked = "None";
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName()/*, TRUE*/);

    mIndexPath = pC->CombineSizer(2, _("File Path"));
    pC->CombineConfItem(mCstrPath);
    pC->CombineButton(btnHFPath, "..", 30);

    pC->CombineSizer(2, _("Offset"));
    pC->CombineConfItem(mCSequenceOffset);
    pC->CombineConfItem(mCbSequence, _("Image Sequence"));

    pC->CombineSizer(2, _("Rim/Channel"));
    pC->CombineConfItem(mCbRim, _("Rim"));
    pC->SetChoiceList(5, GetPCStringProvider()->GetAstr("HFChannel"));
    pC->CombineConfItem(mCChannel);
    
    pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(mCbXRay, _("X-Ray"));
    pC->CombineConfItem(mCbCastShadow, _("Cast Shadow"));

    pC->CombineButton(btnReload, _("Reload"));
    
    pC->FinishSetting();
    return pC;
}

bool CLS::
HideAndShowTexture(MultiInputPanel_cls *pCMNumTex, int indexTexturePath){
    bool bChange = FALSE;
    bool bTex = mpCMaterial->mCstrTexturePath.GetBodySTR().IsSameAs("None") == FALSE;
    int indexTextureOffset = mpCMaterial->mIndexTextureOffset;
    int indexMapping       = mpCMaterial->mIndexMapping;
    int indexInterpolate   = mpCMaterial->mIndexInterpolate;
    
    if (bTex == FALSE)
	pCMNumTex->mVCItem[indexTexturePath+3].mpCWindowLink->Disable();
    else
	pCMNumTex->mVCItem[indexTexturePath+3].mpCWindowLink->Enable();
    bChange |= pCMNumTex->mVCItem[indexTextureOffset].Show(bTex);
    bChange |= pCMNumTex->mVCItem[indexMapping].Show(FALSE);
    bChange |= pCMNumTex->mVCItem[indexInterpolate].Show(bTex);
    return bChange;
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCMaterial->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeMaterial(wxString& strOut){ mpCMaterial->Serialize(strOut); }


