// (setq-c-offset 4)
// "clsIsosurface.cpp"

#include "clsIsosurface.h"
#include "clsVolumeBox.h"
#include "../clsMaterial.h"
#include "../clsSceneObjCtrl.h"
#include "../PROP/clsIsosurfProp.h"
#include "ISOSURFACE/clsMarchingCube.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../../MYGL/clsShadowVolumeCtrl.h"
#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../WIDGETCTRL/INPUTITEM/clsAtoSTextCtrl.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"

#include "../../../clsMain.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#include "../../../wxWidgetEnum.h"

#define CLS Isosurface_cls

CLS::Isosurface_cls(ParentObj_cls *pCParentObj, wxString strName) :
    DataViewerBase_cls(pCParentObj)
{

    mstrClassName = "Isosurface_cls";
    Initialize(strName);

    mpCIsosurfProp = new IsosurfProp_cls(strName, "IsosurfProp_cls");

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";

    mCConfList.Add(&mCstrVolumeLink, "VolumeLink=", "None", strUpdate);
    mCConfList.Add(&mCbCastShadow, "CastShadow=",  "1", strUpdate);
    mCConfList.Add(&mCOptimize, "Optimize=", "0", strUpdate);
    mCConfList.Add(&mCbUseVBO, "UseVBO=", "0", strUpdate);

    mpCMaterial = new Material_cls(strName, mstrClassName);

    mVpCAttribute.push_back(mpCMaterial);
    mVpCAttribute.push_back(mpCIsosurfProp);

    mpCMarchingCube = new MarchingCube_cls;
}

CLS::
~Isosurface_cls(){
    delete mpCMarchingCube;
    //delete mpCIsosurfProp;
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName()/*, TRUE*/);

    pC->CombineSizer(2, _("Link"));
    pC->CombineBitmapPanel(btnDropdown, wxES, 14);
    pC->mVCItem.back().mBitmapTemplate = _BITMAP_BTN_DROPDOWN;
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = mstrClassName;

    pC->CombineConfItem(mCstrVolumeLink);
    
    pC->CombineSizer(2, _("Threshold"));
    mIndexThreshold = pC->CombineConfItem(mpCIsosurfProp->mCThreshold);
    wxString AstrList[2] = {_("No Optimize"), _("Optimize")};
    pC->SetChoiceList(2, AstrList);
    pC->CombineConfItem(mCOptimize);

    pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(mCbXRay, _("X-Ray"));
    pC->CombineConfItem(mCbCastShadow, _("Cast Shadow"));

    pC->CombineSizer(1, " ");
    pC->CombineConfItem(mCbUseVBO, _("Use VBO"));

    pC->FinishSetting();
    return pC;
}

void CLS::
SyncConfListTagIsosurf(){
    wxString strObj = mCstrName.GetBodySTR();
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }

    /*
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }
    */
    
    //mpCParticleFormat->SyncConfListTag(strObj);
}

/*
wxString CLS::
GetStrInitNameCandidate(){
    return "Isosurf";
}
*/

void CLS::
GLDrawSV(){
    if (mCbCastShadow.GetBodyBOOL() == FALSE) return;

    glPushMatrix();

    glTranslated(-1, -1, -1);
    glScaled(2, 2, 2);
    
    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
    
    pCSV->BeginShader();

    for (int i = 0; i < 8*8*8; i++){

	if (pCSV->mCbClipCheck.GetBodyBOOL() == TRUE){
	    glDisable(GL_DEPTH_TEST);    
	    pCSV->SettingFront();
	    pCSV->SetCullingSide(2);
	    mpCMarchingCube->GLDrawSV(i);
	    
	    pCSV->SettingBack();
	    pCSV->SetCullingSide(1);
	    mpCMarchingCube->GLDrawSV(i);
	}
    
	glEnable(GL_DEPTH_TEST);
	pCSV->SettingFront();
	pCSV->SetCullingSide(1);
	mpCMarchingCube->GLDrawSV(i);
	
	pCSV->SettingBack();
	pCSV->SetCullingSide(2);
	mpCMarchingCube->GLDrawSV(i);
    }
	
    pCSV->EndShader();

    glPopMatrix();
}

void CLS::
GLDraw(bool bDrawGuideObject, 
       int lightNum, int ensemble, int composit, int status, bool bHittest){

    if (bDrawGuideObject == TRUE)
	GLDrawBox(status);

    SceneObj_cls *pCObj = GetPCSceneObjCtrl()->GetPCFromName(mCstrVolumeLink.GetBodySTR());
    if (pCObj == NULL) return;
    if (pCObj->mstrClassName.IsSameAs("VolumeBox_cls") == FALSE) return;

    VolumeBox_cls* pCVolume = (VolumeBox_cls*)pCObj;

    pCVolume->PrepareDraw(ensemble, composit);

    Material_cls *pCMaterial = mpCMaterial->GetPCLinked("Primitive_cls::MeshObj_cls::HeightField_cls::Isosurface_cls");
    bool bWireFrame = pCMaterial->mCbWireframe.GetBodyBOOL();
    //bool bSkipMain = FALSE;
    //if (bWireFrame == TRUE && status > 0) bSkipMain = TRUE;

    float time = GetPCTimeCtrl()->GetCurrentNT();
    float threshold = mpCIsosurfProp->GetPCParamGroup("Threshold")->mVCParam[0].GetValue(time);
    bool bOptimize = (mCOptimize.GetBodyINT() == 1);
    bool bUseVBO = mCbUseVBO.GetBodyBOOL();
    bool bUseSV = GetPCShadowVolumeCtrl()->mCbUseShadowVolume.GetBodyBOOL();
    mpCMarchingCube->ConstructVertices(pCVolume->mpCCurrent, threshold, bOptimize);

    if (bUseVBO == TRUE) mpCMarchingCube->SetVBO(bUseSV);

    if (bUseVBO == TRUE && GetPCShadowVolumeCtrl()->mSVState == 1 && bWireFrame == FALSE){
	float alpha = pCMaterial->GetAlpha();
	bool bSkip = GetPCShadowVolumeCtrl()->IsSkip(alpha, ensemble, this);
	if (bSkip == TRUE) return;
	return GLDrawSV();
    }

    if (bWireFrame == TRUE)
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    glEnable(GL_BLEND);
    pCMaterial->SetBlendFunc();
    pCMaterial->SetDepthMask();
    pCMaterial->BeginShader(lightNum, ensemble);
    pCMaterial->SetGLMaterialSetting();

    glEnable(GL_LIGHTING);
    
    glPushMatrix();

    glTranslated(-1, -1, -1);
    glScaled(2, 2, 2);

    mpCMarchingCube->GLDraw(bUseVBO);

    glPopMatrix();

    glDisable(GL_TEXTURE_2D);

    mpCMaterial->EndShader();

    if (bWireFrame == TRUE)
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    glDisable(GL_CULL_FACE);
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCMaterial->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeMaterial(wxString& strOut){ mpCMaterial->Serialize(strOut); }


