// (setq-c-offset 4)
// "clsMaterialParticle.h"

#include "clsMaterialParticle.h"
#include "../clsSceneObjCtrl.h"

#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"
#include "../../COMMON/CommonDefinitionsShader.h"

#include "../../../wxWidgetEnum.h"

#define CLS MaterialParticle_cls

CLS::
MaterialParticle_cls(wxString strName, wxString strObjName) : Attribute_cls("MaterialParticle_cls", strName) {

    wxString strUpdate = "VIEW3D";
    wxString strUpdateMat = "VIEW3D::PROPERTY_MATERIAL";

    SetConfListTagForSceneObjAttribute(strName, strObjName);
    //SyncConfListTag(strObjName);
    mCConfList.Add(&mCCustomName, "CustomName=", "MaterialName");
    mCConfList.Add(&mCColorCurveType, "ColorCurveType=", "0", strUpdateMat);
    mCConfList.Add(&mCColorCurveIndex, "ColorCurveIndex=", "0", strUpdate);
    mCConfList.Add(&mCAlphaType, "AlphaType=", "0", strUpdateMat);
    mCConfList.Add(&mCSizeType, "SizeType=", "0", strUpdate);
    mCConfList.Add(&mCBaseType, "BaseType=", "0", strUpdateMat);
    mCConfList.Add(&mCShaderTypePoint, "ShaderTypePoint=", "0", strUpdateMat);
    mCConfList.Add(&mCShaderTypeSphere, "ShaderTypeSphere=", "1", strUpdateMat);
    mCConfList.Add(&mCShaderTypeBillboard, "ShaderTypeBillboard=", "0", strUpdateMat);
    mCConfList.Add(&mCbShaderFlat, "ShaderFlat=", "0", strUpdate);
    mCConfList.Add(&mCToonThreshold, "ToonThreshold=", "0.1", strUpdate);
    mCConfList.Add(&mCToonBoldness, "ToonBoldness=", "0.1", strUpdate);
    
    mCConfList.Add(&mCClrCurveDependType, "ColorCurveDepend=", "3", strUpdate);
    mCConfList.Add(&mCBlendType, "BlendType=", "0", strUpdate);
    mCConfList.Add(&mCTextureType, "TextureType=", "0", strUpdate);
    mCConfList.Add(&mCTextureIndex, "TextureIndex=", "0", strUpdate);
    
    mCEmit.SetItem("Emit=", "0.5 0.5 0.5");
    mCDiffuse.SetItem("Diffuse=", "0.5 0.5 0.5");
    mCSpecular.SetItem("Specular=", "0.2 0.2 0.2");
    mCSizeFactorPoint.SetItem("SizeFactorPoint=", "1");
    mCSizeFactorSphere.SetItem("SizeFactorSphere=", "0.01");
    mCAlphaFactor.SetItem("AlphaFactor=", "1");
    mCAlphaModBase.SetItem("AlphaModBase=", "1");
    mCAlphaModMax.SetItem("AlphaModMax=", "1");

    ParamGroup_cls *pCEmission = GetPCNewParamGroup("Emission", "R", "G", "B");
    InitializeAndPushParam(pCEmission, 0.5, 0.5, 0.5);

    ParamGroup_cls *pCDiffuse = GetPCNewParamGroup("Diffuse", "R", "G", "B");
    InitializeAndPushParam(pCDiffuse, 0.5, 0.5, 0.5);

    ParamGroup_cls *pCSpecular = GetPCNewParamGroup("Specular", "R", "G", "B");
    InitializeAndPushParam(pCSpecular, 0.2, 0.2, 0.2);

    ParamGroup_cls *pCSizeFactor = GetPCNewParamGroup("SizeFactor", "Pt", "Sp");
    InitializeAndPushParam(pCSizeFactor, 1, 0.01);
    
    //AddOneDimParam("SizeFactor", "S", 1);
    AddOneDimParam("AlphaFactor", "A", 1);
    ParamGroup_cls *pCAlphaMod = GetPCNewParamGroup("Alpha Modifier", "Base", "Max");
    InitializeAndPushParam(pCAlphaMod, 1, 1);
}

CLS::
~MaterialParticle_cls(){
}

MultiInputPanel_cls *CLS::
GetPCNewPanelMaterial(wxWindow *pCWindow, wxString strName, 
		      int indexMat, bool bEnableAdd, bool bEnableDelete){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, strName);
    mstrMNumName = strName;

    pC->CombineSizer(4);
    pC->CombineConfItem(mCCustomName);
    wxString AstrBaseType[3] = {_("Point Material"), _("Billboard Material"), _("Sphere Material")};
    pC->SetChoiceList(3, AstrBaseType);
    pC->CombineConfItem(mCBaseType);
    pC->CombineButton(btnAddMaterial, "+", 30, bEnableAdd);
    int index = pC->CombineBitmapPanel(btnMaterialMaterialDelete, "Delete", 24, bEnableDelete);
    
    pC->CombineSizer(3);
    wxString AstrClrCurve[2] = {_("Constant Color"), _("Use Color Curve")};
    pC->SetChoiceList(2, AstrClrCurve);
    pC->CombineConfItem(mCColorCurveType);
        
    wxString AstrDependency[10] = {"Index depend", "Size depend", "ID depend", "Value0", "Value1", "Value2", "X depend", "Y depend", "Z depend", "Sort index"};
    pC->SetChoiceList(10, AstrDependency);
    mIndexClrCurveDepend = pC->CombineConfItem(mCClrCurveDependType, _("Dependency"));
    pC->CombineConfItem(mCColorCurveIndex);
    
    pC->mTextWidth = 80;

    mIndexShader = pC->CombineSizer(4, _("Shader"));
    pC->SetChoiceList(_MAX_POINTSHADER, GetPCStringProvider()->GetAstr("ShaderParticlePoint"));
    pC->CombineConfItem(mCShaderTypePoint);
    pC->SetChoiceList(_MAX_BILLBOARDSHADER, GetPCStringProvider()->GetAstr("ShaderParticleBillboard"));
    pC->CombineConfItem(mCShaderTypeBillboard);
    pC->SetChoiceList(_MAX_SPHERESHADER, GetPCStringProvider()->GetAstr("ShaderParticleSphere"));
    pC->CombineConfItem(mCShaderTypeSphere);
    

    pC->SetChoiceList(3, GetPCStringProvider()->GetAstr("BlendType"));
    pC->CombineConfItem(mCBlendType);
    
    mIndexShaderFlat = pC->CombineConfItem(mCbShaderFlat, _("Flat"));
    mIndexToonOption = pC->CombineSizer(2, _("Threshold/Bold"));
    pC->CombineConfItem(mCToonThreshold);
    pC->CombineConfItem(mCToonBoldness);
    
    pC->CombineSizer(3, _("SizeFactor"));
    wxString AstrSize[2] = {_("Use Data"), _("Constant(1)")};
    pC->SetChoiceList(2, AstrSize);
    pC->CombineConfItem(mCSizeType);
    mIndexSizeFactor = pC->CombineConfItem(mCSizeFactorPoint);
    pC->CombineConfItem(mCSizeFactorSphere);
    //pC->mVCItem[pC->mVCItem.size()-1].mbHideInit = TRUE;

    mIndexEmit = pC->CombineSizer(2, _("Emit"));
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCEmit);
    pC->mContextMenuType = _ITEM_MENU_COLOR;
    mIndexBtnEmit = pC->CombineBitmapPanel(btnMaterialEmit0+indexMat, " ",  30);

    mIndexDiffuse = pC->CombineSizer(2, _("Diffuse"));
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCDiffuse);
    pC->mContextMenuType = _ITEM_MENU_COLOR;
    mIndexBtnDiffuse = pC->CombineBitmapPanel(btnMaterialDiffuse0+indexMat, " ",  30);

    mIndexSpecular = pC->CombineSizer(2, _("Specular"));
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCSpecular);
    pC->mContextMenuType = _ITEM_MENU_COLOR;
    mIndexBtnSpecular = pC->CombineBitmapPanel(btnMaterialSpecular0+indexMat, " ",  30);
    
    
    pC->CombineSizer(2, _("Alpha"));
    wxString AstrAlpha[2] = {_("No Modify"), _("Square Inverse")};
    pC->SetChoiceList(2, AstrAlpha);
    pC->CombineConfItem(mCAlphaType);
    pC->CombineConfItem(mCAlphaFactor);

    mIndexAlphaDepend = pC->CombineSizer(2, _("Base/Max"));
    pC->CombineConfItem(mCAlphaModBase);
    pC->SetLimit(0, 4);
    pC->CombineConfItem(mCAlphaModMax);

    

    pC->CombineSizer(2, _("Texture"));
    wxString AstrTexture[2] = {_("Not Use"), _("Use Texture")};
    pC->SetChoiceList(2, AstrTexture);
    pC->CombineConfItem(mCTextureType);
    pC->CombineConfItem(mCTextureIndex);
    
    pC->FinishSetting();

    BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[index].mpCWindowLink;
    pCBitmap->mbButton = TRUE;
    pCBitmap->AddBitmapStr("wxART_CLOSE_PANEL", "wxART_CLOSE_PANEL_P");
    pCBitmap->SetBitmapStrDisable("wxART_CLOSE_PANEL_D");
    pCBitmap->mBorderX = 3;
    pCBitmap->mBorderY = 4;

    BitmapPanel_cls *pCBitmapEmit = (BitmapPanel_cls*)pC->mVCItem[mIndexBtnEmit].mpCWindowLink;
    pCBitmapEmit->mbButton = TRUE;
    
    BitmapPanel_cls *pCBitmapDiffuse = (BitmapPanel_cls*)pC->mVCItem[mIndexBtnDiffuse].mpCWindowLink;
    pCBitmapDiffuse->mbButton = TRUE;

    BitmapPanel_cls *pCBitmapSpecular = (BitmapPanel_cls*)pC->mVCItem[mIndexBtnSpecular].mpCWindowLink;
    pCBitmapSpecular->mbButton = TRUE;
    
    return pC;
}

bool CLS::
HideAndShowMaterial(MultiInputPanel_cls *pCMNum){
    bool bChange = FALSE;
    bool bDepend = (mCColorCurveType.GetBodyINT() == 1);
    pCMNum->mVCItem[mIndexClrCurveDepend].Enable(bDepend);
    pCMNum->mVCItem[mIndexClrCurveDepend+1].Enable(bDepend);
    bool bAlphaDepend = (mCAlphaType.GetBodyINT() == 1);
    bChange |= pCMNum->mVCItem[mIndexAlphaDepend].Show(bAlphaDepend);

    int type = mCBaseType.GetBodyINT();

    if (type == 0 || type == 1){
	bChange |= pCMNum->mVCItem[mIndexShaderFlat].Show(FALSE);
    }
    else{
	bChange |= pCMNum->mVCItem[mIndexShaderFlat].Show(TRUE);
    }

    switch(type){
    case _PARTCLE_POINTSHADER:
    	bChange |= pCMNum->mVCItem[mIndexSizeFactor+0].Show(TRUE);
	bChange |= pCMNum->mVCItem[mIndexSizeFactor+1].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexShader+1].Show(TRUE);
	bChange |= pCMNum->mVCItem[mIndexShader+2].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexShader+3].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexToonOption].Show(FALSE);
	break;
    case _PARTCLE_BILLBOARDSHADER:
	bChange |= pCMNum->mVCItem[mIndexSizeFactor+0].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexSizeFactor+1].Show(TRUE);
	bChange |= pCMNum->mVCItem[mIndexShader+1].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexShader+2].Show(TRUE);
	bChange |= pCMNum->mVCItem[mIndexShader+3].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexToonOption].Show(FALSE);
	break;
    case _PARTCLE_SPHERESHADER:
	bChange |= pCMNum->mVCItem[mIndexSizeFactor+0].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexSizeFactor+1].Show(TRUE);
	bChange |= pCMNum->mVCItem[mIndexShader+1].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexShader+2].Show(FALSE);
	bChange |= pCMNum->mVCItem[mIndexShader+3].Show(TRUE);
	if (mCShaderTypeSphere.GetBodyINT() == _SPHERESHADER_TOON)
	    bChange |= pCMNum->mVCItem[mIndexToonOption].Show(TRUE);
	else
	    bChange |= pCMNum->mVCItem[mIndexToonOption].Show(FALSE);
	break;
    }

    bool bSphere = 
	(type == _PARTCLE_POINTSHADER && mCShaderTypePoint.GetBodyINT() == _POINTSHADER_PSEUDO_SPHERE) || 
	(type == _PARTCLE_BILLBOARDSHADER && mCShaderTypeBillboard.GetBodyINT() == _BILLBOARDSHADER_PSEUDO_SPHERE) || 
	(type == _PARTCLE_SPHERESHADER);
    bChange |= pCMNum->mVCItem[mIndexDiffuse].Show(bSphere);
    bChange |= pCMNum->mVCItem[mIndexSpecular].Show(bSphere);
    
    return bChange;
}

/*
void CLS::
SyncConfListTag(wxString strObj){
    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strObj << "::" << GetStrName();
    mCConfList.SetTag(strConfListLabel);
}
*/

void CLS::
SyncMaterialParticle(wxString strObj){
    SyncronizeParamAndConf3(strObj, "Emission", mCEmit);
    SyncronizeParamAndConf3(strObj, "Diffuse", mCDiffuse);
    SyncronizeParamAndConf3(strObj, "Specular", mCSpecular);
    SyncronizeParamAndConf2(strObj, "SizeFactor", mCSizeFactorPoint, mCSizeFactorSphere);
    SyncronizeParamAndConf(strObj, "AlphaFactor", mCAlphaFactor);
    SyncronizeParamAndConf2(strObj, "Alpha Modifier", mCAlphaModBase, mCAlphaModMax);
}

/*
void CLS::
SyncronizeParamMaterial3(wxString strObj, wxString strParamGroup, ConfItemFLOATARRAY_cls &CConf){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    for (int i = 0; i < 3; i++){
	float newValue = CConf.GetBodyFLOATARRAY()[i];
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	float oldValue = CParam.GetValue(time);
	if (oldValue != newValue){
	    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam.GetStrName();
	    CParam.SetValueAtCommand(time, newValue, strTag);
	}
    }
}
*/

/*
void CLS::
SyncronizeParamMaterial2(wxString strObj, wxString strParamGroup, ConfItemFLOAT_cls &CConf0, ConfItemFLOAT_cls&CConf1){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    float newValue = CConf0.GetBodyDOUBLE();
    Parameter_cls &CParam0 = pCParamGroup->mVCParam[0];
    float oldValue = CParam0.GetValue(time);

    if (oldValue != newValue){
	wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam0.GetStrName();
	CParam0.SetValueAtCommand(time, newValue, strTag);
    }

    newValue = CConf1.GetBodyDOUBLE();
    Parameter_cls &CParam1 = pCParamGroup->mVCParam[1];
    oldValue = CParam1.GetValue(time);

    if (oldValue != newValue){
	wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam1.GetStrName();
	CParam1.SetValueAtCommand(time, newValue, strTag);
    }
}
*/

/*
void CLS::
SyncronizeParamMaterial(wxString strObj, wxString strParamGroup, ConfItemFLOAT_cls &CConf){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    float newValue = CConf.GetBodyDOUBLE();
    Parameter_cls &CParam = pCParamGroup->mVCParam[0];
    float oldValue = CParam.GetValue(time);

    if (oldValue != newValue){
	wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam.GetStrName();
	CParam.SetValueAtCommand(time, newValue, strTag);
    }
}
*/

void CLS::
UpdateButtonColor(MNumPainBase_cls *pCPain){
    MultiInputPanel_cls* pCurr = pCPain->GetPCMNum(mstrMNumName);
    if (pCurr == NULL) return;

    pCurr->mVCItem[mIndexBtnEmit].SetBackgroundColour(mCEmit.GetBodyWXColor());
    pCurr->mVCItem[mIndexBtnDiffuse].SetBackgroundColour(mCDiffuse.GetBodyWXColor());
    pCurr->mVCItem[mIndexBtnSpecular].SetBackgroundColour(mCSpecular.GetBodyWXColor());
}

void CLS::
SetMaterialInfo(){
    
    double time = GetPCTimeCtrl()->GetCurrentNT();
    mCEmit.SetBodyFromData
	(GetPCParamGroup("Emission")->GetStrValues(time, FALSE));
    mCDiffuse.SetBodyFromData
	(GetPCParamGroup("Diffuse")->GetStrValues(time, FALSE));
    mCSpecular.SetBodyFromData
	(GetPCParamGroup("Specular")->GetStrValues(time, FALSE));
    mCSizeFactorPoint.SetBodyFromData
	(GetPCParamGroup("SizeFactor")->mVCParam[0].GetValue(time));
    mCSizeFactorSphere.SetBodyFromData
	(GetPCParamGroup("SizeFactor")->mVCParam[1].GetValue(time));
    mCAlphaFactor.SetBodyFromData
	(GetPCParamGroup("AlphaFactor")->GetStrValues(time, FALSE));
    mCAlphaModBase.SetBodyFromData
	(GetPCParamGroup("Alpha Modifier")->mVCParam[0].GetValue(time));
    mCAlphaModMax.SetBodyFromData
	(GetPCParamGroup("Alpha Modifier")->mVCParam[1].GetValue(time));
}

float CLS::
GetAlpha(double NT){
    // X̗qalpha~B͉

    float alphaFactor = GetPCParamGroup("AlphaFactor")->GetVInterpolatedValue(NT)[0];

    return alphaFactor;
}
