// (setq-c-offset 4)
// "clsMaterialStreamline.h"

#include "clsMaterialStreamline.h"
#include "../clsSceneObjCtrl.h"

#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#include "../../../wxWidgetEnum.h"

#define CLS MaterialStreamline_cls

CLS::
MaterialStreamline_cls(wxString strName, wxString strObjName) : Attribute_cls("MaterialStreamline_cls", strName) {
    
    wxString strUpdate = "VIEW3D";
    wxString strUpdateMat = "VIEW3D::PROPERTY_MATERIAL";
    
    //SyncConfListTag(strObjName);
    SetConfListTagForSceneObjAttribute(strName, strObjName);

    MaterialOption_cls::AddItemToConfList(mCConfList, strUpdate);
    MaterialTexture_cls::AddItemToConfList(mCConfList, strUpdate);

    mCConfList.Add(&mCCustomName, "CustomName=", "MaterialName");
    mCConfList.Add(&mCClrCurveDependType, "ColorCurveDepend=", "0", strUpdate);
    mCConfList.Add(&mCBlendType, "BlendType=", "0", strUpdate);
    mCConfList.Add(&mCMappingType, "MappingType=", "0", strUpdate);

    mCEmit.SetItem("Emit=", "1 1 1");
    mCAlphaFactor.SetItem("AlphaFactor=", "1");

    ParamGroup_cls *pCEmission = GetPCNewParamGroup("Emission", "R", "G", "B");
    InitializeAndPushParam(pCEmission, 1, 1, 1);

    AddOneDimParam("AlphaFactor", "A", 1);
};

CLS::
~MaterialStreamline_cls(){
}

MultiInputPanel_cls *CLS::
GetPCNewPanelMaterial(wxWindow *pCWindow, wxString strName, 
		      int indexMat, bool bEnableAdd, bool bEnableDelete){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, strName);
    mstrMNumName = strName;

    pC->mTextWidth = 90;

    pC->CombineSizer(3);
    pC->CombineConfItem(mCCustomName);
    pC->CombineButton(btnAddMaterial, "+", 30, bEnableAdd);
    int index = pC->CombineBitmapPanel(btnMaterialMaterialDelete, "Delete", 24, bEnableDelete);

    mIndexEmit = pC->CombineSizer(2, _("Emit"));
    pC->SetLimit(0, 1);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mCEmit);
    pC->mContextMenuType = _ITEM_MENU_COLOR;
    mIndexBtnEmit = pC->CombineBitmapPanel(btnMaterialEmit0+indexMat, " ",  30);

    pC->CombineSizer(2, _("Alpha Factor"));
    pC->CombineConfItem(mCAlphaFactor);
    pC->SetChoiceList(2, GetPCStringProvider()->GetAstr("BlendType"));
    pC->CombineConfItem(mCBlendType);

    pC->FinishSetting();

    BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[index].mpCWindowLink;
    pCBitmap->mbButton = TRUE;
    pCBitmap->AddBitmapStr("wxART_CLOSE_PANEL", "wxART_CLOSE_PANEL_P");
    pCBitmap->SetBitmapStrDisable("wxART_CLOSE_PANEL_D");
    pCBitmap->mBorderX = 3;
    pCBitmap->mBorderY = 4;

    BitmapPanel_cls *pCBitmapEmit = (BitmapPanel_cls*)pC->mVCItem[mIndexBtnEmit].mpCWindowLink;
    pCBitmapEmit->mbButton = TRUE;
    
    return pC;
}

void CLS::
SyncConfListTag(wxString strObj){
    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strObj << "::" << GetStrName();
    mCConfList.SetTag(strConfListLabel);
}

void CLS::
SyncMaterialStreamline(wxString strObj){
    SyncronizeParamAndConf3(strObj, "Emission", mCEmit);
    SyncronizeParamAndConf(strObj, "AlphaFactor", mCAlphaFactor);
}

void CLS::
SetGLColor(double NT, bool bUseAlpha){
    std::vector < double > VClr = GetPCParamGroup("Emission")->GetVInterpolatedValue(NT);
    std::vector < double > VAlpha = GetPCParamGroup("AlphaFactor")->GetVInterpolatedValue(NT);
    if (bUseAlpha == FALSE) VAlpha[0] = 1;

    glColor4f(VClr[0], VClr[1], VClr[2], VAlpha[0]);
}

void CLS::
SetBlendFunc(){
    if (mCBlendType.GetBodyINT() == 1)
	GetPCGLExtentions()->
	    BlendFuncSeparate(GL_SRC_ALPHA, GL_ONE,
			      GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    else
	GetPCGLExtentions()->
	    BlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
			      GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
}

/*
void CLS::
SetDepthMask(){
    if (mCbDepthMask.GetBodyBOOL() == TRUE)
	glDepthMask(GL_TRUE);
    else
	glDepthMask(GL_FALSE);
}
*/

void CLS::
UpdateButtonColor(MNumPainBase_cls *pCPain){
    MultiInputPanel_cls* pCurr = pCPain->GetPCMNum(mstrMNumName);
    if (pCurr == NULL) return;

    pCurr->mVCItem[mIndexBtnEmit].mpCWindowLink->SetBackgroundColour(mCEmit.GetBodyWXColor());
}

void CLS::
SetMaterialInfo(){
    
    double time = GetPCTimeCtrl()->GetCurrentNT();
    mCEmit.SetBodyFromData
	(GetPCParamGroup("Emission")->GetStrValues(time, FALSE));
    mCAlphaFactor.SetBodyFromData
	(GetPCParamGroup("AlphaFactor")->GetStrValues(time, FALSE));
}

bool CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (strFirstItem.IsSameAs("ReadMaterial") == FALSE) return FALSE;
    
    ASSERT(strLine.IsSameAs("Begin"));
    std::wstring wstrLine;
    while (std::getline(istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");

	if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE) 
	    continue;

	if (strFirstItem.IsSameAs("ReadParamGroup")){
	    wxString paramGroupName = PopStrItemFromStrLine(strLine, "::");
	    ParamGroup_cls *pCParamGroup = GetPCParamGroup(paramGroupName);
	    if (pCParamGroup != NULL){
	    
		std::wstring wstrLine;
		while (std::getline(istream, wstrLine)){
		    wxString strLine = wstrLine.c_str();
		wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
		pCParamGroup->ReadStrLine(istream, strFirstItem, strLine);
		
		if (strFirstItem.IsSameAs("ReadParamGroup")){
		    ASSERT(PopStrItemFromStrLine(strLine, wxT("::")).IsSameAs(wxT("End")));
		    break;
		}
		}
	    }
	}
	
	if (strFirstItem.IsSameAs("ReadMaterial")){
	    ASSERT(strLine.IsSameAs("End"));
	    break;
	}
    }

    return TRUE;
}

void CLS::
Serialize(wxString& strOut){
    strOut << "ReadMaterial::Begin" << ENDL;

    ConfListContainer_cls::Serialize(strOut);

    int numParamGroup = mVpCParamGroup.size();
    for (int i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->Serialize(strOut);

    strOut << "ReadMaterial::End" << ENDL;
}
