// (setq-c-offset 4)
// "clsParticle.cpp"

#include "clsParticle.h"
#include "clsColorCurve.h"
#include "clsDrawParticle.h"
#include "clsDrawParticleLargeVBO.h"
#include "clsMaterialParticle.h"
#include "clsViewerClone.h"
#include "../clsSceneObjCtrl.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../../GLOBJ/clsVBOCameraBox.h"
#include "../../MISC/clsMisc.h"

#include "../../MODIFIER/clsSequenceRemap.h"
#include "../../MODIFIER/clsParticleTweak.h"
#include "../../PREFERENCE/clsPreference.h"
#include "../../SIMDATA/clsParticleFormat.h"
#include "../../SIMDATA/clsParticleForRender.h"
#include "../../SIMDATA/clsParticleSequence.h"
#include "../../SIMDATA/clsParticleSnap.h"
#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../UNDO/clsUndoManager.h"

#include "../../WIDGETCTRL/clsBitmapPanel.h"
#include "../../WIDGETCTRL/INPUTITEM/clsAtoSTextCtrl.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"

#include "../../../clsFrameMain.h"
#include "../../../clsMain.h"

#include <wx/filename.h>

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#include "../../../wxWidgetEnum.h"

#define CLS ThreadParticlePrepareNext_cls

void CLS::
SetParameters(ParticleForRender_cls *pCNext,
	      ParticleSnap_cls *pCSnap1, ParticleSnap_cls *pCSnap2,
	      ParticleSnap_cls *pCSnap0, ParticleSnap_cls *pCSnap3,
	      int frame1, int subframe, int seq, float interpFactor,
	      ViewerClone_cls *pCParticleClone,
	      ViewerClone_cls *pCParticleCloneForMat,
	      double NTNext, PositionAndDirection_cls CSortPos,
	      bool bStoreValue){
    mpCNext = pCNext;
    mpCSnap1 = pCSnap1; mpCSnap2 = pCSnap2; 
    mpCSnap0 = pCSnap0; mpCSnap3 = pCSnap3;
    mFrame1 = frame1; mSubframe = subframe; mSeq = seq;
    mInterpFactor = interpFactor;
    mpCParticleClone = pCParticleClone;
    mpCParticleCloneForMat = pCParticleCloneForMat;
    mNTNext = NTNext;
    mCSortPos = CSortPos;
    mbStoreValue = bStoreValue;

    if (interpFactor != 0)
	ASSERT(pCSnap2->mbAllocated == TRUE);
}

wxThread::ExitCode CLS::
Entry(){
    mpCNext->PrepareData(mpCSnap1, mpCSnap2, mpCSnap0, mpCSnap3,
			 mFrame1, mSubframe, mSeq, mInterpFactor,
			 mpCParticleClone, mpCParticleCloneForMat,
			 mNTNext, mCSortPos, FALSE, mbStoreValue, TRUE);
    
    *mpbFinishedFlagLink = TRUE;
    return 0;
}

#define CLS Particle_cls

CLS::
Particle_cls(ParentObj_cls *pCParentObj, wxString strName) :
    DataViewerBase_cls(pCParentObj)
{
    mSortOrder = 1;

    mstrClassName = "Particle_cls";
    mbPreparingNextFinished = TRUE;
    mbSkipDrawing = FALSE;
    mbUsePrepareNext = FALSE;
    mbExporting = FALSE;

    mpCParticleSequence     = new ParticleSequence_cls;
    mpCDrawParticle         = new DrawParticle_cls;
    mpCDrawParticleLargeVBO = new DrawParticleLargeVBO_cls;
    mpCParticleFormat       = new ParticleFormat_cls(this, strName);
    mpCCurrent              = new ParticleForRender_cls;
    mpCNext                 = new ParticleForRender_cls;
    mpCExporterSequence     = new FileSequence_cls;

    Initialize(strName);

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";
    wxString strUpdateMat = "VIEW3D::PROPERTY_MATERIAL";
    wxString strUpdateMod = "VIEW3D::PROPERTY_OBJECT::PROPERTY_MODIFIER";
    wxString strUpdateText = "VIEW3D::TEXT_CTRL";
    wxString strUpdateFormat = "PROPERTY_FORMAT";
    wxString strUpdateInfo = "PROPERTY_INFO";

    mCConfList.Add(&mCstrLink, "Link=", "None", strUpdate);
    mCConfList.Add(&mCShaderDissolve, "Dissolve=", "0", strUpdate);
    mCConfList.Add(&mCTwoPass, "TwoPass=", "0", strUpdateMat);
    mCConfList.Add(&mCSortType, "SortType=", "0", strUpdateMod);
    mCConfList.Add(&mCSortFrequency, "SortFrequency=", "0.999", strUpdate);
    mCConfList.Add(&mCstrSortObject, "SortObject=", "None", strUpdate);
    mCConfList.Add(&mCClassifyType, "ClassifyType=", "0", strUpdate);
    mCConfList.Add(&mCstrClassify, "Classify=", "return 0", strUpdateText);

    mCConfList.Add(&mCstrPathExporter, "PathExporter=", "None", strUpdate);
    mCConfList.Add(&mCExportType, "ExportType=", "0");
    mCConfList.Add(&mCExportDir, "ExportDir=", "1");
    
    mCConfList.Add(&mCbDepthMask, "DepthMask=", "1", strUpdate);
    mCConfList.Add(&mCInterpType, "InterpType=", "1", strUpdateFormat);
    mCConfList.Add(&mCInterpolateBy, "InterpolateBy=", "0", strUpdateFormat);
    mCConfList.Add(&mCVanishSizeType, "VanishSizeType=", "2", strUpdateFormat);
    mCConfList.Add(&mCExtrapolateType, "ExtrapolateType=", "0", strUpdateFormat);
    mCConfList.Add(&mCstrTexturePath[0], "TexturePath0=", "None", strUpdateMat);
    mCConfList.Add(&mCstrTexturePath[1], "TexturePath1=", "None", strUpdateMat);
    mCConfList.Add(&mCstrTexturePath[2], "TexturePath2=", "None", strUpdateMat);
    mCConfList.Add(&mCDependDistance, "DependDistance=", "0", strUpdateMat);
    mCConfList.Add(&mCPointSizeFactor, "PointSizeFactor=", "10", strUpdateMat);
    mCConfList.Add(&mCDistanceBase, "DistanceBase=", "10", strUpdateMat);
    mCConfList.Add(&mCSphereDepth, "SphereDepth=", "2", strUpdateMat);
    mCConfList.Add(&mCAlphaCrit, "AlphaCrit=", "0.01", strUpdateMat);
    mCConfList.Add(&mCTextureNum, "TextureNum=", "1", strUpdateMat);
    mCConfList.Add(&mCInfoIndex, "InfoIndex=", "0", strUpdateInfo);
    mCConfList.Add(&mCInfoID, "InfoID=", "0", strUpdateInfo);
    mCConfList.Add(&mCbCastShadow, "CastShadow=", "1", strUpdate);
    mCConfList.Add(&mCbTwistRandom, "TwistRandom=", "0", strUpdate);
    
    wxString strAttribName;
    
    strAttribName = _("Material");
    strAttribName << ".0";
    MaterialParticle_cls *pCMaterial = new MaterialParticle_cls(strAttribName, GetStrName());
    mVpCAttribute.push_back(pCMaterial);
    
    strAttribName = _("ColorCurve");
    strAttribName << ".0";
    ColorCurve_cls *pCColorCurve = new ColorCurve_cls(strAttribName, GetPCStringProvider()->GetArrayString("RGBA"));
    
    double clrInit[2][5] = {{0, 0, 1, 0, 0},
			    {1, 1, 0, 0, 1}};
    pCColorCurve->SetInitKeyframes(2, clrInit);
    
    mVpCAttribute.push_back(pCColorCurve);
}

CLS::
~Particle_cls(){
    delete mpCParticleSequence;
    delete mpCDrawParticle;
    delete mpCDrawParticleLargeVBO;
    delete mpCParticleFormat;
    delete mpCCurrent;
    delete mpCNext;
    delete mpCExporterSequence;
}

void CLS::
SyncConfListTagParticle(){
    wxString strObj = mCstrName.GetBodySTR();
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }

    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }
    
    mpCParticleFormat->SyncConfListTag(strObj);
}

/*
void CLS::
Clone(ViewerClone_cls &CClone){
    CClone.CloneConfListFrom(*this);

    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	AttributeClone_cls *pCAttribClone = new AttributeClone_cls;
	Attribute_cls* pCTgt = GetPCAttributeFromIndex(i);
	if (pCTgt->GetStrAttributeName().IsSameAs("MaterialParticle_cls"))
	    ((MaterialParticle_cls*)pCTgt)->SetMaterialInfo();
	pCTgt->Clone(*pCAttribClone);
	CClone.mVpCAttribClone.push_back(pCAttribClone);
    }
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	AttributeClone_cls *pCAttribClone = new AttributeClone_cls;
	Attribute_cls* pCTgt = GetPCDataProcFromIndex(i);
	pCTgt->Clone(*pCAttribClone);
	CClone.mVpCAttribClone.push_back(pCAttribClone);
    }

    double matrix16[16];
    GetMatrix16(matrix16, NT);
    memcpy(CClone.mMatrix16, matrix16, sizeof(double)*16);
}
*/

int CLS::
GetCurrentFrame(){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frame, subframe);

    int seqLen = mpCParticleSequence->mSeqLen;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");
    if (pCSeqRemap != NULL)
	frame = (int)pCSeqRemap->GetRemappedFrame(NT, frame, seqLen);

    return frame;
}

ParticleSnap_cls* CLS::
GetPCCurrentSnap(){
    return mpCParticleSequence->GetPCSnap(GetCurrentFrame());
}

wxString CLS::
GetStrInfo(){
    
    wxString strInfo;
    strInfo << mpCParticleSequence->GetStrInfo(FALSE);

    //int frame = GetCurrentFrame();
    ParticleSnap_cls *pCSnap = GetPCCurrentSnap();
    if (pCSnap != NULL)
	strInfo << pCSnap->GetStrInfo(mCInfoIndex.GetBodyINT());
    if (mpCCurrent != NULL)
	strInfo << mpCCurrent->GetStrInfo(mCInfoIndex.GetBodyINT());

    return strInfo;
}

wxString CLS::
GetStrPathCandid(){
    return mpCParticleSequence->GetFormat(); }

long long CLS::
SearchID(){
    if (mpCCurrent == NULL) return -1;
    return mpCCurrent->SearchID(mCInfoID.GetBodyINT());
}

bool CLS::
IsDataValid(){
    return mpCParticleSequence->IsDataValid(); }

bool CLS::
IsExportValid(){
    wxFileName CFile = mCstrPathExporter.GetBodySTR();
    wxString strDir = CFile.GetPath();
    return wxDirExists(strDir); // !! tH_݂邩
}

bool CLS::
DoesUseIDLongLong(){
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->IsAttributeValidAndSameAs("ParticleTweak_cls"))
	    if (pCAttrib->GetPCItem("StoreIDLongLong=")->GetBodyBOOL() == TRUE) return TRUE;
	
    }
    /* // ɂɂ̓^[Qbgł`FbNȂƂȂ
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	if (pCAttrib->IsAttributeValidAndSameAs("ParticleExclude_cls")){
	    SceneObj_cls *pCTgt = pCAttrib->GetPCTargetSceneObj();
	    if (pCTgt != NULL && pCTgt->IsSameAs("Particle_cls"))
		return TRUE;
	}
    }
    */
    return FALSE;
}

bool CLS::
DoesStoreValue(){
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->IsAttributeValidAndSameAs("ParticleTweak_cls"))
	    if (pCAttrib->GetPCItem("StoreValue=")->GetBodyBOOL() == TRUE) return TRUE;
	
    }
    return FALSE;
}

int CLS::
GetNumMaterialLink(){
    SceneObj_cls* pCObjLink = GetPCSceneObjCtrl()->GetPCFromName(mCstrLink.GetBodySTR());
    if (pCObjLink != NULL && pCObjLink->IsSameAs("Particle_cls")){
	Particle_cls* pCParticleLink = (Particle_cls*)pCObjLink;
	return pCParticleLink->GetNumAttribute("MaterialParticle_cls");
    }
    else
	return GetNumAttribute("MaterialParticle_cls");
}

MaterialParticle_cls* CLS::
GetPCMaterialLink(int i){
    SceneObj_cls* pCObjLink = GetPCSceneObjCtrl()->GetPCFromName(mCstrLink.GetBodySTR());
    if (pCObjLink != NULL && pCObjLink->IsSameAs("Particle_cls")){
	Particle_cls* pCParticleLink = (Particle_cls*)pCObjLink;
	return (MaterialParticle_cls*)(pCParticleLink->GetPCAttributeFromAttribName("MaterialParticle_cls", i));
    }
    else
	return (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", i);
}

bool CLS::
IsPointAll(){
    bool bAllPoint = TRUE;
    int num = GetNumMaterialLink();
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = GetPCMaterialLink(i);
	if (pCMat->mCBaseType.GetBodyINT() != _PARTICLE_POINT) bAllPoint = FALSE;
    }
    /*
    int num = GetNumAttribute("MaterialParticle_cls");
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", i);
	if (pCMat->mCBaseType.GetBodyINT() != _PARTICLE_POINT) bAllPoint = FALSE;
    }
    */
    return bAllPoint;
}

bool CLS::
IsBillboardAll(){
    bool bAllBillboard = TRUE;
    int num = GetNumMaterialLink();
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = GetPCMaterialLink(i);
	if (pCMat->mCBaseType.GetBodyINT() != _PARTICLE_BILLBOARD) bAllBillboard = FALSE;
    }
    /*
    int num = GetNumAttribute("MaterialParticle_cls");
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", i);
	if (pCMat->mCBaseType.GetBodyINT() != _PARTICLE_BILLBOARD) bAllBillboard = FALSE;
    }
    */
    return bAllBillboard;
}

bool CLS::
IsSphereAll(){
    bool bAllSphere = TRUE;
    int num = GetNumMaterialLink();
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = GetPCMaterialLink(i);
	if (pCMat->mCBaseType.GetBodyINT() != _PARTICLE_SPHERE) bAllSphere = FALSE;
    }
    /*
    int num = GetNumAttribute("MaterialParticle_cls");
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", i);
	if (pCMat->mCBaseType.GetBodyINT() != _PARTICLE_SPHERE) bAllSphere = FALSE;
    }
    */
    return bAllSphere;
}

bool CLS::
DoesMatereialBaseExist(int typeTgt){
    bool bTgt = FALSE;
    int num = GetNumAttribute("MaterialParticle_cls");
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", i);
	if (pCMat->mCBaseType.GetBodyINT() == typeTgt) bTgt = TRUE;
    }
    return bTgt;
}

void CLS::
InitializeStatusForDrawingExe(){
    mbSkipDrawing = FALSE;
    mbUsePrepareNext = FALSE; // !! oOȂTRUEɂ !!
    mSortOrder = 1;
}

void CLS::
CheckStatusForDrawingExe(){
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	if (pCAttrib->IsAttributeValidAndSameAs("ParticleMerge_cls")){
	    SceneObj_cls *pCTgt = pCAttrib->GetPCTargetSceneObj();
	    if (pCTgt != NULL && pCTgt->IsSameAs("Particle_cls")){
		Particle_cls *pCParticle = ((Particle_cls*)pCTgt);
		pCParticle->mbSkipDrawing = TRUE;
		pCParticle->mSortOrder = 2;
		mbUsePrepareNext = FALSE;
	    }
	}
	    
	if (pCAttrib->IsAttributeValidAndSameAs("ParticleExclude_cls")){
	    SceneObj_cls *pCTgt = pCAttrib->GetPCTargetSceneObj();
	    if (pCTgt != NULL && pCTgt->IsSameAs("Particle_cls")){
		Particle_cls *pCParticle = ((Particle_cls*)pCTgt);
		pCParticle->mSortOrder = 2;
		mbUsePrepareNext = FALSE;
	    }
	}
    }
}

void CLS::
CheckFileSequence(){
    int offset = mCSequenceOffset.GetBodyINT();
    mpCParticleSequence->Set(mCstrPath.GetBodySTR(), offset);
    mpCParticleSequence->CalcMaxOffset(mCbSequence.GetBodyBOOL());
    if (mpCParticleSequence->mSeqLen != 1)
	mpCParticleSequence->mbTimeDepend = TRUE;

    mpCCurrent->Clear();
    mstrPathChecked = mCstrPath.GetBodySTR();
    if (mpCParticleSequence->IsDataValid() == TRUE)
	GetPCMain()->mCLastOpenedFilePARTICLE.SetBodyFromData(mstrPathChecked);
}

bool CLS::
CheckFormat(int dataType){ return mpCParticleFormat->CheckFormat(dataType); }

void CLS::
ExportSnap(ParticleSnap_cls *pCSnap){
    if (pCSnap == NULL) return;
    
    if (IsExportValid() == FALSE){
	wxMessageBox(_("Filepath is invalid"), _("Warning"), wxICON_EXCLAMATION);
	mbExporting = FALSE; 
	GetPCTimeCtrl()->StopPlayCommand();
	return;
    }

    wxString strFile = mpCExporterSequence->GetModifiedStr(GetPCTimeCtrl()->GetCurrentStep());
    if (wxFileExists(strFile) == TRUE){
	mbExporting = FALSE; 
	wxMessageBox(_("A file with same name exists.\nOverwrite is forbidden for safety."), _("Warning"), wxICON_EXCLAMATION);
	GetPCTimeCtrl()->StopPlayCommand();
	return;
    }
 
    if (mCExportType.GetBodyINT() == 0)
	pCSnap->ExportSnap(strFile, mCExportDir.GetBodyINT());
    else
	mpCCurrent->ExportSnap(strFile, mCExportDir.GetBodyINT());
}

void CLS::
GLDrawBox(int status){
    glDisable(GL_LIGHTING);

    glColor4f(0.2f, 0.7f, 0.2f, 1.0f);
    if (status == 1) ColorProvider_cls::SetGLColor("Selected");
    if (status == 2) ColorProvider_cls::SetGLColor("Active");

    GCClassProvider.GetPCVBOCameraBox()->Draw();
}

void CLS::
PrepareFilename(ParticleSnap_cls *pCSnap, int frame){
    wxString strFile;
    if (mpCParticleSequence->mbTimeDepend == TRUE)
	strFile = mpCParticleSequence->GetModifiedStr(frame);
    else
	strFile = mpCParticleSequence->mstrPath;
    pCSnap->SetStrFile(strFile);
    pCSnap->SetFrame(frame);
}

bool CLS::
PrepareDataNext(ViewerClone_cls *pCParticleClone,
		ViewerClone_cls *pCParticleCloneForMat,
		int composit, double NTNext, PositionAndDirection_cls CSortPos,
		bool bStoreValue){

    int frame0, frame1, frame2, frame3, subframe;
    double interpFactor = GetFrameInfo(frame0, frame1, frame2, frame3, subframe, NTNext);

    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);
    
    GetPCTimeCtrl()->AdjustFrameForComposer
	(&frame1, &frame2, &interpFactor, composit);

    bool bTimeDepend = mpCParticleSequence->mbTimeDepend;
    ParticleSnap_cls *pCSnap1 = mpCParticleSequence->GenerateIfNot(frame1);
    //ParticleSnap_cls *pCSnap1 = mpCParticleSequence->GetPCSnap(frame1);
    if (pCSnap1 == NULL) return FALSE;
    
    // Prefetchłɂŏ߂č쐬ꍇ̂
    // ̏ꍇ͑ޔ
    if (pCSnap1->mValidData == FALSE ||
	pCSnap1->mbAllocated == FALSE ||
	pCSnap1->mbReadFinished == FALSE) return FALSE;
    
    ParticleSnap_cls *pCSnap2 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0){
	pCSnap2 = mpCParticleSequence->GenerateIfNot(frame2);
	//pCSnap2 = mpCParticleSequence->GetPCSnap(frame2);
	if (pCSnap2 == NULL) return FALSE;

	// Prefetchłɂŏ߂č쐬ꍇ̂
	// ̏ꍇ͑ޔ
	if (pCSnap2->mValidData == FALSE ||
	    pCSnap2->mbAllocated == FALSE ||
	    pCSnap2->mbReadFinished == FALSE) return FALSE;
    }

    ParticleSnap_cls *pCSnap0 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0){
	pCSnap0 = mpCParticleSequence->GenerateIfNot(frame0);
	//pCSnap0 = mpCParticleSequence->GetPCSnap(frame0);
	if (pCSnap0 == NULL) return FALSE;

	// Prefetchłɂŏ߂č쐬ꍇ̂
	// ̏ꍇ͑ޔ
	if (pCSnap0->mValidData == FALSE ||
	    pCSnap0->mbAllocated == FALSE ||
	    pCSnap0->mbReadFinished == FALSE) return FALSE;
    }

    bool bNeedSnap3 = FALSE;
    if (mCInterpType.GetBodyINT() == 3 || 
	mCInterpType.GetBodyINT() == 4) bNeedSnap3 = TRUE; //Bezier
    ParticleSnap_cls *pCSnap3 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0 && bNeedSnap3){
	pCSnap3 = mpCParticleSequence->GenerateIfNot(frame3);
	//pCSnap0 = mpCParticleSequence->GetPCSnap(frame3);
	if (pCSnap3 == NULL) return FALSE;
				 
	// Prefetchłɂŏ߂č쐬ꍇ̂
	// ̏ꍇ͑ޔ
	if (pCSnap3->mValidData == FALSE ||
	    pCSnap3->mbAllocated == FALSE ||
	    pCSnap3->mbReadFinished == FALSE) return FALSE;
    }
    
    mbPreparingNextFinished = FALSE;
    ThreadParticlePrepareNext_cls *pCThreadPrepareNext = 
	new ThreadParticlePrepareNext_cls;
    pCThreadPrepareNext->SetParameters
	(mpCNext, pCSnap1, pCSnap2, pCSnap0, pCSnap3,
	 frame1, subframe, seq, interpFactor, 
	 pCParticleClone, pCParticleClone,
	 NTNext, CSortPos, bStoreValue);
    pCThreadPrepareNext->CreateThread(&mbPreparingNextFinished, 0, 0);
    pCThreadPrepareNext->Run();

    //WaitThreadsNOTDeleteFlag(&mbPreparingNextFinished, 1, 3);
    /*
    mpCNext->PrepareData(pCSnap1, pCSnap2,
			 frame1, subframe, seq, interpFactor,
			 pCParticleClone,
			 NTNext, CSortPos);
    */

    return TRUE;
}

void CLS::
GLDrawPointAll(int packVBO, int numParticle, bool bFullUpdateVBO,
	       int lightNum, int ensemble){
    bool bTwoPass = mCTwoPass.GetBodyBOOL();
    int dissolve = mCShaderDissolve.GetBodyINT();

    if (bFullUpdateVBO == TRUE){

	mpCDrawParticleLargeVBO->Allocate(packVBO, bTwoPass);

	for (int offset = 0; offset < numParticle; offset += packVBO){
	    GetPCFrameMain()->CheckAbortion();
	    if (GetPCTimeCtrl()->msbForceAbort == TRUE) continue;
	    
	    mpCDrawParticleLargeVBO->PrepareNextDataPos(mpCCurrent, offset);
	    mpCDrawParticleLargeVBO->DrawPoints(mpCCurrent, lightNum, ensemble, dissolve);
	    mpCDrawParticleLargeVBO->FinishPrepareNextPos();
	    
	    if (mpCDrawParticleLargeVBO->mNumNext == 0){
		mpCDrawParticleLargeVBO->mNumValidNext[0] = 0;
		mpCDrawParticleLargeVBO->mNumValidNext[1] = 0;
	    }
	    else{
		mpCDrawParticleLargeVBO->PrepareNextDataAlpha(mpCCurrent, offset);
		mpCDrawParticleLargeVBO->SetNextVBOPos();
		mpCDrawParticleLargeVBO->FinishPrepareNextAlpha();
		mpCDrawParticleLargeVBO->PrepareNextDataST(mpCCurrent, offset);
		mpCDrawParticleLargeVBO->PrepareNextDataElement(bTwoPass);// AlphaΓɍs
		mpCDrawParticleLargeVBO->SetNextVBOAlpha();
		mpCDrawParticleLargeVBO->FinishPrepareNextST();
		mpCDrawParticleLargeVBO->PrepareNextDataOther(mpCCurrent, offset);
		mpCDrawParticleLargeVBO->SetNextVBOST();
		mpCDrawParticleLargeVBO->FinishPrepareNextOther();
		mpCDrawParticleLargeVBO->SetNextVBOOther(bTwoPass);
		mpCDrawParticleLargeVBO->FinishPrepareNextElement();
	    }
	    mpCDrawParticleLargeVBO->SetNextVBOElement(bTwoPass);
	    
	}
	mpCDrawParticleLargeVBO->DrawPoints(mpCCurrent, lightNum, ensemble, dissolve);
	mpCDrawParticleLargeVBO->ClearVBO();
	mpCDrawParticleLargeVBO->ReleaseMemoryData();
    }
    else{
	if (ensemble == 0 && mProcessedInInstanceLoop == 0)
	    mpCDrawParticle->SetVBOBuffer(mpCCurrent, bTwoPass);
	
	mpCDrawParticle->DrawPoints(mpCCurrent, lightNum, ensemble, dissolve);
    }
    
}

void CLS::
GLDrawBillboardAll(int packVBO, int numParticle, bool bFullUpdateVBO,
		   int lightNum, int ensemble,
		   float baseMat16[6], float baseMat16Inv[16], 
		   float viewMat16[16], float viewMat16Inv[16]){
    bool bTwoPass = mCTwoPass.GetBodyBOOL();
    int dissolve = mCShaderDissolve.GetBodyINT();
    int twist = mCbTwistRandom.GetBodyBOOL();
    if (bFullUpdateVBO == TRUE){
	mpCDrawParticleLargeVBO->Allocate(packVBO, bTwoPass);
	
	int numValidLog = 0;
	for (int offset = 0; offset < numParticle; offset += packVBO){
	    GetPCFrameMain(0)->CheckAbortion();
	    if (GetPCTimeCtrl()->msbForceAbort == TRUE) continue;

	    mpCDrawParticleLargeVBO->PrepareNextDataPos(mpCCurrent, offset);
	    mpCDrawParticleLargeVBO->DrawBillboards
		(mpCCurrent, lightNum, ensemble, dissolve, twist,
		 baseMat16, baseMat16Inv,
		 viewMat16, viewMat16Inv);
	    mpCDrawParticleLargeVBO->FinishPrepareNextPos();

	    if (mpCDrawParticleLargeVBO->mNumNext == 0){
		mpCDrawParticleLargeVBO->mNumValidNext[0] = 0;
		mpCDrawParticleLargeVBO->mNumValidNext[1] = 0;
	    }
	    else{
		mpCDrawParticleLargeVBO->PrepareNextDataAlpha(mpCCurrent, offset);
		mpCDrawParticleLargeVBO->SetNextVBOPos();
		mpCDrawParticleLargeVBO->FinishPrepareNextAlpha();
		
		mpCDrawParticleLargeVBO->PrepareNextDataST(mpCCurrent, offset);
		mpCDrawParticleLargeVBO->PrepareNextDataElement(bTwoPass); // AlphaΓɍs
		mpCDrawParticleLargeVBO->SetNextVBOAlpha();
		mpCDrawParticleLargeVBO->FinishPrepareNextST();
		
		mpCDrawParticleLargeVBO->PrepareNextDataOther(mpCCurrent, offset);
		mpCDrawParticleLargeVBO->SetNextVBOST();
		mpCDrawParticleLargeVBO->FinishPrepareNextOther();
		
		mpCDrawParticleLargeVBO->SetNextVBOOther(bTwoPass);
		mpCDrawParticleLargeVBO->FinishPrepareNextElement();
	    }
	    mpCDrawParticleLargeVBO->SetNextVBOElement(bTwoPass);
	    numValidLog += mpCDrawParticleLargeVBO->mNumValid[0];

	}
	mpCDrawParticleLargeVBO->DrawBillboards
	    (mpCCurrent, lightNum, ensemble, dissolve, twist, 
	     baseMat16, baseMat16Inv,
	     viewMat16, viewMat16Inv);
	mpCDrawParticleLargeVBO->ReleaseMemoryData();
	mpCDrawParticleLargeVBO->ClearVBO();

	wxString strLog;
	strLog << "mNumValidSum = " << numValidLog << ENDL;
	wxLogMessage(strLog);
    }
    else{
	if (ensemble == 0 && mProcessedInInstanceLoop == 0)
	    mpCDrawParticle->SetVBOBuffer(mpCCurrent, bTwoPass);
	
	mpCDrawParticle->DrawBillboards
	    (mpCCurrent, lightNum, ensemble, dissolve, twist,
	     baseMat16, baseMat16Inv,
	     viewMat16, viewMat16Inv);
    }

}

double CLS::
GetFrameInfo(int &frame0, int &frame1, int &frame2, int &frame3, int &subframe,
	     double NT){
    
    int seqLen = mpCParticleSequence->mSeqLen;

    //double NT = GetPCTimeCtrl()->GetCurrentNT();
    //int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NT, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);

    bool bRepeat = FALSE;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    if (pCSeqRemap != NULL){
	frame1 = pCSeqRemap->GetRemappedFrame(NT, frame1, seqLen);
	interpFactor = pCSeqRemap->GetInterpFactor(NT, frame1, interpFactor);
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    if (interpFactor < 1e-8) interpFactor = 0;

    frame0 = frame1-1;
    frame2 = frame1+1;
    frame3 = frame1+2;
    ModFrame(frame0, seqLen, bRepeat);
    ModFrame(frame1, seqLen, bRepeat);
    ModFrame(frame2, seqLen, bRepeat);
    ModFrame(frame3, seqLen, bRepeat);

    return interpFactor;
}

std::vector < Particle_cls* > CLS::
GetVpCDependency(){
    std::vector < Particle_cls* > VpCDepend;
    
    uint numAttribute = mVpCDataProc.size();
    for (uint i = 0; i < numAttribute; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	if (pCAttrib->IsAttributeValidAndSameAs("ParticleMerge_cls") ||
	    pCAttrib->IsAttributeValidAndSameAs("ParticleExclude_cls")) {
	    SceneObj_cls *pCTgt = pCAttrib->GetPCTargetSceneObj();
	    if (pCTgt != NULL && pCTgt->IsSameAs("Particle_cls") &&
		pCTgt->mCbVisible.GetBodyBOOL() == TRUE)
		VpCDepend.push_back((Particle_cls*)pCTgt);
	}
    }

    return VpCDepend;
}

void CLS::
PrepareDraw(int ensemble, int composit, bool bPrepareNext){
    if (ensemble == 0 && mProcessedInInstanceLoop == 0)
	if (mstrPathChecked != mCstrPath.GetBodySTR())
	    CheckFileSequence();

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame0, frame1, frame2, frame3, subframe;
    double interpFactor = GetFrameInfo(frame0, frame1, frame2, frame3, subframe, NT);
        
    int seqLen = mpCParticleSequence->mSeqLen;
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);

    bool bForceInterpolate = TRUE;
    if (bForceInterpolate == TRUE)
	if (interpFactor == 0) interpFactor = 1e-7;
  
        
    GetPCTimeCtrl()->StoreInterpFactorForComposit(interpFactor);
    GetPCTimeCtrl()->AdjustFrameForComposer
	(&frame1, &frame2, &interpFactor, composit);
    
    if (mpCParticleSequence->mbTimeDepend == TRUE)
	if (frame1 < 0 || frame1 >= seqLen) return;

    bool bTimeDepend = mpCParticleSequence->mbTimeDepend;
    ParticleSnap_cls *pCSnap1 = mpCParticleSequence->GenerateIfNot(frame1);
    if (pCSnap1 == NULL) return;
    
    ParticleSnap_cls *pCSnap2 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0){
	pCSnap2 = mpCParticleSequence->GenerateIfNot(frame2);
	if (pCSnap2 == NULL) return;
    }

    ParticleSnap_cls *pCSnap0 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0){
	pCSnap0 = mpCParticleSequence->GenerateIfNot(frame0);
	if (pCSnap0 == NULL) return;
    }

    bool bNeedSnap3 = FALSE;
    if (mCInterpType.GetBodyINT() == 3||
	mCInterpType.GetBodyINT() == 4) bNeedSnap3 = TRUE; //Bezier
    ParticleSnap_cls *pCSnap3 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0 && bNeedSnap3){
	pCSnap3 = mpCParticleSequence->GenerateIfNot(frame3);
	if (pCSnap3 == NULL) return;
    }

    bool bUseIDLongLong = DoesUseIDLongLong() & (pCSnap1->mpIDLongLong != NULL);
    bool bStoreValue = DoesStoreValue();
    
    bool bInterpByID = (mCInterpolateBy.GetBodyINT() == _P_INTERP_BY_ID);
    if (ensemble == 0 && mProcessedInInstanceLoop == 0){
	PrepareFilename(pCSnap1, frame1);
	
	ParticleFormatClone_cls *pCFormatClone1 = new ParticleFormatClone_cls;
	pCFormatClone1->CloneFromFormat(*mpCParticleFormat, bInterpByID);
	
	pCSnap1->ReadHeader(pCFormatClone1);
	if (pCSnap1->ReadData(pCFormatClone1) == TRUE)
	    GetPCMain()->UpdatePanels("PROPERTY_INFO");
	else
	    delete pCFormatClone1;
	
	if (pCSnap2 != NULL){
	    ParticleFormatClone_cls *pCFormatClone2 = new ParticleFormatClone_cls;
	    pCFormatClone2->CloneFromFormat(*mpCParticleFormat, bInterpByID);

	    if (pCSnap2->mbReadStarted == TRUE && pCSnap2->mbReadFinished == FALSE){
		while(pCSnap2->mbReadFinished == FALSE) wxMilliSleep(3);
	    }
	    
	    PrepareFilename(pCSnap2, frame2);
	    pCSnap2->ReadHeader(pCFormatClone2);
	    if (pCSnap2->ReadData(pCFormatClone2) == FALSE)
		delete pCFormatClone2;
	}

	if (pCSnap0 != NULL){
	    ParticleFormatClone_cls *pCFormatClone0 = new ParticleFormatClone_cls;
	    pCFormatClone0->CloneFromFormat(*mpCParticleFormat, bInterpByID);

	    if (pCSnap0->mbReadStarted == TRUE && pCSnap0->mbReadFinished == FALSE){
		while(pCSnap0->mbReadFinished == FALSE) wxMilliSleep(3);
	    }
	    
	    PrepareFilename(pCSnap0, frame0);
	    pCSnap0->ReadHeader(pCFormatClone0);
	    if (pCSnap0->ReadData(pCFormatClone0) == FALSE)
		delete pCFormatClone0;
	}

	if (pCSnap3 != NULL){
	    ParticleFormatClone_cls *pCFormatClone3 = new ParticleFormatClone_cls;
	    pCFormatClone3->CloneFromFormat(*mpCParticleFormat, bInterpByID);

	    if (pCSnap3->mbReadStarted == TRUE && pCSnap3->mbReadFinished == FALSE){
		while(pCSnap3->mbReadFinished == FALSE) wxMilliSleep(3);
	    }
	    
	    PrepareFilename(pCSnap3, frame3);
	    pCSnap3->ReadHeader(pCFormatClone3);
	    if (pCSnap3->ReadData(pCFormatClone3) == FALSE)
		delete pCFormatClone3;
	}
	
	PrefetchData();
    }

    //if (mbExporting == TRUE) ExportSnap(pCSnap1);
    
    // CCamPos ƂAǂgׂ
    double modelview[16];
    double invert[16];
    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
    GetInvMatrix4x4(modelview, invert);
    PositionAndDirection_cls CCamDummy(invert, TRUE);
    PositionAndDirection_cls CCamDummyNext(invert, TRUE);

    CCamDummy = mCCamPos; //
    CCamDummyNext = mCCamPosNext;

    if (mCSortType.GetBodyINT() == _SORT_OBJECT){
	SceneObj_cls* pCObj = GetPCSceneObjCtrl()->GetPCFromName(mCstrSortObject.GetBodySTR());
	if (pCObj != NULL){
	    double matrix16[16];
	    double NT = GetPCTimeCtrl()->GetCurrentNT();
	    pCObj->GetMatrix16(matrix16, NT);
	    CCamDummy = PositionAndDirection_cls(matrix16, FALSE);

	    double NTNext = GetPCTimeCtrl()->GetNextNT();
	    pCObj->GetMatrix16(matrix16, NTNext);
	    CCamDummyNext = PositionAndDirection_cls(matrix16, FALSE);
	}
    }
		
    // f[^̍\ensemble = 0, ŏ̃CX^X ̂Ƃł悢B
    if (ensemble == 0 && mProcessedInInstanceLoop == 0){
	
	if (GetPCTimeCtrl()->mPlayDirection == 1){
	    if (mpCNext->mFrame == frame1 &&
		(float)mpCNext->mFactor == (float)interpFactor){
		ParticleForRender_cls *pCTmp = mpCCurrent;
		mpCCurrent = mpCNext;
		mpCNext = pCTmp;
		wxString strLog;
		strLog << "Swap Render Snap" << ENDL;
		strLog << "mpCCurrent at " << mpCCurrent->mFrame << ":" << mpCCurrent->mFactor << ENDL;
		strLog << "mpCNext at " << mpCNext->mFrame << ":" << mpCNext->mFactor << ENDL;
		//wxLogMessage (strLog);
		
	    }
	    else{
		wxString strLog;
		strLog << "Not Swap" << ENDL;
		strLog << "Now is " << frame1 << ":" << interpFactor << ENDL;
		strLog << "mpCNext at " << mpCNext->mFrame << ":" << mpCNext->mFactor << ENDL;
		//wxLogMessage (strLog);
	    }
	}
	
	// NextStepPrepareData ͗ōƂ̂
	// Kvȃf[^̃N[nׂA
	ViewerClone_cls *pCClone = new ViewerClone_cls;
	Clone(*pCClone, NT);
	ViewerClone_cls *pCCloneForMat = pCClone;
	SceneObj_cls *pCObjLinked = GetPCSceneObjCtrl()->GetPCFromName(mCstrLink.GetBodySTR());
	if (pCObjLinked != NULL && pCObjLinked->IsSameAs("Particle_cls")){
	    Particle_cls *pCParticleLinked = (Particle_cls*)pCObjLinked;
	    pCCloneForMat = new ViewerClone_cls;
	    pCParticleLinked->Clone(*pCCloneForMat, NT);
	}

	if (bTimeDepend == TRUE && interpFactor != 0){
	    mpCCurrent->PrepareData(pCSnap1, pCSnap2, pCSnap0, pCSnap3,
				    frame1, subframe, seq, interpFactor,
				    pCClone, pCCloneForMat, NT, CCamDummy,
				    bUseIDLongLong, bStoreValue);
	}
	else
	    mpCCurrent->PrepareData(pCSnap1, NULL, NULL, NULL,
				    frame1, subframe, seq, interpFactor, 
				    pCClone, pCCloneForMat, NT, CCamDummy,
				    bUseIDLongLong, bStoreValue);
	
	mpCCurrent->SetColorCurveTexture();
	
	// pCClone PrepareDataŔj
	bPrepareNext = bPrepareNext && mbUsePrepareNext;
	if (GetPCTimeCtrl()->mPlayDirection == 1 && bPrepareNext == TRUE){
	    double NTNext = GetPCTimeCtrl()->GetNextNT();
	    ViewerClone_cls *pCCloneNext = new ViewerClone_cls;
	    Clone(*pCCloneNext, NTNext);
	    ViewerClone_cls *pCCloneNextForMat = pCCloneNext;
	    if (pCObjLinked != NULL && pCObjLinked->IsSameAs("Particle_cls")){
		Particle_cls *pCParticleLinked = (Particle_cls*)pCObjLinked;
		pCCloneNextForMat = new ViewerClone_cls;
		pCParticleLinked->Clone(*pCCloneNextForMat, NT);
	    }
	    
	    if (PrepareDataNext(pCCloneNext, pCCloneNextForMat, 
				composit, NTNext, CCamDummyNext, 
				bStoreValue) == FALSE)
		delete pCCloneNext;
	}
    }
}

void CLS::
GLDraw(bool bDrawData, bool bDrawGuideObject, 
       int lightNum, int ensemble, int composit, int status,  
       float baseMat16[16], float baseMat16Inv[16], 
       float viewMat16[16], float viewMat16Inv[16],
       bool bHittest){

    /*
    wxString strLog;
    strLog << baseMat16[0] << ":" << baseMat16[1] << ":" << baseMat16[2] << ":" << baseMat16[3] << ENDL;
    strLog << baseMat16[4] << ":" << baseMat16[5] << ":" << baseMat16[6] << ":" << baseMat16[7] << ENDL;
    strLog << baseMat16[8] << ":" << baseMat16[9] << ":" << baseMat16[10] << ":" << baseMat16[11] << ENDL;
    strLog << baseMat16[12] << ":" << baseMat16[13] << ":" << baseMat16[14] << ":" << baseMat16[15] << ENDL;
    */

    /*
    strLog << CCamTmp.mPos[0] << ":" << CCamTmp.mPos[1] << ":" << CCamTmp.mPos[2] << ENDL;
    strLog << baseMat16[0] << ":" << baseMat16[1] << ":" << baseMat16[2] << ":" << baseMat16[3] << ENDL;
    strLog << baseMat16[4] << ":" << baseMat16[5] << ":" << baseMat16[6] << ":" << baseMat16[7] << ENDL;
    strLog << baseMat16[8] << ":" << baseMat16[9] << ":" << baseMat16[10] << ":" << baseMat16[11] << ENDL;
    strLog << baseMat16[12] << ":" << baseMat16[13] << ":" << baseMat16[14] << ":" << baseMat16[15] << ENDL;
    wxLogMessage(strLog);
    */

    int frame = GetCurrentFrame();

    if (bDrawGuideObject == TRUE) GLDrawBox(status);
    if (bDrawData == FALSE) return;
    if (mbExporting == TRUE){
	ExportSnap(GetPCCurrentSnap());
	GetPCTimeCtrl()->mbExportConfirmed = TRUE;
	return;
    }
    
    float billboardFocusYaw = 0;
    float billboardFocusPitch = -60;
    float billboardFocusRow = 0;
    int rotOrder = 0;
    ParticleTweak_cls* pCTweak = (ParticleTweak_cls*)GetPCAttributeFromAttribName("ParticleTweak_cls", 0);
    if (pCTweak != NULL)
	pCTweak->GetBillboartFocus(billboardFocusYaw, billboardFocusPitch, billboardFocusRow, rotOrder);

    float baseMat16tmp[16], baseMat16Invtmp[16];
    for (int i = 0; i < 16; i++) baseMat16tmp[i] = baseMat16[i];
    PositionAndDirection_cls CCamTmp;
    CCamTmp.FromModelView(baseMat16tmp);
    switch(rotOrder){
    case 0:
	CCamTmp.RotateYaw(billboardFocusYaw);
	CCamTmp.RotatePitch(billboardFocusPitch);
	CCamTmp.RotateRoll(billboardFocusRow);
	break;
    case 1:
	CCamTmp.RotateRoll(billboardFocusRow);
	CCamTmp.RotatePitch(billboardFocusPitch);
	CCamTmp.RotateYaw(billboardFocusYaw);
	break;
    }
    CCamTmp.ToModelView(baseMat16tmp);
    TensorMath_cls::GetInvMatrix4x4(baseMat16tmp, baseMat16Invtmp);
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    bool bDepthMask = mCbDepthMask.GetBodyBOOL();
    if (bDepthMask == FALSE)
	glDepthMask(GL_FALSE);
    
    int dissolve = mCShaderDissolve.GetBodyINT();

    int renderMode;
    glGetIntegerv(GL_RENDER_MODE, &renderMode);

    // xɕ`s\ȏꍇ́A͂VBO𖈉쐬邵Ȃ
    // AArrayIndexɂёւőΉł킯ł͂Ȃ̂ŁA
    // ArrayIndexɏ]ĕבւf[^VBOō쐬ȂƂȂB 
    bool bFullUpdateVBO = FALSE;
    int packVBO = GetPCPreference()->mCParticleVBOMax.GetBodyINT() * 1000;
    int numParticle = mpCCurrent->mPNum;
    if (numParticle > packVBO) bFullUpdateVBO = TRUE;

    if (renderMode == GL_RENDER){

	bool bSereal = TRUE;

	if (IsPointAll() == TRUE){
	    GLDrawPointAll(packVBO, numParticle, bFullUpdateVBO,
			   lightNum, ensemble);
	    bSereal = FALSE;
	}
	
	if (IsBillboardAll() == TRUE){
	    GLDrawBillboardAll(packVBO, numParticle, bFullUpdateVBO,
			       lightNum, ensemble,
			       baseMat16tmp, baseMat16Invtmp,
			       viewMat16, viewMat16Inv);
	    bSereal = FALSE;
	}
	
	if (IsSphereAll() == TRUE){
	    int sphereDepth = mCSphereDepth.GetBodyINT();
	    mpCCurrent->mbCastShadow = mCbCastShadow.GetBodyBOOL();

	    bool bSingle = FALSE;
	    
	    int num = GetNumMaterialLink();
	    if (num == 1) bSingle = TRUE;

	    MaterialParticle_cls *pCMatParticle[_MAX_PARTICLE_TYPE];
	    
	    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
		pCMatParticle[i] = GetPCMaterialLink(i);

	    mpCDrawParticle->DrawSphere(mpCCurrent, lightNum, ensemble, dissolve, sphereDepth, bSingle, pCMatParticle);
	    bSereal = FALSE;
	}

	if (bSereal == TRUE){
	    int sphereDepth = mCSphereDepth.GetBodyINT();
	    MaterialParticle_cls *pCMatParticle[_MAX_PARTICLE_TYPE];
	    
	    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
		pCMatParticle[i] = GetPCMaterialLink(i);

	    mpCDrawParticle->DrawSereal(mpCCurrent, lightNum, ensemble, dissolve, sphereDepth, 
					mCbTwistRandom.GetBodyBOOL(),
					pCMatParticle,
					baseMat16tmp, baseMat16tmp,
					viewMat16, viewMat16Inv);
	}
    }
    
    glPopAttrib();

    WaitThreadsNOTDeleteFlag(&mbPreparingNextFinished, 1, 1);
}

void CLS::
DrawTL(wxDC& dc, TLPanelGeom_cls* pCGeom,
       wxBitmap &CBitmapBuffer,
       int sashX, int posY, int panelWidth, int height,
       double normalizedWAMin, double normalizedWAMax){

    ParamGroup_cls *pCRemapParamGroup = NULL;
    SequenceRemap_cls* pCSeqRemap
	= (SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    
    bool bRepeat = FALSE;
    if (pCSeqRemap != NULL){
	pCRemapParamGroup = pCSeqRemap->GetPCParamGroup("Remap");
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    mpCParticleSequence->DrawTL(dc, pCGeom, CBitmapBuffer,
				sashX, posY, panelWidth, height,
				normalizedWAMin, normalizedWAMax,
				pCRemapParamGroup, bRepeat);
}

int CLS::
GetUsedMemoryMB(){
    return mpCParticleSequence->GetTotalUsedMemoryMB();
}

void CLS::
PrefetchData(){
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");

    // N[𑗂
    bool bInterpByID = (mCInterpolateBy.GetBodyINT() == _P_INTERP_BY_ID);
    ParticleFormatClone_cls *pCFormatClone = new ParticleFormatClone_cls;
    pCFormatClone->CloneFromFormat(*mpCParticleFormat, bInterpByID);

    // FALSE łΏȂ̂ delete 
    if (mpCParticleSequence->PrefetchData(pCFormatClone,
					  pCSeqRemap) == FALSE)
	delete pCFormatClone;
}

void CLS::
ReleaseMemoryData(){
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");
    
    mpCParticleSequence->ReleaseMemoryData(pCSeqRemap);
}

void CLS::
SetPathGUI(MultiInputPanel_cls *pCMNum, wxString strPath){
    InputItem_cls* pCInputItem = pCMNum->GetPCInputItem(&mCstrPath);
    if (pCInputItem == NULL) return;
    ((AtoSTextCtrl_cls*)(pCInputItem->mpCWindowLink))->SetValue(strPath);
    pCInputItem->OnChanged(FALSE);
}

void CLS::
SetPathGUIExporter(MultiInputPanel_cls *pCMNum, wxString strPath){
    InputItem_cls* pCInputItem = pCMNum->GetPCInputItem(&mCstrPathExporter);
    if (pCInputItem == NULL) return;
    ((AtoSTextCtrl_cls*)(pCInputItem->mpCWindowLink))->SetValue(strPath);
    pCInputItem->OnChanged(FALSE);
}

MultiInputPanel_cls *CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, GetStrName()/*, TRUE*/);

    pC->CombineSizer(2, _("Classify"));
    wxString AstrClassify[2] = {_("Use Type"), _("Expression")};
    pC->SetChoiceList(2, AstrClassify);
    pC->CombineConfItem(mCClassifyType);
    pC->CombineButton(btnClassify, _("Edit"), 120);

    pC->CombineSizer(2, _("Sort"));
    wxString AstrSort[8] = {_("Not Use"), "View-Z", "View-Z inverse", "Distance", "Distance inverse", "type", _("Object"), _("Distance (skip sort)")};
    pC->SetChoiceList(8, AstrSort);
    pC->CombineConfItem(mCSortType, _("Sort"));
    pC->CombineConfItem(mCbDepthMask, _("Depth Mask"));

    mIndexSortFreq = pC->CombineConfItem(mCSortFrequency, _("Sort Frequency"));

    mIndexSortObject = pC->CombineConfItem(mCstrSortObject, _("Sort Object"));
    
    mIndexPath = pC->CombineSizer(2, _("File Path"));
    pC->CombineConfItem(mCstrPath);
    pC->CombineButton(btnParticlePath, "..", 30);

    pC->CombineSizer(2, _("Offset"));
    pC->CombineConfItem(mCSequenceOffset, _("Offset"));
    pC->CombineConfItem(mCbSequence, _("Data Sequence"));

    pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(mCbCastShadow, _("Cast Shadow"));
    pC->CombineConfItem(mCbTwistRandom, _("Random Twist"));

    pC->CombineButton(btnReload, _("Reload"));

    pC->FinishSetting();
    
    return pC;
}

MultiInputPanel_cls *CLS::
GetPCNewPanelExporter(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Exporter"/*, TRUE*/);

    mIndexPathExporter = pC->CombineSizer(2, _("File Path"));
    pC->CombineConfItem(mCstrPathExporter);
    pC->CombineButton(btnParticlePathExporter, "..", 30);

    pC->CombineSizer(2);
    wxString strType[2] = {_("Snap Data"), _("Interpolated Data")};
    pC->SetChoiceList(2, strType);
    pC->CombineConfItem(mCExportType);

    wxString strDir[2] = {_("Horizontal"), _("Vertical")};
    pC->SetChoiceList(2, strDir);
    pC->CombineConfItem(mCExportDir);

    pC->CombineSizer(2);
    mIndexBtnExporter = pC->CombineButton(btnExporter, _("Export"));
    mIndexBtnExportFormat = pC->CombineButton(btnExportFormat, _("Format"), 90);

    pC->FinishSetting();

    return pC;
}

MultiInputPanel_cls *CLS::
GetPCNewPanelCommon(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Material.Common"/*, TRUE*/);
    
    pC->mTextWidth = 80;

    pC->CombineSizer(2, _("Shader"));
    pC->SetChoiceList(3, GetPCStringProvider()->GetAstr("ShaderDissolve"));
    pC->CombineConfItem(mCShaderDissolve);
    wxString AstrPass[2] = {_("1 pass"), _("2 pass")};
    pC->SetChoiceList(2, AstrPass);
    pC->CombineConfItem(mCTwoPass);
    
    int numTexture = mCTextureNum.GetBodyINT();
    std::vector < int > VIndexDelete;
    for (int i = 0; i < numTexture; i++){
	wxString strLabel = _("Texture"); strLabel << " " << i;
	mIndexParticleTexturePath[i] = pC->CombineSizer(4, strLabel);
	pC->CombineConfItem(mCstrTexturePath[i]);
	pC->CombineButton(btnParticleTexture0+i, "..", 30);
	bool bEnable = (i == numTexture - 1);
	pC->CombineButton(btnParticleTextureAdd, "+", 30, bEnable);
	if (i == 0) bEnable = FALSE;
	int index = pC->CombineBitmapPanel(btnParticleTextureDelete, "Delete", 24, bEnable);
	VIndexDelete.push_back(index);
    }
    
    mIndexPointSizeDepend = pC->CombineSizer(2, _("Point Size"));
    wxString AstrListDistance[3] = {_("Constant size"), _("Depend on distance")};
    pC->SetChoiceList(2, AstrListDistance);
    pC->CombineConfItem(mCDependDistance);
    pC->SetLimit(0, 1000);
    pC->mDragScale = 1;
    mIndexDistanceBase = pC->CombineConfItem(mCDistanceBase);

    pC->mDragScale = 1;
    pC->SetLimit(0, 100);
    mIndexPointSize = pC->CombineConfItem(mCPointSizeFactor, _("Point Size"));

    pC->mDragScale = 1;
    pC->SetLimit(0, 1);
    mIndexAlphaCrit = pC->CombineConfItem(mCAlphaCrit, _("Alpha Crit"));

    pC->mDragScale = 4;
    pC->SetLimit(0, 4);
    mIndexSphereDepth = pC->CombineConfItem(mCSphereDepth, _("Division Depth"));
    
    pC->FinishSetting();

    for (int i = 0; i < numTexture; i++){
	BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[VIndexDelete[i]].mpCWindowLink;
	pCBitmap->mbButton = TRUE;
	pCBitmap->AddBitmapStr("wxART_CLOSE_PANEL", "wxART_CLOSE_PANEL_P");
	pCBitmap->SetBitmapStrDisable("wxART_CLOSE_PANEL_D");
	pCBitmap->mBorderX = 3;
	pCBitmap->mBorderY = 4;
    }

    pC->mVCItem[mIndexDistanceBase].mpCWindowLink->SetToolTip(_("Show particles at this distance as n pixeled dot"));
    pC->mVCItem[mIndexAlphaCrit].mpCWindowLink->SetToolTip(_("Ignore particles with size-depend-alpha less than this value"));
    
    return pC;
}

MultiInputPanel_cls *CLS::
GetPCNewPanelInfo(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Particle.InfoIndex"/*, TRUE*/);

    pC->CombineConfItem(mCInfoIndex, "Particle Index");
    pC->CombineSizer(2, _("ID Search"));
    pC->CombineConfItem(mCInfoID, "Particle ID");
    pC->CombineButton(btnInfoIDSearch, _("Search"), 60);
    
    pC->FinishSetting();

    return pC;
}

MultiInputPanel_cls *CLS::
GetPCNewPanelMaterial(wxWindow *pCWindow, int indexMat){
    
    MaterialParticle_cls *pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", indexMat);

    wxString strName = "Material";
    strName << "." << indexMat;

    int numMat = GetNumAttribute("MaterialParticle_cls");
    bool bEnableAdd = FALSE, bEnableDelete = FALSE;
    if (indexMat == numMat-1) bEnableAdd = TRUE;
    if ((indexMat == numMat-1) && numMat > 1) bEnableDelete = TRUE;

    return pCMat->GetPCNewPanelMaterial(pCWindow, strName, indexMat,
					bEnableAdd, bEnableDelete);
}

MultiInputPanel_cls *CLS::
GetPCNewPanelColorCurve(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "ColorCurve"/*, TRUE*/);
    
    pC->mTextWidth = 100;

    int curveNum = GetNumAttribute("ColorCurve_cls");
    std::vector < int > VIndexClrCurve;
    std::vector < int > VIndexDelete;

    uint numAttrib = mVpCAttribute.size();
    int curveIdx = 0;
    for (int i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->GetStrAttributeName().IsSameAs("ColorCurve_cls")){
	    wxString strName = pCAttrib->mCstrName.GetBodySTR();

	    pC->CombineSizer(3, strName);
	    
	    pC->mContextMenuType = _ITEM_MENU_COLOR_CURVE;
	    int indexClrCurve = pC->CombineBitmapPanel(btnMaterialColorCurve0+curveIdx, "Send to Editor");
	    VIndexClrCurve.push_back(indexClrCurve);

	    bool bAdd = FALSE;
	    if (curveIdx == curveNum-1 && curveNum < _MAX_COLOR_CURVE) bAdd = TRUE;
	    pC->CombineButton(btnMaterialColorCurveAdd, "+", 30, bAdd);

	    bool bEnable = FALSE;
	    if ((curveIdx == curveNum-1) && curveNum > 1) bEnable = TRUE;

	    int indexDelete = pC->CombineBitmapPanel(btnMaterialColorCurveDelete, "Delete", 24, bEnable);
	    VIndexDelete.push_back(indexDelete);
	    curveIdx++;
	}
    }

    /*
    bool bAdd = TRUE;
    if (curveNum > 9) bAdd = FALSE;
    pC->CombineButton(btnMaterialColorCurveAdd, "Add Color Curve", -1, bAdd);
    */
    
    pC->FinishSetting();

    uint numClrCurve = VIndexClrCurve.size();
    for (uint i = 0; i < numClrCurve; i++){
	BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[VIndexClrCurve[i]].mpCWindowLink;
	pCBitmap->mbButton = TRUE;
	
	ColorCurve_cls *pCClrCurve = (ColorCurve_cls*)GetPCAttributeFromAttribName("ColorCurve_cls", i);
	pCBitmap->SetColorCurveLink(pCClrCurve);
	pCBitmap->mColorCurveAlpha = _COLOR_CURVE_RGBA;
	pCBitmap->mBorderX = 1;
	pCBitmap->mBorderY = 1;
    }

    uint numDelete = VIndexDelete.size();
    for (uint i = 0; i < numDelete; i++){
	BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[VIndexDelete[i]].mpCWindowLink;
	pCBitmap->mbButton = TRUE;
	pCBitmap->AddBitmapStr("wxART_CLOSE_PANEL", "wxART_CLOSE_PANEL_P");
	pCBitmap->SetBitmapStrDisable("wxART_CLOSE_PANEL_D");
	pCBitmap->mBorderX = 3;
	pCBitmap->mBorderY = 3;
    }

    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelFormatMain(wxWindow *pCWindow){
    return mpCParticleFormat->GetPCNewPanelFormatMain(pCWindow); }

MultiInputPanel_cls* CLS::
GetPCNewPanelFormatHeader(wxWindow *pCWindow){
    return mpCParticleFormat->GetPCNewPanelFormatHeader(pCWindow); }

MultiInputPanel_cls* CLS::
GetPCNewPanelFormatBody(wxWindow *pCWindow){
    return mpCParticleFormat->GetPCNewPanelFormatBody(pCWindow); }

void CLS::
Reload(){
    mpCParticleSequence->ClearSequence();
    mstrPathChecked.Clear();// = "None";
    mpCCurrent->mDrawCounter++;
    mpCCurrent->mstrStatusLog = wxES;
}

void CLS::
Export(){
    mbExporting = TRUE;
    mpCExporterSequence->Set(mCstrPathExporter.GetBodySTR(), 0);
}

void CLS::
ExportFormat(){
    //int frame = GetCurrentFrame();
    
    wxString strFormat;
    switch(mCExportType.GetBodyINT()){
    case 0: {
	ParticleSnap_cls *pCSnap = GetPCCurrentSnap();
	if (pCSnap != NULL)
	    strFormat = pCSnap->GetStrExportFormat(mCExportDir.GetBodyINT());
	break; }
    case 1: 
	strFormat = mpCCurrent->GetStrExportFormat(mCExportDir.GetBodyINT());
	break; 
    }
    
    if (strFormat.IsEmpty() == FALSE)
	wxMessageBox(strFormat, _("Format"), wxOK_DEFAULT);
}

void CLS::
Expression(){
    wxString strCommand = "SelectExpression";
    wxString strItem;
    strItem << GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << mCstrName.GetBodySTR() << "::" << "Classify";
    GetPCMain()->ProcAfterMath(strCommand, strItem);
}

int CLS::
GetIndexColorCurveInit(){
    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++){
	wxString strAttrib = GetPCAttributeFromIndex(i)->GetStrAttributeName();
	if (strAttrib.IsSameAs("ColorCurve_cls") == TRUE) return i;
    }
    return numAttribute;
}

void CLS::
AddTextureCommand(){
    int num = mCTextureNum.GetBodyINT();
    if (num < 3) num++;
    else return;
    wxString strFrom, strTo;
    mCTextureNum.SerializeSingleItem(strFrom);
    mCTextureNum.SetBodyFromData((long long)num);
    mCTextureNum.SerializeSingleItem(strTo);
    GetPCUndo()->PushCommandPair(strFrom, strTo);
    GetPCUndo()->Skip(); //łɕύXĂ̂SkipB
    GetPCMain()->UpdatePanels(mCTextureNum.mstrLinkedPanel.ToStdString());
}

void CLS::
DeleteTextureCommand(){
    int num = mCTextureNum.GetBodyINT();
    if (num > 1) num--;
    else return;
    wxString strFrom, strTo;
    mCTextureNum.SerializeSingleItem(strFrom);
    mCTextureNum.SetBodyFromData((long long)num);
    mCTextureNum.SerializeSingleItem(strTo);
    GetPCUndo()->PushCommandPair(strFrom, strTo);
    GetPCUndo()->Skip(); //łɕύXĂ̂SkipB
    GetPCMain()->UpdatePanels(mCTextureNum.mstrLinkedPanel.ToStdString());
}

void CLS::
AddMaterialCommand(){
    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    wxString strUndo(strTag), strRedo(strTag);

    int numMat = GetNumAttribute("MaterialParticle_cls");
    wxString strName = _("Material");
    strName << "." << numMat;
    
    strUndo << "::Delete::MaterialParticle_cls::" << strName;
    strRedo << "::New::MaterialParticle_cls::" << strName;
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
DeleteHeaderCommand(int index){
    mpCParticleFormat->DeleteHeaderCommand(index, GetStrName()); }

void CLS::
AddHeaderCommand(int index){
    mpCParticleFormat->AddHeaderCommand(index, GetStrName()); }

void CLS::
UpDownHeaderCommand(int index, int direction){
    mpCParticleFormat->UpDownHeaderCommand(index, GetStrName(), direction); }

void CLS::
UpdateHeaderCommand(){
    mpCParticleFormat->UpdateHeaderCommand(GetStrName()); }

void CLS::
DeleteBodyCommand(int index){
    mpCParticleFormat->DeleteBodyCommand(index, GetStrName()); }

void CLS::
AddBodyCommand(int index){
    mpCParticleFormat->AddBodyCommand(index, GetStrName()); }

void CLS::
UpDownBodyCommand(int index, int direction){
    mpCParticleFormat->UpDownBodyCommand(index, GetStrName(), direction); }

void CLS::
UpdateBodyCommand(){
    mpCParticleFormat->UpdateBodyCommand(GetStrName()); }

bool CLS::
ProcNewMaterial(wxString strName){
    if (IsAttributeExist(strName) == TRUE) return FALSE;

    int tgtIdx = GetIndexColorCurveInit();
    std::list < Attribute_cls* >::iterator itr = GetIterAttributeFromIndex(tgtIdx);

    MaterialParticle_cls *pCMat = new MaterialParticle_cls(strName, GetStrName());

    mVpCAttribute.insert(itr, pCMat);
    
    return TRUE;
}

bool CLS::
ProcNewColorCurve(wxString strName){

    if (IsAttributeExist(strName) == TRUE) return FALSE;

    int tgtIdx = GetIndexModifierInit();
    std::list < Attribute_cls* >::iterator itr = GetIterAttributeFromIndex(tgtIdx);
    
    ColorCurve_cls *pCColorCurve = new ColorCurve_cls(strName, GetPCStringProvider()->GetArrayString("RGBA")/*, strName*/);
    double clrInit[2][5] = {{0, 0, 1, 0, 0},
				{1, 1, 0, 0, 1}};
    pCColorCurve->SetInitKeyframes(2, clrInit);
    
    mVpCAttribute.insert(itr, pCColorCurve);

    return TRUE;
}

void CLS::
ProcDeleteMaterial(int index){

    wxString strTag = GetStrTag();

    MaterialParticle_cls *pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", index);

    wxString strName = pCMat->GetStrName();

    wxString strUndo1(strTag), strRedo1(strTag);
    wxString strUndo2(strTag), strRedo2(strTag);

    GetPCUndo()->OpenBlock("MaterialDelete");

    strUndo1 << "::"; pCMat->Serialize(strUndo1); strUndo1.Trim(TRUE);
    strRedo1 << "::Skip";
    GetPCUndo()->PushCommandPair(strUndo1, strRedo1);
    GetPCUndo()->Redo();

    strUndo2 << "::New::MaterialParticle_cls::" << strName;
    strRedo2 << "::Delete::MaterialParticle_cls::" << strName;
    GetPCUndo()->PushCommandPair(strUndo2, strRedo2);
    GetPCUndo()->Redo();
    
    GetPCUndo()->CloseBlock("MaterialDelete");
}

void CLS::
SyncMaterialParticle(){
    int num = GetNumAttribute("MaterialParticle_cls");
    wxString strObj = GetStrName();
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", i);
	pCMat->SyncMaterialParticle(strObj);
    }
}

bool CLS::
HideAndShowMaterialCommon(MNumPainBase_cls *pCPain){
    MultiInputPanel_cls* pCMNum = pCPain->mVpCMNumPanelLink[0];

    bool bChanged = FALSE;
    bool bSizeDepend = (mCDependDistance.GetBodyINT() == 1);
    pCMNum->mVCItem[mIndexDistanceBase].Enable(bSizeDepend);
    
    bool bPoint = DoesMatereialBaseExist(_PARTICLE_POINT);
    bChanged |= pCMNum->mVCItem[mIndexPointSizeDepend].Show(bPoint);
    bChanged |= pCMNum->mVCItem[mIndexPointSize].Show(bPoint);
    
    bool bSphere = DoesMatereialBaseExist(_PARTICLE_SPHERE);
    bChanged |= pCMNum->mVCItem[mIndexSphereDepth].Show(bSphere);

    return bChanged;
}

bool CLS::
HideAndShow(MNumPainBase_cls* pCPain){
    MultiInputPanel_cls* pCMNum = pCPain->mVpCMNumPanelLink[0];

    bool bChanged = FALSE;
    bool bSortFreq = (mCSortType.GetBodyINT() == _SORT_Z ||
		      mCSortType.GetBodyINT() == _SORT_Z_INV);
    bChanged |= pCMNum->mVCItem[mIndexSortFreq].Show(bSortFreq);
    bool bSortObject = (mCSortType.GetBodyINT() == _SORT_OBJECT);
    bChanged |= pCMNum->mVCItem[mIndexSortObject].Show(bSortObject);

    return bChanged;
}

void CLS::
UpdateButtonColorExporter(MNumPainBase_cls* pCPain){
    MultiInputPanel_cls* pCMNum = pCPain->mVpCMNumPanelLink[1];
    if (GetPCTimeCtrl()->mbExport == TRUE && mbExporting == TRUE)
	pCMNum->mVCItem[mIndexBtnExporter].mpCWindowLink->SetBackgroundColour(GetPCColorProvider()->GetColor("Active"));
    else{
	pCMNum->mVCItem[mIndexBtnExporter].mpCWindowLink->SetBackgroundColour(GetPCColorProvider()->GetColor("Null"));
	// Export IΎIɃ[hIt
	mbExporting = FALSE;
    }
}

void CLS::
UpdatePathColorExporter(MNumPainBase_cls *pCPain){
    MultiInputPanel_cls* pCMNum = pCPain->mVpCMNumPanelLink[1];
    int index = mIndexPathExporter+1;
    if (IsExportValid() == FALSE){
	wxColor clr(GetPCColorProvider()->GetColor("Error"));
	pCMNum->mVCItem[index].SetBackgroundColour(clr);
    }
    else
	pCMNum->mVCItem[index].SetBackgroundColour(*wxWHITE);
    pCMNum->mVCItem[index].mpCWindowLink->Refresh();
}

void CLS::
UpdateButtonColorMaterial(MNumPainBase_cls* pCPain){
    int num = GetNumAttribute("MaterialParticle_cls");
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", i);
	pCMat->UpdateButtonColor(pCPain);
    }
}

void CLS::
UpdateFormatPanel(MNumPainBase_cls* pCPain){
    mpCParticleFormat->UpdateFormatPanel(pCPain); }

void CLS::
UpdateHeaderPanel(MNumPainBase_cls* pCPain){
    mpCParticleFormat->UpdateHeaderPanel(pCPain); }

void CLS::
UpdateBodyPanel(MNumPainBase_cls* pCPain){
    mpCParticleFormat->UpdateBodyPanel(pCPain); }

void CLS::
SetMaterialInfo(MNumPainBase_cls *pCPain){
    int num = GetNumAttribute("MaterialParticle_cls");
    for (int i = 0; i < num; i++){
	MaterialParticle_cls* pCMat = (MaterialParticle_cls*)GetPCAttributeFromAttribName("MaterialParticle_cls", i);
	pCMat->SetMaterialInfo(/*pCPain*/);
    }
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    bool bRet = FALSE;
    for (int i = 0; i < 4; i++){
	wxString strName = _("ColorCurve");
	strName << "." << i;

	Attribute_cls *pCClrCurve = GetPCAttribute(strName);
	
	if (pCClrCurve != NULL)
	
	    bRet |= ((ColorCurve_cls*)pCClrCurve)->ReadStrLine(istream, strFirstItem, strLine);
    }
    return bRet;
}

bool CLS::
ReadStrLineFormat(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCParticleFormat->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeMaterial(wxString& strOut){
    //!! IMPLEMENT_ME !!
}

void CLS::
SerializeFormat(wxString& strOut){
    mpCParticleFormat->Serialize(strOut);
}

