// (setq-c-offset 4)
// "clsStreamline.cpp"

#include "clsStreamline.h"
#include "clsColorCurve.h"
#include "clsDrawStreamline.h"
#include "clsMaterialStreamline.h"
#include "clsViewerClone.h"
#include "../clsSceneObjCtrl.h"
#include "../PROP/clsStreamlineProp.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../../GLSL/clsShaderCtrl.h"
#include "../../GLOBJ/clsVBOCameraBox.h"
#include "../../MODIFIER/clsSequenceRemap.h"
#include "../../MODIFIER/clsTrimming.h"
#include "../../MYGL/clsTextureProvider.h"
#include "../../MYGL/clsTextureProvider1D.h"
#include "../../SIMDATA/clsStreamlineFormat.h"
#include "../../SIMDATA/clsStreamlineForRender.h"
#include "../../SIMDATA/clsStreamlineSequence.h"
#include "../../SIMDATA/clsStreamlineSnap.h"
#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../UNDO/clsUndoManager.h"

#include "../../WIDGETCTRL/INPUTITEM/clsAtoSTextCtrl.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"

#include "../../../clsMain.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#include "../../../wxWidgetEnum.h"

#define CLS Streamline_cls

CLS::
Streamline_cls(ParentObj_cls *pCParentObj, wxString strName) :
    DataViewerBase_cls(pCParentObj)
{
    mstrClassName = "Streamline_cls";

    mpCStreamlineSequence = new StreamlineSequence_cls;
    mpCDrawStreamline = new DrawStreamline_cls;

    //mpCMaterial = new Material_cls(strName, mstrClassName);

    mpCStreamlineFormat = new StreamlineFormat_cls(this, strName);
    mpCCurrent = new StreamlineForRender_cls;
    mpCNext = new StreamlineForRender_cls;

    Initialize(strName);

    
    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";
    wxString strUpdateMat = "VIEW3D::PROPERTY_MATERIAL";

    mCConfList.Add(&mCShaderDissolve, "Dissolve=", "0", strUpdate);
    mCConfList.Add(&mCbUseClrCurve, "UseColorCurve=", "0", strUpdate);
    mCConfList.Add(&mCbXRay, "XRay=", "0", strUpdate);
    mCConfList.Add(&mCbRepeatCC, "RepeatCC=", "0", strUpdateMat);
    mCConfList.Add(&mCStreamlineShader, "StreamlineShader=", "0", strUpdateMat);
    
    wxString strAttribName;
    
    strAttribName = _("Material");
    strAttribName << ".0";
    
    MaterialStreamline_cls *pCMaterial = new MaterialStreamline_cls(strAttribName,  GetStrName());
    mVpCAttribute.push_back(pCMaterial);
    
    strAttribName = _("ColorCurve");
    strAttribName << ".0";
    ColorCurve_cls *pCColorCurve = new ColorCurve_cls(strAttribName, GetPCStringProvider()->GetArrayString("RGBA"));
    
    double clrInit[2][5] = {{0, 0, 1, 0, 0},
			    {1, 1, 0, 0, 1}};
    pCColorCurve->SetInitKeyframes(2, clrInit);

    mVpCAttribute.push_back(pCColorCurve);

    mpCStreamlineProp = new StreamlineProp_cls(strName, "StreamlineProp_cls");
    mVpCAttribute.push_back(mpCStreamlineProp);
}

CLS::
~Streamline_cls(){
    delete mpCStreamlineSequence;
    delete mpCDrawStreamline;
    delete mpCStreamlineFormat;
    delete mpCCurrent;
    delete mpCNext;
    //delete mpCStreamlineProp;
}

wxString CLS::
GetStrInfo(){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frame, subframe);

    wxString strInfo;
    strInfo << mpCStreamlineSequence->GetStrInfo(FALSE);

    int seqLen = mpCStreamlineSequence->mSeqLen;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");
    if (pCSeqRemap != NULL)
	frame = (int)pCSeqRemap->GetRemappedFrame(NT, frame, seqLen);

    StreamlineSnap_cls *pCSnap = mpCStreamlineSequence->GetPCSnap(frame);
    if (pCSnap == NULL) return strInfo;
    strInfo << pCSnap->GetStrInfo(/*mCInfoIndex.GetBodyINT()*/);
    
    //strInfo << "Streamline";
    return strInfo;
}

wxString CLS::
GetStrPathCandid(){
    return mpCStreamlineSequence->GetFormat(); }

bool CLS::
IsDataValid(){
    return mpCStreamlineSequence->IsDataValid(); }
    

void CLS::
SyncConfListTagStreamline(){
    wxString strObj = mCstrName.GetBodySTR();
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }

    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }

    //mpCParticleFormat->SyncConfListTag(strObj);
}

void CLS::
Clone(ViewerClone_cls &CClone){
    CClone.CloneConfListFrom(*this);

    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	AttributeClone_cls *pCAttribClone = new AttributeClone_cls;
	Attribute_cls* pCTgt = GetPCAttributeFromIndex(i);
	if (pCTgt->GetStrAttributeName().IsSameAs("MaterialStreamline_cls"))
	    ((MaterialStreamline_cls*)pCTgt)->SetMaterialInfo();
	pCTgt->Clone(*pCAttribClone);
	CClone.mVpCAttribClone.push_back(pCAttribClone);
    }
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	AttributeClone_cls *pCAttribClone = new AttributeClone_cls;
	Attribute_cls* pCTgt = GetPCDataProcFromIndex(i);
	pCTgt->Clone(*pCAttribClone);
	CClone.mVpCAttribClone.push_back(pCAttribClone);
    }
}

void CLS::
CheckFileSequence(){
    int offset = mCSequenceOffset.GetBodyINT();
    mpCStreamlineSequence->Set(mCstrPath.GetBodySTR(), offset);
    mpCStreamlineSequence->CalcMaxOffset(mCbSequence.GetBodyBOOL());
    if (mpCStreamlineSequence->mSeqLen != 1)
	mpCStreamlineSequence->mbTimeDepend = TRUE;

    mpCCurrent->Clear();
    mstrPathChecked = mCstrPath.GetBodySTR();
    if (mpCStreamlineSequence->IsDataValid() == TRUE)
	GetPCMain()->mCLastOpenedFileSTREAM.SetBodyFromData(mstrPathChecked);
}

void CLS::
ProcTrimming(double clipBoxMin[3], double clipBoxMax[3],
	     double matrix16[16], double NT, 
	     float &smoothWidth, float &thinningFactor){
    uint attribNum = mVpCAttribute.size();
    for (uint i = 0; i < attribNum; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->GetStrAttributeName() == "Trimming_cls"){
	    Trimming_cls *pCTrim = (Trimming_cls*)pCAttrib;
	    pCTrim->Execute(clipBoxMin, clipBoxMax);
	    if (pCTrim->mCbValidity.GetBodyBOOL() == TRUE){
		std::vector < double > VThinning = pCTrim->GetPCParamGroup("Thinning")->GetVInterpolatedValue(NT);
		smoothWidth = VThinning[0];
		thinningFactor = VThinning[1];
		if (pCTrim->mCbSmooth.GetBodyBOOL() == FALSE)
		    smoothWidth = 0;
		if (pCTrim->mCbGlobal.GetBodyBOOL() == TRUE)
		    GetMatrix16(matrix16, NT);
	    }
	}
    }
}

void CLS::
GLDrawBox(int status){
    glDisable(GL_LIGHTING);

    glColor4f(0.2f, 0.7f, 0.2f, 1.0f);
    if (status == 1) ColorProvider_cls::SetGLColor("Selected");
    if (status == 2) ColorProvider_cls::SetGLColor("Active");

    GCClassProvider.GetPCVBOCameraBox()->Draw();
}

bool CLS::
PrepareDataNext(ViewerClone_cls *pCStreamlineClone,
		int composit, double NTNext){

    int seqLen = mpCStreamlineSequence->mSeqLen;

    int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NTNext, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);

    bool bRepeat = FALSE;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    if (pCSeqRemap != NULL){
	frame1 = pCSeqRemap->GetRemappedFrame(NTNext, frame1);
	interpFactor = pCSeqRemap->GetInterpFactor(NTNext, frame1, interpFactor);
	if (frame1 < 0 || frame1 >= seqLen) return FALSE;
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    int frame2 = frame1+1;
    ModFrame(frame1, seqLen, bRepeat);
    ModFrame(frame2, seqLen, bRepeat);
    /*
    if (frame1 >= seqLen){
	if (bRepeat) frame1 = Misc_cls::Surplus(frame1, seqLen);
	else         frame1 = seqLen - 1;
    }
    if (frame2 >= seqLen){
	if (bRepeat) frame2 = Misc_cls::Surplus(frame2, seqLen);
	else         frame2 = seqLen - 1;
    }
    */
    
    GetPCTimeCtrl()->AdjustFrameForComposer
	(&frame1, &frame2, &interpFactor, composit);

    bool bTimeDepend = mpCStreamlineSequence->mbTimeDepend;
    StreamlineSnap_cls *pCSnap1 = mpCStreamlineSequence->GenerateIfNot(frame1);
    if (pCSnap1 == NULL) return FALSE;
    
    StreamlineSnap_cls *pCSnap2 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0){
	pCSnap2 = mpCStreamlineSequence->GenerateIfNot(frame2);
	if (pCSnap2 == NULL) return FALSE;
    }

    return FALSE; // ȉ

    /*
    mbPreparingNextFinished = FALSE;
    ThreadStreamlinePrepareNext_cls *pCThreadPrepareNext = 
	new ThreadStreamlinePrepareNext_cls;
    pCThreadPrepareNext->SetParameters
	(mpCNext, pCSnap1, pCSnap2,
	 frame1, subframe, seq, interpFactor, pCStreamlineClone,
	 NTNext, CSortPos);
    pCThreadPrepareNext->CreateThread(&mbPreparingNextFinished, 0, 0);
    pCThreadPrepareNext->Run();
    */
    
    return TRUE;
}

void CLS::
PrepareFilename(StreamlineSnap_cls *pCSnap, int frame){
    wxString strFile;
    if (mpCStreamlineSequence->mbTimeDepend == TRUE)
	strFile = mpCStreamlineSequence->GetModifiedStr(frame);
    else
	strFile = mpCStreamlineSequence->mstrPath;
    pCSnap->SetStrFile(strFile);
    //pCSnap->SetFileSequenceDual(NULL, frame);
    pCSnap->SetFrame(frame);
}

void CLS::
PrepareDraw(int ensemble, int composit){
    if (ensemble == 0 && mProcessedInInstanceLoop == 0)
	if (mstrPathChecked != mCstrPath.GetBodySTR())
	    CheckFileSequence();

    int seqLen = mpCStreamlineSequence->mSeqLen;

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NT, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);

    bool bRepeat = FALSE;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    if (pCSeqRemap != NULL){
	frame1 = pCSeqRemap->GetRemappedFrame(NT, frame1, seqLen);
	interpFactor = pCSeqRemap->GetInterpFactor(NT, frame1, interpFactor);
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    if (interpFactor < 1e-8) interpFactor = 0;

    int frame2 = frame1+1;
    ModFrame(frame1, seqLen, bRepeat);
    ModFrame(frame2, seqLen, bRepeat);

    GetPCTimeCtrl()->StoreInterpFactorForComposit(interpFactor);
    GetPCTimeCtrl()->AdjustFrameForComposer
	(&frame1, &frame2, &interpFactor, composit);

    bool bTimeDepend = mpCStreamlineSequence->mbTimeDepend;
    if (bTimeDepend == TRUE)
	if (frame1 < 0 || frame1 >= seqLen) return;

    StreamlineSnap_cls *pCSnap1 = mpCStreamlineSequence->GenerateIfNot(frame1);
    if (pCSnap1 == NULL) return;

    StreamlineSnap_cls *pCSnap2 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0){
	pCSnap2 = mpCStreamlineSequence->GenerateIfNot(frame2);
	if (pCSnap2 == NULL) return;
    }

    if (ensemble == 0 && mProcessedInInstanceLoop == 0){
	PrepareFilename(pCSnap1, frame1);

	StreamlineFormatClone_cls *pCFormatClone1 = new StreamlineFormatClone_cls;
	pCFormatClone1->CloneFromFormat(*mpCStreamlineFormat);

	pCSnap1->ReadHeader(pCFormatClone1);
	if (pCSnap1->ReadData(pCFormatClone1) == TRUE)
	    GetPCMain()->UpdatePanels("PROPERTY_INFO");
	else
	    delete pCFormatClone1;

	if (pCSnap2 != NULL){
	    StreamlineFormatClone_cls *pCFormatClone2 = new StreamlineFormatClone_cls;
	    pCFormatClone2->CloneFromFormat(*mpCStreamlineFormat);
	    
	    PrepareFilename(pCSnap2, frame2);
	    pCSnap2->ReadHeader(pCFormatClone2);
	    if (pCSnap2->ReadData(pCFormatClone2) == FALSE)
		delete pCFormatClone2;
	}
	
	PrefetchData();
    }

    // f[^̍\ensemble = 0, ŏ̃CX^X ̂Ƃł悢B
    if (ensemble == 0 && mProcessedInInstanceLoop == 0){
	
	if (GetPCTimeCtrl()->mPlayDirection == 1){
	    if (mpCNext->mFrame == frame1 &&
		mpCNext->mFactor == interpFactor){
		StreamlineForRender_cls *pCTmp = mpCCurrent;
		mpCCurrent = mpCNext;
		mpCNext = pCTmp;
	    }
	}

	// NextStepPrepareData ͗ōƂ̂
	// Kvȃf[^̃N[nׂA
	ViewerClone_cls *pCClone = new ViewerClone_cls;
	Clone(*pCClone);
	
	if (bTimeDepend == TRUE && interpFactor != 0){
	    mpCCurrent->PrepareData(pCSnap1, pCSnap2, 
				    frame1, subframe, seq, interpFactor,
				    pCClone, NT);
	}
	else
	    mpCCurrent->PrepareData(pCSnap1, NULL, 
				    frame1, subframe, seq, 0, 
				    pCClone, NT);

	//mpCCurrent->SetColorCurveTexture();

	// pCClone PrepareDataŔj
	if (GetPCTimeCtrl()->mPlayDirection == 1){
	    double NTNext = GetPCTimeCtrl()->GetNextNT();
	    ViewerClone_cls *pCCloneNext = new ViewerClone_cls;
	    Clone(*pCCloneNext);
	    
	    if (PrepareDataNext(pCCloneNext, composit, NTNext) == FALSE)
		delete pCCloneNext;
	}
    }
}

void CLS::
GLDraw(bool bDrawGuideObject, 
       int lightNum, int ensemble, int composit, int status, 
       bool bHittest){

    MaterialStreamline_cls* pCMat = (MaterialStreamline_cls*)GetPCAttributeFromAttribName("MaterialStreamline_cls", 0);

    if (ensemble == 0 && mProcessedInInstanceLoop == 0){
	pCMat->CheckTextureSequence();
    }
    
    if (bDrawGuideObject == TRUE) GLDrawBox(status);

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NT, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);

    int textureMode = 0;
    std::vector < double > VAlpha = pCMat->GetPCParamGroup("AlphaFactor")->GetVInterpolatedValue(NT);
    if (mCbUseClrCurve.GetBodyBOOL() == TRUE){
	ColorCurve_cls *pCColorCurve = GetPCColorCurve(0);
	pCColorCurve->PrepareTexture(mCbRepeatCC.GetBodyBOOL(), VAlpha[0]);
	pCColorCurve->BindTexture(1, mCbRepeatCC.GetBodyBOOL());
	textureMode = 1;
    }
    else{
	GetPCTextureProvider1D()->BindColorTexture1D(1, 1, 1, VAlpha[0], 1);
	textureMode = 1;
	//GetPCTextureProvider()->BindBlankTexture(TRUE, 1);
    }
    
    pCMat->SetDepthMask();

    double matrix16[16];
    LoadIdentity(matrix16);
    double clipBoxMin[3] = {-9999, -9999, -9999};
    double clipBoxMax[3] = {+9999, +9999, +9999};
    float smoothWidth = 0, thinningFactor = 0;
    ProcTrimming(clipBoxMin, clipBoxMax, matrix16, NT, smoothWidth, thinningFactor);
        
    int dissolve = mCShaderDissolve.GetBodyINT();
    int shaderType = mCStreamlineShader.GetBodyINT();
    int mappingType = pCMat->mCMappingType.GetBodyINT();

    if (shaderType == 1)
	pCMat->BindTexture(frame1);

    int renderMode;
    glGetIntegerv(GL_RENDER_MODE, &renderMode);

    if (renderMode == GL_RENDER){
	float time = GetPCTimeCtrl()->GetCurrentNT();
	float lineWidth = mpCStreamlineProp->GetPCParamGroup("Width")->mVCParam[0].GetValue(time);//mCLineWidth.GetBodyDOUBLE();
	
	ProcTexSpaceModBegin();
	
	if (ensemble == 0 && mProcessedInInstanceLoop == 0)
	    mpCDrawStreamline->SetVBOBuffer(mpCCurrent);

	//MaterialStreamline_cls *pCMat = (MaterialStreamline_cls*)GetPCAttribute("Material.0");
	
	glEnable(GL_BLEND);
	// J[J[ugƂ̓VF[_[ŎgQleNX`ɃAt@Ă
	// łȂ΁AGLColorŃAt@Kp
	// Fix ME !! 
	// J[J[uȂ̎łQleNX`ŃAt@𔽉fȂƁAfBUłȂ
	pCMat->SetGLColor(NT, FALSE);
	pCMat->SetBlendFunc();

	mpCDrawStreamline->Draw(mpCCurrent, shaderType, lineWidth, mappingType,
				textureMode, lightNum, ensemble, dissolve,
				clipBoxMin, clipBoxMax, 
				matrix16, smoothWidth, thinningFactor);
	
	ProcTexSpaceModEnd();
    }

    glPopAttrib();
    
    //WaitThreadsNOTDeleteFlag(&mbPreparingNextFinished, 1, 1);
}

void CLS::
DrawTL(wxDC& dc, TLPanelGeom_cls* pCGeom,
       wxBitmap &CBitmapBuffer,
       int sashX, int posY, int panelWidth, int height,
       double normalizedWAMin, double normalizedWAMax){

    ParamGroup_cls *pCRemapParamGroup = NULL;
    SequenceRemap_cls* pCSeqRemap
	= (SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    
    bool bRepeat = FALSE;
    if (pCSeqRemap != NULL){
	pCRemapParamGroup = pCSeqRemap->GetPCParamGroup("Remap");
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    mpCStreamlineSequence->DrawTL(dc, pCGeom, CBitmapBuffer,
				  sashX, posY, panelWidth, height,
				  normalizedWAMin, normalizedWAMax,
				  pCRemapParamGroup, bRepeat);
}

void CLS::
PrefetchData(){
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");

    // N[𑗂
    //bool bInterpByID = (mCInterpolateBy.GetBodyINT() == TRUE);
    StreamlineFormatClone_cls *pCFormatClone = new StreamlineFormatClone_cls;
    pCFormatClone->CloneFromFormat(*mpCStreamlineFormat);
    
    // FALSE łΏȂ̂ delete 
    if (mpCStreamlineSequence->PrefetchData(pCFormatClone,
					    pCSeqRemap) == FALSE)
	delete pCFormatClone;
}

void CLS::
SetPathGUI(MultiInputPanel_cls *pCMNum, wxString strPath){
    InputItem_cls* pCInputItem = pCMNum->GetPCInputItem(&mCstrPath);
    if (pCInputItem == NULL) return;
    ((AtoSTextCtrl_cls*)(pCInputItem->mpCWindowLink))->SetValue(strPath);
    pCInputItem->OnChanged(FALSE);
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, GetStrName()/*, TRUE*/);
    
    mIndexPath = pC->CombineSizer(2, _("File Path"));
    pC->CombineConfItem(mCstrPath);
    pC->CombineButton(btnStreamlinePath, "..", 30);

    pC->CombineSizer(2, _("Offset"));
    pC->CombineConfItem(mCSequenceOffset);
    pC->CombineConfItem(mCbSequence, _("Sequential"));

    pC->CombineSizer(1, _("Option"));
    //pC->CombineConfItem(mCbDepthMask, _("Depth Mask"));
    pC->CombineConfItem(mCbXRay, _("XRay"));

    pC->CombineButton(btnReload, _("Reload"));

    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelLineProp(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, "LineProperty"/*, TRUE*/);

    pC->mTextWidth = 120;
    pC->CombineSizer(2, _("Line Width/Dissolve"));
    pC->SetLimit(0, 128);
    pC->mDragScale = 10;
    pC->CombineConfItem(mpCStreamlineProp->mCLineWidth);
    
    pC->SetChoiceList(3, GetPCStringProvider()->GetAstr("ShaderDissolve"));
    pC->CombineConfItem(mCShaderDissolve);

    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelShader(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, "StreamlineShader"/*, TRUE*/);
    
    pC->mTextWidth = 120;
    wxString AstrStreamlineShader[2] =
	{"Streamline Basic", "Streamline Polygon Line"};
    pC->SetChoiceList(2, AstrStreamlineShader);
    pC->CombineConfItem(mCStreamlineShader, _("Streamline Shader"));

    wxString AstrMapping[2] = 
	{"Segment UV", "Streamline UV"};
    pC->SetChoiceList(2, AstrMapping);

    MaterialStreamline_cls* pCMat = (MaterialStreamline_cls*)GetPCAttributeFromAttribName("MaterialStreamline_cls", 0);
    
    mIndexMappingType = pC->CombineConfItem(pCMat->mCMappingType, _("Mapping Type"));
    
    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelMaterial(wxWindow *pCWindow, int indexMat){
    MaterialStreamline_cls *pCMat = (MaterialStreamline_cls*)GetPCAttributeFromAttribName("MaterialStreamline_cls", indexMat);

    wxString strName = "Material";
    strName << "." << indexMat;

    int numMat = GetNumAttribute("MaterialStreamline_cls");
    bool bEnableAdd = FALSE, bEnableDelete = FALSE;
    if (indexMat == numMat-1) bEnableAdd = TRUE;
    if ((indexMat == numMat-1) && numMat > 1) bEnableDelete = TRUE;

    return pCMat->GetPCNewPanelMaterial(pCWindow, strName, indexMat,
					bEnableAdd, bEnableDelete);
}

MultiInputPanel_cls* CLS::
GetPCNewPanelFormatMain(wxWindow *pCWindow){
    return mpCStreamlineFormat->GetPCNewPanelFormatMain(pCWindow); }


int CLS::
GetIndexColorCurveInit(){
    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++){
	wxString strAttrib = GetPCAttributeFromIndex(i)->GetStrAttributeName();
	if (strAttrib.IsSameAs("ColorCurve_cls") == TRUE) return i;
    }
    return numAttribute;
}

void CLS::
AddMaterialCommand(){
    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    wxString strUndo(strTag), strRedo(strTag);

    int numMat = GetNumAttribute("MaterialStreamline_cls");
    wxString strName = _("Material");
    strName << "." << numMat;
    
    strUndo << "::Delete::MaterialStreamline_cls::" << strName;
    strRedo << "::New::MaterialStreamline_cls::" << strName;
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

bool CLS::
ProcNewMaterial(wxString strName){
    if (IsAttributeExist(strName) == TRUE) return FALSE;

    int tgtIdx = GetIndexColorCurveInit();
    std::list < Attribute_cls* >::iterator itr = GetIterAttributeFromIndex(tgtIdx);

    MaterialStreamline_cls *pCMat = new MaterialStreamline_cls(strName, GetStrName());

    mVpCAttribute.insert(itr, pCMat);
    
    return TRUE;
}

void CLS::
ProcDeleteMaterial(int index){
    wxString strTag = GetStrTag();

    MaterialStreamline_cls *pCMat = (MaterialStreamline_cls*)GetPCAttributeFromAttribName("MaterialStreamline_cls", index);

    wxString strName = pCMat->GetStrName();

    wxString strUndo1(strTag), strRedo1(strTag);
    wxString strUndo2(strTag), strRedo2(strTag);

    GetPCUndo()->OpenBlock("MaterialDelete");

    strUndo1 << "::"; pCMat->Serialize(strUndo1); strUndo1.Trim(TRUE);
    strRedo1 << "::Skip";
    GetPCUndo()->PushCommandPair(strUndo1, strRedo1);
    GetPCUndo()->Redo();

    strUndo2 << "::New::MaterialStreamline_cls::" << strName;
    strRedo2 << "::Delete::MaterialStreamline_cls::" << strName;
    GetPCUndo()->PushCommandPair(strUndo2, strRedo2);
    GetPCUndo()->Redo();
    
    GetPCUndo()->CloseBlock("MaterialDelete");
}

bool CLS::
HideAndShowShader(MultiInputPanel_cls *pCMNumShader){
    bool bChange = FALSE;
    int shaderType = mCStreamlineShader.GetBodyINT();
    bChange |= pCMNumShader->mVCItem[mIndexMappingType].Show(shaderType == 1);
    
    return bChange;
}

bool CLS::
HideAndShowTexture(MultiInputPanel_cls *pCMNumTex, int indexTexturePath){

    MaterialStreamline_cls *pCMat = (MaterialStreamline_cls*)GetPCAttributeFromAttribName("MaterialStreamline_cls", 0);

    bool bChange = FALSE;
    bool bTex = pCMat->mCstrTexturePath.GetBodySTR().IsSameAs("None") == FALSE;
    int indexTextureOffset = pCMat->mIndexTextureOffset;
    int indexMapping       = pCMat->mIndexMapping;
    int indexInterpolate   = pCMat->mIndexInterpolate;
    if (bTex == FALSE) {
	pCMNumTex->mVCItem[indexTexturePath+3].mpCWindowLink->Disable();
	bChange |= pCMNumTex->mVCItem[indexInterpolate].Show(FALSE);
    }
    else{
	pCMNumTex->mVCItem[indexTexturePath+3].mpCWindowLink->Enable();
	bChange |= pCMNumTex->mVCItem[indexInterpolate].Show(TRUE);
    }
    bChange |= pCMNumTex->mVCItem[indexTextureOffset].Show(bTex);
    bChange |= pCMNumTex->mVCItem[indexMapping].Show(FALSE);

    return bChange;
}

void CLS::
SyncMaterialStreamline(){
    int num = GetNumAttribute("MaterialStreamline_cls");
    wxString strObj = GetStrName();
    for (int i = 0; i < num; i++){
	MaterialStreamline_cls* pCMat = (MaterialStreamline_cls*)GetPCAttributeFromAttribName("MaterialStreamline_cls", i);
	pCMat->SyncMaterialStreamline(strObj);
    }
    mpCStreamlineProp->SyncronizeParam(strObj);
}

void CLS::
UpdateButtonColor(MNumPainBase_cls* pCPain){
    int num = GetNumAttribute("MaterialStreamline_cls");
    for (int i = 0; i < num; i++){
	MaterialStreamline_cls* pCMat = (MaterialStreamline_cls*)GetPCAttributeFromAttribName("MaterialStreamline_cls", i);
	pCMat->UpdateButtonColor(pCPain);
    }
}

void CLS::
SetMaterialInfo(MNumPainBase_cls *pCPain){
    int num = GetNumAttribute("MaterialStreamline_cls");
    for (int i = 0; i < num; i++){
	MaterialStreamline_cls* pCMat = (MaterialStreamline_cls*)GetPCAttributeFromAttribName("MaterialStreamline_cls", i);
	pCMat->SetMaterialInfo(/*pCPain*/);
    }
}

void CLS::
Reload(){
    mpCStreamlineSequence->ClearSequence();
    mstrPathChecked.Clear();
    mpCCurrent->mFrame = -1;
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    bool bRet = FALSE;
    //bRet |= mpCMaterial->ReadStrLine(istream, strFirstItem, strLine);
    return bRet;
}

bool CLS::
ReadStrLineFormat(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCStreamlineFormat->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeMaterial(wxString& strOut){
    //!! IMPLEMENT_ME !!
    //mpCMaterial->Serialize(strOut);
}

void CLS::
SerializeFormat(wxString& strOut){
    mpCStreamlineFormat->Serialize(strOut);
}
