// (setq-c-offset 4)
// "clsVolumeBox.cpp"

#include "clsVolumeBox.h"
#include "clsColorCurve.h"
#include "clsDrawVolumeBox.h"
#include "clsViewerClone.h"
#include "../clsMaterial.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../../GLOBJ/clsVBOCameraBox.h"
#include "../../GLSL/clsShaderCtrl.h"
#include "../../MISC/clsMisc.h"
#include "../../MODIFIER/clsTrimming.h"
#include "../../MODIFIER/clsTexSpaceMod.h"
#include "../../MODIFIER/clsSequenceRemap.h"
#include "../../MODIFIER/clsVolumeBoolean.h"
#include "../../MODIFIER/clsVolumeCoord.h"
#include "../../MODIFIER/clsVolumeMask.h"
#include "../../SIMDATA/clsVolumeForRender.h"
#include "../../SIMDATA/clsVolumeSequence.h"
#include "../../SIMDATA/clsVolumeSnap.h"
#include "../../TIMELINE/clsTimeCtrl.h"

#include "../../WIDGETCTRL/clsBitmapPanel.h"
#include "../../WIDGETCTRL/INPUTITEM/clsAtoSTextCtrl.h"
#include "../../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"

#include "../../../clsMain.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#include "../../../wxWidgetEnum.h"

#define CLS ThreadVolumePrepareNext_cls

void CLS::
SetParameters(VolumeFoRender_cls *pCNext,
	      VolumeSnap_cls *pCSnap1, VolumeSnap_cls *pCSnap2,
	      int frame1, float interpFactor, double NTNext,
	      ViewerClone_cls *pCVolumeBoxClone){
    mpCNext = pCNext;
    mpCSnap1 = pCSnap1; mpCSnap2 = pCSnap2;
    mFrame1 = frame1; 
    mInterpFactor = interpFactor;
    mNTNext = NTNext;
    mpCVolumeBoxClone = pCVolumeBoxClone;
}

wxThread::ExitCode CLS::
Entry(){
    mpCNext->PrepareData(mpCSnap1, mpCSnap2,
			 mFrame1, mInterpFactor,
			 mpCVolumeBoxClone, FALSE, mNTNext);

    *mpbFinishedFlagLink = TRUE;
    return 0;
}

#define CLS VolumeBox_cls

CLS::
VolumeBox_cls(ParentObj_cls *pCParentObj, wxString strName) :
    DataViewerBase_cls(pCParentObj)
{
    mSortOrder = 0;

    mstrClassName = "VolumeBox_cls";
    mbPreparingNextFinished = TRUE;
    mbSkipDrawing = FALSE;

    mpCMaterial = new Material_cls(strName, mstrClassName);
    mpCMaterial->mbUseBlendType = TRUE;
    Initialize(strName);
    mpCVolumeSequence = new VolumeSequence_cls;
    mpCDrawVolumeBox = new DrawVolumeBox_cls;
    mpCCurrent = new VolumeFoRender_cls;
    mpCNext = new VolumeFoRender_cls;

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";
    wxString strUpdateMat = "VIEW3D::PROPERTY_MATERIAL";
    wxString strUpdateInfo = "PROPERTY_INFO";
    
    mCConfList.Add(&mCSequenceOffsetZ, "SeqOffsetZ=", "0", strUpdate);
    mCConfList.Add(&mCQuality, "Quality=", "128", strUpdate);
    mCConfList.Add(&mCSubQuality, "SubQuality=", "4", strUpdate);
    mCConfList.Add(&mCTest, "Test=", "1", strUpdate);
    mCConfList.Add(&mCbZStack, "ZStack=", "0", strUpdate);

    mCConfList.Add(&mCVolumeShader, "Shader=", "1", strUpdateMat);
    mCConfList.Add(&mCbDiscard0Value, "Discard0=", "0", strUpdateMat);
    //mCConfList.Add(&mCBlendType, "BlendType=", "0", strUpdateMat);
    mCConfList.Add(&mCbSelfShadow, "SelfShadow=", "0", strUpdateMat);
    mCConfList.Add(&mCShadowFactor, "ShadowFactor=", "1", strUpdateMat);
    mCConfList.Add(&mCShaderQuality, "ShaderQuality=", "16", strUpdateMat);
    mCConfList.Add(&mCShaderFactor, "ShaderFactor=", "-1", strUpdateMat);
    mCConfList.Add(&mCShaderDelta, "ShaderDelta=", "0.01", strUpdateMat);
    mCConfList.Add(&mCbRepeatCC, "RepeatCC=", "0", strUpdateMat);
    mCConfList.Add(&mCInfoIndex, "InfoIndex=", "0 0 0", strUpdateInfo);
    
    ColorCurve_cls *pCColorCurve = new ColorCurve_cls(_("ColorCurve"), GetPCStringProvider()->GetArrayString("RGBA"));

    double clrInit[2][5] = {{0, 0, 1, 0, 0},
			    {1, 1, 0, 0, 1}};
    pCColorCurve->SetInitKeyframes(2, clrInit);
    
    mVpCAttribute.push_back(mpCMaterial);
    mVpCAttribute.push_back(pCColorCurve);
    
    //mpCVolumeSequence->SetMaxInex(mCMaxIndex.GetBodyINT());

    mbZStackChecked = FALSE;
}

CLS::
~VolumeBox_cls(){
    delete mpCVolumeSequence;
    delete mpCDrawVolumeBox;
    delete mpCCurrent;
    delete mpCNext;
}

void CLS::
SyncConfListTagVolumeBox(){
    wxString strObj = mCstrName.GetBodySTR();
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }

    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }
    
    mpCMaterial->SyncConfListTag(strObj);
}

/*
void CLS::
Clone(ViewerClone_cls &CClone, double NT){
    CClone.CloneConfListFrom(*this);
    
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	AttributeClone_cls *pCAttribClone = new AttributeClone_cls;
	Attribute_cls* pCTgt = GetPCAttributeFromIndex(i);
	pCTgt->Clone(*pCAttribClone);
	CClone.mVpCAttribClone.push_back(pCAttribClone);
    }
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	AttributeClone_cls *pCAttribClone = new AttributeClone_cls;
	Attribute_cls* pCTgt = GetPCDataProcFromIndex(i);
	pCTgt->Clone(*pCAttribClone);
	CClone.mVpCAttribClone.push_back(pCAttribClone);
    }

    double matrix16[16];
    GetMatrix16(matrix16, NT);
    memcpy(CClone.mMatrix16, matrix16, sizeof(double)*16);
}
*/

wxString CLS::
GetStrInfo(){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frame, subframe);

    wxString strInfo;
    strInfo << mpCVolumeSequence->GetStrInfo(mCbZStack.GetBodyBOOL());

    int seqLen = mpCVolumeSequence->mSeqLen;
    if (mCbZStack.GetBodyBOOL()) seqLen = mpCVolumeSequence->mSeqLenT;
    
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");
    if (pCSeqRemap != NULL)
	frame = (int)pCSeqRemap->GetRemappedFrame(NT, frame, seqLen);

    VolumeSnap_cls *pCSnap = mpCVolumeSequence->GetPCSnap(frame);
    if (pCSnap == NULL) return strInfo;

    std::vector < int > VInfo = mCInfoIndex.GetBodyINTARRAY();
    strInfo << pCSnap->GetStrInfo(VInfo[0], VInfo[1], VInfo[2]);

    return strInfo;
}

wxString CLS::
GetStrPathCandid(){
    if (mCbZStack.GetBodyBOOL() == TRUE)
	return mpCVolumeSequence->GetFormatDual();
    else 
	return mpCVolumeSequence->GetFormat();
}

bool CLS::
IsDataValid(){
    return mpCVolumeSequence->IsDataValid();
}

void CLS::
InitializeStatusForDrawingExe(){
    mbSkipDrawing = FALSE;
    mbUsePrepareNext = FALSE; // !! oOȂTRUEɂ !!
}

void CLS::
CheckStatusForDrawingExe(){
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->IsAttributeValidAndSameAs("NestedVolume_cls")){
	    SceneObj_cls *pCTgt = pCAttrib->GetPCTargetSceneObj();
	    if (pCTgt != NULL && pCTgt->IsSameAs("VolumeBox_cls")){
		VolumeBox_cls *pCVolumeBox = ((VolumeBox_cls*)pCTgt);
		pCVolumeBox->mbSkipDrawing = TRUE;
		pCVolumeBox->mSortOrder = 2;
		mbUsePrepareNext = FALSE;
	    }
	}
    }
}


void CLS::
CheckFileSequence(){
    int offset = mCSequenceOffset.GetBodyINT();
    int offsetZ = mCSequenceOffsetZ.GetBodyINT();
    if (mCbZStack.GetBodyBOOL() == TRUE){
	mpCVolumeSequence->SetDual(mCstrPath.GetBodySTR(), offset, offsetZ);
	mpCVolumeSequence->CalcMaxOffsetDual();
	if (mpCVolumeSequence->mSeqLenT != 1)
	    mpCVolumeSequence->mbTimeDepend = TRUE;
    }
    else{
	mpCVolumeSequence->Set(mCstrPath.GetBodySTR(), offset);
	mpCVolumeSequence->CalcMaxOffset(mCbSequence.GetBodyBOOL());
	if (mpCVolumeSequence->mSeqLen != 1)
	    mpCVolumeSequence->mbTimeDepend = TRUE;
    }

    mpCCurrent->Clear();
    mstrPathChecked = mCstrPath.GetBodySTR();
    mbZStackChecked = mCbZStack.GetBodyBOOL();
    if (mpCVolumeSequence->IsDataValid() == TRUE)
	GetPCMain()->mCLastOpenedFileVOLUME.SetBodyFromData(mstrPathChecked);
}

void CLS::
ProcTrimming(double clipBoxMin[3], double clipBoxMax[3]){
    uint attribNum = mVpCAttribute.size();
    for (uint i = 0; i < attribNum; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->GetStrAttributeName() == "Trimming_cls")
	    ((Trimming_cls*)pCAttrib)->Execute(clipBoxMin, clipBoxMax);
    }
}

void CLS::
PrepareFilename(VolumeSnap_cls *pCSnap, int frame, bool bZStack){
    wxString strFile;
    if (bZStack == TRUE){
	strFile = mpCVolumeSequence->GetModifiedStrDual(frame, 0);
	pCSnap->SetStrFile(strFile);
	pCSnap->SetFileSequenceDual(mpCVolumeSequence, frame);
    }
    else{
	if (mpCVolumeSequence->mbTimeDepend == TRUE)
	    strFile = mpCVolumeSequence->GetModifiedStr(frame);
	else
	    strFile = mpCVolumeSequence->mstrPath;
	pCSnap->SetStrFile(strFile);
	pCSnap->SetFileSequenceDual(NULL, frame);
    }
}

void CLS::
PrepareDataNext(int composit){

    double NTNext = GetPCTimeCtrl()->GetNextNT(TRUE);
    
    int seqLen = mpCVolumeSequence->mSeqLen;
    if (mCbZStack.GetBodyBOOL()) seqLen = mpCVolumeSequence->mSeqLenT;
    
    int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NTNext, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);
        
    bool bRepeat = FALSE;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    if (pCSeqRemap != NULL){
	frame1 = pCSeqRemap->GetRemappedFrame(NTNext, frame1);
	interpFactor = pCSeqRemap->GetInterpFactor(NTNext, frame1, interpFactor);
	if (frame1 < 0 || frame1 >= seqLen) return;
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    int frame2 = frame1+1;
    ModFrame(frame1, seqLen, bRepeat);
    ModFrame(frame2, seqLen, bRepeat);

    GetPCTimeCtrl()->AdjustFrameForComposer
	(&frame1, &frame2, &interpFactor, composit, TRUE);

    bool bTimeDepend = mpCVolumeSequence->mbTimeDepend;
    VolumeSnap_cls *pCSnap1 = mpCVolumeSequence->GenerateIfNot(frame1);
    if (pCSnap1 == NULL ||
	pCSnap1->mAllocatedSize[X] == 0) return;

    bool bZStack = mCbZStack.GetBodyBOOL();
    PrepareFilename(pCSnap1, frame1, bZStack);
    pCSnap1->ReadHeader(bZStack);
    pCSnap1->ReadData();

    VolumeSnap_cls *pCSnap2 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0){
	pCSnap2 = mpCVolumeSequence->GenerateIfNot(frame2);
	if (pCSnap2 == NULL ||
	    pCSnap2->mAllocatedSize[X] == 0) return;

	PrepareFilename(pCSnap2, frame2, bZStack);
	pCSnap2->ReadHeader(bZStack);
	pCSnap2->ReadData();
    }

    if (mpCNext->IsSameStep(frame1, interpFactor) == TRUE) return;

    // pCClone PrepareDataŔj
    ViewerClone_cls *pCCloneNext = new ViewerClone_cls;
    Clone(*pCCloneNext, NTNext);
    
    mbPreparingNextFinished = FALSE;
    ThreadVolumePrepareNext_cls *pCThreadPrepareNext = 
	new ThreadVolumePrepareNext_cls;
    pCThreadPrepareNext->SetParameters
	(mpCNext, pCSnap1, pCSnap2,
	 frame1, interpFactor, NTNext, pCCloneNext);
    pCThreadPrepareNext->CreateThread(&mbPreparingNextFinished, 0, 0);
    pCThreadPrepareNext->Run();

    //WaitThreadsNOTDeleteFlag(&mbPreparingNextFinished, 1, 3);
    /*
    mpCNext->PrepareData(pCSnap1, pCSnap2,
			 frame1, interpFactor, pCVolumeBoxClone,
			 TRUE);
    */
    return;
}

void CLS::
SetTexture3DParameter(Material_cls *pCMaterial){
    if (pCMaterial->mCTexFilter.GetBodyINT() == 0){
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    }
    else{
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    }
    switch(pCMaterial->mCTexWrap.GetBodyINT()){
    case 0:
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	break;
    case 1:
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_REPEAT);
	break;
    case 2:
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	break;
    }
	
}

VolumeCoord_cls* CLS::
GetPCVolumeCoord(){
    VolumeCoord_cls *pCVolumeCoord = NULL;
    uint numModifier = mVpCAttribute.size();
    for (uint i = 0; i < numModifier; i++){
	if (GetPCAttributeFromIndex(i)->GetStrAttributeName().IsSameAs("VolumeCoord_cls")){
	    pCVolumeCoord = (VolumeCoord_cls*)GetPCAttributeFromIndex(i);
	    if (pCVolumeCoord->mCbValidity.GetBodyBOOL() == TRUE)
		break;
	    else
		pCVolumeCoord = NULL;
	}
    }
    return pCVolumeCoord;
}

VolumeMask_cls* CLS::
GetPCVolumeMask(){
    VolumeMask_cls *pCVolumeMask = NULL;
    uint numModifier = mVpCAttribute.size();
    for (uint i = 0; i < numModifier; i++){
	if (GetPCAttributeFromIndex(i)->GetStrAttributeName().IsSameAs("VolumeMask_cls")){
	    pCVolumeMask = (VolumeMask_cls*)GetPCAttributeFromIndex(i);
	    if (pCVolumeMask->mCbValidity.GetBodyBOOL() == TRUE)
		break;
	    else
		pCVolumeMask = NULL;
	}
    }
    return pCVolumeMask;
}

ShaderItem_cls* CLS::
GetPCShaderItem(int shaderType, bool bSelfShadow){
    ShaderCtrl_cls *pCSC = GetPCShaderCtrl();
    ShaderItem_cls *pCShaderItemLink = NULL;
    switch(shaderType){
    case 1:
	pCShaderItemLink = &pCSC->mVCShader[_SHADER_VOLUME_BASIC];
	break;
    case 2:
	if (bSelfShadow == FALSE)
	    pCShaderItemLink = &pCSC->mVCShader[_SHADER_VOLUME_PHONG];
	else
	    pCShaderItemLink = &pCSC->mVCShader[_SHADER_VOLUME_SHADOW];
	break;
    case 3:
	if (bSelfShadow == FALSE)
	    pCShaderItemLink = &pCSC->mVCShader[_SHADER_VOLUME_SCATTER];
	else
	    pCShaderItemLink = &pCSC->mVCShader[_SHADER_VOLUME_SCATTER_SHADOW];
	break;
    }
    return pCShaderItemLink;
}

void CLS::
SetVolumeShader(int shaderType, Material_cls *pCMaterial, 
		int lightNum, double clipBoxMin[3], double clipBoxMax[3],
		bool bSelfShadow, float alpha, float intervalGlobal){
    ShaderCtrl_cls *pCSC = GetPCShaderCtrl();
    int trim = pCMaterial->mCbVolumeTrim.GetBodyBOOL();
    
    /*
    if (mpCMaterial->mCTexWrap.GetBodyINT() == 0 ||
	mpCMaterial->mCTexWrap.GetBodyINT() == 1) trim = FALSE;
    */
    
    int discard0 = mCbDiscard0Value.GetBodyINT();
    int subQuality = mCSubQuality.GetBodyINT();
    
    switch(shaderType){
    case 1:{
	VolumeBasicParams_cls *pCParam = new VolumeBasicParams_cls;
	LOOP_XYZ pCParam->mClipMin[xyz] = clipBoxMin[xyz];
	LOOP_XYZ pCParam->mClipMax[xyz] = clipBoxMax[xyz];
	pCParam->mDiscard0 = mCbDiscard0Value.GetBodyINT();
	pCParam->mTrim = trim;
	pCParam->mIntervalGlobal = intervalGlobal;// * mCTest.GetBodyDOUBLE();
	pCParam->mSubQuality = subQuality;
	pCParam->mBlendType = mpCMaterial->mCBlendType.GetBodyINT();
	pCSC->BeginVolumeBasic(pCParam);
	delete pCParam;
	break; }
    case 2: 
	if (bSelfShadow == FALSE){
	    VolumePhongParams_cls *pCParam = new VolumePhongParams_cls;
	    pCParam->mLightNumber = lightNum;
	    LOOP_XYZ pCParam->mClipMin[xyz] = clipBoxMin[xyz];
	    LOOP_XYZ pCParam->mClipMax[xyz] = clipBoxMax[xyz];
	    pCParam->mDiscard0 = discard0;
	    pCParam->mTrim = trim;
	    pCParam->mIntervalGlobal = intervalGlobal;
	    pCParam->mSubQuality = subQuality;
	    pCParam->mDelta = mCShaderDelta.GetBodyDOUBLE();
	    pCParam->mBlendType = mpCMaterial->mCBlendType.GetBodyINT();
	    pCMaterial->GetColorInfo(pCParam->mAmbient, pCParam->mDiffise, pCParam->mSpecular, pCParam->mShininess);
	    pCSC->BeginVolumePhong(pCParam);
	    delete pCParam;
	}
	else{
	    VolumePhongShadowParams_cls *pCParam = new VolumePhongShadowParams_cls;
	    pCParam->mLightNumber = lightNum;
	    LOOP_XYZ pCParam->mClipMin[xyz] = clipBoxMin[xyz];
	    LOOP_XYZ pCParam->mClipMax[xyz] = clipBoxMax[xyz];
	    pCParam->mBaseAlpha = alpha;
	    pCParam->mShadowQuality = mCShaderQuality.GetBodyINT();
	    pCParam->mShadowFactor = mCShadowFactor.GetBodyDOUBLE();
	    pCParam->mDiscard0 = discard0;
	    pCParam->mTrim = trim;
	    pCParam->mIntervalGlobal = intervalGlobal;
	    pCParam->mSubQuality = subQuality;
	    pCParam->mDelta = mCShaderDelta.GetBodyDOUBLE();
	    pCParam->mBlendType = mpCMaterial->mCBlendType.GetBodyINT();
	    LOOP_XYZ pCParam->mClipMin[xyz] = clipBoxMin[xyz];
	    LOOP_XYZ pCParam->mClipMax[xyz] = clipBoxMax[xyz];
	    pCMaterial->GetColorInfo(pCParam->mAmbient, pCParam->mDiffise, pCParam->mSpecular, pCParam->mShininess);
	    pCSC->BeginVolumePhongShadow(pCParam);
	    
	    delete pCParam;
	}
	break; 
    case 3: 
	if (bSelfShadow == FALSE){
	    VolumeScatterParams_cls *pCParam = new VolumeScatterParams_cls;
	    pCParam->mLightNumber = lightNum;
	    LOOP_XYZ pCParam->mClipMin[xyz] = clipBoxMin[xyz];
	    LOOP_XYZ pCParam->mClipMax[xyz] = clipBoxMax[xyz];
	    pCParam->mScatterFatcor = mCShaderFactor.GetBodyDOUBLE();
	    pCParam->mDiscard0 = discard0;
	    pCParam->mTrim = trim;
	    pCParam->mIntervalGlobal = intervalGlobal;
	    pCParam->mSubQuality = subQuality;
	    pCParam->mBlendType = mpCMaterial->mCBlendType.GetBodyINT();
	    pCMaterial->GetColorInfo(pCParam->mAmbient, pCParam->mDiffise, pCParam->mSpecular, pCParam->mShininess);
	    pCSC->BeginVolumeScatter(pCParam);
	    delete pCParam;
	}
	else{
	    VolumeScatterShadowParams_cls *pCParam = new VolumeScatterShadowParams_cls;
	    pCParam->mLightNumber = lightNum;
	    pCParam->mScatterFatcor = mCShaderFactor.GetBodyDOUBLE();
	    pCParam->mDiscard0 = discard0;
	    pCParam->mTrim = trim;
	    pCParam->mIntervalGlobal = intervalGlobal;
	    pCParam->mSubQuality = subQuality;
	    pCParam->mBaseAlpha = alpha;
	    pCParam->mShadowQuality = mCShaderQuality.GetBodyINT();
	    pCParam->mShadowFactor = mCShadowFactor.GetBodyDOUBLE();
	    pCParam->mBlendType = mpCMaterial->mCBlendType.GetBodyINT();
	    LOOP_XYZ pCParam->mClipMin[xyz] = clipBoxMin[xyz];
	    LOOP_XYZ pCParam->mClipMax[xyz] = clipBoxMax[xyz];
	    pCMaterial->GetColorInfo(pCParam->mAmbient, pCParam->mDiffise, pCParam->mSpecular, pCParam->mShininess);
	    pCSC->BeginVolumeScatterShadow(pCParam);
	    
	    delete pCParam;
	}
	break; 
    }
}

std::vector < VolumeBox_cls* > CLS::
GetVpCDependency(){
    std::vector < VolumeBox_cls* > VpCDepend;
    
    uint numAttribute = mVpCDataProc.size();
    for (uint i = 0; i < numAttribute; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	if (pCAttrib->IsAttributeValidAndSameAs("NestedVolume_cls")){
	    SceneObj_cls *pCTgt = pCAttrib->GetPCTargetSceneObj();
	    if (pCTgt != NULL && pCTgt->IsSameAs("VolumeBox_cls") &&
		pCTgt->mCbVisible.GetBodyBOOL() == TRUE)
		VpCDepend.push_back((VolumeBox_cls*)pCTgt);
	}
    }

    return VpCDepend;
}

void CLS::
PrepareDraw(int ensemble, int composit, bool bPrepareNext){
    if (ensemble == 0 && mProcessedInInstanceLoop == 0)
	if (mstrPathChecked != mCstrPath.GetBodySTR() ||
	    mbZStackChecked != mCbZStack.GetBodyBOOL())
	    CheckFileSequence();

    int seqLen = mpCVolumeSequence->mSeqLen;
    if (mpCVolumeSequence->mSeqLenT != -1)
	seqLen = mpCVolumeSequence->mSeqLenT;
    
    double NT = GetPCTimeCtrl()->GetCurrentNT(TRUE);
    int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NT, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);
    
    bool bRepeat = FALSE;
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    if (pCSeqRemap != NULL){
	frame1 = pCSeqRemap->GetRemappedFrame(NT, frame1, seqLen);
	interpFactor = pCSeqRemap->GetInterpFactor(NT, frame1, interpFactor);
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    wxString strLog;
    strLog << "VolumeTime NT=" << NT << " frame1=" << frame1 << " interp=" << interpFactor << ENDL;

    if (interpFactor < 1e-8) interpFactor = 0;
    
    int frame2 = frame1+1;
    ModFrame(frame1, seqLen, bRepeat);
    ModFrame(frame2, seqLen, bRepeat);

    GetPCTimeCtrl()->StoreInterpFactorForComposit(interpFactor);
    GetPCTimeCtrl()->AdjustFrameForComposer
	(&frame1, &frame2, &interpFactor, composit);

    if (mpCVolumeSequence->mbTimeDepend == TRUE)
	if (frame1 < 0 || frame1 >= seqLen) return;

    strLog << "Adjusted frame1=" << frame1 << " frame2=" << frame2 << "interp=" << interpFactor << " composit=" << composit << ENDL;
    wxLogMessage(strLog);
    
    bool bTimeDepend = mpCVolumeSequence->mbTimeDepend;
    VolumeSnap_cls *pCSnap1 = mpCVolumeSequence->GenerateIfNot(frame1);
    if (pCSnap1 == NULL) return;

    VolumeSnap_cls *pCSnap2 = NULL;
    if (bTimeDepend == TRUE && interpFactor != 0){
	pCSnap2 = mpCVolumeSequence->GenerateIfNot(frame2);
	if (pCSnap2 == NULL) return;
    }
    
    if (ensemble == 0 && mProcessedInInstanceLoop == 0 || TRUE){
	bool bZStack = mCbZStack.GetBodyBOOL();
	PrepareFilename(pCSnap1, frame1, bZStack);
	pCSnap1->ReadHeader(bZStack);
	if (pCSnap1->ReadData() == TRUE)
	    GetPCMain()->UpdatePanels("PROPERTY_INFO");
	
	if (pCSnap2 != NULL){
	    PrepareFilename(pCSnap2, frame2, bZStack);
	    pCSnap2->ReadHeader(bZStack);
	    pCSnap2->ReadData();
	}

	/*
	wxString strLog = "Prepare Data frame";
	strLog << frame1 << "::" << frame2 << "::" << interpFactor << ENDL;
	wxLogMessage(strLog);
	*/
	
	PrefetchData();
    }

    // f[^̍\ensemble = 0, ŏ̃CX^X ̂Ƃł悢B
    if (ensemble == 0 && mProcessedInInstanceLoop == 0){

	if (GetPCTimeCtrl()->mPlayDirection == 1){
	    if (mpCNext->mFrame == frame1 &&
		Misc_cls::IsNearlySame((double)mpCNext->mFactor, interpFactor, 1e-5)){
		VolumeFoRender_cls *pCTmp = mpCCurrent;
		mpCCurrent = mpCNext;
		mpCNext = pCTmp;
	    }
	}
	
	// NextStepPrepareData ͗ōƂ̂
	// Kvȃf[^̃N[nׂA
	bool bForceRecalcCurrent = FALSE;
	wxString strProcCounterForProcess = mpCCurrent->GetStrProcCounterForProcess(this, NT);
	if (strProcCounterForProcess.IsSameAs(mpCCurrent->mstrProcCounterForProcess) == FALSE) {
	    wxLogMessage(strProcCounterForProcess+"\n");
	    wxLogMessage(mpCCurrent->mstrProcCounterForProcess+"\n");
	    bForceRecalcCurrent = TRUE;
	    
	}
	
	if (bForceRecalcCurrent == TRUE || 
	    mpCCurrent->IsSameStep(frame1, interpFactor) == FALSE){
	    
	    ViewerClone_cls *pCClone = new ViewerClone_cls;
	    Clone(*pCClone, NT);
	    if (bTimeDepend == TRUE && interpFactor != 0){
		mpCCurrent->PrepareData(pCSnap1, pCSnap2, 
					frame1, interpFactor, pCClone, 
					bForceRecalcCurrent, NT);
	    }
	    else
		mpCCurrent->PrepareData(pCSnap1, NULL, frame1, 0, pCClone, 
					bForceRecalcCurrent, NT);

	    
	}

	/*bPrepareNext = bPrepareNext && mbUsePrepareNext*/;
	if (GetPCTimeCtrl()->mPlayDirection == 1 && bPrepareNext == TRUE)
	    PrepareDataNext(composit);
    }
}

void CLS::
GLDraw(bool bDrawGuideObject, 
       int lightNum, int ensemble, int composit, int status, bool bHittest){

    if (bDrawGuideObject == TRUE) GLDrawBox(status);

    double NT = GetPCTimeCtrl()->GetCurrentNT(TRUE);
    
    double modelview[16], invert[16];
    //glGetDoublev(GL_TRANSPOSE_MODELVIEW_MATRIX, matrix);
    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);

    std::vector < double > VScale = GetPCParamGroup("Scale")->GetVInterpolatedValue(NT);
    
    GetInvMatrix4x4(modelview, invert);

    PositionAndDirection_cls CCamDummy(invert, TRUE);

    // 2|B
    LOOP_XYZ{
	CCamDummy.mTgt[xyz] *= VScale[xyz];
	CCamDummy.mRgt[xyz] *= VScale[xyz];
	CCamDummy.mTop[xyz] *= VScale[xyz]; }
    LOOP_XYZ{
	CCamDummy.mTgt[xyz] *= VScale[xyz];
	CCamDummy.mRgt[xyz] *= VScale[xyz];
	CCamDummy.mTop[xyz] *= VScale[xyz]; }

    Material_cls *pCMaterial = mpCMaterial->GetPCLinked("VolumeBox_cls");
    int quality = mCQuality.GetBodyINT();
    float alpha = pCMaterial->GetPCParamGroup("Alpha")->GetVInterpolatedValue(NT)[0];

    double clipBoxMin[3] = {0, 0, 0}, clipBoxMax[3] = {1, 1, 1};
    ProcTrimming(clipBoxMin, clipBoxMax);

    float intervalGlobal = 
	mpCDrawVolumeBox->CalcAllSlices(CCamDummy, quality,
					clipBoxMin, clipBoxMax);
    
    int shaderType = mCVolumeShader.GetBodyINT();
    bool bSelfShadow = mCbSelfShadow.GetBodyBOOL();

    ShaderItem_cls *pCShaderItemLink = GetPCShaderItem(shaderType, bSelfShadow);

    VolumeCoord_cls   *pCVolumeCoord = GetPCVolumeCoord();
    VolumeMask_cls    *pCVolumeMask = GetPCVolumeMask();
    
    if (pCShaderItemLink != NULL){
	bool bChanged = FALSE;
	
	ShaderStatusVolumeCoord_cls &CStatusVolumeCoord = 
	    pCShaderItemLink->mCStatusVolumeCoord;
	ShaderStatusVolumeMask_cls &CStatusVolumeMask = 
	    pCShaderItemLink->mCStatusVolumeMask;

	if (pCVolumeCoord == NULL && CStatusVolumeCoord.mbUseCompiled == TRUE)
	    bChanged = TRUE;
	if (pCVolumeMask == NULL && CStatusVolumeMask.mbUseCompiled == TRUE)
	    bChanged = TRUE;

	if (pCVolumeCoord != NULL && pCVolumeCoord->IsStatusChanged(CStatusVolumeCoord) == TRUE)
	    bChanged = TRUE;
	if (pCVolumeMask != NULL && pCVolumeMask->IsStatusChanged(CStatusVolumeMask) == TRUE)
	    bChanged = TRUE;

	if (bChanged == TRUE){
	    pCShaderItemLink->Clear();
	    pCShaderItemLink->mstrFrag = pCShaderItemLink->mstrFragSrc;
	    if (pCVolumeCoord != NULL){
		pCVolumeCoord->SetStatus(CStatusVolumeCoord);
		pCShaderItemLink->mstrFrag = 
		    pCVolumeCoord->GetStrModifiedShader(pCShaderItemLink->mstrFrag);
	    }
	    if (pCVolumeMask != NULL){
		pCVolumeMask->SetStatus(CStatusVolumeMask);
		pCShaderItemLink->mstrFrag = 
		    pCVolumeMask->GetStrModifiedShader(pCShaderItemLink->mstrFrag);
	    }
	}
    }

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glDisable(GL_LIGHTING);
    glDepthMask(mpCMaterial->mCbDepthMask.GetBodyBOOL() ? GL_TRUE : GL_FALSE);
    
    glPushMatrix();

    mpCCurrent->PrepareTexture();
    mpCCurrent->BindTexture();
    SetTexture3DParameter(pCMaterial);

    ColorCurve_cls* pCColorCurve = GetPCColorCurve(0);
    
    pCColorCurve->PrepareTexture(mCbRepeatCC.GetBodyBOOL());
    pCColorCurve->BindTexture();
    
    SetVolumeShader(shaderType, pCMaterial,
		    lightNum, clipBoxMin, clipBoxMax,
		    bSelfShadow, alpha, intervalGlobal);

    int matrixMode;
    glGetIntegerv(GL_MATRIX_MODE, &matrixMode);
    int modeModelView = GL_MODELVIEW;
    int modeTexture = GL_TEXTURE;
    int modeProjection = GL_PROJECTION;

    glScaled(2, 2, 2);

    ProcTexSpaceModBegin();

    glEnable(GL_BLEND);
    if (shaderType != 0)
	mpCMaterial->SetBlendFunc();
    else
	mpCMaterial->SetAlphaBlendFunc();

    mpCDrawVolumeBox->DrawAllSlice(quality, alpha);

    if (mCVolumeShader.GetBodyINT() != 0) 
	GetPCShaderCtrl()->EndShader();

    ProcTexSpaceModEnd();
    
    glScaled(0.5, 0.5, 0.5);

    glDisable(GL_TEXTURE_3D);

    glPopMatrix();
    glPopAttrib();

    WaitThreadsNOTDeleteFlag(&mbPreparingNextFinished, 1, 3);

    CheckGLError("VolumeBox::End");
}

void CLS::
DrawTL(wxDC& dc, TLPanelGeom_cls* pCGeom,
       wxBitmap &CBitmapBuffer,
       int sashX, int posY, int panelWidth, int height,
       double normalizedWAMin, double normalizedWAMax){

    ParamGroup_cls *pCRemapParamGroup = NULL;
    SequenceRemap_cls* pCSeqRemap
	= (SequenceRemap_cls*)GetPCValidModifier("SequenceRemap");
    
    bool bRepeat = FALSE;
    if (pCSeqRemap != NULL){
	pCRemapParamGroup = pCSeqRemap->GetPCParamGroup("Remap");
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }

    mpCVolumeSequence->DrawTL(dc, pCGeom, CBitmapBuffer,
			      sashX, posY, panelWidth, height,
			      normalizedWAMin, normalizedWAMax,
			      pCRemapParamGroup, bRepeat);
}

int CLS::
GetUsedMemoryMB(){
    return mpCVolumeSequence->GetTotalUsedMemoryMB();
}

void CLS::
PrefetchData(){
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");
    
    bool bZStack = mCbZStack.GetBodyBOOL();
    mpCVolumeSequence->PrefetchData(pCSeqRemap, bZStack);
}

void CLS::
ReleaseMemoryData(){
    SequenceRemap_cls* pCSeqRemap = 
	(SequenceRemap_cls*)GetPCAttribute("SequenceRemap");

    mpCVolumeSequence->ReleaseMemoryData(pCSeqRemap);
}

void CLS::
SetPathGUI(MultiInputPanel_cls *pCMNum, wxString strPath){
    InputItem_cls* pCInputItem = pCMNum->GetPCInputItem(&mCstrPath);
    if (pCInputItem == NULL) return;
    ((AtoSTextCtrl_cls*)(pCInputItem->mpCWindowLink))->SetValue(strPath);
    pCInputItem->OnChanged(FALSE);
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName()/*, TRUE*/);

    pC->CombineSizer(3, _("Quality"));
    pC->CombineConfItem(mCQuality);
    pC->CombineConfItem(mCSubQuality);
    pC->CombineConfItem(mCTest);
    pC->CombineSizer(2, _("Data"));
    pC->CombineConfItem(mCbZStack, _("ZStack"));
    pC->CombineConfItem(mCbSequence, _("Data Sequence"));

    mIndexPath = pC->CombineSizer(2, _("File Path"));
    pC->CombineConfItem(mCstrPath);
    pC->CombineButton(btnVolumePath, "..", 30);

    pC->CombineSizer(2, _("Offset T/Z"));
    pC->CombineConfItem(mCSequenceOffset);
    pC->CombineConfItem(mCSequenceOffsetZ);

    pC->CombineSizer(2);
    pC->CombineButton(btnAutoscale, _("Auto Scale"));
    pC->CombineButton(btnReload, _("Reload"));

    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelShader(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, "VolumeShader"/*, TRUE*/);

    pC->mTextWidth = 80;
    pC->CombineSizer(2, _("Type/Delta"));
    wxString AstrVolumeShader[4] =
	{_("Original"), "Volume Basic", "Volume Phong", "Volume Scatter"};
    pC->SetChoiceList(4, AstrVolumeShader);
    pC->CombineConfItem(mCVolumeShader);
    pC->CombineConfItem(mCShaderDelta);

    pC->CombineSizer(2, _("Qual./Scat."));
    pC->CombineConfItem(mCShaderQuality);
    pC->CombineConfItem(mCShaderFactor);
    pC->CombineSizer(2, _("Shadow"));
    pC->CombineConfItem(mCbSelfShadow, _("Self Shadow"));
    pC->SetLimit(0, 32);
    pC->mDragScale = 0.3;
    pC->CombineConfItem(mCShadowFactor);
    
    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls *CLS::
GetPCNewPanelTexture(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Texture"/*, TRUE*/);

    pC->mTextWidth = 80;
    pC->CombineSizer(2, _("Interpolate"));
    wxString AstrInterp[2] = {"Nearest", "Linear"};
    pC->SetChoiceList(2, AstrInterp);
    pC->CombineConfItem(mpCMaterial->mCTexFilter);
    wxString AstrWrap[3] = {"Clamp to Edge", _("Repeat"), "Clamp"};
    pC->SetChoiceList(3, AstrWrap);
    pC->CombineConfItem(mpCMaterial->mCTexWrap);
    pC->CombineConfItem(mpCMaterial->mCbVolumeTrim, _("Clipping"));

    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls *CLS::
GetPCNewPanelOption(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Option"/*, TRUE*/);

    pC->CombineSizer(2);
    pC->CombineConfItem(mCbDiscard0Value, _("Discard 0"));
    /*
    pC->CombineConfItem(mpCMaterial->mCbAdditive, _("Additive"));
    */
    pC->SetChoiceList(3, GetPCStringProvider()->GetAstr("BlendType"));
    pC->CombineConfItem(mpCMaterial->mCBlendType);
    pC->CombineSizer(1);
    pC->CombineConfItem(mpCMaterial->mCbDepthMask, _("Depth Mask"));

    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls *CLS::
GetPCNewPanelInfo(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Volume.InfoIndex"/*, TRUE*/);
    
    pC->CombineConfItem(mCInfoIndex, "x-y-z");

    pC->FinishSetting();
    return pC;
}

void CLS::
Rescale(){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frame, subframe);

    VolumeSnap_cls *pCSnap = mpCVolumeSequence->GetPCSnap(frame);
    if (pCSnap == NULL) return;
    double scaleCandid[3];
    pCSnap->GetScaleCandid(scaleCandid);

    std::vector < double > VScale = GetPCParamGroup("Scale")->GetVInterpolatedValue(NT, FALSE);
    bool bDiff = FALSE;
    LOOP_XYZ if (VScale[xyz] != scaleCandid[xyz]) bDiff = TRUE;
    if (bDiff == FALSE) return;

    SetParam3Command("Scale", scaleCandid);
}

void CLS::
Reload(){
    mpCVolumeSequence->ClearSequence();
    mstrPathChecked.Clear();
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    bool bRet = FALSE;
    bRet |= mpCMaterial->ReadStrLine(istream, strFirstItem, strLine);
    //bRet |= ((ColorCurve_cls*)GetPCModifierFromIndex(1))->ReadStrLine(istream, strFirstItem, strLine);
    return bRet;
}

void CLS::
SerializeMaterial(wxString& strOut){
    mpCMaterial->Serialize(strOut); 
    // !! FIX_ME !!
    // J[J[u̐𑝂₵Ƃ
    // 2߈ȍ~ new ŃCX^XĂƂȂƂ͂
    //((ColorCurve_cls*)GetPCModifierFromIndex(1))->Serialize(strOut);
}
