// (setq-c-offset 4)
// "clsFocalBlurProp.h"

#include "clsFocalBlurProp.h"
#include "clsStereoProp.h"

#include "../../COMMON/clsHalton.h"
#include "../../MYGL/clsPosition.h"
#include "../../PARAMETER/clsParameter.h"
#include "../../PARAMETER/clsParamGroup.h"
#include "../../SCENEOBJ/clsSceneObj.h"
#include "../../SCENEOBJ/clsSceneObjCtrl.h"
#include "../../TIMELINE/clsTimeCtrl.h"
#include "../../COMMON/CommonDefinitions.h"

#define CLS FocalBlurProp_cls

CLS::
FocalBlurProp_cls(wxString strName, wxString strParentClassName,
		  StereoProp_cls *pCSteroProp) : Attribute_cls("FocalBlurProp_cls", "FocalBlurProp") {
    mpCStereoPropLink = pCSteroProp;

    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strName;
    strConfListLabel << wxT("::FocalBlurProp");

    mCConfList.SetTag(strConfListLabel);

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";
    
    mCConfList.Add(&mCbUseFocalBlur, "FocalBlur=", "0", strUpdate);
    mCConfList.Add(&mCSize, "BlurSize=", "0.3", strUpdate);
    mCConfList.Add(&mCDistance, "BlurDistance=", "10", strUpdate);
    mCConfList.Add(&mCFocusType, "BlurFocusType=", "0", strUpdate);
    mCConfList.Add(&mCstrFocusObj, "BlurFocusObj=", "None", strUpdate);

    AddOneDimParam("Size", "S", 0.3);
    AddOneDimParam("Distance", "D", 10);
}


float CLS::
GetWidthBase(double nt){
    return GetPCParamGroup("Size")->GetVInterpolatedValue(nt)[0];
}

float CLS::
GetFocalDistance(PositionAndDirection_cls &CCam, double nt){
    //float softSize = GetPCParamGroup("Size")->GetVInterpolatedValue(nt)[0];

    float dist = GetPCParamGroup("Distance")->GetVInterpolatedValue(nt)[0];

    int focusType = mCFocusType.GetBodyINT();
    switch(focusType){
    case _FOCUS_ORIGIN:
	dist = sqrt(CCam.mPos[X]*CCam.mPos[X] + CCam.mPos[Y]*CCam.mPos[Y] + CCam.mPos[Z]*CCam.mPos[Z]);
	break;
    case _FOCUS_OBJECT: {
	SceneObj_cls *pCTgt = GetPCSceneObjCtrl()->GetPCFromName(mCstrFocusObj.GetBodySTR());
	if (pCTgt == NULL){
	    dist = sqrt(CCam.mPos[X]*CCam.mPos[X] + CCam.mPos[Y]*CCam.mPos[Y] + CCam.mPos[Z]*CCam.mPos[Z]);
	    break;
	}

	double NT = GetPCTimeCtrl()->GetCurrentNT();
	std::vector < double > VPosGlobal = pCTgt->GetPositionGlobal(NT);

	dist = sqrt((CCam.mPos[X]-VPosGlobal[X])*(CCam.mPos[X]-VPosGlobal[X])+
		    (CCam.mPos[Y]-VPosGlobal[Y])*(CCam.mPos[Y]-VPosGlobal[Y])+
		    (CCam.mPos[Z]-VPosGlobal[Z])*(CCam.mPos[Z]-VPosGlobal[Z]));
	break; }
    case _FOCUS_STEROFOCUS: {
	dist = mpCStereoPropLink->GetStereoFocusDistance(nt);

	break; }
    }

    return dist;
}

void CLS::
BlurRotate(PositionAndDirection_cls &CCam, int ensemble, double nt){

    if (mCbUseFocalBlur.GetBodyBOOL() == FALSE) return;
    if (ensemble == 0) return;

    //double nt = GetPCTimeCtrl()->GetCurrentNT();
    //float softSize = GetPCParamGroup("Size")->GetVInterpolatedValue(nt)[0];
    float softSize = GetWidthBase(nt);
    float dist = GetFocalDistance(CCam, nt);

    /*
    float dist = GetPCParamGroup("Distance")->GetVInterpolatedValue(nt)[0];

    int focusType = mCFocusType.GetBodyINT();
    switch(focusType){
    case _FOCUS_ORIGIN:
	dist = sqrt(CCam.mPos[X]*CCam.mPos[X] + CCam.mPos[Y]*CCam.mPos[Y] + CCam.mPos[Z]*CCam.mPos[Z]);
	break;
    case _FOCUS_OBJECT: {
	SceneObj_cls *pCTgt = GetPCSceneObjCtrl()->GetPCFromName(mCstrFocusObj.GetBodySTR());
	if (pCTgt == NULL){
	    dist = sqrt(CCam.mPos[X]*CCam.mPos[X] + CCam.mPos[Y]*CCam.mPos[Y] + CCam.mPos[Z]*CCam.mPos[Z]);
	    break;
	}

	double NT = GetPCTimeCtrl()->GetCurrentNT();
	std::vector < double > VPosGlobal = pCTgt->GetPositionGlobal(NT);

	dist = sqrt((CCam.mPos[X]-VPosGlobal[X])*(CCam.mPos[X]-VPosGlobal[X])+
		    (CCam.mPos[Y]-VPosGlobal[Y])*(CCam.mPos[Y]-VPosGlobal[Y])+
		    (CCam.mPos[Z]-VPosGlobal[Z])*(CCam.mPos[Z]-VPosGlobal[Z]));
	break; }
    case _FOCUS_STEROFOCUS: {
	dist = mpCStereoPropLink->GetStereoFocusDistance(nt);

	break; }
    }
    */

    float x = (GetPCHalton()->GetX(ensemble-1)-0.5) * softSize;
    float y = (GetPCHalton()->GetY(ensemble-1)-0.5) * softSize;
    
    CCam.TranslateInWC(x, y);

    float degx = atan(x/dist) * 180 / 3.14159;
    float degy = atan(y/dist) * 180 / 3.14159;
    CCam.RotateYaw(-degx);
    CCam.RotatePitch(-degy);
}

void CLS::
SyncronizeParam(wxString strObj){
    SyncronizeParamAndConf(strObj, "Size", mCSize);
    SyncronizeParamAndConf(strObj, "Distance", mCDistance);
}

void CLS::
SyncronizeConf(double time){
    SyncronizeConfAndParam(time, "Size", mCSize);
    SyncronizeConfAndParam(time, "Distance", mCDistance);
}
