// (setq-c-offset 4)
// "clsStereoProp.h"

#include "clsStereoProp.h"

#include "../../MYGL/clsPosition.h"
#include "../../PARAMETER/clsParameter.h"
#include "../../PARAMETER/clsParamGroup.h"
#include "../../TIMELINE/clsTimeCtrl.h"

#include "../../COMMON/CommonDefinitions.h"

#define CLS StereoProp_cls

CLS::
StereoProp_cls(wxString strName, wxString strParentClassName) : Attribute_cls("StereoProp_cls", "StereoProp") {
    
    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strName;
    strConfListLabel << wxT("::StereoProp");

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";

    mCConfList.SetTag(strConfListLabel);

    mCConfList.Add(&mCStereoType, "StereoType=", "0", strUpdate);
    mCConfList.Add(&mCEyeType, "EyeType=", "0", strUpdate);
    mCConfList.Add(&mCParallax, "Parallax=", "0.05", strUpdate);
    mCConfList.Add(&mCToeInAngle, "ToeAngle=", "0.5", strUpdate);

    AddOneDimParam("Parallax", "P", 0.05);
};

void CLS::
StereoRotate(PositionAndDirection_cls &CCam, int stereo, double nt){
    if (mCEyeType.GetBodyINT() == _STEREO_LFT) stereo = 0;
    if (mCEyeType.GetBodyINT() == _STEREO_RGT) stereo = 1;

    //double nt = GetPCTimeCtrl()->GetCurrentNT();
    double parallax = GetPCParamGroup("Parallax")->GetVInterpolatedValue(nt)[0];
    
    switch(stereo){
    case 0:
	CCam.TranslateInWC(parallax, 0); break;
    case 1:
	CCam.TranslateInWC(-parallax, 0); break;
    }
    
    if (mCStereoType.GetBodyINT() == _STEREO_TOEIN)
	switch(stereo){
	case 0:
	    CCam.RotateYaw(-mCToeInAngle.GetBodyDOUBLE()); break;
	case 1:
	    CCam.RotateYaw(mCToeInAngle.GetBodyDOUBLE()); break;
	    
	}
    
}

int CLS::
GetStereoNum(){
    if (mCStereoType.GetBodyINT() == _STEREO_NONE) return 1;
    switch(mCEyeType.GetBodyINT()){
    case _STEREO_NONE: return 1;
    case _STEREO_LFT: case _STEREO_RGT: return 1;
    case _STEREO_SCANLINE: case _STEREO_SIDEBYSIDE: return 2;
    case _STEREO_RED_CYAN: case _STEREO_YELLOW_BLUE: return 2;
    case _STEREO_QUADBUFFER: return 2;
    }
    return 1;
}

float CLS::
GetStereoFocusDistance(double nt){
    if (mCStereoType.GetBodyINT() != 2) return 1e7;
    if (mCToeInAngle.GetBodyDOUBLE() == 0) return 1e7;
    
    double parallax = GetPCParamGroup("Parallax")->GetVInterpolatedValue(nt)[0];
    return parallax / atan(mCToeInAngle.GetBodyDOUBLE()/360*3.14159);
}

int CLS::
GetEyeType(){
    int eyeType = mCEyeType.GetBodyINT();
    if (mCStereoType.GetBodyINT() == 0) eyeType = 0;
    return eyeType;
}

void CLS::
SyncronizeParam(wxString strObj){
    SyncronizeParamAndConf(strObj, "Parallax", mCParallax);
}

void CLS::
SyncronizeConf(double time){
    SyncronizeConfAndParam(time, "Parallax", mCParallax);
}
