// (setq-c-offset 4)
// "clsAttribute.h"

#include "clsAttribute.h"
#include "clsMaterial.h"
#include "clsMaterialLight.h"
#include "DATAVIEWER/clsColorCurve.h"
#include "DATAVIEWER/clsMaterialParticle.h"
#include "DATAVIEWER/clsMaterialStreamline.h"
#include "PROP/clsCameraProp.h"
#include "PROP/clsFogProp.h"
#include "PROP/clsFocalBlurProp.h"
#include "PROP/clsIsosurfProp.h"
#include "PROP/clsStereoProp.h"
#include "PROP/clsStreamlineProp.h"
#include "PROP/clsTextProp.h"

#include "../MISC/clsMisc.h"
#include "../FILTER/clsBlur.h"
#include "../FILTER/clsFocalBlur.h"
#include "../FILTER/clsExposure.h"
#include "../FILTER/clsGlare.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../FILTER/clsRemoveAlpha.h"
#include "../FILTER/clsFilterOverlay.h"
#include "../FILTER/clsFilterLegend.h"
#include "../FILTER/clsSSAO.h"
#include "../FILTER/clsFilterDepth.h"
#include "../MODIFIER/clsTracking.h"
#include "../MODIFIER/clsArrayCopy.h"
#include "../MODIFIER/clsModifyApparentSize.h"
#include "../MODIFIER/clsModifyParticleSize.h"
#include "../MODIFIER/clsNestedVolume.h"
#include "../MODIFIER/clsTrimming.h"
#include "../MODIFIER/clsTexSpaceMod.h"
#include "../MODIFIER/clsSequenceRemap.h"
#include "../MODIFIER/clsParticleBoolean.h"
#include "../MODIFIER/clsParticleExclude.h"
#include "../MODIFIER/clsParticleFiltering.h"
#include "../MODIFIER/clsParticleTweak.h"
#include "../MODIFIER/clsParticleDownsize.h"
#include "../MODIFIER/clsParticleMerge.h"
#include "../MODIFIER/clsStreamFiltering.h"
#include "../MODIFIER/clsStreamSetAlpha.h"
#include "../MODIFIER/clsStreamTransform.h"
#include "../MODIFIER/clsVolumeBoolean.h"
#include "../MODIFIER/clsVolumeCoord.h"
#include "../MODIFIER/clsVolumeDownsize.h"
#include "../MODIFIER/clsVolumeMask.h"

#include "../PARAMETER/clsParamGroup.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS AttributeClone_cls

CLS::
~AttributeClone_cls(){
    mCConfList.DeleteAllItems();
}

bool CLS::
IsParamGroupTimeDepend(){
    uint numParamGroup = mVCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	if (mVCParamGroup[i].IsTimeDepend() == TRUE) return TRUE;
    return FALSE;
}

wxString CLS::
GetStrProcCounter(){
    wxString str;
    str << mCConfList.GetStrProcCounter();
    uint numParamGroup = mVCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	str << mVCParamGroup[i].GetStrProcCounter();
    return str;
}

wxString CLS::
GetStrAttributeName(){
    return mCstrAttributeName.GetBodySTR();
}

bool CLS::
IsAttributeSameAs(wxString strAttribName){
    return GetStrAttributeName().IsSameAs(strAttribName);
}



bool CLS::
IsAttributeValidAndSameAs(wxString strAttribName){
    return GetStrAttributeName().IsSameAs(strAttribName) &&
	(GetPCItem("Validity")->GetBodyBOOL() == TRUE);
}

ParamGroup_cls* CLS::
GetPCParamGroup(wxString strParamGroupName){
    uint num = mVCParamGroup.size();
    for (uint i = 0; i < num; i++){
	if (mVCParamGroup[i].mCstrName.GetBodySTR().IsSameAs(strParamGroupName))
	    return &mVCParamGroup[i];
    }
    
    /*
    wxString strName = mCstrName.GetBodySTR();
    if (strParamGroupName.IsSameAs(strName) &&
	GetStrAttributeName().IsSameAs("ColorCurve_cls"))
	return (ColorCurve_cls*)this;
    */
    
    return NULL;
}

#define CLS Attribute_cls

CLS::
Attribute_cls(wxString strAttribName, wxString strName){
    mCstrAttributeName.SetItem("AttributeName=", strAttribName);
    mCstrName.SetItem("Name=", strName);
    mCConfList.Add(&mCbShy, "Shy=", "0", "TIMELINE");
}

CLS::
Attribute_cls(Attribute_cls &CSrc){
    mCstrAttributeName = CSrc.mCstrAttributeName;
    mCstrName = CSrc.mCstrName;
    mCbShy = CSrc.mCbShy;
}

CLS::
~Attribute_cls(){
    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++) delete mVpCParamGroup[i];
}

void CLS::
Delete(Attribute_cls *pCAttribute){
    wxString strAttrName = pCAttribute->GetStrAttributeName();
    if (strAttrName.IsSameAs("Tracking_cls"))
	delete (Tracking_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ArrayCopy_cls"))
	delete (ArrayCopy_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ModifyApparentSize_cls"))
	delete (ModifyApparentSize_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ModifyParticleSize_cls"))
	delete (ModifyParticleSize_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("NestedVolume_cls"))
	delete (NestedVolume_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("Trimming_cls"))
	delete (Trimming_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("TexSpaceMod_cls"))
	delete (TexSpaceMod_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("SequenceRemap_cls"))
	delete (SequenceRemap_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleBoolean_cls"))
	delete (ParticleBoolean_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleExclude_cls"))
	delete (ParticleExclude_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleFiltering_cls"))
	delete (ParticleFiltering_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleTweak_cls"))
	delete (ParticleTweak_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleDownsizing_cls"))
	delete (ParticleDownsizing_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleMerge_cls"))
	delete (ParticleMerge_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("StreamFiltering_cls"))
	delete (StreamFiltering_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("StreamSetAlpha_cls"))
	delete (StreamSetAlpha_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("StreamTransform_cls"))
	delete (StreamTransform_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("VolumeBoolean_cls"))
	delete (VolumeBoolean_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("VolumeDownsize_cls"))
	delete (VolumeDownsize_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("VolumeCoord_cls"))
	delete (VolumeCoord_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("VolumeMask_cls"))
	delete (VolumeMask_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("Material_cls"))
	delete (Material_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("MaterialLight_cls"))
	delete (MaterialLight_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("CameraProp_cls"))
	delete (CameraProp_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("FogProp_cls"))
	delete (FogProp_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("FocalBlurProp_cls"))
	delete (FocalBlurProp_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("IsosurfProp_cls"))
	delete (IsosurfProp_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("StereoProp_cls"))
	delete (StereoProp_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("StreamlineProp_cls"))
	delete (StreamlineProp_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("TextProp_cls"))
	delete (TextProp_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("MaterialParticle_cls"))
	delete (MaterialParticle_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("MaterialStreamline_cls"))
	delete (MaterialStreamline_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ColorCurve_cls"))
	delete (ColorCurve_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("Exposure_cls"))
	delete (Exposure_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("RemoveAlpha_cls"))
	delete (RemoveAlpha_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("Blur_cls"))
	delete (Blur_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("FocalBlur_cls"))
	delete (FocalBlur_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("Glare_cls"))
	delete (Glare_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("Overlay_cls"))
	delete (Overlay_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("Legend_cls"))
	delete (Legend_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("SSAO_cls"))
	delete (SSAO_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("FilterDepth_cls"))
	delete (FilterDepth_cls*)pCAttribute;
    else 
	delete pCAttribute;
}

void CLS::
DeleteBuffer(Attribute_cls *pCAttribute){
    wxString strAttrName = pCAttribute->GetStrAttributeName();
    if (strAttrName.IsSameAs("Tracking_cls"))
	delete (TrackingBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ArrayCopy_cls"))
	delete (ArrayCopyBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ModifyApparentSize_cls"))
	delete (ModifyApparentSizeBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ModifyParticleSize_cls"))
	delete (ModifyParticleSizeBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("NestedVolume_cls"))
	delete (NestedVolumeBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("Trimming_cls"))
	delete (TrimmingBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("TexSpaceMod_cls"))
	delete (TexSpaceModBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("SequenceRemap_cls"))
	delete (SequenceRemapBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleBoolean_cls"))
	delete (ParticleBooleanBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleExclude_cls"))
	delete (ParticleExcludeBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleFiltering_cls"))
	delete (ParticleFilteringBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleTweak_cls"))
	delete (ParticleTweakBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleDownsizing_cls"))
	delete (ParticleDownsizingBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("ParticleMerge_cls"))
	delete (ParticleMergeBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("StreamFiltering_cls"))
	delete (StreamFilteringBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("StreamSetAlpha_cls"))
	delete (StreamSetAlphaBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("StreamTransform_cls"))
	delete (StreamTransformBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("VolumeBoolean_cls"))
	delete (VolumeBooleanBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("VolumeDownsize_cls"))
	delete (VolumeDownsizeBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("VolumeCoord_cls"))
	delete (VolumeCoordBuffer_cls*)pCAttribute;
    else if (strAttrName.IsSameAs("VolumeMask_cls"))
	delete (VolumeMaskBuffer_cls*)pCAttribute;
    else
	delete pCAttribute;
}

void CLS::
SyncBufferFromSrc(){
    wxString strAttribName = mCstrAttributeName.GetBodySTR();
    if (strAttribName.IsSameAs("Tracking_cls"))
	((TrackingBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ArrayCopy_cls"))
	((ArrayCopyBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ModifyApparentSize_cls"))
	((ModifyApparentSizeBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ModifyParticleSize_cls"))
	((ModifyParticleSizeBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("NestedVolume_cls"))
	((NestedVolumeBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("Trimming_cls"))
	((TrimmingBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("TexSpaceMod_cls"))
	((TexSpaceModBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("SequenceRemap_cls"))
	((SequenceRemapBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("VolumeMask_cls"))
	((VolumeMaskBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("VolumeCoord_cls"))
	((VolumeCoordBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("VolumeBoolean_cls"))
	((VolumeBooleanBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("VolumeDownsize_cls"))
	((VolumeDownsizeBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ParticleBoolean_cls"))
	((ParticleBooleanBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ParticleExclude_cls"))
	((ParticleExcludeBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ParticleFiltering_cls"))
	((ParticleFilteringBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ParticleTweak_cls"))
	((ParticleTweakBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ParticleDownsizing_cls"))
	((ParticleDownsizingBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("ParticleMerge_cls"))
	((ParticleMergeBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("StreamFiltering_cls"))
	((StreamFilteringBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("StreamSetAlpha_cls"))
	((StreamSetAlphaBuffer_cls*)this)->SyncFromSrc();
    if (strAttribName.IsSameAs("StreamTransform_cls"))
	    ((StreamTransformBuffer_cls*)this)->SyncFromSrc();
}

void CLS::
SyncSrcFromBuffer(SceneObj_cls *pCObj){
    wxString strAttribName = mCstrAttributeName.GetBodySTR();
    if (strAttribName.IsSameAs("Tracking_cls"))
	((TrackingBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ArrayCopy_cls"))
	((ArrayCopyBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ModifyApparentSize_cls"))
	((ModifyApparentSizeBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ModifyParticleSize_cls"))
	((ModifyParticleSizeBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("NestedVolume_cls"))
	((NestedVolumeBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("Trimming_cls"))
	((TrimmingBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("TexSpaceMod_cls"))
	((TexSpaceModBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("SequenceRemap_cls"))
	((SequenceRemapBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("VolumeMask_cls"))
	((VolumeMaskBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("VolumeCoord_cls"))
	((VolumeCoordBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("VolumeBoolean_cls"))
	((VolumeBooleanBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("VolumeDownsize_cls"))
	((VolumeDownsizeBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ParticleBoolean_cls"))
	((ParticleBooleanBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ParticleExclude_cls"))
	((ParticleExcludeBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ParticleFiltering_cls"))
	((ParticleFilteringBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ParticleTweak_cls"))
	((ParticleTweakBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ParticleDownsizing_cls"))
	((ParticleDownsizingBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("ParticleMerge_cls"))
	((ParticleMergeBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("StreamFiltering_cls"))
	((StreamFilteringBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("StreamSetAlpha_cls"))
	((StreamSetAlphaBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
    if (strAttribName.IsSameAs("StreamTransform_cls"))
	((StreamTransformBuffer_cls*)this)->ForceSrcToSyncBuffer(pCObj);
}

void CLS::
SetConfListTagForSceneObjAttribute(wxString strName, wxString strObjName){
    wxString strConfListLabel = "SceneObj::"; strConfListLabel << strObjName << "::" << strName;
    mCConfList.SetTag(strConfListLabel);
}

void CLS::
Clone(AttributeClone_cls &CClone){

    CClone.CloneConfListFrom(*this);
    
    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++){
	CClone.mVCParamGroup.push_back(*mVpCParamGroup[i]);
	CClone.mVCParamGroup[CClone.mVCParamGroup.size()-1].mstrExpressionDefined.clear();
    }

    if (mCstrAttributeName.GetBodySTR().IsSameAs("ColorCurve_cls"))
	CClone.mVCParamGroup.push_back(*(ColorCurve_cls*)this);

    CClone.mCstrAttributeName = mCstrAttributeName;
    CClone.mCstrName = mCstrName;
}

bool CLS::
IsParamGroupTimeDepend(){
    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	if (mVpCParamGroup[i]->IsTimeDepend() == TRUE) return TRUE;
    return FALSE;
}

void CLS::
GetVstrParamTag(wxString strParent, std::vector < wxString > &VstrTag){
    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++){
	uint numParam = mVpCParamGroup[i]->mVCParam.size();
	for (uint j = 0; j < numParam; j++)
	    VstrTag.push_back(strParent + "::" + GetStrName() + "::" + mVpCParamGroup[i]->GetStrName() + "::" + mVpCParamGroup[i]->mVCParam[j].GetStrName());
    }
}

ParamGroup_cls* CLS::
GetPCParamGroup(wxString strParamGroupName){
    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++){
	if (mVpCParamGroup[i]->mCstrName.GetBodySTR().IsSameAs(strParamGroupName))
	    return mVpCParamGroup[i];
    }

    wxString strName = mCstrName.GetBodySTR();
    if ((strParamGroupName.IsSameAs(strName) /*||
					       strParamGroupName.IsSameAs("ColorCurve")*/) &&
	GetStrAttributeName().IsSameAs("ColorCurve_cls"))
	return (ColorCurve_cls*)this;

    return NULL;
}

/*
wxString CLS::
GetStrProcCounter(){
    wxString str;
    str << mCConfList.GetStrProcCounter();
    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	str << mVpCParamGroup[i]->GetStrProcCounter();
    return str;
}
*/

wxString CLS::
GetStrAttributeName(){ return mCstrAttributeName.mstrBody; }

bool CLS::
IsAttributeSameAs(wxString strAttribName){
    return GetStrAttributeName().IsSameAs(strAttribName);
}

SceneObj_cls* CLS::
GetPCTargetSceneObj(){
    return GetPCSceneObjCtrl()->GetPCFromName(GetPCItem("Target=")->GetBodySTR());
}

bool CLS::
IsAttributeValidAndSameAs(wxString strAttribName){
    return GetStrAttributeName().IsSameAs(strAttribName) &&
	(GetPCItem("Validity")->GetBodyBOOL() == TRUE);
}

/*
wxString CLS::
GetStrName() { return mCstrName.mstrBody; }
*/

ParamGroup_cls* CLS::
GetPCNewParamGroup(wxString strParamGroupName, 
		   wxString str1, wxString str2, wxString str3){
    wxArrayString mAstrItem;
    mAstrItem.push_back(str1);
    if (str2.IsEmpty() == FALSE) mAstrItem.push_back(str2);
    if (str3.IsEmpty() == FALSE) mAstrItem.push_back(str3);
    return new ParamGroup_cls(strParamGroupName, mAstrItem);
}

void CLS::
InitializeAndPushParam(ParamGroup_cls *pCParamGroup,
		       double v0, double v1, double v2){
    pCParamGroup->SetInitValue(v0, v1, v2);
    mVpCParamGroup.push_back(pCParamGroup);
}

void CLS::
AddOneDimParam(wxString strGroupName, wxString strParamName, double ini){
    ParamGroup_cls *pCParamGroup = GetPCNewParamGroup(strGroupName, strParamName);
    InitializeAndPushParam(pCParamGroup, ini);
}

void CLS::
SyncronizeParamForFilter(wxString strParamGroup, std::vector < float > VValueNew){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    float time = GetPCTimeCtrl()->GetCurrentNT();
    int dim = VValueNew.size();
    ASSERT(dim == pCParamGroup->mVCParam.size());

    for (int i = 0; i < dim; i++){
	double newValue = VValueNew[i];
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	if (Misc_cls::IsNearlySame(CParam.GetValue(time), newValue, 5e-7) == FALSE){
	    wxString strTag = GetPCFilterCtrl()->GetStrSelf() << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam.GetStrName();
	    CParam.SetValueAtCommand(time, newValue, strTag);
	}
    }
}

void CLS::
SyncronizeParamForModifier(wxString strParamGroup, std::vector < float > VValueNew, SceneObj_cls *pCObj){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    float time = GetPCTimeCtrl()->GetCurrentNT();
    int dim = VValueNew.size();
    ASSERT(dim == pCParamGroup->mVCParam.size());

    for (int i = 0; i < dim; i++){
	double newValue = VValueNew[i];
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	if (Misc_cls::IsNearlySame(CParam.GetValue(time), newValue, 5e-7) == FALSE){
	    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << pCObj->GetStrName() << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam.GetStrName();
	    CParam.SetValueAtCommand(time, newValue, strTag);
	}
    }
}

void CLS::
DeselectKeyFrame(wxString strException){
    wxString strName = PopStrItemFromStrLine(strException, wxT("::"));
    if (GetStrName().IsSameAs(strName) == FALSE)
	strException.Clear();

    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++){
	mVpCParamGroup[i]->DeselectKey(strException);
    }

    // J[J[úAAttribute  ParamGroupȂ̂ŁA
    // Oɑׂ͎̖OG
    if (GetStrAttributeName().IsSameAs("ColorCurve_cls"))
	((ColorCurve_cls*)this)->DeselectKey(strName);
}

bool CLS::
IsSelectedTLKeyExist(){
    bool bExist = FALSE;

    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++)
	bExist |= mVpCParamGroup[i]->IsSelectedKeyExist();

    if (GetStrAttributeName().IsSameAs("ColorCurve_cls"))
	((ColorCurve_cls*)this)->ParamGroup_cls::IsSelectedKeyExist();

    return bExist;
}

void CLS::
DeleteSelectedTLKeyCommand(wxString strTag){
    strTag << "::" << GetStrName();
    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++)
	mVpCParamGroup[i]->DeleteSelectedTLKeyCommand(strTag);

    if (GetStrAttributeName().IsSameAs("ColorCurve_cls"))
	((ColorCurve_cls*)this)->ParamGroup_cls::DeleteSelectedTLKeyCommand(strTag);
}

void CLS::
SetSelectedTLKeyTypeCommand(wxString strTag, int keyType, double ease){
    strTag << "::" << GetStrName();
    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++)
	mVpCParamGroup[i]->SetSelectedTLKeyTypeCommand(strTag, keyType, ease);
    
    if (GetStrAttributeName().IsSameAs("ColorCurve_cls"))
	((ColorCurve_cls*)this)->ParamGroup_cls::SetSelectedTLKeyTypeCommand(strTag, keyType, ease);
}

void CLS::
MoveSelectedTLKeyCommand(wxString strTag, double dNX, double dY){
    strTag << "::" << GetStrName();
    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++){
	mVpCParamGroup[i]->MoveSelectedTLKeyCommand(dNX, dY, strTag);
    }

    if (GetStrAttributeName().IsSameAs("ColorCurve_cls"))
	((ColorCurve_cls*)this)->ParamGroup_cls::MoveSelectedTLKeyCommand(dNX, dY, strTag);
}

void CLS::
RescaleParametersCommand(wxString strTag, double scale){
    strTag << "::" << GetStrName();
    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++){
	mVpCParamGroup[i]->RescaleParametersCommand(scale, strTag);
    }
}

void CLS::
SyncronizeParamAndConf3(wxString strObj, wxString strParamGroup, ConfItemFLOATARRAY_cls &CConf){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    int keyType;
    double easeFactor;
    GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);

    for (int i = 0; i < 3; i++){
	float newValue = CConf.GetBodyFLOATARRAY()[i];
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	float oldValue = CParam.GetValue(time);
	if (oldValue != newValue){
	    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam.GetStrName();
	    CParam.SetValueAtCommand(time, newValue, strTag,
				     keyType, easeFactor);
	}
    }
}

void CLS::
SyncronizeParamAndConf2(wxString strObj, wxString strParamGroup, ConfItemFLOAT_cls &CConf0, ConfItemFLOAT_cls&CConf1){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    int keyType;
    double easeFactor;
    GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);

    float newValue = CConf0.GetBodyDOUBLE();
    Parameter_cls &CParam0 = pCParamGroup->mVCParam[0];
    float oldValue = CParam0.GetValue(time);

    if (oldValue != newValue){
	wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam0.GetStrName();
	CParam0.SetValueAtCommand(time, newValue, strTag,
				  keyType, easeFactor);
    }

    newValue = CConf1.GetBodyDOUBLE();
    Parameter_cls &CParam1 = pCParamGroup->mVCParam[1];
    oldValue = CParam1.GetValue(time);

    if (oldValue != newValue){
	wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam1.GetStrName();
	CParam1.SetValueAtCommand(time, newValue, strTag,
				  keyType, easeFactor);
    }
}

void CLS::
SyncronizeParamAndConf(wxString strObj, wxString strParamGroup, ConfItemFLOAT_cls &CConf){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    double time = GetPCTimeCtrl()->GetCurrentNT();

    int keyType;
    double easeFactor;
    GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);

    float newValue = CConf.GetBodyDOUBLE();
    Parameter_cls &CParam = pCParamGroup->mVCParam[0];
    float oldValue = CParam.GetValue(time);

    if (oldValue != newValue){
	wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::" << GetStrName() << "::" << pCParamGroup->GetStrName() << "::" << CParam.GetStrName();
	CParam.SetValueAtCommand(time, newValue, strTag,
				 keyType, easeFactor);
    }
}

void CLS::
SyncronizeConfAndParam(double time, wxString strParamGroup, ConfItemFLOAT_cls &CConf){
    CConf.SetBodyFromData(GetPCParamGroup(strParamGroup)->GetStrValues(time, FALSE));
}

void CLS::
SyncronizeConfAndParam2(double time, wxString strParamGroup, ConfItemFLOAT_cls &CConf0, ConfItemFLOAT_cls &CConf1){
    CConf0.SetBodyFromData(GetPCParamGroup(strParamGroup)->GetVInterpolatedValue(time, FALSE)[0]);
    CConf1.SetBodyFromData(GetPCParamGroup(strParamGroup)->GetVInterpolatedValue(time, FALSE)[1]);
}

bool CLS::	
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){

    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE) 
	return TRUE;

    if (strFirstItem.IsSameAs("ReadAttribute") == FALSE) return FALSE;

    wxString strAttribName = PopStrItemFromStrLine(strLine, "::");
    wxString strName = PopStrItemFromStrLine(strLine, "::");
        
    if (strAttribName.IsSameAs(GetStrAttributeName()) != TRUE ||
	strName.IsSameAs(GetStrName()) != TRUE) return FALSE;
	
    ASSERT(strLine.IsSameAs("Begin"));
    std::wstring wstrLine;
    while (std::getline(istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");

	if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE) 
	    continue;

	if (strFirstItem.IsSameAs("ReadParamGroup")){
	    wxString paramGroupName = PopStrItemFromStrLine(strLine, "::");
	    ParamGroup_cls *pCParamGroup = GetPCParamGroup(paramGroupName);
	    if (pCParamGroup != NULL){
	    
		std::wstring wstrLine;
		while (std::getline(istream, wstrLine)){
		    wxString strLine = wstrLine.c_str();
		wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
		pCParamGroup->ReadStrLine(istream, strFirstItem, strLine);
		
		if (strFirstItem.IsSameAs("ReadParamGroup")){
		    ASSERT(PopStrItemFromStrLine(strLine, wxT("::")).IsSameAs(wxT("End")));
		    break;
		}
		}
	    }
	}

	if (strFirstItem.IsSameAs("ReadAttribute")){
	    ASSERT(strLine.IsSameAs("End"));
	    break;
	}
    }
    
    return TRUE;
}

void CLS::
Serialize(wxString& strOut){
    strOut << "ReadAttribute::" << GetStrAttributeName() << "::" << GetStrName() << "::Begin" << ENDL;
    
    ConfListContainer_cls::Serialize(strOut);

    int numParamGroup = mVpCParamGroup.size();
    for (int i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->Serialize(strOut);
    
    strOut << "ReadAttribute::End" << ENDL;
}
