// (setq-c-offset 4)
// "clsPrimitive.cpp"

#include "clsCamera.h"
#include "PROP/clsCameraProp.h"
#include "PROP/clsFocalBlurProp.h"
#include "PROP/clsStereoProp.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../GLOBJ/clsVBOCameraBox.h"
#include "../GRAPHIC/clsImageOutput.h"
#include "../MYGL/clsGLExtention.h"
#include "../MYGL/clsWideCamera.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"
#include "../COMMON/CommonDefinitionsBuf.h"

#define CLS Camera_cls

#include "../../wxWidgetEnum.h"

CLS::
Camera_cls(ParentObj_cls *pCParentObj, wxString strName) : 
    SceneObj_cls(pCParentObj)
{
    int type = 0;
    wxString strTypeNum; strTypeNum << type;

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";
    mCConfList.Add(&mCType, "Type=", strTypeNum, strUpdate);
    
    mstrClassName = "Camera_cls";
    Initialize(strName);
    
    mpCCameraProp = new CameraProp_cls(strName, "CameraProp_cls");
    mpCStereoProp = new StereoProp_cls(strName, "StereoProp_cls");
    mpCFocalBlurProp = new FocalBlurProp_cls(strName, "FocalBlurProp_cls",
					     mpCStereoProp);
    
    mVpCAttribute.push_back(mpCCameraProp);
    mVpCAttribute.push_back(mpCFocalBlurProp);
    mVpCAttribute.push_back(mpCStereoProp);

    ParamGroup_cls *pCPos = GetPCParamGroup("Position");
    pCPos->mVCParam[0].mVCKeys[0].mValue = -5;
    pCPos->mVCParam[1].mVCKeys[0].mValue = -5;
    pCPos->mVCParam[2].mVCKeys[0].mValue = +5;

    ParamGroup_cls *pCRot = GetPCParamGroup("RotationEuler");
    pCRot->mVCParam[1].mVCKeys[0].mValue = 35;
    pCRot->mVCParam[2].mVCKeys[0].mValue = 45;

    mbSkipDraw = FALSE;
}

CLS::
~Camera_cls(){
}

wxSize CLS::
GetSize(){
    return wxSize(mpCCameraProp->mCWidthOverwrite.GetBodyINT(),
		  mpCCameraProp->mCHeightOverwrite.GetBodyINT());
}

float CLS::
GetAngleL(){
    double aspect = GetAspectRatioWonH();
    double angleV = mpCCameraProp->mCAngleV.GetBodyDOUBLE();
    if (aspect <= 1) return angleV;
    return atan(tan(angleV*3.14159/180 /2) * aspect) * 2 * 180/3.14159;
}

float CLS::
GetSizeOrthoL(){
    double aspect = GetAspectRatioWonH();
    double sizeOthro = mpCCameraProp->mCSizeOrthoV.GetBodyDOUBLE();
    if (aspect <= 1) return sizeOthro;
    return sizeOthro * aspect;
}

std::vector < double > CLS::
GetVMinMax(double NT){
    return mpCCameraProp->GetPCParamGroup("Clip Min/Max")->GetVInterpolatedValue(NT);
}

float CLS::
GetAspectRatioWonH(){
    if (mpCCameraProp->mCbOverwriteGeometry.GetBodyBOOL() == FALSE)
	return GCClassProvider.GetPCImageOutput()->GetAspectRatioWonH();
    return mpCCameraProp->GetAspectRatioWonH();
}

void CLS::
GLDraw(bool bDrawLightCam, int status, bool bHittest){
    if (bDrawLightCam == FALSE) return;
    if (mbSkipDraw == TRUE) return;

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    ColorProvider_cls::SetGLColor("GuideObject");
    if (status == 1) ColorProvider_cls::SetGLColor("Selected");
    if (status == 2) ColorProvider_cls::SetGLColor("Active");

    glScaled(0.2, 0.2, 0.2);
    GCClassProvider.GetPCVBOCameraBox()->Draw();
    glScaled(5, 5, 5);

    int type = mCType.GetBodyINT();
    switch(type){
    case _WC_PERSPECTIVE:
	GLDrawPerspective(); break;
    case _WC_ORTHOGONAL:
	GLDrawPerspective(); break;
    case _WC_FISHEYE3:
    case _WC_FISHEYE4:
	GLDrawFisheye(); break;
    }
}

void CLS::
GLDrawPerspective(){
    float time = GetPCTimeCtrl()->GetCurrentNT();
    float angle = mpCCameraProp->GetPCParamGroup("AngleV")->mVCParam[0].GetValue(time);
    if (angle > 120) angle = 120;

    float angleH = tan(angle*3.14159/180 / 2);
    float angleW = angleH;

    angleW *= GetAspectRatioWonH();

    glBegin(GL_LINES);
    glVertex3d(0, 0, 0); glVertex3d(1, -angleW, -angleH);
    glVertex3d(0, 0, 0); glVertex3d(1, +angleW, -angleH);
    glVertex3d(0, 0, 0); glVertex3d(1, +angleW, +angleH);
    glVertex3d(0, 0, 0); glVertex3d(1, -angleW, +angleH);
    
    glVertex3d(1, -0.2, +angleH);
    glVertex3d(1, 0,    +angleH+0.1);
    glVertex3d(1, 0,    +angleH+0.1);
    glVertex3d(1, +0.2, +angleH);

    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3d(1, -angleW, -angleH);
    glVertex3d(1, +angleW, -angleH);
    glVertex3d(1, +angleW, +angleH);
    glVertex3d(1, -angleW, +angleH);
    glEnd();
}

void CLS::
GLDrawOrthogonal(){
}

void CLS::
GLDrawFisheye(){
    float time = GetPCTimeCtrl()->GetCurrentNT();
    float angle = mpCCameraProp->GetPCParamGroup("AngleV")->mVCParam[0].GetValue(time);

    double rad = angle * 3.14159 / 180 / 2;
    double x = cos(rad);
    double z = sin(rad);

    glBegin(GL_LINE_LOOP);
    for (int phi = 0; phi < 360; phi+=10){
	double radphi = phi * 3.14159 / 180;
	glVertex3d(x, z * cos(radphi), z * sin(radphi));
    }
    glEnd();
    
    glBegin(GL_LINE_STRIP);
    for (int phi = -90; phi <= 90; phi+= 10){
	double radphi = phi * 3.14159 / 180 * angle / 180;
	glVertex3d(cos(radphi), sin(radphi), 0); 
    }
    glEnd();

    glBegin(GL_LINE_STRIP);
    for (int phi = -90; phi <= 90; phi+= 10){
	double radphi = phi * 3.14159 / 180 * angle / 180;
	glVertex3d(cos(radphi), 0, sin(radphi)); 
    }
    glEnd();
    
    if (angle < 90){
	glBegin(GL_LINES);
	glVertex3d(0, 0, 0); glVertex3d(x, -z, 0);
	glVertex3d(0, 0, 0); glVertex3d(x, +z, 0);
	glVertex3d(0, 0, 0); glVertex3d(x, 0, +z);
	glVertex3d(0, 0, 0); glVertex3d(x, 0, -z);
	glEnd();
    }
    else{
	glBegin(GL_LINES);
	glVertex3d(0, 0, 0); glVertex3d(1, 0, 0);
	glEnd();
    }
}

std::vector < float > CLS::
GetVPasspartoutInfo(){
    std::vector < float > VInfo;

    if (mCType.GetBodyINT() != _WC_PERSPECTIVE){
	VInfo.push_back(-1); VInfo.push_back(-1);
	return VInfo;
    }

    float time = GetPCTimeCtrl()->GetCurrentNT();
    float angle = mpCCameraProp->GetPCParamGroup("AngleV")->GetVInterpolatedValue(time)[0];
    if (angle > 120) angle = 120;
    float angleH = tan(angle*3.14159/180 / 2);
    float angleW = angleH;
    angleW *= GetAspectRatioWonH();

    VInfo.push_back(angleW); VInfo.push_back(angleH);
    return VInfo;
}

void CLS::
GLDrawPassepartoutMain(float angleW, float angleH, float angleCanvas){
    
    int status = GetPCSceneObjCtrl()->GetStatus(this);
    glColor4f(0.2f, 0.7f, 0.2f, 1.0f);
    if (status == 1) ColorProvider_cls::SetGLColor("Selected");
    if (status == 2) ColorProvider_cls::SetGLColor("Active");

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);

    GetPCGLExtentions()->
	BlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
			  GL_ONE_MINUS_DST_ALPHA, GL_ONE);

    glPushMatrix();
    glLoadIdentity();
    
    glBegin(GL_LINE_LOOP);
    glVertex3d(-angleW, -angleH, -1);
    glVertex3d(+angleW, -angleH, -1);
    glVertex3d(+angleW, +angleH, -1);
    glVertex3d(-angleW, +angleH, -1);
    glEnd();

    glColor4f(0, 0, 0, 0.5);

    glBegin(GL_QUADS);
    if (angleCanvas > angleH){
	glVertex3d(-angleW, -angleCanvas, -1);
	glVertex3d(+angleW, -angleCanvas, -1);
	glVertex3d(+angleW, -angleH, -1);
	glVertex3d(-angleW, -angleH, -1);

	glVertex3d(-angleW, +angleH, -1);
	glVertex3d(+angleW, +angleH, -1);
	glVertex3d(+angleW, +angleCanvas, -1);
	glVertex3d(-angleW, +angleCanvas, -1);
    }

    if (angleCanvas > angleW){
	glVertex3d(-angleCanvas, -angleCanvas, -1);
	glVertex3d(-angleW, -angleCanvas, -1);
	glVertex3d(-angleW, +angleCanvas, -1);
	glVertex3d(-angleCanvas, +angleCanvas, -1);

	glVertex3d(+angleW, -angleCanvas, -1);
	glVertex3d(+angleCanvas, -angleCanvas, -1);
	glVertex3d(+angleCanvas, +angleCanvas, -1);
	glVertex3d(+angleW, +angleCanvas, -1);
    }
    glEnd();

    glPopMatrix();

    glDisable(GL_BLEND);
}

void CLS::
GLDrawPassepartout(double canvasAngle){

    if (mpCCameraProp->mCbPassepartout.GetBodyBOOL() == FALSE)  return;

    std::vector < float > VPasspartout = GetVPasspartoutInfo();
    float angleW = VPasspartout[0];
    float angleH = VPasspartout[1];

    float angleCanvas = tan(canvasAngle*3.14159/180 / 2);

    GLDrawPassepartoutMain(angleW, angleH, angleCanvas);
}

void CLS::
GLDrawPassepartoutOrtho(double canvasSizeOrtho){
    
    if (mpCCameraProp->mCbPassepartout.GetBodyBOOL() == FALSE)  return;

    int status = GetPCSceneObjCtrl()->GetStatus(this);
    glColor4f(0.2f, 0.7f, 0.2f, 1.0f);
    if (status == 1) ColorProvider_cls::SetGLColor("Selected");
    if (status == 2) ColorProvider_cls::SetGLColor("Active");

    double angleW = mpCCameraProp->mCSizeOrthoV.GetBodyDOUBLE() * GetAspectRatioWonH();
    double angleH = mpCCameraProp->mCSizeOrthoV.GetBodyDOUBLE();

    GLDrawPassepartoutMain(angleW, angleH, canvasSizeOrtho);

    /*
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);

    GetPCGLExtentions()->
	BlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
			  GL_ONE_MINUS_DST_ALPHA, GL_ONE);

    glPushMatrix();
    glLoadIdentity();
    
    glBegin(GL_LINE_LOOP);
    glVertex3d(-angleW, -angleH, -1);
    glVertex3d(+angleW, -angleH, -1);
    glVertex3d(+angleW, +angleH, -1);
    glVertex3d(-angleW, +angleH, -1);
    glEnd();

    glBegin(GL_QUADS);
    
    if (angleCanvas > angleH){
	glVertex3d(-angleW, -angleCanvas, -1);
	glVertex3d(+angleW, -angleCanvas, -1);
	glVertex3d(+angleW, -angleH, -1);
	glVertex3d(-angleW, -angleH, -1);

	glVertex3d(-angleW, +angleH, -1);
	glVertex3d(+angleW, +angleH, -1);
	glVertex3d(+angleW, +angleCanvas, -1);
	glVertex3d(-angleW, +angleCanvas, -1);
    }

    if (angleCanvas > angleW){
	glVertex3d(-angleCanvas, -angleCanvas, -1);
	glVertex3d(-angleW, -angleCanvas, -1);
	glVertex3d(-angleW, +angleCanvas, -1);
	glVertex3d(-angleCanvas, +angleCanvas, -1);

	glVertex3d(+angleW, -angleCanvas, -1);
	glVertex3d(+angleCanvas, -angleCanvas, -1);
	glVertex3d(+angleCanvas, +angleCanvas, -1);
	glVertex3d(+angleW, +angleCanvas, -1);
    }
    glEnd();

    glPopMatrix();

    glDisable(GL_BLEND);
    
    //glPushMatrix();
    //glLoadIdentity();
    */
}

MultiInputPanel_cls* CLS::
GetPCNewPanelCamera(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, _("Camera"));
        
    mIndexAngleV = pC->CombineSizer(2, _("Type/AngleV"));
    pC->SetChoiceList(_CAMERA_MAX, GetPCStringProvider()->GetAstr("CameraType"));
    pC->CombineConfItem(mCType);

    pC->mDragScale = 10;
    pC->SetLimit(1, 360);
    pC->mbItemVolatile = TRUE;
    pC->CombineConfItem(mpCCameraProp->mCAngleV);

    mIndexSizeOrtho = pC->CombineSizer(2, _("Type/SizeOrthoV"));
    pC->SetChoiceList(_CAMERA_MAX, GetPCStringProvider()->GetAstr("CameraType"));
    pC->CombineConfItem(mCType);

    pC->mDragScale = 10;
    pC->SetLimit(1, 360);
    pC->mbItemVolatile = TRUE;
    pC->CombineConfItem(mpCCameraProp->mCSizeOrthoV);
    
    pC->SetChoiceList(_BUFFER_INDEX_MAX-2, GetPCStringProvider()->GetAstr("BufferSize"));
    mIndex1stBuffer = 
	pC->CombineConfItem(mpCCameraProp->mC1stBufferSizeIndex, _("1st Buffer"));

    pC->mDragScale = 10;
    pC->SetLimit(-180, 180);
    pC->mbItemVolatile = TRUE;
    mIndexTilt = pC->CombineConfItem(mpCCameraProp->mCTiltAngle, _("Tilt Angle"));

    mIndexClipMinMax = pC->CombineSizer(2, _("Clip Min/Max"));
    pC->CombineConfItem(mpCCameraProp->mCClipMin);
    pC->CombineConfItem(mpCCameraProp->mCClipMax);

    pC->CombineConfItem(mpCCameraProp->mCbPassepartout, _("Passepartout"));
    pC->CombineConfItem(mpCCameraProp->mCbOverwriteGeometry, _("Overwrite"));
    
    pC->CombineSizer(2, _("width/height"));
    pC->SetLimit(4, 16384);
    pC->CombineConfItem(mpCCameraProp->mCWidthOverwrite);
    pC->SetLimit(4, 16384);
    pC->CombineConfItem(mpCCameraProp->mCHeightOverwrite);

    pC->FinishSetting();
    HideAndShowCamera(pC);

    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelFocalBlur(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, _("Focal Blur"));

    pC->CombineSizer(1, _("Focal Blur"));
    pC->CombineConfItem(mpCFocalBlurProp->mCbUseFocalBlur, _("Ensemble Blur"));

    pC->SetChoiceList(_FOCAL_BLUR_MAX, GetPCStringProvider()->GetAstr("FocalBlurType"));
    mIndexBlurType = pC->CombineConfItem(mpCFocalBlurProp->mCFocusType, _("Type"));
    
    mIndexTarget = pC->CombineSizer(2, _("Target"));
    pC->CombineBitmapPanel(btnDropdown, wxES, 14);
    pC->mVCItem.back().mBitmapTemplate = _BITMAP_BTN_DROPDOWN;
    pC->mVCItem.back().mbUseVCenter = TRUE;
    pC->mVCItem.back().mstrCommandStr = "FocalBlurObj";
    pC->CombineConfItem(mpCFocalBlurProp->mCstrFocusObj);
    
    mIndexBlurSize = pC->CombineSizer(2, _("Size/Distance"));
    pC->CombineConfItem(mpCFocalBlurProp->mCSize);
    pC->CombineConfItem(mpCFocalBlurProp->mCDistance);

    pC->FinishSetting();
    HideAndShowFocalBlur(pC);
    
    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelStereo(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, _("Stereogram"));

    pC->SetChoiceList(3, GetPCStringProvider()->GetAstr("StereoType"));
    pC->CombineConfItem(mpCStereoProp->mCStereoType, _("Camera Setting"));

    pC->SetChoiceList(8, GetPCStringProvider()->GetAstr("EyeType"));
    mIndexEyeType = pC->CombineConfItem(mpCStereoProp->mCEyeType, _("Stereogram Type"));
  
    mIndexParallax = pC->CombineSizer(2, _("Angle/Parallax"));
    pC->CombineConfItem(mpCStereoProp->mCToeInAngle);
    pC->CombineConfItem(mpCStereoProp->mCParallax);
    
    pC->FinishSetting();
    HideAndShowStereo(pC);

    return pC;
}

bool CLS::
HideAndShowCamera(MultiInputPanel_cls *pC){
    bool bChange = FALSE;
    if (mCType.GetBodyINT() == _WC_ORTHOGONAL){
	bChange |= pC->mVCItem[mIndexAngleV].Hide();
	bChange |= pC->mVCItem[mIndexSizeOrtho].Show();
    }
    else{
	bChange |= pC->mVCItem[mIndexAngleV].Show();
	bChange |= pC->mVCItem[mIndexSizeOrtho].Hide();
    }

    if (mCType.GetBodyINT() > 1)
	bChange |= pC->mVCItem[mIndex1stBuffer].Show();
    else
	bChange |= pC->mVCItem[mIndex1stBuffer].Hide();

    return bChange;
}

bool CLS::
HideAndShowStereo(MultiInputPanel_cls *pC){
    bool bChange = FALSE;
    
    if (mpCStereoProp->mCStereoType.GetBodyINT() > 0){
	bChange |= pC->mVCItem[mIndexParallax].Show();
	bChange |= pC->mVCItem[mIndexEyeType].Show();
    }
    else{
	bChange |= pC->mVCItem[mIndexParallax].Hide();
	bChange |= pC->mVCItem[mIndexEyeType].Hide();
    }
	
    return bChange;
}

bool CLS::
HideAndShowFocalBlur(MultiInputPanel_cls *pC){
    bool bChange = FALSE;

    bChange |= pC->mVCItem[mIndexBlurType].Show();
    bChange |= pC->mVCItem[mIndexBlurSize].Show();
    if (mpCFocalBlurProp->mCFocusType.GetBodyINT() == _FOCUS_OBJECT)
	bChange |= pC->mVCItem[mIndexTarget].Show();
    else
	bChange |= pC->mVCItem[mIndexTarget].Hide();

    return bChange;
}
