// (setq-c-offset 4)
// "clsPrimitive.cpp"

#include "clsLightSource.h"
#include "clsMaterialLight.h"

#include "../COMMON/clsHalton.h"
#include "../MYGL/clsGLInfo.h"
#include "../MYGL/clsGLPerspective.h"
#include "../MYGL/clsShadowVolumeCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#define CLS LightSource_cls

CLS::
LightSource_cls(ParentObj_cls *pCParentObj, wxString strType, wxString strName) : 
    SceneObj_cls(pCParentObj)
{
    int type = GetTypeFromStr(strType);
    wxString strTypeNum; strTypeNum << type;

    mSortOrder = 4;

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";
    
    mCConfList.Add(&mCType, "Type=", strTypeNum, strUpdate);
    mCConfList.Add(&mCbValid, "Valid=", "1", strUpdate);
    mCConfList.Add(&mCbXRay, "XRay=", "1", strUpdate);
    mCConfList.Add(&mCbCastShadow, "CastShadow=", "1", strUpdate);
    mCConfList.Add(&mCFadeType, "FadeType=", "0", strUpdate);
    mCConfList.Add(&mCFadeDistance, "FadeDistance=", "1", strUpdate);
    mCConfList.Add(&mCSpotAngle, "SpotAngle=", "35", strUpdate);
    mCConfList.Add(&mCSpotExponent, "SpotExponent=", "0", strUpdate);
    mCConfList.Add(&mCSoftSize, "SoftSize=", "0", strUpdate);
    
    mstrClassName = "LightSource_cls";
    
    mpCMaterial = new MaterialLight_cls(strName, "MaterialLight_cls");

    Initialize(strName);
    
    mVpCAttribute.push_back(mpCMaterial);
}

CLS::
~LightSource_cls(){
}

wxString CLS::
GetStrInitName(){
    return msstrTypeName[mCType.GetBodyINT()]; }

MaterialLight_cls* CLS::
GetPCMaterial(){
    return mpCMaterial; }

wxString CLS::
GetStrInitNameCandidate(int type){
    if (type >= 0 && type < _LIGHT_MAX) return msstrTypeName[type];
    return wxT("InvalidType");
}

int CLS::
GetTypeFromStr(wxString strType){
    for (int i = 0; i < _LIGHT_MAX; i++)
	if (msstrTypeName[i].IsSameAs(strType)) return i;
    return -1;
}

void CLS::
SyncConfListTagLightSource(){
    wxString strObj = mCstrName.GetBodySTR();
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }

    mpCMaterial->SyncConfListTag(strObj);
}

void CLS::
GLSetAttenuation(int &lightCurrent){
    switch(mCFadeType.GetBodyINT()){
    case _LIGHT_FADE_OFF:
	glLightf(lightCurrent, GL_CONSTANT_ATTENUATION, 1);
	glLightf(lightCurrent, GL_LINEAR_ATTENUATION, 0);
	glLightf(lightCurrent, GL_QUADRATIC_ATTENUATION, 0);
	break;
    case _LIGHT_FADE_1: {
	float factor = 1.0 / mCFadeDistance.GetBodyDOUBLE();
	glLightf(lightCurrent, GL_CONSTANT_ATTENUATION, 1);
	glLightf(lightCurrent, GL_LINEAR_ATTENUATION, factor);
	glLightf(lightCurrent, GL_QUADRATIC_ATTENUATION, 0);
    	break; }
    case _LIGHT_FADE_2: {
	float factor = 1.0 / mCFadeDistance.GetBodyDOUBLE() / mCFadeDistance.GetBodyDOUBLE();;
	glLightf(lightCurrent, GL_CONSTANT_ATTENUATION, 1);
	glLightf(lightCurrent, GL_LINEAR_ATTENUATION, 0);
	glLightf(lightCurrent, GL_QUADRATIC_ATTENUATION, factor);
	break; }
    }
}

void CLS::
GLSetSpotEffect(int &lightCurrent){
    if (mCType.GetBodyINT() != _LIGHT_SPOT){
	glLightf(lightCurrent, GL_SPOT_CUTOFF, 180.0);
	return;
    }

    float lightDir[4] = {-1, 0, 0, 0};
    float lightZero[3] = {0, 0, 0};

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < double > VSpot = mpCMaterial->GetPCParamGroup("SpotParam")->GetVInterpolatedValue(NT);
        
    glLightf(lightCurrent, GL_SPOT_CUTOFF, VSpot[0]);
    glLightfv(lightCurrent, GL_SPOT_DIRECTION, lightDir);
    glLightf(lightCurrent, GL_SPOT_EXPONENT, VSpot[1]);
}

void CLS::
SetGLLight(int &lightCurrent, int ensemble){
    glEnable(lightCurrent);

    float softSize = mCSoftSize.GetBodyDOUBLE();
    if (mCType.GetBodyINT() == _LIGHT_PARALLEL)
	if (softSize > 0.5) softSize = 0.5;

    float x = (GetPCHalton()->GetX(ensemble)-0.5) * softSize;
    float y = (GetPCHalton()->GetY(ensemble)-0.5) * softSize;
    float z = (GetPCHalton()->GetZ(ensemble)-0.5) * softSize;
    float w = 1;
    if (mCType.GetBodyINT() == _LIGHT_PARALLEL){
	x = 1;
	w = 0;
    }

    float light0pos[4] = {x, y, z, w};
    glLightfv(lightCurrent, GL_POSITION, light0pos);

    GLSetAttenuation(lightCurrent);
    GLSetSpotEffect(lightCurrent);

    mpCMaterial->SetGLLightSetting(lightCurrent);

    lightCurrent++;
}

void CLS::
GLDrawMain(bool bDrawLightCam, int status, bool bHittest){
    
    if (bDrawLightCam == FALSE) return;
    if (mCbVisible.GetBodyBOOL() == FALSE) return;

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    GLfloat matrixA[16], matrixB[16];
	double scale = GLPerspective_cls::GetScaleForFixSizedObject
	    (matrixA, matrixB);
    /*
    GLfloat matrixA[16];
    glGetFloatv(GL_PROJECTION_MATRIX, matrixA);

    GLfloat matrixB[16];
	glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, matrixB);

    double scale = 1;
    if (matrixA[15] == 1)
	scale = 1.0/matrixA[5] * 15;
    else
	scale = -matrixB[11];
    */
    
    glEnable(GL_LINE_STIPPLE);
    glColor4f(0.7f, 0.7f, 0.4f, 1.0f);
    if (status == 1) ColorProvider_cls::SetGLColor("Selected");
    if (status == 2) ColorProvider_cls::SetGLColor("Active");

    // ⏕
    if (bHittest == FALSE){
	int type = mCType.GetBodyINT();
	if (type == _LIGHT_PARALLEL || type == _LIGHT_SPOT){
	    glLineStipple(1,0xf0f0);
	    glBegin(GL_LINES); 
	    glVertex3d(0, 0, 0);
	    glVertex3d(-10, 0, 0);
	    glEnd();
	}
    }
    
    glDisable(GL_LINE_STIPPLE);

    if (bHittest == FALSE){

	glPushMatrix(); 

	double matrix[16] = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};
		
	for (int j = 0; j < 3; j++){
	    matrix[0+j] = matrixB[0+j];
	    matrix[4+j] = matrixB[4+j];
	    matrix[8+j] = matrixB[8+j];
	}
		
	glMultMatrixd(matrix); 
	glScaled(scale*0.02, scale*0.02, scale*0.02);
	
	glBegin(GL_LINES);
	for (int i = 0; i < 360; i+=10){
	    double rad = 3.14159 * i / 180;
	    glVertex3d(cos(rad), sin(rad), 0);
	}
	glEnd();
	
	glBegin(GL_LINE_STRIP);
	for (int i = 0; i <= 360; i+=10){
	    double rad = 3.14159 * i / 180;
	    glVertex3d(0.7 * cos(rad), 0.7 * sin(rad), 0);
	}
	glEnd();

	glPopMatrix();

    }
}

void CLS::
GLDraw(bool bDrawLightCam, int &lightCurrent, int ensemble, int status, bool bHittest){

    wxString strName = mCstrName.GetBodySTR();

    if (mCbValid.GetBodyBOOL() == TRUE){
	ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
	int lightMax = GetPCGLInfo()->mMaxLightNum;
	if (pCSV->mCbUseShadowVolume.GetBodyBOOL() == TRUE){
	    if (pCSV->mSVStep == 0){
		if (mCbCastShadow.GetBodyBOOL() == FALSE)
		    SetGLLight(lightCurrent, ensemble);
	    }
	    else{
		if (mCbCastShadow.GetBodyBOOL() == TRUE){
		    if (pCSV->mSVStep-1 == lightCurrent - GL_LIGHT0){
			int singleLight = GL_LIGHT0;
			SetGLLight(singleLight, ensemble);
		    }
		    lightCurrent++;
		}
	    }
	}
	else
	    if (lightCurrent - GL_LIGHT0 < lightMax)
		SetGLLight(lightCurrent, ensemble);
    }
    
    if (mCbXRay.GetBodyBOOL() == FALSE)
	GLDrawMain(bDrawLightCam, status, bHittest);
    

    //if (bHittest == FALSE) GLDrawPath();
}

void CLS::
GLDrawXray(bool bDrawLightCam, int status){
    GLDrawMain(bDrawLightCam, status, FALSE); 
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, "LightSource"/*, TRUE*/);

    wxString AstrLight[_LIGHT_MAX] = {_("Parallel Light"), _("Point Light"), _("SpotLight")};
    pC->SetChoiceList(_LIGHT_MAX, AstrLight/*LightSource_cls::msstrTypeName*/);
    
    pC->CombineSizer(2, _("Type/Fade"));
    pC->CombineConfItem(mCType);
    wxString AstrFade[3] = {_("No Fade"), _("power 1"), _("power 2")};
    pC->SetChoiceList(3, AstrFade);
    pC->CombineConfItem(mCFadeType);
    mIndexFadeDistance = pC->CombineConfItem(mCFadeDistance, _("Fade Distance"));

    mIndexSpot = pC->CombineSizer(2, _("Spot Angle/Exp."));
    pC->CombineConfItem(mCSpotAngle);
    pC->CombineConfItem(mCSpotExponent);

    pC->SetLimit(0, 10);
    pC->mDragScale = 1;
    pC->CombineConfItem(mCSoftSize, _("Soft Size"));
    
    pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(mCbXRay, _("X-Ray"));
    pC->CombineConfItem(mCbCastShadow, _("Cast Shadow"));

    pC->FinishSetting();
    return pC;
}

bool CLS::
HideAndShow(MultiInputPanel_cls* pC){
    bool bChange = FALSE;
    bool bSpot = (mCType.GetBodyINT() == _LIGHT_SPOT);
    bChange |= pC->mVCItem[mIndexSpot].Show(bSpot);
    bool bFade = (mCFadeType.GetBodyINT() != 0);
    bChange |= pC->mVCItem[mIndexFadeDistance].Show(bFade);
    return bChange;
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCMaterial->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeMaterial(wxString& strOut){ mpCMaterial->Serialize(strOut); }

wxString LightSource_cls::msstrTypeName[_LIGHT_MAX] = 
{ "ParallelLight", "PointLight", "SpotLight" };
