// (setq-c-offset 4)
// "clsMaterial.h"

#include "clsMaterial.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../SCENEOBJ/clsMeshObj.h"
#include "../SCENEOBJ/clsPrimitive.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/DATAVIEWER/clsHeightField.h"
#include "../SCENEOBJ/DATAVIEWER/clsIsosurface.h"
#include "../SCENEOBJ/DATAVIEWER/clsVolumeBox.h"
#include "../SIMDATA/clsImageSequence.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#define CLS Material_cls

CLS::
Material_cls(wxString strObjName, wxString strParentClassName) : /*MaterialBase_cls(strName, "Material_cls")*/Attribute_cls("Material_cls", "Material") {
    
    mstrParentClassName = strParentClassName;

    wxString strConfListLabel = "SceneObj::";  
    strConfListLabel << strObjName << "::Material";

    wxString strUpdate = "VIEW3D::PROPERTY_MATERIAL";

    mCConfList.SetTag(strConfListLabel);
    MaterialOption_cls::AddItemToConfList(mCConfList, strUpdate);
    MaterialTexture_cls::AddItemToConfList(mCConfList, strUpdate);

    mCConfList.Add(&mCstrLink, "Link=", "None", strUpdate);
    mCConfList.Add(&mCShaderType, "ShaderType=", "2", strUpdate);
    mCConfList.Add(&mCShaderDissolve, "ShaderDissolve=", "0", strUpdate);
    mCConfList.Add(&mCbShaderFlat, "ShaderFlat=", "0", strUpdate);
    mCConfList.Add(&mCbVolumeTrim, "VolumeTrim=", "1", strUpdate);
    mCConfList.Add(&mCToonThreshold, "ToonThreshold=", "0.1", strUpdate);
    mCConfList.Add(&mCToonBoldness, "ToonBoldness=", "0", strUpdate);

    double ambient = 0.2, diffuse = 0.8, specular = 0.2;
    if (strParentClassName.IsSameAs("LightSource_cls")){
	ambient = 1;
	diffuse = 1;
	specular = 1;
    }
    
    ParamGroup_cls *pCAmbient = GetPCNewParamGroup("Ambient", "R", "G", "B");
    InitializeAndPushParam(pCAmbient, ambient, ambient, ambient);

    ParamGroup_cls *pCDiffuse = GetPCNewParamGroup("Diffuse", "R", "G", "B");
    InitializeAndPushParam(pCDiffuse, diffuse, diffuse, diffuse);
    
    ParamGroup_cls *pCSpecular = GetPCNewParamGroup("Specular", "R", "G", "B");
    InitializeAndPushParam(pCSpecular, specular, specular, specular);

    if (strParentClassName.IsSameAs("Primitive_cls") ||
	strParentClassName.IsSameAs("MeshObj_cls") ||
	strParentClassName.IsSameAs("TextObj_cls") ||
	strParentClassName.IsSameAs("VolumeBox_cls") ||
	strParentClassName.IsSameAs("Isosurface_cls") ||
	strParentClassName.IsSameAs("HeightField_cls")){
	AddOneDimParam("Shininess", "S", 32);
	AddOneDimParam("Alpha", "A", 1);
    }

    mReferenceCounter = 0;
}

CLS::
~Material_cls(){
}

void CLS::
SyncConfListTag(wxString strName){
    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strName;
    strConfListLabel << wxT("::Material");
    mCConfList.SetTag(strConfListLabel);
}

Material_cls* CLS::
GetPCLinked(wxString strObjectClassName){
    
    wxString strLink = mCstrLink.GetBodySTR();

    SceneObj_cls *pCSceneObj = GetPCSceneObjCtrl()->GetPCFromName(strLink);
    if (pCSceneObj == NULL) return this;
    if (strObjectClassName.Find(pCSceneObj->mstrClassName) == wxNOT_FOUND)
	return this;
    
    wxString strClassName = pCSceneObj->mstrClassName;
    if (strClassName.IsSameAs("Empty")) return this;
    if (strClassName.IsSameAs("Particle_cls")) return this;
    
    mReferenceCounter++;
    if (mReferenceCounter > 1){
	mReferenceCounter--;
	return this;
    }
    
    Material_cls *pCLinked = this;
    if (strClassName.IsSameAs("Primitive_cls"))
	pCLinked = ((Primitive_cls*)pCSceneObj)->mpCMaterial->GetPCLinked(strObjectClassName);
    if (strClassName.IsSameAs("MeshObj_cls"))
	pCLinked = ((MeshObj_cls*)pCSceneObj)->mpCMaterial->GetPCLinked(strObjectClassName);
    if (strClassName.IsSameAs("HeightField_cls"))
	pCLinked = ((HeightField_cls*)pCSceneObj)->mpCMaterial->GetPCLinked(strObjectClassName);
    if (strClassName.IsSameAs("VolumeBox_cls"))
	pCLinked = ((VolumeBox_cls*)pCSceneObj)->mpCMaterial->GetPCLinked(strObjectClassName);
    if (strClassName.IsSameAs("Isosurface_cls"))
	pCLinked = ((Isosurface_cls*)pCSceneObj)->mpCMaterial->GetPCLinked(strObjectClassName);
    
    mReferenceCounter--;
    return pCLinked;
}



bool CLS::
DoesUseTexture(){
    wxString strPath = mCstrTexturePath.GetBodySTR();
    if (strPath.IsSameAs("None")) return FALSE;
    return TRUE;
}

float CLS::
GetAlpha(){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    return GetPCParamGroup("Alpha")->GetVInterpolatedValue((float)NT)[0];
}

float CLS::
GetShininess(){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    return GetPCParamGroup("Shininess")->GetVInterpolatedValue((float)NT)[0];
}

std::vector < float > CLS::
GetVColor(wxString strTag){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    
    std::vector < float > VClr;

    if (strTag.IsSameAs("Diffuse"))
	VClr = GetPCParamGroup("Diffuse")->GetVInterpolatedValue((float)NT);
    else if (strTag.IsSameAs("Specular"))
	VClr = GetPCParamGroup("Specular")->GetVInterpolatedValue((float)NT);
    else
        VClr = GetPCParamGroup("Ambient")->GetVInterpolatedValue((float)NT);

    VClr.push_back(1);
    return VClr;
}


void CLS::
GetColorInfo(float ambient[3], float diffuse[3], float specular[3], float &shininess){
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VAmb = GetVColor("Ambient");
    std::vector < float > VDif = GetVColor("Diffuse");
    std::vector < float > VSpe = GetVColor("Specular");
    
    shininess = GetPCParamGroup("Shininess")->GetVInterpolatedValue((float)NT)[0];

    for (int i = 0; i < 3; i++) ambient[i] = VAmb[i];
    for (int i = 0; i < 3; i++) diffuse[i] = VDif[i];
    for (int i = 0; i < 3; i++) specular[i] = VSpe[i];
}

void CLS::
SetGLMaterialSetting(){
    double time = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VAmb = GetPCParamGroup("Ambient")->GetVInterpolatedValue((float)time);
    std::vector < float > VDif = GetPCParamGroup("Diffuse")->GetVInterpolatedValue((float)time);
    std::vector < float > VSpe = GetPCParamGroup("Specular")->GetVInterpolatedValue((float)time);
    float shininess = GetPCParamGroup("Shininess")->GetVInterpolatedValue((float)time)[0];
    float alpha = GetPCParamGroup("Alpha")->GetVInterpolatedValue((float)time)[0];
    
    VAmb.push_back(alpha);
    VDif.push_back(1);
    VSpe.push_back(1);

    glMaterialfv(GL_FRONT, GL_AMBIENT, VAmb.data());
    glMaterialfv(GL_FRONT, GL_DIFFUSE, VDif.data());
    glMaterialfv(GL_FRONT, GL_SPECULAR, VSpe.data());
    glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);

    if (mCbCullBack.GetBodyBOOL() == TRUE){
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
    }
}

void CLS::
BeginShader(int lightNum, int ensemble, bool bForceUseTexture){
    int shaderType = mCShaderType.GetBodyINT();
    ShaderCtrl_cls *pCSC = GetPCShaderCtrl();
    ShaderItem_cls *pCShaderItemLink = NULL;
    switch(shaderType){
    case _MATERIAL_SHADER_GOURAUD:
	pCShaderItemLink = &pCSC->mVCShader[_SHADER_OBJECT_GOURAUD]; break;
    case _MATERIAL_SHADER_PHONG:
	pCShaderItemLink = &pCSC->mVCShader[_SHADER_OBJECT_PHONG]; break;
    case _MATERIAL_SHADER_TOON:
	pCShaderItemLink = &pCSC->mVCShader[_SHADER_OBJECT_TOON]; break;
    case _MATERIAL_SHADER_NORMAL:
	pCShaderItemLink = &pCSC->mVCShader[_SHADER_OBJECT_NORMAL]; break;
    }

    if (pCShaderItemLink != NULL){
	// JX^VF[_[ōăRpCKvȂ珈
    }
    
    bool bTexture = bForceUseTexture;
    if (bTexture == FALSE) bTexture = DoesUseTexture();
    int dissolve = mCShaderDissolve.GetBodyINT();
    int textureMode = (int)bTexture;
    ShaderParamEnsemble_cls CParamEnsemble(ensemble, dissolve,
					   lightNum, textureMode);
    switch(shaderType){
    case _MATERIAL_SHADER_GOURAUD: {
	ObjectGouraudParam_cls *pCParam = new ObjectGouraudParam_cls;
	pCParam->ShaderParamEnsemble_cls::CopyFrom(CParamEnsemble);
	pCSC->BeginObjectGouraud(pCParam);
	delete pCParam;
	break;}
    case _MATERIAL_SHADER_PHONG: {
	ObjectPhongParam_cls *pCParam = new ObjectPhongParam_cls;
	pCParam->ShaderParamEnsemble_cls::CopyFrom(CParamEnsemble);
	pCParam->mbFlat = mCbShaderFlat.GetBodyBOOL();
	pCParam->mUseAlpha = 1;
	pCSC->BeginObjectPhong(pCParam);
	delete pCParam;
	break;}
    case _MATERIAL_SHADER_TOON: {
	ObjectToonParam_cls *pCParam = new ObjectToonParam_cls;
	pCParam->ShaderParamEnsemble_cls::CopyFrom(CParamEnsemble);
	pCParam->mBoldness = 0.1;
	pCParam->mThreshold = mCToonThreshold.GetBodyDOUBLE();
	pCParam->mBoldness = mCToonBoldness.GetBodyDOUBLE();
	pCParam->mbFlat = mCbShaderFlat.GetBodyBOOL();
	pCSC->BeginObjectToon(pCParam);
	delete pCParam;
	break;}
    case _MATERIAL_SHADER_NORMAL:
	ObjectNormalParam_cls *pCParam = new ObjectNormalParam_cls;
	pCParam->mbFlat = mCbShaderFlat.GetBodyBOOL();
	pCSC->BeginObjectProjectNormal(pCParam);
	delete pCParam;
	break; 
    }
}

void CLS::
BeginLineShader(int ensemble){
    LineBasicParam_cls CParam;
    CParam.mEnsemble = ensemble;
    CParam.mDissolveMode = mCShaderDissolve.GetBodyINT();
    GetPCShaderCtrl()->BeginLineBasic(&CParam);
}

void CLS::
EndShader(){
    int shaderType = mCShaderType.GetBodyINT();
    if (shaderType != 0)
	GetPCShaderCtrl()->EndShader();
}

void CLS::
EndLineShader(){
    GetPCShaderCtrl()->EndShader();
}

bool CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (strFirstItem.IsSameAs("ReadMaterial") == FALSE) return FALSE;
    
    ASSERT(strLine.IsSameAs("Begin"));
    std::wstring wstrLine;
    while (std::getline(istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");

	if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE) 
	    continue;

	if (strFirstItem.IsSameAs("ReadParamGroup")){
	    wxString paramGroupName = PopStrItemFromStrLine(strLine, "::");
	    ParamGroup_cls *pCParamGroup = GetPCParamGroup(paramGroupName);
	    if (pCParamGroup != NULL){
	    
		std::wstring wstrLine;
		while (std::getline(istream, wstrLine)){
		    wxString strLine = wstrLine.c_str();
		wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
		pCParamGroup->ReadStrLine(istream, strFirstItem, strLine);
		
		if (strFirstItem.IsSameAs("ReadParamGroup")){
		    ASSERT(PopStrItemFromStrLine(strLine, wxT("::")).IsSameAs(wxT("End")));
		    break;
		}
		}
	    }
	}
	
	if (strFirstItem.IsSameAs("ReadMaterial")){
	    ASSERT(strLine.IsSameAs("End"));
	    break;
	}
    }

    return TRUE;
}

void CLS::
Serialize(wxString& strOut){
    strOut << "ReadMaterial::Begin" << ENDL;

    ConfListContainer_cls::Serialize(strOut);

    int numParamGroup = mVpCParamGroup.size();
    for (int i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->Serialize(strOut);

    strOut << "ReadMaterial::End" << ENDL;
}
