// (setq-c-offset 4)
// "clsMaterialBase.h"

#include "clsMaterialBase.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsGLExtention.h"
#include "../MYGL/clsTextureProvider.h"
#include "../SIMDATA/clsImageSequence.h"

#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include "../../wxWidgetEnum.h"

#define CLS MaterialBase_cls

CLS::
MaterialBase_cls(wxString strName, wxString strClassName, wxString strParentClassName) : Attribute_cls(strClassName, strParentClassName.IsSameAs("LightSource_cls") ? "Property" : "Material") {
    mstrParentClassName = strParentClassName;
    mpCImageSequence = new ImageSequence_cls;

    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strName;
    if (strParentClassName.IsSameAs("LightSource_cls")){
	strConfListLabel << wxT("::Property");
    }
    else
	strConfListLabel << wxT("::Material");
	
    wxString strUpdate = "VIEW3D::PROPERTY_MATERIAL";
    
    mCConfList.SetTag(strConfListLabel);

    mCConfList.Add(&mCbAdditive, "Additive=", "0", strUpdate);
    mCConfList.Add(&mCbCullBack, "CullBack=", "1", strUpdate);
    mCConfList.Add(&mCbDepthMask, "DepthMask=", "1", strUpdate);

    mCConfList.Add(&mCstrTexturePath, "TexturePath=", "None", strUpdate);
    mCConfList.Add(&mCTextureOffset, "TextureOffset=", "0", strUpdate);
    mCConfList.Add(&mCbUseTexture, "TextureUse=", "1", strUpdate);
    mCConfList.Add(&mCbTextureSequence, "TextureSequence=", "1", strUpdate);
    mCConfList.Add(&mCMapping, "Mapping=", "0", strUpdate);
    mCConfList.Add(&mCTexFilter, "TexFilter=", "1", strUpdate);
    mCConfList.Add(&mCTexWrap, "TexWrap=", "1", strUpdate);

    mCstrTexturePath.mbFilePath = TRUE;
}

CLS::
~MaterialBase_cls(){
    delete mpCImageSequence;
}

void CLS::
CheckFileSequence(){
    int offset = mCTextureOffset.GetBodyINT();
    mpCImageSequence->Set(mCstrTexturePath.GetBodySTR(), offset);

    mpCImageSequence->CalcMaxOffset(mCbTextureSequence.GetBodyBOOL());
    if (mpCImageSequence->mSeqLen != 1)
	mpCImageSequence->mbTimeDepend = TRUE;

    mstrPathChecked = mCstrTexturePath.GetBodySTR();
}

void CLS::
CheckTextureSequence(){
    if (mstrPathChecked != mCstrTexturePath.GetBodySTR())
	CheckFileSequence();
}

bool CLS::
BindTexture(int frame){
    wxString strPath = mCstrTexturePath.GetBodySTR();

    if (mCbTextureSequence.GetBodyINT() == TRUE && frame != -1){
	int frameImge;
	//frameImge = mpCImageSequence->GetIndexClamp(frame);
	frameImge = mpCImageSequence->GetIndexRepeat(frame);
	strPath = mpCImageSequence->GetModifiedStr(frameImge);
    }
    
    if (mCbUseTexture.GetBodyBOOL() == FALSE || 
	strPath.IsSameAs("None")){
	glEnable(GL_TEXTURE_2D);
	GetPCTextureProvider()->BindBlankTexture(TRUE);
	return FALSE;
    }

    glEnable(GL_TEXTURE_2D);
    GetPCTextureProvider()->BindTexture(strPath);
    if (mCTexFilter.GetBodyINT() == 0){
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    }
    else{
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    }
    switch(mCTexWrap.GetBodyINT()){
    case 0:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP_TO_EDGE);
	break;
    case 1:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_REPEAT);
	break;
    case 2:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S , GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T , GL_CLAMP);
	break;
    }
    return TRUE;
}

/*
void CLS::
SetBlendFunc(){
    if (mCbAdditive.GetBodyBOOL() == TRUE)
	GetPCGLExtentions()->
	    BlendFuncSeparate(GL_SRC_ALPHA, GL_ONE,
			      GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    else
	GetPCGLExtentions()->
	    BlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
			      GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
}
*/

void CLS::
SetDepthMask(){
    if (mCbDepthMask.GetBodyBOOL() == TRUE)
	glDepthMask(GL_TRUE);
    else
	glDepthMask(GL_FALSE);
}

MultiInputPanel_cls* CLS::
GetPCNewPanelTexture(wxWindow *pCWindow, ConfItemSTR_cls &CDummySTR,
		     int &indexTexturePath){

    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Texture");

    pC->mTextWidth = 80;

    indexTexturePath = pC->CombineSizer(4, _("TexturePath"));
    pC->CombineConfItem(mCstrTexturePath);
    pC->CombineButton(btnMaterialTexture, "..", 30);
    int indexTexClear = pC->CombineBitmapPanel(btnMaterialClearTexture, "Clear", 24);
    pC->mItemSize = 0;
    pC->CombineConfItem(mCbUseTexture);
    pC->mItemSize = 1;

    mIndexTextureOffset = pC->CombineSizer(2, _("Offset"));
    pC->CombineConfItem(mCTextureOffset, _("Offset"));
    pC->CombineConfItem(mCbTextureSequence, _("Image Sequence"));

    wxString AstrMapping[4] =
	{"Along X", "Along Y", "Along Z", "UV"};
    pC->SetChoiceList(4, AstrMapping);
    mIndexMapping = pC->CombineConfItem(mCMapping, _("Mapping"));

    mIndexInterpolate = pC->CombineSizer(2, _("Interpolate"));
    wxString AstrFilter[3] = {"Nearest", "Linear"};
    pC->SetChoiceList(2, AstrFilter);
    pC->CombineConfItem(mCTexFilter);
    wxString AstrRepeat[3] = {"Clamp at edge", _("Repeat"), "Clamp"};
    pC->SetChoiceList(3, AstrRepeat);
    pC->CombineConfItem(mCTexWrap);

    pC->FinishSetting();

    BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[indexTexClear].mpCWindowLink;
    pCBitmap->mbButton = TRUE;
    pCBitmap->AddBitmapStr("wxART_CLEAR_PANEL", "wxART_CLEAR_PANEL_P");
    pCBitmap->SetBitmapStrDisable("wxART_CLEAR_PANEL_D");
    pCBitmap->mBorderX = 4;
    pCBitmap->mBorderY = 4;

    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelOption(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Option"/*, TRUE*/);
    
    pC->mTextWidth = 100;
    pC->CombineSizer(2);
    pC->CombineConfItem(mCbCullBack, _("Cull Back"));
    pC->CombineConfItem(mCbAdditive, _("Additive"));
    pC->CombineSizer(1);
    pC->CombineConfItem(mCbDepthMask, _("Depth Mask"));
    
    pC->FinishSetting();
    return pC;
}
