// (setq-c-offset 4)
// "clsMaterial.h"

#include "clsMaterialLight.h"

#include "../MYGL/clsShadowVolumeCtrl.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#define CLS MaterialLight_cls

CLS::
MaterialLight_cls(wxString strName, wxString strClassName) : Attribute_cls(strClassName, "Property") {

    wxString strConfListLabel = "SceneObj::";  
    strConfListLabel << strName << "::Property";

    wxString strUpdate = "VIEW3D::PROPERTY_MATERIAL";

    double ambient = 1, diffuse = 1, specular = 1;

    ParamGroup_cls *pCAmbient = GetPCNewParamGroup("Ambient", "R", "G", "B");
    InitializeAndPushParam(pCAmbient, ambient, ambient, ambient);

    ParamGroup_cls *pCDiffuse = GetPCNewParamGroup("Diffuse", "R", "G", "B");
    InitializeAndPushParam(pCDiffuse, diffuse, diffuse, diffuse);
    
    ParamGroup_cls *pCSpecular = GetPCNewParamGroup("Specular", "R", "G", "B");
    InitializeAndPushParam(pCSpecular, specular, specular, specular);

    AddOneDimParam("Strength", "S", 1);
        
    ParamGroup_cls *pCSpot = GetPCNewParamGroup("SpotParam", "Ang.", "Exp.");
    InitializeAndPushParam(pCSpot, 35, 0);
    
    mReferenceCounter = 0;
}

void CLS::
SyncConfListTag(wxString strName){
    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strName;
    strConfListLabel << wxT("::Property");
    mCConfList.SetTag(strConfListLabel);
}

void CLS::
SetGLLightSetting(int lightCurrent){
    double time = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < float > VAmb = GetPCParamGroup("Ambient")->GetVInterpolatedValue((float)time);
    std::vector < float > VDif = GetPCParamGroup("Diffuse")->GetVInterpolatedValue((float)time);
    std::vector < float > VSpe = GetPCParamGroup("Specular")->GetVInterpolatedValue((float)time);
    float strength = GetPCParamGroup("Strength")->GetVInterpolatedValue((float)time)[0];

    for (int i = 0; i < 3; i++){
	VAmb[i] *= strength;
	VDif[i] *= strength;
	VSpe[i] *= strength;
    }

    if (GetPCShadowVolumeCtrl()->mSVState != 0){
	LOOP_RGB VDif[rgb] = 0;
	LOOP_RGB VSpe[rgb] = 0;
    }
	
    VAmb.push_back(1);
    VDif.push_back(1);
    VSpe.push_back(1);

    glLightfv(lightCurrent, GL_AMBIENT, VAmb.data());
    glLightfv(lightCurrent, GL_DIFFUSE, VDif.data());
    glLightfv(lightCurrent, GL_SPECULAR, VSpe.data());
}

bool CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (strFirstItem.IsSameAs("ReadMaterial") == FALSE) return FALSE;
    
    ASSERT(strLine.IsSameAs("Begin"));
    std::wstring wstrLine;
    while (std::getline(istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");

	if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE) 
	    continue;

	if (strFirstItem.IsSameAs("ReadParamGroup")){
	    wxString paramGroupName = PopStrItemFromStrLine(strLine, "::");
	    ParamGroup_cls *pCParamGroup = GetPCParamGroup(paramGroupName);
	    if (pCParamGroup != NULL){
	    
		std::wstring wstrLine;
		while (std::getline(istream, wstrLine)){
		    wxString strLine = wstrLine.c_str();
		wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
		pCParamGroup->ReadStrLine(istream, strFirstItem, strLine);
		
		if (strFirstItem.IsSameAs("ReadParamGroup")){
		    ASSERT(PopStrItemFromStrLine(strLine, wxT("::")).IsSameAs(wxT("End")));
		    break;
		}
		}
	    }
	}
	
	if (strFirstItem.IsSameAs("ReadMaterial")){
	    ASSERT(strLine.IsSameAs("End"));
	    break;
	}
    }

    return TRUE;
}

void CLS::
Serialize(wxString& strOut){
    strOut << "ReadMaterial::Begin" << ENDL;

    ConfListContainer_cls::Serialize(strOut);

    int numParamGroup = mVpCParamGroup.size();
    for (int i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->Serialize(strOut);

    strOut << "ReadMaterial::End" << ENDL;
}

