// (setq-c-offset 4)
// "clsPrimitive.cpp"

#include "clsMeshObj.h"
#include "clsMaterial.h"
#include "clsSceneObjCtrl.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../MESH/clsWavefrontMaterial.h"
#include "../MYGL/clsShadowVolumeCtrl.h"
#include "../MYGL/clsTextureProvider.h"
#include "../WIDGETCTRL/INPUTITEM/clsAtoSTextCtrl.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"

#include "../../clsMain.h"

#include <wx/file.h>
#include <wx/filename.h>

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include "../../wxWidgetEnum.h"

#define CLS MeshObj_cls

CLS::MeshObj_cls(ParentObj_cls *pCParentObj, wxString strName) :
    SceneObj_cls(pCParentObj)/*, mCMesh(this)*/
{

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";
    wxString strUpdateFormat = "VIEW3D::PROPERTY_FORMAT";

    mCConfList.Add(&mCstrPath, "Path=", "None", strUpdate);
    mCConfList.Add(&mCUseMaterial, "UseMaterial=", "1", strUpdate);
    mCConfList.Add(&mCbCastShadow, "CastShadow=",  "1", strUpdate);
    mCConfList.Add(&mCbXRay, "XRay=", "0", strUpdate);
   
    mCConfList.Add(&mCWavefrontCoord, "WavefrontCoord=", "0", strUpdateFormat);
    mCConfList.Add(&mCbWavefrontUseEmit, "WavefrontUseEmit=", "1", strUpdateFormat);
    mCstrPath.mbFilePath = TRUE;
    
    mstrClassName = wxT("MeshObj_cls");
    Initialize(strName);

    mpCMaterial = new Material_cls(strName, mstrClassName);
    
    mVpCAttribute.push_back(mpCMaterial);
}

CLS::
~MeshObj_cls(){
}

wxString CLS::
GetStrInitNameCandidate(){
    return "Mesh";
}

void CLS::
ClearMesh(){
    mCMesh2.ClearAll();
    mstrLoaded.Clear();
    mstrVBOMade.Clear();
}

wxString CLS::
GetStrInfo(){
    int numVertex = mCMesh2.mVVertex.size() / 3;
    int numFace = mCMesh2.mVFaceIndexVertex.size() / 3;
    int numMaterial = mCMesh2.mCWFMaterial.mVCItems.size();

    wxString strInfo;
    strInfo << _("Vertex Number") << ":" << numVertex << ENDL;
    strInfo << _("Face Number") << ":" << numFace << ENDL;
    strInfo << _("Matreial Number") << ":" << numMaterial << ENDL;

    int VBONum = mCMesh2.mCVBO.mVIndiceTriangle.size();

    strInfo << "VBO Num" << ":" << VBONum << ENDL;
    
    for (int i = 0; i < VBONum; i++)
	strInfo << mCMesh2.mCVBO.mVIndiceTriangle[i] << ENDL;
    
    return strInfo;
}

void CLS::
CheckLoad(){
    if (mCstrPath.GetBodySTR().IsSameAs(mstrLoaded) == FALSE){
	ReadMesh();
	wxString strName = mstrLoadedObjectName;
	if (strName.IsEmpty() == FALSE &&
	    mCstrName.GetBodySTR().IsSameAs(strName) == FALSE){
	    GetPCSceneObjCtrl()->RenameCommand(mCstrName.GetBodySTR(), strName);
	}
    }
}

void CLS::
CheckVBO(){
    if (mstrLoaded.IsSameAs(mstrVBOMade) == FALSE){
	mCMesh2.SetVBO();
	mstrVBOMade = mstrLoaded;
    }
}

void CLS::
CheckTree(){
    mCMesh2.CheckTree();
}

void CLS::
ReadMesh(){
    wxString strFile = mCstrPath.GetBodySTR();
    if (wxFile::Exists(strFile) == FALSE) return;
    
    if (mCMesh2.ReadObjFile(strFile, 
			    mCWavefrontCoord.GetBodyINT(), 
			    mCbWavefrontUseEmit.GetBodyBOOL()) == TRUE){
    }
    
    mstrLoaded = strFile;

    GetPCMain()->mCLastOpenedFileMESH.SetBodyFromData(strFile);
    GetPCMain()->UpdatePanels("PROPERTY_INFO");


}

void CLS::
GLDrawSV(){
    if (mCbCastShadow.GetBodyBOOL() == FALSE) return;
    
    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
    pCSV->BeginShader();
    
    if (pCSV->mCbClipCheck.GetBodyBOOL() == TRUE){
	glDisable(GL_DEPTH_TEST);
	pCSV->SettingFront();
	pCSV->SetCullingSide(2);
	mCMesh2.mCVBO.Draw(_VBO_SHADOW_VOLUME, TRUE);
	
	pCSV->SettingBack();
	pCSV->SetCullingSide(1);
	mCMesh2.mCVBO.Draw(_VBO_SHADOW_VOLUME, TRUE);
    }
	
    glEnable(GL_DEPTH_TEST);
    pCSV->SettingFront();
    pCSV->SetCullingSide(1);
    mCMesh2.mCVBO.Draw(_VBO_SHADOW_VOLUME, TRUE);

    pCSV->SettingBack();
    pCSV->SetCullingSide(2);
    mCMesh2.mCVBO.Draw(_VBO_SHADOW_VOLUME, TRUE);

    pCSV->EndShader();
}

//#include "../MYGL/clsTextureProvider.h"

void CLS::
GLDraw(int lightNum, int ensemble, int status){
    
    if (mCstrPath.GetBodySTR().IsSameAs("None")) return;
    CheckLoad();
    CheckVBO();
    
    if (mCMesh2.mVVertex.size() == 0) return;

    glEnable(GL_DEPTH_TEST);

    Material_cls *pCMaterial = mpCMaterial->GetPCLinked("Primitive_cls::MeshObj_cls::HeightField_cls::Isosurface_cls");
    bool bWireFrame = pCMaterial->mCbWireframe.GetBodyBOOL();
    bool bSkipMain = FALSE;
    if (bWireFrame == TRUE && status > 0) bSkipMain = TRUE;

    if (GetPCShadowVolumeCtrl()->mSVState == 1 && bWireFrame == FALSE){
	float alpha = pCMaterial->GetAlpha();
	bool bSkip = GetPCShadowVolumeCtrl()->IsSkip(alpha, ensemble, this);
	if (bSkip == TRUE) return;
	return GLDrawSV();
    }

    bool bUseNornal = TRUE;
    bool bUseTexture = TRUE;
    
    //if (mCMesh.mVNormal.size() == 0) bUseNornal = FALSE;
    //if (mCMesh.mVUV2.size() == 0) bUseTexture = FALSE;
    if (mCMesh2.mVUV2.size() == 0) bUseTexture = FALSE;

    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    GetPCTextureProvider()->BindBlankTexture(TRUE);

    glEnable(GL_BLEND);
    pCMaterial->SetBlendFunc();
    pCMaterial->SetDepthMask();

    if (bSkipMain == FALSE){

	if (bWireFrame == TRUE && status == 0)
	    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	pCMaterial->BeginShader(lightNum, ensemble, TRUE);
	
	pCMaterial->SetGLMaterialSetting();
	float alphaFactor = pCMaterial->GetAlpha();
	float shininess = pCMaterial->GetShininess();
	
	mCMesh2.GLDraw(bUseNornal, bUseTexture, mCUseMaterial.GetBodyBOOL(),
		       alphaFactor, shininess);
	
	/*
	  bool bUseVBO = FALSE;
	  if (bUseTexture == FALSE &&
	  pCMaterial->mCbShaderFlat.GetBodyBOOL() == TRUE)
	  bUseVBO = TRUE;
	  if (bUseVBO == TRUE){
	  mCVBOForSV.Draw(_VBO_OBJECT, TRUE);
	  }
	  else{
	  mCMesh.GLDraw(bUseNornal, bUseTexture);
	  }
	  
	  glDisable(GL_TEXTURE_2D);
	*/
	
	pCMaterial->EndShader();

	if (bWireFrame == TRUE && status == 0)
	    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    if (status > 0 || bWireFrame == TRUE){
	float lineWidth;
	glGetFloatv(GL_LINE_WIDTH, &lineWidth);
	//glLineWidth(lineWidth*2);

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	std::vector < float > VClr;
	if (status == 0) VClr = pCMaterial->GetVColor("Diffuse");
	if (status == 1) VClr = ColorProvider_cls::GetVGLColor("Selected");
	if (status == 2) VClr = ColorProvider_cls::GetVGLColor("Active");
	VClr[3] = pCMaterial->GetAlpha();
	
	glColor4fv(VClr.data());
	
	glEnable(GL_POLYGON_OFFSET_LINE);
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	glPolygonOffset(30, 30);
	pCMaterial->BeginLineShader(ensemble);

	/*mCVBOForSV*/mCMesh2.mCVBO.Draw(_VBO_OBJECT, TRUE);
	//mCMesh.GLDraw(bUseNornal, bUseTexture);
	
	pCMaterial->EndLineShader();
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glDisable(GL_POLYGON_OFFSET_LINE);
	
	//glLineWidth(lineWidth);
    }

    glDisable(GL_CULL_FACE);
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
}

void CLS::
SetPathGUI(MultiInputPanel_cls *pCMNum, wxString strPath){
    InputItem_cls* pCInputItem = pCMNum->GetPCInputItem(&mCstrPath);
    if (pCInputItem == NULL) return;
    ((AtoSTextCtrl_cls*)(pCInputItem->mpCWindowLink))->SetValue(strPath);
    pCInputItem->OnChanged(FALSE);
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName()/*, TRUE*/);

    pC->CombineSizer(2, _("File Path"));
    pC->CombineConfItem(mCstrPath);
    pC->CombineButton(btnMeshPath, "..", 30);

    wxString AstrMeshMaterial[2] = 
	{_("Material Color"), _("Original Color")};
    pC->SetChoiceList(2, AstrMeshMaterial);
    pC->CombineConfItem(mCUseMaterial, _("Base Color"));

    pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(mCbXRay, _("X-Ray"));
    pC->CombineConfItem(mCbCastShadow, _("Cast Shadow"));

    pC->CombineButton(btnReload, _("Reload"));

    pC->FinishSetting();
    return pC;
}

MultiInputPanel_cls* CLS::
GetPCNewPanelFormatWavefrontObj(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, "Format.Wavefront"/*, TRUE*/);
    
    wxString AstrCoord[6] = {"x-y-z", "y-z-x", "z-x-y", "x-z-y (*)", "z-y-x (*)", "y-x-z (*)"};
    pC->SetChoiceList(6, AstrCoord);
    pC->CombineConfItem(mCWavefrontCoord, _("Coordinates"));

    pC->CombineSizer(1, _("Option"));
    pC->CombineConfItem(mCbWavefrontUseEmit, _("Use emission as ambient"));

    pC->FinishSetting();
    return pC;
}

void CLS::
Reload(){
    ClearMesh();
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCMaterial->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeMaterial(wxString& strOut){ mpCMaterial->Serialize(strOut); }
