// (setq c-basic-offset 4)
// "clsOutliner.h"

#include "clsOutliner.h"

#include "../INPUTMETHOD/clsMouseCtrl.h"
#include "../MISC/clsMisc.h"
#include "../WIDGETCTRL/EX/clsPanelBar.h"
#include "../WIDGETCTRL/WX/clsFontProvider.h"
#include "../WIDGETCTRL/WX/myArtProv.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../SCENEOBJ/clsLightSource.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../PREFERENCE/clsShortcut.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/EX/clsPanelBar.h"
#include "../WIDGETCTRL/INPUTITEM/clsTextButton.h"
#include "../WIDGETCTRL/INPUTITEM/clsVCenter.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsIconProvider.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsWX.h"

#include "../../clsMain.h"

#include "../../wxWidgetEnumMenu.h"
#include "../../wxWidgetEnumMenu.h"

#include <wx/dcbuffer.h>

#define CLS Outliner_cls

BEGIN_EVENT_TABLE(Outliner_cls, wxPanelEX)
EVT_BUTTON(menuPanelView, CLS::OnMenuView)
EVT_BUTTON(menuPanelSelect, CLS::OnMenuSelect)
EVT_MENU(menuObjectDeselect, CLS::OnDeselect)
EVT_SIZE(Outliner_cls::OnSize)
END_EVENT_TABLE()

CLS::
Outliner_cls(wxWindow *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
    wxPanelEX(pCParent, id, pos, size, style)
{
    SetToDefaultParam();
    SetPanelType(_PANEL_OUTLINER);
    Freeze();
    
    mpCPanelBar = new PanelBar_cls(this);
    mpCPanelBar->Freeze();
    mpCPanelBar->SetPanelType(_PANEL_CHILD, _PANEL_OUTLINER);

    wxBoxSizer *pBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerBar = new wxBoxSizer(wxHORIZONTAL);
    
    ConstructTypeSelector(pBoxSizerBar);
    
    mpCTextMenuView = new TextButton_cls
	(mpCPanelBar, menuPanelView, GetPCStringProvider()->GetStrMenu("View"));
    mpCTextMenuView->SetEventTarget(this);
    VCenter_cls *pCMenuView = new VCenter_cls(mpCTextMenuView);

    mpCTextMenuSelect = new TextButton_cls
	(mpCPanelBar, menuPanelSelect, GetPCStringProvider()->GetStrMenu("Select"));
    mpCTextMenuSelect->SetEventTarget(this);
    VCenter_cls *pCMenuSelect = new VCenter_cls(mpCTextMenuSelect);

    pBoxSizerBar->Add(pCMenuView, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuSelect, 0, wxEXPAND);
    pBoxSizerBar->AddStretchSpacer(1);
    mpCPanelBar->ConstructBarItemZoom(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemDefault(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemHideShow(pBoxSizerBar);

    mpCPanelBar->SetSizer(pBoxSizerBar);

    mpCOutlinerMain = new OutlinerMain_cls(this, wxDefIDPS, 0);

    pBoxSizerSrc->Add(mpCPanelBar, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCOutlinerMain, 1, wxEXPAND);
    pBoxSizerSrc->AddSpacer(1);
    
    SetSizer(pBoxSizerSrc);

    mpCPanelBar->Thaw();
    Thaw();
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();
    mCConfList.Add(&mCFontSize, "FontSize=", "1");
}

int CLS::
GetLineHeight(){
    int fontType = mCFontSize.GetBodyINT();
    return FontProvider_cls::GetHeight(fontType);
}

void CLS::
Zoom(int direction){
    long long fontsize = mCFontSize.GetBodyINT() + direction;
    if (fontsize < 0) fontsize = 0;
    if (fontsize >= _FONT_MAX) fontsize = _FONT_MAX - 1;
    mCFontSize.SetBodyFromData(fontsize);
    UpdatePanels(wxT("ALL"));
}

void CLS::
ProcAfterMath(const wxString& strCommand, const wxString& strItem){
    mpCOutlinerMain->ProcAfterMath(strCommand, strItem);
} 

void CLS::
OnMenuView(wxComEvt& event){
    wxMenu CMenu;
    bool bBarItem = mpCPanelBar->mbShowItem;
    CMenu.AppendCheckItem(menuShowPanelManage, GetPCShortcut()->GetMenuString(menuShowPanelManage))->Check(bBarItem);
    
    mpCTextMenuView->PopMenu(this, &CMenu);
}

void CLS::
OnMenuSelect(wxComEvt& event){
    wxMenu CMenu;
    bool bBarItem = mpCPanelBar->mbShowItem;
    CMenu.Append(menuObjectDeselect, GetPCShortcut()->GetMenuString(menuObjectDeselect));

    mpCTextMenuSelect->PopMenu(this, &CMenu);
}

void CLS::
OnDeselect(wxComEvt& event){
    GetPCSceneObjCtrl()->Select(wxES, FALSE);
    //GetPCMain()->UpdatePanels("OUTLINER::TIMELINE::VIEW3D");
}

void CLS::
OnSize(wxSizeEvent& event){
    UpdatePanels(wxT("ALL"));
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs(wxT("ALL"))){
	mpCOutlinerMain->Refresh();
    }
}

#undef CLS
#define CLS OutlinerMain_cls

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
EVT_PAINT(CLS::OnPaint)
EVT_MENU(menuPopRename, CLS::OnRename)
EVT_LEFT_UP    (CLS::OnLeftUp)
EVT_LEFT_DOWN  (CLS::OnLeftDown)
EVT_RIGHT_DOWN (CLS::OnRightDown)
EVT_MOUSEWHEEL (CLS::OnMouseWheel)
EVT_MOTION     (CLS::OnMouseMotion)
EVT_LEFT_DCLICK(CLS::OnDoubleClick)
EVT_LEAVE_WINDOW(CLS::OnMouseLeave)
EVT_KEY_DOWN   (CLS::OnKeyDown)
EVT_IDLE(CLS::OnIdle)
END_EVENT_TABLE()

CLS::
OutlinerMain_cls(Outliner_cls *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxPanelEX(pCParent, id, pos, size, style)
{
    mFuncRenamer = wxComEH(CLS::OnRenamer);
    mFuncRenamerFocus = wxFocusEH(CLS::OnRenamerFocus);
    mFuncRenamerKey = wxKeyEH(CLS::OnRenamerKey);

    mpCParent = pCParent;
    SetTgtPanel(this);

    IconProvider_cls *pCIcon = GetPCIconProvider();
    for (int i = 0; i < _FONT_MAX; i++){
	mpCIconOpen[i] = pCIcon->GetPCIcon("GROUP_OPEN", i);
	mpCIconClose[i] = pCIcon->GetPCIcon("GROUP_CLOSE", i);
	mpCIconEyeP[i] = pCIcon->GetPCIcon("EYE_P", i);
	mpCIconEyeN[i] = pCIcon->GetPCIcon("EYE_N", i);
	mpCIconValidP[i] = pCIcon->GetPCIcon("EYE_P", i);
	mpCIconValidN[i] = pCIcon->GetPCIcon("EYE_N", i);
	mpCIconArrow[i] = pCIcon->GetPCIcon("ARROW", i);
    }
    
    mCDisplayItem.SetLink(&mCMemoryDCWorkSpace, &mADisplayedRect);
    mCOpenClose.SetLink(&mAIconRectOpenClose);
    mCVisible.SetLink(&mAIconRectVisible);
    mCValid.SetLink(&mAIconRectValid);

    mCBitmapBuffer.Create(240, 36);
    mCMemoryDCWorkSpace.SelectObject(mCBitmapBuffer);

    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    ColorProvider_cls::SetBGColor(this, "EditPanel");

    mLastSelectIndex = -1;

    mpCPopupMenu = new wxMenu; 
    mpCPopupMenu->Append(menuPopRename, _("Rename"));
    mpCPopupMenu->Append(menuDelete, _("Delete"));
}

CLS::
~OutlinerMain_cls(){
    delete mpCPopupMenu;
}

void CLS::
ClearDisplayedItems(){
    mAstrDisplayedItem.clear();
    mADisplayedDepth.clear();
    mADisplayedRight.clear();
    mADisplayedRect.clear();
    mAIconRectOpenClose.clear();
    mAIconRectVisible.clear();
    mAIconRectValid.clear();
}

int CLS::
GetYPos(int line){
    return line * mpCParent->GetLineHeight() + mOffsetY;
}

bool CLS::
IsInsertPointAdditional(int index){
    bool bDepthChasm = FALSE;
    if (index > 0 && index < mADisplayedDepth.size() &&
	mADisplayedDepth[index-1] > mADisplayedDepth[index]) 
	bDepthChasm = TRUE;
    return (mMouseXCurrent > 160) & bDepthChasm;
}

void CLS::
DCDisplayChildTarget(wxAutoBufferedPaintDC& dc, int index,
		     wxString strActiveObjectId){
    if (index >= mADisplayedDepth.size()) return;
    int lineHeight = mpCParent->GetLineHeight();
    if (mAstrDisplayedItem[index].IsSameAs(strActiveObjectId) == TRUE) return;
    
    int fSize = GetFontSize();
    dc.DrawIcon(*mpCIconArrow[fSize], mADisplayedRight[index], index*lineHeight+mOffsetY + lineHeight/8);
    /*
    dc.SetBrush(wxColour(64, 0, 0));
    dc.DrawRectangle(mADisplayedRight[index], index*lineHeight+mOffsetY,
		     lineHeight, lineHeight);
    */
}

void CLS::
DCDisplayInsertPoint(wxAutoBufferedPaintDC& dc, int index, 
		       wxString strActiveObjectId){
    dc.SetBrush(wxColour(64, 0, 0));
    bool bInsertPointAdditional = IsInsertPointAdditional(index);
    
    if (GetDragDistance() < 7) return;

    int lineHeight = mpCParent->GetLineHeight();
    if (bInsertPointAdditional == TRUE)
	dc.DrawRectangle(120, index*lineHeight+mOffsetY,
			 mTgtPanelWidth - 60 - 120, 2);
    else
	if (index < mAstrDisplayedItem.GetCount() &&    
	    mAstrDisplayedItem[index].IsSameAs(strActiveObjectId)== FALSE)
	    dc.DrawRectangle(0, index*lineHeight+mOffsetY,
			     mTgtPanelWidth - 60, 2);
}

void CLS::
DrawItem(wxAutoBufferedPaintDC& dc, ParentObj_cls* pCParent, int& itemCount,
	 SceneObj_cls *pCActiveObject, int depth){
    
    for (std::list< SceneObj_cls*>::iterator itr = pCParent->mVpCObj.begin();
	 itr != pCParent->mVpCObj.end(); itr++){

	SceneObj_cls *pCObj = (*itr);
	
	wxCoord w, h;
	wxString strItem = (*itr)->mCstrName.mstrBody.Clone();
	//strItem << wxT(":") << (*itr)->mReferenceCounter;
	int indentPos = 10+24+19+depth*24;
	int lineHeight =mpCParent->GetLineHeight();
	int yPos = GetYPos(itemCount);
	
	mCMemoryDCWorkSpace.GetMultiLineTextExtent(strItem, &w, &h);

	if (pCObj == pCActiveObject)
	    dc.DrawRectangle(indentPos-4, yPos, w+8, lineHeight);
	
	bool bSelected = GetPCSceneObjCtrl()->IsSelected(*itr);
	bool bClosed = Misc_cls::DoesExistInArray(mAstrClosedItem, strItem);
	
	ColorProvider_cls::SetTextColorDC(dc, bSelected);
	
	if (! (mbRename == TRUE && *itr == pCActiveObject))
	    mCDisplayItem.Draw(dc, strItem, indentPos, yPos);
	
	wxString strIcon = wxES; //pCObjCtrl->GetPCFromIndex(i)->mstrIconName;
	if (strIcon.IsEmpty() == FALSE){
	    dc.DrawBitmap(ARTS(strIcon, wxSize(24, 24)), 3, yPos);
	}

	wxSize sIcon = wxSize(lineHeight*0.8, lineHeight*0.8);
	int fSize = GetFontSize();
	int eyePos = mTgtPanelWidth - 60 + 4 - lineHeight*0.0;
	
	if (pCObj->mCbVisible.GetBodyBOOL() == TRUE)
	    mCVisible.Draw(dc, mpCIconEyeP[fSize], eyePos, yPos+lineHeight*0.0);
	else
	    mCVisible.Draw(dc, mpCIconEyeN[fSize], eyePos, yPos+lineHeight*0.0);

	if (pCObj->mstrClassName.IsSameAs("LightSource_cls")){
	    if (((LightSource_cls*)pCObj)->mCbValid.GetBodyBOOL() == TRUE)
		mCValid.Draw(dc, mpCIconValidP[fSize], eyePos + 16 + fSize, yPos + lineHeight*0.0);
	    else
		mCValid.Draw(dc, mpCIconValidN[fSize], eyePos + 16 + fSize, yPos + lineHeight*0.0);
	}
	else
	    mCValid.DrawEmpty();

	if (pCObj->mVpCObj.size() > 0){
	    int folderPosX = indentPos-2-lineHeight*0.5;
	    int folderPosY = yPos+lineHeight*0.2;
	    if (bClosed == TRUE)
		mCOpenClose.Draw(dc,mpCIconClose[fSize],folderPosX,folderPosY);
	    else
		mCOpenClose.Draw(dc,mpCIconOpen[fSize],folderPosX,folderPosY);
	}
	else
	    mCOpenClose.DrawEmpty();
	
	mAstrDisplayedItem.push_back((*itr)->mCstrName.mstrBody);
	mADisplayedDepth.push_back(depth);
	mADisplayedRight.push_back(w + 24 + 19 + 18 + depth*24);
	
	itemCount++;
	
	if (bClosed == FALSE)
	    DrawItem(dc, (*itr), itemCount, pCActiveObject, depth+1);
    }
}

void CLS::
DCDisplayList(wxAutoBufferedPaintDC& dc){
    
    dc.SetTextForeground(wxColour(255, 255, 255));
    dc.SetFont(FontProvider_cls::GetFont(GetFontSize()));
    mCMemoryDCWorkSpace.SetFont(FontProvider_cls::GetFont(GetFontSize()));

    dc.SetPen(*wxTRANSPARENT_PEN);
    dc.SetBrush(GetBackgroundColour());
    dc.DrawRectangle(0, 0, mTgtPanelWidth, mTgtPanelHeight);

    dc.SetBrush(ColorProvider_cls::GetColor("Active"));

    SceneObjCtrl_cls *pCObjCtrl = GetPCSceneObjCtrl();

    int itemCount = 0;
    ClearDisplayedItems();
    
    DrawItem(dc, pCObjCtrl, itemCount, pCObjCtrl->mpCActiveObject, 0);

    mAstrDisplayedItem.push_back(wxT("End"));

    int lineHeight = mpCParent->GetLineHeight();
    if (mbItemSelecting == TRUE && pCObjCtrl->mpCActiveObject != NULL){
	int index = GetIndex(lineHeight);

	wxString strActiveObjectName = pCObjCtrl->mpCActiveObject->GetStrName();
	if (mpCMouseCtrl->mbAModifier[_MOUSE_CTRL] == TRUE &&
	    mpCMouseCtrl->mbAModifier[_MOUSE_SHIFT] == FALSE)
	    DCDisplayChildTarget(dc, index, strActiveObjectName);
	else
	    DCDisplayInsertPoint(dc, index, strActiveObjectName);
    }

    dc.SetPen(wxColour(128, 128, 128));
    dc.DrawLine(mTgtPanelWidth-60, 0, mTgtPanelWidth-60, mTgtPanelHeight);
    dc.SetPen(*wxTRANSPARENT_PEN);

    mVirtualHeight = mpCParent->GetLineHeight() * itemCount;
}

int CLS::
GetFontSize(){ return mpCParent->mCFontSize.GetBodyINT(); }

int CLS::
GetItemIndexByMouse(){
    int lineHeight = mpCParent->GetLineHeight();
    if (mMouseXAtDown > mTgtPanelWidth - lineHeight) return -1;

    int index = GetIndex(lineHeight);
    if (index >= (int)GetPCSceneObjCtrl()->mVpCObj.size()) return -1;
    if (mMouseXAtDown < 200 + lineHeight) return index;

    return -1;
}


void CLS::
SelectItemByMouse(bool bAdditional, bool bMultiLine){

    int lineHeight = mpCParent->GetLineHeight();

    mbItemSelecting = FALSE;
    mbBarSelecting  = FALSE;
    if (mMouseXAtDown > mTgtPanelWidth - _SCROLL_BARWIDTH){
	mbBarSelecting = TRUE;
	Refresh(FALSE);
	return;
    }

    int index = GetIndex(lineHeight);
    if (index >= (int)mADisplayedRect.size()) {
	GetPCSceneObjCtrl()->Select(wxES, bAdditional);
	mLastSelectIndex = -1;
	return;
    }

    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    if (mAIconRectVisible[index].Contains(CAtDown))
	GetPCSceneObjCtrl()->ToggleVisibleCommand(mAstrDisplayedItem[index]);

    if (mAIconRectOpenClose[index].Contains(CAtDown))
	ToggleOpenClose(mAstrDisplayedItem[index]);

    if (mAIconRectValid[index].Contains(CAtDown))
	GetPCSceneObjCtrl()->ToggleValidCommand(mAstrDisplayedItem[index]);
    
    if (mADisplayedRect[index].Contains(CAtDown)){
	if (bMultiLine == TRUE && mLastSelectIndex != -1){
	    int min = mLastSelectIndex, max = index;
	    if (min > max) Misc_cls::Swap(&min, &max);
	    for (int i = min; i <= max; i++)
		GetPCSceneObjCtrl()->Select(mAstrDisplayedItem[i], bAdditional, i == index);
	}
	else
	    GetPCSceneObjCtrl()->Select(mAstrDisplayedItem[index], bAdditional);
	mbItemSelecting = TRUE;
	mLastSelectIndex = index;
    }
    
    if (mbItemSelecting == FALSE) SetCursor(mCursorA);
 
}



void CLS::
ToggleOpenClose(wxString strItem){
    if (Misc_cls::DoesExistInArray(mAstrClosedItem, strItem) == TRUE)
	mAstrClosedItem.Remove(strItem);
    else
	mAstrClosedItem.Add(strItem);
}

/*
void CLS::
ProcRename(const wxString& strNameOld, const wxString& strNameNew){
    if (Misc_cls::DoesExistInArray(mAstrClosedItem, strNameOld) == FALSE) return;
    mAstrClosedItem.Remove(strNameOld);
    mAstrClosedItem.Add(strNameNew);
}

void CLS::
ProcDelete(const wxString& strName){
    if (Misc_cls::DoesExistInArray(mAstrClosedItem, strName) == FALSE) return;
    mAstrClosedItem.Remove(strName);
}
*/

void CLS::
ProcAfterMath(const wxString& strCommand, wxString strItem){
    if (strCommand.IsSameAs("RenameObject")){
	wxString strNameOld = PopStrItemFromStrLine(strItem, "::");
	wxString strNameNew = strItem;
	if (Misc_cls::DoesExistInArray(mAstrClosedItem, strNameOld) == FALSE) return;
	mAstrClosedItem.Remove(strNameOld);
	mAstrClosedItem.Add(strNameNew);
    }
    if (strCommand.IsSameAs("DeleteObject")){
	if (Misc_cls::DoesExistInArray(mAstrClosedItem, strItem) == FALSE) return;
	mAstrClosedItem.Remove(strItem);
    }
}

void CLS::
OnRenamer(wxComEvt& event){ ProcRenamer(event); }

void CLS::
OnRenamerFocus(wxFocusEvent& event){ ProcRenamerFocus(event); }

void CLS::
OnRenamerKey(wxKeyEvent& event){ ProcRenamerKey(event); }

void CLS::
OnRename(wxComEvt& event){
    int index = GetItemIndexByName(GetPCSceneObjCtrl()->GetActiveObjectName());
    if (index == -1) return;
    ConstructRenameTextCtrl(this, GetFontSize(), this);
}

void CLS::
OnPaint(wxPaintEvent& event){
    wxAutoBufferedPaintDC dc(this);
    GetClientSize(&mTgtPanelWidth, &mTgtPanelHeight);
    DCDisplayList(dc);
    DrawScrollBar(dc);
}

void CLS::
OnLeftDown(wxMouseEvent& event){
    if (mbRename == TRUE){
	SendEvent(this, wxEVT_TEXT_ENTER, mpCRenamer->GetId());
	return; 
    }
    
    LeftDownBase(event);
    bool bAdditional = event.m_shiftDown;
    bool bMultiLine = event.m_controlDown;
    SelectItemByMouse(bAdditional, bMultiLine);

    SetFocus();
}

void CLS::
OnRightDown(wxMouseEvent& event){
    if (mbRename == TRUE){ mbKillRenamer = TRUE; return; }
    
    int index = GetItemIndexByName(GetPCSceneObjCtrl()->GetActiveObjectName());
    if (index == -1) return;
    PopupMenu(mpCPopupMenu);//, mMouseXAtDown, mMouseYAtDown);
    /*
    LeftDownBase(event);
    int index = GetItemIndexByMouse();
    if (index != -1){
	SelectItemByMouse();
	
    }
    */
    /*
    else
	SendMenuEvent(wxGetTopLevelParent(this), menuDropPrmitive);
    */
}

void CLS::
OnLeftUp(wxMouseEvent& event){

    SceneObjCtrl_cls *pCObjCtrl = GetPCSceneObjCtrl();
    int lineHeight = mpCParent->GetLineHeight();

    LeftUpBase(event);
    if (mbItemSelecting == TRUE){
	int indexTo = (mMouseYCurrent-mOffsetY) / lineHeight;
	bool bAdditional = IsInsertPointAdditional(indexTo);

	if (bAdditional == TRUE && indexTo > 0) indexTo--;
	
	if (indexTo < mAstrDisplayedItem.GetCount() &&
	    GetDragDistance() > 7){
	    if (mpCMouseCtrl->mbAModifier[_MOUSE_CTRL] == TRUE &&
		mpCMouseCtrl->mbAModifier[_MOUSE_SHIFT] == FALSE)
		pCObjCtrl->SetToChildCommand(mAstrDisplayedItem[indexTo]);
	    else
		pCObjCtrl->MoveItemToCommand(mAstrDisplayedItem[indexTo], bAdditional);
	}
    }
    mbItemSelecting = FALSE;
    mbBarSelecting = FALSE;

    SetCursor(mCursorB);
    Refresh(FALSE);

    event.Skip();
}

void CLS::
OnDoubleClick(wxMouseEvent& event){

    bool bAdditional = event.m_shiftDown;
    LeftDownBase(event);
    int index = GetIndex(mpCParent->GetLineHeight());
    if (index >= mADisplayedRect.size()) return;
    
    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    if (mADisplayedRect[index].Contains(CAtDown) == TRUE){
	GetPCSceneObjCtrl()->Select(mAstrDisplayedItem[index], bAdditional);
        ConstructRenameTextCtrl(this, GetFontSize(), this);
    }

    event.Skip();
}

void CLS::
OnMouseWheel(wxMouseEvent& event){
    if (WheelBase(event) == TRUE)
	Refresh(FALSE);
}

void CLS::
OnMouseMotion(wxMouseEvent& event){

    mMouseXCurrent = event.GetX();
    mMouseYCurrent = event.GetY();
    
    if (event.Dragging() == TRUE){
	MotionBase(event, (mbItemSelecting == FALSE), mbBarSelecting);
	
	CheckOffsetBase();
	Refresh(FALSE);
    }

    if (mbBarFloating == FALSE && event.GetX() > mTgtPanelWidth - 30){
	mbBarFloating = TRUE;
	Refresh(FALSE);
    }
    if (mbBarFloating == TRUE && event.GetX() < mTgtPanelWidth - 30){
	mbBarFloating = FALSE;
	Refresh(FALSE);
    }
    
    event.Skip();
}

void CLS::
OnMouseLeave(wxMouseEvent& event){
    LeaveBase(event);
    event.Skip();
}

void CLS::
OnKeyDown(wxKeyEvent& event){

    switch(event.GetKeyCode()){
    case WXK_F2:
	ConstructRenameTextCtrl(this, GetFontSize(), this);
	return;
	/*
    case 'X':
    case WXK_DELETE:
	mpCCtrlLink->RemoveSceneObjCommand();
	event.Skip();
	return; 
	*/
    }
    //SendKeyEventToTop(this, event);
}

void CLS::
OnIdle(wxIdleEvent& event){
    ProcIdle(this, this);
    event.Skip();
}
