// (setq c-basic-offset 4)
// "clsOutliner.cpp"

#ifndef _OUTLINER_H_
#define _OUTLINER_H_

#include "clsSceneObjViewerBase.h"

#include "../COMMON/clsGetPC.h"
#include "../CONFITEM/clsConfListContainer.h"
#include "../FILEIO/clsParser.h"
#include "../WIDGETCTRL/SCROLL/clsIconButtonCtrl.h"
#include "../WIDGETCTRL/SCROLL/clsScrollPanel.h"
#include "../WIDGETCTRL/EX/wxPanelEX.h"
#include "../WIDGETCTRL/WX/clsFontProvider.h"
#include "../COMMON/CommonDefinitionsWX.h"

#include <vector>
#include <wx/dcmemory.h>

class ParentObj_cls;
class SceneObj_cls;
class SceneObjCtrl_cls;
class TextButton_cls;
class wxAutoBufferedPaintDC;

class OutlinerMain_cls;

class Outliner_cls : public wxPanelEX, public GetPC_cls, public ConfListContainer_cls {
 public:
    Outliner_cls(wxWindow *pCParent, wxWindowID id=wxID_ANY, const wxPoint &pos=wxDP, const wxSize &size=wxDS, long style=wxTAB_TRAVERSAL);

    OutlinerMain_cls *mpCOutlinerMain;
    TextButton_cls *mpCTextMenuView, *mpCTextMenuSelect;
    
    
    ConfItemINT_cls mCFontSize;
    
    void SetToDefaultParam();    
    int GetLineHeight();
    void Zoom(int direction);
    void ProcAfterMath(const wxString& strCommand, const wxString& strItem);
    void OnMenuView(wxComEvt& event);
    void OnMenuSelect(wxComEvt& event);
    void OnDeselect(wxComEvt& event);
    void OnSize(wxSizeEvent& event);

    void UpdatePanels(const wxString& strPanelTag);

private:
    DECLARE_EVENT_TABLE()
};

class OutlinerMain_cls : public wxPanelEX, public ScrollPanel_cls, public SceneObjViewerBase_cls, public Parser_cls , public GetPC_cls {
public:
    OutlinerMain_cls(Outliner_cls *pCParent, wxWindowID id=wxID_ANY, const wxPoint &pos=wxDP, const wxSize &size=wxDS, long style=wxTAB_TRAVERSAL);
    ~OutlinerMain_cls();

    Outliner_cls* mpCParent;

    wxIcon *mpCIconOpen[_FONT_MAX], *mpCIconClose[_FONT_MAX];
    wxIcon *mpCIconEyeP[_FONT_MAX], *mpCIconEyeN[_FONT_MAX];
    wxIcon *mpCIconValidP[_FONT_MAX], *mpCIconValidN[_FONT_MAX];
    wxIcon *mpCIconArrow[_FONT_MAX];
    
    wxBitmap mCBitmapBuffer;
    wxMemoryDC mCMemoryDCWorkSpace;

    std::vector < int > mADisplayedDepth;
    std::vector < int > mADisplayedRight;

    std::vector < wxRect > mAIconRectOpenClose;
    std::vector < wxRect > mAIconRectVisible;
    std::vector < wxRect > mAIconRectValid;

    StringButtonCtrl_cls mCDisplayItem;
    IconButtonCtrl_cls mCOpenClose, mCVisible, mCValid;

    wxArrayString mAstrClosedItem;
    
    wxCursor mCursorA, mCursorB;

    bool mbItemSelecting;
    int mLastSelectIndex;

    void ClearDisplayedItems();
    int GetYPos(int line);
    bool IsInsertPointAdditional(int index);
    void DCDisplayChildTarget(wxAutoBufferedPaintDC& dc, int index,
			      wxString strActiveObjectId);
    void DCDisplayInsertPoint(wxAutoBufferedPaintDC& dc, int index, 
			      wxString strActiveObjectId);
    void DrawItem(wxAutoBufferedPaintDC& dc, ParentObj_cls* pCParent, 
		  int& itemCount, SceneObj_cls *pCActiveObject, int depth);
    void DCDisplayList(wxAutoBufferedPaintDC& dc);
    int  GetFontSize();
    int  GetItemIndexByMouse();
    void SelectItemByMouse(bool bAdditional, bool bMultiLine);

    void ToggleOpenClose(wxString strItem);
    void ProcAfterMath(const wxString& strCommand, wxString strItem);

    void OnRename(wxComEvt& event);
    void OnDelete(wxComEvt& event);
    void OnPaint(wxPaintEvent& event);
    void OnLeftDown(wxMouseEvent& event);
    void OnRightDown(wxMouseEvent& event);
    void OnLeftUp(wxMouseEvent& event);
    void OnMouseWheel(wxMouseEvent& event);
    void OnMouseMotion(wxMouseEvent& event);
    void OnDoubleClick(wxMouseEvent& event);
    void OnMouseLeave(wxMouseEvent& event);
    void OnKeyDown(wxKeyEvent& event);
    void OnRenamer(wxComEvt& event);
    void OnRenamerFocus(wxFocusEvent& event);
    void OnRenamerKey(wxKeyEvent& event);
    void OnIdle(wxIdleEvent& event);
    
private:
    DECLARE_EVENT_TABLE()
};

#endif
