// (setq-c-offset 4)
// "clsPrimitive.cpp"

#include "clsPrimitive.h"
#include "clsMaterial.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../GLOBJ/clsVBOCameraBox.h"
#include "../GLOBJ/clsVBOCone.h"
#include "../GLOBJ/clsVBOCube.h"
#include "../GLOBJ/clsVBOCylinder.h"
#include "../GLOBJ/clsVBOEmpty.h"
#include "../GLOBJ/clsVBOPlane.h"
#include "../GLOBJ/clsVBOSphere.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../MYGL/clsShadowVolumeCtrl.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#define CLS Primitive_cls

CLS::
Primitive_cls(ParentObj_cls *pCParentObj, wxString strType, wxString strName) : 
    SceneObj_cls(pCParentObj)
{
    int type = GetTypeFromStr(strType);
    wxString strTypeNum; strTypeNum << type;

    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";
    
    mCConfList.Add(&mCType, "Type=", strTypeNum, strUpdate);
    mCConfList.Add(&mCSphereDepth, "SphereDepth=", "2", strUpdate);
    mCConfList.Add(&mCbCastShadow, "CastShadow=",  "1", strUpdate);
    mCConfList.Add(&mCbXRay, "XRay=", "0", strUpdate);
        
    mstrClassName = "Primitive_cls";
    mpCMaterial = new Material_cls(strName, mstrClassName);

    Initialize(strName);
    
    mVpCAttribute.push_back(mpCMaterial);
}

CLS::
~Primitive_cls(){
}

wxString CLS::
GetStrInitName(){
    return msstrTypeName[mCType.GetBodyINT()];
}

wxString CLS::
GetStrInitNameCandidate(int type){
    if (type >= 0 && type < _PRIMITIVE_MAX) return msstrTypeName[type];
    return wxT("InvalidType");
}

int CLS::
GetTypeFromStr(wxString strType){
    for (int i = 0; i < _PRIMITIVE_MAX; i++)
	if (msstrTypeName[i].IsSameAs(strType)) return i;
    return -1;
}

void CLS::
SyncConfListTagPrimitive(){
    wxString strObj = mCstrName.GetBodySTR();
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }
    
    mpCMaterial->SyncConfListTag(strObj);
}

bool CLS::
IsInside(float fX, float fY, float fZ){
    switch(mCType.GetBodyINT()){
    case _PRIMITIVE_PLANE:
	if (fZ > 0) return FALSE;
	break;
    case _PRIMITIVE_BOX:
	if (fX < -1 || fX > 1) return FALSE;
	if (fY < -1 || fY > 1) return FALSE;
	if (fZ < -1 || fZ > 1) return FALSE;
	break;
    case _PRIMITIVE_SPHERE:
	if (fX*fX+fY*fY + fZ*fZ > 1) return FALSE;
	break;
    case _PRIMITIVE_CONE:
	if (fZ > 1 || fZ < -1) return FALSE;
	if (fX*fX+fY*fY > (1-fZ)*(1-fZ)/4) return FALSE;
	break;
    case _PRIMITIVE_CYLINDER:
	if (fZ > 1 || fZ < -1) return FALSE;
	if (fX*fX+fY*fY > 1) return FALSE;
	break;
    }
    return TRUE;
}

void CLS::
GLDrawEachType(bool bDrawGuideObject, int mapping, int objectState){

    switch(mCType.GetBodyINT()){
    case _PRIMITIVE_EMPTY: {
	if (bDrawGuideObject == FALSE) break;
	glDisable(GL_LIGHTING);
	GCClassProvider.GetPCVBOEmpty()->Draw();
	break; }
    case _PRIMITIVE_PLANE:
	GCClassProvider.GetPCVBOPlane()->Draw(objectState, mapping);
	break;
    case _PRIMITIVE_BOX:
	GCClassProvider.GetPCVBOCube()->Draw(objectState, mapping);
	break;
    case _PRIMITIVE_SPHERE: {
	int depth = mCSphereDepth.GetBodyINT();
	GCClassProvider.GetPCVBOSphere()->Draw(objectState, mapping, depth);
	break; }
    case _PRIMITIVE_CONE:
	GCClassProvider.GetPCVBOCone()->Draw(objectState, mapping);
	break;
    case _PRIMITIVE_CYLINDER:
	GCClassProvider.GetPCVBOCylinder()->Draw(objectState, mapping);
	break;
    }
}

void CLS::
GLDrawSV(){
    if (mCbCastShadow.GetBodyBOOL() == FALSE) return;
    
    ShadowVolumeCtrl_cls *pCSV = GetPCShadowVolumeCtrl();
    
    pCSV->BeginShader();

    if (pCSV->mCbClipCheck.GetBodyBOOL() == TRUE){
	glDisable(GL_DEPTH_TEST);    
	pCSV->SettingFront();
	pCSV->SetCullingSide(2);
	GLDrawEachType(FALSE, 0, _VBO_SHADOW_VOLUME);
	
	pCSV->SettingBack();
	pCSV->SetCullingSide(1);
	GLDrawEachType(FALSE, 0, _VBO_SHADOW_VOLUME);
    }
	
    glEnable(GL_DEPTH_TEST);
    pCSV->SettingFront();
    pCSV->SetCullingSide(1);
    GLDrawEachType(FALSE, 0, _VBO_SHADOW_VOLUME);

    pCSV->SettingBack();
    pCSV->SetCullingSide(2);
    GLDrawEachType(FALSE, 0, _VBO_SHADOW_VOLUME);

    pCSV->EndShader();
}

void CLS::
GLDrawBox(){ GCClassProvider.GetPCVBOCameraBox()->Draw(); }

void CLS::
GLDraw(bool bDrawGuideObject,
       int lightNum, int ensemble, int composit, int status,
       bool bHitTest){
    
    glEnable(GL_DEPTH_TEST);

    Material_cls *pCMaterial = mpCMaterial->GetPCLinked("Primitive_cls::MeshObj_cls::HeightField_cls::Isosurface_cls");
    bool bWireFrame = pCMaterial->mCbWireframe.GetBodyBOOL();
    bool bSkipMain = FALSE;
    if (bWireFrame == TRUE && status > 0) bSkipMain = TRUE;
    
    if (ensemble == 0 && mProcessedInInstanceLoop == 0){
	pCMaterial->CheckTextureSequence();
    }

    if (GetPCShadowVolumeCtrl()->mSVState == 1 && bWireFrame == FALSE){
	float alpha = pCMaterial->GetAlpha();
	bool bSkip = GetPCShadowVolumeCtrl()->IsSkip(alpha, ensemble, this);
	if (bSkip == TRUE) return;
	return GLDrawSV();
    }

    // qbgeXgɃtŕ\Ƃ߂ȏꍇiˑHjB́H
    if (bHitTest == TRUE)
	return GLDrawBox();

    int mapping = mpCMaterial->mCMapping.GetBodyINT();
    int type = mCType.GetBodyINT();

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    int frame1, subframe;
    double interpFactor = 
	GetPCTimeCtrl()->GetFrameFromNT(NT, frame1, subframe);
    int seq = GetPCTimeCtrl()->GetStep(frame1, subframe);
    int frame2 = frame1+1;
    
    GetPCTimeCtrl()->StoreInterpFactorForComposit(interpFactor);
    GetPCTimeCtrl()->AdjustFrameForComposer
	(&frame1, &frame2, &interpFactor, composit);
    
    ColorProvider_cls::SetGLColor("GuideObject");
    
    if (bSkipMain == FALSE){

	if (bWireFrame == TRUE && status == 0)
	    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	
	pCMaterial->BeginShader(lightNum, ensemble);
	
	glEnable(GL_BLEND);
	pCMaterial->SetBlendFunc();
	pCMaterial->SetDepthMask();
	
	pCMaterial->SetGLMaterialSetting();
	pCMaterial->BindTexture(frame1);
	
	ProcTexSpaceModBegin();
	
	glEnable(GL_LIGHTING);
	if (! (type == _PRIMITIVE_EMPTY && status != 0))
	    GLDrawEachType(bDrawGuideObject, mapping);
	glDisable(GL_TEXTURE_2D);
	
	mpCMaterial->EndShader();
	
	if (bWireFrame == TRUE && status == 0)
	    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    if (status > 0){
	float lineWidth;
	glGetFloatv(GL_LINE_WIDTH, &lineWidth);
	
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	std::vector < float > VClr;
	if (status == 1) VClr = ColorProvider_cls::GetVGLColor("Selected");
	if (status == 2) VClr = ColorProvider_cls::GetVGLColor("Active");
	VClr[3] = pCMaterial->GetAlpha();
	
	glColor4fv(VClr.data());
	
	glEnable(GL_POLYGON_OFFSET_LINE);
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	if (mCType.GetBodyINT() != _PRIMITIVE_EMPTY)
	    glPolygonOffset(30, 30);
	
	pCMaterial->BeginLineShader(ensemble);

	GLDrawEachType(bDrawGuideObject, mapping);

	pCMaterial->EndLineShader();
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glDisable(GL_POLYGON_OFFSET_LINE);
    }

    ProcTexSpaceModEnd();

    glDisable(GL_CULL_FACE);
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, "Primitive");
    
    pC->SetChoiceList(_PRIMITIVE_MAX, GetPCStringProvider()->GetAstr("PrimitiveType"));
    pC->CombineConfItem(mCType, _("Type"));

    /*
    wxString AstrMapping[4] =
	{"Along X", "Along Y", "Along Z", "UV"};
    pC->SetChoiceList(4, AstrMapping);
    pC->CombineConfItem(mCMapping, _("Mapping"));
    */

    pC->mDragScale = 4;
    pC->SetLimit(0, 4);
    mIndexDepth = pC->CombineConfItem(mCSphereDepth, _("Division Depth"));

    pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(mCbXRay, _("X-Ray"));
    pC->CombineConfItem(mCbCastShadow, _("Cast Shadow"));

    pC->FinishSetting();
    HideAndShow(pC);
    
    return pC;
}

bool CLS::
HideAndShow(MultiInputPanel_cls *pC){
    bool bChange = FALSE;
    int type = mCType.GetBodyINT();

    bool bSphere = (type == _PRIMITIVE_SPHERE) ? TRUE : FALSE;
    bChange |= pC->mVCItem[mIndexDepth].Show(bSphere);

    return bChange;
}

bool CLS::
HideAndShowTexture(MultiInputPanel_cls *pCMNumTex, int indexTexturePath){
    bool bChange = FALSE;
    bool bTex = mpCMaterial->mCstrTexturePath.GetBodySTR().IsSameAs("None") == FALSE;
    int indexTextureOffset = mpCMaterial->mIndexTextureOffset;
    int indexMapping       = mpCMaterial->mIndexMapping;
    int indexInterpolate   = mpCMaterial->mIndexInterpolate;

    if (bTex == FALSE) 
	pCMNumTex->mVCItem[indexTexturePath+3].mpCWindowLink->Disable();
    else
	pCMNumTex->mVCItem[indexTexturePath+3].mpCWindowLink->Enable();
    bChange |= pCMNumTex->mVCItem[indexTextureOffset].Show(bTex);
    bChange |= pCMNumTex->mVCItem[indexMapping].Show(bTex);
    bChange |= pCMNumTex->mVCItem[indexInterpolate].Show(bTex);
    return bChange;
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCMaterial->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeMaterial(wxString& strOut){ mpCMaterial->Serialize(strOut); }

wxString Primitive_cls::msstrTypeName[_PRIMITIVE_MAX] = 
{ "Empty", "Box", "Sphere", "Plane", "Cone", "Cylinder" };
