// (setq c-basic-offset 4)
// "clsSceneObj.cpp"

#include "clsSceneObj.h"
#include "clsSceneObjCtrl.h"

#include "clsAttribute.h"
#include "clsCamera.h"
#include "clsLightSource.h"
#include "clsMaterial.h"
#include "clsMeshObj.h"
#include "clsPrimitive.h"
#include "clsTextObj.h"
#include "clsWorld.h"
#include "DATAVIEWER/clsHeightField.h"
#include "DATAVIEWER/clsIsosurface.h"
#include "DATAVIEWER/clsColorCurve.h"
#include "DATAVIEWER/clsMaterialParticle.h"
#include "DATAVIEWER/clsMaterialStreamline.h"
#include "DATAVIEWER/clsVolumeBox.h"
#include "DATAVIEWER/clsParticle.h"
#include "DATAVIEWER/clsStreamline.h"
//#include "PROP/clsCameraProp.h"
//#include "PROP/clsFogProp.h"
//#include "PROP/clsFocalBlurProp.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../MISC/clsMisc.h"
#include "../MODIFIER/clsModifierBase.h"
#include "../MODIFIER/clsTracking.h"
#include "../MODIFIER/clsArrayCopy.h"
#include "../MODIFIER/clsModifyApparentSize.h"
#include "../MODIFIER/clsModifyParticleSize.h"
#include "../MODIFIER/clsNestedVolume.h"
#include "../MODIFIER/clsTrimming.h"
#include "../MODIFIER/clsTexSpaceMod.h"
#include "../MODIFIER/clsSequenceRemap.h"
#include "../MODIFIER/clsVolumeMask.h"
#include "../MODIFIER/clsParticleBoolean.h"
#include "../MODIFIER/clsParticleExclude.h"
#include "../MODIFIER/clsParticleFiltering.h"
#include "../MODIFIER/clsParticleTweak.h"
#include "../MODIFIER/clsParticleDownsize.h"
#include "../MODIFIER/clsParticleMerge.h"
#include "../MODIFIER/clsStreamFiltering.h"
#include "../MODIFIER/clsStreamSetAlpha.h"
#include "../MODIFIER/clsStreamTransform.h"
#include "../MODIFIER/clsVolumeBoolean.h"
#include "../MODIFIER/clsVolumeCoord.h"
#include "../MODIFIER/clsVolumeDownsize.h"
#include "../MYGL/clsPosition.h"
#include "../MYGL/clsShadowVolumeCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../UNDO/clsUndoManager.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

ParentObj_cls::
ParentObj_cls(int type){
    mParentObjType = type;
    mProcessedInInstanceLoop = 0;
}

ParentObj_cls::
~ParentObj_cls(){
    Clear();
}

void ParentObj_cls::
Clear(){
    while( mVpCObj.empty() == FALSE){
	SceneObj_cls *pCLastOne = mVpCObj.back();
	mVpCObj.pop_back();
	GetPCSceneObjCtrl()->DeleteSceneObj(pCLastOne);
    }
}

void ParentObj_cls::
GetVPCList(std::vector < SceneObj_cls* > &VpCList){
    for (uint i = 0; i < mVpCObj.size(); i++){
	SceneObj_cls *pCChild = *GetIterFromIndex(i);
	pCChild->GetVPCList(VpCList);
	VpCList.push_back(pCChild);
    }
}

std::list< SceneObj_cls* >::iterator ParentObj_cls::
GetIterFromP(SceneObj_cls *pC){
    for (std::list< SceneObj_cls* >::iterator itr = mVpCObj.begin();
	 itr != mVpCObj.end(); itr++){
	if (*itr == pC)
	    return itr;
    }
    return mVpCObj.end();
}

std::list< SceneObj_cls* >::iterator ParentObj_cls::
GetIterFromIndex(int index){
    int numObj = mVpCObj.size();
    if (index >= numObj || index < 0) return mVpCObj.end();
    std::list< SceneObj_cls* >::iterator itr = mVpCObj.begin();
    for (int i = 0; i < index; i++){
	if (itr == mVpCObj.end()) return itr;
	itr++;
    }
    return itr;
}

SceneObj_cls* ParentObj_cls::
GetPCFromIndex(int index){
    int numObj = mVpCObj.size();
    if (index >= numObj || index < 0) return NULL;
    return *GetIterFromIndex(index);
}

int ParentObj_cls::
GetIndexFromPC(SceneObj_cls *pC){
    int index = 0;
    for (std::list< SceneObj_cls* >::iterator itr = mVpCObj.begin();
	 itr != mVpCObj.end(); itr++){
	if (*itr == pC) return index;
	index++;
    }
    return -1;
}

wxString ParentObj_cls::
GetStrNewCandidate(wxString strCandidate){
    int pos = strCandidate.Find(wxUniChar('.'), TRUE);
    if (pos == wxNOT_FOUND)
	return strCandidate+wxT(".001");
    wxString strEnum = strCandidate.Mid(pos+1);
    if (strEnum.Matches(wxT("???"))){
	long num; strEnum.ToLong(&num);
	int numInt = num;
	wxString strDigit = wxString::Format(wxT("%03d"), numInt);
	if (strEnum.IsSameAs(strDigit) == TRUE){
	    return strCandidate.Mid(0, pos+1) + wxString::Format(wxT("%03d"), numInt+1);
	}
    }
    return strCandidate+wxT(".001");
}

void ParentObj_cls::
GetAllObjectsName(wxArrayString &CAString){
    for (std::list< SceneObj_cls* >::iterator itr = mVpCObj.begin();
	 itr != mVpCObj.end(); itr++){
	(*itr)->GetAllObjectsName(CAString);
    }
    if (mParentObjType == _PARENTOBJ_OBJ)
	CAString.Add(((SceneObj_cls*)this)->mCstrName.mstrBody);
}

void ParentObj_cls::
GetDirectChildrensName(wxArrayString &CAString){
    for (std::list< SceneObj_cls* >::iterator itr = mVpCObj.begin();
	 itr != mVpCObj.end(); itr++){
	CAString.Add((*itr)->mCstrName.mstrBody);
    }
}

int ParentObj_cls::
GetActiveLightNumRoot(bool bForShadow){
    int lightNum = 0;
    uint numObj = mVpCObj.size();
    for (uint i = 0; i < numObj; i++)
	lightNum += GetPCFromIndex(i)->GetActiveLightNum(bForShadow);
    if (bForShadow == FALSE && lightNum == 0) lightNum = 1;
    return lightNum;
}

bool ParentObj_cls::
IsSameAs(wxString strClassName){
    return mstrClassName.IsSameAs(strClassName);
}

bool ParentObj_cls::
IsObjExist(wxString strName, SceneObj_cls *pCException){
    for (uint i = 0; i < mVpCObj.size(); i++){
	SceneObj_cls *pCChild = *GetIterFromIndex(i);
	if (pCChild != pCException &&
	    pCChild->GetStrName().IsSameAs(strName) == TRUE) return TRUE;
	if (pCChild->IsObjExist(strName) == TRUE) return TRUE;
    }
    return FALSE;
}

bool ParentObj_cls::
IsAncestor(wxString strName){
    bool bRet = FALSE;
    if (mParentObjType == 0) return bRet;
    else{
	SceneObj_cls *pCObj = (SceneObj_cls*)this;
	if (strName.IsSameAs(pCObj->mCstrName.mstrBody)) return TRUE;
	bRet |= pCObj->mpCParentObj->IsAncestor(strName);
    }
    return bRet;
}

SceneObj_cls* ParentObj_cls::
GetPCFromName(const wxString& strName){
    uint numObj = mVpCObj.size();
    for (uint i = 0; i < numObj; i++){
	SceneObj_cls *pCChild = GetPCFromIndex(i);
	SceneObj_cls *pCGrandCHild = pCChild->GetPCFromName(strName);
	if (pCGrandCHild != NULL) return pCGrandCHild;
	
	if (pCChild->mCstrName.mstrBody.IsSameAs(strName) == TRUE) return pCChild;
	
    }
    return NULL;
}

SceneObj_cls* ParentObj_cls::
GetPCWorld(){
    uint numObj = mVpCObj.size();
    for (uint i = 0; i < numObj; i++){
	SceneObj_cls *pCChild = GetPCFromIndex(i);
	SceneObj_cls *pCGrandCHild = pCChild->GetPCWorld();
	if (pCGrandCHild != NULL) return pCGrandCHild;
	
	if (pCChild->IsSameAs("World_cls") == TRUE &&
	    pCChild->mCbVisible.GetBodyBOOL() == TRUE) return pCChild;
	
    }
    return NULL;
}

Attribute_cls* ParentObj_cls::
GetPCAttributeFromAddress(wxString strAddress){
    wxString strObj = PopStrItemFromStrLine(strAddress, "::");
    SceneObj_cls* pCObj = GetPCFromName(strObj);
    if (pCObj != NULL){
	Attribute_cls* pCAttrib = pCObj->GetPCAttribute/*FromName*/(strAddress);
	return pCAttrib;
    }
    return NULL;
}

wxString ParentObj_cls::
GetStrName(){
    if (mParentObjType == _PARENTOBJ_ROOT) return wxT("ROOT");
    return ((SceneObj_cls*)this)->mCstrName.mstrBody;
}

void ParentObj_cls::
ClearProcessedMarkInInstanceLoop(){
    mProcessedInInstanceLoop = 0;
    for (uint i = 0; i < mVpCObj.size(); i++)
	(*GetIterFromIndex(i))->ClearProcessedMarkInInstanceLoop();
}

void ParentObj_cls::
InitializeStatusForDrawing(){
    if (IsSameAs("Particle_cls"))
	((Particle_cls*)this)->InitializeStatusForDrawingExe();
    if (IsSameAs("VolumeBox_cls"))
	((VolumeBox_cls*)this)->InitializeStatusForDrawingExe();
    for (uint i = 0; i < mVpCObj.size(); i++)
	(*GetIterFromIndex(i))->InitializeStatusForDrawing();
}

void ParentObj_cls::
CheckStatusForDrawing(){
    if (IsSameAs("Particle_cls"))
	((Particle_cls*)this)->CheckStatusForDrawingExe();
    if (IsSameAs("VolumeBox_cls"))
	((VolumeBox_cls*)this)->CheckStatusForDrawingExe();
    for (uint i = 0; i < mVpCObj.size(); i++)
	(*GetIterFromIndex(i))->CheckStatusForDrawing();
}

void ParentObj_cls::
InsertSceneObjNative(SceneObj_cls *pCObj, int pos){
    mVpCObj.insert(GetIterFromIndex(pos), pCObj);
}

void ParentObj_cls::
ProcessSortSelection(std::vector < SceneObj_cls* > &VpCNewList,
		     std::vector < SceneObj_cls* > VpCSelectiedObject){
    for (uint i = 0; i < VpCSelectiedObject.size(); i++){
	wxString strSelfName = GetStrName();
	if (VpCSelectiedObject[i]->mCstrName.mstrBody.IsSameAs(strSelfName))
	    VpCNewList.push_back(VpCSelectiedObject[i]);
    }
    for (std::list< SceneObj_cls* >::iterator itr = mVpCObj.begin();
	 itr != mVpCObj.end(); itr++){
	(*itr)->ProcessSortSelection(VpCNewList, VpCSelectiedObject);
    }
}


void ParentObj_cls::
InheritChilds(SceneObj_cls* pCFrom){
    std::list < SceneObj_cls* >::iterator itrNext = GetIterFromP(pCFrom);
    itrNext++;
    
    for (std::list< SceneObj_cls* >::iterator itr = pCFrom->mVpCObj.begin();
	 itr != pCFrom->mVpCObj.end(); itr++){
	mVpCObj.insert(itrNext, *itr);
	(*itr)->mpCParentObj = this;
    }
    pCFrom->mVpCObj.clear();
}

void ParentObj_cls::
RescaleParametersRoot(double scale){
    uint numObj = mVpCObj.size();

    for (uint i = 0; i < numObj; i++)
	GetPCFromIndex(i)->RescaleParameters(scale);
}

void ParentObj_cls::
GetTotalUsedMemoryMB(int &memoryMB){
    uint numObj = mVpCObj.size();

    for (uint i = 0; i < numObj; i++)
      GetPCFromIndex(i)->GetTotalUsedMemoryMB(memoryMB);

    if (mParentObjType == _PARENTOBJ_OBJ)
      memoryMB += ((SceneObj_cls*)this)->GetUsedMemoryMB();
}

void ParentObj_cls::
Prefetch(){
    uint numObj = mVpCObj.size();
    for (uint i = 0; i < numObj; i++)
	GetPCFromIndex(i)->Prefetch();

    if (mParentObjType == _PARENTOBJ_OBJ){
	SceneObj_cls* pCThis = (SceneObj_cls*)this;
	if (pCThis->mCbVisible.GetBodyBOOL() == FALSE) return;
	pCThis->PrefetchData();
    }
}

void ParentObj_cls::
ReleaseMemory(){
    uint numObj = mVpCObj.size();
    for (uint i = 0; i < numObj; i++)
	GetPCFromIndex(i)->ReleaseMemory();

    if (mParentObjType == _PARENTOBJ_OBJ){
	((SceneObj_cls*)this)->ReleaseMemoryData();
    }
}

// ----

#define CLS SceneObj_cls

CLS::
SceneObj_cls(ParentObj_cls *pCParentObj) : ParentObj_cls(_PARENTOBJ_OBJ) {

    mSortOrder = 2;

    mpCParentObj = pCParentObj;

    wxArrayString mAstrItemPosition;
    mAstrItemPosition.push_back("x");
    mAstrItemPosition.push_back("y");
    mAstrItemPosition.push_back("z");
    mVpCParamGroup.push_back(new ParamGroup_cls("Position", mAstrItemPosition));
    
    wxArrayString mAstrItemRotationEuler;
    mAstrItemRotationEuler.push_back("x");
    mAstrItemRotationEuler.push_back("y");
    mAstrItemRotationEuler.push_back("z");
    mVpCParamGroup.push_back(new ParamGroup_cls("RotationEuler", mAstrItemRotationEuler));
    
    wxArrayString mAstrItemRotationScale;
    mAstrItemRotationScale.push_back("x");
    mAstrItemRotationScale.push_back("y");
    mAstrItemRotationScale.push_back("z");
    ParamGroup_cls* pCScale = new ParamGroup_cls("Scale", mAstrItemRotationEuler);
    pCScale->mVCParam[0].mVCKeys[0].mValue = 1;
    pCScale->mVCParam[1].mVCKeys[0].mValue = 1;
    pCScale->mVCParam[2].mVCKeys[0].mValue = 1;
    mVpCParamGroup.push_back(pCScale);

    mGeneratedDepth = 0;
}

CLS::
~SceneObj_cls(){
    ASSERT(mReferenceCounter == 0);
    uint numParamGroup = mVpCParamGroup.size();
    for (int i = 0; i < numParamGroup; i++)
	delete mVpCParamGroup[i];
    while( mVpCAttribute.empty() == FALSE){
	Attribute_cls *pCLastOne = mVpCAttribute.back();
	mVpCAttribute.pop_back();
	Attribute_cls::Delete(pCLastOne);
    }
    while (mVpCDataProc.empty() == FALSE){
	Attribute_cls *pCLastOne = mVpCDataProc.back();
	mVpCDataProc.pop_back();
	Attribute_cls::Delete(pCLastOne);
    }
}

void CLS::
Initialize(wxString strName){
    if (mstrClassName.IsEmpty() == TRUE)
	mstrClassName = wxT("BaseSceneObj");
    
    if (strName.IsEmpty() == TRUE){
	strName = GCClassProvider.GetPCSceneObjCtrl()->GetStrUniqueName(mstrClassName);
    }
    
    mCstrName.SetItem("Name=", strName);
    mCConfList.Add(&mCbVisible, "Visible=", "1");
    mCConfList.Add(&mCbShy, "Shy=", "0", "TIMELINE"); 
    
    SyncConfListTag();

    mReferenceCounter = 0;
}

void CLS::
ProcRename(wxString strNewName){
    mCstrName.SetBodyFromData(strNewName);
    SyncConfListTag();
}

void CLS::
SyncConfListTag(){
    wxString strName = mCstrName.GetBodySTR();
    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strName;
    mCConfList.SetTag(strConfListLabel);
    if (IsSameAs("LightSource_cls"))
	((LightSource_cls*)this)->SyncConfListTagLightSource();
    if (IsSameAs("VolumeBox_cls"))
	((VolumeBox_cls*)this)->SyncConfListTagVolumeBox();
    if (IsSameAs("Particle_cls"))
	((Particle_cls*)this)->SyncConfListTagParticle();
    if (IsSameAs("VolumeBox_cls"))
	((VolumeBox_cls*)this)->SyncConfListTagVolumeBox();
    if (IsSameAs("Isosurface_cls"))
	((Isosurface_cls*)this)->SyncConfListTagIsosurf();
    if (IsSameAs("Streamline_cls"))
	((Streamline_cls*)this)->SyncConfListTagStreamline();
    if (IsSameAs("Primitive_cls"))
	((Primitive_cls*)this)->SyncConfListTagPrimitive();
    if (IsSameAs("TextObj_cls"))
	((TextObj_cls*)this)->SyncConfListTagTextObj();
    if (IsSameAs("HeightField_cls"))
	((HeightField_cls*)this)->SyncConfListTagHeightField();
}

wxArrayString CLS::
GetAstrAttribList(){
    wxArrayString AstrList;
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++) 
	AstrList.push_back(GetPCAttributeFromIndex(i)->GetStrName());
    return AstrList;
}

std::vector < SceneObjInstance_cls* > CLS::
GetVNewPCInstance(){
    std::vector < SceneObjInstance_cls* > VPCInstance;
    
    // ́AsłglLight̃ZbgpɏKv
    if (IsSameAs("LightSource_cls") == FALSE)
	if (mCbVisible.GetBodyBOOL() == FALSE) return VPCInstance;
    
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++) {
	if (GetPCAttributeFromIndex(i)->GetStrAttributeName() == "ArrayCopy_cls")
	    ((ArrayCopy_cls*)GetPCAttributeFromIndex(i))->Execute(this, VPCInstance, i);
    }

    SceneObjInstance_cls *pCInstance = new SceneObjInstance_cls(this);
    VPCInstance.push_back(pCInstance);

    return VPCInstance;
}

int CLS::
GetChildIndex(){ 
    return mpCParentObj->GetIndexFromPC(this); }

int CLS::
GetActiveLightNum(bool bForShadow){
    int lightNum = 0;
    if (IsSameAs("LightSource_cls")){
	LightSource_cls *pCLight = (LightSource_cls*)this;
	if (pCLight->mCbValid.GetBodyBOOL() == TRUE){
	    if (bForShadow == FALSE ||
		pCLight->mCbCastShadow.GetBodyBOOL() == TRUE) lightNum++;
	}
    }
    uint numObj = mVpCObj.size();
    for (uint i = 0; i < numObj; i++)
	lightNum += GetPCFromIndex(i)->GetActiveLightNum(bForShadow);
    return lightNum;
}

void CLS::
GetVstrParamTag(std::vector < wxString > &VstrTag){
    uint numParamGroup = mVpCParamGroup.size();
    for (int i = 0; i < numParamGroup; i++){
	uint numParam = mVpCParamGroup[i]->mVCParam.size();
	for (int j = 0; j < numParam; j++)
	    VstrTag.push_back(GetStrName() + "::" + mVpCParamGroup[i]->GetStrName() + "::" + mVpCParamGroup[i]->mVCParam[j].GetStrName());
    }

    uint numAttrib = mVpCAttribute.size();
    for (int i = 0; i < numAttrib; i++)
	(*GetIterAttributeFromIndex(i))->GetVstrParamTag(GetStrName(), VstrTag);
}

ParamGroup_cls* CLS::
GetPCParamGroup(wxString strFirstItem, wxString strTag){
    uint num = mVpCParamGroup.size();
    for (uint i = 0; i < num; i++){
	if (mVpCParamGroup[i]->mCstrName.GetBodySTR().IsSameAs(strFirstItem))
	    return mVpCParamGroup[i];
    }

    if (strTag.IsEmpty() == TRUE){
	if (strFirstItem.Find("::") != wxNOT_FOUND){
	    wxString strFirstItemOrg = strFirstItem;
	    strFirstItem = PopStrItemFromStrLine(strFirstItemOrg, "::");
	    strTag = strFirstItemOrg;
	}
    }

    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	
	if (GetPCAttributeFromIndex(i)->mCstrName.GetBodySTR().IsSameAs(strFirstItem)){
	    //J[J[u̎ strFirstItem == attrib = paramGroup
	    ParamGroup_cls *pCParamSelf = GetPCAttributeFromIndex(i)->GetPCParamGroup(strFirstItem);
	    
	    //łȂƂ́ÃACe paramGroup
	    strFirstItem = PopStrItemFromStrLine(strTag, "::");
	    ParamGroup_cls *pCParamGroup = GetPCAttributeFromIndex(i)->GetPCParamGroup(strFirstItem);
	    if (pCParamGroup != NULL) return pCParamGroup;
	    return pCParamSelf;
	}
    }
    
    uint procNum = mVpCDataProc.size();
    for (uint i = 0; i < procNum; i++){
	if (GetPCDataProcFromIndex(i)->mCstrName.GetBodySTR().IsSameAs(strFirstItem)){
	    strFirstItem = PopStrItemFromStrLine(strTag, "::");
	    return GetPCDataProcFromIndex(i)->GetPCParamGroup(strFirstItem);
	}
    }
    
    return NULL;
}

ConfItem_cls* CLS::
GetPCItemInAttrib(wxString strAttribName, wxString strTag){

    Attribute_cls* pCAttrib = GetPCAttribute(strAttribName);
    if (pCAttrib == NULL) return NULL;

    return pCAttrib->GetPCItem(strTag);
}

Material_cls* CLS::
GetPCMaterial(){
    if (IsSameAs("Primitive_cls"))  return ((Primitive_cls*)this)->mpCMaterial;
    if (IsSameAs("MeshObj_cls"))    return ((MeshObj_cls*)this)->mpCMaterial;
    if (IsSameAs("TextObj_cls"))    return ((TextObj_cls*)this)->mpCMaterial;
    if (IsSameAs("VolumeBox_cls"))  return ((VolumeBox_cls*)this)->mpCMaterial;
    if (IsSameAs("Isosurface_cls")) return ((Isosurface_cls*)this)->mpCMaterial;
    if (IsSameAs("HeightField_cls")) return ((HeightField_cls*)this)->mpCMaterial;
    return NULL;
}

void CLS::
ClearModifierMatrix(){
    LoadIdentity(mModifierMatrix);
}

std::vector < double > CLS::
GetPositionGlobal(double NT, double *pPosOffset){

    double matrix16[16];
    GetMatrix16(matrix16, NT, pPosOffset);
    std::vector < double > VPosGlobal;
    VPosGlobal.push_back(matrix16[12]);
    VPosGlobal.push_back(matrix16[13]);
    VPosGlobal.push_back(matrix16[14]);
    
    return VPosGlobal;

    /*
    ParamGroup_cls *pCPos = GetPCParamGroup("Position");
    std::vector < double > VPosLocal = pCPos->GetVInterpolatedValue(NT);
    
    if (pPosOffset != NULL)
	LOOP_XYZ VPosLocal[xyz] += pPosOffset[xyz];
    
    double matrix[16];
    LoadIdentity(matrix);
    matrix[12] = VPosLocal[0];
    matrix[13] = VPosLocal[1];
    matrix[14] = VPosLocal[2];

    double matrixParent[16];
    LoadIdentity(matrixParent);

    if (mpCParentObj->mParentObjType != _PARENTOBJ_ROOT)
	((SceneObj_cls*)mpCParentObj)->GetMatrix16(matrixParent, NT);

    Multiply(matrixParent, matrix);
    
    std::vector < double > VPosGlobal;
    VPosGlobal.push_back(matrix[12]);
    VPosGlobal.push_back(matrix[13]);
    VPosGlobal.push_back(matrix[14]);

    return VPosGlobal;
    */
}

void CLS::
GetMatrix16(double matrix[16], double NT, double *pPosOffset){

    double matrixParent[16];
    LoadIdentity(matrixParent);
    
    if (mpCParentObj != NULL)
	if (mpCParentObj->mParentObjType == _PARENTOBJ_OBJ)
	    ((SceneObj_cls*)mpCParentObj)->GetMatrix16(matrixParent, NT);

    std::vector < double > VScale = GetPCParamGroup("Scale")->GetVInterpolatedValue(NT);

    double matrixLocal[16];
    double matrixScale[16], matrixRotate[16], matrixTranslate[16];
    double matrixRotTrans[16];
    LoadIdentity(matrixLocal);
    LoadIdentity(matrixScale);
    LoadIdentity(matrixRotate);
    LoadIdentity(matrixTranslate);
    LoadIdentity(matrixRotTrans);

    Scale(matrixScale, VScale.data());
    
    ApplyRotate(matrixRotate, NT);

    ApplyTranslate(matrixTranslate, NT);
    if (pPosOffset != NULL)
	Translate(matrixTranslate, pPosOffset);
    
    Multiply(matrixRotate, matrixRotTrans);
    Multiply(matrixTranslate, matrixRotTrans);

    ProcModifier("TransformRot", matrixRotTrans, matrixParent);

    Multiply(matrixScale, matrixLocal);
    Multiply(matrixRotTrans, matrixLocal);
    
    Multiply(matrixParent, matrixLocal);

    for (int i = 0; i < 16; i++)
	matrix[i] = matrixLocal[i];
}

void CLS::
ApplyTranslate(double matrix[16], double NT){
    std::vector < double > VPos = GetPCParamGroup("Position")->GetVInterpolatedValue(NT);
    Translate(matrix, VPos.data());
}

void CLS::
ApplyRotate(double matrix[16], double NT){
    std::vector < double > VRot = GetPCParamGroup("RotationEuler")->GetVInterpolatedValue(NT);
    
    double rotX = VRot[0] * 3.14159 / 180;
    double rotY = VRot[1] * 3.14159 / 180;
    double rotZ = VRot[2] * 3.14159 / 180;

    RotateX(matrix, rotX);
    RotateY(matrix, rotY);    
    RotateZ(matrix, rotZ);
}

void CLS::
ProcTexSpaceModBegin(){
    glMatrixMode(GL_TEXTURE);
    glLoadIdentity();
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    std::vector < double > VScale = GetPCParamGroup("Scale")->GetVInterpolatedValue(NT);
    
    glTranslated(+0.5, +0.5, +0.5);
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->GetStrAttributeName() == "TexSpaceMod_cls")
	    ((TexSpaceMod_cls*)pCAttrib)->Execute(VScale);
    }
    glTranslated(-0.5, -0.5, -0.5);
    glMatrixMode(GL_MODELVIEW);
}

void CLS::
ProcTexSpaceModEnd(){
    glMatrixMode(GL_TEXTURE);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
}

void CLS::
BasicGLMatrix(){
    
    double NT = GetPCTimeCtrl()->GetCurrentNT();

    double matrix[16];
    LoadIdentity(matrix);
    GetMatrix16(matrix, NT);

    glMultMatrixd(matrix);
}

void CLS::
GLDraw(bool bDrawGuideObject, bool bDrawLightCam, bool bDrawData,
       int &lightCurrent, int ensemble, int composit, int status,
       float baseMat16[16], float baseMat16Inv[16],
       float viewMat16[16], float viewMat16Inv[16],
       bool bHitTest){

    if (IsSameAs("Primitive_cls")){
	Primitive_cls *pCThis = (Primitive_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE) return; }
    if (IsSameAs("MeshObj_cls")){
	MeshObj_cls *pCThis = (MeshObj_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE) return; }
    if (IsSameAs("TextObj_cls")){
	TextObj_cls *pCThis = (TextObj_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE) return; }
    if (IsSameAs("HeightField_cls")){
	HeightField_cls *pCThis = (HeightField_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE) return; }
    if (IsSameAs("Streamline_cls")){
	Streamline_cls *pCThis = (Streamline_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE) return; }

    //CheckGLError(mstrClassName.char_str());
    
    glPushMatrix();
    BasicGLMatrix();

    if (IsSameAs("LightSource_cls") == FALSE &&
	lightCurrent == GL_LIGHT0)
	lightCurrent++; // Enable default light;

    int lightNum = lightCurrent - GL_LIGHT0;
    if (GetPCShadowVolumeCtrl()->mSVStep != 0) lightNum = 1;

    if (IsSameAs("Camera_cls"))
	((Camera_cls*)this)->GLDraw(bDrawLightCam, status);
    if (IsSameAs("LightSource_cls"))
	((LightSource_cls*)this)->
	    GLDraw(bDrawLightCam, lightCurrent, ensemble, status);
    if (IsSameAs("Primitive_cls"))
	((Primitive_cls*)this)->
	    GLDraw(bDrawGuideObject, lightNum, ensemble, composit, status,
		   bHitTest);
    if (IsSameAs("MeshObj_cls"))
	((MeshObj_cls*)this)->GLDraw(lightNum, ensemble, status);
    if (IsSameAs("TextObj_cls"))
	((TextObj_cls*)this)->GLDraw(lightNum, ensemble, status);
    if (IsSameAs("World_cls"))
	((World_cls*)this)->GLDraw(FALSE);

    if (bDrawData == TRUE){
	if (IsSameAs("VolumeBox_cls")){
	    VolumeBox_cls *pCThis = (VolumeBox_cls*)this;
	    std::vector < VolumeBox_cls* > VpCDepend = pCThis->GetVpCDependency();
	    int num = VpCDepend.size();
	    for (int i = 0; i < num; i++)
		VpCDepend[i]->PrepareDraw(ensemble, composit, FALSE);
	    
	    pCThis->PrepareDraw(ensemble, composit);
	    if (pCThis->mbSkipDrawing == FALSE)
		pCThis->GLDraw(bDrawGuideObject, lightNum, ensemble, composit, status);
	}
	if (IsSameAs("Isosurface_cls"))
	    ((Isosurface_cls*)this)->
		GLDraw(bDrawGuideObject, lightNum, ensemble, composit, status);
	if (IsSameAs("Particle_cls")){
	    Particle_cls *pCThis = (Particle_cls*)this;
	    std::vector < Particle_cls* > VpCDepend = pCThis->GetVpCDependency();
	    int num = VpCDepend.size();
	    for (int i = 0; i < num; i++)
		VpCDepend[i]->PrepareDraw(ensemble, composit, FALSE);
	    	    
	    pCThis->PrepareDraw(ensemble, composit);
	    if (pCThis->mbSkipDrawing == FALSE)
		pCThis->GLDraw(bDrawData, bDrawGuideObject, lightNum, ensemble, composit, status,
			       baseMat16, baseMat16Inv, viewMat16, viewMat16Inv);
	}
	if (IsSameAs("Streamline_cls")){
	    Streamline_cls *pCThis = (Streamline_cls*)this;
	    pCThis->PrepareDraw(ensemble, composit);
	pCThis->GLDraw(bDrawGuideObject, lightNum, ensemble, composit, status,
		       bHitTest);
	}
	if (IsSameAs("HeightField_cls"))
	    ((HeightField_cls*)this)->
		GLDraw(bDrawGuideObject, lightNum, ensemble, composit, status,
		       bHitTest);
    }
    
    mProcessedInInstanceLoop = 1;

    glPopMatrix();
}

void CLS::
GLDrawXray(bool bDrawGuideObject, bool bDrawLightCam, bool bDrawData,
	   int &lightCurrent, int ensemble, int composit, int status,
	   float baseMat16[16], float baseMat16Inv[16],
	   float viewMat16[16], float viewMat16Inv[16],
	   bool bHittest){

    int lightNum = lightCurrent - GL_LIGHT0;
    
    if (IsSameAs("LightSource_cls")){
	LightSource_cls *pCThis = (LightSource_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE){
	    glPushMatrix();
	    BasicGLMatrix();
	    pCThis->GLDrawXray(bDrawLightCam, status);
	    mProcessedInInstanceLoop = 1;
	    glPopMatrix();
	    
	}
    }

    if (IsSameAs("Primitive_cls")){
	Primitive_cls *pCThis = (Primitive_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE){
	    glPushMatrix();
	    BasicGLMatrix();
	    pCThis->GLDraw(bDrawGuideObject, lightNum, ensemble, composit,
			   status, bHittest);
	    mProcessedInInstanceLoop = 1;
	    glPopMatrix();
	}
    }
    
    if (IsSameAs("TextObj_cls")){
	TextObj_cls *pCThis = (TextObj_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE){
	    glPushMatrix();
	    BasicGLMatrix();
	    pCThis->GLDraw(lightNum, ensemble, status);
	    mProcessedInInstanceLoop = 1;
	    glPopMatrix();
	}
    }

    if (IsSameAs("Streamline_cls")){
	Streamline_cls *pCThis = (Streamline_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE){
	    glPushMatrix();
	    BasicGLMatrix();
	    pCThis->GLDraw(bDrawGuideObject, lightNum, ensemble, composit, status, bHittest);
	    mProcessedInInstanceLoop = 1;
	    glPopMatrix();
	}
    }

    if (IsSameAs("MeshObj_cls")){
	MeshObj_cls *pCThis = (MeshObj_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE){
	    glPushMatrix();
	    BasicGLMatrix();
	    ((MeshObj_cls*)this)->
		GLDraw(lightNum, ensemble, status);
	    mProcessedInInstanceLoop = 1;
	    glPopMatrix();
	}
    }

    if (IsSameAs("HeightField_cls")){
	HeightField_cls *pCThis = (HeightField_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE){
	    glPushMatrix();
	    BasicGLMatrix();
	    ((HeightField_cls*)this)->
		GLDraw(bDrawGuideObject, lightNum, ensemble, composit,
		       status, bHittest);
	    mProcessedInInstanceLoop = 1;
	    glPopMatrix();
	}
    }

    if (IsSameAs("World_cls"))
	((World_cls*)this)->GLDraw(TRUE);

    // ParticleXrayݒ́i܂j
    /*
    if (mstrClassName.IsSameAs("Particle_cls")){
	Particle_cls *pCThis = (Particle_cls*)this;
	if (pCThis->mCbXRay.GetBodyBOOL() == TRUE){
	    int lightNum = lightCurrent - GL_LIGHT0;
	    glPushMatrix();
	    BasicGLMatrix();
	    ((Particle_cls*)this)->
		GLDraw(bDrawGuideObject, lightNum, ensemble, composit, status, 
		       baseMat16, baseMat16Inv,
		       bHittest);
	    mProcessedInInstanceLoop = 1;
	    glPopMatrix();
	}
    }
    */
}

std::list< Attribute_cls* >::iterator CLS::
GetIterAttributeFromIndex(int index){
    int numAttrib = mVpCAttribute.size();
    if (index >= numAttrib || index < 0) return mVpCAttribute.end();
    std::list< Attribute_cls* >::iterator itr = mVpCAttribute.begin();
    for (int i = 0; i < index; i++){
	if (itr == mVpCAttribute.end()) return itr;
	itr++;
    }
    return itr;
}

std::list< Attribute_cls* >::iterator CLS::
GetIterAttributeFromPC(Attribute_cls *pCAttrib){
    return GetIterAttributeFromIndex(GetIndexModifier(pCAttrib));
}

/*
Attribute_cls* CLS::
GetPCAttributeFromName(wxString strName){
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	if ((*GetIterAttributeFromIndex(i))->GetStrName().IsSameAs(strName))
	    return *GetIterAttributeFromIndex(i);
    }
    return NULL;
}
*/

Attribute_cls* CLS::
GetPCAttributeFromAttribName(wxString strAttribName, int skip){
    int count = 0;
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	if (GetPCAttributeFromIndex(i)->GetStrAttributeName().IsSameAs(strAttribName)){
	    Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	    if (count >= skip)
		return pCAttrib;
	    count++;
	}
    }
    return NULL;
}

Attribute_cls* CLS::
GetPCAttribute(wxString strName, int skip){
    int count = 0;
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	wxString strAttribName = GetPCAttributeFromIndex(i)->GetStrName();
	if (GetPCAttributeFromIndex(i)->GetStrName().IsSameAs(strName)){
	    Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	    if (count >= skip)
		return pCAttrib;
	    count++;
	}
    }
    return NULL;
}

ModifierBase_cls* CLS::
GetPCValidModifier(wxString strName){
    int skip = 1;
    ModifierBase_cls *pCModifier = (ModifierBase_cls*)GetPCAttribute(strName);
    while (pCModifier != NULL){
	if (pCModifier->mCbValidity.GetBodyBOOL() == TRUE) break;
	if (pCModifier->mCbValidity.GetBodyBOOL() == FALSE){
	    pCModifier = (SequenceRemap_cls*)GetPCAttribute("SequenceRemap", skip);
	    skip++;
	}
    }
    return pCModifier;
}

/*
ModifierBase_cls* CLS::
GetPCAttributeFromIndex(int index){
    int numAttrib = mVpCAttribute.size();
    if (index >= numAttrib || index < 0) return NULL;
    return (ModifierBase_cls*)*GetIterAttributeFromIndex(index);
}
*/

std::list< Attribute_cls* >::iterator CLS::
GetIterDataProcFromIndex(int index){
    int numAttrib = mVpCDataProc.size();
    if (index >= numAttrib || index < 0) return mVpCDataProc.end();
    std::list< Attribute_cls* >::iterator itr = mVpCDataProc.begin();
    for (int i = 0; i < index; i++){
	if (itr == mVpCDataProc.end()) return itr;
	itr++;
    }
    return itr;
}

ModifierBase_cls* CLS::
GetPCDataProc(wxString strName, int skip){
    int count = 0;
    uint numAttrib = mVpCDataProc.size();
    for (uint i = 0; i < numAttrib; i++){
	if (GetPCDataProcFromIndex(i)->GetStrName().IsSameAs(strName)){
	    ModifierBase_cls *pCModifier = GetPCDataProcFromIndex(i);
	    if (count >= skip)
		return pCModifier;
	    count++;
	}
    }
    return NULL;
}

ModifierBase_cls* CLS::
GetPCDataProcFromIndex(int index){
    int numAttrib = mVpCDataProc.size();
    if (index >= numAttrib || index < 0) return NULL;
    return (ModifierBase_cls*)*GetIterDataProcFromIndex(index);
}

Attribute_cls* CLS::
GetPCAttributeFromIndex(int index){
    return *GetIterAttributeFromIndex(index);
}

int CLS::
GetIndexModifierInit(){
    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++){
	wxString strAttrib = GetPCAttributeFromIndex(i)->GetStrAttributeName();
	if (strAttrib.IsSameAs("Material_cls") == FALSE &&
	    strAttrib.IsSameAs("MaterialLight_cls") == FALSE &&
	    strAttrib.IsSameAs("MaterialParticle_cls") == FALSE &&
	    strAttrib.IsSameAs("MaterialStreamline_cls") == FALSE &&
	    strAttrib.IsSameAs("CameraProp_cls") == FALSE &&
	    strAttrib.IsSameAs("FocalBlurProp_cls") == FALSE &&
	    strAttrib.IsSameAs("StereoProp_cls") == FALSE &&
	    strAttrib.IsSameAs("FogProp_cls") == FALSE &&
	    strAttrib.IsSameAs("IsosurfProp_cls") == FALSE &&
	    strAttrib.IsSameAs("StreamlineProp_cls") == FALSE &&
	    strAttrib.IsSameAs("TextProp_cls") == FALSE &&
	    strAttrib.IsSameAs("ColorCurve_cls") == FALSE) return i;
    }
    return numAttribute;
}

int CLS::
GetIndexModifier(Attribute_cls* pCAttribute){
    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++)
	if (GetPCAttributeFromIndex(i) == pCAttribute) return i;
    return -1;
}

int CLS::
GetIndexDProc(Attribute_cls* pCAttribute){
    uint numAttribute = mVpCDataProc.size();
    for (uint i = 0; i < numAttribute; i++)
	if (GetPCDataProcFromIndex(i) == pCAttribute) return i;
    return -1;
}

int CLS::
GetNumAttribute(wxString strAttribName){
    int count = 0;
    uint numAttrib = mVpCAttribute.size();
    for (int i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	if (pCAttrib->GetStrAttributeName().IsSameAs(strAttribName))
	    count++;
    }
    return count;
}

bool CLS::
IsAttributesTimeDepend(){
    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++)
	if (GetPCAttributeFromIndex(i)->IsParamGroupTimeDepend() == TRUE)
	    return TRUE;
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++)
	if (GetPCDataProcFromIndex(i)->IsParamGroupTimeDepend() == TRUE)
	    return TRUE;
    return FALSE;
}

bool CLS::
IsParamGroupTimeDepend(){
    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++){
	if (mVpCParamGroup[i]->IsTimeDepend() == TRUE) return TRUE;
    }
    return FALSE;
}

bool CLS::
IsAttributeExist(wxString strName){
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	if (GetPCAttributeFromIndex(i)->GetStrName().IsSameAs(strName))
	    return TRUE;
    }
    return FALSE;
}

bool CLS::
IsDataProcExist(wxString strName){
    uint numAttrib = mVpCDataProc.size();
    for (uint i = 0; i < numAttrib; i++){
	if (GetPCDataProcFromIndex(i)->GetStrName().IsSameAs(strName))
	    return TRUE;
    }
    return FALSE;
}

bool CLS::
IsDataProcClsExist(wxString strName){
    uint numAttrib = mVpCDataProc.size();
    for (uint i = 0; i < numAttrib; i++){
	if (GetPCDataProcFromIndex(i)->GetStrAttributeName().IsSameAs(strName))
	    return TRUE;
    }
    return FALSE;
}

wxString CLS::
GetStrBaseName(){
    if (IsSameAs(wxT("Primitive_cls")))
	return ((Primitive_cls*)this)->GetStrInitName();
    if (IsSameAs(wxT("LightSource_cls")))
	return ((LightSource_cls*)this)->GetStrInitName();
    return wxES;
}

wxString CLS::
GetStrName(){ return mCstrName.mstrBody; }

wxString CLS::
GetStrTag(){
    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    return strTag;
}

wxString CLS::
GetStrTag(ParamGroup_cls *pCParamGroup){
    wxString strTag = GetStrTag();
    strTag << "::" << pCParamGroup->GetStrName();
    return strTag;
}

wxString CLS::
GetStrParentName(){ return mpCParentObj->GetStrName(); }

wxString CLS::
GetStrUniqueName(wxString strName){
    wxString strCandidate = strName;
    while (IsAttributeExist(strCandidate))
	strCandidate = GetStrNewCandidate(strCandidate);
    while (IsDataProcExist(strCandidate))
	strCandidate = GetStrNewCandidate(strCandidate);
    return strCandidate;
}

int CLS::
GetUsedMemoryMB(){
    if (IsSameAs("VolumeBox_cls") == TRUE)
	return ((VolumeBox_cls*)this)->GetUsedMemoryMB();
    if (IsSameAs("Particle_cls") == TRUE)
	return ((Particle_cls*)this)->GetUsedMemoryMB();
    return 0;
}

void CLS::
PrefetchData(){
    if (IsSameAs("VolumeBox_cls") == TRUE)
	((VolumeBox_cls*)this)->PrefetchData();
    if (IsSameAs("Particle_cls") == TRUE)
	((Particle_cls*)this)->PrefetchData();
    if (IsSameAs("Streamline_cls") == TRUE)
	((Streamline_cls*)this)->PrefetchData();
}

void CLS::
ReleaseMemoryData(){
    if (IsSameAs("VolumeBox_cls") == TRUE)
	((VolumeBox_cls*)this)->ReleaseMemoryData();
    if (IsSameAs("Particle_cls") == TRUE)
	((Particle_cls*)this)->ReleaseMemoryData();
}

void CLS::
ProcModifier(wxString strType, void *pTgt, void *pParam){
    if (strType.IsSameAs("TransformRot")){
	uint numAttrib = mVpCAttribute.size();
	int init = GetIndexModifierInit();
	for (uint i = init; i < numAttrib; i++) {
	    if (i == mGeneratedDepth){
		double *matrixParent = (double*)pParam;
		double matrixParentTmp[16];
		memcpy(matrixParentTmp, mModifierMatrix, sizeof(double)*16);
		Multiply(matrixParent, matrixParentTmp);
		memcpy(matrixParent, matrixParentTmp, sizeof(double)*16);
	    }
		
	    if (GetPCAttributeFromIndex(i)->GetStrAttributeName().IsSameAs("Tracking_cls"))
		
		((Tracking_cls*)GetPCAttributeFromIndex(i))->Execute((double*)pTgt, (double *)pParam);	    
	}
    }
}

void CLS::
DeselectKeyFrame(wxString strException){
    for (uint i = 0; i < mVpCObj.size(); i++){
	SceneObj_cls* pCObj = GetPCFromIndex(i);
	if (pCObj->GetStrName().IsSameAs(strException) == FALSE)
	    pCObj->DeselectKeyFrame(strException);
    }

    wxString strName = PopStrItemFromStrLine(strException, wxT("::"));
    if (GetStrName().IsSameAs(strName) == FALSE)
	strException.Clear();
    
    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->DeselectKey(strException);

    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++)
	GetPCAttributeFromIndex(i)->DeselectKeyFrame(strException);

    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++)
	GetPCDataProcFromIndex(i)->DeselectKeyFrame(strException);
}

void CLS::
AddModifierCommand(wxString strClassName, wxString strNameCandid){
    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    wxString strUndo(strTag), strRedo(strTag);
    wxString strName = GetStrUniqueName(strNameCandid);
    
    strUndo << "::Delete::Modifier::" << strName;    
    strRedo << "::New::Modifier::" << strClassName << "::" << strName;
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
DeleteModifierCommand(wxString strName){
    ASSERT(IsAttributeExist(strName) == TRUE);

    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    wxString strUndo2(strTag), strRedo2(strTag);
    wxString strUndo3(strTag), strRedo3(strTag);
    Attribute_cls *pCAttribute = GetPCAttribute(strName);
    wxString strClassName = pCAttribute->GetStrAttributeName();
    int index = GetIndexModifier(pCAttribute);
    ASSERT(index != -1);

    GetPCUndo()->OpenBlock("ModifierDelete");
    if (index != mVpCAttribute.size()-1){
	
	wxString strUndo1(strTag), strRedo1(strTag);
	strUndo1 << "::Move::Modifier::" << strName << "::To::" << index;
	strRedo1 << "::Skip";    

	GetPCUndo()->PushCommandPair(strUndo1, strRedo1);
	GetPCUndo()->Redo();
    }

    strUndo2 << "::"; pCAttribute->Serialize(strUndo2); strUndo2.Trim(TRUE);
    strRedo2 << "::Skip";
    GetPCUndo()->PushCommandPair(strUndo2, strRedo2);
    GetPCUndo()->Redo();
    
    strUndo3 << "::New::Modifier::" << strClassName << "::" << strName;
    strRedo3 << "::Delete::Modifier::" << strName;    

    GetPCUndo()->PushCommandPair(strUndo3, strRedo3);
    GetPCUndo()->Redo();
    
    GetPCUndo()->CloseBlock("ModifierDelete");
}

void CLS::
MoveModifierCommand(wxString strName, int indexTo){
    ASSERT(IsAttributeExist(strName) == TRUE);

    ModifierBase_cls *pCModifier = (ModifierBase_cls*)GetPCAttribute(strName);
    int indexFrom = GetIndexModifier(pCModifier);

    indexTo += GetIndexModifierInit();

    if (indexTo == indexFrom+1 || indexTo == indexFrom) return; 
    if (indexTo < indexFrom) indexFrom++;

    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    wxString strUndo(strTag), strRedo(strTag);
    strUndo << "::Move::Modifier::" << strName << "::To::" << indexFrom;
    strRedo << "::Move::Modifier::" << strName << "::To::" << indexTo;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetModifierValidityCommand(wxString strName, bool bValidity){
    ASSERT(IsAttributeExist(strName) == TRUE);

    ModifierBase_cls *pCModifier = (ModifierBase_cls*)GetPCAttribute(strName);
    
    if (pCModifier->mCbValidity.GetBodyBOOL() == bValidity) return;

    wxString strFrom, strTo;
    pCModifier->mCbValidity.SerializeSingleItem(strFrom);
    pCModifier->mCbValidity.SetBodyFromData((long long)bValidity);
    pCModifier->mCbValidity.SerializeSingleItem(strTo);

    GetPCUndo()->PushCommandPair(strFrom, strTo);
    GetPCUndo()->Redo(); 
}

void CLS::
AddDataProcCommand(wxString strClassName, wxString strNameCandid){
    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    wxString strUndo(strTag), strRedo(strTag);
    wxString strName = GetStrUniqueName(strNameCandid);
    
    strUndo << "::Delete::DataProc::" << strName;    
    strRedo << "::New::DataProc::" << strClassName << "::" << strName;
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
DeleteDataProcCommand(wxString strName){
    ASSERT(IsDataProcExist(strName) == TRUE);

    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    wxString strUndo2(strTag), strRedo2(strTag);
    wxString strUndo3(strTag), strRedo3(strTag);
    Attribute_cls *pCAttribute = GetPCDataProc(strName);
    wxString strClassName = pCAttribute->GetStrAttributeName();
    int index = GetIndexDProc(pCAttribute);
    ASSERT(index != -1);

    GetPCUndo()->OpenBlock("DProcDelete");
    if (index != mVpCAttribute.size()-1){
	
	wxString strUndo1(strTag), strRedo1(strTag);
	strUndo1 << "::Move::DataProc::" << strName << "::To::" << index;
	strRedo1 << "::Skip";    

	GetPCUndo()->PushCommandPair(strUndo1, strRedo1);
	GetPCUndo()->Redo();
    }

    strUndo2 << "::"; pCAttribute->Serialize(strUndo2); strUndo2.Trim(TRUE);
    strRedo2 << "::Skip";
    GetPCUndo()->PushCommandPair(strUndo2, strRedo2);
    GetPCUndo()->Redo();
    
    strUndo3 << "::New::DataProc::" << strClassName << "::" << strName;
    strRedo3 << "::Delete::DataProc::" << strName;    

    GetPCUndo()->PushCommandPair(strUndo3, strRedo3);
    GetPCUndo()->Redo();
    
    GetPCUndo()->CloseBlock("DProcDelete");
}

void CLS::
MoveDataProcCommand(wxString strName, int indexTo){
    ASSERT(IsDataProcExist(strName) == TRUE);

    ModifierBase_cls *pCModifier = GetPCDataProc(strName);
    int indexFrom = GetIndexDProc(pCModifier);

    if (indexTo == indexFrom+1 || indexTo == indexFrom) return; 
    if (indexTo < indexFrom) indexFrom++;

    wxString strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << GetStrName();
    wxString strUndo(strTag), strRedo(strTag);
    strUndo << "::Move::DataProc::" << strName << "::To::" << indexFrom;
    strRedo << "::Move::DataProc::" << strName << "::To::" << indexTo;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetDataProcValidityCommand(wxString strName, bool bValidity){
    ASSERT(IsDataProcExist(strName) == TRUE);

    ModifierBase_cls *pCModifier = GetPCDataProc(strName);
    
    if (pCModifier->mCbValidity.GetBodyBOOL() == bValidity) return;
    
    wxString strFrom, strTo;
    pCModifier->mCbValidity.SerializeSingleItem(strFrom);
    pCModifier->mCbValidity.SetBodyFromData((long long)bValidity);
    pCModifier->mCbValidity.SerializeSingleItem(strTo);

    GetPCUndo()->PushCommandPair(strFrom, strTo);
    GetPCUndo()->Redo(); 
}

void CLS::
AddMaterialCommand(){
    if (IsSameAs("Particle_cls") == TRUE)
	((Particle_cls*)this)->AddMaterialCommand();
    if (IsSameAs("Streamline_cls") == TRUE)
	((Streamline_cls*)this)->AddMaterialCommand();
}

bool CLS::
IsSelectedTLKeyExist(){
    bool bExist = FALSE;
    uint numChild = mVpCObj.size();
    for (uint i = 0; i < numChild; i++)
	bExist |= GetPCFromIndex(i)->IsSelectedTLKeyExist();

    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	bExist |= mVpCParamGroup[i]->IsSelectedKeyExist();

    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++)
	bExist |= GetPCAttributeFromIndex(i)->IsSelectedTLKeyExist();

    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++)
	bExist |= GetPCDataProcFromIndex(i)->IsSelectedTLKeyExist();

    return bExist;
}

void CLS::
DeleteSelectedTLKeyCommand(){
    uint numChild = mVpCObj.size();
    for (uint i = 0; i < numChild; i++)
	GetPCFromIndex(i)->DeleteSelectedTLKeyCommand();

    wxString strTag = GetStrTag();

    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->DeleteSelectedTLKeyCommand(strTag);

    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++)
	GetPCAttributeFromIndex(i)->DeleteSelectedTLKeyCommand(strTag);

    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++)
	GetPCDataProcFromIndex(i)->DeleteSelectedTLKeyCommand(strTag);
}

void CLS::
SetSelectedTLKeyTypeCommand(int keyType, double ease){
    uint numChild = mVpCObj.size();
    for (uint i = 0; i < numChild; i++)
	GetPCFromIndex(i)->SetSelectedTLKeyTypeCommand(keyType, ease);
    
    wxString strTag = GetStrTag();
    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->SetSelectedTLKeyTypeCommand(strTag, keyType, ease);

    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++)
	GetPCAttributeFromIndex(i)->SetSelectedTLKeyTypeCommand(strTag, keyType, ease);

    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++)
	GetPCDataProcFromIndex(i)->SetSelectedTLKeyTypeCommand(strTag, keyType, ease);
}

void CLS::
MoveSelectedTLKeyCommand(double dNX, double dY){
    uint numChild = mVpCObj.size();
    for (uint i = 0; i < numChild; i++)
	GetPCFromIndex(i)->MoveSelectedTLKeyCommand(dNX, dY);
    
    wxString strTag = GetStrTag();
    uint numParamGroup = mVpCParamGroup.size();
    for (uint i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->MoveSelectedTLKeyCommand(dNX, dY, strTag);

    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++)
	GetPCAttributeFromIndex(i)->MoveSelectedTLKeyCommand(strTag, dNX, dY);

    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++)
	GetPCDataProcFromIndex(i)->MoveSelectedTLKeyCommand(strTag, dNX, dY);
}

void CLS::
SetParamD3Command(wxString strParamGroup, double dValue[3]){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    if (pCParamGroup == NULL) return;
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    double newValue[3];
    for (int i = 0; i < 3; i++)
	newValue[i] = pCParamGroup->mVCParam[i].GetValue(NT) + dValue[i];
    SetParam3Command(strParamGroup, newValue);
}

void CLS::
SetParam3Command(wxString strParamGroup, double value[3]){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strParamGroup);
    if (pCParamGroup == NULL) return;
    double NT = GetPCTimeCtrl()->GetCurrentNT();

    for (int i = 0; i < 3; i++){
	double newValue = value[i];
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];
	if (Misc_cls::IsNearlySame(CParam.GetValue(NT), newValue, 1e-6) == FALSE){
	    wxString strTag = GetStrTag() << "::" << pCParamGroup->GetStrName() << "::" << CParam.GetStrName();
	    CParam.SetValueAtCommand(NT, newValue, strTag);
	}
    }
}

bool CLS::
ProcNewModifier(Attribute_cls *pCAttrib){
    mVpCAttribute.push_back(pCAttrib);
    GetPCMain()->ProcAfterMath("ModifierChange", GetStrName());
    return TRUE;
}

bool CLS::
ProcNewTracking(wxString strName){
    return ProcNewModifier(new Tracking_cls(strName, GetStrName())); }

bool CLS::
ProcNewArrayCopy(wxString strName){
    return ProcNewModifier(new ArrayCopy_cls(strName, GetStrName())); }

bool CLS::
ProcNewModifyApparentSize(wxString strName){
    return ProcNewModifier(new ModifyApparentSize_cls(strName, GetStrName())); }

bool CLS::
ProcNewModifyParticleSize(wxString strName){
    return ProcNewModifier(new ModifyParticleSize_cls(strName, GetStrName())); }

bool CLS::
ProcNewNestedVolume(wxString strName){
    return ProcNewModifier(new NestedVolume_cls(strName, GetStrName())); }

bool CLS::
ProcNewParticleBoolean(wxString strName){
    return ProcNewModifier(new ParticleBoolean_cls(strName, GetStrName())); }

bool CLS::
ProcNewParticleFiltering(wxString strName){
    return ProcNewModifier(new ParticleFiltering_cls(strName, GetStrName())); }

bool CLS::
ProcNewParticleTweak(wxString strName){
    return ProcNewModifier(new ParticleTweak_cls(strName, GetStrName())); }

bool CLS::
ProcNewTrimming(wxString strName){
    return ProcNewModifier(new Trimming_cls(strName, GetStrName())); }

bool CLS::
ProcNewTexSpaceMod(wxString strName){
    return ProcNewModifier(new TexSpaceMod_cls(strName, GetStrName())); }

bool CLS::
ProcNewSequenceRemapMod(wxString strName){
    return ProcNewModifier(new SequenceRemap_cls(strName, GetStrName())); }

bool CLS::
ProcNewVolumeMask(wxString strName){
    return ProcNewModifier(new VolumeMask_cls(strName, GetStrName())); }

bool CLS::
ProcNewVolumeCoord(wxString strName){
    return ProcNewModifier(new VolumeCoord_cls(strName, GetStrName())); }

bool CLS::
ProcNewDProc(Attribute_cls *pCAttrib){
    mVpCDataProc.push_back(pCAttrib);
    GetPCMain()->ProcAfterMath("DataProcChange", GetStrName());
    return TRUE;
}

bool CLS::
ProcNewVolumeBoolean(wxString strName){
    return ProcNewDProc(new VolumeBoolean_cls(strName, GetStrName())); }

bool CLS::
ProcNewVolumeDownsize(wxString strName){
    return ProcNewDProc(new VolumeDownsize_cls(strName, GetStrName())); }

bool CLS::
ProcNewParticleDownsize(wxString strName){
    return ProcNewDProc(new ParticleDownsizing_cls(strName, GetStrName())); }

bool CLS::
ProcNewParticleExclude(wxString strName){
    return ProcNewDProc(new ParticleExclude_cls(strName, GetStrName())); }

bool CLS::
ProcNewParticleMerge(wxString strName){
    return ProcNewDProc(new ParticleMerge_cls(strName, GetStrName())); }

bool CLS::
ProcNewStreamFiltering(wxString strName){
    return ProcNewDProc(new StreamFiltering_cls(strName, GetStrName())); }

bool CLS::
ProcNewStreamSetAlpha(wxString strName){
    return ProcNewDProc(new StreamSetAlpha_cls(strName, GetStrName())); }

bool CLS::
ProcNewStreamTransform(wxString strName){
    return ProcNewDProc(new StreamTransform_cls(strName, GetStrName())); }

bool CLS::
ProcNewMaterial(wxString strName){
    if (mstrClassName.IsSameAs("Particle_cls"))
	return ((Particle_cls*)this)->ProcNewMaterial(strName);
    if (mstrClassName.IsSameAs("Streamline_cls"))
	return ((Streamline_cls*)this)->ProcNewMaterial(strName);
    return FALSE;
}

bool CLS::
ProcNewColorCurve(wxString strName){
    if (mstrClassName.IsSameAs("Particle_cls"))
	return ((Particle_cls*)this)->ProcNewColorCurve(strName);
    return FALSE;
}

bool CLS::
ProcNew(wxString strFirstItem, wxString strLine){
    if (strFirstItem.IsSameAs("Modifier")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	if (strFirstItem.IsSameAs("Tracking_cls")) 
	    return ProcNewTracking(strLine);
	if (strFirstItem.IsSameAs("ArrayCopy_cls"))
	    return ProcNewArrayCopy(strLine);
	if (strFirstItem.IsSameAs("ModifyApparentSize_cls"))
	    return ProcNewModifyApparentSize(strLine);
	if (strFirstItem.IsSameAs("ModifyParticleSize_cls"))
	    return ProcNewModifyParticleSize(strLine);
	if (strFirstItem.IsSameAs("NestedVolume_cls"))
	    return ProcNewNestedVolume(strLine);
	if (strFirstItem.IsSameAs("ParticleBoolean_cls"))
	    return ProcNewParticleBoolean(strLine);
	if (strFirstItem.IsSameAs("ParticleFiltering_cls"))
	    return ProcNewParticleFiltering(strLine);
	if (strFirstItem.IsSameAs("ParticleTweak_cls"))
	    return ProcNewParticleTweak(strLine);
	if (strFirstItem.IsSameAs("Trimming_cls"))
	    return ProcNewTrimming(strLine);
	if (strFirstItem.IsSameAs("TexSpaceMod_cls"))
	    return ProcNewTexSpaceMod(strLine);
	if (strFirstItem.IsSameAs("SequenceRemap_cls"))
	    return ProcNewSequenceRemapMod(strLine);
	if (strFirstItem.IsSameAs("VolumeMask_cls"))
	    return ProcNewVolumeMask(strLine);
	if (strFirstItem.IsSameAs("VolumeCoord_cls")) 
	    return ProcNewVolumeCoord(strLine);
    }
    if (strFirstItem.IsSameAs("DataProc")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	if (strFirstItem.IsSameAs("VolumeBoolean_cls")) 
	    return ProcNewVolumeBoolean(strLine);
	if (strFirstItem.IsSameAs("VolumeDownsize_cls"))
	    return ProcNewVolumeDownsize(strLine);
	if (strFirstItem.IsSameAs("ParticleDownsizing_cls"))
	    return ProcNewParticleDownsize(strLine);
	if (strFirstItem.IsSameAs("ParticleExclude_cls"))
	    return ProcNewParticleExclude(strLine);
	if (strFirstItem.IsSameAs("ParticleMerge_cls"))
	    return ProcNewParticleMerge(strLine);
	if (strFirstItem.IsSameAs("StreamFiltering_cls"))
	    return ProcNewStreamFiltering(strLine);
	if (strFirstItem.IsSameAs("StreamSetAlpha_cls"))
	    return ProcNewStreamSetAlpha(strLine);
	if (strFirstItem.IsSameAs("StreamTransform_cls"))
	    return ProcNewStreamTransform(strLine);
    }
    if (strFirstItem.IsSameAs("ColorCurve")){
	return ProcNewColorCurve(strLine);
    }
    if (strFirstItem.IsSameAs("MaterialParticle_cls") ||
	strFirstItem.IsSameAs("MaterialStreamline_cls"))
	return ProcNewMaterial(strLine);
    
    return FALSE;
}

bool CLS::
ProcDelete(wxString strFirstItem, wxString strLine){
    if (strFirstItem.IsSameAs("Modifier")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	Attribute_cls *pCAttribute = GetPCAttribute(strFirstItem);
	if (pCAttribute == NULL) return FALSE;
	int indexTgt = GetIndexModifier(pCAttribute);
	if (indexTgt == -1) return FALSE;
	mVpCAttribute.erase(GetIterAttributeFromIndex(indexTgt));

	wxString strDeletedItem = GetStrName() << "::" << pCAttribute->GetStrName();
	GetPCMain()->ProcAfterMath("DeleteModifier", strDeletedItem);
	Attribute_cls::Delete(pCAttribute);
	GetPCMain()->ProcAfterMath("ModifierChange", GetStrName());
	return TRUE;
    }
    if (strFirstItem.IsSameAs("DataProc")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	Attribute_cls *pCAttribute = GetPCDataProc(strFirstItem);
	if (pCAttribute == NULL) return FALSE;
	int indexTgt = GetIndexDProc(pCAttribute);
	if (indexTgt == -1) return FALSE;
	mVpCDataProc.erase(GetIterDataProcFromIndex(indexTgt));
	
	wxString strDeletedItem = GetStrName() << "::" << pCAttribute->GetStrName();
	GetPCMain()->ProcAfterMath("DeleteDataProc", strDeletedItem);
	Attribute_cls::Delete(pCAttribute);
	GetPCMain()->ProcAfterMath("DataProcChange", GetStrName());
	return TRUE;
    }
    if (strFirstItem.IsSameAs("ColorCurve")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	Attribute_cls *pCAttribute = GetPCAttribute(strFirstItem);
	if (pCAttribute == NULL) return FALSE;
	int indexTgt = GetIndexModifier(pCAttribute);
	if (indexTgt == -1) return FALSE;
	mVpCAttribute.erase(GetIterAttributeFromIndex(indexTgt));
	wxString strDeletedItem = GetStrName() << "::" << pCAttribute->GetStrName();
	GetPCMain()->ProcAfterMath("DeleteColorCurve", strDeletedItem);
	Attribute_cls::Delete(pCAttribute);
	GetPCMain()->ProcAfterMath("ColorCurveChange", GetStrName());
	return TRUE;
    }
    if (strFirstItem.IsSameAs("MaterialParticle_cls")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	Attribute_cls *pCAttribute = GetPCAttribute(strFirstItem);
	if (pCAttribute == NULL) return FALSE;
	int indexTgt = GetIndexModifier(pCAttribute);
	if (indexTgt == -1) return FALSE;
	mVpCAttribute.erase(GetIterAttributeFromIndex(indexTgt));
	wxString strDeletedItem = GetStrName() << "::" << pCAttribute->GetStrName();
	Attribute_cls::Delete(pCAttribute);
	return TRUE;
    }
    if (strFirstItem.IsSameAs("MaterialStreamline_cls")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	Attribute_cls *pCAttribute = GetPCAttribute(strFirstItem);
	if (pCAttribute == NULL) return FALSE;
	int indexTgt = GetIndexModifier(pCAttribute);
	if (indexTgt == -1) return FALSE;
	mVpCAttribute.erase(GetIterAttributeFromIndex(indexTgt));
	wxString strDeletedItem = GetStrName() << "::" << pCAttribute->GetStrName();
	Attribute_cls::Delete(pCAttribute);
	return TRUE;
    }
    return FALSE;
}

bool CLS::
ProcMove(wxString strFirstItem, wxString strLine){
    if (strFirstItem.IsSameAs("Modifier")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	Attribute_cls *pCAttribute = GetPCAttribute(strFirstItem);
	if (pCAttribute == NULL) return FALSE;
	int indexTgt = GetIndexModifier(pCAttribute);
	if (indexTgt == -1) return FALSE;
	
	PopStrItemFromStrLine(strLine, "::");
	int indexTo = wxAtoi(strLine);

	mVpCAttribute.erase(GetIterAttributeFromIndex(indexTgt));
	if (indexTo > indexTgt) indexTo++;
	mVpCAttribute.insert(GetIterAttributeFromIndex(indexTo), pCAttribute);
	//GetPCMain()->ProcAfterMath("ModifierChange", GetStrName());
	return TRUE;
    }
    if (strFirstItem.IsSameAs("DataProc")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	Attribute_cls *pCAttribute = GetPCDataProc(strFirstItem);
	if (pCAttribute == NULL) return FALSE;
	int indexTgt = GetIndexDProc(pCAttribute);
	if (indexTgt == -1) return FALSE;

	PopStrItemFromStrLine(strLine, "::");
	int indexTo = wxAtoi(strLine);

	mVpCDataProc.erase(GetIterDataProcFromIndex(indexTgt));
	if (indexTo > indexTgt) indexTo++;
	mVpCDataProc.insert(GetIterDataProcFromIndex(indexTo), pCAttribute);
	return TRUE;
    }
    return FALSE;
}

void CLS::
RescaleParameters(double scale){
    ParentObj_cls::RescaleParametersRoot(scale);
    
    wxString strTag = GetStrTag();
    int numParamGroup = mVpCParamGroup.size();
    for (int i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->RescaleParametersCommand(scale, strTag);

    int numAttrib = mVpCAttribute.size();
    for (int i = 0; i < numAttrib; i++)
	(*GetIterAttributeFromIndex(i))->RescaleParametersCommand(strTag, scale);
}

bool CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){

    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE)
	return TRUE;

    if (IsSameAs("Isosurface_cls")){
	if (((Isosurface_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    if (IsSameAs("Primitive_cls")){
	if (((Primitive_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    if (IsSameAs("MeshObj_cls")){
	if (((MeshObj_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    if (IsSameAs("TextObj_cls")){
	if (((TextObj_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    if (IsSameAs("LightSource_cls")){
	if (((LightSource_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    if (IsSameAs("World_cls")){
	if (((World_cls*)this)->ReadStrLineFog(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    if (IsSameAs("VolumeBox_cls")){
	if (((VolumeBox_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    if (IsSameAs("Particle_cls")){
	if (((Particle_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
	if (((Particle_cls*)this)->ReadStrLineFormat(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }

    if (IsSameAs("Streamline_cls")){
	if (((Streamline_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
	if (((Streamline_cls*)this)->ReadStrLineFormat(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }


    if (IsSameAs("HeightField_cls")){
	if (((HeightField_cls*)this)->ReadStrLineMaterial(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    
    uint numAttribute = mVpCAttribute.size();
    for (int i = 0; i < numAttribute; i++){
	Attribute_cls* pCAttribute = GetPCAttributeFromIndex(i);
	if (pCAttribute->GetStrAttributeName().IsSameAs("ColorCurve_cls")){
	    if (((ColorCurve_cls*)pCAttribute)->ReadStrLine(istream, strFirstItem, strLine) == TRUE)
		return TRUE;
	}
	else{
	    if (pCAttribute->ReadStrLine(istream, strFirstItem, strLine) == TRUE)
		return TRUE;
	}
    }
    uint numDataProc = mVpCDataProc.size();
    for (int i = 0; i < numDataProc; i++){
	if (GetPCDataProcFromIndex(i)->ReadStrLine(istream, strFirstItem, strLine) == TRUE)
	    return TRUE;
    }
    
    // 1sR}h
    ParamGroup_cls* pCParamGroup = GetPCParamGroup(strFirstItem, strLine);
    if (pCParamGroup != NULL){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	if (strFirstItem.IsSameAs(pCParamGroup->mCstrName.GetBodySTR()))
	    strFirstItem = PopStrItemFromStrLine(strLine, "::");
	return pCParamGroup->ReadStrOneLiner(strFirstItem, strLine);
    }

    if (strFirstItem.IsSameAs("New")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	if (ProcNew(strFirstItem, strLine) == TRUE) return TRUE;
    }

    if (strFirstItem.IsSameAs("Delete")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	return ProcDelete(strFirstItem, strLine);
    }

    if (strFirstItem.IsSameAs("Move")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	return ProcMove(strFirstItem, strLine);
    }
    
    // Sp[^ǂ
    if (strFirstItem.IsSameAs("ReadParamGroup")){
	wxString paramGroupName = PopStrItemFromStrLine(strLine, "::");
	pCParamGroup = GetPCParamGroup(paramGroupName);
	if (pCParamGroup != NULL){
	    
	    std::wstring wstrLine;
	    while (std::getline(istream, wstrLine)){
		wxString strLine = wstrLine.c_str();
		wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
		pCParamGroup->ReadStrLine(istream, strFirstItem, strLine);
		
		if (strFirstItem.IsSameAs("ReadParamGroup")){
		    ASSERT(PopStrItemFromStrLine(strLine, wxT("::")).IsSameAs(wxT("End")));
		    break;
		}
	    }
	    return TRUE;
	}
    }

    return FALSE;
}

void CLS::
SerializePropertiesOnly(wxString& strOut){
    strOut << "SceneObjCtrl::ReadProperties::" << mCstrName.mstrBody << ENDL;
    ConfListContainer_cls::Serialize(strOut);

    if (IsSameAs("Isosurface_cls")) 
	((Isosurface_cls*)this)->SerializeMaterial(strOut);
    if (IsSameAs("Primitive_cls"))
	((Primitive_cls*)this)->SerializeMaterial(strOut);
    if (IsSameAs("MeshObj_cls"))
	((MeshObj_cls*)this)->SerializeMaterial(strOut);
    if (IsSameAs("TextObj_cls"))
	((TextObj_cls*)this)->SerializeMaterial(strOut);
    if (IsSameAs("LightSource_cls"))
	((LightSource_cls*)this)->SerializeMaterial(strOut);
    if (IsSameAs("World_cls"))
	((World_cls*)this)->SerializeFog(strOut);
    if (IsSameAs("VolumeBox_cls"))
	((VolumeBox_cls*)this)->SerializeMaterial(strOut);
    if (IsSameAs("HeightField_cls"))
	((HeightField_cls*)this)->SerializeMaterial(strOut);
    if (IsSameAs("Particle_cls")){
	((Particle_cls*)this)->SerializeMaterial(strOut);
	((Particle_cls*)this)->SerializeFormat(strOut);
    }
    if (IsSameAs("Streamline_cls")){
	((Streamline_cls*)this)->SerializeMaterial(strOut);
	((Streamline_cls*)this)->SerializeFormat(strOut);
    }
    
    uint numAttribute = mVpCAttribute.size();
    for (uint i = 0; i < numAttribute; i++){
	Attribute_cls *pCAttr = GetPCAttributeFromIndex(i);
	if (pCAttr->IsAttributeSameAs("Material_cls") == FALSE &&
	    pCAttr->IsAttributeSameAs("ColorCurve_cls") == FALSE){
	    strOut << "New::Modifier::" << pCAttr->GetStrAttributeName() << "::" << pCAttr->GetStrName() << ENDL;
	    pCAttr->Serialize(strOut);
	}

	if (pCAttr->IsAttributeSameAs("ColorCurve_cls") == TRUE){
	    strOut << "New::ColorCurve::" << pCAttr->GetStrName() << ENDL;
	    ((ColorCurve_cls*)pCAttr)->Serialize(strOut);
	}
    }
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCAttr = GetPCDataProcFromIndex(i);
	strOut << "New::DataProc::" << pCAttr->GetStrAttributeName() << "::" << pCAttr->GetStrName() << ENDL;
	pCAttr->Serialize(strOut);
    }
    
    int numParamGroup = mVpCParamGroup.size();
    for (int i = 0; i < numParamGroup; i++)
	mVpCParamGroup[i]->Serialize(strOut);

    strOut << "ReadProperties::End";
}

void CLS::
Serialize(wxString& strOut){
    strOut << "SceneObjCtrl::New::" << mstrClassName << "::ANY::" << mCstrName.mstrBody << ENDL;
    SerializePropertiesOnly(strOut); strOut << ENDL;
    for (int i = 0; i < mVpCObj.size(); i++){
	SceneObj_cls *pCChild = GetPCFromIndex(i);
	pCChild->Serialize(strOut);
	strOut << "SceneObjCtrl::SetToChild::" << pCChild->mCstrName.mstrBody << "::To::" << mCstrName.mstrBody << ENDL;
    }
}

void CLS::
Refer(){
    mReferenceCounter++;
}

void CLS::
ReferRelease(){
    mReferenceCounter--;
}

bool operator<(const SceneObj_cls& a, const SceneObj_cls& b){
    return FALSE;
}

bool operator>(const SceneObj_cls& a, const SceneObj_cls& b){

    if (a.mSortOrder > b.mSortOrder) return TRUE;

    return FALSE;
}

bool SceneObj_cls::
ComparePointer(const SceneObj_cls* a, const SceneObj_cls* b){
    return *a > *b;
}

// -----

#define CLS SceneObjInstance_cls

CLS::
SceneObjInstance_cls(SceneObj_cls *pCSceneObj){
    mpCSceneObjLink = pCSceneObj;
    LoadIdentity(mMatrix);
    mStatus = 0;
    mModifierDepth = 0;
}

void CLS::
GLDraw(bool bDrawGuideObject, bool bDrawLightCam, bool bDrawData,
       int &lightCurrent, int ensemble, int composit, 
       float baseMat16[16], float baseMat16Inv[16],
       float viewMat16[16], float viewMat16Inv[16],
       bool bHittest){

    memcpy(mpCSceneObjLink->mModifierMatrix, mMatrix, sizeof(double)*16);
    mpCSceneObjLink->mGeneratedDepth = mModifierDepth;
    mpCSceneObjLink->GLDraw(bDrawGuideObject, bDrawLightCam, bDrawData,
			    lightCurrent, ensemble, composit, mStatus,
			    baseMat16, baseMat16Inv,
			    viewMat16, viewMat16Inv,
			    bHittest);
}

void CLS::
GLDrawXray(bool bDrawGuideObject, bool bDrawLightCam, bool bDrawData,
	   int &lightCurrent, int ensemble, int composit, 
	   float baseMat16[16], float baseMat16Inv[16],
	   float viewMat16[16], float viewMat16Inv[16],
	   bool bHittest){

    memcpy(mpCSceneObjLink->mModifierMatrix, mMatrix, sizeof(double)*16);
    mpCSceneObjLink->mGeneratedDepth = mModifierDepth;
    mpCSceneObjLink->GLDrawXray(bDrawGuideObject, bDrawLightCam, bDrawData,
				lightCurrent, ensemble, composit, mStatus,
				baseMat16, baseMat16Inv,
				viewMat16, viewMat16Inv,
				bHittest);
}

void CLS::
SetDistance(Position_cls mCPos, double NT){

    memcpy(mpCSceneObjLink->mModifierMatrix, mMatrix, sizeof(double)*16);
	
    std::vector < double > VPosGlobal = mpCSceneObjLink->GetPositionGlobal(NT);
    
    mDist2ForSort = 
	(VPosGlobal[X]-mCPos.mPos[X])*(VPosGlobal[X]-mCPos.mPos[X]) +
	(VPosGlobal[Y]-mCPos.mPos[Y])*(VPosGlobal[Y]-mCPos.mPos[Y]) +
	(VPosGlobal[Z]-mCPos.mPos[Z])*(VPosGlobal[Z]-mCPos.mPos[Z]);
}

bool operator<(const SceneObjInstance_cls& a, const SceneObjInstance_cls& b){
    return FALSE;
}

bool operator>(const SceneObjInstance_cls& a, const SceneObjInstance_cls& b){
    wxString AstrComp[6] = {"LightSource_cls", "VolumeBox_cls", "Particle_cls", "Streamline_cls", "World_cls", "Text_cls"};

    /*
    if (a.mpCSceneObjLink->mstrClassName.IsSameAs("World_cls") &&
	!(b.mpCSceneObjLink->mstrClassName.IsSameAs("World_cls")) ||
	!a.mpCSceneObjLink->mstrClassName.IsSameAs("World_cls") &&
	(b.mpCSceneObjLink->mstrClassName.IsSameAs("World_cls")))
	return SceneObj_cls::ComparePointer(a.mpCSceneObjLink, b.mpCSceneObjLink);
    */
    for (int i = 0; i < 6; i++)
    if (a.mpCSceneObjLink->mstrClassName.IsSameAs(AstrComp[i]) &&
	!(b.mpCSceneObjLink->mstrClassName.IsSameAs(AstrComp[i])) ||
	!a.mpCSceneObjLink->mstrClassName.IsSameAs(AstrComp[i]) &&
	(b.mpCSceneObjLink->mstrClassName.IsSameAs(AstrComp[i])))
	return SceneObj_cls::ComparePointer(a.mpCSceneObjLink, b.mpCSceneObjLink);
    
    if ((a.mDist2ForSort) > (b.mDist2ForSort)) return TRUE;
    return FALSE;
}

bool CLS::
ComparePointer(const SceneObjInstance_cls* a, const SceneObjInstance_cls* b){
    return *a > *b;
}
