// (setq c-basic-offset 4)
// "clsSceneObjCtrl.h"

#include "clsSceneObjCtrl.h"
#include "clsSceneObj.h"

#include "clsCamera.h"
#include "clsLightSource.h"
#include "clsMeshObj.h"
#include "clsPrimitive.h"
#include "clsTextObj.h"
#include "clsWorld.h"
#include "DATAVIEWER/clsVolumeBox.h"
#include "DATAVIEWER/clsIsosurface.h"
#include "DATAVIEWER/clsParticle.h"
#include "DATAVIEWER/clsStreamline.h"
#include "DATAVIEWER/clsHeightField.h"

#include "../FILTER/clsFilterCtrl.h"
#include "../GLCANVAS/NAVIGATORS/clsHitTest.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../UNDO/clsUndoManager.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include <algorithm>

#define CLS SceneObjCtrl_cls

CLS::
SceneObjCtrl_cls() : ParentObj_cls(_PARENTOBJ_ROOT) {
    SetToDefault();
}

CLS::
~SceneObjCtrl_cls(){ Clear(); }

void CLS::
SetToDefault(){
   
    mpCActiveObject = NULL;
    mCConfList.Clear();

    mCConfList.SetTag("SceneObjCtrl");
    mCConfList.Add(&mCstrActiveItemName, "DUMMY");
}

void CLS::
Clear(){
    ClearSelection();
    Select(wxES, FALSE, FALSE);
    ParentObj_cls::Clear();
}

void CLS::
GLDrawObjects(bool bDrawGuideObject, bool bDrawLightCam, bool bDrawData,
	      int &lightCurrent, int ensemble, int composit,
	      PositionAndDirection_cls &CCamPos, 
	      PositionAndDirection_cls &CCamPosNext,
	      float baseMat16[16], float baseMat16Inv[16],
	      float viewMat16[16], float viewMat16Inv[16],
	      Hittest_cls *pCHitTest, bool bHittest){
    
    ClearProcessedMarkInInstanceLoop();
    InitializeStatusForDrawing();
    CheckStatusForDrawing();

    std::vector < SceneObj_cls* > VpCObj;
    GetVPCList(VpCObj);

    uint numObj = VpCObj.size();
    //`XLbv闱qn}[N
    /*
    for (uint i = 0; i < numObj; i++){
	if (VpCObj[i]->IsSameAs("Particle_cls")){
	    Particle_cls *pCParticle = (Particle_cls*)VpCObj[i];
	    if (pCParticle->mCstrName.GetBodySTR().IsSameAs("Skip"))
		pCParticle->mbSkipDrawing = TRUE;
	    else
		pCParticle->mbSkipDrawing = FALSE;
	}
    }
    */

    std::vector < SceneObjInstance_cls* > VpCInstance;
    for (uint i = 0; i < numObj; i++){
	int status = GetStatus(VpCObj[i]);
	std::vector < SceneObjInstance_cls* > VpCInstanceEach = VpCObj[i]->GetVNewPCInstance();
	uint numInstance = VpCInstanceEach.size();
	for (uint i = 0; i < numInstance; i++){
	    SceneObjInstance_cls *pCInstance = VpCInstanceEach[i];
	    pCInstance->mStatus = status;
	    VpCInstance.push_back(pCInstance);
	}
    }

    uint numInstance = VpCInstance.size();
    pCHitTest->ClearReferTable();

    // CX^XPʂŃ\[g
    // ArrayŃRs[ꂽCX^Ẍʒu̔f͂܂
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    for (int i = 0; i < numInstance; i++)
	VpCInstance[i]->SetDistance(CCamPos, NT);
    std::sort(VpCInstance.begin(), VpCInstance.end(), SceneObjInstance_cls::ComparePointer);

    CheckGLError("SceneObjCtrl::GLDrawBegin");
    
    unsigned int hittestCurrent = 0;
    for (uint i = 0; i < numInstance; i++){
	SceneObj_cls *pCObj = VpCInstance[i]->mpCSceneObjLink;
	if (pCObj->mstrClassName.IsSameAs("Particle_cls")){
	    ((Particle_cls*)pCObj)->mCCamPos = CCamPos;
	    ((Particle_cls*)pCObj)->mCCamPosNext = CCamPosNext;
	}

	glPushName(hittestCurrent);
	VpCInstance[i]->GLDraw(bDrawGuideObject, bDrawLightCam, bDrawData,
			       lightCurrent, ensemble, composit, 
			       baseMat16, baseMat16Inv,
			       viewMat16, viewMat16Inv,
			       bHittest);
	glPopName();
	pCHitTest->SetReferTable(VpCInstance[i]->mpCSceneObjLink);
	hittestCurrent++;
    }

    CheckGLError("SceneObjCtrl::GLDrawEnd_XRayBegin");

    glDisable(GL_DEPTH_TEST);
    for (uint i = 0; i < numInstance; i++){
	glPushName(hittestCurrent);
	VpCInstance[i]->GLDrawXray(bDrawGuideObject, bDrawLightCam, bDrawData,
				   lightCurrent, ensemble, composit,
				   baseMat16, baseMat16Inv,
				   viewMat16, viewMat16Inv,
				   bHittest);
	glPopName();
	pCHitTest->SetReferTable(VpCInstance[i]->mpCSceneObjLink);
	hittestCurrent++;
	delete VpCInstance[i];
    }
    glEnable(GL_DEPTH_TEST);

    CheckGLError("SceneObjCtrl::GLDrawEnd_XRayEnd");
}

void CLS::
ClearSelection(){
    for (uint i = 0; i < mVpCSelectedObject.size(); i++)
	mVpCSelectedObject[i]->ReferRelease();
    mVpCSelectedObject.clear();
}

wxString CLS::
GetStrSelf(){
    return "SceneObjCtrl::";
}

wxString CLS::
GetActiveObjectName(){
    if (mpCActiveObject == NULL) return wxES;
    return mpCActiveObject->GetStrName();
}

ConfItem_cls* CLS::
GetPCConfItem(wxString strTag){
    wxString strName = PopStrItemFromStrLine(strTag, "::");
    if (strName.IsSameAs("SceneObjCtrl")){
	strName = PopStrItemFromStrLine(strTag, "::");
	ASSERT(strName.IsSameAs("Obj"));
	strName = PopStrItemFromStrLine(strTag, "::");
    }
    if (strName.IsSameAs("SceneObj")){
	strName = PopStrItemFromStrLine(strTag, "::");
    }

    SceneObj_cls *pCObj = GetPCFromName(strName);
    if (pCObj == NULL) return NULL;
    wxString strFirstItem = PopStrItemFromStrLine(strTag, wxT("::"));

    if (strTag.IsEmpty() == TRUE)
	return pCObj->GetPCItem(strFirstItem);
    
    return pCObj->GetPCItemInAttrib(strFirstItem, strTag);
}

ParamGroup_cls* CLS::
GetPCParamGroup(wxString strTag){
    wxString strName = PopStrItemFromStrLine(strTag, "::");
    if (strName.IsSameAs("SceneObjCtrl")){
	strName = PopStrItemFromStrLine(strTag, "::");
	ASSERT(strName.IsSameAs("Obj"));
	strName = PopStrItemFromStrLine(strTag, "::");
    }
    
    SceneObj_cls *pCObj = GetPCFromName(strName);
    if (pCObj == NULL) return NULL;
    wxString strFirstItem = PopStrItemFromStrLine(strTag, wxT("::"));
    return pCObj->GetPCParamGroup(strFirstItem, strTag);
}

Parameter_cls* CLS::
GetPCParameter(wxString strTag){
    wxString strName = PopStrItemFromStrLine(strTag, wxT("::"));
    SceneObj_cls *pCObj = GetPCFromName(strName);
    if (pCObj == NULL) return NULL;
    wxString strFirstItem = PopStrItemFromStrLine(strTag, wxT("::"));
    ParamGroup_cls *pCParamGroup = pCObj->GetPCParamGroup(strFirstItem, strTag);
    if (pCParamGroup == NULL) return NULL;

    if (strTag.Find(pCParamGroup->mCstrName.GetBodySTR()+"::") == 0)
	strTag.Replace(pCParamGroup->mCstrName.GetBodySTR()+"::", wxES);

    return pCParamGroup->GetPCParameter(strTag);
}

int CLS::
GetStatus(SceneObj_cls *pCObj){
    if (pCObj == mpCActiveObject) return 2;
    if (IsSelected(pCObj) == TRUE) return 1;
    return 0;
}

bool CLS::
IsSelected(SceneObj_cls* pCObj){
    uint numObj = mVpCSelectedObject.size();
    for (uint i = 0; i < numObj; i++)
	if (mVpCSelectedObject[i] == pCObj) return TRUE;
    return FALSE;
}

void CLS::
RemoveSelectedObject(int index){
    for (int i = index; i < mVpCSelectedObject.size() - 1; i++){
	mVpCSelectedObject[i] = mVpCSelectedObject[i+1];
    }
    mVpCSelectedObject.pop_back();
}

void CLS::
Detatch(SceneObj_cls *pCObj){
    if (mpCActiveObject == pCObj){
	mpCActiveObject->ReferRelease();
	mpCActiveObject = NULL;
    }
    for (uint i = 0; i < mVpCSelectedObject.size(); i++){
	if (mVpCSelectedObject[i] == pCObj){
	    mVpCSelectedObject[i]->ReferRelease();
	    RemoveSelectedObject(i);
	    i--;
	}
    }
}

void CLS::
Select(wxString strObjName, bool bAdditional, bool bUpdateGUI, bool bToggle){
    
    if (strObjName.IsEmpty() == TRUE){
	if (mpCActiveObject != NULL) mpCActiveObject->ReferRelease();
	mpCActiveObject = NULL;
    }
    else{
	SceneObj_cls *pCPrev = mpCActiveObject;
	
	if (mpCActiveObject != NULL) mpCActiveObject->ReferRelease();
	mpCActiveObject = GetPCFromName(strObjName);
	if (mpCActiveObject != NULL) mpCActiveObject->Refer();

	if (bToggle == FALSE)
	    if (mpCActiveObject == pCPrev) bAdditional = TRUE;
	
	if (bToggle == TRUE)
	    if (mpCActiveObject == pCPrev){
		mpCActiveObject->ReferRelease();
		Deselect(mpCActiveObject->GetStrName(), bUpdateGUI);
		mpCActiveObject = NULL;
	    }
    }

    if (bAdditional == FALSE) ClearSelection();

    if (mpCActiveObject != NULL && IsSelected(mpCActiveObject) == FALSE){
	mVpCSelectedObject.push_back(mpCActiveObject);
	mpCActiveObject->Refer();
	wxString strTagLong = GetPCSceneObjCtrl()->GetStrSelf() << wxT("Obj::") << strObjName;
	GetPCMain()->ProcAfterMath("SelectObject", strTagLong);
    }

    if (bUpdateGUI == TRUE)
	GetPCMain()->UpdatePanels("OUTLINER::TIMELINE::VIEW3D");
}

void CLS::
Deselect(wxString strObjName, bool bUpdateGUI){

    SceneObj_cls *pCObj = GetPCFromName(strObjName);
    if (pCObj != NULL && IsSelected(pCObj) == TRUE){
	mVpCSelectedObject.erase(std::remove(mVpCSelectedObject.begin(), mVpCSelectedObject.end(), pCObj), mVpCSelectedObject.end());
	pCObj->ReferRelease();
    }

    if (bUpdateGUI == TRUE)
	GetPCMain()->UpdatePanels("OUTLINER::TIMELINE");
}

void CLS::
DeselectKeyFrame(wxString strException){
    RemoveTagHeader(strException);
    for (uint i = 0; i < mVpCObj.size(); i++){
	SceneObj_cls* pCObj = GetPCFromIndex(i);
	if (pCObj->GetStrName().IsSameAs(strException) == FALSE)
	    pCObj->DeselectKeyFrame(strException);
    }
}

void CLS::
SelectActitiveObjectOnly(){
    ClearSelection();
    if (mpCActiveObject == NULL) return;
    mVpCSelectedObject.push_back(mpCActiveObject);
    
    GetPCMain()->UpdatePanels("OUTLINER::TIMELINE");
}

bool CLS::
ChangeKeyTypeByModKey(wxString strTag, int keyIndex, bool bCtrl, bool bShift, bool bAlt){
    if (bCtrl == FALSE && bShift == FALSE && bAlt == FALSE) return FALSE;

    RemoveTagHeader(strTag);
    Parameter_cls *pCParam = GetPCParameter(strTag);

    if (pCParam != NULL){
	strTag = GetStrSelf() << "Obj::" << strTag;
	return pCParam->ChangeKeyTypeByModKey(strTag, keyIndex, bCtrl, bShift, bAlt);
    }
    return FALSE;
}

bool CLS::
SelectKeyFrame(wxString strTag, int keyIndex, bool bSelect, bool bAdditional,
	       double resolution){
    RemoveTagHeader(strTag);
    Parameter_cls *pCParam = GetPCParameter(strTag);
    if (pCParam != NULL) 
	return pCParam->SelectKeyByIndex(keyIndex, bSelect, bAdditional,
					 resolution);
    return FALSE;
}

void CLS::
RemoveTagHeader(wxString& strTag){
    wxString strAttempt = strTag.Clone();
    wxString strHead = PopStrItemFromStrLine(strAttempt, "::");
    if (strHead.IsSameAs("SceneObjCtrl")){
	strHead = PopStrItemFromStrLine(strAttempt, "::");
	ASSERT(strHead.IsSameAs("Obj"));
	strTag = strAttempt;
    }
}

wxString CLS::
GetStrUniqueName(wxString strName){
    wxString strCandidate = strName;
    while(IsObjExist(strCandidate))
	strCandidate = GetStrNewCandidate(strCandidate);
    return strCandidate;
}

void CLS::
SortSelectionList(){
    std::vector < SceneObj_cls* > VpCNewList;
    ProcessSortSelection(VpCNewList, mVpCSelectedObject);
    mVpCSelectedObject = VpCNewList;
}

bool CLS::
IsSelectedTLKeyExist(){
    bool bExit = FALSE;
    uint num = mVpCObj.size();
    for (uint i = 0; i < num; i++)
	bExit |= GetPCFromIndex(i)->IsSelectedTLKeyExist();
    return bExit;
}

void CLS::
DeleteSelectedTLKeyCommand(){
    GetPCUndo()->OpenBlock("DeleteSelectedTLKey");
    uint num = mVpCObj.size();
    for (uint i = 0; i < num; i++)
	GetPCFromIndex(i)->DeleteSelectedTLKeyCommand();
    GetPCUndo()->CloseBlock("DeleteSelectedTLKey");
}

void CLS::
SetSelectedTLKeyTypeCommand(int keyType, double ease){
    GetPCUndo()->OpenBlock("SetSelectedTLKeyType");
    uint num = mVpCObj.size();
    for (uint i = 0; i < num; i++)
	GetPCFromIndex(i)->SetSelectedTLKeyTypeCommand(keyType, ease);
    GetPCUndo()->CloseBlock("SetSelectedTLKeyType");
}

void CLS::
MoveSelectedTLKeyCommand(double dNX, double dY){
    GetPCUndo()->OpenBlock("MoveSelectedTLKey");
    uint num = mVpCObj.size();
    for (uint i = 0; i < num; i++)
	GetPCFromIndex(i)->MoveSelectedTLKeyCommand(dNX, dY);
    GetPCUndo()->CloseBlock("MoveSelectedTLKey");
}

void CLS::
SetVisibleMultipleCommand(bool bVisibleNew){
    SortSelectionList();
    GetPCUndo()->OpenBlock("SetVisible");
    for (uint i = 0; i < mVpCSelectedObject.size(); i++){
	SceneObj_cls *pCTgt = mVpCSelectedObject[i];

	bool bVisibleOrg = pCTgt->mCbVisible.GetBodyBOOL();
	if (bVisibleOrg == bVisibleNew) continue;

	wxString strData; strData << bVisibleNew;
	pCTgt->mCbVisible.PushCommandPair(strData);
    }

    GetPCUndo()->CloseBlock("SetVisible");
}

void CLS::
ToggleVisibleCommand(wxString strName){
    SceneObj_cls *pCTgt = GetPCFromName(strName);
    if (pCTgt == NULL) return ;

    if (IsSelected(pCTgt) == TRUE) {
	bool bVisibleOrg = pCTgt->mCbVisible.GetBodyBOOL();
	bool bVisibleNew = !bVisibleOrg;
	return SetVisibleMultipleCommand(bVisibleNew);
    }

    bool bVisibleOrg = pCTgt->mCbVisible.GetBodyBOOL();
    wxString strData; strData << !bVisibleOrg;
    pCTgt->mCbVisible.PushCommandPair(strData);
}

void CLS::
SetValidMultipleCommand(bool bValidNew){
    SortSelectionList();
    GetPCUndo()->OpenBlock("SetValid");
    for (uint i = 0; i < mVpCSelectedObject.size(); i++){
	SceneObj_cls *pCTgt = mVpCSelectedObject[i];

	if (pCTgt->mstrClassName.IsSameAs("LightSource_cls") == FALSE) continue;

	bool bValidOrg = ((LightSource_cls*)pCTgt)->mCbValid.GetBodyBOOL();
	if (bValidOrg == bValidNew) continue;

	wxString strData; strData << bValidNew;
	((LightSource_cls*)pCTgt)->mCbValid.PushCommandPair(strData);
    }

    GetPCUndo()->CloseBlock("SetValid");
}

void CLS::
ToggleValidCommand(wxString strName){
    SceneObj_cls *pCTgt = GetPCFromName(strName);
    if (pCTgt == NULL) return;

    if (IsSelected(pCTgt) == TRUE) {
	bool bValidOrg = ((LightSource_cls*)pCTgt)->mCbValid.GetBodyBOOL();
	bool bValidNew = !bValidOrg;
	return SetValidMultipleCommand(bValidNew);
    }

    if (pCTgt->mstrClassName.IsSameAs("LightSource_cls") == FALSE) return;

    bool bValidOrg = ((LightSource_cls*)pCTgt)->mCbValid.GetBodyBOOL();
    wxString strData; strData << !bValidOrg;
    ((LightSource_cls*)pCTgt)->mCbValid.PushCommandPair(strData);
}

/*
void CLS::
SetValidCommand(wxString strName, bool bValidNew){
    SceneObj_cls *pCTgt = GetPCFromName(strName);
    if (pCTgt == NULL) return ;
    if (pCTgt->mstrClassName.IsSameAs("LightSource_cls") == FALSE) return;
    bool bValidOrg = ((LightSource_cls*)pCTgt)->mCbValid.GetBodyBOOL();

    if (bValidNew == bValidOrg) return;
    ToggleValidCommand(strName);
}
*/

void CLS::
MoveItemMultipleToCommand(wxString strNameTo, bool bAdditional){

    SortSelectionList();
    GetPCUndo()->OpenBlock(wxT("MoveMultiple"));

    for (uint i = 0; i < mVpCSelectedObject.size(); i++){
	SceneObj_cls *pCTgt = mVpCSelectedObject[i];

	wxString strNameFrom = pCTgt->GetStrName();
	if (strNameFrom == strNameTo) continue;
	wxString strParentNameFrom = pCTgt->GetStrParentName();
	int indexFrom = pCTgt->GetChildIndex();

	SceneObj_cls* pCTo = GetPCFromName(strNameTo);
	wxString strParentNameTo;
	int indexTo;
	if (strNameTo.IsSameAs(wxT("End"))){
	    strParentNameTo = wxT("ROOT");
	    indexTo = mVpCObj.size();
	}
	else{
	    strParentNameTo = pCTo->GetStrParentName();
	    indexTo = pCTo->GetChildIndex();
	    if (bAdditional == TRUE) indexTo += i+1;
	}
	
	if (strParentNameFrom.IsSameAs(strParentNameTo)){
	    // ӖȂ̃`FbN
	    if (indexTo == indexFrom+1 || indexTo == indexFrom) continue; 
	    if (indexTo < indexFrom) indexFrom++;
	}

	if (pCTo != NULL && pCTo->IsAncestor(strNameFrom) == TRUE) continue; // Q
    
	wxString strUndo = GetStrSelf() << wxT("Move::") << strNameFrom << wxT("::To::") << strParentNameFrom << wxT("::") << indexFrom;
	wxString strRedo = GetStrSelf() << wxT("Move::") << strNameFrom << wxT("::To::") << strParentNameTo << wxT("::") << indexTo;
	GetPCUndo()->PushCommandPair(strUndo, strRedo);
	GetPCUndo()->Redo();
    }

    GetPCUndo()->CloseBlock(wxT("MoveMultiple"));
}

void CLS::
MoveItemToCommand(wxString strNameTo, bool bAdditional){

    if (mpCActiveObject == NULL) return;

    wxString strNameFrom = mpCActiveObject->GetStrName();
    if (strNameFrom == strNameTo) return;

    if (mVpCSelectedObject.size() > 1) 
	return MoveItemMultipleToCommand(strNameTo, bAdditional);
    
    wxString strParentNameFrom = mpCActiveObject->GetStrParentName();
    int indexFrom = mpCActiveObject->GetChildIndex();
    
    SceneObj_cls* pCTo = GetPCFromName(strNameTo);
    wxString strParentNameTo;
    int indexTo;
    if (strNameTo.IsSameAs(wxT("End"))){
	strParentNameTo = wxT("ROOT");
	indexTo = mVpCObj.size();
    }
    else{
	strParentNameTo = pCTo->GetStrParentName();
	indexTo = pCTo->GetChildIndex();
	if (bAdditional == TRUE) indexTo++;
    }
    
    if (strParentNameFrom.IsSameAs(strParentNameTo)){
	// ӖȂ̃`FbN
	if (indexTo == indexFrom+1 || indexTo == indexFrom) return; 
	if (indexTo < indexFrom) indexFrom++;
    }

    if (pCTo != NULL && pCTo->IsAncestor(strNameFrom) == TRUE) return; // Q
    
    wxString strTag = GetStrSelf();
    wxString strUndo(strTag), strRedo(strTag);
    strUndo << wxT("Move::") << strNameFrom << wxT("::To::") << strParentNameFrom << wxT("::") << indexFrom;
    strRedo << wxT("Move::") << strNameFrom << wxT("::To::") << strParentNameTo << wxT("::") << indexTo;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
MoveModifierCommand(wxString strTag, int indexTo){
    wxString strName = PopStrItemFromStrLine(strTag, "::");
    SceneObj_cls *pCObj = GetPCFromName(strName);
    if (pCObj == NULL) return;

    if (pCObj->IsAttributeExist(strTag))
	pCObj->MoveModifierCommand(strTag, indexTo);
    if (pCObj->IsDataProcExist(strTag))
	pCObj->MoveDataProcCommand(strTag, indexTo);
}

void CLS::
SetMultipleToChildCommand(wxString strNameTo){

    SortSelectionList();
    GetPCUndo()->OpenBlock(wxT("ToChildMultiple"));

    for (uint i = 0; i < mVpCSelectedObject.size(); i++){
	SceneObj_cls *pCTgt = mVpCSelectedObject[i];

	wxString strNameFrom = pCTgt->GetStrName();
	if (strNameFrom == strNameTo) continue;
	wxString strParentNameFrom = pCTgt->GetStrParentName();
	int indexFrom = pCTgt->GetChildIndex();

	SceneObj_cls *pCTo = GetPCFromName(strNameTo);
	
	if (pCTo != NULL && pCTo->IsAncestor(strNameFrom) == TRUE) continue; // Q

	wxString strUndo = GetStrSelf() << wxT("Move::") << strNameFrom << wxT("::To::") << strParentNameFrom << wxT("::") << indexFrom;
	wxString strRedo = GetStrSelf() << wxT("SetToChild::") << strNameFrom << wxT("::To::") << strNameTo;
	GetPCUndo()->PushCommandPair(strUndo, strRedo);
	GetPCUndo()->Redo();
    }
    
    GetPCUndo()->CloseBlock(wxT("ToChildMultiple"));
}

void CLS::
SetToChildCommand(wxString strNameTo){
    if (mpCActiveObject == NULL) return;
    
    wxString strNameFrom = mpCActiveObject->GetStrName();
    if (strNameFrom == strNameTo) return;

    if (mVpCSelectedObject.size() > 1)
	return SetMultipleToChildCommand(strNameTo);

    wxString strParentNameFrom = mpCActiveObject->GetStrParentName();
    int indexFrom = mpCActiveObject->GetChildIndex();

    SceneObj_cls* pCTo = GetPCFromName(strNameTo);
    
    if (pCTo != NULL && pCTo->IsAncestor(strNameFrom) == TRUE) return; // Q

    wxString strUndo = GetStrSelf() << wxT("Move::") << strNameFrom << wxT("::To::") << strParentNameFrom << wxT("::") << indexFrom;
    wxString strRedo = GetStrSelf() << wxT("SetToChild::") << strNameFrom << wxT("::To::") << strNameTo;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
RenameCommand(wxString strNameOld, wxString strNameNew){
    if (strNameNew == strNameOld) return;

    if (strNameNew.Find(wxT("::")) != wxNOT_FOUND){
	return; // !! IMPLEMENT_ME !! Warning Message:
    }

    SceneObj_cls* pCSelf = GetPCFromName(strNameOld);

    while (IsObjExist(strNameNew, pCSelf) == TRUE)
	strNameNew = GetStrNewCandidate(strNameNew);

    if (strNameNew.IsSameAs(strNameOld)) return;
    
    wxString strUndo = GetStrSelf() << wxT("Rename::") << strNameNew << wxT("::To::") << strNameOld;
    wxString strRedo = GetStrSelf() << wxT("Rename::") << strNameOld << wxT("::To::") << strNameNew;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
AppendCommandBase(wxString strName, wxString strClassName, 
		  wxString strBaseName){
    wxString strUndo = GetStrSelf() << wxT("Delete::") << strName;
    wxString strRedo = GetStrSelf() << wxT("New::") << strClassName << wxT("::") << strBaseName << wxT("::") << strName;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
    
}

wxString CLS::
AppendPrimitiveCommand(int type, wxString strNewName){
    if (type == _PRIMITIVE_INVALID) return wxES;
    wxString strBaseName = Primitive_cls::GetStrInitNameCandidate(type);
    wxString strName = GetStrUniqueName(strBaseName);
    if (strNewName.IsEmpty() == FALSE)
	strName = GetStrUniqueName(strNewName);
    
    AppendCommandBase(strName, "Primitive_cls", strBaseName);
    return strName;
}

wxString CLS::
AppendLightCommand(int type, wxString strNewName){
    if (type == _LIGHT_INVALID) return wxES;
    wxString strBaseName = LightSource_cls::GetStrInitNameCandidate(type);
    wxString strName = GetStrUniqueName(strBaseName);
    if (strNewName.IsEmpty() == FALSE)
	strName = GetStrUniqueName(strNewName);

    AppendCommandBase(strName, "LightSource_cls", strBaseName);
    return strName;
}

wxString CLS::
AppendSceneObjCommand(wxString strBaseName, wxString strNewName){
    wxString strName = GetStrUniqueName(strBaseName);
    wxString strClassName = strBaseName + "_cls";
    if (strNewName.IsEmpty() == FALSE)
	strName = GetStrUniqueName(strNewName);
    
    AppendCommandBase(strName, strClassName, strBaseName);
    return strName;
}

void CLS::
DuplicateObjectCommand(){
    // selectԂς̂ŃRs[
    std::vector < SceneObj_cls* > VpCSelectedObject = mVpCSelectedObject;

    uint numSelected = VpCSelectedObject.size();
    if (numSelected == 0) return;
    GetPCUndo()->OpenBlock(wxT("DuplicateItems"));

    std::vector < wxString > VstrFrom;
    std::vector < wxString > VstrParent;
    std::vector < wxString > VstrNew;
    
    for (uint i = 0; i < numSelected; i++){
	SceneObj_cls *pCTgt = VpCSelectedObject[i];
	int tgtChildIndex = pCTgt->GetChildIndex();
	wxString strClassName = pCTgt->mstrClassName;
	wxString strName = pCTgt->GetStrName();
	wxString strParentName = pCTgt->GetStrParentName();

	//wxString strNewName = GetStrUniqueName(strName);

	wxString strNewName;
	if (strClassName.IsSameAs("Primitive_cls"))
	    strNewName = AppendPrimitiveCommand(0, strName);
	if (strClassName.IsSameAs("MeshObj_cls"))
	    strNewName = AppendSceneObjCommand("MeshObj", strName);
	if (strClassName.IsSameAs("TextObj_cls"))
	    strNewName = AppendSceneObjCommand("TextObj", strName);
	if (strClassName.IsSameAs("LightSource_cls"))
	    strNewName = AppendLightCommand(0, strName);
	if (strClassName.IsSameAs("Camera_cls"))
	    strNewName = AppendSceneObjCommand("Camera", strName);
	if (strClassName.IsSameAs("World_cls"))
	    strNewName = AppendSceneObjCommand("World", strName);
	if (strClassName.IsSameAs("VolumeBox_cls"))
	    strNewName = AppendSceneObjCommand("VolumeBox", strName);
	if (strClassName.IsSameAs("Isosurface_cls"))
	    strNewName = AppendSceneObjCommand("Isosurface", strName);
	if (strClassName.IsSameAs("Particle_cls"))
	    strNewName = AppendSceneObjCommand("Particle", strName);
	if (strClassName.IsSameAs("Streamline_cls"))
	    strNewName = AppendSceneObjCommand("Streamline", strName);
	if (strClassName.IsSameAs("HeightField_cls"))
	    strNewName = AppendSceneObjCommand("HeightField", strName);

	//GetPCFromName(strTmp)->ProcRename(strNewName);
	
	// j[N擾̎זɂȂ̂ňꎞl[
	//GetPCFromName(strTmp)->ProcRename("TemporaryCopyObject");
	//GetPCFromName("TemporaryCopyObject")->ProcRename(strNewName);

	//GetPCFromName("TemporaryCopyObject")->ProcRename(strTmp);

	//mbSkipAfterMath = TRUE;
	//RenameCommand(strTmp, strNewName);
	//mbSkipAfterMath = FALSE;

	pCTgt->ProcRename(strNewName); // VACYpɈꎞl[

	wxString strUndo1 = "Skip";
	wxString strRedo1;

	pCTgt->SerializePropertiesOnly(strRedo1);
	pCTgt->ProcRename(strName);  // l[߂

	GetPCUndo()->PushCommandPairAndDo(strUndo1, strRedo1);

	// Rs[̒
	int newChildIndex = GetPCFromName(strNewName)->GetChildIndex();
	wxString strTag = GetStrSelf();
	wxString strUndo2(strTag), strRedo2(strTag);
	strUndo2 << wxT("Move::") << strNewName << wxT("::To::") << strParentName << wxT("::") << newChildIndex;
	strRedo2 << wxT("Move::") << strNewName << wxT("::To::") << strParentName << wxT("::") << tgtChildIndex+1;
	GetPCUndo()->PushCommandPairAndDo(strUndo2, strRedo2);
	//GetPCUndo()->Redo();
	
	VstrFrom.push_back(strName);
	VstrParent.push_back(strParentName);
	VstrNew.push_back(strNewName);
    }

    // eqƃRs[̎Aeq֌Wč\z
    for (uint i = 0; i < numSelected; i++)
	for (uint j = 0; j < numSelected; j++)
	    if (VstrParent[i].IsSameAs(VstrFrom[j])){
		VstrParent[i] = VstrNew[j];
		break;
	    }
    
    for (int i = 0; i < numSelected; i++){
	if (VstrParent[i].IsEmpty() == TRUE) continue;
	if (GetPCFromName(VstrNew[i])->GetStrParentName().IsSameAs(VstrParent[i]) == FALSE){
	    Select(VstrNew[i], FALSE);
	    SetToChildCommand(VstrParent[i]);
	}
    }
    
    GetPCUndo()->CloseBlock(wxT("DuplicateItems"));
}

void CLS::
DeleteItemsCommand(){
    // [vdelete邽߃Rs[
    std::vector < SceneObj_cls* > VpCSelectedObject = mVpCSelectedObject;
    
    uint numSelected = VpCSelectedObject.size();

    if (numSelected == 0) return;
    GetPCUndo()->OpenBlock(wxT("DeleteItems"));

    for (uint i = 0; i < numSelected; i++){
	SceneObj_cls *pCTgt = VpCSelectedObject[i];

	wxString strNameFrom = pCTgt->GetStrName();
	wxString strParentNameFrom = pCTgt->GetStrParentName();
	int indexFrom = pCTgt->GetChildIndex();

	wxString strBaseName = pCTgt->GetStrBaseName();
	wxString strClassName = pCTgt->mstrClassName;

	// *Serialize BEFORE Deletion*
	wxString strUndo2;
	pCTgt->SerializePropertiesOnly(strUndo2);
	wxString strRedo2 = (wxString)wxT("Skip");

	wxArrayString AstrUndoChild, AstrRedoChild;
	for (std::list< SceneObj_cls* >::iterator itr = pCTgt->mVpCObj.begin();
	     itr != pCTgt->mVpCObj.end(); itr++){
	    AstrUndoChild.Add(GetStrSelf() << "SetToChild::" << (*itr)->GetStrName() << "::To::" << strNameFrom);
	    AstrRedoChild.Add("Skip");
	}
	
	wxString strUndo1 = GetStrSelf() << "Move::" << strNameFrom << "::To::" << strParentNameFrom << wxT("::") << indexFrom;
	wxString strRedo1 = GetStrSelf() << "Delete::" << strNameFrom;
	GetPCUndo()->PushCommandPair(strUndo1, strRedo1);
	GetPCUndo()->Redo();

	GetPCUndo()->PushCommandPair(strUndo2, strRedo2);
	GetPCUndo()->Redo();

	for (uint i = 0; i < AstrRedoChild.GetCount(); i++){
	    GetPCUndo()->PushCommandPair(AstrUndoChild[i], AstrRedoChild[i]);
	    GetPCUndo()->Redo();
	}

	wxString strUndo3 = GetStrSelf() << "New::" << strClassName << "::" << strBaseName << "::" << strNameFrom;
	wxString strRedo3 = "Skip";
	GetPCUndo()->PushCommandPair(strUndo3, strRedo3);
	GetPCUndo()->Redo();
    }
    
    GetPCUndo()->CloseBlock(wxT("DeleteItems"));
}

wxString CLS::
GetStrCommandParamGroup(wxString strTag, ParamGroup_cls *pCParamGroup){
    wxString strName = PopStrItemFromStrLine(strTag, "::");
    wxString strParamGroup = PopStrItemFromStrLine(strTag, "::");
    if (strParamGroup.IsSameAs(pCParamGroup->mCstrName.GetBodySTR()) == FALSE)
	strParamGroup += "::" + PopStrItemFromStrLine(strTag, "::");
    return GetStrSelf() << "Obj::" << strName << "::" << strParamGroup << wxT("::");
}

/*
bool CLS::
ProcSetVisible(std::wistream &istream, wxString strFirstItem, wxString strLine){
    SceneObj_cls *pCTgt = GetPCFromName(strFirstItem);
    if (pCTgt == NULL) return FALSE;

    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    pCTgt->mCbVisible.SetBodyFromData(strFirstItem);
    return TRUE;
}
*/

void CLS::
MoveItemNative(SceneObj_cls *pCFrom, wxString strToParentName, int index1){

    ParentObj_cls* pCToParent = GetPCFromName(strToParentName);
    if (pCToParent == NULL) pCToParent = this;

    std::list < SceneObj_cls*>::iterator itrTo = pCToParent->GetIterFromIndex(index1);
    std::list < SceneObj_cls*>::iterator itrFrom = pCFrom->mpCParentObj->GetIterFromP(pCFrom);

    if (pCToParent == pCFrom->mpCParentObj && itrTo == itrFrom) return;

    pCFrom->mpCParentObj->mVpCObj.erase(itrFrom);
    pCToParent->mVpCObj.insert(itrTo, pCFrom);
    pCFrom->mpCParentObj = pCToParent;
}

bool CLS::
ProcMove(std::wistream &istream, wxString strFirstItem, wxString strLine){
    SceneObj_cls *pCFrom = GetPCFromName(strFirstItem);
    if (pCFrom == NULL) return FALSE;
    
    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    wxString strToParentName = PopStrItemFromStrLine(strLine, wxT("::"));
    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    long indexTo;
    strFirstItem.ToLong(&indexTo);
    MoveItemNative(pCFrom, strToParentName, indexTo);
    return TRUE;
}

bool CLS::
ProcSetToChild(std::wistream &istream, wxString strFirstItem, wxString strLine){
    SceneObj_cls *pCFrom = GetPCFromName(strFirstItem);
    if (pCFrom == NULL) return FALSE;
    
    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    wxString strTo = PopStrItemFromStrLine(strLine, wxT("::"));
    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    MoveItemNative(pCFrom, strTo, -1);
    return TRUE;
}

bool CLS::
ProcRename(std::wistream &istream, wxString strFirstItem, wxString strLine){
    SceneObj_cls *pCTgt = GetPCFromName(strFirstItem);
    if (pCTgt == NULL) return FALSE;
    
    wxString strNameOld = strFirstItem;
    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    wxString strNameNew = PopStrItemFromStrLine(strLine, wxT("::"));
    pCTgt->ProcRename(strNameNew);

    //pCTgt->mCstrName.mstrBody = strNameNew;
    //GetPCMain()->ProcRename(strNameOld, strNameNew);
    GetPCMain()->ProcAfterMath("RenameObject",
			       strNameOld + "::" + strNameNew);
    
    return TRUE;
}

bool CLS::
ProcNew(std::wistream &istream, wxString strFirstItem, wxString strLine){
    wxString strClassName = strFirstItem;
    wxString strType = PopStrItemFromStrLine(strLine, wxT("::"));
    wxString strNewName = PopStrItemFromStrLine(strLine, wxT("::"));

    SceneObj_cls *pCTgt = NULL;
    if (strClassName.IsSameAs("Primitive_cls"))
	pCTgt = new Primitive_cls(this, strType, strNewName);
    if (strClassName.IsSameAs("MeshObj_cls"))
	pCTgt = new MeshObj_cls(this, strNewName);
    if (strClassName.IsSameAs("TextObj_cls"))
	pCTgt = new TextObj_cls(this, strNewName);
    if (strClassName.IsSameAs("LightSource_cls"))
	pCTgt = new LightSource_cls(this, strType, strNewName);
    if (strClassName.IsSameAs("Camera_cls"))
	pCTgt = new Camera_cls(this, strNewName);
    if (strClassName.IsSameAs("VolumeBox_cls"))
	pCTgt = new VolumeBox_cls(this, strNewName);
    if (strClassName.IsSameAs("Isosurface_cls"))
	pCTgt = new Isosurface_cls(this, strNewName);
    if (strClassName.IsSameAs("Particle_cls"))
	pCTgt = new Particle_cls(this, strNewName);
    if (strClassName.IsSameAs("Streamline_cls"))
	pCTgt = new Streamline_cls(this, strNewName);
    if (strClassName.IsSameAs("HeightField_cls"))
	pCTgt = new HeightField_cls(this, strNewName);
    if (strClassName.IsSameAs("World_cls"))
	pCTgt = new World_cls(this, strNewName);
    if (pCTgt != NULL){
	mVpCObj.push_back(pCTgt);
	Select(pCTgt->mCstrName.mstrBody, FALSE, FALSE);
	return TRUE;
    }
    
    return FALSE;
}

void CLS::
DeleteSceneObj(SceneObj_cls *pCTgt){
    wxString strName = pCTgt->GetStrName();
    Detatch(pCTgt);
    delete pCTgt;
    GetPCMain()->ProcAfterMath("DeleteObject", strName);
}

bool CLS::
ProcDelete(std::wistream &istream, wxString strFirstItem, wxString strLine){
    SceneObj_cls *pCTgt = GetPCFromName(strFirstItem);
    if (pCTgt == NULL) return FALSE;

    ParentObj_cls *pCParent = pCTgt->mpCParentObj;

    pCParent->InheritChilds(pCTgt);
    pCParent->mVpCObj.erase(pCParent->GetIterFromP(pCTgt));

    DeleteSceneObj(pCTgt);
    /*
    Detatch(pCTgt);
    delete pCTgt;
    GetPCMain()->ProcAfterMath("DeleteObject", strFirstItem);
    */

    return TRUE;
}

bool CLS::
ProcReadProperties(std::wistream &istream, wxString strFirstItem, wxString strLine){
    SceneObj_cls *pCTgt = GetPCFromName(strFirstItem);
    if (pCTgt == NULL) return FALSE;

    std::wstring wstrLine;
    while( std::getline(istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	
	strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));

	pCTgt->ReadStrLine(istream, strFirstItem, strLine);
	if (strFirstItem.IsSameAs("ReadProperties")){
	    ASSERT(PopStrItemFromStrLine(strLine, wxT("::")).IsSameAs(wxT("End")));
	    break;
	}
    }

    return TRUE;
}

bool CLS::
ProcReadObj(std::wistream &istream, wxString strFirstItem, wxString strLine){
    SceneObj_cls *pCTgt = GetPCFromName(strFirstItem);
    if (pCTgt == NULL) return FALSE;

    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    return pCTgt->ReadStrLine(istream, strFirstItem, strLine);
}

bool CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){

    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE){
	Select(mCstrActiveItemName.GetBodySTR(), FALSE);
	return TRUE;
    }

    // SceneObject̃vpeBύX
    std::vector < SceneObj_cls* > VpCList;
    GetVPCList(VpCList);
    int objNum = VpCList.size();
    for (int i = 0; i < objNum; i++)
	if (VpCList[i]->ReadStrLine(istream, strFirstItem, strLine) == TRUE){
	    GetPCMain()->UpdatePanels("OUTLINER::CONSOLE::VIEW3D::PROPERTY_OBJECT::PROPERTY_MATERIAL::PROPERTY_MODIFIER::PROPERTY_DATAPROC");
	    return TRUE;
	}

    // SceneObjectւ̑
    if (strFirstItem.IsSameAs("SceneObjCtrl") == FALSE) return FALSE;

    strFirstItem = PopStrItemFromStrLine(strLine, "::");
    bool bRet = FALSE;
    /*
    if (strFirstItem.IsSameAs("SetVisible")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcSetVisible(istream, strFirstItem, strLine);
    }
    */
    if (strFirstItem.IsSameAs("Move")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcMove(istream, strFirstItem, strLine);
    }
    if (strFirstItem.IsSameAs("SetToChild")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcSetToChild(istream, strFirstItem, strLine);
    }
    if (strFirstItem.IsSameAs("Rename")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcRename(istream, strFirstItem, strLine);
    }
    if (strFirstItem.IsSameAs("New")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcNew(istream, strFirstItem, strLine);
    }
    if (strFirstItem.IsSameAs("Delete")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcDelete(istream, strFirstItem, strLine);
    }
    if (strFirstItem.IsSameAs("ReadProperties")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcReadProperties(istream, strFirstItem, strLine);
    }
    if (strFirstItem.IsSameAs("Obj")){
	strFirstItem = PopStrItemFromStrLine(strLine, "::");
	bRet |= ProcReadObj(istream, strFirstItem, strLine);
    }
    
    if (bRet == TRUE) GetPCMain()->UpdatePanels("OUTLINER::TIMELINE::CONSOLE::GRAPH::COLOR_CURVE::VIEW3D::PROPERTY_OBJECT::PROPERTY_MATERIAL::PROPERTY_MODIFIER::PROPERTY_DATAPROC");
    return TRUE;
}

void CLS::
Serialize(wxString& strOut){
    uint num = mVpCObj.size();
    for (uint i = 0; i < num; i++){
	GetPCFromIndex(i)->Serialize(strOut);
    }

    mCstrActiveItemName.SetBodyFromData(GetActiveObjectName());
    ConfListContainer_cls::Serialize(strOut);
}

SceneObjCtrl_cls* GpCSceneObjCtrl = NULL;

