// (setq-c-offset 4)
// "clsTextObj.h"

#include "clsTextObj.h"
#include "clsMaterial.h"
#include "PROP/clsTextProp.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;

#include "../GLOBJ/clsVBOPlane.h"
#include "../MYGL/clsShadowVolumeCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/MISC/clsInputStream.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include "../../wxWidgetEnum.h"

#include <wx/sstream.h>

TextItem_cls::
TextItem_cls(wxString strText){
    mstrText = strText;
    for (int i = 0; i < 3; i++) mOffset[i] = 0;
    for (int i = 0; i < 3; i++) mScale[i] = 1;
}

float TextItem_cls::
GetAspectRatio(){
    if (mHeight == 0) return 0;
    return mWidth / mHeight;
}

#define CLS TextObj_cls

CLS::
TextObj_cls(ParentObj_cls *pCParentObj, wxString strName) : 
SceneObj_cls(pCParentObj)
{
    mSortOrder = 1;
    
    wxString strUpdate = "VIEW3D::PROPERTY_OBJECT";

    mCConfList.Add(&mCbXRay, "XRay=", "0", strUpdate);
    mCConfList.Add(&mCbDepthTest, "DepthTest=", "1", strUpdate);

    mstrClassName = "TextObj_cls";
    mpCMaterial = new Material_cls(strName, mstrClassName);

    Initialize(strName);
    
    mpCTextProp = new TextProp_cls(strName, "TextProp_cls");

    mVpCAttribute.push_back(mpCTextProp);
    mVpCAttribute.push_back(mpCMaterial);
}   

CLS::~TextObj_cls(){}

void CLS::
SyncConfListTagTextObj(){
    wxString strObj = mCstrName.GetBodySTR();
    uint numAttrib = mVpCAttribute.size();
    for (uint i = 0; i < numAttrib; i++){
	Attribute_cls *pCAttrib = GetPCAttributeFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }

    /*
    uint numDataProc = mVpCDataProc.size();
    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCAttrib = GetPCDataProcFromIndex(i);
	pCAttrib->SetConfListTagForSceneObjAttribute(pCAttrib->GetStrName(), strObj);
    }
    */
    
    //mpCParticleFormat->SyncConfListTag(strObj);
}

void CLS::
ReadString(wxString strIn){
    InputStream_cls CStream(strIn);

    wxString strAccum;

    for (wxString str = CStream.GetLine(); !str.IsEmpty() || !CStream.Eof();
	 str = CStream.GetLine() ){

	wxString strOffset, strScale;

	int pos = str.Find(wxT("//"));
	if (pos != wxNOT_FOUND){
	    str = str.Mid(0, pos);
	    if (str.IsEmpty() == TRUE) continue;
	}

	pos = str.Find(wxT('|'));
	if (pos != wxNOT_FOUND){
	    strOffset = str.Mid(pos+1);
	    int pos2 = strOffset.Find(wxT('|'));
	    if (pos2 != wxNOT_FOUND){
		strScale = strOffset.Mid(pos2+1);
		strOffset = strOffset.Mid(0, pos2-1);
	    }
	    str = str.Mid(0, pos-1);
	    str.Trim();
	}
	
	if (strAccum.IsEmpty() == FALSE)
	    str = strAccum + wxT("\n") + str;
	strAccum.Clear();
	
	if (strOffset.IsEmpty() == FALSE){
	    TextItem_cls CText(str);
	    
	    if (strOffset.IsEmpty() == FALSE)
		sscanf(strOffset.ToAscii(), "%g%*[ ,]%g%*[ ,]%g", &CText.mOffset[0], &CText.mOffset[1], &CText.mOffset[2]);
	    if (strScale.IsEmpty() == FALSE)
		sscanf(strScale.ToAscii(), "%g%*[ ,]%g*[ ,]%g", &CText.mScale[0], &CText.mScale[1], &CText.mScale[2]);
	    mATextMade.push_back(CText);
	}
	else
	    strAccum << str;
    }

    if (strAccum.IsEmpty() == FALSE) 
	mATextMade.push_back(TextItem_cls(strAccum));
}

void CLS::
MakeTextTexture(){
    long style = wxFONTWEIGHT_NORMAL;
    long weight = wxFONTWEIGHT_NORMAL;
    int fontSize = 36;
    wxFont CFont;
    CFont.SetNativeFontInfoUserDesc(mpCTextProp->mCstrFontDesc.GetBodySTR());

    mACGLTexture.clear();
    mATextMade.clear();

    bool mbListModeMade = FALSE;

    if (mbListModeMade == FALSE && mpCTextProp->mCstrText.GetBodySTR().IsEmpty() == FALSE){
	ReadString(mpCTextProp->mCstrText.GetBodySTR());
    }
	
    /*
      if (mbListModeMade == TRUE) {
	ReadList();
    }
    */

    mACGLTexture.resize(mATextMade.size());
    for (int i = 0; i < mATextMade.size(); i++){
	//mACGLTexture.push_back(GLTexture_cls());
	mACGLTexture[i].SetEmptyImage();
	mACGLTexture[i].ClearImage();
    }

    bool bSucceed = TRUE; // ȂsƂ̂sc
    wxScreenDC DC;
    DC.SetFont(CFont);
    for (int i = 0; i < mATextMade.size(); i++){

	wxString strText = mATextMade[i].mstrText;
	wxCoord singleW, singleH, singleD;
	wxCoord w,h;
	
	DC.GetTextExtent(strText, &singleW, &singleH, &singleD);
	DC.GetMultiLineTextExtent(strText, &w, &h, &singleH);

	wxRect CRectCanvas(0, 0, w, h);
	wxRect CRectBounding;
	wxMemoryDC *pMemDC = mACGLTexture[i].BeginDC(w, h);
	
	pMemDC->SetBackground(wxColour(0, 0, 0));
	pMemDC->Clear();
	pMemDC->SetBackground(wxColour(255, 0, 0));
	pMemDC->SetTextForeground(*wxWHITE);
	pMemDC->SetFont(CFont);
	pMemDC->DrawLabel(strText, wxNullBitmap, CRectCanvas, wxALIGN_LEFT | wxALIGN_TOP, -1, &CRectBounding);
	
	mACGLTexture[i].EndDC(TRUE, TRUE);

	mATextMade[i].mWidth = w;
	mATextMade[i].mHeight = h;
	mATextMade[i].mDescend = (float)singleD/singleH;

	if (i != 0){
	    if (mACGLTexture[i].mTextureID == mACGLTexture[i-1].mTextureID)
		bSucceed = FALSE;
	}
    }
    
    if (bSucceed == TRUE){
	if (mbListModeMade == FALSE) mstrTextMade = mpCTextProp->mCstrText.GetBodySTR();
	//if (mbListModeMade == TRUE) mstrPathImported = pCData->mCstrImportPath.mstrBody;
	mstrFontMade = mpCTextProp->mCstrFontDesc.GetBodySTR();
    }
}

void CLS::
CheckAndPrepare(){
    bool bMakeTexture = FALSE;
    if (mstrTextMade.IsSameAs(mpCTextProp->mCstrText.GetBodySTR()) == FALSE) 
	bMakeTexture = TRUE;
    if (mstrFontMade.IsSameAs(mpCTextProp->mCstrFontDesc.GetBodySTR()) == FALSE)
	bMakeTexture = TRUE;

    if (bMakeTexture == TRUE) MakeTextTexture();
}

void CLS::
GLDraw(int lightNum, int ensemble, int status){

    if (ensemble == 0 && mProcessedInInstanceLoop == 0){
	mpCMaterial->CheckTextureSequence();
    }
    
    if (GetPCShadowVolumeCtrl()->mSVState == 1)
	return;

    CheckAndPrepare();
    if (mACGLTexture.size() == 0) return;

    mpCMaterial->BeginShader(lightNum, ensemble, TRUE);

    if (mCbDepthTest.GetBodyBOOL() == TRUE)
	glEnable(GL_DEPTH_TEST);
    
    glEnable(GL_BLEND);
    mpCMaterial->SetBlendFunc();
    mpCMaterial->SetDepthMask();
    
    mpCMaterial->SetGLMaterialSetting();

    glEnable(GL_TEXTURE_2D);
    mpCMaterial->BindTexture(0);

    glEnable(GL_LIGHTING);
    
    double NT = GetPCTimeCtrl()->GetCurrentNT();
    float scaleBase = mpCTextProp->GetPCParamGroup("Scale")->GetVInterpolatedValue(NT)[0];
    if (scaleBase != 0){
	uint textNum = mACGLTexture.size();
	for (uint i = 0; i < textNum; i++){
	    mACGLTexture[i].BindTexture();
	    float aspect = mATextMade[0].GetAspectRatio();
	    if (aspect != 0){
		glScalef(scaleBase, scaleBase, scaleBase);
		glScalef(aspect, 1, 1);
		GCClassProvider.GetPCVBOPlane()->Draw(_VBO_OBJECT, 2);
		glScalef(1.0/aspect, 1, 1);
		glScalef(1.0/scaleBase, 1.0/scaleBase, 1.0/scaleBase);
	    }
	}
    }

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_CULL_FACE);
    glDisable(GL_BLEND);

    mpCMaterial->EndShader();

}

MultiInputPanel_cls* CLS::
GetPCNewPanel(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, GetStrName()/*, TRUE*/);
    
    mIndexExternal = pC->CombineConfItem(mpCTextProp->mCbExternal, _("External Link"));
    pC->CombineSizer(2, _("Text"));
    pC->CombineConfItem(mpCTextProp->mCstrText, _("Text"));
    int indexFont = pC->CombineBitmapPanel(btnTextFont, "Font", 24);

    mIndexScale = pC->CombineConfItem(mpCTextProp->mCTextScale, _("Scale"));
    
    pC->CombineSizer(2, _("Option"));
    pC->CombineConfItem(mCbXRay, _("X-Ray"));
    pC->CombineConfItem(mCbDepthTest, _("Depth Test"));

    pC->FinishSetting();

    BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[indexFont].mpCWindowLink;
    pCBitmap->mbButton = TRUE;
    pCBitmap->AddBitmapStr("wxART_FONT", "wxART_FONT_P");
    pCBitmap->mBorderX = 4;
    pCBitmap->mBorderY = 4;
	
    return pC;
}

bool CLS::
ReadStrLineMaterial(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCMaterial->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeMaterial(wxString& strOut){ mpCMaterial->Serialize(strOut); }
