// (setq-c-offset 4)
// "clsWodld.h"

#include "clsWorld.h"
#include "PROP/clsFogProp.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include "../../wxWidgetEnum.h"

#define CLS World_cls

CLS::
World_cls(ParentObj_cls *pCParentObj, wxString strName) : 
SceneObj_cls(pCParentObj)
{
    mSortOrder = 3;

    mstrClassName = "World_cls";
    Initialize(strName);
    mpCFogProp = new FogProp_cls;

    mVpCAttribute.push_back(mpCFogProp);
}

CLS::~World_cls(){}

void CLS::
SyncPropWorld(){
    mpCFogProp->SyncProp(GetStrName());
}

void CLS::
SetPropInfo(){
    mpCFogProp->SetPropInfo();
}

MultiInputPanel_cls* CLS::
GetPCNewPanelFog(wxWindow *pCWindow){
    MultiInputPanel_cls *pC =
	new MultiInputPanel_cls(pCWindow, _("Fog"));

    pC->CombineSizer(2, _("Type/Density"));
    pC->SetChoiceList(4, GetPCStringProvider()->GetAstr("FogType"));
    pC->CombineConfItem(mpCFogProp->mCFogType);
    pC->SetLimit(0, 10);
    pC->mDragScale = 0.1;
    pC->CombineConfItem(mpCFogProp->mCFogDensity);

    mIndexColor = pC->CombineSizer(2, _("Fog Color"));
    pC->CombineConfItem(mpCFogProp->mCFogColor);
    mIndexBtnColor = pC->CombineButton(btnFogColor, " ", 30);
    
    mIndexDistance = pC->CombineSizer(2, _("Start/End"));
    pC->SetLimit(0, 1000);
    pC->mDragScale = 1;
    pC->CombineConfItem(mpCFogProp->mCFogStart);
    pC->SetLimit(0, 10000);
    pC->mDragScale = 100;
    pC->CombineConfItem(mpCFogProp->mCFogEnd);
    
    mIndexOverwriteColor = pC->CombineSizer(1, _("Option"));
    pC->CombineConfItem(mpCFogProp->mCbOverwriteBG, _("Overwrite BG"));

    pC->FinishSetting();
    HideAndShow(pC);

    return pC;
}

bool CLS::
HideAndShow(MultiInputPanel_cls *pC){
    bool bChange = FALSE;
    bChange |= HideAndShowFog(pC);
    return bChange;
}

bool CLS::
HideAndShowFog(MultiInputPanel_cls *pC){
    bool bChange = FALSE;
    int fogType = mpCFogProp->mCFogType.GetBodyINT();
    if (fogType == _FOG_NONE){
	bChange |= pC->mVCItem[mIndexColor].Hide();
	bChange |= pC->mVCItem[mIndexOverwriteColor].Hide();
    }
    else{
	bChange |= pC->mVCItem[mIndexColor].Show();
	bChange |= pC->mVCItem[mIndexOverwriteColor].Show();
    }
    
    if (mpCFogProp->mCFogType.GetBodyINT() == _FOG_LINEAR)
	bChange |= pC->mVCItem[mIndexDistance].Show();
    else
	bChange |= pC->mVCItem[mIndexDistance].Hide();
    return bChange;
}

void CLS::
UpdateButtonColor(MNumPainBase_cls *pCPain){
    MultiInputPanel_cls* pCurr = pCPain->GetPCMNum(_("Fog"));
    if (pCurr == NULL) return;

    pCurr->mVCItem[mIndexBtnColor].mpCWindowLink->SetBackgroundColour(mpCFogProp->mCFogColor.GetBodyWXColor());
}

void CLS::
GetFogColor(float clr[4]){
    
    double NT = GetPCTimeCtrl()->GetCurrentNT();

    std::vector < float > VClr = mpCFogProp->GetPCParamGroup("FogColor")->GetVInterpolatedValue((float)NT);

    clr[0] = VClr[0];
    clr[1] = VClr[1];
    clr[2] = VClr[2];
    clr[3] = 1;
}

void CLS::
SetFog(bool bXRay){
    int fogType = mpCFogProp->mCFogType.GetBodyINT();
    if (fogType == _FOG_NONE) return;

    glEnable(GL_FOG);
    GetPCShaderCtrl()->mFogMode = fogType;

    float fogClr[4];
    GetFogColor(fogClr);

    if (msSVSampleNum > 2){
	fogClr[0] /= (msSVSampleNum-1);
	fogClr[1] /= (msSVSampleNum-1);
	fogClr[2] /= (msSVSampleNum-1);
    }

    //fogClr[0] = pow(pow(fogClr[0], 2.2) / msSVSampleNum, 0.45);
    //fogClr[1] = pow(pow(fogClr[1], 2.2) / msSVSampleNum, 0.45);
    //fogClr[2] = pow(pow(fogClr[2], 2.2) / msSVSampleNum, 0.45);

    glFogfv(GL_FOG_COLOR, fogClr);
    
    double NT = GetPCTimeCtrl()->GetCurrentNT();

    std::vector < float > VDensity = mpCFogProp->GetPCParamGroup("Density")->GetVInterpolatedValue((float)NT);
    glFogf(GL_FOG_DENSITY, VDensity[0]);
    
    std::vector < float > VDistance = mpCFogProp->GetPCParamGroup("Distance")->GetVInterpolatedValue((float)NT);
    glFogf(GL_FOG_START, VDistance[0]);
    glFogf(GL_FOG_END, VDistance[1]);
}

void CLS::
GLDraw(bool bXRay){
    SetFog(bXRay);

    /*
    if (bXRay == FALSE)
	SetBGObject();
    */
}

bool CLS::
ReadStrLineFog(std::wistream &istream, wxString strFirstItem, wxString strLine){
    return mpCFogProp->ReadStrLine(istream, strFirstItem, strLine);
}

void CLS::
SerializeFog(wxString& strOut){
    mpCFogProp->Serialize(strOut);
}

int World_cls::msSVSampleNum = 1;
