// (setq-c-offset 4)
// "clsParticleForRender.cpp"

#ifndef _PARTICLE_FOR_RENDER_INTERP_THREAD_H
#define _PARTICLE_FOR_RENDER_INTERP_THREAD_H

#include "../clsArrayAndIndex.h"

#include "../../COMMON/clsThreadCommon.h"

#include "../../MYGL/clsPosition.h"

class ArrayAndIndex_cls;
class LuaBase_cls;
class ParticleSnap_cls;
class ParamGroup_cls;

#define _DOWNSIZE_BY_INDEX       0
#define _DOWNSIZE_BY_ID          1
#define _DOWNSIZE_BY_ID_LONGLONG 2

#define _INTERP_STEP    0
#define _INTERP_LINEAR  1
#define _INTERP_HERMITE 2
#define _INTERP_BEZIER  3
#define _INTERP_BEZIERB 4

class ThreadInterpStep_cls : public ThreadInstance_cls {
public:
    void SetParameters
	(ParticleSnap_cls *pCSnap1, ParticleSnap_cls *pCSnap2,
	 ParticleSnap_cls *pCSnap0, ParticleSnap_cls *pCSnap3,
	 int classifyType, wxString strClassify, LuaBase_cls *pCLuaForClassify,
	 int interpType, int vanishSizeType, int extrapolateType,
	 double factor, bool bUseID, 
	 int numMat, float alphaCrit, float *sizeFactor, int *sizeType, 
	 int *sizeAlphaType, float *sizeAlphaBase, float *sizeAlphaMax, 
	 int *clrCurveType, int *clrCurveDepend,
	 float (*clrEmit)[3], unsigned char *paramBit,
	 ParamGroup_cls** ppCClr,
	 std::vector < double > *VRange,
	 float (*pX)[3], unsigned char *pType, float *pFilterAlpha,
	 float *pSize, float *pSizeAlpha,
	 float *pClrCurveS, float *pClrCurveT,
	 float *pValue0, float *pValue1, float *pValue2,
	 unsigned char *pParamBit, int *pID, long long *pIDLongLong,
	 ArrayAndIndex_cls *pIndexAndZ);
    void SetParametersTime
	(double NT, int frame, int subFrame, int seq);
    void SetParametersDownsize
	(int tgtIni, int downsize, int downsizeType);
    
    LuaBase_cls *mpCLuaForClassify;
    ParticleSnap_cls *mpCSnap1, *mpCSnap2, *mpCSnap0, *mpCSnap3;
    int mTgtIni, mDownsize, mDownsizeType;

    bool mbSearchNextIndexOnly;

    double mNT, mFactor;
    bool mbUseID;
    int mFrame, mSubFrame, mSeq;
    int mClassifyType;
    wxString mstrClassify;
    int mInterpType, mVanishSizeTpe, mExtrapolateType, mSortType, mNumMat;
    float mAlphaCrit, *mSizeFactor;
    int *mSizeType, *mClrCurveType, *mClrCurveDepend, *mSizeAlphaType;
    float *mpFilterAlpha, *mSizeAlphaBase, *mSizeAlphaMax;
    unsigned char *mParamBit;
    float (*mClrEmit)[3];
    ParamGroup_cls **mppCClr;
    std::vector < double > *mVRange;
    float (*mpX)[3];
    unsigned char *mpType;
    float *mpSize;
    float *mpSizeAlpha;
    float *mpClrCurveS, *mpClrCurveT;
    unsigned char *mpParamBit;
    int   *mpID;
    long long *mpIDLongLong;
    float *mpValue0, *mpValue1, *mpValue2;
    ArrayAndIndex_cls *mpIndexAndZ;

    void ClassifyByExpression(unsigned char * const pType,
			      const int *pCSnap1ID,
			      const long long *pCSnap1IDLongLong,
			      const int PIni, const int PFin);
    double GetRemapX(std::vector < double > &VRange, double x0/*, double NT*/);
    
    int SearchIDFull(const int * const pID2, int num2, int index, int id);
    int SearchIDNear10(const int * const pID2, int num2, int index, int id);
    int SearchIDNear(const int * const pID2, int num2, int index, int id, int range);
    int SearchIDFrom(const int * const pID2, int num2, int index, int id);

    int SearchIDLongFull(const long long * const pID2, int num2, int index, long long id);
    int SearchIDLongNear10(const long long * const pID2, int num2, int index, long long id);
    int SearchIDLongNear(const long long * const pID2, int num2, int index, long long id, int range);
    int SearchIDLongFrom(const long long * const pID2, int num2, int index, long long id);
    


    void SearchNextIndexByID();
    void InterpolateXYZLinearByID();
    void InterpolateXYZLinear();
    void InterpolateXYZStep();
    void InterpolateXYZHermiteByID();
    void InterpolateXYZHermite();
    void InterpolateXYZBezierByID(int bezierType);
    void InterpolateXYZBezier(int bezierType);
    void ExecBezier(float *tgt, const float* x0, const float* x1, const float* x2, const float* x3, float t);
    void ExecBezier2(float *tgt, const float* x0, const float* x1, const float* x2, const float* x3, float t);

    void InterpolateSizeLinearByID();
    void InterpolateSizeLinear();
    void InterpolateSizeStep();
    void InterpolateValueLinearByID(float *pValue, int valueIndex);
    void InterpolateValueLinear(float *pValue, int valueIndex);
    void InterpolateValueStep(float *pValue, int valueIndex);
    virtual ExitCode Entry();
};

class ThreadSetArrayZ_cls : public ThreadInstance_cls {
 public:
    void SetParameters
	(float (*pX)[3], unsigned char *pType, float *pFilterAlpha,
	 float *pSizeAlpha, unsigned char *pParamBit,
	 ArrayAndIndex_cls *pIndexAndZ,
	 PositionAndDirection_cls CSortPos,
	 int sortType);
    
    float (*mpX)[3];
    unsigned char *mpType;
    float *mpFilterAlpha, *mpSizeAlpha;
    unsigned char *mpParamBit;
    ArrayAndIndex_cls *mpIndexAndZ;
    PositionAndDirection_cls mCSortPos;
    int mSortType;

    virtual ExitCode Entry();
};

#endif
