// -*- c -*- 
// "clsParticleAsciiRead.h"

#include "clsParticleAsciiRead.h"

#include "../clsParticleFormat.h"
#include "../clsParticleSnap.h"
#include "../clsParticleSequence.h"

#include "../../COMMON/clsUseGZ.h"

#include "../../../clsFrameMain.h"

#include <stdio.h>
#include <fstream>
#include <sstream>

#include <wx/filename.h>

#include "../../COMMON/CommonDefinitions.h"

#define CLS ThreadParticleAsciiRead

bool CLS::
SetParameters(bool *pbFinishedFlagLink, ParticleSnap_cls *pCSnap,
	      ParticleFormatClone_cls *pCFormatClone){

    SetBaseParam(pbFinishedFlagLink, pCSnap);
    mstrFile = pCSnap->mstrFile.Clone();
    mpCFormatClone = pCFormatClone;
    
    if (Create() != wxTHREAD_NO_ERROR) return FALSE;
    return TRUE;
}

void CLS::
ReadBodyStreamVertical(ParticleFormatClone_cls *pCFormatClone,
		       std::istream &istream){
    std::string strBuf;

    std::vector < int > VFormat = pCFormatClone->GetVFormatBody(_PFORMAT_ASCII);
    uint numFormat = VFormat.size();

    for (uint j = 0; j < numFormat; j++){
	switch(VFormat[j]){
	case _PFORMAT_ASCII_XYZ:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		const char *ptr = strBuf.data();
		mpXLink[i][X] = ReadFloat(&ptr, ' '); ptr++;
		mpXLink[i][Y] = ReadFloat(&ptr, ' '); ptr++;
		mpXLink[i][Z] = ReadFloat(&ptr, ' '); 
	    }
	    break;
	case _PFORMAT_ASCII_SIZE:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		mpSizeLink[i] = atof(strBuf.data());
	    }
	    break;
	case _PFORMAT_ASCII_TYPE:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		mpTypeLink[i] = atoi(strBuf.data());
	    }
	    break;
	case _PFORMAT_ASCII_VALUE0:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		mpValue0Link[i] = atof(strBuf.data());
	    }
	    break;
	case _PFORMAT_ASCII_VALUE1:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		mpValue1Link[i] = atof(strBuf.data());
	    }
	    break;
	case _PFORMAT_ASCII_VALUE2:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		mpValue2Link[i] = atof(strBuf.data());
	    }
	    break;
	case _PFORMAT_ASCII_DUMMY:
	    for (int i = 0; i < mPNum; i++) std::getline(istream, strBuf);
	    break;
	case _PFORMAT_ASCII_DUMMYx3:
	    for (int i = 0; i < mPNum*3; i++) std::getline(istream, strBuf);
	    break;
	case _PFORMAT_ASCII_DUMMYx10:
	    for (int i = 0; i < mPNum*10; i++) std::getline(istream, strBuf);
	    break;
	case _PFORMAT_ASCII_X:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		mpXLink[i][X] = atof(strBuf.data());
	    }
	    break;
	case _PFORMAT_ASCII_Y:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		mpXLink[i][Y] = atof(strBuf.data());
	    }
	    break;
	case _PFORMAT_ASCII_Z:
	    for (int i = 0; i < mPNum; i++){
		std::getline(istream, strBuf); strBuf += " ";
		mpXLink[i][Z] = atof(strBuf.data());
	    }
	    break;
	}
    }
}

void CLS::
ReadBodyStream(ParticleFormatClone_cls *pCFormatClone,
	       std::istream &istream){
    
    if (pCFormatClone->GetPCItem("ArrayDirection=")->GetBodyINT() == _PFORMAT_ARRAY_VERTICLAL)
	return ReadBodyStreamVertical(pCFormatClone, istream);

    std::string strBuf;

    std::vector < int > VFormat = pCFormatClone->GetVFormatBody(_PFORMAT_ASCII);
    uint numFormat = VFormat.size();

    for (int i = 0; i < mPNum; i++){
	std::getline(istream, strBuf); strBuf += " ";
	const char *ptr = strBuf.data();
	for (uint j = 0; j < numFormat; j++){
	    switch(VFormat[j]){
	    case _PFORMAT_ASCII_XYZ:
		mpXLink[i][X] = ReadFloat(&ptr, ' '); ptr++;
		mpXLink[i][Y] = ReadFloat(&ptr, ' '); ptr++;
		mpXLink[i][Z] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_VXVYVZ:
		mpVLink[i][X] = ReadFloat(&ptr, ' '); ptr++;
		mpVLink[i][Y] = ReadFloat(&ptr, ' '); ptr++;
		mpVLink[i][Z] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_SIZE:
		mpSizeLink[i] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_ID:
		mpIDLink[i] = ReadInt(&ptr, ' '); break;
	    case _PFORMAT_ASCII_ID_LONGLOG:
		mpIDLongLongLink[i] = ReadLongLong(&ptr, ' '); break;
	    case _PFORMAT_ASCII_TYPE:
		mpTypeLink[i] = ReadInt(&ptr, ' '); break;
	    case _PFORMAT_ASCII_DUMMY:
		ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_DUMMYx3:
		ReadFloat(&ptr, ' '); ptr++;
		ReadFloat(&ptr, ' '); ptr++;
		ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_DUMMYx10:
		for (int k = 0; k < 9; k++)
		    ReadFloat(&ptr, ' '); ptr++;
		ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_VALUE0:
		mpValue0Link[i] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_VALUE1:
		mpValue1Link[i] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_VALUE2:
		mpValue2Link[i] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_X:
		mpXLink[i][X] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_Y:
		mpXLink[i][X] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_Z:
		mpXLink[i][X] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_VX:
		mpVLink[i][X] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_VY:
		mpVLink[i][X] = ReadFloat(&ptr, ' '); break;
	    case _PFORMAT_ASCII_VZ:
		mpVLink[i][X] = ReadFloat(&ptr, ' '); break;
	    }
	    ptr++;
	}
    }
}

void CLS::
ReadBodyAscii(ParticleFormatClone_cls *pCFormatClone){
    wxFileName CName(mstrFile);
    
    int headerLine = pCFormatClone->GetHeaderLine();
    std::string strBuf;

    if (CName.GetExt().IsSameAs(wxT("gz"))){
	std::string strBuf;
	UseGZ_cls::CopyToString(&strBuf, mstrFile);
	
	std::istringstream isstream(strBuf);
	for (int i = 0; i < headerLine; i++)
	    std::getline(isstream, strBuf);
	
	ReadBodyStream(pCFormatClone, isstream);
    }
    else{
#ifdef WIN32
	std::fstream ifstream(mstrFile.ToStdWstring(),
			      std::ios::in | std::ios::binary);
#else
	std::fstream ifstream(mstrFile.ToAscii(),
			      std::ios::in | std::ios::binary);
#endif
	if (ifstream.fail()){
	    printf("Read Header Fails!!\n");
	    return ;//exit(0);
	}

	int bufferSize = 10*1024*1024;//5*1024*1024;
	char *pBuffer = new char [bufferSize];
	ifstream.rdbuf()->pubsetbuf(pBuffer, bufferSize);

	for (int i = 0; i < headerLine; i++)
	    std::getline(ifstream, strBuf);

	ReadBodyStream(pCFormatClone, ifstream);
	
	delete [] pBuffer;
	
	ifstream.close();
    }
}

wxThread::ExitCode CLS::
Entry(){
    wxStopWatch CSW;
    CSW.Start();

    mpCDataSequenceLink->PushLock();
    DummyEntry();
    *mpbFinishedFlagLink = TRUE;
    mpCDataSequenceLink->PopLock();

    SendMenuEvent(GetPCFrameMain(0), menuFromSubThread);

    wxString strLog;
    strLog << "ReadLog :: " << mstrFile << "::" << CSW.Time() << "ms" << ENDL;
    wxLogMessage(strLog);
  
    return 0;
}

void CLS::
DummyEntry(){
    ReadBodyAscii(mpCFormatClone);
    
    bool bSort = mpCFormatClone->GetPCItem("PreSortByID=")->GetBodyBOOL();
    bool bRemoveDupli = mpCFormatClone->GetPCItem("RemoveDupli=")->GetBodyBOOL();
    if (bSort || bRemoveDupli){
	SortDataByID(bSort, bRemoveDupli);
    }

    delete mpCFormatClone;
}
