// -*- c -*- 
// "clsParticleBinaryRead.h"

#include "clsParticleBinaryRead.h"

#include "../clsParticleFormat.h"
#include "../clsParticleSnap.h"
#include "../clsParticleSequence.h"

#include "../../COMMON/clsUseGZ.h"

#include "../../../clsFrameMain.h"

#include <stdio.h>
#include <fstream>
#include <sstream>

#include <wx/filename.h>

#include "../../COMMON/CommonDefinitions.h"

#define CLS ThreadParticleBinaryRead

bool CLS::
SetParameters(bool *pbFinishedFlagLink, ParticleSnap_cls *pCSnap,
	      ParticleFormatClone_cls *pCFormatClone){

    SetBaseParam(pbFinishedFlagLink, pCSnap);
    mstrFile = pCSnap->mstrFile.Clone();
    mpCFormatClone = pCFormatClone;
    
    if (Create() != wxTHREAD_NO_ERROR) return FALSE;
    return TRUE;
}

void CLS::
ReadBodyStreamVertical(ParticleFormatClone_cls *pCFormatClone,
		       std::istream &istream){

    std::vector < int > VFormat = pCFormatClone->GetVFormatBody(_PFORMAT_BINARY);
    uint numFormat = VFormat.size();

    for (uint j = 0; j < numFormat; j++){
	switch(VFormat[j]){
	case _PFORMAT_XYZ_FLOAT:
	    istream.read((char*)mpXLink[0], sizeof(float)*3*mPNum); break; 
	case _PFORMAT_SIZE_FLOAT:
	    istream.read((char*)mpSizeLink, sizeof(float)*mPNum); break;
	case _PFORMAT_TYPE_INT:{
	    int *pBufferInt = new int [mPNum];
	    istream.read((char*)pBufferInt, sizeof(int)*mPNum); 
	    for (int i = 0; i < mPNum; i++)
		mpTypeLink[i] = (unsigned char)pBufferInt[i];
	    delete [] pBufferInt;
	    break; }
	case _PFORMAT_ID_INT:
	    istream.read((char*)mpIDLink, sizeof(int)*mPNum); break;
	case _PFORMAT_ID_LONGLONG:
	    istream.read((char*)mpIDLongLongLink, sizeof(long long)*mPNum); break;
	case _PFORMAT_VALUE_0:
	    istream.read((char*)mpValue0Link, sizeof(float)*mPNum); break;
	case _PFORMAT_VALUE_1:
	    istream.read((char*)mpValue1Link, sizeof(float)*mPNum); break;
	case _PFORMAT_VALUE_2:
	    istream.read((char*)mpValue2Link, sizeof(float)*mPNum); break;
	case _PFORMAT_DUMMY:
	    istream.ignore(sizeof(float)*mPNum); break;
	case _PFORMAT_DUMMYx3:
	    istream.ignore(sizeof(float)*mPNum*3); break;
	case _PFORMAT_DUMMYx10:
	    istream.ignore(sizeof(float)*mPNum*10); break;
	}
    }
}

void CLS::
ReadBodyStream(ParticleFormatClone_cls *pCFormatClone,
	       std::istream &istream){

    if (pCFormatClone->GetPCItem("ArrayDirection=")->GetBodyINT() == _PFORMAT_ARRAY_VERTICLAL)
	return ReadBodyStreamVertical(pCFormatClone, istream);

    int eachItemSize = pCFormatClone->GetBodyParticleByte();
    int packSize = 200000;
    
    char *pItemBuffer[2];
    pItemBuffer[0] = new char [eachItemSize * packSize];
    pItemBuffer[1] = new char [eachItemSize * packSize];
    char *pCurrent;

    bool bFinished[2] = {FALSE, FALSE};
    int currentMode = 0;

    for (int i_ = 0; i_ < mPNum; i_ += packSize){
	
	int currentSize = packSize;
	int nextSize    = packSize;
	if (i_ + currentSize > mPNum){
	    currentSize = mPNum - i_;
	    nextSize = -1;
	}
	else
	    if (i_ + currentSize + nextSize > mPNum)
		nextSize = mPNum - i_ - currentSize;

	if (i_ == 0){
	    bFinished[currentMode] = FALSE;
	    ThreadReadBuffer_cls *pCReadBuffer = new ThreadReadBuffer_cls;
	    pCReadBuffer->SetParameters(&bFinished[currentMode], &istream, 
					pItemBuffer[currentMode],
					eachItemSize*currentSize);
	    pCReadBuffer->Run();
	    while (bFinished[currentMode] == FALSE){
		wxMilliSleep(2);
	    }
	}

	if (nextSize != -1){
	    bFinished[1-currentMode] = FALSE;
	    ThreadReadBuffer_cls *pCReadBuffer = new ThreadReadBuffer_cls;
	    pCReadBuffer->SetParameters(&bFinished[1-currentMode], &istream, 
					pItemBuffer[1-currentMode],
					eachItemSize*nextSize);
	    pCReadBuffer->Run();
	}
	else
	    bFinished[1-currentMode] = TRUE;

	pCurrent = pItemBuffer[currentMode];
	int iFin = i_ + currentSize;

	std::vector < int > VFormat = pCFormatClone->GetVFormatBody(_PFORMAT_BINARY);
	uint numFormat = VFormat.size();

	for (int i = i_; i < iFin; i++){
	    for (uint j = 0; j < numFormat; j++){
		switch(VFormat[j]){
		case _PFORMAT_XYZ_FLOAT:
		    memcpy(mpXLink[i], pCurrent, 4*3);
		    pCurrent += 4*3; break; 
		case _PFORMAT_SIZE_FLOAT:
		    memcpy(&mpSizeLink[i], pCurrent, 4);
		    pCurrent += 4; break; 
		case _PFORMAT_TYPE_INT:
		    int type;
		    memcpy(&type, pCurrent, 4);
		    mpTypeLink[i] = (unsigned char) type;
		    pCurrent += 4; break;
		case _PFORMAT_ID_INT:
		    memcpy(&mpIDLink[i], pCurrent, 4);
		    pCurrent += 4; break;
		case _PFORMAT_ID_LONGLONG:
		    memcpy(&mpIDLongLongLink[i], pCurrent, 8);
		    pCurrent += 8; break;
		case _PFORMAT_VALUE_0:
		    memcpy(&mpValue0Link[i], pCurrent, 4);
		    pCurrent += 4; break;
		case _PFORMAT_VALUE_1:
		    memcpy(&mpValue1Link[i], pCurrent, 4);
		    pCurrent += 4; break;
		case _PFORMAT_VALUE_2:
		    memcpy(&mpValue2Link[i], pCurrent, 4);
		    pCurrent += 4; break;
		case _PFORMAT_DUMMY:
		    pCurrent += 4; break;
		case _PFORMAT_DUMMYx3:
		    pCurrent += 12; break;
		case _PFORMAT_DUMMYx10:
		    pCurrent += 40; break;
		}
	    }
	}

	while (bFinished[1-currentMode] == FALSE) wxMilliSleep(2);
	
	currentMode = 1-currentMode;
    }

    delete [] pItemBuffer[0];
    delete [] pItemBuffer[1];
}


void CLS::
ReadBodyBinary(ParticleFormatClone_cls *pCFormatClone){
    wxFileName CName(mstrFile);
    
    int headerSize = pCFormatClone->GetHeaderByte();

    if (CName.GetExt().IsSameAs(wxT("gz"))){
	std::string strBuf;
	UseGZ_cls::CopyToString(&strBuf, mstrFile);
	
	std::istringstream isstream(strBuf);
	isstream.seekg(headerSize, std::ios::beg);
	
	ReadBodyStream(pCFormatClone, isstream);
    }
    else{
#ifdef WIN32
	std::fstream ifstream(mstrFile.ToStdWstring(),
			      std::ios::in | std::ios::binary);
#else
	std::fstream ifstream(mstrFile.ToAscii(),
			      std::ios::in | std::ios::binary);
#endif
	if (ifstream.fail()){
	    printf("Read Header Fails!!\n");
	    return ;//exit(0);
	}

	ifstream.seekg(headerSize, std::ios::beg);
 
	int bufferSize = 10*1024*1024;//5*1024*1024;
	char *pBuffer = new char [bufferSize];
	ifstream.rdbuf()->pubsetbuf(pBuffer, bufferSize);

	ReadBodyStream(pCFormatClone, ifstream);
	
	delete [] pBuffer;
	
	ifstream.close();
    }
}

wxThread::ExitCode CLS::
Entry(){
    wxStopWatch CSW;
    CSW.Start();

    mpCDataSequenceLink->PushLock();
    DummyEntry();
    *mpbFinishedFlagLink = TRUE;
    mpCDataSequenceLink->PopLock();

    SendMenuEvent(GetPCFrameMain(0), menuFromSubThread);

    wxString strLog;
    strLog << "ReadLog :: " << mstrFile << "::" << CSW.Time() << "ms" << ENDL;
    wxLogMessage(strLog);
    
    return 0;
}

void CLS::
DummyEntry(){
    ReadBodyBinary(mpCFormatClone);

    bool bSort = mpCFormatClone->GetPCItem("PreSortByID=")->GetBodyBOOL();
    bool bRemoveDupli = mpCFormatClone->GetPCItem("RemoveDupli=")->GetBodyBOOL();
    if (bSort || bRemoveDupli){
	SortDataByID(bSort, bRemoveDupli);
    }
    
    delete mpCFormatClone;
}
