// -*- c -*- 
// "clsParticleReadBase.h"

#include "clsParticleReadBase.h"

#include "../clsArrayAndIndex.h"
#include "../clsParticleSnap.h"
#include "../../COMMON/CommonDefinitions.h"

#include <algorithm>

#define CLS ParticleReadBase_cls

CLS::
ParticleReadBase_cls() : wxThread(wxTHREAD_DETACHED){
}

void CLS::
SetBaseParam(bool *pbFinishedFlagLink, ParticleSnap_cls *pCSnap){
    mpbFinishedFlagLink = pbFinishedFlagLink;
    mpCSnap = pCSnap;
    mPNum = pCSnap->mPNum;
    mpXLink = pCSnap->mpX;
    mpVLink = pCSnap->mpV;
    mpTypeLink = pCSnap->mpType;
    mpSizeLink = pCSnap->mpSize;
    mpIDLink = pCSnap->mpID;
    mpIDLongLongLink = pCSnap->mpIDLongLong;
    mpCDataSequenceLink = pCSnap->mpCParent;
    mpValue0Link = pCSnap->mpValue[0];
    mpValue1Link = pCSnap->mpValue[1];
    mpValue2Link = pCSnap->mpValue[2];
}

void CLS::
SortDataByID(bool bSort, bool bRemoveDupli){
    if (mpIDLink == NULL && mpIDLongLongLink == NULL) return;

    bool bUseIDLongLong = (mpIDLink == NULL) ? TRUE : FALSE;

    LongLongArrayAndIndex_cls *pCIdxAndID = new LongLongArrayAndIndex_cls [mPNum];
    if (bUseIDLongLong == TRUE)
	for (long long i = 0; i < mPNum; i++){
	    pCIdxAndID[i].mIndex = i;
	    pCIdxAndID[i].mZ = mpIDLongLongLink[i];
	}
    else
	for (long long i = 0; i < mPNum; i++){
	    pCIdxAndID[i].mIndex = i;
	    pCIdxAndID[i].mZ = mpIDLink[i];
	}

    float (*pXBuf)[3] = new float [mPNum][3];
    memcpy(pXBuf, mpXLink, sizeof(float)*3*mPNum);
    
    unsigned char *pTypeBuf = NULL;
    if (mpTypeLink != NULL){
	pTypeBuf = new unsigned char [mPNum];
	memcpy(pTypeBuf, mpTypeLink, sizeof(unsigned char)*mPNum);
    }

    int *pIDBuf = NULL;
    if (mpIDLink != NULL){
	pIDBuf = new int [mPNum];
	memcpy(pIDBuf, mpIDLink, sizeof(int)*mPNum);
    }

    long long *pIDLongLongBuf = NULL;
    if (mpIDLongLongLink != NULL){
	pIDLongLongBuf = new long long [mPNum];
	memcpy(pIDLongLongBuf, mpIDLongLongLink, sizeof(long long)*mPNum);
    }

    float *pSizeBuf = NULL;
    if (mpSizeLink != NULL){
	pSizeBuf = new float [mPNum];
	memcpy(pSizeBuf, mpSizeLink, sizeof(float)*mPNum);
    }

    float *pValue0Buf = NULL;
    if (mpValue0Link != NULL){
	pValue0Buf = new float [mPNum];
	memcpy(pValue0Buf, mpValue0Link, sizeof(float)*mPNum);
    }

    float *pValue1Buf = NULL;
    if (mpValue1Link != NULL){
	pValue1Buf = new float [mPNum];
	memcpy(pValue1Buf, mpValue1Link, sizeof(float)*mPNum);
    }

    float *pValue2Buf = NULL;
    if (mpValue2Link != NULL){
	pValue2Buf = new float [mPNum];
	memcpy(pValue2Buf, mpValue2Link, sizeof(float)*mPNum);
    }
    
    if (bSort == TRUE)
	std::sort(pCIdxAndID, pCIdxAndID + mPNum);

    if (bRemoveDupli == TRUE){
	bool *pbValid = new bool [mPNum];
	for (long long i = 0; i < mPNum; i++) pbValid[i] = TRUE;
	int invalidTag = SetTagForRemoveDupli(pCIdxAndID, pbValid);

	int current;
	current = 0;
	for (long long i = 0; i < mPNum; i++){
	    if (pbValid[i] == TRUE){
		memcpy(mpXLink[current], pXBuf[pCIdxAndID[i].mIndex], sizeof(float)*3);
		current++;
	    }
	}
	if (pTypeBuf != NULL){
	    current = 0;
	    for (long long i = 0; i < mPNum; i++)
		if (pbValid[i] == TRUE){
		    mpTypeLink[current] = pTypeBuf[pCIdxAndID[i].mIndex];
		    current++;
		}
	}
	if (pIDBuf != NULL){
	    current = 0;
	    for (long long i = 0; i < mPNum; i++)
		if (pbValid[i] == TRUE){
		    mpIDLink[current] = pIDBuf[pCIdxAndID[i].mIndex];
		    current++;
		}
	}
	if (pIDLongLongBuf != NULL){
	    current = 0;
	    for (long long i = 0; i < mPNum; i++)
		if (pbValid[i] == TRUE){
		    mpIDLongLongLink[current] = pIDLongLongBuf[pCIdxAndID[i].mIndex];
		    current++;
		}
	}
	if (pSizeBuf != NULL){
	    current = 0;
	    for (long long i = 0; i < mPNum; i++)
		if (pbValid[i] == TRUE){
		    mpSizeLink[current] = pSizeBuf[pCIdxAndID[i].mIndex];
		    current++;
		}
	}
	if (pValue0Buf != NULL){
	    current = 0;
	    for (long long i = 0; i < mPNum; i++)
		if (pbValid[i] == TRUE){
		    mpValue0Link[current] = pValue0Buf[pCIdxAndID[i].mIndex];
		    current++;
		}
	}
	if (pValue1Buf != NULL){
	    current = 0;
	    for (long long i = 0; i < mPNum; i++)
		if (pbValid[i] == TRUE){
		    mpValue1Link[current] = pValue1Buf[pCIdxAndID[i].mIndex];
		    current++;
		}
	}
	if (pValue2Buf != NULL){
	    current = 0;
	    for (long long i = 0; i < mPNum; i++)
		if (pbValid[i] == TRUE){
		    mpValue2Link[current] = pValue2Buf[pCIdxAndID[i].mIndex];
		    current++;
		}
	}

	if (invalidTag != 0){
	    wxString strLog;
	    strLog << "Detect duplicated ID: num= " << invalidTag << ENDL;
	    wxLogMessage(strLog);
	}
	mpCSnap->mPNum = mPNum - invalidTag;
	// ŁA ParticleSnap ̔zTCY 
	// mPNumȂꍇ

	delete [] pbValid;
    }
    else{
	for (long long i = 0; i < mPNum; i++){
	    memcpy(mpXLink[i], pXBuf[pCIdxAndID[i].mIndex], sizeof(float)*3);
	}
	if (pTypeBuf != NULL)
	    for (long long i = 0; i < mPNum; i++)
		mpTypeLink[i] = pTypeBuf[pCIdxAndID[i].mIndex];
	if (pIDBuf != NULL)
	    for (long long i = 0; i < mPNum; i++)
		mpIDLink[i] = pIDBuf[pCIdxAndID[i].mIndex];
	if (pIDLongLongBuf != NULL)
	    for (long long i = 0; i < mPNum; i++)
	    mpIDLongLongLink[i] = pIDLongLongBuf[pCIdxAndID[i].mIndex];
	if (pSizeBuf != NULL)
	    for (long long i = 0; i < mPNum; i++)
		mpSizeLink[i] = pSizeBuf[pCIdxAndID[i].mIndex];
	if (pValue0Buf != NULL)
	    for (long long i = 0; i < mPNum; i++)
	    mpValue0Link[i] = pValue0Buf[pCIdxAndID[i].mIndex];
	if (pValue1Buf != NULL)
	    for (long long i = 0; i < mPNum; i++)
		mpValue1Link[i] = pValue1Buf[pCIdxAndID[i].mIndex];
	if (pValue2Buf != NULL)
	    for (long long i = 0; i < mPNum; i++)
		mpValue2Link[i] = pValue2Buf[pCIdxAndID[i].mIndex];
    }

    delete [] pXBuf;
    if (pTypeBuf != NULL)       delete [] pTypeBuf;
    if (pIDBuf != NULL)         delete [] pIDBuf;
    if (pIDLongLongBuf != NULL) delete [] pIDLongLongBuf;
    if (pSizeBuf != NULL)       delete [] pSizeBuf;
    if (pValue0Buf != NULL)     delete [] pValue0Buf;
    if (pValue1Buf != NULL)     delete [] pValue1Buf;
    if (pValue2Buf != NULL)     delete [] pValue2Buf;
    delete [] pCIdxAndID;
}

int CLS::
SetTagForRemoveDupli(LongLongArrayAndIndex_cls *pCIdxAndID, bool *pbValid){
    int invalid = 0;
    int prevID = -1;
    if (pCIdxAndID[0].mZ == -1) prevID = -2;
    for (long long i = 1; i < mPNum; i++){
	if (pCIdxAndID[i].mZ == prevID){
	    pbValid[i] = FALSE;
	    invalid++;
	    if (pbValid[i-1] == TRUE){
		pbValid[i-1] = FALSE;
		invalid++;
	    }
	    
	}
	prevID = pCIdxAndID[i].mZ;
    }
    
    return invalid;
}
