// -*- c -*- 
// "clsStreamlineAsciiRead.h"

#include "clsStreamlineAsciiRead.h"

#include "../clsStreamlineFormat.h"
#include "../clsStreamlineSnap.h"
#include "../clsStreamlineSequence.h"

#include "../../COMMON/clsUseGZ.h"

#include "../../../clsFrameMain.h"

#include <stdio.h>
#include <fstream>
#include <sstream>

#include <wx/filename.h>

#include "../../COMMON/CommonDefinitions.h"

#define CLS ThreadStreamlineAsciiRead

bool CLS::
SetParameters(bool *pbFinishedFlagLink, StreamlineSnap_cls *pCSnap,
	      StreamlineFormatClone_cls *pCFormatClone){

    SetBaseParam(pbFinishedFlagLink, pCSnap);
    mstrFile = pCSnap->mstrFile.Clone();
    mpCFormatClone = pCFormatClone;
    
    if (Create() != wxTHREAD_NO_ERROR) return FALSE;
    return TRUE;
}

void CLS::
ReadBodyStreamVer0(StreamlineFormatClone_cls *pCFormatClone,
		   std::istream &istream){

    /*
    if (pCFormatClone->GetPCItem("ArrayDirection=")->GetBodyINT() == _PFORMAT_ARRAY_VERTICLAL)
	return ReadBodyStreamVertical(pCFormatClone, istream);
    */

    std::string strBuf;

    //std::vector < int > VFormat = pCFormatClone->GetVFormatBody(_PFORMAT_ASCII);
    //uint numFormat = VFormat.size();

    int lineIndex = 0;
    for (int i = 0; i < mSNum; i++){
	std::getline(istream, strBuf); strBuf += " ";
	const char *ptr = strBuf.data();
	mpPosLink[i][X] = ReadFloat(&ptr, ' '); ptr++;
	mpPosLink[i][Y] = ReadFloat(&ptr, ' '); ptr++;
	mpPosLink[i][Z] = ReadFloat(&ptr, ' '); ptr++;
	mpULink[i] = ReadFloat(&ptr, ' '); ptr++;
	int delim = ReadInt(&ptr, ' ');
	if (delim != 0){
	    if (delim == -1){
		mpLineIndexLink[i] = -1;
		if (i == 0) lineIndex = 0;
		else        lineIndex++;
	    }
	    else{
		lineIndex = delim;
		mpLineIndexLink[i] = lineIndex;
	    }
	}
	else
	    mpLineIndexLink[i] = lineIndex;
    }
}

void CLS::
ReadBodyStreamVer1(StreamlineFormatClone_cls *pCFormatClone,
		   std::istream &istream){

    /*
    if (pCFormatClone->GetPCItem("ArrayDirection=")->GetBodyINT() == _PFORMAT_ARRAY_VERTICLAL)
	return ReadBodyStreamVertical(pCFormatClone, istream);
    */

    std::string strBuf;

    //std::vector < int > VFormat = pCFormatClone->GetVFormatBody(_PFORMAT_ASCII);
    //uint numFormat = VFormat.size();

    int lineIndex = 0;
    for (int i = 0; i < mSNum; i++){
	std::getline(istream, strBuf); strBuf += " ";
	const char *ptr = strBuf.data();
	mpPosLink[i][X] = ReadFloat(&ptr, ' '); ptr++;
	mpPosLink[i][Y] = ReadFloat(&ptr, ' '); ptr++;
	mpPosLink[i][Z] = ReadFloat(&ptr, ' '); ptr++;
	mpULink[i] = ReadFloat(&ptr, ' '); ptr++;
	mpValueLink[i] = ReadFloat(&ptr, ' '); ptr++;
	int delim = ReadInt(&ptr, ' ');
	if (delim != 0){
	    if (delim == -1){
		mpLineIndexLink[i] = -1;
		if (i == 0) lineIndex = 0;
		else        lineIndex++;
	    }
	    else{
		lineIndex = delim;
		mpLineIndexLink[i] = lineIndex;
	    }
	}
	else
	    mpLineIndexLink[i] = lineIndex;
    }

}

void CLS::
ReadBodyStream(StreamlineFormatClone_cls *pCFormatClone,
	       std::istream &istream){

    int version = pCFormatClone->GetPCItem("Version=")->GetBodyINT();

    switch(version){
    case 0:
	ReadBodyStreamVer0(pCFormatClone, istream);
	break;
    case 1:
	ReadBodyStreamVer1(pCFormatClone, istream);
	break;
    }
}

void CLS::
ReadBodyAscii(StreamlineFormatClone_cls *pCFormatClone){
    wxFileName CName(mstrFile);
    
    int headerLine = 0;//pCFormatClone->GetHeaderLine();
    std::string strBuf;

    if (CName.GetExt().IsSameAs(wxT("gz"))){
	std::string strBuf;
	UseGZ_cls::CopyToString(&strBuf, mstrFile);
	
	std::istringstream isstream(strBuf);
	for (int i = 0; i < headerLine; i++)
	    std::getline(isstream, strBuf);
	
	ReadBodyStream(pCFormatClone, isstream);
    }
    else{
#ifdef WIN32
	std::fstream ifstream(mstrFile.ToStdWstring(),
			      std::ios::in | std::ios::binary);
#else
	std::fstream ifstream(mstrFile.ToAscii(),
			      std::ios::in | std::ios::binary);
#endif
	if (ifstream.fail()){
	    printf("Read Header Fails!!\n");
	    return ;//exit(0);
	}

	int bufferSize = 10*1024*1024;//5*1024*1024;
	char *pBuffer = new char [bufferSize];
	ifstream.rdbuf()->pubsetbuf(pBuffer, bufferSize);

	for (int i = 0; i < headerLine; i++)
	    std::getline(ifstream, strBuf);

	ReadBodyStream(pCFormatClone, ifstream);
	
	delete [] pBuffer;
	
	ifstream.close();
    }
}

wxThread::ExitCode CLS::
Entry(){
    mpCDataSequenceLink->PushLock();
    DummyEntry();
    *mpbFinishedFlagLink = TRUE;
    mpCDataSequenceLink->PopLock();

    SendMenuEvent(GetPCFrameMain(0), menuFromSubThread);
    
    return 0;
}

void CLS::
DummyEntry(){
    ReadBodyAscii(mpCFormatClone);
    delete mpCFormatClone;
}
