// -*- c -*- 
// "clsDataSequence.h"

#include "clsDataSequence.h"
#include "clsVolumeSnap.h"
#include "clsParticleSnap.h"
#include "clsStreamlineSnap.h"
#include "clsDataSnap.h"

#include "../MISC/clsMisc.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../PREFERENCE/clsPreference.h"
#include "../TIMELINE/clsTimelineBase.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"

#include <wx/filename.h>

#include <stdlib.h>

#define CLS DataSequence_cls

CLS::
DataSequence_cls(int type){
    mDataType = type;
    mApCSnap = NULL;
    mReadingLock = 0;
    mDrawCounter = 0;
    mTotalUsedMemoryMB = 0;
    mDrawCounter = -1;

    mCColorBase = wxColor(204, 204, 164);
    mCColorWA = wxColor(248, 248, 196);
}

CLS::
~DataSequence_cls(){
    //ClearSequence(); 
    // hŃNAĂ͂
}

void CLS::
ClearSequence(){
    if (mApCSnap != NULL){
	if (mDataType == _DATA_VOLUME){
	    int seqLen = mSeqLen;
	    if (seqLen == -1) seqLen = mSeqLenT;
	    for (int i = 0; i < seqLen; i++)
		if (mApCSnap[i] != NULL){
		    delete (VolumeSnap_cls*)mApCSnap[i];
		    mApCSnap[i] = NULL;
		}
	}
	else if (mDataType == _DATA_PARTICLE){
	    int seqLen = mSeqLen;
	    for (int i = 0; i < seqLen; i++)
		if (mApCSnap[i] != NULL){
		    delete (ParticleSnap_cls*)mApCSnap[i];
		    mApCSnap[i] = NULL;
		}
	}
	else if (mDataType == _DATA_STREAMLINE){
	    int seqLen = mSeqLen;
	    for (int i = 0; i < seqLen; i++)
		if (mApCSnap[i] != NULL){
		    delete (StreamlineSnap_cls*)mApCSnap[i];
		    mApCSnap[i] = NULL;
		}
	}
	else{
	    for (int i = 0; i < mSeqLen; i++)
		if (mApCSnap[i] != NULL){
		    delete mApCSnap[i];
		    mApCSnap[i] = NULL;
		}
	}
    }
    delete [] mApCSnap;
    mApCSnap = NULL;
    mSeqLen = -1;
    mSeqLenT = -1;
    mSeqLenZ = -1;
}

int CLS::
GetRemappedIndex(int index, ParamGroup_cls *pCRemapParamGroup, double NT,
		 bool bRepeat){
    if (pCRemapParamGroup != NULL){
	index = floor(pCRemapParamGroup->GetVInterpolatedValue(NT)[0] + 1e-6);
    }
    
    if (mSeqLen != -1)
	if (index >= mSeqLen){
	    if (bRepeat) index = Misc_cls::Surplus(index, mSeqLen);
	    else         index = mSeqLen - 1;
	}
    if (mSeqLenT != -1)
	if (index >= mSeqLenT){
	    if (bRepeat) index = Misc_cls::Surplus(index, mSeqLenT);
	    else         index = mSeqLenT - 1;
	}
    
    return index;
}

void CLS::
MarkPriority(ParamGroup_cls *pCRemapParamGroup, bool bRepeat){
    if (mApCSnap == NULL) return;

    if (mSeqLen == 1 || mSeqLenT == 1){
	if (mApCSnap[0] != NULL)
	    mApCSnap[0]->mbPriority = FALSE;
	return;
    }

    int maxFrame = GetPCTimeCtrl()->mCFrameMax.GetBodyINT();
    int priorityLength = GetPCPreference()->mCPrefetchLength.GetBodyINT() + 1;
    int currentFrame = GetPCTimeCtrl()->mCFrameCurrent.GetBodyINT();
    for (int i = 0; i < maxFrame; i++){
	double NT = (double)i / maxFrame;
	int index = GetRemappedIndex(i, pCRemapParamGroup, NT, bRepeat);
	if (index < 0) continue;
	DataSnap_cls *pCSnap = mApCSnap[index];
	if (pCSnap == NULL) continue;
	if (i >= currentFrame && i < currentFrame + priorityLength)
	    pCSnap->mbPriority = TRUE;
	else
	    pCSnap->mbPriority = FALSE;
	pCSnap->mDistanceToCurrent = abs(i - currentFrame);
    }
}

int CLS::
GetOldestDrawnIndex(){
    
    if (mApCSnap == NULL) return -1;
    if (mSeqLen == 1) return -1;
    int oldest = 999999999;
    int oldestIndex = -1;
    for (int i = 0; i < mSeqLen; i++){
	DataSnap_cls *pCSnap = mApCSnap[i];
	if (pCSnap == NULL) continue;
	if (pCSnap->mbPriority == TRUE) continue;
	int drawCounter = pCSnap->mDrawCounter - pCSnap->mDistanceToCurrent * 0.5;
	
	if (drawCounter < oldest){
	    oldest = drawCounter;
	    oldestIndex = i;
	}
    }
    return oldestIndex;
}

bool CLS::
IsDataValid(){
    if (mbReadStatus == FALSE) return FALSE;
    if (mbReadStatusDual == FALSE) return FALSE;
    return TRUE;
}

void CLS::
DrawTL(wxDC& dc, TLPanelGeom_cls* pCGeom,
       wxBitmap &CBitmapBuffer,
       int sashX, int posY, int panelWidth, int height,
       double normalizedWAMin, double normalizedWAMax,
       ParamGroup_cls *pCRemapParamGroup, bool bRepeat){
    
    if (CBitmapBuffer.GetWidth() != panelWidth ||
	CBitmapBuffer.GetHeight() != height)
	CBitmapBuffer.Create(panelWidth, height); 
    wxMemoryDC CMemDC(CBitmapBuffer);

    DrawTLBackground(CMemDC, pCGeom, sashX, panelWidth, height, 
		     normalizedWAMin, normalizedWAMax);

    int maxFrame = GetPCTimeCtrl()->mCFrameMax.GetBodyINT();
    int frameMin = pCGeom->mTLRegionMin * maxFrame;
    int frameMax = pCGeom->mTLRegionMax * maxFrame;

    MarkPriority(pCRemapParamGroup, bRepeat);
    int oldestIndex = GetOldestDrawnIndex();
    
    if (mApCSnap != NULL)
	for (int i = frameMin; i < frameMax; i++){
	    double NT = (double)i / maxFrame;
	    int index = GetRemappedIndex(i, pCRemapParamGroup, NT, bRepeat);
	    int state = 0;
	    if (index >= 0){
		if (mApCSnap[index] != NULL && mApCSnap[index]->mbReadFinished == TRUE) state++;
		
		if (index < mSeqLen - 1)
		    if (mApCSnap[index+1] != NULL && mApCSnap[index+1]->mbReadFinished == TRUE) state++;
		if (index < mSeqLenT - 1)
		    if (mApCSnap[index+1] != NULL && mApCSnap[index+1]->mbReadFinished == TRUE) state++;
		
		if (index == oldestIndex) state = 3;
	    }

	    if (state > 0){
		double XNormal0  = GetPCTimeCtrl()->GetNormalizedValue(i, 0);
		double XNormal1 = GetPCTimeCtrl()->GetNormalizedValue(i+1, 0);
		int posX0 = pCGeom->GetDevCoordFromN(XNormal0, sashX);
		int posX1 = pCGeom->GetDevCoordFromN(XNormal1, sashX);
		switch(state){
		case 1: CMemDC.SetPen(wxColor(128,255,128)); break;
		case 2: CMemDC.SetPen(wxColor(0, 255, 0)); break;
		case 3: CMemDC.SetPen(wxColor(255, 0, 0)); break;
		}

		// foOp
		if (mDataType == _DATA_PARTICLE){
		    ParticleSnap_cls *pCSnap = (ParticleSnap_cls*)mApCSnap[index];
		    if (pCSnap != NULL && pCSnap->mbNextIndexCalced == TRUE)
			CMemDC.SetPen(wxColor(255, 255, 0));
		    else
			CMemDC.SetPen(wxColor(0, 255, 255));
		}

		    
		CMemDC.DrawRectangle(posX0, 0, posX1-posX0, height);
	    }
	}

    //int memoryMB = GetTotalUsedMemoryMB();
    wxString strMemoryMB; strMemoryMB << mTotalUsedMemoryMB;
    int posX = pCGeom->GetDevCoordFromN(0, sashX);
    CMemDC.DrawText(strMemoryMB, posX, 0);

    dc.Blit(sashX, posY, panelWidth-sashX, height, &CMemDC, sashX, 0);

}

void CLS::
CalcMaxOffset(bool bSequence){
    ClearSequence();
    mbReadStatusDual = TRUE;
    if (bSequence == FALSE){
	mbTimeDepend = FALSE;
	mSeqLen = 1;
	CheckSingleFileValidity();
    }
    else 
	FileSequence_cls::CalcMaxOffset();
    if (mSeqLen == 0) return;
    
    mApCSnap = new DataSnap_cls* [mSeqLen];
    for (int i = 0; i < mSeqLen; i++) mApCSnap[i] = NULL;

    GetPCMain()->UpdatePanels("PROPERTY_OBJECT");
}

void CLS::
CalcMaxOffsetDual(){
    ClearSequence();
    mbReadStatus = TRUE;
    FileSequenceDual_cls::CalcMaxOffsetDual();
    if (mSeqLenT == 0) return;

    mApCSnap = new DataSnap_cls* [mSeqLenT];
    for (int i = 0; i < mSeqLenT; i++) mApCSnap[i] = NULL;

    GetPCMain()->UpdatePanels("PROPERTY_OBJECT");
}

wxString CLS::
GetStrInfo(bool bZStack){
    if (mstrFormat.IsEmpty() == TRUE) return wxES;
    wxString strInfo, strFormat;
    if (bZStack == TRUE) strFormat = mstrFormatDual;
    else                 strFormat = mstrFormat;
	
    wxFileName CFormat(strFormat);
    strInfo << CFormat.GetPath(wxPATH_GET_VOLUME | wxPATH_GET_SEPARATOR) << ENDL;
    strInfo << CFormat.GetFullName() << ENDL;
    strInfo << mInit << " - " << mInit + mSeqLen - 1 << ENDL; 
    return strInfo;
}

void CLS::
PushLock(){
    mReadingLock++;
}

void CLS::
PopLock(){
    ASSERT(mReadingLock > 0);
    mReadingLock--;
}
