// (setq-c-offset 4)
// "clsParticleForRender.h"

#include "clsParticleForRender.h"
#include "clsParticleSnap.h"
#include "INTERPTHREAD/clsParticleForRenderInterpThread.h"

#include "../COMMON/clsHalfFloat.h"
#include "../COMMON/clsSortParallel.h"
#include "../MODIFIER/clsModifyApparentSize.h"
#include "../MODIFIER/clsModifyParticleSize.h"
#include "../MODIFIER/clsParticleBoolean.h"
#include "../MODIFIER/clsParticleExclude.h"
#include "../MODIFIER/clsParticleFiltering.h"
#include "../MODIFIER/clsParticleMerge.h"
#include "../SCENEOBJ/clsMeshObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/DATAVIEWER/clsMaterialParticle.h"
#include "../SCENEOBJ/DATAVIEWER/clsParticle.h"
#include "../SCENEOBJ/DATAVIEWER/clsViewerClone.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"
#include "../COMMON/CommonDefinitionsShader.h"

#include <algorithm>

#define CLS ThreadFillInitDataParticle_cls

void CLS::
SetParameters(float *pFilterAlpha, ArrayAndIndex_cls *pIndexAndZ, float *pSizeMod, bool bInitializeAtProcIndexAndZ){
    mpFilterAlpha = pFilterAlpha;
    mpIndexAndZ = pIndexAndZ;
    mpSizeMod = pSizeMod;
    mbInitializeAtProcIndexAndZ = bInitializeAtProcIndexAndZ;
}

wxThread::ExitCode CLS::
Entry(){
    DummyEntry();
    
    return 0;
}

void CLS::
DummyEntry(){
    const int PIni = mPIni;
    const int PFin = mPFin;

    float * const pFilterAlpha = mpFilterAlpha;
    ArrayAndIndex_cls * const pIndexAndZ = mpIndexAndZ;
    float * const pSizeMod = mpSizeMod;
    
    if (mbInitializeAtProcIndexAndZ == TRUE){
	for (int i = PIni; i < PFin; i++) 
	    if (pFilterAlpha[i] != 0) pFilterAlpha[i] = 1;
    }
    else
	for (int i = PIni; i < PFin; i++) pFilterAlpha[i] = 1;
    
    // ProcIndexAndZ łΏ[Ȃ̂
    if (mbInitializeAtProcIndexAndZ == TRUE){
	if (pSizeMod != NULL)
	    for (int i = PIni; i < PFin; i++) mpSizeMod[i] = 1;
    }
    
    if (mbInitializeAtProcIndexAndZ == FALSE)
	for (int i = PIni; i < PFin; i++)
	    pIndexAndZ[i].mIndex = i;
    
    *mpbFinishedFlagLink = TRUE;
}

#define CLS ThreadCurveDependOnSort_cls

void CLS::
SetParameters(float *pClrCurveS, float *pClrCurveT, 
	      int *clrCurveType, int *clrCurveDepend,
	      unsigned char *pType, ArrayAndIndex_cls *pIndexAndZ,
	      std::vector < double > *VRange,
	      int numMat, double NT){
    mpClrCurveS = pClrCurveS;
    mpClrCurveT = pClrCurveT;
    mpClrCurveType = clrCurveType;
    mpClrCurveDepend = clrCurveDepend;
    mpType = pType;
    mpIndexAndZ = pIndexAndZ;
    mpVRange = VRange;
    mNumMat = numMat;
    mNT = NT;
}

double CLS::
GetRemapX(std::vector < double > &VRange, double x0/*, double NT*/){
    double width = VRange[1] - VRange[0];
    if (width == 0) return 1;
    return (x0 - VRange[0])/width;
}

wxThread::ExitCode CLS::
Entry(){
    const int PIni = mPIni;
    const int PFin = mPFin;

    float * const pClrCurveS = mpClrCurveS;
    float * const pClrCurveT = mpClrCurveT;
    int * const clrCurveType = mpClrCurveType;
    int * const clrCurveDepend = mpClrCurveDepend;
    unsigned char * const pType = mpType;
    ArrayAndIndex_cls * const pIndexAndZ = mpIndexAndZ;
    std::vector < double > * const VRange = mpVRange;
    const int numMat = mNumMat;
    const double NT = mNT;

    for (int i = PIni; i < PFin; i++){
	int type = pType[i];
	if (clrCurveType[type] == 1 &&
	    clrCurveDepend[type] == _COLOR_CURVE_DEPEND_TYPE_SORT){
	    double value = 0, x0;
	    value = -pIndexAndZ[i].mZ;
	    x0 = GetRemapX(VRange[type], value/*, NT*/);
	    
	    pClrCurveS[i] = x0;
	    pClrCurveT[i] = ((float)type + 0.5) / numMat / 3;
	}
    }

    *mpbFinishedFlagLink = TRUE;
    return 0;
}

#define CLS ParticleForRender_cls

CLS::
ParticleForRender_cls(){
    mFrame = -1;
    mPNum = -1;
    mPNumAllocated = -1;
    mpX = NULL;
    mpSize = NULL;
    mpSizeMod = NULL;
    mpClrCurveS = mpClrCurveT = NULL;
    mpFilterAlpha = NULL;
    mpSizeAlpha = NULL;
    mpParamBit = NULL;
    mpType = NULL;
    mpID = NULL; mpIDLongLong = NULL;
    for (int i = 0; i < _MAX_PVALUE; i++) mpValue[i] = NULL;
    mpIndexAndZ = NULL;
    mSortTypeUsed = -1;
    mDrawCounter = -1;
    mDrawCounterIndex = -1;
    mRangeMin0 = -1; mRangeMax0 = -1;
    mpTextureData = NULL;
    mpTextureDataFloat = NULL;
    mTextureDataProcCounter = 0;
    mTextureDataProcCounterStored = 0;
    mTextureDataSize = 0;
}

CLS::
~ParticleForRender_cls(){ 
    Clear(); 
    if (mpTextureData != NULL) { delete [] mpTextureData; mpTextureData = NULL;}
    if (mpTextureDataFloat != NULL){
	delete [] mpTextureDataFloat; mpTextureDataFloat = NULL; }
}

void CLS::
DeleteAndSetNullFloatArray(float **p){
    if (*p != NULL){ delete [] (*p); *p = NULL; }
}
    
void CLS::
Clear(){
    if (mpX != NULL){ 
	delete [] mpX; mpX = NULL; }
    DeleteAndSetNullFloatArray(&mpSize);
    DeleteAndSetNullFloatArray(&mpSizeMod);
    DeleteAndSetNullFloatArray(&mpClrCurveS);
    DeleteAndSetNullFloatArray(&mpClrCurveT);
    DeleteAndSetNullFloatArray(&mpFilterAlpha);
    DeleteAndSetNullFloatArray(&mpSizeAlpha);
    
    if (mpParamBit != NULL) {
	delete [] mpParamBit; mpParamBit = NULL; }
    if (mpType != NULL){ 
	delete [] mpType; mpType = NULL; }
    if (mpID != NULL){
	delete [] mpID; mpID = NULL; }
    if (mpIDLongLong != NULL){
	delete [] mpIDLongLong; mpIDLongLong = NULL; }
    if (mpIndexAndZ != NULL) {
	delete [] mpIndexAndZ; mpIndexAndZ = NULL; }
    
    for (int i = 0; i < _MAX_PVALUE; i++)
	DeleteAndSetNullFloatArray(&mpValue[i]);
	
    mPNum = 0;
    mPNumAllocated = -1;
}

wxString CLS::
GetStrInfo(int infoIndex){
    wxString strInfo;
    strInfo << "PNum(current):" << mPNum << ENDL;
    if (infoIndex >= 0 && infoIndex < mPNum){
	if (mpValue[0] != NULL)
	    strInfo << "Value0(current): " << mpValue[0][infoIndex] << ENDL; 
	if (mpValue[1] != NULL)
	    strInfo << "Value1(current): " << mpValue[1][infoIndex] << ENDL; 
	if (mpValue[2] != NULL)
	    strInfo << "Value2(current): " << mpValue[2][infoIndex] << ENDL; 
    }
    return strInfo;
}

long long CLS::
SearchID(long long id){
    // ID\[gĂ邱ƂOƂB
    if (mpIDLongLong != NULL){
	for (int i = 0; i < mPNum; i++){
	    if (i < mPNum - 1000)
		if (mpIDLongLong[i+1000] < id) i+= 1000;
	    if (mpIDLongLong[i] == id) return i;
	    if (mpIDLongLong[i]  > id) return -1;
	}
    }

    if (mpID != NULL){
	for (int i = 0; i < mPNum; i++){
	    if (i < mPNum - 1000)
		if (mpID[i+1000] < id) i+= 1000;
	    if (mpID[i] == id) return i;
	    if (mpID[i]  > id) return -1;
	}
    }

    return -1;
}

bool CLS::
CheckUpdate(int frame, float factor, wxString strStatusLog){
    bool bUpdate = TRUE;
    if (mFrame == frame && mFactor == factor &&
	mstrStatusLog.IsSameAs(strStatusLog) ) bUpdate = FALSE;
    else{
	/*
	wxString strLog;
	strLog << "frame=" << mFrame << "::" << frame << "factor=" << mFactor << "::" << factor << ENDL;
	strLog << "statusLog A::" << mstrStatusLog << ENDL;
	strLog << "statusLog B::" << strStatusLog << ENDL;
	wxLogMessage(strLog);
	*/
    }
    
    mFrame = frame;
    mFactor = factor;
    mstrStatusLog = strStatusLog;
    return bUpdate;
}

bool CLS::
CheckUpdateSort(int sortType, PositionAndDirection_cls CSortPos, float sortFreq, wxString &strLog){
    bool bUpdate = FALSE;

    if (mSortTypeUsed != sortType) bUpdate = TRUE;

    if (sortType == _SORT_Z &&
	(mCSortPos.IsSimilarTgtDirection(CSortPos, sortFreq) == FALSE ||
	 sortFreq == 1))
	bUpdate = TRUE;

    if ((sortType == _SORT_DISTANCE || sortType == _SORT_OBJECT ||
	 sortType == _SORT_DISTANCE_INV || sortType == _SORT_DISTANCE_ONLY) &&
	mCSortPos.IsSimilarPosition(CSortPos, 1e-6) == FALSE) bUpdate = TRUE;
    
    if (bUpdate == TRUE){

	//strLog << "Sort Update Required" << ENDL;
	
	mCSortPos = CSortPos;
	mSortTypeUsed = sortType;
    }

    return bUpdate;
}

void CLS::
PrepareColorCurveTexture(ViewerClone_cls *pCParticleClone, double NT){
    int numClrCurve = pCParticleClone->GetNumAttribute("ColorCurve_cls");
    int numMat = pCParticleClone->GetNumAttribute("MaterialParticle_cls");
    
    mCColorCurveTexture.ClearImage();
    //float *pDataFloat = new float [numMat * 255 * 4 * 3];
    
    if (mpTextureData != NULL && mTextureDataSize != numMat){
	delete [] mpTextureData; mpTextureData = NULL;
	delete [] mpTextureDataFloat; mpTextureDataFloat = NULL;
    }
    
    if (mpTextureData == NULL){
	mpTextureData = new unsigned short [numMat * 255 * 4 * 3];
	mpTextureDataFloat = new float [numMat * 255 * 4 * 3];
    }
    float * const pDataFloat = mpTextureDataFloat;
    
    mTextureDataSize = numMat;
    
    for (int i = 0; i < numMat; i++){
	AttributeClone_cls *pCMat = pCParticleClone->GetPCAttribCloneFromAttribName("MaterialParticle_cls", i);
	wxString strMat = pCMat->mCstrName.GetBodySTR();
	
	int clrCurveType = pCMat->GetPCItem("ColorCurveType=")->GetBodyINT();

	std::vector < double > VEmit = pCParticleClone->GetPCParamGroup(strMat, "Emission")->GetVInterpolatedValue(NT);
	std::vector < double > VDiffuse = pCParticleClone->GetPCParamGroup(strMat, "Diffuse")->GetVInterpolatedValue(NT);
	std::vector < double > VSpecular = pCParticleClone->GetPCParamGroup(strMat, "Specular")->GetVInterpolatedValue(NT);
	
	float alphaFactor = pCParticleClone->GetPCParamGroup(strMat, "AlphaFactor")->GetVInterpolatedValue(NT)[0];
	
	int clrCurveIndex = pCMat->GetPCItem("ColorCurveIndex=")->GetBodyINT();
	if (clrCurveIndex >= numClrCurve) clrCurveIndex = numClrCurve - 1;

	AttributeClone_cls *pCClrCurve = pCParticleClone->GetPCAttribCloneFromAttribName("ColorCurve_cls", clrCurveIndex);
	wxString strClr = pCClrCurve->mCstrName.GetBodySTR();
	
	ParamGroup_cls *pCClr = pCParticleClone->GetPCParamGroup(strClr, wxES);
	
	int indexy = 255 * 4 * i;
	if (clrCurveType == 0){
	    for (int ix = 0; ix < 255; ix++){
		int index = indexy + ix * 4;
		pDataFloat[index+0] = VEmit[0];
		pDataFloat[index+1] = VEmit[1];
		pDataFloat[index+2] = VEmit[2];
		pDataFloat[index+3] = alphaFactor;
	    }
	}
	else{
	    for (int ix = 0; ix < 255; ix++){
		double x = (double)ix/255;
		std::vector < double > VClr = pCClr->GetVInterpolatedValue(x);
		int index = indexy + ix * 4;
		pDataFloat[index+0] = VClr[0]*VEmit[0];
		pDataFloat[index+1] = VClr[1]*VEmit[1];
		pDataFloat[index+2] = VClr[2]*VEmit[2];
		pDataFloat[index+3] = VClr[3]*alphaFactor;
	    }
	}

	indexy = 255 * 4 * i + (255 * 4 * numMat);
	for (int ix = 0; ix < 255; ix++){
	    int index = indexy + ix * 4;
	    pDataFloat[index+0] = VDiffuse[0];
	    pDataFloat[index+1] = VDiffuse[1];
	    pDataFloat[index+2] = VDiffuse[2];
	    pDataFloat[index+3] = alphaFactor;
	}
	
	indexy = 255 * 4 * i + (255 * 4 * numMat)*2;
	for (int ix = 0; ix < 255; ix++){
	    int index = indexy + ix * 4;
	    pDataFloat[index+0] = VSpecular[0];
	    pDataFloat[index+1] = VSpecular[1];
	    pDataFloat[index+2] = VSpecular[2];
	    pDataFloat[index+3] = alphaFactor;
	}
    }

    mCColorCurveTexture.mMagFilter = GL_NEAREST;
    mCColorCurveTexture.mMinFilter = GL_NEAREST;
    
    HalfFloat_cls::Singles2Half(mpTextureData, pDataFloat, numMat*255 * 4 * 3);
    mTextureDataProcCounter++;

    //delete [] pDataFloat;
    //delete [] pData;
}

void CLS::
SetColorCurveTexture(){
    if (mTextureDataProcCounterStored == mTextureDataProcCounter) return;

    mCColorCurveTexture.CopyFromDataRGBA16HF(mpTextureData, 255, mTextureDataSize * 3);
    mTextureDataProcCounterStored = mTextureDataProcCounter;
}

int CLS::
GetDownsize(ViewerClone_cls *pCParticleClone, int &downsizeType){
    int downsize = 1;
    int numAttribute = pCParticleClone->mVpCAttribClone.size();
    for (uint i = 0; i < numAttribute; i++){
	AttributeClone_cls *pCAttribute = pCParticleClone->mVpCAttribClone[i];
	if (pCAttribute->IsAttributeValidAndSameAs("ParticleDownsizing_cls")){
	    downsize *= pCAttribute->GetPCItem("Downsize=")->GetBodyINT();
	    downsizeType = pCAttribute->GetPCItem("DownsizeType=")->GetBodyINT();
	}
    }
    return downsize;
}

int CLS::
GetNumDownsized(ParticleSnap_cls *pCSnap1, int downsize, int downsizeType){
    
    if (downsizeType == 1){
	return pCSnap1->mPNum;
    }
    
    if (downsizeType == 0){
	// Jグ
	return (pCSnap1->mPNum + (downsize - 1) )/ downsize;
    }
}

int CLS::
GetDownsizeIncrement(ParticleSnap_cls *pCSnap1, int downsize, int downsizeType,
		     int PFin, int PIni){
    if (downsizeType == 1)
	return PFin - PIni;
    
    if (downsizeType == 0){
	int increment = 0;
	for (int i = PIni; i < PFin; i++)
	    if (i % downsize == 0) increment++;
	/*
	return (PFin / downsize - PIni / downsize);
	*/
	return increment;
    }
}

void CLS::
AllocateFull(long long numParticle, bool bUseSizeAlpha, bool bUseFilterAlpha, bool bUseSizeMod, bool bUseID, bool bUseIDLongLong, bool bUseValue[_MAX_PVALUE]){
    if (mpX != NULL){
	int PNum = mPNum;
	Clear(); // mPNum  0 ɂȂ̂Œ
	mPNum = PNum;
    }
    
    mpX = new float [mPNum][3];
    mpType = new unsigned char[mPNum];
    mpSize = new float [mPNum];
    mpClrCurveS = new float [mPNum];
    mpClrCurveT = new float [mPNum];
    if (bUseSizeAlpha == TRUE) mpSizeAlpha = new float [mPNum];
    if (bUseFilterAlpha == TRUE) mpFilterAlpha = new float [mPNum];
    if (bUseSizeMod == TRUE) mpSizeMod = new float [mPNum];
    if (bUseID == TRUE) mpID = new int [mPNum];
    if (bUseIDLongLong == TRUE) mpIDLongLong = new long long [mPNum];
    for (int i = 0; i < _MAX_PVALUE; i++)
	if (bUseValue[i] == TRUE) mpValue[i] = new float [mPNum];
    
    mpParamBit = new unsigned char [mPNum];
    mpIndexAndZ = new ArrayAndIndex_cls [mPNum];
}

void CLS::
ReallocationFloat(bool bUse, float **p){
    if (bUse == TRUE){
	if (*p == NULL) *p = new float [mPNum]; }
    else
	if (*p != NULL) DeleteAndSetNullFloatArray(p);
}

void CLS::
Reallocation(long long numParticle, bool bUseSizeAlpha, bool bUseFilterAlpha, bool bUseSizeMod, bool bUseID, bool bUseIDLongLong, bool bUseValue[_MAX_PVALUE]){
    ReallocationFloat(bUseSizeAlpha,   &mpSizeAlpha);
    ReallocationFloat(bUseFilterAlpha, &mpFilterAlpha);
    ReallocationFloat(bUseSizeMod,     &mpSizeMod);
        
    if (bUseID == TRUE){
	if (mpID == NULL) mpID = new int [mPNum]; }
    else
	if (mpID != NULL){
	    delete [] mpID; mpID = NULL; }

    if (bUseIDLongLong == TRUE){
	if (mpIDLongLong == NULL) mpIDLongLong = new long long [mPNum]; }
    else
	if (mpIDLongLong != NULL){
	    delete [] mpIDLongLong; mpIDLongLong = NULL; }

    for (int i = 0; i < _MAX_PVALUE; i++)
	ReallocationFloat(bUseValue[i], &mpValue[i]);
    
}

void CLS::
Allocate(long long numParticle, bool bUseSizeAlpha, bool bUseFilterAlpha, bool bUseSizeMod, bool bUseID, bool bUseIDLongLong, bool bUseValue[_MAX_PVALUE]){
    
    mPNum = numParticle;
    
    if (numParticle != mPNumAllocated) 
	AllocateFull(numParticle, bUseSizeAlpha, bUseFilterAlpha, bUseSizeMod, bUseID, bUseIDLongLong, bUseValue);
    else{
	Reallocation(numParticle, bUseSizeAlpha, bUseFilterAlpha, bUseSizeMod, bUseID, bUseIDLongLong, bUseValue);
    }

    int threadNumber;
    bool *bAFinished = GetNewPCbFinishList(threadNumber, mPNum);
    ThreadFillInitDataParticle_cls **ppCThread = 
	new ThreadFillInitDataParticle_cls* [threadNumber];
    
    for (int core = 0; core < threadNumber; core++){
	ppCThread[core] = 
	    new ThreadFillInitDataParticle_cls;

	int PIni, PFin;
	GetIniAndFin(PIni, PFin, core, threadNumber, mPNum);

	ppCThread[core]->SetParameters(mpFilterAlpha, mpIndexAndZ, mpSizeMod);
	ppCThread[core]->CreateThread(&bAFinished[core], PIni, PFin);
	if (threadNumber == 1){
	    ppCThread[core]->DummyEntry();
	    delete ppCThread[core];
	}
	else
	    ppCThread[core]->Run();
    }

    WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 1);
    delete [] ppCThread;

    /*
    for (int i = 0; i < numParticle; i++){
	mpFilterAlpha[i] = 1;
	mpIndexAndZ[i].mIndex = i;
    }
    if (mpSizeMod != NULL)
	for (int i = 0; i < numParticle; i++) mpSizeMod[i] = 1;
    */
    
    mPNumAllocated = mPNum;
}

void CLS::
PrepareCommonRenderParameter(ViewerClone_cls* pCParticleClone){
    ConfItemList_cls &CConfList = pCParticleClone->mCConfList;
    mstrParticleTexture[0] = CConfList.GetPCItem("TexturePath0=")->GetBodySTR();
    mstrParticleTexture[1] = CConfList.GetPCItem("TexturePath1=")->GetBodySTR();
    mstrParticleTexture[2] = CConfList.GetPCItem("TexturePath2=")->GetBodySTR();
    int textureNum = CConfList.GetPCItem("TextureNum=")->GetBodyINT();
    for (int i = textureNum; i < 3; i++)
	mstrParticleTexture[i] = "None";

    bool bDependDistance = CConfList.GetPCItem("DependDistance=")->GetBodyBOOL();
    if (bDependDistance == TRUE)
	mDistanceBase = CConfList.GetPCItem("DistanceBase=")->GetBodyDOUBLE();
    else
	mDistanceBase = 0;
}

bool CLS::
IsHide(int type, int shaderPoint, int shaderSphere, int shaderBillboard){
    return 
	(type == _PARTICLE_POINT && shaderPoint == _POINTSHADER_HIDE) ||
	(type == _PARTICLE_SPHERE && shaderSphere == _SPHERESHADER_HIDE) ||
	(type == _PARTICLE_BILLBOARD && shaderBillboard == _BILLBOARDSHADER_HIDE); 
}

void CLS::
PrepareData(ParticleSnap_cls *pCSnap1, ParticleSnap_cls *pCSnap2,
	    ParticleSnap_cls *pCSnap0, ParticleSnap_cls *pCSnap3,
	    int frame, int subFrame, int seq, double factor, 
	    ViewerClone_cls *pCParticleClone,
	    ViewerClone_cls *pCParticleCloneForMat,
	    double NT, PositionAndDirection_cls CSortPosOrg, 
	    bool bUseIDLongLong, bool bStoreValue, bool bNext){

    uint numAttribute = pCParticleClone->mVpCAttribClone.size();
    int numClrCurve = pCParticleClone->GetNumAttribute("ColorCurve_cls");
    int numMat = pCParticleCloneForMat->GetNumAttribute("MaterialParticle_cls");
    if (numMat > _MAX_PARTICLE_TYPE) numMat = _MAX_PARTICLE_TYPE;

    ASSERT(pCSnap1 != NULL);

    wxStopWatch CSW;
    CSW.Start();

    // CSortPos[JWɏ悹ȂƂȂB
    double invMatrix16[16];
    TensorMath_cls::GetInvMatrix4x4(pCParticleClone->mMatrix16, invMatrix16);
    double sortPosMatrix[16];
    CSortPosOrg.GetMatrix16(sortPosMatrix);
    TensorMath_cls::Multiply(invMatrix16, sortPosMatrix);
    PositionAndDirection_cls CSortPos(sortPosMatrix);

    wxString strLog, strTime;
    if (bNext == TRUE) 
	strTime << "Next";
    else
	strTime << "Now";

    strLog << "Prepare for " << strTime << ENDL;

    wxString strStatusLog;

    strStatusLog << pCSnap1->mstrLoadedFileHeader << "::";

    PrepareCommonRenderParameter(pCParticleCloneForMat);
    
    int clrCurveType[_MAX_PARTICLE_TYPE] = {};
    int clrCurveDepend[_MAX_PARTICLE_TYPE] = {};
    unsigned char paramBit[_MAX_PARTICLE_TYPE] = {};
    float clrEmit[_MAX_PARTICLE_TYPE][3] = {};
    ParamGroup_cls* pCClr[_MAX_PARTICLE_TYPE] = {};
    std::vector < double > VRange[_MAX_PARTICLE_TYPE];
    
    ConfItem_cls *pCClassifyType = pCParticleClone->GetPCItem("ClassifyType=");
    ConfItem_cls *pCClassify = pCParticleClone->GetPCItem("Classify=");
    int classifyType = pCClassifyType->GetBodyINT();
    wxString strClassify = pCClassify->GetBodySTR();
    strStatusLog << pCClassify->mProcCounter;
    strStatusLog << pCClassifyType->mProcCounter;
    strStatusLog << pCParticleCloneForMat->GetPCItem("PointSizeFactor=")->GetBodySTR();
    strStatusLog << pCParticleCloneForMat->GetPCItem("AlphaCrit=")->GetBodySTR();
    
    int interpType = pCParticleClone->GetPCItem("InterpType=")->GetBodyINT();
    int vanishSizeType = pCParticleClone->GetPCItem("VanishSizeType=")->GetBodyINT();
    int extrapolateType = pCParticleClone->GetPCItem("ExtrapolateType=")->GetBodyINT();
    int sortType = pCParticleClone->GetPCItem("SortType=")->GetBodyINT();
    float sortFreq = pCParticleClone->GetPCItem("SortFrequency=")->GetBodyDOUBLE();
    float alphaCrit = pCParticleClone->GetPCItem("AlphaCrit=")->GetBodyDOUBLE();
    
    int sizeType       [_MAX_PARTICLE_TYPE] = {};
    float sizeFactor   [_MAX_PARTICLE_TYPE] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1};
    int sizeAlphaType  [_MAX_PARTICLE_TYPE] = {};
    float sizeAlphaBase[_MAX_PARTICLE_TYPE] = {};
    float sizeAlphaMax [_MAX_PARTICLE_TYPE] = {};

    bool bClrCurveToUse[_MAX_COLOR_CURVE] = {FALSE};
    
    bool bTimeDepend = FALSE;
    if (classifyType == 1) bTimeDepend = TRUE;

    strStatusLog << GetStrProcCounterClone(pCParticleClone, NT);

    bool bUseSizeMod = FALSE;
    for (uint i = 0; i < numAttribute; i++){
	AttributeClone_cls *pCAttribute = pCParticleClone->mVpCAttribClone[i];
	if (pCAttribute->IsAttributeSameAs("ModifyApparentSize_cls"))
	    bUseSizeMod = TRUE;
	if (pCAttribute->IsAttributeSameAs("ModifyParticleSize_cls")){
	    bUseSizeMod = TRUE;
	    int clrCurveIndex = pCAttribute->GetPCItem("ColorCurveIndex=")->GetBodyINT();
	    bClrCurveToUse[clrCurveIndex] = TRUE;
	}
    }
    
    for (int i = 0; i < numMat; i++){
	AttributeClone_cls *pCMat = pCParticleCloneForMat->GetPCAttribCloneFromAttribName("MaterialParticle_cls", i);
	wxString strMat = pCMat->mCstrName.GetBodySTR();

	ConfItemList_cls &CConfListMat = pCMat->mCConfList;

	int clrCurveIndex = CConfListMat.GetPCItem("ColorCurveIndex=")->GetBodyINT();
	if (clrCurveIndex >= numClrCurve) clrCurveIndex = numClrCurve - 1;
	bClrCurveToUse[clrCurveIndex] = TRUE;
	
	clrCurveType[i] = CConfListMat.GetPCItem("ColorCurveType=")->GetBodyINT();
	clrCurveDepend[i] = CConfListMat.GetPCItem("ColorCurveDepend=")->GetBodyINT();
	int blendType = CConfListMat.GetPCItem("BlendType=")->GetBodyINT();
	int baseType = CConfListMat.GetPCItem("BaseType=")->GetBodyINT();
	int shaderTypePoint = CConfListMat.GetPCItem("ShaderTypePoint=")->GetBodyINT();
	int shaderTypeBillboard = CConfListMat.GetPCItem("ShaderTypeBillboard=")->GetBodyINT();
	int shaderTypeSphere = CConfListMat.GetPCItem("ShaderTypeSphere=")->GetBodyINT();
	bool bPseudoSphere = 
	    (baseType == _PARTICLE_POINT && shaderTypePoint == _POINTSHADER_PSEUDO_SPHERE) ||
	    (baseType == _PARTICLE_BILLBOARD && shaderTypeBillboard == _BILLBOARDSHADER_PSEUDO_SPHERE);
	bool bBoxPoint = (shaderTypePoint == _POINTSHADER_POINT);
	bool bHide = IsHide(baseType, shaderTypePoint, shaderTypeSphere, shaderTypeBillboard);
	
	bool bTextureUse = CConfListMat.GetPCItem("TextureType=")->GetBodyINT();
	int textureIndex = CConfListMat.GetPCItem("TextureIndex=")->GetBodyINT();

	paramBit[i] = (unsigned char)(bBoxPoint * 128 + textureIndex * 32 + bTextureUse * 16 + bHide * _PARAMBIT_HIDE + bPseudoSphere * _PARAMBIT_PSEUDO_SPHERE + blendType);
	unsigned char dummy = paramBit[i];
	
	sizeType[i] = CConfListMat.GetPCItem("SizeType=")->GetBodyINT();
	sizeAlphaType[i] = CConfListMat.GetPCItem("AlphaType=")->GetBodyINT();
	AttributeClone_cls *pCClrCurve = pCParticleCloneForMat->GetPCAttribCloneFromAttribName("ColorCurve_cls", clrCurveIndex);
	wxString strClr = pCClrCurve->mCstrName.GetBodySTR();

	pCClr[i] = pCParticleCloneForMat->GetPCParamGroup(strClr, wxES);
	ParamGroup_cls* pCRange = pCParticleCloneForMat->GetPCParamGroup(strClr, "Range");
	VRange[i] = pCRange->GetVInterpolatedValue(NT);
	bTimeDepend |= pCRange->IsTimeDepend();

	std::vector < double > VEmit = pCParticleCloneForMat->GetPCParamGroup(strMat, "Emission")->GetVInterpolatedValue(NT);
	for (int rgb = 0; rgb < 3; rgb++) clrEmit[i][rgb] = VEmit[rgb];
	
	ParamGroup_cls *pCAlphaModifier = pCParticleCloneForMat->GetPCParamGroup(strMat, "Alpha Modifier");
	sizeAlphaBase[i] = pCAlphaModifier->GetVInterpolatedValue(NT)[0];
	sizeAlphaMax[i] = pCAlphaModifier->GetVInterpolatedValue(NT)[1];
	
	ParamGroup_cls* pCSizeFactor = pCParticleCloneForMat->GetPCParamGroup(strMat, "SizeFactor");

	if (baseType == _PARTICLE_POINT){
	    sizeFactor[i] *= pCParticleCloneForMat->GetPCItem("PointSizeFactor=")->GetBodyDOUBLE() * pCSizeFactor->GetVInterpolatedValue(NT)[0];
	}
	if (baseType == _PARTICLE_BILLBOARD || baseType == _PARTICLE_SPHERE){
	    sizeFactor[i] = pCSizeFactor->GetVInterpolatedValue(NT)[1];
	    sizeFactor[i] = pCSizeFactor->GetVInterpolatedValue(NT)[1];
	}
	
	bTimeDepend |= pCMat->IsParamGroupTimeDepend();

	//strStatusLog << pCClr[i]->GetStrProcCounter() << "::" << pCRange->GetStrProcCounter() << "::";
	strStatusLog << pCMat->GetStrProcCounter();
    }

    // pJ[J[ȕ
    for (int i = 0; i < numClrCurve; i++){
	if (bClrCurveToUse[i] == TRUE){
	    AttributeClone_cls *pCClrCurveRef = pCParticleClone->GetPCAttribCloneFromAttribName("ColorCurve_cls", i);
	    wxString strClr = pCClrCurveRef->mCstrName.GetBodySTR();
	    ParamGroup_cls *pCClrRef = pCParticleClone->GetPCParamGroup(strClr, wxES);
	    ParamGroup_cls *pCRangeRef = pCParticleClone->GetPCParamGroup(strClr, "Range");
	    strStatusLog << pCClrRef->GetStrProcCounter() << "::" << pCRangeRef->GetStrProcCounter() << "::";
	}
    }

    if (bTimeDepend == TRUE) //!!
	strStatusLog << "::" << NT; 

    bool bUpdateIndexAndZ = FALSE;
    
    if (CheckUpdate(frame, factor, strStatusLog) == TRUE){
	
	PrepareColorCurveTexture(pCParticleCloneForMat, NT);

	bUpdateIndexAndZ = TRUE;

	bool bUseSizeAlpha = TRUE;
	bool bUseFilterAlpha = TRUE; // No Render łĝŏTRUEKv
	bool bUseID = TRUE; // TRUE 
	bool bUseValue[_MAX_PVALUE];
	for (int i = 0; i < _MAX_PVALUE; i++)
	    bUseValue[i] = (pCSnap1->mpValue[i] != NULL) & bStoreValue;
	
	int numParticleSRC = pCSnap1->mPNum;
	int downsizeType = 0;
	int downsize = GetDownsize(pCParticleClone, downsizeType);
	int numDownsized = GetNumDownsized(pCSnap1, downsize, downsizeType);
	
	Allocate(numDownsized, bUseFilterAlpha, bUseFilterAlpha, bUseSizeMod, bUseID, bUseIDLongLong, bUseValue);
	
	if (numDownsized == 0){
	    if (pCParticleCloneForMat != pCParticleClone)
		delete pCParticleCloneForMat;
	    delete pCParticleClone; 
	    mDrawCounter++;
	    return; 
	}
	
	bool bInterpByID = (pCParticleClone->GetPCItem("InterpolateBy=")->GetBodyINT() == 1);
	pCSnap1->CheckAllocateNextIndex(bInterpByID);
	if (pCSnap2 != NULL)
	    pCSnap2->CheckAllocateNextIndex(bInterpByID);
	if (pCSnap3 != NULL){
	    pCSnap3->CheckAllocateNextIndex(bInterpByID);

	    //pCSnap3gꍇ́A
	    //pCSnap2\ Next ID vZς݂łKv

	    if (bInterpByID == TRUE && pCSnap2->mbNextIndexCalced == FALSE){
		if (pCSnap2->mbNextIndexCalcStarted == TRUE){
		    CSW.Start();
		    while (pCSnap2->mbNextIndexCalced == FALSE){
			wxMilliSleep(3);
			if (CSW.Time() > 5000) {
			    wxLogMessage("Wait ID Calc Timeout B");
			    break;
			}
		    }
		}
	    }

	    if (bInterpByID == TRUE && pCSnap2->mbNextIndexCalced == FALSE){

		//pCSnap2->mbNextIndexCalcStarted = TRUE;
		
		int threadNumber;
		bool *bAFinished = GetNewPCbFinishList(threadNumber, pCSnap2->mPNum, FALSE);
		
		ThreadInterpStep_cls **ppCSearchIndex = 
		    new ThreadInterpStep_cls* [threadNumber];
		
		int tgtIni = 0;
		
		for (int core = 0; core < threadNumber; core++){
		    ppCSearchIndex[core] = 
			new ThreadInterpStep_cls;
		    
		    int PIni, PFin;
		    GetIniAndFin(PIni, PFin, core, threadNumber, pCSnap2->mPNum);
		    ppCSearchIndex[core]->SetParameters
			(pCSnap2, pCSnap3, NULL, NULL,
			 classifyType, strClassify, &mCLuaClassify[core],
			 interpType, vanishSizeType, extrapolateType,
			 factor, bInterpByID,
			 numMat, alphaCrit, sizeFactor, sizeType, 
			 sizeAlphaType, sizeAlphaBase, sizeAlphaMax, 
			 clrCurveType, clrCurveDepend,
			 clrEmit, paramBit, 
			 pCClr, VRange,
			 mpX, mpType, mpFilterAlpha, mpSize, mpSizeAlpha,
			 mpClrCurveS, mpClrCurveT, 
			 mpValue[0], mpValue[1], mpValue[2],
			 mpParamBit, mpID, mpIDLongLong,
			 mpIndexAndZ);
		    ppCSearchIndex[core]->mbSearchNextIndexOnly = TRUE;
		    ppCSearchIndex[core]->SetParametersTime
			(NT, frame, subFrame, seq);
		    ppCSearchIndex[core]->SetParametersDownsize
			(tgtIni, downsize, downsizeType);
		    ppCSearchIndex[core]->CreateThread
			(&bAFinished[core], PIni, PFin);
		    ppCSearchIndex[core]->Run();
		    
		    tgtIni += GetDownsizeIncrement
			(pCSnap1, downsize, downsizeType, PFin, PIni);
		}
		
		WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 3);

		delete [] ppCSearchIndex;

		if (pCSnap2->mbNextIndexCalcedThread == TRUE){
		    pCSnap2->mbNextIndexCalced = TRUE;
		    if (pCSnap3 != NULL)
			pCSnap3->mbPrevIndexCalced = TRUE;
		}
	    }
	}

	if (pCSnap1->mbNextIndexCalcStarted == TRUE){
	    wxStopWatch CSW;
	    CSW.Start();
	    while (pCSnap1->mbNextIndexCalced == FALSE){
		wxMilliSleep(3);
		if (CSW.Time() > 5000) {
		    wxLogMessage("Wait ID Calc Timeout B\n");
		    wxString strLog;
		    strLog << pCSnap1->mstrLoadedFileBody << ENDL;
		    wxLogMessage("Wait ID Calc Timeout B\n");
		    break;
		}
	    }
	}

	int threadNumber;
	bool *bAFinished = GetNewPCbFinishList(threadNumber, numParticleSRC, bNext);
	
	ThreadInterpStep_cls **ppCThreadInterpStep = 
	    new ThreadInterpStep_cls* [threadNumber];

	int tgtIni = 0;

	for (int core = 0; core < threadNumber; core++){
	    ppCThreadInterpStep[core] = 
		new ThreadInterpStep_cls;
	    
	    int PIni, PFin;
	    GetIniAndFin(PIni, PFin, core, threadNumber, numParticleSRC);
	    
	    ppCThreadInterpStep[core]->SetParameters
		(pCSnap1, pCSnap2, pCSnap0, pCSnap3,
		 classifyType, strClassify, &mCLuaClassify[core],
		 interpType, vanishSizeType, extrapolateType,
		 factor, bInterpByID,
		 numMat, alphaCrit, sizeFactor, sizeType, 
		 sizeAlphaType, sizeAlphaBase, sizeAlphaMax, 
		 clrCurveType, clrCurveDepend,
		 clrEmit, paramBit, 
		 pCClr, VRange,
		 mpX, mpType, mpFilterAlpha, mpSize, mpSizeAlpha,
		 mpClrCurveS, mpClrCurveT, 
		 mpValue[0], mpValue[1], mpValue[2],
		 mpParamBit, mpID, mpIDLongLong,
		 mpIndexAndZ);
	    ppCThreadInterpStep[core]->SetParametersTime
		(NT, frame, subFrame, seq);
	    ppCThreadInterpStep[core]->SetParametersDownsize
		(tgtIni, downsize, downsizeType);
	    ppCThreadInterpStep[core]->CreateThread
		(&bAFinished[core], PIni, PFin);
	    ppCThreadInterpStep[core]->Run();

	    tgtIni += GetDownsizeIncrement
		(pCSnap1, downsize, downsizeType, PFin, PIni);
	    
	}
	
	WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 3);

	delete [] ppCThreadInterpStep;

	if (pCSnap1->mbNextIndexCalcedThread == TRUE){
	    pCSnap1->mbNextIndexCalced = TRUE;
	    if (pCSnap2 != NULL)
		pCSnap2->mbPrevIndexCalced = TRUE;
	}
	
	
	pCSnap1->MarkDraw();
	mDrawCounter++;
	mDrawCounterIndex++;

	strLog << "PrepareData " << strTime << " NT=" << NT << "frame=" << mFrame << "::" << mFactor << "::" << CSW.Time() << " ms" << ENDL;
	
	ProcDataModifier(pCParticleClone);

	strLog << "ModifyData::" << CSW.Time() << ENDL;

	// Merge, Exclude
	uint numAttribClone = pCParticleClone->mVpCAttribClone.size();
	for (uint i = 0; i < numAttribClone; i++){
	    AttributeClone_cls* pCAttribClone = pCParticleClone->mVpCAttribClone[i];
	    if (((pCAttribClone->IsAttributeSameAs("ParticleMerge_cls")) ||
		 (pCAttribClone->IsAttributeSameAs("ParticleExclude_cls"))) &&
		pCAttribClone->GetPCItem("Validity=")->GetBodyBOOL() == TRUE){
		
		wxString strTgt = pCAttribClone->GetPCItem("Target=")->GetBodySTR();
		SceneObj_cls *pCTgt = GetPCSceneObjCtrl()->GetPCFromName(strTgt);
		
		if (pCTgt != NULL && pCTgt->IsSameAs("Particle_cls") &&
		    pCTgt->mCbVisible.GetBodyBOOL() == TRUE){
		    
		    // Merge, Exclude ĂƂNext̏͂Ȃ
		    ASSERT(bNext == FALSE);
		    if (pCAttribClone->IsAttributeSameAs("ParticleMerge_cls")){
			ParticleMergeExe_cls CExe = ParticleMergeExe_cls(pCAttribClone, this, ((Particle_cls*)pCTgt)->mpCCurrent);
			CExe.Execute(); 
		    }
		    if (pCAttribClone->IsAttributeSameAs("ParticleExclude_cls")){
			ParticleExcludeExe_cls CExe = ParticleExcludeExe_cls(pCAttribClone, this, ((Particle_cls*)pCTgt)->mpCCurrent);
			CExe.Execute(); 
		    }
		}
	    }
	}
    }
    else{
	//strLog << "Skip Data Prepare " << strTime << " NT=" << NT << ENDL;
    }

    


    bUpdateIndexAndZ |= CheckUpdateSort(sortType, CSortPos, sortFreq, strLog);
    
    if (bUpdateIndexAndZ == TRUE){
	ProcIndexAndZ(NT, pCParticleClone, CSortPos, sortType, numMat, clrCurveType, clrCurveDepend, VRange, factor, bNext);
	
	if (sortType != _SORT_NONE && sortType != _SORT_DISTANCE_ONLY){
	if (mPNum > 640000){
	    SortParallel_cls CSortParallel;
	    CSortParallel.SortParallel(mpIndexAndZ, mPNum);
	}
	else
	    std::sort(mpIndexAndZ, mpIndexAndZ+mPNum);
    }
    mDrawCounterIndex++;

    }

    if (pCParticleCloneForMat != pCParticleClone)
	delete pCParticleCloneForMat;
    delete pCParticleClone;

    //wxLogMessage(strLog);
}

wxString CLS::
GetStrProcCounterClone(ViewerClone_cls *pCParticleClone, double NT){
    wxString strCounter;
    
    uint numAttribClone = pCParticleClone->mVpCAttribClone.size();

    for (uint i = 0; i < numAttribClone; i++){
	AttributeClone_cls *pCMod = pCParticleClone->mVpCAttribClone[i];
	if (pCMod->IsAttributeSameAs("ParticleBoolean_cls"))
	    strCounter << BooleanBase_cls::GetStrProcCounterClone(pCParticleClone, pCMod, NT);
	if (pCMod->IsAttributeSameAs("ModifyApparentSize_cls"))
	    strCounter << ModifyApparentSize_cls::GetStrProcCounterClone(pCMod, NT);
	if (pCMod->IsAttributeValidAndSameAs("ModifyParticleSize_cls"))
	    strCounter << pCMod->mCConfList.GetStrProcCounter();

	if (pCMod->IsAttributeValidAndSameAs("ParticleFiltering_cls"))
	    strCounter << pCMod->mCConfList.GetStrProcCounter();

	if (pCMod->IsAttributeValidAndSameAs("ParticleDownsizing_cls")) 
	    strCounter << pCMod->mCConfList.GetStrProcCounter();

	if (pCMod->IsAttributeValidAndSameAs("ParticleMerge_cls"))
	    strCounter << pCMod->mCConfList.GetStrProcCounter();
    }
    return strCounter;
}

void CLS::
ProcDataModifier(ViewerClone_cls *pCParticle){
    //bool bFiltering = pCVolumeBox->IsDataProcClsExist("VolumeFiltering_cls");
    //if (bFiltering == TRUE)
    //ProcDataFiltering(pCVolumeBox);
}

double CLS::
GetRemapX(std::vector < double > &VRange, double x0/*, double NT*/){
    double width = VRange[1] - VRange[0];
    if (width == 0) return 1;
    return (x0 - VRange[0])/width;
}

void CLS::
ProcIndexAndZ(double NT, 
	      ViewerClone_cls *pCParticle, PositionAndDirection_cls CSortPos,
	      int sortType, int numMat, int *clrCurveType, int *clrCurveDepend, 
	      std::vector < double > *VRange,
	      //ParticleSnap_cls *pCSnap1, ParticleSnap_cls*pCSnap2,
	      double factor, bool bNext){
    
    wxStopWatch CSW;
    CSW.Start();
    wxString strLog;
    
    // FilterAlpha, SizeMod ̏
    {
	int threadNumber;
	bool *bAFinished = GetNewPCbFinishList(threadNumber, mPNum, bNext);
	ThreadFillInitDataParticle_cls **ppCThread = 
	    new ThreadFillInitDataParticle_cls* [threadNumber];
	
	for (int core = 0; core < threadNumber; core++){
	    ppCThread[core] = 
		new ThreadFillInitDataParticle_cls;
	    
	    int PIni, PFin;
	    GetIniAndFin(PIni, PFin, core, threadNumber, mPNum);
	    
	    ppCThread[core]->SetParameters(mpFilterAlpha, mpIndexAndZ, mpSizeMod, TRUE);
	    ppCThread[core]->CreateThread(&bAFinished[core], PIni, PFin);
	    if (threadNumber == 1) {
		ppCThread[core]->DummyEntry();
		delete ppCThread[core];
	    }
	    else
		ppCThread[core]->Run();
	}
	
	WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 1);
	delete [] ppCThread;
    }

    // ܂boolean
    /*
    if (mpSizeMod != NULL)
	for (int i = 0; i < mPNum; i++) mpSizeMod[i] = 1;
    for (int i = 0; i < mPNum; i++) mpFilterAlpha[i] = 1;
    */

    uint numAttribClone = pCParticle->mVpCAttribClone.size();
    for (int i = 0; i < numAttribClone; i++){
	AttributeClone_cls* pCAttribClone = pCParticle->mVpCAttribClone[i];
	if (pCAttribClone->IsAttributeSameAs("ParticleFiltering_cls")){
	    ParticleFilteringExec_cls *pCExe = new ParticleFilteringExec_cls(pCAttribClone);
	    pCExe->CheckFiltering(mpFilterAlpha, mpIndexAndZ, this);
	    delete pCExe;
	}
	if (pCAttribClone->IsAttributeSameAs("ParticleBoolean_cls")){

	    double matrix16TgtInv[16];
	    BooleanBase_cls::GetMatrix16TgtInv(pCAttribClone, matrix16TgtInv, NT);
	    wxString strTgt  = pCAttribClone->GetPCItem("BooleanObject=")->GetBodySTR();
	    SceneObj_cls *pCObj = GetPCSceneObjCtrl()->GetPCFromName(strTgt);

	    if (pCObj != NULL && pCObj->mstrClassName.IsSameAs("MeshObj_cls")){
		MeshObj_cls *pCMesh = (MeshObj_cls*)pCObj;
		pCMesh->CheckLoad();
		pCMesh->CheckTree();
	    }
	    
	    ParticleBooleanExec_cls *pCExe = new ParticleBooleanExec_cls(pCAttribClone, NT, pCParticle->mMatrix16, matrix16TgtInv, pCObj);
	    pCExe->CheckBoolean(mpFilterAlpha, mpX, mPNum, bNext);
	    delete pCExe;
	}
    }

    // \[gpZbgƓ
    // pFilterAlph  pSizeAlpha, bitParam ̌vZSetArrayZ_clsłĂ܂

    /*
    if (sortType == _SORT_NONE){
	for (int i = 0; i < mPNum; i++){
	    mpIndexAndZ[i].mIndex = i;
	    mpIndexAndZ[i].mZ = 0;
	}
	
	mDrawCounterIndex++;
    }
    */

    int threadNumber;
    bool *bAFinished = GetNewPCbFinishList(threadNumber, mPNum, bNext);

    ThreadSetArrayZ_cls **ppCThreadSetArray = 
	new ThreadSetArrayZ_cls* [threadNumber];

    for (int core = 0; core < threadNumber; core++){
	ppCThreadSetArray[core] = new ThreadSetArrayZ_cls;

	int PIni, PFin;
	GetIniAndFin(PIni, PFin, core, threadNumber, mPNum);
	
	ppCThreadSetArray[core]->SetParameters(mpX, mpType, mpFilterAlpha, mpSizeAlpha, mpParamBit, mpIndexAndZ, CSortPos, sortType);
	ppCThreadSetArray[core]->CreateThread(&bAFinished[core], PIni, PFin);
	ppCThreadSetArray[core]->Run();
    }
    
    WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 1);
    delete [] ppCThreadSetArray;

    for (int i = 0; i < numAttribClone; i++){
	AttributeClone_cls* pCAttribClone = pCParticle->mVpCAttribClone[i];
	if (pCAttribClone->IsAttributeSameAs("ModifyApparentSize_cls")){
	    ASSERT(mpSizeMod != NULL);
	    ModifyApparentSizeExec_cls *pCExe = new ModifyApparentSizeExec_cls(pCAttribClone, NT);
	    pCExe->Execute(mpFilterAlpha, mpSize, mpSizeMod, mpType, mpIndexAndZ, sortType, mPNum, bNext);
	    mDrawCounter++;
	    delete pCExe;
	}

	if (pCAttribClone->IsAttributeSameAs("ModifyParticleSize_cls")){
	    ASSERT(mpSizeMod != NULL);
	    ModifyParticleSizeExec_cls *pCExe = new ModifyParticleSizeExec_cls(pCAttribClone);
	    pCExe->Execute(pCParticle, this, factor, NT, bNext);
	    mDrawCounter++;
	    delete pCExe;
	}
    }

    if (sortType != _SORT_NONE){
	bool bRemakeCurve = FALSE;
	for (int i = 0; i < 10; i++){
	    if (clrCurveType[i] == 1 &&
		clrCurveDepend[i] == _COLOR_CURVE_DEPEND_TYPE_SORT)
		bRemakeCurve = TRUE;
	}
	
	if (bRemakeCurve == TRUE){
	    int threadNumber;
	    bool *bAFinished = GetNewPCbFinishList(threadNumber, mPNum, bNext);
	    
	    ThreadCurveDependOnSort_cls **ppCThread = 
		new ThreadCurveDependOnSort_cls* [threadNumber];
	    
	    for (int core = 0; core < threadNumber; core++){
		ppCThread[core] = new ThreadCurveDependOnSort_cls;
		
		int PIni, PFin;
		GetIniAndFin(PIni, PFin, core, threadNumber, mPNum);
		
		ppCThread[core]->SetParameters(mpClrCurveS, mpClrCurveT, clrCurveType, clrCurveDepend, mpType, mpIndexAndZ, VRange, numMat, NT);
		ppCThread[core]->CreateThread(&bAFinished[core], PIni, PFin);
		ppCThread[core]->Run();
	    }
	    
	    WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 1);
	    delete [] ppCThread;
	}
    }

    strLog << "ProcIndexAndZ Internal Time=" << CSW.Time() << " ms" << ENDL;

    //strLog << " :: " << CSW.Time() << " ms" << ENDL;
    //wxLogMessage(strLog);
}

void CLS::
ExportSnap(wxString strFile, int direction){
    FILE *fp = fopen(strFile.char_str(), "wb");

    if (direction == 0){
	for (int i = 0; i < mPNum; i++){
	    if (mpX != NULL) fwrite(mpX[i], sizeof(float), 3, fp);
	    if (mpID != NULL) fwrite(&mpID[i], sizeof(float), 1, fp);
	    if (mpIDLongLong != NULL) 
		fwrite(&mpIDLongLong[i], sizeof(long long), 1, fp);
	    if (mpType != NULL){
		int tmp = (int)mpType[i];
		fwrite(&tmp, sizeof(int), 1, fp);
	    }
	    if (mpSize != NULL)
		fwrite(&mpSize[i], sizeof(float), 1, fp);
	    for (int j = 0; j < _MAX_PVALUE; j++)
		if (mpValue[j] != NULL)
		    fwrite(&mpValue[j][i], sizeof(float), mPNum, fp);
	}
    }

    if (direction == 1){
	if (mpX != NULL) fwrite(mpX, sizeof(float), mPNum*3, fp);
	if (mpID != NULL) fwrite(mpID, sizeof(int), mPNum, fp);
	if (mpIDLongLong != NULL)
	    fwrite(mpIDLongLong, sizeof(long long), mPNum, fp);
	if (mpType != NULL){
	    int *pTmp = new int [mPNum];
	    for (int i = 0; i < mPNum; i++) pTmp[i] = (int)mpType[i];
	    fwrite(pTmp, sizeof(int), mPNum, fp);
	    delete [] pTmp;
	}
	if (mpSize != NULL) fwrite(mpSize, sizeof(float), mPNum, fp);
	for (int i = 0; i < _MAX_PVALUE; i++)
	    if (mpValue[i] != NULL)
		fwrite(mpValue[i], sizeof(float), mPNum, fp);
    }
    
    fclose(fp);
}

wxString CLS::
GetStrExportFormat(int direction){
    wxString strFormat;

    if (direction == 0){
	if (mpX != NULL) strFormat << "x y z ";
	if (mpID != NULL) strFormat << "ID ";
	if (mpIDLongLong != NULL) strFormat << "ID(long long) ";
	if (mpType != NULL) strFormat << "type(int) ";
	if (mpSize != NULL) strFormat << "size ";
	for (int i = 0; i < _MAX_PVALUE; i++)
	    if (mpValue[i] != NULL)
		strFormat << "value" << i << " ";
	strFormat << ENDL;
	strFormat << strFormat;
	strFormat << "..." << ENDL;
    }
    if (direction == 1){
	if (mpX != NULL)
	    strFormat << "xyz : float[" << mPNum << "][3]" << ENDL;
	if (mpID != NULL)
	    strFormat << "ID : int[" << mPNum << "]" << ENDL;
	if (mpIDLongLong != NULL)
	    strFormat << "ID(long long) : long long[" << mPNum << "]" << ENDL;
	if (mpType != NULL)
	    strFormat << "type : int[" << mPNum << "]" << ENDL;
	if (mpSize != NULL)
	    strFormat << "size : float[" << mPNum << "]" << ENDL;
	for (int i = 0; i < _MAX_PVALUE; i++)
	    if (mpValue[i] != NULL)
		strFormat << "value" << i << " : float[" << mPNum << "]" << ENDL;
    }

    return strFormat;
}
