// (setq-c-offset 4)
// "clsParticleForRender.cpp"

#ifndef _PARTICLE_FOR_RENDER_H
#define _PARTICLE_FOR_RENDER_H

#include "clsArrayAndIndex.h"
#include "../LUA/clsLuaBase.h"

#include "../COMMON/clsThreadCommon.h"
#include "../COMMON/clsThreadCommonTypeB.h"
#include "../COMMON/clsGetPC.h"
#include "../COMMON/CommonDefinitionsForH.h"

#include "../MYGL/clsGLTexture.h"
#include "../MYGL/clsPosition.h"

#define _PARAMBIT_PSEUDO_SPHERE 4
#define _PARAMBIT_HIDE 8

class ParticleSnap_cls;
class ViewerClone_cls;

class ThreadFillInitDataParticle_cls : public ThreadInstance_cls {
 public:
    float *mpFilterAlpha, *mpSizeMod;
    ArrayAndIndex_cls *mpIndexAndZ;
    bool   mbInitializeAtProcIndexAndZ;
        
    void SetParameters(float *pFilterAlpha, ArrayAndIndex_cls *pIndexAndZ, float *pSizeMod, bool bInitializeAtProcIndexAndZ = FALSE);
    
    virtual ExitCode Entry();
    void DummyEntry();
};

class ThreadCurveDependOnSort_cls : public ThreadInstance_cls {
 public:
    float *mpClrCurveS, *mpClrCurveT;
    int *mpClrCurveType, *mpClrCurveDepend;
    unsigned char *mpType;
    ArrayAndIndex_cls *mpIndexAndZ;
    std::vector < double > *mpVRange;
    int mNumMat;
    double mNT;
    
    void SetParameters(float *pClrCurveS, float *pClrCurveT, 
		       int *clrCurveType, int *clrCurveDepend,
		       unsigned char *pType, ArrayAndIndex_cls *pIndexAndZ,
		       std::vector < double > *VRange,
		       int numMat, double NT);
    double GetRemapX(std::vector < double > &VRange, double x0/* double NT*/);
    virtual ExitCode Entry();
};

class ParticleForRender_cls : public ThreadCommon_cls, public GetPC_cls {
public:
    ParticleForRender_cls();
    ~ParticleForRender_cls();
    
    long long mPNum, mPNumAllocated;
    float (*mpX)[3];
    float *mpClrCurveS, *mpClrCurveT;
    float *mpFilterAlpha;
    float *mpSizeAlpha;
    unsigned char *mpParamBit;
    ArrayAndIndex_cls *mpIndexAndZ;

    float *mpSize, *mpSizeMod;
    unsigned char *mpType;
    int *mpID;
    long long *mpIDLongLong; // TwistID͓ق
    float *mpValue[_MAX_PVALUE];
    
    int mFrame;
    float mFactor;
    int mSortTypeUsed;
    PositionAndDirection_cls mCSortPos;

    float *mpTextureDataFloat;
    unsigned short *mpTextureData;
    int mTextureDataProcCounter;
    int mTextureDataProcCounterStored;
    int mTextureDataSize;
    
    wxString mstrParticleTexture[3];
    float mDistanceBase;

    wxString mstrStatusLog;

    GLTexture_cls mCColorCurveTexture;
    double mRangeMin0, mRangeMax0;

    bool mbCastShadow;
    
    LuaBase_cls mCLuaClassify[_MAX_THREAD_LUA_AND_INTERPOLATE];
    
    int mDrawCounter;
    int mDrawCounterIndex;

    void DeleteAndSetNullFloatArray(float **p);
    void Clear();
    wxString GetStrInfo(int infoIndex);
    long long SearchID(long long id);
    
    bool CheckUpdate(int frame, float factor, wxString strStatusLog);
    bool CheckUpdateSort(int sortType, PositionAndDirection_cls CSortPos,
			 float sortFreq, wxString &strLog);
    void PrepareCommonRenderParameter(ViewerClone_cls *pCParticleClone);
    void PrepareColorCurveTexture(ViewerClone_cls *pCParticleClone,
				  double NT);
    void SetColorCurveTexture();
    void AllocateFull(long long numParticle, bool bUseSizeAlpha, bool bUseFilterAlpha, bool bUseSizeMod, bool bUseTwistID, bool bUseIDLongLong, bool bUseValue[_MAX_PVALUE]);
    void ReallocationFloat(bool bUse, float **p);
    void Reallocation(long long numParticle, bool bUseSizeAlpha, bool bUseFilterAlpha, bool bUseSizeMod, bool bUseTwistID, bool bUseIDLongLong, bool bUseValue[_MAX_PVALUE]);
    void Allocate(long long numParticle, bool bUseSizeAlpha, bool bUseFilterAlpha, bool bUseSizeMod, bool bUseTwistID, bool bUseIDL, bool bUseValue[_MAX_PVALUE]);
    int  GetDownsize(ViewerClone_cls *pCParticleClone, int &downsizeTypeype);
    int  GetNumDownsized(ParticleSnap_cls *pCSnap1,
			 int downsize, int downsizeTypey);
    int  GetDownsizeIncrement(ParticleSnap_cls *pCSnap1,
			      int downsize, int downsizeType,
			      int PFin, int pIni);
    bool IsHide(int type, int shaderPoint, int shaderSphere, int shaderBillboard);
    void PrepareData(ParticleSnap_cls *pCSnap1, ParticleSnap_cls *pCSnap2,
		     ParticleSnap_cls *pCSnap0, ParticleSnap_cls *pCSnap3,
		     int frame, int subFrame, int seq, double factor, 
		     ViewerClone_cls *pCParticleClone,
		     ViewerClone_cls *pCParticleCloneForMat,
		     double NT, PositionAndDirection_cls CSortPosOrg,
		     bool bUseIDLongLong, bool bStoreValue,
		     bool bNext = FALSE);

    wxString GetStrProcCounterClone(ViewerClone_cls *pCParticleClone, double NT);

    void ProcDataModifier(ViewerClone_cls *pCParticle);
    double GetRemapX(std::vector < double > &VRange, double x0/*, double NT*/);
    void ProcIndexAndZ(double NT, ViewerClone_cls *pCParticle,
		       PositionAndDirection_cls CSortPos,
		       int sortType, 
		       int numMat, int *clrCurveType, int *clrCurveDepend, 
		       std::vector < double > *VRange,
		       double factor, bool bNext);

    void ExportSnap(wxString strFile, int direction);
    wxString GetStrExportFormat(int direction);
};

#endif
