// -*- c -*- 
// "clsParticleFormat.h"

#include "clsParticleFormat.h"

#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/DATAVIEWER/clsParticle.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMultiInputPanel.h"
#include "../WIDGETCTRL/MULTIINPUT/clsMNumPainBase.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"
#include "../UNDO/clsUndoManager.h"

#include "../COMMON/CommonDefinitions.h"


#include "../../wxWidgetEnum.h"

#define CLS ParticleFormatClone_cls

CLS::
~ParticleFormatClone_cls(){
    mCConfList.DeleteAllItems();
}

bool CLS::
DoesExistInBody(wxString strItem, int dataType){
    if (dataType == _PFORMAT_BINARY){
	uint numBody = mCAstrBodyBinary.size();
	for (uint i = 0; i < numBody; i++)
	    if (mCAstrBodyBinary[i].IsSameAs(strItem)) return TRUE;
    }
    if (dataType == _PFORMAT_ASCII){
	uint numBody = mCAstrBodyAscii.size();
	for (uint i = 0; i < numBody; i++)
	    if (mCAstrBodyAscii[i].IsSameAs(strItem)) return TRUE;
    }
    return FALSE;
}

int CLS::
GetBodyParticleByte(){
    int byte = 0;
    uint numBody = mCAstrBodyBinary.size();
    for (uint i = 0; i < numBody; i++){
	wxString strItem = mCAstrBodyBinary[i];
	if (strItem.IsSameAs("XYZ(float)"))    byte+=12;
	if (strItem.IsSameAs("VXVYVZ(float)")) byte+=12;
	if (strItem.IsSameAs("Type(int)"))     byte+=4;
	if (strItem.IsSameAs("Size(float)"))   byte+=4;
	if (strItem.IsSameAs("ID(int)"))       byte+=4;
	if (strItem.IsSameAs("ID(long long)"))  byte+=8;
	if (strItem.IsSameAs("Value0(float)")) byte+=4;
	if (strItem.IsSameAs("Value1(float)")) byte+=4;
	if (strItem.IsSameAs("Value2(float)")) byte+=4;
	if (strItem.IsSameAs("Dummy(4byte)"))  byte+=4;
	if (strItem.IsSameAs("Dummyx3(12byte)")) byte+=12;
	if (strItem.IsSameAs("Dummyx10(40byte)")) byte+=40;
    }
    return byte;
}

int CLS::
GetBodyParticleLineAsciiVertical(){
    int line = 0;
    uint numBody = mCAstrBodyAscii.size();
    for (uint i = 0; i < numBody; i++){
	wxString strItem = mCAstrBodyAscii[i];
	if (strItem.IsSameAs("XYZ"))      line++;
	if (strItem.IsSameAs("VXVYVZ"))   line++;
	if (strItem.IsSameAs("Type"))     line++;
	if (strItem.IsSameAs("Size"))     line++;
	if (strItem.IsSameAs("ID"))       line++;
	if (strItem.IsSameAs("ID(long long)")) line++;
	if (strItem.IsSameAs("Value0"))   line++;
	if (strItem.IsSameAs("Value1"))   line++;
	if (strItem.IsSameAs("Value2"))   line++;
	if (strItem.IsSameAs("Dummy"))    line++;
	if (strItem.IsSameAs("Dummyx3"))  line+=3;
	if (strItem.IsSameAs("Dummyx10")) line+=10;
	if (strItem.IsSameAs("X"))        line++;
	if (strItem.IsSameAs("Y"))        line++;
	if (strItem.IsSameAs("Z"))        line++;
	if (strItem.IsSameAs("VX"))       line++;
	if (strItem.IsSameAs("VY"))       line++;
	if (strItem.IsSameAs("VZ"))       line++;
    }
    return line;
}

std::vector < int > CLS::
GetVFormatHeader(){
    std::vector < int > VFormat;
    uint numHeader = mCAstrHeader.size();
    for (uint i = 0; i < numHeader; i++){
	wxString strItem = mCAstrHeader[i];
	for (int j = 0; j < _PFORMAT_HEADER_MAX; j++)
	    if (strItem.IsSameAs(GetStrHeaderFromIndex(j)))
		VFormat.push_back(j);
    }
    return VFormat;
}

std::vector < int > CLS::
GetVFormatBody(int dataType){
    std::vector < int > VFormat;
    if (dataType == _PFORMAT_BINARY){
	uint numBody = mCAstrBodyBinary.size();
	for (uint i = 0; i < numBody; i++){
	    wxString strItem = mCAstrBodyBinary[i];
	    for (int j = 0; j < _PFORMAT_MAX; j++)
		if (strItem.IsSameAs(GetStrBodyFromIndex(j, _PFORMAT_BINARY)))
		    VFormat.push_back(j);
	    }  
    }
    if (dataType == _PFORMAT_ASCII){
	uint numBody = mCAstrBodyAscii.size();
	for (uint i = 0; i < numBody; i++){
	    wxString strItem = mCAstrBodyAscii[i];
	    for (int j = 0; j < _PFORMAT_ASCII_MAX; j++)
		if (strItem.IsSameAs(GetStrBodyFromIndex(j, _PFORMAT_ASCII)))
		    VFormat.push_back(j);
	} 
    }
    return VFormat;
}

int CLS::
GetHeaderByte(){
    int byte = 0;
    uint numHeader = mCAstrHeader.size();
    for (uint i = 0; i < numHeader; i++){
	if (mCAstrHeader[i].IsSameAs("ParticleNumber")) byte+=4;
	if (mCAstrHeader[i].IsSameAs("Time")) byte+=4;
	if (mCAstrHeader[i].IsSameAs("Dummy")) byte+=4;
	if (mCAstrHeader[i].IsSameAs("Dummyx3")) byte+=12;
	if (mCAstrHeader[i].IsSameAs("Dummyx10")) byte+=40;
    }
    return byte;
}

int CLS::
GetHeaderLine(){
    int line = 0;
    int numHeader = mCAstrHeader.size();
    for (uint i = 0; i < numHeader; i++){
	if (mCAstrHeader[i].IsSameAs("ParticleNumber")) line++;
	if (mCAstrHeader[i].IsSameAs("Time")) line++;
	if (mCAstrHeader[i].IsSameAs("Dummy")) line++;
	if (mCAstrHeader[i].IsSameAs("Dummyx3")) line+=3;
	if (mCAstrHeader[i].IsSameAs("Dummyx10")) line+=10;
    }
    return line;
}

void CLS::
GetInfoBinary(bool *pbUseVelocity, bool *pbUseType, bool *pbUseSize, 
	      bool *pbUseID, bool *pbUseIDLongLong, bool *pbInterpolateByID,
	      bool *pbUseValue0, bool *pbUseValue1, bool *pbUseValue2){
    int dataType = _PFORMAT_BINARY;
    if (DoesExistInBody("VXVYVZ(float)", dataType)) *pbUseVelocity = TRUE;
    else *pbUseVelocity = FALSE;
    if (DoesExistInBody("Type(int)", dataType)) *pbUseType = TRUE;
    else *pbUseType = FALSE;
    if (DoesExistInBody("Size(float)", dataType)) *pbUseSize = TRUE;
    else *pbUseSize = FALSE;
    if (DoesExistInBody("ID(int)", dataType)) *pbUseID = TRUE;
    else *pbUseID = FALSE;
    if (DoesExistInBody("ID(long long)", dataType)) *pbUseIDLongLong = TRUE;
    else *pbUseIDLongLong = FALSE;
    if (DoesExistInBody("Value0(float)", dataType)) *pbUseValue0 = TRUE;
    else *pbUseValue0 = FALSE;
    if (DoesExistInBody("Value1(float)", dataType)) *pbUseValue1 = TRUE;
    else *pbUseValue1 = FALSE;
    if (DoesExistInBody("Value2(float)", dataType)) *pbUseValue2 = TRUE;
    else *pbUseValue2 = FALSE;

    *pbInterpolateByID = mbInterpByID;
}

void CLS::
GetInfoAscii(bool *pbUseVelocity, bool *pbUseType, bool *pbUseSize, 
	     bool *pbUseID, bool *pbUseIDLongLong, bool *pbInterpolateByID, 
	     bool *pbUseValue0, bool *pbUseValue1, bool *pbUseValue2){
    int dataType = _PFORMAT_ASCII;
    if (DoesExistInBody("VXVYVZ", dataType) ||
	DoesExistInBody("VX", dataType) ||
	DoesExistInBody("VY", dataType) ||
	DoesExistInBody("VZ", dataType)) *pbUseVelocity = TRUE;
    else *pbUseVelocity = FALSE;
    if (DoesExistInBody("Type", dataType)) *pbUseType = TRUE;
    else *pbUseType = FALSE;
    if (DoesExistInBody("Size", dataType)) *pbUseSize = TRUE;
    else *pbUseSize = FALSE;
    if (DoesExistInBody("ID", dataType)) *pbUseID = TRUE;
    else *pbUseID = FALSE;
    if (DoesExistInBody("ID(long long)", dataType)) *pbUseIDLongLong = TRUE;
    else *pbUseIDLongLong = FALSE;
    if (DoesExistInBody("Value0", dataType)) *pbUseValue0 = TRUE;
    else *pbUseValue0 = FALSE;
    if (DoesExistInBody("Value1", dataType)) *pbUseValue1 = TRUE;
    else *pbUseValue1 = FALSE;
    if (DoesExistInBody("Value2", dataType)) *pbUseValue2 = TRUE;
    else *pbUseValue2 = FALSE;

    *pbInterpolateByID = mbInterpByID;
}

wxString CLS::
GetStrHeaderFromIndex(int index){
    wxString strItem;
    switch(index){
    case _PFORMAT_HEADER_NUMBER:   strItem = "ParticleNumber"; break;
    case _PFORMAT_HEADER_TIME:     strItem = "Time";           break;
    case _PFORMAT_HEADER_DUMMY:    strItem = "Dummy";          break;
    case _PFORMAT_HEADER_DUMMYx3:  strItem = "Dummyx3";        break;
    case _PFORMAT_HEADER_DUMMYx10: strItem = "Dummyx10";       break;
    }
    return strItem;
}

wxString CLS::
GetStrBodyFromIndex(int index, int dataType){
    wxString strItem;
    if (dataType == _PFORMAT_BINARY)
	switch(index){
	case _PFORMAT_XYZ_FLOAT:   strItem = "XYZ(float)";       break;
	case _PFORMAT_VXVYVZ:      strItem = "VXVYVZ(float)";    break;
	case _PFORMAT_TYPE_INT:    strItem = "Type(int)";        break;
	case _PFORMAT_SIZE_FLOAT:  strItem = "Size(float)";      break;
	case _PFORMAT_ID_INT:      strItem = "ID(int)";          break;
	case _PFORMAT_ID_LONGLONG: strItem = "ID(long long)";    break;
	case _PFORMAT_VALUE_0:     strItem = "Value0(float)";    break;
	case _PFORMAT_VALUE_1:     strItem = "Value1(float)";    break;
	case _PFORMAT_VALUE_2:     strItem = "Value2(float)";    break;
	case _PFORMAT_DUMMY:       strItem = "Dummy(4byte)";     break;
	case _PFORMAT_DUMMYx3:     strItem = "Dummyx3(12byte)";  break;
	case _PFORMAT_DUMMYx10:    strItem = "Dummyx10(40byte)"; break;
	}
    if (dataType == _PFORMAT_ASCII)
	switch(index){
	case _PFORMAT_ASCII_XYZ:      strItem = "XYZ";      break;
	case _PFORMAT_ASCII_VXVYVZ:   strItem = "VXVYVZ";   break;
	case _PFORMAT_ASCII_TYPE:     strItem = "Type";     break;
	case _PFORMAT_ASCII_SIZE:     strItem = "Size";     break;
	case _PFORMAT_ASCII_ID:       strItem = "ID";       break;
	case _PFORMAT_ASCII_ID_LONGLOG: strItem = "ID(long long)";    break;
	case _PFORMAT_ASCII_VALUE0:   strItem = "Value0";   break;
	case _PFORMAT_ASCII_VALUE1:   strItem = "Value1";   break;
	case _PFORMAT_ASCII_VALUE2:   strItem = "Value2";   break;
	case _PFORMAT_ASCII_DUMMY:    strItem = "Dummy";    break;
	case _PFORMAT_ASCII_DUMMYx3:  strItem = "Dummyx3";  break;
	case _PFORMAT_ASCII_DUMMYx10: strItem = "Dummyx10"; break;
	case _PFORMAT_ASCII_X:        strItem = "X"; break;
	case _PFORMAT_ASCII_Y:        strItem = "Y"; break;
	case _PFORMAT_ASCII_Z:        strItem = "Z"; break;
	}
    return strItem;
}

void CLS::
CloneFromFormat(ParticleFormat_cls &CFormat, bool bInterpByID){
    mCAstrHeader = CFormat.mCAstrHeader;
    mCAstrBodyBinary = CFormat.mCAstrBodyBinary;
    mCAstrBodyAscii = CFormat.mCAstrBodyAscii;

    mbInterpByID = bInterpByID;
    CloneConfListFrom(CFormat);
}

#define CLS ParticleFormat_cls

CLS::
ParticleFormat_cls(Particle_cls *pCParentParticle, wxString strObj){

    mpCParentParticle = pCParentParticle;
    
    wxString strUpdateFormat = "PROPERTY_FORMAT";

    SyncConfListTag(strObj);
    mCConfList.Add(&mCDataType, "DataType=", "1", strUpdateFormat);
    mCConfList.Add(&mCArrayDirection, "ArrayDirection=", "0");
    mCConfList.Add(&mCbPreSortByID, "PreSortByID=", "0", strUpdateFormat);
    mCConfList.Add(&mCbRemoveDupli, "RemoveDupli=", "0", strUpdateFormat);
    
    
    mCAstrHeader.push_back("ParticleNumber");
    mCAstrHeader.push_back("Time");
    
    mCAstrBodyBinary.push_back("XYZ(float)");
    mCAstrBodyBinary.push_back("Type(int)");
    mCAstrBodyBinary.push_back("Size(float)");
    
    mCAstrBodyAscii.push_back("XYZ");
    mCAstrBodyAscii.push_back("Type");
    mCAstrBodyAscii.push_back("Size");

    mCAstrHeaderFormat.SetLink(&mCAstrHeader);
    mCConfList.Add(&mCAstrHeaderFormat, "ParticleFormatHeader=", "Dummy", "PROPERTY_FORMAT");
    
    mCAstrBodyFormatBinary.SetLink(&mCAstrBodyBinary);
    mCConfList.Add(&mCAstrBodyFormatBinary, "ParticleFormatBodyBinary=", "Dummy", "PROPERTY_FORMAT");
    
    mCAstrBodyFormatAscii.SetLink(&mCAstrBodyAscii);
    mCConfList.Add(&mCAstrBodyFormatAscii, "ParticleFormatBodyAscii=", "Dummy", "PROPERTY_FORMAT");
}

void CLS::
SyncConfListTag(wxString strObj){
    wxString strConfListLabel = "SceneObj::";  strConfListLabel << strObj << "::" << "Format";
    mCConfList.SetTag(strConfListLabel);
}

bool CLS::
DoesExistInBody(wxString strItem, int dataType){
    if (dataType == _PFORMAT_BINARY){
	uint numBody = mCAstrBodyBinary.size();
	for (uint i = 0; i < numBody; i++)
	    if (mCAstrBodyBinary[i].IsSameAs(strItem)) return TRUE;
    }
    if (dataType == _PFORMAT_ASCII){
	uint numBody = mCAstrBodyAscii.size();
	for (uint i = 0; i < numBody; i++)
	    if (mCAstrBodyAscii[i].IsSameAs(strItem)) return TRUE;
    }
    return FALSE;
}

bool CLS::
CheckFormat(int dataType){
    if (dataType != mCDataType.GetBodyINT()) return FALSE;
    if (mCAstrHeader.size() != mHeaderPanelItemNum) return FALSE;
    if (mCDataType.GetBodyINT() == _PFORMAT_BINARY)
	if (mCAstrBodyBinary.size() != mBodyPanelItemNum) return FALSE;
    if (mCDataType.GetBodyINT() == _PFORMAT_ASCII)
	if (mCAstrBodyAscii.size() != mBodyPanelItemNum) return FALSE;
    return TRUE;
}

int CLS::
GetBodyParticleByte(){
    int byte = 0;
    uint numBody = mCAstrBodyBinary.size();
    for (uint i = 0; i < numBody; i++){
	if (mCAstrBodyBinary[i].IsSameAs("XYZ(float)"))    byte+=12;
	if (mCAstrBodyBinary[i].IsSameAs("VXVYVZ(float)")) byte+=12;
	if (mCAstrBodyBinary[i].IsSameAs("Type(int)"))     byte+=4;
	if (mCAstrBodyBinary[i].IsSameAs("Size(float)"))   byte+=4;
	if (mCAstrBodyBinary[i].IsSameAs("ID(int)"))       byte+=4;
	if (mCAstrBodyBinary[i].IsSameAs("ID(long long"))  byte+=8;
	if (mCAstrBodyBinary[i].IsSameAs("Value0(float)")) byte+=4;
	if (mCAstrBodyBinary[i].IsSameAs("Value1(float)")) byte+=4;
	if (mCAstrBodyBinary[i].IsSameAs("Value2(float)")) byte+=4;
	if (mCAstrBodyBinary[i].IsSameAs("Dummy(4byte)"))  byte+=4;
	if (mCAstrBodyBinary[i].IsSameAs("Dummyx3(12byte)")) byte+=12;
	if (mCAstrBodyBinary[i].IsSameAs("Dummyx10(40byte)")) byte+=40;
    }
    return byte;
}

int CLS::
GetBodyParticleLineAsciiVertical(){
    int line = 0;
    uint numBody = mCAstrBodyAscii.size();
    for (uint i = 0; i < numBody; i++){
	if (mCAstrBodyAscii[i].IsSameAs("XYZ"))      line++;
	if (mCAstrBodyAscii[i].IsSameAs("VXVYVZ"))   line++;
	if (mCAstrBodyAscii[i].IsSameAs("Type"))     line++;
	if (mCAstrBodyAscii[i].IsSameAs("Size"))     line++;
	if (mCAstrBodyAscii[i].IsSameAs("ID"))       line++;
	if (mCAstrBodyAscii[i].IsSameAs("ID(long long)")) line++;
	if (mCAstrBodyAscii[i].IsSameAs("Value0"))   line++;
	if (mCAstrBodyAscii[i].IsSameAs("Value1"))   line++;
	if (mCAstrBodyAscii[i].IsSameAs("Value2"))   line++;
	if (mCAstrBodyAscii[i].IsSameAs("Dummy"))    line++;
	if (mCAstrBodyAscii[i].IsSameAs("Dummyx3"))  line+=3;
	if (mCAstrBodyAscii[i].IsSameAs("Dummyx10")) line+=10;
	if (mCAstrBodyAscii[i].IsSameAs("X"))        line++;
	if (mCAstrBodyAscii[i].IsSameAs("Y"))        line++;
	if (mCAstrBodyAscii[i].IsSameAs("Z"))        line++;
	if (mCAstrBodyAscii[i].IsSameAs("VX"))       line++;
	if (mCAstrBodyAscii[i].IsSameAs("VY"))       line++;
	if (mCAstrBodyAscii[i].IsSameAs("VZ"))       line++;
    }
    return line;
}

std::vector < int > CLS::
GetVFormatHeader(){
    std::vector < int > VFormat;
    uint numHeader = mCAstrHeader.size();
    for (uint i = 0; i < numHeader; i++){
	wxString strItem = mCAstrHeader[i];
	for (int j = 0; j < _PFORMAT_HEADER_MAX; j++)
	    if (strItem.IsSameAs(GetStrHeaderFromIndex(j)))
		VFormat.push_back(j);
    }
    return VFormat;
}

std::vector < int > CLS::
GetVFormatBody(int dataType){
    std::vector < int > VFormat;
    if (dataType == _PFORMAT_BINARY){
	uint numBody = mCAstrBodyBinary.size();
	for (uint i = 0; i < numBody; i++){
	    wxString strItem = mCAstrBodyBinary[i];
	    for (int j = 0; j < _PFORMAT_MAX; j++)
		if (strItem.IsSameAs(GetStrBodyFromIndex(j, _PFORMAT_BINARY)))
		    VFormat.push_back(j);
	    }  
    }
    if (dataType == _PFORMAT_ASCII){
	uint numBody = mCAstrBodyAscii.size();
	for (uint i = 0; i < numBody; i++){
	    wxString strItem = mCAstrBodyAscii[i];
	    for (int j = 0; j < _PFORMAT_ASCII_MAX; j++)
		if (strItem.IsSameAs(GetStrBodyFromIndex(j, _PFORMAT_ASCII)))
		    VFormat.push_back(j);
	} 
    }
    return VFormat;
}

int CLS::
GetHeaderByte(){
    int byte = 0;
    uint numHeader = mCAstrHeader.size();
    for (uint i = 0; i < numHeader; i++){
	if (mCAstrHeader[i].IsSameAs("ParticleNumber")) byte+=4;
	if (mCAstrHeader[i].IsSameAs("Time")) byte+=4;
	if (mCAstrHeader[i].IsSameAs("Dummy")) byte+=4;
	if (mCAstrHeader[i].IsSameAs("Dummyx3")) byte+=12;
	if (mCAstrHeader[i].IsSameAs("Dummyx10")) byte+=40;
    }
    return byte;
}

int CLS::
GetHeaderLine(){
    int line = 0;
    int numHeader = mCAstrHeader.size();
    for (uint i = 0; i < numHeader; i++){
	if (mCAstrHeader[i].IsSameAs("ParticleNumber")) line++;
	if (mCAstrHeader[i].IsSameAs("Time")) line++;
	if (mCAstrHeader[i].IsSameAs("Dummy")) line++;
	if (mCAstrHeader[i].IsSameAs("Dummyx3")) line+=3;
	if (mCAstrHeader[i].IsSameAs("Dummyx10")) line+=10;
    }
    return line;
}

wxString CLS::
GetStrTag(wxString strObj){
    return GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strObj << "::";
}

void CLS::
DeleteHeaderCommand(int index, wxString strObj){
    wxString strTag = GetStrTag(strObj);
    wxString strUndo(strTag), strRedo(strTag);
    mCAstrHeaderFormat.SerializeSingleItem(strUndo);
    mCAstrHeader.RemoveAt(index, 1);
    mCAstrHeaderFormat.SerializeSingleItem(strRedo);
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Skip();
}

void CLS::
AddHeaderCommand(int index, wxString strObj){
    wxString strTag = GetStrTag(strObj);
    wxString strUndo(strTag), strRedo(strTag);
    mCAstrHeaderFormat.SerializeSingleItem(strUndo);
    wxString strItem = "Time";
    if (mCAstrHeader.size() > 0)
	strItem = mCAstrHeader[index];
    mCAstrHeader.Insert(strItem, index);
    mCAstrHeaderFormat.SerializeSingleItem(strRedo);
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Skip();
}

void CLS::
UpDownHeaderCommand(int index, wxString strObj, int direction){
    wxString strTag = GetStrTag(strObj);
    wxString strUndo(strTag), strRedo(strTag);
    mCAstrHeaderFormat.SerializeSingleItem(strUndo);
    wxString strItem = mCAstrHeader[index+direction];
    mCAstrHeader[index+direction] = mCAstrHeader[index];
    mCAstrHeader[index] = strItem;
    mCAstrHeaderFormat.SerializeSingleItem(strRedo);
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Skip();
}

void CLS::
UpdateHeaderCommand(wxString strObj){
    wxString strTag = GetStrTag(strObj);
    wxString strUndo(strTag), strRedo(strTag);
    mCAstrHeaderFormat.SerializeSingleItem(strUndo);
    UpdateHeader();
    mCAstrHeaderFormat.SerializeSingleItem(strRedo);
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Skip();
}

void CLS::
DeleteBodyCommand(int index, wxString strObj){
    wxString strTag = GetStrTag(strObj);
    wxString strUndo(strTag), strRedo(strTag);

    if (mCDataType.GetBodyINT() == _PFORMAT_BINARY){
	mCAstrBodyFormatBinary.SerializeSingleItem(strUndo);
	mCAstrBodyBinary.RemoveAt(index, 1);
	mCAstrBodyFormatBinary.SerializeSingleItem(strRedo);
    }
    if (mCDataType.GetBodyINT() == _PFORMAT_ASCII){
	mCAstrBodyFormatAscii.SerializeSingleItem(strUndo);
	mCAstrBodyAscii.RemoveAt(index, 1);
	mCAstrBodyFormatAscii.SerializeSingleItem(strRedo);
    }
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Skip();
}

void CLS::
AddBodyCommand(int index, wxString strObj){
    wxString strTag = GetStrTag(strObj);
    wxString strUndo(strTag), strRedo(strTag);

    if (mCDataType.GetBodyINT() == _PFORMAT_BINARY){
	mCAstrBodyFormatBinary.SerializeSingleItem(strUndo);
	wxString strItem = "XYZ(float)";
	if (mCAstrBodyBinary.size() > 0)
	    strItem = mCAstrBodyBinary[index];
	mCAstrBodyBinary.Insert(strItem, index);
	mCAstrBodyFormatBinary.SerializeSingleItem(strRedo);
    }
    if (mCDataType.GetBodyINT() == _PFORMAT_ASCII){
	mCAstrBodyFormatAscii.SerializeSingleItem(strUndo);
	wxString strItem = "XYZ";
	if (mCAstrBodyAscii.size() > 0)
	    strItem = mCAstrBodyAscii[index];
	mCAstrBodyAscii.Insert(strItem, index);
	mCAstrBodyFormatAscii.SerializeSingleItem(strRedo);
    }
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Skip();
}

void CLS::
UpDownBodyCommand(int index, wxString strObj, int direction){
    wxString strTag = GetStrTag(strObj);
    wxString strUndo(strTag), strRedo(strTag);

    if (mCDataType.GetBodyINT() == _PFORMAT_BINARY){
	mCAstrBodyFormatBinary.SerializeSingleItem(strUndo);
	wxString strItem = mCAstrBodyBinary[index+direction];
	mCAstrBodyBinary[index+direction] = mCAstrBodyBinary[index];
	mCAstrBodyBinary[index] = strItem;
	mCAstrBodyFormatBinary.SerializeSingleItem(strRedo);
    }
    if (mCDataType.GetBodyINT() == _PFORMAT_ASCII){
	mCAstrBodyFormatAscii.SerializeSingleItem(strUndo);
	wxString strItem = mCAstrBodyAscii[index+direction];
	mCAstrBodyAscii[index+direction] = mCAstrBodyAscii[index];
	mCAstrBodyAscii[index] = strItem;
	mCAstrBodyFormatAscii.SerializeSingleItem(strRedo);
    }
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Skip();
}

void CLS::
UpdateBodyCommand(wxString strObj){
    wxString strTag = GetStrTag(strObj);
    wxString strUndo(strTag), strRedo(strTag);

    if (mCDataType.GetBodyINT() == _PFORMAT_BINARY){
	mCAstrBodyFormatBinary.SerializeSingleItem(strUndo);
	UpdateBody(_PFORMAT_BINARY);
	mCAstrBodyFormatBinary.SerializeSingleItem(strRedo);
    }
    if (mCDataType.GetBodyINT() == _PFORMAT_ASCII){
	mCAstrBodyFormatAscii.SerializeSingleItem(strUndo);
	UpdateBody(_PFORMAT_ASCII);
	mCAstrBodyFormatAscii.SerializeSingleItem(strRedo);
    }
    
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Skip();
    
}

int CLS::
GetTypeIndexHeader(wxString strItem){
    if (strItem.IsSameAs("ParticleNumber")) return _PFORMAT_HEADER_NUMBER;
    if (strItem.IsSameAs("Time"))           return _PFORMAT_HEADER_TIME;
    if (strItem.IsSameAs("Dummy"))          return _PFORMAT_HEADER_DUMMY;
    if (strItem.IsSameAs("Dummyx3"))        return _PFORMAT_HEADER_DUMMYx3;
    if (strItem.IsSameAs("Dummyx10"))       return _PFORMAT_HEADER_DUMMYx10;
}

wxString CLS::
GetStrHeaderFromIndex(int index){
    wxString strItem;
    switch(index){
    case _PFORMAT_HEADER_NUMBER:   strItem = "ParticleNumber"; break;
    case _PFORMAT_HEADER_TIME:     strItem = "Time";           break;
    case _PFORMAT_HEADER_DUMMY:    strItem = "Dummy";          break;
    case _PFORMAT_HEADER_DUMMYx3:  strItem = "Dummyx3";        break;
    case _PFORMAT_HEADER_DUMMYx10: strItem = "Dummyx10";       break;
    }
    return strItem;
}

int CLS::
GetTypeIndexBody(wxString strItem, int  dataType){
    if (dataType == _PFORMAT_BINARY){
	if (strItem.IsSameAs("XYZ(float)"))    return _PFORMAT_XYZ_FLOAT;
	if (strItem.IsSameAs("VXVYVZ(float)")) return _PFORMAT_VXVYVZ;
	if (strItem.IsSameAs("Type(int)"))     return _PFORMAT_TYPE_INT;
	if (strItem.IsSameAs("Size(float)"))   return _PFORMAT_SIZE_FLOAT;
	if (strItem.IsSameAs("ID(int)"))       return _PFORMAT_ID_INT;
	if (strItem.IsSameAs("ID(long long)")) return _PFORMAT_ID_LONGLONG;
	if (strItem.IsSameAs("Value0(float)")) return _PFORMAT_VALUE_0;
	if (strItem.IsSameAs("Value1(float)")) return _PFORMAT_VALUE_1;
	if (strItem.IsSameAs("Value2(float)")) return _PFORMAT_VALUE_2;
	if (strItem.IsSameAs("Dummy(4byte)"))     return _PFORMAT_DUMMY;
	if (strItem.IsSameAs("Dummyx3(12byte)"))  return _PFORMAT_DUMMYx3;
	if (strItem.IsSameAs("Dummyx10(40byte)")) return _PFORMAT_DUMMYx10;
    }
    if (dataType == _PFORMAT_ASCII){
	if (strItem.IsSameAs("XYZ"))       return _PFORMAT_ASCII_XYZ;
	if (strItem.IsSameAs("VXVYVZ"))    return _PFORMAT_ASCII_VXVYVZ;
	if (strItem.IsSameAs("Type"))      return _PFORMAT_ASCII_TYPE;
	if (strItem.IsSameAs("Size"))      return _PFORMAT_ASCII_SIZE;
	if (strItem.IsSameAs("ID"))        return _PFORMAT_ASCII_ID;
	if (strItem.IsSameAs("ID(long long)")) return _PFORMAT_ASCII_ID_LONGLOG;
	if (strItem.IsSameAs("Value0"))    return _PFORMAT_ASCII_VALUE0;
	if (strItem.IsSameAs("Value1"))    return _PFORMAT_ASCII_VALUE1;
	if (strItem.IsSameAs("Value2"))    return _PFORMAT_ASCII_VALUE2;
	if (strItem.IsSameAs("Dummy"))     return _PFORMAT_ASCII_DUMMY;
	if (strItem.IsSameAs("Dummyx3"))   return _PFORMAT_ASCII_DUMMYx3;
	if (strItem.IsSameAs("Dummyx10"))  return _PFORMAT_ASCII_DUMMYx10;
	if (strItem.IsSameAs("X"))         return _PFORMAT_ASCII_X;
	if (strItem.IsSameAs("Y"))         return _PFORMAT_ASCII_Y;
	if (strItem.IsSameAs("Z"))         return _PFORMAT_ASCII_Z;
	if (strItem.IsSameAs("VX"))        return _PFORMAT_ASCII_VX;
	if (strItem.IsSameAs("VY"))        return _PFORMAT_ASCII_VY;
	if (strItem.IsSameAs("VZ"))        return _PFORMAT_ASCII_VZ;
    }
}

wxString CLS::
GetStrBodyFromIndex(int index, int dataType){
    wxString strItem;
    if (dataType == _PFORMAT_BINARY)
	switch(index){
	case _PFORMAT_XYZ_FLOAT:   strItem = "XYZ(float)";       break;
	case _PFORMAT_TYPE_INT:    strItem = "Type(int)";        break;
	case _PFORMAT_SIZE_FLOAT:  strItem = "Size(float)";      break;
	case _PFORMAT_ID_INT:      strItem = "ID(int)";          break;
	case _PFORMAT_ID_LONGLONG: strItem = "ID(long long)";    break;
	case _PFORMAT_VALUE_0:     strItem = "Value0(float)";    break;
	case _PFORMAT_VALUE_1:     strItem = "Value1(float)";    break;
	case _PFORMAT_VALUE_2:     strItem = "Value2(float)";    break;
	case _PFORMAT_VXVYVZ:      strItem = "VXVYVZ(float)";    break;
	case _PFORMAT_DUMMY:       strItem = "Dummy(4byte)";     break;
	case _PFORMAT_DUMMYx3:     strItem = "Dummyx3(12byte)";  break;
	case _PFORMAT_DUMMYx10:    strItem = "Dummyx10(40byte)"; break;
	}
    if (dataType == _PFORMAT_ASCII)
	switch(index){
	case _PFORMAT_ASCII_XYZ:      strItem = "XYZ";      break;
	case _PFORMAT_ASCII_TYPE:     strItem = "Type";     break;
	case _PFORMAT_ASCII_SIZE:     strItem = "Size";     break;
	case _PFORMAT_ASCII_ID:       strItem = "ID";       break;
	case _PFORMAT_ASCII_ID_LONGLOG: strItem = "ID(long long)";    break;
	case _PFORMAT_ASCII_VALUE0:   strItem = "Value0";   break;
	case _PFORMAT_ASCII_VALUE1:   strItem = "Value1";   break;
	case _PFORMAT_ASCII_VALUE2:   strItem = "Value2";   break;
	case _PFORMAT_ASCII_VXVYVZ:   strItem = "VXVYVZ";   break;
	case _PFORMAT_ASCII_DUMMY:    strItem = "Dummy";    break;
	case _PFORMAT_ASCII_DUMMYx3:  strItem = "Dummyx3";  break;
	case _PFORMAT_ASCII_DUMMYx10: strItem = "Dummyx10"; break;
	case _PFORMAT_ASCII_X:        strItem = "X"; break;
	case _PFORMAT_ASCII_Y:        strItem = "Y"; break;
	case _PFORMAT_ASCII_Z:        strItem = "Z"; break;
	case _PFORMAT_ASCII_VX:       strItem = "VX"; break;
	case _PFORMAT_ASCII_VY:       strItem = "VY"; break;
	case _PFORMAT_ASCII_VZ:       strItem = "VZ"; break;
	}
    return strItem;
}

void CLS::
UpdateHeader(){
    mCAstrHeader.clear();

    uint numItem = mVCDummyHeaderItem.size();
    for (uint i = 0; i < numItem; i++)
	mCAstrHeader.push_back(GetStrHeaderFromIndex(mVCDummyHeaderItem[i].GetBodyINT()));
}

void CLS::
UpdateFormatPanel(MNumPainBase_cls *pCPain){
    MultiInputPanel_cls *pC = pCPain->GetPCMNum("Format.Main");

    bool bExistID = FALSE;
    int type = mCDataType.GetBodyINT();
    if (type == _PFORMAT_ASCII){
	bExistID |= DoesExistInBody("ID", _PFORMAT_ASCII);
	bExistID |= DoesExistInBody("ID(long long)", _PFORMAT_ASCII);
    }
    if (type == _PFORMAT_BINARY){
	bExistID |= DoesExistInBody("ID(int)", _PFORMAT_BINARY);
	bExistID |= DoesExistInBody("ID(long long)", _PFORMAT_BINARY);
    }

    bool bWarn = FALSE;
    if (bExistID == TRUE && mpCParentParticle->mCInterpolateBy.GetBodyINT() == _P_INTERP_BY_INDEX){
	bWarn = TRUE;
    }
    if (bExistID == FALSE && mpCParentParticle->mCInterpolateBy.GetBodyINT() == _P_INTERP_BY_ID){
	bWarn = TRUE;
    }
    
    wxWindow *pCItem = pC->mVCItem[mIndexWarningInterp].mpCWindowLink;
    if (bWarn == TRUE){
	if (pCItem->IsShown() == FALSE){
	    pCItem->Show();
	    pCPain->Layout();
	}
    }
    else{
	if (pCItem->IsShown() == TRUE){
	    pCItem->Hide();
	    pCPain->Layout();
	}
    }
}

void CLS::
CheckWarning(MNumPainBase_cls *pCPain, MultiInputPanel_cls *pC, bool bWarn){
    wxWindow *pCItem = pC->mpCStatusLink;
    if (bWarn == TRUE){
	if (pCItem->IsShown() == FALSE){
	    pCItem->Show();
	    pCPain->Layout();
	}
    }
    else{
	if (pCItem->IsShown() == TRUE){
	    pCItem->Hide();
	    pCPain->Layout();
	}
    }
}

void CLS::
UpdateHeaderPanel(MNumPainBase_cls *pCPain){
    MultiInputPanel_cls *pC = pCPain->GetPCMNum("Header");

    uint numItem = mVCDummyHeaderItem.size();
    for (uint i = 0; i < numItem; i++){
	wxString strItem = mCAstrHeader[i];
	mVCDummyHeaderItem[i].SetBodyFromData((long long)GetTypeIndexHeader(strItem));
    }

    bool bWarn = FALSE;
    for (uint i = 0; i < numItem; i++){
	wxString strItem = mCAstrHeader[i];
	if (strItem.Contains("Dummy")) continue;
	for (uint j = 0; j < numItem; j++){
	    if (j == i) continue;
	    if (mCAstrHeader[j].IsSameAs(strItem)) bWarn = TRUE;
	}
    }
    
    CheckWarning(pCPain, pC, bWarn);

    pC->UpdateValues(_MN_ALLITEM);
}

void CLS::
UpdateBody(int dataType){
    uint numItem = mVCDummyBodyItem.size();
    if (dataType == _PFORMAT_BINARY){
	mCAstrBodyBinary.clear();
	for (uint i = 0; i < numItem; i++)
	    mCAstrBodyBinary.push_back(GetStrBodyFromIndex(mVCDummyBodyItem[i].GetBodyINT(), _PFORMAT_BINARY));
    }
    if (dataType == _PFORMAT_ASCII){
	mCAstrBodyAscii.clear();
	for (uint i = 0; i < numItem; i++)
	    mCAstrBodyAscii.push_back(GetStrBodyFromIndex(mVCDummyBodyItem[i].GetBodyINT(), _PFORMAT_ASCII));
    }
}

void CLS::
UpdateBodyPanel(MNumPainBase_cls *pCPain){
    MultiInputPanel_cls *pC = pCPain->GetPCMNum("Body");

    bool bWarn = FALSE;
    uint numItem = mVCDummyBodyItem.size();
    if (mCDataType.GetBodyINT() == _PFORMAT_BINARY){
	for (uint i = 0; i < numItem; i++){
	    wxString strItem = mCAstrBodyBinary[i];
	    mVCDummyBodyItem[i].SetBodyFromData((long long)GetTypeIndexBody(strItem, _PFORMAT_BINARY)); 
	}
	for (uint i = 0; i < numItem; i++){
	    wxString strItem = mCAstrBodyBinary[i];
	    if (strItem.Contains("Dummy")) continue;
	    for (uint j = 0; j < numItem; j++){
		if (j == i) continue;
		if (mCAstrBodyBinary[j].IsSameAs(strItem)) bWarn = TRUE;
	    }
	}
    }
    if (mCDataType.GetBodyINT() == _PFORMAT_ASCII){
	for (uint i = 0; i < numItem; i++){
	    wxString strItem = mCAstrBodyAscii[i];
	    mVCDummyBodyItem[i].SetBodyFromData((long long)GetTypeIndexBody(strItem, _PFORMAT_ASCII));
	}
	for (uint i = 0; i < numItem; i++){
	    wxString strItem = mCAstrBodyAscii[i];
	    if (strItem.Contains("Dummy")) continue;
	    for (uint j = 0; j < numItem; j++){
		if (j == i) continue;
		if (mCAstrBodyAscii[j].IsSameAs(strItem)) bWarn = TRUE;
	    }
	}
    }

    CheckWarning(pCPain, pC, bWarn);

    pC->UpdateValues(_MN_ALLITEM);
}

MultiInputPanel_cls* CLS::
GetPCNewPanelFormatMain(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Format.Main", TRUE); 
    
    pC->mTextWidth = 100;

    pC->CombineSizer(2, _("Type/Direction"));
    wxString AstrType[2] = {_("Ascii"), _("Binary")};
    pC->SetChoiceList(2, AstrType);
    pC->CombineConfItem(mCDataType);
    wxString AstrDirection[2] = {_("Horizontal"), _("Vertical")};
    pC->SetChoiceList(2, AstrDirection);
    pC->CombineConfItem(mCArrayDirection);

    pC->CombineSizer(3, _("Interpolation"));
    pC->SetChoiceList(5, GetPCStringProvider()->GetAstr("InterpType"));
    pC->CombineConfItem(mpCParentParticle->mCInterpType);
    wxString AstrInterpolateBy[2] = {_("by index"), _("by ID") };
    pC->SetChoiceList(2, AstrInterpolateBy);
    pC->CombineConfItem(mpCParentParticle->mCInterpolateBy);
    
    mIndexWarningInterp = pC->CombineBitmapPanel(wxID_ANY, wxES, 24);

    pC->CombineSizer(2, _("Disappear"));
    wxString AstrInterpolateSizeType[4] = {_("Constant size"), _("Decrease size"), _("Set to 0"), _("Stocastic")};
    pC->SetChoiceList(4, AstrInterpolateSizeType);
    pC->CombineConfItem(mpCParentParticle->mCVanishSizeType);
    wxString AstrExtrapolate[2] = {_("Constant location"), _("Extrapolate location")};
    pC->SetChoiceList(2, AstrExtrapolate);
    pC->CombineConfItem(mpCParentParticle->mCExtrapolateType);

    pC->CombineSizer(2);
    pC->CombineConfItem(mCbPreSortByID, _("Do Sort by ID"));
    pC->CombineConfItem(mCbRemoveDupli, _("Remove Dupli ID"));
    
    pC->FinishSetting();

    BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[mIndexWarningInterp].mpCWindowLink;
    pCBitmap->AddBitmapStr("wxART_WARNING");
    pCBitmap->mBorderX = 4;
    pCBitmap->mBorderY = 4;
    pCBitmap->SetToolTip(_("Warning: Sorting data with ID by index, or sorting data without ID by ID."));
    
    return pC;
}

void CLS::
SetBitmapInfo(MultiInputPanel_cls *pC,
	      std::vector < int > VIndexDelete,
	      std::vector < int > VIndexUp,
	      std::vector < int > VIndexDown){
    
    uint numDelete = VIndexDelete.size();
    for (uint i = 0; i < numDelete; i++){
	BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[VIndexDelete[i]].mpCWindowLink;
	pCBitmap->mbButton = TRUE;
	pCBitmap->AddBitmapStr("wxART_CLOSE_PANEL", "wxART_CLOSE_PANEL_P");
	pCBitmap->SetBitmapStrDisable("wxART_CLOSE_PANEL_D");
	pCBitmap->mBorderX = 4;
	pCBitmap->mBorderY = 4;
    }

    uint numUp = VIndexUp.size();
    for (uint i = 0; i < numUp; i++){
	BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[VIndexUp[i]].mpCWindowLink;
	pCBitmap->mbButton = TRUE;
	pCBitmap->AddBitmapStr("wxART_ARROW_UP", "wxART_ARROW_UP_P");
	pCBitmap->SetBitmapStrDisable("wxART_ARROW_UP_D");
	pCBitmap->mBorderX = 3;
	pCBitmap->mBorderY = 4;
	if (i == 0) pCBitmap->Disable();
    }

    uint numDown = VIndexUp.size();
    for (uint i = 0; i < numDown; i++){
	BitmapPanel_cls *pCBitmap = (BitmapPanel_cls*)pC->mVCItem[VIndexDown[i]].mpCWindowLink;
	pCBitmap->mbButton = TRUE;
	pCBitmap->AddBitmapStr("wxART_ARROW_DOWN", "wxART_ARROW_DOWN_P");
	pCBitmap->SetBitmapStrDisable("wxART_ARROW_DOWN_D");
	pCBitmap->mBorderX = 3;
	pCBitmap->mBorderY = 4;
	if (i == numDown - 1) pCBitmap->Disable();
    }
}

MultiInputPanel_cls* CLS::
GetPCNewPanelFormatHeader(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Header", TRUE);
    pC->mbUseStatus = TRUE;
    
    pC->mTextWidth = 80;

    int numItem = mCAstrHeader.size();
    mVCDummyHeaderItem.clear();

    for (int i = 0; i < numItem; i++){
	ConfItemUINT_cls CItem;
	wxString strItem = mCAstrHeader[i];
	int typeIndex = GetTypeIndexHeader(strItem);
	CItem.mstrLinkedPanel = "PROPERTY_FORMAT";
	CItem.SetBodyFromData((long long)typeIndex);
	mVCDummyHeaderItem.push_back(CItem);
    }

    std::vector < int > VIndexUp, VIndexDown, VIndexDelete;
    for (int i = 0; i < numItem; i++){
	wxString strIndex; strIndex << i;
	pC->CombineSizer(5, strIndex);
	wxString AstrList[5];
	for (int j = 0; j < 5; j++) AstrList[j] = GetStrHeaderFromIndex(j);
	pC->SetChoiceList(5, AstrList);
	pC->CombineConfItem(mVCDummyHeaderItem[i], strIndex);
	pC->CombineButton(btnFormatHeaderAdd0+i, "+", 30);
	int indexUp = pC->CombineBitmapPanel(btnFormatHeaderUp0+i, "Up", 22);
	int indexDown = pC->CombineBitmapPanel(btnFormatHeaderDown0+i, "Down", 22);
	int indexDelete = pC->CombineBitmapPanel(btnFormatHeaderDelete0+i, "Delete", 24);
	VIndexUp.push_back(indexUp);
	VIndexDown.push_back(indexDown);
	VIndexDelete.push_back(indexDelete);
    }
    if (numItem == 0){
	pC->CombineButton(btnFormatHeaderAdd0, "+", 30);
    }

    mHeaderPanelItemNum = numItem;
    pC->FinishSetting();

    SetBitmapInfo(pC, VIndexDelete, VIndexUp, VIndexDown);

    return pC;
}

MultiInputPanel_cls* CLS::GetPCNewPanelFormatBody(wxWindow *pCWindow){
    MultiInputPanel_cls *pC = 
	new MultiInputPanel_cls(pCWindow, "Body", TRUE);
    pC->mbUseStatus = TRUE;

    pC->mTextWidth = 80;

    int dataType = mCDataType.GetBodyINT();

    mVCDummyBodyItem.clear();

    int numItem = (dataType == _PFORMAT_BINARY) ? 
	mCAstrBodyBinary.size() : mCAstrBodyAscii.size();

    if (dataType == _PFORMAT_BINARY){
	for (int i = 0; i < numItem; i++){
	    ConfItemUINT_cls CItem;
	    wxString strItem = mCAstrBodyBinary[i];
	    int typeIndex = GetTypeIndexBody(strItem, _PFORMAT_BINARY);
	    CItem.mstrLinkedPanel = "PROPERTY_FORMAT";
	    CItem.SetBodyFromData((long long)typeIndex);
	    mVCDummyBodyItem.push_back(CItem);
	}
    }
    if (dataType == _PFORMAT_ASCII)
	for (int i = 0; i < numItem; i++){
	    ConfItemUINT_cls CItem;
	    wxString strItem = mCAstrBodyAscii[i];
	    int typeIndex = GetTypeIndexBody(strItem, _PFORMAT_ASCII);
	    CItem.mstrLinkedPanel = "PROPERTY_FORMAT";
	    CItem.SetBodyFromData((long long)typeIndex);
	    mVCDummyBodyItem.push_back(CItem);
	}
    if (numItem == 0){
	pC->CombineButton(btnFormatBodyAdd0, "+", 30);
    }
    
    std::vector < int > VIndexUp, VIndexDown, VIndexDelete;
    for (int i = 0; i < numItem; i++){
	wxString strIndex; strIndex << i;
	pC->CombineSizer(5, strIndex);
	wxString AstrList[_PFORMAT_ASCII_MAX]; // BinaryAsciȋ傫ق
	int max = (dataType == _PFORMAT_BINARY) ?
	    _PFORMAT_MAX : _PFORMAT_ASCII_MAX;
	for (int j = 0; j < max; j++) 
	    AstrList[j] = GetStrBodyFromIndex(j, dataType);
	pC->SetChoiceList(max, AstrList);
	pC->CombineConfItem(mVCDummyBodyItem[i], strIndex);
	pC->CombineButton(btnFormatBodyAdd0+i, "+", 30);
	int indexUp = pC->CombineBitmapPanel(btnFormatBodyUp0+i, "Up", 22);
	int indexDown = pC->CombineBitmapPanel(btnFormatBodyDown0+i, "Down", 22);
	int indexDelete = pC->CombineBitmapPanel(btnFormatBodyDelete0+i, "Delete", 24);
	VIndexUp.push_back(indexUp);
	VIndexDown.push_back(indexDown);
	VIndexDelete.push_back(indexDelete);
    }

    mBodyPanelItemNum = numItem;
    pC->FinishSetting();

    SetBitmapInfo(pC, VIndexDelete, VIndexUp, VIndexDown);

    return pC;
}
