// -*- c -*- 
// "clsParticleFormat.cpp"

#ifndef _PARTICLE_FORMAT_H
#define _PARTICLE_FORMAT_H

#include "../CONFITEM/clsConfItem.h"
#include "../CONFITEM/clsConfListContainer.h"

#include "../COMMON/clsGetPC.h"

#define _PFORMAT_ASCII  0
#define _PFORMAT_BINARY 1

#define _PFORMAT_ARRAY_HORIZONTAL 0
#define _PFORMAT_ARRAY_VERTICLAL  1

#define _PFORMAT_HEADER_NUMBER   0
#define _PFORMAT_HEADER_TIME     1
#define _PFORMAT_HEADER_DUMMY    2
#define _PFORMAT_HEADER_DUMMYx3  3
#define _PFORMAT_HEADER_DUMMYx10 4
#define _PFORMAT_HEADER_MAX      5

#define _PFORMAT_XYZ_FLOAT   0
#define _PFORMAT_VXVYVZ      1
#define _PFORMAT_TYPE_INT    2
#define _PFORMAT_SIZE_FLOAT  3
#define _PFORMAT_ID_INT      4
#define _PFORMAT_ID_LONGLONG 5
#define _PFORMAT_VALUE_0     6
#define _PFORMAT_VALUE_1     7
#define _PFORMAT_VALUE_2     8
#define _PFORMAT_DUMMY       9
#define _PFORMAT_DUMMYx3     10
#define _PFORMAT_DUMMYx10    11
#define _PFORMAT_MAX 12

#define _PFORMAT_ASCII_XYZ    0
#define _PFORMAT_ASCII_VXVYVZ 1
#define _PFORMAT_ASCII_TYPE   2
#define _PFORMAT_ASCII_SIZE   3
#define _PFORMAT_ASCII_ID     4
#define _PFORMAT_ASCII_ID_LONGLOG 5
#define _PFORMAT_ASCII_VALUE0 6
#define _PFORMAT_ASCII_VALUE1 7
#define _PFORMAT_ASCII_VALUE2 8
#define _PFORMAT_ASCII_DUMMY  9
#define _PFORMAT_ASCII_DUMMYx3  10
#define _PFORMAT_ASCII_DUMMYx10 11
#define _PFORMAT_ASCII_X      12
#define _PFORMAT_ASCII_Y      13
#define _PFORMAT_ASCII_Z      14
#define _PFORMAT_ASCII_VX     15
#define _PFORMAT_ASCII_VY     16
#define _PFORMAT_ASCII_VZ     17
#define _PFORMAT_ASCII_MAX    18

class MultiInputPanel_cls;
class MNumPainBase_cls;

class Particle_cls;
class ParticleFormat_cls;

class ParticleFormatClone_cls : public ConfListContainer_cls, public GetPC_cls {
public:
    ~ParticleFormatClone_cls();

    wxArrayString mCAstrHeader;
    wxArrayString mCAstrBodyBinary;
    wxArrayString mCAstrBodyAscii;

    bool mbInterpByID;

    bool DoesExistInBody(wxString strItem, int dataType);

    int GetBodyParticleByte();
    int GetBodyParticleLineAsciiVertical();

    std::vector < int > GetVFormatHeader();
    std::vector < int > GetVFormatBody(int dataType);
    
    int GetHeaderByte();
    int GetHeaderLine();

    void GetInfoBinary
	(bool *pbUseVelocity, bool *pbUseType, bool *pbUseSize, 
	 bool *pbUseID, bool *pbUseIDLongLong, bool *pbInterpolateByID,
	 bool *pbUseValue0, bool *pbUseValue1, bool *pbUseValue2);
    void GetInfoAscii
	(bool *pbUseVelocity, bool *pbUseType, bool *pbUseSize, 
	 bool *pbUseID, bool *pbUseIDLongLong, bool *pbInterpolateByID,
	 bool *pbUseValue0, bool *pbUseValue1, bool *pbUseValue2);

    wxString GetStrHeaderFromIndex(int index);
    wxString GetStrBodyFromIndex(int index, int type);

    void CloneFromFormat(ParticleFormat_cls &CFormat, bool bInterpByID);
};

class ParticleFormat_cls : public ConfListContainer_cls, public GetPC_cls {
 public:
    ParticleFormat_cls(Particle_cls *pCParentParticle, wxString strObj);

    Particle_cls *mpCParentParticle;
 
    wxArrayString mCAstrHeader;
    wxArrayString mCAstrBodyBinary;
    wxArrayString mCAstrBodyAscii;
    
    ConfItemUINT_cls mCDataType; //Binary Ascii;
    ConfItemUINT_cls mCArrayDirection;
    ConfItemBOOL_cls mCbPreSortByID;
    ConfItemBOOL_cls mCbRemoveDupli;
    //ConfItemUINT_cls mCInterpolateBy; //by index or ID;
    
    ConfItemSTRINGARRAY_cls mCAstrHeaderFormat;
    ConfItemSTRINGARRAY_cls mCAstrBodyFormatBinary;
    ConfItemSTRINGARRAY_cls mCAstrBodyFormatAscii;

    std::vector < ConfItemUINT_cls > mVCDummyHeaderItem;
    std::vector < ConfItemUINT_cls > mVCDummyBodyItem;
    
    void SyncConfListTag(wxString strObj);
    bool DoesExistInBody(wxString strItem, int dataType);
    bool CheckFormat(int dataType);
    
    int GetBodyParticleByte();
    int GetBodyParticleLineAsciiVertical();
    std::vector < int > GetVFormatHeader();
    std::vector < int > GetVFormatBody(int dataType);
        
    int GetHeaderByte();
    int GetHeaderLine();
    /*
    void GetInfoBinary(bool *pbUseType, bool *pbUseSize, 
		       bool *pbUseID, bool *pbUseIDLongLong, 
		       bool *pbUseValue0, bool *pbUseValue1, bool *pbUseValue2);
    void GetInfoAscii(bool *pbUseType, bool *pbUseSize, 
		      bool *pbUseID, bool *pbUseIDLongLong, 
		      bool *pbUseValue0, bool *pbUseValue1, bool *pbUseValue2);
    */
    
    wxString GetStrTag(wxString strObj);
    void DeleteHeaderCommand(int index, wxString strObj);
    void AddHeaderCommand(int index, wxString strObj);
    void UpDownHeaderCommand(int index, wxString strObj, int direction);
    void UpdateHeaderCommand(wxString strObj);
    void DeleteBodyCommand(int index, wxString strObj);
    void AddBodyCommand(int index, wxString strObj);
    void UpDownBodyCommand(int index, wxString strObj, int direction);
    void UpdateBodyCommand(wxString strObj);

    int GetTypeIndexHeader(wxString strItem);
    wxString GetStrHeaderFromIndex(int index);
    int GetTypeIndexBody(wxString strItem, int type);
    wxString GetStrBodyFromIndex(int index, int type);
    
    void CheckWarning(MNumPainBase_cls *pCPain,
		      MultiInputPanel_cls *pC, bool bWarn);
    void UpdateFormatPanel(MNumPainBase_cls *pCPain);
    void UpdateHeader();
    void UpdateHeaderPanel(MNumPainBase_cls *pCPain);
    void UpdateBody(int type);
    void UpdateBodyPanel(MNumPainBase_cls *pCPain);

    int mHeaderPanelItemNum, mBodyPanelItemNum;
    int mIndexWarningInterp;
    void SetBitmapInfo(MultiInputPanel_cls* pC,
		       std::vector < int > VIndexDelete,
		       std::vector < int > VIndexUp,
		       std::vector < int > VIndexDown);
    MultiInputPanel_cls *GetPCNewPanelFormatMain(wxWindow *pCWindow);
    MultiInputPanel_cls *GetPCNewPanelFormatHeader(wxWindow *pCWindow);
    MultiInputPanel_cls *GetPCNewPanelFormatBody(wxWindow *pCWindow);
};


#endif
