// -*- c -*- 
// "clsParticleSequence.h"

#include "clsParticleSequence.h"

#include "clsDataSnap.h"
#include "clsParticleSnap.h"

#include "../MODIFIER/clsSequenceRemap.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../PREFERENCE/clsPreference.h"
#include "../TIMELINE/clsTimeCtrl.h"

#include "../COMMON/CommonDefinitions.h"

#define CLS ParticleSequence_cls

CLS::
ParticleSequence_cls() : DataSequence_cls(_DATA_PARTICLE){
    mbTimeDepend = FALSE;
}

CLS::
~ParticleSequence_cls(){
    if (mApCSnap != NULL){
	for (int i = 0; i < mSeqLen; i++){
	    if (mApCSnap[i] != NULL){
		delete (ParticleSnap_cls*)mApCSnap[i];
		mApCSnap[i] = NULL;
	    }
	}
	delete mApCSnap;
	mApCSnap = NULL;
    }
}

int CLS:: 
GetTotalUsedMemoryMB(){
    long long memoryByte = 0;
    if (mApCSnap == NULL) return 0;
    for (int i = 0; i < mSeqLen; i++){
	if (mApCSnap[i] == NULL) continue;
	memoryByte += ((ParticleSnap_cls*)mApCSnap[i])->GetUsedMemoryByte();
    }

    mTotalUsedMemoryMB = (int)(memoryByte / 1024 / 1024);
    
    return mTotalUsedMemoryMB;
}

ParticleSnap_cls* CLS::
GenerateIfNot(int index){
    if (index >= mSeqLen) index = mSeqLen - 1;
    if (index < 0) index = 0;
    if (mbTimeDepend == FALSE) index = 0;
    
    if (mApCSnap == NULL) return NULL;

    if (mApCSnap[index] == NULL){
	mApCSnap[index] = new ParticleSnap_cls(this, index);
    }

    return (ParticleSnap_cls*)mApCSnap[index];
}

ParticleSnap_cls* CLS::
GetPCSnap(int index){
    if (mApCSnap == NULL) return NULL;

    if (index >= mSeqLen) index = mSeqLen - 1;
    if (index < 0) index = 0;
    if (mbTimeDepend == FALSE) index = 0;
    
    return (ParticleSnap_cls*)mApCSnap[index];
}

bool CLS::
PrefetchData(ParticleFormatClone_cls *pCFormatClone,
	     SequenceRemap_cls *pCSeqRemap){
    
    if (mReadingLock > 0){
	return FALSE; // 1Xbĥ
    }

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    double dNT = GetPCTimeCtrl()->GetDNT();
    int frameBase, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frameBase, subframe);

    int seqLen = mSeqLen;
    
    int prefetch = GetPCPreference()->mCPrefetchLength.GetBodyINT();

    for (int i = 0; i < prefetch; i++){
	int frame1 = frameBase + i;
	if (pCSeqRemap != NULL)
	    frame1 = pCSeqRemap->GetRemappedFrame(NT+dNT*i, frame1, seqLen);
	    
	if (frame1 < 0 || frame1 >= seqLen) continue;

	GenerateIfNot(frame1);
	if (mbTimeDepend == TRUE)
	    GetPCSnap(frame1)->SetStrFile(GetModifiedStr(frame1));
	else
	    GetPCSnap(frame1)->SetStrFile(mstrPath);
	
	if (GetPCSnap(frame1)->PrefetchData(pCFormatClone/*, pCParticleFormat*/) == TRUE) 
	    return TRUE;

	frame1++; // ԗpɎXebvvtFb`
	if (frame1 >= seqLen) continue;
	GenerateIfNot(frame1);
	GetPCSnap(frame1)->SetStrFile(GetModifiedStr(frame1));
	if (GetPCSnap(frame1)->PrefetchData(pCFormatClone/*, pCParticleFormat*/) == TRUE) 
	    return TRUE;
    }

    return FALSE;
}

void CLS::
ReleaseMemoryData(SequenceRemap_cls *pCSeqRemap){
    
    ParamGroup_cls *pCRemapParamGroup = NULL;
    bool bRepeat = FALSE;
    if (pCSeqRemap != NULL){
	pCRemapParamGroup = pCSeqRemap->GetPCParamGroup("Remap");
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }
    MarkPriority(pCRemapParamGroup, bRepeat);
    int oldestIndex = GetOldestDrawnIndex();
    if (oldestIndex == -1) return;

    ParticleSnap_cls *pCSnap = GetPCSnap(oldestIndex);
    if (pCSnap == NULL)
	return;
    
    wxString strLog;
    strLog << "ReleaseLog :: " << pCSnap->mstrFile << ENDL;
    delete pCSnap;
    mApCSnap[oldestIndex] = NULL;
}
