// -*- c -*- 
// "clsParticleSnap.h"

#include "clsParticleSnap.h"

#include "clsParticleFormat.h"
#include "clsParticleSequence.h"

#include "../PREFERENCE/clsPreference.h"

#include "READTHREAD/clsParticleAsciiRead.h"
#include "READTHREAD/clsParticleBinaryRead.h"

#include "../COMMON/clsUseGZ.h"

#include "../../clsFrameMain.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/MyEndian.h"

#include "stdio.h"
#include <fstream>
#include <sstream>

#include <wx/filename.h>
#include <wx/filefn.h>

ThreadReadBuffer_cls::
ThreadReadBuffer_cls() :
    wxThread(wxTHREAD_DETACHED) {
}

void ThreadReadBuffer_cls::
SetParameters(bool *pbFinishedFlagLink,
	      std::istream *pifstream, char *pBuf, int size){
    mpbFinishedFlagLink = pbFinishedFlagLink;
    mpifstream = pifstream;
    mpBuf = pBuf;
    mSize = size;

    if ( Create() != wxTHREAD_NO_ERROR) 
	return;
	//MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadReadBuffer_cls::
Entry(){
    mpifstream->read(mpBuf, mSize);
    
    *mpbFinishedFlagLink = TRUE;
    return 0;
}

// -----

#define CLS ParticleSnap_cls

CLS::
ParticleSnap_cls(ParticleSequence_cls *pCParent, int index) :
DataSnap_cls(pCParent, _DATA_PARTICLE, index){
    mbAllocated = FALSE;
    mPNum = 0;
    mpX = NULL;
    mpV = NULL;
    mpID = NULL;
    mpIDLongLong = NULL;
    mpNextIndex = NULL;
    mpPrevIndex = NULL;
    mpSize = NULL;
    mpType = NULL;
    mValidData = FALSE;
    mbNextIndexCalcStarted = FALSE;
    mbNextIndexCalced = FALSE;
    mbNextIndexCalcedThread = FALSE;
    mbPrevIndexCalced = FALSE;
    for (int i = 0; i < _MAX_PVALUE; i++) mpValue[i] = NULL;

    mbUseVelocity = FALSE;
    mbUseType = FALSE;
    mbUseID = FALSE;
    mbUseIDLongLong = FALSE;
    mbInterpolateByID = FALSE;
    mbUseSize = FALSE;
    for (int i = 0; i < _MAX_PVALUE; i++)
	mbUseValue[i] = FALSE;
}

CLS::
~ParticleSnap_cls(){ Clear(); }

void CLS::
Clear(){
    WaitReadingThread(3);

    if (mpX != NULL)         { delete [] mpX; mpX = NULL; }
    if (mpV != NULL)         { delete [] mpV; mpV = NULL; }
    if (mpType != NULL)      { delete [] mpType; mpType = NULL; }
    if (mpSize != NULL)      { delete [] mpSize; mpSize = NULL; }
    for (int i = 0; i < _MAX_PVALUE; i++)
	if (mpValue[i] != NULL) { delete [] mpValue[i]; mpValue[i] = NULL; }
    if (mpID != NULL)        { delete [] mpID; mpID = NULL; };
    if (mpIDLongLong != NULL){ delete [] mpIDLongLong; mpIDLongLong = NULL; }
    if (mpNextIndex != NULL) { delete [] mpNextIndex; mpNextIndex = NULL; }
    if (mpPrevIndex != NULL) { delete [] mpPrevIndex; mpPrevIndex = NULL; }
    
    mbAllocated = FALSE;

    mbReadStarted = FALSE;
    mbReadFinished = FALSE;
}

wxString CLS::
GetStrInfo(int infoIndex){
    wxString strInfo;
    wxFileName CName(mstrLoadedFileHeader);
    strInfo << CName.GetFullName() << ENDL;
    strInfo << "PNum:" << mPNum << ENDL;
    strInfo << "Time:" << mSnapTime << ENDL;
    strInfo << "particle:" << infoIndex << ENDL;
    if (infoIndex >= 0 && infoIndex < mPNum){
	strInfo << "XYZ:" << mpX[infoIndex][0] << ", " << mpX[infoIndex][1] << ", " << mpX[infoIndex][2] << ENDL;
	if (mpV != NULL)
	    strInfo << "VXVYVZ:" << mpV[infoIndex][0] << ", " << mpV[infoIndex][1] << ", " << mpV[infoIndex][2] << ENDL;
	if (mpID != NULL)
	    strInfo << "ID:" << mpID[infoIndex] << ENDL;
	if (mpIDLongLong != NULL)
	    strInfo << "IDLongLong:" << mpIDLongLong[infoIndex] << ENDL;
	if (mpType != NULL)
	    strInfo << "Type:" << (int)mpType[infoIndex] << ENDL;
	if (mpSize != NULL)
	    strInfo << "Size:" << mpSize[infoIndex] << ENDL;
	if (mpNextIndex != NULL)
	    strInfo << "Next:" << mpNextIndex[infoIndex] << ENDL;
	if (mpPrevIndex != NULL)
	    strInfo << "Prev:" << mpPrevIndex[infoIndex] << ENDL;
	for (int i = 0; i < _MAX_PVALUE; i++)
	    if (mpValue[i] != NULL){
		strInfo << "Value[" << i << "]:" << mpValue[i][infoIndex] << ENDL;
	    }
    }
    else
	strInfo << "Out of range" << ENDL;
    return strInfo;
}

int CLS::
GetUsedMemoryByte(){

    if (mbAllocated == FALSE) return 0;

    long long byte = 0;
    if (mpX != NULL)    byte += mPNum * 4 * 3;
    if (mpV != NULL)    byte += mPNum * 4 * 3;
    if (mpSize != NULL) byte += mPNum * 4;
    if (mpType != NULL) byte += mPNum;
    if (mpID != NULL)   byte += mPNum * 4;
    if (mpIDLongLong != NULL) byte += mPNum * 8;
    if (mpNextIndex != NULL) byte += mPNum * 4;
    if (mpPrevIndex != NULL) byte += mPNum * 4;
	
    for (int i = 0; i < _MAX_PVALUE; i++)
	if (mpValue[i] != NULL)  byte += mPNum * 4;
    
    return byte; 
}

void CLS::
AllocateIfNot(int PNum){
    if (mPNum != PNum) Clear();

    if (mbAllocated == TRUE) return;//Clear();

    mPNum = PNum;
    if (mPNum == 0) return;

    mpX = new float [PNum][3];
    for (int i = 0; i < PNum; i++){
	mpX[i][X] = 0;
	mpX[i][Y] = 0;
	mpX[i][Z] = 0;
    }

    if (mbUseVelocity == TRUE){
	mpV = new float [mPNum][3];
	for (int i = 0; i < PNum; i++){
	    mpV[i][X] = 0;
	    mpV[i][Y] = 0;
	    mpV[i][Z] = 0;
	}
    }
	
    if (mbUseType == TRUE){
	mpType = new unsigned char [PNum];
	for (int i = 0; i < PNum; i++) mpType[i] = 0;
    }

    if (mbUseID == TRUE){
	mpID = new int [PNum];
	for (int i = 0; i < PNum; i++) mpID[i] = i;
    }

    if (mbUseIDLongLong == TRUE){
	mpIDLongLong = new long long [PNum];
	for (int i = 0; i < PNum; i++) mpIDLongLong[i] = i;
    }

    if (mbUseSize == TRUE){
	mpSize = new float [PNum];
	for (int i = 0; i < PNum; i++) mpSize[i] = 1;
    }
    
    for (int pvidx = 0; pvidx < _MAX_PVALUE; pvidx++){
	if (mbUseValue[pvidx] == TRUE){
	    mpValue[pvidx] = new float [PNum];
	    for (int i = 0; i < PNum; i++){
		float *pValue = mpValue[pvidx];
		pValue[i] = 0;
	    }
	}
    }
    
    mbAllocated = TRUE;
}

void CLS::
CheckAllocateNextIndex(bool bInterpolateById){
    mbInterpolateByID = bInterpolateById;
    if (mbInterpolateByID == TRUE && mpNextIndex == NULL){
	mpNextIndex = new int [mPNum];
	mpPrevIndex = new int [mPNum];
	for (int i = 0; i < mPNum; i++) mpNextIndex[i] = i;
	for (int i = 0; i < mPNum; i++) mpPrevIndex[i] = -1;
	mbNextIndexCalcStarted = FALSE;
	mbNextIndexCalced = FALSE;
	mbNextIndexCalcedThread = FALSE;
	mbPrevIndexCalced = FALSE;
    }

    if (mbInterpolateByID == FALSE && mpNextIndex != NULL){
	delete [] mpNextIndex;
	delete [] mpPrevIndex;
	mpNextIndex = NULL;
	mpPrevIndex = NULL;
	mbNextIndexCalcStarted = FALSE;
	mbNextIndexCalced = FALSE;
	mbNextIndexCalcedThread = FALSE;
	mbPrevIndexCalced = FALSE;
    }
}

void CLS::
SetStrFile(wxString strFile){

    mstrFile = strFile;//Clone(); 

    if (wxFileExists(mstrFile) == FALSE){
	wxFileName CName(mstrFile);
	if (CName.GetExt().IsSameAs("gz")){
	    wxString strCandit = mstrFile;
	    strCandit.RemoveLast(3);
	    if (wxFileExists(strCandit) == TRUE)
		mstrFile = strCandit;
	}
	else{
	    wxString strCandit = mstrFile;
	    strCandit << ".gz";
	    if (wxFileExists(strCandit) == TRUE)
		mstrFile = strCandit;
	}
    }
}

void CLS::
SetFrame(int frame){
    mFrame = frame;
}

void CLS::
ReadHeaderBinarySTREAM(ParticleFormatClone_cls *pCFormatClone,
		       std::istream &istream,
		       int &PNum, double &snapTime){
    
    std::streampos bodyPos = istream.tellg();
    istream.seekg (0, std::ios::end);
    std::streampos endPos = istream.tellg();
    long long length = endPos - bodyPos;
    int headerSize = pCFormatClone->GetHeaderByte();
    int bodyParticleSize = pCFormatClone->GetBodyParticleByte();
    PNum = (length - headerSize)/ bodyParticleSize;

    istream.seekg (0, std::ios::beg);
    std::vector < int > VFormatHeader = pCFormatClone->GetVFormatHeader();
    
    bool bTime = FALSE;
    int num = -1;
    float time = 0;
    int itemNumber = VFormatHeader.size();
    for (int i = 0; i < itemNumber; i++){
	switch(VFormatHeader[i]){
	case _PFORMAT_HEADER_NUMBER:
	    istream.read((char*)&num, sizeof(int)); break;
	case _PFORMAT_HEADER_TIME:
	    bTime = TRUE;
	    istream.read((char*)&time, sizeof(float)); break;
	case _PFORMAT_HEADER_DUMMY:
	    istream.ignore(sizeof(float)); break;
	case _PFORMAT_HEADER_DUMMYx3:
	    istream.ignore(sizeof(float)*3); break;
	case _PFORMAT_HEADER_DUMMYx10:
	    istream.ignore(sizeof(float)*10); break;
	}
    }
    
    if (num != -1 && num < PNum) PNum = num;
    if (bTime == TRUE) snapTime = (double)time;
}

void CLS::
ReadHeaderBinary(ParticleFormatClone_cls *pCFormatClone,
		 int &PNum, double &snapTime){
    wxFileName CName(mstrFile);
    if (CName.GetExt().IsSameAs(wxT("gz"))){
	std::string strBuf;
	UseGZ_cls::CopyToString(&strBuf, mstrFile);
	
	std::istringstream isstream(strBuf);
	
	return ReadHeaderBinarySTREAM(pCFormatClone, isstream, 
				      PNum, snapTime);
    }

    else{
#ifdef WIN32
	std::fstream ifstream(mstrFile.ToStdWstring(),
			      std::ios::in | std::ios::binary);
#else
	std::fstream ifstream(mstrFile.ToAscii(),
			      std::ios::in | std::ios::binary);
#endif
	if (ifstream.fail()){
	    printf("Read Header Fails!!\n");
	    PNum = -1;
	    return;
	}
	
	return ReadHeaderBinarySTREAM(pCFormatClone, ifstream,
				      PNum, snapTime);
    }

    PNum = 0;
    snapTime = 0;//return 0;
}

void CLS::
ReadHeaderAsciiSTREAM(ParticleFormatClone_cls *pCFormatClone,
		      std::istream &istream, 
		      int &PNum, double &snapTime){

    std::string strBuf;

    std::vector < int > VFormatHeader = pCFormatClone->GetVFormatHeader();

    int num = -1;
    int itemNumber = VFormatHeader.size();
    for (int i = 0; i < itemNumber; i++){
	if (! std::getline(istream, strBuf)) goto READEND;
	switch(VFormatHeader[i]){
	case _PFORMAT_HEADER_NUMBER:
	    num = atoi(strBuf.data());  break;
	case _PFORMAT_HEADER_TIME:{
	    snapTime = atof(strBuf.data());  break; }
	case _PFORMAT_HEADER_DUMMY:
	    break;
	case _PFORMAT_HEADER_DUMMYx3:
	    if (! getline(istream, strBuf)) goto READEND;
	    if (! getline(istream, strBuf)) goto READEND;
	    break;
	case _PFORMAT_HEADER_DUMMYx10:
	    for (int i = 0; i < 9; i++)
		if (getline(istream, strBuf)) goto READEND;
	    break;
	}
    }
READEND:

    if (num < 0){
	int lineNum = 0;
	istream.seekg(0, std::ios_base::beg);
	while (std::getline(istream, strBuf))
	    lineNum++;

	int arrayDirection = pCFormatClone->GetPCItem("ArrayDirection=")->GetBodyINT();
	if (arrayDirection == _PFORMAT_ARRAY_HORIZONTAL)
	    num = lineNum - pCFormatClone->GetHeaderLine();
	else{
	    num = (lineNum - pCFormatClone->GetHeaderLine()) 
		/ pCFormatClone->GetBodyParticleLineAsciiVertical();
	}
    }

    PNum = num;
    // return num;
}

void CLS::
ReadHeaderAscii(ParticleFormatClone_cls *pCFormatClone,
		int &PNum, double &snapTime){
    
    wxFileName CName(mstrFile);
    
    if (CName.GetExt().IsSameAs("gz")){
	std::string strBuf;
	UseGZ_cls::CopyToString(&strBuf, mstrFile);
	std::istringstream isstream(strBuf);
	
	return ReadHeaderAsciiSTREAM(pCFormatClone, isstream,
				     PNum, snapTime);
    }
    else{
#ifdef WIN32
	std::fstream ifstream(mstrFile.ToStdWstring(),
			      std::ios::in | std::ios::binary);
#else
	std::fstream ifstream(mstrFile.ToAscii(),
			      std::ios::in | std::ios::binary);
#endif
	if (ifstream.fail()){
	    printf("Read Header Fails!!\n");
	    PNum = 0;
	    //return;
	}
	
	return ReadHeaderAsciiSTREAM(pCFormatClone, ifstream,
				     PNum, snapTime);
    }

    PNum = 0;
    //return 0;
}

bool CLS::
ReadHeader(ParticleFormatClone_cls *pCFormatClone){
    if (mstrLoadedFileHeader.IsSameAs(mstrFile) == TRUE)
	return FALSE;

    int dataType = pCFormatClone->GetPCItem("DataType=")->GetBodyINT();
    int PNum = -1;
    mSnapTime = (double)mSelfIndex;
    if (dataType == _PFORMAT_BINARY) 
	ReadHeaderBinary(pCFormatClone, PNum, mSnapTime);
    else
	ReadHeaderAscii(pCFormatClone, PNum, mSnapTime);
    /*
    int PNum = (dataType == _PFORMAT_BINARY) ?
	ReadHeaderBinary(pCFormatClone) : ReadHeaderAscii(pCFormatClone); 
    */
    
    // Ƃ肠̂̈Su
    // 1疜
    int numMax = GetPCPreference()->mCParticleMax.GetBodyINT() * 1000;
    if (PNum > numMax) PNum = numMax;
        
    mstrLoadedFileHeader = mstrFile;

    mValidData = TRUE;
    if (PNum == -1){
	mValidData = FALSE;
	PNum = 1024 * 8 * 4 /** 4*/;
    }

    if (dataType == _PFORMAT_BINARY)
	pCFormatClone->GetInfoBinary
	    (&mbUseVelocity, &mbUseType, &mbUseSize, 
	     &mbUseID, &mbUseIDLongLong, &mbInterpolateByID,
	     &mbUseValue[0], &mbUseValue[1], &mbUseValue[2]);
    if (dataType == _PFORMAT_ASCII)
	pCFormatClone->GetInfoAscii
	    (&mbUseVelocity, &mbUseType, &mbUseSize, 
	     &mbUseID, &mbUseIDLongLong, &mbInterpolateByID,
	     &mbUseValue[0], &mbUseValue[1], &mbUseValue[2]);
    
    AllocateIfNot(PNum);
    return TRUE;
}

bool CLS::
DefaultData(){

    if (mstrLoadedFileBody.IsSameAs("#DefaultData")) return FALSE;

    // Allocateɂς
    // \eXgp̃_~[f[^ɕKvȂ݂̂̂ŗǂB
    for (int i = 0; i < mPNum; i++)
	LOOP_XYZ
	    mpX[i][xyz] = (float)rand()/RAND_MAX * 2 - 1;

    if (mbUseVelocity == TRUE)
	for (int i = 0; i < mPNum; i++)
	    LOOP_XYZ
		mpV[i][xyz] = (float)rand()/RAND_MAX * 2 - 1;
    
    if (mbUseType == TRUE)
	for (int i = 0; i < mPNum; i++)
	    mpType[i] = (unsigned char)((float)rand()/RAND_MAX * 4);

    if (mbUseSize == TRUE)
	for (int i = 0; i < mPNum; i++){
	    mpSize[i] = (float)rand()/RAND_MAX * 3 + 1;
	}

    // @ςă[hValue^
    for (int j = 0; j < 3; j++){
	if (mbUseValue[j] == TRUE)
	    for (int i = 0; i < mPNum; i++)
		mpValue[j][i] = (float)rand()/RAND_MAX;
	
    }

    mstrLoadedFileBody = "#DefaultData";
    
    return FALSE;
}

bool CLS::
ReadData(ParticleFormatClone_cls *pCFormatClone, bool bParallel){
    //if (mPNum == 0) return TRUE; // EmptyData
    if (mbAllocated == FALSE) return FALSE;
    
    if (mbReadFinished == TRUE)
	if (mstrLoadedFileBody.IsSameAs(mstrFile) == TRUE) return FALSE;
    //if (mstrFile.IsSameAs("None")) return FALSE;
    //if (mstrFile.IsEmpty() == TRUE) return FALSE;

    if (mValidData == FALSE) return DefaultData();

    if (bParallel == FALSE){
	wxFileName CFile(mstrFile);
	GetPCFrameMain()->SetStatusText(CFile.GetFullName() + " Reading");
    }

    mbReadStarted = TRUE;
    mbReadFinished = FALSE;

    int dataType = pCFormatClone->GetPCItem("DataType=")->GetBodyINT();
    if (dataType == _PFORMAT_BINARY){
	ThreadParticleBinaryRead *pCThread = new ThreadParticleBinaryRead();
	pCThread->SetParameters(&mbReadFinished, this, pCFormatClone);
	
	wxString strLog;
	strLog << "Read Stert :: index = " << mSelfIndex << ENDL;
	wxLogMessage(strLog);

	mstrLoadedFileBody = mstrFile;
	if (bParallel == TRUE)
	    pCThread->Run();
	else{
	    pCThread->DummyEntry();
	    delete pCThread;
	    mbReadFinished = TRUE;
	}
    }
    
    if (dataType == _PFORMAT_ASCII){
	ThreadParticleAsciiRead *pCThread = new ThreadParticleAsciiRead();
	pCThread->SetParameters(&mbReadFinished, this, pCFormatClone);

	wxString strLog;
	strLog << "Read Stert :: index = " << mSelfIndex << ENDL;
	wxLogMessage(strLog);
	
	mstrLoadedFileBody = mstrFile;
	if (bParallel == TRUE)
	    pCThread->Run();
	else{
	    pCThread->DummyEntry();
	    delete pCThread;
	    mbReadFinished = TRUE;
	}
    }

    return TRUE;
}

bool CLS::
PrefetchData(ParticleFormatClone_cls *pCFormatClone){
    if (ReadHeader(pCFormatClone) == FALSE) return FALSE;
    if (ReadData(pCFormatClone, TRUE) == FALSE) return FALSE;
    
    return TRUE;
}

void CLS::
ExportSnap(wxString strFile, int direction){
    FILE *fp = fopen(strFile.char_str(), "wb");

    if (direction == 0){
	for (int i = 0; i < mPNum; i++){
	    if (mpX != NULL) fwrite(mpX[i], sizeof(float), 3, fp);
	    if (mpV != NULL) fwrite(mpV[i], sizeof(float), 3, fp);
	    if (mpID != NULL) fwrite(&mpID[i], sizeof(float), 1, fp);
	    if (mpIDLongLong != NULL) 
		fwrite(&mpIDLongLong[i], sizeof(long long), 1, fp);
	    if (mpType != NULL){
		int tmp = (int)mpType[i];
		fwrite(&tmp, sizeof(int), 1, fp);
	    }
	    if (mpSize != NULL)
		fwrite(&mpSize[i], sizeof(float), 1, fp);
	    for (int j = 0; j < _MAX_PVALUE; j++)
		if (mpValue[j] != NULL)
		    fwrite(&mpValue[j][i], sizeof(float), mPNum, fp);
	}
    }

    if (direction == 1){
	if (mpX != NULL) fwrite(mpX, sizeof(float), mPNum*3, fp);
	if (mpV != NULL) fwrite(mpV, sizeof(float), mPNum*3, fp);
	if (mpID != NULL) fwrite(mpID, sizeof(int), mPNum, fp);
	if (mpIDLongLong != NULL)
	    fwrite(mpIDLongLong, sizeof(long long), mPNum, fp);
	if (mpType != NULL){
	    int *pTmp = new int [mPNum];
	    for (int i = 0; i < mPNum; i++) pTmp[i] = (int)mpType[i];
	    fwrite(pTmp, sizeof(int), mPNum, fp);
	    delete [] pTmp;
	}
	if (mpSize != NULL) fwrite(mpSize, sizeof(float), mPNum, fp);
	for (int i = 0; i < _MAX_PVALUE; i++)
	    if (mpValue[i] != NULL)
		fwrite(mpValue[i], sizeof(float), mPNum, fp);
    }
	
    fclose(fp);
}

wxString CLS::
GetStrExportFormat(int direction){
    wxString strFormat;

    if (direction == 0){
	if (mpX != NULL) strFormat << "x y z ";
	if (mpV != NULL) strFormat << "vx vy vz ";
	if (mpID != NULL) strFormat << "ID(int) ";
	if (mpIDLongLong != NULL) strFormat << "ID(long long) "; 
	if (mpType != NULL) strFormat << "type(int) ";
	if (mpSize != NULL) strFormat << "size ";
	for (int i = 0; i < _MAX_PVALUE; i++)
	    if (mpValue[i] != NULL)
		strFormat << "value" << i << " ";
	strFormat << ENDL;
	strFormat << strFormat;
	strFormat << "..." << ENDL;
    }

    if (direction == 1){
	if (mpX != NULL)
	    strFormat << "xyz : float[" << mPNum << "][3]" << ENDL;
	if (mpV != NULL)
	    strFormat << "vxvyvz : float[" << mPNum << "][3]" << ENDL;
	if (mpID != NULL)
	    strFormat << "ID : int[" << mPNum << "]" << ENDL;
	if (mpIDLongLong != NULL)
	    strFormat << "ID(long long) : long long[" << mPNum << "]" << ENDL;
	if (mpType != NULL)
	    strFormat << "type : int[" << mPNum << "]" << ENDL;
	if (mpSize != NULL)
	    strFormat << "size : float[" << mPNum << "]" << ENDL;
	for (int i = 0; i < _MAX_PVALUE; i++)
	    if (mpValue[i] != NULL)
		strFormat << "value" << i << " : float[" << mPNum << "]" << ENDL;
    }
    
    return strFormat;
}
