// (setq-c-offset 4)
// "clsStreamlineForRender.h"

#include "clsStreamlineForRender.h"
#include "clsStreamlineSnap.h"

#include "INTERPTHREAD/clsStreamlineForRenderInterpThread.h"

#include "../MODIFIER/clsStreamFiltering.h"
#include "../MODIFIER/clsStreamSetAlpha.h"
#include "../MODIFIER/clsStreamTransform.h"
#include "../SCENEOBJ/DATAVIEWER/clsStreamline.h"
#include "../SCENEOBJ/DATAVIEWER/clsViewerClone.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"
#include "../COMMON/CommonDefinitionsShader.h"

#define CLS StreamlineForRender_cls

CLS::
StreamlineForRender_cls(){
    mFrame = -1;
    mSNum = -1;
    mpPos = NULL;
    mpU = NULL;
    mpValue = NULL;
    mpLineIndex = NULL;
    mpFilterAlpha = NULL;
}

CLS::
~StreamlineForRender_cls(){ 
    Clear(); 
    //if (mpTextureData != NULL) { delete [] mpTextureData; mpTextureData = NULL;}
}

void CLS::
Clear(){
    if (mpPos != NULL){
	delete [] mpPos; mpPos = NULL; }
    if (mpU != NULL){
	delete [] mpU; mpU = NULL; }
    if (mpValue != NULL){
	delete [] mpValue; mpValue = NULL; }
    if (mpLineIndex != NULL){
	delete [] mpLineIndex; mpLineIndex = NULL; }
    if (mpFilterAlpha != NULL){
	delete [] mpFilterAlpha; mpFilterAlpha = NULL; }
    mSNum = 0;
}

bool CLS::
CheckUpdate(int frame, float factor, wxString strStatusLog){
    bool bUpdate = TRUE;
    if (mFrame == frame && mFactor == factor &&
	mstrStatusLog.IsSameAs(strStatusLog) ) bUpdate = FALSE;
    
    mFrame = frame;
    mFactor = factor;
    mstrStatusLog = strStatusLog;
    return bUpdate;
}

void CLS::
PrepareCommonRenderParameter(ViewerClone_cls* pCStreamlineClone){
    /*
    ConfItemList_cls &CConfList = pCStreamlineClone->mCConfList;
    mstrStreamlineTexture[0] = CConfList.GetPCItem("TexturePath0=")->GetBodySTR();
    mstrStreamlineTexture[1] = CConfList.GetPCItem("TexturePath1=")->GetBodySTR();
    mstrStreamlineTexture[2] = CConfList.GetPCItem("TexturePath2=")->GetBodySTR();
    int textureNum = CConfList.GetPCItem("TextureNum=")->GetBodyINT();
    for (int i = textureNum; i < 3; i++)
	mstrStreamlineTexture[i] = "None";

    bool bDependDistance = CConfList.GetPCItem("DependDistance=")->GetBodyBOOL();
    if (bDependDistance == TRUE)
	mDistanceBase = CConfList.GetPCItem("DistanceBase=")->GetBodyDOUBLE();
    else
	mDistanceBase = 0;
    */
}

void CLS::
Allocate(long long SNum, bool bUseFilterAlpha){
    if (mpPos != NULL) Clear();

    mSNum = SNum;

    mpPos = new float [mSNum][3];
    mpU = new float [mSNum];
    mpValue = new float [mSNum];
    mpLineIndex = new int [mSNum];
    if (bUseFilterAlpha == TRUE) mpFilterAlpha = new float [mSNum];
}

void CLS::
PrepareData(StreamlineSnap_cls *pCSnap1, StreamlineSnap_cls *pCSnap2,
	    int frame, int subFrame, int seq, double factor, 
	    ViewerClone_cls *pCStreamlineClone,
	    double NT){

    uint numAttribute = pCStreamlineClone->mVpCAttribClone.size();

    wxString strStatusLog;

    strStatusLog << pCSnap1->mstrLoadedFileHeader << "::";

    for (uint i = 0; i < numAttribute; i++){
	AttributeClone_cls *pCAttribute = pCStreamlineClone->mVpCAttribClone[i];
	wxString str = pCAttribute->GetStrAttributeName();
	if (pCAttribute->IsAttributeSameAs("StreamFiltering_cls"))
	    strStatusLog << pCAttribute->GetStrProcCounter();
	if (pCAttribute->IsAttributeSameAs("StreamSetAlpha_cls"))
	    strStatusLog << pCAttribute->GetStrProcCounter();
	if (pCAttribute->IsAttributeSameAs("StreamTransform_cls"))
	    strStatusLog << pCAttribute->GetStrProcCounter();
    }

    PrepareCommonRenderParameter(pCStreamlineClone);

    bool bTimeDepend = FALSE;

    if (bTimeDepend == TRUE) //!!
	strStatusLog << "::" << NT; 

    bool bUpdateFiltering = FALSE;
    bool bUpdateSetAlpha = FALSE;
    bool bUpdateTransform = FALSE;

    if (CheckUpdate(frame, factor, strStatusLog) == TRUE){

	bUpdateFiltering = TRUE;
	bUpdateSetAlpha = TRUE;
	bUpdateTransform = TRUE;

	bool bUseFilterAlpha = TRUE;

	Allocate(pCSnap1->mSNum, bUseFilterAlpha);
	
	if (mSNum == 0){
	    delete pCStreamlineClone; return; 
	}

	int threadNumber;
	bool *bAFinished = GetNewPCbFinishList(threadNumber, mSNum);

	ThreadStreamlineInterp_cls  **ppCThreadStreamlineInterp = 
	    new ThreadStreamlineInterp_cls* [threadNumber];

	for (int  core = 0; core < threadNumber; core++){
	    ppCThreadStreamlineInterp[core] = 
		new ThreadStreamlineInterp_cls;
	    
	    int SIni, SFin;
	    GetIniAndFin(SIni, SFin, core, threadNumber, mSNum);

	    ppCThreadStreamlineInterp[core]->SetParameters
		(pCSnap1, pCSnap2, factor, mpPos, mpU, mpValue, mpLineIndex);
	    /*
	    ppCThreadStreamlineInterp[core]->SetParametersTime
		(NT, frame, subFrame, seq);
	    */
	    ppCThreadStreamlineInterp[core]->CreateThread
		(&bAFinished[core], SIni, SFin);
	    ppCThreadStreamlineInterp[core]->Run();
	}

	WaitThreadsAndDeleteFlag(bAFinished, threadNumber, 3);
	delete [] ppCThreadStreamlineInterp;

	pCSnap1->MarkDraw();
	mDrawCounter++;

	//ProcDataModifier(pCStreamlineClone);
    }

    if (bUpdateFiltering == TRUE || bUpdateSetAlpha == TRUE)
	for (int i = 0; i < mSNum; i++) mpFilterAlpha[i] = 1;

    if (bUpdateFiltering == TRUE)
	ProcFiltering(NT, pCStreamlineClone);
    if (bUpdateSetAlpha == TRUE)
	ProcSetAlpha(NT, pCStreamlineClone);
    if (bUpdateTransform == TRUE)
	ProcTransform(NT, pCStreamlineClone);

    delete pCStreamlineClone;
}

void CLS::
ProcFiltering(double NT, ViewerClone_cls *pCStreamline){

    //for (int i = 0; i < mSNum; i++) mpFilterAlpha[i] = 1;

    uint numFiltering = pCStreamline->GetNumAttribute("StreamFiltering_cls");
    for (uint i = 0; i < numFiltering; i++){
	AttributeClone_cls* pCFiltering = pCStreamline->GetPCAttribCloneFromAttribName("StreamFiltering_cls", i);
	StreamFilteringExec_cls *pCExe = new StreamFilteringExec_cls(pCFiltering);
	pCExe->CheckFiltering(mpFilterAlpha, mpLineIndex, mSNum);

	delete pCExe;
    }
}

void CLS::
ProcSetAlpha(double NT, ViewerClone_cls *pCStreamline){
    
    uint numSetAlpha = pCStreamline->GetNumAttribute("StreamSetAlpha_cls");
    for (uint i = 0; i < numSetAlpha; i++){
	AttributeClone_cls* pCSetAlpha = pCStreamline->GetPCAttribCloneFromAttribName("StreamSetAlpha_cls", i);
	StreamSetAlphaExec_cls *pCExe = new StreamSetAlphaExec_cls(pCSetAlpha);
	pCExe->SetAlpha(mpFilterAlpha, mpValue, mSNum);
	
	delete pCExe;
    }
}

void CLS::
ProcTransform(double NT, ViewerClone_cls *pCStreamline){
    
    uint numFiltering = pCStreamline->GetNumAttribute("StreamTransform_cls");
    for (uint i = 0; i < numFiltering; i++){
	AttributeClone_cls* pCTransform = pCStreamline->GetPCAttribCloneFromAttribName("StreamTransform_cls", i);
	StreamTransformExec_cls *pCExe = new StreamTransformExec_cls(pCTransform);
	pCExe->Transform(mpPos, mSNum);
	
	delete pCExe;
    }
}
