// -*- c -*- 
// "clsStreamlineSequence.h"

#include "clsStreamlineSequence.h"

#include "clsDataSnap.h"
#include "clsStreamlineSnap.h"

#include "../MODIFIER/clsSequenceRemap.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../PREFERENCE/clsPreference.h"
#include "../TIMELINE/clsTimeCtrl.h"

#define CLS StreamlineSequence_cls

CLS::
StreamlineSequence_cls() : DataSequence_cls(_DATA_STREAMLINE){
    mbTimeDepend = FALSE;
}

CLS::
~StreamlineSequence_cls(){
    if (mApCSnap != NULL){
	if (mSeqLen == -1){
	    for (int i = 0; i < mSeqLenT; i++){
		if (mApCSnap[i] != NULL){
		    delete (StreamlineSnap_cls*)mApCSnap[i];
		    mApCSnap[i] = NULL;
		}
	    }
	}
	for (int i = 0; i < mSeqLen; i++){
	    if (mApCSnap[i] != NULL){
		delete (StreamlineSnap_cls*)mApCSnap[i];
		mApCSnap[i] = NULL;
	    }
	}
	delete mApCSnap;
	mApCSnap = NULL;
    }
}

int CLS::
GetTotalUsedMemoryMB(){
    long long memoryByte = 0;
    if (mApCSnap == NULL) return 0;
    for (int i = 0; i < mSeqLen; i++){
	if (mApCSnap[i] == NULL) continue;
	memoryByte += ((StreamlineSnap_cls*)mApCSnap[i])->GetUsedMemoryByte();
    }
    for (int i = 0; i < mSeqLenT; i++){
	if (mApCSnap[i] == NULL) continue;
	memoryByte += ((StreamlineSnap_cls*)mApCSnap[i])->GetUsedMemoryByte();
    }
    
    mTotalUsedMemoryMB = (int)(memoryByte / 1024 / 1024);
    
    return mTotalUsedMemoryMB;
}

StreamlineSnap_cls* CLS::
GenerateIfNot(int index){

    int seqLen = mSeqLen;
    if (mSeqLenT >= 0) seqLen = mSeqLenT;

    if (index >= seqLen) index = seqLen - 1;
    if (index < 0) index = 0;
    if (mbTimeDepend == FALSE) index = 0;
    
    if (mApCSnap == NULL) return NULL;

    if (mApCSnap[index] == NULL){
	mApCSnap[index] = new StreamlineSnap_cls(this, index);
    }

    return (StreamlineSnap_cls*)mApCSnap[index];
}

StreamlineSnap_cls* CLS::
GetPCSnap(int index){
    if (mApCSnap == NULL) return NULL;

    int seqLen = mSeqLen;
    if (mSeqLenT >= 1) seqLen = mSeqLenT;

    if (index >= seqLen) index = seqLen - 1;
    if (index < 0) index = 0;
    if (mbTimeDepend == FALSE) index = 0;
    
    return (StreamlineSnap_cls*)mApCSnap[index];
}

bool CLS::
PrefetchData(StreamlineFormatClone_cls *pCFormatClone,
	     SequenceRemap_cls *pCSeqRemap){
    if (mReadingLock > 0) return FALSE; // 1Xbĥ

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    double dNT = GetPCTimeCtrl()->GetDNT();
    int frameBase, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frameBase, subframe);

    int seqLen = mSeqLen;
    
    int prefetch = GetPCPreference()->mCPrefetchLength.GetBodyINT();

    for (int i = 0; i < prefetch; i++){
	int frame1 = frameBase + i;
	if (pCSeqRemap != NULL)
	    frame1 = pCSeqRemap->GetRemappedFrame(NT+dNT*i, frame1, seqLen);
	    
	if (frame1 < 0 || frame1 >= seqLen) continue;

	GenerateIfNot(frame1);
	if (mbTimeDepend == TRUE)
	    GetPCSnap(frame1)->SetStrFile(GetModifiedStr(frame1));
	else
	    GetPCSnap(frame1)->SetStrFile(mstrPath);
	
	if (GetPCSnap(frame1)->PrefetchData(pCFormatClone) == TRUE) 
	    return TRUE;

	frame1++; // ԗpɎXebvvtFb`
	if (frame1 >= seqLen) continue;
	GenerateIfNot(frame1);
	GetPCSnap(frame1)->SetStrFile(GetModifiedStr(frame1));
	if (GetPCSnap(frame1)->PrefetchData(pCFormatClone) == TRUE) 
	    return TRUE;
    }

    return FALSE;
}

void CLS::
ReleaseMemoryData(SequenceRemap_cls *pCSeqRemap){
    
    ParamGroup_cls *pCRemapParamGroup = NULL;
    bool bRepeat = FALSE;
    if (pCSeqRemap != NULL){
	pCRemapParamGroup = pCSeqRemap->GetPCParamGroup("Remap");
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }
    MarkPriority(pCRemapParamGroup, bRepeat);
    int oldestIndex = GetOldestDrawnIndex();
    if (oldestIndex == -1) return;

    StreamlineSnap_cls *pCSnap = GetPCSnap(oldestIndex);
    if (pCSnap == NULL)
	return;
    
    delete pCSnap;
    mApCSnap[oldestIndex] = NULL;
}

