// -*- c -*- 
// "clsStreamlineSnap.h"

#include "clsStreamlineSnap.h"

#include "clsStreamlineFormat.h"
#include "clsStreamlineSequence.h"

#include "READTHREAD/clsStreamlineAsciiRead.h"
//#include "READTHREAD/clsStreamlineBinaryRead.h"

#include "../COMMON/clsUseGZ.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/MyEndian.h"

#include "stdio.h"
#include <fstream>
#include <sstream>

#include <wx/filename.h>

//wxString GstrVolumeSnap;

#define CLS StreamlineSnap_cls

CLS::
StreamlineSnap_cls(StreamlineSequence_cls *pCParent, int index) : 
DataSnap_cls(pCParent, _DATA_STREAMLINE, index){
    mSNum = 0;
    mMaxIndex = -1;
    mAllocatedSize = 0;
    
    mpPos = NULL;
    mpU = NULL;
    mpValue = NULL;
    mpLineIndex = NULL;

    mValidData = FALSE;
    
    Clear();
}

CLS::
~StreamlineSnap_cls(){ Clear(); }

void CLS::
Clear(){
    WaitReadingThread(3);

    if (mpPos != NULL){
	delete [] mpPos; mpPos = NULL; }
    if (mpU != NULL){
	delete [] mpU; mpU = NULL; }
    if (mpValue != NULL){
	delete [] mpValue; mpValue = NULL; }
    if (mpLineIndex != NULL){
	delete [] mpLineIndex; mpLineIndex = NULL; }
    
    mbReadStarted = FALSE;
    mbReadFinished = FALSE;
}

wxString CLS::
GetStrInfo(){
    wxString strInfo;
    wxFileName CName(mstrLoadedFileHeader);
    strInfo << CName.GetFullName() << ENDL;
    strInfo << "vertexNum = " << mSNum << ENDL;
    strInfo << "maxIndex = " << mMaxIndex << ENDL;
    return strInfo;
}

int CLS::
GetUsedMemoryByte(){
    if (mpPos == NULL) return 0;
    return mSNum * (12 + 4 + 4);
}

bool CLS::
DefaultData(){
    if (mstrLoadedFileBody.IsSameAs("None")) return FALSE;

    for (int i = 0; i < mSNum; i++){
	mpPos[i][0] = sin(i*0.1);
	mpPos[i][1] = cos(i*0.1);
	mpPos[i][2] = i*0.1*0.25;
	mpU[i] = ((float)i)/10;//;
	mpValue[i] = ((float)i)/10;
	mpLineIndex[i] = i / 10;
	//if (i % 10 == 0) mpLineIndex[i] = -1;
    }

    mstrLoadedFileBody = "None";
    return FALSE;
}

void CLS::
AllocateIfNot(int SNum){

    mSNum = SNum;

    if (mSNum != mAllocatedSize) Clear();

    if (mpPos != NULL) return; 

    mpPos = new float [mSNum][3];
    mpU = new float [mSNum];
    mpValue = new float [mSNum];
    mpLineIndex = new int [mSNum];

    mAllocatedSize = mSNum;
    mMaxIndex = -1;
}

void CLS::
SetStrFile(wxString strFile){

    mstrFile = strFile;//Clone(); 

    if (wxFileExists(mstrFile) == FALSE){
	wxFileName CName(mstrFile);
	if (CName.GetExt().IsSameAs("gz")){
	    wxString strCandit = mstrFile;
	    strCandit.RemoveLast(3);
	    if (wxFileExists(strCandit) == TRUE)
		mstrFile = strCandit;
	}
	else{
	    wxString strCandit = mstrFile;
	    strCandit << ".gz";
	    if (wxFileExists(strCandit) == TRUE)
		mstrFile = strCandit;
	}
    }
}

void CLS::
SetFrame(int frame){
    mFrame = frame;
}

int CLS::
ReadHeaderBinary(StreamlineFormatClone_cls *pCFormatClone){
    return 0;
}

int CLS::
ReadHeaderAsciiSTREAM(StreamlineFormatClone_cls *pCFormatClone,
		      std::istream &istream){

    std::string strBuf;
    
    int num = -1;

    // Read Predefined Info

    if (num < 0){
	int lineNum = 0;
	istream.seekg(0, std::ios_base::beg);
	while (std::getline(istream, strBuf))
	    lineNum++;

	int arrayDirection = _SFORMAT_ARRAY_HORIZONTAL;//pCFormatClone->GetPCItem("ArrayDirection=")->GetBodyINT();
	if (arrayDirection == _SFORMAT_ARRAY_HORIZONTAL)
	    num = lineNum - 0/*pCFormatClone->GetHeaderLine()*/;
	else{
	    num = (lineNum - 0/*pCFormatClone->GetHeaderLine()*/) 
		/ 3/*pCFormatClone->GetBodyParticleLineAsciiVertical()*/;
	}
    }
    return num;
}

int CLS::
ReadHeaderAscii(StreamlineFormatClone_cls *pCFormatClone){
    wxFileName CName(mstrFile);
    
    if (CName.GetExt().IsSameAs("gz")){
	std::string strBuf;
	UseGZ_cls::CopyToString(&strBuf, mstrFile);
	std::istringstream isstream(strBuf);
	
	return ReadHeaderAsciiSTREAM(pCFormatClone, isstream);
    }
    else{
#ifdef WIN32
	std::fstream ifstream(mstrFile.ToStdWstring(),
			      std::ios::in | std::ios::binary);
#else
	std::fstream ifstream(mstrFile.ToAscii(),
			      std::ios::in | std::ios::binary);
#endif
	if (ifstream.fail()){
	    printf("Read Header Fails!!\n");
	    return 0;//exit(0);
	}

	return ReadHeaderAsciiSTREAM(pCFormatClone, ifstream);

    }

    return 0;
}

bool CLS::
ReadHeader(StreamlineFormatClone_cls *pCFormatClone){
    if (mstrLoadedFileHeader.IsSameAs(mstrFile) == TRUE) return FALSE;

    int dataType = _SFORMAT_ASCII;
    int SNum = (dataType == _SFORMAT_BINARY) ?
	ReadHeaderBinary(pCFormatClone) : ReadHeaderAscii(pCFormatClone);

    FILE *pFile = NULL;

    mstrLoadedFileHeader = mstrFile;

    mValidData = TRUE;
    if (SNum == 0){
	mValidData = FALSE;
	SNum = 128;
    }

    AllocateIfNot(SNum);
    return TRUE;
}

bool CLS::
ReadData(StreamlineFormatClone_cls *pCFormatClone, bool bParallel){

    if (mpPos == NULL) return FALSE;

    if (mbReadFinished == TRUE)
	if (mstrLoadedFileBody.IsSameAs(mstrFile) == TRUE) return FALSE;

    if (mValidData == FALSE) return DefaultData();

    mbReadStarted = TRUE;
    mbReadFinished = FALSE;

    int dataType = _SFORMAT_ASCII;

    if (dataType == _SFORMAT_ASCII){

	ThreadStreamlineAsciiRead *pCThread = new ThreadStreamlineAsciiRead();
	pCThread->SetParameters(&mbReadFinished, this, pCFormatClone);
	
	mstrLoadedFileBody = mstrFile;
	if (bParallel == TRUE)
	    pCThread->Run();
	else{
	    pCThread->DummyEntry();
	    delete pCThread;
	    mbReadFinished = TRUE;
	}
    }

    for (int i = 0; i < mSNum; i++)
	if (mpLineIndex[i] > mMaxIndex)
	    mMaxIndex = mpLineIndex[i];
    
    return TRUE;
}

bool CLS::
PrefetchData(StreamlineFormatClone_cls *pCFormatClone){

    if (ReadHeader(pCFormatClone) == FALSE) return FALSE;
    if (ReadData(pCFormatClone, TRUE) == FALSE) return FALSE;
    
    return TRUE;
}
