// (setq-c-offset 4)
// "clsVolumeForRender.h"

#include "clsVolumeForRender.h"
#include "clsVolumeSnap.h"

#include "../MISC/clsMisc.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/DATAVIEWER/clsVolumeBox.h"
#include "../SCENEOBJ/DATAVIEWER/clsViewerClone.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../MODIFIER/clsVolumeBoolean.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

//#include <wx/stopwatch.h>

#ifdef WIN32
extern PFNGLTEXIMAGE3DPROC glTexImage3D;
#endif

#define CLS ThreadCallBoolean_cls

void CLS::
SetParameters(ViewerClone_cls *pCVolumeBoxClone, char *pMask,
	      int sizeX, int sizeY, int sizeZ, double NT){
    mpCVolumeBoxClone = pCVolumeBoxClone;
    mpMask = pMask;
    mSizeX = sizeX; mSizeY = sizeY; mSizeZ = sizeZ;
    mNT = NT;
}

bool CLS::
CreateThread(bool *pbFinishedFlag){
    mpbFinishedFlagLink = pbFinishedFlag;

    if ( Create() != wxTHREAD_NO_ERROR) 
	return FALSE;
    return TRUE;
}

wxThread::ExitCode CLS::
Entry(){

    uint numAttribClone = mpCVolumeBoxClone->mVpCAttribClone.size();
    for (uint i = 0; i < numAttribClone; i++){
	AttributeClone_cls *pCMod = mpCVolumeBoxClone->mVpCAttribClone[i];
	if (pCMod->IsAttributeValidAndSameAs("VolumeBoolean_cls")){
	    int booleanShape = pCMod->GetPCItem("BooleanShape=")->GetBodyINT();
	    if (booleanShape == _BOOL_BOX)
		VolumeBoolean_cls::SetBooleanMaskBox
		    (*pCMod, mpMask, mSizeX, mSizeY, mSizeZ, mNT);
	    if (booleanShape == _BOOL_SPHERE)
		VolumeBoolean_cls::SetBooleanMaskSphere
		    (*pCMod, mpMask, mSizeX, mSizeY, mSizeZ, mNT);
	    if (booleanShape == _BOOL_OBJECT){
		double matrix16TgtInv[16];
		BooleanBase_cls::GetMatrix16TgtInv(pCMod, matrix16TgtInv, mNT);
		//pCMod->GetMatrix16TgtInv(matrix16TgtInv, mNT);
		
		VolumeBoolean_cls::SetBooleanMaskObject
		    (*pCMod, mpMask, mSizeX, mSizeY, mSizeZ,
		     mpCVolumeBoxClone->mMatrix16, matrix16TgtInv);
	    }
	}   
    }

    *mpbFinishedFlagLink = TRUE;
    return 0;
}

#define CLS VolumeFoRender_cls

CLS::
VolumeFoRender_cls(){
    mFrame = -1;
    mFactor = -1;
    //mSubFrame = -1;
    mpData_uchar = NULL;
    mpData_ushort = NULL;
    mpMask = NULL;
    mSize[X] = mSize[Y] = mSize[Z] = 0;
    mSizeAllocatedData = 0;
    mSizeAllocatedMask = 0;
    mTexture3D = 0;
    mbDataBakedToTexture = FALSE;
    mbFinishedFlagBoolean = TRUE;
    mPrepareCounter = 0;
}

CLS::
~VolumeFoRender_cls(){ 
    Clear(); 
    ClearTexture();
}

void CLS::
Clear(){
    ClearData();
    ClearMask();

    mSize[X] = mSize[Y] = mSize[Z] = 0;
    
    mFrame = -1;
    mFactor = -1;
}

void CLS::
ClearData(){
    if (mpData_uchar != NULL){
	delete [] mpData_uchar; mpData_uchar = NULL;
    }
    if (mpData_ushort != NULL){
	delete [] mpData_ushort; mpData_ushort = NULL;
    }
    mSizeAllocatedData = 0;
}

void CLS::
ClearMask(){
    if (mpMask != NULL){
	delete [] mpMask; mpMask = NULL;
	mSizeAllocatedMask = 0;
    }
}

void CLS::
ClearTexture(){
    if (mTexture3D != 0){
	glDeleteTextures(1, &mTexture3D);
	mTexture3D = 0;
    }
}

bool CLS::
IsSameStep(int frame, double factor){
    if (mFrame == frame && Misc_cls::IsNearlySame((double)mFactor, factor, 1e-5))
	return TRUE;
    return FALSE;
}

void CLS::
CopyDataFromSnap_EqualSize(VolumeSnap_cls *pCSnap1, VolumeSnap_cls *pCSnap2, double factor, int size){

    if (pCSnap1->mbUseUShort == TRUE){
	if (factor == 0 || pCSnap2 == NULL){
	    memcpy(mpData_ushort, pCSnap1->mpData_ushort, size*2);
	    pCSnap1->MarkDraw();
	}
	else{ 
	    unsigned short * const pData1 = pCSnap1->mpData_ushort;
	    unsigned short * const pData2 = pCSnap2->mpData_ushort;
	    const double factorInv = 1.0 - factor;
	    for (int z = 0; z < mSize[Z]; z++){
		const int indexz = z * mSize[X] * mSize[Y];
		for (int y = 0; y < mSize[Y]; y++){
		    const int indexy = indexz + y * mSize[X];
		    for (int x = 0; x < mSize[X]; x++){
			const int index = indexy + x;
			mpData_ushort[index] = (unsigned short)
			    ((double)pData1[index] * factorInv +
			     (double)pData2[index] * factor);
		    }
		}
	    }
	    pCSnap1->MarkDraw();
	    pCSnap2->MarkDraw();
	}
    }
    else{
	if (factor == 0 || pCSnap2 == NULL){
	    memcpy(mpData_uchar, pCSnap1->mpData_uchar, size);
	    pCSnap1->MarkDraw();
	}
	else{ 
	    unsigned char * const pData1 = pCSnap1->mpData_uchar;
	    unsigned char * const pData2 = pCSnap2->mpData_uchar;
	    const double factorInv = 1.0 - factor;
	    for (int z = 0; z < mSize[Z]; z++){
		const int indexz = z * mSize[X] * mSize[Y];
		for (int y = 0; y < mSize[Y]; y++){
		    const int indexy = indexz + y * mSize[X];
		    for (int x = 0; x < mSize[X]; x++){
			const int index = indexy + x;
			mpData_uchar[index] = (unsigned char)
			    ((double)pData1[index] * factorInv +
			     (double)pData2[index] * factor);
		    }
		}
	    }
	    pCSnap1->MarkDraw();
	    pCSnap2->MarkDraw();
	}
    }
}

void CLS::
CopyDataFromSnap_Resized(VolumeSnap_cls *pCSnap1, VolumeSnap_cls *pCSnap2, double factor, int size){
    
    const double factorInv = 1.0 - factor;
    const int snapSizeX = pCSnap1->mSize[X];
    const int snapSizeY = pCSnap1->mSize[Y];
    const int snapSizeZ = pCSnap1->mSize[Z];
    const int factorX = pCSnap1->mSize[X] / mSize[X];
    const int factorY = pCSnap1->mSize[Y] / mSize[Y];
    const int factorZ = pCSnap1->mSize[Z] / mSize[Z];

    if (pCSnap1->mbUseUShort == TRUE){
	unsigned short * const pData1 = pCSnap1->mpData_ushort;
	if (factor == 0 || pCSnap2 == NULL){
	    for (int z = 0; z < mSize[Z]; z++){
		const int indexz = z * mSize[X] * mSize[Y];
		const int indexSnapZ = z * factorZ * snapSizeX * snapSizeY; 
		for (int y = 0; y < mSize[Y]; y++){
		    const int indexy = indexz + y * mSize[X];
		    const int indexSnapY = indexSnapZ + y * factorY * snapSizeX;
		    for (int x = 0; x < mSize[X]; x++){
			const int index = indexy + x;
			const int indexSnap = indexSnapY + x * factorX;
			mpData_ushort[index] = (unsigned short)pData1[indexSnap];
		    }
		}
	    }
	}
	else{
	    unsigned short * const pData2 = pCSnap2->mpData_ushort;
	    for (int z = 0; z < mSize[Z]; z++){
		const int indexz = z * mSize[X] * mSize[Y];
		const int indexSnapZ = z * factorZ * snapSizeX * snapSizeY; 
		for (int y = 0; y < mSize[Y]; y++){
		    const int indexy = indexz + y * mSize[X];
		    const int indexSnapY = indexSnapZ + y * factorY * snapSizeX;
		    for (int x = 0; x < mSize[X]; x++){
			const int index = indexy + x;
			const int indexSnap = indexSnapY + x * factorX;
			mpData_ushort[index] = (unsigned short)
			    ((double)pData1[indexSnap] * factorInv +
			     (double)pData2[indexSnap] * factor);
		    }
		}
	    }
	}
    }
    else{
	unsigned char * const pData1 = pCSnap1->mpData_uchar;
	if (factor == 0 || pCSnap2 == NULL){
	    for (int z = 0; z < mSize[Z]; z++){
		const int indexz = z * mSize[X] * mSize[Y];
		const int indexSnapZ = z * factorZ * snapSizeX * snapSizeY; 
		for (int y = 0; y < mSize[Y]; y++){
		    const int indexy = indexz + y * mSize[X];
		    const int indexSnapY = indexSnapZ + y * factorY * snapSizeX;
		    for (int x = 0; x < mSize[X]; x++){
			const int index = indexy + x;
			const int indexSnap = indexSnapY + x * factorX;
			mpData_uchar[index] = (unsigned char)pData1[indexSnap];
		    }
		}
	    }
	}
	else{
	    unsigned char * const pData2 = pCSnap2->mpData_uchar;
	    for (int z = 0; z < mSize[Z]; z++){
		const int indexz = z * mSize[X] * mSize[Y];
		const int indexSnapZ = z * factorZ * snapSizeX * snapSizeY; 
		for (int y = 0; y < mSize[Y]; y++){
		    const int indexy = indexz + y * mSize[X];
		    const int indexSnapY = indexSnapZ + y * factorY * snapSizeX;
		    for (int x = 0; x < mSize[X]; x++){
			const int index = indexy + x;
			const int indexSnap = indexSnapY + x * factorX;
			mpData_uchar[index] = (unsigned char)
			    ((double)pData1[indexSnap] * factorInv +
			     (double)pData2[indexSnap] * factor);
		    }
		}
	    }
	}
    }
}

void CLS::
PrepareData(VolumeSnap_cls *pCSnap1, VolumeSnap_cls *pCSnap2,
	    int frame, double factor, 
	    ViewerClone_cls *pCVolumeBoxClone,
	    bool bForce, double NT){

    // ܂ŏResizeĂ邩`FbN
    
    int currentSize[3];
    LOOP_XYZ currentSize[xyz] = pCSnap1->mSize[xyz];
    ProcDownSize(currentSize, pCVolumeBoxClone);

    if (currentSize[X] == mSize[X] &&
	currentSize[Y] == mSize[Y] &&
	currentSize[Z] == mSize[Z])
	if (bForce == FALSE)
	    if (IsSameStep(frame, factor) == TRUE){
		delete pCVolumeBoxClone;
		return;
	    }

       
    LOOP_XYZ mSize[xyz] = currentSize[xyz];

    wxStopWatch CStopWatch;
    CStopWatch.Start();

    //if (mpData_uchar != NULL) Clear();
        
    mNT = NT;
    mFrame = frame;
    mFactor = factor;
    mstrSnap1LoadedBody = pCSnap1->mstrLoadedFileBody;
    if (pCSnap2 != NULL)
	mstrSnap2LoadedBody = pCSnap2->mstrLoadedFileBody;

    // pCSnap1  pCSnap2 ŃTCYႤꍇ͖`FbNŊ댯

    // mSize Ō߂ꂽTCYɏ]
    // ProcDataModifierFirst  Boolean ̃}XNʃXbhŌvZ 
    // ̊Ԃ {f[^̃Rs[XbhŎs
    // ProcDataModifierLaste œ
    
    ProcDataModifierFirst(pCVolumeBoxClone);
    
    int size = mSize[X] * mSize[Y] * mSize[Z];
    if (mSizeAllocatedData != size){
	ClearData();
	if (pCSnap1->mbUseUShort == TRUE)
	    mpData_ushort = new unsigned short [size];
	else
	    mpData_uchar = new unsigned char [size];
	mSizeAllocatedData = size;
    }
    mbDataBakedToTexture = FALSE;

    if (mSize[X] == pCSnap1->mSize[X] &&
	mSize[Y] == pCSnap1->mSize[Y] &&
	mSize[Z] == pCSnap1->mSize[Z])
	CopyDataFromSnap_EqualSize(pCSnap1, pCSnap2, factor, size);
    else
	CopyDataFromSnap_Resized(pCSnap1, pCSnap2, factor, size);
    
    ProcDataModifierLast(pCVolumeBoxClone);
    //ClearTexture();

    mstrProcCounterForProcess = 
	GetStrProcCounterForProcessClone(pCVolumeBoxClone, mNT);

    wxString strLog; strLog << "Prepare Volume Data";
    strLog << CStopWatch.Time() << "ms " << ENDL;
    /*
    strLog << "frame=" << frame << " factor=" << factor << ENDL;
    wxString strLog2; 
    strLog2 << mstrSnap1LoadedBody << (long long)pCSnap1 << "::" << (long long)pCSnap1->mpData_ushort << ENDL;
    if (pCSnap2 != NULL)
	strLog2 << mstrSnap2LoadedBody << (long long)pCSnap2 << "::" << (long long)pCSnap2->mpData_ushort << ENDL;
    else
	strLog2 << "pCSnap2 = NULL" << ENDL;
    strLog2 << ENDL;
    */

    wxLogMessage(strLog);

    mPrepareCounter++;

    delete pCVolumeBoxClone;
}

void CLS::
ProcDownSize(int currentSize[3], ViewerClone_cls *pCVolumeBoxClone){
    uint numAttribClone = pCVolumeBoxClone->mVpCAttribClone.size();
    for (uint i = 0; i < numAttribClone; i++){
	AttributeClone_cls *pCMod = pCVolumeBoxClone->mVpCAttribClone[i];
	if (pCMod->IsAttributeValidAndSameAs("VolumeDownsize_cls")){
	    currentSize[X] /= pCMod->GetPCItem("DownsizeX")->GetBodyINT();
	    currentSize[Y] /= pCMod->GetPCItem("DownsizeY")->GetBodyINT();
	    currentSize[Z] /= pCMod->GetPCItem("DownsizeZ")->GetBodyINT();
	}
    }
    LOOP_XYZ if (currentSize[xyz] < 1) currentSize[xyz] = 1;
}

void CLS::
ProcDataModifierFirst(ViewerClone_cls *pCVolumeBoxClone){
    int numBoolean = pCVolumeBoxClone->GetNumAttribute("VolumeBoolean_cls");
    if (numBoolean != 0)
	ProcDataBooleanFirst(pCVolumeBoxClone);
}

void CLS::
ProcDataModifierLast(ViewerClone_cls *pCVolumeBoxClone){
    int numBoolean = pCVolumeBoxClone->GetNumAttribute("VolumeBoolean_cls");
    if (numBoolean != 0)
	ProcDataBooleanLast(pCVolumeBoxClone);
}

wxString CLS::
GetStrProcCounterForProcess(VolumeBox_cls *pCVolumeBox, double NT){
    //double NT = GetPCTimeCtrl()->GetCurrentNT();
    wxString strCounter;
    uint numDataProc = pCVolumeBox->mVpCDataProc.size();

    for (uint i = 0; i < numDataProc; i++){
	Attribute_cls *pCMod = pCVolumeBox->GetPCDataProcFromIndex(i);
	if (pCMod->IsAttributeValidAndSameAs("VolumeBoolean_cls")){
	    strCounter << pCMod->mCConfList.GetStrProcCounter();
	    strCounter << ((VolumeBoolean_cls*)pCMod)->GetStrProcCounterParam(pCVolumeBox, NT);
	    
	}
	if (pCMod->IsAttributeValidAndSameAs("VolumeDownsize_cls"))
	    strCounter << pCMod->mCConfList.GetStrProcCounter();
    }
    return strCounter;
}

wxString CLS::
GetStrProcCounterForProcessClone(ViewerClone_cls *pCVolumeBoxClone, double NT){
    
    //double NT = GetPCTimeCtrl()->GetCurrentNT();
    wxString strCounter;
    uint numAttribClone = pCVolumeBoxClone->mVpCAttribClone.size();
    
    for (uint i = 0; i < numAttribClone; i++){
	AttributeClone_cls *pCMod = pCVolumeBoxClone->mVpCAttribClone[i];
	if (pCMod->IsAttributeSameAs("VolumeBoolean_cls"))
	    strCounter << BooleanBase_cls::GetStrProcCounterClone(pCVolumeBoxClone, pCMod, NT);
	    
	if (pCMod->IsAttributeValidAndSameAs("VolumeDownsize_cls"))
	    strCounter << pCMod->mCConfList.GetStrProcCounter();
    }
    return strCounter;
}

void CLS::
ProcDataBooleanFirst(ViewerClone_cls *pCVolumeBoxClone){

    //wxStopWatch CStopWatch;
    //CStopWatch.Start();
    //wxString strLog;
    
    int size = mSize[X] * mSize[Y] * mSize[Z];
    if (size != mSizeAllocatedMask){
	ClearMask();
	mpMask = new char [size];
	mSizeAllocatedMask = size;
    }
    memset(mpMask, 1, size);

    mbFinishedFlagBoolean = FALSE;
    ThreadCallBoolean_cls *pCThread = new ThreadCallBoolean_cls;
    pCThread->SetParameters(pCVolumeBoxClone, mpMask,
			    mSize[X], mSize[Y], mSize[Z], mNT);
    pCThread->CreateThread(&mbFinishedFlagBoolean);
    pCThread->Run();
}

void CLS::
ProcDataBooleanLast(ViewerClone_cls *pCVolumeBoxClone){

    WaitThreadsNOTDeleteFlag(&mbFinishedFlagBoolean, 1, 3);
    VolumeBoolean_cls::GatherBooleanMask
	(mpData_uchar, mpData_ushort, mpMask, mSize[X], mSize[Y], mSize[Z]);
}

void CLS::
PrepareTexture(){
    if (mbDataBakedToTexture == TRUE)
	return;

    if (mTexture3D != 0) ClearTexture();
    
    glGenTextures(1, &mTexture3D);

    int alignment = 8;
    if (mSize[X] % 8 == 0) alignment = 8;
    else if (mSize[X] % 4 == 0) alignment = 4;
    else if (mSize[X] % 2 == 0) alignment = 2;
    else alignment = 1;
    
    glEnable(GL_TEXTURE_3D);
    glBindTexture(GL_TEXTURE_3D, mTexture3D);
    
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    glPixelStorei(GL_UNPACK_ALIGNMENT, alignment);

    bool bCompress = FALSE;
    int format16 = GL_LUMINANCE16;
    int format = GL_LUMINANCE;
    if (mpData_ushort != NULL){
	glTexImage3D(GL_TEXTURE_3D, 0, format16, 
		     mSize[X], mSize[Y], mSize[Z], 0,
		     format, GL_UNSIGNED_SHORT, mpData_ushort);
    }
    if (mpData_uchar != NULL){
	glTexImage3D(GL_TEXTURE_3D, 0, format, 
		     mSize[X], mSize[Y], mSize[Z], 0,
		     format, GL_UNSIGNED_BYTE, mpData_uchar);
    }
    
    mbDataBakedToTexture = TRUE;
}

void CLS::
BindTexture(){
    glEnable(GL_TEXTURE_3D);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glBindTexture(GL_TEXTURE_3D, mTexture3D);
}
