// -*- c -*- 
// "clsVolumeSequence.h"

#include "clsVolumeSequence.h"

#include "clsDataSnap.h"
#include "clsVolumeSnap.h"

#include "../MISC/clsMisc.h"
#include "../MODIFIER/clsSequenceRemap.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../PREFERENCE/clsPreference.h"
#include "../TIMELINE/clsTimeCtrl.h"

#define CLS VolumeSequence_cls

CLS::
VolumeSequence_cls() : DataSequence_cls(_DATA_VOLUME){
    mbTimeDepend = FALSE;
}

CLS::
~VolumeSequence_cls(){
    if (mApCSnap != NULL){
	if (mSeqLen == -1){
	    for (int i = 0; i < mSeqLenT; i++){
		if (mApCSnap[i] != NULL){
		    delete (VolumeSnap_cls*)mApCSnap[i];
		    mApCSnap[i] = NULL;
		}
	    }
	}
	for (int i = 0; i < mSeqLen; i++){
	    if (mApCSnap[i] != NULL){
		delete (VolumeSnap_cls*)mApCSnap[i];
		mApCSnap[i] = NULL;
	    }
	}
	delete mApCSnap;
	mApCSnap = NULL;
    }
}

int CLS::
GetTotalUsedMemoryMB(){
    long long memoryByte = 0;
    if (mApCSnap == NULL) return 0;
    for (int i = 0; i < mSeqLen; i++){
	if (mApCSnap[i] == NULL) continue;
	memoryByte += ((VolumeSnap_cls*)mApCSnap[i])->GetUsedMemoryByte();
    }
    for (int i = 0; i < mSeqLenT; i++){
	if (mApCSnap[i] == NULL) continue;
	memoryByte += ((VolumeSnap_cls*)mApCSnap[i])->GetUsedMemoryByte();
    }
    
    mTotalUsedMemoryMB = (int)(memoryByte / 1024 / 1024);
    
    return mTotalUsedMemoryMB;
}

VolumeSnap_cls* CLS::
GenerateIfNot(int index){

    int seqLen = mSeqLen;
    if (mSeqLenT >= 0) seqLen = mSeqLenT;

    if (index >= seqLen) index = seqLen - 1;
    if (index < 0) index = 0;
    if (mbTimeDepend == FALSE) index = 0;
    
    if (mApCSnap == NULL) return NULL;

    if (mApCSnap[index] == NULL){
	mApCSnap[index] = new VolumeSnap_cls(this, index);
    }

    return (VolumeSnap_cls*)mApCSnap[index];
}

VolumeSnap_cls* CLS::
GetPCSnap(int index){
    if (mApCSnap == NULL) return NULL;

    int seqLen = mSeqLen;
    if (mSeqLenT >= 1) seqLen = mSeqLenT;

    if (index >= seqLen) index = seqLen - 1;
    if (index < 0) index = 0;
    if (mbTimeDepend == FALSE) index = 0;
    
    return (VolumeSnap_cls*)mApCSnap[index];
}

void CLS::
PrefetchData(SequenceRemap_cls *pCSeqRemap, bool bZStack){
    if (mReadingLock > 0) return; // 1Xbĥ

    double NT = GetPCTimeCtrl()->GetCurrentNT();
    double dNT = GetPCTimeCtrl()->GetDNT();
    int frameBase, subframe;
    GetPCTimeCtrl()->GetFrameFromNT(NT, frameBase, subframe);

    int seqLen = mSeqLen;
    if (bZStack == TRUE) seqLen = mSeqLenT;

    int prefetch = GetPCPreference()->mCPrefetchLength.GetBodyINT();
    for (int i = 0; i < prefetch; i++){
	int frame1 = frameBase + i;
	if (pCSeqRemap != NULL)
	    pCSeqRemap->GetRemappedFrame(NT+dNT*i, frame1, seqLen);
	    /*
	    frame1 = pCSeqRemap->GetRemappedFrame(NT+dNT*i, frame1);
	    if (pCSeqRemap->mCbRepeat.GetBodyBOOL() == TRUE)
		frame1 = Misc_cls::Surplus(frame1, seqLen);
	    */
	
	if (frame1 < 0 || frame1 >= seqLen) continue;
	
	VolumeSnap_cls *pCSnap1 = GenerateIfNot(frame1);
	if (bZStack == TRUE){
	    pCSnap1->SetFileSequenceDual(this, frame1);
	    pCSnap1->SetStrFile(GetModifiedStrDual(frame1, 0));
	}
	else{
	    if (mbTimeDepend == TRUE)
		pCSnap1->SetStrFile(GetModifiedStr(frame1));
	    else
		pCSnap1->SetStrFile(mstrPath);
	}
	if (pCSnap1->PrefetchData(bZStack) == TRUE) return;
	
	/*
	int frame2 = frame1 + 1; // ԗpɎXebvvtFb`
	if (frame2 >= seqLen) continue;
	VolumeSnap_cls *pCSnap2 = GenerateIfNot(frame2);
	
	if (bZStack == TRUE){
	    pCSnap2->SetFileSequenceDual(this, frame2);
	    pCSnap2->SetStrFile(GetModifiedStrDual(frame2, 0));
	}
	else
	    pCSnap2->SetStrFile(GetModifiedStr(frame2));
	if (pCSnap2->PrefetchData(bZStack) == TRUE) return;
	*/
    }
}

void CLS::
ReleaseMemoryData(SequenceRemap_cls *pCSeqRemap){
    
    ParamGroup_cls *pCRemapParamGroup = NULL;
    bool bRepeat = FALSE;
    if (pCSeqRemap != NULL){
	pCRemapParamGroup = pCSeqRemap->GetPCParamGroup("Remap");
	bRepeat = pCSeqRemap->mCbRepeat.GetBodyBOOL();
    }
    MarkPriority(pCRemapParamGroup, bRepeat);
    int oldestIndex = GetOldestDrawnIndex();
    if (oldestIndex == -1) return;

    VolumeSnap_cls *pCSnap = GetPCSnap(oldestIndex);
    if (pCSnap == NULL)
	return;
    
    delete pCSnap;
    mApCSnap[oldestIndex] = NULL;
}
