// -*- c -*- 
// "clsVolumeSnap.h"

#include "clsVolumeSnap.h"
#include "clsVolumeSequence.h"

#include "READTHREAD/clsDF3Read.h"
#include "READTHREAD/clsImgSeqRead.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../GRAPHIC/clsReadPNG.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/MyEndian.h"

#include <wx/filename.h>

wxString GstrVolumeSnap;

#define CLS VolumeSnap_cls

CLS::
VolumeSnap_cls(VolumeSequence_cls *pCParent, int index) : 
DataSnap_cls(pCParent, _DATA_VOLUME, index){
    LOOP_XYZ mSize[xyz] = 0;
    LOOP_XYZ mAllocatedSize[xyz] = 0;
    
    //mDataFormat = _VOLUME_UCHAR;
    mpData_uchar = NULL;
    mpData_ushort = NULL;
    mbZStack = FALSE;
    mbUseUShort = FALSE;
    mbUseFloat = FALSE;
    
    Clear();
}

CLS::
~VolumeSnap_cls(){ Clear(); }

void CLS::
ClearUChar(){
    if (mpData_uchar != NULL){
	delete [] mpData_uchar; mpData_uchar = NULL; }
}

void CLS::
ClearUShort(){
    if (mpData_ushort != NULL){
	delete [] mpData_ushort; mpData_ushort = NULL; }
}

void CLS::
Clear(){
    WaitReadingThread(3);

    ClearUChar();
    ClearUShort();

    //mDataFormat = _VOLUME_UCHAR;
    //mbUseUShort = FALSE;
    //mbUseFloat = FALSE;
    
    mbReadStarted = FALSE;
    mbReadFinished = FALSE;
}

bool CLS::
IsSizeSame(VolumeSnap_cls* pCTgt){
    LOOP_XYZ if (mSize[xyz] != pCTgt->mSize[xyz]) return FALSE;
    return TRUE;
}

wxString CLS::
GetStrInfo(int x, int y, int z){
    wxString strInfo;
    wxFileName CName(mstrLoadedFileHeader);
    strInfo << CName.GetFullName() << ENDL;
    strInfo << mSize[X] << " x " << mSize[Y] << " x " << mSize[Z] << ENDL;
    if (mbUseUShort == TRUE)
	strInfo << "16bit" << ENDL;
    else
	strInfo << "8bit" << ENDL;
    if (x < 0 || x >= mAllocatedSize[X] ||
	y < 0 || y >= mAllocatedSize[Y] ||
	z < 0 || z >= mAllocatedSize[Z]){ 
	strInfo << "-" << ENDL; 
	return strInfo; 
    }
    int index = z * mAllocatedSize[X] * mAllocatedSize[Y] + y * mAllocatedSize[X] + x;
    strInfo << "raw value: ";
    if (mpData_ushort != NULL)
	strInfo << mpData_ushort[index] << ENDL;
    else if (mpData_uchar != NULL)
	strInfo << (int)mpData_uchar[index] << ENDL;
    return strInfo;
}

int CLS::
GetUsedMemoryByte(){
    if (mpData_ushort != NULL) 
	return mAllocatedSize[X] * mAllocatedSize[Y] * mAllocatedSize[Z] * 2;
    if (mpData_uchar == NULL) return 0;
    return mAllocatedSize[X] * mAllocatedSize[Y] * mAllocatedSize[Z];
}

void CLS::
AllocateIfNot(){
    if (mSize[X] != mAllocatedSize[X] ||
	mSize[Y] != mAllocatedSize[Y] ||
	mSize[Z] != mAllocatedSize[Z]) 	Clear();
    
    int size = mSize[X] * mSize[Y] * mSize[Z];
    if (mbUseUShort == TRUE){
	if (mpData_uchar != NULL) ClearUChar();
	if (mpData_ushort != NULL) return;
	mpData_ushort = new unsigned short [size];
    }
    else{
	if (mpData_ushort != NULL) ClearUShort();
	if (mpData_uchar != NULL) return;
	mpData_uchar = new uchar [size];
    }
    mAllocatedSize[X] = mSize[X];
    mAllocatedSize[Y] = mSize[Y];
    mAllocatedSize[Z] = mSize[Z];
}

void CLS::
GetScaleCandid(double scaleCandid[3]){
    int maxScale = 0;
    LOOP_XYZ if (mSize[xyz] > maxScale) maxScale = mSize[xyz];
    if (maxScale == 0){
	LOOP_XYZ scaleCandid[xyz] = 1;
	return;
    }
    LOOP_XYZ scaleCandid[xyz] = (double)mSize[xyz] / maxScale;
    return;
}

void CLS::
SetStrFile(wxString strFile){ mstrFile = strFile.Clone(); }

void CLS::
SetFileSequenceDual(FileSequenceDual_cls *pCFSDual, int frame){
    mFrame = frame;
    if (pCFSDual != NULL)
	mCFileSequenceDual = *pCFSDual;
}

bool CLS::
ReadHeaderZStack(){
    VolumeSequence_cls* pCParent = (VolumeSequence_cls*)mpCParent;

    FILE *fp = fopen(mstrFile.char_str(), "rb");
    wxImage CImage;
    wxImage CImageDW;
    if (fp != NULL){
	if (GCClassProvider.GetPCReadPNG()->ReadPNG(fp, mstrFile, CImage, CImageDW) == FALSE)
	    CImage.LoadFile(mstrFile);
	if (fp != NULL) fclose(fp);
    }

    //wxImage CImage(mstrFile);
    if (CImage.IsOk() == FALSE){
	mSize[X] = 128; mSize[Y] = 128; mSize[Z] = 128;
	
    }
    else{
	mSize[X] = CImage.GetWidth();
	mSize[Y] = CImage.GetHeight();
	pCParent->CalcMaxOffsetZ(mSelfIndex);
	mSize[Z] = pCParent->mSeqLenZ;
    }

    if (CImageDW.IsOk() == TRUE)
	mbUseUShort = TRUE;
    else
	mbUseUShort = FALSE;

    mstrLoadedFileHeader = mstrFile;

    AllocateIfNot();
    return TRUE;
}

bool CLS::
ReadHeader(bool bZStack){
    if (mstrLoadedFileHeader.IsSameAs(mstrFile) == TRUE &&
	mbZStack == bZStack) return FALSE;

    mbZStack = bZStack;
    if (mbZStack == TRUE) return ReadHeaderZStack();
	
    short buffer[3];
    FILE *pFile = NULL;
    
    wxFileOffset fileSize = 0;
    wxFileName CFileName(mstrFile);
    if (CFileName.GetExt().IsSameAs("df3") ||
	CFileName.GetExt().IsSameAs("DF3")){

	wxFile CFile(mstrFile);
	fileSize = CFile.Length();

	pFile = fopen(mstrFile.char_str(), "rb");
    }

    if (pFile == NULL){
	mSize[X] = 64; mSize[Y] = 64; mSize[Z] = 64;
    }
    else{
	fread(buffer, sizeof(short), 3, pFile);
	fclose(pFile);
	
	LOOP_XYZ{
	    MyEndian::SwapByte((void*)&buffer[xyz], 2);
	    mSize[xyz] = buffer[xyz];
	}

	long long dataSize = ((long long)mSize[X]) * ((long long)mSize[Y]) * ((long long)mSize[Z]);
	if (fileSize == dataSize*4 + 6){
	    mbUseUShort = TRUE;
	    mbUseFloat = TRUE;
	}
	else if (fileSize == dataSize*2 + 6){
	    mbUseUShort = TRUE;
	    mbUseFloat = FALSE;
	}
	else{
	    mbUseUShort = FALSE;
	    mbUseFloat = FALSE;
	}
    }    
    
    mstrLoadedFileHeader = mstrFile;

    AllocateIfNot();
    return TRUE;
}

bool CLS::
DefaultData(){
    if (mstrLoadedFileBody.IsSameAs("None")) return FALSE;

    const int sizeX = mSize[X], sizeY = mSize[Y], sizeZ = mSize[Z];
    GstrVolumeSnap = "DefultData Setted:";
    GstrVolumeSnap << sizeX << ":" << sizeY << ":" << sizeZ ;
    for (int z = 0; z < sizeZ; z++){
	int indexz = z * sizeX * sizeY;
	for (int y = 0; y < sizeY; y++){
	    int indexy = indexz + y * sizeX;
	    for (int x = 0; x < sizeX; x++){
		int index = indexy + x;
		float fvalue = 0.5 + sin(3.14159*x/sizeX * 2 +3.14159/2) * 0.5;
		mpData_uchar[index] = fvalue * 254 + 1;
	    }
	}
    }
     
    mstrLoadedFileBody = "None";
    return FALSE;
}

bool CLS::
ReadData(bool bParallel){
    WaitReadingThread(3);

    if (mbUseUShort == TRUE){
	if (mpData_ushort == NULL){
	    wxLogMessage("mpData_ushort == NULL");
	    return FALSE;
	}
    }
    else{
	if (mpData_uchar == NULL){
	    wxLogMessage("mpData_uchar == NULL");
	    return FALSE;
	}
    }

    if (mbReadFinished == TRUE)
	if (mstrLoadedFileBody.IsSameAs(mstrFile) == TRUE) return FALSE;
    if (mstrFile.IsSameAs("None")) return DefaultData();
    if (mstrFile.IsEmpty() == TRUE) return FALSE;
    if (mFrame < 0) return FALSE;

    mbReadStarted = TRUE;
    mbReadFinished = FALSE;
    
    if (mbZStack == TRUE){
	ThreadImgSeqRead *pCThread = new ThreadImgSeqRead();
	pCThread->SetParameters(&mbReadFinished, this);
	
	mstrLoadedFileBody = mstrFile;
	if (bParallel == TRUE){
	    pCThread->Run();
	}
	else{
	    pCThread->DummyEntry();
	    delete pCThread;
	    mbReadFinished = TRUE;
	}
    }
    else{
	ThreadDF3Read *pCThread = new ThreadDF3Read();
	pCThread->SetParameters(&mbReadFinished, this);
	
	mstrLoadedFileBody = mstrFile;
	if (bParallel == TRUE){
	    pCThread->Run();
	}
	else{
	    pCThread->DummyEntry();
	    delete pCThread;
	    mbReadFinished = TRUE;
	}
    }
    
    return TRUE;
}

bool CLS::
PrefetchData(bool bZStack){

    if (ReadHeader(bZStack) == FALSE) return FALSE;

    ReadData(TRUE);
    
    return TRUE;
}
