// -*- c++ -*-
// (setq c-basic-offset 4)
// "clsEnsembleCtrl.h"

#include "clsEnsembleCtrl.h"

#include "../GLSL/clsShaderCtrl.h"
#include "../MYGL/clsFBO.h"

#define CLS EnsembleCtrl_cls

CLS::
EnsembleCtrl_cls(){
    int mEnsembleStep = 0;
    mpCFBOBuffer = NULL;
    SetToDefaultParam();
}

CLS::
~EnsembleCtrl_cls(){
    ClearFBOBuffer();
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();

    wxString strUpdate = "VIEW3D::PROPERTY_RENDER";

    mCConfList.SetTag("EnsembleCtrl");
    mCConfList.Add(&mCbUseEnsemble, "UseEnsemble=", "0", strUpdate);
    mCConfList.Add(&mCEnsembleSambleNum, "EnsembleSampleNum=", "4", strUpdate);
    mCConfList.Add(&mCbUseGammaForEnsemble, "UseGammaForEnsemble=", "1", strUpdate);
    //mCConfList.Add(&mCbTransparentShadowForEnsemble, "UseTransparentShadow=", "0", strUpdate);
}

void CLS::
ClearFBOBuffer(){
    if (mpCFBOBuffer != NULL){
	delete mpCFBOBuffer;
	mpCFBOBuffer = NULL;
    }
}

void CLS::
SetEnsembleStep(int ensembleStep){
    mEnsembleStep = ensembleStep;
}

int CLS::
GetEnsembleSampleNum(){
    if (mCbUseEnsemble.GetBodyBOOL() == 0) return 1;
    return mCEnsembleSambleNum.GetBodyINT();
}

void CLS::
PrepareBuffer(FBO_cls *pCFBOSrc){
    int size = pCFBOSrc->mWidthFBO;
    int depth = pCFBOSrc->mColorDepth;
    
    if (mpCFBOBuffer == NULL){
	mpCFBOBuffer = new FBO_cls(size, size, NULL);
	mpCFBOBuffer->SetColorDepth(depth);
    }
    else{
	mpCFBOBuffer->SetGeometryFBO(size, size);
	mpCFBOBuffer->SetColorDepth(depth);
    }

    mpCFBOBuffer->MakeFBO(_FBO_USE_RGBA, FALSE);
}

void CLS::
DrawOnBuffer(FBO_cls *pCFBOSrc, int sampleNum){
    
    int size = pCFBOSrc->mWidthTex;

    bool bClear = (mEnsembleStep == 0) ? TRUE : FALSE;
    mpCFBOBuffer->Begin(bClear);

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    GetPCGLExtentions()->ClampColor(GL_FALSE);

    //glColor4f(1.0, 1.0, 1.0, 1.0f/sampleNum);
    glColor4f(1.0, 1.0, 1.0, 1.0f/sampleNum);

    glEnable(GL_TEXTURE_2D);

    ExposureParams_cls CParam;
    if (mCbUseGammaForEnsemble.GetBodyBOOL() == TRUE) CParam.mGamma = 2.2;
    GetPCShaderCtrl()->BeginExposure(&CParam);

    glEnable(GL_BLEND);
    glDisable(GL_LIGHTING);
    glBlendFunc(GL_ONE, GL_ONE);

    pCFBOSrc->DrawOnScreen2D(size, size,
			     size, size,
			     0, 0, TRUE);

    GetPCShaderCtrl()->EndShader();

    glPopAttrib();

    mpCFBOBuffer->End();
}

void CLS::
DrawFromBuffer(FBO_cls *pCFBOTgt){
    
    int size = pCFBOTgt->mWidthTex;

    //bool bClear = (mBlurStep == 0) ? TRUE : FALSE;
    pCFBOTgt->Begin(TRUE);
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glColor4f(1.0, 1.0, 1.0, 1.0f);

    glEnable(GL_TEXTURE_2D);

    ExposureParams_cls CParam;
    if (mCbUseGammaForEnsemble.GetBodyBOOL() == TRUE) CParam.mGamma = 0.4545;
    GetPCShaderCtrl()->BeginExposure(&CParam);

    glEnable(GL_BLEND);
    glDisable(GL_LIGHTING);
    glBlendFunc(GL_ONE, GL_ZERO);

    mpCFBOBuffer->DrawOnScreen2D(size, size,
				 size, size,
				 0, 0, TRUE);

    GetPCShaderCtrl()->EndShader();

    glPopAttrib();

    pCFBOTgt->End();
}

EnsembleCtrl_cls *GpCEnsembleCtrl = NULL;
