// -*- c++ -*-
// (setq c-basic-offset 4)

#include "clsTimeCtrl.h"
#include "clsMotionBlurCtrl.h"

#include "../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../COMMON/clsQueue.h"
#include "../GRAPHIC/clsImageOutput.h"
#include "../MISC/clsMisc.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../UNDO/clsUndoManager.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsWX.h"

#include "../UNDO/clsUndoManager.h"

#define CLS TimeCtrl_cls

CLS::
TimeCtrl_cls() {
    mPlayDirection = 0;
    mbRecordRender = FALSE;
    mbRecordConfirmed = FALSE;
    mbRecordSnap = FALSE;
    mbExport = FALSE;
    mbExportConfirmed = FALSE;
    SetToDefaultParam();
    //mCStopWatch.Start();
}

CLS::
~TimeCtrl_cls(){
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();

    mCConfList.SetTag("TimeCtrl");
    mCConfList.Add(&mCFrameMax, "FrameMax=", "100");
    mCConfList.Add(&mCFrameCurrent, "FrameCurrent=", "0");
    mCConfList.Add(&mCWAMin, "WAMin=", "0");
    mCConfList.Add(&mCWAMax, "WAMax=", "100");
    mCConfList.Add(&mCWASubMin, "WASubMin=", "0");
    mCConfList.Add(&mCWASubMax, "WASubMax=", "0");
    mCConfList.Add(&mCSubFrameMax, "SubFrameMax=", "2", "TIMELINE");
    mCConfList.Add(&mCSubFrameStepping, "SubFrameStepping=", "1", "TIMELINE");
    mCConfList.Add(&mCSubFrameCurrent, "SubFrameCurrent=", "0");
    mCConfList.Add(&mCInterpolateComposite, "InterpolateComposite=", "0", "TIMELINE::VIEW3D");
    mCConfList.Add(&mCDefaultKey, "DefaultKey=", "4", "TIMELINE");
    mCConfList.Add(&mCbRepeat, "Repeat=", "0");
}

wxString CLS::
GetStrSelf(){ return wxT("TimeCtrl::"); }

wxString CLS::
GetStrRecordInfo(double percentage){
    wxString strRecordInfo = GetStrCurrentLapseTime();
    strRecordInfo << _("passed") << "(" << wxString::Format("%.1f", percentage) << "%)" << " " << _("Expected Time") << GetStrExpectedFinishTime(percentage);
    return strRecordInfo;
}

double CLS::
GetOutputPercent(){
    int WAMin = mCWAMin.GetBodyINT();
    int WAMax = mCWAMax.GetBodyINT();
    double NormalzeWAMin = GetNormalizedValue(WAMin, 0, 0);
    double NormalzeWAMax = GetNormalizedValue(WAMax, 0, 0);
    if (NormalzeWAMin == NormalzeWAMax) return 0;
    double NT = GetCurrentNT();
    double ratio = (NT - NormalzeWAMin)/(NormalzeWAMax - NormalzeWAMin);
    return ratio * 100;
}

int CLS::
GetStep(int frame, int subframe){
    return frame * mCSubFrameMax.GetBodyINT() + subframe;
}

int CLS::
GetCurrentStep(){
    return GetStep(mCFrameCurrent.GetBodyINT(), mCSubFrameCurrent.GetBodyINT());
}

int CLS::
GetImageCompositeNum(){
    if (mCInterpolateComposite.GetBodyINT() == 0) return 1;
    //҂̎ł^C}bvł҂肶ȂƂ̂
    //2Ԃ
    return 2;
}

void CLS::
GetDefaultKeyTypeAndEase(int *pKeyType, double *pEaseFactor){
    switch(mCDefaultKey.GetBodyINT()){
    case 0:
	*pKeyType = 1; *pEaseFactor = 1; break;
    case 1:
	*pKeyType = 2; *pEaseFactor = 1; break;
    case 2:
	*pKeyType = 2; *pEaseFactor = 0; break;
    case 3:
	*pKeyType = 0; *pEaseFactor = 1; break;
    case 4:
	*pKeyType = -1; *pEaseFactor = 1; break;
    }
}

double CLS::
GetDefaultResolution(){
    return 1.0 / mCFrameMax.GetBodyINT() / 4;
}

double CLS::
GetQuantized(double normed, int type){
    double frameMax = mCFrameMax.GetBodyINT();
    double frameDouble = normed * frameMax;
    int frame = Misc_cls::Round(frameDouble, type);
    return (float)frame / frameMax;
}

double CLS::
GetQuantizedForColor(double normed, int type){
    double frameMax = _COLOR_CURVE_RESOLUTION;
    double frameDouble = normed * frameMax;
    int frame = Misc_cls::Round(frameDouble, type);
    return (float)frame / frameMax;
}

double CLS::
GetFrameFromNT(double normed, int &frame, int &subframe){
    int frameMax = mCFrameMax.GetBodyINT();
    frame = floor(normed * frameMax + 1e-6);
    double diff = (normed - (double)frame / frameMax) * frameMax;
    subframe = (int)(diff * mCSubFrameMax.GetBodyINT() + 0.5);
    // ԗp̃t@N^[Ԃ
    //return (double)subframe / mCSubFrameMax.GetBodyINT();
    return diff;
}

double CLS::
GetNormalizedValue(int frame, int subFrame, int blurStep){
    double frameMax = (double)mCFrameMax.GetBodyINT();
    double subFrameMax = (double)mCSubFrameMax.GetBodyINT();
    
    double blurFactor = GetPCBlurCtrl()->mCBlurLength.GetBodyDOUBLE() * blurStep / GetPCBlurCtrl()->mCBlurQuality.GetBodyINT();
    if (frameMax == 0) return 0;
    return (frame + ((double)subFrame + blurFactor)/subFrameMax) / frameMax;
}

double CLS::
GetCurrentNT(bool bForData){
    bool bBlurAffect = (!bForData) | (!mCInterpolateComposite.GetBodyBOOL());
    return GetNormalizedValue(mCFrameCurrent.GetBodyINT(), mCSubFrameCurrent.GetBodyINT(), GetPCBlurCtrl()->mBlurStep * bBlurAffect);
}

double CLS::
GetNextNT(bool bForData){
    bool bBlurAffect = (!bForData) | (!mCInterpolateComposite.GetBodyBOOL());

    bool bMotionBlur = GetPCBlurCtrl()->mCbUseMotionBlur.GetBodyBOOL() * bBlurAffect;

    if (bMotionBlur == TRUE){
	int blurStep = GetPCBlurCtrl()->mBlurStep + 1;
	if (blurStep >= GetPCBlurCtrl()->mCBlurQuality.GetBodyINT())
	    goto BREAK;
	return 
	    GetNormalizedValue(mCFrameCurrent.GetBodyINT(), mCSubFrameCurrent.GetBodyINT(), blurStep);
    }
BREAK:
    
    int frame, subFrame;
    GetNextTimeInfo(frame, subFrame, +1);
    return GetNormalizedValue(frame, subFrame, 0);
}

double CLS::
GetDNT(){
    return 1.0 / mCFrameMax.GetBodyINT() / mCSubFrameMax.GetBodyINT();
}

void CLS::
AdjustFrameForComposer(int *pFrame1, int *pFrame2, double *pInterpFactor,
		       int composit, bool bForNext){
    if (mCInterpolateComposite.GetBodyINT() == 0) return;
    if (bForNext == FALSE && composit == 1)
	*pFrame1 = *pFrame2;
    *pInterpFactor = 0;
}

void CLS::
StoreInterpFactorForComposit(double interpFactor){
    mInterpFactorForComposit = interpFactor;
}

double CLS::
GetNormalizedWAMin(){
    return GetNormalizedValue(mCWAMin.GetBodyINT(),
			      mCWASubMin.GetBodyINT()); }

double CLS::
GetNormalizedWAMax(){
    if (mCWAMax.GetBodyINT() == mCFrameMax.GetBodyINT())
	return GetNormalizedValue(mCWAMax.GetBodyINT(), 0);
    return GetNormalizedValue(mCWAMax.GetBodyINT(),
			      mCWASubMax.GetBodyINT()); 
}

double CLS::
GetNormalizedWARegion(){
    int WAmin = mCWAMin.GetBodyINT();
    int WAmax = mCWAMax.GetBodyINT();
    int WASubmin = mCWASubMin.GetBodyINT();
    int WASubmax = mCWASubMax.GetBodyINT();
    int subFrameMax = mCSubFrameMax.GetBodyINT();
    if (subFrameMax == 0){
	printf("WAmin == %d\n", WAmin);
	printf("WAmax == %d\n", WAmax);
	printf("subFrameMax == %d\n", subFrameMax);
	subFrameMax = 1;
    }
    
    if (WAmax == mCFrameMax.GetBodyINT()) WASubmax = 0;

    int WASubSum = (WAmin + WAmax)*subFrameMax + WASubmin + WASubmax;
    int WASubSumMid = WASubSum / 2;

    int WAmid = WASubSumMid / subFrameMax;
    int WASubmid = WASubSumMid % subFrameMax;
    
    return GetNormalizedValue(WAmid, WASubmid);
}

void CLS::
GetNextTimeInfo(int &frame, int &subFrame, int direction){
    frame = mCFrameCurrent.GetBodyINT();
    subFrame = mCSubFrameCurrent.GetBodyINT();
    int subFrameMax = mCSubFrameMax.GetBodyINT();
    
    subFrame += direction * mCSubFrameStepping.GetBodyINT();
    while (subFrame < 0){
	subFrame += subFrameMax;//subFrameMax - 1;
	frame--;
    }

    while (subFrame >= subFrameMax){
	subFrame -= subFrameMax;//0;
	frame++;
    }
}

bool CLS::
ProcPlay(){
    if (mPlayDirection == 0) return FALSE;

    if (mbRecordRender == TRUE && mbRecordConfirmed == FALSE)
	return FALSE;
    if (mbExport == TRUE && mbExportConfirmed == FALSE)
	return FALSE;

    if (msbForceAbort == TRUE){
	msbForceAbort = FALSE;
	StopPlayCommand();
	return TRUE;
    }

    StartSWIfNotYet();

    GetPCQueue()->mbQueueSyncronizeGraphNavi = TRUE;
    
    int WAMax = mCWAMax.GetBodyINT();
    int WAMin = mCWAMin.GetBodyINT();
    int frame, subFrame;
    GetNextTimeInfo(frame, subFrame, mPlayDirection);
    bool bRepeat = mCbRepeat.GetBodyBOOL();
    if (mbRecordRender == TRUE) bRepeat = FALSE;

    if (frame <= 0 && subFrame <= 0) {
	mCSubFrameCurrent.SetBodyFromData((long long)0);
	mCFrameCurrent.SetBodyFromData((long long)0);
	StopPlayCommand(); return TRUE; }

    if (frame <= WAMin && subFrame <= 0) {
	if (bRepeat == FALSE){
	    mCSubFrameCurrent.SetBodyFromData((long long)0);
	    mCFrameCurrent.SetBodyFromData((long long)WAMin);
	    StopPlayCommand(); return TRUE; }
	mCSubFrameCurrent.SetBodyFromData((long long)0);
	mCFrameCurrent.SetBodyFromData((long long)WAMax);
	SendUpdate(); return TRUE;
    }

    if (frame >= WAMax) {
	if (bRepeat == FALSE){
	    mCSubFrameCurrent.SetBodyFromData((long long)0);
	    mCFrameCurrent.SetBodyFromData((long long)WAMax);
	    StopPlayCommand(); return TRUE; }
	mCSubFrameCurrent.SetBodyFromData((long long)0);
	mCFrameCurrent.SetBodyFromData((long long)WAMin);
	SendUpdate(); return TRUE;
    }

    mCSubFrameCurrent.SetBodyFromData((long long)subFrame);
    mCFrameCurrent.SetBodyFromData((long long)frame);
        
    SendUpdate();

    return TRUE;
}

void CLS::
SendUpdate(){
    GCClassProvider.GetPCMain()->UpdatePanels("TIMELINE::GRAPH::COLOR_CURVE::VIEW3D::PROPERTY_OBJECT::PROPERTY_MATERIAL::PROPERTY_MODIFIER::PROPERTY_DATAPROC::PROPERTY_SCENE::PROPERTY_INFO");
}

/*
void CLS::
PrepareBlurBuffer(FBO_cls *pCFBOSrc){
    mCBlurCtrl.PrepareBuffer(pCFBOSrc);
}

void CLS::
SetBlurStep(int blurStep){
    mCBlurCtrl.mBlurStep = blurStep;
}
*/

void CLS::
SetWAMinCommand(int min){
    if (min == mCWAMin.GetBodyINT()) return;
    wxString strUndo = GetStrSelf() << wxT("SetWAMin::") << mCWAMin.GetBodyINT();
    wxString strRedo = GetStrSelf() << wxT("SetWAMin::") << min;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetWAMaxCommand(int max){
    if (max == mCWAMax.GetBodyINT()) return;
    wxString strUndo = GetStrSelf() << wxT("SetWAMax::") << mCWAMax.GetBodyINT();
    wxString strRedo = GetStrSelf() << wxT("SetWAMax::") << max;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetWARegionCommand(int region, int regionL, int regionR){
    GetPCUndo()->OpenBlock(wxT("TimeWARegion"));
    SetWAMinCommand(region - regionL);
    SetWAMaxCommand(region + regionR);
    GetPCUndo()->CloseBlock(wxT("TimeWARegion"));
}

void CLS::
SetCurrentFrameToEndCommand(int direction){
    int newFrame = 0, newSubFrame = 0;
    if (direction == +1){
	newFrame = mCWAMax.GetBodyINT();
	newSubFrame = mCWASubMax.GetBodyINT();
    }
    if (direction == -1){
	newFrame = mCWAMin.GetBodyINT();
	newSubFrame = mCWASubMin.GetBodyINT();
    }
    GetPCUndo()->OpenBlock("ToEnd");
    SetCurrentFrameCommand(newFrame);
    SetCurrentSubFrameCommand(newSubFrame);
    GetPCUndo()->CloseBlock("ToEnd");
}

void CLS::
SetCurrentFrameToNextCommand(int direction){
    int newFrame = 0, newSubFrame = 0;
    GetNextTimeInfo(newFrame, newSubFrame, direction);
    
    int frameMax = mCFrameMax.GetBodyINT();
    if (newFrame < 0){
	newFrame = 0; newSubFrame = 0; }
    if (newFrame >= frameMax)
	newFrame = frameMax;
    if (newFrame == frameMax)
	newSubFrame = 0;
    
    GetPCUndo()->OpenBlock("ToNext");
    SetCurrentFrameCommand(newFrame);
    SetCurrentSubFrameCommand(newSubFrame);
    GetPCUndo()->CloseBlock("ToNext");
}

void CLS::
SetCurrentFrameCommand(int currentFrame){
    if (currentFrame == mCFrameCurrent.GetBodyINT()) return;
    wxString strUndo = GetStrSelf() << wxT("SetFrameCurrent::") << mCFrameCurrent.GetBodyINT();
    wxString strRedo = GetStrSelf() << wxT("SetFrameCurrent::") << currentFrame;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetCurrentSubFrameCommand(int currentSubFrame){
    if (currentSubFrame == mCSubFrameCurrent.GetBodyINT()) return;
    wxString strUndo = GetStrSelf() << wxT("SetSubFrameCurrent::") << mCSubFrameCurrent.GetBodyINT();
    wxString strRedo = GetStrSelf() << wxT("SetSubFrameCurrent::") << currentSubFrame;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetCurrentNTCommand(double newNT){
    int newFrame = 0, newSubFrame = 0;
    GetFrameFromNT(newNT, newFrame, newSubFrame);
    if (newFrame == mCFrameCurrent.GetBodyINT() &&
	newSubFrame == mCSubFrameCurrent.GetBodyINT()) return;

    if (newSubFrame == mCSubFrameCurrent.GetBodyINT()){
	SetCurrentFrameCommand(newFrame); return;
    }
    
    if (newFrame == mCFrameCurrent.GetBodyINT()){
	SetCurrentSubFrameCommand(newSubFrame); return;
    }
	
    GetPCUndo()->OpenBlock("NT");
    SetCurrentFrameCommand(newFrame); 
    SetCurrentSubFrameCommand(newSubFrame);
    GetPCUndo()->CloseBlock("NT");
}

void CLS::
SetFrameMaxCommand(int frameMax){
    if (frameMax == mCFrameMax.GetBodyINT()) return;
    wxString strUndo = GetStrSelf() << wxT("SetFrameMax::") << mCFrameMax.GetBodyINT();
    wxString strRedo = GetStrSelf() << wxT("SetFrameMax::") << frameMax;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
SetRepeatCommand(bool bRepeat){
    if (bRepeat == mCbRepeat.GetBodyBOOL()) return;
    wxString strUndo = GetStrSelf() << "SetRepeat::" << mCbRepeat.GetBodyBOOL();
    wxString strRedo = GetStrSelf() << "SetRepeat::" << bRepeat;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
AnimCommand(bool bSnap){
    if (mbRecordRender == TRUE) 
	return StopPlayCommand();
    
    if (bSnap == FALSE && 
	GetPCImageOutput()->mCbRenderFromWorkAreaIn.GetBodyBOOL() == TRUE)
	SetCurrentFrameToEndCommand(-1);
    StartPlayCommand(+1);
    mbRecordSnap = bSnap;
    mbRecordRender = TRUE;
    mbRecordConfirmed = FALSE;
    StartSWIfNotYet();
    StopSW(); // U 0 
    GetPCMain()->UpdatePanels("PROPERTY_RENDER::VIEW3D");
}

void CLS::
ExportCommand(){
    SetCurrentFrameToEndCommand(-1);
    StartPlayCommand(+1);
    mbExport = TRUE;
    mbExportConfirmed = FALSE;
    StartSWIfNotYet();
    StopSW();
    GetPCMain()->UpdatePanels("PROPERTY_RENDER::VIEW3D");
}

void CLS::
StartPlayCommand(int direction){
    if (mPlayDirection == direction) return StopPlayCommand();
    GetPCUndo()->OpenBlock("StartPlay");
    wxString strUndo1 = GetStrSelf() << "SetFrameCurrent::" << mCFrameCurrent.GetBodyINT();
    wxString strUndo2 = GetStrSelf() << "SetFrameSubCurrent::" << mCSubFrameCurrent.GetBodyINT();
    wxString strRedo1 = GetStrSelf() << wxT("StartPlay(dummy)");
    wxString strRedo2 = GetStrSelf() << wxT("Skip");
    GetPCUndo()->PushCommandPairAndDo(strUndo1, strRedo1); 
    GetPCUndo()->PushCommandPairAndDo(strUndo2, strRedo2); 
    GetPCUndo()->CloseBlock("StartPlay");
    mPlayDirection = direction;
}

void CLS::
StopPlayCommand(){
    GetPCUndo()->OpenBlock("StopPlay");
    wxString strUndo1 = GetStrSelf() << wxT("Skip");
    wxString strUndo2 = GetStrSelf() << wxT("StopPlay(dummy)");
    wxString strRedo1 = GetStrSelf() << "SetFrameCurrent::" << mCFrameCurrent.GetBodyINT();
    wxString strRedo2 = GetStrSelf() << "SetFrameSubCurrent::" << mCSubFrameCurrent.GetBodyINT();
    
    GetPCUndo()->PushCommandPairAndDo(strUndo1, strRedo1); 
    GetPCUndo()->PushCommandPairAndDo(strUndo2, strRedo2); 
    GetPCUndo()->CloseBlock("StopPlay");
    mPlayDirection = 0;
    mbRecordSnap = FALSE;
    if (mbRecordRender == TRUE){
	mbRecordRender = FALSE;
	mbRecordConfirmed = FALSE;
	GetPCMain()->UpdatePanels("PROPERTY_RENDER");
	StopSW();
    }
    if (mbExport == TRUE){
	mbExport = FALSE;
	GetPCMain()->UpdatePanels("PROPERTY_RENDER");
	StopSW();
    }
}

bool CLS::
ProcSetWAMin(wxString strFirstItem){
    mCWAMin.SetBodyFromData(strFirstItem);
    return TRUE;
}

bool CLS::
ProcSetWAMax(wxString strFirstItem){
    mCWAMax.SetBodyFromData(strFirstItem);
    return TRUE;
}

bool CLS::
ProcSetFrameMax(wxString strFirstItem){
    GetPCQueue()->mQueueFrameMaxPrev = mCFrameMax.GetBodyINT();
    mCFrameMax.SetBodyFromData(strFirstItem);
    GetPCQueue()->mbQueueAdjustAADuetoFrameMax = TRUE;
    return TRUE;
}

bool CLS::
ProcSetFrameCurrent(wxString strFirstItem){
    mCFrameCurrent.SetBodyFromData(strFirstItem);
    GetPCQueue()->mbQueueSyncronizeGraphNavi = TRUE;
    return TRUE;
}

bool CLS::
ProcSetSubFrameCurrent(wxString strFirstItem){
    mCSubFrameCurrent.SetBodyFromData(strFirstItem);
    GetPCQueue()->mbQueueSyncronizeGraphNavi = TRUE;
    return TRUE;
}

bool CLS::
ProcSetRepeat(wxString strFirstItem){
    mCbRepeat.SetBodyFromData(strFirstItem);
    return TRUE;
}

bool CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == TRUE)
	return TRUE;

    if (strFirstItem.IsSameAs(wxT("TimeCtrl")) == FALSE) return FALSE;
		
    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
    bool bRet = FALSE;
    if (strFirstItem.IsSameAs("SetWAMin")){
	strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
	bRet = ProcSetWAMin(strFirstItem); }
    if (strFirstItem.IsSameAs("SetWAMax")){
	strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
	bRet = ProcSetWAMax(strFirstItem); }
    if (strFirstItem.IsSameAs("SetFrameMax")){
	strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
	bRet = ProcSetFrameMax(strFirstItem); }
    if (strFirstItem.IsSameAs("SetFrameCurrent")){
	strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
	bRet = ProcSetFrameCurrent(strFirstItem); }
    if (strFirstItem.IsSameAs("SetSubFrameCurrent")){
	strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
	bRet = ProcSetSubFrameCurrent(strFirstItem); }
    if (strFirstItem.IsSameAs("SetRepeat")){
	strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
	bRet = ProcSetRepeat(strFirstItem); }
    
    if (bRet == TRUE) SendUpdate();
    return TRUE;
}

TimeCtrl_cls *GpCTimeCtrl = NULL;
bool TimeCtrl_cls::msbForceAbort = FALSE;
