// (setq c-basic-offset 4)
// "clsOutliner.h"

#include "clsTimeline.h"

#include "clsPlayCtrl.h"
#include "clsTimelineTics.h"
#include "clsTimelineParam.h"
#include "clsTimelineAA.h"
#include "clsTimelineWA.h"
#include "clsTimeCtrl.h"

#include "../COMMON/clsQueue.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../MISC/clsMisc.h"
#include "../PREFERENCE/clsShortcut.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/EX/clsPanelBar.h"
#include "../WIDGETCTRL/WX/clsFontProvider.h"
#include "../WIDGETCTRL/WX/clsStringProvider.h"
#include "../WIDGETCTRL/INPUTITEM/clsAtoSTextCtrl.h"
#include "../WIDGETCTRL/INPUTITEM/clsTextButton.h"
#include "../WIDGETCTRL/INPUTITEM/clsVCenter.h"
#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"

#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#define _TIMELINE_DEFALTSPACER 136

#define CLS Timeline_cls

BEGIN_EVENT_TABLE(CLS, wxPanelEX)
EVT_BUTTON(btnTicsMode, CLS::OnTicsMode)
EVT_BUTTON(btnKilroy, CLS::OnKilroy)
EVT_BUTTON(btnTimeDepend, CLS::OnTimeDepend)
EVT_BUTTON(btnSelectOnly, CLS::OnSelectOnly)
EVT_BUTTON(menuPanelView, CLS::OnMenuView)
EVT_BUTTON(menuPanelSelect, CLS::OnMenuSelect)
EVT_BUTTON(menuPanelTimeline, CLS::OnMenuTimeline)
EVT_BUTTON(menuPanelChannel, CLS::OnMenuChannel)
EVT_BUTTON(menuPanelKey, CLS::OnMenuKey)
EVT_MENU(menuObjectDeselect, CLS::OnDeselect)
EVT_MENU_RANGE(menuExtrapolateOff, menuExtrapolateOn, CLS::OnMenuExtrapolate)
EVT_MENU_RANGE(menuZoomPlus, menuZoomMinus, CLS::OnMenuZoom)
EVT_MENU(menuTLTimeDepend, CLS::OnMenuTimeDepend)
EVT_MENU(menuTLShy, CLS::OnMenuShy)
EVT_MENU(menuTLIn, CLS::OnMenuTLIn)
EVT_MENU(menuTLOut, CLS::OnMenuTLOut)
EVT_MENU(menuDelete, CLS::OnMenuDelete)
EVT_TEXT_ENTER(txtCtrlFrame, CLS::OnFrame)
EVT_TEXT_ENTER(txtCtrlSubFrame, CLS::OnFrame)
EVT_SIZE(CLS::OnSize)
//EVT_CHAR_HOOK (CLS::OnKeyDown)
END_EVENT_TABLE()

CLS::
Timeline_cls(wxWindow *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
    wxPanelEX(pCParent, id, pos, size, style)
{
    SetToDefaultParam();
    mDraggingItem = 0;
    
    mCBitmapBuffer.Create(240, 36);
    mCMemoryDCWorkSpace.SelectObject(mCBitmapBuffer);
    
    SetPanelType(_PANEL_TIMELINE);
    
    mpCPanelBar = new PanelBar_cls(this);
    mpCPanelBar->Freeze();
    mpCPanelBar->SetPanelType(_PANEL_CHILD, _PANEL_TIMELINE);

    wxBoxSizer *pBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerBar = new wxBoxSizer(wxHORIZONTAL);
    
    ConstructTypeSelector(pBoxSizerBar);

    mpCTextMenuView = new TextButton_cls
	(mpCPanelBar, menuPanelView, GetPCStringProvider()->GetStrMenu("View"));
    mpCTextMenuView->SetEventTarget(this);
    VCenter_cls *pCMenuView = new VCenter_cls(mpCTextMenuView);

    mpCTextMenuSelect = new TextButton_cls
	(mpCPanelBar, menuPanelSelect, GetPCStringProvider()->GetStrMenu("Select"));
    mpCTextMenuSelect->SetEventTarget(this);
    VCenter_cls *pCMenuSelect = new VCenter_cls(mpCTextMenuSelect);

    mpCTextMenuTimeline = new TextButton_cls
	(mpCPanelBar, menuPanelTimeline, GetPCStringProvider()->GetStrMenu("Timeline"));
    mpCTextMenuTimeline->SetEventTarget(this);
    VCenter_cls *pCMenuTimeline = new VCenter_cls(mpCTextMenuTimeline);
    
    mpCTextMenuChannel = new TextButton_cls
	(mpCPanelBar, menuPanelChannel, GetPCStringProvider()->GetStrMenu("Channel"));
    mpCTextMenuChannel->SetEventTarget(this);
    VCenter_cls *pCMenuChannel = new VCenter_cls(mpCTextMenuChannel);

    mpCTextMenuKey = new TextButton_cls
	(mpCPanelBar, menuPanelKey, GetPCStringProvider()->GetStrMenu("Key"));
    mpCTextMenuKey->SetEventTarget(this);
    VCenter_cls *pCMenuKey = new VCenter_cls(mpCTextMenuKey);

    pBoxSizerBar->Add(pCMenuView, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuSelect, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuTimeline, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuChannel, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuKey, 0, wxEXPAND);
    pBoxSizerBar->AddStretchSpacer(1);
    mpCPanelBar->ConstructBarItemZoom(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemDefault(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemHideShow(pBoxSizerBar);

    mpCPanelBar->SetSizer(pBoxSizerBar);

    wxBoxSizer *pBoxSizerAABase = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerTicsBase = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerWABase = new wxBoxSizer(wxHORIZONTAL);
    mpBoxSizerAA = new wxBoxSizer(wxHORIZONTAL);
    mpBoxSizerTics = new wxBoxSizer(wxHORIZONTAL);
    mpBoxSizerWA = new wxBoxSizer(wxHORIZONTAL);

    //wxBoxSizer *pBoxSizerAA = new wxBoxSizer(wxHORIZONTAL);
    //wxBoxSizer *pBoxSizerWA = new wxBoxSizer(wxHORIZONTAL);
    
    mpCTimelineAA = new TimelineAA_cls(this);
    mpCTimelineTics = new TimelineTics_cls(this);
    mpCTimelineWA = new TimelineWA_cls(this);
    mpCTimelineParam = new TimelineParam_cls(this);
    mpCPlayCtrl = new PlayCtrl_cls(this);

    mpCBtnKilroy = new BitmapPanel_cls(this, btnKilroy, wxDP, wxSize(18, 18));
    mpCBtnKilroy->AddBitmapStr("wxART_KILROY_S_N");
    mpCBtnKilroy->AddBitmapStr("wxART_KILROY_S_P");
    mpCBtnKilroy->mbButton = TRUE;
    mpCBtnKilroy->SetBorder(1, 1);

    mpCBtnTimeDepend = new BitmapPanel_cls(this, btnTimeDepend, wxDP, wxSize(18, 18));
    mpCBtnTimeDepend->AddBitmapStr("wxART_TIME_N");
    mpCBtnTimeDepend->AddBitmapStr("wxART_TIME_P");
    mpCBtnTimeDepend->mbButton = TRUE;
    mpCBtnTimeDepend->SetBorder(1, 1);

    mpCBtnSelectOnly = new BitmapPanel_cls(this, btnSelectOnly, wxDP, wxSize(18, 18));
    mpCBtnSelectOnly->AddBitmapStr("wxART_ARROW_N");
    mpCBtnSelectOnly->AddBitmapStr("wxART_ARROW_P");
    mpCBtnSelectOnly->mbButton = TRUE;
    mpCBtnSelectOnly->SetBorder(1, 1);
    
    BitmapPanel_cls *pCBtnTicsMode = new BitmapPanel_cls(this, btnTicsMode, wxDP, wxSize(16, 16));
    pCBtnTicsMode->AddBitmapStr("wxART_SELECTOR_N", "wxART_SELECTOR_P");
    pCBtnTicsMode->mbButton = TRUE;
    pCBtnTicsMode->SetBorder(1, 1);
    mpCTicsMode = new wxStaticText(this, wxID_ANY, Tics_cls::msAstrTicsMode[0], wxDP, wxSize(104, 16));
    
    pBoxSizerTicsBase->AddSpacer(16);
    pBoxSizerTicsBase->Add(pCBtnTicsMode, 0, wxEXPAND);
    pBoxSizerTicsBase->Add(mpCTicsMode, 0, wxEXPAND);
    
    mpCTicsSpacer = mpBoxSizerTics->AddSpacer(0);
    mpBoxSizerTics->Add(mpCTimelineTics, 1, wxEXPAND);
    pBoxSizerTicsBase->Add(mpBoxSizerTics, 1, wxEXPAND);
    //pBoxSizerTicsBase->AddSpacer(12);

    pBoxSizerAABase->AddSpacer(_TIMELINE_DEFALTSPACER - 48 - 6);
    pBoxSizerAABase->Add(mpCBtnSelectOnly, 0, wxEXPAND);
    pBoxSizerAABase->Add(mpCBtnTimeDepend, 0, wxEXPAND);
    pBoxSizerAABase->Add(mpCBtnKilroy, 0, wxEXPAND);
    mpCAASpacer = mpBoxSizerAA->AddSpacer(0);
    mpBoxSizerAA->Add(mpCTimelineAA, 1, wxEXPAND);
    pBoxSizerAABase->Add(mpBoxSizerAA, 1, wxEXPAND);
    pBoxSizerAABase->AddSpacer(12);

    mpCTextCtrlFrame = new AtoSTextCtrl_cls(this, txtCtrlFrame, "0", wxDP, wxSize(80, -1));
    mpCTextCtrlSubFrame = new AtoSTextCtrl_cls(this, txtCtrlSubFrame, "0", wxDP, wxSize(56, -1));
    
    pBoxSizerWABase->Add(mpCTextCtrlFrame, 0, wxEXPAND);
    pBoxSizerWABase->Add(mpCTextCtrlSubFrame, 0, wxEXPAND);

    //pBoxSizerWABase->AddSpacer(_TIMELINE_DEFALTSPACER);
    mpCWASpacer = mpBoxSizerWA->AddSpacer(0);
    mpBoxSizerWA->Add(mpCTimelineWA, 1, wxEXPAND);
    pBoxSizerWABase->Add(mpBoxSizerWA, 1, wxEXPAND);
    pBoxSizerWABase->AddSpacer(12);
    
    pBoxSizerSrc->Add(mpCPanelBar, 0, wxEXPAND);
    
    pBoxSizerSrc->Add(pBoxSizerAABase, 0, wxEXPAND);    
    pBoxSizerSrc->Add(pBoxSizerTicsBase, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerWABase, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCTimelineParam, 1, wxEXPAND);
    pBoxSizerSrc->Add(mpCPlayCtrl, 0, wxEXPAND);

    Freeze();
    
    SetSizer(pBoxSizerSrc);

    mpCPanelBar->Thaw();
    Thaw();
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();
    mCConfList.SetTag("TimeLine");
    mCConfList.Add(&mCAAMin, "AAMin=", "0");
    
    int AAMax = GetPCTimeCtrl()->mCFrameMax.GetBodyINT();
    wxString strAAMax; strAAMax << AAMax;
    mCConfList.Add(&mCAAMax, "AAMax=", strAAMax);
    mCConfList.Add(&mCTicsMode, "TicMode=", "0");
    wxString strSash; strSash << _TL_SASH_MIN;
    mCConfList.Add(&mCSashX, "SashX=", strSash);
    mCConfList.Add(&mCFontSize, "FontSize=", "1");
    mCConfList.Add(&mCbKilroy, "Kilroy=", "0");
    mCConfList.Add(&mCbTimeDepend, "TimeDepend=", "0");
    mCConfList.Add(&mCbSelectOnly, "SelectOnly=", "0");
    mPrevSash = 136;
}

int CLS::
GetLineHeight(){
    int fontType = mCFontSize.GetBodyINT();
    return FontProvider_cls::GetHeight(fontType);
}

void CLS::
Zoom(int direction){
    long long fontsize = mCFontSize.GetBodyINT() + direction;
    if (fontsize < 0) fontsize = 0;
    if (fontsize >= _FONT_MAX) fontsize = _FONT_MAX - 1;
    mCFontSize.SetBodyFromData(fontsize);
    UpdatePanels(wxT("ALL"));
}

void CLS::
OnFrame(wxComEvt& event){
    long frameNew, subFrameNew;
    mpCTextCtrlFrame->GetValue().ToLong(&frameNew);
    mpCTextCtrlSubFrame->GetValue().ToLong(&subFrameNew);
    double newNT = GetPCTimeCtrl()->GetNormalizedValue(frameNew, subFrameNew);
    GetPCTimeCtrl()->SetCurrentNTCommand(newNT);
}

void CLS::
OnTicsMode(wxComEvt& event){
    long long mode = mCTicsMode.GetBodyINT();
    mode++;
    if (mode > 2) mode = 0;
    mpCTicsMode->SetLabel(Tics_cls::msAstrTicsMode[mode]);
    mCTicsMode.SetBodyFromData(mode);
    mpCTimelineTics->Refresh();
    mpCTimelineWA->Refresh();
}

void CLS::
OnKilroy(wxComEvt& event){
    mCbKilroy.SetBodyFromData((long long)mpCBtnKilroy->mState);
    mpCTimelineParam->Refresh(FALSE);
}

void CLS::
OnTimeDepend(wxComEvt& event){
    mCbTimeDepend.SetBodyFromData((long long)mpCBtnTimeDepend->mState);
    mpCTimelineParam->Refresh(FALSE);
}

void CLS::
OnSelectOnly(wxComEvt& event){
    mCbSelectOnly.SetBodyFromData((long long)mpCBtnSelectOnly->mState);
    mpCTimelineParam->Refresh(FALSE);
}

void CLS::
OnMenuView(wxComEvt& event){
    Shortcut_cls *pCSC = GetPCShortcut();
    wxMenu CMenu;
    pCSC->SetMenuItem(&CMenu, menuZoomPlus);
    pCSC->SetMenuItem(&CMenu, menuZoomMinus);
    pCSC->SetMenuCheckItem(&CMenu, menuShowPanelManage, mpCPanelBar->mbShowItem);
    
    mpCTextMenuView->PopMenu(this, &CMenu);
}

void CLS::
OnMenuSelect(wxComEvt& event){
    wxMenu CMenu;

    CMenu.Append(menuObjectDeselect, GetPCShortcut()->GetMenuString(menuObjectDeselect));
    bool bBarItem = mpCPanelBar->mbShowItem;
    mpCTextMenuSelect->PopMenu(this, &CMenu);
}

void CLS::
OnMenuTimeline(wxComEvt& event){
    Shortcut_cls *pCSC = GetPCShortcut();
    wxMenu CMenu;
    pCSC->SetMenuItem(&CMenu, menuTLIn);
    pCSC->SetMenuItem(&CMenu, menuTLOut);
    
    mpCTextMenuTimeline->PopMenu(this, &CMenu);
}

void CLS::
OnDeselect(wxComEvt& event){
    GetPCSceneObjCtrl()->Select(wxES, FALSE); }

void CLS::
OnMenuChannel(wxComEvt& event){
    wxMenu CMenu;

    wxMenu *pCMenuExtrapolate = new wxMenu;
    pCMenuExtrapolate->Append(menuExtrapolateOff, "Off");
    pCMenuExtrapolate->Append(menuExtrapolateOn, "On");
    CMenu.Append(wxID_ANY, _("Extrapolate"), pCMenuExtrapolate);

    mpCTextMenuChannel->PopMenu(this, &CMenu);
}

void CLS::
OnMenuKey(wxCommandEvent& event){
    wxMenu CMenu;
    CMenu.Append(menuDelete, GetPCShortcut()->GetMenuString(menuDelete));
    
    mpCTextMenuKey->PopMenu(this, &CMenu);
}

void CLS::
OnMenuDelete(wxCommandEvent& event){
    
    if (GetPCSceneObjCtrl()->IsSelectedTLKeyExist() == TRUE)
	GetPCSceneObjCtrl()->DeleteSelectedTLKeyCommand();

    else 
	event.Skip();
	
}

void CLS::
OnMenuExtrapolate(wxComEvt& event){
    bool bOn = (event.GetId() == menuExtrapolateOn);
    
    std::vector < SceneObj_cls* > VpCObj;
    GetPCSceneObjCtrl()->GetVPCList(VpCObj);
    
    std::vector < wxString > VstrParamTag;
    uint numObj = VpCObj.size();
    for (uint i = 0; i < numObj; i++) 
	VpCObj[i]->GetVstrParamTag(VstrParamTag);
    GetPCFilterCtrl()->GetVstrParamTag(VstrParamTag);

    GetPCUndo()->OpenBlock("SetExtrapolate");
    uint numParamTag = VstrParamTag.size();
    for (uint i = 0; i < numParamTag; i++)
	GetPCMain()->SetParameterConfCommand(VstrParamTag[i], "Extrapolate=", bOn, TRUE);
    GetPCUndo()->CloseBlock("SetExtrapolate");
}

void CLS::
OnMenuZoom(wxComEvt& event){
    bool bExpand = (event.GetId() == menuZoomPlus);
    ExpandAA(bExpand);
    UpdatePanels("ALL");
}

void CLS::
OnMenuTimeDepend(wxComEvt& event){
    mCbTimeDepend.SetBodyFromData((long long)!mCbTimeDepend.GetBodyBOOL());
    mpCBtnTimeDepend->mState = mCbTimeDepend.GetBodyINT();
    mpCBtnTimeDepend->Refresh(FALSE);
    mpCTimelineParam->Refresh(FALSE);
}

void CLS::
OnMenuShy(wxComEvt& event){
    mCbKilroy.SetBodyFromData((long long)!mCbKilroy.GetBodyBOOL());
    mpCBtnKilroy->mState = mCbKilroy.GetBodyINT();
    mpCBtnKilroy->Refresh(FALSE);
    mpCTimelineParam->Refresh(FALSE);
}

void CLS::
OnMenuTLIn(wxComEvt& event){
    GetPCTimeCtrl()->SetWAMinCommand(GetPCTimeCtrl()->mCFrameCurrent.GetBodyINT());
}

void CLS::
OnMenuTLOut(wxComEvt& event){
    GetPCTimeCtrl()->SetWAMaxCommand(GetPCTimeCtrl()->mCFrameCurrent.GetBodyINT());
}

void CLS::
OnSize(wxSizeEvent& event){
    UpdatePanels("ALL");
    event.Skip();
}

/*
void CLS::
OnKeyDown(wxKeyEvent& event){
    if (GetPCShortcut()->ProcShortcutKey(event, this) == FALSE) 
	event.Skip();
}
*/

/*
void CLS::
ProcRename(const wxString& strNameOld, const wxString& strNameNew){
    mpCTimelineParam->ProcRename(strNameOld, strNameNew); }

void CLS::
ProcDelete(const wxString& strName){
    mpCTimelineParam->ProcDelete(strName); }
*/

void CLS::
ProcAfterMath(const wxString& strCommand, const wxString& strItem){
    mpCTimelineParam->ProcAfterMath(strCommand, strItem);
}

/*
void CLS::
CheckAbortClicked(const wxPoint& mousePos){
    mpCPlayCtrl->CheckAbortClicked(mousePos);
}
*/

void CLS::
UpdateStatusButtons(){
    mpCBtnKilroy->SetStatus(mCbKilroy.GetBodyINT());
    mpCBtnTimeDepend->SetStatus(mCbTimeDepend.GetBodyINT());
    mpCBtnSelectOnly->SetStatus(mCbSelectOnly.GetBodyINT());
}

void CLS::
UpdateTextCtrls(){
    int frame, subFrame;
    GetPCTimeCtrl()->GetFrameFromNT(GetPCTimeCtrl()->GetCurrentNT(),
				    frame, subFrame);
    wxString strFrame; strFrame << frame;
    wxString strSubFrame; strSubFrame << subFrame;
    mpCTextCtrlFrame->SetValue(strFrame);
    mpCTextCtrlSubFrame->SetValue(strSubFrame);
}

void CLS::
UpdatePanels(const wxString& strPanelTag, bool bUpdateBar){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs(wxT("ALL"))){

	if (GetPCQueue()->mbQueueAdjustAADuetoFrameMax == TRUE){
	    long long frameMax = GetPCTimeCtrl()->mCFrameMax.GetBodyINT();
	    if (mCAAMax.GetBodyINT() == GetPCQueue()->mQueueFrameMaxPrev)
		mCAAMax.SetBodyFromData(frameMax);
	    else if (mCAAMax.GetBodyINT() > frameMax)
		mCAAMax.SetBodyFromData(frameMax);
	}
		
	int sash = mCSashX.GetBodyINT();
	if (mPrevSash != sash){
	    mpCAASpacer->SetMinSize(sash - _TIMELINE_DEFALTSPACER, 0);
	    mpCTicsSpacer->SetMinSize( sash - _TIMELINE_DEFALTSPACER, 0);
	    mpCWASpacer->SetMinSize( sash - _TIMELINE_DEFALTSPACER, 0);
	    mpBoxSizerAA->Layout();
	    mpBoxSizerTics->Layout();
	    mpBoxSizerWA->Layout();
	    mPrevSash = sash;
	}
	mpCTimelineParam->Refresh(); 
	mpCTimelineTics->Refresh();
	mpCTimelineAA->Refresh();
	mpCTimelineWA->Refresh();
	mpCPlayCtrl->UpdatePanels();
	if (bUpdateBar == TRUE) mpCPanelBar->Refresh();
	UpdateTextCtrls();
	UpdateStatusButtons();
    }
}

void CLS::
Serialize(wxString& strOut){
    //mCSelectorMode.SetBodyFromData((long long)mpCIconSelector->mValue);
    ConfListContainer_cls::Serialize(strOut);
}

void CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == FALSE)
	return;
    //mpCIconSelector->mValue = mCSelectorMode.GetBodyINT();
    //HideAndShow();
    
}
