// (setq c-basic-offset 4)
// "clsTimelineAA.h"

#include "clsTimelineAA.h"

#include "clsTimeCtrl.h"
#include "clsTimeline.h"

#include "../GRAPH/clsGraphEditor.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#define CLS TimelineAA_cls

BEGIN_EVENT_TABLE(CLS, TimelineBase_cls)
EVT_PAINT       (CLS::OnPaint)
EVT_MOUSE_EVENTS(CLS::OnMouse)
END_EVENT_TABLE()

CLS::
TimelineAA_cls(Timeline_cls *pCParent) : TimelineBase_cls(pCParent, _TL_AA, wxSize(-1, 16)){
    ConstructorBase(); }

CLS::
TimelineAA_cls(GraphEditor_cls *pCParent) : TimelineBase_cls(pCParent, _TL_AA, wxSize(-1, 16)){
    ConstructorBase(); }

void CLS::
ConstructorBase(){
    mCIcon.CopyFromBitmap(ART("wxART_TIMESLIDER_SMALL"));
    mCIconMin.CopyFromBitmap(ART("wxART_MIN_SMALL"));
    mCIconMax.CopyFromBitmap(ART("wxART_MAX_SMALL"));
    mCIconRegion.CopyFromBitmap(ART("wxART_REGION"));
}

void CLS::
LimitDragging(){
    if (mSelected == _TL_SEL_MIN){
	int min = mFrameDragging;
	int max = mpCAALink->mCAAMax.GetBodyINT();
	if (min > max - 1) min = max - 1;
	mFrameDragging = min;
    }
    if (mSelected == _TL_SEL_MAX){
	int min = mpCAALink->mCAAMin.GetBodyINT();
	int max = mFrameDragging;
	if (max < min + 1) max = min + 1;
	mFrameDragging = max;
    }
    if (mSelected == _TL_SEL_REGION){
	int LWidth = mpCAALink->mRegionLforDrag;
	int RWidth = mpCAALink->mRegionRforDrag;
	if (mFrameDragging - LWidth < 0) mFrameDragging = LWidth;
	if (mFrameDragging + RWidth > mFrameMax) mFrameDragging = mFrameMax - RWidth;
    }
}

void CLS::
UpdateDisplay(wxAutoBufferedPaintDC &dc){
    
    dc.SetBackground(wxBrush(wxColour(216,216,216), wxSOLID));
    dc.Clear();

    UpdateState();
    SetTLRegion(0, 1);
    
    DrawWorkingArea(dc, 0);
    if (mbForColorCurve == FALSE) DrawCurrentTime(dc, 0);
    
    int draggingX = 0;
    int sliderX = GetDevCoordFromN(mTimeNormalized, 0)-4, sliderY = 2;
    int minX = GetDevCoordFromN(mAAMinNormalized, 0)-5, minY = 1;
    int maxX = GetDevCoordFromN(mAAMaxNormalized, 0), maxY = 1;
    int regionX = GetDevCoordFromN(mAARegionNormalized, 0)-11, regionY = 9;

    if (mSelected == _TL_SEL_SLIDER){
	sliderX = GetDevCoordFromN(mQuantizedTimeDragging, 0)-4;
	draggingX = sliderX; }
    if (mSelected == _TL_SEL_MIN){
	minX = GetDevCoordFromN(mQuantizedTimeDragging, 0)-5;
	draggingX = minX; }
    if (mSelected == _TL_SEL_MAX){
	maxX = GetDevCoordFromN(mQuantizedTimeDragging, 0);
	draggingX = maxX; }
    if (mSelected == _TL_SEL_REGION){
	regionX = GetDevCoordFromN(mQuantizedTimeDragging, 0)-11;
	draggingX = regionX; }
    
    dc.DrawIcon(mCIconRegion, regionX, regionY);
    dc.DrawIcon(mCIconMin, minX, minY);
    dc.DrawIcon(mCIconMax, maxX, maxY);
    mRectRegion = wxRect(regionX, regionY, 16, 8);
    mRectMin = wxRect(minX, minY, 6, 15);
    mRectMax = wxRect(maxX, maxY, 6, 15);

    if (mbForColorCurve == FALSE){
	dc.DrawIcon(mCIcon, sliderX, sliderY);
	mRectSlider = wxRect(sliderX, sliderY, 9, 11);
    }
    // bZ[W`
    int offset = 0;
    if (draggingX > mPanelWidth * 0.7) offset = -40;

    if (HasCapture() == TRUE){
	wxString Mes;
	Mes << mFrameDragging;
	dc.SetFont(*wxNORMAL_FONT);
	dc.SetTextForeground(*wxRED);
	dc.DrawText(Mes, draggingX + 15 + offset, 0);
    }
}

void CLS::
OnPaint(wxPaintEvent& event){
    wxAutoBufferedPaintDC dc(this);
    UpdateDisplay(dc);
}
 
void CLS::
OnMouse(wxMouseEvent& event){
    
    ProcMouseEventBase(event, 0);

    if (event.ButtonUp(wxMOUSE_BTN_LEFT)) ProcMouseRelease();

    if (event.Dragging()){
	if (mSelected == _TL_SEL_MIN) MoveAAMin();
	if (mSelected == _TL_SEL_MAX) MoveAAMax();
	if (mSelected == _TL_SEL_REGION) MoveAARegion();
    }

    event.Skip();
}

void CLS::
ProcMouseRelease(){
    if (mSelected == _TL_SEL_SLIDER) MoveTimeSlider();
    if (mSelected == _TL_SEL_MIN) MoveAAMin();
    if (mSelected == _TL_SEL_MAX) MoveAAMax();
    if (mSelected == _TL_SEL_REGION) MoveAARegion();

    ProcMouseReleaseBase();
}
