// (setq c-basic-offset 4)
// "clsTimelineBase.h"

#include "clsTimelineBase.h"

#include "clsAACtrl.h"
#include "clsTimeCtrl.h"
#include "clsTimeline.h"
#include "../GRAPH/clsGraphEditor.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../UNDO/clsUndoManager.h"

#include "../MISC/clsMisc.h"

TLPanelGeom_cls::
TLPanelGeom_cls(){
    mBorderXL = 6;
    mBorderXR = 6;
}

bool TLPanelGeom_cls::
IsWithinRange(int devX, int margin){
    if (devX -  margin > 0 && devX + margin < mPanelWidth)
	return TRUE;
    return FALSE;
}

int TLPanelGeom_cls::
GetDevCoordFromN(double normedValue, int offsetX){
    double remapedValue = (normedValue-mTLRegionMin) / (mTLRegionMax-mTLRegionMin);
    return (int)(remapedValue * (mPanelWidth-(offsetX + mBorderXL + mBorderXR)) +
		 offsetX + mBorderXL);
}

double TLPanelGeom_cls::
GetNFromDevCoord(int devX, int offsetX){
    double NValueOnCanves = (double)(devX-offsetX-mBorderXL)/(mPanelWidth-(offsetX+mBorderXL+mBorderXR));
    return NValueOnCanves * (mTLRegionMax - mTLRegionMin) + mTLRegionMin;
}

void TLPanelGeom_cls::
SetTLRegion(double regionMin, double regionMax){
    mTLRegionMin = regionMin;
    mTLRegionMax = regionMax;
}

#define CLS TimelineBase_cls

BEGIN_EVENT_TABLE(CLS, wxPanel)
EVT_SIZE        (CLS::OnSize)
END_EVENT_TABLE()

CLS::
TimelineBase_cls(Timeline_cls *pCParent, int type, wxSize size) : 
wxPanel(pCParent, wxID_ANY, wxDP, size)
{
    mbForColorCurve = FALSE;
    mpCParent = pCParent;
    mpCAALink = pCParent;
    ConstructorBase(type);
}

CLS::
TimelineBase_cls(GraphEditor_cls *pCParent, int type, wxSize size) : 
wxPanel(pCParent, wxID_ANY, wxDP, size)
{
    mpCParent = pCParent;
    mpCAALink = pCParent;
    ConstructorBase(type);
}

void CLS::
ConstructorBase(int type){
    mType = type;
    SetBackgroundColour(wxColour(216, 216, 216));
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);

    mSelected = 0;
    mQuantizedTimeDragging = 0;
}

void CLS::
UpdateStateForCC(){
    mFrameCurrent = 0;
    mFrameMax = _COLOR_CURVE_RESOLUTION;
    mSubFrameCurrent = 0;
    mTimeNormalized = 0;
    mAAMinNormalized = (float)mpCAALink->mCAAMin.GetBodyINT() / mFrameMax;
    mAAMaxNormalized = (float)mpCAALink->mCAAMax.GetBodyINT() / mFrameMax;
    mAARegionNormalized = (float)(mpCAALink->mCAAMax.GetBodyINT() + mpCAALink->mCAAMin.GetBodyINT())/2 / mFrameMax;
    mNormalizedWAMin = 0;//mpCAALink->GetWAMinNormalized();
    mNormalizedWAMax = 1;//mpCAALink->GetWAMaxNormalized();
    mNormalizedWARegion = mpCAALink->GetWARegionNormalizedQ();
}

void CLS::
UpdateState(){
    if (mbForColorCurve == TRUE) return UpdateStateForCC();

    TimeCtrl_cls *pCTime = GetPCTimeCtrl();
    mFrameCurrent = pCTime->mCFrameCurrent.GetBodyINT();
    mFrameMax = pCTime->mCFrameMax.GetBodyINT();
    mSubFrameCurrent = pCTime->mCSubFrameCurrent.GetBodyINT();
    mTimeNormalized = mpCAALink->GetTimeNormalized();
    mAAMinNormalized = pCTime->GetNormalizedValue
	(mpCAALink->mCAAMin.GetBodyINT(), 0);
    mAAMaxNormalized = pCTime->GetNormalizedValue
	(mpCAALink->mCAAMax.GetBodyINT(), 0);
    mAARegionNormalized = pCTime->GetNormalizedValue
	((mpCAALink->mCAAMax.GetBodyINT() + mpCAALink->mCAAMin.GetBodyINT())/2, 0);
    mNormalizedWAMin = mpCAALink->GetWAMinNormalized();
    mNormalizedWAMax = mpCAALink->GetWAMaxNormalized();
    mNormalizedWARegion = mpCAALink->GetWARegionNormalizedQ();
}

double CLS::
GetQuantized(double posNormed, int type){
    if (mbForColorCurve == TRUE)
	return GetPCTimeCtrl()->GetQuantizedForColor(posNormed, type);
    return GetPCTimeCtrl()->GetQuantized(posNormed, type);
}

void CLS::
CalcFrameDraggingFromPosition(int offsetX){
    //if (offsetX == -1) offsetX = GetOffsetX();
    double normalizedPos = 
	GetNFromDevCoord(mMousePosCurrentX, offsetX);
    double quantizedPos = GetQuantized(normalizedPos, _MISC_ROUND_ROUND);
    if (quantizedPos < 0) quantizedPos = 0;
    if (quantizedPos > 1) quantizedPos = 1;
    mQuantizedTimeDragging = quantizedPos;

    mFrameDragging = mQuantizedTimeDragging * mFrameMax;
    LimitDragging();
    mQuantizedTimeDragging = (double)mFrameDragging / mFrameMax;
}

void CLS::
DrawWorkingArea(wxAutoBufferedPaintDC &dc, int offsetX){
    //if (offsetX == -1) offsetX = GetOffsetX();
    int x0 = GetDevCoordFromN(mNormalizedWAMin, offsetX);
    int x1 = GetDevCoordFromN(mNormalizedWAMax, offsetX);
    if (x0 < offsetX) x0 =offsetX;
    dc.SetBrush(*wxWHITE_BRUSH);
    dc.DrawRectangle(x0, 0, x1-x0+1, mPanelHeight);
}

void CLS::
DrawCurrentTime(wxDC &dc, int offsetX){
    //if (offsetX == -1) offsetX = GetOffsetX();
    dc.SetPen(*wxRED_PEN);
    int x0 = GetDevCoordFromN(mTimeNormalized, offsetX);
    if (x0 > offsetX)
	dc.DrawRectangle(x0, 0, 1, mPanelHeight);
}

bool CLS::
SelectItem(int offsetX){
    mSelected = 0;
    CalcFrameDraggingFromPosition(offsetX);
    mpCAALink->mDraggingNormalTime = mQuantizedTimeDragging;
    
    int posX = mMousePosCurrentX;
    int posY = mMousePosCurrentY;

    if (mRectSlider.Contains(posX, posY)) mSelected      = _TL_SEL_SLIDER;
    else if (mRectMin.Contains(posX, posY)) mSelected    = _TL_SEL_MIN;
    else if (mRectMax.Contains(posX, posY)) mSelected    = _TL_SEL_MAX;
    else if (mRectRegion.Contains(posX, posY)) mSelected = _TL_SEL_REGION;
    
    if (mSelected == _TL_SEL_REGION) mpCAALink->SetRegion(mType);
    
    if (mSelected == _TL_SEL_SLIDER){
	if (mType != _TL_GRAPH)
	    mpCAALink->mDraggingItem = _TL_SEL_TIME;
	else
	    mpCAALink->mDraggingItem = _TL_SEL_GRAPH_NAVI;
    }
    if (mSelected == _TL_SEL_MIN && mType == _TL_WA)
	mpCAALink->mDraggingItem = _TL_SEL_WA_MIN;
    if (mSelected == _TL_SEL_MAX && mType == _TL_WA)
	mpCAALink->mDraggingItem = _TL_SEL_WA_MAX;
    if (mSelected == _TL_SEL_REGION && mType == _TL_WA)
	mpCAALink->mDraggingItem = _TL_SEL_WA_REGION;
    
	
    return (mSelected != 0);
}

void CLS::LimitDragging(){}

void CLS::
ProcMouseEventBase(wxMouseEvent& event, int offsetX){
    mMousePosCurrentX = event.GetX();
    mMousePosCurrentY = event.GetY();
    
    if (event.ButtonDown(wxMOUSE_BTN_LEFT)){
	if (SelectItem(offsetX) == TRUE)
	    CaptureMouse();
	else{
	    //
	}
    }

    if (event.Dragging()){
	CalcFrameDraggingFromPosition(offsetX);
	
	LimitDragging();
	
	if (mpCAALink->mDraggingItem != 0)
	    if (mpCAALink->mDraggingNormalTime != mQuantizedTimeDragging){
		mpCAALink->mDraggingNormalTime = mQuantizedTimeDragging;
		mpCParent->UpdatePanels(wxT("ALL"));
	    }
    }
}

void CLS::
ProcMouseReleaseBase(){
    if (HasCapture() == TRUE){
	ReleaseMouse();
	SetCursor(wxNullCursor);
    }
    mSelected = 0;
    mpCAALink->mDraggingItem = 0;
}

void CLS::
MoveTimeSlider(){
    if (mFrameDragging == mFrameCurrent) return ;
    if (GetPCTimeCtrl()->mCSubFrameCurrent.GetBodyINT() != 0){
	GetPCUndo()->OpenBlock("MoveTimeSlider");
	GetPCTimeCtrl()->SetCurrentFrameCommand(mFrameDragging);
	GetPCTimeCtrl()->SetCurrentSubFrameCommand(0);
	GetPCUndo()->CloseBlock("MoveTimeSlider");
    }
    else
	GetPCTimeCtrl()->SetCurrentFrameCommand(mFrameDragging);
}

void CLS::
MoveWAMin(){
    GetPCTimeCtrl()->SetWAMinCommand(mFrameDragging);
}

void CLS::
MoveWAMax(){
    GetPCTimeCtrl()->SetWAMaxCommand(mFrameDragging);
}

void CLS::
MoveWARegion(){
    GetPCTimeCtrl()->SetWARegionCommand(mFrameDragging, mpCAALink->mRegionLforDrag, mpCAALink->mRegionRforDrag);
}

void CLS::
MoveAAMin(){ // \݂̂ɊւϐȂ̂ Undo,Redo Ȃ
    mpCAALink->mCAAMin.SetBodyFromData((long long)mFrameDragging);
    mpCParent->UpdatePanels(wxT("ALL"));
}

void CLS::
MoveAAMax(){ // \݂̂ɊւϐȂ̂ Undo,Redo Ȃ
    mpCAALink->mCAAMax.SetBodyFromData((long long)mFrameDragging);
    mpCParent->UpdatePanels(wxT("ALL"));
}

void CLS::
MoveAARegion(){
    int width = mpCAALink->mCAAMax.GetBodyINT() - mpCAALink->mCAAMin.GetBodyINT();
    if (width < 2) width = 2;
    int frameMax = GetPCTimeCtrl()->mCFrameMax.GetBodyINT();
    if (mbForColorCurve == TRUE)
	frameMax = 1024;
    long long min = mFrameDragging - width / 2;
    long long max = mFrameDragging + width / 2;

    if (min < 0) min = 0;
    if (max > frameMax) max = frameMax;
    mpCAALink->mCAAMin.SetBodyFromData(min);
    mpCAALink->mCAAMax.SetBodyFromData(max);
    mpCParent->UpdatePanels(wxT("ALL"));
}

void CLS::
OnSize(wxSizeEvent &event){
    mPanelWidth = event.GetSize().GetWidth();
    mPanelHeight = event.GetSize().GetHeight();
    event.Skip();
}
