// (setq c-basic-offset 4)
// "clsTimelineParam.cpp"

#include "clsTimelineParam.h"
#include "clsTimeline.h"

#include "../DIALOG/clsParamDialog.h"
#include "../DIALOG/clsParamGroupDialog.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../FILTER/clsFilterBase.h"
#include "../LUA/clsLuaMiniCalc.h"
#include "../MISC/clsMisc.h"
#include "../PREFERENCE/clsShortcut.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../SCENEOBJ/clsAttribute.h"
#include "../SCENEOBJ/clsSceneObjCtrl.h"
#include "../SCENEOBJ/clsSceneObj.h"
#include "../SCENEOBJ/DATAVIEWER/clsVolumeBox.h"
#include "../SCENEOBJ/DATAVIEWER/clsParticle.h"
#include "../SCENEOBJ/DATAVIEWER/clsStreamline.h"

#include "../WIDGETCTRL/INPUTITEM/clsTextCtrlWithDrag.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"
#include "../WIDGETCTRL/WX/clsContainerObject.h"
#include "../WIDGETCTRL/WX/clsFontProvider.h"
#include "../WIDGETCTRL/WX/clsIconProvider.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"

#include "../../wxWidgetEnumMenu.h"

#define CLS TimelineParam_cls

BEGIN_EVENT_TABLE(TimelineParam_cls, TimelineBase_cls)
EVT_PAINT(CLS::OnPaint)
EVT_MENU(menuPopRename, CLS::OnRename)
EVT_MENU(menuPopInsertKey, CLS::OnInsertKey)
EVT_LEFT_UP    (CLS::OnLeftUp)
EVT_LEFT_DOWN  (CLS::OnLeftDown)
EVT_RIGHT_DOWN (CLS::OnRightDown)
EVT_MOUSEWHEEL (CLS::OnMouseWheel)
EVT_MOTION     (CLS::OnMouseMotion)
EVT_LEFT_DCLICK(CLS::OnDoubleClick)
EVT_LEAVE_WINDOW(CLS::OnMouseLeave)
EVT_KEY_DOWN   (CLS::OnKeyDown)
EVT_IDLE(CLS::OnIdle)
END_EVENT_TABLE()

CLS::
TimelineParam_cls(Timeline_cls *pCParent) : 
TimelineBase_cls(pCParent, _TL_PARAM){

    mpCParent = pCParent;
    
    mpCTgtPanel = this;
    mCursorCurrent = _CURSOR_NORMAL;

    mFuncRenamer = wxComEH(CLS::OnRenamer);
    mFuncRenamerFocus = wxFocusEH(CLS::OnRenamerFocus);
    mFuncRenamerKey = wxKeyEH(CLS::OnRenamerKey);


    mpCValueInput = NULL;
    mbValueInput = FALSE;
    mbKillValueInput = FALSE;
    mQueOpenNextValueInput = -1;
    
    IconProvider_cls *pCIconProv = GetPCIconProvider();
    for (int i = 0; i < _FONT_MAX; i++){
	mpCIconOpen[i] = pCIconProv->GetPCIcon("GROUP_OPEN", i);
	mpCIconClose[i] = pCIconProv->GetPCIcon("GROUP_CLOSE", i);
	mpCIconLink[i] = pCIconProv->GetPCIcon("LINK", i);
	mpCIconUnlink[i] = pCIconProv->GetPCIcon("UNLINK", i);
	mpCIconTimeP[i] = pCIconProv->GetPCIcon("TIME_P", i);
	mpCIconTimeN[i] = pCIconProv->GetPCIcon("TIME_N", i);
	mpCIconLeftP[i] = pCIconProv->GetPCIcon("LEFT_P", i);
	mpCIconLeftN[i] = pCIconProv->GetPCIcon("LEFT_N", i);
	mpCIconRightP[i] = pCIconProv->GetPCIcon("RIGHT_P", i);
	mpCIconRightN[i] = pCIconProv->GetPCIcon("RIGHT_N", i);
	mpCIconCheck[i] = pCIconProv->GetPCIcon("CHECK", i);
	mpCIconUncheck[i] = pCIconProv->GetPCIcon("UNCHECK", i);
	mpCIconKilroyP[i] = pCIconProv->GetPCIcon("KILROY_P", i);
	mpCIconKilroyN[i] = pCIconProv->GetPCIcon("KILROY_N", i);
	mpCIconExprP[i] = pCIconProv->GetPCIcon("EXPRESSION_P", i);
	mpCIconExprN[i] = pCIconProv->GetPCIcon("EXPRESSION_N", i);
	mpCIconExprE[i] = pCIconProv->GetPCIcon("EXPRESSION_E", i);
	mpCIconTextP[i] = pCIconProv->GetPCIcon("TEXT_BTN_P", i);
	mpCIconTextN[i] = pCIconProv->GetPCIcon("TEXT_BTN_N", i);
    }

    mCDisplayItem.SetLink(&mpCParent->mCMemoryDCWorkSpace, &mADisplayedRect);
    mCDisplayValue.SetLink(&mpCParent->mCMemoryDCWorkSpace, &mAValueRect);
    mCOpenClose.SetLink(&mAIconRectOpenClose);
    mCLink.SetLink(&mAIconRectLink);
    mCKilroy.SetLink(&mAIconRectKilroy);
    mCExpr.SetLink(&mAIconRectExpr);
    mCTextBtn.SetLink(&mAIconRectText);
    mCTimeDepend.SetLink(&mAIconRectTimeDepend);
    mCKeyLeft.SetLink(&mAIconRectKeyLeft);
    mCKeyCheck.SetLink(&mAIconRectKeyCheck);
    mCKeyRight.SetLink(&mAIconRectKeyRight);
    
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    ColorProvider_cls::SetBGColor(this, "EditPanel");

    //mOffsetX = 136;
    mBorderXR = 6 + 12;

    mbKillRenamer = FALSE;
    mbRename = FALSE;
    mbSashSelecting = FALSE;
    mbKeySelecting = FALSE;

    mpCPopupMenu = new wxMenu; 
    mpCPopupMenu->Append(menuPopRename, _("Rename"));
    mpCPopupMenu->Append(menuDelete, _("Delete"));

    mpCPopupMenuInsertKey = new wxMenu;
    mpCPopupMenuInsertKey->Append(menuPopInsertKey, _("Insert Key at Current Time"));
}

CLS::
~TimelineParam_cls(){
    delete mpCPopupMenu;
    delete mpCPopupMenuInsertKey;
}

void CLS::
ClearDisplayedItems(){
    mAstrDisplayedItem.clear();
    mADisplayedRect.clear();
    mAIconRectOpenClose.clear();
    mAIconRectLink.clear();
    mAIconRectKilroy.clear();
    mAIconRectExpr.clear();
    mAIconRectText.clear();
    mAIconRectTimeDepend.clear();
    mAIconRectKeyLeft.clear();
    mAIconRectKeyRight.clear();
    mAIconRectKeyCheck.clear();
    mAValueRect.clear();
    mAAKeyFrameRect.clear();
    mAAKeyFrameIndex.clear();
}

wxRect CLS::
GetTimelineRect(int Xexpand){
    int sashX = mpCParent->mCSashX.GetBodyINT();
    return wxRect(wxPoint(sashX + mBorderXL - Xexpand, 0), wxPoint(mPanelWidth - mBorderXR + Xexpand, mPanelHeight-1));
}

int CLS::
GetYPos(int line){ return line * mpCParent->GetLineHeight() + mOffsetY; }

double CLS::
GetMouseDNT(){
    double NTMouseCurrent = GetNFromDevCoord(mMouseXCurrent, GetOffsetX());
    double NTMouseAtDown = GetNFromDevCoord(mMouseXAtDown, GetOffsetX());
    return NTMouseCurrent - NTMouseAtDown;
}

int CLS::
GetOffsetX(){
    return mpCParent->mCSashX.GetBodyINT();
}

void CLS::
DrawParameter(wxDC& dc, wxString strItem, ParamGroup_cls* pCParamGroup, int& itemCount){
    wxString strParamGroup = pCParamGroup->mCstrName.GetBodySTR();
    
    uint paramNum = pCParamGroup->mVCParam.size();
    for (uint i = 0; i < paramNum; i++){
	Parameter_cls &CParam = pCParamGroup->mVCParam[i];

	wxString strParam = pCParamGroup->mVCParam[i].GetStrName();
	wxString strParamTag = strItem + wxT("::") + strParamGroup + wxT("::") + strParam;
	
	int indentPos = 30;
	int lineHeight =mpCParent->GetLineHeight();
	int yPos = GetYPos(itemCount);

	if (CParam.IsSelectedKeyExist() == TRUE)
	    dc.SetTextForeground(wxColour(255, 255, 0));
	else
	    dc.SetTextForeground(wxColour(0, 0, 0));
	mCDisplayItem.Draw(dc, strParam, indentPos, yPos);

	double time = mpCParent->GetTimeNormalized();
	if (mpCParent->mDraggingItem == _TL_SEL_TIME)
	    time = mpCParent->mDraggingNormalTime;
	
	bool bOnKey = CParam.IsKeyExist(time);
	bool bTimeDepend = CParam.mCbTimeDepend.GetBodyBOOL();
	if (bTimeDepend == TRUE){
	    if (bOnKey == TRUE) dc.SetTextForeground(wxColour(0, 0, 0));
	    else                dc.SetTextForeground(wxColour(0, 128, 0));
	}
	else dc.SetTextForeground(wxColour(64, 64, 64));

	int fSize = GetFontSize();
	wxString strValue; strValue << CParam.GetValue(time);
	mCDisplayValue.Draw(dc, strValue, indentPos + 28 + fSize*2, yPos);

	dc.SetTextForeground(wxColour(0, 0, 0));

	int sashX = mpCParent->mCSashX.GetBodyINT();
	wxSize sIcon = wxSize(lineHeight*0.8, lineHeight*0.8);
	int iconWidth = lineHeight*0.8;
	int posTimeDepend = indentPos-15-fSize*2;
	int posLeft = indentPos + 124 + Misc_cls::Max(sashX - _TL_SASH_MIN, 0) - lineHeight - fSize*2;
	int posCheck = posLeft + iconWidth;
	int posRight = posCheck + iconWidth;

	if (CParam.mCbTimeDepend.GetBodyBOOL() == TRUE){
	    mCTimeDepend.Draw(dc, mpCIconTimeP[fSize], posTimeDepend, yPos);
	    mCKeyLeft.Draw(dc, mpCIconLeftP[fSize], posLeft, yPos);
	    if (CParam.IsKeyExist(time))
		mCKeyCheck.Draw(dc, mpCIconCheck[fSize], posCheck, yPos);
	    else
		mCKeyCheck.Draw(dc, mpCIconUncheck[fSize], posCheck, yPos);
	    mCKeyRight.Draw(dc, mpCIconRightP[fSize], posRight, yPos);

	    if (CParam.mCbExtrapolate.GetBodyBOOL() == TRUE){
		dc.DrawText("E", posLeft - 12, yPos);
	    }

	}
	else{
	    mCTimeDepend.Draw(dc, mpCIconTimeN[fSize], posTimeDepend, yPos);
	    mCKeyLeft.Draw(dc, mpCIconLeftN[fSize], posLeft, yPos);
	    mCKeyCheck.Draw(dc, mpCIconUncheck[fSize], posCheck, yPos);
	    mCKeyRight.Draw(dc, mpCIconRightN[fSize], posRight, yPos);
	}

	mCOpenClose.DrawEmpty();
	mCLink.DrawEmpty();
	mCKilroy.DrawEmpty();
	mCExpr.DrawEmpty();
	mCTextBtn.DrawEmpty();
		
	mAstrDisplayedItem.push_back(strParamTag);
	
	std::vector < wxRect > AKeyFrameRect;
	std::vector < int > AKeyFrameIndex;
	
	int panelWidth = mPanelWidth-12;
	if (yPos + lineHeight > 0 && yPos < mPanelHeight){
	    double dX = 0;
	    if (mbKeySelecting == TRUE)	dX = GetMouseDNT();

	    CParam.DrawTL(dc, this, 
			  mpCParent->mCBitmapBuffer, dX,
			  sashX, yPos, panelWidth, lineHeight,
			  mNormalizedWAMin, mNormalizedWAMax,
			  AKeyFrameRect, AKeyFrameIndex);
	}
		
	mAAKeyFrameRect.push_back(AKeyFrameRect);
	mAAKeyFrameIndex.push_back(AKeyFrameIndex);

	itemCount++;
    }
}

void CLS::
DrawAttribute(wxDC& dc, wxString strItem,
	      std::list < Attribute_cls* > &VpCAttribute,
	      int& itemCount, int indentPos){
    dc.SetTextForeground(wxColour(0, 0, 0));

    indentPos+=16;
    for (std::list< Attribute_cls*>::iterator itr = VpCAttribute.begin();
	 itr != VpCAttribute.end(); itr++){
	wxString strAttribName = (*itr)->mCstrName.GetBodySTR();
	wxString strAttribTag = strItem + wxT("::") + strAttribName;

	bool bShy = (*itr)->mCbShy.GetBodyBOOL();
	bool bTimeDepend = (*itr)->IsParamGroupTimeDepend();
	if (mpCParent->mCbKilroy.GetBodyBOOL() && bShy == TRUE)
	    continue;
	if (mpCParent->mCbTimeDepend.GetBodyBOOL() && bTimeDepend == FALSE)
	    continue;
	
	int lineHeight = mpCParent->GetLineHeight();
	int yPos = GetYPos(itemCount);
	
	bool bOpend = Misc_cls::DoesExistInArray(mAstrOpenedItem, strAttribTag);
	
	int w = mCDisplayItem.Draw(dc, strAttribName, indentPos + lineHeight, yPos);
	mCDisplayValue.DrawEmpty();

	wxSize sIcon = wxSize(lineHeight*0.8, lineHeight*0.8);
	int fSize = mpCParent->mCFontSize.GetBodyINT();
	int folderPosX = indentPos-2-lineHeight*0.5;
	int folderPosY = yPos+lineHeight*0.2;
	if (bOpend == FALSE)
	    mCOpenClose.Draw(dc, mpCIconClose[fSize], folderPosX, folderPosY);
	else
	    mCOpenClose.Draw(dc, mpCIconOpen[fSize], folderPosX, folderPosY);

	mCTimeDepend.DrawEmpty();
	mCLink.DrawEmpty();
	
	int groupPosX = indentPos/*+w+4*/;
	
	if (bShy == FALSE)
	    mCKilroy.Draw(dc, mpCIconKilroyN[fSize], groupPosX, folderPosY);
	else
	    mCKilroy.Draw(dc, mpCIconKilroyP[fSize], groupPosX, folderPosY);
	
	mCExpr.DrawEmpty();
	mCTextBtn.DrawEmpty();
	mCKeyLeft.DrawEmpty();
	mCKeyCheck.DrawEmpty();
	mCKeyRight.DrawEmpty();
	mAAKeyFrameRect.push_back(msAEmptyRect);
	mAAKeyFrameIndex.push_back(msAEmptyIndex);

	mAstrDisplayedItem.push_back(strAttribTag);

	itemCount++;
	if (bOpend == TRUE)
	    DrawParamGroup(dc, strAttribTag, (*itr)->mVpCParamGroup, itemCount, indentPos);
    }
}

void CLS::
DrawParamGroup(wxDC& dc, wxString strItem, std::vector < ParamGroup_cls* > &VpCParamGroup, int& itemCount, int indentPos){
    dc.SetTextForeground(wxColour(0, 0, 0));

    indentPos += 16;
    uint paramGroupNum = VpCParamGroup.size();
    for (uint i = 0; i < paramGroupNum; i++){
	wxString strParamGroup = VpCParamGroup[i]->mCstrName.GetBodySTR();
	wxString strParamTag = strItem + wxT("::") + strParamGroup;

	int lineHeight = mpCParent->GetLineHeight();
	int yPos = GetYPos(itemCount);

	bool bShy = VpCParamGroup[i]->mCbShy.GetBodyBOOL();
	bool bTimeDepend = VpCParamGroup[i]->IsTimeDepend();
	bool bExpress = VpCParamGroup[i]->mCbExpression.GetBodyBOOL();
	if (mpCParent->mCbKilroy.GetBodyBOOL() && bShy == TRUE)
	    continue;
	if (mpCParent->mCbTimeDepend.GetBodyBOOL() && bTimeDepend == FALSE)
	    continue;
    
	bool bOpend = Misc_cls::DoesExistInArray(mAstrOpenedItem, strParamTag);
	bool bLinked = VpCParamGroup[i]->mCbLinked.GetBodyBOOL();
		
	int w = mCDisplayItem.Draw(dc, strParamGroup, indentPos + lineHeight, yPos);
	mCDisplayValue.DrawEmpty();

	wxSize sIcon = wxSize(lineHeight*0.8, lineHeight*0.8);
	int fSize = mpCParent->mCFontSize.GetBodyINT();
	int folderPosX = indentPos-2-lineHeight*0.5;
	int folderPosY = yPos+lineHeight*0.2;
	if (bOpend == FALSE)
	    mCOpenClose.Draw(dc, mpCIconClose[fSize], folderPosX, folderPosY);
	else
	    mCOpenClose.Draw(dc, mpCIconOpen[fSize], folderPosX, folderPosY);

	if (bShy == FALSE)
	    mCKilroy.Draw(dc, mpCIconKilroyN[fSize], indentPos, folderPosY);
	else
	    mCKilroy.Draw(dc, mpCIconKilroyP[fSize], indentPos, folderPosY);
	
	int groupPosX = indentPos+lineHeight+w+4;
	if (bLinked == FALSE)
	    mCLink.Draw(dc, mpCIconUnlink[fSize], groupPosX, folderPosY);
	else
	    mCLink.Draw(dc, mpCIconLink[fSize], groupPosX, folderPosY);

	if (bExpress == FALSE){
	    mCExpr.Draw(dc, mpCIconExprN[fSize], groupPosX+14+fSize*2, folderPosY);
	    mCTextBtn.DrawEmpty();
	}
	else{
	    if (VpCParamGroup[i]->mpLua == NULL)
		mCExpr.Draw(dc, mpCIconExprE[fSize], groupPosX+14+fSize*2, folderPosY);
	    else
		mCExpr.Draw(dc, mpCIconExprP[fSize], groupPosX+14+fSize*2, folderPosY);
		
	    if (strParamTag == mstrPushedItem)
		mCTextBtn.Draw(dc, mpCIconTextP[fSize], groupPosX+28+fSize*3, folderPosY);
	    else
		mCTextBtn.Draw(dc, mpCIconTextN[fSize], groupPosX+28+fSize*3, folderPosY);
	}

	mCTimeDepend.DrawEmpty();
	mCKeyLeft.DrawEmpty();
	mCKeyCheck.DrawEmpty();
	mCKeyRight.DrawEmpty();
	mAAKeyFrameRect.push_back(msAEmptyRect);
	mAAKeyFrameIndex.push_back(msAEmptyIndex);
	
	mAstrDisplayedItem.push_back(strParamTag);

	int sashX = mpCParent->mCSashX.GetBodyINT();
	int panelWidth = mPanelWidth-12;
	if (yPos + lineHeight > 0 && yPos < mPanelHeight)
	    VpCParamGroup[i]->DrawTL
		(dc, this, mpCParent->mCBitmapBuffer,
		 sashX, yPos, panelWidth, lineHeight,
		 mNormalizedWAMin, mNormalizedWAMax);
	
	itemCount++;
	if (bOpend == TRUE)
	    DrawParameter(dc, strItem, VpCParamGroup[i], itemCount);
    }
}

void CLS::
DrawItem(wxDC& dc, ParentObj_cls* pCParent, int& itemCount,
	 SceneObj_cls *pCActiveObject, int depth){
    
    for (std::list< SceneObj_cls*>::iterator itr = pCParent->mVpCObj.begin();
	 itr != pCParent->mVpCObj.end(); itr++){
	
	SceneObj_cls* pCObj = *itr;

	bool bShy = pCObj->mCbShy.GetBodyBOOL();
	bool bTimeDepend = 
	    pCObj->IsAttributesTimeDepend() |
	    pCObj->IsParamGroupTimeDepend();
	bool bSelected = GetPCSceneObjCtrl()->IsSelected(pCObj);

	// q̕\ւ͐i܂ȂƑʖ
	if (mpCParent->mCbKilroy.GetBodyBOOL() && bShy == TRUE){
	    DrawItem(dc, pCObj, itemCount, pCActiveObject, depth+1);
	    continue;
	}

	if (mpCParent->mCbTimeDepend.GetBodyBOOL() && bTimeDepend == FALSE){
	    DrawItem(dc, pCObj, itemCount, pCActiveObject, depth+1);
	    continue;
	}
	if (mpCParent->mCbSelectOnly.GetBodyBOOL() && bSelected == FALSE){
	    
	    DrawItem(dc, pCObj, itemCount, pCActiveObject, depth+1);
	    continue;
	}
		
	wxString strItem = pCObj->mCstrName.mstrBody.Clone();
	int indentPos = 16;
	int lineHeight =mpCParent->GetLineHeight();
	int yPos = GetYPos(itemCount);
	
	bool bEnhance = FALSE;
	if (pCObj == pCActiveObject) bEnhance = TRUE;

	
	bool bOpend = Misc_cls::DoesExistInArray(mAstrOpenedItem, strItem);
	
	ColorProvider_cls::SetTextColorDC(dc, bSelected);
	
	int w = 0;
	if (! (mbRename == TRUE && pCObj == pCActiveObject)){
	    w = mCDisplayItem.Draw(dc, strItem, indentPos + lineHeight, yPos, bEnhance);
	}
	else
	    mCDisplayItem.DrawEmpty();
	mCDisplayValue.DrawEmpty();

	wxString strIcon = wxES; 
	if (strIcon.IsEmpty() == FALSE){
	    dc.DrawBitmap(ARTS(strIcon, wxSize(24, 24)), 3, yPos);
	}

	wxSize sIcon = wxSize(lineHeight*0.8, lineHeight*0.8);
	int fSize = mpCParent->mCFontSize.GetBodyINT();
	int folderPosX = indentPos-2-lineHeight*0.5;
	int folderPosY = yPos+lineHeight*0.2;
	if (bOpend == FALSE)
	    mCOpenClose.Draw(dc, mpCIconClose[fSize], folderPosX, folderPosY);
	else
	    mCOpenClose.Draw(dc, mpCIconOpen[fSize], folderPosX, folderPosY);
	
	mAstrDisplayedItem.push_back(strItem);
	mCLink.DrawEmpty();
	int kilroyPosX = indentPos/*+w+4*/;
	if (bShy == FALSE)
	    mCKilroy.Draw(dc, mpCIconKilroyN[fSize], kilroyPosX, folderPosY);
	else
	    mCKilroy.Draw(dc, mpCIconKilroyP[fSize], kilroyPosX, folderPosY);
	
	mCExpr.DrawEmpty();
	mCTextBtn.DrawEmpty();
	mCTimeDepend.DrawEmpty();
	mCKeyLeft.DrawEmpty();
	mCKeyCheck.DrawEmpty();
	mCKeyRight.DrawEmpty();
	mAAKeyFrameRect.push_back(msAEmptyRect);
	mAAKeyFrameIndex.push_back(msAEmptyIndex);

	int sashX = mpCParent->mCSashX.GetBodyINT();
	int panelWidth = mPanelWidth-12;
	if (yPos + lineHeight > 0 && yPos < mPanelHeight){
	    wxString strClassName = pCObj->mstrClassName;
	    if (strClassName.IsSameAs("VolumeBox_cls"))
		((VolumeBox_cls*)pCObj)->DrawTL
		    (dc, this, mpCParent->mCBitmapBuffer,
		     sashX, yPos, panelWidth, lineHeight,
		     mNormalizedWAMin, mNormalizedWAMax);
	    if (strClassName.IsSameAs("Particle_cls"))
		((Particle_cls*)pCObj)->DrawTL
		    (dc, this, mpCParent->mCBitmapBuffer,
		     sashX, yPos, panelWidth, lineHeight,
		     mNormalizedWAMin, mNormalizedWAMax);
	    if (strClassName.IsSameAs("Streamline_cls"))
		((Streamline_cls*)pCObj)->DrawTL
		    (dc, this, mpCParent->mCBitmapBuffer,
		     sashX, yPos, panelWidth, lineHeight,
		     mNormalizedWAMin, mNormalizedWAMax);
	}
	    

	itemCount++;
	if (bOpend == TRUE) DrawParamGroup(dc, strItem, pCObj->mVpCParamGroup, itemCount, indentPos);
	if (bOpend == TRUE){
	    DrawAttribute(dc, strItem, pCObj->mVpCAttribute, itemCount, indentPos);
	    DrawAttribute(dc, strItem, pCObj->mVpCDataProc, itemCount, indentPos);
	}
	
	DrawItem(dc, pCObj, itemCount, pCActiveObject, depth+1);
    }
}

void CLS::
DrawFilter(wxDC& dc, int& itemCount, int depth){

    if (GetPCFilterCtrl()->mVpCFilter.size() == 0) return;

    wxString strItem = "Scene Filter";
    int indentPos = 16;
    int lineHeight =mpCParent->GetLineHeight();
    int yPos = GetYPos(itemCount);
	
    bool bEnhance = FALSE;
    bool bSelected = FALSE;
    bool bOpend = Misc_cls::DoesExistInArray(mAstrOpenedItem, strItem);

    ColorProvider_cls::SetTextColorDC(dc, bSelected);

    mCDisplayItem.Draw(dc, strItem, indentPos, yPos, bEnhance);
    mCDisplayValue.DrawEmpty();

    wxString strIcon = wxES; 
    if (strIcon.IsEmpty() == FALSE){
	dc.DrawBitmap(ARTS(strIcon, wxSize(24, 24)), 3, yPos);
    }

    wxSize sIcon = wxSize(lineHeight*0.8, lineHeight*0.8);
    int fSize = mpCParent->mCFontSize.GetBodyINT();
    int folderPosX = indentPos-2-lineHeight*0.5;
    int folderPosY = yPos+lineHeight*0.2;
    if (bOpend == FALSE)
	mCOpenClose.Draw(dc, mpCIconClose[fSize], folderPosX, folderPosY);
    else
	mCOpenClose.Draw(dc, mpCIconOpen[fSize], folderPosX, folderPosY);
    
    mAstrDisplayedItem.push_back(strItem);
    mCLink.DrawEmpty();
    mCKilroy.DrawEmpty();
    mCExpr.DrawEmpty();
    mCTextBtn.DrawEmpty();
    mCTimeDepend.DrawEmpty();
    mCKeyLeft.DrawEmpty();
    mCKeyCheck.DrawEmpty();
    mCKeyRight.DrawEmpty();
    mAAKeyFrameRect.push_back(msAEmptyRect);
    mAAKeyFrameIndex.push_back(msAEmptyIndex);
	
    itemCount++;

    if (bOpend == TRUE) DrawAttribute(dc, strItem, GetPCFilterCtrl()->mVpCFilter, itemCount, indentPos);
}

int CLS::
GetFontSize(){
    return mpCParent->mCFontSize.GetBodyINT();
}

wxString CLS::
GetParamGroupTagLong(wxString strTag){
    ParamGroup_cls *pCParamGroup = GetPCMain()->GetPCParamGroup(strTag);
    if (pCParamGroup == NULL) return wxES;

    wxString strTagLong;
    if (strTag.Find("Scene Filter") == 0){
	PopStrItemFromStrLine(strTag, "::");
	strTagLong = GetPCFilterCtrl()->GetStrSelf() << strTag;
    }
    else
	strTagLong = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strTag;
    return strTagLong;
}

void CLS::
SelectIfObject(wxString strTag, bool bAdditional){
    if (strTag.Find("::") == wxNOT_FOUND)
	GetPCSceneObjCtrl()->Select(strTag, bAdditional);
}

void CLS::
SelectParameter(wxString strTag){
    Parameter_cls *pCParam = GetPCMain()->GetPCParameter(strTag);
    if (pCParam == NULL){
	/*
	ParamGroup_cls *pCParamGroup = GetPCMain()->GetPCParamGroup(strTag);
	if (pCParamGroup == NULL) return;

	wxString strTagLong;
	if (strTag.Find("Scene Filter") == 0){
	    PopStrItemFromStrLine(strTag, "::");
	    strTagLong = GetPCFilterCtrl()->GetStrSelf() << strTag;
	}
	else
	    strTagLong = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strTag;
	*/
	wxString strTagLong = GetParamGroupTagLong(strTag);
	if (strTagLong.IsEmpty() == FALSE)
	    GetPCMain()->ProcAfterMath("SelectParamGroup", strTagLong);
	return;
    }
    pCParam->SelectKeyAll(FALSE);
}

bool CLS::
SelectItemByMouse(bool bAdditional, bool bMultiLine){

    mbKeySelecting = FALSE;
    mbBarSelecting = FALSE;
    if (mMouseXAtDown > mTgtPanelWidth - _SCROLL_BARWIDTH){
	mbBarSelecting = TRUE;
	Refresh(FALSE);
	return TRUE;
    }

    if (mMouseXAtDown > GetOffsetX() - 4 && mMouseXAtDown < GetOffsetX() + 4){
	mbSashSelecting = TRUE;
	Refresh(FALSE);
	return TRUE;
    }

    int index = GetIndex(mpCParent->GetLineHeight());
    if (index < 0 || index >= (int)mADisplayedRect.size()) {
	return TRUE;
    }

    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    if (mAIconRectOpenClose[index].Contains(CAtDown)){
	ToggleOpenClose(mAstrDisplayedItem[index]); return TRUE; }
    
    if (mAIconRectLink[index].Contains(CAtDown)){
	ToggleLink(mAstrDisplayedItem[index]); return TRUE; }

    if (mAIconRectKilroy[index].Contains(CAtDown)){
	ToggleKilroy(mAstrDisplayedItem[index]); return TRUE; }

    if (mAIconRectExpr[index].Contains(CAtDown)){
	ToggleExpr(mAstrDisplayedItem[index]); return TRUE; }

    if (mAIconRectText[index].Contains(CAtDown)){
	PushText(mAstrDisplayedItem[index]); return TRUE; }

    if (mAIconRectTimeDepend[index].Contains(CAtDown)){
	ToggleTimeDepend(mAstrDisplayedItem[index]); return TRUE; }

    if (mAIconRectKeyLeft[index].Contains(CAtDown)){
	MoveKey(-1, mAstrDisplayedItem[index]); return TRUE; }
    if (mAIconRectKeyRight[index].Contains(CAtDown)){
	MoveKey(+1, mAstrDisplayedItem[index]); return TRUE; }
    
    if (mAIconRectKeyCheck[index].Contains(CAtDown)){
	ToggleKey(mAstrDisplayedItem[index]); return TRUE; }

    if (mADisplayedRect[index].Contains(CAtDown)){
	if (bMultiLine == TRUE && mLastSelectIndex != -1){
	    int min = mLastSelectIndex, max = index;
	    if (min > max) Misc_cls::Swap(&min, &max);
	    for (int i = min; i <= max; i++)
		SelectIfObject(mAstrDisplayedItem[i], bAdditional);
	}
	SelectIfObject(mAstrDisplayedItem[index], bAdditional);
	mLastSelectIndex = index;

	SelectParameter(mAstrDisplayedItem[index]); 
	return TRUE;
    }

    wxString strTag = mAstrDisplayedItem[index];

    bool bSelectedKeySelected = FALSE;
    bool bKeySelected = FALSE;
    uint keyNum = mAAKeyFrameRect[index].size();
    Parameter_cls *pCSelected = GetPCMain()->GetPCParameter(strTag);
    ParamGroup_cls *pCParamGroup = NULL;
    if (pCSelected != NULL) pCParamGroup = pCSelected->mpCParent;

    for (int i = 0; i < keyNum; i++){
	if (mAAKeyFrameRect[index][i].Contains(CAtDown)){
	    int keyIndex = mAAKeyFrameIndex[index][i]; 
	    bool bSelectedAlready = pCSelected->mVCKeys[keyIndex].mbSelected;
	    if (bSelectedAlready == TRUE && bAdditional == FALSE){
		bKeySelected = TRUE;
		bSelectedKeySelected = TRUE;
	    }
	    else
		bKeySelected |= GetPCMain()->
		    SelectKeyFrame(strTag, keyIndex, ! bSelectedAlready, bAdditional);
	    

	    GetPCMain()->
		ChangeKeyTypeByModKey(strTag, keyIndex, mbControlDown, mbShiftDown, mbAltDown);
	       

	    if (pCSelected->mVCKeys[keyIndex].mbSelected == FALSE){
		Refresh();
		return TRUE;
	    }
	}
    }
    
    if (bKeySelected == FALSE){
	if (bAdditional == FALSE){
	    GetPCMain()->DeselectKeyFrame();
	    SendUpdate();
	}
	return FALSE; 
    }
    else{
	if (bAdditional == FALSE && bSelectedKeySelected == FALSE){
	    wxString strItem = mAstrDisplayedItem[index];
	    wxString strObjName = PopStrItemFromStrLine(strItem, wxT("::"));
	    wxString strParamGroup = PopStrItemFromStrLine(strItem, wxT("::"));
	    if (strParamGroup.IsSameAs(pCParamGroup->GetStrName()) == FALSE)
		strParamGroup << "::" << PopStrItemFromStrLine(strItem, wxT("::"));
	    GetPCMain()->DeselectKeyFrame(strObjName + wxT("::") + strParamGroup);
	}
	SendUpdate();
    }

    mbKeySelecting = bKeySelected;
    
    return TRUE;
}

void CLS::
SelectItemByRegion(){
    uint lineNum = mAAKeyFrameIndex.size();
    for (uint index = 0; index < lineNum; index++){
	uint keyNum = mAAKeyFrameRect[index].size();
	for (uint i = 0; i < keyNum; i++){
	    wxRect CRectKey = mAAKeyFrameRect[index][i];
	    CRectKey.Deflate(1, 1);
	    if (mCRectSelectRegion.Contains(CRectKey)){
		int keyIndex = mAAKeyFrameIndex[index][i]; 
		GetPCMain()->
		    SelectKeyFrame(mAstrDisplayedItem[index], keyIndex, TRUE, TRUE);
	    }
	}
    }
    SendUpdate();
}

void CLS::
ConstructValueTextCtrl(int index){
    if (index >= mAstrDisplayedItem.size()) return;
    wxString strItem = mAstrDisplayedItem[index];
    Parameter_cls *pCParamer = GetPCMain()->GetPCParameter(strItem);
    if (pCParamer == NULL) return;

    strItem << wxT("::") << index;
    
    mbValueInput = TRUE;
    wxString strIni; strIni << pCParamer->GetCurrentValue();
    wxPoint pos = mAValueRect[index].GetPosition();
    wxSize size = mAValueRect[index].GetSize();
    int x = pos.x - 2, y = pos.y - 2;
    int width = size.GetWidth() + 16;
    int height = size.GetHeight();
    int fSize = GetFontSize();
    if (width < 46 + fSize * 4) width = 46 + fSize * 4;

    mpCValueInput = new TextCtrlWithDrag_cls
	(this, wxID_ANY, strIni,
	 wxPoint(x, y), wxSize(width, height),
	 wxTE_PROCESS_ENTER | wxTE_PROCESS_TAB);
    
    mpCValueInput->mDragScale = 1;
    /*
    mpCValueInput = new wxTextCtrl
	(this, wxID_ANY, strIni,
	 wxPoint(x, y), wxSize(width, height),
	 wxTE_PROCESS_ENTER | wxTE_PROCESS_TAB);
    */
    
    mpCValueInput->SetFont(FontProvider_cls::GetFont(GetFontSize()));
    mpCValueInput->SetFocus();
    mpCValueInput->SelectAll();

    Connect(mpCValueInput->GetId(), wxEVT_TEXT_ENTER, wxComEH(CLS::OnValueInput),  new ContainerObject_cls(strItem));
    mpCValueInput->Connect(wxEVT_KILL_FOCUS, wxFocusEH(CLS::OnValueInputFocus), new ContainerObject_cls(strItem), this);
    mpCValueInput->Connect(wxEVT_KEY_DOWN, wxKeyEH(CLS::OnValueInputKey), new ContainerObject_cls(strItem), this);

    mpCParent->mbSkipKeyHook = TRUE;
}

void CLS::
ToggleOpenClose(wxString strItem){
    if (Misc_cls::DoesExistInArray(mAstrOpenedItem, strItem) == TRUE)
	mAstrOpenedItem.Remove(strItem);
    else
	mAstrOpenedItem.Add(strItem);
}

void CLS::
ToggleLink(wxString strItem){
    ParamGroup_cls *pCParamGroup = GetPCMain()->GetPCParamGroup(strItem);
    bool bLinkOld = pCParamGroup->mCbLinked.GetBodyBOOL();
    GetPCMain()->SetParamGroupConfCommand(strItem, "Linked=", !bLinkOld);
}

void CLS::
ToggleKilroy(wxString strItem){
    SceneObj_cls *pCObj = GetPCSceneObjCtrl()->GetPCFromName(strItem);
    if (pCObj != NULL){
	bool bShy = pCObj->mCbShy.GetBodyBOOL();
	wxString strData; strData << !bShy;
	pCObj->mCbShy.PushCommandPair(strData);
	return;
    }

    ParamGroup_cls *pCParamGroup = GetPCMain()->GetPCParamGroup(strItem);
    if (pCParamGroup != NULL){
	bool bShyOld = pCParamGroup->mCbShy.GetBodyBOOL();
	GetPCMain()->SetParamGroupConfCommand(strItem, "Shy=", !bShyOld);
	return;
    }

    Attribute_cls *pCAttrib = GetPCSceneObjCtrl()->GetPCAttributeFromAddress(strItem);
    if (pCAttrib != NULL){
	bool bShy = pCAttrib->mCbShy.GetBodyBOOL();
	wxString strData; strData << !bShy;
	pCAttrib->mCbShy.PushCommandPair(strData);
	return;
    }

    FilterBase_cls *pCFilter = GetPCFilterCtrl()->GetPCFromAddress(strItem);
    if (pCFilter != NULL){
	bool bShy = pCFilter->mCbShy.GetBodyBOOL();
	wxString strData; strData << !bShy;
	pCFilter->mCbShy.PushCommandPair(strData);
	return;
    }
}

void CLS::
ToggleExpr(wxString strItem){
    ParamGroup_cls *pCParamGroup = GetPCMain()->GetPCParamGroup(strItem);
    bool bExprOld = pCParamGroup->mCbExpression.GetBodyBOOL();
    GetPCMain()->SetParamGroupConfCommand(strItem, "UseExpr=", !bExprOld);
}

void CLS::
PushText(wxString strItem){
    mstrPushedItem = strItem;

    wxString strTagLong = GetParamGroupTagLong(strItem);
    if (strTagLong.IsEmpty() == FALSE)
	GetPCMain()->ProcAfterMath("SelectExpression", strTagLong);

    Refresh();
}

void CLS::
ToggleTimeDepend(wxString strItem){
    Parameter_cls *pCParamer = GetPCMain()->GetPCParameter(strItem);
    
    bool bTimeDependOld = pCParamer->mCbTimeDepend.GetBodyBOOL();
    GetPCMain()->SetParameterConfCommand(strItem, "TimeDepend=", !bTimeDependOld);
}

void CLS::
ToggleKey(wxString strItem){
    Parameter_cls *pCParamer = GetPCMain()->GetPCParameter(strItem);
    if (pCParamer == NULL) return;
    if (pCParamer->mCbTimeDepend.GetBodyBOOL() == FALSE) return;
    double currentNT = GetPCTimeCtrl()->GetCurrentNT();
    bool bKey = pCParamer->IsKeyExist(currentNT);

    wxString strTag;
    if (strItem.Find("Scene Filter") == 0){
	PopStrItemFromStrLine(strItem, "::");
	strTag = GetPCFilterCtrl()->GetStrSelf() << strItem;
    }
    else
	strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strItem;
   
    if (bKey == FALSE){
	int keyType; 
	double easeFactor;
		GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);
	pCParamer->InsertKeyCommand(currentNT, strTag, 
				    keyType, easeFactor, FALSE);
    }
    else
	pCParamer->DeleteKeyCommand(currentNT, strTag, FALSE);
}

void CLS::
MoveKey(int direction, wxString strItem){
    Parameter_cls *pCParamer = GetPCMain()->GetPCParameter(strItem);
    if (pCParamer == NULL) return;
    double currentNT = GetPCTimeCtrl()->GetCurrentNT();
    int keyIndex = pCParamer->GetAdjuscentKeyIndex(direction);
    
    double newNT = pCParamer->GetKeyXNormalizeEX(keyIndex);
    if (currentNT != newNT)
	GetPCTimeCtrl()->SetCurrentNTCommand(newNT);
}

/*
void CLS::
ProcRename(const wxString& strNameOld, const wxString& strNameNew){
    uint num = mAstrOpenedItem.size();
    for (uint i = 0; i < num; i++){
	if (mAstrOpenedItem[i].Find(strNameOld + wxT("::")) == 0)
	    mAstrOpenedItem[i].Replace(strNameOld + wxT("::"), strNameNew + wxT("::"));
    }
    
    if (Misc_cls::DoesExistInArray(mAstrOpenedItem, strNameOld) == FALSE) return;
    mAstrOpenedItem.Remove(strNameOld);
    mAstrOpenedItem.Add(strNameNew);
}

void CLS::
ProcDelete(const wxString& strTag){
    if (Misc_cls::DoesExistInArray(mAstrOpenedItem, strTag) == FALSE) return;
    mAstrOpenedItem.Remove(strTag);
}
*/

void CLS::
ProcAfterMath(const wxString& strCommand, wxString strItem){
    if (strCommand.IsSameAs("RenameObject")){
	wxString strNameOld = PopStrItemFromStrLine(strItem, "::");
	wxString strNameNew = strItem;
	uint num = mAstrOpenedItem.size();
	for (uint i = 0; i < num; i++){
	    if (mAstrOpenedItem[i].Find(strNameOld + wxT("::")) == 0)
		mAstrOpenedItem[i].Replace(strNameOld + wxT("::"), strNameNew + wxT("::"));
	}
	
	if (Misc_cls::DoesExistInArray(mAstrOpenedItem, strNameOld) == FALSE) return;
	mAstrOpenedItem.Remove(strNameOld);
	mAstrOpenedItem.Add(strNameNew);
    }
    if (strCommand.IsSameAs("DeleteObject") ||
	strCommand.IsSameAs("DeleteModifier")){
	if (Misc_cls::DoesExistInArray(mAstrOpenedItem, strItem) == FALSE) return;
	mAstrOpenedItem.Remove(strItem);
    }
}

void CLS::
MotionSash(wxMouseEvent & event){
    long long sash = (long long)event.GetX();
    if (sash < _TL_SASH_MIN) sash = _TL_SASH_MIN;
    if (sash > mPanelWidth - _TL_SASH_MIN) sash = mPanelWidth - _TL_SASH_MIN;
    mpCParent->mCSashX.SetBodyFromData(sash);
    mpCParent->UpdatePanels(wxT("ALL"));
}

void CLS::
SendUpdate(){
    GetPCMain()->UpdatePanels("TIMELINE::GRAPH::PROPERTY_MATERIAL");
}

void CLS::
UpdateDisplay(wxDC& dc){
    UpdateState();
    SetTLRegion(mAAMinNormalized, mAAMaxNormalized);

    dc.SetTextForeground(wxColour(255, 255, 255));
    dc.SetFont(FontProvider_cls::GetFont(GetFontSize()));
    mpCParent->mCMemoryDCWorkSpace.SetFont(FontProvider_cls::GetFont(GetFontSize()));

    dc.SetPen(*wxTRANSPARENT_PEN);
    dc.SetBrush(GetBackgroundColour());
    //dc.DrawRectangle(0, 0, mTgtPanelWidth, mTgtPanelHeight);
    dc.SetBackground(wxBrush(GetBackgroundColour()));
    dc.Clear();
    
    dc.SetBrush(ColorProvider_cls::GetColor("Active"));

    SceneObjCtrl_cls *pCObjCtrl = GetPCSceneObjCtrl();

    int itemCount = 0;
    ClearDisplayedItems();
    
    DrawItem(dc, pCObjCtrl, itemCount, pCObjCtrl->mpCActiveObject, 0);
    DrawFilter(dc, itemCount, 0);
    mVirtualHeight = mpCParent->GetLineHeight() * itemCount;

    mAstrDisplayedItem.push_back(wxT("End"));

    dc.SetPen(*wxTRANSPARENT_PEN);
    dc.SetBrush(GetBackgroundColour());
    
    if (mbSashSelecting == TRUE)
	dc.SetPen(wxColour(240, 255, 255));
    else
	dc.SetPen(wxColour(0, 0, 0));
	
    dc.DrawRectangle(GetOffsetX(), 0, 1, mPanelHeight);

    DrawScrollBar(dc, mbDraggingRegion);
    DrawDraggingRegion(dc, GetTimelineRect(4));
    DrawCurrentTime(dc, GetOffsetX());
}

void CLS::
OnPaint(wxPaintEvent& event){
    wxAutoBufferedPaintDC dc(this);
    
    GetClientSize(&mPanelWidth, &mPanelHeight);
    GetClientSize(&mTgtPanelWidth, &mTgtPanelHeight);
    
    UpdateDisplay(dc);
}

void CLS::
OnLeftDown(wxMouseEvent& event){
    if (mbValueInput == TRUE){ 	
	SendEvent(this, wxEVT_TEXT_ENTER, mpCValueInput->GetId());
	return ; 
    }
    if (mbRename == TRUE){
	SendEvent(this, wxEVT_TEXT_ENTER, mpCRenamer->GetId());
	return ;
    }

    LeftDownBase(event);

    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    int index = GetIndex(mpCParent->GetLineHeight());
    if (index > 0 && index < mAValueRect.size())
	if (mAValueRect[index].Contains(CAtDown) == TRUE){
	    ConstructValueTextCtrl(index);
	    return;
	}
    
    bool bAdditional = event.m_shiftDown;
    bool bMultiLine = event.m_controlDown;
    mbDraggingRegion = FALSE;
    if (SelectItemByMouse(bAdditional, bMultiLine) == FALSE)
	mbDraggingRegion = TRUE;

    SetFocus();
}

bool CLS::
ProcKeyDialogCheck(int index, wxPoint CAtDown){

    if (index >= mAAKeyFrameIndex.size()) return FALSE;

    uint keyNum = mAAKeyFrameRect[index].size();
    wxString strItem = mAstrDisplayedItem[index];

    Parameter_cls *pCParam = GetPCMain()->GetPCParameter(strItem);
    if (pCParam == NULL) return FALSE;
    
    for (int i = 0; i < keyNum; i++){
	if (mAAKeyFrameRect[index][i].Contains(CAtDown)){
	    int keyIndex = mAAKeyFrameIndex[index][i]; 
		
	    GetPCParamDialog()->mCPos = 
		GetScreenPosition() + GetPointAtDown() - wxPoint(20, 30);
	    GetPCParamGroupDialog()->mCPos = 
		GetScreenPosition() + GetPointAtDown() - wxPoint(20, 30);
	    
	    if (pCParam->IsKeyLinkedAll(keyIndex) == TRUE)
		GetPCParamGroupDialog()->OpenWindow
		    (this, pCParam, keyIndex, strItem);
	    else
		GetPCParamDialog()->OpenWindow
		    (this, pCParam, keyIndex, strItem);
	    return TRUE;
	}
    }
    return FALSE;
}

void CLS::
OnRightDown(wxMouseEvent& event){
    if (mbRename == TRUE) { mbKillRenamer = TRUE; return; }
    if (mbValueInput == TRUE) { mbKillValueInput = TRUE;; return; }

    LeftDownBase(event);

    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    int indexActiveObj = GetItemIndexByName(GetPCSceneObjCtrl()->GetActiveObjectName());
    int index = GetIndex(mpCParent->GetLineHeight());
    if (index == indexActiveObj &&
	mADisplayedRect[index].Contains(CAtDown) == TRUE){
	PopupMenu(mpCPopupMenu);
	return;
    }

    if (ProcKeyDialogCheck(index, CAtDown) == TRUE) return;
    
    if (index >= mAstrDisplayedItem.size() ) return;

    wxString strItem = mAstrDisplayedItem[index];
    Parameter_cls *pCParam = GetPCMain()->GetPCParameter(strItem);
    if (pCParam != NULL){
	PopupMenu(mpCPopupMenuInsertKey);
    }
}

void CLS::
OnLeftUp(wxMouseEvent& event){

    LeftUpBase(event);
    if (mbDraggingRegion == TRUE){
	SelectItemByRegion();
	mbDraggingRegion = FALSE;
    }
    if (mbKeySelecting == TRUE){
	GetPCMain()->MoveSelectedTLKeyCommand(GetMouseDNT(), 0);
	mbKeySelecting = FALSE;
    }
    /*
    if (mbItemSelecting == TRUE){
	int indexTo = (mMouseYCurrent-mOffsetY) / LINEHEIGHT;
	bool bAdditional = IsInsertPointAdditional(indexTo);

	if (bAdditional == TRUE && indexTo > 0) indexTo--;
	
	if (indexTo < mAstrDisplayedItem.GetCount() &&
	    GetDragDistance() > 7){
	    if (mpCMouseCtrl->mbAModifier[_MOUSE_CTRL] == TRUE &&
		mpCMouseCtrl->mbAModifier[_MOUSE_SHIFT] == FALSE)
		pCObjCtrl->SetToChildCommand(mAstrDisplayedItem[indexTo]);
	    else
		pCObjCtrl->MoveItemToCommand(mAstrDisplayedItem[indexTo], bAdditional);
	}
    }
    mbItemSelecting = FALSE;
    */
    mbBarSelecting = FALSE;
    mbSashSelecting = FALSE;

    if (mstrPushedItem.IsEmpty() == FALSE) mstrPushedItem.clear();

    //SetCursor(mCursorB);
    Refresh(FALSE);

    event.Skip();
}

void CLS::
OnDoubleClick(wxMouseEvent& event){

    LeftDownBase(event);
    int index = GetIndex(mpCParent->GetLineHeight());
    if (index >= mADisplayedRect.size()) return;

    wxPoint CAtDown(mMouseXAtDown, mMouseYAtDown);
    if (mbShiftDown == FALSE && mbControlDown == FALSE && mbAltDown == FALSE)
	if (ProcKeyDialogCheck(index, CAtDown) == TRUE) return;
    
    bool bAdditional = event.m_shiftDown;    
    bool bMultiLine = event.m_controlDown;
    if (mADisplayedRect[index].Contains(CAtDown) == TRUE){
	GetPCSceneObjCtrl()->Select(mAstrDisplayedItem[index], bAdditional);
        ConstructRenameTextCtrl(this, GetFontSize(), mpCParent);
    }

    mbDraggingRegion = FALSE;
    if (SelectItemByMouse(bAdditional, bMultiLine) == FALSE)
	mbDraggingRegion = TRUE;

    event.Skip();
}

void CLS::
OnMouseWheel(wxMouseEvent& event){
    if (WheelBase(event) == TRUE)
	Refresh(FALSE);
}

void CLS::
OnMouseMotion(wxMouseEvent& event){

    mMouseXCurrent = event.GetX();
    mMouseYCurrent = event.GetY();

    if (event.Dragging() == TRUE){
	if (mbSashSelecting == TRUE){ MotionSash(event); return; }
	MotionBase(event, FALSE, mbBarSelecting);
	if (mbBarSelecting == TRUE){ Refresh(FALSE); return; }
	Refresh(FALSE);
    }

    if (mbBarFloating == FALSE && event.GetX() > mTgtPanelWidth - 20){
	mbBarFloating = TRUE;
	Refresh(FALSE);
    }
    if (mbBarFloating == TRUE && event.GetX() < mTgtPanelWidth - 20){
	mbBarFloating = FALSE;
	Refresh(FALSE);
    }
    
    if (mMouseXCurrent > GetOffsetX() - 4 && mMouseXCurrent < GetOffsetX() + 4){
	if (mCursorCurrent != _CURSOR_SIZE_WE){
	    SetCursor(wxCURSOR_SIZEWE);
	    mCursorCurrent = _CURSOR_SIZE_WE;
	}
    }
    else
	if (mCursorCurrent != _CURSOR_NORMAL){
	    SetCursor(wxNullCursor);
	    mCursorCurrent = _CURSOR_NORMAL;
	}
	
    event.Skip();
}

void CLS::
OnMouseLeave(wxMouseEvent& event){
    LeaveBase(event);
    event.Skip();
}

void CLS::
OnRenamer(wxCommandEvent& event){ ProcRenamer(event); }

void CLS::
OnRenamerFocus(wxFocusEvent& event){ ProcRenamerFocus(event); }

void CLS::
OnRenamerKey(wxKeyEvent& event){ ProcRenamerKey(event); }

void CLS::
OnRename(wxCommandEvent& event){
    int index = GetItemIndexByName(GetPCSceneObjCtrl()->GetActiveObjectName());
    if (index == -1) return;
    ConstructRenameTextCtrl(this, GetFontSize(), mpCParent);
}

void CLS::
OnInsertKey(wxCommandEvent& event){
    int index = GetIndex(mpCParent->GetLineHeight());
    wxString strItem = mAstrDisplayedItem[index];
    Parameter_cls *pCParam = GetPCMain()->GetPCParameter(strItem);
    if (pCParam == NULL) return;

    wxString strTag;
    if (strItem.Find("Scene Filter") == 0){
	PopStrItemFromStrLine(strItem, "::");
	strTag = GetPCFilterCtrl()->GetStrSelf() << strItem;
    }
    else
	strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strItem;

    double currentNT = GetPCTimeCtrl()->GetCurrentNT();
    int keyType;
    double easeFactor;
    GetPCTimeCtrl()->GetDefaultKeyTypeAndEase(&keyType, &easeFactor);
    pCParam->InsertKeyCommand(currentNT, strTag, keyType, easeFactor, FALSE);
}

void CLS::
ProcessInput(wxString strItem, bool bOpenNext){
    //if (mbKillValueInput == TRUE) return;
    if (mpCValueInput == NULL) return;

    mbKillValueInput = TRUE;
    wxString strNewValue = mpCValueInput->GetValue();

    double newValue = 0; strNewValue.ToDouble(&newValue);
    if (GetPCLuaMinicalc()->IsStringExpression(strNewValue) == TRUE)
	newValue = GetPCLuaMinicalc()->GetValue(strNewValue);
    
    wxString strIndex = PopLastItemFromStrLine(strItem, "::");
    long index; strIndex.ToLong(&index);
    
    Parameter_cls *pCParamer = GetPCMain()->GetPCParameter(strItem);
    
    wxString strTag;
    if (strItem.Find("Scene Filter") == 0){
	PopStrItemFromStrLine(strItem, "::");
	strTag = GetPCFilterCtrl()->GetStrSelf() << strItem;
    }
    else
	strTag = GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strItem;

    pCParamer->SetValueCommand(newValue, strTag);

    if (bOpenNext == TRUE) mQueOpenNextValueInput = index + 1;
}

void CLS::
OnValueInput(wxCommandEvent& event){
    wxString strItem = ((ContainerObject_cls*)event.GetEventUserData())->mstrUserData;
    ProcessInput(strItem, FALSE);
    event.Skip();
}

void CLS::
OnValueInputFocus(wxFocusEvent& event){
    wxString strItem = ((ContainerObject_cls*)event.GetEventUserData())->mstrUserData;
    ProcessInput(strItem, FALSE);
    event.Skip();
}

void CLS::
OnValueInputKey(wxKeyEvent& event){
    if (event.GetKeyCode() == WXK_ESCAPE) mbKillValueInput = TRUE;
    if (event.GetKeyCode() == WXK_TAB){
	wxString strItem = ((ContainerObject_cls*)event.GetEventUserData())->mstrUserData;
	ProcessInput(strItem, TRUE);
    }

    int shortcutEnum = 
	GetPCTextShortcut()->GetShortcutEnum(event, GetName());
    if (shortcutEnum != -1){
	int lineLength = mpCValueInput->GetLineLength(0);
	if (shortcutEnum == _SC_TEXT_BEGINNING_LINE)
	    mpCValueInput->SetInsertionPoint(0);
	if (shortcutEnum == _SC_TEXT_END_LINE)
	    mpCValueInput->SetInsertionPoint(lineLength);
	if (shortcutEnum == _SC_TEXT_KILL_LINE){
	    mpCValueInput->SetSelection(mpCValueInput->GetInsertionPoint(),lineLength);
	    mpCValueInput->Cut();
	}
	if (shortcutEnum == _SC_TEXT_SELECT_ALL)
	    mpCValueInput->SelectAll();
	return;
    }
    event.Skip();
}

void CLS::
OnKeyDown(wxKeyEvent& event){

    switch(event.GetKeyCode()){
    case WXK_F2:
	ConstructRenameTextCtrl(this, GetFontSize(), mpCParent);
	return;
	/*
    case 'X':
	*/
    case WXK_DELETE:
	GetPCSceneObjCtrl()->DeleteSelectedTLKeyCommand();
	event.Skip();
	return; 
    }
    //SendKeyEventToTop(this, event);
}

void CLS::
OnIdle(wxIdleEvent& event){
    ProcIdle(this, mpCParent);
    if (mbKillValueInput == TRUE){
	mbValueInput = FALSE;
	mbKillValueInput = FALSE;
	if (mpCValueInput != NULL){
	    mpCValueInput->Destroy();
	    mpCParent->mbSkipKeyHook = FALSE;
	}
	mpCValueInput = NULL;
	Refresh();
	if (mQueOpenNextValueInput != -1){
	    ConstructValueTextCtrl(mQueOpenNextValueInput);
	    mQueOpenNextValueInput = -1;
	}
    }
    event.Skip();
}

std::vector < wxRect > TimelineParam_cls::msAEmptyRect;
std::vector < int > TimelineParam_cls::msAEmptyIndex;
