// (setq c-basic-offset 4)
// "clsTimelineWA.h"

#include "clsTimelineWA.h"

#include "clsTimeCtrl.h"
#include "clsTimeline.h"

#include "../WIDGETCTRL/WX/myArtProv.h"
#include "../WIDGETCTRL/WX/clsColorProvider.h"

#define CLS TimelineWA_cls

BEGIN_EVENT_TABLE(CLS, TimelineBase_cls)
EVT_PAINT       (CLS::OnPaint)
EVT_MOUSE_EVENTS(CLS::OnMouse)
END_EVENT_TABLE()

CLS::
TimelineWA_cls(Timeline_cls *pCParent) : TimelineBase_cls(pCParent, _TL_WA, wxSize(-1, 23)), mCBitmapTextBuffer(96, 24), mCMemoryDCTextBuffer(mCBitmapTextBuffer) {
    mCIcon.CopyFromBitmap(ART("wxART_TIMESLIDER"));
    mCIconMin.CopyFromBitmap(ART("wxART_MIN"));
    mCIconMax.CopyFromBitmap(ART("wxART_MAX"));
    mCIconRegion.CopyFromBitmap(ART("wxART_REGION"));
}

void CLS::
LimitDragging(){
    if (mSelected == _TL_SEL_MIN){
	int min = mFrameDragging;
	int max = GetPCTimeCtrl()->mCWAMax.GetBodyINT();
	if (min > max) min = max;
	mFrameDragging = min;
    }
    if (mSelected == _TL_SEL_MAX){
	int min = GetPCTimeCtrl()->mCWAMin.GetBodyINT();
	int max = mFrameDragging;
	if (max < min) max = min;
	mFrameDragging = max;
    }
    if (mSelected == _TL_SEL_REGION){
	int LWidth = mpCAALink->mRegionLforDrag;
	int RWidth = mpCAALink->mRegionRforDrag;
	if (mFrameDragging - LWidth < 0) mFrameDragging = LWidth;
	if (mFrameDragging + RWidth > mFrameMax) mFrameDragging = mFrameMax - RWidth;
    }
}

void CLS::
DrawStr(wxAutoBufferedPaintDC &dc, int pos, wxString strMes){
    wxCoord singleW, singleH, singleD;
    mCMemoryDCTextBuffer.GetTextExtent(strMes, &singleW, &singleH, &singleD);
    int offset = 0;
    if (pos > mPanelWidth * 0.7) offset = - 14 - singleW;
    dc.DrawText(strMes, pos + offset, 0);
}

void CLS::
UpdateDisplay(wxAutoBufferedPaintDC &dc){
    dc.SetBackground(wxBrush(wxColour(216,216,216), wxSOLID));
    dc.Clear();
    
    UpdateState();
    SetTLRegion(mAAMinNormalized, mAAMaxNormalized);
    
    DrawWorkingArea(dc, 0);
    if (mbForColorCurve == FALSE) DrawCurrentTime(dc, 0);

    int draggingX = 0;
    int sliderX = GetDevCoordFromN(mTimeNormalized, 0)-7, sliderY = 4;
    int minX = GetDevCoordFromN(mNormalizedWAMin, 0)-5, minY = 1;
    int maxX = GetDevCoordFromN(mNormalizedWAMax, 0), maxY = 1;
    int regionX = GetDevCoordFromN(mNormalizedWARegion, 0)-11, regionY = 16;

    if (mSelected == _TL_SEL_SLIDER){
	sliderX = GetDevCoordFromN(mQuantizedTimeDragging, 0)-7;
	draggingX = sliderX; }
    if (mSelected == _TL_SEL_MIN){
	minX = GetDevCoordFromN(mQuantizedTimeDragging, 0)-5;
	draggingX = minX; }
    if (mSelected == _TL_SEL_MAX){
	maxX = GetDevCoordFromN(mQuantizedTimeDragging, 0);
	draggingX = maxX; }
    if (mSelected == _TL_SEL_REGION){
	regionX = GetDevCoordFromN(mQuantizedTimeDragging, 0)-11;
	draggingX = regionX; }
    
    dc.DrawIcon(mCIconRegion, regionX, regionY);
    dc.DrawIcon(mCIconMin, minX, minY);
    dc.DrawIcon(mCIconMax, maxX, maxY);
    dc.DrawIcon(mCIcon, sliderX, sliderY);
    mRectRegion = wxRect(regionX, regionY, 16, 8);
    mRectSlider = wxRect(sliderX, sliderY, 15, 18);
    mRectMin = wxRect(minX, minY, 6, 22);
    mRectMax = wxRect(maxX, maxY, 6, 22);

    // bZ[W`
    //int offset = 0;
    //if (draggingX > mPanelWidth * 0.7) offset = -40;

    int framePos = sliderX;
    int frameCurrent = mFrameCurrent;
    int subFrameCurrent = mSubFrameCurrent;
    if (HasCapture() == TRUE){
	frameCurrent = mFrameDragging;
	framePos = draggingX;
	dc.SetTextForeground(*wxRED);
    }
    else
	dc.SetTextForeground(GetPCColorProvider()->GetColor("TextWeak"));

    wxString strMes;
    int ticsMode = ((Timeline_cls*)mpCParent)->mCTicsMode.GetBodyINT();
    switch(ticsMode){
    case 0:
	strMes << frameCurrent;
	break;
    case 1:
	strMes << GetPCTimeCtrl()->GetStep(frameCurrent, subFrameCurrent);
	break;
    case 2:
	strMes << GetPCTimeCtrl()->GetNormalizedValue(frameCurrent, subFrameCurrent);
	break;
    }
    
    dc.SetFont(*wxNORMAL_FONT);
    
    DrawStr(dc, framePos + 15, strMes);
    //dc.DrawText(Mes, draggingX + 15 + offset, 0);
}

void CLS::
OnPaint(wxPaintEvent& event){
    wxAutoBufferedPaintDC dc(this);
    UpdateDisplay(dc);
}
 
void CLS::
OnMouse(wxMouseEvent& event){

    ProcMouseEventBase(event, 0);
    
    if (event.ButtonUp(wxMOUSE_BTN_LEFT)) ProcMouseRelease();

    event.Skip();
    //Refresh(FALSE);
}

void CLS::
ProcMouseRelease(){
    if (mSelected == _TL_SEL_SLIDER) MoveTimeSlider();
    if (mSelected == _TL_SEL_MIN) MoveWAMin();
    if (mSelected == _TL_SEL_MAX) MoveWAMax();
    if (mSelected == _TL_SEL_REGION) MoveWARegion();
    ProcMouseReleaseBase();
}

