// -*- c -*- 
// "clsUndoManager.h"

#include "clsUndoManager.h"

#include "../../clsMain.h"

#include <fstream>
#include <sstream>

#include "../COMMON/CommonDefinitions.h"

UndoManager_cls::
UndoManager_cls(){
    Clear();
}

void UndoManager_cls::
Clear(){
    mVCRedoCommand.clear();
    mVCUndoCommand.clear();
    mCurrent = 0;
    mBlockDepth = 0;
}

wxString UndoManager_cls::
GetStrRedoAll(int lineNum){
    const std::string endl = "\n";
    wxString strOut;
    int numItem = (int)mVCRedoCommand.size();
    for (int i = 0; i < numItem; i++){
	if (i < (int)numItem - lineNum && 
	    (i < mCurrent - lineNum || i > mCurrent + lineNum)) continue;
	if (i == mCurrent){
	    std::string strCurrent =
		(std::string)"<-- mCurrent=" + std::to_string(mCurrent) + endl;
	    strOut << strCurrent;
	}
	strOut << mVCRedoCommand[i];
    }
    if (mVCRedoCommand.size() == mCurrent){
	std::string strCurrent =
		(std::string)"<-- mCurrent=" + std::to_string(mCurrent) + endl;
	strOut << strCurrent;
    }
    
    return strOut;
}

wxString UndoManager_cls::
GetStrUndoAll(int lineNum){
    const std::string endl = "\n";
    wxString strOut;
    int numItem = (int)mVCUndoCommand.size();
    for (int i = 0; i < numItem; i++){
	if (i < (int)numItem - lineNum && 
	    (i < mCurrent - lineNum || i > mCurrent + lineNum)) continue;
	if (i == mCurrent){
	    std::string strCurrent =
		(std::string)"<-- mCurrent=" + std::to_string(mCurrent) + endl;
	    strOut << strCurrent;
	}
	strOut << mVCUndoCommand[i];
    }
    if (mVCRedoCommand.size() == mCurrent){
	std::string strCurrent = 
	    (std::string)"<-- mCurrent=" + std::to_string(mCurrent) + endl;
	strOut << strCurrent;
    }
    
    return strOut;
}

void UndoManager_cls::
PushCommandPair(wxString strUndo, wxString strRedo){
    mVCRedoCommand.resize(mCurrent);
    mVCUndoCommand.resize(mCurrent);

    mVCRedoCommand.push_back(strRedo << ENDL);
    mVCUndoCommand.push_back(strUndo << ENDL);
}

void UndoManager_cls::
PushCommandPairAndDo(wxString strUndo, wxString strRedo){
    PushCommandPair(strUndo, strRedo);
    Redo();
}

void UndoManager_cls::
PopCommandPair(){
    mVCRedoCommand.pop_back();
    mVCUndoCommand.pop_back();
    mCurrent--;
}

void UndoManager_cls::
Execute(int direction){

    if (direction == -1 && mCurrent == 0) return;
    if (direction == +1 && mCurrent >= mVCRedoCommand.size()) return;

    do {
	wxString strCommand;
	if (direction == 1) strCommand = mVCRedoCommand[mCurrent];
	if (direction == -1) strCommand = mVCUndoCommand[mCurrent-1];
	
	std::wstring strBuf(strCommand.c_str());
	std::wistringstream sstream(strBuf);
	std::wstring wstrLine;
	while ( std::getline (sstream, wstrLine)){
	    wxString strLine = wstrLine.c_str();
	    wxString strFirstItem = PopStrItemFromStrLine(strLine, "::");
	    if (strFirstItem.IsSameAs(wxT("OpenBlock"))){ 
		mBlockDepth++; break; }
		
	    if (strFirstItem.IsSameAs(wxT("CloseBlock"))){
		mBlockDepth--; break; }
	    GetPCMain()->ReadStrCommand(sstream, strFirstItem, strLine);
	}
	mCurrent += direction; 
	
    } while (mBlockDepth > 0);

    GetPCMain()->UpdatePanels("CONSOLE");
}

void UndoManager_cls::
Undo(){ Execute(-1); }

void UndoManager_cls::
Redo(){ Execute(+1); }


void UndoManager_cls::
Skip(){
    mCurrent++;
}

void UndoManager_cls::
OpenBlock(wxString strTag){
    wxString strComUndo = wxT("CloseBlock");
    wxString strComRedo = wxT("OpenBlock");
    if (strTag.IsEmpty() == FALSE){
	strComUndo << wxT("::") << strTag;
	strComRedo << wxT("::") << strTag;
    }
    PushCommandPair(strComUndo, strComRedo);
    mCurrent++;
}

void UndoManager_cls::
CloseBlock(wxString strTag){
    wxString strComUndo = wxT("OpenBlock");
    wxString strComRedo = wxT("CloseBlock");
    if (strTag.IsEmpty() == FALSE){
	strComUndo << wxT("::") << strTag;
	strComRedo << wxT("::") << strTag;
    }

    if (mVCRedoCommand[mCurrent-1].IsSameAs(strComUndo+"\n") &&
	mVCUndoCommand[mCurrent-1].IsSameAs(strComRedo+"\n")){
	PopCommandPair();
	return;
    }
	
    PushCommandPair(strComUndo, strComRedo);
    mCurrent++;
}

UndoManager_cls* GpCUndo = NULL;
