// -*- c++ -*- 
// (setq c-basic-offset 4)
// "wxPanelEX.h"

#include "wxPanelEX.h"
#include "clsPanelBar.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../../GLCANVAS/clsView3D.h"
#include "../../GRAPH/clsGraphEditor.h"
#include "../../GRAPH/clsColorCurveEditor.h"
#include "../../INPUTMETHOD/clsMouseCtrl.h"
#include "../../MAINCANVAS/clsProperty.h"
#include "../../PREFERENCE/clsShortcut.h"
#include "../../SCENEOBJ/clsOutliner.h"
#include "../../TIMELINE/clsTimeline.h"
#include "../../WIDGETCTRL/clsBitmapPanel.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"
#include "../../WIDGETCTRL/WX/myArtProv.h"
#include "../../WIDGETCTRL/WX/clsSplittableWindow.h"
#include "../../WIDGETCTRL/TEXT/clsConsoleCtrl.h"
#include "../../WIDGETCTRL/TEXT/clsTextEntryCtrl.h"

#include "../../COMMON/CommonDefinitions.h"

#include "../../../wxWidgetEnumMenu.h"
#include "../../../wxWidgetEnum.h"

#include <wx/bmpcbox.h>

#define CLS wxPanelEX

#define EVT_BUTTON_RANGE(id1, id2, func) wx__DECLARE_EVT2(wxEVT_COMMAND_BUTTON_CLICKED, id1, id2, wxCommandEventHandler(func))

BEGIN_EVENT_TABLE(wxPanelEX, wxPanel)
EVT_MOUSE_EVENTS(CLS::OnMouse)
EVT_CHAR_HOOK (CLS::OnKeyDown)
EVT_MENU_RANGE(menuPopTimeline, menuPopConsole, CLS::OnMenuChangePanelType)
EVT_BUTTON_RANGE(btnSeparatePanelH, btnSeparatePanelV, CLS::OnBtnSeparatePanel)
EVT_BUTTON(btnDestroyPanel, CLS::OnBtnDestroyPanel)
EVT_BUTTON(btnHidePanel, CLS::OnBtnHideBarItem)
EVT_BUTTON(btnShowPanel, CLS::OnBtnShowBarItem)
EVT_BUTTON(btnZoomPanelP, CLS::OnBtnZoomP)
EVT_BUTTON(btnZoomPanelM, CLS::OnBtnZoomM)
EVT_COMBOBOX(cmbTypeSelector, CLS::OnTypeSelector)
EVT_MENU(menuPopSeparationDirection, CLS::OnMenuSeparationDirection)
EVT_MENU(menuPopSeparationSwap, CLS::OnMenuSeparationSwap)
EVT_MENU(menuPopDestroyPanel, CLS::OnMenuDestroyPanel)
EVT_MENU(menuShowPanelManage, CLS::OnShowPanelManage)
END_EVENT_TABLE()

wxPanelEX::
wxPanelEX(wxWindow *parent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxScrolled < wxPanel > (parent, id, pos, size, style)
{
    mPanelType = _PANEL_UNDEFINED;
    mpCParentSWindow = NULL;
    mbSkipKeyHook = FALSE;
    
    mpCMouseCtrl = new MouseCtrl_cls;

    if (((style & wxHSCROLL) != 0) || ((style & wxVSCROLL) != 0)){
	SetSize(wxDefaultSize);
	SetVirtualSize(size);
	ShowScrollbars(wxSHOW_SB_ALWAYS, wxSHOW_SB_ALWAYS);
	EnableScrolling(TRUE, TRUE);
	SetScrollbars(10, 10, 50, 50);
    }

    ColorProvider_cls::SetBGColor(this, "BasePanel");
}

CLS::
~wxPanelEX(){
    delete mpCMouseCtrl;
}

void CLS::
InitializeStatic(){
    msVstrPanelNameLocale[GetPanelTypeFromStr("TIMELINE")] = _("Timeline");
    msVstrPanelNameLocale[GetPanelTypeFromStr("OUTLINER")] = _("Outliner");
    msVstrPanelNameLocale[GetPanelTypeFromStr("PROPERTY")] = _("Property");
    msVstrPanelNameLocale[GetPanelTypeFromStr("VIEW3D")] = _("View 3D");
    msVstrPanelNameLocale[GetPanelTypeFromStr("GRAPH")] = _("Graph Editor");
    msVstrPanelNameLocale[GetPanelTypeFromStr("COLOR_CURVE")] = _("Color Curve");
    msVstrPanelNameLocale[GetPanelTypeFromStr("TEXT_EDITOR")] = _("Text Editor");
    msVstrPanelNameLocale[GetPanelTypeFromStr("CONSOLE")] = _("Console");
    msVstrPanelNameLocale[GetPanelTypeFromStr("PROPERTY_RENDER")] = _("Property Render");
    msVstrPanelNameLocale[GetPanelTypeFromStr("PROPERTY_OBJECT")] = _("Property Object");
    msVstrPanelNameLocale[GetPanelTypeFromStr("PROPERTY_FORMAT")] = _("Property Format");
    msVstrPanelNameLocale[GetPanelTypeFromStr("PROPERTY_INFO")] = _("Property Info");
    msVstrPanelNameLocale[GetPanelTypeFromStr("CHILDPANEL")] = _("ChildPanel");
    msVstrPanelNameLocale[GetPanelTypeFromStr("DUMMY")] = _("Dummy");
}

int CLS::
GetPanelTypeFromStr(wxString strPanelType){
    for (int i = 0; i < _PANEL_TYPE_MAX; i++)
	if (strPanelType.IsSameAs(msVstrPanelName[i])) return i;
    return -1;
}


int CLS::
GetPanelTypeFromStrLocale(wxString strPanelType){
    for (int i = 0; i < _PANEL_TYPE_MAX; i++)
	if (strPanelType.IsSameAs(msVstrPanelNameLocale[i])) return i;
    return -1;
}

void CLS::
SetPanelType(int panelType, int parentPanelType){
    mPanelType = panelType;
    SetName(msVstrPanelName[panelType]);
    if (mPanelType == _PANEL_CHILD){
	assert(parentPanelType != _PANEL_UNDEFINED);
	mParentPanelType = parentPanelType;
    }
}

void CLS::
ConstructSelfIcon(wxBoxSizer *pBoxSizerBar){
    wxString strIconTag = msVstrPanelIconTag[mPanelType];
    if (strIconTag.IsEmpty() == TRUE) return;

    wxBoxSizer *pSizerV = new wxBoxSizer(wxVERTICAL);
    BitmapPanel_cls *pCSelfIcon = new BitmapPanel_cls(mpCPanelBar, wxID_ANY, wxDP, wxSize(16, 16));
    pCSelfIcon->SetBorder(1, 1);
    pCSelfIcon->AddBitmapStr(strIconTag);
    
    pSizerV->AddStretchSpacer(1);
    pSizerV->Add(pCSelfIcon, 0, 0);
    pSizerV->AddStretchSpacer(1);
    pBoxSizerBar->Add(pSizerV, 0, wxEXPAND);
}

void CLS::
ConstructTypeSelector(wxBoxSizer *pBoxSizerBar){

    ConstructSelfIcon(pBoxSizerBar);
    
    wxBitmapComboBox *pCComboBox = new wxBitmapComboBox (mpCPanelBar, cmbTypeSelector, msVstrPanelNameLocale[mPanelType], wxDPS, _PANEL_BASIC_MAX, msVstrPanelNameLocale);
    pCComboBox->SetItemBitmap(0, ART("wxART_TIMELINE"));
    pCComboBox->SetItemBitmap(1, ART("wxART_OUTLINER"));
    pCComboBox->SetItemBitmap(2, ART("wxART_PROPERTY"));
    pCComboBox->SetItemBitmap(3, ART("wxART_VIEW3D"));
    pCComboBox->SetItemBitmap(4, ART("wxART_GRAPH"));
    pCComboBox->SetItemBitmap(5, ART("wxART_COLOR_CURVE"));
    pCComboBox->SetItemBitmap(6, ART("wxART_TEXT"));
    pCComboBox->SetItemBitmap(7, ART("wxART_CONSOLE"));
    
    pBoxSizerBar->Add(pCComboBox, 0, wxEXPAND);
}

void CLS::
OnMouse(wxMouseEvent& event){

    mpCMouseCtrl->SetMouseInformation(event);
    
    if (mpCMouseCtrl->IsShowMenu() == FALSE) return;

    if (HasCapture() == TRUE) ReleaseMouse();

    wxMenu *pCPopMenu = new wxMenu;
    wxMenu *pCPopSumbenu = new wxMenu;

    int panelType = mPanelType;
    if (mPanelType == _PANEL_CHILD) panelType = mParentPanelType;

    switch(panelType){
    case _PANEL_TEXT_EDITOR:
	pCPopMenu->Append(menuPopTextShowUndo, _("Show Undo"));
	break;
    }

    pCPopMenu->AppendSeparator();
    //pCPopSumbenu->Append(menuPopTimeline, GCClassProvider.GetPCShortcut()->GetMenuString(menuPopTimeline));
    //pCPopSumbenu->Append(menuPopOutliner, GCClassProvider.GetPCShortcut()->GetMenuString(menuPopOutliner));
    
    for (int i = 0; i < _PANEL_BASIC_MAX; i++)
	pCPopSumbenu->Append(menuPopTimeline+i, GCClassProvider.GetPCShortcut()->GetMenuString(menuPopTimeline+i)/*msVstrPanelNameLocale[i]*/);
    
    pCPopSumbenu->AppendSeparator();
    //pCPopSumbenu->Append(menuPopSeparatePanelH, _("Separate Panel H"));
    //pCPopSumbenu->Append(menuPopSeparatePanelV, _("Separate Panel V"));

    pCPopSumbenu->Append(menuPopSeparationDirection, _("Change Separate Direction"));
    pCPopSumbenu->Append(menuPopSeparationSwap, _("Swap Separate windows"));
    pCPopSumbenu->Append(menuPopDestroyPanel, _("Destroy Panel"));
    pCPopMenu->Append(wxID_ANY, _("ManagePanel"), pCPopSumbenu, _("B")); 

    PopupMenu(pCPopMenu);

    delete pCPopMenu;
}

void CLS::
OnKeyDown(wxKeyEvent& event){
    if (mbSkipKeyHook == TRUE) return event.Skip();
    if (GCClassProvider.GetPCShortcut()->ProcShortcutKey(event, this) == FALSE)
	event.Skip();
    //event.Skip();
}

void CLS::
OnMenuChangePanelType(wxCommandEvent& event){
    // ComboboxEventőɍXVłȂ̂Ńj[oRŕύX
    if (mpCParentSWindow != NULL)
	mpCParentSWindow->ChangePanelType(event.GetId() - menuPopTimeline);
    else event.Skip();
}

void CLS::
OnMenuSeparationDirection(wxCommandEvent& event){
    if (mpCParentSWindow != NULL)
	mpCParentSWindow->SeparationDirection();
    else event.Skip();
}

void CLS::
OnMenuSeparationSwap(wxCommandEvent& event){
    if (mpCParentSWindow != NULL)
	mpCParentSWindow->SeparationSwap();
    else event.Skip();
}

void CLS::
OnMenuDestroyPanel(wxCommandEvent& event){
    if (mpCParentSWindow != NULL)
	mpCParentSWindow->DestroyPanel(this);
    else event.Skip();
}

void CLS::
OnBtnSeparatePanel(wxCommandEvent& event){
    if (mpCParentSWindow != NULL){
	bool bVertical = FALSE;
	if (event.GetId() == btnSeparatePanelV) bVertical = TRUE;
	mpCParentSWindow->SeparatePanel(bVertical);
    }
    else event.Skip();
}

void CLS::
OnShowPanelManage(wxCommandEvent& event){
    mpCPanelBar->ToggleItem();
}

void CLS::
OnLeftDown(wxMouseEvent& event){
    if (event.LeftDown() == TRUE)
	SetFocus();
    event.Skip();
}

void CLS::
OnBtnDestroyPanel(wxCommandEvent& event){
    if (mpCParentSWindow != NULL)
	mpCParentSWindow->DestroyPanel(this);
    else event.Skip();
}

void CLS::
OnBtnHideBarItem(wxCommandEvent& event){
    if (mpCParentSWindow != NULL)
	mpCPanelBar->HideItem();
    else event.Skip();
}

void CLS::
OnBtnShowBarItem(wxCommandEvent& event){
    if (mpCParentSWindow != NULL)
	mpCPanelBar->ShowItem();
    else event.Skip();
}

void CLS::
OnBtnZoomP(wxCommandEvent& event){
    if (mpCParentSWindow != NULL){
	if (mPanelType == _PANEL_TIMELINE)
	    ((Timeline_cls*)this)->Zoom(+1);
	if (mPanelType == _PANEL_OUTLINER)
	    ((Outliner_cls*)this)->Zoom(+1);
    }
    else event.Skip();
}

void CLS::
OnBtnZoomM(wxCommandEvent& event){
    if (mpCParentSWindow != NULL){
	if (mPanelType == _PANEL_TIMELINE)
	    ((Timeline_cls*)this)->Zoom(-1);
	if (mPanelType == _PANEL_OUTLINER)
	    ((Outliner_cls*)this)->Zoom(-1);
    }
    else event.Skip();
}

void CLS::
OnTypeSelector(wxCommandEvent& event){
    if (mpCParentSWindow != NULL){
	int type = GetPanelTypeFromStrLocale(event.GetString());
	if (type == mPanelType) return;

	// ComboboxEvent IɕύXȂƂȂ
	SendMenuEvent(this, menuPopTimeline + type);
    }
    else event.Skip();
}

bool CLS::
IsInnertiaExist(){
    if (mPanelType == _PANEL_VIEW3D)
	return ((View3D_cls*)this)->IsInnertiaExist();
					 
    return FALSE;
}

void CLS::
UpdatePanels(const wxString& strPanelTag, bool bUpdateBar){
    switch(mPanelType){
    case _PANEL_TIMELINE:
	((Timeline_cls*)this)->UpdatePanels(strPanelTag, bUpdateBar); break;
    case _PANEL_OUTLINER:
	((Outliner_cls*)this)->UpdatePanels(strPanelTag); break;
    case _PANEL_PROPERTY:
	((Property_cls*)this)->UpdatePanels(strPanelTag); break;
    case _PANEL_VIEW3D:
	((View3D_cls*)this)->UpdatePanels(strPanelTag); break;
    case _PANEL_GRAPH:
	((GraphEditor_cls*)this)->UpdatePanels(strPanelTag); break; 
    case _PANEL_COLOR_CURVE:
	((ColorCurveEditor_cls*)this)->UpdatePanels(strPanelTag); break; 
    case _PANEL_TEXT_EDITOR:
	((TextEntryCtrl_cls*)this)->UpdatePanels(strPanelTag); break;
    case _PANEL_CONSOLE:
	((ConsoleCtrl_cls*)this)->UpdatePanels(strPanelTag); break;
    }
}

void CLS::
ProcAfterMath(const wxString& strCommand, const wxString& strItem){
    if (strCommand.IsSameAs("SelectParamGroup"))
	switch(mPanelType){
	case _PANEL_GRAPH:
	    ((GraphEditor_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    if (strCommand.IsSameAs("SelectExpression") ||
	strCommand.IsSameAs("SelectConfItem"))
	switch(mPanelType){
	case _PANEL_TEXT_EDITOR:
	    ((TextEntryCtrl_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    
    if (strCommand.IsSameAs("SelectColorCurve"))
	switch(mPanelType){
	case _PANEL_COLOR_CURVE:
	    ((ColorCurveEditor_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    if (strCommand.IsSameAs("DeleteObject") ||
	strCommand.IsSameAs("RenameObject"))
	switch(mPanelType){
	case _PANEL_OUTLINER:
	    ((Outliner_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	case _PANEL_TIMELINE:
	    ((Timeline_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	case _PANEL_GRAPH:
	    ((GraphEditor_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	case _PANEL_COLOR_CURVE:
	    ((ColorCurveEditor_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	case _PANEL_PROPERTY:
	    ((Property_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	case _PANEL_TEXT_EDITOR:
	    ((TextEntryCtrl_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    if (strCommand.IsSameAs("SelectObject"))
	switch(mPanelType){
	case _PANEL_PROPERTY:
	    ((Property_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    
    if (strCommand.IsSameAs("ModifierChange") ||
	strCommand.IsSameAs("DataProcChange") ||
	strCommand.IsSameAs("FilterChange") ||
	strCommand.IsSameAs("DeleteFilter"))
	switch(mPanelType){
	case _PANEL_PROPERTY:
	    ((Property_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    if (strCommand.IsSameAs("DeleteModifier") ||
	strCommand.IsSameAs("DeleteDataProc") ||
	strCommand.IsSameAs("DeleteColorCurve") ||
	strCommand.IsSameAs("DeleteMaterial"))
	switch(mPanelType){
	case _PANEL_TIMELINE:
	    ((Timeline_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    if (strCommand.IsSameAs("ColorCurveChange"))
	switch(mPanelType){
	case _PANEL_COLOR_CURVE:
	    ((ColorCurveEditor_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    
    if (strCommand.IsSameAs("Quit"))
	switch(mPanelType){
	case _PANEL_PROPERTY:
	    ((Property_cls*)this)->ProcAfterMath(strCommand, strItem); break;
	}
    
}

void CLS::
CheckAbortClicked(const wxPoint& mousePos){
    // XeC^Xo[ɔzu邱ƂɂKvȂȂ
    /*
    if (mPanelType == _PANEL_TIMELINE)
	((Timeline_cls*)this)->CheckAbortClicked(mousePos);
    if (mPanelType == _PANEL_VIEW3D)
	((View3D_cls*)this)->CheckAbortClicked(mousePos);
    */
}

void CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    switch(mPanelType){
    case _PANEL_VIEW3D:
	((View3D_cls*)this)->ReadStrLine(istream, strFirstItem, strLine);
	break;
    case _PANEL_PROPERTY:
	((Property_cls*)this)->ReadStrLine(istream, strFirstItem, strLine);
	break;
    case _PANEL_TIMELINE:
	((Timeline_cls*)this)->ReadStrLine(istream, strFirstItem, strLine);
	break;
    case _PANEL_GRAPH:
	((GraphEditor_cls*)this)->ReadStrLine(istream, strFirstItem, strLine);
	break;
    case _PANEL_COLOR_CURVE:
	((ColorCurveEditor_cls*)this)->ReadStrLine(istream, strFirstItem, strLine);
	break;
    case _PANEL_TEXT_EDITOR:
	((TextEntryCtrl_cls*)this)->ReadStrLine(istream, strFirstItem, strLine);
	break;
    }
}

void CLS::
Serialize(wxString &strOut){
    if (mPanelType == _PANEL_UNDEFINED){
	strOut << wxT("wxPanelEX::Undefiend") << ENDL;
	return; }
    strOut << wxT("wxPanelEX::") << msVstrPanelName[mPanelType] << ENDL;
    switch(mPanelType){
    case _PANEL_VIEW3D:
	((View3D_cls*)this)->Serialize(strOut);break;
    case _PANEL_PROPERTY:
	((Property_cls*)this)->Serialize(strOut); break;
    case _PANEL_TIMELINE:
	((Timeline_cls*)this)->Serialize(strOut); break;
    case _PANEL_GRAPH:
	((GraphEditor_cls*)this)->Serialize(strOut); break;
    case _PANEL_COLOR_CURVE:
	((ColorCurveEditor_cls*)this)->Serialize(strOut); break;
    case _PANEL_TEXT_EDITOR:
	((TextEntryCtrl_cls*)this)->Serialize(strOut); break;
    }
}

wxString wxPanelEX::msVstrPanelName[_PANEL_TYPE_MAX] = 
    {"TIMELINE", "OUTLINER", "PROPERTY", "VIEW3D",
     "GRAPH", "COLOR_CURVE", "TEXT_EDITOR", "CONSOLE",
     "PROPERTY_RENDER", "PROPERTY_SCENE", "PROPERTY_OBJECT",
     "PROPERTY_MATERIAL", "PROPERTY_MODIFIER", "PROPERTY_DATAPROC",
     "PROPERTY_FORMAT", "PROPERTY_INFO", "CHILDPANEL",
     "DUMMY"};
// o^̑Oɏ̂ŁAInitializeStatic() ōď
wxString wxPanelEX::msVstrPanelNameLocale[_PANEL_TYPE_MAX] = 
    {"Timeline", "Outliner", "Property", "View 3D",
     "Graph Editor", "Color Curve", "Text Editor", "Console",
     "Property Render", "Property Scene", "Property Object", 
     "Property Material", "Property Modifier", "Property DataProc", 
     "Property Format", "Property Info", "ChildPanel",
     "Dummy"};
wxString wxPanelEX::msVstrPanelIconTag[_PANEL_TYPE_MAX] = 
    {"wxART_TIMELINE", "wxART_OUTLINER", "wxART_PROPERTY", "wxART_VIEW3D",
     "wxART_GRAPH", "wxART_COLOR_CURVE", "wxART_TEXT", "wxART_CONSOLE",
     "", "", "",
     "", "", "",
     "", "", "",
     ""};
