// -*- c -*- 
// "clsIconSelector.h"

#include "clsIconSelector.h"

#include "../WX/clsColorProvider.h"
#include "../WX/myArtProv.h"

#include <wx/dcbuffer.h>
#include <wx/imaglist.h>

#include "../../COMMON/CommonDefinitions.h"

BEGIN_EVENT_TABLE(IconSelector_cls, wxPanel)
     EVT_PAINT       (IconSelector_cls::OnPaint)
     EVT_MOUSE_EVENTS(IconSelector_cls::OnMouse)
END_EVENT_TABLE()

IconSelector_cls::
IconSelector_cls(wxWindow *parent, wxWindowID id, 
		 unsigned int initvalue,/*, unsigned int max,*/
		 std::vector < wxString > VstrArt,
		 wxPoint pos, wxSize size, long style) :
     wxPanel(parent, id, pos, size, style)
{
    SetBackgroundColour(wxColour(216, 216, 216));
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);

    mStyle = style;
    mValue = initvalue;
    mMax = 0;
#ifdef WIN32
    mFontHeight = 14;
#else
    mFontHeight = 18;
#endif
    mClientHeight = -1;

    mItemHeight = size.GetHeight();
    if (mItemHeight == -1) mItemHeight = 16; 
    mItemWidth = mItemHeight;
    mVstrArt = VstrArt;
}

IconSelector_cls::
~IconSelector_cls(){
}

wxWindow* IconSelector_cls::
GetPWindow(){
    return this;
}

void IconSelector_cls::
SetItemSize(wxSize itemSize){
    mItemWidth = itemSize.GetWidth();
    mItemHeight = itemSize.GetHeight();
}

void IconSelector_cls::
UpdateDisplay(wxAutoBufferedPaintDC &dc){
    mClientWidth = GetClientSize().GetWidth();
    mClientHeight = GetClientSize().GetHeight();
    mMax = (int)mVstrArt.size();

    int offsetY = (mClientHeight - mItemHeight) / 2;
    
    dc.SetBackground(wxBrush(wxColour(216,216,216), wxSOLID));
    dc.Clear();

    dc.SetFont(*wxNORMAL_FONT);

    if ((mStyle | wxNO_BORDER) != 0){
	dc.SetPen(wxColour(128, 128, 164));
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	for ( int i = 0; i < mMax; i++ )
	    dc.DrawRectangle(i * (mItemWidth+1), offsetY-1, mItemWidth+2, mItemHeight+2);
    }

    dc.SetPen(*wxTRANSPARENT_PEN);
    
    for ( int i = 0; i < mMax; i++ ){
	dc.SetBrush(ColorProvider_cls::GetColor("Active"));/*wxColour(242, 196, 0));*/
	if (i == mValue)
	    dc.DrawRectangle(i*(mItemWidth+1)+1, offsetY, mItemWidth, mItemHeight);
	dc.DrawBitmap(ARTS(mVstrArt[i], wxSize(mItemWidth, mItemHeight)), i*(mItemWidth+1)+1, offsetY);
    }
    
    wxString strMes;
    int pos = mMax*mItemWidth + 12;
    /*
    int pos = mClientWidth - 80; 
    if (mMax*mItemWidth+4 > pos) pos = mMax*mItemWidth+4;
    */
    if (mVstrTag.size() == 0){
	if (mMax > 0) strMes << mValue << wxT("/") << mMax;
    }
    else{
	strMes = mVstrTag[mValue];
    }
    dc.DrawText(strMes, pos, (mClientHeight - mFontHeight)/2);
}

void IconSelector_cls::
OnPaint(wxPaintEvent& event){
    event.Skip();
    wxAutoBufferedPaintDC dc(this);
    UpdateDisplay(dc);
}

int IconSelector_cls::
GetSelected(wxMouseEvent& event){
    int posX = event.GetX();
    int mSelected = posX / mItemWidth;
    if (mSelected < mMax) return mSelected;
    return -1;
}

void IconSelector_cls::
OnMouse(wxMouseEvent& event){
    if (event.ButtonDown(wxMOUSE_BTN_LEFT)){
	int selected = GetSelected(event);
	if ( selected == -1) return;
	mValue = selected;
	SendEvent(GetParent(), wxEVT_COMMAND_COMBOBOX_SELECTED, GetId());
	Refresh(FALSE);
    }
}

int IconSelector_cls::
GetValueInt(){
    return mValue;
}

void IconSelector_cls::
SetValueInt(int value){
    mValue = value;
    Refresh(FALSE);
}

