// -*- c -*- 
// "clsInputItem.h"

#include "clsInputItem.h"
#include "clsMultiInputPanel.h"

#include "../clsBitmapPanel.h"
#include "../EX/wxButtonEX.h"
#include "../INPUTITEM/clsAtoiComboCtrl.h"
#include "../INPUTITEM/clsAtoSTextCtrl.h"
#include "../INPUTITEM/clsVCenter.h"

#include "../../LUA/clsLuaMiniCalc.h"
#include "../../UNDO/clsUndoManager.h"

#include "../WX/clsFontProvider.h"

#include "../../../clsMain.h"

#undef min
#undef max
#include "wx/valnum.h"

#include "../../COMMON/CommonDefinitions.h"

InputItem_cls::
InputItem_cls(int type, wxString strLabel, bool bEnable, 
	      bool bVolatile, int size) : 
mpCConfItemLink(NULL){
    mInputType = type;
    mstrLabel = strLabel;
    mbEnable = bEnable;
    mbVolatile = bVolatile;
    mSize = size;
    mpCWindowLink = NULL;
    mpCTextLabelLink = NULL;
    //mCConfItemRefer(ConfItemBase_cls);
    mbLimit = FALSE;
    mMin = mMax = -1;
    mContextMenuType = -1;
    mDragScale = 0;
    mpBoxSizerLink = NULL;
    mBitmapTemplate = -1;
    mbUseVCenter = FALSE;
}

bool InputItem_cls::
IsItemIdMatch(int id){
    if (mpCWindowLink != NULL && mpCWindowLink->GetId() == id) return TRUE;
    for (uint i = 0; i < mVpCWindowLinkArray.size(); i++)
	if (mVpCWindowLinkArray[i]->GetId() == id) return TRUE;
    return FALSE;
}

std::vector < wxString > InputItem_cls::
GetVstrLabel(){
    std::vector < wxString > VstrLabel;
    wxString strLabel = mstrLabel;
    if (strLabel.IsEmpty()) return VstrLabel;
    while(TRUE){
	int pos = strLabel.Find(wxT(" "));
	wxString strItem = strLabel.Mid(0, pos);
	VstrLabel.push_back(strItem);
	if (pos == wxNOT_FOUND) break;
	strLabel = strLabel.Mid(pos+1);
    }
    return VstrLabel;
}

void InputItem_cls::
CombineChoiceList(std::vector < wxString > VstrChoice){
    mVstrChoice = VstrChoice;
}

void InputItem_cls::
RecombineConfItem(ConfItem_cls& CConfItem, wxString strPanelName, bool bForceShow){
    //mpCConfItemLink->DetachInputPanelName(strPanelName);
    //CConfItem.AppendInputPanelName(strPanelName);
    mpCConfItemLink = &CConfItem;
    if (bForceShow == TRUE) Show();
}

void InputItem_cls::
ConnectConfItem(wxBoxSizer* pBoxSizerCurrent, wxWindow *pCParent){
    int type = mpCConfItemLink->mConfType;
    switch(type){
    case _CONF_ITEM_UINT:{
	unsigned int max = mVstrChoice.size();
	wxString strInit = wxT("dummy");
	unsigned int body = mpCConfItemLink->GetBodyINT();
	if (max > 0 && body < max) strInit = mVstrChoice[body];

	AtoiComboCtrl_cls *pCComboBox = new AtoiComboCtrl_cls(pCParent, wxID_ANY, strInit, max, mVstrChoice.data());
	pBoxSizerCurrent->Add(pCComboBox, mSize, wxEXPAND);
	mpCWindowLink = pCComboBox;
	pCParent->Connect(pCComboBox->GetId(), wxEVT_COMMAND_COMBOBOX_SELECTED,
			  wxComEH(MultiInputPanel_cls::OnItemChange));
	break; }
    case _CONF_ITEM_INT:
    case _CONF_ITEM_FLOAT:
    case _CONF_ITEM_STR: {
	AtoSTextCtrl_cls *pCTextCtrl = new AtoSTextCtrl_cls(pCParent, wxID_ANY, mpCConfItemLink->mstrBody,  wxDP, wxSize(40, -1), 0);
	pCTextCtrl->mDragScale = mDragScale;
	pCTextCtrl->mbLimit = mbLimit;
	pCTextCtrl->mMin = mMin;
	pCTextCtrl->mMax = mMax;
	
	if (type == _CONF_ITEM_INT){
	    pCTextCtrl->mbInteger = TRUE;
	    //pCTextCtrl->SetValidator(wxIntegerValidator<int>());
	}

	pBoxSizerCurrent->Add(pCTextCtrl, mSize, wxEXPAND);
	mpCWindowLink = pCTextCtrl;
	pCParent->Connect(pCTextCtrl->GetId(), wxEVT_COMMAND_TEXT_ENTER,
			  wxComEH(MultiInputPanel_cls::OnItemChange));
	break; }
    case _CONF_ITEM_BOOL: {
	bool bInit = (bool)mpCConfItemLink->GetBodyINT();
	wxCheckBox *pCCheckBox = new wxCheckBox(pCParent, wxID_ANY, mstrLabel, wxDP, wxSize(30, -1), 0);
	pCCheckBox->SetValue(bInit);
	pBoxSizerCurrent->Add(pCCheckBox, mSize, wxEXPAND);
	mpCWindowLink = pCCheckBox;
	pCParent->Connect(pCCheckBox->GetId(), wxEVT_CHECKBOX,
			  wxComEH(MultiInputPanel_cls::OnItemChange));
	break; }
    case _CONF_ITEM_FLOATARRAY: {
	std::vector < float > VInit = mpCConfItemLink->GetBodyFLOATARRAY();
	for (uint i = 0; i < VInit.size(); i++){
	    wxString strInit; strInit << VInit[i];
	    AtoSTextCtrl_cls *pCTextCtrl = new AtoSTextCtrl_cls(pCParent, wxID_ANY, strInit, wxDP, wxSize(30, -1), 0);
	    pCTextCtrl->mDragScale = mDragScale;
	    pCTextCtrl->mbLimit = mbLimit;
	    pCTextCtrl->mMin = mMin;
	    pCTextCtrl->mMax = mMax;
	    pBoxSizerCurrent->Add(pCTextCtrl, mSize, wxEXPAND);
	    mVpCWindowLinkArray.push_back(pCTextCtrl);
	    pCParent->Connect(pCTextCtrl->GetId(), wxEVT_TEXT_ENTER,
			      wxComEH(MultiInputPanel_cls::OnItemChange));
	}
	
	break; }
    case _CONF_ITEM_BOOLARRAY: {
	std::vector < bool > VbInit = mpCConfItemLink->GetBodyBOOLARRAY();
	std::vector < wxString > VstrLabel = GetVstrLabel();
	for (uint i = 0; i < VbInit.size(); i++){
	    wxCheckBox *pCCheckBox = new wxCheckBox(pCParent, wxID_ANY, VstrLabel[i]);
	    pCCheckBox->SetValue(VbInit[i]);
	    pBoxSizerCurrent->Add(pCCheckBox, mSize, wxEXPAND);
	    mVpCWindowLinkArray.push_back(pCCheckBox);
	    pCParent->Connect(pCCheckBox->GetId(), wxEVT_CHECKBOX,
			      wxComEH(MultiInputPanel_cls::OnItemChange));
	}
	break; }
    case _CONF_ITEM_INTARRAY: {
	std::vector < int > VInit = mpCConfItemLink->GetBodyINTARRAY();
	for (uint i = 0; i < VInit.size(); i++){
	    wxString strInit; strInit << VInit[i];
	    AtoSTextCtrl_cls *pCTextCtrl = new AtoSTextCtrl_cls(pCParent, wxID_ANY, strInit, wxDP, wxSize(30, -1), 0);
	    pCTextCtrl->mDragScale = mDragScale;
	    pCTextCtrl->mbLimit = mbLimit;
	    pCTextCtrl->mMin = mMin;
	    pCTextCtrl->mMax = mMax;
	    pBoxSizerCurrent->Add(pCTextCtrl, mSize, wxEXPAND);
	    mVpCWindowLinkArray.push_back(pCTextCtrl);
	    pCParent->Connect(pCTextCtrl->GetId(), wxEVT_TEXT_ENTER,
			      wxComEH(MultiInputPanel_cls::OnItemChange));
	}
	break; }
    }

    if (mpCWindowLink != NULL){
	mpCWindowLink->Enable(mbEnable);
    }
    int num = mVpCWindowLinkArray.size();
    for (int i = 0; i < num; i++)
	mVpCWindowLinkArray[i]->Enable(mbEnable);
}

void InputItem_cls::
ConnectButton(wxBoxSizer* pBoxSizerCurrent, wxWindow *pCParent){
    int buttonHeight = FontProvider_cls::GetButtonHeight();
    wxSize size(mButtonWidth, buttonHeight);
    wxButtonEX *pCButton = new wxButtonEX(pCParent, mButtonId, mstrLabel, wxDP, size);

    mpCWindowLink = pCButton;
    if (mButtonWidth == -1)
	pBoxSizerCurrent->Add(pCButton, mSize, wxEXPAND);
    else
	pBoxSizerCurrent->Add(pCButton, 0, wxEXPAND);
    pCButton->Enable(mbEnable);
}

void InputItem_cls::
ConnectBitmapPanel(wxBoxSizer* pBoxSizerCurrent, wxWindow *pCParent){
    int buttonHeight = FontProvider_cls::GetButtonHeight();
    wxSize size(mButtonWidth, buttonHeight);
    BitmapPanel_cls *pCBitmapPanel;
    if (mBitmapTemplate == _BITMAP_BTN_DROPDOWN)
	pCBitmapPanel = BitmapPanel_cls::GetNewPCButtonDropDown(pCParent, mButtonId);
    else
	pCBitmapPanel = new BitmapPanel_cls(pCParent, mButtonId, wxDP, size);
    pCBitmapPanel->mstrLabel = mstrLabel;
    pCBitmapPanel->mbButton = TRUE;
    pCBitmapPanel->mstrCommandStr = mstrCommandStr;
    mpCWindowLink = pCBitmapPanel;

    if (mbUseVCenter == TRUE){
	VCenter_cls *pCVCenter = new VCenter_cls(pCBitmapPanel);
	if (mButtonWidth == -1)
	    pBoxSizerCurrent->Add(pCVCenter, mSize, wxEXPAND);
	else
	    pBoxSizerCurrent->Add(pCVCenter, 0, wxEXPAND);
    }
    else{
        if (mButtonWidth == -1)
	    pBoxSizerCurrent->Add(pCBitmapPanel, mSize, wxEXPAND);
	else
	    pBoxSizerCurrent->Add(pCBitmapPanel, 0, wxEXPAND);
    }

    pCBitmapPanel->Enable(mbEnable);
}

void InputItem_cls::
ConnectSizer(wxBoxSizer* pBoxSizerCurrent){
    mpBoxSizerLink = pBoxSizerCurrent;
}

void InputItem_cls::
OnChanged(bool bVolatile){
    if (mInputType == _ITEM_CONFITEM){
	bool bMod = FALSE;

	wxString strFrom, strTo;
	mpCConfItemLink->SerializeSingleItem(strFrom);

	switch(mpCConfItemLink->mConfType){
	case _CONF_ITEM_UINT:{
	    long long newValue = ((AtoiComboCtrl_cls*)mpCWindowLink)->GetValueInt();
	    mpCConfItemLink->SetBodyFromData(newValue);
	    break; }
	case _CONF_ITEM_INT:{
	    wxString strNew = ((AtoSTextCtrl_cls*)mpCWindowLink)->GetValue();
	    long long newValue; strNew.ToLongLong(&newValue);
	    if (GetPCLuaMinicalc()->IsStringExpression(strNew) == TRUE)
		newValue = (long long)GetPCLuaMinicalc()->GetValue(strNew);
	    if (mbLimit == TRUE){
		if (newValue < (long long)mMin){
		    newValue = (long long)mMin; bMod = TRUE; }
		if (newValue > (long long)mMax){
		    newValue = (long long)mMax; bMod = TRUE; }
		wxString strLimited; strLimited << newValue;
		((AtoSTextCtrl_cls*)mpCWindowLink)->SetLabel(strLimited);
	    }
	    mpCConfItemLink->SetBodyFromData(newValue);
	    break; }
	case _CONF_ITEM_FLOAT:{
	    wxString strNew = ((AtoSTextCtrl_cls*)mpCWindowLink)->GetValue();
	    double newValue; strNew.ToDouble(&newValue);
	    if (GetPCLuaMinicalc()->IsStringExpression(strNew) == TRUE)
		newValue = GetPCLuaMinicalc()->GetValue(strNew);
	    if (mbLimit == TRUE){
		if (newValue < mMin){ newValue = mMin; bMod = TRUE; }
		if (newValue > mMax){ newValue = mMax; bMod = TRUE; }
	    }
	    mpCConfItemLink->SetBodyFromData(newValue);
	    break; }
	case _CONF_ITEM_STR:{
	    wxString strNew = ((AtoSTextCtrl_cls*)mpCWindowLink)->GetValue();
	    mpCConfItemLink->SetBodyFromData(strNew);
	    break; }
	case _CONF_ITEM_BOOL:{
	    bool newValue = ((wxCheckBox*)mpCWindowLink)->GetValue();
	    mpCConfItemLink->SetBodyFromData((long long)newValue);
	    break; }
	case _CONF_ITEM_FLOATARRAY:{
	    std::vector < float > VNewValue;
	    for (int i = 0; i < mVpCWindowLinkArray.size(); i++){
		wxString strNew = ((AtoSTextCtrl_cls*)mVpCWindowLinkArray[i])->GetValue();
		double newValue; strNew.ToDouble(&newValue);
		if (GetPCLuaMinicalc()->IsStringExpression(strNew) == TRUE)
		    newValue = GetPCLuaMinicalc()->GetValue(strNew);
		if (mbLimit == TRUE){
		    if (newValue < mMin){ newValue = mMin; bMod = TRUE; }
		    if (newValue > mMax){ newValue = mMax; bMod = TRUE; }
		}
		VNewValue.push_back((float)newValue);
	    }
	    mpCConfItemLink->SetBodyFromData(VNewValue);
	    break; }
	case _CONF_ITEM_BOOLARRAY:{
	    std::vector < bool > VNewValue;
	    for (int i = 0; i < mVpCWindowLinkArray.size(); i++)
		VNewValue.push_back(((wxCheckBox*)mVpCWindowLinkArray[i])->GetValue());
	    mpCConfItemLink->SetBodyFromData(VNewValue);
	    break; }
	case _CONF_ITEM_INTARRAY:{
	    std::vector < int > VNewValue;
	    for (int i = 0; i < mVpCWindowLinkArray.size(); i++){
		wxString strNew = ((AtoSTextCtrl_cls*)mVpCWindowLinkArray[i])->GetValue();
		long long newValue; strNew.ToLongLong(&newValue);
		if (GetPCLuaMinicalc()->IsStringExpression(strNew) == TRUE)
		    newValue = GetPCLuaMinicalc()->GetValue(strNew);
		if (mbLimit == TRUE){
		    if (newValue < mMin){ newValue = mMin; bMod = TRUE; }
		    if (newValue > mMax){ newValue = mMax; bMod = TRUE; }
		}
		VNewValue.push_back(newValue);
	    }
	    mpCConfItemLink->SetBodyFromData(VNewValue);
	    break; }

	}
	
	mpCConfItemLink->SerializeSingleItem(strTo);
	
	if (bVolatile == FALSE && mbVolatile == FALSE &&
	    strFrom.IsSameAs(strTo) != TRUE){
	    GetPCUndo()->PushCommandPair(strFrom, strTo);
	    GetPCUndo()->Skip(); //łɕύXĂ̂SkipB
	}
	
	/*
	// plŎgpĂA\XVB
	if (mpCConfItemLink->GetInputParentNumber() > 1)
	    GetPCMain()->UpdatePanels(mpCConfItemLink->mstrInputPanelName.ToStdString());
	*/
	
	GetPCMain()->UpdatePanels("CONSOLE");
	if (mpCConfItemLink->mstrLinkedPanel.IsEmpty() == FALSE)
	    GetPCMain()->UpdatePanels(mpCConfItemLink->mstrLinkedPanel.ToStdString());

	if (bMod == TRUE) UpdateDisplay();
    }
}

/*
bool InputItem_cls::
IsShown(){
    if (mInputType != _ITEM_BOXSIZERHORIZONTAL)
	return mpCWindowLink->IsShown();
    if (mpCTextLabelLink != NULL) 
	return mpCTextLabelLink->IsShown();
    return FALSE;
}
*/

bool InputItem_cls::
Show(bool bShow){
    if (bShow == FALSE) {
	return Hide();
    }

    bool bRetVal = FALSE;
    if (mInputType != _ITEM_BOXSIZER){
	if (mpCWindowLink != NULL){
	    if (mpCWindowLink->IsShown() == FALSE){
		mpCWindowLink->Show();
		bRetVal = TRUE;
	    }
	}
	
	uint num = mVpCWindowLinkArray.size();
	for (uint i = 0; i < num; i++){
	    if (mVpCWindowLinkArray[i]->IsShown() == FALSE){
		mVpCWindowLinkArray[i]->Show(); bRetVal = TRUE;
	    }
	}
    }

    if (mInputType == _ITEM_BOXSIZER){
	uint num = mpBoxSizerLink->GetItemCount();
	for (uint i = 0; i < num; i++){
	    wxSizerItem *pCItem = mpBoxSizerLink->GetItem(i);
	    if (pCItem->IsShown() == FALSE){
		pCItem->Show(TRUE); bRetVal = TRUE; 
	    }
	    /*
	    wxWindow *pCWindow = mpBoxSizerLink->GetItem(i)->GetWindow();
	    wxSizer *pCSizer = mpBoxSizerLink->GetItem(i)->GetSizer();
	    if (pCWindow != NULL){
		if (pCWindow->IsShown() == FALSE){
		    pCWindow->Show(); bRetVal = TRUE; }
	    }
	    if (pCSizer != NULL){
		
	    }
	    */
	}
    }

    if (mpCTextLabelLink != NULL) mpCTextLabelLink->Show();

    return bRetVal;
}

bool InputItem_cls::
Hide(){
    bool bRetVal = FALSE;
    if (mInputType != _ITEM_BOXSIZER){
	if (mpCWindowLink != NULL){
	    if (mpCWindowLink->IsShown() == TRUE){
		mpCWindowLink->Hide();
		bRetVal = TRUE;
	    }
	}
	uint num = mVpCWindowLinkArray.size();
	for (uint i = 0; i < num; i++){
	    if (mVpCWindowLinkArray[i]->IsShown() == TRUE){
		mVpCWindowLinkArray[i]->Hide(); bRetVal = TRUE;
	    }
	}
    }

    if (mInputType == _ITEM_BOXSIZER){
	uint num = mpBoxSizerLink->GetItemCount();
	for (uint i = 0; i < num; i++){
	    wxSizerItem *pCItem = mpBoxSizerLink->GetItem(i);
	    if (pCItem->IsShown() == TRUE){
		pCItem->Show(FALSE); bRetVal = TRUE; 
	    }
	    /*
	    wxWindow *pCWindow = mpBoxSizerLink->GetItem(i)->GetWindow();
	    if (pCWindow != NULL){
		if (pCWindow->IsShown() == TRUE){
		    pCWindow->Hide(); bRetVal = TRUE;
		}
	    }
	    */
	}
    }

    if (mpCTextLabelLink != NULL) mpCTextLabelLink->Hide();

    return bRetVal;
}

void InputItem_cls::
Enable(bool bEnable){
    if (mInputType != _ITEM_BOXSIZER)
	mpCWindowLink->Enable(bEnable);

    if (mpCTextLabelLink != NULL) mpCTextLabelLink->Enable(bEnable);
}

void InputItem_cls::
SetBackgroundColour(std::vector < wxColor > Vclr){
    uint size = Vclr.size();
    if (mVpCWindowLinkArray.size() != size) return;
    for (uint i = 0; i < size; i++)
	if (mVpCWindowLinkArray[i]->GetBackgroundColour() != Vclr[i]){
	    mVpCWindowLinkArray[i]->SetBackgroundColour(Vclr[i]);
	    mVpCWindowLinkArray[i]->Refresh();
	}
}

void InputItem_cls::
SetBackgroundColour(wxColor clr){
    if (mpCWindowLink == NULL) return;
    if (mpCWindowLink->GetBackgroundColour() != clr){
	mpCWindowLink->SetBackgroundColour(clr);
	mpCWindowLink->Refresh();
    }
}

void InputItem_cls::
UpdateLabel(){
    if (mpCTextLabelLink == NULL) return;
    mpCTextLabelLink->SetLabel(mstrLabel);
}

void InputItem_cls::
UpdateDisplay(){
    if (mInputType == _ITEM_CONFITEM)
	switch(mpCConfItemLink->mConfType){
	case _CONF_ITEM_UINT: {
	    long long value = mpCConfItemLink->GetBodyINT();
	    if (((AtoiComboCtrl_cls*)mpCWindowLink)->GetValueInt() != value)
		((AtoiComboCtrl_cls*)mpCWindowLink)->SetValueInt(value);
	    break; }
	case _CONF_ITEM_FLOAT:{
	    wxString strValue = mpCConfItemLink->GetBodySTRShortString();
	    if (((AtoSTextCtrl_cls*)mpCWindowLink)->GetValue() != strValue)
		((AtoSTextCtrl_cls*)mpCWindowLink)->SetValue(strValue);
	    break; }
	case _CONF_ITEM_INT:
	case _CONF_ITEM_STR: {
	    wxString strValue = mpCConfItemLink->GetBodySTR();
	    if (((AtoSTextCtrl_cls*)mpCWindowLink)->GetValue() != strValue)
		((AtoSTextCtrl_cls*)mpCWindowLink)->SetValue(strValue);
	    break; }
	case _CONF_ITEM_BOOL: {
	    bool value = (bool)mpCConfItemLink->GetBodyINT();
	    if (((wxCheckBox*)mpCWindowLink)->GetValue() != value)
		((wxCheckBox*)mpCWindowLink)->SetValue(value);
	    break; }
	case _CONF_ITEM_FLOATARRAY: {
	    std::vector < float > VValue = 
		mpCConfItemLink->GetBodyFLOATARRAY();
	    for (int i = 0; i < mVpCWindowLinkArray.size(); i++){
		float valueData = VValue[i];
		wxString strValue = wxString::Format("%g", valueData);
		((AtoSTextCtrl_cls*)mVpCWindowLinkArray[i])->SetValue(strValue);
	    }
	    break; }
	case _CONF_ITEM_BOOLARRAY: {
	    std::vector < bool > VbValue = 
		mpCConfItemLink->GetBodyBOOLARRAY();
	    for (int i = 0; i < mVpCWindowLinkArray.size(); i++){
		bool value = ((wxCheckBox*)mVpCWindowLinkArray[i])->GetValue();
		if (value != VbValue[i])
		    ((wxCheckBox*)mVpCWindowLinkArray[i])->SetValue(VbValue[i]);
	    }
	    break; }
	case _CONF_ITEM_INTARRAY: {
	    std::vector < int > VValue = 
		mpCConfItemLink->GetBodyINTARRAY();
	    for (int i = 0; i < mVpCWindowLinkArray.size(); i++){
		long long valueData = VValue[i];
		wxString strValue = wxString::Format("%d", valueData);
		((AtoSTextCtrl_cls*)mVpCWindowLinkArray[i])->SetValue(strValue);
	    }
	    break; }
	}
    

    if (mInputType == _ITEM_BITMAPPANEL)
	((BitmapPanel_cls*)mpCWindowLink)->Refresh();
}


