// (setq c-basic-offset 4)
// "clsConsoleCtrl.h"

#include "clsConsoleCtrl.h"
#include "clsAutoComplete.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../../UNDO/clsUndoManager.h"

#include <wx/richtext/richtextctrl.h>

#include "../../COMMON/CommonDefinitions.h"
#include "../../PREFERENCE/clsShortcut.h"
#include "../../WIDGETCTRL/EX/clsPanelBar.h"
#include "../../WIDGETCTRL/EX/wxLogGuiEX.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"
#include "../../WIDGETCTRL/INPUTITEM/clsTextButton.h"
#include "../../WIDGETCTRL/INPUTITEM/clsVCenter.h"

#include "../../../clsFrameMain.h"

#include "../../../wxWidgetEnumMenu.h"
	
#include <iostream>
#include "../../LUA/clsLuaBase.h"
#include "../../LUA/clsLuaInterpreter.h"

#define CLS ConsoleCtrl_cls

BEGIN_EVENT_TABLE(ConsoleCtrl_cls, wxPanelEX)
EVT_BUTTON(menuPanelLog, CLS::OnMenuLog)
EVT_MENU(menuLogClear, CLS::OnMenuLogClear)
EVT_MENU(menuLogUpdate, CLS::OnMenuLogUpdate)
//EVT_MENU(menuPopTextShowUndo, CLS::ShowUndo)
EVT_MENU_RANGE(menuPopText0, menuPopText9, CLS::OnPopMenu)
EVT_TIMER(-1, CLS::OnTimer)
EVT_IDLE(CLS::OnIdle)
END_EVENT_TABLE()

    CLS::
ConsoleCtrl_cls(wxWindow *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxPanelEX(pCParent, id, pos, size, style), 
    TextEntryBase_cls(this) {

    mbSkipKeyHook = TRUE;

    SetPanelType(_PANEL_CONSOLE);
    mCTimerMarkup.SetOwner(GetEventHandler());
    SetPMarkupRichTextCtrl(mpCTextCtrl);

    mpCTextCtrl->Connect
	(wxEVT_KEY_DOWN, wxKeyEventHandler(CLS::OnKey),
	 NULL, this);
    mpCTextCtrl->Connect
	(wxEVT_CHAR, wxKeyEventHandler(CLS::OnChar),
	 NULL, this);
    mpCTextCtrl->Connect
	(wxEVT_CHAR_HOOK, wxKeyEventHandler(CLS::OnCharHook),
	 NULL, this);

    mpCTextCtrl->SetBackgroundColour(wxColour(0, 0, 0));
    mpCTextCtrl->SetForegroundColour(*wxWHITE);
    SetForegroundColour(*wxWHITE);
    //wxRichTextAttr CTextAttr;
    mCTextAttrDefault.SetFlags(wxTEXT_ATTR_ALL);
    mCTextAttrDefault.SetTextColour(wxColour(255, 255, 255));
    mCTextAttrDefault.SetBackgroundColour(wxNullColour);
    //mpCTextCtrl->SetDefaultStyle(CTextAttr);
    //mpCTextCtrl->SetStyleEx(wxRichTextRange(0, 100), mCTextAttrDefault);
    SetToDefaultAttr();
    mpCTextCtrl->SetDefaultStyleToCursorStyle();
    
    mpCPanelBar = new PanelBar_cls(this);
    mpCPanelBar->SetPanelType(_PANEL_CHILD, _PANEL_CONSOLE);

    wxBoxSizer *pBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerBar = new wxBoxSizer(wxHORIZONTAL);

    mAstrSelctor[0] = _("Lua Console");    
    mAstrSelctor[1] = _("Undo");
    mAstrSelctor[2] = _("Redo"); 
    mAstrSelctor[3] = _("Windows");
    mAstrSelctor[4] = wxT("wxLog");
    mpCComboSelector = new wxComboBox(mpCPanelBar, wxID_ANY, mAstrSelctor[0], wxDP, wxSize(120, -1), 5, mAstrSelctor, wxCB_DROPDOWN | wxCB_READONLY);
    
    Connect(mpCComboSelector->GetId(), wxEVT_COMMAND_COMBOBOX_SELECTED,
	    wxCommandEventHandler(CLS::OnComboBox));
    
    ConstructTypeSelector(pBoxSizerBar);

    mpCTextMenuLog = new TextButton_cls
	(mpCPanelBar, menuPanelLog, GetPCStringProvider()->GetStrMenu("Log"));
    mpCTextMenuLog->SetEventTarget(this);
    VCenter_cls *pCMenuLog = new VCenter_cls(mpCTextMenuLog);

    pBoxSizerBar->Add(mpCComboSelector, 0, wxEXPAND);
    pBoxSizerBar->Add(pCMenuLog, 0, wxEXPAND);
    pBoxSizerBar->AddStretchSpacer(1);
    mpCPanelBar->ConstructBarItemHideShow(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemDefault(pBoxSizerBar);

    mpCPanelBar->SetSizer(pBoxSizerBar);

    pBoxSizerSrc->Add(mpCPanelBar, 0, wxEXPAND);
    pBoxSizerSrc->Add(mpCTextCtrl, 1, wxEXPAND);
    
    SetSizer(pBoxSizerSrc);

    mMode = _CONSOLE_LUA_CONSOLE;
    mHistory = 0;
    
    mpCLua = new LuaBase_cls;
}

CLS::
~ConsoleCtrl_cls(){
    mpCLua->CloseLuaBase();
    delete mpCLua;
}

void CLS::
CallHistory(int direction){
    mHistory += direction;
    if (mHistory < 0) mHistory = 0;

    wxString strHistory;
    if (mHistory >= mVstrHistory.size()){
	mHistory = mVstrHistory.size(); 
    }
    else
	strHistory = mVstrHistory[mHistory];

    long posX, posY;
    int lineLength = GetInsertionPointXY(&posX, &posY);
    int pos0 = mpCTextCtrl->XYToPosition(0, posY);
    int pos1 = mpCTextCtrl->XYToPosition(lineLength, posY);
    mpCTextCtrl->Remove(pos0, pos1);
    if (strHistory.IsEmpty() == TRUE) return;

    mpCTextCtrl->WriteText(strHistory);
    lineLength = mpCTextCtrl->GetLineLength(posY);
    int newPos = mpCTextCtrl->XYToPosition(lineLength, posY);
    mpCTextCtrl->SetInsertionPoint(newPos);
}

void CLS::
ProcessOneLiner(){
    wxStreamToTextRedirector CRedirect(mpCTempTextCtrl);
    
    wxString strFunction;

    strFunction << "function Data.ParamGroup(obj, param)" << ENDL
		<< "return { "<< ENDL 
		<< "obj = obj," << ENDL
		<< "param = param," << ENDL
		<< "GetName = function() return Data.GetName(param) end," << ENDL
		<< "GetDimension = function() return ParamGroup.GetDimension(obj, param) end," << ENDL
		<< "GetValueAt = function(nt) return ParamGroup.GetValueAt(obj, param, nt) end," << ENDL
		<< "SetValueAt = function(nt, ...) return ParamGroup.SetValueAt(obj, param, nt, ...) end," << ENDL
		<< "GetLink = function() return ParamGroup.GetLink(obj, param) end," << ENDL
		<< "SetLink = function(link) return ParamGroup.SetLink(obj, param, link) end," << ENDL
		<< "GetShy = function() return ParamGroup.GetShy(obj, param) end," << ENDL
		<< "SetShy = function(shy) return ParamGroup.SetShy(obj, param, shy) end," << ENDL
		<< "GetUseExpression = function() return ParamGroup.GetUseExpression() end," << ENDL
		<< "SetUseExpression = function(use) return ParamGroup.SetUseExpression(obj, param, use) end," << ENDL
		<< "GetUseExpression = function() return ParamGroup.GetExpression() end," << ENDL
		<< "SetExpression = function(expr) return ParamGroup.SetExpression(obj, param, expr) end," << ENDL
		<< "GetTimeDependency = function() return ParamGroup.GetTimeDependency(obj, param) end," << ENDL
		<< "SetTimeDependency = function(...) return ParamGroup.SetTimeDependency(obj, param, ...) end," << ENDL
		<< "}" <<ENDL
		<< "end" << ENDL;

    strFunction << "function Data.Object(obj)" << ENDL
		<< "return {" << ENDL
		<< "obj = obj," << ENDL
		<< "GetShy = function() return Object.GetShy(obj) end," << ENDL
		<< "SetShy = function(shy) return Object.SetShy(obj, shy) end," << ENDL
		<< "GetVisibility = function() return Object.GetVisibility(obj) end," << ENDL
		<< "SetVisibility = function(visible) return Object.SetVisibility(obj, visible) end," << ENDL
		<< "GetValidity = function() return Object.GetValidity(obj) end," << ENDL
		<< "SetValidity = function(valid) return Object.SetValidity(obj, valid) end," << ENDL
		<< "GetConfList = function() return Object.GetConfList(obj) end, " << ENDL
		<< "GetConfItem = function(tag) return Object.GetConfItem(obj, tag) end," << ENDL 
		<< "GetAttribList = function() return Object.GetAttribList(obj) end," << ENDL
		<< "GetAttrib = function (attrib) return {" << ENDL
		<< "attrib = attrib," << ENDL
		<< "GetConfList = function() return Object.GetAttribConfList(obj, attrib) end," << ENDL 
		<< "GetConfItem = function(conf) return Object.GetAttribConfItem(obj, attrib, conf) end," << ENDL
		<< "} end," << ENDL
		<< "}" << ENDL
		<< "end" << ENDL;


    
    strFunction << "function script()" << ENDL
		<< mstrOneLiner << ENDL
		<< "end" << ENDL;

    luaL_Reg *LuaTables[6] = {
	GCClassProvider.GetPCLuaInterpreter()->mVLibTableOps.data(),
	GCClassProvider.GetPCLuaInterpreter()->mVLibTableOpsActiveObj.data(),
	GCClassProvider.GetPCLuaInterpreter()->mVLibTableOpsSelectedObj.data(),
	GCClassProvider.GetPCLuaInterpreter()->mVLibTableData.data(),
	GCClassProvider.GetPCLuaInterpreter()->mVLibTableParamGroup.data(),
	GCClassProvider.GetPCLuaInterpreter()->mVLibTableObject.data()
    };
    
    wxArrayString AstrLibraryName;
    AstrLibraryName.push_back("Ops");
    AstrLibraryName.push_back("Ops.ActiveObj");
    AstrLibraryName.push_back("Ops.SelectedObject");
    AstrLibraryName.push_back("Data");
    AstrLibraryName.push_back("ParamGroup");
    AstrLibraryName.push_back("Object");

    mpCLua->OpenLuaBase(strFunction, 6, AstrLibraryName, LuaTables);
    
    // scriptƂO̊֐̃AhXQbgāAX^bNɐς
    lua_getglobal(mpCLua->mpLua, "script"); 
    if (lua_pcall(mpCLua->mpLua, 0, 0, 0) != 0) mpCLua->COutErrorMessage();

    wxString strOutput = mpCTempTextCtrl->GetValue();
    if (strOutput.IsEmpty() == FALSE) mpCTextCtrl->AppendText(strOutput);
    mpCTempTextCtrl->Clear();

    bool bSetHistory = TRUE;
    if (mVstrHistory.size() != 0)
	if (mVstrHistory[mVstrHistory.size() - 1].IsSameAs(mstrOneLiner))
	    bSetHistory = FALSE;
    if (bSetHistory == TRUE) mVstrHistory.push_back(mstrOneLiner);
    
    mHistory = mVstrHistory.size();
    mstrOneLiner.Clear();

    mpCTextCtrl->SetInsertionPoint(mpCTextCtrl->GetLastPosition());
    mpCTextCtrl->ShowPosition(mpCTextCtrl->GetLastPosition());
}

void CLS::
ProcessCompletion(wxString strComplete){
    if (mMode == _CONSOLE_LUA_CONSOLE)
	ProcessCompletionLua(strComplete);
}

void CLS::
ProcessEnter(){
    if (mMode == _CONSOLE_LUA_CONSOLE){
	int pos = mpCTextCtrl->GetInsertionPoint();
	long posX, posY;
	int lineLength = GetInsertionPointXY(&posX, &posY);
	mstrOneLiner = mpCTextCtrl->GetLineText(posY);
	int newPos = mpCTextCtrl->XYToPosition(lineLength, posY);
	mpCTextCtrl->SetInsertionPoint(newPos);
    }
}

void CLS::
OnKey(wxKeyEvent& event){
    int keyCode = event.GetKeyCode();
    int modifier = event.GetModifiers();

    if (keyCode == WXK_UP && modifier == wxMOD_CONTROL){
	CallHistory(-1); return; }
    if (keyCode == WXK_DOWN && modifier == wxMOD_CONTROL){
	CallHistory(+1); return; }
    if (keyCode == WXK_SPACE && modifier == wxMOD_CONTROL){
	ProcessCompletion(); return; }
    

    if (keyCode == WXK_RETURN) ProcessEnter();

    event.Skip();
}

/*
void CLS::
ShowUndo(wxCommandEvent& event){
    mpCTextCtrl->SetValue(GetPCUndo()->GetStrRedoAll());
}
*/

void CLS::
OnPopMenu(wxCommandEvent& event){
    wxString strPopMenu = mCPopup.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == FALSE)
	ProcessCompletion(strPopMenu);
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs(wxT("ALL"))){

	// !! FIX_ME !! SXVȂ̂łƌ悭XV
	if (mMode == _CONSOLE_UNDO){
	    mpCTextCtrl->SetValue(GetPCUndo()->GetStrUndoAll(20));
	    SetToDefaultAttr();
	}
	if (mMode == _CONSOLE_REDO){
	    mpCTextCtrl->SetValue(GetPCUndo()->GetStrRedoAll(20));
	    SetToDefaultAttr();
	}
	if (mMode == _CONSOLE_WINDOWS){
	    mpCTextCtrl->SetValue(GetPCFrameMain()->GetStrWindows());
	    SetToDefaultAttr();
	}
	if (mMode == _CONSOLE_WXLOG){
	    mpCTextCtrl->Clear();
	    GCClassProvider.GetPCwxLogGuiEX()->Flush();
	    wxArrayString CArray = GCClassProvider.GetPCwxLogGuiEX()->GetLog();
	    int count = CArray.GetCount();
	    if (count > 100) count = 100;
	    wxString strAllLog;
	    for (uint i = 0; i < count; i++)
		strAllLog += CArray[i];
	    mpCTextCtrl->AppendText(strAllLog/*CArray[i]*/);
	    if (count == 100)
		mpCTextCtrl->AppendText("Over 100: To show more info, clear log");
	    SetToDefaultAttr();
	}
    }
}

void CLS::
OnComboBox(wxCommandEvent& event){
    wxString strValue = mpCComboSelector->GetValue();
    if (strValue.IsSameAs(mAstrSelctor[_CONSOLE_UNDO])){
	mMode = _CONSOLE_UNDO;
	mpCTextCtrl->SetValue(GetPCUndo()->GetStrUndoAll());
	SetToDefaultAttr();
    }
    if (strValue.IsSameAs(mAstrSelctor[_CONSOLE_REDO])){
	mMode = _CONSOLE_REDO;
	mpCTextCtrl->SetValue(GetPCUndo()->GetStrRedoAll());
	SetToDefaultAttr();
    }
    if (strValue.IsSameAs(mAstrSelctor[_CONSOLE_LUA_CONSOLE])){
	mMode = _CONSOLE_LUA_CONSOLE;
	//mpCTextCtrl->Clear(); 
	mpCTextCtrl->SetValue(wxES);
	SetToDefaultAttr();
    }
    if (strValue.IsSameAs(mAstrSelctor[_CONSOLE_WINDOWS])){
	mMode = _CONSOLE_WINDOWS;
	mpCTextCtrl->SetValue(GetPCFrameMain()->GetStrWindows());
	SetToDefaultAttr();
    }
    if (strValue.IsSameAs(mAstrSelctor[_CONSOLE_WXLOG])){
	mMode = _CONSOLE_WXLOG;
	//mpCTextCtrl->Clear();
	mpCTextCtrl->SetValue(wxES);
	SetToDefaultAttr();
	UpdatePanels("ALL");
    }
}

void CLS::
OnMenuLog(wxCommandEvent& event){
    wxMenu CMenu;
    CMenu.Append(menuLogClear, _("Clear Log"));
    CMenu.Append(menuLogUpdate, _("Update Log"));
    mpCTextMenuLog->PopMenu(this, &CMenu);
}

void CLS::
OnMenuLogClear(wxCommandEvent& event){
    GCClassProvider.GetPCwxLogGuiEX()->ClearBuffer();
    mpCTextCtrl->SetValue(wxES);
}

void CLS::
OnMenuLogUpdate(wxCommandEvent& event){
    UpdatePanels("ALL");
}

void CLS::
OnCharHook(wxKeyEvent& event){
    // SkipȂƂɂAHookŏ̊KwL[͂̏̂h
    // DoAllowNextEvent() ɂāAʏ̃L[Cxg
    //printf("TextHook!!\n"); fflush(stdout);

    ProcMarkupCharHook();

    int shortcutEnum = 
	GCClassProvider.GetPCTextShortcut()->GetShortcutEnum(event, GetName());
    if (shortcutEnum != -1){
	long posX, posY;
	int lineLength = GetInsertionPointXY(&posX, &posY);
	if (shortcutEnum == _SC_TEXT_BEGINNING_LINE){
	    int pos0 = mpCTextCtrl->XYToPosition(0, posY);
	    mpCTextCtrl->SetInsertionPoint(pos0);
	}
	if (shortcutEnum == _SC_TEXT_END_LINE){
	    int pos0 = mpCTextCtrl->XYToPosition(lineLength, posY);
	    mpCTextCtrl->SetInsertionPoint(pos0);
	}
	if (shortcutEnum == _SC_TEXT_KILL_LINE){
	    int pos0 = mpCTextCtrl->XYToPosition(lineLength, posY);
	    mpCTextCtrl->SetSelection(mpCTextCtrl->GetInsertionPoint(), pos0);
	    mpCTextCtrl->Cut();
	}
	if (shortcutEnum == _SC_TEXT_SELECT_ALL){
	    mpCTextCtrl->SelectAll();
	}
	
	return; 
    }
    
#ifdef WIN32
    event.DoAllowNextEvent();
#else
    event.Skip();
#endif
}

void CLS::
OnChar(wxKeyEvent& event){
    ProcMarkupChar(event);
    event.Skip();
}

void CLS::
OnTimer(wxTimerEvent& event){
    ProcMarkupTimer();
}

void CLS::
OnIdle(wxIdleEvent& event){
    if (mstrOneLiner.IsEmpty() == FALSE) ProcessOneLiner();
    ProcMarkupIdle();
}
