// -*- c -*-
// "clsTextEntryCtrl.h"
// (setq c-basic-offset 4)

#include "clsTextEntryCtrl.h"
#include "clsAutoComplete.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../../COMMON/clsUseGZ.h"

#include <wx/richtext/richtextctrl.h>

#include "../../INPUTMETHOD/clsSelectorPopup.h"
#include "../../PARAMETER/clsParamGroup.h"
#include "../../PREFERENCE/clsShortcut.h"
#include "../../SCENEOBJ/clsSceneObjCtrl.h"
#include "../../UNDO/clsUndoManager.h"
#include "../../WIDGETCTRL/clsBitmapPanel.h"
#include "../../WIDGETCTRL/EX/clsPanelBar.h"
#include "../../WIDGETCTRL/EX/clsFileDialogEX.h"
#include "../../WIDGETCTRL/INPUTITEM/clsVCenter.h"
#include "../../WIDGETCTRL/WX/myArtProv.h"
#include "../../WIDGETCTRL/WX/clsStringProvider.h"

#include "../../../clsFrameMain.h"
#include "../../../clsMain.h"

#include "../../../wxWidgetEnum.h"
#include "../../../wxWidgetEnumMenu.h"

#include <iostream>
#include "../../LUA/clsLuaBase.h"
#include "../../LUA/clsLuaInterpreter.h"
#include "../../WIDGETCTRL/WX/clsColorProvider.h"

#include <wx/file.h>
#include <wx/splitter.h>

#include "../../COMMON/CommonDefinitions.h"


#define CLS TextEntryCtrl_cls

BEGIN_EVENT_TABLE(TextEntryCtrl_cls, wxPanelEX)
EVT_BUTTON(btnGraphPin, CLS::OnPinMode)
EVT_BUTTON(btnTextClear, CLS::OnTextClear)
EVT_BUTTON(btnTextUndo, CLS::OnTextUndo)
EVT_BUTTON(btnTextRedo, CLS::OnTextRedo)
EVT_BUTTON(btnTextCopy, CLS::OnTextCopy)
EVT_BUTTON(btnTextCut, CLS::OnTextCut)
EVT_BUTTON(btnTextPaste, CLS::OnTextPaste)
EVT_BUTTON(btnTextOpen, CLS::OnTextOpen)
EVT_BUTTON(btnTextSave, CLS::OnTextSave)
EVT_BUTTON(btnTextApply, CLS::OnTextApply)
EVT_BUTTON(btnDropdown, CLS::OnDropDown)
EVT_MENU(menuPopTextShowUndo, CLS::ShowUndo)
EVT_MENU_RANGE(menuPopText0, menuPopText9, CLS::OnPopMenuText)
EVT_MENU_RANGE(menuPop0, menuPop9, CLS::OnPopMenu)
EVT_TIMER(-1, CLS::OnTimer)
EVT_IDLE(CLS::OnIdle)
END_EVENT_TABLE()

CLS::
TextEntryCtrl_cls(wxWindow *pCParent, wxWindowID id, const wxPoint &pos, const wxSize &size, long style) :
wxPanelEX(pCParent, id, pos, size, style),
TextEntryBase_cls() {
    
    SetPanelType(_PANEL_TEXT_EDITOR);
    SetToDefaultParam();

    mpCSplitterWindow = new wxSplitterWindow(this);

    /*
    mpCTempTextCtrl = new wxTextCtrl
	(mpCSplitterWindow, wxID_ANY, wxES, wxDP, wxSize(0,0));
    */
    mpCTextCtrl = new wxRichTextCtrl
	(mpCSplitterWindow, wxID_ANY, wxES, wxDPS, 
	 wxTE_MULTILINE | wxTE_PROCESS_TAB | wxWANTS_CHARS);
    ColorProvider_cls::SetBGColor(mpCTextCtrl, "TextPanel");

    wxString strMessage = "message";
    mpCMessage = new wxStaticText(mpCSplitterWindow, wxID_ANY, strMessage);
    mpCSplitterWindow->SplitHorizontally(mpCTextCtrl, mpCMessage, -40);
    mpCSplitterWindow->SetMinimumPaneSize(16);

    mCTimerMarkup.SetOwner(GetEventHandler());
    SetPMarkupRichTextCtrl(mpCTextCtrl);

    mpCPanelBar = new PanelBar_cls(this);
    mpCPanelBar->SetPanelType(_PANEL_CHILD, _PANEL_TEXT_EDITOR);

    wxBoxSizer *pBoxSizerSrc = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerBar = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerInfo = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerBtn = new wxBoxSizer(wxHORIZONTAL);

    BitmapPanel_cls *pCBtnPinMode = BitmapPanel_cls::GetNewPCButtonPin(this);
    BitmapPanel_cls *pCBtnDropDown = BitmapPanel_cls::GetNewPCButtonDropDown(this);
    VCenter_cls *pCVCenterDropDown = new VCenter_cls(pCBtnDropDown);
    

    mpCLinkedParamGroup = new wxStaticText(this, wxID_ANY, "dummy");
    pBoxSizerInfo->Add(pCBtnPinMode, 0, wxEXPAND);
    pBoxSizerInfo->Add(mpCLinkedParamGroup, 0, wxEXPAND);
    pBoxSizerInfo->AddSpacer(4);
    pBoxSizerInfo->Add(pCVCenterDropDown, 0, wxEXPAND);

    mpCTextCtrl->Connect
	(wxEVT_KEY_DOWN, wxKeyEventHandler(CLS::OnKey),	NULL, this);
    mpCTextCtrl->Connect
	(wxEVT_CHAR, wxKeyEventHandler(CLS::OnChar), NULL, this);
    mpCTextCtrl->Connect
	(wxEVT_CHAR_HOOK, wxKeyEventHandler(CLS::OnCharHook), NULL, this);

    wxBitmapButton *pCBtnOpen = new wxBitmapButton
	(this, btnTextOpen, ARTS("wxART_OPEN_FILE", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnOpen->SetToolTip(_("Open Text"));

    wxBitmapButton *pCBtnSave = new wxBitmapButton
	(this, btnTextSave, ARTS("wxART_SAVE_FILE_AS", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnSave->SetToolTip(_("Save Text"));

    wxBitmapButton *pCBtnClear = new wxBitmapButton
	(this, btnTextClear, ARTS("wxART_CLEAR", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnClear->SetToolTip(_("Clear Text"));

    wxBitmapButton *pCBtnUndo = new wxBitmapButton
	(this, btnTextUndo, ARTS("wxART_UNDO", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnUndo->SetToolTip(_("Undo"));

    wxBitmapButton *pCBtnRedo = new wxBitmapButton
	(this, btnTextRedo, ARTS("wxART_REDO", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnRedo->SetToolTip(_("Redo"));
    
    wxBitmapButton *pCBtnCopy = new wxBitmapButton
	(this, btnTextCopy, ARTS("wxART_TEXT_COPY", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnCopy->SetToolTip(_("Copy to clipboard"));
    
    wxBitmapButton *pCBtnCut = new wxBitmapButton
	(this, btnTextCut, ARTS("wxART_TEXT_CUT", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnCut->SetToolTip(_("Cut to clipboard"));

    wxBitmapButton *pCBtnPaste = new wxBitmapButton
	(this, btnTextPaste, ARTS("wxART_TEXT_PASTE", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnPaste->SetToolTip(_("Paste from clipboard"));
    
    wxBitmapButton *pCBtnApply = new wxBitmapButton
	(this, btnTextApply, ARTS("wxART_TEXT_APPLY", wxSize(24, 24)), wxDP, wxSize(32, 32));
    pCBtnApply->SetToolTip(_("Apply Text"));
    
    pBoxSizerBtn->Add(pCBtnOpen, 0, wxEXPAND);
    pBoxSizerBtn->Add(pCBtnSave, 0, wxEXPAND);
    pBoxSizerBtn->AddSpacer(8);
    pBoxSizerBtn->Add(pCBtnClear, 0, wxEXPAND);
    pBoxSizerBtn->AddSpacer(8);
    pBoxSizerBtn->Add(pCBtnUndo, 0, wxEXPAND);
    pBoxSizerBtn->Add(pCBtnRedo, 0, wxEXPAND);
    pBoxSizerBtn->AddSpacer(8);
    pBoxSizerBtn->Add(pCBtnCopy, 0, wxEXPAND);
    pBoxSizerBtn->Add(pCBtnCut, 0, wxEXPAND);
    pBoxSizerBtn->Add(pCBtnPaste, 0, wxEXPAND);
    pBoxSizerBtn->AddStretchSpacer(1);
    pBoxSizerBtn->Add(pCBtnApply, 0, wxEXPAND);

    mAstrSelctor[0] = _("Lua Script");    
    mAstrSelctor[1] = _("GLSL");
    mAstrSelctor[2] = _("Text"); 
    mpCComboSelector = new wxComboBox(mpCPanelBar, wxID_ANY, mAstrSelctor[0], wxDP, wxSize(140, -1), 3, mAstrSelctor, wxCB_DROPDOWN | wxCB_READONLY);
    
    Connect(mpCComboSelector->GetId(), wxEVT_COMMAND_COMBOBOX_SELECTED,
	    wxCommandEventHandler(CLS::OnComboBox));
    
    ConstructTypeSelector(pBoxSizerBar);
    pBoxSizerBar->Add(mpCComboSelector, 0, wxEXPAND);
    pBoxSizerBar->AddStretchSpacer(1);
    mpCPanelBar->ConstructBarItemHideShow(pBoxSizerBar);
    mpCPanelBar->ConstructBarItemDefault(pBoxSizerBar);

    mpCPanelBar->SetSizer(pBoxSizerBar);

    pBoxSizerSrc->Add(mpCPanelBar, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerInfo, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerBtn, 0, wxEXPAND);
    pBoxSizerSrc->Add(/*mpCTextCtrl*/mpCSplitterWindow, 1, wxEXPAND);
    
    SetSizer(pBoxSizerSrc);

    mMode = _TEXT_LUA_SCRIPT;

    mpCLua = new LuaBase_cls;
}

CLS::
~TextEntryCtrl_cls(){
    mpCLua->CloseLuaBase();
    delete mpCLua;
}

void CLS::
SetToDefaultParam(){
    mCConfList.Clear();
    mCConfList.SetTag("TextEntryCtrl");
    mCConfList.Add(&mCstrLinkedItem, "LinkedItem=", "None");
    mCConfList.Add(&mCbPinned, "Pinned=", "0");
}


void CLS::
ProcessCompletion(wxString strComplete){
    if (mMode == _TEXT_LUA_SCRIPT)
	ProcessCompletionLua(strComplete);
    if (mMode == _TEXT_GLSL)
	ProcessCompletionGLSL(strComplete);
}

void CLS::
OnKey(wxKeyEvent& event){
    int keyCode = event.GetKeyCode();
    int modifier = event.GetModifiers();
    
    if (keyCode == WXK_SPACE && modifier == wxMOD_CONTROL){
	ProcessCompletion(); return; }
    	
    event.Skip();
}

void CLS::
OnPinMode(wxCommandEvent& event){
    long long pinned = event.GetInt();
    mCbPinned.SetBodyFromData(pinned);
}

void CLS::
OnDropDown(wxCommandEvent& event){
    mCSelector.PopupParamGroupSelector(this); }

void CLS::
OnTextClear(wxCommandEvent& event){ 
    mpCTextCtrl->SetSelection(-1, -1);
    mpCTextCtrl->Cut();
}

void CLS::
OnTextUndo(wxCommandEvent& event){ mpCTextCtrl->Undo(); }

void CLS::
OnTextRedo(wxCommandEvent& event){ mpCTextCtrl->Redo(); }

void CLS::
OnTextCopy(wxCommandEvent& event){ mpCTextCtrl->Copy(); }

void CLS::
OnTextCut(wxCommandEvent& event){ mpCTextCtrl->Cut(); }

void CLS::
OnTextPaste(wxCommandEvent& event){ mpCTextCtrl->Paste(); }

void CLS::
OnTextOpen(wxCommandEvent& event){
    wxFileDialogEX *pCDLG = GetPCFileDialog();
    
    pCDLG->MakeDialog(this);
    pCDLG->SetMessage("Select text file");
    pCDLG->SetWildcard("text(*.txt)|*.txt|all|*");
    
    if (pCDLG->ShowModal() == wxID_OK){
	wxString strFile = pCDLG->GetPath();
	pCDLG->DestroyDialog();

	unsigned char *pBuf = NULL;
	long long dummy;
	if (UseGZ_cls::CopyToMemory(&pBuf, strFile, dummy) == FALSE) {
	    wxLogMessage(wxT("Read File Fails ") + strFile);
	    return;
	}

	if (pBuf == NULL) return;

	size_t wcharSize = mbstowcs(NULL, (char*)pBuf, 0);
	wchar_t *wc = new wchar_t [wcharSize+1];
	
	mbstowcs(wc, (char*)pBuf, wcharSize+1);

	delete [] pBuf;
	
	mpCTextCtrl->SetValue(wc);

	delete [] wc;
    }
}

void CLS::
OnTextSave(wxCommandEvent& event){
    wxFileDialogEX *pCDLG = GetPCFileDialog();
    
    pCDLG->MakeDialog(this, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
    pCDLG->SetMessage("Select text file");
    pCDLG->SetWildcard("text(*.txt)|*.txt|all|*");

    if (pCDLG->ShowModal() == wxID_OK){
	wxString strFile = pCDLG->GetPath();
	pCDLG->DestroyDialog();

	wxBeginBusyCursor();

	wxFile CFile(strFile, wxFile::write);
	CFile.Write(mpCTextCtrl->GetValue(), wxConvFile);

	wxMilliSleep(500);
	wxEndBusyCursor();
    }
}

void CLS::
OnTextApply(wxCommandEvent& event){
    ParamGroup_cls *pCParamGroup = GetPCSceneObjCtrl()->GetPCParamGroup(mCstrLinkedItem.GetBodySTR());
    if (pCParamGroup != NULL){
	GetPCMain()->SetParamGroupConfCommand(mCstrLinkedItem.GetBodySTR(),
					      "Expression=",
					      mpCTextCtrl->GetValue());
    }
    else{
	ConfItem_cls *pCConfItem = GetPCSceneObjCtrl()->GetPCConfItem(mCstrLinkedItem.GetBodySTR());

	if (pCConfItem != NULL)
	    pCConfItem->PushCommandPair(mpCTextCtrl->GetValue());
    }
}

void CLS::
ShowUndo(wxCommandEvent& event){
    mpCTextCtrl->SetValue(GetPCUndo()->GetStrRedoAll());
}

void CLS::
Reconnect(){
    ParamGroup_cls *pCParamGroup = GetPCSceneObjCtrl()->GetPCParamGroup(mCstrLinkedItem.GetBodySTR());
    if (pCParamGroup == NULL){

	ConfItem_cls *pCConfItem = GetPCSceneObjCtrl()->GetPCConfItem(mCstrLinkedItem.GetBodySTR());
	
	if (pCConfItem == NULL){
	    mpCMessage->SetLabel("message");
	    return mpCTextCtrl->SetValue(wxES);
	}

	wxString strMessage = "message";
	if (pCConfItem->mstrHeader.IsSameAs("Classify="))
	    strMessage = GetPCStringProvider()->GetStrTextCtrl("Classify");
	if (pCConfItem->mstrHeader.IsSameAs("CustomCoord="))
	    strMessage = GetPCStringProvider()->GetStrTextCtrl("CustomCoord");
	
	mpCTextCtrl->SetValue(pCConfItem->GetBodySTR());
	mpCMessage->SetLabel(strMessage);
    }
    else{
	mpCTextCtrl->SetValue(pCParamGroup->mCstrExpression.GetBodySTR());
	mpCMessage->SetLabel(pCParamGroup->GetStrExpressionMessage());
    }

    Layout();
}

void CLS::
ProcAfterMath(const wxString& strCommand, wxString strItem, bool bForce){
    
    if (strCommand.IsSameAs("DeleteObject")){

	wxString strTgt = mCstrLinkedItem.GetBodySTR();
	wxString strController = PopStrItemFromStrLine(strTgt, "::");
	if (strController.IsSameAs("SceneObjCtrl")){
	    PopStrItemFromStrLine(strTgt, "::");
	    wxString strObj = PopStrItemFromStrLine(strTgt, "::");
	    if (strObj.IsSameAs(strItem)){
		mCstrLinkedItem.SetBodyFromData("None");
		mpCLinkedParamGroup->SetLabel(mCstrLinkedItem.GetBodySTR());
	    }
	    Reconnect();
	}
    }
    if (strCommand.IsSameAs("RenameObject")){
	wxString strNameOld = PopStrItemFromStrLine(strItem, "::");
	wxString strNameNew = strItem;
	
	wxString strTgt = mCstrLinkedItem.GetBodySTR();
	wxString strController = PopStrItemFromStrLine(strTgt, "::");
	if (strController.IsSameAs("SceneObjCtrl")){
	    PopStrItemFromStrLine(strTgt, "::");
	    wxString strObj = PopStrItemFromStrLine(strTgt, "::");
	    if (strObj.IsSameAs(strNameOld)){
		mCstrLinkedItem.mstrBody.Replace(strNameOld, strNameNew);
		mpCLinkedParamGroup->SetLabel(mCstrLinkedItem.GetBodySTR());
	    }
	    Reconnect();
	}
    }
    if (strCommand.IsSameAs("SelectExpression") ||
	strCommand.IsSameAs("SelectConfItem")){
	if (bForce == FALSE && mCbPinned.GetBodyBOOL() == TRUE) return;
	mCstrLinkedItem.SetBodyFromData(strItem);
	mpCLinkedParamGroup->SetLabel(strItem);
	Reconnect();
	UpdatePanels("ALL");
    }
}

void CLS::
UpdatePanels(const wxString& strPanelTag){
    if (GetName().IsSameAs(strPanelTag) ||
	strPanelTag.IsSameAs(wxT("ALL"))){

	Reconnect();
    }
}

void CLS::
OnComboBox(wxCommandEvent& event){
    wxString strValue = mpCComboSelector->GetValue();
    if (strValue.IsSameAs(mAstrSelctor[_TEXT_LUA_SCRIPT])){
	mMode = _TEXT_LUA_SCRIPT;
    }
    if (strValue.IsSameAs(mAstrSelctor[_TEXT_GLSL])){
	mMode = _TEXT_GLSL;
    }
    if (strValue.IsSameAs(mAstrSelctor[_TEXT_TEXT])){
	mMode = _TEXT_TEXT;
    }
}

void CLS::
OnCharHook(wxKeyEvent& event){
    // SkipȂƂɂAHookŏ̊KwL[͂̏̂h
    // DoAllowNextEvent() ɂāAʏ̃L[Cxg
    //printf("TextHook!!\n"); fflush(stdout);

    ProcMarkupCharHook();

    int shortcutEnum = 
	GetPCTextShortcut()->GetShortcutEnum(event, GetName());
    if (shortcutEnum != -1){
	long posX, posY;
	int lineLength = GetInsertionPointXY(&posX, &posY);
	if (shortcutEnum == _SC_TEXT_BEGINNING_LINE){
	    int pos0 = mpCTextCtrl->XYToPosition(0, posY);
	    mpCTextCtrl->SetInsertionPoint(pos0);
	}
	if (shortcutEnum == _SC_TEXT_END_LINE){
	    int pos0 = mpCTextCtrl->XYToPosition(lineLength, posY);
	    mpCTextCtrl->SetInsertionPoint(pos0);
	}
	if (shortcutEnum == _SC_TEXT_KILL_LINE){
	    int pos0 = mpCTextCtrl->XYToPosition(lineLength, posY);
	    mpCTextCtrl->SetSelection(mpCTextCtrl->GetInsertionPoint(), pos0);
	    mpCTextCtrl->Cut();
	}
	if (shortcutEnum == _SC_TEXT_SELECT_ALL){
	    mpCTextCtrl->SelectAll();
	}
	
	return; 
    }
    
#ifdef WIN32
    event.DoAllowNextEvent();
#else
    event.Skip();
#endif
}

void CLS::
OnChar(wxKeyEvent& event){
    ProcMarkupChar(event);
    event.Skip();
}

void CLS::
OnPopMenuText(wxCommandEvent& event){
    wxString strPopMenu = mCPopup.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == FALSE)
	ProcessCompletion(strPopMenu);
}

void CLS::
OnPopMenu(wxCommandEvent& event){
    wxString strPopMenu = mCSelector.GetStr(event.GetId());
    if (strPopMenu.IsEmpty() == TRUE) return;
    
    wxString strCommand = "SelectExpression";
    wxString strItem;
    strItem << GetPCSceneObjCtrl()->GetStrSelf() << "Obj::" << strPopMenu;
    ProcAfterMath(strCommand, strItem, TRUE);
}

void CLS::
OnTimer(wxTimerEvent& event){
    ProcMarkupTimer();
}

void CLS::
OnIdle(wxIdleEvent& event){
    ProcMarkupIdle();
}

void CLS::
Serialize(wxString &strOut){
    ConfListContainer_cls::Serialize(strOut);
}

void CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (ConfListContainer_cls::ReadStrLine(istream, strFirstItem, strLine) == FALSE)
	return;

    mpCLinkedParamGroup->SetLabel(mCstrLinkedItem.GetBodySTR());
}
