// (setq c-basic-offset 4)
// "clsColorProvider.cpp"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "clsColorProvider.h"
#include "../../UNDO/clsUndoManager.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#define CLS ColorProvider_cls

CLS::
ColorProvider_cls(){
    SetToDefault();
}

void CLS::
SetToDefault(){

    wxString strUpdateRender = "VIEW3D::PROPERTY_RENDER";

    mCConfList.SetTag("ColorProvider");
    mCConfList.Add(&mCClrGuideObject, "GuideObject=", "51 179 51");
    mCConfList.Add(&mCClrActive, "Active=", "242 196 9");
    mCConfList.Add(&mCClrSelected, "Selected=", "0 128 255");
    mCConfList.Add(&mCClrTextActive, "TextActive=", "250 236 188");
    mCConfList.Add(&mCClrTextSelected, "TextSelected=", "209 226 242");
    mCConfList.Add(&mCClrText, "Text=", "0 0 0");
    mCConfList.Add(&mCClrTextWeak, "TextWeak=", "128 128 128");
    mCConfList.Add(&mCClrError, "Error=", "255 128 128");
    mCConfList.Add(&mCClrBasePanel, "BasePanel=", "216 216 216");
    mCConfList.Add(&mCClrEditPanel, "EditPanel=", "196 196 196");
    mCConfList.Add(&mCClrBarPanel, "BarPanel=", "240 240 240");
    mCConfList.Add(&mCClrTextPanel, "TextPanel=", "255 255 255");
    mCConfList.Add(&mCClrGLBG, "GLBG=", "0 0 51", strUpdateRender);
    mCConfList.Add(&mCClrMaskBG, "MaskBG=", "0 0 0", strUpdateRender);
    mCConfList.Add(&mCClrGLGrid, "GLGrid=", "128 128 128");
    mCConfList.Add(&mCClrX, "ClrX=", "255 244 244");
    mCConfList.Add(&mCClrY, "ClrY=", "244 255 244");
    mCConfList.Add(&mCClrZ, "ClrZ=", "244 244 255");
    mCConfList.Add(&mCBGAlphaType, "GLBGAlphaType=", "0", strUpdateRender);
}

void CLS::
SetColourCommand(wxString strTag, wxString strColor){
    ConfItemINTARRAY_cls *pCConfItemLink = NULL;
    uint num = mCConfList.mVpCList.size();
    for (uint i = 0; i < num; i++){
	wxString strHeadMain = mCConfList.mVpCList[i]->mstrHeader;
	strHeadMain.Replace("=", wxES);
	if (strTag.IsSameAs(strHeadMain)){
	    pCConfItemLink = (ConfItemINTARRAY_cls*)mCConfList.mVpCList[i];
	    break;
	}
    }
    if (pCConfItemLink == NULL) return;

    wxString strFrom, strTo;
    
    pCConfItemLink->SerializeSingleItem(strFrom);
    pCConfItemLink->SetBodyFromData(strColor);
    pCConfItemLink->SerializeSingleItem(strTo);

    if (strFrom.IsSameAs(strTo) == TRUE) return;

    GetPCUndo()->PushCommandPair(strFrom, strTo);
    GetPCUndo()->Skip(); //łɕύXĂ̂SkipB
}

wxColor CLS::
GetColorMain(wxString strTag){
    if (strTag.IsSameAs("Null")) return wxNullColour;
    if (strTag.IsSameAs("Black")) 
	return *wxBLACK;
    uint num = mCConfList.mVpCList.size();
    for (uint i = 0; i < num; i++){
	wxString strHeadMain = mCConfList.mVpCList[i]->mstrHeader;
	strHeadMain.Replace("=", wxES);
	if (strTag.IsSameAs(strHeadMain)){
	    std::vector < int > VClr = mCConfList.mVpCList[i]->GetBodyINTARRAY();
	    return wxColor(VClr[0], VClr[1], VClr[2]);
	}
    }
    return wxColor (128, 128, 128);
}

wxColor CLS::
GetColor(wxString strTag){
    return GCClassProvider.GetPCColorProvider()->GetColorMain(strTag);
}

std::vector < float > CLS::
GetVGLColor(wxString strTag, bool bSetAlpha){
    wxColor CColor = GetColor(strTag);
    std::vector < float > VClr;
    VClr.push_back((float)CColor.Red() / 255);
    VClr.push_back((float)CColor.Green() / 255);
    VClr.push_back((float)CColor.Blue() / 255);
    if (bSetAlpha == TRUE)
	VClr.push_back(1);
    return VClr;
}

void CLS::
SetGLColor(wxString strTag, bool bSetAlpha){
    if (bSetAlpha == TRUE)
	glColor4fv(GetVGLColor(strTag, TRUE).data());
    else
	glColor3fv(GetVGLColor(strTag, FALSE).data());
}

void CLS::
SetGLClearColor(wxString strTag, bool bSetAlphaInData, float alpha){
    if (bSetAlphaInData == TRUE){
	std::vector < float > VClr = GetVGLColor(strTag, TRUE);
	glClearColor(VClr[0], VClr[1], VClr[2], VClr[3]);
    }
    else{
	std::vector < float > VClr = GetVGLColor(strTag, FALSE);
	glClearColor(VClr[0], VClr[1], VClr[2], alpha);
    }
} 

void CLS::
SetBGColor(wxWindow *pC, wxString strTag){
    pC->SetBackgroundColour(GetColor(strTag));
}

void CLS::
SetBGColorDC(wxDC &dc, wxString strTag){
    dc.SetBackground(wxBrush(GetColor(strTag), wxSOLID));
}

void CLS::
SetTextColorDC(wxDC &dc, bool bSelected){
    if (bSelected == TRUE)
	dc.SetTextForeground(GetColor("Selected"));
    else
	dc.SetTextForeground(GetColor("Text"));
}

void CLS::
SetTextColorWeakDC(wxDC &dc){
    dc.SetTextForeground(GetColor("TextWeak"));
}

ColorProvider_cls *GpCColorProvider = NULL;
