// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsSplittableWindow.h"

#include "clsSplittableWindow.h"

#include "../../COMMON/clsClassProvider.h"
extern ClassProvider_cls GCClassProvider;
#include "../../GRAPH/clsGraphEditor.h"
#include "../../GRAPH/clsColorCurveEditor.h"
#include "../../GLCANVAS/clsView3D.h"
#include "../../MAINCANVAS/clsProperty.h"
#include "../../SCENEOBJ/clsOutliner.h"
#include "../../TIMELINE/clsTimeline.h"
#include "../../GLCANVAS/clsView3D.h"
#include "../../WIDGETCTRL/EX/wxPanelEX.h"
#include "../../WIDGETCTRL/TEXT/clsConsoleCtrl.h"
#include "../../WIDGETCTRL/TEXT/clsTextEntryCtrl.h"

#include "../../../clsFrameMain.h"
#include "../../../wxWidgetEnumMenu.h"

#include <wx/splitter.h>

#include "../../COMMON/CommonDefinitions.h"

extern bool GbLockProjectLoading;

#define CLS SplittableWindow_cls

BEGIN_EVENT_TABLE(SplittableWindow_cls, wxSplitterWindow)
END_EVENT_TABLE()

CLS::
SplittableWindow_cls(wxWindow *pCParentLink) : 
wxSplitterWindow(pCParentLink, wxID_ANY, wxDPS, wxSP_BORDER | wxSP_3DSASH, wxT("SplittableWindow")) {
    mbRootWinodow = TRUE;
    
    mpCParentLink = pCParentLink;
    mpCChildSWindow[0] = NULL;
    mpCChildSWindow[1] = NULL;
    
    SetSashGravity(0);
}

CLS::
SplittableWindow_cls(wxSplitterWindow *pCParentLink) : 
        wxSplitterWindow(pCParentLink, wxID_ANY, wxDPS, wxSP_BORDER | wxSP_3DSASH) {
    mbRootWinodow = FALSE;
    
    mpCParentLink = pCParentLink;
    mpCChildSWindow[0] = NULL;
    mpCChildSWindow[1] = NULL;

    SetSashGravity(0);
}

View3D_cls* CLS::
GetPCActiveView3D(){
    if (IsSplit() == FALSE){
	wxPanelEX* pCPanelEX = (wxPanelEX*)GetWindow1();
	if (pCPanelEX == NULL) return NULL;
	if (pCPanelEX->mPanelType == _PANEL_VIEW3D){
	    if (((View3D_cls*)pCPanelEX)->IsActive() == TRUE)
		return (View3D_cls*)pCPanelEX;
	}
	return NULL;
    }

    View3D_cls *pCActiveView3D = 
	mpCChildSWindow[0]->GetPCActiveView3D();
    if (pCActiveView3D != NULL) return pCActiveView3D;
    
    return mpCChildSWindow[1]->GetPCActiveView3D();
}

wxWindow* CLS::
GeneratePanel(wxString strPanelType, bool bInitialize){
    
    if (IsSplit() == TRUE) return NULL;
	
    wxPanelEX *pCNewPanel = NULL;

    int panelType = wxPanelEX::GetPanelTypeFromStr(strPanelType);
    int scroll = wxHSCROLL|wxVSCROLL;
    switch(panelType){
    case _PANEL_TIMELINE: {
	pCNewPanel = new Timeline_cls(this, wxDefIDPS, 0); break; }
    case _PANEL_OUTLINER: {
	pCNewPanel = new Outliner_cls(this, wxDefIDPS, 0); break;}
    case _PANEL_PROPERTY: {
	pCNewPanel = new Property_cls(this, wxDefIDPS, scroll); break; }
    case _PANEL_VIEW3D: {
	pCNewPanel = new View3D_cls(this, wxDefIDPS, 0); break; }
    case _PANEL_GRAPH: {	
	pCNewPanel = new GraphEditor_cls(this, wxDefIDPS, 0); break; }
    case _PANEL_COLOR_CURVE: {	
	pCNewPanel = new ColorCurveEditor_cls(this, wxDefIDPS, 0); break; }
    case _PANEL_TEXT_EDITOR: {
	pCNewPanel = new TextEntryCtrl_cls(this); break; }
    case _PANEL_CONSOLE: {
	pCNewPanel = new ConsoleCtrl_cls(this); break; }
    case _PANEL_DUMMY:
	pCNewPanel = new wxPanelEX(this, wxDefIDPS, scroll);
	pCNewPanel->mPanelType = _PANEL_DUMMY;
	break;
	
    }

    if (pCNewPanel != NULL)
	pCNewPanel->mpCParentSWindow = this;

    if (bInitialize == FALSE){
	return pCNewPanel;
    }

    Initialize(pCNewPanel);
    return this;
}

void CLS::
GenerateWindows(wxString strPanelType0, wxString strPanelType1){
    
    mpCChildSWindow[0] = new SplittableWindow_cls(this);
    mpCChildSWindow[1] = new SplittableWindow_cls(this);
    
    SplitHorizontally
	(mpCChildSWindow[0]->GeneratePanel(strPanelType0),
	 mpCChildSWindow[1]->GeneratePanel(strPanelType1), -160);
    SetMinimumPaneSize(80);
}

bool CLS::
ReplaceWindowTo(SplittableWindow_cls *pCSWindowOrg, 
		SplittableWindow_cls *pCSWindowNew){
    if (mpCChildSWindow[0] == pCSWindowOrg)
	mpCChildSWindow[0] = pCSWindowNew;
    if (mpCChildSWindow[1] == pCSWindowOrg)
	mpCChildSWindow[1] = pCSWindowNew;
    return ReplaceWindow(pCSWindowOrg, pCSWindowNew);
}

void CLS::
DestroyChildSWindows(SplittableWindow_cls *pCSWindow){
    assert(IsSplit() == TRUE);
    
    if (IsSplit() == TRUE){
	SplittableWindow_cls *pCLastOne = NULL;

	if (pCSWindow == mpCChildSWindow[0]){
	    Unsplit(pCSWindow);
	    pCSWindow->Destroy(); // will be deleted at idle time;
	    pCLastOne = mpCChildSWindow[1];
	}

	if (pCSWindow == mpCChildSWindow[1]){
	    Unsplit(pCSWindow);
	    pCSWindow->Destroy(); // will be deleted at idle time;
	    pCLastOne = mpCChildSWindow[0];
	}
	
	mpCChildSWindow[0] = NULL;
	mpCChildSWindow[1] = NULL;
	
	if (pCLastOne->IsSplit() == FALSE){
	    wxPanelEX *pCLastPanel = (wxPanelEX*)pCLastOne->GetWindow1();
	    pCLastPanel->Reparent(this);
	    pCLastPanel->mpCParentSWindow = this;
	    if (ReplaceWindow(pCLastOne, pCLastPanel) == TRUE)
		pCLastOne->Destroy();
	}
	else {
	    if (mbRootWinodow == TRUE)
		GCClassProvider.GetPCFrameMain()->
		    ReplaceTopSplitterbleWindow(pCLastOne);
	    else{
		pCLastOne->Reparent(mpCParentLink);
		pCLastOne->mpCParentLink = mpCParentLink;
		((SplittableWindow_cls*)mpCParentLink)->
		    ReplaceWindowTo(this, pCLastOne);
		Destroy();
	    }
	    
	}

    }
}

void CLS::
DestroyPanel(wxPanelEX *pCPanel){
    
    assert(IsSplit() == FALSE);

    if (mbRootWinodow == TRUE) return;

    wxWindow *pCTop = wxGetTopLevelParent(this);

    Unsplit(pCPanel);
    pCPanel->Destroy();
    
    ((SplittableWindow_cls*)mpCParentLink)->DestroyChildSWindows(this);

    SendMenuEvent(pCTop, menuSplitterSize);	
    //assert(mbSplit == TRUE);
}

void CLS::
SeparatePanel(bool bVertical){
    assert(IsSplit() == FALSE);

    if (IsSplit() == TRUE) return;

    if (bVertical == TRUE && GetClientSize().GetWidth() < 160) return;
    if (bVertical == FALSE && GetClientSize().GetHeight() < 160) return;

    wxPanelEX *pCChild = (wxPanelEX*)GetWindow1();
    int panelType = pCChild->mPanelType;
    
    mpCChildSWindow[0] = new SplittableWindow_cls(this);
    mpCChildSWindow[1] = new SplittableWindow_cls(this);

    ReplaceWindow(pCChild, mpCChildSWindow[0]);

    pCChild->Reparent(mpCChildSWindow[0]);
    pCChild->mpCParentSWindow = mpCChildSWindow[0];
    
    mpCChildSWindow[0]->Initialize(pCChild);

    if (bVertical == TRUE)
	SplitVertically
	    (mpCChildSWindow[0], 
	     mpCChildSWindow[1]->GeneratePanel(wxPanelEX::msVstrPanelName[panelType]));
    else
	SplitHorizontally
	    (mpCChildSWindow[0], 
	     mpCChildSWindow[1]->GeneratePanel(wxPanelEX::msVstrPanelName[panelType]));
    SetMinimumPaneSize(80);
    SendMenuEvent(wxGetTopLevelParent(this), menuSplitterSize);
}

void CLS::
ChangePanelType(int panelType){
    assert(IsSplit() == FALSE);
    if (IsSplit() == TRUE) return;

    wxPanelEX* pCPanelOrg = (wxPanelEX*)GetWindow1();
    if (pCPanelOrg->mPanelType == panelType) return;

    wxPanelEX* pCPanelNew = (wxPanelEX*)GeneratePanel(wxPanelEX::msVstrPanelName[panelType], FALSE);
    if (pCPanelNew != NULL){
	pCPanelNew->mpCParentSWindow = this;
	if (ReplaceWindow(pCPanelOrg, pCPanelNew) == TRUE)
	    pCPanelOrg->Destroy();
	pCPanelNew->SetFocus();
    }
}

int CLS::
SetMinimumSize(){
    if (IsSplit() == FALSE) return 80;
    
    int minSize0 = mpCChildSWindow[0]->SetMinimumSize();
    int minSize1 = mpCChildSWindow[1]->SetMinimumSize();
	
    int minSize = minSize0;
    if (minSize < minSize1) minSize = minSize1;

    SetMinimumPaneSize(minSize);
    
    return minSize0 + minSize1 + 6;
}

void CLS::
SeparationDirection(){
    if (IsSplit() == FALSE){
	if (mbRootWinodow == FALSE)
	    ((SplittableWindow_cls*)mpCParentLink)->SeparationDirection();
	return;
    }

    wxWindow *pCSecondWindow = GetWindow2();
    if (GetSplitMode() == wxSPLIT_HORIZONTAL){
	Unsplit(pCSecondWindow);
	SplitVertically(GetWindow1(), pCSecondWindow);
    }
    else{
	Unsplit(pCSecondWindow);
	SplitHorizontally(GetWindow1(), pCSecondWindow);
    }
}

void CLS::
SeparationSwap(){
    if (IsSplit() == FALSE){
	if (mbRootWinodow == FALSE)
	    ((SplittableWindow_cls*)mpCParentLink)->SeparationSwap();
	return;
    }

    wxWindow *pCFirstWindow = GetWindow1();
    wxWindow *pCSecondWindow = GetWindow2();
    if (GetSplitMode() == wxSPLIT_HORIZONTAL){
	Unsplit(pCFirstWindow);
	SplitHorizontally(pCSecondWindow, pCFirstWindow);
    }
    else{
	Unsplit(pCFirstWindow);
	SplitVertically(pCSecondWindow, pCFirstWindow);
    }
    ASSERT(mpCChildSWindow[0] == pCFirstWindow);
    ASSERT(mpCChildSWindow[1] == pCSecondWindow);
    SplittableWindow_cls *pCTmp = mpCChildSWindow[0];
    mpCChildSWindow[0] = mpCChildSWindow[1];
    mpCChildSWindow[1] = pCTmp;
}

void CLS::
Serialize(wxString &strOut){
    if (IsSplit() == FALSE){
	strOut << wxT("SplittableWindow::Panel") << ENDL;
	((wxPanelEX*)GetWindow1())->Serialize(strOut);
	strOut << wxT("SplittableWindow::Close") << ENDL;
	return;
    }

    wxString strSplitMode = (GetSplitMode() == wxSPLIT_HORIZONTAL) ? wxT("H::") : wxT("V::") ;
    int sashPos = GetSashPosition();
    strOut << wxT("SplittableWindow::SeparatePanel::") << strSplitMode << sashPos << ENDL;
    strOut << wxT("PanelA") << ENDL;
    mpCChildSWindow[0]->Serialize(strOut);
    strOut << wxT("PanelB") << ENDL;
    mpCChildSWindow[1]->Serialize(strOut);
    strOut << wxT("SplittableWindow::Close") << ENDL;
}

void CLS::
ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine){
    if (strFirstItem.IsSameAs(wxT("Panel")))
	ProceedParsePanel(istream);
    
    if (strFirstItem.IsSameAs(wxT("SeparatePanel"))){
	wxString strDirection = PopStrItemFromStrLine(strLine, wxT("::"));
	long pos;
	PopStrItemFromStrLine(strLine, wxT("::")).ToLong(&pos);
	
	SplittableWindow_cls *pCSWindow0 = NULL, *pCSWindow1 = NULL;
	ProceedParseSplit(istream, &pCSWindow0, &pCSWindow1);
	mpCChildSWindow[0] = pCSWindow0;
	mpCChildSWindow[1] = pCSWindow1;
	
	if (strDirection.IsSameAs(wxT("H")))
	    SplitHorizontally(pCSWindow0 ,pCSWindow1, pos);
	if (strDirection.IsSameAs(wxT("V")))
	    SplitVertically(pCSWindow0 , pCSWindow1, pos);
	SetMinimumPaneSize(80);
    }
}

void CLS::
ProceedParsePanel(std::wistream &istream){
    std::wstring wstrLine;
    while ( std::getline (istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));

	if (strFirstItem.IsSameAs(wxT("wxPanelEX"))){
	    wxString strPanelItem = PopStrItemFromStrLine(strLine, wxT("::"));
	    std::getline (istream, wstrLine);
	    strLine = wstrLine.c_str();
	    strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
	    ((wxPanelEX*)((SplittableWindow_cls*)GeneratePanel(strPanelItem, TRUE))->
	     GetWindow1())->ReadStrLine(istream, strFirstItem, strLine);
	}
	
	if (strFirstItem.IsSameAs(wxT("SplittableWindow")) &&
	    strLine.IsSameAs(wxT("Close")))
	    break;
    }
}

void CLS::
ProceedParseSplit(std::wistream &istream, SplittableWindow_cls **ppCSWindow0, SplittableWindow_cls **ppCSWindow1){
    std::wstring wstrLine;
    while ( std::getline (istream, wstrLine)){
	wxString strLine = wstrLine.c_str();
	wxString strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));

	if (strFirstItem.IsSameAs(wxT("PanelA"))){
	    std::getline (istream, wstrLine);
	    strLine = wstrLine.c_str();
	    wxString strFirstItemA = PopStrItemFromStrLine(strLine, wxT("::"));
	    ASSERT(strFirstItemA.IsSameAs(wxT("SplittableWindow")));
	    strFirstItemA = PopStrItemFromStrLine(strLine, wxT("::"));

	    (*ppCSWindow0) = new SplittableWindow_cls(this);
	    (*ppCSWindow0)->Hide();
	    (*ppCSWindow0)->ReadStrLine(istream, strFirstItemA, strLine);
	    (*ppCSWindow0)->Show();
	}

	if (strFirstItem.IsSameAs(wxT("PanelB"))){
	    std::getline (istream, wstrLine);
	    strLine = wstrLine.c_str();
	    wxString strFirstItemB = PopStrItemFromStrLine(strLine, wxT("::"));
	    ASSERT(strFirstItemB.IsSameAs(wxT("SplittableWindow")));
	    strFirstItemB = PopStrItemFromStrLine(strLine, wxT("::"));
	    
	    (*ppCSWindow1) = new SplittableWindow_cls(this);
	    (*ppCSWindow1)->Hide();
	    (*ppCSWindow1)->ReadStrLine(istream, strFirstItemB, strLine);
	    (*ppCSWindow1)->Show();
	}

	if (strFirstItem.IsSameAs(wxT("SplittableWindow")) &&
	    strLine.IsSameAs(wxT("Close")))
	    break;
    }
}

bool CLS::
IsInnertiaExist(){
    bool bExist = FALSE;

    if (IsSplit() == TRUE){
	bExist |= mpCChildSWindow[0]->IsInnertiaExist();
	bExist |= mpCChildSWindow[1]->IsInnertiaExist();
	return bExist;
    }

    wxWindow *pCWindow = GetWindow1();
    if (pCWindow != NULL)
	bExist |= ((wxPanelEX*)pCWindow)->IsInnertiaExist();

    return bExist;
}

void CLS::
UpdatePanels(const wxString& strPanelTag, bool bUpdateBar){
    if (IsSplit() == TRUE){
	mpCChildSWindow[0]->UpdatePanels(strPanelTag, bUpdateBar);
	mpCChildSWindow[1]->UpdatePanels(strPanelTag, bUpdateBar);
	return;
    }
    
    wxWindow *pCWindow = GetWindow1();
    if (pCWindow != NULL)
	((wxPanelEX*)pCWindow)->UpdatePanels(strPanelTag, bUpdateBar);
}

/*
void CLS::
ProcRename(const wxString& strNameOld, const wxString& strNameNew){
    if (IsSplit() == TRUE){
	mpCChildSWindow[0]->ProcRename(strNameOld, strNameNew);
	mpCChildSWindow[1]->ProcRename(strNameOld, strNameNew);
	return;
    }
    
    wxWindow *pCWindow = GetWindow1();
    ((wxPanelEX*)GetWindow1())->ProcRename(strNameOld, strNameNew);
}

void CLS::
ProcDelete(const wxString& strName){
    if (IsSplit() == TRUE){
	mpCChildSWindow[0]->ProcDelete(strName);
	mpCChildSWindow[1]->ProcDelete(strName);
	return;
    }
    
    wxWindow *pCWindow = GetWindow1();
    ((wxPanelEX*)GetWindow1())->ProcDelete(strName);
}
*/

void CLS::
ProcAfterMath(const wxString& strCommand, const wxString& strItem){
    if (IsSplit() == TRUE){
	mpCChildSWindow[0]->ProcAfterMath(strCommand, strItem);
	mpCChildSWindow[1]->ProcAfterMath(strCommand, strItem);
	return;
    }

    wxWindow *pCWindow = GetWindow1();
    if (pCWindow == NULL) return;
    ((wxPanelEX*)GetWindow1())->ProcAfterMath(strCommand, strItem);
}

void CLS::
CheckAbortClicked(const wxPoint& mousePos){
    if (IsSplit() == TRUE){
	mpCChildSWindow[0]->CheckAbortClicked(mousePos);
	mpCChildSWindow[1]->CheckAbortClicked(mousePos);
	return;
    }

    wxWindow *pCWindow = GetWindow1();
    if (pCWindow == NULL) return;
    ((wxPanelEX*)GetWindow1())->CheckAbortClicked(mousePos);
}
