// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsStringProvider.h"

#include "clsStringProvider.h"

#include "../../COMMON/CommonDefinitionsShader.h"

#define CLS StringProvider_cls

CLS::
StringProvider_cls(){
    AstrKeyFrameType[0] = _("Step");
    AstrKeyFrameType[1] = _("Linear");
    AstrKeyFrameType[2] = _("Hermite (AUTO)");

    AstrPrimitiveType[0] = "Empty";
    AstrPrimitiveType[1] = "Box";
    AstrPrimitiveType[2] = "Sphere";
    AstrPrimitiveType[3] = "Plane";
    AstrPrimitiveType[4] = "Cone";
    AstrPrimitiveType[5] = "Cylinder";

    AstrCameraType[0] = _("Perspective");
    AstrCameraType[1] = _("Orthographic");
    AstrCameraType[2] = _("Fisheye4");
    AstrCameraType[3] = _("Fisheye3");
    AstrCameraType[4] = _("Dual Fisheye");
    AstrCameraType[5] = _("Spherical Fisheye");
    AstrCameraType[6] = _("Equirecutangular");

    AstrFocalBlurType[0] = _("Numerical");
    AstrFocalBlurType[1] = _("Origin");
    AstrFocalBlurType[2] = _("Object");
    AstrFocalBlurType[3] = _("Stereo Focus");

    AstrStereoType[0] = _("No Stereo");
    AstrStereoType[1] = _("Off Axis");
    AstrStereoType[2] = _("Toe-in");

    AstrEyeType[0] = _("No Parallax");
    AstrEyeType[1] = _("LEFT");
    AstrEyeType[2] = _("RIGHT");
    AstrEyeType[3] = _("LINE BY LINE");
    AstrEyeType[4] = _("RED-CYAN");
    AstrEyeType[5] = _("YELLOW-BLUE");
    AstrEyeType[6] = _("SIDE BY SIDE");
    AstrEyeType[7] = _("QUAD BUFFERING");

    AstrBufferSize[0] = "256x256";
    AstrBufferSize[1] = "512x512";
    AstrBufferSize[2] = "1024x1024";
    AstrBufferSize[3] = "2048x2048";
    AstrBufferSize[4] = "4096x4096";
    AstrBufferSize[5] = "8192x8192";
    AstrBufferSize[6] = "16384x16384";

    AstrFogType[0] = _("No Fog");
    AstrFogType[1] = "EXP";
    AstrFogType[2] = "EXP2";
    AstrFogType[3] = "LINEAR";

    AstrShaderObject[0] = _("Classic");
    AstrShaderObject[1] = _("Gouroud");
    AstrShaderObject[2] = _("Phong");
    AstrShaderObject[3] = _("Toon");
    AstrShaderObject[4] = _("Normal Map");
    AstrShaderObject[5] = _("Custom");

    AstrShaderParticlePoint[_POINTSHADER_POINT] = _("Point");
    AstrShaderParticlePoint[_POINTSHADER_CIRCLE] = _("Circle");
    AstrShaderParticlePoint[_POINTSHADER_PSEUDO_SPHERE] = _("PseudoSphere");
    AstrShaderParticlePoint[_POINTSHADER_HIDE] = _("No Rendering");
    
    AstrShaderParticleSphere[_SPHERESHADER_GOURAUD] = _("Gouraud");
    AstrShaderParticleSphere[_SPHERESHADER_PHONG]   = _("Phong");
    AstrShaderParticleSphere[_SPHERESHADER_TOON]    = _("Toon");
    AstrShaderParticleSphere[_SPHERESHADER_NORMAL]  = _("Normal");
    AstrShaderParticleSphere[_SPHERESHADER_HIDE]    = _("No Rendering");

    AstrShaderParticleBillboard[_BILLBOARDSHADER_BILLBOARD] = _("Billboard");
    AstrShaderParticleBillboard[_BILLBOARDSHADER_PSEUDO_SPHERE] = _("PseudoSphere");
    AstrShaderParticleBillboard[_BILLBOARDSHADER_PSEUDO_OBJECT] = _("PseudoObject");
    AstrShaderParticleBillboard[_BILLBOARDSHADER_HIDE] = _("No Rendering");

    AstrShaderDissolve[0] = _("No Dissolve");
    AstrShaderDissolve[1] = _("Dissolve");
    AstrShaderDissolve[2] = _("Full Dissolve");

    AstrBlend[0] = _("Alpha Blend");
    AstrBlend[1] = _("Additive");
    AstrBlend[2] = _("Mix");

    AstrInterp[0] = _("Step");
    AstrInterp[1] = _("Linear");
    AstrInterp[2] = _("Hermite");
    AstrInterp[3] = _("Bezier");
    AstrInterp[4] = _("Bezier (Smooth)");
     
    AstrDefaultKey[0] = _("Linear");
    AstrDefaultKey[1] = _("Ease"); AstrDefaultKey[1] << " 1";
    AstrDefaultKey[2] = _("Ease"); AstrDefaultKey[2] << " 0";
    AstrDefaultKey[3] = _("Step");
    AstrDefaultKey[4] = _("Auto");

    AstrHFChannel[0] = "R";
    AstrHFChannel[1] = "G";
    AstrHFChannel[2] = "B";
    AstrHFChannel[3] = _("Luminocity");
    AstrHFChannel[4] = _("Norm");

    AstrTextCtrl[0] = 
	"nt:normalize time, frame:sequence frame, subframe:sub frame, seq:sequence index\nindex:particle index, (x,y,z):position, id:ID, type:type";
    AstrTextCtrl[1] = 
	"#p0#, #p1#, #p2# would be substituted custom parameter 0,1,2";
}

int CLS::
GetIndexFromStr(wxString strTag, wxString strItem){
    int num = -1;
    if (strTag.IsSameAs("DefaultKey")) num = _MAX_DEFAULT_KEY;
    wxString *Astr = GetAstr(strTag);
    if (Astr == NULL) return -1;
    for (int i = 0; i < num; i++)
	if (Astr[i].IsSameAs(strItem)) return i;
    return -1;
}

wxString CLS::
GetStrMenu(wxString strTag){
    wxString strMenu;
    if (strTag.IsSameAs("View"))
	strMenu << "  " << _("View") << "  ";
    if (strTag.IsSameAs("Select"))
	strMenu << "  " << _("Select") << "  ";
    if (strTag.IsSameAs("Timeline"))
	strMenu << " " << _("Timeline") << " ";
    if (strTag.IsSameAs("Channel"))
	strMenu << "  " << _("Channel") << "  ";
    if (strTag.IsSameAs("Camera"))
	strMenu << "  " << _("Camera") << "  ";
    if (strTag.IsSameAs("Navigator"))
	strMenu << "  " << _("Navigator") << "  ";
    if (strTag.IsSameAs("Log"))
	strMenu << "  " << _("Log") << "  ";
    if (strTag.IsSameAs("Key"))
	strMenu << " " << _("Key") << " ";
    if (strTag.IsSameAs("Edit"))
	strMenu << " " << _("Edit") << " ";
    return strMenu;
}

wxString CLS::
GetStrTextCtrl(wxString strTag){
    wxString strMessage;
    if (strTag.IsSameAs("Classify"))
	strMessage << AstrTextCtrl[0];
    if (strTag.IsSameAs("CustomCoord"))
	strMessage << AstrTextCtrl[1];
    return strMessage;
}

wxString* CLS::
GetAstr(wxString strTag){
    if (strTag.IsSameAs("KeyFrameType"))  return AstrKeyFrameType;
    if (strTag.IsSameAs("PrimitiveType")) return AstrPrimitiveType;
    if (strTag.IsSameAs("CameraType"))    return AstrCameraType;
    if (strTag.IsSameAs("FocalBlurType")) return AstrFocalBlurType;
    if (strTag.IsSameAs("StereoType"))    return AstrStereoType;
    if (strTag.IsSameAs("EyeType"))       return AstrEyeType;
    if (strTag.IsSameAs("BufferSize"))    return AstrBufferSize;
    if (strTag.IsSameAs("FogType"))       return AstrFogType;
    if (strTag.IsSameAs("ShaderObject"))  return AstrShaderObject;
    if (strTag.IsSameAs("ShaderDissolve")) return AstrShaderDissolve;
    if (strTag.IsSameAs("BlendType"))      return AstrBlend;
    if (strTag.IsSameAs("InterpType"))     return AstrInterp;
    if (strTag.IsSameAs("DefaultKey"))     return AstrDefaultKey;
    if (strTag.IsSameAs("HFChannel"))      return AstrHFChannel;
    
    if (strTag.IsSameAs("ShaderParticlePoint"))
	return AstrShaderParticlePoint;
    if (strTag.IsSameAs("ShaderParticleBillboard"))
	return AstrShaderParticleBillboard;
    if (strTag.IsSameAs("ShaderParticleSphere"))
	return AstrShaderParticleSphere;
    
    return NULL;
}

wxArrayString CLS::
GetArrayString(wxString strTag){
    wxArrayString Astr;
    if (strTag.IsSameAs("RGBA")){
	Astr.push_back("R"); Astr.push_back("G");
	Astr.push_back("B"); Astr.push_back("A");
    }
    if (strTag.IsSameAs("COORDINATE")){
	Astr.push_back("local");
	Astr.push_back("global");
	Astr.push_back("parent");
    }
    
    return Astr;
}

StringProvider_cls* GpCStringProvider = NULL;
