// -*- c -*- 
// "clsBitmapPanel.h"

#include "clsBitmapPanel.h"
#include "../SCENEOBJ/DATAVIEWER/clsColorCurve.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#include "../COMMON/CommonDefinitions.h"

#include "../../wxWidgetEnum.h"

#define CLS BitmapPanel_cls

BEGIN_EVENT_TABLE(CLS, wxPanel)
     EVT_PAINT       (CLS::OnPaint)
     EVT_MOUSE_EVENTS(CLS::OnMouse)
END_EVENT_TABLE()

CLS::
BitmapPanel_cls(wxWindow *parent, wxWindowID id, 
	       wxPoint pos, wxSize size) :
    wxPanel(parent, id, pos, size, wxNO_BORDER)
{
    SetBackgroundStyle(wxBG_STYLE_PAINT);
    SetBackgroundColour(parent->GetBackgroundColour());
    mbDealPushed = TRUE;
    mbButton = FALSE;
    mbBorder = FALSE;
    mbPushed = FALSE;
    mState = 0;
    mBorderX = 0; mBorderY = 0;
    mColorCurveAlpha = _COLOR_CURVE_RGB;
    mpCColorCurveLink = NULL;
}

void CLS::
SetBorder(int borderX, int borderY){
    mBorderX = borderX;
    mBorderY = borderY;
}

void CLS::
SetBitmapStr(int index, const char *pChar, const char *pCharPushed){
    mVstrBitmap[index] = wxString::FromAscii(pChar);
    if (pCharPushed == NULL)
	mVstrBitmapPushed[index] = wxString::FromAscii(pChar);
    else
	mVstrBitmapPushed[index] = wxString::FromAscii(pCharPushed);
    Refresh();
}

void CLS::
AddBitmapStr(const char *pChar, const char *pCharPushed){
    mVstrBitmap.push_back(wxString::FromAscii(pChar));
    if (pCharPushed == NULL)
	mVstrBitmapPushed.push_back(wxString::FromAscii(pChar));
    else
	mVstrBitmapPushed.push_back(wxString::FromAscii(pCharPushed));
}

void CLS::
SetBitmapStrDisable(wxString strBitmap){
    mstrBitmapDisable = strBitmap;
}

void CLS::
SetStatus(unsigned int status){
    mState = status;
    Refresh(TRUE);
}

void CLS::
SetColorCurveLink(ColorCurve_cls *pCColorCurve){
    mpCColorCurveLink = pCColorCurve;
}

int CLS::
GetDevCoordFromN(double normedValue, int offsetX){
    int panelWidth = mCSize.GetWidth();
    return (int)(normedValue * (panelWidth-(offsetX + mBorderX + mBorderX)) +
		 offsetX + mBorderX);
}

void CLS::
DrawColorCurve(wxAutoBufferedPaintDC& dc){

    int width = mCSize.GetWidth() - mBorderX - mBorderX;
    int height = mCSize.GetHeight();

    mpCColorCurveLink->PrecalcBegin();
    int mode = mColorCurveAlpha;
    if (mode == _COLOR_CURVE_RGB){
	for (int i = 0; i < width; i++){
	    double x0 = (double)i / width;
	    double x1 = (double)(i+1) / width;
	    int posX0 = GetDevCoordFromN(x0, 0);
	    int posX1 = GetDevCoordFromN(x1, 0);
	    std::vector < double > VClr =
		mpCColorCurveLink->GetVInterpolatedValue(x0);
	    wxColour CClr(VClr[0] * 255, VClr[1] * 255, VClr[2] * 255);
	    dc.SetPen(CClr);
	    dc.SetBrush(CClr);
	    dc.DrawRectangle(posX0, 0, posX1-posX0, 24);
	}
    }
    if (mode == _COLOR_CURVE_RGBA){
	for (int i = 0; i < width; i++){
	    double x0 = (double)i / width;
	    double x1 = (double)(i+1) / width;
	    int posX0 = GetDevCoordFromN(x0, 0);
	    int posX1 = GetDevCoordFromN(x1, 0);
	    std::vector < double > VClr = 
		mpCColorCurveLink->GetVInterpolatedValue(x0);
	    double alpha = VClr[3];
	
	    int bg0 = (posX0 / 14) % 2 * 64 + 98;
	    int bg1 = ((posX0 / 14) + 1) % 2 * 64 + 98;
	    
	    wxColour CClr(bg0 * (1.0-alpha) + alpha * VClr[0] * 255,
			  bg0 * (1.0-alpha) + alpha * VClr[1] * 255,
			  bg0 * (1.0-alpha) + alpha * VClr[2] * 255);
	    dc.SetPen(CClr);
	    dc.SetBrush(CClr);
	    dc.DrawRectangle(posX0, mBorderY, posX1-posX0, height/2-mBorderY);
	    
	    wxColour CClr2(bg1 * (1.0-alpha) + alpha * VClr[0] * 255,
			   bg1 * (1.0-alpha) + alpha * VClr[1] * 255,
			   bg1 * (1.0-alpha) + alpha * VClr[2] * 255);
	    dc.SetPen(CClr2);
	    dc.SetBrush(CClr2);
	    dc.DrawRectangle(posX0, height/2, posX1-posX0, height/2-mBorderY);
	}
    }

    if (mbPushed == TRUE){
	if (mstrLabel.IsEmpty() == FALSE){
	    wxRect CRect(GetClientSize());
	    dc.DrawLabel(mstrLabel, CRect);
	}
    }

    mpCColorCurveLink->PrecalcEnd();
}

void CLS::
DrawDefault(wxAutoBufferedPaintDC& dc){
    int width = mCSize.GetWidth();
    int height = mCSize.GetHeight();

    wxColour CClr(196, 196, 196);
    dc.SetPen(CClr);
    dc.SetBrush(*wxTRANSPARENT_BRUSH);
    dc.DrawRectangle(0, 0, width, height);
    wxColour CClr2(128, 128, 128);
    dc.SetPen(CClr2);
    dc.DrawRectangle(1, 1, width-2, height-2);
}

void CLS::
UpdateDisplay(wxAutoBufferedPaintDC& dc){

    wxSize size = GetClientSize();
    mCSize = size;
    size.IncBy(-mBorderX*2, -mBorderY*2);
    if (size.GetWidth() < 0 || size.GetHeight() < 0){
	return;
    }

    if (mpCColorCurveLink != NULL)
	return DrawColorCurve(dc);
        
    if (mVstrBitmap.GetCount() == 0) 
	return DrawDefault(dc);
    
    if ((IsEnabled() == FALSE) && (mstrBitmapDisable.IsEmpty() == FALSE)){
	dc.DrawBitmap(ARTS(mstrBitmapDisable, size), mBorderX, mBorderY, TRUE);
	return;
    }

    if (mState >= mVstrBitmap.GetCount()) return;
    if (mVstrBitmap[mState].IsEmpty() == FALSE){
	if (mbPushed == TRUE)
	    dc.DrawBitmap(ARTS(mVstrBitmapPushed[mState], size), mBorderX, mBorderY, TRUE);
	else
	    dc.DrawBitmap(ARTS(mVstrBitmap[mState], size), mBorderX, mBorderY, TRUE);

    }
    else if (mbBorder == TRUE){
	dc.SetPen(wxPen(wxColour(128, 128, 128)));
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	dc.DrawRectangle(0, 0, size.GetWidth()-1, size.GetHeight()-1);
    }
}

void CLS::
OnPaint(wxPaintEvent& event){
    wxAutoBufferedPaintDC dc(this);
    dc.Clear();
    UpdateDisplay(dc);
    event.Skip();
}

void CLS::
OnMouse(wxMouseEvent& event){
    if (event.Leaving() == TRUE && mbDealPushed == TRUE){
	mbPushed = FALSE;
	Refresh(TRUE);
    }

    if (event.ButtonDown(wxMOUSE_BTN_LEFT) && mbDealPushed){
	mbPushed = TRUE;
	Refresh(TRUE);
    }

    if ((event.ButtonUp(wxMOUSE_BTN_LEFT) && mbPushed == TRUE) ||
	event.ButtonDClick(wxMOUSE_BTN_LEFT)){

	mState++;
	if (mState >= mVstrBitmap.GetCount()) mState = 0;
	mbPushed = FALSE;
	Refresh(TRUE);
	
	if (mbButton == TRUE){
	    if (mstrCommandStr.IsEmpty() == TRUE)
		SendBtnEventInt(GetParent(), GetId(), mState);
	    else
		SendBtnEventIntStr(GetParent(), GetId(), mState, mstrCommandStr);
	} 
    }

    event.Skip();
}

BitmapPanel_cls* CLS::
GetNewPCButtonPin(wxWindow *pCParent){
    BitmapPanel_cls *pCBtnPinMode = new BitmapPanel_cls(pCParent, btnGraphPin, wxDP, wxSize(20, 20));
    pCBtnPinMode->AddBitmapStr("wxART_UNPIN_N", "wxART_UNPIN_P");
    pCBtnPinMode->AddBitmapStr("wxART_PIN_N", "wxART_PIN_P");
    pCBtnPinMode->mbButton = TRUE;
    pCBtnPinMode->SetBorder(1, 1);
    return pCBtnPinMode;
}

BitmapPanel_cls* CLS::
GetNewPCButtonDropDown(wxWindow *pCParent, int id){
    if (id == wxID_ANY) id = btnDropdown;
    BitmapPanel_cls *pCBtnDropDown = new BitmapPanel_cls(pCParent, id, wxDP, wxSize(14, 14));
    pCBtnDropDown->AddBitmapStr("wxART_DROPDOWN");
    pCBtnDropDown->mbButton = TRUE;
    pCBtnDropDown->SetBorder(1, 1);
    return pCBtnDropDown;
}

BitmapPanel_cls* CLS::
GetNewPCButtonPanelHide(wxWindow *pCParent){
    wxSize CSize(20, 20);
    BitmapPanel_cls *pCBtnNumPanelHide = new BitmapPanel_cls(pCParent, btnNumPanelHide, wxDP, CSize);
    pCBtnNumPanelHide->AddBitmapStr("wxART_SHOW", "wxART_SHOW_P");
    pCBtnNumPanelHide->mbButton = TRUE;
    pCBtnNumPanelHide->SetBorder(1,1);
    return pCBtnNumPanelHide;
}

BitmapPanel_cls* CLS::
GetNewPCButtonLink(wxWindow *pCParent){
    BitmapPanel_cls *pCBtnLinked = new BitmapPanel_cls(pCParent, btnGraphLinked, wxDP, wxSize(16, 16));
    pCBtnLinked->AddBitmapStr("wxART_LINK_N");
    pCBtnLinked->AddBitmapStr("wxART_LINK_P");
    pCBtnLinked->SetBitmapStrDisable("wxART_LINK_D");
    pCBtnLinked->mbButton = TRUE;
    pCBtnLinked->Disable();
    return pCBtnLinked;
}

BitmapPanel_cls* CLS::
GetNewPCButtonLock(wxWindow *pCParent, int id){
    BitmapPanel_cls *pCBtnLock = new BitmapPanel_cls(pCParent, id, wxDP, wxSize(16, 16));
    pCBtnLock->AddBitmapStr("wxART_LOCK_N");
    pCBtnLock->AddBitmapStr("wxART_LOCK_P");
    pCBtnLock->SetBitmapStrDisable("wxART_LOCK_D");
    pCBtnLock->mbButton = TRUE;
    pCBtnLock->Disable();
    return pCBtnLock;
}
