// -*- c++ -*- 
// "clsFrameMain.cpp"
// (setq c-basic-offset 4)

#ifndef _FRAME_MAIN_H_
#define _FRAME_MAIN_H_

#include "clsMain.h"

#include "SRC/COMMON/clsGetPC.h"
#include "SRC/FILEIO/clsParser.h"
#include "SRC/WIDGETCTRL/clsWidgetCommon.h"

#include <wx/wx.h>
#include <wx/splitter.h>
#include <wx/ribbon/toolbar.h>

#ifdef WIN32
#include <wx/appprogress.h>
#endif

class Main_cls;
class RibbonUI_cls;
class SceneObjCtrl_cls;
class SplittableWindow_cls;
class View3D_cls;

class FrameMain_cls : public wxFrame, public Parser_cls, public WidgetCommon_cls, public GetPC_cls {
 public:
    FrameMain_cls(const wxString& title, Main_cls *pCMain, int posx, int posy,
		  int width, int height);
    ~FrameMain_cls();

    Main_cls *mpCMainLink;
    RibbonUI_cls *mpCRibbonUI;
    SplittableWindow_cls *mpCSplittableWindow;
    wxString mstrTitleBase;
    
    wxBitmapButton *mpCButtonAbort;
    wxGauge *mpCStatusGauge, *mpCOutputGauge;
#ifdef WIN32
    wxAppProgressIndicator *mpCAppGauge;
#endif
    

    void ConstructRibbon();
    
    void ReplaceTopSplitterbleWindow(SplittableWindow_cls *pCNewWindow);
    void UpdatePanels(const wxString& strPanelTag);
    void UpdateTitle();
    void SetGauge(int percentage, bool bForce = FALSE);
    void SetStatusText(wxString strStatus);
    void SetStatusAborted();
    //void ProcRename(wxString strNameOld, wxString strNameNew);
    //void ProcDelete(wxString strName);
    //void ProcSelectParamGroup(wxString strTag);
    void ProcAfterMath(wxString strCommand, wxString strItem);
    void CheckAbortion();
    
    View3D_cls* GetPCActiveView3D();

    void OnRbnOpenProject  (wxRTBEvt& event);
    void OnRbnOpenRecent   (wxRTBEvt& event);
    void OnRbnSaveProject  (wxRTBEvt& event);
    void OnRbnSaveProjectAs(wxRTBEvt& event);
    void OnRbnScene        (wxRTBEvt& event);
    void OnRbnUndo         (wxRTBEvt& event);
    void OnRbnRedo         (wxRTBEvt& event);
    void OnRbnObjectBox    (wxRTBEvt& event);
    void OnRbnObjectSphere (wxRTBEvt& event);
    void OnRbnObjectPlane  (wxRTBEvt& event);
    void OnRbnObjectCone   (wxRTBEvt& event);
    void OnRbnObjectCylinder(wxRTBEvt& event);
    void OnRbnObjectMesh   (wxRTBEvt& event);
    void OnRbnObjectText   (wxRTBEvt& event);
    void OnRbnLightParallel(wxRTBEvt& event);
    void OnRbnLightPoint   (wxRTBEvt& event);
    void OnRbnLightSpot    (wxRTBEvt& event);
    void OnRbnCamera       (wxRTBEvt& event);
    void OnRbnWorld        (wxRTBEvt& event);
    void OnRbnVolumeBox    (wxRTBEvt& event);
    void OnRbnIsosurface   (wxRTBEvt& event);
    void OnRbnParticle     (wxRTBEvt& event);
    void OnRbnStreamline   (wxRTBEvt& event);
    void OnRbnHeightFiled  (wxRTBEvt& event);
    void OnRbnEmpty        (wxRTBEvt& event);
    void OnRbnDuplicate    (wxRTBEvt& event);
    void OnRbnObjectsDelete(wxRTBEvt& event);
    void OnRbnInfo         (wxRTBEvt& event);
    void OnRbnHelp         (wxRTBEvt& event);
    void OnRbnToolTip      (wxRTBEvt& event);
    void OnRbnInfoGL       (wxRTBEvt& event);
    void OnRbnInfoStatus   (wxRTBEvt& event);
    void OnRbnPreference   (wxRTBEvt& event);
    void OnRbnClearCacheTexture(wxRTBEvt& event);
    void OnRbnClearCacheVBO(wxRTBEvt& event);
    
    void OnDropRecent      (wxRTBEvt& event);
    void OnDropSaveFile    (wxRTBEvt& event);
    void OnDropObjects     (wxRTBEvt& event);
    void OnDropCache       (wxRTBEvt& event);
    void OnDropInfo        (wxRTBEvt& event);

    void OnButtonAbort      (wxComEvt& event);

    void OnMenuCommon       (wxComEvt& event);
    void OnMenuRecent       (wxComEvt& event);
    void OnMenuSplitterSize (wxComEvt& event);
    void OnMenuFromSubThread(wxComEvt& event);
    void OnMenuLog          (wxComEvt& event);
    void OnKeyDown(wxKeyEvent& event);
    void OnSizeFrameMain(wxSizeEvent& event);
    void OnClose(wxCloseEvent& event);
    void MakeSinglePanel();
    void OnIdle(wxIdleEvent& event);
    
    wxString GetStrWindows();
  
    void Serialize(wxString &strOut, int index);
    void ReadStrLine(std::wistream &istream, wxString strFirstItem, wxString strLine);
    void ProceedParse(std::wistream &istream, wxString strLine);

    DECLARE_EVENT_TABLE()
};

#endif
