// -*- c++ -*-
// (setq c-basic-offset 4)
// "clsMain.h"

#include "clsMain.h"
#include "clsFrameMain.h"

#include "SRC/COMMON/clsClassProvider.h"
#include "SRC/COMMON/clsQueue.h"
#include "SRC/COMMON/clsUseGZ.h"
#include "SRC/FILTER/clsFilterCtrl.h"
#include "SRC/GRAPHIC/clsImageOutput.h"
#include "SRC/GLCANVAS/clsGLCanvas.h"
#include "SRC/MAINCANVAS/clsRibbonUI.h"
#include "SRC/MISC/clsInitializeParallel.h"
#include "SRC/MYGL/clsShadowVolumeCtrl.h"
#include "SRC/PREFERENCE/clsPreference.h"
#include "SRC/PREFERENCE/clsShortcut.h"
#include "SRC/PREFERENCE/clsMousePreference.h"
#include "SRC/SCENEOBJ/clsSceneObjCtrl.h"
#include "SRC/TIMELINE/clsEnsembleCtrl.h"
#include "SRC/TIMELINE/clsMotionBlurCtrl.h"
#include "SRC/TIMELINE/clsTimeCtrl.h"
#include "SRC/UNDO/clsUndoManager.h"
#include "SRC/WIDGETCTRL/EX/clsFileDialogEX.h"
#include "SRC/WIDGETCTRL/EX/wxPanelEX.h"
#include "SRC/WIDGETCTRL/WX/myArtProv.h"
#include "SRC/WIDGETCTRL/WX/clsColorProvider.h"
#include "SRC/WIDGETCTRL/WX/clsFontProvider.h"
#include "SRC/WIDGETCTRL/WX/clsIconProvider.h"

#include "wxWidgetString.h"

#include <fstream>
#include <sstream>
#include <wx/filename.h>
#include <wx/file.h>
#include <wx/stdpaths.h>
#include <wx/textfile.h>
#include <wx/tooltip.h>

#include "SRC/COMMON/CommonDefinitions.h"
#include "wxWidgetEnum.h"
#include "wxWidgetEnumMenu.h"

ClassProvider_cls GCClassProvider;
bool GbLockProjectLoading = FALSE;

#define CLS Main_cls

CLS::
Main_cls() 
{
    ASSERT((RBTN_COMMON_FIN - RBTN_COMMON_INI) == (MENU_COMMON_FIN - MENU_COMMON_INI));
    mActiveGLViewID = -1;
    mHungAppTimeoutMS = -1;
    mTdrDelay = -1;

    mCStopWatch.Start();

    mbHasOpenedProject = FALSE;
    GCClassProvider.mpCMainLink = this;

    wxFileName CFileExe(wxStandardPathsBase::Get().GetExecutablePath());
    mstrExecuteDir = CFileExe.GetPath();
    mstrInitWorkingDir = wxGetCwd();

    wxArtProvider::Push(new MyArtProvider(wxGetCwd()));
    FontProvider_cls::InitializeStatic();
    //wxLog::SetLogLevel(0);
    wxPanelEX::InitializeStatic();

    wxToolTip::SetAutoPop(10);
    wxToolTip::SetDelay(1000);
    wxToolTip::SetReshow(10);
    
    // wxFramȅ́A܂wxFrameȂ̂
    // ̊ԎgȂ Singleton ̓pɏłB
    // ColorProvider Ȃǂ̌ɕ񏈗ĕCȂ͂B
    bool bFinished = FALSE;
    InitializeParallel_cls *pCInit = new InitializeParallel_cls;
    pCInit->SetParameters(&bFinished, &mInitializeTimeParallel);
    pCInit->Run();

    mCConfList.SetTag(wxT("MainConfList"));
    mCConfList.Add(&mCLastOpenedFileZJP,  "LastFileZJP=",  wxES);
    mCConfList.Add(&mCLastOpenedFileIMG,  "LastFileIMG=",  wxES);
    mCConfList.Add(&mCLastOpenedFilePARTICLE, "LastFilePARTICLE=", wxES);
    mCConfList.Add(&mCLastOpenedFileVOLUME, "LastFileVOLUME=", wxES);
    mCConfList.Add(&mCLastOpenedFileMESH, "LastFileMESH=", wxES);
    mCConfList.Add(&mCLastOpenedFileHF,   "LastFileHF=",   wxES);
    mCConfList.Add(&mCLastOpenedFileSTREAM, "LastFileSTREAM=", wxES);
    mCConfList.Add(&mCVersionProject, "VersionProject=", _VERSION);
    mCLastOpenedFileZJP.mbFilePath = TRUE;
    mCLastOpenedFileIMG.mbFilePath = TRUE;
    mCLastOpenedFilePARTICLE.mbFilePath = TRUE;
    mCLastOpenedFileVOLUME.mbFilePath = TRUE;
    mCLastOpenedFileMESH.mbFilePath = TRUE;
    mCLastOpenedFileHF.mbFilePath = TRUE;
    mCLastOpenedFileSTREAM.mbFilePath = TRUE;

    wxRect CRectFrame(0, 0, 1280, 960);
    mVpCFrame.push_back
	(new FrameMain_cls(_FRAME_MAIN_APPNAME, this, 
			   CRectFrame.GetX(), CRectFrame.GetY(),
			   CRectFrame.GetWidth(), CRectFrame.GetHeight()));

    while (bFinished == FALSE) wxMilliSleep(3);

    GCClassProvider.CreateSingletonSerial();
    
    mVpCFrame[0]->mpCRibbonUI->UpdateRibbonHelp();

    //GCClassProvider.GetPCFileDialog()->mpCDialog->Reparent(mVpCFrame[0]);
    //wxLogMessage("Test Log::clsMain.cpp\n");

    //mVpCFrame[0]->SetIcon(*(GetPCIconProvider()->GetPCIcon("ARROW")));

    OpenProjectMain(mstrExecuteDir + _WXDELIM_ + _CONFIGURE_FILE, FALSE);
}

CLS::
~Main_cls(){ 
    GCClassProvider.DeleteSingleton(); 
}

bool CLS::
InitLayout(){
    bool bRet = OpenProjectMain(wxT("Resource/InitLayout.zjp"), FALSE);
    mInitializeTimeTotal = mCStopWatch.Time();
    return bRet;
}

ParamGroup_cls* CLS::
GetPCParamGroup(wxString strItem){
    ParamGroup_cls *pCParamGroup = GetPCSceneObjCtrl()->GetPCParamGroup(strItem);
    if (pCParamGroup == NULL){
	//wxString strFirstItem = PopStrItemFromStrLine(strItem, "::");
	pCParamGroup = GetPCFilterCtrl()->GetPCParamGroup(strItem);
	return pCParamGroup;
    }
    return pCParamGroup;
}

Parameter_cls* CLS::
GetPCParameter(wxString strItem){
    Parameter_cls *pCParamer = GetPCSceneObjCtrl()->GetPCParameter(strItem);
    if (pCParamer == NULL){
	pCParamer = GetPCFilterCtrl()->GetPCParameter(strItem);
	return pCParamer;
    }
    return pCParamer;
}

void CLS::
DeselectKeyFrame(wxString strException){
    GetPCSceneObjCtrl()->DeselectKeyFrame(strException);
    GetPCFilterCtrl()->DeselectKeyFrame(strException);
}

bool CLS::
ChangeKeyTypeByModKey(wxString strTag, int keyIndex, 
		      bool bCtrl, bool bShift, bool bAlt){
    wxString strHeader = PopStrItemFromStrLine(strTag, "::");
    if (strHeader.IsSameAs("Scene Filter"))
	return GetPCFilterCtrl()->ChangeKeyTypeByModKey(strTag, keyIndex, bCtrl, bShift, bAlt);
    return GetPCSceneObjCtrl()->ChangeKeyTypeByModKey(strHeader + "::" + strTag, keyIndex, bCtrl, bShift, bAlt);
}

bool CLS::
SelectKeyFrame(wxString strTag, int keyIndex, 
	       bool bSelect, bool bAdditional,
	       double resolution){
    wxString strHeader = PopStrItemFromStrLine(strTag, "::");
    if (strHeader.IsSameAs("Scene Filter"))
	return GetPCFilterCtrl()->SelectKeyFrame(strTag, keyIndex, bSelect, bAdditional, resolution);
    return GetPCSceneObjCtrl()->SelectKeyFrame(strHeader + "::" + strTag, keyIndex, bSelect, bAdditional, resolution);
}

void CLS::
SetSelectedTLKeyTypeCommand(int keyType, double ease){
    GetPCUndo()->OpenBlock("SetSelectedTLKeyType_Main");
    GetPCSceneObjCtrl()->SetSelectedTLKeyTypeCommand(keyType, ease);
    GetPCFilterCtrl()->SetSelectedTLKeyTypeCommand(keyType, ease);
    GetPCUndo()->CloseBlock("SetSelectedTLKeyType_Main");
}

void CLS::
MoveSelectedTLKeyCommand(double dNX, double dY){
    GetPCUndo()->OpenBlock("MoveSelectedTLKey_Main");
    GetPCSceneObjCtrl()->MoveSelectedTLKeyCommand(dNX, dY);
    GetPCFilterCtrl()->MoveSelectedTLKeyCommand(dNX, dY);
    GetPCUndo()->CloseBlock("MoveSelectedTLKey_Main");
}

wxString CLS::
GetStrCommandParamGroup(wxString strTag){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strTag);

    wxString strFirstItem = PopStrItemFromStrLine(strTag, "::");
    if (strFirstItem.IsSameAs("Scene Filter")){
	wxString strFilterName = PopStrItemFromStrLine(strTag, "::");
	return GetPCFilterCtrl()->GetStrSelf() + strFilterName + "::" + pCParamGroup->GetStrName() + "::";
    }

    if (strFirstItem.IsSameAs("SceneObjCtrl")){
	strFirstItem = PopStrItemFromStrLine(strTag, "::");
	ASSERT(strFirstItem.IsSameAs("Obj"));
	strFirstItem = PopStrItemFromStrLine(strTag, "::");
    }
    
    wxString strObjName = strFirstItem;
    wxString strAttribName = PopStrItemFromStrLine(strTag, "::");
    if (strAttribName.IsSameAs("Position") ||	
	strAttribName.IsSameAs("RotationEuler") ||
	strAttribName.IsSameAs("Scale"))
	return GetPCSceneObjCtrl()->GetStrSelf() + "Obj::" + strObjName + "::" + pCParamGroup->GetStrName() + "::";
    return GetPCSceneObjCtrl()->GetStrSelf() + "Obj::" + strObjName + "::" + strAttribName + "::" + pCParamGroup->GetStrName() + "::";
}

bool CLS::
SetParamGroupConfCommand(wxString strTag, wxString strHeader, bool bNew){
    ParamGroup_cls *pCParamGroup = GetPCMain()->GetPCParamGroup(strTag);
    if (pCParamGroup == NULL) return FALSE;
    
    wxString strCommandParamGroup = GetStrCommandParamGroup(strTag);
    bool bOld = pCParamGroup->GetPCItem(strHeader)->GetBodyBOOL();
    wxString strUndo(strCommandParamGroup);
    wxString strRedo(strCommandParamGroup); 
    strUndo << strHeader << bOld;
    strRedo << strHeader << bNew;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
    return TRUE;
}

bool CLS::
SetParamGroupConfCommand(wxString strTag, wxString strHeader, wxString strNew){
    ParamGroup_cls *pCParamGroup = GetPCMain()->GetPCParamGroup(strTag);
    if (pCParamGroup == NULL) return FALSE;
    
    wxString strCommandParamGroup = GetStrCommandParamGroup(strTag);
    wxString strOld = pCParamGroup->GetPCItem(strHeader)->GetBodySTR();
    wxString strNewCompact = strNew.Clone();
    wxString strOldCompact = strOld.Clone();
    strNewCompact.Replace("\n", "\\n", TRUE);
    strOldCompact.Replace("\n", "\\n", TRUE);
    wxString strUndo(strCommandParamGroup);
    wxString strRedo(strCommandParamGroup); 
    strUndo << strHeader << strOldCompact;
    strRedo << strHeader << strNewCompact;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
    return TRUE;
}

void CLS::
SetParameterConfCommand(wxString strTag, wxString strHeader, bool bNew, bool bSelectedOnly){
    ParamGroup_cls *pCParamGroup = GetPCParamGroup(strTag);
    Parameter_cls *pCParam = NULL;

    if (pCParamGroup == NULL) return;
    
    wxString strCommandParamGroup = GetStrCommandParamGroup(strTag);
    wxString strParam = GetLastItem(strTag, "::");
    
    bool bLink = pCParamGroup->mCbLinked.GetBodyBOOL();
    if (bLink == TRUE){
	if (bSelectedOnly == TRUE && 
	    pCParamGroup->IsSelectedKeyExist() == FALSE) return;

	int paramNum = pCParamGroup->mVCParam.size();
	bool bSameCheck = TRUE;
	for (int i = 0; i < paramNum; i++){
	    Parameter_cls& CParam = pCParamGroup->mVCParam[i];
	    bool bOld = CParam.GetPCItem(strHeader)->GetBodyBOOL();
	    if (bOld != bNew) bSameCheck = FALSE;
	}
	if (bSameCheck == TRUE) return;

	GetPCUndo()->OpenBlock(strCommandParamGroup);
	for (int i = 0; i < paramNum; i++){
	    Parameter_cls& CParam = pCParamGroup->mVCParam[i];
	    bool bOld = CParam.GetPCItem(strHeader)->GetBodyBOOL();
	    wxString strUndo(strCommandParamGroup);  
	    wxString strRedo(strCommandParamGroup); 
	    wxString strParam = CParam.GetStrName();
	    strUndo << strParam << "::" << strHeader << bOld;
	    strRedo << strParam << "::" << strHeader << bNew;
	    GetPCUndo()->PushCommandPair(strUndo, strRedo); 
	    GetPCUndo()->Redo();
	}
	GetPCUndo()->CloseBlock(strCommandParamGroup);
	return;
    }

    pCParam = pCParamGroup->GetPCParameter(strParam);
    if (pCParam == NULL) return;
    if (bSelectedOnly == TRUE && 
	pCParam->IsSelectedKeyExist() == FALSE) return;
        
    bool bOld = pCParam->GetPCItem(strHeader)->GetBodyBOOL();
    wxString strUndo(strCommandParamGroup); 
    wxString strRedo(strCommandParamGroup); 
    strUndo << strParam << "::" << strHeader << bOld;
    strRedo << strParam << "::" << strHeader << bNew;
    GetPCUndo()->PushCommandPair(strUndo, strRedo);
    GetPCUndo()->Redo();
}

void CLS::
PushBackRecent(wxString strFile){
    for (int i = 0; i < _MAX_RECENT_FILE; i++){
	if (mAstrRecentFile[i].IsSameAs(strFile) == TRUE){
	    for (int j = i; j > 0; j--)
		mAstrRecentFile[j] = mAstrRecentFile[j-1];
	    mAstrRecentFile[0] = strFile;
	    return;
	}
    }
    
    for (int i = _MAX_RECENT_FILE - 1; i > 0; i--)
	mAstrRecentFile[i] = mAstrRecentFile[i-1];
    mAstrRecentFile[0] = strFile;
}

void Main_cls::
OpenProject(wxString strFile){
    if (strFile.IsEmpty() == TRUE){
	wxFileDialogEX *pCDLG = GetPCFileDialog();

	pCDLG->MakeDialog(NULL);
	pCDLG->SetMessage("Select project file");
#ifdef ZLIB
	wxString strWildcard = "Project file (*.zjp;*.zjp.gz)|*.zjp;*.zjp.gz";
#else
	wxString strWildcard = "Project file (*.zjp)|*.zjp";
#endif
	pCDLG->SetWildcard(strWildcard);

	if (mCLastOpenedFileZJP.mstrBody.IsEmpty() == FALSE){
	    wxFileName CName(mCLastOpenedFileZJP.mstrBody);
	    pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
	}
	
	if (pCDLG->ShowModal() == wxID_OK) strFile = pCDLG->GetPath();
	pCDLG->DestroyDialog();
    }

    if (strFile.IsEmpty() == TRUE) return;
    
    if (OpenProjectMain(strFile) == TRUE){
	wxFileName CName(strFile);
	wxSetWorkingDirectory(CName.GetPath());
    }
}

bool Main_cls::
OpenProjectMain(wxString strFile, bool bFullProject){

    unsigned char *pBuf = NULL;
    wxString strMatch = wxT("*");
    long long dummy;
    if (UseGZ_cls::CopyToMemory(&pBuf, strFile, dummy) == FALSE){
	if (strFile.Matches(strMatch + _CONFIGURE_FILE) == FALSE)
	    wxLogMessage(wxT("Read Configure Fails ") + strFile); 
	return FALSE;
    }
    else{
	if (strFile.Matches(strMatch + _CONFIGURE_FILE) == FALSE &&
	    strFile.Matches(strMatch + "InitLayout.zjp") == FALSE)
	    PushBackRecent(strFile);
    }
    
    if (pBuf == NULL) return FALSE;

    if (bFullProject == TRUE){
	GetPCSceneObjCtrl()->Clear();
	GetPCFilterCtrl()->Clear();
	for (unsigned int i = 1; i < mVpCFrame.size(); i++)
	    mVpCFrame[i]->Destroy();
	mVpCFrame.resize(1);
	/*
	if (GLCanvas_cls::mspCGLContext != NULL){
	    delete GLCanvas_cls::mspCGLContext;
	    GLCanvas_cls::mspCGLContext = NULL;
	}
	*/
    }

    size_t wcharSize = mbstowcs(NULL, (char*)pBuf, 0);
    wchar_t *wc = new wchar_t [wcharSize+1];
    
    mbstowcs(wc, (char*)pBuf, wcharSize+1);
    delete [] pBuf;

    std::wstring strBuf(wc);
    delete [] wc;

    std::wstring::size_type pos = 0;
    while (pos = strBuf.find(L"\r", pos), pos != std::string::npos) {
        strBuf.replace(pos, 1, L"");
        //pos += to.length();
    }
    
    std::wistringstream sstream(strBuf);

    bool bComment = FALSE;
    GbLockProjectLoading = TRUE;
    std::wstring wstrLine;
    while ( std::getline (sstream, wstrLine)){
	wxString strLine = wstrLine.c_str();

	if (bComment == FALSE){
	    int posCommentLine = strLine.Find(wxT("--"));
	    int posCommentRegion = strLine.Find(wxT("/*"));
	    if (posCommentLine == 0) continue;
	    if (posCommentRegion == 0){
		bComment = TRUE;
		continue;
	    }

	    wxString strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));
	    ReadStrCommand(sstream, strFirstItem, strLine);
	    continue;
	}
	
	int posCommentRegionEnd = strLine.Find(wxT("*/"));
	if (posCommentRegionEnd != wxNOT_FOUND)
	    bComment = FALSE;
    }

    GetPCUndo()->Clear();

    UpdatePanels("PROPERTY_OBJECT");
    GetPCPreference()->UpdatePanels("ALL");

    GbLockProjectLoading = FALSE;
    
    if (bFullProject == TRUE){
	mbHasOpenedProject = TRUE;
	mCLastOpenedFileZJP.SetBodyFromData(strFile);
	GetPCFrameMain(0)->UpdateTitle();
    }

    return TRUE;
}

void CLS::
OpenRecentFile(wxString strFile){
    if (wxFile::Exists(strFile) == FALSE) return;
    wxTextFile CFile(strFile);
    if (CFile.Open() == FALSE) return;
    
    int lineCount = 0;
    wxString strLine;
    for (strLine = CFile.GetFirstLine() << ENDL; !CFile.Eof();
	 strLine = CFile.GetNextLine() << ENDL){
	if (strLine.IsEmpty() == TRUE) continue;
	mAstrRecentFile[lineCount] = strLine.Trim();
	lineCount++;
	if (lineCount >= _MAX_RECENT_FILE) break;
    }
}

void Main_cls::
ReadStrCommand(std::wistream &sstream, wxString strFirstItem, wxString strLine){
    /*wxString strFirstItem = PopStrItemFromStrLine(strLine, wxT("::"));*/
    mVpCFrame[0]->ReadStrLine(sstream, strFirstItem, strLine);
    ReadStrLine(sstream, strFirstItem, strLine);
    GetPCColorProvider()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCFilterCtrl()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCImageOutput()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCPreference()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCMousePreference()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCSceneObjCtrl()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCShortcut()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCTextShortcut()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCTimeCtrl()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCBlurCtrl()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCEnsembleCtrl()->ReadStrLine(sstream, strFirstItem, strLine);
    GetPCShadowVolumeCtrl()->ReadStrLine(sstream, strFirstItem, strLine);
}

void Main_cls::
SaveProjectAs(int flag){
    wxFileDialogEX *pCDLG = GetPCFileDialog();

    pCDLG->MakeDialog(NULL, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
    pCDLG->SetMessage(wxT("Select project file name to save"));
#ifdef ZLIB
    wxString strWildcard = wxT("Project file (*.zjp;*.zjp.gz)|*.zjp;*.zjp.gz");
#else
    wxString strWildcard = wxT("Project file (*.zjp)|*.zjp");
#endif
    pCDLG->SetWildcard(strWildcard);
    
    if (mCLastOpenedFileZJP.mstrBody.IsEmpty() == FALSE){
	wxFileName CName = mCLastOpenedFileZJP.mstrBody;
	if (mbHasOpenedProject == TRUE)
	    pCDLG->SetInitPath(CName.GetPath(), CName.GetFullName());
	else
	    pCDLG->SetInitPath(CName.GetPath(), wxT(".zjp"));
    }
    
    if (pCDLG->ShowModal() == wxID_OK){
	wxString strFile = pCDLG->GetPath();
	if (SaveProjectMain(strFile, flag) == TRUE){
	    mbHasOpenedProject = TRUE;
	    mCLastOpenedFileZJP.SetBodyFromData(strFile);
	    
	    GetPCFrameMain(0)->UpdateTitle();
	}
    }

    pCDLG->DestroyDialog();
}

void Main_cls::
SaveProject(int flag){
    if (mCLastOpenedFileZJP.IsEmpty() == TRUE || mbHasOpenedProject == FALSE)
	return SaveProjectAs(flag);
    SaveProjectMain(mCLastOpenedFileZJP.GetBodySTR(), flag);
}

bool Main_cls::
SaveProjectMain(wxString strFileOut, int flag){

    mCVersionProject.SetBodyFromData(_VERSION);

    wxBeginBusyCursor();

    wxString strMatch = wxT("*");
    if (! strFileOut.Matches(strMatch + _CONFIGURE_FILE)){
	mCLastOpenedFileZJP.mstrBody = strFileOut;
	PushBackRecent(strFileOut);

	wxFileName CName(strFileOut);
	wxSetWorkingDirectory(CName.GetPath());
    }

    wxString strOut;
    
    for (unsigned int i = 0; i < mVpCFrame.size(); i++)
	mVpCFrame[i]->Serialize(strOut, i);

    Serialize(strOut);
    GetPCColorProvider()->Serialize(strOut);
    GetPCFilterCtrl()->Serialize(strOut);
    GetPCImageOutput()->Serialize(strOut);
    GetPCPreference()->Serialize(strOut);
    GetPCMousePreference()->Serialize(strOut);
    GetPCSceneObjCtrl()->Serialize(strOut);
    GetPCShortcut()->Serialize(strOut);
    GetPCTextShortcut()->Serialize(strOut);
    GetPCTimeCtrl()->Serialize(strOut);
    GetPCBlurCtrl()->Serialize(strOut);
    GetPCEnsembleCtrl()->Serialize(strOut);
    GetPCShadowVolumeCtrl()->Serialize(strOut);
    
    wxFile CFile(strFileOut, wxFile::write);
    if (CFile.IsOpened() == FALSE){
	//!!_FIX_!! ShowWarining;
	wxEndBusyCursor();
	return FALSE;
    }
    
    CFile.Write(strOut, wxConvFile);

    wxMilliSleep(500);
    wxEndBusyCursor();

    return TRUE;
}

void CLS::
SaveRecentFile(wxString strFileOut){
    wxFile CFile(strFileOut, wxFile::write);
    for (int i = 0; i < _MAX_RECENT_FILE; i++)
	if (mAstrRecentFile[i].IsEmpty() == FALSE){
	    CFile.Write(mAstrRecentFile[i], wxConvFile);
	    CFile.Write(ENDL, wxConvFile);
	}
}

void CLS::
SaveConfigureFile(wxString strFileOut){
    wxFile CFile(strFileOut, wxFile::write);
    wxString strOut;
    Serialize(strOut);
    CFile.Write(strOut, wxConvFile);
}

void Main_cls::
RescaleParameters(double scale){
    GetPCSceneObjCtrl()->RescaleParametersRoot(scale);
}

void Main_cls::
UpdatePanels(wxString/*std::string*/ssPanelList){
    uint num = mAssUpdatePanelQueue.size();
    for (uint i = 0; i < num; i++)
	if (mAssUpdatePanelQueue[i].IsSameAs(ssPanelList)) return;
    mAssUpdatePanelQueue.push_back(ssPanelList);
}

void Main_cls::
UpdatePanelsFlush(){
    while (mAssUpdatePanelQueue.size() != 0){
	wxString strPanelList = mAssUpdatePanelQueue.back();
	mAssUpdatePanelQueue.pop_back();
	
	while(TRUE){
	    wxString strPanelTag = PopStrItemFromStrLine(strPanelList, wxT("::")).Clone();
	    if (strPanelTag.IsEmpty() == TRUE) break;
	    for (unsigned int i = 0; i < mVpCFrame.size(); i++)
		mVpCFrame[i]->UpdatePanels(strPanelTag);
	    GetPCPreference()->UpdatePanels(strPanelTag);
	}
    }

    GetPCQueue()->ClearQueue();
    //Queue_cls::msbQueueSyncronizeGraphNavi = FALSE;
    //Queue_cls::msbQueueAdjustAADuetoFrameMax = FALSE;
}

void Main_cls::
ProcAfterMath(wxString strCommand, wxString strItem){
    for (unsigned int i = 0; i < mVpCFrame.size(); i++)
	mVpCFrame[i]->ProcAfterMath(strCommand, strItem);

    if (strCommand.IsSameAs("Quit")){
	GetPCSceneObjCtrl()->Clear();
	GetPCFilterCtrl()->Clear();
	SaveRecentFile(mstrExecuteDir + _WXDELIM_ + _RECENT_FILE);
	SaveConfigureFile(mstrExecuteDir + _WXDELIM_ + _CONFIGURE_FILE);
    }
}
